(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.2'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[   1850605,      26892]*)
(*NotebookOutlinePosition[   1851250,      26914]*)
(*  CellTagsIndexPosition[   1851206,      26910]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["\<\
This notebook is a tutorial on using the KnotTheory` package to retrieve data \
from third party sources. As an example, I'm using Marc Culler's table of \
A-polynomials, available online at \
http://www.math.uic.edu/~culler/Apolynomials/index.html\
\>", "Text"],

Cell["\<\
To get started, you'll need to download a copy of the KnotTheory` package, \
from http://katlas.math.toronto.edu/
Once you've got that, and unzipped it, modify the path in the following line, \
and load KnotTheory`\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(KnotTheoryPath\  = \ \
"\<c:\\scott\\projects\\svn-checkouts\\KnotTheory\\trunk\\\>";\)\), "\
\[IndentingNewLine]", 
    \(\(AppendTo[$Path, \ KnotTheoryPath];\)\), "\[IndentingNewLine]", 
    \(<< \ KnotTheory`\)}], "Input"],

Cell[BoxData[
    \("Loading KnotTheory` version of May 25, 2006, 14:39:5.78811.\nRead more \
at http://katlas.math.toronto.edu/wiki/KnotTheory."\)], "Print"]
}, Open  ]],

Cell["Quickly verify that it's working:", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(HOMFLYPT[Knot[8, 19]]\)[a, z]\)], "Input"],

Cell[BoxData[
    \(KnotTheory::"loading" \(\(:\)\(\ \)\) 
      "Loading precomputed data in \!\(\"PD4Knots`\"\)."\)], "Message"],

Cell[BoxData[
    \(KnotTheory::"credits" \(\(:\)\(\ \)\) 
      "\!\(\"The HOMFLYPT program was written by Scott Morrison.\"\)"\)], \
"Message"],

Cell[BoxData[
    \(1\/a\^10 - 5\/a\^8 + 5\/a\^6 - \(5\ z\^2\)\/a\^8 + \(10\ z\^2\)\/a\^6 - 
      z\^4\/a\^8 + \(6\ z\^4\)\/a\^6 + z\^6\/a\^6\)], "Output"]
}, Open  ]],

Cell["\<\
Now verify that the KnotTheory` package can successfully retrieve data from \
online:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(RetrieveInvariants[{"\<Khovanov s-Invariant\>"}, 
        AllKnots[{3, 6}], "\<KnotInfo\>"] // TableForm\)], "Input"],

Cell[BoxData[
    TagBox[GridBox[{
          {"\<\"Khovanov s-Invariant\"\>", \(Knot[3, 1]\), \(-2\)},
          {"\<\"Khovanov s-Invariant\"\>", \(Knot[4, 1]\), "0"},
          {"\<\"Khovanov s-Invariant\"\>", \(Knot[5, 1]\), \(-4\)},
          {"\<\"Khovanov s-Invariant\"\>", \(Knot[5, 2]\), \(-2\)},
          {"\<\"Khovanov s-Invariant\"\>", \(Knot[6, 1]\), "0"},
          {"\<\"Khovanov s-Invariant\"\>", \(Knot[6, 2]\), \(-2\)},
          {"\<\"Khovanov s-Invariant\"\>", \(Knot[6, 3]\), "0"}
          },
        RowSpacings->1,
        ColumnSpacings->3,
        RowAlignments->Baseline,
        ColumnAlignments->{Left}],
      Function[ BoxForm`e$, 
        TableForm[ BoxForm`e$]]]], "Output"]
}, Open  ]],

Cell["\<\
Okay, now we can get started. We're going to need to tell KnotTheory` where \
to find the knot invariants online, and then how to parse the data it finds \
there. We construct the URL for a given knot, first writing a function which \
translates KnotTheory`'s syntax for knots into Culler's notation.\
\>", "Text"],

Cell[BoxData[
    \(CullerKnotName[Knot[n_, k_]] := "\<L1\>" <> 
        If[n < 10, "\<0\>", "\<\>"] <> ToString[n] <> 
        If[k < 100, "\<0\>", "\<\>"] <> If[k < 10, "\<0\>", "\<\>"] <> 
        ToString[k]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(CullerKnotName[Knot[5, 1]]\)], "Input"],

Cell[BoxData[
    \("L105001"\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(CullerKnotName[Knot[10, 128]]\)], "Input"],

Cell[BoxData[
    \("L110128"\)], "Output"]
}, Open  ]],

Cell["\<\
With this, it's easy to construct the URL. We need to override the function \
\"KnotInvariantURL\". It takes two arguments, a string naming the invariant \
(in this case, I've chosen \"A-polynomial\"), and the knot.\
\>", "Text"],

Cell[BoxData[
    \(KnotInvariantURL["\<A-polynomial\>", 
        K_] := "\<http://www.math.uic.edu/~culler/Apolynomials/apolys/\>" <> 
        CullerKnotName[K] <> "\<.apoly\>"\)], "Input"],

Cell["\<\
Let's not bother doing any parsing at first, and just see what KnotTheory` \
can do. We use the function RetrieveInvariants, with third argument \
\"url\".\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(RetrieveInvariants[{"\<A-polynomial\>"}, AllKnots[5], "\<url\>"]\  // 
      TableForm\)], "Input"],

Cell[BoxData[
    TagBox[GridBox[{
          {"\<\"A-polynomial\"\>", \(Knot[5, 
              1]\), \("A_L105001 :=\n(1*M^10)\n+ (L^1)*(1);\n"\)},
          {"\<\"A-polynomial\"\>", \(Knot[5, 
              2]\), \("A_L105002 :=\n(1)\n+ (L^1)*(-1 + 2*M^2 + 2*M^4 - 1*M^8 \
+ 1*M^10)\n+ (L^2)*(1*M^4 - 1*M^6 + 2*M^10 + 2*M^12 - 1*M^14)\n+ \
(L^3)*(1*M^14);\n"\)}
          },
        RowSpacings->1,
        ColumnSpacings->3,
        RowAlignments->Baseline,
        ColumnAlignments->{Left}],
      Function[ BoxForm`e$, 
        TableForm[ BoxForm`e$]]]], "Output"]
}, Open  ]],

Cell["\<\
We need to strip off the initial and final parts of the string.  To do this, \
we override the function ParseKnotInvariantFromURL. A little experimentation \
indicates we should use:\
\>", "Text"],

Cell[BoxData[
    \(ParseKnotInvariantFromURL["\<A-polynomial\>", \ K_, \ data_] := 
      StringTake[data, {14, \(-3\)}]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(RetrieveInvariants[{"\<A-polynomial\>"}, AllKnots[5], "\<url\>"] // 
      TableForm\)], "Input"],

Cell[BoxData[
    TagBox[GridBox[{
          {"\<\"A-polynomial\"\>", \(Knot[5, 
              1]\), \("(1*M^10)\n+ (L^1)*(1)"\)},
          {"\<\"A-polynomial\"\>", \(Knot[5, 
              2]\), \("(1)\n+ (L^1)*(-1 + 2*M^2 + 2*M^4 - 1*M^8 + 1*M^10)\n+ \
(L^2)*(1*M^4 - 1*M^6 + 2*M^10 + 2*M^12 - 1*M^14)\n+ (L^3)*(1*M^14)"\)}
          },
        RowSpacings->1,
        ColumnSpacings->3,
        RowAlignments->Baseline,
        ColumnAlignments->{Left}],
      Function[ BoxForm`e$, 
        TableForm[ BoxForm`e$]]]], "Output"]
}, Open  ]],

Cell["\<\
This isn't quite right, however; ParseKnotInvariantFromURL should return a \
Mathematica expression, not just a string. We need to wrap the result with a \
call to ToExpression.

(ToExpression[string, StandardForm] does a reasonable attempt at interpreting \
polynomials. Often you may want to use ToExpression[string, TeXForm] for \
fancier input.)\
\>", "Text"],

Cell[BoxData[
    \(ParseKnotInvariantFromURL["\<A-polynomial\>", \ K_, \ data_] := 
      ToExpression[StringTake[data, {14, \(-3\)}], StandardForm]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(RetrieveInvariants[{"\<A-polynomial\>"}, AllKnots[5], "\<url\>"]\  // 
      TableForm\)], "Input"],

Cell[BoxData[
    TagBox[GridBox[{
          {"\<\"A-polynomial\"\>", \(Knot[5, 1]\), \(L + M\^10\)},
          {"\<\"A-polynomial\"\>", \(Knot[5, 2]\), \(1 + L\^3\ M\^14 + 
              L\ \((\(-1\) + 2\ M\^2 + 2\ M\^4 - M\^8 + M\^10)\) + 
              L\^2\ \((M\^4 - M\^6 + 2\ M\^10 + 2\ M\^12 - M\^14)\)\)}
          },
        RowSpacings->1,
        ColumnSpacings->3,
        RowAlignments->Baseline,
        ColumnAlignments->{Left}],
      Function[ BoxForm`e$, 
        TableForm[ BoxForm`e$]]]], "Output"]
}, Open  ]],

Cell["\<\
Not bad! Let's just make some slight improvements, for safety's sake. The \
first change below ensures that we don't try anything silly if we can't work \
out the Culler notation for the given knot. The second change does some basic \
checking that the downloaded data looks like what we expect.\
\>", "Text"],

Cell[BoxData[
    \(KnotInvariantURL["\<A-polynomial\>", K_] := 
      Module[{knotname = 
            CullerKnotName[K]}, \[IndentingNewLine]If[\(! MatchQ[
              knotname, \ _String]\), 
          Return[$Failed]]; \
"\<http://www.math.uic.edu/~culler/Apolynomials/apolys/\>" <> 
          knotname <> "\<.apoly\>"\[IndentingNewLine]]\)], "Input"],

Cell[BoxData[
    \(ParseKnotInvariantFromURL["\<A-polynomial\>", \ K_, \ data_String] := 
      Module[{results}, \[IndentingNewLine]results = 
          StringCases[
            data, "\<A_\>" ~~ \(CullerKnotName[
                    K] ~~ \("\< :=\n\>" ~~ \(polynomial__ ~~ "\<;\n\>"\)\)\) \
\[RuleDelayed] polynomial]; \[IndentingNewLine]If[Length[results] \[Equal] 0, 
          Return[$Failed]]; \[IndentingNewLine]ToExpression[
          results\[LeftDoubleBracket]1\[RightDoubleBracket], 
          StandardForm]\[IndentingNewLine]]\)], "Input"],

Cell["\<\
Now let's extract all the data Culler has for the Rolfsen table, storing it \
in the variable \"apolynomialdata\".\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\((apolynomialdata = 
          RetrieveInvariants[{"\<A-polynomial\>"}, 
            AllKnots[{3, 10}], "\<url\>"])\)\  // TableForm\)], "Input"],

Cell[BoxData[
    \(FetchURL::"conopen" \(\(:\)\(\ \)\) 
      "The connection to \
\!\(\"http://www.math.uic.edu/~culler/Apolynomials/apolys/L109030.apoly\"\) \
cannot be opened."\)], "Message"],

Cell[BoxData[
    \(FetchURL::"conopen" \(\(:\)\(\ \)\) 
      "The connection to \
\!\(\"http://www.math.uic.edu/~culler/Apolynomials/apolys/L109032.apoly\"\) \
cannot be opened."\)], "Message"],

Cell[BoxData[
    \(FetchURL::"conopen" \(\(:\)\(\ \)\) 
      "The connection to \
\!\(\"http://www.math.uic.edu/~culler/Apolynomials/apolys/L109033.apoly\"\) \
cannot be opened."\)], "Message"],

Cell[BoxData[
    RowBox[{\(General::"stop"\), \(\(:\)\(\ \)\), "\<\"Further output of \
\\!\\(FetchURL :: \\\"conopen\\\"\\) will be suppressed during this \
calculation. \\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", \
ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"General::stop\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    TagBox[GridBox[{
          {"\<\"A-polynomial\"\>", \(Knot[3, 1]\), \(L + M\^6\)},
          {"\<\"A-polynomial\"\>", \(Knot[4, 1]\), \(M\^4 + L\^2\ M\^4 + 
              L\ \((\(-1\) + M\^2 + 2\ M\^4 + M\^6 - M\^8)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[5, 1]\), \(L + M\^10\)},
          {"\<\"A-polynomial\"\>", \(Knot[5, 2]\), \(1 + L\^3\ M\^14 + 
              L\ \((\(-1\) + 2\ M\^2 + 2\ M\^4 - M\^8 + M\^10)\) + 
              L\^2\ \((M\^4 - M\^6 + 2\ M\^10 + 2\ M\^12 - M\^14)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[6, 1]\), \(M\^8 + L\^4\ M\^8 + 
              L\^3\ \((\(-1\) + M\^2 + 3\ M\^8 + 3\ M\^10 - 2\ M\^12)\) + 
              L\ \((\(-2\)\ M\^4 + 3\ M\^6 + 3\ M\^8 + M\^14 - M\^16)\) + 
              L\^2\ \((1 - 3\ M\^2 - M\^4 + 3\ M\^6 + 6\ M\^8 + 3\ M\^10 - 
                    M\^12 - 3\ M\^14 + M\^16)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[6, 2]\), \(M\^4 + L\^5\ M\^26 + 
              L\ \((\(-1\) + 2\ M\^2 - M\^4 - 2\ M\^6 + 5\ M\^8 + 5\ M\^10 - 
                    3\ M\^12)\) + 
              L\^2\ \((\(-M\^2\) + 3\ M\^4 - M\^6 - 5\ M\^8 - 3\ M\^10 + 
                    12\ M\^12 + 13\ M\^14 - 3\ M\^16 - 8\ M\^18 + 
                    3\ M\^20)\) + 
              L\^3\ \((3\ M\^10 - 8\ M\^12 - 3\ M\^14 + 13\ M\^16 + 
                    12\ M\^18 - 3\ M\^20 - 5\ M\^22 - M\^24 + 3\ M\^26 - 
                    M\^28)\) + 
              L\^4\ \((\(-3\)\ M\^18 + 5\ M\^20 + 5\ M\^22 - 2\ M\^24 - 
                    M\^26 + 2\ M\^28 - M\^30)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[6, 3]\), \(M\^14 + L\^6\ M\^14 + 
              L\ \((2\ M\^8 - 5\ M\^10 + M\^12 + 10\ M\^14 + M\^16 - 
                    5\ M\^18 + 2\ M\^20)\) + 
              L\^5\ \((2\ M\^8 - 5\ M\^10 + M\^12 + 10\ M\^14 + M\^16 - 
                    5\ M\^18 + 2\ M\^20)\) + 
              L\^2\ \((M\^2 - 4\ M\^4 + 4\ M\^6 + 2\ M\^8 - 6\ M\^10 + 
                    2\ M\^12 + 17\ M\^14 + 2\ M\^16 - 6\ M\^18 + 2\ M\^20 + 
                    4\ M\^22 - 4\ M\^24 + M\^26)\) + 
              L\^4\ \((M\^2 - 4\ M\^4 + 4\ M\^6 + 2\ M\^8 - 6\ M\^10 + 
                    2\ M\^12 + 17\ M\^14 + 2\ M\^16 - 6\ M\^18 + 2\ M\^20 + 
                    4\ M\^22 - 4\ M\^24 + M\^26)\) + 
              L\^3\ \((1 - 5\ M\^2 + 3\ M\^4 + 9\ M\^6 - 2\ M\^8 - 
                    21\ M\^10 + 8\ M\^12 + 34\ M\^14 + 8\ M\^16 - 21\ M\^18 - 
                    2\ M\^20 + 9\ M\^22 + 3\ M\^24 - 5\ M\^26 + M\^28)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[7, 1]\), \(L + M\^14\)},
          {"\<\"A-polynomial\"\>", \(Knot[7, 2]\), \(L\^5 + M\^22 + 
              L\^4\ \((\(-2\) + 4\ M\^2 + 3\ M\^4 - M\^12 + M\^14)\) + 
              L\^3\ \((1 - 4\ M\^2 + 2\ M\^4 + 5\ M\^6 + 6\ M\^8 - 4\ M\^12 + 
                    M\^14 + 5\ M\^16 - 2\ M\^18)\) + 
              L\ \((M\^8 - M\^10 + 3\ M\^18 + 4\ M\^20 - 2\ M\^22)\) + 
              L\^2\ \((\(-2\)\ M\^4 + 5\ M\^6 + M\^8 - 4\ M\^10 + 6\ M\^14 + 
                    5\ M\^16 + 2\ M\^18 - 4\ M\^20 + M\^22)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[7, 3]\), \(L\^6 + M\^52 + 
              L\^5\ \((\(-1\) + 2\ M\^2 - M\^4 + 5\ M\^8 + 2\ M\^10 - 
                    3\ M\^12 + 2\ M\^14)\) + 
              L\^4\ \((3\ M\^8 - 9\ M\^10 + 2\ M\^12 + 14\ M\^14 + 2\ M\^16 - 
                    4\ M\^18 + 4\ M\^20 + 2\ M\^22 + 3\ M\^24 - 3\ M\^26 + 
                    M\^28)\) + 
              L\^3\ \((\(-3\)\ M\^16 + 10\ M\^18 - 3\ M\^20 - 12\ M\^22 + 
                    6\ M\^24 + 24\ M\^26 + 6\ M\^28 - 12\ M\^30 - 3\ M\^32 + 
                    10\ M\^34 - 3\ M\^36)\) + 
              L\^2\ \((M\^24 - 3\ M\^26 + 3\ M\^28 + 2\ M\^30 + 4\ M\^32 - 
                    4\ M\^34 + 2\ M\^36 + 14\ M\^38 + 2\ M\^40 - 9\ M\^42 + 
                    3\ M\^44)\) + 
              L\ \((2\ M\^38 - 3\ M\^40 + 2\ M\^42 + 5\ M\^44 - M\^48 + 
                    2\ M\^50 - M\^52)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[7, 4]\), \(1 + L\^5\ M\^22 + 
              L\ \((\(-3\) + 7\ M\^2 + 4\ M\^4 - 6\ M\^6 + M\^8 + 3\ M\^10 - 
                    2\ M\^12 + M\^14)\) + 
              L\^4\ \((M\^8 - 2\ M\^10 + 3\ M\^12 + M\^14 - 6\ M\^16 + 
                    4\ M\^18 + 7\ M\^20 - 3\ M\^22)\) + 
              L\^2\ \((3 - 10\ M\^2 + 3\ M\^4 + 21\ M\^6 - 3\ M\^8 - 
                    17\ M\^10 + 6\ M\^12 + 10\ M\^14 - 2\ M\^16 - 3\ M\^18 + 
                    3\ M\^20 - M\^22)\) + 
              L\^3\ \((\(-1\) + 3\ M\^2 - 3\ M\^4 - 2\ M\^6 + 10\ M\^8 + 
                    6\ M\^10 - 17\ M\^12 - 3\ M\^14 + 21\ M\^16 + 3\ M\^18 - 
                    10\ M\^20 + 3\ M\^22)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[7, 5]\), \(L\^8 + M\^68 + 
              L\^7\ \((\(-1\) + 4\ M\^2 - 5\ M\^4 - 2\ M\^6 + 13\ M\^8 + 
                    3\ M\^10 - 7\ M\^12 + 3\ M\^14)\) + 
              L\^6\ \((M\^4 - 6\ M\^6 + 17\ M\^8 - 10\ M\^10 - 35\ M\^12 + 
                    32\ M\^14 + 56\ M\^16 - 24\ M\^18 - 28\ M\^20 + 
                    22\ M\^22 + 14\ M\^24 - 14\ M\^26 + 3\ M\^28)\) + 
              L\^5\ \((2\ M\^10 - 12\ M\^12 + 23\ M\^14 + 6\ M\^16 - 
                    48\ M\^18 - 15\ M\^20 + 82\ M\^22 + 28\ M\^24 - 
                    47\ M\^26 + 13\ M\^28 + 46\ M\^30 - 15\ M\^32 - 
                    15\ M\^34 + 2\ M\^36 + 12\ M\^38 - 7\ M\^40 + M\^42)\) + 
              L\^4\ \((M\^16 - 6\ M\^18 + 11\ M\^20 - 4\ M\^22 - 4\ M\^24 - 
                    4\ M\^26 + 11\ M\^28 - 12\ M\^30 + 16\ M\^32 + 
                    52\ M\^34 + 16\ M\^36 - 12\ M\^38 + 11\ M\^40 - 
                    4\ M\^42 - 4\ M\^44 - 4\ M\^46 + 11\ M\^48 - 6\ M\^50 + 
                    M\^52)\) + 
              L\^3\ \((M\^26 - 7\ M\^28 + 12\ M\^30 + 2\ M\^32 - 15\ M\^34 - 
                    15\ M\^36 + 46\ M\^38 + 13\ M\^40 - 47\ M\^42 + 
                    28\ M\^44 + 82\ M\^46 - 15\ M\^48 - 48\ M\^50 + 
                    6\ M\^52 + 23\ M\^54 - 12\ M\^56 + 2\ M\^58)\) + 
              L\^2\ \((3\ M\^40 - 14\ M\^42 + 14\ M\^44 + 22\ M\^46 - 
                    28\ M\^48 - 24\ M\^50 + 56\ M\^52 + 32\ M\^54 - 
                    35\ M\^56 - 10\ M\^58 + 17\ M\^60 - 6\ M\^62 + M\^64)\) + 
              L\ \((3\ M\^54 - 7\ M\^56 + 3\ M\^58 + 13\ M\^60 - 2\ M\^62 - 
                    5\ M\^64 + 4\ M\^66 - M\^68)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[7, 6]\), \(L\^9\ M\^8 + M\^46 + 
              L\^8\ \((\(-2\)\ M\^4 + 6\ M\^6 - 5\ M\^8 - 5\ M\^10 + 
                    16\ M\^12 + 5\ M\^14 - 9\ M\^16 + 3\ M\^18)\) + 
              L\^7\ \((1 - 6\ M\^2 + 11\ M\^4 + 2\ M\^6 - 16\ M\^8 - 
                    7\ M\^10 + 9\ M\^12 + 8\ M\^14 + 42\ M\^16 + 11\ M\^18 - 
                    37\ M\^20 + 8\ M\^22 + 23\ M\^24 - 16\ M\^26 + 
                    3\ M\^28)\) + 
              L\^6\ \((\(-1\) + 7\ M\^2 - 16\ M\^4 + M\^6 + 34\ M\^8 + 
                    10\ M\^10 - 80\ M\^12 - 9\ M\^14 + 62\ M\^16 + 
                    10\ M\^18 + 34\ M\^20 + 83\ M\^22 - 44\ M\^24 - 
                    48\ M\^26 + 47\ M\^28 + 10\ M\^30 - 29\ M\^32 + 
                    19\ M\^34 - 7\ M\^36 + M\^38)\) + 
              L\^5\ \((M\^4 - 9\ M\^6 + 32\ M\^8 - 30\ M\^10 - 68\ M\^12 + 
                    98\ M\^14 + 164\ M\^16 - 212\ M\^18 - 266\ M\^20 + 
                    196\ M\^22 + 377\ M\^24 + 24\ M\^26 - 237\ M\^28 - 
                    85\ M\^30 + 158\ M\^32 + 52\ M\^34 - 78\ M\^36 - 
                    18\ M\^38 + 41\ M\^40 - 16\ M\^42 + 2\ M\^44)\) + 
              L\ \((3\ M\^36 - 9\ M\^38 + 5\ M\^40 + 16\ M\^42 - 5\ M\^44 - 
                    5\ M\^46 + 6\ M\^48 - 2\ M\^50)\) + 
              L\^4\ \((2\ M\^10 - 16\ M\^12 + 41\ M\^14 - 18\ M\^16 - 
                    78\ M\^18 + 52\ M\^20 + 158\ M\^22 - 85\ M\^24 - 
                    237\ M\^26 + 24\ M\^28 + 377\ M\^30 + 196\ M\^32 - 
                    266\ M\^34 - 212\ M\^36 + 164\ M\^38 + 98\ M\^40 - 
                    68\ M\^42 - 30\ M\^44 + 32\ M\^46 - 9\ M\^48 + M\^50)\) + 
              L\^3\ \((M\^16 - 7\ M\^18 + 19\ M\^20 - 29\ M\^22 + 10\ M\^24 + 
                    47\ M\^26 - 48\ M\^28 - 44\ M\^30 + 83\ M\^32 + 
                    34\ M\^34 + 10\ M\^36 + 62\ M\^38 - 9\ M\^40 - 
                    80\ M\^42 + 10\ M\^44 + 34\ M\^46 + M\^48 - 16\ M\^50 + 
                    7\ M\^52 - M\^54)\) + 
              L\^2\ \((3\ M\^26 - 16\ M\^28 + 23\ M\^30 + 8\ M\^32 - 
                    37\ M\^34 + 11\ M\^36 + 42\ M\^38 + 8\ M\^40 + 9\ M\^42 - 
                    7\ M\^44 - 16\ M\^46 + 2\ M\^48 + 11\ M\^50 - 6\ M\^52 + 
                    M\^54)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[7, 7]\), \(L\^7\ M\^18 + M\^20 + 
              L\^6\ \((3\ M\^12 - 11\ M\^14 + 4\ M\^16 + 20\ M\^18 - 
                    7\ M\^20 - 7\ M\^22 + 7\ M\^24 - 2\ M\^26)\) + 
              L\ \((\(-2\)\ M\^12 + 7\ M\^14 - 7\ M\^16 - 7\ M\^18 + 
                    20\ M\^20 + 4\ M\^22 - 11\ M\^24 + 3\ M\^26)\) + 
              L\^2\ \((M\^4 - 6\ M\^6 + 11\ M\^8 - M\^10 - 8\ M\^12 + M\^14 - 
                    28\ M\^16 + 16\ M\^18 + 84\ M\^20 - 19\ M\^22 - 
                    65\ M\^24 + 27\ M\^26 + 23\ M\^28 - 18\ M\^30 + 
                    3\ M\^32)\) + 
              L\^5\ \((3\ M\^6 - 18\ M\^8 + 23\ M\^10 + 27\ M\^12 - 
                    65\ M\^14 - 19\ M\^16 + 84\ M\^18 + 16\ M\^20 - 
                    28\ M\^22 + M\^24 - 8\ M\^26 - M\^28 + 11\ M\^30 - 
                    6\ M\^32 + M\^34)\) + 
              L\^4\ \((1 - 7\ M\^2 + 15\ M\^4 + M\^6 - 30\ M\^8 - M\^10 + 
                    41\ M\^12 + 7\ M\^14 - 29\ M\^16 - 46\ M\^18 + 
                    35\ M\^20 + 130\ M\^22 - 2\ M\^24 - 123\ M\^26 + 
                    59\ M\^30 - 4\ M\^32 - 19\ M\^34 + 8\ M\^36 - M\^38)\) + 
              L\^3\ \((\(-1\) + 8\ M\^2 - 19\ M\^4 - 4\ M\^6 + 59\ M\^8 - 
                    123\ M\^12 - 2\ M\^14 + 130\ M\^16 + 35\ M\^18 - 
                    46\ M\^20 - 29\ M\^22 + 7\ M\^24 + 41\ M\^26 - M\^28 - 
                    30\ M\^30 + M\^32 + 15\ M\^34 - 7\ M\^36 + M\^38)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[8, 1]\), \(M\^12 + L\^6\ M\^12 + 
              L\^5\ \((\(-1\) + M\^2 + 4\ M\^12 + 5\ M\^14 - 3\ M\^16)\) + 
              L\^4\ \((2 - 6\ M\^2 + 4\ M\^6 + 10\ M\^12 + 12\ M\^14 - 
                    10\ M\^18 + 3\ M\^20)\) + 
              L\ \((\(-3\)\ M\^8 + 5\ M\^10 + 4\ M\^12 + M\^22 - M\^24)\) + 
              L\^3\ \((\(-1\) + 5\ M\^2 - 6\ M\^4 - 5\ M\^6 - 3\ M\^8 + 
                    10\ M\^10 + 20\ M\^12 + 10\ M\^14 - 3\ M\^16 - 5\ M\^18 - 
                    6\ M\^20 + 5\ M\^22 - M\^24)\) + 
              L\^2\ \((3\ M\^4 - 10\ M\^6 + 12\ M\^10 + 10\ M\^12 + 
                    4\ M\^18 - 6\ M\^22 + 2\ M\^24)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[8, 2]\), \(L\^8\ M\^4 + M\^72 + 
              L\^7\ \((\(-1\) + 2\ M\^2 - M\^4 - 3\ M\^10 + 7\ M\^12 + 
                    9\ M\^14 - 5\ M\^16)\) + 
              L\^6\ \((\(-2\)\ M\^6 + 7\ M\^8 - 8\ M\^10 - 4\ M\^12 + 
                    14\ M\^14 - M\^16 - 28\ M\^18 + 17\ M\^20 + 56\ M\^22 - 
                    M\^24 - 32\ M\^26 + 10\ M\^28)\) + 
              L\^5\ \((\(-M\^12\) + 5\ M\^14 - 8\ M\^16 + M\^18 - 2\ M\^20 + 
                    29\ M\^22 - 17\ M\^24 - 77\ M\^26 + 33\ M\^28 + 
                    143\ M\^30 + 29\ M\^32 - 87\ M\^34 - 24\ M\^36 + 
                    42\ M\^38 - 10\ M\^40)\) + 
              L\^4\ \((5\ M\^24 - 24\ M\^26 + 26\ M\^28 + 36\ M\^30 - 
                    43\ M\^32 - 108\ M\^34 + 47\ M\^36 + 192\ M\^38 + 
                    47\ M\^40 - 108\ M\^42 - 43\ M\^44 + 36\ M\^46 + 
                    26\ M\^48 - 24\ M\^50 + 5\ M\^52)\) + 
              L\^3\ \((\(-10\)\ M\^36 + 42\ M\^38 - 24\ M\^40 - 87\ M\^42 + 
                    29\ M\^44 + 143\ M\^46 + 33\ M\^48 - 77\ M\^50 - 
                    17\ M\^52 + 29\ M\^54 - 2\ M\^56 + M\^58 - 8\ M\^60 + 
                    5\ M\^62 - M\^64)\) + 
              L\^2\ \((10\ M\^48 - 32\ M\^50 - M\^52 + 56\ M\^54 + 
                    17\ M\^56 - 28\ M\^58 - M\^60 + 14\ M\^62 - 4\ M\^64 - 
                    8\ M\^66 + 7\ M\^68 - 2\ M\^70)\) + 
              L\ \((\(-5\)\ M\^60 + 9\ M\^62 + 7\ M\^64 - 3\ M\^66 - M\^72 + 
                    2\ M\^74 - M\^76)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[8, 3]\), \(M\^16 + L\^8\ M\^16 + 
              L\ \((\(-2\)\ M\^8 + 3\ M\^10 - 2\ M\^12 + M\^14 + 8\ M\^16 + 
                    M\^18 - 2\ M\^20 + 3\ M\^22 - 2\ M\^24)\) + 
              L\^7\ \((\(-2\)\ M\^8 + 3\ M\^10 - 2\ M\^12 + M\^14 + 
                    8\ M\^16 + M\^18 - 2\ M\^20 + 3\ M\^22 - 2\ M\^24)\) + 
              L\^3\ \((\(-4\) + 16\ M\^2 - 14\ M\^4 - 21\ M\^6 + 24\ M\^8 + 
                    16\ M\^10 - 56\ M\^12 + 17\ M\^14 + 100\ M\^16 + 
                    17\ M\^18 - 56\ M\^20 + 16\ M\^22 + 24\ M\^24 - 
                    21\ M\^26 - 14\ M\^28 + 16\ M\^30 - 4\ M\^32)\) + 
              L\^5\ \((\(-4\) + 16\ M\^2 - 14\ M\^4 - 21\ M\^6 + 24\ M\^8 + 
                    16\ M\^10 - 56\ M\^12 + 17\ M\^14 + 100\ M\^16 + 
                    17\ M\^18 - 56\ M\^20 + 16\ M\^22 + 24\ M\^24 - 
                    21\ M\^26 - 14\ M\^28 + 16\ M\^30 - 4\ M\^32)\) + 
              L\^2\ \((1 - 3\ M\^2 + 3\ M\^4 - M\^6 - 4\ M\^8 - 9\ M\^10 + 
                    15\ M\^12 + 13\ M\^14 - 2\ M\^16 + 13\ M\^18 + 
                    15\ M\^20 - 9\ M\^22 - 4\ M\^24 - M\^26 + 3\ M\^28 - 
                    3\ M\^30 + M\^32)\) + 
              L\^6\ \((1 - 3\ M\^2 + 3\ M\^4 - M\^6 - 4\ M\^8 - 9\ M\^10 + 
                    15\ M\^12 + 13\ M\^14 - 2\ M\^16 + 13\ M\^18 + 
                    15\ M\^20 - 9\ M\^22 - 4\ M\^24 - M\^26 + 3\ M\^28 - 
                    3\ M\^30 + M\^32)\) + 
              L\^4\ \((6 - 26\ M\^2 + 22\ M\^4 + 40\ M\^6 - 59\ M\^8 - 
                    64\ M\^10 + 82\ M\^12 + 50\ M\^14 - 32\ M\^16 + 
                    50\ M\^18 + 82\ M\^20 - 64\ M\^22 - 59\ M\^24 + 
                    40\ M\^26 + 22\ M\^28 - 26\ M\^30 + 6\ M\^32)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[8, 4]\), \(M\^10 + L\^9\ M\^48 + 
              L\ \((\(-M\^2\) + 2\ M\^4 - 3\ M\^6 + 2\ M\^8 + 3\ M\^10 - 
                    6\ M\^12 + 7\ M\^14 + 11\ M\^16 - 6\ M\^18)\) + 
              L\^2\ \((\(-1\) + 4\ M\^2 - 7\ M\^4 + 7\ M\^6 + 2\ M\^8 - 
                    27\ M\^10 + 11\ M\^12 + 46\ M\^14 - 44\ M\^16 - 
                    12\ M\^18 + 88\ M\^20 + 3\ M\^22 - 49\ M\^24 + 
                    15\ M\^26)\) + 
              L\^3\ \((1 - 5\ M\^2 + 8\ M\^4 - 16\ M\^8 + 9\ M\^10 + 
                    55\ M\^12 - 98\ M\^14 - 51\ M\^16 + 178\ M\^18 - 
                    34\ M\^20 - 163\ M\^22 + 224\ M\^24 + 154\ M\^26 - 
                    189\ M\^28 - 55\ M\^30 + 86\ M\^32 - 20\ M\^34)\) + 
              L\^4\ \((\(-6\)\ M\^8 + 31\ M\^10 - 44\ M\^12 - 37\ M\^14 + 
                    136\ M\^16 + 15\ M\^18 - 291\ M\^20 + 105\ M\^22 + 
                    245\ M\^24 - 220\ M\^26 + 91\ M\^28 + 399\ M\^30 - 
                    201\ M\^32 - 263\ M\^34 + 144\ M\^36 + 81\ M\^38 - 
                    74\ M\^40 + 15\ M\^42)\) + 
              L\^5\ \((15\ M\^16 - 74\ M\^18 + 81\ M\^20 + 144\ M\^22 - 
                    263\ M\^24 - 201\ M\^26 + 399\ M\^28 + 91\ M\^30 - 
                    220\ M\^32 + 245\ M\^34 + 105\ M\^36 - 291\ M\^38 + 
                    15\ M\^40 + 136\ M\^42 - 37\ M\^44 - 44\ M\^46 + 
                    31\ M\^48 - 6\ M\^50)\) + 
              L\^8\ \((\(-6\)\ M\^40 + 11\ M\^42 + 7\ M\^44 - 6\ M\^46 + 
                    3\ M\^48 + 2\ M\^50 - 3\ M\^52 + 2\ M\^54 - M\^56)\) + 
              L\^7\ \((15\ M\^32 - 49\ M\^34 + 3\ M\^36 + 88\ M\^38 - 
                    12\ M\^40 - 44\ M\^42 + 46\ M\^44 + 11\ M\^46 - 
                    27\ M\^48 + 2\ M\^50 + 7\ M\^52 - 7\ M\^54 + 4\ M\^56 - 
                    M\^58)\) + 
              L\^6\ \((\(-20\)\ M\^24 + 86\ M\^26 - 55\ M\^28 - 189\ M\^30 + 
                    154\ M\^32 + 224\ M\^34 - 163\ M\^36 - 34\ M\^38 + 
                    178\ M\^40 - 51\ M\^42 - 98\ M\^44 + 55\ M\^46 + 
                    9\ M\^48 - 16\ M\^50 + 8\ M\^54 - 5\ M\^56 + M\^58)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[8, 5]\), \(M\^4 + L\^8\ M\^68 + 
              L\ \((\(-1\) + 3\ M\^2 - 7\ M\^4 + 8\ M\^6 + 3\ M\^8 - 
                    12\ M\^10 + 11\ M\^12 + 7\ M\^14 - 4\ M\^16)\) + 
              L\^2\ \((\(-M\^2\) + 4\ M\^4 - 10\ M\^6 + 12\ M\^8 - 
                    16\ M\^12 + 33\ M\^16 - 11\ M\^18 - 24\ M\^20 + 
                    56\ M\^22 + M\^24 - 22\ M\^26 + 6\ M\^28)\) + 
              L\^3\ \((M\^10 - 6\ M\^12 + 19\ M\^14 - 39\ M\^16 + 18\ M\^18 + 
                    54\ M\^20 - 66\ M\^22 - 31\ M\^24 + 93\ M\^26 + 
                    17\ M\^28 - 48\ M\^30 + 80\ M\^32 - 15\ M\^34 - 
                    41\ M\^36 + 24\ M\^38 - 4\ M\^40)\) + 
              L\^4\ \((M\^20 - 10\ M\^22 + 35\ M\^24 - 47\ M\^26 - 
                    16\ M\^28 + 79\ M\^30 - 53\ M\^32 - 2\ M\^34 + 
                    96\ M\^36 - 2\ M\^38 - 53\ M\^40 + 79\ M\^42 - 
                    16\ M\^44 - 47\ M\^46 + 35\ M\^48 - 10\ M\^50 + 
                    M\^52)\) + 
              L\^5\ \((\(-4\)\ M\^32 + 24\ M\^34 - 41\ M\^36 - 15\ M\^38 + 
                    80\ M\^40 - 48\ M\^42 + 17\ M\^44 + 93\ M\^46 - 
                    31\ M\^48 - 66\ M\^50 + 54\ M\^52 + 18\ M\^54 - 
                    39\ M\^56 + 19\ M\^58 - 6\ M\^60 + M\^62)\) + 
              L\^6\ \((6\ M\^44 - 22\ M\^46 + M\^48 + 56\ M\^50 - 24\ M\^52 - 
                    11\ M\^54 + 33\ M\^56 - 16\ M\^60 + 12\ M\^64 - 
                    10\ M\^66 + 4\ M\^68 - M\^70)\) + 
              L\^7\ \((\(-4\)\ M\^56 + 7\ M\^58 + 11\ M\^60 - 12\ M\^62 + 
                    3\ M\^64 + 8\ M\^66 - 7\ M\^68 + 3\ M\^70 - M\^72)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[8, 6]\), \(L\^11\ M\^8 + M\^54 + 
              L\^10\ \((\(-2\)\ M\^4 + 6\ M\^6 - 8\ M\^8 + 19\ M\^12 - 
                    2\ M\^14 - 6\ M\^16 + 7\ M\^18 - 3\ M\^20)\) + 
              L\^9\ \((1 - 6\ M\^2 + 13\ M\^4 - 11\ M\^6 + 2\ M\^8 - 
                    17\ M\^10 - 2\ M\^12 + 70\ M\^14 - 10\ M\^16 - 
                    47\ M\^18 + 85\ M\^20 + 29\ M\^22 - 64\ M\^24 - 
                    4\ M\^26 + 27\ M\^28 - 14\ M\^30 + 3\ M\^32)\) + 
              L\^8\ \((\(-1\) + 9\ M\^2 - 38\ M\^4 + 63\ M\^6 + 17\ M\^8 - 
                    138\ M\^10 - 28\ M\^12 + 250\ M\^14 + 24\ M\^16 - 
                    343\ M\^18 + 28\ M\^20 + 491\ M\^22 + 52\ M\^24 - 
                    285\ M\^26 + 11\ M\^28 + 120\ M\^30 - 71\ M\^32 - 
                    26\ M\^34 + 28\ M\^36 + 17\ M\^38 - 21\ M\^40 + 
                    7\ M\^42 - M\^44)\) + 
              L\^7\ \((\(-3\)\ M\^2 + 26\ M\^4 - 90\ M\^6 + 104\ M\^8 + 
                    135\ M\^10 - 336\ M\^12 - 252\ M\^14 + 744\ M\^16 + 
                    358\ M\^18 - 1055\ M\^20 - 552\ M\^22 + 1054\ M\^24 + 
                    773\ M\^26 - 207\ M\^28 - 340\ M\^30 - 213\ M\^32 + 
                    45\ M\^34 + 261\ M\^36 - 78\ M\^38 - 104\ M\^40 + 
                    54\ M\^42 + 39\ M\^44 - 48\ M\^46 + 17\ M\^48 - 
                    2\ M\^50)\) + 
              L\^6\ \((\(-3\)\ M\^4 + 27\ M\^6 - 89\ M\^8 + 97\ M\^10 + 
                    113\ M\^12 - 237\ M\^14 - 257\ M\^16 + 363\ M\^18 + 
                    515\ M\^20 + 21\ M\^22 - 1051\ M\^24 - 998\ M\^26 + 
                    1231\ M\^28 + 2288\ M\^30 - 339\ M\^32 - 1945\ M\^34 + 
                    2\ M\^36 + 1031\ M\^38 - 141\ M\^40 - 251\ M\^42 + 
                    106\ M\^44 - 3\ M\^46 - 54\ M\^48 + 60\ M\^50 - 
                    32\ M\^52 + 9\ M\^54 - M\^56)\) + 
              L\^5\ \((\(-M\^6\) + 9\ M\^8 - 32\ M\^10 + 60\ M\^12 - 
                    54\ M\^14 - 3\ M\^16 + 106\ M\^18 - 251\ M\^20 - 
                    141\ M\^22 + 1031\ M\^24 + 2\ M\^26 - 1945\ M\^28 - 
                    339\ M\^30 + 2288\ M\^32 + 1231\ M\^34 - 998\ M\^36 - 
                    1051\ M\^38 + 21\ M\^40 + 515\ M\^42 + 363\ M\^44 - 
                    257\ M\^46 - 237\ M\^48 + 113\ M\^50 + 97\ M\^52 - 
                    89\ M\^54 + 27\ M\^56 - 3\ M\^58)\) + 
              L\ \((\(-3\)\ M\^42 + 7\ M\^44 - 6\ M\^46 - 2\ M\^48 + 
                    19\ M\^50 - 8\ M\^54 + 6\ M\^56 - 2\ M\^58)\) + 
              L\^4\ \((\(-2\)\ M\^12 + 17\ M\^14 - 48\ M\^16 + 39\ M\^18 + 
                    54\ M\^20 - 104\ M\^22 - 78\ M\^24 + 261\ M\^26 + 
                    45\ M\^28 - 213\ M\^30 - 340\ M\^32 - 207\ M\^34 + 
                    773\ M\^36 + 1054\ M\^38 - 552\ M\^40 - 1055\ M\^42 + 
                    358\ M\^44 + 744\ M\^46 - 252\ M\^48 - 336\ M\^50 + 
                    135\ M\^52 + 104\ M\^54 - 90\ M\^56 + 26\ M\^58 - 
                    3\ M\^60)\) + 
              L\^3\ \((\(-M\^18\) + 7\ M\^20 - 21\ M\^22 + 17\ M\^24 + 
                    28\ M\^26 - 26\ M\^28 - 71\ M\^30 + 120\ M\^32 + 
                    11\ M\^34 - 285\ M\^36 + 52\ M\^38 + 491\ M\^40 + 
                    28\ M\^42 - 343\ M\^44 + 24\ M\^46 + 250\ M\^48 - 
                    28\ M\^50 - 138\ M\^52 + 17\ M\^54 + 63\ M\^56 - 
                    38\ M\^58 + 9\ M\^60 - M\^62)\) + 
              L\^2\ \((3\ M\^30 - 14\ M\^32 + 27\ M\^34 - 4\ M\^36 - 
                    64\ M\^38 + 29\ M\^40 + 85\ M\^42 - 47\ M\^44 - 
                    10\ M\^46 + 70\ M\^48 - 2\ M\^50 - 17\ M\^52 + 2\ M\^54 - 
                    11\ M\^56 + 13\ M\^58 - 6\ M\^60 + M\^62)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[8, 7]\), \(L\^11\ M\^12 + M\^58 + 
              L\^10\ \((2\ M\^6 - 5\ M\^8 + 4\ M\^10 - M\^12 - 5\ M\^14 + 
                    17\ M\^16 + 6\ M\^18 - 11\ M\^20 + 4\ M\^22)\) + 
              L\^9\ \((1 - 4\ M\^2 + 6\ M\^4 - 4\ M\^6 - M\^8 - 2\ M\^10 + 
                    23\ M\^12 - M\^14 - 67\ M\^16 + M\^18 + 142\ M\^20 + 
                    23\ M\^22 - 102\ M\^24 + 17\ M\^26 + 47\ M\^28 - 
                    30\ M\^30 + 6\ M\^32)\) + 
              L\^8\ \((3\ M\^2 - 19\ M\^4 + 42\ M\^6 - 17\ M\^8 - 81\ M\^10 + 
                    87\ M\^12 + 115\ M\^14 - 192\ M\^16 - 35\ M\^18 + 
                    130\ M\^20 - 46\ M\^22 + 104\ M\^24 + 188\ M\^26 - 
                    137\ M\^28 - 41\ M\^30 + 124\ M\^32 - 50\ M\^34 - 
                    53\ M\^36 + 66\ M\^38 - 27\ M\^40 + 4\ M\^42)\) + 
              L\^7\ \((3\ M\^4 - 26\ M\^6 + 75\ M\^8 - 47\ M\^10 - 
                    142\ M\^12 + 163\ M\^14 + 223\ M\^16 - 391\ M\^18 - 
                    77\ M\^20 + 512\ M\^22 - 230\ M\^24 - 578\ M\^26 + 
                    711\ M\^28 + 748\ M\^30 - 613\ M\^32 - 455\ M\^34 + 
                    521\ M\^36 + 150\ M\^38 - 275\ M\^40 - M\^42 + 
                    105\ M\^44 - 67\ M\^46 + 28\ M\^48 - 8\ M\^50 + 
                    M\^52)\) + 
              L\^6\ \((M\^6 - 11\ M\^8 + 41\ M\^10 - 54\ M\^12 - 4\ M\^14 + 
                    30\ M\^16 + 79\ M\^18 - 107\ M\^20 - 179\ M\^22 + 
                    276\ M\^24 + 424\ M\^26 - 735\ M\^28 - 527\ M\^30 + 
                    1182\ M\^32 + 795\ M\^34 - 791\ M\^36 - 367\ M\^38 + 
                    446\ M\^40 + 131\ M\^42 - 229\ M\^44 + 25\ M\^46 + 
                    80\ M\^48 - 20\ M\^50 - 68\ M\^52 + 61\ M\^54 - 
                    19\ M\^56 + 2\ M\^58)\) + 
              L\^5\ \((2\ M\^12 - 19\ M\^14 + 61\ M\^16 - 68\ M\^18 - 
                    20\ M\^20 + 80\ M\^22 + 25\ M\^24 - 229\ M\^26 + 
                    131\ M\^28 + 446\ M\^30 - 367\ M\^32 - 791\ M\^34 + 
                    795\ M\^36 + 1182\ M\^38 - 527\ M\^40 - 735\ M\^42 + 
                    424\ M\^44 + 276\ M\^46 - 179\ M\^48 - 107\ M\^50 + 
                    79\ M\^52 + 30\ M\^54 - 4\ M\^56 - 54\ M\^58 + 
                    41\ M\^60 - 11\ M\^62 + M\^64)\) + 
              L\ \((4\ M\^48 - 11\ M\^50 + 6\ M\^52 + 17\ M\^54 - 5\ M\^56 - 
                    M\^58 + 4\ M\^60 - 5\ M\^62 + 2\ M\^64)\) + 
              L\^4\ \((M\^18 - 8\ M\^20 + 28\ M\^22 - 67\ M\^24 + 
                    105\ M\^26 - M\^28 - 275\ M\^30 + 150\ M\^32 + 
                    521\ M\^34 - 455\ M\^36 - 613\ M\^38 + 748\ M\^40 + 
                    711\ M\^42 - 578\ M\^44 - 230\ M\^46 + 512\ M\^48 - 
                    77\ M\^50 - 391\ M\^52 + 223\ M\^54 + 163\ M\^56 - 
                    142\ M\^58 - 47\ M\^60 + 75\ M\^62 - 26\ M\^64 + 
                    3\ M\^66)\) + 
              L\^3\ \((4\ M\^28 - 27\ M\^30 + 66\ M\^32 - 53\ M\^34 - 
                    50\ M\^36 + 124\ M\^38 - 41\ M\^40 - 137\ M\^42 + 
                    188\ M\^44 + 104\ M\^46 - 46\ M\^48 + 130\ M\^50 - 
                    35\ M\^52 - 192\ M\^54 + 115\ M\^56 + 87\ M\^58 - 
                    81\ M\^60 - 17\ M\^62 + 42\ M\^64 - 19\ M\^66 + 
                    3\ M\^68)\) + 
              L\^2\ \((6\ M\^38 - 30\ M\^40 + 47\ M\^42 + 17\ M\^44 - 
                    102\ M\^46 + 23\ M\^48 + 142\ M\^50 + M\^52 - 67\ M\^54 - 
                    M\^56 + 23\ M\^58 - 2\ M\^60 - M\^62 - 4\ M\^64 + 
                    6\ M\^66 - 4\ M\^68 + M\^70)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[8, 8]\), \(M\^30 + L\^12\ M\^30 + 
              L\^11\ \((2\ M\^20 - 6\ M\^22 + 7\ M\^24 - M\^26 - 10\ M\^28 + 
                    20\ M\^30 + 9\ M\^32 - 13\ M\^34 + 4\ M\^36)\) + 
              L\ \((4\ M\^24 - 13\ M\^26 + 9\ M\^28 + 20\ M\^30 - 10\ M\^32 - 
                    M\^34 + 7\ M\^36 - 6\ M\^38 + 2\ M\^40)\) + 
              L\^10\ \((M\^10 - 6\ M\^12 + 17\ M\^14 - 18\ M\^16 - 
                    23\ M\^18 + 45\ M\^20 + 44\ M\^22 - 59\ M\^24 - 
                    78\ M\^26 + 23\ M\^28 + 169\ M\^30 + 41\ M\^32 - 
                    136\ M\^34 + 10\ M\^36 + 66\ M\^38 - 36\ M\^40 + 
                    6\ M\^42)\) + 
              L\^9\ \((M\^4 - 8\ M\^6 + 29\ M\^8 - 46\ M\^10 - 23\ M\^12 + 
                    147\ M\^14 - M\^16 - 364\ M\^18 + 119\ M\^20 + 
                    529\ M\^22 - 119\ M\^24 - 620\ M\^26 + 10\ M\^28 + 
                    668\ M\^30 + 279\ M\^32 - 447\ M\^34 - 125\ M\^36 + 
                    287\ M\^38 - 52\ M\^40 - 113\ M\^42 + 98\ M\^44 - 
                    33\ M\^46 + 4\ M\^48)\) + 
              L\^2\ \((6\ M\^18 - 36\ M\^20 + 66\ M\^22 + 10\ M\^24 - 
                    136\ M\^26 + 41\ M\^28 + 169\ M\^30 + 23\ M\^32 - 
                    78\ M\^34 - 59\ M\^36 + 44\ M\^38 + 45\ M\^40 - 
                    23\ M\^42 - 18\ M\^44 + 17\ M\^46 - 6\ M\^48 + M\^50)\) + 
              L\^8\ \((\(-1\) + 9\ M\^2 - 29\ M\^4 + 32\ M\^6 + 17\ M\^8 + 
                    M\^10 - 175\ M\^12 + 14\ M\^14 + 536\ M\^16 + 7\ M\^18 - 
                    1322\ M\^20 + 188\ M\^22 + 2055\ M\^24 - 354\ M\^26 - 
                    2323\ M\^28 + 721\ M\^30 + 2283\ M\^32 - 541\ M\^34 - 
                    1361\ M\^36 + 649\ M\^38 + 470\ M\^40 - 479\ M\^42 - 
                    12\ M\^44 + 202\ M\^46 - 128\ M\^48 + 45\ M\^50 - 
                    10\ M\^52 + M\^54)\) + 
              L\^3\ \((4\ M\^12 - 33\ M\^14 + 98\ M\^16 - 113\ M\^18 - 
                    52\ M\^20 + 287\ M\^22 - 125\ M\^24 - 447\ M\^26 + 
                    279\ M\^28 + 668\ M\^30 + 10\ M\^32 - 620\ M\^34 - 
                    119\ M\^36 + 529\ M\^38 + 119\ M\^40 - 364\ M\^42 - 
                    M\^44 + 147\ M\^46 - 23\ M\^48 - 46\ M\^50 + 29\ M\^52 - 
                    8\ M\^54 + M\^56)\) + 
              L\^7\ \((1 - 12\ M\^2 + 62\ M\^4 - 142\ M\^6 + 27\ M\^8 + 
                    395\ M\^10 - 168\ M\^12 - 1091\ M\^14 + 568\ M\^16 + 
                    1932\ M\^18 - 757\ M\^20 - 2792\ M\^22 + 799\ M\^24 + 
                    2846\ M\^26 - 447\ M\^28 - 1661\ M\^30 + 881\ M\^32 + 
                    765\ M\^34 - 715\ M\^36 + 228\ M\^38 + 677\ M\^40 - 
                    673\ M\^42 - 289\ M\^44 + 462\ M\^46 + 20\ M\^48 - 
                    225\ M\^50 + 130\ M\^52 - 32\ M\^54 + 3\ M\^56)\) + 
              L\^6\ \((3\ M\^2 - 33\ M\^4 + 143\ M\^6 - 243\ M\^8 - 
                    106\ M\^10 + 767\ M\^12 + 2\ M\^14 - 1841\ M\^16 + 
                    357\ M\^18 + 2670\ M\^20 - 361\ M\^22 - 2690\ M\^24 + 
                    216\ M\^26 + 1370\ M\^28 + 416\ M\^30 + 1370\ M\^32 + 
                    216\ M\^34 - 2690\ M\^36 - 361\ M\^38 + 2670\ M\^40 + 
                    357\ M\^42 - 1841\ M\^44 + 2\ M\^46 + 767\ M\^48 - 
                    106\ M\^50 - 243\ M\^52 + 143\ M\^54 - 33\ M\^56 + 
                    3\ M\^58)\) + 
              L\^4\ \((M\^6 - 10\ M\^8 + 45\ M\^10 - 128\ M\^12 + 
                    202\ M\^14 - 12\ M\^16 - 479\ M\^18 + 470\ M\^20 + 
                    649\ M\^22 - 1361\ M\^24 - 541\ M\^26 + 2283\ M\^28 + 
                    721\ M\^30 - 2323\ M\^32 - 354\ M\^34 + 2055\ M\^36 + 
                    188\ M\^38 - 1322\ M\^40 + 7\ M\^42 + 536\ M\^44 + 
                    14\ M\^46 - 175\ M\^48 + M\^50 + 17\ M\^52 + 32\ M\^54 - 
                    29\ M\^56 + 9\ M\^58 - M\^60)\) + 
              L\^5\ \((3\ M\^4 - 32\ M\^6 + 130\ M\^8 - 225\ M\^10 + 
                    20\ M\^12 + 462\ M\^14 - 289\ M\^16 - 673\ M\^18 + 
                    677\ M\^20 + 228\ M\^22 - 715\ M\^24 + 765\ M\^26 + 
                    881\ M\^28 - 1661\ M\^30 - 447\ M\^32 + 2846\ M\^34 + 
                    799\ M\^36 - 2792\ M\^38 - 757\ M\^40 + 1932\ M\^42 + 
                    568\ M\^44 - 1091\ M\^46 - 168\ M\^48 + 395\ M\^50 + 
                    27\ M\^52 - 142\ M\^54 + 62\ M\^56 - 12\ M\^58 + 
                    M\^60)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[8, 9]\), \(M\^30 + L\^12\ M\^30 + 
              L\ \((\(-3\)\ M\^22 + 8\ M\^24 - 8\ M\^26 - 2\ M\^28 + 
                    22\ M\^30 - 2\ M\^32 - 8\ M\^34 + 8\ M\^36 - 
                    3\ M\^38)\) + 
              L\^11\ \((\(-3\)\ M\^22 + 8\ M\^24 - 8\ M\^26 - 2\ M\^28 + 
                    22\ M\^30 - 2\ M\^32 - 8\ M\^34 + 8\ M\^36 - 
                    3\ M\^38)\) + 
              L\^2\ \((3\ M\^14 - 14\ M\^16 + 26\ M\^18 - 15\ M\^20 - 
                    26\ M\^22 + 29\ M\^24 - 3\ M\^26 + 66\ M\^30 - 3\ M\^34 + 
                    29\ M\^36 - 26\ M\^38 - 15\ M\^40 + 26\ M\^42 - 
                    14\ M\^44 + 3\ M\^46)\) + 
              L\^10\ \((3\ M\^14 - 14\ M\^16 + 26\ M\^18 - 15\ M\^20 - 
                    26\ M\^22 + 29\ M\^24 - 3\ M\^26 + 66\ M\^30 - 3\ M\^34 + 
                    29\ M\^36 - 26\ M\^38 - 15\ M\^40 + 26\ M\^42 - 
                    14\ M\^44 + 3\ M\^46)\) + 
              L\^3\ \((\(-M\^6\) + 6\ M\^8 - 15\ M\^10 + 20\ M\^12 - 
                    32\ M\^14 + 78\ M\^16 - 39\ M\^18 - 220\ M\^20 + 
                    274\ M\^22 + 282\ M\^24 - 722\ M\^26 - 56\ M\^28 + 
                    1070\ M\^30 - 56\ M\^32 - 722\ M\^34 + 282\ M\^36 + 
                    274\ M\^38 - 220\ M\^40 - 39\ M\^42 + 78\ M\^44 - 
                    32\ M\^46 + 20\ M\^48 - 15\ M\^50 + 6\ M\^52 - M\^54)\) + 
              L\^9\ \((\(-M\^6\) + 6\ M\^8 - 15\ M\^10 + 20\ M\^12 - 
                    32\ M\^14 + 78\ M\^16 - 39\ M\^18 - 220\ M\^20 + 
                    274\ M\^22 + 282\ M\^24 - 722\ M\^26 - 56\ M\^28 + 
                    1070\ M\^30 - 56\ M\^32 - 722\ M\^34 + 282\ M\^36 + 
                    274\ M\^38 - 220\ M\^40 - 39\ M\^42 + 78\ M\^44 - 
                    32\ M\^46 + 20\ M\^48 - 15\ M\^50 + 6\ M\^52 - M\^54)\) + 
              L\^4\ \((\(-3\)\ M\^4 + 25\ M\^6 - 78\ M\^8 + 79\ M\^10 + 
                    79\ M\^12 - 157\ M\^14 - 153\ M\^16 + 303\ M\^18 + 
                    180\ M\^20 - 238\ M\^22 - 379\ M\^24 - 38\ M\^26 + 
                    354\ M\^28 + 547\ M\^30 + 354\ M\^32 - 38\ M\^34 - 
                    379\ M\^36 - 238\ M\^38 + 180\ M\^40 + 303\ M\^42 - 
                    153\ M\^44 - 157\ M\^46 + 79\ M\^48 + 79\ M\^50 - 
                    78\ M\^52 + 25\ M\^54 - 3\ M\^56)\) + 
              L\^8\ \((\(-3\)\ M\^4 + 25\ M\^6 - 78\ M\^8 + 79\ M\^10 + 
                    79\ M\^12 - 157\ M\^14 - 153\ M\^16 + 303\ M\^18 + 
                    180\ M\^20 - 238\ M\^22 - 379\ M\^24 - 38\ M\^26 + 
                    354\ M\^28 + 547\ M\^30 + 354\ M\^32 - 38\ M\^34 - 
                    379\ M\^36 - 238\ M\^38 + 180\ M\^40 + 303\ M\^42 - 
                    153\ M\^44 - 157\ M\^46 + 79\ M\^48 + 79\ M\^50 - 
                    78\ M\^52 + 25\ M\^54 - 3\ M\^56)\) + 
              L\^5\ \((\(-3\)\ M\^2 + 32\ M\^4 - 121\ M\^6 + 154\ M\^8 + 
                    147\ M\^10 - 460\ M\^12 - 183\ M\^14 + 1050\ M\^16 - 
                    175\ M\^18 - 1214\ M\^20 + 783\ M\^22 + 748\ M\^24 - 
                    2001\ M\^26 + 86\ M\^28 + 3106\ M\^30 + 86\ M\^32 - 
                    2001\ M\^34 + 748\ M\^36 + 783\ M\^38 - 1214\ M\^40 - 
                    175\ M\^42 + 1050\ M\^44 - 183\ M\^46 - 460\ M\^48 + 
                    147\ M\^50 + 154\ M\^52 - 121\ M\^54 + 32\ M\^56 - 
                    3\ M\^58)\) + 
              L\^7\ \((\(-3\)\ M\^2 + 32\ M\^4 - 121\ M\^6 + 154\ M\^8 + 
                    147\ M\^10 - 460\ M\^12 - 183\ M\^14 + 1050\ M\^16 - 
                    175\ M\^18 - 1214\ M\^20 + 783\ M\^22 + 748\ M\^24 - 
                    2001\ M\^26 + 86\ M\^28 + 3106\ M\^30 + 86\ M\^32 - 
                    2001\ M\^34 + 748\ M\^36 + 783\ M\^38 - 1214\ M\^40 - 
                    175\ M\^42 + 1050\ M\^44 - 183\ M\^46 - 460\ M\^48 + 
                    147\ M\^50 + 154\ M\^52 - 121\ M\^54 + 32\ M\^56 - 
                    3\ M\^58)\) + 
              L\^6\ \((\(-1\) + 13\ M\^2 - 61\ M\^4 + 125\ M\^6 - 68\ M\^8 - 
                    134\ M\^10 + 167\ M\^12 + 94\ M\^14 - 395\ M\^16 + 
                    313\ M\^18 + 408\ M\^20 - 285\ M\^22 - 715\ M\^24 - 
                    558\ M\^26 + 665\ M\^28 + 1788\ M\^30 + 665\ M\^32 - 
                    558\ M\^34 - 715\ M\^36 - 285\ M\^38 + 408\ M\^40 + 
                    313\ M\^42 - 395\ M\^44 + 94\ M\^46 + 167\ M\^48 - 
                    134\ M\^50 - 68\ M\^52 + 125\ M\^54 - 61\ M\^56 + 
                    13\ M\^58 - M\^60)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[8, 10]\), \(M\^14 + L\^11\ M\^60 + 
              L\ \((2\ M\^8 - 7\ M\^10 + 15\ M\^12 - 11\ M\^14 - 14\ M\^16 + 
                    31\ M\^18 + 2\ M\^20 - 11\ M\^22 + 4\ M\^24)\) + 
              L\^2\ \((M\^2 - 6\ M\^4 + 19\ M\^6 - 34\ M\^8 + 30\ M\^10 - 
                    52\ M\^14 + 121\ M\^16 - 33\ M\^18 - 205\ M\^20 + 
                    201\ M\^22 + 147\ M\^24 - 192\ M\^26 + 29\ M\^28 + 
                    57\ M\^30 - 34\ M\^32 + 6\ M\^34)\) + 
              L\^3\ \((1 - 8\ M\^2 + 31\ M\^4 - 75\ M\^6 + 88\ M\^8 + 
                    33\ M\^10 - 253\ M\^12 + 295\ M\^14 + 176\ M\^16 - 
                    903\ M\^18 + 626\ M\^20 + 1016\ M\^22 - 1517\ M\^24 - 
                    197\ M\^26 + 1520\ M\^28 - 501\ M\^30 - 601\ M\^32 + 
                    607\ M\^34 - 103\ M\^36 - 159\ M\^38 + 121\ M\^40 - 
                    36\ M\^42 + 4\ M\^44)\) + 
              L\^4\ \((\(-1\) + 11\ M\^2 - 49\ M\^4 + 98\ M\^6 - 23\ M\^8 - 
                    246\ M\^10 + 310\ M\^12 + 355\ M\^14 - 1279\ M\^16 + 
                    605\ M\^18 + 2265\ M\^20 - 2971\ M\^22 - 1821\ M\^24 + 
                    5219\ M\^26 - 466\ M\^28 - 4704\ M\^30 + 2950\ M\^32 + 
                    1701\ M\^34 - 2716\ M\^36 + 852\ M\^38 + 1052\ M\^40 - 
                    1133\ M\^42 + 190\ M\^44 + 295\ M\^46 - 230\ M\^48 + 
                    79\ M\^50 - 14\ M\^52 + M\^54)\) + 
              L\^5\ \((\(-2\)\ M\^6 + 27\ M\^8 - 130\ M\^10 + 253\ M\^12 - 
                    41\ M\^14 - 597\ M\^16 + 826\ M\^18 + 260\ M\^20 - 
                    2174\ M\^22 + 1852\ M\^24 + 2370\ M\^26 - 4216\ M\^28 - 
                    206\ M\^30 + 3447\ M\^32 - 1998\ M\^34 + 947\ M\^36 + 
                    2159\ M\^38 - 4549\ M\^40 + 339\ M\^42 + 4196\ M\^44 - 
                    1842\ M\^46 - 1610\ M\^48 + 1420\ M\^50 - 55\ M\^52 - 
                    393\ M\^54 + 242\ M\^56 - 75\ M\^58 + 13\ M\^60 - 
                    M\^62)\) + 
              L\^10\ \((4\ M\^50 - 11\ M\^52 + 2\ M\^54 + 31\ M\^56 - 
                    14\ M\^58 - 11\ M\^60 + 15\ M\^62 - 7\ M\^64 + 
                    2\ M\^66)\) + 
              L\^6\ \((\(-M\^12\) + 13\ M\^14 - 75\ M\^16 + 242\ M\^18 - 
                    393\ M\^20 - 55\ M\^22 + 1420\ M\^24 - 1610\ M\^26 - 
                    1842\ M\^28 + 4196\ M\^30 + 339\ M\^32 - 4549\ M\^34 + 
                    2159\ M\^36 + 947\ M\^38 - 1998\ M\^40 + 3447\ M\^42 - 
                    206\ M\^44 - 4216\ M\^46 + 2370\ M\^48 + 1852\ M\^50 - 
                    2174\ M\^52 + 260\ M\^54 + 826\ M\^56 - 597\ M\^58 - 
                    41\ M\^60 + 253\ M\^62 - 130\ M\^64 + 27\ M\^66 - 
                    2\ M\^68)\) + 
              L\^9\ \((6\ M\^40 - 34\ M\^42 + 57\ M\^44 + 29\ M\^46 - 
                    192\ M\^48 + 147\ M\^50 + 201\ M\^52 - 205\ M\^54 - 
                    33\ M\^56 + 121\ M\^58 - 52\ M\^60 + 30\ M\^64 - 
                    34\ M\^66 + 19\ M\^68 - 6\ M\^70 + M\^72)\) + 
              L\^7\ \((M\^20 - 14\ M\^22 + 79\ M\^24 - 230\ M\^26 + 
                    295\ M\^28 + 190\ M\^30 - 1133\ M\^32 + 1052\ M\^34 + 
                    852\ M\^36 - 2716\ M\^38 + 1701\ M\^40 + 2950\ M\^42 - 
                    4704\ M\^44 - 466\ M\^46 + 5219\ M\^48 - 1821\ M\^50 - 
                    2971\ M\^52 + 2265\ M\^54 + 605\ M\^56 - 1279\ M\^58 + 
                    355\ M\^60 + 310\ M\^62 - 246\ M\^64 - 23\ M\^66 + 
                    98\ M\^68 - 49\ M\^70 + 11\ M\^72 - M\^74)\) + 
              L\^8\ \((4\ M\^30 - 36\ M\^32 + 121\ M\^34 - 159\ M\^36 - 
                    103\ M\^38 + 607\ M\^40 - 601\ M\^42 - 501\ M\^44 + 
                    1520\ M\^46 - 197\ M\^48 - 1517\ M\^50 + 1016\ M\^52 + 
                    626\ M\^54 - 903\ M\^56 + 176\ M\^58 + 295\ M\^60 - 
                    253\ M\^62 + 33\ M\^64 + 88\ M\^66 - 75\ M\^68 + 
                    31\ M\^70 - 8\ M\^72 + M\^74)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[8, 11]\), \(L\^11\ M\^12 + M\^54 + 
              L\^10\ \((\(-3\)\ M\^8 + 10\ M\^10 - 11\ M\^12 - 5\ M\^14 + 
                    28\ M\^16 - 3\ M\^18 - 11\ M\^20 + 9\ M\^22 - 
                    3\ M\^24)\) + 
              L\^9\ \((3\ M\^4 - 20\ M\^6 + 48\ M\^8 - 25\ M\^10 - 
                    79\ M\^12 + 103\ M\^14 + 20\ M\^16 - 100\ M\^18 + 
                    93\ M\^20 + 41\ M\^22 - 34\ M\^24 + 35\ M\^26 - 
                    32\ M\^28 - 18\ M\^30 + 33\ M\^32 - 16\ M\^34 + 
                    3\ M\^36)\) + 
              L\^8\ \((\(-1\) + 10\ M\^2 - 39\ M\^4 + 67\ M\^6 - 27\ M\^8 - 
                    81\ M\^10 + 154\ M\^12 - 108\ M\^14 - 124\ M\^16 + 
                    432\ M\^18 - 369\ M\^20 - 419\ M\^22 + 977\ M\^24 + 
                    172\ M\^26 - 821\ M\^28 + 237\ M\^30 + 318\ M\^32 - 
                    242\ M\^34 - 16\ M\^36 + 61\ M\^38 - 30\ M\^40 + 
                    29\ M\^42 - 21\ M\^44 + 7\ M\^46 - M\^48)\) + 
              L\^7\ \((2 - 21\ M\^2 + 87\ M\^4 - 141\ M\^6 - 91\ M\^8 + 
                    579\ M\^10 - 202\ M\^12 - 1403\ M\^14 + 1171\ M\^16 + 
                    2142\ M\^18 - 2370\ M\^20 - 2446\ M\^22 + 2936\ M\^24 + 
                    2057\ M\^26 - 2381\ M\^28 - 1001\ M\^30 + 2032\ M\^32 + 
                    502\ M\^34 - 1608\ M\^36 - 242\ M\^38 + 1091\ M\^40 - 
                    76\ M\^42 - 489\ M\^44 + 156\ M\^46 + 123\ M\^48 - 
                    103\ M\^50 + 29\ M\^52 - 3\ M\^54)\) + 
              L\ \((\(-3\)\ M\^42 + 9\ M\^44 - 11\ M\^46 - 3\ M\^48 + 
                    28\ M\^50 - 5\ M\^52 - 11\ M\^54 + 10\ M\^56 - 
                    3\ M\^58)\) + 
              L\^6\ \((\(-1\) + 11\ M\^2 - 49\ M\^4 + 90\ M\^6 + 19\ M\^8 - 
                    286\ M\^10 + 58\ M\^12 + 903\ M\^14 - 427\ M\^16 - 
                    2136\ M\^18 + 855\ M\^20 + 4252\ M\^22 - 783\ M\^24 - 
                    6431\ M\^26 - 970\ M\^28 + 7391\ M\^30 + 3909\ M\^32 - 
                    6169\ M\^34 - 4818\ M\^36 + 4955\ M\^38 + 3567\ M\^40 - 
                    3742\ M\^42 - 1559\ M\^44 + 2296\ M\^46 + 123\ M\^48 - 
                    890\ M\^50 + 210\ M\^52 + 185\ M\^54 - 131\ M\^56 + 
                    33\ M\^58 - 3\ M\^60)\) + 
              L\^2\ \((3\ M\^30 - 16\ M\^32 + 33\ M\^34 - 18\ M\^36 - 
                    32\ M\^38 + 35\ M\^40 - 34\ M\^42 + 41\ M\^44 + 
                    93\ M\^46 - 100\ M\^48 + 20\ M\^50 + 103\ M\^52 - 
                    79\ M\^54 - 25\ M\^56 + 48\ M\^58 - 20\ M\^60 + 
                    3\ M\^62)\) + 
              L\^3\ \((\(-M\^18\) + 7\ M\^20 - 21\ M\^22 + 29\ M\^24 - 
                    30\ M\^26 + 61\ M\^28 - 16\ M\^30 - 242\ M\^32 + 
                    318\ M\^34 + 237\ M\^36 - 821\ M\^38 + 172\ M\^40 + 
                    977\ M\^42 - 419\ M\^44 - 369\ M\^46 + 432\ M\^48 - 
                    124\ M\^50 - 108\ M\^52 + 154\ M\^54 - 81\ M\^56 - 
                    27\ M\^58 + 67\ M\^60 - 39\ M\^62 + 10\ M\^64 - 
                    M\^66)\) + 
              L\^5\ \((\(-3\)\ M\^6 + 33\ M\^8 - 131\ M\^10 + 185\ M\^12 + 
                    210\ M\^14 - 890\ M\^16 + 123\ M\^18 + 2296\ M\^20 - 
                    1559\ M\^22 - 3742\ M\^24 + 3567\ M\^26 + 4955\ M\^28 - 
                    4818\ M\^30 - 6169\ M\^32 + 3909\ M\^34 + 7391\ M\^36 - 
                    970\ M\^38 - 6431\ M\^40 - 783\ M\^42 + 4252\ M\^44 + 
                    855\ M\^46 - 2136\ M\^48 - 427\ M\^50 + 903\ M\^52 + 
                    58\ M\^54 - 286\ M\^56 + 19\ M\^58 + 90\ M\^60 - 
                    49\ M\^62 + 11\ M\^64 - M\^66)\) + 
              L\^4\ \((\(-3\)\ M\^12 + 29\ M\^14 - 103\ M\^16 + 123\ M\^18 + 
                    156\ M\^20 - 489\ M\^22 - 76\ M\^24 + 1091\ M\^26 - 
                    242\ M\^28 - 1608\ M\^30 + 502\ M\^32 + 2032\ M\^34 - 
                    1001\ M\^36 - 2381\ M\^38 + 2057\ M\^40 + 2936\ M\^42 - 
                    2446\ M\^44 - 2370\ M\^46 + 2142\ M\^48 + 1171\ M\^50 - 
                    1403\ M\^52 - 202\ M\^54 + 579\ M\^56 - 91\ M\^58 - 
                    141\ M\^60 + 87\ M\^62 - 21\ M\^64 + 2\ M\^66)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[8, 12]\), \(M\^36 + L\^14\ M\^36 + 
              L\ \((\(-3\)\ M\^28 + 11\ M\^30 - 15\ M\^32 - 4\ M\^34 + 
                    36\ M\^36 - 4\ M\^38 - 15\ M\^40 + 11\ M\^42 - 
                    3\ M\^44)\) + 
              L\^13\ \((\(-3\)\ M\^28 + 11\ M\^30 - 15\ M\^32 - 4\ M\^34 + 
                    36\ M\^36 - 4\ M\^38 - 15\ M\^40 + 11\ M\^42 - 
                    3\ M\^44)\) + 
              L\^2\ \((3\ M\^20 - 22\ M\^22 + 60\ M\^24 - 47\ M\^26 - 
                    67\ M\^28 + 111\ M\^30 - 35\ M\^32 - 42\ M\^34 + 
                    169\ M\^36 - 42\ M\^38 - 35\ M\^40 + 111\ M\^42 - 
                    67\ M\^44 - 47\ M\^46 + 60\ M\^48 - 22\ M\^50 + 
                    3\ M\^52)\) + 
              L\^12\ \((3\ M\^20 - 22\ M\^22 + 60\ M\^24 - 47\ M\^26 - 
                    67\ M\^28 + 111\ M\^30 - 35\ M\^32 - 42\ M\^34 + 
                    169\ M\^36 - 42\ M\^38 - 35\ M\^40 + 111\ M\^42 - 
                    67\ M\^44 - 47\ M\^46 + 60\ M\^48 - 22\ M\^50 + 
                    3\ M\^52)\) + 
              L\^3\ \((\(-M\^12\) + 11\ M\^14 - 48\ M\^16 + 103\ M\^18 - 
                    120\ M\^20 + 22\ M\^22 + 283\ M\^24 - 440\ M\^26 - 
                    176\ M\^28 + 754\ M\^30 - 469\ M\^32 - 268\ M\^34 + 
                    1062\ M\^36 - 268\ M\^38 - 469\ M\^40 + 754\ M\^42 - 
                    176\ M\^44 - 440\ M\^46 + 283\ M\^48 + 22\ M\^50 - 
                    120\ M\^52 + 103\ M\^54 - 48\ M\^56 + 11\ M\^58 - 
                    M\^60)\) + 
              L\^11\ \((\(-M\^12\) + 11\ M\^14 - 48\ M\^16 + 103\ M\^18 - 
                    120\ M\^20 + 22\ M\^22 + 283\ M\^24 - 440\ M\^26 - 
                    176\ M\^28 + 754\ M\^30 - 469\ M\^32 - 268\ M\^34 + 
                    1062\ M\^36 - 268\ M\^38 - 469\ M\^40 + 754\ M\^42 - 
                    176\ M\^44 - 440\ M\^46 + 283\ M\^48 + 22\ M\^50 - 
                    120\ M\^52 + 103\ M\^54 - 48\ M\^56 + 11\ M\^58 - 
                    M\^60)\) + 
              L\^4\ \((3\ M\^8 - 35\ M\^10 + 154\ M\^12 - 285\ M\^14 + 
                    56\ M\^16 + 552\ M\^18 - 563\ M\^20 - 540\ M\^22 + 
                    1380\ M\^24 - 280\ M\^26 - 1538\ M\^28 + 600\ M\^30 + 
                    874\ M\^32 - 12\ M\^34 + 269\ M\^36 - 12\ M\^38 + 
                    874\ M\^40 + 600\ M\^42 - 1538\ M\^44 - 280\ M\^46 + 
                    1380\ M\^48 - 540\ M\^50 - 563\ M\^52 + 552\ M\^54 + 
                    56\ M\^56 - 285\ M\^58 + 154\ M\^60 - 35\ M\^62 + 
                    3\ M\^64)\) + 
              L\^10\ \((3\ M\^8 - 35\ M\^10 + 154\ M\^12 - 285\ M\^14 + 
                    56\ M\^16 + 552\ M\^18 - 563\ M\^20 - 540\ M\^22 + 
                    1380\ M\^24 - 280\ M\^26 - 1538\ M\^28 + 600\ M\^30 + 
                    874\ M\^32 - 12\ M\^34 + 269\ M\^36 - 12\ M\^38 + 
                    874\ M\^40 + 600\ M\^42 - 1538\ M\^44 - 280\ M\^46 + 
                    1380\ M\^48 - 540\ M\^50 - 563\ M\^52 + 552\ M\^54 + 
                    56\ M\^56 - 285\ M\^58 + 154\ M\^60 - 35\ M\^62 + 
                    3\ M\^64)\) + 
              L\^5\ \((\(-3\)\ M\^4 + 37\ M\^6 - 175\ M\^8 + 365\ M\^10 - 
                    168\ M\^12 - 609\ M\^14 + 869\ M\^16 + 349\ M\^18 - 
                    1793\ M\^20 + 1922\ M\^22 - 6\ M\^24 - 4521\ M\^26 + 
                    5689\ M\^28 + 4728\ M\^30 - 14289\ M\^32 - 1270\ M\^34 + 
                    19752\ M\^36 - 1270\ M\^38 - 14289\ M\^40 + 4728\ M\^42 + 
                    5689\ M\^44 - 4521\ M\^46 - 6\ M\^48 + 1922\ M\^50 - 
                    1793\ M\^52 + 349\ M\^54 + 869\ M\^56 - 609\ M\^58 - 
                    168\ M\^60 + 365\ M\^62 - 175\ M\^64 + 37\ M\^66 - 
                    3\ M\^68)\) + 
              L\^9\ \((\(-3\)\ M\^4 + 37\ M\^6 - 175\ M\^8 + 365\ M\^10 - 
                    168\ M\^12 - 609\ M\^14 + 869\ M\^16 + 349\ M\^18 - 
                    1793\ M\^20 + 1922\ M\^22 - 6\ M\^24 - 4521\ M\^26 + 
                    5689\ M\^28 + 4728\ M\^30 - 14289\ M\^32 - 1270\ M\^34 + 
                    19752\ M\^36 - 1270\ M\^38 - 14289\ M\^40 + 4728\ M\^42 + 
                    5689\ M\^44 - 4521\ M\^46 - 6\ M\^48 + 1922\ M\^50 - 
                    1793\ M\^52 + 349\ M\^54 + 869\ M\^56 - 609\ M\^58 - 
                    168\ M\^60 + 365\ M\^62 - 175\ M\^64 + 37\ M\^66 - 
                    3\ M\^68)\) + 
              L\^7\ \((\(-2\) + 27\ M\^2 - 143\ M\^4 + 348\ M\^6 - 
                    210\ M\^8 - 775\ M\^10 + 1066\ M\^12 + 2476\ M\^14 - 
                    5483\ M\^16 - 4936\ M\^18 + 17933\ M\^20 + 8023\ M\^22 - 
                    47011\ M\^24 - 7629\ M\^26 + 92880\ M\^28 + 3286\ M\^30 - 
                    141994\ M\^32 + 896\ M\^34 + 165928\ M\^36 + 896\ M\^38 - 
                    141994\ M\^40 + 3286\ M\^42 + 92880\ M\^44 - 
                    7629\ M\^46 - 47011\ M\^48 + 8023\ M\^50 + 17933\ M\^52 - 
                    4936\ M\^54 - 5483\ M\^56 + 2476\ M\^58 + 1066\ M\^60 - 
                    775\ M\^62 - 210\ M\^64 + 348\ M\^66 - 143\ M\^68 + 
                    27\ M\^70 - 2\ M\^72)\) + 
              L\^6\ \((1 - 13\ M\^2 + 68\ M\^4 - 188\ M\^6 + 311\ M\^8 - 
                    279\ M\^10 - 314\ M\^12 + 1647\ M\^14 - 957\ M\^16 - 
                    5334\ M\^18 + 8053\ M\^20 + 9216\ M\^22 - 24416\ M\^24 - 
                    11061\ M\^26 + 51479\ M\^28 + 7421\ M\^30 - 
                    80653\ M\^32 - 1409\ M\^34 + 95859\ M\^36 - 1409\ M\^38 - 
                    80653\ M\^40 + 7421\ M\^42 + 51479\ M\^44 - 
                    11061\ M\^46 - 24416\ M\^48 + 9216\ M\^50 + 8053\ M\^52 - 
                    5334\ M\^54 - 957\ M\^56 + 1647\ M\^58 - 314\ M\^60 - 
                    279\ M\^62 + 311\ M\^64 - 188\ M\^66 + 68\ M\^68 - 
                    13\ M\^70 + M\^72)\) + 
              L\^8\ \((1 - 13\ M\^2 + 68\ M\^4 - 188\ M\^6 + 311\ M\^8 - 
                    279\ M\^10 - 314\ M\^12 + 1647\ M\^14 - 957\ M\^16 - 
                    5334\ M\^18 + 8053\ M\^20 + 9216\ M\^22 - 24416\ M\^24 - 
                    11061\ M\^26 + 51479\ M\^28 + 7421\ M\^30 - 
                    80653\ M\^32 - 1409\ M\^34 + 95859\ M\^36 - 1409\ M\^38 - 
                    80653\ M\^40 + 7421\ M\^42 + 51479\ M\^44 - 
                    11061\ M\^46 - 24416\ M\^48 + 9216\ M\^50 + 8053\ M\^52 - 
                    5334\ M\^54 - 957\ M\^56 + 1647\ M\^58 - 314\ M\^60 - 
                    279\ M\^62 + 311\ M\^64 - 188\ M\^66 + 68\ M\^68 - 
                    13\ M\^70 + M\^72)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[8, 13]\), \(M\^36 + L\^14\ M\^36 + 
              L\^13\ \((2\ M\^26 - 7\ M\^28 + 11\ M\^30 - 5\ M\^32 - 
                    15\ M\^34 + 31\ M\^36 + 10\ M\^38 - 19\ M\^40 + 
                    6\ M\^42)\) + 
              L\ \((6\ M\^30 - 19\ M\^32 + 10\ M\^34 + 31\ M\^36 - 
                    15\ M\^38 - 5\ M\^40 + 11\ M\^42 - 7\ M\^44 + 
                    2\ M\^46)\) + 
              L\^12\ \((M\^16 - 6\ M\^18 + 15\ M\^20 - 20\ M\^22 + 
                    22\ M\^24 - 40\ M\^26 + 20\ M\^28 + 166\ M\^30 - 
                    248\ M\^32 - 192\ M\^34 + 531\ M\^36 + 94\ M\^38 - 
                    408\ M\^40 + 86\ M\^42 + 143\ M\^44 - 88\ M\^46 + 
                    15\ M\^48)\) + 
              L\^11\ \((M\^10 - 10\ M\^12 + 43\ M\^14 - 96\ M\^16 + 
                    69\ M\^18 + 131\ M\^20 - 181\ M\^22 - 263\ M\^24 + 
                    237\ M\^26 + 820\ M\^28 - 195\ M\^30 - 2157\ M\^32 + 
                    517\ M\^34 + 3417\ M\^36 - 889\ M\^38 - 2821\ M\^40 + 
                    1442\ M\^42 + 1276\ M\^44 - 1120\ M\^46 - 135\ M\^48 + 
                    420\ M\^50 - 162\ M\^52 + 20\ M\^54)\) + 
              L\^2\ \((15\ M\^24 - 88\ M\^26 + 143\ M\^28 + 86\ M\^30 - 
                    408\ M\^32 + 94\ M\^34 + 531\ M\^36 - 192\ M\^38 - 
                    248\ M\^40 + 166\ M\^42 + 20\ M\^44 - 40\ M\^46 + 
                    22\ M\^48 - 20\ M\^50 + 15\ M\^52 - 6\ M\^54 + M\^56)\) + 
              L\^10\ \((\(-2\)\ M\^6 + 24\ M\^8 - 115\ M\^10 + 258\ M\^12 - 
                    120\ M\^14 - 641\ M\^16 + 939\ M\^18 + 1095\ M\^20 - 
                    3002\ M\^22 - 879\ M\^24 + 5728\ M\^26 + 619\ M\^28 - 
                    8552\ M\^30 - 1018\ M\^32 + 10154\ M\^34 + 2413\ M\^36 - 
                    9136\ M\^38 - 1750\ M\^40 + 7677\ M\^42 + 509\ M\^44 - 
                    5226\ M\^46 + 817\ M\^48 + 2317\ M\^50 - 970\ M\^52 - 
                    514\ M\^54 + 509\ M\^56 - 148\ M\^58 + 15\ M\^60)\) + 
              L\^3\ \((20\ M\^18 - 162\ M\^20 + 420\ M\^22 - 135\ M\^24 - 
                    1120\ M\^26 + 1276\ M\^28 + 1442\ M\^30 - 2821\ M\^32 - 
                    889\ M\^34 + 3417\ M\^36 + 517\ M\^38 - 2157\ M\^40 - 
                    195\ M\^42 + 820\ M\^44 + 237\ M\^46 - 263\ M\^48 - 
                    181\ M\^50 + 131\ M\^52 + 69\ M\^54 - 96\ M\^56 + 
                    43\ M\^58 - 10\ M\^60 + M\^62)\) + 
              L\^4\ \((15\ M\^12 - 148\ M\^14 + 509\ M\^16 - 514\ M\^18 - 
                    970\ M\^20 + 2317\ M\^22 + 817\ M\^24 - 5226\ M\^26 + 
                    509\ M\^28 + 7677\ M\^30 - 1750\ M\^32 - 9136\ M\^34 + 
                    2413\ M\^36 + 10154\ M\^38 - 1018\ M\^40 - 8552\ M\^42 + 
                    619\ M\^44 + 5728\ M\^46 - 879\ M\^48 - 3002\ M\^50 + 
                    1095\ M\^52 + 939\ M\^54 - 641\ M\^56 - 120\ M\^58 + 
                    258\ M\^60 - 115\ M\^62 + 24\ M\^64 - 2\ M\^66)\) + 
              L\^9\ \((M\^2 - 16\ M\^4 + 98\ M\^6 - 288\ M\^8 + 330\ M\^10 + 
                    352\ M\^12 - 1391\ M\^14 + 392\ M\^16 + 3608\ M\^18 - 
                    4196\ M\^20 - 5382\ M\^22 + 11133\ M\^24 + 6576\ M\^26 - 
                    20088\ M\^28 - 6778\ M\^30 + 26483\ M\^32 + 5965\ M\^34 - 
                    26328\ M\^36 - 1692\ M\^38 + 22374\ M\^40 - 264\ M\^42 - 
                    14320\ M\^44 + 877\ M\^46 + 6971\ M\^48 - 282\ M\^50 - 
                    3135\ M\^52 + 211\ M\^54 + 1186\ M\^56 - 161\ M\^58 - 
                    453\ M\^60 + 280\ M\^62 - 67\ M\^64 + 6\ M\^66)\) + 
              L\^5\ \((6\ M\^6 - 67\ M\^8 + 280\ M\^10 - 453\ M\^12 - 
                    161\ M\^14 + 1186\ M\^16 + 211\ M\^18 - 3135\ M\^20 - 
                    282\ M\^22 + 6971\ M\^24 + 877\ M\^26 - 14320\ M\^28 - 
                    264\ M\^30 + 22374\ M\^32 - 1692\ M\^34 - 26328\ M\^36 + 
                    5965\ M\^38 + 26483\ M\^40 - 6778\ M\^42 - 20088\ M\^44 + 
                    6576\ M\^46 + 11133\ M\^48 - 5382\ M\^50 - 4196\ M\^52 + 
                    3608\ M\^54 + 392\ M\^56 - 1391\ M\^58 + 352\ M\^60 + 
                    330\ M\^62 - 288\ M\^64 + 98\ M\^66 - 16\ M\^68 + 
                    M\^70)\) + 
              L\^7\ \((\(-2\) + 26\ M\^2 - 139\ M\^4 + 351\ M\^6 - 
                    232\ M\^8 - 768\ M\^10 + 1076\ M\^12 + 2272\ M\^14 - 
                    4643\ M\^16 - 4160\ M\^18 + 12366\ M\^20 + 6882\ M\^22 - 
                    27465\ M\^24 - 7544\ M\^26 + 47282\ M\^28 + 5444\ M\^30 - 
                    66058\ M\^32 - 787\ M\^34 + 75630\ M\^36 - 787\ M\^38 - 
                    66058\ M\^40 + 5444\ M\^42 + 47282\ M\^44 - 7544\ M\^46 - 
                    27465\ M\^48 + 6882\ M\^50 + 12366\ M\^52 - 4160\ M\^54 - 
                    4643\ M\^56 + 2272\ M\^58 + 1076\ M\^60 - 768\ M\^62 - 
                    232\ M\^64 + 351\ M\^66 - 139\ M\^68 + 26\ M\^70 - 
                    2\ M\^72)\) + 
              L\^6\ \((1 - 12\ M\^2 + 58\ M\^4 - 130\ M\^6 + 113\ M\^8 - 
                    96\ M\^10 + 613\ M\^12 - 806\ M\^14 - 2059\ M\^16 + 
                    4317\ M\^18 + 4419\ M\^20 - 13635\ M\^22 - 4701\ M\^24 + 
                    27191\ M\^26 + 2353\ M\^28 - 42112\ M\^30 + 2505\ M\^32 + 
                    52732\ M\^34 - 6703\ M\^36 - 52559\ M\^38 + 
                    13798\ M\^40 + 46426\ M\^42 - 16089\ M\^44 - 
                    33470\ M\^46 + 14529\ M\^48 + 18153\ M\^50 - 
                    10122\ M\^52 - 6945\ M\^54 + 5681\ M\^56 + 1027\ M\^58 - 
                    1961\ M\^60 + 248\ M\^62 + 472\ M\^64 - 314\ M\^66 + 
                    95\ M\^68 - 15\ M\^70 + M\^72)\) + 
              L\^8\ \((1 - 15\ M\^2 + 95\ M\^4 - 314\ M\^6 + 472\ M\^8 + 
                    248\ M\^10 - 1961\ M\^12 + 1027\ M\^14 + 5681\ M\^16 - 
                    6945\ M\^18 - 10122\ M\^20 + 18153\ M\^22 + 
                    14529\ M\^24 - 33470\ M\^26 - 16089\ M\^28 + 
                    46426\ M\^30 + 13798\ M\^32 - 52559\ M\^34 - 
                    6703\ M\^36 + 52732\ M\^38 + 2505\ M\^40 - 42112\ M\^42 + 
                    2353\ M\^44 + 27191\ M\^46 - 4701\ M\^48 - 13635\ M\^50 + 
                    4419\ M\^52 + 4317\ M\^54 - 2059\ M\^56 - 806\ M\^58 + 
                    613\ M\^60 - 96\ M\^62 + 113\ M\^64 - 130\ M\^66 + 
                    58\ M\^68 - 12\ M\^70 + M\^72)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[8, 14]\), \(M\^12 + L\^15\ M\^74 + 
              L\ \((\(-3\)\ M\^8 + 12\ M\^10 - 17\ M\^12 - 3\ M\^14 + 
                    39\ M\^16 - 8\ M\^18 - 15\ M\^20 + 14\ M\^22 - 
                    4\ M\^24)\) + 
              L\^2\ \((3\ M\^4 - 22\ M\^6 + 59\ M\^8 - 56\ M\^10 - 
                    21\ M\^12 + 69\ M\^14 - 108\ M\^16 + 125\ M\^18 + 
                    155\ M\^20 - 249\ M\^22 + 99\ M\^24 + 219\ M\^26 - 
                    187\ M\^28 - 46\ M\^30 + 99\ M\^32 - 40\ M\^34 + 
                    6\ M\^36)\) + 
              L\^3\ \((\(-1\) + 10\ M\^2 - 42\ M\^4 + 108\ M\^6 - 212\ M\^8 + 
                    203\ M\^10 + 359\ M\^12 - 987\ M\^14 - 104\ M\^16 + 
                    1754\ M\^18 - 525\ M\^20 - 1703\ M\^22 + 843\ M\^24 + 
                    1449\ M\^26 - 29\ M\^28 - 795\ M\^30 - 395\ M\^32 + 
                    655\ M\^34 + 177\ M\^36 - 477\ M\^38 + 73\ M\^40 + 
                    200\ M\^42 - 140\ M\^44 + 38\ M\^46 - 4\ M\^48)\) + 
              L\^4\ \((3 - 37\ M\^2 + 177\ M\^4 - 375\ M\^6 + 199\ M\^8 + 
                    445\ M\^10 - 530\ M\^12 + 97\ M\^14 - 80\ M\^16 - 
                    1740\ M\^18 + 3481\ M\^20 + 2628\ M\^22 - 8635\ M\^24 - 
                    1213\ M\^26 + 12078\ M\^28 - 1385\ M\^30 - 8710\ M\^32 + 
                    4666\ M\^34 + 3664\ M\^36 - 4281\ M\^38 - 190\ M\^40 + 
                    1570\ M\^42 - 353\ M\^44 - 152\ M\^46 + 35\ M\^48 - 
                    71\ M\^50 + 167\ M\^52 - 140\ M\^54 + 58\ M\^56 - 
                    12\ M\^58 + M\^60)\) + 
              L\^5\ \((\(-3\) + 43\ M\^2 - 248\ M\^4 + 671\ M\^6 - 
                    526\ M\^8 - 1430\ M\^10 + 2705\ M\^12 + 2531\ M\^14 - 
                    8259\ M\^16 - 1821\ M\^18 + 15481\ M\^20 - 1339\ M\^22 - 
                    21247\ M\^24 + 7720\ M\^26 + 20156\ M\^28 - 
                    13923\ M\^30 - 12925\ M\^32 + 16830\ M\^34 + 
                    8661\ M\^36 - 8748\ M\^38 - 4646\ M\^40 + 1311\ M\^42 + 
                    2608\ M\^44 + 820\ M\^46 - 2024\ M\^48 + 378\ M\^50 + 
                    530\ M\^52 - 475\ M\^54 + 148\ M\^56 + 199\ M\^58 - 
                    332\ M\^60 + 222\ M\^62 - 78\ M\^64 + 14\ M\^66 - 
                    M\^68)\) + 
              L\^6\ \((1 - 16\ M\^2 + 109\ M\^4 - 405\ M\^6 + 819\ M\^8 - 
                    483\ M\^10 - 1688\ M\^12 + 3439\ M\^14 + 1706\ M\^16 - 
                    11184\ M\^18 + 4654\ M\^20 + 20523\ M\^22 - 
                    18998\ M\^24 - 28928\ M\^26 + 40319\ M\^28 + 
                    33615\ M\^30 - 58438\ M\^32 - 37681\ M\^34 + 
                    62452\ M\^36 + 40302\ M\^38 - 43222\ M\^40 - 
                    26536\ M\^42 + 23612\ M\^44 + 10137\ M\^46 - 
                    8538\ M\^48 - 1893\ M\^50 - 188\ M\^52 + 500\ M\^54 + 
                    3287\ M\^56 - 2079\ M\^58 - 1479\ M\^60 + 1560\ M\^62 + 
                    151\ M\^64 - 751\ M\^66 + 429\ M\^68 - 119\ M\^70 + 
                    17\ M\^72 - M\^74)\) + 
              L\^14\ \((\(-4\)\ M\^62 + 14\ M\^64 - 15\ M\^66 - 8\ M\^68 + 
                    39\ M\^70 - 3\ M\^72 - 17\ M\^74 + 12\ M\^76 - 
                    3\ M\^78)\) + 
              L\^7\ \((\(-3\)\ M\^4 + 47\ M\^6 - 303\ M\^8 + 978\ M\^10 - 
                    1269\ M\^12 - 1456\ M\^14 + 6376\ M\^16 - 1023\ M\^18 - 
                    19764\ M\^20 + 15925\ M\^22 + 40777\ M\^24 - 
                    48584\ M\^26 - 70123\ M\^28 + 98204\ M\^30 + 
                    102643\ M\^32 - 143991\ M\^34 - 131180\ M\^36 + 
                    160673\ M\^38 + 139227\ M\^40 - 139158\ M\^42 - 
                    113292\ M\^44 + 112077\ M\^46 + 81848\ M\^48 - 
                    82687\ M\^50 - 53994\ M\^52 + 55829\ M\^54 + 
                    28995\ M\^56 - 33079\ M\^58 - 10817\ M\^60 + 
                    17004\ M\^62 + 461\ M\^64 - 5859\ M\^66 + 1433\ M\^68 + 
                    1230\ M\^70 - 968\ M\^72 + 302\ M\^74 - 47\ M\^76 + 
                    3\ M\^78)\) + 
              L\^8\ \((3\ M\^8 - 47\ M\^10 + 302\ M\^12 - 968\ M\^14 + 
                    1230\ M\^16 + 1433\ M\^18 - 5859\ M\^20 + 461\ M\^22 + 
                    17004\ M\^24 - 10817\ M\^26 - 33079\ M\^28 + 
                    28995\ M\^30 + 55829\ M\^32 - 53994\ M\^34 - 
                    82687\ M\^36 + 81848\ M\^38 + 112077\ M\^40 - 
                    113292\ M\^42 - 139158\ M\^44 + 139227\ M\^46 + 
                    160673\ M\^48 - 131180\ M\^50 - 143991\ M\^52 + 
                    102643\ M\^54 + 98204\ M\^56 - 70123\ M\^58 - 
                    48584\ M\^60 + 40777\ M\^62 + 15925\ M\^64 - 
                    19764\ M\^66 - 1023\ M\^68 + 6376\ M\^70 - 1456\ M\^72 - 
                    1269\ M\^74 + 978\ M\^76 - 303\ M\^78 + 47\ M\^80 - 
                    3\ M\^82)\) + 
              L\^13\ \((6\ M\^50 - 40\ M\^52 + 99\ M\^54 - 46\ M\^56 - 
                    187\ M\^58 + 219\ M\^60 + 99\ M\^62 - 249\ M\^64 + 
                    155\ M\^66 + 125\ M\^68 - 108\ M\^70 + 69\ M\^72 - 
                    21\ M\^74 - 56\ M\^76 + 59\ M\^78 - 22\ M\^80 + 
                    3\ M\^82)\) + 
              L\^10\ \((\(-M\^18\) + 14\ M\^20 - 78\ M\^22 + 222\ M\^24 - 
                    332\ M\^26 + 199\ M\^28 + 148\ M\^30 - 475\ M\^32 + 
                    530\ M\^34 + 378\ M\^36 - 2024\ M\^38 + 820\ M\^40 + 
                    2608\ M\^42 + 1311\ M\^44 - 4646\ M\^46 - 8748\ M\^48 + 
                    8661\ M\^50 + 16830\ M\^52 - 12925\ M\^54 - 
                    13923\ M\^56 + 20156\ M\^58 + 7720\ M\^60 - 
                    21247\ M\^62 - 1339\ M\^64 + 15481\ M\^66 - 1821\ M\^68 - 
                    8259\ M\^70 + 2531\ M\^72 + 2705\ M\^74 - 1430\ M\^76 - 
                    526\ M\^78 + 671\ M\^80 - 248\ M\^82 + 43\ M\^84 - 
                    3\ M\^86)\) + 
              L\^12\ \((\(-4\)\ M\^38 + 38\ M\^40 - 140\ M\^42 + 200\ M\^44 + 
                    73\ M\^46 - 477\ M\^48 + 177\ M\^50 + 655\ M\^52 - 
                    395\ M\^54 - 795\ M\^56 - 29\ M\^58 + 1449\ M\^60 + 
                    843\ M\^62 - 1703\ M\^64 - 525\ M\^66 + 1754\ M\^68 - 
                    104\ M\^70 - 987\ M\^72 + 359\ M\^74 + 203\ M\^76 - 
                    212\ M\^78 + 108\ M\^80 - 42\ M\^82 + 10\ M\^84 - 
                    M\^86)\) + 
              L\^9\ \((\(-M\^12\) + 17\ M\^14 - 119\ M\^16 + 429\ M\^18 - 
                    751\ M\^20 + 151\ M\^22 + 1560\ M\^24 - 1479\ M\^26 - 
                    2079\ M\^28 + 3287\ M\^30 + 500\ M\^32 - 188\ M\^34 - 
                    1893\ M\^36 - 8538\ M\^38 + 10137\ M\^40 + 23612\ M\^42 - 
                    26536\ M\^44 - 43222\ M\^46 + 40302\ M\^48 + 
                    62452\ M\^50 - 37681\ M\^52 - 58438\ M\^54 + 
                    33615\ M\^56 + 40319\ M\^58 - 28928\ M\^60 - 
                    18998\ M\^62 + 20523\ M\^64 + 4654\ M\^66 - 
                    11184\ M\^68 + 1706\ M\^70 + 3439\ M\^72 - 1688\ M\^74 - 
                    483\ M\^76 + 819\ M\^78 - 405\ M\^80 + 109\ M\^82 - 
                    16\ M\^84 + M\^86)\) + 
              L\^11\ \((M\^26 - 12\ M\^28 + 58\ M\^30 - 140\ M\^32 + 
                    167\ M\^34 - 71\ M\^36 + 35\ M\^38 - 152\ M\^40 - 
                    353\ M\^42 + 1570\ M\^44 - 190\ M\^46 - 4281\ M\^48 + 
                    3664\ M\^50 + 4666\ M\^52 - 8710\ M\^54 - 1385\ M\^56 + 
                    12078\ M\^58 - 1213\ M\^60 - 8635\ M\^62 + 2628\ M\^64 + 
                    3481\ M\^66 - 1740\ M\^68 - 80\ M\^70 + 97\ M\^72 - 
                    530\ M\^74 + 445\ M\^76 + 199\ M\^78 - 375\ M\^80 + 
                    177\ M\^82 - 37\ M\^84 + 3\ M\^86)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[8, 15]\), \(L\^12 + M\^92 + 
              L\^11\ \((\(-2\) + 10\ M\^2 - 12\ M\^4 - 18\ M\^6 + 47\ M\^8 + 
                    4\ M\^10 - 28\ M\^12 + 14\ M\^14 - 3\ M\^16)\) + 
              L\^10\ \((1 - 11\ M\^2 + 47\ M\^4 - 86\ M\^6 + 14\ M\^8 + 
                    234\ M\^10 - 363\ M\^12 - 81\ M\^14 + 564\ M\^16 - 
                    116\ M\^18 - 296\ M\^20 + 207\ M\^22 - M\^24 - 
                    103\ M\^26 + 77\ M\^28 - 24\ M\^30 + 3\ M\^32)\) + 
              L\^9\ \((M\^2 - 14\ M\^4 + 81\ M\^6 - 231\ M\^8 + 262\ M\^10 + 
                    231\ M\^12 - 1024\ M\^14 + 892\ M\^16 + 905\ M\^18 - 
                    2415\ M\^20 + 169\ M\^22 + 3088\ M\^24 - 1082\ M\^26 - 
                    1720\ M\^28 + 1252\ M\^30 + 237\ M\^32 - 678\ M\^34 + 
                    234\ M\^36 + 132\ M\^38 - 169\ M\^40 + 102\ M\^42 - 
                    42\ M\^44 + 10\ M\^46 - M\^48)\) + 
              L\^8\ \((\(-3\)\ M\^6 + 41\ M\^8 - 221\ M\^10 + 536\ M\^12 - 
                    297\ M\^14 - 1150\ M\^16 + 1788\ M\^18 + 788\ M\^20 - 
                    2756\ M\^22 + 363\ M\^24 + 1876\ M\^26 - 1883\ M\^28 + 
                    470\ M\^30 + 2028\ M\^32 - 724\ M\^34 - 134\ M\^36 + 
                    419\ M\^38 - 1327\ M\^40 + 72\ M\^42 + 1209\ M\^44 - 
                    427\ M\^46 - 460\ M\^48 + 264\ M\^50 + 160\ M\^52 - 
                    206\ M\^54 + 83\ M\^56 - 15\ M\^58 + M\^60)\) + 
              L\^7\ \((3\ M\^10 - 42\ M\^12 + 231\ M\^14 - 570\ M\^16 + 
                    317\ M\^18 + 1244\ M\^20 - 1641\ M\^22 - 2006\ M\^24 + 
                    3371\ M\^26 + 3007\ M\^28 - 3822\ M\^30 - 4174\ M\^32 + 
                    906\ M\^34 + 4711\ M\^36 + 2581\ M\^38 - 2505\ M\^40 - 
                    3469\ M\^42 + 2052\ M\^44 + 3346\ M\^46 - 1857\ M\^48 - 
                    3096\ M\^50 + 1481\ M\^52 + 1999\ M\^54 - 1065\ M\^56 - 
                    677\ M\^58 + 427\ M\^60 + 225\ M\^62 - 267\ M\^64 + 
                    97\ M\^66 - 16\ M\^68 + M\^70)\) + 
              L\^6\ \((\(-M\^14\) + 14\ M\^16 - 80\ M\^18 + 231\ M\^20 - 
                    312\ M\^22 + 2\ M\^24 + 508\ M\^26 - 224\ M\^28 - 
                    852\ M\^30 + 595\ M\^32 + 825\ M\^34 + 3\ M\^36 - 
                    359\ M\^38 - 1782\ M\^40 - 761\ M\^42 + 1371\ M\^44 + 
                    2568\ M\^46 + 1371\ M\^48 - 761\ M\^50 - 1782\ M\^52 - 
                    359\ M\^54 + 3\ M\^56 + 825\ M\^58 + 595\ M\^60 - 
                    852\ M\^62 - 224\ M\^64 + 508\ M\^66 + 2\ M\^68 - 
                    312\ M\^70 + 231\ M\^72 - 80\ M\^74 + 14\ M\^76 - 
                    M\^78)\) + 
              L\^5\ \((M\^22 - 16\ M\^24 + 97\ M\^26 - 267\ M\^28 + 
                    225\ M\^30 + 427\ M\^32 - 677\ M\^34 - 1065\ M\^36 + 
                    1999\ M\^38 + 1481\ M\^40 - 3096\ M\^42 - 1857\ M\^44 + 
                    3346\ M\^46 + 2052\ M\^48 - 3469\ M\^50 - 2505\ M\^52 + 
                    2581\ M\^54 + 4711\ M\^56 + 906\ M\^58 - 4174\ M\^60 - 
                    3822\ M\^62 + 3007\ M\^64 + 3371\ M\^66 - 2006\ M\^68 - 
                    1641\ M\^70 + 1244\ M\^72 + 317\ M\^74 - 570\ M\^76 + 
                    231\ M\^78 - 42\ M\^80 + 3\ M\^82)\) + 
              L\^4\ \((M\^32 - 15\ M\^34 + 83\ M\^36 - 206\ M\^38 + 
                    160\ M\^40 + 264\ M\^42 - 460\ M\^44 - 427\ M\^46 + 
                    1209\ M\^48 + 72\ M\^50 - 1327\ M\^52 + 419\ M\^54 - 
                    134\ M\^56 - 724\ M\^58 + 2028\ M\^60 + 470\ M\^62 - 
                    1883\ M\^64 + 1876\ M\^66 + 363\ M\^68 - 2756\ M\^70 + 
                    788\ M\^72 + 1788\ M\^74 - 1150\ M\^76 - 297\ M\^78 + 
                    536\ M\^80 - 221\ M\^82 + 41\ M\^84 - 3\ M\^86)\) + 
              L\^3\ \((\(-M\^44\) + 10\ M\^46 - 42\ M\^48 + 102\ M\^50 - 
                    169\ M\^52 + 132\ M\^54 + 234\ M\^56 - 678\ M\^58 + 
                    237\ M\^60 + 1252\ M\^62 - 1720\ M\^64 - 1082\ M\^66 + 
                    3088\ M\^68 + 169\ M\^70 - 2415\ M\^72 + 905\ M\^74 + 
                    892\ M\^76 - 1024\ M\^78 + 231\ M\^80 + 262\ M\^82 - 
                    231\ M\^84 + 81\ M\^86 - 14\ M\^88 + M\^90)\) + 
              L\ \((\(-3\)\ M\^76 + 14\ M\^78 - 28\ M\^80 + 4\ M\^82 + 
                    47\ M\^84 - 18\ M\^86 - 12\ M\^88 + 10\ M\^90 - 
                    2\ M\^92)\) + 
              L\^2\ \((3\ M\^60 - 24\ M\^62 + 77\ M\^64 - 103\ M\^66 - 
                    M\^68 + 207\ M\^70 - 296\ M\^72 - 116\ M\^74 + 
                    564\ M\^76 - 81\ M\^78 - 363\ M\^80 + 234\ M\^82 + 
                    14\ M\^84 - 86\ M\^86 + 47\ M\^88 - 11\ M\^90 + 
                    M\^92)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[8, 16]\), \(M\^14 - 2\ M\^18 + 
              M\^22 + L\ \((2\ M\^8 - 8\ M\^10 + 11\ M\^12 + 15\ M\^14 - 
                    65\ M\^16 + 39\ M\^18 + 91\ M\^20 - 103\ M\^22 - 
                    43\ M\^24 + 70\ M\^26 + 7\ M\^28 - 15\ M\^30 - 4\ M\^32 + 
                    3\ M\^34 + 2\ M\^36 - M\^38)\) + 
              L\^2\ \((M\^2 - 8\ M\^4 + 30\ M\^6 - 52\ M\^8 + 4\ M\^10 + 
                    144\ M\^12 - 215\ M\^14 - 30\ M\^16 + 405\ M\^18 - 
                    260\ M\^20 - 488\ M\^22 + 698\ M\^24 + 357\ M\^26 - 
                    1067\ M\^28 + 30\ M\^30 + 901\ M\^32 - 266\ M\^34 - 
                    448\ M\^36 + 247\ M\^38 + 133\ M\^40 - 115\ M\^42 - 
                    7\ M\^44 + 21\ M\^46 - 4\ M\^48)\) + 
              L\^3\ \((1 - 11\ M\^2 + 50\ M\^4 - 103\ M\^6 + 21\ M\^8 + 
                    276\ M\^10 - 240\ M\^12 - 640\ M\^14 + 810\ M\^16 + 
                    1410\ M\^18 - 2250\ M\^20 - 2070\ M\^22 + 3801\ M\^24 + 
                    2691\ M\^26 - 4354\ M\^28 - 3289\ M\^30 + 3311\ M\^32 + 
                    3513\ M\^34 - 1210\ M\^36 - 3502\ M\^38 - 124\ M\^40 + 
                    2921\ M\^42 + 320\ M\^44 - 1740\ M\^46 - 3\ M\^48 + 
                    664\ M\^50 - 130\ M\^52 - 114\ M\^54 + 52\ M\^56 - 
                    6\ M\^58)\) + 
              L\^4\ \((M\^2 - 13\ M\^4 + 77\ M\^6 - 240\ M\^8 + 297\ M\^10 + 
                    356\ M\^12 - 1315\ M\^14 - 79\ M\^16 + 3365\ M\^18 - 
                    784\ M\^20 - 6791\ M\^22 + 2314\ M\^24 + 10377\ M\^26 - 
                    2683\ M\^28 - 12964\ M\^30 + 1113\ M\^32 + 12724\ M\^34 + 
                    1337\ M\^36 - 8861\ M\^38 - 2947\ M\^40 + 3409\ M\^42 + 
                    2401\ M\^44 + 1388\ M\^46 - 836\ M\^48 - 3325\ M\^50 + 
                    238\ M\^52 + 2812\ M\^54 - 491\ M\^56 - 1323\ M\^58 + 
                    528\ M\^60 + 244\ M\^62 - 210\ M\^64 + 50\ M\^66 - 
                    4\ M\^68)\) + 
              L\^5\ \((\(-2\)\ M\^6 + 27\ M\^8 - 151\ M\^10 + 407\ M\^12 - 
                    381\ M\^14 - 577\ M\^16 + 1497\ M\^18 + 32\ M\^20 - 
                    2612\ M\^22 + 1375\ M\^24 + 2333\ M\^26 - 3453\ M\^28 + 
                    1732\ M\^30 + 4984\ M\^32 - 9174\ M\^34 - 7199\ M\^36 + 
                    17268\ M\^38 + 11348\ M\^40 - 21699\ M\^42 - 
                    15395\ M\^44 + 19170\ M\^46 + 15968\ M\^48 - 
                    11228\ M\^50 - 11165\ M\^52 + 3438\ M\^54 + 5220\ M\^56 + 
                    990\ M\^58 - 1749\ M\^60 - 2108\ M\^62 + 880\ M\^64 + 
                    1337\ M\^66 - 748\ M\^68 - 288\ M\^70 + 357\ M\^72 - 
                    121\ M\^74 + 18\ M\^76 - M\^78)\) + 
              L\^6\ \((M\^10 - 15\ M\^12 + 94\ M\^14 - 312\ M\^16 + 
                    542\ M\^18 - 166\ M\^20 - 1361\ M\^22 + 2200\ M\^24 + 
                    1791\ M\^26 - 6862\ M\^28 - 1528\ M\^30 + 15381\ M\^32 + 
                    1276\ M\^34 - 27149\ M\^36 - 4559\ M\^38 + 39322\ M\^40 + 
                    13451\ M\^42 - 45491\ M\^44 - 24482\ M\^46 + 
                    40138\ M\^48 + 30245\ M\^50 - 25687\ M\^52 - 
                    25627\ M\^54 + 12155\ M\^56 + 15682\ M\^58 - 
                    5000\ M\^60 - 6708\ M\^62 + 2726\ M\^64 + 1415\ M\^66 - 
                    1815\ M\^68 + 696\ M\^70 + 821\ M\^72 - 989\ M\^74 + 
                    68\ M\^76 + 428\ M\^78 - 299\ M\^80 + 94\ M\^82 - 
                    15\ M\^84 + M\^86)\) + 
              L\^13\ \((M\^80 - 2\ M\^84 + M\^88)\) + 
              L\^7\ \((M\^16 - 15\ M\^18 + 94\ M\^20 - 299\ M\^22 + 
                    428\ M\^24 + 68\ M\^26 - 989\ M\^28 + 821\ M\^30 + 
                    696\ M\^32 - 1815\ M\^34 + 1415\ M\^36 + 2726\ M\^38 - 
                    6708\ M\^40 - 5000\ M\^42 + 15682\ M\^44 + 12155\ M\^46 - 
                    25627\ M\^48 - 25687\ M\^50 + 30245\ M\^52 + 
                    40138\ M\^54 - 24482\ M\^56 - 45491\ M\^58 + 
                    13451\ M\^60 + 39322\ M\^62 - 4559\ M\^64 - 
                    27149\ M\^66 + 1276\ M\^68 + 15381\ M\^70 - 1528\ M\^72 - 
                    6862\ M\^74 + 1791\ M\^76 + 2200\ M\^78 - 1361\ M\^80 - 
                    166\ M\^82 + 542\ M\^84 - 312\ M\^86 + 94\ M\^88 - 
                    15\ M\^90 + M\^92)\) + 
              L\^12\ \((\(-M\^64\) + 2\ M\^66 + 3\ M\^68 - 4\ M\^70 - 
                    15\ M\^72 + 7\ M\^74 + 70\ M\^76 - 43\ M\^78 - 
                    103\ M\^80 + 91\ M\^82 + 39\ M\^84 - 65\ M\^86 + 
                    15\ M\^88 + 11\ M\^90 - 8\ M\^92 + 2\ M\^94)\) + 
              L\^8\ \((\(-M\^24\) + 18\ M\^26 - 121\ M\^28 + 357\ M\^30 - 
                    288\ M\^32 - 748\ M\^34 + 1337\ M\^36 + 880\ M\^38 - 
                    2108\ M\^40 - 1749\ M\^42 + 990\ M\^44 + 5220\ M\^46 + 
                    3438\ M\^48 - 11165\ M\^50 - 11228\ M\^52 + 
                    15968\ M\^54 + 19170\ M\^56 - 15395\ M\^58 - 
                    21699\ M\^60 + 11348\ M\^62 + 17268\ M\^64 - 
                    7199\ M\^66 - 9174\ M\^68 + 4984\ M\^70 + 1732\ M\^72 - 
                    3453\ M\^74 + 2333\ M\^76 + 1375\ M\^78 - 2612\ M\^80 + 
                    32\ M\^82 + 1497\ M\^84 - 577\ M\^86 - 381\ M\^88 + 
                    407\ M\^90 - 151\ M\^92 + 27\ M\^94 - 2\ M\^96)\) + 
              L\^9\ \((\(-4\)\ M\^34 + 50\ M\^36 - 210\ M\^38 + 244\ M\^40 + 
                    528\ M\^42 - 1323\ M\^44 - 491\ M\^46 + 2812\ M\^48 + 
                    238\ M\^50 - 3325\ M\^52 - 836\ M\^54 + 1388\ M\^56 + 
                    2401\ M\^58 + 3409\ M\^60 - 2947\ M\^62 - 8861\ M\^64 + 
                    1337\ M\^66 + 12724\ M\^68 + 1113\ M\^70 - 12964\ M\^72 - 
                    2683\ M\^74 + 10377\ M\^76 + 2314\ M\^78 - 6791\ M\^80 - 
                    784\ M\^82 + 3365\ M\^84 - 79\ M\^86 - 1315\ M\^88 + 
                    356\ M\^90 + 297\ M\^92 - 240\ M\^94 + 77\ M\^96 - 
                    13\ M\^98 + M\^100)\) + 
              L\^11\ \((\(-4\)\ M\^54 + 21\ M\^56 - 7\ M\^58 - 115\ M\^60 + 
                    133\ M\^62 + 247\ M\^64 - 448\ M\^66 - 266\ M\^68 + 
                    901\ M\^70 + 30\ M\^72 - 1067\ M\^74 + 357\ M\^76 + 
                    698\ M\^78 - 488\ M\^80 - 260\ M\^82 + 405\ M\^84 - 
                    30\ M\^86 - 215\ M\^88 + 144\ M\^90 + 4\ M\^92 - 
                    52\ M\^94 + 30\ M\^96 - 8\ M\^98 + M\^100)\) + 
              L\^10\ \((\(-6\)\ M\^44 + 52\ M\^46 - 114\ M\^48 - 130\ M\^50 + 
                    664\ M\^52 - 3\ M\^54 - 1740\ M\^56 + 320\ M\^58 + 
                    2921\ M\^60 - 124\ M\^62 - 3502\ M\^64 - 1210\ M\^66 + 
                    3513\ M\^68 + 3311\ M\^70 - 3289\ M\^72 - 4354\ M\^74 + 
                    2691\ M\^76 + 3801\ M\^78 - 2070\ M\^80 - 2250\ M\^82 + 
                    1410\ M\^84 + 810\ M\^86 - 640\ M\^88 - 240\ M\^90 + 
                    276\ M\^92 + 21\ M\^94 - 103\ M\^96 + 50\ M\^98 - 
                    11\ M\^100 + M\^102)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[8, 17]\), \(M\^58 - 2\ M\^62 + 
              M\^66 + L\^20\ \((M\^58 - 2\ M\^62 + M\^66)\) + 
              L\ \((M\^44 - 2\ M\^46 - 3\ M\^48 + 5\ M\^50 + 14\ M\^52 - 
                    25\ M\^54 - 40\ M\^56 + 111\ M\^58 + 28\ M\^60 - 
                    179\ M\^62 + 28\ M\^64 + 111\ M\^66 - 40\ M\^68 - 
                    25\ M\^70 + 14\ M\^72 + 5\ M\^74 - 3\ M\^76 - 2\ M\^78 + 
                    M\^80)\) + 
              L\^19\ \((M\^44 - 2\ M\^46 - 3\ M\^48 + 5\ M\^50 + 14\ M\^52 - 
                    25\ M\^54 - 40\ M\^56 + 111\ M\^58 + 28\ M\^60 - 
                    179\ M\^62 + 28\ M\^64 + 111\ M\^66 - 40\ M\^68 - 
                    25\ M\^70 + 14\ M\^72 + 5\ M\^74 - 3\ M\^76 - 2\ M\^78 + 
                    M\^80)\) + 
              L\^2\ \((\(-3\)\ M\^36 + 18\ M\^38 - 29\ M\^40 - 51\ M\^42 + 
                    202\ M\^44 - 24\ M\^46 - 588\ M\^48 + 522\ M\^50 + 
                    963\ M\^52 - 1770\ M\^54 - 756\ M\^56 + 3206\ M\^58 + 
                    202\ M\^60 - 3802\ M\^62 + 202\ M\^64 + 3206\ M\^66 - 
                    756\ M\^68 - 1770\ M\^70 + 963\ M\^72 + 522\ M\^74 - 
                    588\ M\^76 - 24\ M\^78 + 202\ M\^80 - 51\ M\^82 - 
                    29\ M\^84 + 18\ M\^86 - 3\ M\^88)\) + 
              L\^18\ \((\(-3\)\ M\^36 + 18\ M\^38 - 29\ M\^40 - 51\ M\^42 + 
                    202\ M\^44 - 24\ M\^46 - 588\ M\^48 + 522\ M\^50 + 
                    963\ M\^52 - 1770\ M\^54 - 756\ M\^56 + 3206\ M\^58 + 
                    202\ M\^60 - 3802\ M\^62 + 202\ M\^64 + 3206\ M\^66 - 
                    756\ M\^68 - 1770\ M\^70 + 963\ M\^72 + 522\ M\^74 - 
                    588\ M\^76 - 24\ M\^78 + 202\ M\^80 - 51\ M\^82 - 
                    29\ M\^84 + 18\ M\^86 - 3\ M\^88)\) + 
              L\^3\ \((3\ M\^28 - 31\ M\^30 + 123\ M\^32 - 170\ M\^34 - 
                    250\ M\^36 + 1073\ M\^38 - 536\ M\^40 - 2396\ M\^42 + 
                    3463\ M\^44 + 1789\ M\^46 - 7683\ M\^48 + 4319\ M\^50 + 
                    9243\ M\^52 - 16447\ M\^54 - 6809\ M\^56 + 30667\ M\^58 + 
                    2446\ M\^60 - 37761\ M\^62 + 2446\ M\^64 + 30667\ M\^66 - 
                    6809\ M\^68 - 16447\ M\^70 + 9243\ M\^72 + 4319\ M\^74 - 
                    7683\ M\^76 + 1789\ M\^78 + 3463\ M\^80 - 2396\ M\^82 - 
                    536\ M\^84 + 1073\ M\^86 - 250\ M\^88 - 170\ M\^90 + 
                    123\ M\^92 - 31\ M\^94 + 3\ M\^96)\) + 
              L\^17\ \((3\ M\^28 - 31\ M\^30 + 123\ M\^32 - 170\ M\^34 - 
                    250\ M\^36 + 1073\ M\^38 - 536\ M\^40 - 2396\ M\^42 + 
                    3463\ M\^44 + 1789\ M\^46 - 7683\ M\^48 + 4319\ M\^50 + 
                    9243\ M\^52 - 16447\ M\^54 - 6809\ M\^56 + 30667\ M\^58 + 
                    2446\ M\^60 - 37761\ M\^62 + 2446\ M\^64 + 30667\ M\^66 - 
                    6809\ M\^68 - 16447\ M\^70 + 9243\ M\^72 + 4319\ M\^74 - 
                    7683\ M\^76 + 1789\ M\^78 + 3463\ M\^80 - 2396\ M\^82 - 
                    536\ M\^84 + 1073\ M\^86 - 250\ M\^88 - 170\ M\^90 + 
                    123\ M\^92 - 31\ M\^94 + 3\ M\^96)\) + 
              L\^4\ \((\(-M\^20\) + 16\ M\^22 - 108\ M\^24 + 383\ M\^26 - 
                    630\ M\^28 - 307\ M\^30 + 3195\ M\^32 - 3456\ M\^34 - 
                    6602\ M\^36 + 15714\ M\^38 + 6124\ M\^40 - 36835\ M\^42 + 
                    974\ M\^44 + 57615\ M\^46 - 8306\ M\^48 - 60622\ M\^50 + 
                    8010\ M\^52 + 28048\ M\^54 + 11\ M\^56 + 25719\ M\^58 - 
                    3075\ M\^60 - 52550\ M\^62 - 3075\ M\^64 + 25719\ M\^66 + 
                    11\ M\^68 + 28048\ M\^70 + 8010\ M\^72 - 60622\ M\^74 - 
                    8306\ M\^76 + 57615\ M\^78 + 974\ M\^80 - 36835\ M\^82 + 
                    6124\ M\^84 + 15714\ M\^86 - 6602\ M\^88 - 3456\ M\^90 + 
                    3195\ M\^92 - 307\ M\^94 - 630\ M\^96 + 383\ M\^98 - 
                    108\ M\^100 + 16\ M\^102 - M\^104)\) + 
              L\^16\ \((\(-M\^20\) + 16\ M\^22 - 108\ M\^24 + 383\ M\^26 - 
                    630\ M\^28 - 307\ M\^30 + 3195\ M\^32 - 3456\ M\^34 - 
                    6602\ M\^36 + 15714\ M\^38 + 6124\ M\^40 - 36835\ M\^42 + 
                    974\ M\^44 + 57615\ M\^46 - 8306\ M\^48 - 60622\ M\^50 + 
                    8010\ M\^52 + 28048\ M\^54 + 11\ M\^56 + 25719\ M\^58 - 
                    3075\ M\^60 - 52550\ M\^62 - 3075\ M\^64 + 25719\ M\^66 + 
                    11\ M\^68 + 28048\ M\^70 + 8010\ M\^72 - 60622\ M\^74 - 
                    8306\ M\^76 + 57615\ M\^78 + 974\ M\^80 - 36835\ M\^82 + 
                    6124\ M\^84 + 15714\ M\^86 - 6602\ M\^88 - 3456\ M\^90 + 
                    3195\ M\^92 - 307\ M\^94 - 630\ M\^96 + 383\ M\^98 - 
                    108\ M\^100 + 16\ M\^102 - M\^104)\) + 
              L\^5\ \((\(-M\^14\) + 17\ M\^16 - 134\ M\^18 + 603\ M\^20 - 
                    1490\ M\^22 + 1025\ M\^24 + 4598\ M\^26 - 10942\ M\^28 - 
                    4735\ M\^30 + 40539\ M\^32 - 12290\ M\^34 - 
                    100363\ M\^36 + 58397\ M\^38 + 201005\ M\^40 - 
                    120180\ M\^42 - 347559\ M\^44 + 130537\ M\^46 + 
                    517005\ M\^48 + 23815\ M\^50 - 634820\ M\^52 - 
                    382738\ M\^54 + 563488\ M\^56 + 800903\ M\^58 - 
                    229998\ M\^60 - 996424\ M\^62 - 229998\ M\^64 + 
                    800903\ M\^66 + 563488\ M\^68 - 382738\ M\^70 - 
                    634820\ M\^72 + 23815\ M\^74 + 517005\ M\^76 + 
                    130537\ M\^78 - 347559\ M\^80 - 120180\ M\^82 + 
                    201005\ M\^84 + 58397\ M\^86 - 100363\ M\^88 - 
                    12290\ M\^90 + 40539\ M\^92 - 4735\ M\^94 - 
                    10942\ M\^96 + 4598\ M\^98 + 1025\ M\^100 - 
                    1490\ M\^102 + 603\ M\^104 - 134\ M\^106 + 17\ M\^108 - 
                    M\^110)\) + 
              L\^15\ \((\(-M\^14\) + 17\ M\^16 - 134\ M\^18 + 603\ M\^20 - 
                    1490\ M\^22 + 1025\ M\^24 + 4598\ M\^26 - 10942\ M\^28 - 
                    4735\ M\^30 + 40539\ M\^32 - 12290\ M\^34 - 
                    100363\ M\^36 + 58397\ M\^38 + 201005\ M\^40 - 
                    120180\ M\^42 - 347559\ M\^44 + 130537\ M\^46 + 
                    517005\ M\^48 + 23815\ M\^50 - 634820\ M\^52 - 
                    382738\ M\^54 + 563488\ M\^56 + 800903\ M\^58 - 
                    229998\ M\^60 - 996424\ M\^62 - 229998\ M\^64 + 
                    800903\ M\^66 + 563488\ M\^68 - 382738\ M\^70 - 
                    634820\ M\^72 + 23815\ M\^74 + 517005\ M\^76 + 
                    130537\ M\^78 - 347559\ M\^80 - 120180\ M\^82 + 
                    201005\ M\^84 + 58397\ M\^86 - 100363\ M\^88 - 
                    12290\ M\^90 + 40539\ M\^92 - 4735\ M\^94 - 
                    10942\ M\^96 + 4598\ M\^98 + 1025\ M\^100 - 
                    1490\ M\^102 + 603\ M\^104 - 134\ M\^106 + 17\ M\^108 - 
                    M\^110)\) + 
              L\^6\ \((3\ M\^10 - 52\ M\^12 + 385\ M\^14 - 1494\ M\^16 + 
                    2751\ M\^18 + 316\ M\^20 - 10616\ M\^22 + 10575\ M\^24 + 
                    25053\ M\^26 - 50199\ M\^28 - 31976\ M\^30 + 
                    130086\ M\^32 + 8792\ M\^34 - 244141\ M\^36 + 
                    75827\ M\^38 + 347946\ M\^40 - 233410\ M\^42 - 
                    385836\ M\^44 + 424868\ M\^46 + 356823\ M\^48 - 
                    579188\ M\^50 - 287533\ M\^52 + 609105\ M\^54 + 
                    201277\ M\^56 - 530229\ M\^58 - 82052\ M\^60 + 
                    477270\ M\^62 - 82052\ M\^64 - 530229\ M\^66 + 
                    201277\ M\^68 + 609105\ M\^70 - 287533\ M\^72 - 
                    579188\ M\^74 + 356823\ M\^76 + 424868\ M\^78 - 
                    385836\ M\^80 - 233410\ M\^82 + 347946\ M\^84 + 
                    75827\ M\^86 - 244141\ M\^88 + 8792\ M\^90 + 
                    130086\ M\^92 - 31976\ M\^94 - 50199\ M\^96 + 
                    25053\ M\^98 + 10575\ M\^100 - 10616\ M\^102 + 
                    316\ M\^104 + 2751\ M\^106 - 1494\ M\^108 + 385\ M\^110 - 
                    52\ M\^112 + 3\ M\^114)\) + 
              L\^14\ \((3\ M\^10 - 52\ M\^12 + 385\ M\^14 - 1494\ M\^16 + 
                    2751\ M\^18 + 316\ M\^20 - 10616\ M\^22 + 10575\ M\^24 + 
                    25053\ M\^26 - 50199\ M\^28 - 31976\ M\^30 + 
                    130086\ M\^32 + 8792\ M\^34 - 244141\ M\^36 + 
                    75827\ M\^38 + 347946\ M\^40 - 233410\ M\^42 - 
                    385836\ M\^44 + 424868\ M\^46 + 356823\ M\^48 - 
                    579188\ M\^50 - 287533\ M\^52 + 609105\ M\^54 + 
                    201277\ M\^56 - 530229\ M\^58 - 82052\ M\^60 + 
                    477270\ M\^62 - 82052\ M\^64 - 530229\ M\^66 + 
                    201277\ M\^68 + 609105\ M\^70 - 287533\ M\^72 - 
                    579188\ M\^74 + 356823\ M\^76 + 424868\ M\^78 - 
                    385836\ M\^80 - 233410\ M\^82 + 347946\ M\^84 + 
                    75827\ M\^86 - 244141\ M\^88 + 8792\ M\^90 + 
                    130086\ M\^92 - 31976\ M\^94 - 50199\ M\^96 + 
                    25053\ M\^98 + 10575\ M\^100 - 10616\ M\^102 + 
                    316\ M\^104 + 2751\ M\^106 - 1494\ M\^108 + 385\ M\^110 - 
                    52\ M\^112 + 3\ M\^114)\) + 
              L\^7\ \((\(-3\)\ M\^6 + 56\ M\^8 - 434\ M\^10 + 1741\ M\^12 - 
                    3392\ M\^14 + 567\ M\^16 + 10010\ M\^18 - 10719\ M\^20 - 
                    22944\ M\^22 + 47900\ M\^24 + 20915\ M\^26 - 
                    109816\ M\^28 + 28577\ M\^30 + 161436\ M\^32 - 
                    196500\ M\^34 - 78535\ M\^36 + 565804\ M\^38 - 
                    301270\ M\^40 - 1228349\ M\^42 + 1026959\ M\^44 + 
                    2267977\ M\^46 - 1892430\ M\^48 - 3626086\ M\^50 + 
                    2382035\ M\^52 + 5075006\ M\^54 - 2027581\ M\^56 - 
                    6203868\ M\^58 + 799657\ M\^60 + 6608010\ M\^62 + 
                    799657\ M\^64 - 6203868\ M\^66 - 2027581\ M\^68 + 
                    5075006\ M\^70 + 2382035\ M\^72 - 3626086\ M\^74 - 
                    1892430\ M\^76 + 2267977\ M\^78 + 1026959\ M\^80 - 
                    1228349\ M\^82 - 301270\ M\^84 + 565804\ M\^86 - 
                    78535\ M\^88 - 196500\ M\^90 + 161436\ M\^92 + 
                    28577\ M\^94 - 109816\ M\^96 + 20915\ M\^98 + 
                    47900\ M\^100 - 22944\ M\^102 - 10719\ M\^104 + 
                    10010\ M\^106 + 567\ M\^108 - 3392\ M\^110 + 
                    1741\ M\^112 - 434\ M\^114 + 56\ M\^116 - 3\ M\^118)\) + 
              L\^13\ \((\(-3\)\ M\^6 + 56\ M\^8 - 434\ M\^10 + 1741\ M\^12 - 
                    3392\ M\^14 + 567\ M\^16 + 10010\ M\^18 - 10719\ M\^20 - 
                    22944\ M\^22 + 47900\ M\^24 + 20915\ M\^26 - 
                    109816\ M\^28 + 28577\ M\^30 + 161436\ M\^32 - 
                    196500\ M\^34 - 78535\ M\^36 + 565804\ M\^38 - 
                    301270\ M\^40 - 1228349\ M\^42 + 1026959\ M\^44 + 
                    2267977\ M\^46 - 1892430\ M\^48 - 3626086\ M\^50 + 
                    2382035\ M\^52 + 5075006\ M\^54 - 2027581\ M\^56 - 
                    6203868\ M\^58 + 799657\ M\^60 + 6608010\ M\^62 + 
                    799657\ M\^64 - 6203868\ M\^66 - 2027581\ M\^68 + 
                    5075006\ M\^70 + 2382035\ M\^72 - 3626086\ M\^74 - 
                    1892430\ M\^76 + 2267977\ M\^78 + 1026959\ M\^80 - 
                    1228349\ M\^82 - 301270\ M\^84 + 565804\ M\^86 - 
                    78535\ M\^88 - 196500\ M\^90 + 161436\ M\^92 + 
                    28577\ M\^94 - 109816\ M\^96 + 20915\ M\^98 + 
                    47900\ M\^100 - 22944\ M\^102 - 10719\ M\^104 + 
                    10010\ M\^106 + 567\ M\^108 - 3392\ M\^110 + 
                    1741\ M\^112 - 434\ M\^114 + 56\ M\^116 - 3\ M\^118)\) + 
              L\^8\ \((M\^2 - 21\ M\^4 + 188\ M\^6 - 924\ M\^8 + 
                    2586\ M\^10 - 3371\ M\^12 - 1717\ M\^14 + 11234\ M\^16 - 
                    3258\ M\^18 - 37243\ M\^20 + 47397\ M\^22 + 
                    67532\ M\^24 - 175294\ M\^26 - 82388\ M\^28 + 
                    486442\ M\^30 + 10629\ M\^32 - 1047727\ M\^34 + 
                    163026\ M\^36 + 1799870\ M\^38 - 305426\ M\^40 - 
                    2366813\ M\^42 + 154875\ M\^44 + 1989076\ M\^46 + 
                    457144\ M\^48 - 28066\ M\^50 - 1179848\ M\^52 - 
                    3341745\ M\^54 + 1353494\ M\^56 + 6703365\ M\^58 - 
                    624625\ M\^60 - 8128610\ M\^62 - 624625\ M\^64 + 
                    6703365\ M\^66 + 1353494\ M\^68 - 3341745\ M\^70 - 
                    1179848\ M\^72 - 28066\ M\^74 + 457144\ M\^76 + 
                    1989076\ M\^78 + 154875\ M\^80 - 2366813\ M\^82 - 
                    305426\ M\^84 + 1799870\ M\^86 + 163026\ M\^88 - 
                    1047727\ M\^90 + 10629\ M\^92 + 486442\ M\^94 - 
                    82388\ M\^96 - 175294\ M\^98 + 67532\ M\^100 + 
                    47397\ M\^102 - 37243\ M\^104 - 3258\ M\^106 + 
                    11234\ M\^108 - 1717\ M\^110 - 3371\ M\^112 + 
                    2586\ M\^114 - 924\ M\^116 + 188\ M\^118 - 21\ M\^120 + 
                    M\^122)\) + 
              L\^12\ \((M\^2 - 21\ M\^4 + 188\ M\^6 - 924\ M\^8 + 
                    2586\ M\^10 - 3371\ M\^12 - 1717\ M\^14 + 11234\ M\^16 - 
                    3258\ M\^18 - 37243\ M\^20 + 47397\ M\^22 + 
                    67532\ M\^24 - 175294\ M\^26 - 82388\ M\^28 + 
                    486442\ M\^30 + 10629\ M\^32 - 1047727\ M\^34 + 
                    163026\ M\^36 + 1799870\ M\^38 - 305426\ M\^40 - 
                    2366813\ M\^42 + 154875\ M\^44 + 1989076\ M\^46 + 
                    457144\ M\^48 - 28066\ M\^50 - 1179848\ M\^52 - 
                    3341745\ M\^54 + 1353494\ M\^56 + 6703365\ M\^58 - 
                    624625\ M\^60 - 8128610\ M\^62 - 624625\ M\^64 + 
                    6703365\ M\^66 + 1353494\ M\^68 - 3341745\ M\^70 - 
                    1179848\ M\^72 - 28066\ M\^74 + 457144\ M\^76 + 
                    1989076\ M\^78 + 154875\ M\^80 - 2366813\ M\^82 - 
                    305426\ M\^84 + 1799870\ M\^86 + 163026\ M\^88 - 
                    1047727\ M\^90 + 10629\ M\^92 + 486442\ M\^94 - 
                    82388\ M\^96 - 175294\ M\^98 + 67532\ M\^100 + 
                    47397\ M\^102 - 37243\ M\^104 - 3258\ M\^106 + 
                    11234\ M\^108 - 1717\ M\^110 - 3371\ M\^112 + 
                    2586\ M\^114 - 924\ M\^116 + 188\ M\^118 - 21\ M\^120 + 
                    M\^122)\) + 
              L\^10\ \((\(-2\) + 39\ M\^2 - 318\ M\^4 + 1363\ M\^6 - 
                    2959\ M\^8 + 1440\ M\^10 + 5839\ M\^12 - 1619\ M\^14 - 
                    33450\ M\^16 + 23992\ M\^18 + 119966\ M\^20 - 
                    97513\ M\^22 - 407029\ M\^24 + 345289\ M\^26 + 
                    1113641\ M\^28 - 906347\ M\^30 - 2538819\ M\^32 + 
                    1957890\ M\^34 + 4659712\ M\^36 - 3741520\ M\^38 - 
                    6634081\ M\^40 + 6979626\ M\^42 + 6841859\ M\^44 - 
                    13338023\ M\^46 - 4134900\ M\^48 + 24483742\ M\^50 - 
                    206284\ M\^52 - 39672954\ M\^54 + 2862683\ M\^56 + 
                    53803932\ M\^58 - 1670168\ M\^60 - 59678674\ M\^62 - 
                    1670168\ M\^64 + 53803932\ M\^66 + 2862683\ M\^68 - 
                    39672954\ M\^70 - 206284\ M\^72 + 24483742\ M\^74 - 
                    4134900\ M\^76 - 13338023\ M\^78 + 6841859\ M\^80 + 
                    6979626\ M\^82 - 6634081\ M\^84 - 3741520\ M\^86 + 
                    4659712\ M\^88 + 1957890\ M\^90 - 2538819\ M\^92 - 
                    906347\ M\^94 + 1113641\ M\^96 + 345289\ M\^98 - 
                    407029\ M\^100 - 97513\ M\^102 + 119966\ M\^104 + 
                    23992\ M\^106 - 33450\ M\^108 - 1619\ M\^110 + 
                    5839\ M\^112 + 1440\ M\^114 - 2959\ M\^116 + 
                    1363\ M\^118 - 318\ M\^120 + 39\ M\^122 - 2\ M\^124)\) + 
              L\^9\ \((1 - 20\ M\^2 + 175\ M\^4 - 856\ M\^6 + 2404\ M\^8 - 
                    3139\ M\^10 - 1361\ M\^12 + 7915\ M\^14 + 4526\ M\^16 - 
                    33845\ M\^18 + 7476\ M\^20 + 78684\ M\^22 - 
                    36013\ M\^24 - 135728\ M\^26 + 126434\ M\^28 + 
                    16044\ M\^30 - 320363\ M\^32 + 756600\ M\^34 + 
                    797993\ M\^36 - 2988606\ M\^38 - 1970513\ M\^40 + 
                    7319228\ M\^42 + 4279307\ M\^44 - 13302878\ M\^46 - 
                    7500156\ M\^48 + 19237876\ M\^50 + 9859600\ M\^52 - 
                    23202163\ M\^54 - 8850832\ M\^56 + 24794668\ M\^58 + 
                    3601322\ M\^60 - 25131318\ M\^62 + 3601322\ M\^64 + 
                    24794668\ M\^66 - 8850832\ M\^68 - 23202163\ M\^70 + 
                    9859600\ M\^72 + 19237876\ M\^74 - 7500156\ M\^76 - 
                    13302878\ M\^78 + 4279307\ M\^80 + 7319228\ M\^82 - 
                    1970513\ M\^84 - 2988606\ M\^86 + 797993\ M\^88 + 
                    756600\ M\^90 - 320363\ M\^92 + 16044\ M\^94 + 
                    126434\ M\^96 - 135728\ M\^98 - 36013\ M\^100 + 
                    78684\ M\^102 + 7476\ M\^104 - 33845\ M\^106 + 
                    4526\ M\^108 + 7915\ M\^110 - 1361\ M\^112 - 
                    3139\ M\^114 + 2404\ M\^116 - 856\ M\^118 + 175\ M\^120 - 
                    20\ M\^122 + M\^124)\) + 
              L\^11\ \((1 - 20\ M\^2 + 175\ M\^4 - 856\ M\^6 + 2404\ M\^8 - 
                    3139\ M\^10 - 1361\ M\^12 + 7915\ M\^14 + 4526\ M\^16 - 
                    33845\ M\^18 + 7476\ M\^20 + 78684\ M\^22 - 
                    36013\ M\^24 - 135728\ M\^26 + 126434\ M\^28 + 
                    16044\ M\^30 - 320363\ M\^32 + 756600\ M\^34 + 
                    797993\ M\^36 - 2988606\ M\^38 - 1970513\ M\^40 + 
                    7319228\ M\^42 + 4279307\ M\^44 - 13302878\ M\^46 - 
                    7500156\ M\^48 + 19237876\ M\^50 + 9859600\ M\^52 - 
                    23202163\ M\^54 - 8850832\ M\^56 + 24794668\ M\^58 + 
                    3601322\ M\^60 - 25131318\ M\^62 + 3601322\ M\^64 + 
                    24794668\ M\^66 - 8850832\ M\^68 - 23202163\ M\^70 + 
                    9859600\ M\^72 + 19237876\ M\^74 - 7500156\ M\^76 - 
                    13302878\ M\^78 + 4279307\ M\^80 + 7319228\ M\^82 - 
                    1970513\ M\^84 - 2988606\ M\^86 + 797993\ M\^88 + 
                    756600\ M\^90 - 320363\ M\^92 + 16044\ M\^94 + 
                    126434\ M\^96 - 135728\ M\^98 - 36013\ M\^100 + 
                    78684\ M\^102 + 7476\ M\^104 - 33845\ M\^106 + 
                    4526\ M\^108 + 7915\ M\^110 - 1361\ M\^112 - 
                    3139\ M\^114 + 2404\ M\^116 - 856\ M\^118 + 175\ M\^120 - 
                    20\ M\^122 + M\^124)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[8, 18]\), \(M\^32 + L\^8\ M\^32 + 
              L\ \((M\^18 - 2\ M\^20 - 3\ M\^22 + 18\ M\^26 - 12\ M\^28 - 
                    30\ M\^30 + 64\ M\^32 - 30\ M\^34 - 12\ M\^36 + 
                    18\ M\^38 - 3\ M\^42 - 2\ M\^44 + M\^46)\) + 
              L\^7\ \((M\^18 - 2\ M\^20 - 3\ M\^22 + 18\ M\^26 - 12\ M\^28 - 
                    30\ M\^30 + 64\ M\^32 - 30\ M\^34 - 12\ M\^36 + 
                    18\ M\^38 - 3\ M\^42 - 2\ M\^44 + M\^46)\) + 
              L\^2\ \((\(-2\)\ M\^10 + 16\ M\^12 - 32\ M\^14 - 38\ M\^16 + 
                    162\ M\^18 + 13\ M\^20 - 370\ M\^22 + 170\ M\^24 + 
                    200\ M\^26 + 20\ M\^28 - 174\ M\^30 + 98\ M\^32 - 
                    174\ M\^34 + 20\ M\^36 + 200\ M\^38 + 170\ M\^40 - 
                    370\ M\^42 + 13\ M\^44 + 162\ M\^46 - 38\ M\^48 - 
                    32\ M\^50 + 16\ M\^52 - 2\ M\^54)\) + 
              L\^6\ \((\(-2\)\ M\^10 + 16\ M\^12 - 32\ M\^14 - 38\ M\^16 + 
                    162\ M\^18 + 13\ M\^20 - 370\ M\^22 + 170\ M\^24 + 
                    200\ M\^26 + 20\ M\^28 - 174\ M\^30 + 98\ M\^32 - 
                    174\ M\^34 + 20\ M\^36 + 200\ M\^38 + 170\ M\^40 - 
                    370\ M\^42 + 13\ M\^44 + 162\ M\^46 - 38\ M\^48 - 
                    32\ M\^50 + 16\ M\^52 - 2\ M\^54)\) + 
              L\^3\ \((M\^2 - 14\ M\^4 + 75\ M\^6 - 184\ M\^8 + 173\ M\^10 + 
                    32\ M\^12 - 67\ M\^14 - 22\ M\^16 + 49\ M\^18 - 
                    890\ M\^20 + 849\ M\^22 + 1838\ M\^24 - 704\ M\^26 - 
                    4126\ M\^28 + 678\ M\^30 + 4680\ M\^32 + 678\ M\^34 - 
                    4126\ M\^36 - 704\ M\^38 + 1838\ M\^40 + 849\ M\^42 - 
                    890\ M\^44 + 49\ M\^46 - 22\ M\^48 - 67\ M\^50 + 
                    32\ M\^52 + 173\ M\^54 - 184\ M\^56 + 75\ M\^58 - 
                    14\ M\^60 + M\^62)\) + 
              L\^5\ \((M\^2 - 14\ M\^4 + 75\ M\^6 - 184\ M\^8 + 173\ M\^10 + 
                    32\ M\^12 - 67\ M\^14 - 22\ M\^16 + 49\ M\^18 - 
                    890\ M\^20 + 849\ M\^22 + 1838\ M\^24 - 704\ M\^26 - 
                    4126\ M\^28 + 678\ M\^30 + 4680\ M\^32 + 678\ M\^34 - 
                    4126\ M\^36 - 704\ M\^38 + 1838\ M\^40 + 849\ M\^42 - 
                    890\ M\^44 + 49\ M\^46 - 22\ M\^48 - 67\ M\^50 + 
                    32\ M\^52 + 173\ M\^54 - 184\ M\^56 + 75\ M\^58 - 
                    14\ M\^60 + M\^62)\) + 
              L\^4\ \((1 - 16\ M\^2 + 101\ M\^4 - 300\ M\^6 + 324\ M\^8 + 
                    284\ M\^10 - 499\ M\^12 - 1264\ M\^14 + 1677\ M\^16 + 
                    2184\ M\^18 - 1954\ M\^20 - 4120\ M\^22 + 2202\ M\^24 + 
                    3536\ M\^26 - 234\ M\^28 - 1888\ M\^30 + 2\ M\^32 - 
                    1888\ M\^34 - 234\ M\^36 + 3536\ M\^38 + 2202\ M\^40 - 
                    4120\ M\^42 - 1954\ M\^44 + 2184\ M\^46 + 1677\ M\^48 - 
                    1264\ M\^50 - 499\ M\^52 + 284\ M\^54 + 324\ M\^56 - 
                    300\ M\^58 + 101\ M\^60 - 16\ M\^62 + M\^64)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[8, 19]\), \(1 + L\ M\^12\)},
          {"\<\"A-polynomial\"\>", \(Knot[8, 20]\), \(L\^5\ M\^8 + M\^10 + 
              L\ \((1 - M\^2 + 2\ M\^4 - 2\ M\^6 - M\^8 + 5\ M\^10 + 
                    M\^12)\) + 
              L\^2\ \((\(-1\) + 5\ M\^2 - 3\ M\^6 + 3\ M\^8 + 4\ M\^12 + 
                    2\ M\^14)\) + 
              L\^3\ \((2\ M\^4 + 4\ M\^6 + 3\ M\^10 - 3\ M\^12 + 5\ M\^16 - 
                    M\^18)\) + 
              L\^4\ \((M\^6 + 5\ M\^8 - M\^10 - 2\ M\^12 + 2\ M\^14 - M\^16 + 
                    M\^18)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[8, 21]\), \(L\^5\ M\^2 + M\^20 + 
              L\^4\ \((\(-2\)\ M\^2 - 2\ M\^4 + 13\ M\^6 + M\^8 - 8\ M\^10 + 
                    4\ M\^12 - M\^14)\) + 
              L\ \((\(-M\^8\) + 4\ M\^10 - 8\ M\^12 + M\^14 + 13\ M\^16 - 
                    2\ M\^18 - 2\ M\^20)\) + 
              L\^3\ \((\(-1\) + 7\ M\^2 - 11\ M\^4 - 13\ M\^6 + 25\ M\^8 + 
                    9\ M\^10 + 4\ M\^12 - 6\ M\^14 - 10\ M\^16 + 7\ M\^18 - 
                    M\^20)\) + 
              L\^2\ \((\(-M\^2\) + 7\ M\^4 - 10\ M\^6 - 6\ M\^8 + 4\ M\^10 + 
                    9\ M\^12 + 25\ M\^14 - 13\ M\^16 - 11\ M\^18 + 7\ M\^20 - 
                    M\^22)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 1]\), \(L + M\^18\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 2]\), \(L\^7 + M\^30 + 
              L\^6\ \((\(-3\) + 6\ M\^2 + 4\ M\^4 - M\^16 + M\^18)\) + 
              L\^5\ \((3 - 12\ M\^2 + 5\ M\^4 + 15\ M\^6 + 10\ M\^8 - 
                    5\ M\^16 + 8\ M\^20 - 3\ M\^22)\) + 
              L\^4\ \((\(-1\) + 6\ M\^2 - 12\ M\^4 + 2\ M\^6 + 5\ M\^8 + 
                    15\ M\^10 + 20\ M\^12 - 15\ M\^16 + 16\ M\^20 + 
                    9\ M\^22 - 13\ M\^24 + 3\ M\^26)\) + 
              L\ \((M\^12 - M\^14 + 4\ M\^26 + 6\ M\^28 - 3\ M\^30)\) + 
              L\^3\ \((3\ M\^4 - 13\ M\^6 + 9\ M\^8 + 16\ M\^10 - 15\ M\^14 + 
                    20\ M\^18 + 15\ M\^20 + 5\ M\^22 + 2\ M\^24 - 12\ M\^26 + 
                    6\ M\^28 - M\^30)\) + 
              L\^2\ \((\(-3\)\ M\^8 + 8\ M\^10 - 5\ M\^14 + 10\ M\^22 + 
                    15\ M\^24 + 5\ M\^26 - 12\ M\^28 + 3\ M\^30)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 3]\), \(1 + L\^9\ M\^114 + 
              L\ \((\(-1\) + 2\ M\^2 - M\^4 + 7\ M\^12 + 4\ M\^14 - 
                    5\ M\^16 + 3\ M\^18)\) + 
              L\^2\ \((5\ M\^12 - 17\ M\^14 + 10\ M\^16 + 15\ M\^18 - 
                    17\ M\^20 + 4\ M\^22 + 28\ M\^24 + 5\ M\^26 - 7\ M\^28 + 
                    3\ M\^30 + 14\ M\^32 - 10\ M\^34 + 3\ M\^36)\) + 
              L\^3\ \((\(-10\)\ M\^24 + 44\ M\^26 - 40\ M\^28 - 61\ M\^30 + 
                    77\ M\^32 + 74\ M\^34 - 55\ M\^36 - 15\ M\^38 + 
                    69\ M\^40 + 12\ M\^42 - 26\ M\^44 + 19\ M\^46 - 
                    10\ M\^48 + 10\ M\^50 - 5\ M\^52 + M\^54)\) + 
              L\^4\ \((10\ M\^36 - 50\ M\^38 + 61\ M\^40 + 60\ M\^42 - 
                    110\ M\^44 - 95\ M\^46 + 170\ M\^48 + 165\ M\^50 - 
                    90\ M\^52 - 85\ M\^54 + 65\ M\^56 + 50\ M\^58 - 
                    6\ M\^60 - 40\ M\^62 + 26\ M\^64 - 5\ M\^66)\) + 
              L\^5\ \((\(-5\)\ M\^48 + 26\ M\^50 - 40\ M\^52 - 6\ M\^54 + 
                    50\ M\^56 + 65\ M\^58 - 85\ M\^60 - 90\ M\^62 + 
                    165\ M\^64 + 170\ M\^66 - 95\ M\^68 - 110\ M\^70 + 
                    60\ M\^72 + 61\ M\^74 - 50\ M\^76 + 10\ M\^78)\) + 
              L\^6\ \((M\^60 - 5\ M\^62 + 10\ M\^64 - 10\ M\^66 + 19\ M\^68 - 
                    26\ M\^70 + 12\ M\^72 + 69\ M\^74 - 15\ M\^76 - 
                    55\ M\^78 + 74\ M\^80 + 77\ M\^82 - 61\ M\^84 - 
                    40\ M\^86 + 44\ M\^88 - 10\ M\^90)\) + 
              L\^7\ \((3\ M\^78 - 10\ M\^80 + 14\ M\^82 + 3\ M\^84 - 
                    7\ M\^86 + 5\ M\^88 + 28\ M\^90 + 4\ M\^92 - 17\ M\^94 + 
                    15\ M\^96 + 10\ M\^98 - 17\ M\^100 + 5\ M\^102)\) + 
              L\^8\ \((3\ M\^96 - 5\ M\^98 + 4\ M\^100 + 7\ M\^102 - M\^110 + 
                    2\ M\^112 - M\^114)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 4]\), \(1 + L\^10\ M\^84 + 
              L\ \((\(-2\) + 4\ M\^2 - 3\ M\^4 + 2\ M\^6 + 9\ M\^8 - M\^12 + 
                    2\ M\^14 - 3\ M\^16 + 2\ M\^18)\) + 
              L\^2\ \((1 - 4\ M\^2 + 6\ M\^4 - 4\ M\^6 - 3\ M\^8 - 5\ M\^10 + 
                    20\ M\^12 + M\^14 + 6\ M\^16 + 41\ M\^18 - 6\ M\^20 - 
                    28\ M\^22 + 17\ M\^24 + 3\ M\^26 - M\^30 + 3\ M\^32 - 
                    3\ M\^34 + M\^36)\) + 
              L\^3\ \((\(-6\)\ M\^8 + 29\ M\^10 - 43\ M\^12 - 8\ M\^14 + 
                    77\ M\^16 - 44\ M\^18 - 114\ M\^20 + 178\ M\^22 + 
                    130\ M\^24 - 158\ M\^26 + 5\ M\^28 + 124\ M\^30 - 
                    69\ M\^32 - 12\ M\^34 + 54\ M\^36 - 14\ M\^38 - 
                    28\ M\^40 + 25\ M\^42 - 6\ M\^44)\) + 
              L\^4\ \((15\ M\^16 - 79\ M\^18 + 116\ M\^20 + 68\ M\^22 - 
                    292\ M\^24 + 37\ M\^26 + 402\ M\^28 - 209\ M\^30 - 
                    146\ M\^32 + 383\ M\^34 - 73\ M\^36 - 224\ M\^38 + 
                    306\ M\^40 + 21\ M\^42 - 228\ M\^44 + 77\ M\^46 + 
                    95\ M\^48 - 74\ M\^50 + 15\ M\^52)\) + 
              L\^5\ \((\(-20\)\ M\^24 + 106\ M\^26 - 149\ M\^28 - 
                    118\ M\^30 + 395\ M\^32 + 4\ M\^34 - 511\ M\^36 + 
                    274\ M\^38 + 285\ M\^40 - 280\ M\^42 + 285\ M\^44 + 
                    274\ M\^46 - 511\ M\^48 + 4\ M\^50 + 395\ M\^52 - 
                    118\ M\^54 - 149\ M\^56 + 106\ M\^58 - 20\ M\^60)\) + 
              L\^6\ \((15\ M\^32 - 74\ M\^34 + 95\ M\^36 + 77\ M\^38 - 
                    228\ M\^40 + 21\ M\^42 + 306\ M\^44 - 224\ M\^46 - 
                    73\ M\^48 + 383\ M\^50 - 146\ M\^52 - 209\ M\^54 + 
                    402\ M\^56 + 37\ M\^58 - 292\ M\^60 + 68\ M\^62 + 
                    116\ M\^64 - 79\ M\^66 + 15\ M\^68)\) + 
              L\^7\ \((\(-6\)\ M\^40 + 25\ M\^42 - 28\ M\^44 - 14\ M\^46 + 
                    54\ M\^48 - 12\ M\^50 - 69\ M\^52 + 124\ M\^54 + 
                    5\ M\^56 - 158\ M\^58 + 130\ M\^60 + 178\ M\^62 - 
                    114\ M\^64 - 44\ M\^66 + 77\ M\^68 - 8\ M\^70 - 
                    43\ M\^72 + 29\ M\^74 - 6\ M\^76)\) + 
              L\^9\ \((2\ M\^66 - 3\ M\^68 + 2\ M\^70 - M\^72 + 9\ M\^76 + 
                    2\ M\^78 - 3\ M\^80 + 4\ M\^82 - 2\ M\^84)\) + 
              L\^8\ \((M\^48 - 3\ M\^50 + 3\ M\^52 - M\^54 + 3\ M\^58 + 
                    17\ M\^60 - 28\ M\^62 - 6\ M\^64 + 41\ M\^66 + 6\ M\^68 + 
                    M\^70 + 20\ M\^72 - 5\ M\^74 - 3\ M\^76 - 4\ M\^78 + 
                    6\ M\^80 - 4\ M\^82 + M\^84)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 5]\), \(1 + L\^11\ M\^46 + 
              L\ \((\(-7\) + 14\ M\^2 + 8\ M\^4 - 7\ M\^6 + 3\ M\^8 - 
                    2\ M\^12 + 3\ M\^14 - 2\ M\^16 + M\^18)\) + 
              L\^2\ \((21 - 76\ M\^2 + 19\ M\^4 + 137\ M\^6 - 35\ M\^8 - 
                    57\ M\^10 + 64\ M\^12 - 18\ M\^14 - 30\ M\^16 + 
                    32\ M\^18 + 5\ M\^20 - 12\ M\^22 + 8\ M\^24 - 5\ M\^26 + 
                    3\ M\^28 - M\^30)\) + 
              L\^3\ \((\(-35\) + 170\ M\^2 - 157\ M\^4 - 354\ M\^6 + 
                    439\ M\^8 + 426\ M\^10 - 419\ M\^12 - 72\ M\^14 + 
                    242\ M\^16 - 158\ M\^18 - 12\ M\^20 + 164\ M\^22 - 
                    65\ M\^24 - 22\ M\^26 + 29\ M\^28 - 7\ M\^30 - 
                    13\ M\^32 + 16\ M\^34 - 9\ M\^36 + 2\ M\^38)\) + 
              L\^8\ \((2\ M\^8 - 9\ M\^10 + 16\ M\^12 - 13\ M\^14 - 
                    7\ M\^16 + 29\ M\^18 - 22\ M\^20 - 65\ M\^22 + 
                    164\ M\^24 - 12\ M\^26 - 158\ M\^28 + 242\ M\^30 - 
                    72\ M\^32 - 419\ M\^34 + 426\ M\^36 + 439\ M\^38 - 
                    354\ M\^40 - 157\ M\^42 + 170\ M\^44 - 35\ M\^46)\) + 
              L\^6\ \((7 - 44\ M\^2 + 91\ M\^4 - 7\ M\^6 - 222\ M\^8 + 
                    130\ M\^10 + 533\ M\^12 - 634\ M\^14 - 733\ M\^16 + 
                    985\ M\^18 + 899\ M\^20 - 350\ M\^22 - 413\ M\^24 - 
                    894\ M\^26 - 87\ M\^28 + 1903\ M\^30 + 456\ M\^32 - 
                    1583\ M\^34 - 88\ M\^36 + 753\ M\^38 - 111\ M\^40 - 
                    238\ M\^42 + 130\ M\^44 - 21\ M\^46)\) + 
              L\^10\ \((M\^28 - 2\ M\^30 + 3\ M\^32 - 2\ M\^34 + 3\ M\^38 - 
                    7\ M\^40 + 8\ M\^42 + 14\ M\^44 - 7\ M\^46)\) + 
              L\^4\ \((35 - 200\ M\^2 + 290\ M\^4 + 334\ M\^6 - 952\ M\^8 - 
                    398\ M\^10 + 1538\ M\^12 + 364\ M\^14 - 976\ M\^16 - 
                    10\ M\^18 + 112\ M\^20 - 100\ M\^22 + 438\ M\^24 + 
                    49\ M\^26 - 300\ M\^28 + 55\ M\^30 + 129\ M\^32 - 
                    92\ M\^34 + 2\ M\^36 + 12\ M\^38 + 8\ M\^40 - 13\ M\^42 + 
                    6\ M\^44 - M\^46)\) + 
              L\^5\ \((\(-21\) + 130\ M\^2 - 238\ M\^4 - 111\ M\^6 + 
                    753\ M\^8 - 88\ M\^10 - 1583\ M\^12 + 456\ M\^14 + 
                    1903\ M\^16 - 87\ M\^18 - 894\ M\^20 - 413\ M\^22 - 
                    350\ M\^24 + 899\ M\^26 + 985\ M\^28 - 733\ M\^30 - 
                    634\ M\^32 + 533\ M\^34 + 130\ M\^36 - 222\ M\^38 - 
                    7\ M\^40 + 91\ M\^42 - 44\ M\^44 + 7\ M\^46)\) + 
              L\^9\ \((\(-M\^16\) + 3\ M\^18 - 5\ M\^20 + 8\ M\^22 - 
                    12\ M\^24 + 5\ M\^26 + 32\ M\^28 - 30\ M\^30 - 
                    18\ M\^32 + 64\ M\^34 - 57\ M\^36 - 35\ M\^38 + 
                    137\ M\^40 + 19\ M\^42 - 76\ M\^44 + 21\ M\^46)\) + 
              L\^7\ \((\(-1\) + 6\ M\^2 - 13\ M\^4 + 8\ M\^6 + 12\ M\^8 + 
                    2\ M\^10 - 92\ M\^12 + 129\ M\^14 + 55\ M\^16 - 
                    300\ M\^18 + 49\ M\^20 + 438\ M\^22 - 100\ M\^24 + 
                    112\ M\^26 - 10\ M\^28 - 976\ M\^30 + 364\ M\^32 + 
                    1538\ M\^34 - 398\ M\^36 - 952\ M\^38 + 334\ M\^40 + 
                    290\ M\^42 - 200\ M\^44 + 35\ M\^46)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 6]\), \(L\^12 + M\^144 + 
              L\^11\ \((\(-1\) + 4\ M\^2 - 5\ M\^4 + 2\ M\^6 - 6\ M\^10 + 
                    19\ M\^12 + 8\ M\^14 - 13\ M\^16 + 4\ M\^18)\) + 
              L\^10\ \((M\^4 - 4\ M\^6 + 6\ M\^8 - 8\ M\^10 + 26\ M\^12 - 
                    39\ M\^14 - 14\ M\^16 + 100\ M\^18 - 86\ M\^20 - 
                    97\ M\^22 + 226\ M\^24 + 75\ M\^26 - 173\ M\^28 + 
                    12\ M\^30 + 74\ M\^32 - 39\ M\^34 + 6\ M\^36)\) + 
              L\^9\ \((4\ M\^14 - 25\ M\^16 + 58\ M\^18 - 45\ M\^20 - 
                    24\ M\^22 + 36\ M\^24 + 10\ M\^26 + 110\ M\^28 - 
                    110\ M\^30 - 477\ M\^32 + 336\ M\^34 + 891\ M\^36 - 
                    266\ M\^38 - 712\ M\^40 + 304\ M\^42 + 369\ M\^44 - 
                    222\ M\^46 - 108\ M\^48 + 126\ M\^50 - 39\ M\^52 + 
                    4\ M\^54)\) + 
              L\^8\ \((6\ M\^24 - 48\ M\^26 + 144\ M\^28 - 153\ M\^30 - 
                    111\ M\^32 + 316\ M\^34 - 22\ M\^36 - 159\ M\^38 + 
                    253\ M\^40 - 758\ M\^42 - 356\ M\^44 + 1634\ M\^46 + 
                    572\ M\^48 - 1144\ M\^50 - 120\ M\^52 + 612\ M\^54 + 
                    54\ M\^56 - 290\ M\^58 - 44\ M\^60 + 127\ M\^62 + 
                    56\ M\^64 - 124\ M\^66 + 62\ M\^68 - 13\ M\^70 + 
                    M\^72)\) + 
              L\^7\ \((4\ M\^34 - 37\ M\^36 + 132\ M\^38 - 213\ M\^40 + 
                    108\ M\^42 + 162\ M\^44 - 538\ M\^46 + 561\ M\^48 + 
                    1114\ M\^50 - 2208\ M\^52 - 1186\ M\^54 + 3106\ M\^56 + 
                    742\ M\^58 - 2181\ M\^60 + 656\ M\^62 + 1580\ M\^64 - 
                    746\ M\^66 - 750\ M\^68 + 524\ M\^70 + 51\ M\^72 - 
                    194\ M\^74 + 121\ M\^76 + 78\ M\^78 - 169\ M\^80 + 
                    96\ M\^82 - 23\ M\^84 + 2\ M\^86)\) + 
              L\^6\ \((M\^44 - 10\ M\^46 + 43\ M\^48 - 121\ M\^50 + 
                    256\ M\^52 - 240\ M\^54 - 408\ M\^56 + 975\ M\^58 + 
                    364\ M\^60 - 1926\ M\^62 + 477\ M\^64 + 1707\ M\^66 - 
                    1739\ M\^68 - 385\ M\^70 + 2936\ M\^72 - 385\ M\^74 - 
                    1739\ M\^76 + 1707\ M\^78 + 477\ M\^80 - 1926\ M\^82 + 
                    364\ M\^84 + 975\ M\^86 - 408\ M\^88 - 240\ M\^90 + 
                    256\ M\^92 - 121\ M\^94 + 43\ M\^96 - 10\ M\^98 + 
                    M\^100)\) + 
              L\^5\ \((2\ M\^58 - 23\ M\^60 + 96\ M\^62 - 169\ M\^64 + 
                    78\ M\^66 + 121\ M\^68 - 194\ M\^70 + 51\ M\^72 + 
                    524\ M\^74 - 750\ M\^76 - 746\ M\^78 + 1580\ M\^80 + 
                    656\ M\^82 - 2181\ M\^84 + 742\ M\^86 + 3106\ M\^88 - 
                    1186\ M\^90 - 2208\ M\^92 + 1114\ M\^94 + 561\ M\^96 - 
                    538\ M\^98 + 162\ M\^100 + 108\ M\^102 - 213\ M\^104 + 
                    132\ M\^106 - 37\ M\^108 + 4\ M\^110)\) + 
              L\^4\ \((M\^72 - 13\ M\^74 + 62\ M\^76 - 124\ M\^78 + 
                    56\ M\^80 + 127\ M\^82 - 44\ M\^84 - 290\ M\^86 + 
                    54\ M\^88 + 612\ M\^90 - 120\ M\^92 - 1144\ M\^94 + 
                    572\ M\^96 + 1634\ M\^98 - 356\ M\^100 - 758\ M\^102 + 
                    253\ M\^104 - 159\ M\^106 - 22\ M\^108 + 316\ M\^110 - 
                    111\ M\^112 - 153\ M\^114 + 144\ M\^116 - 48\ M\^118 + 
                    6\ M\^120)\) + 
              L\^3\ \((4\ M\^90 - 39\ M\^92 + 126\ M\^94 - 108\ M\^96 - 
                    222\ M\^98 + 369\ M\^100 + 304\ M\^102 - 712\ M\^104 - 
                    266\ M\^106 + 891\ M\^108 + 336\ M\^110 - 477\ M\^112 - 
                    110\ M\^114 + 110\ M\^116 + 10\ M\^118 + 36\ M\^120 - 
                    24\ M\^122 - 45\ M\^124 + 58\ M\^126 - 25\ M\^128 + 
                    4\ M\^130)\) + 
              L\^2\ \((6\ M\^108 - 39\ M\^110 + 74\ M\^112 + 12\ M\^114 - 
                    173\ M\^116 + 75\ M\^118 + 226\ M\^120 - 97\ M\^122 - 
                    86\ M\^124 + 100\ M\^126 - 14\ M\^128 - 39\ M\^130 + 
                    26\ M\^132 - 8\ M\^134 + 6\ M\^136 - 4\ M\^138 + 
                    M\^140)\) + 
              L\ \((4\ M\^126 - 13\ M\^128 + 8\ M\^130 + 19\ M\^132 - 
                    6\ M\^134 + 2\ M\^138 - 5\ M\^140 + 4\ M\^142 - 
                    M\^144)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 7]\), \(L\^14 + M\^116 + 
              L\^13\ \((\(-2\) + 8\ M\^2 - 12\ M\^4 + 2\ M\^6 + 23\ M\^8 - 
                    5\ M\^10 - 2\ M\^12 + 6\ M\^14 - 7\ M\^16 + 3\ M\^18)\) + 
              L\^12\ \((1 - 8\ M\^2 + 26\ M\^4 - 38\ M\^6 + 12\ M\^8 + 
                    25\ M\^10 - 9\ M\^12 - 19\ M\^14 - 28\ M\^16 + 
                    177\ M\^18 + 109\ M\^20 - 215\ M\^22 - 28\ M\^24 + 
                    120\ M\^26 - 22\ M\^28 - 28\ M\^30 + 27\ M\^32 - 
                    14\ M\^34 + 3\ M\^36)\) + 
              L\^11\ \((\(-2\)\ M\^4 + 20\ M\^6 - 91\ M\^8 + 202\ M\^10 - 
                    104\ M\^12 - 391\ M\^14 + 500\ M\^16 + 644\ M\^18 - 
                    1183\ M\^20 - 898\ M\^22 + 1939\ M\^24 + 1101\ M\^26 - 
                    1601\ M\^28 - 411\ M\^30 + 714\ M\^32 - 97\ M\^34 + 
                    3\ M\^36 + 199\ M\^38 - 219\ M\^40 - 31\ M\^42 + 
                    86\ M\^44 + 4\ M\^46 - 35\ M\^48 + 21\ M\^50 - 7\ M\^52 + 
                    M\^54)\) + 
              L\^10\ \((M\^8 - 16\ M\^10 + 112\ M\^12 - 391\ M\^14 + 
                    593\ M\^16 + 273\ M\^18 - 2022\ M\^20 + 738\ M\^22 + 
                    4448\ M\^24 - 2886\ M\^26 - 7089\ M\^28 + 4441\ M\^30 + 
                    8444\ M\^32 - 2871\ M\^34 - 5980\ M\^36 + 565\ M\^38 + 
                    3381\ M\^40 + 865\ M\^42 - 1676\ M\^44 - 579\ M\^46 + 
                    898\ M\^48 - 255\ M\^50 - 166\ M\^52 + 282\ M\^54 - 
                    47\ M\^56 - 138\ M\^58 + 101\ M\^60 - 28\ M\^62 + 
                    3\ M\^64)\) + 
              L\^9\ \((4\ M\^14 - 51\ M\^16 + 289\ M\^18 - 818\ M\^20 + 
                    811\ M\^22 + 1396\ M\^24 - 3546\ M\^26 - 1591\ M\^28 + 
                    8595\ M\^30 + 1890\ M\^32 - 14218\ M\^34 - 5297\ M\^36 + 
                    17199\ M\^38 + 11712\ M\^40 - 13537\ M\^42 - 
                    14043\ M\^44 + 9480\ M\^46 + 11536\ M\^48 - 5242\ M\^50 - 
                    5412\ M\^52 + 2101\ M\^54 + 893\ M\^56 + 197\ M\^58 - 
                    159\ M\^60 - 449\ M\^62 + 247\ M\^64 + 176\ M\^66 - 
                    270\ M\^68 + 139\ M\^70 - 33\ M\^72 + 3\ M\^74)\) + 
              L\^8\ \((6\ M\^20 - 74\ M\^22 + 380\ M\^24 - 932\ M\^26 + 
                    647\ M\^28 + 1665\ M\^30 - 2467\ M\^32 - 2976\ M\^34 + 
                    4676\ M\^36 + 6768\ M\^38 - 4799\ M\^40 - 16223\ M\^42 + 
                    1881\ M\^44 + 26298\ M\^46 + 4628\ M\^48 - 27211\ M\^50 - 
                    6975\ M\^52 + 20943\ M\^54 + 8308\ M\^56 - 10261\ M\^58 - 
                    6851\ M\^60 + 3730\ M\^62 + 4615\ M\^64 - 2475\ M\^66 - 
                    1429\ M\^68 + 1497\ M\^70 - 137\ M\^72 - 526\ M\^74 + 
                    455\ M\^76 - 211\ M\^78 + 64\ M\^80 - 12\ M\^82 + 
                    M\^84)\) + 
              L\^7\ \((4\ M\^26 - 49\ M\^28 + 244\ M\^30 - 600\ M\^32 + 
                    586\ M\^34 + 326\ M\^36 - 852\ M\^38 + 9\ M\^40 - 
                    809\ M\^42 + 1134\ M\^44 + 6452\ M\^46 - 5749\ M\^48 - 
                    13890\ M\^50 + 7885\ M\^52 + 21264\ M\^54 - 2956\ M\^56 - 
                    22566\ M\^58 - 2956\ M\^60 + 21264\ M\^62 + 7885\ M\^64 - 
                    13890\ M\^66 - 5749\ M\^68 + 6452\ M\^70 + 1134\ M\^72 - 
                    809\ M\^74 + 9\ M\^76 - 852\ M\^78 + 326\ M\^80 + 
                    586\ M\^82 - 600\ M\^84 + 244\ M\^86 - 49\ M\^88 + 
                    4\ M\^90)\) + 
              L\^6\ \((M\^32 - 12\ M\^34 + 64\ M\^36 - 211\ M\^38 + 
                    455\ M\^40 - 526\ M\^42 - 137\ M\^44 + 1497\ M\^46 - 
                    1429\ M\^48 - 2475\ M\^50 + 4615\ M\^52 + 3730\ M\^54 - 
                    6851\ M\^56 - 10261\ M\^58 + 8308\ M\^60 + 20943\ M\^62 - 
                    6975\ M\^64 - 27211\ M\^66 + 4628\ M\^68 + 26298\ M\^70 + 
                    1881\ M\^72 - 16223\ M\^74 - 4799\ M\^76 + 6768\ M\^78 + 
                    4676\ M\^80 - 2976\ M\^82 - 2467\ M\^84 + 1665\ M\^86 + 
                    647\ M\^88 - 932\ M\^90 + 380\ M\^92 - 74\ M\^94 + 
                    6\ M\^96)\) + 
              L\^5\ \((3\ M\^42 - 33\ M\^44 + 139\ M\^46 - 270\ M\^48 + 
                    176\ M\^50 + 247\ M\^52 - 449\ M\^54 - 159\ M\^56 + 
                    197\ M\^58 + 893\ M\^60 + 2101\ M\^62 - 5412\ M\^64 - 
                    5242\ M\^66 + 11536\ M\^68 + 9480\ M\^70 - 14043\ M\^72 - 
                    13537\ M\^74 + 11712\ M\^76 + 17199\ M\^78 - 
                    5297\ M\^80 - 14218\ M\^82 + 1890\ M\^84 + 8595\ M\^86 - 
                    1591\ M\^88 - 3546\ M\^90 + 1396\ M\^92 + 811\ M\^94 - 
                    818\ M\^96 + 289\ M\^98 - 51\ M\^100 + 4\ M\^102)\) + 
              L\^4\ \((3\ M\^52 - 28\ M\^54 + 101\ M\^56 - 138\ M\^58 - 
                    47\ M\^60 + 282\ M\^62 - 166\ M\^64 - 255\ M\^66 + 
                    898\ M\^68 - 579\ M\^70 - 1676\ M\^72 + 865\ M\^74 + 
                    3381\ M\^76 + 565\ M\^78 - 5980\ M\^80 - 2871\ M\^82 + 
                    8444\ M\^84 + 4441\ M\^86 - 7089\ M\^88 - 2886\ M\^90 + 
                    4448\ M\^92 + 738\ M\^94 - 2022\ M\^96 + 273\ M\^98 + 
                    593\ M\^100 - 391\ M\^102 + 112\ M\^104 - 16\ M\^106 + 
                    M\^108)\) + 
              L\^3\ \((M\^62 - 7\ M\^64 + 21\ M\^66 - 35\ M\^68 + 4\ M\^70 + 
                    86\ M\^72 - 31\ M\^74 - 219\ M\^76 + 199\ M\^78 + 
                    3\ M\^80 - 97\ M\^82 + 714\ M\^84 - 411\ M\^86 - 
                    1601\ M\^88 + 1101\ M\^90 + 1939\ M\^92 - 898\ M\^94 - 
                    1183\ M\^96 + 644\ M\^98 + 500\ M\^100 - 391\ M\^102 - 
                    104\ M\^104 + 202\ M\^106 - 91\ M\^108 + 20\ M\^110 - 
                    2\ M\^112)\) + 
              L\ \((3\ M\^98 - 7\ M\^100 + 6\ M\^102 - 2\ M\^104 - 
                    5\ M\^106 + 23\ M\^108 + 2\ M\^110 - 12\ M\^112 + 
                    8\ M\^114 - 2\ M\^116)\) + 
              L\^2\ \((3\ M\^80 - 14\ M\^82 + 27\ M\^84 - 28\ M\^86 - 
                    22\ M\^88 + 120\ M\^90 - 28\ M\^92 - 215\ M\^94 + 
                    109\ M\^96 + 177\ M\^98 - 28\ M\^100 - 19\ M\^102 - 
                    9\ M\^104 + 25\ M\^106 + 12\ M\^108 - 38\ M\^110 + 
                    26\ M\^112 - 8\ M\^114 + M\^116)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 8]\), \(L\^15\ M\^20 + M\^82 + 
              L\^14\ \((\(-2\)\ M\^12 + 6\ M\^14 - 7\ M\^16 + 3\ M\^18 + 
                    3\ M\^20 - 12\ M\^22 + 23\ M\^24 + 13\ M\^26 - 
                    17\ M\^28 + 5\ M\^30)\) + 
              L\^13\ \((M\^4 - 6\ M\^6 + 17\ M\^8 - 28\ M\^10 + 9\ M\^12 + 
                    72\ M\^14 - 88\ M\^16 - 116\ M\^18 + 201\ M\^20 + 
                    105\ M\^22 - 245\ M\^24 - 119\ M\^26 + 336\ M\^28 + 
                    173\ M\^30 - 269\ M\^32 - 17\ M\^34 + 133\ M\^36 - 
                    64\ M\^38 + 10\ M\^40)\) + 
              L\^12\ \((\(-1\) + 7\ M\^2 - 22\ M\^4 + 45\ M\^6 - 54\ M\^8 - 
                    45\ M\^10 + 245\ M\^12 - 88\ M\^14 - 616\ M\^16 + 
                    694\ M\^18 + 696\ M\^20 - 1401\ M\^22 - 427\ M\^24 + 
                    1647\ M\^26 - 216\ M\^28 - 1190\ M\^30 + 995\ M\^32 + 
                    1071\ M\^34 - 1044\ M\^36 - 408\ M\^38 + 905\ M\^40 - 
                    185\ M\^42 - 371\ M\^44 + 298\ M\^46 - 90\ M\^48 + 
                    10\ M\^50)\) + 
              L\^11\ \((1 - 11\ M\^2 + 47\ M\^4 - 85\ M\^6 + 12\ M\^8 + 
                    113\ M\^10 + 179\ M\^12 - 605\ M\^14 - 431\ M\^16 + 
                    1804\ M\^18 + 422\ M\^20 - 3704\ M\^22 + 1478\ M\^24 + 
                    4096\ M\^26 - 4308\ M\^28 - 2743\ M\^30 + 5102\ M\^32 + 
                    325\ M\^34 - 1927\ M\^36 + 2915\ M\^38 - 270\ M\^40 - 
                    3127\ M\^42 + 2003\ M\^44 + 1492\ M\^46 - 1948\ M\^48 + 
                    208\ M\^50 + 744\ M\^52 - 622\ M\^54 + 256\ M\^56 - 
                    56\ M\^58 + 5\ M\^60)\) + 
              L\^10\ \((\(-2\)\ M\^4 + 26\ M\^6 - 146\ M\^8 + 408\ M\^10 - 
                    373\ M\^12 - 810\ M\^14 + 1734\ M\^16 + 1649\ M\^18 - 
                    5292\ M\^20 - 2440\ M\^22 + 10109\ M\^24 + 3740\ M\^26 - 
                    13908\ M\^28 - 2656\ M\^30 + 11794\ M\^32 - 913\ M\^34 - 
                    7333\ M\^36 + 2445\ M\^38 + 5976\ M\^40 + 4363\ M\^42 - 
                    3225\ M\^44 - 7893\ M\^46 + 3152\ M\^48 + 7419\ M\^50 - 
                    4151\ M\^52 - 3387\ M\^54 + 3464\ M\^56 - 4\ M\^58 - 
                    1439\ M\^60 + 973\ M\^62 - 347\ M\^64 + 82\ M\^66 - 
                    13\ M\^68 + M\^70)\) + 
              L\^9\ \((M\^8 - 19\ M\^10 + 155\ M\^12 - 652\ M\^14 + 
                    1300\ M\^16 - 68\ M\^18 - 4334\ M\^20 + 3727\ M\^22 + 
                    10183\ M\^24 - 12851\ M\^26 - 17902\ M\^28 + 
                    21976\ M\^30 + 28451\ M\^32 - 21393\ M\^34 - 
                    34691\ M\^36 + 3322\ M\^38 + 29348\ M\^40 + 
                    15552\ M\^42 - 15649\ M\^44 - 13280\ M\^46 + 
                    13708\ M\^48 + 4210\ M\^50 - 11902\ M\^52 + 4211\ M\^54 + 
                    7148\ M\^56 - 7348\ M\^58 - 409\ M\^60 + 4397\ M\^62 - 
                    2347\ M\^64 - 748\ M\^66 + 1624\ M\^68 - 981\ M\^70 + 
                    317\ M\^72 - 55\ M\^74 + 4\ M\^76)\) + 
              L\^8\ \((4\ M\^14 - 61\ M\^16 + 404\ M\^18 - 1374\ M\^20 + 
                    1971\ M\^22 + 1398\ M\^24 - 7488\ M\^26 + 1252\ M\^28 + 
                    17897\ M\^30 - 6873\ M\^32 - 31806\ M\^34 + 4000\ M\^36 + 
                    49455\ M\^38 + 19864\ M\^40 - 58202\ M\^42 - 
                    65255\ M\^44 + 42831\ M\^46 + 97601\ M\^48 - 
                    5314\ M\^50 - 78858\ M\^52 - 13595\ M\^54 + 
                    39698\ M\^56 + 16212\ M\^58 - 10294\ M\^60 - 
                    12735\ M\^62 - 1373\ M\^64 + 10117\ M\^66 + 572\ M\^68 - 
                    5751\ M\^70 + 1336\ M\^72 + 1914\ M\^74 - 1547\ M\^76 + 
                    515\ M\^78 - 86\ M\^80 + 6\ M\^82)\) + 
              L\^7\ \((6\ M\^20 - 86\ M\^22 + 515\ M\^24 - 1547\ M\^26 + 
                    1914\ M\^28 + 1336\ M\^30 - 5751\ M\^32 + 572\ M\^34 + 
                    10117\ M\^36 - 1373\ M\^38 - 12735\ M\^40 - 
                    10294\ M\^42 + 16212\ M\^44 + 39698\ M\^46 - 
                    13595\ M\^48 - 78858\ M\^50 - 5314\ M\^52 + 
                    97601\ M\^54 + 42831\ M\^56 - 65255\ M\^58 - 
                    58202\ M\^60 + 19864\ M\^62 + 49455\ M\^64 + 
                    4000\ M\^66 - 31806\ M\^68 - 6873\ M\^70 + 17897\ M\^72 + 
                    1252\ M\^74 - 7488\ M\^76 + 1398\ M\^78 + 1971\ M\^80 - 
                    1374\ M\^82 + 404\ M\^84 - 61\ M\^86 + 4\ M\^88)\) + 
              L\ \((5\ M\^72 - 17\ M\^74 + 13\ M\^76 + 23\ M\^78 - 
                    12\ M\^80 + 3\ M\^82 + 3\ M\^84 - 7\ M\^86 + 6\ M\^88 - 
                    2\ M\^90)\) + 
              L\^6\ \((4\ M\^26 - 55\ M\^28 + 317\ M\^30 - 981\ M\^32 + 
                    1624\ M\^34 - 748\ M\^36 - 2347\ M\^38 + 4397\ M\^40 - 
                    409\ M\^42 - 7348\ M\^44 + 7148\ M\^46 + 4211\ M\^48 - 
                    11902\ M\^50 + 4210\ M\^52 + 13708\ M\^54 - 
                    13280\ M\^56 - 15649\ M\^58 + 15552\ M\^60 + 
                    29348\ M\^62 + 3322\ M\^64 - 34691\ M\^66 - 
                    21393\ M\^68 + 28451\ M\^70 + 21976\ M\^72 - 
                    17902\ M\^74 - 12851\ M\^76 + 10183\ M\^78 + 
                    3727\ M\^80 - 4334\ M\^82 - 68\ M\^84 + 1300\ M\^86 - 
                    652\ M\^88 + 155\ M\^90 - 19\ M\^92 + M\^94)\) + 
              L\^5\ \((M\^32 - 13\ M\^34 + 82\ M\^36 - 347\ M\^38 + 
                    973\ M\^40 - 1439\ M\^42 - 4\ M\^44 + 3464\ M\^46 - 
                    3387\ M\^48 - 4151\ M\^50 + 7419\ M\^52 + 3152\ M\^54 - 
                    7893\ M\^56 - 3225\ M\^58 + 4363\ M\^60 + 5976\ M\^62 + 
                    2445\ M\^64 - 7333\ M\^66 - 913\ M\^68 + 11794\ M\^70 - 
                    2656\ M\^72 - 13908\ M\^74 + 3740\ M\^76 + 10109\ M\^78 - 
                    2440\ M\^80 - 5292\ M\^82 + 1649\ M\^84 + 1734\ M\^86 - 
                    810\ M\^88 - 373\ M\^90 + 408\ M\^92 - 146\ M\^94 + 
                    26\ M\^96 - 2\ M\^98)\) + 
              L\^2\ \((10\ M\^62 - 64\ M\^64 + 133\ M\^66 - 17\ M\^68 - 
                    269\ M\^70 + 173\ M\^72 + 336\ M\^74 - 119\ M\^76 - 
                    245\ M\^78 + 105\ M\^80 + 201\ M\^82 - 116\ M\^84 - 
                    88\ M\^86 + 72\ M\^88 + 9\ M\^90 - 28\ M\^92 + 
                    17\ M\^94 - 6\ M\^96 + M\^98)\) + 
              L\^3\ \((10\ M\^52 - 90\ M\^54 + 298\ M\^56 - 371\ M\^58 - 
                    185\ M\^60 + 905\ M\^62 - 408\ M\^64 - 1044\ M\^66 + 
                    1071\ M\^68 + 995\ M\^70 - 1190\ M\^72 - 216\ M\^74 + 
                    1647\ M\^76 - 427\ M\^78 - 1401\ M\^80 + 696\ M\^82 + 
                    694\ M\^84 - 616\ M\^86 - 88\ M\^88 + 245\ M\^90 - 
                    45\ M\^92 - 54\ M\^94 + 45\ M\^96 - 22\ M\^98 + 
                    7\ M\^100 - M\^102)\) + 
              L\^4\ \((5\ M\^42 - 56\ M\^44 + 256\ M\^46 - 622\ M\^48 + 
                    744\ M\^50 + 208\ M\^52 - 1948\ M\^54 + 1492\ M\^56 + 
                    2003\ M\^58 - 3127\ M\^60 - 270\ M\^62 + 2915\ M\^64 - 
                    1927\ M\^66 + 325\ M\^68 + 5102\ M\^70 - 2743\ M\^72 - 
                    4308\ M\^74 + 4096\ M\^76 + 1478\ M\^78 - 3704\ M\^80 + 
                    422\ M\^82 + 1804\ M\^84 - 431\ M\^86 - 605\ M\^88 + 
                    179\ M\^90 + 113\ M\^92 + 12\ M\^94 - 85\ M\^96 + 
                    47\ M\^98 - 11\ M\^100 + M\^102)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 9]\), \(L\^15 + M\^186 + 
              L\^14\ \((\(-1\) + 4\ M\^2 - 8\ M\^4 + 8\ M\^6 + 2\ M\^8 - 
                    16\ M\^10 + 24\ M\^12 + 13\ M\^14 - 17\ M\^16 + 
                    6\ M\^18)\) + 
              L\^13\ \((\(-2\)\ M\^6 + 14\ M\^8 - 42\ M\^10 + 74\ M\^12 - 
                    44\ M\^14 - 136\ M\^16 + 279\ M\^18 - 26\ M\^20 - 
                    402\ M\^22 + 366\ M\^24 + 280\ M\^26 - 360\ M\^28 + 
                    16\ M\^30 + 158\ M\^32 - 85\ M\^34 + 15\ M\^36)\) + 
              L\^12\ \((\(-M\^12\) + 12\ M\^14 - 63\ M\^16 + 180\ M\^18 - 
                    254\ M\^20 - 32\ M\^22 + 756\ M\^24 - 912\ M\^26 - 
                    670\ M\^28 + 2374\ M\^30 - 948\ M\^32 - 2740\ M\^34 + 
                    3192\ M\^36 + 1664\ M\^38 - 3408\ M\^40 + 338\ M\^42 + 
                    1967\ M\^44 - 954\ M\^46 - 401\ M\^48 + 505\ M\^50 - 
                    170\ M\^52 + 20\ M\^54)\) + 
              L\^11\ \((3\ M\^20 - 30\ M\^22 + 135\ M\^24 - 332\ M\^26 + 
                    363\ M\^28 + 271\ M\^30 - 1237\ M\^32 + 677\ M\^34 + 
                    1930\ M\^36 - 3066\ M\^38 + 3\ M\^40 + 4678\ M\^42 - 
                    5200\ M\^44 - 3486\ M\^46 + 11298\ M\^48 + 222\ M\^50 - 
                    11765\ M\^52 + 4562\ M\^54 + 7854\ M\^56 - 5902\ M\^58 - 
                    2243\ M\^60 + 3679\ M\^62 - 491\ M\^64 - 1103\ M\^66 + 
                    700\ M\^68 - 170\ M\^70 + 15\ M\^72)\) + 
              L\^10\ \((\(-3\)\ M\^28 + 30\ M\^30 - 129\ M\^32 + 291\ M\^34 - 
                    281\ M\^36 - 145\ M\^38 + 408\ M\^40 + 429\ M\^42 - 
                    1075\ M\^44 - 503\ M\^46 + 1131\ M\^48 + 1580\ M\^50 - 
                    255\ M\^52 - 2703\ M\^54 - 4350\ M\^56 + 3940\ M\^58 + 
                    11451\ M\^60 - 3982\ M\^62 - 12657\ M\^64 + 8269\ M\^66 + 
                    10067\ M\^68 - 10044\ M\^70 - 3951\ M\^72 + 8427\ M\^74 - 
                    1089\ M\^76 - 3815\ M\^78 + 1921\ M\^80 + 768\ M\^82 - 
                    1103\ M\^84 + 455\ M\^86 - 85\ M\^88 + 6\ M\^90)\) + 
              L\^9\ \((M\^36 - 10\ M\^38 + 43\ M\^40 - 100\ M\^42 + 
                    150\ M\^44 - 275\ M\^46 + 658\ M\^48 - 621\ M\^50 - 
                    1260\ M\^52 + 2629\ M\^54 + 2308\ M\^56 - 8000\ M\^58 - 
                    887\ M\^60 + 13115\ M\^62 - 347\ M\^64 - 14515\ M\^66 - 
                    1967\ M\^68 + 8319\ M\^70 + 10777\ M\^72 + 1410\ M\^74 - 
                    15969\ M\^76 + 967\ M\^78 + 19712\ M\^80 - 6615\ M\^82 - 
                    14470\ M\^84 + 9563\ M\^86 + 6400\ M\^88 - 7795\ M\^90 - 
                    44\ M\^92 + 3105\ M\^94 - 1119\ M\^96 - 723\ M\^98 + 
                    900\ M\^100 - 437\ M\^102 + 118\ M\^104 - 17\ M\^106 + 
                    M\^108)\) + 
              L\^8\ \((3\ M\^50 - 37\ M\^52 + 185\ M\^54 - 437\ M\^56 + 
                    316\ M\^58 + 538\ M\^60 - 446\ M\^62 - 2282\ M\^64 + 
                    2812\ M\^66 + 3668\ M\^68 - 5708\ M\^70 - 6290\ M\^72 + 
                    9379\ M\^74 + 8608\ M\^76 - 8295\ M\^78 - 13385\ M\^80 + 
                    2819\ M\^82 + 18772\ M\^84 + 3939\ M\^86 - 17929\ M\^88 + 
                    3111\ M\^90 + 19994\ M\^92 - 9948\ M\^94 - 15764\ M\^96 + 
                    13898\ M\^98 + 7344\ M\^100 - 11102\ M\^102 - 
                    1388\ M\^104 + 6582\ M\^106 - 1556\ M\^108 - 
                    2023\ M\^110 + 879\ M\^112 + 656\ M\^114 - 691\ M\^116 + 
                    254\ M\^118 - 44\ M\^120 + 3\ M\^122)\) + 
              L\^7\ \((3\ M\^64 - 44\ M\^66 + 254\ M\^68 - 691\ M\^70 + 
                    656\ M\^72 + 879\ M\^74 - 2023\ M\^76 - 1556\ M\^78 + 
                    6582\ M\^80 - 1388\ M\^82 - 11102\ M\^84 + 7344\ M\^86 + 
                    13898\ M\^88 - 15764\ M\^90 - 9948\ M\^92 + 
                    19994\ M\^94 + 3111\ M\^96 - 17929\ M\^98 + 
                    3939\ M\^100 + 18772\ M\^102 + 2819\ M\^104 - 
                    13385\ M\^106 - 8295\ M\^108 + 8608\ M\^110 + 
                    9379\ M\^112 - 6290\ M\^114 - 5708\ M\^116 + 
                    3668\ M\^118 + 2812\ M\^120 - 2282\ M\^122 - 
                    446\ M\^124 + 538\ M\^126 + 316\ M\^128 - 437\ M\^130 + 
                    185\ M\^132 - 37\ M\^134 + 3\ M\^136)\) + 
              L\^6\ \((M\^78 - 17\ M\^80 + 118\ M\^82 - 437\ M\^84 + 
                    900\ M\^86 - 723\ M\^88 - 1119\ M\^90 + 3105\ M\^92 - 
                    44\ M\^94 - 7795\ M\^96 + 6400\ M\^98 + 9563\ M\^100 - 
                    14470\ M\^102 - 6615\ M\^104 + 19712\ M\^106 + 
                    967\ M\^108 - 15969\ M\^110 + 1410\ M\^112 + 
                    10777\ M\^114 + 8319\ M\^116 - 1967\ M\^118 - 
                    14515\ M\^120 - 347\ M\^122 + 13115\ M\^124 - 
                    887\ M\^126 - 8000\ M\^128 + 2308\ M\^130 + 
                    2629\ M\^132 - 1260\ M\^134 - 621\ M\^136 + 658\ M\^138 - 
                    275\ M\^140 + 150\ M\^142 - 100\ M\^144 + 43\ M\^146 - 
                    10\ M\^148 + M\^150)\) + 
              L\^5\ \((6\ M\^96 - 85\ M\^98 + 455\ M\^100 - 1103\ M\^102 + 
                    768\ M\^104 + 1921\ M\^106 - 3815\ M\^108 - 
                    1089\ M\^110 + 8427\ M\^112 - 3951\ M\^114 - 
                    10044\ M\^116 + 10067\ M\^118 + 8269\ M\^120 - 
                    12657\ M\^122 - 3982\ M\^124 + 11451\ M\^126 + 
                    3940\ M\^128 - 4350\ M\^130 - 2703\ M\^132 - 
                    255\ M\^134 + 1580\ M\^136 + 1131\ M\^138 - 503\ M\^140 - 
                    1075\ M\^142 + 429\ M\^144 + 408\ M\^146 - 145\ M\^148 - 
                    281\ M\^150 + 291\ M\^152 - 129\ M\^154 + 30\ M\^156 - 
                    3\ M\^158)\) + 
              L\^4\ \((15\ M\^114 - 170\ M\^116 + 700\ M\^118 - 
                    1103\ M\^120 - 491\ M\^122 + 3679\ M\^124 - 
                    2243\ M\^126 - 5902\ M\^128 + 7854\ M\^130 + 
                    4562\ M\^132 - 11765\ M\^134 + 222\ M\^136 + 
                    11298\ M\^138 - 3486\ M\^140 - 5200\ M\^142 + 
                    4678\ M\^144 + 3\ M\^146 - 3066\ M\^148 + 1930\ M\^150 + 
                    677\ M\^152 - 1237\ M\^154 + 271\ M\^156 + 363\ M\^158 - 
                    332\ M\^160 + 135\ M\^162 - 30\ M\^164 + 3\ M\^166)\) + 
              L\^3\ \((20\ M\^132 - 170\ M\^134 + 505\ M\^136 - 401\ M\^138 - 
                    954\ M\^140 + 1967\ M\^142 + 338\ M\^144 - 3408\ M\^146 + 
                    1664\ M\^148 + 3192\ M\^150 - 2740\ M\^152 - 
                    948\ M\^154 + 2374\ M\^156 - 670\ M\^158 - 912\ M\^160 + 
                    756\ M\^162 - 32\ M\^164 - 254\ M\^166 + 180\ M\^168 - 
                    63\ M\^170 + 12\ M\^172 - M\^174)\) + 
              L\^2\ \((15\ M\^150 - 85\ M\^152 + 158\ M\^154 + 16\ M\^156 - 
                    360\ M\^158 + 280\ M\^160 + 366\ M\^162 - 402\ M\^164 - 
                    26\ M\^166 + 279\ M\^168 - 136\ M\^170 - 44\ M\^172 + 
                    74\ M\^174 - 42\ M\^176 + 14\ M\^178 - 2\ M\^180)\) + 
              L\ \((6\ M\^168 - 17\ M\^170 + 13\ M\^172 + 24\ M\^174 - 
                    16\ M\^176 + 2\ M\^178 + 8\ M\^180 - 8\ M\^182 + 
                    4\ M\^184 - M\^186)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 10]\), \(1 + L\^11\ M\^90 + 
              L\ \((\(-3\) + 10\ M\^2 - 13\ M\^4 + 2\ M\^6 + 29\ M\^8 - 
                    16\ M\^10 - 7\ M\^12 + 13\ M\^14 - 6\ M\^16 + 
                    2\ M\^18)\) + 
              L\^2\ \((3 - 20\ M\^2 + 58\ M\^4 - 51\ M\^6 - 97\ M\^8 + 
                    184\ M\^10 + 37\ M\^12 - 213\ M\^14 + 110\ M\^16 + 
                    131\ M\^18 - 109\ M\^20 - 16\ M\^22 + 35\ M\^24 + 
                    11\ M\^26 - 20\ M\^30 + 17\ M\^32 - 6\ M\^34 + M\^36)\) + 
              L\^3\ \((\(-1\) + 10\ M\^2 - 45\ M\^4 + 84\ M\^6 - 5\ M\^8 - 
                    201\ M\^10 + 157\ M\^12 + 383\ M\^14 - 589\ M\^16 - 
                    495\ M\^18 + 1247\ M\^20 + 407\ M\^22 - 1579\ M\^24 + 
                    127\ M\^26 + 1371\ M\^28 - 532\ M\^30 - 689\ M\^32 + 
                    581\ M\^34 + 154\ M\^36 - 325\ M\^38 + 58\ M\^40 + 
                    103\ M\^42 - 76\ M\^44 + 23\ M\^46 - 3\ M\^48)\) + 
              L\^4\ \((\(-3\)\ M\^6 + 33\ M\^8 - 131\ M\^10 + 214\ M\^12 + 
                    78\ M\^14 - 785\ M\^16 + 457\ M\^18 + 1657\ M\^20 - 
                    1668\ M\^22 - 2299\ M\^24 + 2689\ M\^26 + 2562\ M\^28 - 
                    2559\ M\^30 - 2273\ M\^32 + 1812\ M\^34 + 2002\ M\^36 - 
                    1002\ M\^38 - 1476\ M\^40 + 754\ M\^42 + 895\ M\^44 - 
                    635\ M\^46 - 286\ M\^48 + 389\ M\^50 - 13\ M\^52 - 
                    155\ M\^54 + 96\ M\^56 - 26\ M\^58 + 3\ M\^60)\) + 
              L\^5\ \((\(-3\)\ M\^12 + 32\ M\^14 - 126\ M\^16 + 197\ M\^18 + 
                    51\ M\^20 - 509\ M\^22 + 190\ M\^24 + 889\ M\^26 - 
                    522\ M\^28 - 1282\ M\^30 + 511\ M\^32 + 2222\ M\^34 - 
                    303\ M\^36 - 3338\ M\^38 + 504\ M\^40 + 4183\ M\^42 - 
                    757\ M\^44 - 3533\ M\^46 + 1251\ M\^48 + 2003\ M\^50 - 
                    1168\ M\^52 - 452\ M\^54 + 561\ M\^56 - 94\ M\^58 - 
                    78\ M\^60 + 63\ M\^62 - 74\ M\^64 + 72\ M\^66 - 
                    36\ M\^68 + 9\ M\^70 - M\^72)\) + 
              L\^6\ \((\(-M\^18\) + 9\ M\^20 - 36\ M\^22 + 72\ M\^24 - 
                    74\ M\^26 + 63\ M\^28 - 78\ M\^30 - 94\ M\^32 + 
                    561\ M\^34 - 452\ M\^36 - 1168\ M\^38 + 2003\ M\^40 + 
                    1251\ M\^42 - 3533\ M\^44 - 757\ M\^46 + 4183\ M\^48 + 
                    504\ M\^50 - 3338\ M\^52 - 303\ M\^54 + 2222\ M\^56 + 
                    511\ M\^58 - 1282\ M\^60 - 522\ M\^62 + 889\ M\^64 + 
                    190\ M\^66 - 509\ M\^68 + 51\ M\^70 + 197\ M\^72 - 
                    126\ M\^74 + 32\ M\^76 - 3\ M\^78)\) + 
              L\^7\ \((3\ M\^30 - 26\ M\^32 + 96\ M\^34 - 155\ M\^36 - 
                    13\ M\^38 + 389\ M\^40 - 286\ M\^42 - 635\ M\^44 + 
                    895\ M\^46 + 754\ M\^48 - 1476\ M\^50 - 1002\ M\^52 + 
                    2002\ M\^54 + 1812\ M\^56 - 2273\ M\^58 - 2559\ M\^60 + 
                    2562\ M\^62 + 2689\ M\^64 - 2299\ M\^66 - 1668\ M\^68 + 
                    1657\ M\^70 + 457\ M\^72 - 785\ M\^74 + 78\ M\^76 + 
                    214\ M\^78 - 131\ M\^80 + 33\ M\^82 - 3\ M\^84)\) + 
              L\^10\ \((2\ M\^72 - 6\ M\^74 + 13\ M\^76 - 7\ M\^78 - 
                    16\ M\^80 + 29\ M\^82 + 2\ M\^84 - 13\ M\^86 + 
                    10\ M\^88 - 3\ M\^90)\) + 
              L\^8\ \((\(-3\)\ M\^42 + 23\ M\^44 - 76\ M\^46 + 103\ M\^48 + 
                    58\ M\^50 - 325\ M\^52 + 154\ M\^54 + 581\ M\^56 - 
                    689\ M\^58 - 532\ M\^60 + 1371\ M\^62 + 127\ M\^64 - 
                    1579\ M\^66 + 407\ M\^68 + 1247\ M\^70 - 495\ M\^72 - 
                    589\ M\^74 + 383\ M\^76 + 157\ M\^78 - 201\ M\^80 - 
                    5\ M\^82 + 84\ M\^84 - 45\ M\^86 + 10\ M\^88 - M\^90)\) + 
              L\^9\ \((M\^54 - 6\ M\^56 + 17\ M\^58 - 20\ M\^60 + 11\ M\^64 + 
                    35\ M\^66 - 16\ M\^68 - 109\ M\^70 + 131\ M\^72 + 
                    110\ M\^74 - 213\ M\^76 + 37\ M\^78 + 184\ M\^80 - 
                    97\ M\^82 - 51\ M\^84 + 58\ M\^86 - 20\ M\^88 + 
                    3\ M\^90)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 11]\), \(M\^8 + L\^16\ M\^140 + 
              L\ \((\(-2\)\ M\^4 + 6\ M\^6 - 8\ M\^8 + 6\ M\^10 + 3\ M\^12 - 
                    17\ M\^14 + 26\ M\^16 + 15\ M\^18 - 19\ M\^20 + 
                    6\ M\^22)\) + 
              L\^2\ \((1 - 6\ M\^2 + 13\ M\^4 - 14\ M\^6 + 10\ M\^8 + 
                    2\ M\^10 + 3\ M\^12 - 53\ M\^14 - 20\ M\^16 + 
                    203\ M\^18 - 55\ M\^20 - 348\ M\^22 + 335\ M\^24 + 
                    326\ M\^26 - 359\ M\^28 - 21\ M\^30 + 177\ M\^32 - 
                    89\ M\^34 + 15\ M\^36)\) + 
              L\^3\ \((\(-4\)\ M\^4 + 31\ M\^6 - 98\ M\^8 + 147\ M\^10 - 
                    26\ M\^12 - 320\ M\^14 + 455\ M\^16 + 238\ M\^18 - 
                    896\ M\^20 + 200\ M\^22 + 432\ M\^24 + 66\ M\^26 + 
                    30\ M\^28 - 1060\ M\^30 + 940\ M\^32 + 1890\ M\^34 - 
                    1836\ M\^36 - 661\ M\^38 + 1727\ M\^40 - 502\ M\^42 - 
                    558\ M\^44 + 511\ M\^46 - 166\ M\^48 + 20\ M\^50)\) + 
              L\^4\ \((6\ M\^8 - 57\ M\^10 + 231\ M\^12 - 455\ M\^14 + 
                    175\ M\^16 + 1057\ M\^18 - 1614\ M\^20 - 938\ M\^22 + 
                    3556\ M\^24 - 1198\ M\^26 - 1912\ M\^28 + 2811\ M\^30 - 
                    4554\ M\^32 - 1134\ M\^34 + 10937\ M\^36 - 4775\ M\^38 - 
                    8939\ M\^40 + 11249\ M\^42 + 2082\ M\^44 - 9246\ M\^46 + 
                    4387\ M\^48 + 2649\ M\^50 - 4042\ M\^52 + 1437\ M\^54 + 
                    863\ M\^56 - 1246\ M\^58 + 629\ M\^60 - 154\ M\^62 + 
                    15\ M\^64)\) + 
              L\^5\ \((\(-4\)\ M\^12 + 45\ M\^14 - 220\ M\^16 + 538\ M\^18 - 
                    429\ M\^20 - 861\ M\^22 + 1658\ M\^24 + 1389\ M\^26 - 
                    4497\ M\^28 - 193\ M\^30 + 5481\ M\^32 - 2854\ M\^34 + 
                    952\ M\^36 + 8076\ M\^38 - 19979\ M\^40 - 7682\ M\^42 + 
                    38245\ M\^44 - 5805\ M\^46 - 36498\ M\^48 + 
                    30004\ M\^50 + 18482\ M\^52 - 36630\ M\^54 + 
                    5592\ M\^56 + 24148\ M\^58 - 15490\ M\^60 - 5571\ M\^62 + 
                    9765\ M\^64 - 2728\ M\^66 - 1894\ M\^68 + 2124\ M\^70 - 
                    1093\ M\^72 + 362\ M\^74 - 71\ M\^76 + 6\ M\^78)\) + 
              L\^6\ \((M\^16 - 13\ M\^18 + 74\ M\^20 - 220\ M\^22 + 
                    327\ M\^24 - 213\ M\^26 + 396\ M\^28 - 1283\ M\^30 + 
                    259\ M\^32 + 3520\ M\^34 - 2084\ M\^36 - 4174\ M\^38 + 
                    4292\ M\^40 - 9642\ M\^42 + 6385\ M\^44 + 43643\ M\^46 - 
                    37034\ M\^48 - 83178\ M\^50 + 77937\ M\^52 + 
                    83838\ M\^54 - 99318\ M\^56 - 26750\ M\^58 + 
                    91593\ M\^60 - 29475\ M\^62 - 45471\ M\^64 + 
                    47525\ M\^66 + 752\ M\^68 - 28909\ M\^70 + 17165\ M\^72 + 
                    4458\ M\^74 - 10832\ M\^76 + 4181\ M\^78 + 2064\ M\^80 - 
                    2882\ M\^82 + 1415\ M\^84 - 413\ M\^86 + 86\ M\^88 - 
                    13\ M\^90 + M\^92)\) + 
              L\^7\ \((4\ M\^26 - 55\ M\^28 + 311\ M\^30 - 866\ M\^32 + 
                    942\ M\^34 + 748\ M\^36 - 2178\ M\^38 - 1355\ M\^40 + 
                    6397\ M\^42 - 4483\ M\^44 - 4092\ M\^46 + 21608\ M\^48 - 
                    22241\ M\^50 - 53386\ M\^52 + 91893\ M\^54 + 
                    80391\ M\^56 - 173391\ M\^58 - 77250\ M\^60 + 
                    194611\ M\^62 + 26888\ M\^64 - 112371\ M\^66 + 
                    50021\ M\^68 + 2178\ M\^70 - 71095\ M\^72 + 
                    72339\ M\^74 + 36605\ M\^76 - 76648\ M\^78 + 
                    3277\ M\^80 + 42768\ M\^82 - 17474\ M\^84 - 
                    10536\ M\^86 + 10132\ M\^88 - 1071\ M\^90 - 2601\ M\^92 + 
                    2202\ M\^94 - 1050\ M\^96 + 319\ M\^98 - 55\ M\^100 + 
                    4\ M\^102)\) + 
              L\^8\ \((6\ M\^36 - 84\ M\^38 + 477\ M\^40 - 1358\ M\^42 + 
                    1779\ M\^44 + 83\ M\^46 - 3558\ M\^48 + 4693\ M\^50 + 
                    1807\ M\^52 - 19967\ M\^54 + 23898\ M\^56 + 
                    35906\ M\^58 - 85441\ M\^60 - 35563\ M\^62 + 
                    163041\ M\^64 + 6271\ M\^66 - 181044\ M\^68 + 
                    27107\ M\^70 + 85470\ M\^72 - 34176\ M\^74 + 
                    85470\ M\^76 + 27107\ M\^78 - 181044\ M\^80 + 
                    6271\ M\^82 + 163041\ M\^84 - 35563\ M\^86 - 
                    85441\ M\^88 + 35906\ M\^90 + 23898\ M\^92 - 
                    19967\ M\^94 + 1807\ M\^96 + 4693\ M\^98 - 3558\ M\^100 + 
                    83\ M\^102 + 1779\ M\^104 - 1358\ M\^106 + 477\ M\^108 - 
                    84\ M\^110 + 6\ M\^112)\) + 
              L\^9\ \((4\ M\^46 - 55\ M\^48 + 319\ M\^50 - 1050\ M\^52 + 
                    2202\ M\^54 - 2601\ M\^56 - 1071\ M\^58 + 10132\ M\^60 - 
                    10536\ M\^62 - 17474\ M\^64 + 42768\ M\^66 + 
                    3277\ M\^68 - 76648\ M\^70 + 36605\ M\^72 + 
                    72339\ M\^74 - 71095\ M\^76 + 2178\ M\^78 + 
                    50021\ M\^80 - 112371\ M\^82 + 26888\ M\^84 + 
                    194611\ M\^86 - 77250\ M\^88 - 173391\ M\^90 + 
                    80391\ M\^92 + 91893\ M\^94 - 53386\ M\^96 - 
                    22241\ M\^98 + 21608\ M\^100 - 4092\ M\^102 - 
                    4483\ M\^104 + 6397\ M\^106 - 1355\ M\^108 - 
                    2178\ M\^110 + 748\ M\^112 + 942\ M\^114 - 866\ M\^116 + 
                    311\ M\^118 - 55\ M\^120 + 4\ M\^122)\) + 
              L\^10\ \((M\^56 - 13\ M\^58 + 86\ M\^60 - 413\ M\^62 + 
                    1415\ M\^64 - 2882\ M\^66 + 2064\ M\^68 + 4181\ M\^70 - 
                    10832\ M\^72 + 4458\ M\^74 + 17165\ M\^76 - 
                    28909\ M\^78 + 752\ M\^80 + 47525\ M\^82 - 45471\ M\^84 - 
                    29475\ M\^86 + 91593\ M\^88 - 26750\ M\^90 - 
                    99318\ M\^92 + 83838\ M\^94 + 77937\ M\^96 - 
                    83178\ M\^98 - 37034\ M\^100 + 43643\ M\^102 + 
                    6385\ M\^104 - 9642\ M\^106 + 4292\ M\^108 - 
                    4174\ M\^110 - 2084\ M\^112 + 3520\ M\^114 + 
                    259\ M\^116 - 1283\ M\^118 + 396\ M\^120 - 213\ M\^122 + 
                    327\ M\^124 - 220\ M\^126 + 74\ M\^128 - 13\ M\^130 + 
                    M\^132)\) + 
              L\^11\ \((6\ M\^70 - 71\ M\^72 + 362\ M\^74 - 1093\ M\^76 + 
                    2124\ M\^78 - 1894\ M\^80 - 2728\ M\^82 + 9765\ M\^84 - 
                    5571\ M\^86 - 15490\ M\^88 + 24148\ M\^90 + 5592\ M\^92 - 
                    36630\ M\^94 + 18482\ M\^96 + 30004\ M\^98 - 
                    36498\ M\^100 - 5805\ M\^102 + 38245\ M\^104 - 
                    7682\ M\^106 - 19979\ M\^108 + 8076\ M\^110 + 
                    952\ M\^112 - 2854\ M\^114 + 5481\ M\^116 - 193\ M\^118 - 
                    4497\ M\^120 + 1389\ M\^122 + 1658\ M\^124 - 
                    861\ M\^126 - 429\ M\^128 + 538\ M\^130 - 220\ M\^132 + 
                    45\ M\^134 - 4\ M\^136)\) + 
              L\^12\ \((15\ M\^84 - 154\ M\^86 + 629\ M\^88 - 1246\ M\^90 + 
                    863\ M\^92 + 1437\ M\^94 - 4042\ M\^96 + 2649\ M\^98 + 
                    4387\ M\^100 - 9246\ M\^102 + 2082\ M\^104 + 
                    11249\ M\^106 - 8939\ M\^108 - 4775\ M\^110 + 
                    10937\ M\^112 - 1134\ M\^114 - 4554\ M\^116 + 
                    2811\ M\^118 - 1912\ M\^120 - 1198\ M\^122 + 
                    3556\ M\^124 - 938\ M\^126 - 1614\ M\^128 + 
                    1057\ M\^130 + 175\ M\^132 - 455\ M\^134 + 231\ M\^136 - 
                    57\ M\^138 + 6\ M\^140)\) + 
              L\^13\ \((20\ M\^98 - 166\ M\^100 + 511\ M\^102 - 558\ M\^104 - 
                    502\ M\^106 + 1727\ M\^108 - 661\ M\^110 - 1836\ M\^112 + 
                    1890\ M\^114 + 940\ M\^116 - 1060\ M\^118 + 30\ M\^120 + 
                    66\ M\^122 + 432\ M\^124 + 200\ M\^126 - 896\ M\^128 + 
                    238\ M\^130 + 455\ M\^132 - 320\ M\^134 - 26\ M\^136 + 
                    147\ M\^138 - 98\ M\^140 + 31\ M\^142 - 4\ M\^144)\) + 
              L\^15\ \((6\ M\^126 - 19\ M\^128 + 15\ M\^130 + 26\ M\^132 - 
                    17\ M\^134 + 3\ M\^136 + 6\ M\^138 - 8\ M\^140 + 
                    6\ M\^142 - 2\ M\^144)\) + 
              L\^14\ \((15\ M\^112 - 89\ M\^114 + 177\ M\^116 - 21\ M\^118 - 
                    359\ M\^120 + 326\ M\^122 + 335\ M\^124 - 348\ M\^126 - 
                    55\ M\^128 + 203\ M\^130 - 20\ M\^132 - 53\ M\^134 + 
                    3\ M\^136 + 2\ M\^138 + 10\ M\^140 - 14\ M\^142 + 
                    13\ M\^144 - 6\ M\^146 + M\^148)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 12]\), \(M\^16 + L\^17\ M\^86 + 
              L\ \((\(-4\)\ M\^12 + 14\ M\^14 - 17\ M\^16 - 3\ M\^18 + 
                    37\ M\^20 - 8\ M\^22 - 7\ M\^24 + 11\ M\^26 - 9\ M\^28 + 
                    3\ M\^30)\) + 
              L\^2\ \((6\ M\^8 - 42\ M\^10 + 113\ M\^12 - 102\ M\^14 - 
                    130\ M\^16 + 341\ M\^18 - 111\ M\^20 - 375\ M\^22 + 
                    428\ M\^24 + 243\ M\^26 - 251\ M\^28 - 31\ M\^30 + 
                    31\ M\^32 + 18\ M\^34 + 14\ M\^36 - 36\ M\^38 + 
                    33\ M\^40 - 16\ M\^42 + 3\ M\^44)\) + 
              L\^3\ \((\(-4\)\ M\^4 + 42\ M\^6 - 178\ M\^8 + 346\ M\^10 - 
                    153\ M\^12 - 562\ M\^14 + 857\ M\^16 + 125\ M\^18 - 
                    1065\ M\^20 + 655\ M\^22 - 277\ M\^24 - 290\ M\^26 + 
                    1746\ M\^28 - 376\ M\^30 - 529\ M\^32 + 1102\ M\^34 - 
                    1021\ M\^36 - 409\ M\^38 + 1105\ M\^40 - 292\ M\^42 - 
                    348\ M\^44 + 277\ M\^46 - 91\ M\^48 + 40\ M\^50 - 
                    35\ M\^52 + 21\ M\^54 - 7\ M\^56 + M\^58)\) + 
              L\^4\ \((1 - 14\ M\^2 + 85\ M\^4 - 284\ M\^6 + 548\ M\^8 - 
                    511\ M\^10 - 334\ M\^12 + 2000\ M\^14 - 2344\ M\^16 - 
                    2672\ M\^18 + 9079\ M\^20 - 2058\ M\^22 - 14107\ M\^24 + 
                    11364\ M\^26 + 9035\ M\^28 - 18740\ M\^30 + 6554\ M\^32 + 
                    17956\ M\^34 - 15451\ M\^36 - 6057\ M\^38 + 
                    13832\ M\^40 - 3230\ M\^42 - 6194\ M\^44 + 4055\ M\^46 + 
                    1483\ M\^48 - 1936\ M\^50 - 180\ M\^52 + 626\ M\^54 + 
                    104\ M\^56 - 420\ M\^58 + 256\ M\^60 - 78\ M\^62 + 
                    13\ M\^64 - M\^66)\) + 
              L\^5\ \((\(-3\) + 43\ M\^2 - 260\ M\^4 + 802\ M\^6 - 
                    1061\ M\^8 - 672\ M\^10 + 3704\ M\^12 - 1314\ M\^14 - 
                    8977\ M\^16 + 11444\ M\^18 + 8516\ M\^20 - 29920\ M\^22 + 
                    9350\ M\^24 + 49576\ M\^26 - 50830\ M\^28 - 
                    50264\ M\^30 + 92704\ M\^32 + 19221\ M\^34 - 
                    104314\ M\^36 + 35575\ M\^38 + 87004\ M\^40 - 
                    66320\ M\^42 - 46567\ M\^44 + 64802\ M\^46 + 
                    10146\ M\^48 - 43022\ M\^50 + 7395\ M\^52 + 
                    20568\ M\^54 - 10496\ M\^56 - 5143\ M\^58 + 5862\ M\^60 - 
                    461\ M\^62 - 1773\ M\^64 + 1201\ M\^66 - 394\ M\^68 + 
                    72\ M\^70 - 6\ M\^72)\) + 
              L\^6\ \((3 - 44\ M\^2 + 276\ M\^4 - 891\ M\^6 + 1245\ M\^8 + 
                    733\ M\^10 - 4109\ M\^12 - 594\ M\^14 + 15011\ M\^16 - 
                    8202\ M\^18 - 34409\ M\^20 + 34438\ M\^22 + 
                    60987\ M\^24 - 94703\ M\^26 - 73977\ M\^28 + 
                    189143\ M\^30 + 56842\ M\^32 - 299776\ M\^34 - 
                    3977\ M\^36 + 371998\ M\^38 - 73306\ M\^40 - 
                    354214\ M\^42 + 173862\ M\^44 + 275906\ M\^46 - 
                    229136\ M\^48 - 158193\ M\^50 + 216275\ M\^52 + 
                    45335\ M\^54 - 147639\ M\^56 + 20249\ M\^58 + 
                    68390\ M\^60 - 33871\ M\^62 - 14145\ M\^64 + 
                    17592\ M\^66 - 3044\ M\^68 - 3833\ M\^70 + 3034\ M\^72 - 
                    1058\ M\^74 + 193\ M\^76 - 15\ M\^78)\) + 
              L\^7\ \((\(-1\) + 15\ M\^2 - 97\ M\^4 + 335\ M\^6 - 635\ M\^8 + 
                    706\ M\^10 - 1179\ M\^12 + 2950\ M\^14 - 533\ M\^16 - 
                    13372\ M\^18 + 15634\ M\^20 + 33586\ M\^22 - 
                    75226\ M\^24 - 32552\ M\^26 + 193413\ M\^28 - 
                    33672\ M\^30 - 360739\ M\^32 + 200745\ M\^34 + 
                    529141\ M\^36 - 425434\ M\^38 - 660931\ M\^40 + 
                    607781\ M\^42 + 711176\ M\^44 - 667025\ M\^46 - 
                    616708\ M\^48 + 658997\ M\^50 + 433400\ M\^52 - 
                    576197\ M\^54 - 216283\ M\^56 + 429427\ M\^58 + 
                    35472\ M\^60 - 250776\ M\^62 + 52819\ M\^64 + 
                    101631\ M\^66 - 58765\ M\^68 - 16267\ M\^70 + 
                    26756\ M\^72 - 6191\ M\^74 - 5119\ M\^76 + 4484\ M\^78 - 
                    1575\ M\^80 + 277\ M\^82 - 20\ M\^84)\) + 
              L\^8\ \((\(-6\)\ M\^6 + 90\ M\^8 - 558\ M\^10 + 1754\ M\^12 - 
                    2410\ M\^14 - 1043\ M\^16 + 6672\ M\^18 + 294\ M\^20 - 
                    21100\ M\^22 + 9654\ M\^24 + 54596\ M\^26 - 
                    45976\ M\^28 - 132158\ M\^30 + 173510\ M\^32 + 
                    241841\ M\^34 - 449781\ M\^36 - 346568\ M\^38 + 
                    871505\ M\^40 + 400126\ M\^42 - 1303528\ M\^44 - 
                    410715\ M\^46 + 1545319\ M\^48 + 393558\ M\^50 - 
                    1448791\ M\^52 - 280983\ M\^54 + 1145076\ M\^56 + 
                    129400\ M\^58 - 771867\ M\^60 - 1811\ M\^62 + 
                    442789\ M\^64 - 68986\ M\^66 - 205888\ M\^68 + 
                    77083\ M\^70 + 70683\ M\^72 - 53476\ M\^74 - 
                    8414\ M\^76 + 21305\ M\^78 - 5132\ M\^80 - 4494\ M\^82 + 
                    3837\ M\^84 - 1301\ M\^86 + 219\ M\^88 - 15\ M\^90)\) + 
              L\^16\ \((3\ M\^72 - 9\ M\^74 + 11\ M\^76 - 7\ M\^78 - 
                    8\ M\^80 + 37\ M\^82 - 3\ M\^84 - 17\ M\^86 + 14\ M\^88 - 
                    4\ M\^90)\) + 
              L\^15\ \((3\ M\^58 - 16\ M\^60 + 33\ M\^62 - 36\ M\^64 + 
                    14\ M\^66 + 18\ M\^68 + 31\ M\^70 - 31\ M\^72 - 
                    251\ M\^74 + 243\ M\^76 + 428\ M\^78 - 375\ M\^80 - 
                    111\ M\^82 + 341\ M\^84 - 130\ M\^86 - 102\ M\^88 + 
                    113\ M\^90 - 42\ M\^92 + 6\ M\^94)\) + 
              L\^9\ \((\(-15\)\ M\^12 + 219\ M\^14 - 1301\ M\^16 + 
                    3837\ M\^18 - 4494\ M\^20 - 5132\ M\^22 + 21305\ M\^24 - 
                    8414\ M\^26 - 53476\ M\^28 + 70683\ M\^30 + 
                    77083\ M\^32 - 205888\ M\^34 - 68986\ M\^36 + 
                    442789\ M\^38 - 1811\ M\^40 - 771867\ M\^42 + 
                    129400\ M\^44 + 1145076\ M\^46 - 280983\ M\^48 - 
                    1448791\ M\^50 + 393558\ M\^52 + 1545319\ M\^54 - 
                    410715\ M\^56 - 1303528\ M\^58 + 400126\ M\^60 + 
                    871505\ M\^62 - 346568\ M\^64 - 449781\ M\^66 + 
                    241841\ M\^68 + 173510\ M\^70 - 132158\ M\^72 - 
                    45976\ M\^74 + 54596\ M\^76 + 9654\ M\^78 - 
                    21100\ M\^80 + 294\ M\^82 + 6672\ M\^84 - 1043\ M\^86 - 
                    2410\ M\^88 + 1754\ M\^90 - 558\ M\^92 + 90\ M\^94 - 
                    6\ M\^96)\) + 
              L\^14\ \((M\^44 - 7\ M\^46 + 21\ M\^48 - 35\ M\^50 + 
                    40\ M\^52 - 91\ M\^54 + 277\ M\^56 - 348\ M\^58 - 
                    292\ M\^60 + 1105\ M\^62 - 409\ M\^64 - 1021\ M\^66 + 
                    1102\ M\^68 - 529\ M\^70 - 376\ M\^72 + 1746\ M\^74 - 
                    290\ M\^76 - 277\ M\^78 + 655\ M\^80 - 1065\ M\^82 + 
                    125\ M\^84 + 857\ M\^86 - 562\ M\^88 - 153\ M\^90 + 
                    346\ M\^92 - 178\ M\^94 + 42\ M\^96 - 4\ M\^98)\) + 
              L\^12\ \((\(-6\)\ M\^30 + 72\ M\^32 - 394\ M\^34 + 
                    1201\ M\^36 - 1773\ M\^38 - 461\ M\^40 + 5862\ M\^42 - 
                    5143\ M\^44 - 10496\ M\^46 + 20568\ M\^48 + 7395\ M\^50 - 
                    43022\ M\^52 + 10146\ M\^54 + 64802\ M\^56 - 
                    46567\ M\^58 - 66320\ M\^60 + 87004\ M\^62 + 
                    35575\ M\^64 - 104314\ M\^66 + 19221\ M\^68 + 
                    92704\ M\^70 - 50264\ M\^72 - 50830\ M\^74 + 
                    49576\ M\^76 + 9350\ M\^78 - 29920\ M\^80 + 8516\ M\^82 + 
                    11444\ M\^84 - 8977\ M\^86 - 1314\ M\^88 + 3704\ M\^90 - 
                    672\ M\^92 - 1061\ M\^94 + 802\ M\^96 - 260\ M\^98 + 
                    43\ M\^100 - 3\ M\^102)\) + 
              L\^10\ \((\(-20\)\ M\^18 + 277\ M\^20 - 1575\ M\^22 + 
                    4484\ M\^24 - 5119\ M\^26 - 6191\ M\^28 + 26756\ M\^30 - 
                    16267\ M\^32 - 58765\ M\^34 + 101631\ M\^36 + 
                    52819\ M\^38 - 250776\ M\^40 + 35472\ M\^42 + 
                    429427\ M\^44 - 216283\ M\^46 - 576197\ M\^48 + 
                    433400\ M\^50 + 658997\ M\^52 - 616708\ M\^54 - 
                    667025\ M\^56 + 711176\ M\^58 + 607781\ M\^60 - 
                    660931\ M\^62 - 425434\ M\^64 + 529141\ M\^66 + 
                    200745\ M\^68 - 360739\ M\^70 - 33672\ M\^72 + 
                    193413\ M\^74 - 32552\ M\^76 - 75226\ M\^78 + 
                    33586\ M\^80 + 15634\ M\^82 - 13372\ M\^84 - 533\ M\^86 + 
                    2950\ M\^88 - 1179\ M\^90 + 706\ M\^92 - 635\ M\^94 + 
                    335\ M\^96 - 97\ M\^98 + 15\ M\^100 - M\^102)\) + 
              L\^13\ \((\(-M\^36\) + 13\ M\^38 - 78\ M\^40 + 256\ M\^42 - 
                    420\ M\^44 + 104\ M\^46 + 626\ M\^48 - 180\ M\^50 - 
                    1936\ M\^52 + 1483\ M\^54 + 4055\ M\^56 - 6194\ M\^58 - 
                    3230\ M\^60 + 13832\ M\^62 - 6057\ M\^64 - 15451\ M\^66 + 
                    17956\ M\^68 + 6554\ M\^70 - 18740\ M\^72 + 9035\ M\^74 + 
                    11364\ M\^76 - 14107\ M\^78 - 2058\ M\^80 + 9079\ M\^82 - 
                    2672\ M\^84 - 2344\ M\^86 + 2000\ M\^88 - 334\ M\^90 - 
                    511\ M\^92 + 548\ M\^94 - 284\ M\^96 + 85\ M\^98 - 
                    14\ M\^100 + M\^102)\) + 
              L\^11\ \((\(-15\)\ M\^24 + 193\ M\^26 - 1058\ M\^28 + 
                    3034\ M\^30 - 3833\ M\^32 - 3044\ M\^34 + 17592\ M\^36 - 
                    14145\ M\^38 - 33871\ M\^40 + 68390\ M\^42 + 
                    20249\ M\^44 - 147639\ M\^46 + 45335\ M\^48 + 
                    216275\ M\^50 - 158193\ M\^52 - 229136\ M\^54 + 
                    275906\ M\^56 + 173862\ M\^58 - 354214\ M\^60 - 
                    73306\ M\^62 + 371998\ M\^64 - 3977\ M\^66 - 
                    299776\ M\^68 + 56842\ M\^70 + 189143\ M\^72 - 
                    73977\ M\^74 - 94703\ M\^76 + 60987\ M\^78 + 
                    34438\ M\^80 - 34409\ M\^82 - 8202\ M\^84 + 
                    15011\ M\^86 - 594\ M\^88 - 4109\ M\^90 + 733\ M\^92 + 
                    1245\ M\^94 - 891\ M\^96 + 276\ M\^98 - 44\ M\^100 + 
                    3\ M\^102)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 13]\), \(1 + L\^18\ M\^148 + 
              L\ \((\(-4\) + 14\ M\^2 - 19\ M\^4 + 2\ M\^6 + 40\ M\^8 - 
                    17\ M\^10 - 7\ M\^12 + 16\ M\^14 - 10\ M\^16 + 
                    3\ M\^18)\) + 
              L\^2\ \((6 - 38\ M\^2 + 99\ M\^4 - 108\ M\^6 - 41\ M\^8 + 
                    210\ M\^10 - 179\ M\^12 - 19\ M\^14 + 287\ M\^16 - 
                    69\ M\^18 + 23\ M\^20 + 3\ M\^22 - 117\ M\^24 + 
                    121\ M\^26 + 15\ M\^28 - 80\ M\^30 + 57\ M\^32 - 
                    20\ M\^34 + 3\ M\^36)\) + 
              L\^3\ \((\(-4\) + 34\ M\^2 - 125\ M\^4 + 238\ M\^6 - 
                    242\ M\^8 + 183\ M\^10 - 110\ M\^12 - 603\ M\^14 + 
                    1882\ M\^16 - 1014\ M\^18 - 3370\ M\^20 + 4628\ M\^22 + 
                    2204\ M\^24 - 5364\ M\^26 + 1383\ M\^28 + 3074\ M\^30 - 
                    2560\ M\^32 - 104\ M\^34 + 1218\ M\^36 - 571\ M\^38 - 
                    31\ M\^40 + 127\ M\^42 - 134\ M\^44 + 145\ M\^46 - 
                    101\ M\^48 + 42\ M\^50 - 10\ M\^52 + M\^54)\) + 
              L\^4\ \((1 - 10\ M\^2 + 45\ M\^4 - 126\ M\^6 + 318\ M\^8 - 
                    850\ M\^10 + 1539\ M\^12 - 459\ M\^14 - 3697\ M\^16 + 
                    5125\ M\^18 + 3038\ M\^20 - 10180\ M\^22 + 1159\ M\^24 + 
                    10380\ M\^26 - 5013\ M\^28 - 8726\ M\^30 + 10409\ M\^32 + 
                    11240\ M\^34 - 13405\ M\^36 - 9600\ M\^38 + 
                    15542\ M\^40 + 3984\ M\^42 - 12236\ M\^44 + 1733\ M\^46 + 
                    6450\ M\^48 - 3814\ M\^50 - 1257\ M\^52 + 2199\ M\^54 - 
                    507\ M\^56 - 615\ M\^58 + 600\ M\^60 - 269\ M\^62 + 
                    73\ M\^64 - 12\ M\^66 + M\^68)\) + 
              L\^5\ \((6\ M\^6 - 74\ M\^8 + 388\ M\^10 - 1074\ M\^12 + 
                    1580\ M\^14 - 1127\ M\^16 + 521\ M\^18 + 845\ M\^20 - 
                    7898\ M\^22 + 11468\ M\^24 + 13359\ M\^26 - 
                    40842\ M\^28 - 319\ M\^30 + 69622\ M\^32 - 25802\ M\^34 - 
                    88556\ M\^36 + 42762\ M\^38 + 105799\ M\^40 - 
                    27719\ M\^42 - 113324\ M\^44 + 13085\ M\^46 + 
                    112420\ M\^48 - 12385\ M\^50 - 89083\ M\^52 + 
                    24120\ M\^54 + 50394\ M\^56 - 26746\ M\^58 - 
                    17301\ M\^60 + 18411\ M\^62 - 149\ M\^64 - 6586\ M\^66 + 
                    2691\ M\^68 + 1008\ M\^70 - 1490\ M\^72 + 746\ M\^74 - 
                    216\ M\^76 + 37\ M\^78 - 3\ M\^80)\) + 
              L\^6\ \((15\ M\^12 - 214\ M\^14 + 1266\ M\^16 - 3828\ M\^18 + 
                    5184\ M\^20 + 1877\ M\^22 - 15205\ M\^24 + 11315\ M\^26 + 
                    19460\ M\^28 - 35667\ M\^30 + 4219\ M\^32 + 
                    40576\ M\^34 - 56470\ M\^36 + 2796\ M\^38 + 
                    137673\ M\^40 - 120476\ M\^42 - 239318\ M\^44 + 
                    291284\ M\^46 + 339816\ M\^48 - 431600\ M\^50 - 
                    349404\ M\^52 + 493278\ M\^54 + 266409\ M\^56 - 
                    430359\ M\^58 - 131179\ M\^60 + 288753\ M\^62 + 
                    31560\ M\^64 - 148206\ M\^66 + 8615\ M\^68 + 
                    57307\ M\^70 - 10140\ M\^72 - 19717\ M\^74 + 
                    7027\ M\^76 + 5051\ M\^78 - 3163\ M\^80 - 1155\ M\^82 + 
                    1927\ M\^84 - 973\ M\^86 + 269\ M\^88 - 42\ M\^90 + 
                    3\ M\^92)\) + 
              L\^7\ \((20\ M\^18 - 317\ M\^20 + 2076\ M\^22 - 6989\ M\^24 + 
                    10747\ M\^26 + 4111\ M\^28 - 42586\ M\^30 + 
                    42101\ M\^32 + 72673\ M\^34 - 177721\ M\^36 - 
                    4701\ M\^38 + 360189\ M\^40 - 252336\ M\^42 - 
                    460104\ M\^44 + 717489\ M\^46 + 325422\ M\^48 - 
                    1263900\ M\^50 + 31659\ M\^52 + 1694766\ M\^54 - 
                    433054\ M\^56 - 1805129\ M\^58 + 731767\ M\^60 + 
                    1598631\ M\^62 - 804126\ M\^64 - 1144932\ M\^66 + 
                    697698\ M\^68 + 651047\ M\^70 - 494591\ M\^72 - 
                    268096\ M\^74 + 280236\ M\^76 + 61706\ M\^78 - 
                    121522\ M\^80 + 10096\ M\^82 + 30873\ M\^84 - 
                    10460\ M\^86 - 2673\ M\^88 + 2500\ M\^90 - 1649\ M\^92 + 
                    1712\ M\^94 - 1192\ M\^96 + 485\ M\^98 - 117\ M\^100 + 
                    16\ M\^102 - M\^104)\) + 
              L\^8\ \((15\ M\^24 - 258\ M\^26 + 1860\ M\^28 - 7058\ M\^30 + 
                    13227\ M\^32 - 981\ M\^34 - 47548\ M\^36 + 69564\ M\^38 + 
                    69789\ M\^40 - 273708\ M\^42 + 63034\ M\^44 + 
                    594049\ M\^46 - 520253\ M\^48 - 900017\ M\^50 + 
                    1402529\ M\^52 + 981125\ M\^54 - 2577094\ M\^56 - 
                    756438\ M\^58 + 3714387\ M\^60 + 304367\ M\^62 - 
                    4384016\ M\^64 + 238659\ M\^66 + 4353506\ M\^68 - 
                    655821\ M\^70 - 3564021\ M\^72 + 875545\ M\^74 + 
                    2384516\ M\^76 - 833314\ M\^78 - 1259196\ M\^80 + 
                    596314\ M\^82 + 508080\ M\^84 - 324390\ M\^86 - 
                    148218\ M\^88 + 130458\ M\^90 + 34542\ M\^92 - 
                    46249\ M\^94 - 3561\ M\^96 + 14068\ M\^98 - 
                    1357\ M\^100 - 4783\ M\^102 + 3325\ M\^104 - 
                    1110\ M\^106 + 211\ M\^108 - 22\ M\^110 + M\^112)\) + 
              L\^9\ \((6\ M\^30 - 112\ M\^32 + 902\ M\^34 - 3939\ M\^36 + 
                    9210\ M\^38 - 6429\ M\^40 - 22163\ M\^42 + 52290\ M\^44 + 
                    16369\ M\^46 - 181843\ M\^48 + 104091\ M\^50 + 
                    407315\ M\^52 - 477357\ M\^54 - 713085\ M\^56 + 
                    1291165\ M\^58 + 975724\ M\^60 - 2599503\ M\^62 - 
                    1046810\ M\^64 + 4192754\ M\^66 + 812818\ M\^68 - 
                    5537535\ M\^70 - 295929\ M\^72 + 6092742\ M\^74 - 
                    295929\ M\^76 - 5537535\ M\^78 + 812818\ M\^80 + 
                    4192754\ M\^82 - 1046810\ M\^84 - 2599503\ M\^86 + 
                    975724\ M\^88 + 1291165\ M\^90 - 713085\ M\^92 - 
                    477357\ M\^94 + 407315\ M\^96 + 104091\ M\^98 - 
                    181843\ M\^100 + 16369\ M\^102 + 52290\ M\^104 - 
                    22163\ M\^106 - 6429\ M\^108 + 9210\ M\^110 - 
                    3939\ M\^112 + 902\ M\^114 - 112\ M\^116 + 6\ M\^118)\) + 
              L\^10\ \((M\^36 - 22\ M\^38 + 211\ M\^40 - 1110\ M\^42 + 
                    3325\ M\^44 - 4783\ M\^46 - 1357\ M\^48 + 14068\ M\^50 - 
                    3561\ M\^52 - 46249\ M\^54 + 34542\ M\^56 + 
                    130458\ M\^58 - 148218\ M\^60 - 324390\ M\^62 + 
                    508080\ M\^64 + 596314\ M\^66 - 1259196\ M\^68 - 
                    833314\ M\^70 + 2384516\ M\^72 + 875545\ M\^74 - 
                    3564021\ M\^76 - 655821\ M\^78 + 4353506\ M\^80 + 
                    238659\ M\^82 - 4384016\ M\^84 + 304367\ M\^86 + 
                    3714387\ M\^88 - 756438\ M\^90 - 2577094\ M\^92 + 
                    981125\ M\^94 + 1402529\ M\^96 - 900017\ M\^98 - 
                    520253\ M\^100 + 594049\ M\^102 + 63034\ M\^104 - 
                    273708\ M\^106 + 69789\ M\^108 + 69564\ M\^110 - 
                    47548\ M\^112 - 981\ M\^114 + 13227\ M\^116 - 
                    7058\ M\^118 + 1860\ M\^120 - 258\ M\^122 + 
                    15\ M\^124)\) + 
              L\^11\ \((\(-M\^44\) + 16\ M\^46 - 117\ M\^48 + 485\ M\^50 - 
                    1192\ M\^52 + 1712\ M\^54 - 1649\ M\^56 + 2500\ M\^58 - 
                    2673\ M\^60 - 10460\ M\^62 + 30873\ M\^64 + 
                    10096\ M\^66 - 121522\ M\^68 + 61706\ M\^70 + 
                    280236\ M\^72 - 268096\ M\^74 - 494591\ M\^76 + 
                    651047\ M\^78 + 697698\ M\^80 - 1144932\ M\^82 - 
                    804126\ M\^84 + 1598631\ M\^86 + 731767\ M\^88 - 
                    1805129\ M\^90 - 433054\ M\^92 + 1694766\ M\^94 + 
                    31659\ M\^96 - 1263900\ M\^98 + 325422\ M\^100 + 
                    717489\ M\^102 - 460104\ M\^104 - 252336\ M\^106 + 
                    360189\ M\^108 - 4701\ M\^110 - 177721\ M\^112 + 
                    72673\ M\^114 + 42101\ M\^116 - 42586\ M\^118 + 
                    4111\ M\^120 + 10747\ M\^122 - 6989\ M\^124 + 
                    2076\ M\^126 - 317\ M\^128 + 20\ M\^130)\) + 
              L\^12\ \((3\ M\^56 - 42\ M\^58 + 269\ M\^60 - 973\ M\^62 + 
                    1927\ M\^64 - 1155\ M\^66 - 3163\ M\^68 + 5051\ M\^70 + 
                    7027\ M\^72 - 19717\ M\^74 - 10140\ M\^76 + 
                    57307\ M\^78 + 8615\ M\^80 - 148206\ M\^82 + 
                    31560\ M\^84 + 288753\ M\^86 - 131179\ M\^88 - 
                    430359\ M\^90 + 266409\ M\^92 + 493278\ M\^94 - 
                    349404\ M\^96 - 431600\ M\^98 + 339816\ M\^100 + 
                    291284\ M\^102 - 239318\ M\^104 - 120476\ M\^106 + 
                    137673\ M\^108 + 2796\ M\^110 - 56470\ M\^112 + 
                    40576\ M\^114 + 4219\ M\^116 - 35667\ M\^118 + 
                    19460\ M\^120 + 11315\ M\^122 - 15205\ M\^124 + 
                    1877\ M\^126 + 5184\ M\^128 - 3828\ M\^130 + 
                    1266\ M\^132 - 214\ M\^134 + 15\ M\^136)\) + 
              L\^13\ \((\(-3\)\ M\^68 + 37\ M\^70 - 216\ M\^72 + 746\ M\^74 - 
                    1490\ M\^76 + 1008\ M\^78 + 2691\ M\^80 - 6586\ M\^82 - 
                    149\ M\^84 + 18411\ M\^86 - 17301\ M\^88 - 26746\ M\^90 + 
                    50394\ M\^92 + 24120\ M\^94 - 89083\ M\^96 - 
                    12385\ M\^98 + 112420\ M\^100 + 13085\ M\^102 - 
                    113324\ M\^104 - 27719\ M\^106 + 105799\ M\^108 + 
                    42762\ M\^110 - 88556\ M\^112 - 25802\ M\^114 + 
                    69622\ M\^116 - 319\ M\^118 - 40842\ M\^120 + 
                    13359\ M\^122 + 11468\ M\^124 - 7898\ M\^126 + 
                    845\ M\^128 + 521\ M\^130 - 1127\ M\^132 + 1580\ M\^134 - 
                    1074\ M\^136 + 388\ M\^138 - 74\ M\^140 + 6\ M\^142)\) + 
              L\^17\ \((3\ M\^130 - 10\ M\^132 + 16\ M\^134 - 7\ M\^136 - 
                    17\ M\^138 + 40\ M\^140 + 2\ M\^142 - 19\ M\^144 + 
                    14\ M\^146 - 4\ M\^148)\) + 
              L\^15\ \((M\^94 - 10\ M\^96 + 42\ M\^98 - 101\ M\^100 + 
                    145\ M\^102 - 134\ M\^104 + 127\ M\^106 - 31\ M\^108 - 
                    571\ M\^110 + 1218\ M\^112 - 104\ M\^114 - 2560\ M\^116 + 
                    3074\ M\^118 + 1383\ M\^120 - 5364\ M\^122 + 
                    2204\ M\^124 + 4628\ M\^126 - 3370\ M\^128 - 
                    1014\ M\^130 + 1882\ M\^132 - 603\ M\^134 - 110\ M\^136 + 
                    183\ M\^138 - 242\ M\^140 + 238\ M\^142 - 125\ M\^144 + 
                    34\ M\^146 - 4\ M\^148)\) + 
              L\^14\ \((M\^80 - 12\ M\^82 + 73\ M\^84 - 269\ M\^86 + 
                    600\ M\^88 - 615\ M\^90 - 507\ M\^92 + 2199\ M\^94 - 
                    1257\ M\^96 - 3814\ M\^98 + 6450\ M\^100 + 1733\ M\^102 - 
                    12236\ M\^104 + 3984\ M\^106 + 15542\ M\^108 - 
                    9600\ M\^110 - 13405\ M\^112 + 11240\ M\^114 + 
                    10409\ M\^116 - 8726\ M\^118 - 5013\ M\^120 + 
                    10380\ M\^122 + 1159\ M\^124 - 10180\ M\^126 + 
                    3038\ M\^128 + 5125\ M\^130 - 3697\ M\^132 - 
                    459\ M\^134 + 1539\ M\^136 - 850\ M\^138 + 318\ M\^140 - 
                    126\ M\^142 + 45\ M\^144 - 10\ M\^146 + M\^148)\) + 
              L\^16\ \((3\ M\^112 - 20\ M\^114 + 57\ M\^116 - 80\ M\^118 + 
                    15\ M\^120 + 121\ M\^122 - 117\ M\^124 + 3\ M\^126 + 
                    23\ M\^128 - 69\ M\^130 + 287\ M\^132 - 19\ M\^134 - 
                    179\ M\^136 + 210\ M\^138 - 41\ M\^140 - 108\ M\^142 + 
                    99\ M\^144 - 38\ M\^146 + 6\ M\^148)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 14]\), \(M\^48 + L\^18\ M\^48 + 
              L\^17\ \((\(-2\)\ M\^36 + 7\ M\^38 - 11\ M\^40 + 9\ M\^42 + 
                    2\ M\^44 - 22\ M\^46 + 37\ M\^48 + 16\ M\^50 - 
                    26\ M\^52 + 8\ M\^54)\) + 
              L\ \((8\ M\^42 - 26\ M\^44 + 16\ M\^46 + 37\ M\^48 - 
                    22\ M\^50 + 2\ M\^52 + 9\ M\^54 - 11\ M\^56 + 7\ M\^58 - 
                    2\ M\^60)\) + 
              L\^16\ \((M\^24 - 6\ M\^26 + 15\ M\^28 - 22\ M\^30 + 
                    26\ M\^32 - 34\ M\^34 + 66\ M\^36 - 70\ M\^38 - 
                    174\ M\^40 + 488\ M\^42 - 141\ M\^44 - 764\ M\^46 + 
                    843\ M\^48 + 516\ M\^50 - 832\ M\^52 + 64\ M\^54 + 
                    321\ M\^56 - 172\ M\^58 + 28\ M\^60)\) + 
              L\^15\ \((\(-M\^16\) + 8\ M\^18 - 28\ M\^20 + 69\ M\^22 - 
                    139\ M\^24 + 159\ M\^26 + 32\ M\^28 - 286\ M\^30 + 
                    64\ M\^32 + 594\ M\^34 - 49\ M\^36 - 2094\ M\^38 + 
                    916\ M\^40 + 5128\ M\^42 - 5176\ M\^44 - 6306\ M\^46 + 
                    11177\ M\^48 + 3630\ M\^50 - 11495\ M\^52 + 1685\ M\^54 + 
                    6229\ M\^56 - 3279\ M\^58 - 1044\ M\^60 + 1452\ M\^62 - 
                    486\ M\^64 + 56\ M\^66)\) + 
              L\^2\ \((28\ M\^36 - 172\ M\^38 + 321\ M\^40 + 64\ M\^42 - 
                    832\ M\^44 + 516\ M\^46 + 843\ M\^48 - 764\ M\^50 - 
                    141\ M\^52 + 488\ M\^54 - 174\ M\^56 - 70\ M\^58 + 
                    66\ M\^60 - 34\ M\^62 + 26\ M\^64 - 22\ M\^66 + 
                    15\ M\^68 - 6\ M\^70 + M\^72)\) + 
              L\^14\ \((4\ M\^12 - 41\ M\^14 + 183\ M\^16 - 451\ M\^18 + 
                    579\ M\^20 + 32\ M\^22 - 1538\ M\^24 + 2159\ M\^26 + 
                    1165\ M\^28 - 6951\ M\^30 + 4724\ M\^32 + 10954\ M\^34 - 
                    16354\ M\^36 - 12228\ M\^38 + 33260\ M\^40 + 
                    7714\ M\^42 - 53831\ M\^44 + 3953\ M\^46 + 71203\ M\^48 - 
                    18533\ M\^50 - 64627\ M\^52 + 33658\ M\^54 + 
                    37932\ M\^56 - 33975\ M\^58 - 8828\ M\^60 + 
                    18825\ M\^62 - 3910\ M\^64 - 4356\ M\^66 + 3028\ M\^68 - 
                    760\ M\^70 + 70\ M\^72)\) + 
              L\^13\ \((\(-6\)\ M\^8 + 75\ M\^10 - 408\ M\^12 + 1188\ M\^14 - 
                    1656\ M\^16 - 386\ M\^18 + 5355\ M\^20 - 7063\ M\^22 - 
                    3637\ M\^24 + 23941\ M\^26 - 24152\ M\^28 - 
                    28971\ M\^30 + 81168\ M\^32 - 4155\ M\^34 - 
                    141583\ M\^36 + 83433\ M\^38 + 163646\ M\^40 - 
                    178913\ M\^42 - 143490\ M\^44 + 244899\ M\^46 + 
                    106755\ M\^48 - 244761\ M\^50 - 54529\ M\^52 + 
                    208639\ M\^54 + 8093\ M\^56 - 148841\ M\^58 + 
                    24778\ M\^60 + 82330\ M\^62 - 35879\ M\^64 - 
                    26896\ M\^66 + 23387\ M\^68 + 541\ M\^70 - 7132\ M\^72 + 
                    3452\ M\^74 - 710\ M\^76 + 56\ M\^78)\) + 
              L\^3\ \((56\ M\^30 - 486\ M\^32 + 1452\ M\^34 - 1044\ M\^36 - 
                    3279\ M\^38 + 6229\ M\^40 + 1685\ M\^42 - 11495\ M\^44 + 
                    3630\ M\^46 + 11177\ M\^48 - 6306\ M\^50 - 5176\ M\^52 + 
                    5128\ M\^54 + 916\ M\^56 - 2094\ M\^58 - 49\ M\^60 + 
                    594\ M\^62 + 64\ M\^64 - 286\ M\^66 + 32\ M\^68 + 
                    159\ M\^70 - 139\ M\^72 + 69\ M\^74 - 28\ M\^76 + 
                    8\ M\^78 - M\^80)\) + 
              L\^4\ \((70\ M\^24 - 760\ M\^26 + 3028\ M\^28 - 4356\ M\^30 - 
                    3910\ M\^32 + 18825\ M\^34 - 8828\ M\^36 - 33975\ M\^38 + 
                    37932\ M\^40 + 33658\ M\^42 - 64627\ M\^44 - 
                    18533\ M\^46 + 71203\ M\^48 + 3953\ M\^50 - 
                    53831\ M\^52 + 7714\ M\^54 + 33260\ M\^56 - 
                    12228\ M\^58 - 16354\ M\^60 + 10954\ M\^62 + 
                    4724\ M\^64 - 6951\ M\^66 + 1165\ M\^68 + 2159\ M\^70 - 
                    1538\ M\^72 + 32\ M\^74 + 579\ M\^76 - 451\ M\^78 + 
                    183\ M\^80 - 41\ M\^82 + 4\ M\^84)\) + 
              L\^12\ \((4\ M\^4 - 59\ M\^6 + 377\ M\^8 - 1311\ M\^10 + 
                    2447\ M\^12 - 1333\ M\^14 - 4446\ M\^16 + 11122\ M\^18 - 
                    6005\ M\^20 - 24001\ M\^22 + 57003\ M\^24 - 5199\ M\^26 - 
                    142676\ M\^28 + 125710\ M\^30 + 207112\ M\^32 - 
                    347768\ M\^34 - 143606\ M\^36 + 585894\ M\^38 - 
                    92425\ M\^40 - 718197\ M\^42 + 398335\ M\^44 + 
                    655279\ M\^46 - 585161\ M\^48 - 415140\ M\^50 + 
                    599400\ M\^52 + 180433\ M\^54 - 453985\ M\^56 - 
                    41783\ M\^58 + 271720\ M\^60 - 6378\ M\^62 - 
                    137110\ M\^64 + 12244\ M\^66 + 64757\ M\^68 - 
                    15101\ M\^70 - 24006\ M\^72 + 11860\ M\^74 + 
                    4572\ M\^76 - 5876\ M\^78 + 2229\ M\^80 - 396\ M\^82 + 
                    28\ M\^84)\) + 
              L\^5\ \((56\ M\^18 - 710\ M\^20 + 3452\ M\^22 - 7132\ M\^24 + 
                    541\ M\^26 + 23387\ M\^28 - 26896\ M\^30 - 35879\ M\^32 + 
                    82330\ M\^34 + 24778\ M\^36 - 148841\ M\^38 + 
                    8093\ M\^40 + 208639\ M\^42 - 54529\ M\^44 - 
                    244761\ M\^46 + 106755\ M\^48 + 244899\ M\^50 - 
                    143490\ M\^52 - 178913\ M\^54 + 163646\ M\^56 + 
                    83433\ M\^58 - 141583\ M\^60 - 4155\ M\^62 + 
                    81168\ M\^64 - 28971\ M\^66 - 24152\ M\^68 + 
                    23941\ M\^70 - 3637\ M\^72 - 7063\ M\^74 + 5355\ M\^76 - 
                    386\ M\^78 - 1656\ M\^80 + 1188\ M\^82 - 408\ M\^84 + 
                    75\ M\^86 - 6\ M\^88)\) + 
              L\^11\ \((\(-1\) + 17\ M\^2 - 128\ M\^4 + 561\ M\^6 - 
                    1586\ M\^8 + 2891\ M\^10 - 2220\ M\^12 - 5221\ M\^14 + 
                    17477\ M\^16 - 5694\ M\^18 - 55416\ M\^20 + 
                    76452\ M\^22 + 81811\ M\^24 - 237024\ M\^26 - 
                    23524\ M\^28 + 465423\ M\^30 - 235445\ M\^32 - 
                    608447\ M\^34 + 720990\ M\^36 + 491861\ M\^38 - 
                    1286816\ M\^40 - 66437\ M\^42 + 1613076\ M\^44 - 
                    524581\ M\^46 - 1480310\ M\^48 + 1049449\ M\^50 + 
                    1029162\ M\^52 - 1187556\ M\^54 - 474763\ M\^56 + 
                    958242\ M\^58 + 72651\ M\^60 - 570049\ M\^62 + 
                    83638\ M\^64 + 247416\ M\^66 - 81630\ M\^68 - 
                    71110\ M\^70 + 31806\ M\^72 + 16028\ M\^74 - 
                    6555\ M\^76 - 4636\ M\^78 + 343\ M\^80 + 3463\ M\^82 - 
                    2438\ M\^84 + 768\ M\^86 - 122\ M\^88 + 8\ M\^90)\) + 
              L\^6\ \((28\ M\^12 - 396\ M\^14 + 2229\ M\^16 - 5876\ M\^18 + 
                    4572\ M\^20 + 11860\ M\^22 - 24006\ M\^24 - 
                    15101\ M\^26 + 64757\ M\^28 + 12244\ M\^30 - 
                    137110\ M\^32 - 6378\ M\^34 + 271720\ M\^36 - 
                    41783\ M\^38 - 453985\ M\^40 + 180433\ M\^42 + 
                    599400\ M\^44 - 415140\ M\^46 - 585161\ M\^48 + 
                    655279\ M\^50 + 398335\ M\^52 - 718197\ M\^54 - 
                    92425\ M\^56 + 585894\ M\^58 - 143606\ M\^60 - 
                    347768\ M\^62 + 207112\ M\^64 + 125710\ M\^66 - 
                    142676\ M\^68 - 5199\ M\^70 + 57003\ M\^72 - 
                    24001\ M\^74 - 6005\ M\^76 + 11122\ M\^78 - 4446\ M\^80 - 
                    1333\ M\^82 + 2447\ M\^84 - 1311\ M\^86 + 377\ M\^88 - 
                    59\ M\^90 + 4\ M\^92)\) + 
              L\^9\ \((\(-3\) + 49\ M\^2 - 345\ M\^4 + 1298\ M\^6 - 
                    2485\ M\^8 + 994\ M\^10 + 4227\ M\^12 - 1371\ M\^14 - 
                    19703\ M\^16 + 19501\ M\^18 + 50330\ M\^20 - 
                    83806\ M\^22 - 92134\ M\^24 + 265318\ M\^26 + 
                    53028\ M\^28 - 583643\ M\^30 + 207978\ M\^32 + 
                    974540\ M\^34 - 847583\ M\^36 - 1198605\ M\^38 + 
                    1815346\ M\^40 + 1005105\ M\^42 - 2779959\ M\^44 - 
                    375070\ M\^46 + 3222606\ M\^48 - 375070\ M\^50 - 
                    2779959\ M\^52 + 1005105\ M\^54 + 1815346\ M\^56 - 
                    1198605\ M\^58 - 847583\ M\^60 + 974540\ M\^62 + 
                    207978\ M\^64 - 583643\ M\^66 + 53028\ M\^68 + 
                    265318\ M\^70 - 92134\ M\^72 - 83806\ M\^74 + 
                    50330\ M\^76 + 19501\ M\^78 - 19703\ M\^80 - 
                    1371\ M\^82 + 4227\ M\^84 + 994\ M\^86 - 2485\ M\^88 + 
                    1298\ M\^90 - 345\ M\^92 + 49\ M\^94 - 3\ M\^96)\) + 
              L\^7\ \((8\ M\^6 - 122\ M\^8 + 768\ M\^10 - 2438\ M\^12 + 
                    3463\ M\^14 + 343\ M\^16 - 4636\ M\^18 - 6555\ M\^20 + 
                    16028\ M\^22 + 31806\ M\^24 - 71110\ M\^26 - 
                    81630\ M\^28 + 247416\ M\^30 + 83638\ M\^32 - 
                    570049\ M\^34 + 72651\ M\^36 + 958242\ M\^38 - 
                    474763\ M\^40 - 1187556\ M\^42 + 1029162\ M\^44 + 
                    1049449\ M\^46 - 1480310\ M\^48 - 524581\ M\^50 + 
                    1613076\ M\^52 - 66437\ M\^54 - 1286816\ M\^56 + 
                    491861\ M\^58 + 720990\ M\^60 - 608447\ M\^62 - 
                    235445\ M\^64 + 465423\ M\^66 - 23524\ M\^68 - 
                    237024\ M\^70 + 81811\ M\^72 + 76452\ M\^74 - 
                    55416\ M\^76 - 5694\ M\^78 + 17477\ M\^80 - 5221\ M\^82 - 
                    2220\ M\^84 + 2891\ M\^86 - 1586\ M\^88 + 561\ M\^90 - 
                    128\ M\^92 + 17\ M\^94 - M\^96)\) + 
              L\^10\ \((3 - 50\ M\^2 + 359\ M\^4 - 1398\ M\^6 + 2910\ M\^8 - 
                    1820\ M\^10 - 5278\ M\^12 + 10305\ M\^14 + 7159\ M\^16 - 
                    41492\ M\^18 + 27615\ M\^20 + 82406\ M\^22 - 
                    154117\ M\^24 - 75611\ M\^26 + 447725\ M\^28 - 
                    144327\ M\^30 - 859183\ M\^32 + 701848\ M\^34 + 
                    1195792\ M\^36 - 1531175\ M\^38 - 1162329\ M\^40 + 
                    2289145\ M\^42 + 598927\ M\^44 - 2626485\ M\^46 + 
                    308907\ M\^48 + 2433260\ M\^50 - 1016520\ M\^52 - 
                    1729167\ M\^54 + 1300132\ M\^56 + 894011\ M\^58 - 
                    1144615\ M\^60 - 268929\ M\^62 + 756635\ M\^64 - 
                    40346\ M\^66 - 373194\ M\^68 + 103537\ M\^70 + 
                    135269\ M\^72 - 72547\ M\^74 - 26791\ M\^76 + 
                    27474\ M\^78 - 136\ M\^80 - 6634\ M\^82 + 3454\ M\^84 - 
                    1577\ M\^86 + 923\ M\^88 - 412\ M\^90 + 110\ M\^92 - 
                    16\ M\^94 + M\^96)\) + 
              L\^8\ \((1 - 16\ M\^2 + 110\ M\^4 - 412\ M\^6 + 923\ M\^8 - 
                    1577\ M\^10 + 3454\ M\^12 - 6634\ M\^14 - 136\ M\^16 + 
                    27474\ M\^18 - 26791\ M\^20 - 72547\ M\^22 + 
                    135269\ M\^24 + 103537\ M\^26 - 373194\ M\^28 - 
                    40346\ M\^30 + 756635\ M\^32 - 268929\ M\^34 - 
                    1144615\ M\^36 + 894011\ M\^38 + 1300132\ M\^40 - 
                    1729167\ M\^42 - 1016520\ M\^44 + 2433260\ M\^46 + 
                    308907\ M\^48 - 2626485\ M\^50 + 598927\ M\^52 + 
                    2289145\ M\^54 - 1162329\ M\^56 - 1531175\ M\^58 + 
                    1195792\ M\^60 + 701848\ M\^62 - 859183\ M\^64 - 
                    144327\ M\^66 + 447725\ M\^68 - 75611\ M\^70 - 
                    154117\ M\^72 + 82406\ M\^74 + 27615\ M\^76 - 
                    41492\ M\^78 + 7159\ M\^80 + 10305\ M\^82 - 5278\ M\^84 - 
                    1820\ M\^86 + 2910\ M\^88 - 1398\ M\^90 + 359\ M\^92 - 
                    50\ M\^94 + 3\ M\^96)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 15]\), \(M\^16 + L\^19\ M\^94 + 
              L\ \((\(-4\)\ M\^12 + 16\ M\^14 - 24\ M\^16 + M\^18 + 
                    49\ M\^20 - 19\ M\^22 - 10\ M\^24 + 18\ M\^26 - 
                    11\ M\^28 + 3\ M\^30)\) + 
              L\^2\ \((6\ M\^8 - 48\ M\^10 + 150\ M\^12 - 174\ M\^14 - 
                    116\ M\^16 + 436\ M\^18 - 194\ M\^20 - 326\ M\^22 + 
                    358\ M\^24 + 213\ M\^26 + 54\ M\^28 - 202\ M\^30 - 
                    185\ M\^32 + 247\ M\^34 + 22\ M\^36 - 124\ M\^38 + 
                    73\ M\^40 - 22\ M\^42 + 3\ M\^44)\) + 
              L\^3\ \((\(-4\)\ M\^4 + 48\ M\^6 - 232\ M\^8 + 532\ M\^10 - 
                    479\ M\^12 - 271\ M\^14 + 1095\ M\^16 - 1273\ M\^18 + 
                    265\ M\^20 + 2155\ M\^22 - 2950\ M\^24 - 938\ M\^26 + 
                    3478\ M\^28 + 1046\ M\^30 - 302\ M\^32 - 1863\ M\^34 - 
                    1104\ M\^36 + 2438\ M\^38 + 79\ M\^40 - 1239\ M\^42 + 
                    545\ M\^44 + 16\ M\^46 - 215\ M\^48 + 260\ M\^50 - 
                    165\ M\^52 + 57\ M\^54 - 11\ M\^56 + M\^58)\) + 
              L\^4\ \((1 - 16\ M\^2 + 110\ M\^4 - 426\ M\^6 + 1045\ M\^8 - 
                    1623\ M\^10 + 899\ M\^12 + 2659\ M\^14 - 6410\ M\^16 + 
                    2448\ M\^18 + 8525\ M\^20 - 12719\ M\^22 + 3901\ M\^24 + 
                    14812\ M\^26 - 27000\ M\^28 - 4399\ M\^30 + 
                    44351\ M\^32 - 5568\ M\^34 - 37087\ M\^36 + 
                    12560\ M\^38 + 20756\ M\^40 - 11979\ M\^42 - 
                    7628\ M\^44 + 8270\ M\^46 + 1538\ M\^48 - 5438\ M\^50 + 
                    1510\ M\^52 + 2279\ M\^54 - 1728\ M\^56 - 298\ M\^58 + 
                    933\ M\^60 - 536\ M\^62 + 158\ M\^64 - 26\ M\^66 + 
                    2\ M\^68)\) + 
              L\^5\ \((\(-4\) + 67\ M\^2 - 460\ M\^4 + 1630\ M\^6 - 
                    2948\ M\^8 + 1630\ M\^10 + 2937\ M\^12 - 4806\ M\^14 + 
                    1885\ M\^16 - 1124\ M\^18 - 601\ M\^20 + 12816\ M\^22 - 
                    31851\ M\^24 + 21312\ M\^26 + 80944\ M\^28 - 
                    138537\ M\^30 - 97834\ M\^32 + 281746\ M\^34 + 
                    48908\ M\^36 - 320860\ M\^38 + 46889\ M\^40 + 
                    244854\ M\^42 - 92414\ M\^44 - 120292\ M\^46 + 
                    70841\ M\^48 + 37880\ M\^50 - 31724\ M\^52 - 
                    10277\ M\^54 + 10969\ M\^56 + 6593\ M\^58 - 8044\ M\^60 - 
                    1602\ M\^62 + 4880\ M\^64 - 1107\ M\^66 - 1737\ M\^68 + 
                    1567\ M\^70 - 619\ M\^72 + 137\ M\^74 - 17\ M\^76 + 
                    M\^78)\) + 
              L\^6\ \((6 - 105\ M\^2 + 759\ M\^4 - 2862\ M\^6 + 5558\ M\^8 - 
                    3117\ M\^10 - 7965\ M\^12 + 13930\ M\^14 + 1443\ M\^16 - 
                    20688\ M\^18 + 14011\ M\^20 + 15188\ M\^22 - 
                    17803\ M\^24 - 70527\ M\^26 + 65038\ M\^28 + 
                    285670\ M\^30 - 268331\ M\^32 - 655961\ M\^34 + 
                    670844\ M\^36 + 943008\ M\^38 - 1110681\ M\^40 - 
                    860704\ M\^42 + 1367164\ M\^44 + 471001\ M\^46 - 
                    1221325\ M\^48 - 20819\ M\^50 + 784591\ M\^52 - 
                    224444\ M\^54 - 328714\ M\^56 + 215932\ M\^58 + 
                    61328\ M\^60 - 103392\ M\^62 + 17680\ M\^64 + 
                    19558\ M\^66 - 8028\ M\^68 + 608\ M\^70 - 1692\ M\^72 + 
                    110\ M\^74 + 2261\ M\^76 - 2145\ M\^78 + 955\ M\^80 - 
                    239\ M\^82 + 33\ M\^84 - 2\ M\^86)\) + 
              L\^7\ \((\(-4\) + 73\ M\^2 - 562\ M\^4 + 2367\ M\^6 - 
                    5848\ M\^8 + 7970\ M\^10 - 2017\ M\^12 - 17332\ M\^14 + 
                    40700\ M\^16 - 12616\ M\^18 - 122212\ M\^20 + 
                    175296\ M\^22 + 182132\ M\^24 - 496593\ M\^26 - 
                    189319\ M\^28 + 911488\ M\^30 + 206815\ M\^32 - 
                    1244347\ M\^34 - 403011\ M\^36 + 1393253\ M\^38 + 
                    794751\ M\^40 - 1425545\ M\^42 - 1192496\ M\^44 + 
                    1506240\ M\^46 + 1411865\ M\^48 - 1500290\ M\^50 - 
                    1219068\ M\^52 + 1365241\ M\^54 + 746676\ M\^56 - 
                    1061716\ M\^58 - 272398\ M\^60 + 677249\ M\^62 - 
                    23704\ M\^64 - 327603\ M\^66 + 110419\ M\^68 + 
                    104991\ M\^70 - 84447\ M\^72 - 3160\ M\^74 + 
                    28159\ M\^76 - 11436\ M\^78 - 2362\ M\^80 + 5067\ M\^82 - 
                    3486\ M\^84 + 1662\ M\^86 - 566\ M\^88 + 128\ M\^90 - 
                    17\ M\^92 + M\^94)\) + 
              L\^18\ \((3\ M\^80 - 11\ M\^82 + 18\ M\^84 - 10\ M\^86 - 
                    19\ M\^88 + 49\ M\^90 + M\^92 - 24\ M\^94 + 16\ M\^96 - 
                    4\ M\^98)\) + 
              L\^8\ \((1 - 19\ M\^2 + 161\ M\^4 - 833\ M\^6 + 3033\ M\^8 - 
                    8004\ M\^10 + 12885\ M\^12 - 658\ M\^14 - 46726\ M\^16 + 
                    69313\ M\^18 + 76337\ M\^20 - 284974\ M\^22 + 
                    28851\ M\^24 + 650382\ M\^26 - 403196\ M\^28 - 
                    1104621\ M\^30 + 1156770\ M\^32 + 1490127\ M\^34 - 
                    2230105\ M\^36 - 1713132\ M\^38 + 3474559\ M\^40 + 
                    1727799\ M\^42 - 4620101\ M\^44 - 1581101\ M\^46 + 
                    5269679\ M\^48 + 1320176\ M\^50 - 4935663\ M\^52 - 
                    785504\ M\^54 + 3837308\ M\^56 + 226766\ M\^58 - 
                    2411054\ M\^60 + 122142\ M\^62 + 1195454\ M\^64 - 
                    208830\ M\^66 - 446375\ M\^68 + 128962\ M\^70 + 
                    128716\ M\^72 - 42127\ M\^74 - 35231\ M\^76 - 
                    2785\ M\^78 + 26486\ M\^80 - 1651\ M\^82 - 13751\ M\^84 + 
                    4507\ M\^86 + 5147\ M\^88 - 5376\ M\^90 + 2325\ M\^92 - 
                    555\ M\^94 + 72\ M\^96 - 4\ M\^98)\) + 
              L\^9\ \((\(-4\)\ M\^4 + 76\ M\^6 - 620\ M\^8 + 2800\ M\^10 - 
                    7356\ M\^12 + 9812\ M\^14 + 945\ M\^16 - 26531\ M\^18 + 
                    36738\ M\^20 + 19296\ M\^22 - 141410\ M\^24 + 
                    146034\ M\^26 + 241883\ M\^28 - 610424\ M\^30 - 
                    225815\ M\^32 + 1502992\ M\^34 - 19376\ M\^36 - 
                    2724634\ M\^38 + 298882\ M\^40 + 4020560\ M\^42 - 
                    185629\ M\^44 - 5050131\ M\^46 - 621061\ M\^48 + 
                    5586790\ M\^50 + 1741819\ M\^52 - 5528536\ M\^54 - 
                    2252516\ M\^56 + 5133634\ M\^58 + 1927885\ M\^60 - 
                    4267939\ M\^62 - 1016481\ M\^64 + 3049877\ M\^66 + 
                    120430\ M\^68 - 1790635\ M\^70 + 334506\ M\^72 + 
                    833624\ M\^74 - 385188\ M\^76 - 278516\ M\^78 + 
                    245762\ M\^80 + 54572\ M\^82 - 119111\ M\^84 + 
                    15969\ M\^86 + 36338\ M\^88 - 16026\ M\^90 - 
                    6679\ M\^92 + 8816\ M\^94 - 3830\ M\^96 + 892\ M\^98 - 
                    112\ M\^100 + 6\ M\^102)\) + 
              L\^17\ \((3\ M\^66 - 22\ M\^68 + 73\ M\^70 - 124\ M\^72 + 
                    22\ M\^74 + 247\ M\^76 - 185\ M\^78 - 202\ M\^80 + 
                    54\ M\^82 + 213\ M\^84 + 358\ M\^86 - 326\ M\^88 - 
                    194\ M\^90 + 436\ M\^92 - 116\ M\^94 - 174\ M\^96 + 
                    150\ M\^98 - 48\ M\^100 + 6\ M\^102)\) + 
              L\^16\ \((M\^52 - 11\ M\^54 + 57\ M\^56 - 165\ M\^58 + 
                    260\ M\^60 - 215\ M\^62 + 16\ M\^64 + 545\ M\^66 - 
                    1239\ M\^68 + 79\ M\^70 + 2438\ M\^72 - 1104\ M\^74 - 
                    1863\ M\^76 - 302\ M\^78 + 1046\ M\^80 + 3478\ M\^82 - 
                    938\ M\^84 - 2950\ M\^86 + 2155\ M\^88 + 265\ M\^90 - 
                    1273\ M\^92 + 1095\ M\^94 - 271\ M\^96 - 479\ M\^98 + 
                    532\ M\^100 - 232\ M\^102 + 48\ M\^104 - 4\ M\^106)\) + 
              L\^10\ \((6\ M\^8 - 112\ M\^10 + 892\ M\^12 - 3830\ M\^14 + 
                    8816\ M\^16 - 6679\ M\^18 - 16026\ M\^20 + 36338\ M\^22 + 
                    15969\ M\^24 - 119111\ M\^26 + 54572\ M\^28 + 
                    245762\ M\^30 - 278516\ M\^32 - 385188\ M\^34 + 
                    833624\ M\^36 + 334506\ M\^38 - 1790635\ M\^40 + 
                    120430\ M\^42 + 3049877\ M\^44 - 1016481\ M\^46 - 
                    4267939\ M\^48 + 1927885\ M\^50 + 5133634\ M\^52 - 
                    2252516\ M\^54 - 5528536\ M\^56 + 1741819\ M\^58 + 
                    5586790\ M\^60 - 621061\ M\^62 - 5050131\ M\^64 - 
                    185629\ M\^66 + 4020560\ M\^68 + 298882\ M\^70 - 
                    2724634\ M\^72 - 19376\ M\^74 + 1502992\ M\^76 - 
                    225815\ M\^78 - 610424\ M\^80 + 241883\ M\^82 + 
                    146034\ M\^84 - 141410\ M\^86 + 19296\ M\^88 + 
                    36738\ M\^90 - 26531\ M\^92 + 945\ M\^94 + 9812\ M\^96 - 
                    7356\ M\^98 + 2800\ M\^100 - 620\ M\^102 + 76\ M\^104 - 
                    4\ M\^106)\) + 
              L\^14\ \((M\^32 - 17\ M\^34 + 137\ M\^36 - 619\ M\^38 + 
                    1567\ M\^40 - 1737\ M\^42 - 1107\ M\^44 + 4880\ M\^46 - 
                    1602\ M\^48 - 8044\ M\^50 + 6593\ M\^52 + 10969\ M\^54 - 
                    10277\ M\^56 - 31724\ M\^58 + 37880\ M\^60 + 
                    70841\ M\^62 - 120292\ M\^64 - 92414\ M\^66 + 
                    244854\ M\^68 + 46889\ M\^70 - 320860\ M\^72 + 
                    48908\ M\^74 + 281746\ M\^76 - 97834\ M\^78 - 
                    138537\ M\^80 + 80944\ M\^82 + 21312\ M\^84 - 
                    31851\ M\^86 + 12816\ M\^88 - 601\ M\^90 - 1124\ M\^92 + 
                    1885\ M\^94 - 4806\ M\^96 + 2937\ M\^98 + 1630\ M\^100 - 
                    2948\ M\^102 + 1630\ M\^104 - 460\ M\^106 + 67\ M\^108 - 
                    4\ M\^110)\) + 
              L\^12\ \((M\^16 - 17\ M\^18 + 128\ M\^20 - 566\ M\^22 + 
                    1662\ M\^24 - 3486\ M\^26 + 5067\ M\^28 - 2362\ M\^30 - 
                    11436\ M\^32 + 28159\ M\^34 - 3160\ M\^36 - 
                    84447\ M\^38 + 104991\ M\^40 + 110419\ M\^42 - 
                    327603\ M\^44 - 23704\ M\^46 + 677249\ M\^48 - 
                    272398\ M\^50 - 1061716\ M\^52 + 746676\ M\^54 + 
                    1365241\ M\^56 - 1219068\ M\^58 - 1500290\ M\^60 + 
                    1411865\ M\^62 + 1506240\ M\^64 - 1192496\ M\^66 - 
                    1425545\ M\^68 + 794751\ M\^70 + 1393253\ M\^72 - 
                    403011\ M\^74 - 1244347\ M\^76 + 206815\ M\^78 + 
                    911488\ M\^80 - 189319\ M\^82 - 496593\ M\^84 + 
                    182132\ M\^86 + 175296\ M\^88 - 122212\ M\^90 - 
                    12616\ M\^92 + 40700\ M\^94 - 17332\ M\^96 - 
                    2017\ M\^98 + 7970\ M\^100 - 5848\ M\^102 + 
                    2367\ M\^104 - 562\ M\^106 + 73\ M\^108 - 4\ M\^110)\) + 
              L\^11\ \((\(-4\)\ M\^12 + 72\ M\^14 - 555\ M\^16 + 
                    2325\ M\^18 - 5376\ M\^20 + 5147\ M\^22 + 4507\ M\^24 - 
                    13751\ M\^26 - 1651\ M\^28 + 26486\ M\^30 - 2785\ M\^32 - 
                    35231\ M\^34 - 42127\ M\^36 + 128716\ M\^38 + 
                    128962\ M\^40 - 446375\ M\^42 - 208830\ M\^44 + 
                    1195454\ M\^46 + 122142\ M\^48 - 2411054\ M\^50 + 
                    226766\ M\^52 + 3837308\ M\^54 - 785504\ M\^56 - 
                    4935663\ M\^58 + 1320176\ M\^60 + 5269679\ M\^62 - 
                    1581101\ M\^64 - 4620101\ M\^66 + 1727799\ M\^68 + 
                    3474559\ M\^70 - 1713132\ M\^72 - 2230105\ M\^74 + 
                    1490127\ M\^76 + 1156770\ M\^78 - 1104621\ M\^80 - 
                    403196\ M\^82 + 650382\ M\^84 + 28851\ M\^86 - 
                    284974\ M\^88 + 76337\ M\^90 + 69313\ M\^92 - 
                    46726\ M\^94 - 658\ M\^96 + 12885\ M\^98 - 8004\ M\^100 + 
                    3033\ M\^102 - 833\ M\^104 + 161\ M\^106 - 19\ M\^108 + 
                    M\^110)\) + 
              L\^15\ \((2\ M\^42 - 26\ M\^44 + 158\ M\^46 - 536\ M\^48 + 
                    933\ M\^50 - 298\ M\^52 - 1728\ M\^54 + 2279\ M\^56 + 
                    1510\ M\^58 - 5438\ M\^60 + 1538\ M\^62 + 8270\ M\^64 - 
                    7628\ M\^66 - 11979\ M\^68 + 20756\ M\^70 + 
                    12560\ M\^72 - 37087\ M\^74 - 5568\ M\^76 + 
                    44351\ M\^78 - 4399\ M\^80 - 27000\ M\^82 + 
                    14812\ M\^84 + 3901\ M\^86 - 12719\ M\^88 + 8525\ M\^90 + 
                    2448\ M\^92 - 6410\ M\^94 + 2659\ M\^96 + 899\ M\^98 - 
                    1623\ M\^100 + 1045\ M\^102 - 426\ M\^104 + 110\ M\^106 - 
                    16\ M\^108 + M\^110)\) + 
              L\^13\ \((\(-2\)\ M\^24 + 33\ M\^26 - 239\ M\^28 + 955\ M\^30 - 
                    2145\ M\^32 + 2261\ M\^34 + 110\ M\^36 - 1692\ M\^38 + 
                    608\ M\^40 - 8028\ M\^42 + 19558\ M\^44 + 17680\ M\^46 - 
                    103392\ M\^48 + 61328\ M\^50 + 215932\ M\^52 - 
                    328714\ M\^54 - 224444\ M\^56 + 784591\ M\^58 - 
                    20819\ M\^60 - 1221325\ M\^62 + 471001\ M\^64 + 
                    1367164\ M\^66 - 860704\ M\^68 - 1110681\ M\^70 + 
                    943008\ M\^72 + 670844\ M\^74 - 655961\ M\^76 - 
                    268331\ M\^78 + 285670\ M\^80 + 65038\ M\^82 - 
                    70527\ M\^84 - 17803\ M\^86 + 15188\ M\^88 + 
                    14011\ M\^90 - 20688\ M\^92 + 1443\ M\^94 + 
                    13930\ M\^96 - 7965\ M\^98 - 3117\ M\^100 + 
                    5558\ M\^102 - 2862\ M\^104 + 759\ M\^106 - 105\ M\^108 + 
                    6\ M\^110)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 16]\), \(1 + L\^14\ M\^176 + 
              L\ \((\(-1\) + 6\ M\^2 - 18\ M\^4 + 26\ M\^6 - M\^8 - 
                    40\ M\^10 + 42\ M\^12 + 14\ M\^14 - 20\ M\^16 + 
                    6\ M\^18)\) + 
              L\^2\ \((M\^4 - 9\ M\^6 + 42\ M\^8 - 112\ M\^10 + 155\ M\^12 + 
                    2\ M\^14 - 374\ M\^16 + 463\ M\^18 + 178\ M\^20 - 
                    810\ M\^22 + 403\ M\^24 + 571\ M\^26 - 551\ M\^28 + 
                    17\ M\^30 + 198\ M\^32 - 98\ M\^34 + 15\ M\^36)\) + 
              L\^3\ \((2\ M\^10 - 20\ M\^12 + 96\ M\^14 - 259\ M\^16 + 
                    350\ M\^18 + 55\ M\^20 - 968\ M\^22 + 1087\ M\^24 + 
                    849\ M\^26 - 2957\ M\^28 + 1591\ M\^30 + 2837\ M\^32 - 
                    4778\ M\^34 + 618\ M\^36 + 4994\ M\^38 - 3471\ M\^40 - 
                    1669\ M\^42 + 3153\ M\^44 - 889\ M\^46 - 722\ M\^48 + 
                    634\ M\^50 - 189\ M\^52 + 20\ M\^54)\) + 
              L\^4\ \((M\^16 - 11\ M\^18 + 57\ M\^20 - 174\ M\^22 + 
                    318\ M\^24 - 288\ M\^26 - 80\ M\^28 + 587\ M\^30 - 
                    669\ M\^32 - 444\ M\^34 + 2801\ M\^36 - 2877\ M\^38 - 
                    3733\ M\^40 + 9805\ M\^42 + 367\ M\^44 - 17167\ M\^46 + 
                    7780\ M\^48 + 19272\ M\^50 - 15463\ M\^52 - 
                    10506\ M\^54 + 16051\ M\^56 - 136\ M\^58 - 8653\ M\^60 + 
                    4421\ M\^62 + 920\ M\^64 - 1862\ M\^66 + 849\ M\^68 - 
                    180\ M\^70 + 15\ M\^72)\) + 
              L\^5\ \((M\^26 - 13\ M\^28 + 78\ M\^30 - 276\ M\^32 + 
                    562\ M\^34 - 436\ M\^36 - 653\ M\^38 + 1530\ M\^40 + 
                    641\ M\^42 - 4897\ M\^44 + 3472\ M\^46 + 7134\ M\^48 - 
                    12889\ M\^50 - 3220\ M\^52 + 25499\ M\^54 - 
                    10682\ M\^56 - 33620\ M\^58 + 29933\ M\^60 + 
                    29681\ M\^62 - 40160\ M\^64 - 13985\ M\^66 + 
                    39786\ M\^68 - 3036\ M\^70 - 26309\ M\^72 + 
                    11502\ M\^74 + 9867\ M\^76 - 9623\ M\^78 + 615\ M\^80 + 
                    2838\ M\^82 - 1796\ M\^84 + 538\ M\^86 - 86\ M\^88 + 
                    6\ M\^90)\) + 
              L\^6\ \((\(-M\^38\) + 17\ M\^40 - 115\ M\^42 + 393\ M\^44 - 
                    644\ M\^46 + 174\ M\^48 + 660\ M\^50 + 756\ M\^52 - 
                    3647\ M\^54 - 851\ M\^56 + 11668\ M\^58 - 2744\ M\^60 - 
                    27728\ M\^62 + 17827\ M\^64 + 46512\ M\^66 - 
                    46271\ M\^68 - 54800\ M\^70 + 74813\ M\^72 + 
                    45013\ M\^74 - 85231\ M\^76 - 16437\ M\^78 + 
                    72605\ M\^80 - 5826\ M\^82 - 46103\ M\^84 + 
                    15994\ M\^86 + 20266\ M\^88 - 14322\ M\^90 - 
                    4455\ M\^92 + 8199\ M\^94 - 1895\ M\^96 - 2180\ M\^98 + 
                    1962\ M\^100 - 744\ M\^102 + 155\ M\^104 - 18\ M\^106 + 
                    M\^108)\) + 
              L\^7\ \((\(-M\^52\) + 16\ M\^54 - 111\ M\^56 + 444\ M\^58 - 
                    1103\ M\^60 + 1447\ M\^62 + 396\ M\^64 - 4236\ M\^66 + 
                    1924\ M\^68 + 11019\ M\^70 - 10445\ M\^72 - 
                    22576\ M\^74 + 31447\ M\^76 + 34137\ M\^78 - 
                    65796\ M\^80 - 34800\ M\^82 + 101889\ M\^84 + 
                    15341\ M\^86 - 114552\ M\^88 + 15341\ M\^90 + 
                    101889\ M\^92 - 34800\ M\^94 - 65796\ M\^96 + 
                    34137\ M\^98 + 31447\ M\^100 - 22576\ M\^102 - 
                    10445\ M\^104 + 11019\ M\^106 + 1924\ M\^108 - 
                    4236\ M\^110 + 396\ M\^112 + 1447\ M\^114 - 
                    1103\ M\^116 + 444\ M\^118 - 111\ M\^120 + 16\ M\^122 - 
                    M\^124)\) + 
              L\^8\ \((M\^68 - 18\ M\^70 + 155\ M\^72 - 744\ M\^74 + 
                    1962\ M\^76 - 2180\ M\^78 - 1895\ M\^80 + 8199\ M\^82 - 
                    4455\ M\^84 - 14322\ M\^86 + 20266\ M\^88 + 
                    15994\ M\^90 - 46103\ M\^92 - 5826\ M\^94 + 
                    72605\ M\^96 - 16437\ M\^98 - 85231\ M\^100 + 
                    45013\ M\^102 + 74813\ M\^104 - 54800\ M\^106 - 
                    46271\ M\^108 + 46512\ M\^110 + 17827\ M\^112 - 
                    27728\ M\^114 - 2744\ M\^116 + 11668\ M\^118 - 
                    851\ M\^120 - 3647\ M\^122 + 756\ M\^124 + 660\ M\^126 + 
                    174\ M\^128 - 644\ M\^130 + 393\ M\^132 - 115\ M\^134 + 
                    17\ M\^136 - M\^138)\) + 
              L\^9\ \((6\ M\^86 - 86\ M\^88 + 538\ M\^90 - 1796\ M\^92 + 
                    2838\ M\^94 + 615\ M\^96 - 9623\ M\^98 + 9867\ M\^100 + 
                    11502\ M\^102 - 26309\ M\^104 - 3036\ M\^106 + 
                    39786\ M\^108 - 13985\ M\^110 - 40160\ M\^112 + 
                    29681\ M\^114 + 29933\ M\^116 - 33620\ M\^118 - 
                    10682\ M\^120 + 25499\ M\^122 - 3220\ M\^124 - 
                    12889\ M\^126 + 7134\ M\^128 + 3472\ M\^130 - 
                    4897\ M\^132 + 641\ M\^134 + 1530\ M\^136 - 653\ M\^138 - 
                    436\ M\^140 + 562\ M\^142 - 276\ M\^144 + 78\ M\^146 - 
                    13\ M\^148 + M\^150)\) + 
              L\^10\ \((15\ M\^104 - 180\ M\^106 + 849\ M\^108 - 
                    1862\ M\^110 + 920\ M\^112 + 4421\ M\^114 - 
                    8653\ M\^116 - 136\ M\^118 + 16051\ M\^120 - 
                    10506\ M\^122 - 15463\ M\^124 + 19272\ M\^126 + 
                    7780\ M\^128 - 17167\ M\^130 + 367\ M\^132 + 
                    9805\ M\^134 - 3733\ M\^136 - 2877\ M\^138 + 
                    2801\ M\^140 - 444\ M\^142 - 669\ M\^144 + 587\ M\^146 - 
                    80\ M\^148 - 288\ M\^150 + 318\ M\^152 - 174\ M\^154 + 
                    57\ M\^156 - 11\ M\^158 + M\^160)\) + 
              L\^11\ \((20\ M\^122 - 189\ M\^124 + 634\ M\^126 - 
                    722\ M\^128 - 889\ M\^130 + 3153\ M\^132 - 1669\ M\^134 - 
                    3471\ M\^136 + 4994\ M\^138 + 618\ M\^140 - 
                    4778\ M\^142 + 2837\ M\^144 + 1591\ M\^146 - 
                    2957\ M\^148 + 849\ M\^150 + 1087\ M\^152 - 968\ M\^154 + 
                    55\ M\^156 + 350\ M\^158 - 259\ M\^160 + 96\ M\^162 - 
                    20\ M\^164 + 2\ M\^166)\) + 
              L\^12\ \((15\ M\^140 - 98\ M\^142 + 198\ M\^144 + 17\ M\^146 - 
                    551\ M\^148 + 571\ M\^150 + 403\ M\^152 - 810\ M\^154 + 
                    178\ M\^156 + 463\ M\^158 - 374\ M\^160 + 2\ M\^162 + 
                    155\ M\^164 - 112\ M\^166 + 42\ M\^168 - 9\ M\^170 + 
                    M\^172)\) + 
              L\^13\ \((6\ M\^158 - 20\ M\^160 + 14\ M\^162 + 42\ M\^164 - 
                    40\ M\^166 - M\^168 + 26\ M\^170 - 18\ M\^172 + 
                    6\ M\^174 - M\^176)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 17]\), \(L\^13\ M\^22 + M\^72 + 
              L\^12\ \((\(-2\)\ M\^14 + 8\ M\^16 - 13\ M\^18 + 8\ M\^20 + 
                    7\ M\^22 - 27\ M\^24 + 33\ M\^26 + 18\ M\^28 - 
                    25\ M\^30 + 6\ M\^32)\) + 
              L\^11\ \((M\^6 - 8\ M\^8 + 30\ M\^10 - 58\ M\^12 + 16\ M\^14 + 
                    144\ M\^16 - 147\ M\^18 - 227\ M\^20 + 305\ M\^22 + 
                    342\ M\^24 - 514\ M\^26 - 437\ M\^28 + 792\ M\^30 + 
                    327\ M\^32 - 664\ M\^34 + 30\ M\^36 + 244\ M\^38 - 
                    113\ M\^40 + 15\ M\^42)\) + 
              L\^10\ \((\(-M\^2\) + 11\ M\^4 - 55\ M\^6 + 149\ M\^8 - 
                    155\ M\^10 - 275\ M\^12 + 938\ M\^14 - 346\ M\^16 - 
                    1913\ M\^18 + 2445\ M\^20 + 1253\ M\^22 - 4605\ M\^24 + 
                    1761\ M\^26 + 4592\ M\^28 - 5465\ M\^30 - 2458\ M\^32 + 
                    7405\ M\^34 + 347\ M\^36 - 6029\ M\^38 + 1306\ M\^40 + 
                    3021\ M\^42 - 1580\ M\^44 - 558\ M\^46 + 680\ M\^48 - 
                    202\ M\^50 + 20\ M\^52)\) + 
              L\^9\ \((\(-1\) + 12\ M\^2 - 60\ M\^4 + 145\ M\^6 - 102\ M\^8 - 
                    274\ M\^10 + 489\ M\^12 + 428\ M\^14 - 1373\ M\^16 - 
                    333\ M\^18 + 2531\ M\^20 + 232\ M\^22 - 4092\ M\^24 - 
                    448\ M\^26 + 6236\ M\^28 + 2846\ M\^30 - 10377\ M\^32 - 
                    8168\ M\^34 + 15191\ M\^36 + 12940\ M\^38 - 
                    15920\ M\^40 - 11550\ M\^42 + 13484\ M\^44 + 
                    6607\ M\^46 - 9372\ M\^48 - 1673\ M\^50 + 4775\ M\^52 - 
                    836\ M\^54 - 1231\ M\^56 + 772\ M\^58 - 178\ M\^60 + 
                    15\ M\^62)\) + 
              L\^8\ \((3\ M\^4 - 39\ M\^6 + 209\ M\^8 - 526\ M\^10 + 
                    291\ M\^12 + 1550\ M\^14 - 2726\ M\^16 - 2481\ M\^18 + 
                    8794\ M\^20 + 1899\ M\^22 - 17616\ M\^24 - 1858\ M\^26 + 
                    26521\ M\^28 + 6867\ M\^30 - 30272\ M\^32 - 
                    18647\ M\^34 + 25691\ M\^36 + 28484\ M\^38 - 
                    16256\ M\^40 - 27751\ M\^42 + 9657\ M\^44 + 
                    21514\ M\^46 - 4119\ M\^48 - 12746\ M\^50 - 62\ M\^52 + 
                    5475\ M\^54 + 2306\ M\^56 - 2555\ M\^58 - 1937\ M\^60 + 
                    1775\ M\^62 + 407\ M\^64 - 884\ M\^66 + 390\ M\^68 - 
                    77\ M\^70 + 6\ M\^72)\) + 
              L\ \((6\ M\^62 - 25\ M\^64 + 18\ M\^66 + 33\ M\^68 - 
                    27\ M\^70 + 7\ M\^72 + 8\ M\^74 - 13\ M\^76 + 8\ M\^78 - 
                    2\ M\^80)\) + 
              L\^7\ \((\(-3\)\ M\^8 + 40\ M\^10 - 220\ M\^12 + 583\ M\^14 - 
                    485\ M\^16 - 1081\ M\^18 + 2233\ M\^20 + 1321\ M\^22 - 
                    4988\ M\^24 - 1806\ M\^26 + 7566\ M\^28 + 4896\ M\^30 - 
                    9945\ M\^32 - 10065\ M\^34 + 10400\ M\^36 + 
                    16503\ M\^38 - 6649\ M\^40 - 24733\ M\^42 - 5571\ M\^44 + 
                    32309\ M\^46 + 24991\ M\^48 - 29883\ M\^50 - 
                    33709\ M\^52 + 22721\ M\^54 + 27609\ M\^56 - 
                    16039\ M\^58 - 14954\ M\^60 + 10212\ M\^62 + 
                    4142\ M\^64 - 4398\ M\^66 + 282\ M\^68 + 707\ M\^70 - 
                    475\ M\^72 + 356\ M\^74 - 211\ M\^76 + 72\ M\^78 - 
                    13\ M\^80 + M\^82)\) + 
              L\^6\ \((M\^12 - 13\ M\^14 + 72\ M\^16 - 211\ M\^18 + 
                    356\ M\^20 - 475\ M\^22 + 707\ M\^24 + 282\ M\^26 - 
                    4398\ M\^28 + 4142\ M\^30 + 10212\ M\^32 - 14954\ M\^34 - 
                    16039\ M\^36 + 27609\ M\^38 + 22721\ M\^40 - 
                    33709\ M\^42 - 29883\ M\^44 + 24991\ M\^46 + 
                    32309\ M\^48 - 5571\ M\^50 - 24733\ M\^52 - 6649\ M\^54 + 
                    16503\ M\^56 + 10400\ M\^58 - 10065\ M\^60 - 
                    9945\ M\^62 + 4896\ M\^64 + 7566\ M\^66 - 1806\ M\^68 - 
                    4988\ M\^70 + 1321\ M\^72 + 2233\ M\^74 - 1081\ M\^76 - 
                    485\ M\^78 + 583\ M\^80 - 220\ M\^82 + 40\ M\^84 - 
                    3\ M\^86)\) + 
              L\^2\ \((15\ M\^52 - 113\ M\^54 + 244\ M\^56 + 30\ M\^58 - 
                    664\ M\^60 + 327\ M\^62 + 792\ M\^64 - 437\ M\^66 - 
                    514\ M\^68 + 342\ M\^70 + 305\ M\^72 - 227\ M\^74 - 
                    147\ M\^76 + 144\ M\^78 + 16\ M\^80 - 58\ M\^82 + 
                    30\ M\^84 - 8\ M\^86 + M\^88)\) + 
              L\^5\ \((6\ M\^22 - 77\ M\^24 + 390\ M\^26 - 884\ M\^28 + 
                    407\ M\^30 + 1775\ M\^32 - 1937\ M\^34 - 2555\ M\^36 + 
                    2306\ M\^38 + 5475\ M\^40 - 62\ M\^42 - 12746\ M\^44 - 
                    4119\ M\^46 + 21514\ M\^48 + 9657\ M\^50 - 27751\ M\^52 - 
                    16256\ M\^54 + 28484\ M\^56 + 25691\ M\^58 - 
                    18647\ M\^60 - 30272\ M\^62 + 6867\ M\^64 + 
                    26521\ M\^66 - 1858\ M\^68 - 17616\ M\^70 + 1899\ M\^72 + 
                    8794\ M\^74 - 2481\ M\^76 - 2726\ M\^78 + 1550\ M\^80 + 
                    291\ M\^82 - 526\ M\^84 + 209\ M\^86 - 39\ M\^88 + 
                    3\ M\^90)\) + 
              L\^3\ \((20\ M\^42 - 202\ M\^44 + 680\ M\^46 - 558\ M\^48 - 
                    1580\ M\^50 + 3021\ M\^52 + 1306\ M\^54 - 6029\ M\^56 + 
                    347\ M\^58 + 7405\ M\^60 - 2458\ M\^62 - 5465\ M\^64 + 
                    4592\ M\^66 + 1761\ M\^68 - 4605\ M\^70 + 1253\ M\^72 + 
                    2445\ M\^74 - 1913\ M\^76 - 346\ M\^78 + 938\ M\^80 - 
                    275\ M\^82 - 155\ M\^84 + 149\ M\^86 - 55\ M\^88 + 
                    11\ M\^90 - M\^92)\) + 
              L\^4\ \((15\ M\^32 - 178\ M\^34 + 772\ M\^36 - 1231\ M\^38 - 
                    836\ M\^40 + 4775\ M\^42 - 1673\ M\^44 - 9372\ M\^46 + 
                    6607\ M\^48 + 13484\ M\^50 - 11550\ M\^52 - 
                    15920\ M\^54 + 12940\ M\^56 + 15191\ M\^58 - 
                    8168\ M\^60 - 10377\ M\^62 + 2846\ M\^64 + 6236\ M\^66 - 
                    448\ M\^68 - 4092\ M\^70 + 232\ M\^72 + 2531\ M\^74 - 
                    333\ M\^76 - 1373\ M\^78 + 428\ M\^80 + 489\ M\^82 - 
                    274\ M\^84 - 102\ M\^86 + 145\ M\^88 - 60\ M\^90 + 
                    12\ M\^92 - M\^94)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 18]\), \(L\^20 + M\^164 + 
              L\^19\ \((\(-4\) + 18\ M\^2 - 28\ M\^4 + 57\ M\^8 - 22\ M\^10 - 
                    14\ M\^12 + 21\ M\^14 - 11\ M\^16 + 3\ M\^18)\) + 
              L\^18\ \((6 - 54\ M\^2 + 189\ M\^4 - 256\ M\^6 - 110\ M\^8 + 
                    681\ M\^10 - 498\ M\^12 - 497\ M\^14 + 963\ M\^16 + 
                    34\ M\^18 - 409\ M\^20 + 148\ M\^22 - 78\ M\^24 + 
                    77\ M\^26 + 54\ M\^28 - 111\ M\^30 + 70\ M\^32 - 
                    22\ M\^34 + 3\ M\^36)\) + 
              L\^17\ \((\(-4\) + 54\ M\^2 - 299\ M\^4 + 820\ M\^6 - 
                    963\ M\^8 - 483\ M\^10 + 2899\ M\^12 - 2810\ M\^14 - 
                    1432\ M\^16 + 5442\ M\^18 - 4181\ M\^20 - 1745\ M\^22 + 
                    6675\ M\^24 - 3369\ M\^26 - 892\ M\^28 + 4159\ M\^30 - 
                    4039\ M\^32 - 209\ M\^34 + 3036\ M\^36 - 1611\ M\^38 - 
                    294\ M\^40 + 618\ M\^42 - 366\ M\^44 + 223\ M\^46 - 
                    129\ M\^48 + 50\ M\^50 - 11\ M\^52 + M\^54)\) + 
              L\^16\ \((1 - 18\ M\^2 + 143\ M\^4 - 656\ M\^6 + 1882\ M\^8 - 
                    3145\ M\^10 + 1171\ M\^12 + 7657\ M\^14 - 16158\ M\^16 + 
                    2641\ M\^18 + 33116\ M\^20 - 37770\ M\^22 - 
                    21375\ M\^24 + 68816\ M\^26 - 23471\ M\^28 - 
                    59069\ M\^30 + 62628\ M\^32 + 19382\ M\^34 - 
                    51573\ M\^36 + 14164\ M\^38 + 17272\ M\^40 - 
                    14648\ M\^42 + 2373\ M\^44 + 2807\ M\^46 - 762\ M\^48 - 
                    314\ M\^50 - 1686\ M\^52 + 1982\ M\^54 + 89\ M\^56 - 
                    1371\ M\^58 + 1087\ M\^60 - 443\ M\^62 + 107\ M\^64 - 
                    15\ M\^66 + M\^68)\) + 
              L\^15\ \((\(-5\)\ M\^4 + 92\ M\^6 - 716\ M\^8 + 2997\ M\^10 - 
                    6772\ M\^12 + 5220\ M\^14 + 12223\ M\^16 - 33675\ M\^18 + 
                    12216\ M\^20 + 64121\ M\^22 - 91786\ M\^24 - 
                    34199\ M\^26 + 178624\ M\^28 - 71397\ M\^30 - 
                    193432\ M\^32 + 184383\ M\^34 + 111639\ M\^36 - 
                    222526\ M\^38 - 5926\ M\^40 + 206063\ M\^42 - 
                    29853\ M\^44 - 158705\ M\^46 + 35146\ M\^48 + 
                    118094\ M\^50 - 44875\ M\^52 - 67476\ M\^54 + 
                    53447\ M\^56 + 14774\ M\^58 - 40741\ M\^60 + 
                    17296\ M\^62 + 11839\ M\^64 - 15678\ M\^66 + 
                    3521\ M\^68 + 4309\ M\^70 - 4111\ M\^72 + 1749\ M\^74 - 
                    434\ M\^76 + 62\ M\^78 - 4\ M\^80)\) + 
              L\^14\ \((10\ M\^8 - 188\ M\^10 + 1500\ M\^12 - 6393\ M\^14 + 
                    14148\ M\^16 - 7122\ M\^18 - 40149\ M\^20 + 
                    81095\ M\^22 + 29464\ M\^24 - 252373\ M\^26 + 
                    135520\ M\^28 + 450508\ M\^30 - 503428\ M\^32 - 
                    567160\ M\^34 + 996030\ M\^36 + 543977\ M\^38 - 
                    1423550\ M\^40 - 429569\ M\^42 + 1654334\ M\^44 + 
                    206187\ M\^46 - 1641870\ M\^48 + 221356\ M\^50 + 
                    1432462\ M\^52 - 692982\ M\^54 - 942366\ M\^56 + 
                    964003\ M\^58 + 323938\ M\^60 - 837599\ M\^62 + 
                    146844\ M\^64 + 442868\ M\^66 - 281918\ M\^68 - 
                    93532\ M\^70 + 181184\ M\^72 - 53199\ M\^74 - 
                    45462\ M\^76 + 44133\ M\^78 - 7710\ M\^80 - 
                    10334\ M\^82 + 8951\ M\^84 - 3570\ M\^86 + 822\ M\^88 - 
                    106\ M\^90 + 6\ M\^92)\) + 
              L\^13\ \((\(-10\)\ M\^12 + 192\ M\^14 - 1575\ M\^16 + 
                    6964\ M\^18 - 16349\ M\^20 + 11024\ M\^22 + 
                    40466\ M\^24 - 91261\ M\^26 - 33890\ M\^28 + 
                    302748\ M\^30 - 126103\ M\^32 - 672630\ M\^34 + 
                    549396\ M\^36 + 1318362\ M\^38 - 1432559\ M\^40 - 
                    2369193\ M\^42 + 3017903\ M\^44 + 3765322\ M\^46 - 
                    5428163\ M\^48 - 4846663\ M\^50 + 8141993\ M\^52 + 
                    4582475\ M\^54 - 10001820\ M\^56 - 2441687\ M\^58 + 
                    10011863\ M\^60 - 565371\ M\^62 - 7855110\ M\^64 + 
                    2896571\ M\^66 + 4549138\ M\^68 - 3479610\ M\^70 - 
                    1575170\ M\^72 + 2588188\ M\^74 - 117236\ M\^76 - 
                    1249265\ M\^78 + 543167\ M\^80 + 318878\ M\^82 - 
                    354747\ M\^84 + 41606\ M\^86 + 101287\ M\^88 - 
                    61199\ M\^90 - 298\ M\^92 + 18198\ M\^94 - 11432\ M\^96 + 
                    3788\ M\^98 - 747\ M\^100 + 83\ M\^102 - 4\ M\^104)\) + 
              L\^12\ \((5\ M\^16 - 98\ M\^18 + 824\ M\^20 - 3802\ M\^22 + 
                    9896\ M\^24 - 11508\ M\^26 - 7235\ M\^28 + 32233\ M\^30 + 
                    5705\ M\^32 - 88510\ M\^34 - 24941\ M\^36 + 
                    299017\ M\^38 + 112581\ M\^40 - 1170967\ M\^42 + 
                    95538\ M\^44 + 3407215\ M\^46 - 1570936\ M\^48 - 
                    7403543\ M\^50 + 5634501\ M\^52 + 12316937\ M\^54 - 
                    12579629\ M\^56 - 16002786\ M\^58 + 20623367\ M\^60 + 
                    16010142\ M\^62 - 26372146\ M\^64 - 11497193\ M\^66 + 
                    27224054\ M\^68 + 4571917\ M\^70 - 22672505\ M\^72 + 
                    1625007\ M\^74 + 15120786\ M\^76 - 4764204\ M\^78 - 
                    7737948\ M\^80 + 4776979\ M\^82 + 2683426\ M\^84 - 
                    3176311\ M\^86 - 261643\ M\^88 + 1484533\ M\^90 - 
                    376982\ M\^92 - 456467\ M\^94 + 302795\ M\^96 + 
                    39724\ M\^98 - 108161\ M\^100 + 31690\ M\^102 + 
                    16572\ M\^104 - 18011\ M\^106 + 7738\ M\^108 - 
                    1967\ M\^110 + 307\ M\^112 - 27\ M\^114 + M\^116)\) + 
              L\^11\ \((\(-M\^20\) + 20\ M\^22 - 173\ M\^24 + 857\ M\^26 - 
                    2712\ M\^28 + 5886\ M\^30 - 9679\ M\^32 + 13156\ M\^34 - 
                    7795\ M\^36 - 35450\ M\^38 + 109262\ M\^40 - 
                    23345\ M\^42 - 409321\ M\^44 + 553363\ M\^46 + 
                    768778\ M\^48 - 2159473\ M\^50 - 638352\ M\^52 + 
                    5566631\ M\^54 - 1261508\ M\^56 - 10773677\ M\^58 + 
                    6270383\ M\^60 + 16469119\ M\^62 - 14542620\ M\^64 - 
                    20106222\ M\^66 + 23812538\ M\^68 + 19504666\ M\^70 - 
                    30316862\ M\^72 - 14448935\ M\^74 + 31424045\ M\^76 + 
                    7473366\ M\^78 - 26804586\ M\^80 - 1289728\ M\^82 + 
                    19036005\ M\^84 - 2241330\ M\^86 - 11215833\ M\^88 + 
                    3146855\ M\^90 + 5374720\ M\^92 - 2488704\ M\^94 - 
                    2025938\ M\^96 + 1454103\ M\^98 + 533922\ M\^100 - 
                    653921\ M\^102 - 62909\ M\^104 + 245749\ M\^106 - 
                    45339\ M\^108 - 61886\ M\^110 + 31651\ M\^112 + 
                    8526\ M\^114 - 15293\ M\^116 + 7908\ M\^118 - 
                    2339\ M\^120 + 424\ M\^122 - 44\ M\^124 + 2\ M\^126)\) + 
              L\^10\ \((M\^28 - 21\ M\^30 + 200\ M\^32 - 1123\ M\^34 + 
                    4013\ M\^36 - 8957\ M\^38 + 10223\ M\^40 + 2082\ M\^42 - 
                    19053\ M\^44 - 1493\ M\^46 + 56998\ M\^48 - 9121\ M\^50 - 
                    171049\ M\^52 + 43104\ M\^54 + 556309\ M\^56 - 
                    371091\ M\^58 - 1217169\ M\^60 + 1250851\ M\^62 + 
                    1945570\ M\^64 - 2736771\ M\^66 - 2259065\ M\^68 + 
                    4322608\ M\^70 + 1948329\ M\^72 - 5432957\ M\^74 - 
                    1203736\ M\^76 + 5858512\ M\^78 + 440807\ M\^80 - 
                    5831246\ M\^82 + 440807\ M\^84 + 5858512\ M\^86 - 
                    1203736\ M\^88 - 5432957\ M\^90 + 1948329\ M\^92 + 
                    4322608\ M\^94 - 2259065\ M\^96 - 2736771\ M\^98 + 
                    1945570\ M\^100 + 1250851\ M\^102 - 1217169\ M\^104 - 
                    371091\ M\^106 + 556309\ M\^108 + 43104\ M\^110 - 
                    171049\ M\^112 - 9121\ M\^114 + 56998\ M\^116 - 
                    1493\ M\^118 - 19053\ M\^120 + 2082\ M\^122 + 
                    10223\ M\^124 - 8957\ M\^126 + 4013\ M\^128 - 
                    1123\ M\^130 + 200\ M\^132 - 21\ M\^134 + M\^136)\) + 
              L\^9\ \((2\ M\^38 - 44\ M\^40 + 424\ M\^42 - 2339\ M\^44 + 
                    7908\ M\^46 - 15293\ M\^48 + 8526\ M\^50 + 31651\ M\^52 - 
                    61886\ M\^54 - 45339\ M\^56 + 245749\ M\^58 - 
                    62909\ M\^60 - 653921\ M\^62 + 533922\ M\^64 + 
                    1454103\ M\^66 - 2025938\ M\^68 - 2488704\ M\^70 + 
                    5374720\ M\^72 + 3146855\ M\^74 - 11215833\ M\^76 - 
                    2241330\ M\^78 + 19036005\ M\^80 - 1289728\ M\^82 - 
                    26804586\ M\^84 + 7473366\ M\^86 + 31424045\ M\^88 - 
                    14448935\ M\^90 - 30316862\ M\^92 + 19504666\ M\^94 + 
                    23812538\ M\^96 - 20106222\ M\^98 - 14542620\ M\^100 + 
                    16469119\ M\^102 + 6270383\ M\^104 - 10773677\ M\^106 - 
                    1261508\ M\^108 + 5566631\ M\^110 - 638352\ M\^112 - 
                    2159473\ M\^114 + 768778\ M\^116 + 553363\ M\^118 - 
                    409321\ M\^120 - 23345\ M\^122 + 109262\ M\^124 - 
                    35450\ M\^126 - 7795\ M\^128 + 13156\ M\^130 - 
                    9679\ M\^132 + 5886\ M\^134 - 2712\ M\^136 + 
                    857\ M\^138 - 173\ M\^140 + 20\ M\^142 - M\^144)\) + 
              L\^8\ \((M\^48 - 27\ M\^50 + 307\ M\^52 - 1967\ M\^54 + 
                    7738\ M\^56 - 18011\ M\^58 + 16572\ M\^60 + 
                    31690\ M\^62 - 108161\ M\^64 + 39724\ M\^66 + 
                    302795\ M\^68 - 456467\ M\^70 - 376982\ M\^72 + 
                    1484533\ M\^74 - 261643\ M\^76 - 3176311\ M\^78 + 
                    2683426\ M\^80 + 4776979\ M\^82 - 7737948\ M\^84 - 
                    4764204\ M\^86 + 15120786\ M\^88 + 1625007\ M\^90 - 
                    22672505\ M\^92 + 4571917\ M\^94 + 27224054\ M\^96 - 
                    11497193\ M\^98 - 26372146\ M\^100 + 16010142\ M\^102 + 
                    20623367\ M\^104 - 16002786\ M\^106 - 12579629\ M\^108 + 
                    12316937\ M\^110 + 5634501\ M\^112 - 7403543\ M\^114 - 
                    1570936\ M\^116 + 3407215\ M\^118 + 95538\ M\^120 - 
                    1170967\ M\^122 + 112581\ M\^124 + 299017\ M\^126 - 
                    24941\ M\^128 - 88510\ M\^130 + 5705\ M\^132 + 
                    32233\ M\^134 - 7235\ M\^136 - 11508\ M\^138 + 
                    9896\ M\^140 - 3802\ M\^142 + 824\ M\^144 - 98\ M\^146 + 
                    5\ M\^148)\) + 
              L\^7\ \((\(-4\)\ M\^60 + 83\ M\^62 - 747\ M\^64 + 3788\ M\^66 - 
                    11432\ M\^68 + 18198\ M\^70 - 298\ M\^72 - 61199\ M\^74 + 
                    101287\ M\^76 + 41606\ M\^78 - 354747\ M\^80 + 
                    318878\ M\^82 + 543167\ M\^84 - 1249265\ M\^86 - 
                    117236\ M\^88 + 2588188\ M\^90 - 1575170\ M\^92 - 
                    3479610\ M\^94 + 4549138\ M\^96 + 2896571\ M\^98 - 
                    7855110\ M\^100 - 565371\ M\^102 + 10011863\ M\^104 - 
                    2441687\ M\^106 - 10001820\ M\^108 + 4582475\ M\^110 + 
                    8141993\ M\^112 - 4846663\ M\^114 - 5428163\ M\^116 + 
                    3765322\ M\^118 + 3017903\ M\^120 - 2369193\ M\^122 - 
                    1432559\ M\^124 + 1318362\ M\^126 + 549396\ M\^128 - 
                    672630\ M\^130 - 126103\ M\^132 + 302748\ M\^134 - 
                    33890\ M\^136 - 91261\ M\^138 + 40466\ M\^140 + 
                    11024\ M\^142 - 16349\ M\^144 + 6964\ M\^146 - 
                    1575\ M\^148 + 192\ M\^150 - 10\ M\^152)\) + 
              L\^6\ \((6\ M\^72 - 106\ M\^74 + 822\ M\^76 - 3570\ M\^78 + 
                    8951\ M\^80 - 10334\ M\^82 - 7710\ M\^84 + 44133\ M\^86 - 
                    45462\ M\^88 - 53199\ M\^90 + 181184\ M\^92 - 
                    93532\ M\^94 - 281918\ M\^96 + 442868\ M\^98 + 
                    146844\ M\^100 - 837599\ M\^102 + 323938\ M\^104 + 
                    964003\ M\^106 - 942366\ M\^108 - 692982\ M\^110 + 
                    1432462\ M\^112 + 221356\ M\^114 - 1641870\ M\^116 + 
                    206187\ M\^118 + 1654334\ M\^120 - 429569\ M\^122 - 
                    1423550\ M\^124 + 543977\ M\^126 + 996030\ M\^128 - 
                    567160\ M\^130 - 503428\ M\^132 + 450508\ M\^134 + 
                    135520\ M\^136 - 252373\ M\^138 + 29464\ M\^140 + 
                    81095\ M\^142 - 40149\ M\^144 - 7122\ M\^146 + 
                    14148\ M\^148 - 6393\ M\^150 + 1500\ M\^152 - 
                    188\ M\^154 + 10\ M\^156)\) + 
              L\^5\ \((\(-4\)\ M\^84 + 62\ M\^86 - 434\ M\^88 + 1749\ M\^90 - 
                    4111\ M\^92 + 4309\ M\^94 + 3521\ M\^96 - 15678\ M\^98 + 
                    11839\ M\^100 + 17296\ M\^102 - 40741\ M\^104 + 
                    14774\ M\^106 + 53447\ M\^108 - 67476\ M\^110 - 
                    44875\ M\^112 + 118094\ M\^114 + 35146\ M\^116 - 
                    158705\ M\^118 - 29853\ M\^120 + 206063\ M\^122 - 
                    5926\ M\^124 - 222526\ M\^126 + 111639\ M\^128 + 
                    184383\ M\^130 - 193432\ M\^132 - 71397\ M\^134 + 
                    178624\ M\^136 - 34199\ M\^138 - 91786\ M\^140 + 
                    64121\ M\^142 + 12216\ M\^144 - 33675\ M\^146 + 
                    12223\ M\^148 + 5220\ M\^150 - 6772\ M\^152 + 
                    2997\ M\^154 - 716\ M\^156 + 92\ M\^158 - 5\ M\^160)\) + 
              L\ \((3\ M\^146 - 11\ M\^148 + 21\ M\^150 - 14\ M\^152 - 
                    22\ M\^154 + 57\ M\^156 - 28\ M\^160 + 18\ M\^162 - 
                    4\ M\^164)\) + 
              L\^3\ \((M\^110 - 11\ M\^112 + 50\ M\^114 - 129\ M\^116 + 
                    223\ M\^118 - 366\ M\^120 + 618\ M\^122 - 294\ M\^124 - 
                    1611\ M\^126 + 3036\ M\^128 - 209\ M\^130 - 
                    4039\ M\^132 + 4159\ M\^134 - 892\ M\^136 - 
                    3369\ M\^138 + 6675\ M\^140 - 1745\ M\^142 - 
                    4181\ M\^144 + 5442\ M\^146 - 1432\ M\^148 - 
                    2810\ M\^150 + 2899\ M\^152 - 483\ M\^154 - 963\ M\^156 + 
                    820\ M\^158 - 299\ M\^160 + 54\ M\^162 - 4\ M\^164)\) + 
              L\^4\ \((M\^96 - 15\ M\^98 + 107\ M\^100 - 443\ M\^102 + 
                    1087\ M\^104 - 1371\ M\^106 + 89\ M\^108 + 1982\ M\^110 - 
                    1686\ M\^112 - 314\ M\^114 - 762\ M\^116 + 2807\ M\^118 + 
                    2373\ M\^120 - 14648\ M\^122 + 17272\ M\^124 + 
                    14164\ M\^126 - 51573\ M\^128 + 19382\ M\^130 + 
                    62628\ M\^132 - 59069\ M\^134 - 23471\ M\^136 + 
                    68816\ M\^138 - 21375\ M\^140 - 37770\ M\^142 + 
                    33116\ M\^144 + 2641\ M\^146 - 16158\ M\^148 + 
                    7657\ M\^150 + 1171\ M\^152 - 3145\ M\^154 + 
                    1882\ M\^156 - 656\ M\^158 + 143\ M\^160 - 18\ M\^162 + 
                    M\^164)\) + 
              L\^2\ \((3\ M\^128 - 22\ M\^130 + 70\ M\^132 - 111\ M\^134 + 
                    54\ M\^136 + 77\ M\^138 - 78\ M\^140 + 148\ M\^142 - 
                    409\ M\^144 + 34\ M\^146 + 963\ M\^148 - 497\ M\^150 - 
                    498\ M\^152 + 681\ M\^154 - 110\ M\^156 - 256\ M\^158 + 
                    189\ M\^160 - 54\ M\^162 + 6\ M\^164)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 19]\), \(M\^52 + L\^20\ M\^52 + 
              L\ \((4\ M\^42 - 14\ M\^44 + 18\ M\^46 - 7\ M\^48 - 21\ M\^50 + 
                    52\ M\^52 + 2\ M\^54 - 27\ M\^56 + 17\ M\^58 - 
                    4\ M\^60)\) + 
              L\^19\ \((\(-4\)\ M\^44 + 17\ M\^46 - 27\ M\^48 + 2\ M\^50 + 
                    52\ M\^52 - 21\ M\^54 - 7\ M\^56 + 18\ M\^58 - 
                    14\ M\^60 + 4\ M\^62)\) + 
              L\^2\ \((6\ M\^32 - 40\ M\^34 + 115\ M\^36 - 162\ M\^38 - 
                    39\ M\^40 + 458\ M\^42 - 313\ M\^44 - 440\ M\^46 + 
                    367\ M\^48 + 181\ M\^50 + 227\ M\^52 - 119\ M\^54 - 
                    296\ M\^56 + 369\ M\^58 - 35\ M\^60 - 199\ M\^62 + 
                    152\ M\^64 - 48\ M\^66 + 6\ M\^68)\) + 
              L\^18\ \((6\ M\^36 - 48\ M\^38 + 152\ M\^40 - 199\ M\^42 - 
                    35\ M\^44 + 369\ M\^46 - 296\ M\^48 - 119\ M\^50 + 
                    227\ M\^52 + 181\ M\^54 + 367\ M\^56 - 440\ M\^58 - 
                    313\ M\^60 + 458\ M\^62 - 39\ M\^64 - 162\ M\^66 + 
                    115\ M\^68 - 40\ M\^70 + 6\ M\^72)\) + 
              L\^3\ \((4\ M\^22 - 38\ M\^24 + 162\ M\^26 - 372\ M\^28 + 
                    336\ M\^30 + 385\ M\^32 - 1124\ M\^34 + 427\ M\^36 + 
                    1270\ M\^38 - 2032\ M\^40 - 27\ M\^42 + 4126\ M\^44 - 
                    1215\ M\^46 - 8096\ M\^48 + 2368\ M\^50 + 11951\ M\^52 - 
                    3215\ M\^54 - 9331\ M\^56 + 5074\ M\^58 + 3158\ M\^60 - 
                    3704\ M\^62 + 858\ M\^64 + 626\ M\^66 - 820\ M\^68 + 
                    540\ M\^70 - 212\ M\^72 + 45\ M\^74 - 4\ M\^76)\) + 
              L\^17\ \((\(-4\)\ M\^28 + 45\ M\^30 - 212\ M\^32 + 540\ M\^34 - 
                    820\ M\^36 + 626\ M\^38 + 858\ M\^40 - 3704\ M\^42 + 
                    3158\ M\^44 + 5074\ M\^46 - 9331\ M\^48 - 3215\ M\^50 + 
                    11951\ M\^52 + 2368\ M\^54 - 8096\ M\^56 - 1215\ M\^58 + 
                    4126\ M\^60 - 27\ M\^62 - 2032\ M\^64 + 1270\ M\^66 + 
                    427\ M\^68 - 1124\ M\^70 + 385\ M\^72 + 336\ M\^74 - 
                    372\ M\^76 + 162\ M\^78 - 38\ M\^80 + 4\ M\^82)\) + 
              L\^4\ \((M\^12 - 12\ M\^14 + 66\ M\^16 - 208\ M\^18 + 
                    385\ M\^20 - 394\ M\^22 + 256\ M\^24 - 303\ M\^26 + 
                    126\ M\^28 + 1162\ M\^30 - 2764\ M\^32 + 1350\ M\^34 + 
                    5892\ M\^36 - 9007\ M\^38 - 4890\ M\^40 + 7888\ M\^42 + 
                    13108\ M\^44 + 8183\ M\^46 - 42623\ M\^48 - 
                    22818\ M\^50 + 73112\ M\^52 + 17935\ M\^54 - 
                    60982\ M\^56 + 7246\ M\^58 + 22849\ M\^60 - 
                    17258\ M\^62 + 6724\ M\^64 + 6228\ M\^66 - 10545\ M\^68 + 
                    3372\ M\^70 + 2680\ M\^72 - 2937\ M\^74 + 1356\ M\^76 - 
                    413\ M\^78 + 93\ M\^80 - 14\ M\^82 + M\^84)\) + 
              L\^5\ \((\(-3\)\ M\^8 + 43\ M\^10 - 270\ M\^12 + 913\ M\^14 - 
                    1571\ M\^16 + 506\ M\^18 + 2646\ M\^20 - 2433\ M\^22 - 
                    5621\ M\^24 + 9614\ M\^26 + 7253\ M\^28 - 24320\ M\^30 - 
                    3385\ M\^32 + 43026\ M\^34 - 12258\ M\^36 - 
                    23670\ M\^38 - 3712\ M\^40 - 62292\ M\^42 + 
                    105481\ M\^44 + 177088\ M\^46 - 273823\ M\^48 - 
                    213512\ M\^50 + 376010\ M\^52 + 120087\ M\^54 - 
                    286700\ M\^56 + 58866\ M\^58 + 102146\ M\^60 - 
                    142086\ M\^62 + 38406\ M\^64 + 92248\ M\^66 - 
                    72122\ M\^68 - 13276\ M\^70 + 35155\ M\^72 - 
                    13791\ M\^74 - 2396\ M\^76 + 6353\ M\^78 - 4685\ M\^80 + 
                    2060\ M\^82 - 546\ M\^84 + 80\ M\^86 - 5\ M\^88)\) + 
              L\^16\ \((M\^20 - 14\ M\^22 + 93\ M\^24 - 413\ M\^26 + 
                    1356\ M\^28 - 2937\ M\^30 + 2680\ M\^32 + 3372\ M\^34 - 
                    10545\ M\^36 + 6228\ M\^38 + 6724\ M\^40 - 17258\ M\^42 + 
                    22849\ M\^44 + 7246\ M\^46 - 60982\ M\^48 + 
                    17935\ M\^50 + 73112\ M\^52 - 22818\ M\^54 - 
                    42623\ M\^56 + 8183\ M\^58 + 13108\ M\^60 + 7888\ M\^62 - 
                    4890\ M\^64 - 9007\ M\^66 + 5892\ M\^68 + 1350\ M\^70 - 
                    2764\ M\^72 + 1162\ M\^74 + 126\ M\^76 - 303\ M\^78 + 
                    256\ M\^80 - 394\ M\^82 + 385\ M\^84 - 208\ M\^86 + 
                    66\ M\^88 - 12\ M\^90 + M\^92)\) + 
              L\^6\ \((3\ M\^4 - 50\ M\^6 + 357\ M\^8 - 1381\ M\^10 + 
                    2964\ M\^12 - 2825\ M\^14 - 1429\ M\^16 + 6398\ M\^18 - 
                    4620\ M\^20 - 4539\ M\^22 + 13229\ M\^24 - 15853\ M\^26 + 
                    8117\ M\^28 + 29343\ M\^30 - 63018\ M\^32 - 
                    37901\ M\^34 + 142302\ M\^36 + 98548\ M\^38 - 
                    208962\ M\^40 - 339720\ M\^42 + 268152\ M\^44 + 
                    738630\ M\^46 - 329760\ M\^48 - 1025207\ M\^50 + 
                    377502\ M\^52 + 908486\ M\^54 - 301526\ M\^56 - 
                    366487\ M\^58 + 182505\ M\^60 - 100457\ M\^62 - 
                    26945\ M\^64 + 193984\ M\^66 - 83767\ M\^68 - 
                    66549\ M\^70 + 82654\ M\^72 - 38147\ M\^74 - 
                    16628\ M\^76 + 38035\ M\^78 - 15376\ M\^80 - 
                    8708\ M\^82 + 11652\ M\^84 - 5422\ M\^86 + 1344\ M\^88 - 
                    178\ M\^90 + 10\ M\^92)\) + 
              L\^7\ \((\(-1\) + 19\ M\^2 - 157\ M\^4 + 756\ M\^6 - 
                    2426\ M\^8 + 5502\ M\^10 - 7943\ M\^12 + 1605\ M\^14 + 
                    21911\ M\^16 - 41142\ M\^18 - 5552\ M\^20 + 
                    126294\ M\^22 - 147332\ M\^24 - 149166\ M\^26 + 
                    474548\ M\^28 - 27225\ M\^30 - 884533\ M\^32 + 
                    472141\ M\^34 + 1026572\ M\^36 - 951913\ M\^38 - 
                    500171\ M\^40 + 1086064\ M\^42 - 857492\ M\^44 - 
                    669335\ M\^46 + 2553389\ M\^48 - 34672\ M\^50 - 
                    3650688\ M\^52 + 579883\ M\^54 + 3622189\ M\^56 - 
                    553174\ M\^58 - 2552588\ M\^60 + 328535\ M\^62 + 
                    1372750\ M\^64 - 205806\ M\^66 - 689093\ M\^68 + 
                    211447\ M\^70 + 364796\ M\^72 - 227376\ M\^74 - 
                    159276\ M\^76 + 185298\ M\^78 + 12420\ M\^80 - 
                    84448\ M\^82 + 28364\ M\^84 + 16733\ M\^86 - 
                    18062\ M\^88 + 7306\ M\^90 - 1615\ M\^92 + 194\ M\^94 - 
                    10\ M\^96)\) + 
              L\^15\ \((\(-5\)\ M\^16 + 80\ M\^18 - 546\ M\^20 + 
                    2060\ M\^22 - 4685\ M\^24 + 6353\ M\^26 - 2396\ M\^28 - 
                    13791\ M\^30 + 35155\ M\^32 - 13276\ M\^34 - 
                    72122\ M\^36 + 92248\ M\^38 + 38406\ M\^40 - 
                    142086\ M\^42 + 102146\ M\^44 + 58866\ M\^46 - 
                    286700\ M\^48 + 120087\ M\^50 + 376010\ M\^52 - 
                    213512\ M\^54 - 273823\ M\^56 + 177088\ M\^58 + 
                    105481\ M\^60 - 62292\ M\^62 - 3712\ M\^64 - 
                    23670\ M\^66 - 12258\ M\^68 + 43026\ M\^70 - 
                    3385\ M\^72 - 24320\ M\^74 + 7253\ M\^76 + 9614\ M\^78 - 
                    5621\ M\^80 - 2433\ M\^82 + 2646\ M\^84 + 506\ M\^86 - 
                    1571\ M\^88 + 913\ M\^90 - 270\ M\^92 + 43\ M\^94 - 
                    3\ M\^96)\) + 
              L\^14\ \((10\ M\^12 - 178\ M\^14 + 1344\ M\^16 - 5422\ M\^18 + 
                    11652\ M\^20 - 8708\ M\^22 - 15376\ M\^24 + 
                    38035\ M\^26 - 16628\ M\^28 - 38147\ M\^30 + 
                    82654\ M\^32 - 66549\ M\^34 - 83767\ M\^36 + 
                    193984\ M\^38 - 26945\ M\^40 - 100457\ M\^42 + 
                    182505\ M\^44 - 366487\ M\^46 - 301526\ M\^48 + 
                    908486\ M\^50 + 377502\ M\^52 - 1025207\ M\^54 - 
                    329760\ M\^56 + 738630\ M\^58 + 268152\ M\^60 - 
                    339720\ M\^62 - 208962\ M\^64 + 98548\ M\^66 + 
                    142302\ M\^68 - 37901\ M\^70 - 63018\ M\^72 + 
                    29343\ M\^74 + 8117\ M\^76 - 15853\ M\^78 + 
                    13229\ M\^80 - 4539\ M\^82 - 4620\ M\^84 + 6398\ M\^86 - 
                    1429\ M\^88 - 2825\ M\^90 + 2964\ M\^92 - 1381\ M\^94 + 
                    357\ M\^96 - 50\ M\^98 + 3\ M\^100)\) + 
              L\^8\ \((4 - 79\ M\^2 + 664\ M\^4 - 3073\ M\^6 + 8273\ M\^8 - 
                    11466\ M\^10 - 106\ M\^12 + 28384\ M\^14 - 35707\ M\^16 - 
                    25612\ M\^18 + 99358\ M\^20 - 42537\ M\^22 - 
                    60064\ M\^24 + 9365\ M\^26 - 181286\ M\^28 + 
                    456806\ M\^30 + 698627\ M\^32 - 1688435\ M\^34 - 
                    1440577\ M\^36 + 3389843\ M\^38 + 2418589\ M\^40 - 
                    4572706\ M\^42 - 3501472\ M\^44 + 4100108\ M\^46 + 
                    4213328\ M\^48 - 1890015\ M\^50 - 3891886\ M\^52 - 
                    826705\ M\^54 + 2669110\ M\^56 + 2907287\ M\^58 - 
                    1023448\ M\^60 - 3509115\ M\^62 + 67646\ M\^64 + 
                    2940045\ M\^66 - 78715\ M\^68 - 1942178\ M\^70 + 
                    437986\ M\^72 + 952854\ M\^74 - 536627\ M\^76 - 
                    265203\ M\^78 + 355469\ M\^80 - 46306\ M\^82 - 
                    113511\ M\^84 + 68296\ M\^86 + 4482\ M\^88 - 
                    24604\ M\^90 + 14883\ M\^92 - 4850\ M\^94 + 945\ M\^96 - 
                    104\ M\^98 + 5\ M\^100)\) + 
              L\^11\ \((\(-1\) + 22\ M\^2 - 221\ M\^4 + 1359\ M\^6 - 
                    5706\ M\^8 + 16578\ M\^10 - 29271\ M\^12 + 9034\ M\^14 + 
                    94190\ M\^16 - 187396\ M\^18 - 82897\ M\^20 + 
                    709097\ M\^22 - 421638\ M\^24 - 1540486\ M\^26 + 
                    1884740\ M\^28 + 2525290\ M\^30 - 4601955\ M\^32 - 
                    3569124\ M\^34 + 8161481\ M\^36 + 5157495\ M\^38 - 
                    11344602\ M\^40 - 7760710\ M\^42 + 12261588\ M\^44 + 
                    10731201\ M\^46 - 9681845\ M\^48 - 12184454\ M\^50 + 
                    4079064\ M\^52 + 10775355\ M\^54 + 2607011\ M\^56 - 
                    6553719\ M\^58 - 7245308\ M\^60 + 2050508\ M\^62 + 
                    8491031\ M\^64 + 637488\ M\^66 - 7088999\ M\^68 - 
                    1216809\ M\^70 + 4655682\ M\^72 + 713406\ M\^74 - 
                    2530249\ M\^76 - 109135\ M\^78 + 1130756\ M\^80 - 
                    156028\ M\^82 - 399286\ M\^84 + 173722\ M\^86 + 
                    71919\ M\^88 - 79573\ M\^90 + 10226\ M\^92 + 
                    19480\ M\^94 - 14624\ M\^96 + 5236\ M\^98 - 
                    1080\ M\^100 + 123\ M\^102 - 6\ M\^104)\) + 
              L\^13\ \((\(-10\)\ M\^8 + 194\ M\^10 - 1615\ M\^12 + 
                    7306\ M\^14 - 18062\ M\^16 + 16733\ M\^18 + 
                    28364\ M\^20 - 84448\ M\^22 + 12420\ M\^24 + 
                    185298\ M\^26 - 159276\ M\^28 - 227376\ M\^30 + 
                    364796\ M\^32 + 211447\ M\^34 - 689093\ M\^36 - 
                    205806\ M\^38 + 1372750\ M\^40 + 328535\ M\^42 - 
                    2552588\ M\^44 - 553174\ M\^46 + 3622189\ M\^48 + 
                    579883\ M\^50 - 3650688\ M\^52 - 34672\ M\^54 + 
                    2553389\ M\^56 - 669335\ M\^58 - 857492\ M\^60 + 
                    1086064\ M\^62 - 500171\ M\^64 - 951913\ M\^66 + 
                    1026572\ M\^68 + 472141\ M\^70 - 884533\ M\^72 - 
                    27225\ M\^74 + 474548\ M\^76 - 149166\ M\^78 - 
                    147332\ M\^80 + 126294\ M\^82 - 5552\ M\^84 - 
                    41142\ M\^86 + 21911\ M\^88 + 1605\ M\^90 - 7943\ M\^92 + 
                    5502\ M\^94 - 2426\ M\^96 + 756\ M\^98 - 157\ M\^100 + 
                    19\ M\^102 - M\^104)\) + 
              L\^9\ \((\(-6\) + 123\ M\^2 - 1080\ M\^4 + 5236\ M\^6 - 
                    14624\ M\^8 + 19480\ M\^10 + 10226\ M\^12 - 
                    79573\ M\^14 + 71919\ M\^16 + 173722\ M\^18 - 
                    399286\ M\^20 - 156028\ M\^22 + 1130756\ M\^24 - 
                    109135\ M\^26 - 2530249\ M\^28 + 713406\ M\^30 + 
                    4655682\ M\^32 - 1216809\ M\^34 - 7088999\ M\^36 + 
                    637488\ M\^38 + 8491031\ M\^40 + 2050508\ M\^42 - 
                    7245308\ M\^44 - 6553719\ M\^46 + 2607011\ M\^48 + 
                    10775355\ M\^50 + 4079064\ M\^52 - 12184454\ M\^54 - 
                    9681845\ M\^56 + 10731201\ M\^58 + 12261588\ M\^60 - 
                    7760710\ M\^62 - 11344602\ M\^64 + 5157495\ M\^66 + 
                    8161481\ M\^68 - 3569124\ M\^70 - 4601955\ M\^72 + 
                    2525290\ M\^74 + 1884740\ M\^76 - 1540486\ M\^78 - 
                    421638\ M\^80 + 709097\ M\^82 - 82897\ M\^84 - 
                    187396\ M\^86 + 94190\ M\^88 + 9034\ M\^90 - 
                    29271\ M\^92 + 16578\ M\^94 - 5706\ M\^96 + 1359\ M\^98 - 
                    221\ M\^100 + 22\ M\^102 - M\^104)\) + 
              L\^10\ \((4 - 85\ M\^2 + 786\ M\^4 - 4124\ M\^6 + 13194\ M\^8 - 
                    24128\ M\^10 + 11219\ M\^12 + 60450\ M\^14 - 
                    139115\ M\^16 - 10015\ M\^18 + 483086\ M\^20 - 
                    472805\ M\^22 - 975707\ M\^24 + 1774970\ M\^26 + 
                    1581600\ M\^28 - 4238332\ M\^30 - 2543966\ M\^32 + 
                    7639657\ M\^34 + 4972515\ M\^36 - 10975190\ M\^38 - 
                    10175635\ M\^40 + 12346275\ M\^42 + 18046761\ M\^44 - 
                    10087391\ M\^46 - 25770109\ M\^48 + 3990718\ M\^50 + 
                    29175490\ M\^52 + 3990718\ M\^54 - 25770109\ M\^56 - 
                    10087391\ M\^58 + 18046761\ M\^60 + 12346275\ M\^62 - 
                    10175635\ M\^64 - 10975190\ M\^66 + 4972515\ M\^68 + 
                    7639657\ M\^70 - 2543966\ M\^72 - 4238332\ M\^74 + 
                    1581600\ M\^76 + 1774970\ M\^78 - 975707\ M\^80 - 
                    472805\ M\^82 + 483086\ M\^84 - 10015\ M\^86 - 
                    139115\ M\^88 + 60450\ M\^90 + 11219\ M\^92 - 
                    24128\ M\^94 + 13194\ M\^96 - 4124\ M\^98 + 786\ M\^100 - 
                    85\ M\^102 + 4\ M\^104)\) + 
              L\^12\ \((5\ M\^4 - 104\ M\^6 + 945\ M\^8 - 4850\ M\^10 + 
                    14883\ M\^12 - 24604\ M\^14 + 4482\ M\^16 + 
                    68296\ M\^18 - 113511\ M\^20 - 46306\ M\^22 + 
                    355469\ M\^24 - 265203\ M\^26 - 536627\ M\^28 + 
                    952854\ M\^30 + 437986\ M\^32 - 1942178\ M\^34 - 
                    78715\ M\^36 + 2940045\ M\^38 + 67646\ M\^40 - 
                    3509115\ M\^42 - 1023448\ M\^44 + 2907287\ M\^46 + 
                    2669110\ M\^48 - 826705\ M\^50 - 3891886\ M\^52 - 
                    1890015\ M\^54 + 4213328\ M\^56 + 4100108\ M\^58 - 
                    3501472\ M\^60 - 4572706\ M\^62 + 2418589\ M\^64 + 
                    3389843\ M\^66 - 1440577\ M\^68 - 1688435\ M\^70 + 
                    698627\ M\^72 + 456806\ M\^74 - 181286\ M\^76 + 
                    9365\ M\^78 - 60064\ M\^80 - 42537\ M\^82 + 
                    99358\ M\^84 - 25612\ M\^86 - 35707\ M\^88 + 
                    28384\ M\^90 - 106\ M\^92 - 11466\ M\^94 + 8273\ M\^96 - 
                    3073\ M\^98 + 664\ M\^100 - 79\ M\^102 + 4\ M\^104)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 20]\), \(L\^20\ M\^8 + M\^172 + 
              L\^19\ \((\(-2\)\ M\^4 + 8\ M\^6 - 15\ M\^8 + 14\ M\^10 + 
                    5\ M\^12 - 34\ M\^14 + 43\ M\^16 + 23\ M\^18 - 
                    31\ M\^20 + 9\ M\^22)\) + 
              L\^18\ \((1 - 8\ M\^2 + 28\ M\^4 - 50\ M\^6 + 39\ M\^8 + 
                    38\ M\^12 - 110\ M\^14 - 204\ M\^16 + 702\ M\^18 - 
                    149\ M\^20 - 1230\ M\^22 + 1107\ M\^24 + 916\ M\^26 - 
                    1182\ M\^28 + 14\ M\^30 + 476\ M\^32 - 234\ M\^34 + 
                    36\ M\^36)\) + 
              L\^17\ \((\(-1\) + 12\ M\^2 - 67\ M\^4 + 214\ M\^6 - 
                    374\ M\^8 + 138\ M\^10 + 824\ M\^12 - 1367\ M\^14 - 
                    787\ M\^16 + 3939\ M\^18 - 656\ M\^20 - 7875\ M\^22 + 
                    4153\ M\^24 + 12938\ M\^26 - 11584\ M\^28 - 
                    14875\ M\^30 + 20413\ M\^32 + 9914\ M\^34 - 
                    20622\ M\^36 + 525\ M\^38 + 11706\ M\^40 - 4980\ M\^42 - 
                    2235\ M\^44 + 2473\ M\^46 - 770\ M\^48 + 84\ M\^50)\) + 
              L\^16\ \((\(-2\)\ M\^2 + 27\ M\^4 - 169\ M\^6 + 615\ M\^8 - 
                    1279\ M\^10 + 863\ M\^12 + 2773\ M\^14 - 7402\ M\^16 + 
                    1837\ M\^18 + 18679\ M\^20 - 25460\ M\^22 - 
                    16802\ M\^24 + 65853\ M\^26 - 19401\ M\^28 - 
                    98552\ M\^30 + 91407\ M\^32 + 92222\ M\^34 - 
                    167047\ M\^36 - 46693\ M\^38 + 205177\ M\^40 - 
                    9291\ M\^42 - 175913\ M\^44 + 56909\ M\^46 + 
                    104015\ M\^48 - 67505\ M\^50 - 30477\ M\^52 + 
                    41018\ M\^54 - 5180\ M\^56 - 10219\ M\^58 + 6158\ M\^60 - 
                    1442\ M\^62 + 126\ M\^64)\) + 
              L\^15\ \((\(-M\^4\) + 16\ M\^6 - 122\ M\^8 + 560\ M\^10 - 
                    1610\ M\^12 + 2632\ M\^14 - 889\ M\^16 - 6192\ M\^18 + 
                    13344\ M\^20 - 3875\ M\^22 - 31487\ M\^24 + 
                    55751\ M\^26 + 5755\ M\^28 - 133945\ M\^30 + 
                    118442\ M\^32 + 167136\ M\^34 - 352271\ M\^36 - 
                    53136\ M\^38 + 615827\ M\^40 - 230912\ M\^42 - 
                    783319\ M\^44 + 559430\ M\^46 + 748925\ M\^48 - 
                    726860\ M\^50 - 506203\ M\^52 + 689092\ M\^54 + 
                    224235\ M\^56 - 499730\ M\^58 - 20526\ M\^60 + 
                    276803\ M\^62 - 69198\ M\^64 - 97742\ M\^66 + 
                    61087\ M\^68 + 7605\ M\^70 - 20309\ M\^72 + 8745\ M\^74 - 
                    1680\ M\^76 + 126\ M\^78)\) + 
              L\^14\ \((\(-3\)\ M\^10 + 53\ M\^12 - 428\ M\^14 + 
                    2006\ M\^16 - 5594\ M\^18 + 7554\ M\^20 + 4316\ M\^22 - 
                    34498\ M\^24 + 40647\ M\^26 + 46853\ M\^28 - 
                    172801\ M\^30 + 87344\ M\^32 + 291004\ M\^34 - 
                    458050\ M\^36 - 136502\ M\^38 + 939950\ M\^40 - 
                    554947\ M\^42 - 1131947\ M\^44 + 1703826\ M\^46 + 
                    744235\ M\^48 - 2828102\ M\^50 + 66136\ M\^52 + 
                    3318713\ M\^54 - 852102\ M\^56 - 2936777\ M\^58 + 
                    1340113\ M\^60 + 2063215\ M\^62 - 1338451\ M\^64 - 
                    1133176\ M\^66 + 996813\ M\^68 + 475904\ M\^70 - 
                    585220\ M\^72 - 142650\ M\^74 + 293850\ M\^76 - 
                    1652\ M\^78 - 116100\ M\^80 + 38076\ M\^82 + 
                    22649\ M\^84 - 21823\ M\^86 + 7488\ M\^88 - 1246\ M\^90 + 
                    84\ M\^92)\) + 
              L\^13\ \((\(-3\)\ M\^16 + 60\ M\^18 - 548\ M\^20 + 
                    2905\ M\^22 - 9283\ M\^24 + 15548\ M\^26 + 629\ M\^28 - 
                    59474\ M\^30 + 93852\ M\^32 + 60520\ M\^34 - 
                    354414\ M\^36 + 236365\ M\^38 + 620477\ M\^40 - 
                    1068551\ M\^42 - 440914\ M\^44 + 2414775\ M\^46 - 
                    796372\ M\^48 - 3746741\ M\^50 + 3273148\ M\^52 + 
                    4314079\ M\^54 - 6316982\ M\^56 - 3544267\ M\^58 + 
                    8506278\ M\^60 + 1493443\ M\^62 - 8765528\ M\^64 + 
                    964955\ M\^66 + 7257511\ M\^68 - 2433350\ M\^70 - 
                    4777956\ M\^72 + 2593222\ M\^74 + 2461674\ M\^76 - 
                    1903812\ M\^78 - 936862\ M\^80 + 1036489\ M\^82 + 
                    226752\ M\^84 - 402828\ M\^86 - 46102\ M\^88 + 
                    129166\ M\^90 + 17877\ M\^92 - 50226\ M\^94 + 
                    533\ M\^96 + 21375\ M\^98 - 13193\ M\^100 + 
                    3831\ M\^102 - 574\ M\^104 + 36\ M\^106)\) + 
              L\^12\ \((\(-M\^22\) + 26\ M\^24 - 298\ M\^26 + 1944\ M\^28 - 
                    7723\ M\^30 + 17762\ M\^32 - 14715\ M\^34 - 
                    35851\ M\^36 + 106222\ M\^38 - 18310\ M\^40 - 
                    313150\ M\^42 + 366851\ M\^44 + 513462\ M\^46 - 
                    1288107\ M\^48 - 369859\ M\^50 + 3101750\ M\^52 - 
                    911073\ M\^54 - 5617479\ M\^56 + 4100840\ M\^58 + 
                    7792666\ M\^60 - 9092187\ M\^62 - 7949728\ M\^64 + 
                    14305771\ M\^66 + 5224036\ M\^68 - 17627812\ M\^70 - 
                    711562\ M\^72 + 17609396\ M\^74 - 3172667\ M\^76 - 
                    14059318\ M\^78 + 5188012\ M\^80 + 8940304\ M\^82 - 
                    5049101\ M\^84 - 4436370\ M\^86 + 3604273\ M\^88 + 
                    1626692\ M\^90 - 1992730\ M\^92 - 353340\ M\^94 + 
                    874754\ M\^96 - 64788\ M\^98 - 258906\ M\^100 + 
                    82360\ M\^102 + 30903\ M\^104 - 19552\ M\^106 + 
                    7046\ M\^108 - 10486\ M\^110 + 9301\ M\^112 - 
                    4225\ M\^114 + 1078\ M\^116 - 150\ M\^118 + 
                    9\ M\^120)\) + 
              L\^11\ \((3\ M\^30 - 62\ M\^32 + 572\ M\^34 - 3028\ M\^36 + 
                    9665\ M\^38 - 16695\ M\^40 + 4392\ M\^42 + 38872\ M\^44 - 
                    42679\ M\^46 - 99336\ M\^48 + 202581\ M\^50 + 
                    203137\ M\^52 - 737770\ M\^54 - 215907\ M\^56 + 
                    2178681\ M\^58 - 618850\ M\^60 - 4673586\ M\^62 + 
                    3362901\ M\^64 + 7540357\ M\^66 - 8753205\ M\^68 - 
                    9144401\ M\^70 + 16199679\ M\^72 + 7997181\ M\^74 - 
                    23450730\ M\^76 - 3942666\ M\^78 + 27133213\ M\^80 - 
                    1531150\ M\^82 - 25127648\ M\^84 + 6459616\ M\^86 + 
                    19038088\ M\^88 - 8550767\ M\^90 - 11723612\ M\^92 + 
                    7635180\ M\^94 + 5656085\ M\^96 - 5131998\ M\^98 - 
                    1963451\ M\^100 + 2662361\ M\^102 + 378836\ M\^104 - 
                    1052939\ M\^106 + 1345\ M\^108 + 356848\ M\^110 - 
                    50705\ M\^112 - 106030\ M\^114 + 39814\ M\^116 + 
                    21560\ M\^118 - 23895\ M\^120 + 11142\ M\^122 - 
                    4267\ M\^124 + 1678\ M\^126 - 562\ M\^128 + 128\ M\^130 - 
                    17\ M\^132 + M\^134)\) + 
              L\^10\ \((\(-3\)\ M\^38 + 56\ M\^40 - 460\ M\^42 + 
                    2141\ M\^44 - 5967\ M\^46 + 9407\ M\^48 - 6900\ M\^50 + 
                    6172\ M\^52 - 28315\ M\^54 + 23804\ M\^56 + 
                    137684\ M\^58 - 270703\ M\^60 - 268494\ M\^62 + 
                    1094369\ M\^64 + 34228\ M\^66 - 2788526\ M\^68 + 
                    1215498\ M\^70 + 5522354\ M\^72 - 4597428\ M\^74 - 
                    8593732\ M\^76 + 10801040\ M\^78 + 10327507\ M\^80 - 
                    18827052\ M\^82 - 8824397\ M\^84 + 25626497\ M\^86 + 
                    3583926\ M\^88 - 28160656\ M\^90 + 3583926\ M\^92 + 
                    25626497\ M\^94 - 8824397\ M\^96 - 18827052\ M\^98 + 
                    10327507\ M\^100 + 10801040\ M\^102 - 8593732\ M\^104 - 
                    4597428\ M\^106 + 5522354\ M\^108 + 1215498\ M\^110 - 
                    2788526\ M\^112 + 34228\ M\^114 + 1094369\ M\^116 - 
                    268494\ M\^118 - 270703\ M\^120 + 137684\ M\^122 + 
                    23804\ M\^124 - 28315\ M\^126 + 6172\ M\^128 - 
                    6900\ M\^130 + 9407\ M\^132 - 5967\ M\^134 + 
                    2141\ M\^136 - 460\ M\^138 + 56\ M\^140 - 3\ M\^142)\) + 
              L\^9\ \((M\^46 - 17\ M\^48 + 128\ M\^50 - 562\ M\^52 + 
                    1678\ M\^54 - 4267\ M\^56 + 11142\ M\^58 - 23895\ M\^60 + 
                    21560\ M\^62 + 39814\ M\^64 - 106030\ M\^66 - 
                    50705\ M\^68 + 356848\ M\^70 + 1345\ M\^72 - 
                    1052939\ M\^74 + 378836\ M\^76 + 2662361\ M\^78 - 
                    1963451\ M\^80 - 5131998\ M\^82 + 5656085\ M\^84 + 
                    7635180\ M\^86 - 11723612\ M\^88 - 8550767\ M\^90 + 
                    19038088\ M\^92 + 6459616\ M\^94 - 25127648\ M\^96 - 
                    1531150\ M\^98 + 27133213\ M\^100 - 3942666\ M\^102 - 
                    23450730\ M\^104 + 7997181\ M\^106 + 16199679\ M\^108 - 
                    9144401\ M\^110 - 8753205\ M\^112 + 7540357\ M\^114 + 
                    3362901\ M\^116 - 4673586\ M\^118 - 618850\ M\^120 + 
                    2178681\ M\^122 - 215907\ M\^124 - 737770\ M\^126 + 
                    203137\ M\^128 + 202581\ M\^130 - 99336\ M\^132 - 
                    42679\ M\^134 + 38872\ M\^136 + 4392\ M\^138 - 
                    16695\ M\^140 + 9665\ M\^142 - 3028\ M\^144 + 
                    572\ M\^146 - 62\ M\^148 + 3\ M\^150)\) + 
              L\^8\ \((9\ M\^60 - 150\ M\^62 + 1078\ M\^64 - 4225\ M\^66 + 
                    9301\ M\^68 - 10486\ M\^70 + 7046\ M\^72 - 19552\ M\^74 + 
                    30903\ M\^76 + 82360\ M\^78 - 258906\ M\^80 - 
                    64788\ M\^82 + 874754\ M\^84 - 353340\ M\^86 - 
                    1992730\ M\^88 + 1626692\ M\^90 + 3604273\ M\^92 - 
                    4436370\ M\^94 - 5049101\ M\^96 + 8940304\ M\^98 + 
                    5188012\ M\^100 - 14059318\ M\^102 - 3172667\ M\^104 + 
                    17609396\ M\^106 - 711562\ M\^108 - 17627812\ M\^110 + 
                    5224036\ M\^112 + 14305771\ M\^114 - 7949728\ M\^116 - 
                    9092187\ M\^118 + 7792666\ M\^120 + 4100840\ M\^122 - 
                    5617479\ M\^124 - 911073\ M\^126 + 3101750\ M\^128 - 
                    369859\ M\^130 - 1288107\ M\^132 + 513462\ M\^134 + 
                    366851\ M\^136 - 313150\ M\^138 - 18310\ M\^140 + 
                    106222\ M\^142 - 35851\ M\^144 - 14715\ M\^146 + 
                    17762\ M\^148 - 7723\ M\^150 + 1944\ M\^152 - 
                    298\ M\^154 + 26\ M\^156 - M\^158)\) + 
              L\^7\ \((36\ M\^74 - 574\ M\^76 + 3831\ M\^78 - 13193\ M\^80 + 
                    21375\ M\^82 + 533\ M\^84 - 50226\ M\^86 + 17877\ M\^88 + 
                    129166\ M\^90 - 46102\ M\^92 - 402828\ M\^94 + 
                    226752\ M\^96 + 1036489\ M\^98 - 936862\ M\^100 - 
                    1903812\ M\^102 + 2461674\ M\^104 + 2593222\ M\^106 - 
                    4777956\ M\^108 - 2433350\ M\^110 + 7257511\ M\^112 + 
                    964955\ M\^114 - 8765528\ M\^116 + 1493443\ M\^118 + 
                    8506278\ M\^120 - 3544267\ M\^122 - 6316982\ M\^124 + 
                    4314079\ M\^126 + 3273148\ M\^128 - 3746741\ M\^130 - 
                    796372\ M\^132 + 2414775\ M\^134 - 440914\ M\^136 - 
                    1068551\ M\^138 + 620477\ M\^140 + 236365\ M\^142 - 
                    354414\ M\^144 + 60520\ M\^146 + 93852\ M\^148 - 
                    59474\ M\^150 + 629\ M\^152 + 15548\ M\^154 - 
                    9283\ M\^156 + 2905\ M\^158 - 548\ M\^160 + 60\ M\^162 - 
                    3\ M\^164)\) + 
              L\^6\ \((84\ M\^88 - 1246\ M\^90 + 7488\ M\^92 - 21823\ M\^94 + 
                    22649\ M\^96 + 38076\ M\^98 - 116100\ M\^100 - 
                    1652\ M\^102 + 293850\ M\^104 - 142650\ M\^106 - 
                    585220\ M\^108 + 475904\ M\^110 + 996813\ M\^112 - 
                    1133176\ M\^114 - 1338451\ M\^116 + 2063215\ M\^118 + 
                    1340113\ M\^120 - 2936777\ M\^122 - 852102\ M\^124 + 
                    3318713\ M\^126 + 66136\ M\^128 - 2828102\ M\^130 + 
                    744235\ M\^132 + 1703826\ M\^134 - 1131947\ M\^136 - 
                    554947\ M\^138 + 939950\ M\^140 - 136502\ M\^142 - 
                    458050\ M\^144 + 291004\ M\^146 + 87344\ M\^148 - 
                    172801\ M\^150 + 46853\ M\^152 + 40647\ M\^154 - 
                    34498\ M\^156 + 4316\ M\^158 + 7554\ M\^160 - 
                    5594\ M\^162 + 2006\ M\^164 - 428\ M\^166 + 53\ M\^168 - 
                    3\ M\^170)\) + 
              L\ \((9\ M\^158 - 31\ M\^160 + 23\ M\^162 + 43\ M\^164 - 
                    34\ M\^166 + 5\ M\^168 + 14\ M\^170 - 15\ M\^172 + 
                    8\ M\^174 - 2\ M\^176)\) + 
              L\^5\ \((126\ M\^102 - 1680\ M\^104 + 8745\ M\^106 - 
                    20309\ M\^108 + 7605\ M\^110 + 61087\ M\^112 - 
                    97742\ M\^114 - 69198\ M\^116 + 276803\ M\^118 - 
                    20526\ M\^120 - 499730\ M\^122 + 224235\ M\^124 + 
                    689092\ M\^126 - 506203\ M\^128 - 726860\ M\^130 + 
                    748925\ M\^132 + 559430\ M\^134 - 783319\ M\^136 - 
                    230912\ M\^138 + 615827\ M\^140 - 53136\ M\^142 - 
                    352271\ M\^144 + 167136\ M\^146 + 118442\ M\^148 - 
                    133945\ M\^150 + 5755\ M\^152 + 55751\ M\^154 - 
                    31487\ M\^156 - 3875\ M\^158 + 13344\ M\^160 - 
                    6192\ M\^162 - 889\ M\^164 + 2632\ M\^166 - 
                    1610\ M\^168 + 560\ M\^170 - 122\ M\^172 + 16\ M\^174 - 
                    M\^176)\) + 
              L\^4\ \((126\ M\^116 - 1442\ M\^118 + 6158\ M\^120 - 
                    10219\ M\^122 - 5180\ M\^124 + 41018\ M\^126 - 
                    30477\ M\^128 - 67505\ M\^130 + 104015\ M\^132 + 
                    56909\ M\^134 - 175913\ M\^136 - 9291\ M\^138 + 
                    205177\ M\^140 - 46693\ M\^142 - 167047\ M\^144 + 
                    92222\ M\^146 + 91407\ M\^148 - 98552\ M\^150 - 
                    19401\ M\^152 + 65853\ M\^154 - 16802\ M\^156 - 
                    25460\ M\^158 + 18679\ M\^160 + 1837\ M\^162 - 
                    7402\ M\^164 + 2773\ M\^166 + 863\ M\^168 - 
                    1279\ M\^170 + 615\ M\^172 - 169\ M\^174 + 27\ M\^176 - 
                    2\ M\^178)\) + 
              L\^3\ \((84\ M\^130 - 770\ M\^132 + 2473\ M\^134 - 
                    2235\ M\^136 - 4980\ M\^138 + 11706\ M\^140 + 
                    525\ M\^142 - 20622\ M\^144 + 9914\ M\^146 + 
                    20413\ M\^148 - 14875\ M\^150 - 11584\ M\^152 + 
                    12938\ M\^154 + 4153\ M\^156 - 7875\ M\^158 - 
                    656\ M\^160 + 3939\ M\^162 - 787\ M\^164 - 1367\ M\^166 + 
                    824\ M\^168 + 138\ M\^170 - 374\ M\^172 + 214\ M\^174 - 
                    67\ M\^176 + 12\ M\^178 - M\^180)\) + 
              L\^2\ \((36\ M\^144 - 234\ M\^146 + 476\ M\^148 + 14\ M\^150 - 
                    1182\ M\^152 + 916\ M\^154 + 1107\ M\^156 - 
                    1230\ M\^158 - 149\ M\^160 + 702\ M\^162 - 204\ M\^164 - 
                    110\ M\^166 + 38\ M\^168 + 39\ M\^172 - 50\ M\^174 + 
                    28\ M\^176 - 8\ M\^178 + M\^180)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 21]\), \(M\^24 + L\^21\ M\^110 + 
              L\ \((\(-6\)\ M\^20 + 22\ M\^22 - 27\ M\^24 - 6\ M\^26 + 
                    60\ M\^28 - 20\ M\^30 - 15\ M\^32 + 22\ M\^34 - 
                    12\ M\^36 + 3\ M\^38)\) + 
              L\^2\ \((15\ M\^16 - 106\ M\^18 + 295\ M\^20 - 264\ M\^22 - 
                    417\ M\^24 + 1042\ M\^26 - 307\ M\^28 - 1099\ M\^30 + 
                    1203\ M\^32 + 367\ M\^34 - 752\ M\^36 + 198\ M\^38 + 
                    27\ M\^40 - 7\ M\^42 + 75\ M\^44 - 109\ M\^46 + 
                    68\ M\^48 - 22\ M\^50 + 3\ M\^52)\) + 
              L\^3\ \((\(-20\)\ M\^12 + 204\ M\^14 - 857\ M\^16 + 
                    1586\ M\^18 - 142\ M\^20 - 4339\ M\^22 + 5415\ M\^24 + 
                    3393\ M\^26 - 11288\ M\^28 + 2146\ M\^30 + 10721\ M\^32 - 
                    4892\ M\^34 - 5674\ M\^36 + 3529\ M\^38 + 4944\ M\^40 - 
                    1168\ M\^42 - 5381\ M\^44 + 2203\ M\^46 + 3020\ M\^48 - 
                    2517\ M\^50 - 13\ M\^52 + 852\ M\^54 - 587\ M\^56 + 
                    287\ M\^58 - 128\ M\^60 + 45\ M\^62 - 10\ M\^64 + 
                    M\^66)\) + 
              L\^4\ \((15\ M\^8 - 196\ M\^10 + 1076\ M\^12 - 2924\ M\^14 + 
                    2909\ M\^16 + 4280\ M\^18 - 13012\ M\^20 + 1846\ M\^22 + 
                    25635\ M\^24 - 17248\ M\^26 - 36771\ M\^28 + 
                    41901\ M\^30 + 47070\ M\^32 - 87577\ M\^34 - 
                    40037\ M\^36 + 140177\ M\^38 + 2220\ M\^40 - 
                    147931\ M\^42 + 55830\ M\^44 + 100958\ M\^46 - 
                    77048\ M\^48 - 33013\ M\^50 + 52124\ M\^52 - 
                    5399\ M\^54 - 15438\ M\^56 + 7121\ M\^58 - 595\ M\^60 + 
                    34\ M\^62 + 642\ M\^64 - 1448\ M\^66 + 1207\ M\^68 - 
                    555\ M\^70 + 156\ M\^72 - 26\ M\^74 + 2\ M\^76)\) + 
              L\^5\ \((\(-6\)\ M\^4 + 94\ M\^6 - 627\ M\^8 + 2216\ M\^10 - 
                    3968\ M\^12 + 1332\ M\^14 + 7214\ M\^16 - 6205\ M\^18 - 
                    18764\ M\^20 + 25673\ M\^22 + 47248\ M\^24 - 
                    105331\ M\^26 - 58020\ M\^28 + 287858\ M\^30 - 
                    66150\ M\^32 - 463690\ M\^34 + 369208\ M\^36 + 
                    423934\ M\^38 - 716637\ M\^40 - 104624\ M\^42 + 
                    875298\ M\^44 - 261094\ M\^46 - 712694\ M\^48 + 
                    453103\ M\^50 + 404205\ M\^52 - 395995\ M\^54 - 
                    157337\ M\^56 + 236531\ M\^58 + 43891\ M\^60 - 
                    112477\ M\^62 - 9593\ M\^64 + 53188\ M\^66 - 
                    6861\ M\^68 - 20555\ M\^70 + 10033\ M\^72 + 3356\ M\^74 - 
                    5565\ M\^76 + 2874\ M\^78 - 857\ M\^80 + 160\ M\^82 - 
                    18\ M\^84 + M\^86)\) + 
              L\^6\ \((1 - 18\ M\^2 + 141\ M\^4 - 628\ M\^6 + 1737\ M\^8 - 
                    3010\ M\^10 + 3103\ M\^12 - 2565\ M\^14 + 6566\ M\^16 - 
                    11549\ M\^18 - 16673\ M\^20 + 82702\ M\^22 - 
                    39770\ M\^24 - 215836\ M\^26 + 293546\ M\^28 + 
                    266072\ M\^30 - 770286\ M\^32 - 54006\ M\^34 + 
                    1368935\ M\^36 - 538861\ M\^38 - 1771642\ M\^40 + 
                    1390712\ M\^42 + 1537611\ M\^44 - 2174474\ M\^46 - 
                    470483\ M\^48 + 2506444\ M\^50 - 864056\ M\^52 - 
                    2042808\ M\^54 + 1771817\ M\^56 + 1078193\ M\^58 - 
                    1826066\ M\^60 - 175904\ M\^62 + 1273749\ M\^64 - 
                    292659\ M\^66 - 611227\ M\^68 + 342234\ M\^70 + 
                    171587\ M\^72 - 209539\ M\^74 + 13929\ M\^76 + 
                    68982\ M\^78 - 34159\ M\^80 - 5652\ M\^82 + 
                    13560\ M\^84 - 7379\ M\^86 + 2291\ M\^88 - 448\ M\^90 + 
                    53\ M\^92 - 3\ M\^94)\) + 
              L\^7\ \((\(-1\) + 20\ M\^2 - 170\ M\^4 + 803\ M\^6 - 
                    2301\ M\^8 + 4141\ M\^10 - 5303\ M\^12 + 8005\ M\^14 - 
                    13224\ M\^16 - 2426\ M\^18 + 63377\ M\^20 - 
                    63931\ M\^22 - 160059\ M\^24 + 290364\ M\^26 + 
                    279413\ M\^28 - 772538\ M\^30 - 422973\ M\^32 + 
                    1752997\ M\^34 + 230945\ M\^36 - 3256298\ M\^38 + 
                    987965\ M\^40 + 4759530\ M\^42 - 3610723\ M\^44 - 
                    5157232\ M\^46 + 6668400\ M\^48 + 3593438\ M\^50 - 
                    8361157\ M\^52 - 309074\ M\^54 + 7824238\ M\^56 - 
                    2832418\ M\^58 - 5305179\ M\^60 + 4384949\ M\^62 + 
                    2235189\ M\^64 - 3987872\ M\^66 + 50618\ M\^68 + 
                    2475496\ M\^70 - 1013955\ M\^72 - 934510\ M\^74 + 
                    910935\ M\^76 + 46177\ M\^78 - 432645\ M\^80 + 
                    193767\ M\^82 + 68759\ M\^84 - 104259\ M\^86 + 
                    31316\ M\^88 + 16589\ M\^90 - 20455\ M\^92 + 
                    10011\ M\^94 - 2950\ M\^96 + 548\ M\^98 - 60\ M\^100 + 
                    3\ M\^102)\) + 
              L\^8\ \((6\ M\^4 - 123\ M\^6 + 1074\ M\^8 - 5112\ M\^10 + 
                    13596\ M\^12 - 15514\ M\^14 - 13974\ M\^16 + 
                    53844\ M\^18 + 19775\ M\^20 - 210554\ M\^22 + 
                    54076\ M\^24 + 680769\ M\^26 - 519433\ M\^28 - 
                    1778363\ M\^30 + 2299884\ M\^32 + 3256928\ M\^34 - 
                    6374551\ M\^36 - 3963718\ M\^38 + 12799911\ M\^40 + 
                    1789006\ M\^42 - 19457097\ M\^44 + 4968617\ M\^46 + 
                    22643005\ M\^48 - 15783207\ M\^50 - 19192852\ M\^52 + 
                    26549755\ M\^54 + 8978841\ M\^56 - 31465075\ M\^58 + 
                    4677775\ M\^60 + 27955543\ M\^62 - 15456503\ M\^64 - 
                    17825847\ M\^66 + 19068047\ M\^68 + 6444837\ M\^70 - 
                    15578776\ M\^72 + 1291426\ M\^74 + 8846522\ M\^76 - 
                    3876103\ M\^78 - 3050828\ M\^80 + 2956746\ M\^82 + 
                    139203\ M\^84 - 1223554\ M\^86 + 490810\ M\^88 + 
                    172343\ M\^90 - 230238\ M\^92 + 68877\ M\^94 + 
                    25008\ M\^96 - 34852\ M\^98 + 18664\ M\^100 - 
                    6427\ M\^102 + 1522\ M\^104 - 241\ M\^106 + 23\ M\^108 - 
                    M\^110)\) + 
              L\^20\ \((3\ M\^96 - 12\ M\^98 + 22\ M\^100 - 15\ M\^102 - 
                    20\ M\^104 + 60\ M\^106 - 6\ M\^108 - 27\ M\^110 + 
                    22\ M\^112 - 6\ M\^114)\) + 
              L\^9\ \((\(-15\)\ M\^8 + 315\ M\^10 - 2845\ M\^12 + 
                    14079\ M\^14 - 38673\ M\^16 + 41251\ M\^18 + 
                    76353\ M\^20 - 282857\ M\^22 + 85566\ M\^24 + 
                    907653\ M\^26 - 1200662\ M\^28 - 1667975\ M\^30 + 
                    4553857\ M\^32 + 1376783\ M\^34 - 11730899\ M\^36 + 
                    2928793\ M\^38 + 22713136\ M\^40 - 14821056\ M\^42 - 
                    33997257\ M\^44 + 35433535\ M\^46 + 38232951\ M\^48 - 
                    59626121\ M\^50 - 28435752\ M\^52 + 76814845\ M\^54 + 
                    4146514\ M\^56 - 77235077\ M\^58 + 25317629\ M\^60 + 
                    59280500\ M\^62 - 45495587\ M\^64 - 30103330\ M\^66 + 
                    48212313\ M\^68 + 2820538\ M\^70 - 35479509\ M\^72 + 
                    12785887\ M\^74 + 17118122\ M\^76 - 15057175\ M\^78 - 
                    2956030\ M\^80 + 9529776\ M\^82 - 3135144\ M\^84 - 
                    3182771\ M\^86 + 3166089\ M\^88 - 163586\ M\^90 - 
                    1328364\ M\^92 + 764186\ M\^94 + 111427\ M\^96 - 
                    320973\ M\^98 + 126553\ M\^100 + 29070\ M\^102 - 
                    52051\ M\^104 + 26296\ M\^106 - 7587\ M\^108 + 
                    1338\ M\^110 - 135\ M\^112 + 6\ M\^114)\) + 
              L\^10\ \((20\ M\^12 - 430\ M\^14 + 4008\ M\^16 - 20667\ M\^18 + 
                    60131\ M\^20 - 72757\ M\^22 - 110531\ M\^24 + 
                    511086\ M\^26 - 353095\ M\^28 - 1380503\ M\^30 + 
                    2779538\ M\^32 + 1300351\ M\^34 - 8704061\ M\^36 + 
                    3312858\ M\^38 + 17977948\ M\^40 - 17501368\ M\^42 - 
                    26441952\ M\^44 + 43327853\ M\^46 + 27236993\ M\^48 - 
                    75291675\ M\^50 - 15731535\ M\^52 + 99599561\ M\^54 - 
                    4555330\ M\^56 - 101499887\ M\^58 + 22266778\ M\^60 + 
                    76038880\ M\^62 - 25594626\ M\^64 - 33953695\ M\^66 + 
                    11771608\ M\^68 - 3526054\ M\^70 + 12116759\ M\^72 + 
                    21241953\ M\^74 - 32193780\ M\^76 - 18194228\ M\^78 + 
                    38524291\ M\^80 + 5231249\ M\^82 - 31020879\ M\^84 + 
                    5353914\ M\^86 + 17513715\ M\^88 - 8404096\ M\^90 - 
                    6204586\ M\^92 + 5963460\ M\^94 + 574052\ M\^96 - 
                    2528477\ M\^98 + 760618\ M\^100 + 501221\ M\^102 - 
                    420529\ M\^104 + 47942\ M\^106 + 79791\ M\^108 - 
                    53308\ M\^110 + 17040\ M\^112 - 3168\ M\^114 + 
                    330\ M\^116 - 15\ M\^118)\) + 
              L\^19\ \((3\ M\^82 - 22\ M\^84 + 68\ M\^86 - 109\ M\^88 + 
                    75\ M\^90 - 7\ M\^92 + 27\ M\^94 + 198\ M\^96 - 
                    752\ M\^98 + 367\ M\^100 + 1203\ M\^102 - 1099\ M\^104 - 
                    307\ M\^106 + 1042\ M\^108 - 417\ M\^110 - 264\ M\^112 + 
                    295\ M\^114 - 106\ M\^116 + 15\ M\^118)\) + 
              L\^18\ \((M\^68 - 10\ M\^70 + 45\ M\^72 - 128\ M\^74 + 
                    287\ M\^76 - 587\ M\^78 + 852\ M\^80 - 13\ M\^82 - 
                    2517\ M\^84 + 3020\ M\^86 + 2203\ M\^88 - 5381\ M\^90 - 
                    1168\ M\^92 + 4944\ M\^94 + 3529\ M\^96 - 5674\ M\^98 - 
                    4892\ M\^100 + 10721\ M\^102 + 2146\ M\^104 - 
                    11288\ M\^106 + 3393\ M\^108 + 5415\ M\^110 - 
                    4339\ M\^112 - 142\ M\^114 + 1586\ M\^116 - 857\ M\^118 + 
                    204\ M\^120 - 20\ M\^122)\) + 
              L\^11\ \((\(-15\)\ M\^16 + 330\ M\^18 - 3168\ M\^20 + 
                    17040\ M\^22 - 53308\ M\^24 + 79791\ M\^26 + 
                    47942\ M\^28 - 420529\ M\^30 + 501221\ M\^32 + 
                    760618\ M\^34 - 2528477\ M\^36 + 574052\ M\^38 + 
                    5963460\ M\^40 - 6204586\ M\^42 - 8404096\ M\^44 + 
                    17513715\ M\^46 + 5353914\ M\^48 - 31020879\ M\^50 + 
                    5231249\ M\^52 + 38524291\ M\^54 - 18194228\ M\^56 - 
                    32193780\ M\^58 + 21241953\ M\^60 + 12116759\ M\^62 - 
                    3526054\ M\^64 + 11771608\ M\^66 - 33953695\ M\^68 - 
                    25594626\ M\^70 + 76038880\ M\^72 + 22266778\ M\^74 - 
                    101499887\ M\^76 - 4555330\ M\^78 + 99599561\ M\^80 - 
                    15731535\ M\^82 - 75291675\ M\^84 + 27236993\ M\^86 + 
                    43327853\ M\^88 - 26441952\ M\^90 - 17501368\ M\^92 + 
                    17977948\ M\^94 + 3312858\ M\^96 - 8704061\ M\^98 + 
                    1300351\ M\^100 + 2779538\ M\^102 - 1380503\ M\^104 - 
                    353095\ M\^106 + 511086\ M\^108 - 110531\ M\^110 - 
                    72757\ M\^112 + 60131\ M\^114 - 20667\ M\^116 + 
                    4008\ M\^118 - 430\ M\^120 + 20\ M\^122)\) + 
              L\^12\ \((6\ M\^20 - 135\ M\^22 + 1338\ M\^24 - 7587\ M\^26 + 
                    26296\ M\^28 - 52051\ M\^30 + 29070\ M\^32 + 
                    126553\ M\^34 - 320973\ M\^36 + 111427\ M\^38 + 
                    764186\ M\^40 - 1328364\ M\^42 - 163586\ M\^44 + 
                    3166089\ M\^46 - 3182771\ M\^48 - 3135144\ M\^50 + 
                    9529776\ M\^52 - 2956030\ M\^54 - 15057175\ M\^56 + 
                    17118122\ M\^58 + 12785887\ M\^60 - 35479509\ M\^62 + 
                    2820538\ M\^64 + 48212313\ M\^66 - 30103330\ M\^68 - 
                    45495587\ M\^70 + 59280500\ M\^72 + 25317629\ M\^74 - 
                    77235077\ M\^76 + 4146514\ M\^78 + 76814845\ M\^80 - 
                    28435752\ M\^82 - 59626121\ M\^84 + 38232951\ M\^86 + 
                    35433535\ M\^88 - 33997257\ M\^90 - 14821056\ M\^92 + 
                    22713136\ M\^94 + 2928793\ M\^96 - 11730899\ M\^98 + 
                    1376783\ M\^100 + 4553857\ M\^102 - 1667975\ M\^104 - 
                    1200662\ M\^106 + 907653\ M\^108 + 85566\ M\^110 - 
                    282857\ M\^112 + 76353\ M\^114 + 41251\ M\^116 - 
                    38673\ M\^118 + 14079\ M\^120 - 2845\ M\^122 + 
                    315\ M\^124 - 15\ M\^126)\) + 
              L\^17\ \((2\ M\^58 - 26\ M\^60 + 156\ M\^62 - 555\ M\^64 + 
                    1207\ M\^66 - 1448\ M\^68 + 642\ M\^70 + 34\ M\^72 - 
                    595\ M\^74 + 7121\ M\^76 - 15438\ M\^78 - 5399\ M\^80 + 
                    52124\ M\^82 - 33013\ M\^84 - 77048\ M\^86 + 
                    100958\ M\^88 + 55830\ M\^90 - 147931\ M\^92 + 
                    2220\ M\^94 + 140177\ M\^96 - 40037\ M\^98 - 
                    87577\ M\^100 + 47070\ M\^102 + 41901\ M\^104 - 
                    36771\ M\^106 - 17248\ M\^108 + 25635\ M\^110 + 
                    1846\ M\^112 - 13012\ M\^114 + 4280\ M\^116 + 
                    2909\ M\^118 - 2924\ M\^120 + 1076\ M\^122 - 
                    196\ M\^124 + 15\ M\^126)\) + 
              L\^16\ \((M\^48 - 18\ M\^50 + 160\ M\^52 - 857\ M\^54 + 
                    2874\ M\^56 - 5565\ M\^58 + 3356\ M\^60 + 10033\ M\^62 - 
                    20555\ M\^64 - 6861\ M\^66 + 53188\ M\^68 - 9593\ M\^70 - 
                    112477\ M\^72 + 43891\ M\^74 + 236531\ M\^76 - 
                    157337\ M\^78 - 395995\ M\^80 + 404205\ M\^82 + 
                    453103\ M\^84 - 712694\ M\^86 - 261094\ M\^88 + 
                    875298\ M\^90 - 104624\ M\^92 - 716637\ M\^94 + 
                    423934\ M\^96 + 369208\ M\^98 - 463690\ M\^100 - 
                    66150\ M\^102 + 287858\ M\^104 - 58020\ M\^106 - 
                    105331\ M\^108 + 47248\ M\^110 + 25673\ M\^112 - 
                    18764\ M\^114 - 6205\ M\^116 + 7214\ M\^118 + 
                    1332\ M\^120 - 3968\ M\^122 + 2216\ M\^124 - 
                    627\ M\^126 + 94\ M\^128 - 6\ M\^130)\) + 
              L\^13\ \((\(-M\^24\) + 23\ M\^26 - 241\ M\^28 + 1522\ M\^30 - 
                    6427\ M\^32 + 18664\ M\^34 - 34852\ M\^36 + 
                    25008\ M\^38 + 68877\ M\^40 - 230238\ M\^42 + 
                    172343\ M\^44 + 490810\ M\^46 - 1223554\ M\^48 + 
                    139203\ M\^50 + 2956746\ M\^52 - 3050828\ M\^54 - 
                    3876103\ M\^56 + 8846522\ M\^58 + 1291426\ M\^60 - 
                    15578776\ M\^62 + 6444837\ M\^64 + 19068047\ M\^66 - 
                    17825847\ M\^68 - 15456503\ M\^70 + 27955543\ M\^72 + 
                    4677775\ M\^74 - 31465075\ M\^76 + 8978841\ M\^78 + 
                    26549755\ M\^80 - 19192852\ M\^82 - 15783207\ M\^84 + 
                    22643005\ M\^86 + 4968617\ M\^88 - 19457097\ M\^90 + 
                    1789006\ M\^92 + 12799911\ M\^94 - 3963718\ M\^96 - 
                    6374551\ M\^98 + 3256928\ M\^100 + 2299884\ M\^102 - 
                    1778363\ M\^104 - 519433\ M\^106 + 680769\ M\^108 + 
                    54076\ M\^110 - 210554\ M\^112 + 19775\ M\^114 + 
                    53844\ M\^116 - 13974\ M\^118 - 15514\ M\^120 + 
                    13596\ M\^122 - 5112\ M\^124 + 1074\ M\^126 - 
                    123\ M\^128 + 6\ M\^130)\) + 
              L\^14\ \((3\ M\^32 - 60\ M\^34 + 548\ M\^36 - 2950\ M\^38 + 
                    10011\ M\^40 - 20455\ M\^42 + 16589\ M\^44 + 
                    31316\ M\^46 - 104259\ M\^48 + 68759\ M\^50 + 
                    193767\ M\^52 - 432645\ M\^54 + 46177\ M\^56 + 
                    910935\ M\^58 - 934510\ M\^60 - 1013955\ M\^62 + 
                    2475496\ M\^64 + 50618\ M\^66 - 3987872\ M\^68 + 
                    2235189\ M\^70 + 4384949\ M\^72 - 5305179\ M\^74 - 
                    2832418\ M\^76 + 7824238\ M\^78 - 309074\ M\^80 - 
                    8361157\ M\^82 + 3593438\ M\^84 + 6668400\ M\^86 - 
                    5157232\ M\^88 - 3610723\ M\^90 + 4759530\ M\^92 + 
                    987965\ M\^94 - 3256298\ M\^96 + 230945\ M\^98 + 
                    1752997\ M\^100 - 422973\ M\^102 - 772538\ M\^104 + 
                    279413\ M\^106 + 290364\ M\^108 - 160059\ M\^110 - 
                    63931\ M\^112 + 63377\ M\^114 - 2426\ M\^116 - 
                    13224\ M\^118 + 8005\ M\^120 - 5303\ M\^122 + 
                    4141\ M\^124 - 2301\ M\^126 + 803\ M\^128 - 170\ M\^130 + 
                    20\ M\^132 - M\^134)\) + 
              L\^15\ \((\(-3\)\ M\^40 + 53\ M\^42 - 448\ M\^44 + 
                    2291\ M\^46 - 7379\ M\^48 + 13560\ M\^50 - 5652\ M\^52 - 
                    34159\ M\^54 + 68982\ M\^56 + 13929\ M\^58 - 
                    209539\ M\^60 + 171587\ M\^62 + 342234\ M\^64 - 
                    611227\ M\^66 - 292659\ M\^68 + 1273749\ M\^70 - 
                    175904\ M\^72 - 1826066\ M\^74 + 1078193\ M\^76 + 
                    1771817\ M\^78 - 2042808\ M\^80 - 864056\ M\^82 + 
                    2506444\ M\^84 - 470483\ M\^86 - 2174474\ M\^88 + 
                    1537611\ M\^90 + 1390712\ M\^92 - 1771642\ M\^94 - 
                    538861\ M\^96 + 1368935\ M\^98 - 54006\ M\^100 - 
                    770286\ M\^102 + 266072\ M\^104 + 293546\ M\^106 - 
                    215836\ M\^108 - 39770\ M\^110 + 82702\ M\^112 - 
                    16673\ M\^114 - 11549\ M\^116 + 6566\ M\^118 - 
                    2565\ M\^120 + 3103\ M\^122 - 3010\ M\^124 + 
                    1737\ M\^126 - 628\ M\^128 + 141\ M\^130 - 18\ M\^132 + 
                    M\^134)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 22]\), \(M\^28 + L\^25\ M\^134 + 
              L\ \((\(-2\)\ M\^20 + 9\ M\^22 - 24\ M\^24 + 40\ M\^26 - 
                    16\ M\^28 - 47\ M\^30 + 68\ M\^32 + 11\ M\^34 - 
                    26\ M\^36 + 8\ M\^38)\) + 
              L\^2\ \((M\^12 - 8\ M\^14 + 32\ M\^16 - 82\ M\^18 + 
                    152\ M\^20 - 243\ M\^22 + 335\ M\^24 - 126\ M\^26 - 
                    735\ M\^28 + 1328\ M\^30 - 1937\ M\^34 + 1319\ M\^36 + 
                    1137\ M\^38 - 1283\ M\^40 + 109\ M\^42 + 361\ M\^44 - 
                    180\ M\^46 + 28\ M\^48)\) + 
              L\^3\ \((\(-2\)\ M\^8 + 23\ M\^10 - 123\ M\^12 + 399\ M\^14 - 
                    828\ M\^16 + 1012\ M\^18 - 490\ M\^20 - 235\ M\^22 + 
                    429\ M\^24 - 1487\ M\^26 + 3575\ M\^28 + 178\ M\^30 - 
                    11608\ M\^32 + 11946\ M\^34 + 11052\ M\^36 - 
                    26661\ M\^38 + 4374\ M\^40 + 25234\ M\^42 - 
                    15909\ M\^44 - 8090\ M\^46 + 12496\ M\^48 - 2978\ M\^50 - 
                    2471\ M\^52 + 1933\ M\^54 - 537\ M\^56 + 56\ M\^58)\) + 
              L\^4\ \((M\^4 - 17\ M\^6 + 133\ M\^8 - 628\ M\^10 + 
                    1930\ M\^12 - 3729\ M\^14 + 3468\ M\^16 + 1944\ M\^18 - 
                    8448\ M\^20 + 3683\ M\^22 + 13348\ M\^24 - 20335\ M\^26 + 
                    4366\ M\^28 + 11673\ M\^30 - 11744\ M\^32 + 
                    20850\ M\^34 - 37567\ M\^36 - 5534\ M\^38 + 
                    107609\ M\^40 - 103034\ M\^42 - 86042\ M\^44 + 
                    205199\ M\^46 - 34388\ M\^48 - 157645\ M\^50 + 
                    112836\ M\^52 + 33525\ M\^54 - 73821\ M\^56 + 
                    26546\ M\^58 + 9521\ M\^60 - 12020\ M\^62 + 4714\ M\^64 - 
                    898\ M\^66 + 70\ M\^68)\) + 
              L\^5\ \((2\ M\^2 - 35\ M\^4 + 284\ M\^6 - 1389\ M\^8 + 
                    4384\ M\^10 - 8607\ M\^12 + 7975\ M\^14 + 5096\ M\^16 - 
                    20341\ M\^18 + 3526\ M\^20 + 53109\ M\^22 - 
                    78424\ M\^24 - 6415\ M\^26 + 144019\ M\^28 - 
                    146120\ M\^30 - 88381\ M\^32 + 334586\ M\^34 - 
                    184643\ M\^36 - 292295\ M\^38 + 426540\ M\^40 - 
                    27221\ M\^42 - 216900\ M\^44 + 128228\ M\^46 - 
                    280321\ M\^48 + 329060\ M\^50 + 374248\ M\^52 - 
                    797166\ M\^54 + 126996\ M\^56 + 619616\ M\^58 - 
                    463796\ M\^60 - 91035\ M\^62 + 284777\ M\^64 - 
                    131261\ M\^66 - 20021\ M\^68 + 48491\ M\^70 - 
                    24430\ M\^72 + 6433\ M\^74 - 915\ M\^76 + 56\ M\^78)\) + 
              L\^6\ \((1 - 21\ M\^2 + 197\ M\^4 - 1096\ M\^6 + 3994\ M\^8 - 
                    9845\ M\^10 + 16180\ M\^12 - 16107\ M\^14 + 4822\ M\^16 + 
                    13171\ M\^18 - 21814\ M\^20 - 14260\ M\^22 + 
                    114324\ M\^24 - 152073\ M\^26 - 79473\ M\^28 + 
                    361054\ M\^30 - 46003\ M\^32 - 622214\ M\^34 + 
                    266403\ M\^36 + 943209\ M\^38 - 665902\ M\^40 - 
                    1181652\ M\^42 + 1365891\ M\^44 + 790090\ M\^46 - 
                    1775945\ M\^48 + 381259\ M\^50 + 696190\ M\^52 - 
                    869770\ M\^54 + 1309474\ M\^56 - 456377\ M\^58 - 
                    1756299\ M\^60 + 2139694\ M\^62 + 98971\ M\^64 - 
                    1822549\ M\^66 + 1196011\ M\^68 + 271340\ M\^70 - 
                    786048\ M\^72 + 398895\ M\^74 + 25426\ M\^76 - 
                    137200\ M\^78 + 82810\ M\^80 - 26942\ M\^82 + 
                    5236\ M\^84 - 576\ M\^86 + 28\ M\^88)\) + 
              L\^7\ \((\(-1\) + 19\ M\^2 - 175\ M\^4 + 1034\ M\^6 - 
                    4358\ M\^8 + 13595\ M\^10 - 30774\ M\^12 + 45432\ M\^14 - 
                    27076\ M\^16 - 36847\ M\^18 + 63819\ M\^20 + 
                    65549\ M\^22 - 237827\ M\^24 + 118632\ M\^26 + 
                    235667\ M\^28 - 221940\ M\^30 - 279766\ M\^32 + 
                    429601\ M\^34 + 194954\ M\^36 - 612745\ M\^38 + 
                    73656\ M\^40 + 501976\ M\^42 - 412930\ M\^44 + 
                    22119\ M\^46 + 757913\ M\^48 - 843005\ M\^50 - 
                    1041384\ M\^52 + 2096735\ M\^54 + 595250\ M\^56 - 
                    3977363\ M\^58 + 2199848\ M\^60 + 4174879\ M\^62 - 
                    5984919\ M\^64 - 303081\ M\^66 + 6241789\ M\^68 - 
                    4323699\ M\^70 - 1731758\ M\^72 + 4310191\ M\^74 - 
                    2054069\ M\^76 - 915166\ M\^78 + 1671325\ M\^80 - 
                    801407\ M\^82 - 57035\ M\^84 + 298097\ M\^86 - 
                    195556\ M\^88 + 72465\ M\^90 - 17076\ M\^92 + 
                    2531\ M\^94 - 215\ M\^96 + 8\ M\^98)\) + 
              L\^8\ \((\(-M\^2\) + 23\ M\^4 - 251\ M\^6 + 1695\ M\^8 - 
                    7816\ M\^10 + 25704\ M\^12 - 60703\ M\^14 + 
                    97601\ M\^16 - 79865\ M\^18 - 55665\ M\^20 + 
                    217384\ M\^22 - 40315\ M\^24 - 611340\ M\^26 + 
                    761478\ M\^28 + 594231\ M\^30 - 1569362\ M\^32 - 
                    495728\ M\^34 + 2445191\ M\^36 + 263577\ M\^38 - 
                    2716786\ M\^40 - 487811\ M\^42 + 2501272\ M\^44 + 
                    1398782\ M\^46 - 3262977\ M\^48 - 1490473\ M\^50 + 
                    5468623\ M\^52 - 893825\ M\^54 - 6929778\ M\^56 + 
                    5576502\ M\^58 + 4753465\ M\^60 - 10194397\ M\^62 + 
                    732190\ M\^64 + 12813186\ M\^66 - 8036586\ M\^68 - 
                    10653964\ M\^70 + 14405370\ M\^72 + 2353531\ M\^74 - 
                    13996581\ M\^76 + 6736215\ M\^78 + 5530863\ M\^80 - 
                    7906774\ M\^82 + 2296509\ M\^84 + 2409995\ M\^86 - 
                    2807324\ M\^88 + 1077300\ M\^90 + 232216\ M\^92 - 
                    528640\ M\^94 + 337512\ M\^96 - 131537\ M\^98 + 
                    34458\ M\^100 - 6064\ M\^102 + 678\ M\^104 - 42\ M\^106 + 
                    M\^108)\) + 
              L\^9\ \((4\ M\^6 - 90\ M\^8 + 943\ M\^10 - 6019\ M\^12 + 
                    25728\ M\^14 - 76057\ M\^16 + 153329\ M\^18 - 
                    190139\ M\^20 + 69364\ M\^22 + 226206\ M\^24 - 
                    413169\ M\^26 - 74030\ M\^28 + 1333130\ M\^30 - 
                    1546574\ M\^32 - 1665696\ M\^34 + 4480840\ M\^36 + 
                    1606800\ M\^38 - 9472492\ M\^40 - 879640\ M\^42 + 
                    15731138\ M\^44 + 890\ M\^46 - 20315751\ M\^48 + 
                    9733\ M\^50 + 19059442\ M\^52 + 853574\ M\^54 - 
                    8607788\ M\^56 - 4320907\ M\^58 - 7104557\ M\^60 + 
                    13843246\ M\^62 + 17266697\ M\^64 - 28715692\ M\^66 - 
                    13599994\ M\^68 + 39532551\ M\^70 - 329299\ M\^72 - 
                    38849760\ M\^74 + 15104034\ M\^76 + 27686481\ M\^78 - 
                    23909238\ M\^80 - 10420025\ M\^82 + 22556768\ M\^84 - 
                    5329553\ M\^86 - 10992940\ M\^88 + 10044237\ M\^90 - 
                    837523\ M\^92 - 4324539\ M\^94 + 3497345\ M\^96 - 
                    861619\ M\^98 - 611966\ M\^100 + 761239\ M\^102 - 
                    435057\ M\^104 + 165100\ M\^106 - 44431\ M\^108 + 
                    8446\ M\^110 - 1082\ M\^112 + 84\ M\^114 - 3\ M\^116)\) + 
              L\^10\ \((\(-6\)\ M\^10 + 140\ M\^12 - 1518\ M\^14 + 
                    10020\ M\^16 - 44244\ M\^18 + 134516\ M\^20 - 
                    275067\ M\^22 + 331952\ M\^24 - 81646\ M\^26 - 
                    431191\ M\^28 + 611465\ M\^30 + 132926\ M\^32 - 
                    1443817\ M\^34 + 1837701\ M\^36 + 635619\ M\^38 - 
                    4807917\ M\^40 + 1933057\ M\^42 + 10059940\ M\^44 - 
                    6361612\ M\^46 - 19688960\ M\^48 + 10247868\ M\^50 + 
                    36125943\ M\^52 - 6937709\ M\^54 - 61464977\ M\^56 - 
                    7617695\ M\^58 + 91233746\ M\^60 + 27749632\ M\^62 - 
                    115197666\ M\^64 - 35497741\ M\^66 + 123482150\ M\^68 + 
                    16120943\ M\^70 - 110444136\ M\^72 + 22566446\ M\^74 + 
                    75229425\ M\^76 - 53832076\ M\^78 - 30126120\ M\^80 + 
                    59193186\ M\^82 - 5526061\ M\^84 - 43382689\ M\^86 + 
                    23214443\ M\^88 + 20498604\ M\^90 - 24945037\ M\^92 - 
                    577785\ M\^94 + 14944043\ M\^96 - 8371158\ M\^98 - 
                    2272764\ M\^100 + 5408322\ M\^102 - 2879228\ M\^104 + 
                    35165\ M\^106 + 1007080\ M\^108 - 833254\ M\^110 + 
                    409689\ M\^112 - 142368\ M\^114 + 36421\ M\^116 - 
                    6827\ M\^118 + 899\ M\^120 - 75\ M\^122 + 3\ M\^124)\) + 
              L\^11\ \((4\ M\^14 - 99\ M\^16 + 1156\ M\^18 - 8344\ M\^20 + 
                    41044\ M\^22 - 142584\ M\^24 + 346915\ M\^26 - 
                    546245\ M\^28 + 376027\ M\^30 + 454883\ M\^32 - 
                    1301801\ M\^34 + 573889\ M\^36 + 1960928\ M\^38 - 
                    3135396\ M\^40 - 518936\ M\^42 + 6303932\ M\^44 - 
                    3942332\ M\^46 - 11786556\ M\^48 + 16918160\ M\^50 + 
                    20817379\ M\^52 - 47515561\ M\^54 - 32071165\ M\^56 + 
                    99264696\ M\^58 + 45065654\ M\^60 - 162541347\ M\^62 - 
                    55797325\ M\^64 + 208444229\ M\^66 + 54673735\ M\^68 - 
                    203851641\ M\^70 - 31011357\ M\^72 + 141492964\ M\^74 - 
                    11603704\ M\^76 - 49543232\ M\^78 + 48259283\ M\^80 - 
                    33028758\ M\^82 - 54453428\ M\^84 + 77607765\ M\^86 + 
                    31546279\ M\^88 - 78709190\ M\^90 - 3094640\ M\^92 + 
                    55166838\ M\^94 - 13535866\ M\^96 - 28897512\ M\^98 + 
                    17984847\ M\^100 + 8572318\ M\^102 - 13525093\ M\^104 + 
                    2891999\ M\^106 + 4706068\ M\^108 - 4340253\ M\^110 + 
                    1094320\ M\^112 + 848567\ M\^114 - 1074811\ M\^116 + 
                    643441\ M\^118 - 263168\ M\^120 + 79472\ M\^122 - 
                    17991\ M\^124 + 3009\ M\^126 - 356\ M\^128 + 27\ M\^130 - 
                    M\^132)\) + 
              L\^12\ \((\(-M\^18\) + 29\ M\^20 - 398\ M\^22 + 3395\ M\^24 - 
                    19966\ M\^26 + 84771\ M\^28 - 262834\ M\^30 + 
                    579232\ M\^32 - 808479\ M\^34 + 345715\ M\^36 + 
                    1178010\ M\^38 - 2480234\ M\^40 + 743258\ M\^42 + 
                    4141222\ M\^44 - 5677615\ M\^46 - 2578781\ M\^48 + 
                    12756637\ M\^50 - 4345522\ M\^52 - 23111746\ M\^54 + 
                    22777104\ M\^56 + 36495502\ M\^58 - 59417213\ M\^60 - 
                    49301882\ M\^62 + 106834958\ M\^64 + 64373039\ M\^66 - 
                    141305229\ M\^68 - 87285577\ M\^70 + 132037226\ M\^72 + 
                    114012421\ M\^74 - 64182042\ M\^76 - 128358410\ M\^78 - 
                    37217197\ M\^80 + 119105088\ M\^82 + 119235455\ M\^84 - 
                    96328291\ M\^86 - 147736636\ M\^88 + 76880058\ M\^90 + 
                    125484285\ M\^92 - 64747803\ M\^94 - 78170598\ M\^96 + 
                    52021562\ M\^98 + 34738105\ M\^100 - 35715239\ M\^102 - 
                    9220701\ M\^104 + 21300051\ M\^106 - 2401163\ M\^108 - 
                    10206125\ M\^110 + 5940929\ M\^112 + 1975964\ M\^114 - 
                    3965007\ M\^116 + 1663479\ M\^118 + 493234\ M\^120 - 
                    1042932\ M\^122 + 714470\ M\^124 - 317425\ M\^126 + 
                    102961\ M\^128 - 25176\ M\^130 + 4602\ M\^132 - 
                    600\ M\^134 + 50\ M\^136 - 2\ M\^138)\) + 
              L\^24\ \((8\ M\^124 - 26\ M\^126 + 11\ M\^128 + 68\ M\^130 - 
                    47\ M\^132 - 16\ M\^134 + 40\ M\^136 - 24\ M\^138 + 
                    9\ M\^140 - 2\ M\^142)\) + 
              L\^13\ \((\(-2\)\ M\^24 + 50\ M\^26 - 600\ M\^28 + 
                    4602\ M\^30 - 25176\ M\^32 + 102961\ M\^34 - 
                    317425\ M\^36 + 714470\ M\^38 - 1042932\ M\^40 + 
                    493234\ M\^42 + 1663479\ M\^44 - 3965007\ M\^46 + 
                    1975964\ M\^48 + 5940929\ M\^50 - 10206125\ M\^52 - 
                    2401163\ M\^54 + 21300051\ M\^56 - 9220701\ M\^58 - 
                    35715239\ M\^60 + 34738105\ M\^62 + 52021562\ M\^64 - 
                    78170598\ M\^66 - 64747803\ M\^68 + 125484285\ M\^70 + 
                    76880058\ M\^72 - 147736636\ M\^74 - 96328291\ M\^76 + 
                    119235455\ M\^78 + 119105088\ M\^80 - 37217197\ M\^82 - 
                    128358410\ M\^84 - 64182042\ M\^86 + 114012421\ M\^88 + 
                    132037226\ M\^90 - 87285577\ M\^92 - 141305229\ M\^94 + 
                    64373039\ M\^96 + 106834958\ M\^98 - 49301882\ M\^100 - 
                    59417213\ M\^102 + 36495502\ M\^104 + 22777104\ M\^106 - 
                    23111746\ M\^108 - 4345522\ M\^110 + 12756637\ M\^112 - 
                    2578781\ M\^114 - 5677615\ M\^116 + 4141222\ M\^118 + 
                    743258\ M\^120 - 2480234\ M\^122 + 1178010\ M\^124 + 
                    345715\ M\^126 - 808479\ M\^128 + 579232\ M\^130 - 
                    262834\ M\^132 + 84771\ M\^134 - 19966\ M\^136 + 
                    3395\ M\^138 - 398\ M\^140 + 29\ M\^142 - M\^144)\) + 
              L\^14\ \((\(-M\^30\) + 27\ M\^32 - 356\ M\^34 + 3009\ M\^36 - 
                    17991\ M\^38 + 79472\ M\^40 - 263168\ M\^42 + 
                    643441\ M\^44 - 1074811\ M\^46 + 848567\ M\^48 + 
                    1094320\ M\^50 - 4340253\ M\^52 + 4706068\ M\^54 + 
                    2891999\ M\^56 - 13525093\ M\^58 + 8572318\ M\^60 + 
                    17984847\ M\^62 - 28897512\ M\^64 - 13535866\ M\^66 + 
                    55166838\ M\^68 - 3094640\ M\^70 - 78709190\ M\^72 + 
                    31546279\ M\^74 + 77607765\ M\^76 - 54453428\ M\^78 - 
                    33028758\ M\^80 + 48259283\ M\^82 - 49543232\ M\^84 - 
                    11603704\ M\^86 + 141492964\ M\^88 - 31011357\ M\^90 - 
                    203851641\ M\^92 + 54673735\ M\^94 + 208444229\ M\^96 - 
                    55797325\ M\^98 - 162541347\ M\^100 + 45065654\ M\^102 + 
                    99264696\ M\^104 - 32071165\ M\^106 - 47515561\ M\^108 + 
                    20817379\ M\^110 + 16918160\ M\^112 - 11786556\ M\^114 - 
                    3942332\ M\^116 + 6303932\ M\^118 - 518936\ M\^120 - 
                    3135396\ M\^122 + 1960928\ M\^124 + 573889\ M\^126 - 
                    1301801\ M\^128 + 454883\ M\^130 + 376027\ M\^132 - 
                    546245\ M\^134 + 346915\ M\^136 - 142584\ M\^138 + 
                    41044\ M\^140 - 8344\ M\^142 + 1156\ M\^144 - 
                    99\ M\^146 + 4\ M\^148)\) + 
              L\^23\ \((28\ M\^114 - 180\ M\^116 + 361\ M\^118 + 
                    109\ M\^120 - 1283\ M\^122 + 1137\ M\^124 + 
                    1319\ M\^126 - 1937\ M\^128 + 1328\ M\^132 - 
                    735\ M\^134 - 126\ M\^136 + 335\ M\^138 - 243\ M\^140 + 
                    152\ M\^142 - 82\ M\^144 + 32\ M\^146 - 8\ M\^148 + 
                    M\^150)\) + 
              L\^15\ \((3\ M\^38 - 75\ M\^40 + 899\ M\^42 - 6827\ M\^44 + 
                    36421\ M\^46 - 142368\ M\^48 + 409689\ M\^50 - 
                    833254\ M\^52 + 1007080\ M\^54 + 35165\ M\^56 - 
                    2879228\ M\^58 + 5408322\ M\^60 - 2272764\ M\^62 - 
                    8371158\ M\^64 + 14944043\ M\^66 - 577785\ M\^68 - 
                    24945037\ M\^70 + 20498604\ M\^72 + 23214443\ M\^74 - 
                    43382689\ M\^76 - 5526061\ M\^78 + 59193186\ M\^80 - 
                    30126120\ M\^82 - 53832076\ M\^84 + 75229425\ M\^86 + 
                    22566446\ M\^88 - 110444136\ M\^90 + 16120943\ M\^92 + 
                    123482150\ M\^94 - 35497741\ M\^96 - 115197666\ M\^98 + 
                    27749632\ M\^100 + 91233746\ M\^102 - 7617695\ M\^104 - 
                    61464977\ M\^106 - 6937709\ M\^108 + 36125943\ M\^110 + 
                    10247868\ M\^112 - 19688960\ M\^114 - 6361612\ M\^116 + 
                    10059940\ M\^118 + 1933057\ M\^120 - 4807917\ M\^122 + 
                    635619\ M\^124 + 1837701\ M\^126 - 1443817\ M\^128 + 
                    132926\ M\^130 + 611465\ M\^132 - 431191\ M\^134 - 
                    81646\ M\^136 + 331952\ M\^138 - 275067\ M\^140 + 
                    134516\ M\^142 - 44244\ M\^144 + 10020\ M\^146 - 
                    1518\ M\^148 + 140\ M\^150 - 6\ M\^152)\) + 
              L\^22\ \((56\ M\^104 - 537\ M\^106 + 1933\ M\^108 - 
                    2471\ M\^110 - 2978\ M\^112 + 12496\ M\^114 - 
                    8090\ M\^116 - 15909\ M\^118 + 25234\ M\^120 + 
                    4374\ M\^122 - 26661\ M\^124 + 11052\ M\^126 + 
                    11946\ M\^128 - 11608\ M\^130 + 178\ M\^132 + 
                    3575\ M\^134 - 1487\ M\^136 + 429\ M\^138 - 235\ M\^140 - 
                    490\ M\^142 + 1012\ M\^144 - 828\ M\^146 + 399\ M\^148 - 
                    123\ M\^150 + 23\ M\^152 - 2\ M\^154)\) + 
              L\^16\ \((\(-3\)\ M\^46 + 84\ M\^48 - 1082\ M\^50 + 
                    8446\ M\^52 - 44431\ M\^54 + 165100\ M\^56 - 
                    435057\ M\^58 + 761239\ M\^60 - 611966\ M\^62 - 
                    861619\ M\^64 + 3497345\ M\^66 - 4324539\ M\^68 - 
                    837523\ M\^70 + 10044237\ M\^72 - 10992940\ M\^74 - 
                    5329553\ M\^76 + 22556768\ M\^78 - 10420025\ M\^80 - 
                    23909238\ M\^82 + 27686481\ M\^84 + 15104034\ M\^86 - 
                    38849760\ M\^88 - 329299\ M\^90 + 39532551\ M\^92 - 
                    13599994\ M\^94 - 28715692\ M\^96 + 17266697\ M\^98 + 
                    13843246\ M\^100 - 7104557\ M\^102 - 4320907\ M\^104 - 
                    8607788\ M\^106 + 853574\ M\^108 + 19059442\ M\^110 + 
                    9733\ M\^112 - 20315751\ M\^114 + 890\ M\^116 + 
                    15731138\ M\^118 - 879640\ M\^120 - 9472492\ M\^122 + 
                    1606800\ M\^124 + 4480840\ M\^126 - 1665696\ M\^128 - 
                    1546574\ M\^130 + 1333130\ M\^132 - 74030\ M\^134 - 
                    413169\ M\^136 + 226206\ M\^138 + 69364\ M\^140 - 
                    190139\ M\^142 + 153329\ M\^144 - 76057\ M\^146 + 
                    25728\ M\^148 - 6019\ M\^150 + 943\ M\^152 - 90\ M\^154 + 
                    4\ M\^156)\) + 
              L\^21\ \((70\ M\^94 - 898\ M\^96 + 4714\ M\^98 - 
                    12020\ M\^100 + 9521\ M\^102 + 26546\ M\^104 - 
                    73821\ M\^106 + 33525\ M\^108 + 112836\ M\^110 - 
                    157645\ M\^112 - 34388\ M\^114 + 205199\ M\^116 - 
                    86042\ M\^118 - 103034\ M\^120 + 107609\ M\^122 - 
                    5534\ M\^124 - 37567\ M\^126 + 20850\ M\^128 - 
                    11744\ M\^130 + 11673\ M\^132 + 4366\ M\^134 - 
                    20335\ M\^136 + 13348\ M\^138 + 3683\ M\^140 - 
                    8448\ M\^142 + 1944\ M\^144 + 3468\ M\^146 - 
                    3729\ M\^148 + 1930\ M\^150 - 628\ M\^152 + 133\ M\^154 - 
                    17\ M\^156 + M\^158)\) + 
              L\^17\ \((M\^54 - 42\ M\^56 + 678\ M\^58 - 6064\ M\^60 + 
                    34458\ M\^62 - 131537\ M\^64 + 337512\ M\^66 - 
                    528640\ M\^68 + 232216\ M\^70 + 1077300\ M\^72 - 
                    2807324\ M\^74 + 2409995\ M\^76 + 2296509\ M\^78 - 
                    7906774\ M\^80 + 5530863\ M\^82 + 6736215\ M\^84 - 
                    13996581\ M\^86 + 2353531\ M\^88 + 14405370\ M\^90 - 
                    10653964\ M\^92 - 8036586\ M\^94 + 12813186\ M\^96 + 
                    732190\ M\^98 - 10194397\ M\^100 + 4753465\ M\^102 + 
                    5576502\ M\^104 - 6929778\ M\^106 - 893825\ M\^108 + 
                    5468623\ M\^110 - 1490473\ M\^112 - 3262977\ M\^114 + 
                    1398782\ M\^116 + 2501272\ M\^118 - 487811\ M\^120 - 
                    2716786\ M\^122 + 263577\ M\^124 + 2445191\ M\^126 - 
                    495728\ M\^128 - 1569362\ M\^130 + 594231\ M\^132 + 
                    761478\ M\^134 - 611340\ M\^136 - 40315\ M\^138 + 
                    217384\ M\^140 - 55665\ M\^142 - 79865\ M\^144 + 
                    97601\ M\^146 - 60703\ M\^148 + 25704\ M\^150 - 
                    7816\ M\^152 + 1695\ M\^154 - 251\ M\^156 + 23\ M\^158 - 
                    M\^160)\) + 
              L\^20\ \((56\ M\^84 - 915\ M\^86 + 6433\ M\^88 - 24430\ M\^90 + 
                    48491\ M\^92 - 20021\ M\^94 - 131261\ M\^96 + 
                    284777\ M\^98 - 91035\ M\^100 - 463796\ M\^102 + 
                    619616\ M\^104 + 126996\ M\^106 - 797166\ M\^108 + 
                    374248\ M\^110 + 329060\ M\^112 - 280321\ M\^114 + 
                    128228\ M\^116 - 216900\ M\^118 - 27221\ M\^120 + 
                    426540\ M\^122 - 292295\ M\^124 - 184643\ M\^126 + 
                    334586\ M\^128 - 88381\ M\^130 - 146120\ M\^132 + 
                    144019\ M\^134 - 6415\ M\^136 - 78424\ M\^138 + 
                    53109\ M\^140 + 3526\ M\^142 - 20341\ M\^144 + 
                    5096\ M\^146 + 7975\ M\^148 - 8607\ M\^150 + 
                    4384\ M\^152 - 1389\ M\^154 + 284\ M\^156 - 35\ M\^158 + 
                    2\ M\^160)\) + 
              L\^18\ \((8\ M\^64 - 215\ M\^66 + 2531\ M\^68 - 17076\ M\^70 + 
                    72465\ M\^72 - 195556\ M\^74 + 298097\ M\^76 - 
                    57035\ M\^78 - 801407\ M\^80 + 1671325\ M\^82 - 
                    915166\ M\^84 - 2054069\ M\^86 + 4310191\ M\^88 - 
                    1731758\ M\^90 - 4323699\ M\^92 + 6241789\ M\^94 - 
                    303081\ M\^96 - 5984919\ M\^98 + 4174879\ M\^100 + 
                    2199848\ M\^102 - 3977363\ M\^104 + 595250\ M\^106 + 
                    2096735\ M\^108 - 1041384\ M\^110 - 843005\ M\^112 + 
                    757913\ M\^114 + 22119\ M\^116 - 412930\ M\^118 + 
                    501976\ M\^120 + 73656\ M\^122 - 612745\ M\^124 + 
                    194954\ M\^126 + 429601\ M\^128 - 279766\ M\^130 - 
                    221940\ M\^132 + 235667\ M\^134 + 118632\ M\^136 - 
                    237827\ M\^138 + 65549\ M\^140 + 63819\ M\^142 - 
                    36847\ M\^144 - 27076\ M\^146 + 45432\ M\^148 - 
                    30774\ M\^150 + 13595\ M\^152 - 4358\ M\^154 + 
                    1034\ M\^156 - 175\ M\^158 + 19\ M\^160 - M\^162)\) + 
              L\^19\ \((28\ M\^74 - 576\ M\^76 + 5236\ M\^78 - 26942\ M\^80 + 
                    82810\ M\^82 - 137200\ M\^84 + 25426\ M\^86 + 
                    398895\ M\^88 - 786048\ M\^90 + 271340\ M\^92 + 
                    1196011\ M\^94 - 1822549\ M\^96 + 98971\ M\^98 + 
                    2139694\ M\^100 - 1756299\ M\^102 - 456377\ M\^104 + 
                    1309474\ M\^106 - 869770\ M\^108 + 696190\ M\^110 + 
                    381259\ M\^112 - 1775945\ M\^114 + 790090\ M\^116 + 
                    1365891\ M\^118 - 1181652\ M\^120 - 665902\ M\^122 + 
                    943209\ M\^124 + 266403\ M\^126 - 622214\ M\^128 - 
                    46003\ M\^130 + 361054\ M\^132 - 79473\ M\^134 - 
                    152073\ M\^136 + 114324\ M\^138 - 14260\ M\^140 - 
                    21814\ M\^142 + 13171\ M\^144 + 4822\ M\^146 - 
                    16107\ M\^148 + 16180\ M\^150 - 9845\ M\^152 + 
                    3994\ M\^154 - 1096\ M\^156 + 197\ M\^158 - 21\ M\^160 + 
                    M\^162)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 23]\), \(L\^14 + M\^116 + 
              L\^13\ \((\(-3\) + 17\ M\^2 - 32\ M\^4 + 3\ M\^6 + 61\ M\^8 - 
                    34\ M\^10 - 14\ M\^12 + 25\ M\^14 - 12\ M\^16 + 
                    3\ M\^18)\) + 
              L\^12\ \((3 - 30\ M\^2 + 110\ M\^4 - 153\ M\^6 - 26\ M\^8 + 
                    276\ M\^10 - 231\ M\^12 - 18\ M\^14 + 140\ M\^16 - 
                    49\ M\^18 + 295\ M\^20 - 192\ M\^22 - 288\ M\^24 + 
                    327\ M\^26 + 2\ M\^28 - 137\ M\^30 + 83\ M\^32 - 
                    24\ M\^34 + 3\ M\^36)\) + 
              L\^11\ \((\(-1\) + 13\ M\^2 - 69\ M\^4 + 202\ M\^6 - 
                    399\ M\^8 + 524\ M\^10 + 150\ M\^12 - 2233\ M\^14 + 
                    2632\ M\^16 + 2673\ M\^18 - 6889\ M\^20 - 324\ M\^22 + 
                    8666\ M\^24 - 2357\ M\^26 - 5504\ M\^28 + 2885\ M\^30 + 
                    1380\ M\^32 - 1251\ M\^34 + 285\ M\^36 + 135\ M\^38 - 
                    182\ M\^40 - 69\ M\^42 + 79\ M\^44 + 106\ M\^46 - 
                    136\ M\^48 + 59\ M\^50 - 12\ M\^52 + M\^54)\) + 
              L\^10\ \((3\ M\^4 - 45\ M\^6 + 270\ M\^8 - 782\ M\^10 + 
                    876\ M\^12 + 752\ M\^14 - 2536\ M\^16 + 481\ M\^18 + 
                    2081\ M\^20 + 357\ M\^22 + 1510\ M\^24 - 7235\ M\^26 - 
                    4694\ M\^28 + 16147\ M\^30 + 3855\ M\^32 - 16796\ M\^34 + 
                    1054\ M\^36 + 9501\ M\^38 - 3760\ M\^40 - 960\ M\^42 + 
                    2872\ M\^44 - 2577\ M\^46 - 517\ M\^48 + 1745\ M\^50 - 
                    361\ M\^52 - 336\ M\^54 + 53\ M\^56 - 52\ M\^58 + 
                    224\ M\^60 - 187\ M\^62 + 70\ M\^64 - 13\ M\^66 + 
                    M\^68)\) + 
              L\^9\ \((\(-3\)\ M\^8 + 49\ M\^10 - 327\ M\^12 + 1096\ M\^14 - 
                    1599\ M\^16 - 810\ M\^18 + 5745\ M\^20 - 2888\ M\^22 - 
                    12102\ M\^24 + 11627\ M\^26 + 19856\ M\^28 - 
                    24497\ M\^30 - 26531\ M\^32 + 39932\ M\^34 + 
                    21960\ M\^36 - 53203\ M\^38 + 1957\ M\^40 + 
                    57125\ M\^42 - 30850\ M\^44 - 43236\ M\^46 + 
                    46874\ M\^48 + 20189\ M\^50 - 41780\ M\^52 - 629\ M\^54 + 
                    25679\ M\^56 - 7479\ M\^58 - 10972\ M\^60 + 6968\ M\^62 + 
                    2462\ M\^64 - 3471\ M\^66 + 328\ M\^68 + 1031\ M\^70 - 
                    667\ M\^72 + 196\ M\^74 - 30\ M\^76 + 2\ M\^78)\) + 
              L\^8\ \((M\^12 - 17\ M\^14 + 126\ M\^16 - 524\ M\^18 + 
                    1245\ M\^20 - 1273\ M\^22 - 1093\ M\^24 + 4085\ M\^26 - 
                    1827\ M\^28 - 4196\ M\^30 + 4653\ M\^32 - 3851\ M\^34 + 
                    3653\ M\^36 + 17747\ M\^38 - 26930\ M\^40 - 
                    33316\ M\^42 + 54451\ M\^44 + 46853\ M\^46 - 
                    73567\ M\^48 - 47204\ M\^50 + 87241\ M\^52 + 
                    29745\ M\^54 - 90340\ M\^56 + 3594\ M\^58 + 
                    76837\ M\^60 - 30883\ M\^62 - 45925\ M\^64 + 
                    36522\ M\^66 + 14654\ M\^68 - 24386\ M\^70 + 
                    2014\ M\^72 + 9059\ M\^74 - 4118\ M\^76 - 1195\ M\^78 + 
                    1765\ M\^80 - 741\ M\^82 + 162\ M\^84 - 19\ M\^86 + 
                    M\^88)\) + 
              L\^7\ \((\(-2\)\ M\^20 + 35\ M\^22 - 260\ M\^24 + 1027\ M\^26 - 
                    2034\ M\^28 + 554\ M\^30 + 5628\ M\^32 - 7781\ M\^34 - 
                    7393\ M\^36 + 22265\ M\^38 + 790\ M\^40 - 36162\ M\^42 + 
                    12355\ M\^44 + 42897\ M\^46 - 23790\ M\^48 - 
                    44637\ M\^50 + 23963\ M\^52 + 47557\ M\^54 - 
                    9257\ M\^56 - 48078\ M\^58 - 9257\ M\^60 + 47557\ M\^62 + 
                    23963\ M\^64 - 44637\ M\^66 - 23790\ M\^68 + 
                    42897\ M\^70 + 12355\ M\^72 - 36162\ M\^74 + 790\ M\^76 + 
                    22265\ M\^78 - 7393\ M\^80 - 7781\ M\^82 + 5628\ M\^84 + 
                    554\ M\^86 - 2034\ M\^88 + 1027\ M\^90 - 260\ M\^92 + 
                    35\ M\^94 - 2\ M\^96)\) + 
              L\^6\ \((M\^28 - 19\ M\^30 + 162\ M\^32 - 741\ M\^34 + 
                    1765\ M\^36 - 1195\ M\^38 - 4118\ M\^40 + 9059\ M\^42 + 
                    2014\ M\^44 - 24386\ M\^46 + 14654\ M\^48 + 
                    36522\ M\^50 - 45925\ M\^52 - 30883\ M\^54 + 
                    76837\ M\^56 + 3594\ M\^58 - 90340\ M\^60 + 
                    29745\ M\^62 + 87241\ M\^64 - 47204\ M\^66 - 
                    73567\ M\^68 + 46853\ M\^70 + 54451\ M\^72 - 
                    33316\ M\^74 - 26930\ M\^76 + 17747\ M\^78 + 
                    3653\ M\^80 - 3851\ M\^82 + 4653\ M\^84 - 4196\ M\^86 - 
                    1827\ M\^88 + 4085\ M\^90 - 1093\ M\^92 - 1273\ M\^94 + 
                    1245\ M\^96 - 524\ M\^98 + 126\ M\^100 - 17\ M\^102 + 
                    M\^104)\) + 
              L\^5\ \((2\ M\^38 - 30\ M\^40 + 196\ M\^42 - 667\ M\^44 + 
                    1031\ M\^46 + 328\ M\^48 - 3471\ M\^50 + 2462\ M\^52 + 
                    6968\ M\^54 - 10972\ M\^56 - 7479\ M\^58 + 25679\ M\^60 - 
                    629\ M\^62 - 41780\ M\^64 + 20189\ M\^66 + 46874\ M\^68 - 
                    43236\ M\^70 - 30850\ M\^72 + 57125\ M\^74 + 
                    1957\ M\^76 - 53203\ M\^78 + 21960\ M\^80 + 
                    39932\ M\^82 - 26531\ M\^84 - 24497\ M\^86 + 
                    19856\ M\^88 + 11627\ M\^90 - 12102\ M\^92 - 
                    2888\ M\^94 + 5745\ M\^96 - 810\ M\^98 - 1599\ M\^100 + 
                    1096\ M\^102 - 327\ M\^104 + 49\ M\^106 - 3\ M\^108)\) + 
              L\^4\ \((M\^48 - 13\ M\^50 + 70\ M\^52 - 187\ M\^54 + 
                    224\ M\^56 - 52\ M\^58 + 53\ M\^60 - 336\ M\^62 - 
                    361\ M\^64 + 1745\ M\^66 - 517\ M\^68 - 2577\ M\^70 + 
                    2872\ M\^72 - 960\ M\^74 - 3760\ M\^76 + 9501\ M\^78 + 
                    1054\ M\^80 - 16796\ M\^82 + 3855\ M\^84 + 16147\ M\^86 - 
                    4694\ M\^88 - 7235\ M\^90 + 1510\ M\^92 + 357\ M\^94 + 
                    2081\ M\^96 + 481\ M\^98 - 2536\ M\^100 + 752\ M\^102 + 
                    876\ M\^104 - 782\ M\^106 + 270\ M\^108 - 45\ M\^110 + 
                    3\ M\^112)\) + 
              L\ \((3\ M\^98 - 12\ M\^100 + 25\ M\^102 - 14\ M\^104 - 
                    34\ M\^106 + 61\ M\^108 + 3\ M\^110 - 32\ M\^112 + 
                    17\ M\^114 - 3\ M\^116)\) + 
              L\^3\ \((M\^62 - 12\ M\^64 + 59\ M\^66 - 136\ M\^68 + 
                    106\ M\^70 + 79\ M\^72 - 69\ M\^74 - 182\ M\^76 + 
                    135\ M\^78 + 285\ M\^80 - 1251\ M\^82 + 1380\ M\^84 + 
                    2885\ M\^86 - 5504\ M\^88 - 2357\ M\^90 + 8666\ M\^92 - 
                    324\ M\^94 - 6889\ M\^96 + 2673\ M\^98 + 2632\ M\^100 - 
                    2233\ M\^102 + 150\ M\^104 + 524\ M\^106 - 399\ M\^108 + 
                    202\ M\^110 - 69\ M\^112 + 13\ M\^114 - M\^116)\) + 
              L\^2\ \((3\ M\^80 - 24\ M\^82 + 83\ M\^84 - 137\ M\^86 + 
                    2\ M\^88 + 327\ M\^90 - 288\ M\^92 - 192\ M\^94 + 
                    295\ M\^96 - 49\ M\^98 + 140\ M\^100 - 18\ M\^102 - 
                    231\ M\^104 + 276\ M\^106 - 26\ M\^108 - 153\ M\^110 + 
                    110\ M\^112 - 30\ M\^114 + 3\ M\^116)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 24]\), \(M\^54 + L\^19\ M\^60 + 
              L\ \((3\ M\^44 - 13\ M\^46 + 33\ M\^48 - 33\ M\^50 - 
                    24\ M\^52 + 78\ M\^54 - 15\ M\^56 - 25\ M\^58 + 
                    20\ M\^60 - 5\ M\^62)\) + 
              L\^18\ \((\(-5\)\ M\^52 + 20\ M\^54 - 25\ M\^56 - 15\ M\^58 + 
                    78\ M\^60 - 24\ M\^62 - 33\ M\^64 + 33\ M\^66 - 
                    13\ M\^68 + 3\ M\^70)\) + 
              L\^2\ \((3\ M\^34 - 24\ M\^36 + 91\ M\^38 - 199\ M\^40 + 
                    245\ M\^42 - 101\ M\^44 - 289\ M\^46 + 897\ M\^48 - 
                    842\ M\^50 - 674\ M\^52 + 1835\ M\^54 - 543\ M\^56 - 
                    907\ M\^58 + 1030\ M\^60 - 252\ M\^62 - 273\ M\^64 + 
                    243\ M\^66 - 79\ M\^68 + 10\ M\^70)\) + 
              L\^3\ \((M\^24 - 11\ M\^26 + 59\ M\^28 - 203\ M\^30 + 
                    497\ M\^32 - 935\ M\^34 + 1234\ M\^36 - 422\ M\^38 - 
                    2535\ M\^40 + 6387\ M\^42 - 5619\ M\^44 - 5847\ M\^46 + 
                    19995\ M\^48 - 11757\ M\^50 - 18196\ M\^52 + 
                    27784\ M\^54 - 2217\ M\^56 - 18376\ M\^58 + 
                    15158\ M\^60 - 667\ M\^62 - 7947\ M\^64 + 6383\ M\^66 - 
                    1259\ M\^68 - 1487\ M\^70 + 1410\ M\^72 - 567\ M\^74 + 
                    116\ M\^76 - 10\ M\^78)\) + 
              L\^17\ \((10\ M\^44 - 79\ M\^46 + 243\ M\^48 - 273\ M\^50 - 
                    252\ M\^52 + 1030\ M\^54 - 907\ M\^56 - 543\ M\^58 + 
                    1835\ M\^60 - 674\ M\^62 - 842\ M\^64 + 897\ M\^66 - 
                    289\ M\^68 - 101\ M\^70 + 245\ M\^72 - 199\ M\^74 + 
                    91\ M\^76 - 24\ M\^78 + 3\ M\^80)\) + 
              L\^4\ \((2\ M\^18 - 28\ M\^20 + 186\ M\^22 - 755\ M\^24 + 
                    1969\ M\^26 - 3132\ M\^28 + 2018\ M\^30 + 3102\ M\^32 - 
                    9493\ M\^34 + 11065\ M\^36 - 5142\ M\^38 - 12302\ M\^40 + 
                    46515\ M\^42 - 55362\ M\^44 - 34299\ M\^46 + 
                    147946\ M\^48 - 82077\ M\^50 - 124349\ M\^52 + 
                    186668\ M\^54 - 30947\ M\^56 - 132273\ M\^58 + 
                    146652\ M\^60 - 13022\ M\^62 - 101277\ M\^64 + 
                    77765\ M\^66 + 4801\ M\^68 - 39260\ M\^70 + 
                    23857\ M\^72 - 2320\ M\^74 - 5823\ M\^76 + 4758\ M\^78 - 
                    2012\ M\^80 + 516\ M\^82 - 76\ M\^84 + 5\ M\^86)\) + 
              L\^16\ \((\(-10\)\ M\^36 + 116\ M\^38 - 567\ M\^40 + 
                    1410\ M\^42 - 1487\ M\^44 - 1259\ M\^46 + 6383\ M\^48 - 
                    7947\ M\^50 - 667\ M\^52 + 15158\ M\^54 - 18376\ M\^56 - 
                    2217\ M\^58 + 27784\ M\^60 - 18196\ M\^62 - 
                    11757\ M\^64 + 19995\ M\^66 - 5847\ M\^68 - 5619\ M\^70 + 
                    6387\ M\^72 - 2535\ M\^74 - 422\ M\^76 + 1234\ M\^78 - 
                    935\ M\^80 + 497\ M\^82 - 203\ M\^84 + 59\ M\^86 - 
                    11\ M\^88 + M\^90)\) + 
              L\^5\ \((M\^12 - 19\ M\^14 + 178\ M\^16 - 1015\ M\^18 + 
                    3694\ M\^20 - 8162\ M\^22 + 7635\ M\^24 + 11406\ M\^26 - 
                    43599\ M\^28 + 32973\ M\^30 + 67477\ M\^32 - 
                    147093\ M\^34 + 5587\ M\^36 + 247428\ M\^38 - 
                    165305\ M\^40 - 270074\ M\^42 + 331996\ M\^44 + 
                    281141\ M\^46 - 543051\ M\^48 - 317374\ M\^50 + 
                    926222\ M\^52 + 186813\ M\^54 - 1321177\ M\^56 + 
                    270071\ M\^58 + 1288151\ M\^60 - 733456\ M\^62 - 
                    695881\ M\^64 + 775723\ M\^66 + 38320\ M\^68 - 
                    402955\ M\^70 + 200725\ M\^72 + 46685\ M\^74 - 
                    105368\ M\^76 + 52111\ M\^78 - 460\ M\^80 - 
                    15048\ M\^82 + 10608\ M\^84 - 4205\ M\^86 + 1095\ M\^88 - 
                    189\ M\^90 + 20\ M\^92 - M\^94)\) + 
              L\^15\ \((5\ M\^28 - 76\ M\^30 + 516\ M\^32 - 2012\ M\^34 + 
                    4758\ M\^36 - 5823\ M\^38 - 2320\ M\^40 + 23857\ M\^42 - 
                    39260\ M\^44 + 4801\ M\^46 + 77765\ M\^48 - 
                    101277\ M\^50 - 13022\ M\^52 + 146652\ M\^54 - 
                    132273\ M\^56 - 30947\ M\^58 + 186668\ M\^60 - 
                    124349\ M\^62 - 82077\ M\^64 + 147946\ M\^66 - 
                    34299\ M\^68 - 55362\ M\^70 + 46515\ M\^72 - 
                    12302\ M\^74 - 5142\ M\^76 + 11065\ M\^78 - 9493\ M\^80 + 
                    3102\ M\^82 + 2018\ M\^84 - 3132\ M\^86 + 1969\ M\^88 - 
                    755\ M\^90 + 186\ M\^92 - 28\ M\^94 + 2\ M\^96)\) + 
              L\^6\ \((\(-3\)\ M\^8 + 56\ M\^10 - 488\ M\^12 + 2522\ M\^14 - 
                    8112\ M\^16 + 14786\ M\^18 - 5515\ M\^20 - 41612\ M\^22 + 
                    91245\ M\^24 - 10726\ M\^26 - 245488\ M\^28 + 
                    329170\ M\^30 + 213227\ M\^32 - 906741\ M\^34 + 
                    353261\ M\^36 + 1398486\ M\^38 - 1517702\ M\^40 - 
                    1384323\ M\^42 + 2879991\ M\^44 + 848225\ M\^46 - 
                    3972796\ M\^48 - 49986\ M\^50 + 4512594\ M\^52 - 
                    894885\ M\^54 - 4159096\ M\^56 + 1877235\ M\^58 + 
                    2712789\ M\^60 - 2361986\ M\^62 - 601374\ M\^64 + 
                    1881180\ M\^66 - 948938\ M\^68 - 733330\ M\^70 + 
                    1214651\ M\^72 - 201354\ M\^74 - 623707\ M\^76 + 
                    412621\ M\^78 + 72895\ M\^80 - 192791\ M\^82 + 
                    77022\ M\^84 + 12136\ M\^86 - 26750\ M\^88 + 
                    14815\ M\^90 - 5124\ M\^92 + 1237\ M\^94 - 205\ M\^96 + 
                    21\ M\^98 - M\^100)\) + 
              L\^14\ \((\(-M\^20\) + 20\ M\^22 - 189\ M\^24 + 1095\ M\^26 - 
                    4205\ M\^28 + 10608\ M\^30 - 15048\ M\^32 - 460\ M\^34 + 
                    52111\ M\^36 - 105368\ M\^38 + 46685\ M\^40 + 
                    200725\ M\^42 - 402955\ M\^44 + 38320\ M\^46 + 
                    775723\ M\^48 - 695881\ M\^50 - 733456\ M\^52 + 
                    1288151\ M\^54 + 270071\ M\^56 - 1321177\ M\^58 + 
                    186813\ M\^60 + 926222\ M\^62 - 317374\ M\^64 - 
                    543051\ M\^66 + 281141\ M\^68 + 331996\ M\^70 - 
                    270074\ M\^72 - 165305\ M\^74 + 247428\ M\^76 + 
                    5587\ M\^78 - 147093\ M\^80 + 67477\ M\^82 + 
                    32973\ M\^84 - 43599\ M\^86 + 11406\ M\^88 + 
                    7635\ M\^90 - 8162\ M\^92 + 3694\ M\^94 - 1015\ M\^96 + 
                    178\ M\^98 - 19\ M\^100 + M\^102)\) + 
              L\^7\ \((3\ M\^4 - 59\ M\^6 + 521\ M\^8 - 2662\ M\^10 + 
                    8327\ M\^12 - 14504\ M\^14 + 4332\ M\^16 + 40110\ M\^18 - 
                    78968\ M\^20 - 8053\ M\^22 + 239382\ M\^24 - 
                    282143\ M\^26 - 250368\ M\^28 + 890978\ M\^30 - 
                    321216\ M\^32 - 1476912\ M\^34 + 1740481\ M\^36 + 
                    1272328\ M\^38 - 3471079\ M\^40 + 198703\ M\^42 + 
                    4313576\ M\^44 - 2532416\ M\^46 - 3362836\ M\^48 + 
                    4769453\ M\^50 + 768458\ M\^52 - 6205978\ M\^54 + 
                    2578526\ M\^56 + 6623661\ M\^58 - 5661932\ M\^60 - 
                    5627333\ M\^62 + 7525636\ M\^64 + 3302892\ M\^66 - 
                    7290352\ M\^68 - 541914\ M\^70 + 5084759\ M\^72 - 
                    1219685\ M\^74 - 2320862\ M\^76 + 1427687\ M\^78 + 
                    433771\ M\^80 - 757899\ M\^82 + 207119\ M\^84 + 
                    156810\ M\^86 - 157343\ M\^88 + 41733\ M\^90 + 
                    22526\ M\^92 - 26413\ M\^94 + 12540\ M\^96 - 
                    3567\ M\^98 + 632\ M\^100 - 65\ M\^102 + 3\ M\^104)\) + 
              L\^13\ \((\(-M\^14\) + 21\ M\^16 - 205\ M\^18 + 1237\ M\^20 - 
                    5124\ M\^22 + 14815\ M\^24 - 26750\ M\^26 + 
                    12136\ M\^28 + 77022\ M\^30 - 192791\ M\^32 + 
                    72895\ M\^34 + 412621\ M\^36 - 623707\ M\^38 - 
                    201354\ M\^40 + 1214651\ M\^42 - 733330\ M\^44 - 
                    948938\ M\^46 + 1881180\ M\^48 - 601374\ M\^50 - 
                    2361986\ M\^52 + 2712789\ M\^54 + 1877235\ M\^56 - 
                    4159096\ M\^58 - 894885\ M\^60 + 4512594\ M\^62 - 
                    49986\ M\^64 - 3972796\ M\^66 + 848225\ M\^68 + 
                    2879991\ M\^70 - 1384323\ M\^72 - 1517702\ M\^74 + 
                    1398486\ M\^76 + 353261\ M\^78 - 906741\ M\^80 + 
                    213227\ M\^82 + 329170\ M\^84 - 245488\ M\^86 - 
                    10726\ M\^88 + 91245\ M\^90 - 41612\ M\^92 - 
                    5515\ M\^94 + 14786\ M\^96 - 8112\ M\^98 + 2522\ M\^100 - 
                    488\ M\^102 + 56\ M\^104 - 3\ M\^106)\) + 
              L\^8\ \((\(-1\) + 20\ M\^2 - 181\ M\^4 + 974\ M\^6 - 
                    3420\ M\^8 + 7898\ M\^10 - 10423\ M\^12 - 55\ M\^14 + 
                    30410\ M\^16 - 48571\ M\^18 - 29475\ M\^20 + 
                    207746\ M\^22 - 184259\ M\^24 - 382260\ M\^26 + 
                    868332\ M\^28 + 176198\ M\^30 - 2166851\ M\^32 + 
                    1301637\ M\^34 + 3518027\ M\^36 - 4949815\ M\^38 - 
                    3579780\ M\^40 + 10838404\ M\^42 + 1102322\ M\^44 - 
                    17664776\ M\^46 + 3673531\ M\^48 + 23250770\ M\^50 - 
                    8795064\ M\^52 - 25668578\ M\^54 + 12216889\ M\^56 + 
                    23975844\ M\^58 - 13018204\ M\^60 - 18416468\ M\^62 + 
                    11731741\ M\^64 + 10771196\ M\^66 - 8843572\ M\^68 - 
                    3701319\ M\^70 + 5118642\ M\^72 - 611534\ M\^74 - 
                    1711262\ M\^76 + 1760627\ M\^78 - 302098\ M\^80 - 
                    1118234\ M\^82 + 803286\ M\^84 + 279043\ M\^86 - 
                    513772\ M\^88 + 87977\ M\^90 + 156031\ M\^92 - 
                    93205\ M\^94 - 5739\ M\^96 + 28665\ M\^98 - 
                    15458\ M\^100 + 4430\ M\^102 - 753\ M\^104 + 72\ M\^106 - 
                    3\ M\^108)\) + 
              L\^12\ \((3\ M\^10 - 65\ M\^12 + 632\ M\^14 - 3567\ M\^16 + 
                    12540\ M\^18 - 26413\ M\^20 + 22526\ M\^22 + 
                    41733\ M\^24 - 157343\ M\^26 + 156810\ M\^28 + 
                    207119\ M\^30 - 757899\ M\^32 + 433771\ M\^34 + 
                    1427687\ M\^36 - 2320862\ M\^38 - 1219685\ M\^40 + 
                    5084759\ M\^42 - 541914\ M\^44 - 7290352\ M\^46 + 
                    3302892\ M\^48 + 7525636\ M\^50 - 5627333\ M\^52 - 
                    5661932\ M\^54 + 6623661\ M\^56 + 2578526\ M\^58 - 
                    6205978\ M\^60 + 768458\ M\^62 + 4769453\ M\^64 - 
                    3362836\ M\^66 - 2532416\ M\^68 + 4313576\ M\^70 + 
                    198703\ M\^72 - 3471079\ M\^74 + 1272328\ M\^76 + 
                    1740481\ M\^78 - 1476912\ M\^80 - 321216\ M\^82 + 
                    890978\ M\^84 - 250368\ M\^86 - 282143\ M\^88 + 
                    239382\ M\^90 - 8053\ M\^92 - 78968\ M\^94 + 
                    40110\ M\^96 + 4332\ M\^98 - 14504\ M\^100 + 
                    8327\ M\^102 - 2662\ M\^104 + 521\ M\^106 - 59\ M\^108 + 
                    3\ M\^110)\) + 
              L\^9\ \((1 - 25\ M\^2 + 266\ M\^4 - 1579\ M\^6 + 5588\ M\^8 - 
                    10888\ M\^10 + 5071\ M\^12 + 24672\ M\^14 - 
                    38011\ M\^16 - 59402\ M\^18 + 191324\ M\^20 + 
                    36132\ M\^22 - 611706\ M\^24 + 283516\ M\^26 + 
                    1532825\ M\^28 - 1584916\ M\^30 - 2912561\ M\^32 + 
                    4809191\ M\^34 + 4227542\ M\^36 - 10513754\ M\^38 - 
                    4577554\ M\^40 + 17803105\ M\^42 + 3458573\ M\^44 - 
                    24074656\ M\^46 - 1346091\ M\^48 + 26332390\ M\^50 - 
                    899474\ M\^52 - 23082395\ M\^54 + 2957002\ M\^56 + 
                    15442884\ M\^58 - 4944367\ M\^60 - 5929640\ M\^62 + 
                    6592679\ M\^64 - 2278258\ M\^66 - 6780111\ M\^68 + 
                    7163140\ M\^70 + 4964390\ M\^72 - 8178933\ M\^74 - 
                    2138798\ M\^76 + 6381861\ M\^78 - 113377\ M\^80 - 
                    3622243\ M\^82 + 984851\ M\^84 + 1426666\ M\^86 - 
                    823977\ M\^88 - 314003\ M\^90 + 397338\ M\^92 - 
                    29550\ M\^94 - 108530\ M\^96 + 46216\ M\^98 + 
                    12011\ M\^100 - 18709\ M\^102 + 8522\ M\^104 - 
                    2174\ M\^106 + 331\ M\^108 - 28\ M\^110 + M\^112)\) + 
              L\^11\ \((\(-3\)\ M\^6 + 72\ M\^8 - 753\ M\^10 + 4430\ M\^12 - 
                    15458\ M\^14 + 28665\ M\^16 - 5739\ M\^18 - 
                    93205\ M\^20 + 156031\ M\^22 + 87977\ M\^24 - 
                    513772\ M\^26 + 279043\ M\^28 + 803286\ M\^30 - 
                    1118234\ M\^32 - 302098\ M\^34 + 1760627\ M\^36 - 
                    1711262\ M\^38 - 611534\ M\^40 + 5118642\ M\^42 - 
                    3701319\ M\^44 - 8843572\ M\^46 + 10771196\ M\^48 + 
                    11731741\ M\^50 - 18416468\ M\^52 - 13018204\ M\^54 + 
                    23975844\ M\^56 + 12216889\ M\^58 - 25668578\ M\^60 - 
                    8795064\ M\^62 + 23250770\ M\^64 + 3673531\ M\^66 - 
                    17664776\ M\^68 + 1102322\ M\^70 + 10838404\ M\^72 - 
                    3579780\ M\^74 - 4949815\ M\^76 + 3518027\ M\^78 + 
                    1301637\ M\^80 - 2166851\ M\^82 + 176198\ M\^84 + 
                    868332\ M\^86 - 382260\ M\^88 - 184259\ M\^90 + 
                    207746\ M\^92 - 29475\ M\^94 - 48571\ M\^96 + 
                    30410\ M\^98 - 55\ M\^100 - 10423\ M\^102 + 
                    7898\ M\^104 - 3420\ M\^106 + 974\ M\^108 - 181\ M\^110 + 
                    20\ M\^112 - M\^114)\) + 
              L\^10\ \((M\^2 - 28\ M\^4 + 331\ M\^6 - 2174\ M\^8 + 
                    8522\ M\^10 - 18709\ M\^12 + 12011\ M\^14 + 
                    46216\ M\^16 - 108530\ M\^18 - 29550\ M\^20 + 
                    397338\ M\^22 - 314003\ M\^24 - 823977\ M\^26 + 
                    1426666\ M\^28 + 984851\ M\^30 - 3622243\ M\^32 - 
                    113377\ M\^34 + 6381861\ M\^36 - 2138798\ M\^38 - 
                    8178933\ M\^40 + 4964390\ M\^42 + 7163140\ M\^44 - 
                    6780111\ M\^46 - 2278258\ M\^48 + 6592679\ M\^50 - 
                    5929640\ M\^52 - 4944367\ M\^54 + 15442884\ M\^56 + 
                    2957002\ M\^58 - 23082395\ M\^60 - 899474\ M\^62 + 
                    26332390\ M\^64 - 1346091\ M\^66 - 24074656\ M\^68 + 
                    3458573\ M\^70 + 17803105\ M\^72 - 4577554\ M\^74 - 
                    10513754\ M\^76 + 4227542\ M\^78 + 4809191\ M\^80 - 
                    2912561\ M\^82 - 1584916\ M\^84 + 1532825\ M\^86 + 
                    283516\ M\^88 - 611706\ M\^90 + 36132\ M\^92 + 
                    191324\ M\^94 - 59402\ M\^96 - 38011\ M\^98 + 
                    24672\ M\^100 + 5071\ M\^102 - 10888\ M\^104 + 
                    5588\ M\^106 - 1579\ M\^108 + 266\ M\^110 - 25\ M\^112 + 
                    M\^114)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 25]\), \(L\^27\ M\^18 + M\^124 + 
              L\^26\ \((\(-4\)\ M\^14 + 16\ M\^16 - 22\ M\^18 - 16\ M\^20 + 
                    88\ M\^22 - 22\ M\^24 - 44\ M\^26 + 41\ M\^28 - 
                    18\ M\^30 + 4\ M\^32)\) + 
              L\^25\ \((6\ M\^10 - 49\ M\^12 + 168\ M\^14 - 274\ M\^16 + 
                    57\ M\^18 + 680\ M\^20 - 1464\ M\^22 + 639\ M\^24 + 
                    1996\ M\^26 - 1931\ M\^28 - 139\ M\^30 + 1331\ M\^32 - 
                    1130\ M\^34 + 158\ M\^36 + 536\ M\^38 - 504\ M\^40 + 
                    217\ M\^42 - 52\ M\^44 + 6\ M\^46)\) + 
              L\^24\ \((\(-4\)\ M\^6 + 51\ M\^8 - 294\ M\^10 + 993\ M\^12 - 
                    2035\ M\^14 + 1786\ M\^16 + 2753\ M\^18 - 10400\ M\^20 + 
                    11092\ M\^22 + 3862\ M\^24 - 25431\ M\^26 + 
                    21048\ M\^28 + 17866\ M\^30 - 29659\ M\^32 + 
                    7206\ M\^34 + 10422\ M\^36 - 15450\ M\^38 + 9288\ M\^40 + 
                    3460\ M\^42 - 8950\ M\^44 + 4729\ M\^46 + 771\ M\^48 - 
                    2850\ M\^50 + 2259\ M\^52 - 1038\ M\^54 + 296\ M\^56 - 
                    50\ M\^58 + 4\ M\^60)\) + 
              L\^23\ \((M\^2 - 19\ M\^4 + 172\ M\^6 - 958\ M\^8 + 
                    3536\ M\^10 - 8577\ M\^12 + 12027\ M\^14 - 1694\ M\^16 - 
                    33928\ M\^18 + 75146\ M\^20 - 47122\ M\^22 - 
                    97534\ M\^24 + 218017\ M\^26 - 66744\ M\^28 - 
                    293835\ M\^30 + 311187\ M\^32 + 184331\ M\^34 - 
                    347970\ M\^36 - 9404\ M\^38 + 170952\ M\^40 - 
                    56925\ M\^42 + 5106\ M\^44 + 6126\ M\^46 - 44137\ M\^48 + 
                    41291\ M\^50 + 190\ M\^52 - 24656\ M\^54 + 18770\ M\^56 - 
                    1813\ M\^58 - 8464\ M\^60 + 8841\ M\^62 - 5138\ M\^64 + 
                    2072\ M\^66 - 606\ M\^68 + 124\ M\^70 - 16\ M\^72 + 
                    M\^74)\) + 
              L\^22\ \((1 - 24\ M\^2 + 261\ M\^4 - 1714\ M\^6 + 7529\ M\^8 - 
                    22598\ M\^10 + 43422\ M\^12 - 37437\ M\^14 - 
                    46666\ M\^16 + 192226\ M\^18 - 254252\ M\^20 + 
                    60124\ M\^22 + 445186\ M\^24 - 944591\ M\^26 + 
                    422963\ M\^28 + 1615831\ M\^30 - 2491848\ M\^32 - 
                    1022001\ M\^34 + 4331951\ M\^36 - 444076\ M\^38 - 
                    4291466\ M\^40 + 1578367\ M\^42 + 2578583\ M\^44 - 
                    1609643\ M\^46 - 799691\ M\^48 + 920967\ M\^50 - 
                    64127\ M\^52 - 275128\ M\^54 + 171420\ M\^56 - 
                    6262\ M\^58 - 40751\ M\^60 + 11886\ M\^62 - 3613\ M\^64 + 
                    11450\ M\^66 - 2452\ M\^68 - 14175\ M\^70 + 
                    19932\ M\^72 - 14532\ M\^74 + 6844\ M\^76 - 2158\ M\^78 + 
                    443\ M\^80 - 54\ M\^82 + 3\ M\^84)\) + 
              L\^21\ \((\(-6\) + 139\ M\^2 - 1435\ M\^4 + 8703\ M\^6 - 
                    34290\ M\^8 + 90155\ M\^10 - 146713\ M\^12 + 
                    75466\ M\^14 + 283032\ M\^16 - 747982\ M\^18 + 
                    552714\ M\^20 + 634291\ M\^22 - 1439434\ M\^24 + 
                    931103\ M\^26 - 734536\ M\^28 - 664352\ M\^30 + 
                    6955290\ M\^32 - 6400190\ M\^34 - 12669396\ M\^36 + 
                    18598199\ M\^38 + 12246855\ M\^40 - 26203046\ M\^42 - 
                    5250737\ M\^44 + 22730409\ M\^46 - 2125010\ M\^48 - 
                    12158397\ M\^50 + 4671816\ M\^52 + 2808216\ M\^54 - 
                    2985583\ M\^56 + 1115173\ M\^58 + 709880\ M\^60 - 
                    1194051\ M\^62 + 203008\ M\^64 + 458344\ M\^66 - 
                    236428\ M\^68 - 47195\ M\^70 + 60324\ M\^72 - 
                    7459\ M\^74 - 9661\ M\^76 + 22735\ M\^78 - 38855\ M\^80 + 
                    39575\ M\^82 - 25423\ M\^84 + 10799\ M\^86 - 
                    3061\ M\^88 + 562\ M\^90 - 61\ M\^92 + 3\ M\^94)\) + 
              L\^20\ \((15 - 369\ M\^2 + 4022\ M\^4 - 25401\ M\^6 + 
                    101498\ M\^8 - 258879\ M\^10 + 378551\ M\^12 - 
                    99600\ M\^14 - 851735\ M\^16 + 1895649\ M\^18 - 
                    1275579\ M\^20 - 2270178\ M\^22 + 5367348\ M\^24 - 
                    968332\ M\^26 - 6562770\ M\^28 + 743576\ M\^30 + 
                    5814762\ M\^32 + 15987127\ M\^34 - 16429402\ M\^36 - 
                    48543679\ M\^38 + 45637735\ M\^40 + 74247169\ M\^42 - 
                    75487088\ M\^44 - 69101721\ M\^46 + 80412417\ M\^48 + 
                    37767348\ M\^50 - 55952381\ M\^52 - 5949309\ M\^54 + 
                    20790770\ M\^56 - 8227093\ M\^58 + 3462227\ M\^60 + 
                    6875396\ M\^62 - 10689100\ M\^64 - 1101892\ M\^66 + 
                    7727205\ M\^68 - 1995658\ M\^70 - 2948470\ M\^72 + 
                    1795865\ M\^74 + 453791\ M\^76 - 791162\ M\^78 + 
                    201294\ M\^80 + 118177\ M\^82 - 70448\ M\^84 - 
                    50825\ M\^86 + 101224\ M\^88 - 88185\ M\^90 + 
                    53172\ M\^92 - 23567\ M\^94 + 7606\ M\^96 - 1731\ M\^98 + 
                    263\ M\^100 - 24\ M\^102 + M\^104)\) + 
              L\^19\ \((\(-20\) + 534\ M\^2 - 6364\ M\^4 + 44236\ M\^6 - 
                    195252\ M\^8 + 548702\ M\^10 - 872738\ M\^12 + 
                    252626\ M\^14 + 1981412\ M\^16 - 3923089\ M\^18 + 
                    1737646\ M\^20 + 4688452\ M\^22 - 9499066\ M\^24 + 
                    4669593\ M\^26 + 12770607\ M\^28 - 19192636\ M\^30 - 
                    15676725\ M\^32 + 27011248\ M\^34 + 52715215\ M\^36 - 
                    37107627\ M\^38 - 151540493\ M\^40 + 74011071\ M\^42 + 
                    271822103\ M\^44 - 135197077\ M\^46 - 318399710\ M\^48 + 
                    180593769\ M\^50 + 245308159\ M\^52 - 168212279\ M\^54 - 
                    105532994\ M\^56 + 98821911\ M\^58 - 8405708\ M\^60 - 
                    19037474\ M\^62 + 50940156\ M\^64 - 28265409\ M\^66 - 
                    40580750\ M\^68 + 36043830\ M\^70 + 17397642\ M\^72 - 
                    23702929\ M\^74 - 3335811\ M\^76 + 10672720\ M\^78 - 
                    1070907\ M\^80 - 3327332\ M\^82 + 1054840\ M\^84 + 
                    780451\ M\^86 - 583526\ M\^88 + 39457\ M\^90 + 
                    41578\ M\^92 + 105275\ M\^94 - 181909\ M\^96 + 
                    147948\ M\^98 - 79993\ M\^100 + 31294\ M\^102 - 
                    9015\ M\^104 + 1880\ M\^106 - 270\ M\^108 + 24\ M\^110 - 
                    M\^112)\) + 
              L\^18\ \((15 - 442\ M\^2 + 5888\ M\^4 - 46421\ M\^6 + 
                    236667\ M\^8 - 790133\ M\^10 + 1600187\ M\^12 - 
                    1225589\ M\^14 - 2617867\ M\^16 + 7871146\ M\^18 - 
                    4752065\ M\^20 - 10314617\ M\^22 + 18508699\ M\^24 - 
                    3300800\ M\^26 - 19446762\ M\^28 + 30560685\ M\^30 - 
                    7041720\ M\^32 - 68857103\ M\^34 + 45549523\ M\^36 + 
                    163360128\ M\^38 - 79586084\ M\^40 - 378671215\ M\^42 + 
                    129042342\ M\^44 + 670060875\ M\^46 - 226435701\ M\^48 - 
                    838385917\ M\^50 + 353261790\ M\^52 + 711068813\ M\^54 - 
                    414678061\ M\^56 - 331934479\ M\^58 + 331696631\ M\^60 - 
                    68433692\ M\^62 - 144099165\ M\^64 + 284222013\ M\^66 - 
                    33412547\ M\^68 - 285697115\ M\^70 + 118469429\ M\^72 + 
                    180188900\ M\^74 - 113852339\ M\^76 - 77780221\ M\^78 + 
                    70743166\ M\^80 + 21249502\ M\^82 - 31635835\ M\^84 - 
                    1986509\ M\^86 + 10531764\ M\^88 - 1377393\ M\^90 - 
                    2305146\ M\^92 + 488367\ M\^94 + 565384\ M\^96 - 
                    312949\ M\^98 + 155803\ M\^100 - 252762\ M\^102 + 
                    270350\ M\^104 - 171746\ M\^106 + 72156\ M\^108 - 
                    21048\ M\^110 + 4287\ M\^112 - 590\ M\^114 + 50\ M\^116 - 
                    2\ M\^118)\) + 
              L\^17\ \((\(-6\) + 203\ M\^2 - 3127\ M\^4 + 28788\ M\^6 - 
                    174147\ M\^8 + 711334\ M\^10 - 1901184\ M\^12 + 
                    2794848\ M\^14 + 143539\ M\^16 - 9184478\ M\^18 + 
                    14277817\ M\^20 + 3758586\ M\^22 - 35203519\ M\^24 + 
                    28350776\ M\^26 + 28993211\ M\^28 - 67581590\ M\^30 + 
                    42482223\ M\^32 + 57913769\ M\^34 - 196526283\ M\^36 + 
                    33233706\ M\^38 + 463225587\ M\^40 - 183975160\ M\^42 - 
                    882683219\ M\^44 + 359363107\ M\^46 + 1376355056\ M\^48 - 
                    561243061\ M\^50 - 1654085637\ M\^52 + 778485916\ M\^54 + 
                    1406582508\ M\^56 - 880103774\ M\^58 - 626169656\ M\^60 + 
                    701701478\ M\^62 - 300158584\ M\^64 - 260251826\ M\^66 + 
                    885544976\ M\^68 - 224462217\ M\^70 - 938221828\ M\^72 + 
                    508712996\ M\^74 + 639119841\ M\^76 - 519689172\ M\^78 - 
                    289624488\ M\^80 + 362916225\ M\^82 + 68574544\ M\^84 - 
                    186554798\ M\^86 + 13556525\ M\^88 + 70639750\ M\^90 - 
                    22079526\ M\^92 - 18328864\ M\^94 + 11318021\ M\^96 + 
                    2873154\ M\^98 - 4184798\ M\^100 + 524617\ M\^102 + 
                    706056\ M\^104 + 9702\ M\^106 - 564944\ M\^108 + 
                    508987\ M\^110 - 257725\ M\^112 + 88313\ M\^114 - 
                    21436\ M\^116 + 3668\ M\^118 - 424\ M\^120 + 30\ M\^122 - 
                    M\^124)\) + 
              L\ \((4\ M\^110 - 18\ M\^112 + 41\ M\^114 - 44\ M\^116 - 
                    22\ M\^118 + 88\ M\^120 - 16\ M\^122 - 22\ M\^124 + 
                    16\ M\^126 - 4\ M\^128)\) + 
              L\^16\ \((1 - 45\ M\^2 + 874\ M\^4 - 9907\ M\^6 + 73526\ M\^8 - 
                    374749\ M\^10 + 1316440\ M\^12 - 3010280\ M\^14 + 
                    3414384\ M\^16 + 2600134\ M\^18 - 15653244\ M\^20 + 
                    17547073\ M\^22 + 16306529\ M\^24 - 60271289\ M\^26 + 
                    33813242\ M\^28 + 71219962\ M\^30 - 133112226\ M\^32 + 
                    52267838\ M\^34 + 181356717\ M\^36 - 380878213\ M\^38 - 
                    40926022\ M\^40 + 907064742\ M\^42 - 328318776\ M\^44 - 
                    1550060749\ M\^46 + 813365558\ M\^48 + 
                    2134579752\ M\^50 - 1255411116\ M\^52 - 
                    2347798000\ M\^54 + 1530780776\ M\^56 + 
                    1864858394\ M\^58 - 1476079436\ M\^60 - 
                    653879339\ M\^62 + 936199735\ M\^64 - 801703351\ M\^66 + 
                    16767180\ M\^68 + 1756273169\ M\^70 - 1013055713\ M\^72 - 
                    1810027040\ M\^74 + 1589791805\ M\^76 + 
                    1190755693\ M\^78 - 1552954115\ M\^80 - 
                    446484099\ M\^82 + 1094444211\ M\^84 - 24281031\ M\^86 - 
                    567209127\ M\^88 + 166515781\ M\^90 + 204234154\ M\^92 - 
                    129046397\ M\^94 - 38460019\ M\^96 + 59102578\ M\^98 - 
                    6650967\ M\^100 - 17402053\ M\^102 + 8930219\ M\^104 + 
                    1568059\ M\^106 - 3031925\ M\^108 + 504420\ M\^110 + 
                    1054195\ M\^112 - 1056557\ M\^114 + 558331\ M\^116 - 
                    199558\ M\^118 + 50902\ M\^120 - 9232\ M\^122 + 
                    1139\ M\^124 - 86\ M\^126 + 3\ M\^128)\) + 
              L\^15\ \((3\ M\^2 - 103\ M\^4 + 1644\ M\^6 - 16061\ M\^8 + 
                    106178\ M\^10 - 493707\ M\^12 + 1609864\ M\^14 - 
                    3452185\ M\^16 + 3615313\ M\^18 + 3323959\ M\^20 - 
                    17445368\ M\^22 + 18845392\ M\^24 + 19276682\ M\^26 - 
                    71329361\ M\^28 + 46600400\ M\^30 + 84908753\ M\^32 - 
                    193991006\ M\^34 + 95537323\ M\^36 + 285039833\ M\^38 - 
                    591049487\ M\^40 - 61485816\ M\^42 + 1345378213\ M\^44 - 
                    629578362\ M\^46 - 2110666204\ M\^48 + 
                    1611250425\ M\^50 + 2585883096\ M\^52 - 
                    2455552446\ M\^54 - 2530161200\ M\^56 + 
                    2757905224\ M\^58 + 1804727430\ M\^60 - 
                    2287461571\ M\^62 - 467044149\ M\^64 + 
                    1046155768\ M\^66 - 1031710033\ M\^68 + 
                    676190082\ M\^70 + 1954528442\ M\^72 - 
                    2311158904\ M\^74 - 1850162935\ M\^76 + 
                    3195269504\ M\^78 + 949502298\ M\^80 - 
                    3023394860\ M\^82 + 59987245\ M\^84 + 2094249644\ M\^86 - 
                    613398151\ M\^88 - 1031881916\ M\^90 + 631924360\ M\^92 + 
                    300848657\ M\^94 - 384286318\ M\^96 + 10867680\ M\^98 + 
                    147927322\ M\^100 - 63940147\ M\^102 - 26439418\ M\^104 + 
                    34647378\ M\^106 - 7630001\ M\^108 - 7400572\ M\^110 + 
                    5893205\ M\^112 - 460712\ M\^114 - 1937052\ M\^116 + 
                    1678610\ M\^118 - 813097\ M\^120 + 270118\ M\^122 - 
                    64429\ M\^124 + 10974\ M\^126 - 1276\ M\^128 + 
                    91\ M\^130 - 3\ M\^132)\) + 
              L\^2\ \((6\ M\^96 - 52\ M\^98 + 217\ M\^100 - 504\ M\^102 + 
                    536\ M\^104 + 158\ M\^106 - 1130\ M\^108 + 1331\ M\^110 - 
                    139\ M\^112 - 1931\ M\^114 + 1996\ M\^116 + 639\ M\^118 - 
                    1464\ M\^120 + 680\ M\^122 + 57\ M\^124 - 274\ M\^126 + 
                    168\ M\^128 - 49\ M\^130 + 6\ M\^132)\) + 
              L\^3\ \((4\ M\^82 - 50\ M\^84 + 296\ M\^86 - 1038\ M\^88 + 
                    2259\ M\^90 - 2850\ M\^92 + 771\ M\^94 + 4729\ M\^96 - 
                    8950\ M\^98 + 3460\ M\^100 + 9288\ M\^102 - 
                    15450\ M\^104 + 10422\ M\^106 + 7206\ M\^108 - 
                    29659\ M\^110 + 17866\ M\^112 + 21048\ M\^114 - 
                    25431\ M\^116 + 3862\ M\^118 + 11092\ M\^120 - 
                    10400\ M\^122 + 2753\ M\^124 + 1786\ M\^126 - 
                    2035\ M\^128 + 993\ M\^130 - 294\ M\^132 + 51\ M\^134 - 
                    4\ M\^136)\) + 
              L\^14\ \((3\ M\^4 - 97\ M\^6 + 1473\ M\^8 - 13858\ M\^10 + 
                    89400\ M\^12 - 411456\ M\^14 + 1350931\ M\^16 - 
                    3002317\ M\^18 + 3617481\ M\^20 + 1311971\ M\^22 - 
                    13356712\ M\^24 + 18525327\ M\^26 + 8761585\ M\^28 - 
                    61075922\ M\^30 + 63551742\ M\^32 + 53478886\ M\^34 - 
                    213921253\ M\^36 + 164070727\ M\^38 + 291336720\ M\^40 - 
                    716555067\ M\^42 + 19065184\ M\^44 + 1504399992\ M\^46 - 
                    952571141\ M\^48 - 2160621485\ M\^50 + 
                    2320526238\ M\^52 + 2299678529\ M\^54 - 
                    3490884195\ M\^56 - 1813129579\ M\^58 + 
                    3779817874\ M\^60 + 899568738\ M\^62 - 
                    2884728376\ M\^64 + 136333387\ M\^66 + 
                    1022247384\ M\^68 - 953184232\ M\^70 + 
                    1248958356\ M\^72 + 1188733776\ M\^74 - 
                    3230910424\ M\^76 - 681913008\ M\^78 + 
                    4224090994\ M\^80 - 316274200\ M\^82 - 
                    3910345046\ M\^84 + 1219325813\ M\^86 + 
                    2641065782\ M\^88 - 1525137549\ M\^90 - 
                    1195715389\ M\^92 + 1209521823\ M\^94 + 
                    213877178\ M\^96 - 642314643\ M\^98 + 158074545\ M\^100 + 
                    198593136\ M\^102 - 156678414\ M\^104 + 226703\ M\^106 + 
                    61039816\ M\^108 - 33225348\ M\^110 - 4135398\ M\^112 + 
                    13756192\ M\^114 - 6044515\ M\^116 - 1235701\ M\^118 + 
                    2970047\ M\^120 - 1914145\ M\^122 + 765612\ M\^124 - 
                    214325\ M\^126 + 43170\ M\^128 - 6188\ M\^130 + 
                    603\ M\^132 - 36\ M\^134 + M\^136)\) + 
              L\^13\ \((M\^6 - 36\ M\^8 + 603\ M\^10 - 6188\ M\^12 + 
                    43170\ M\^14 - 214325\ M\^16 + 765612\ M\^18 - 
                    1914145\ M\^20 + 2970047\ M\^22 - 1235701\ M\^24 - 
                    6044515\ M\^26 + 13756192\ M\^28 - 4135398\ M\^30 - 
                    33225348\ M\^32 + 61039816\ M\^34 + 226703\ M\^36 - 
                    156678414\ M\^38 + 198593136\ M\^40 + 158074545\ M\^42 - 
                    642314643\ M\^44 + 213877178\ M\^46 + 1209521823\ M\^48 - 
                    1195715389\ M\^50 - 1525137549\ M\^52 + 
                    2641065782\ M\^54 + 1219325813\ M\^56 - 
                    3910345046\ M\^58 - 316274200\ M\^60 + 
                    4224090994\ M\^62 - 681913008\ M\^64 - 
                    3230910424\ M\^66 + 1188733776\ M\^68 + 
                    1248958356\ M\^70 - 953184232\ M\^72 + 
                    1022247384\ M\^74 + 136333387\ M\^76 - 
                    2884728376\ M\^78 + 899568738\ M\^80 + 
                    3779817874\ M\^82 - 1813129579\ M\^84 - 
                    3490884195\ M\^86 + 2299678529\ M\^88 + 
                    2320526238\ M\^90 - 2160621485\ M\^92 - 
                    952571141\ M\^94 + 1504399992\ M\^96 + 19065184\ M\^98 - 
                    716555067\ M\^100 + 291336720\ M\^102 + 
                    164070727\ M\^104 - 213921253\ M\^106 + 
                    53478886\ M\^108 + 63551742\ M\^110 - 61075922\ M\^112 + 
                    8761585\ M\^114 + 18525327\ M\^116 - 13356712\ M\^118 + 
                    1311971\ M\^120 + 3617481\ M\^122 - 3002317\ M\^124 + 
                    1350931\ M\^126 - 411456\ M\^128 + 89400\ M\^130 - 
                    13858\ M\^132 + 1473\ M\^134 - 97\ M\^136 + 
                    3\ M\^138)\) + 
              L\^4\ \((M\^68 - 16\ M\^70 + 124\ M\^72 - 606\ M\^74 + 
                    2072\ M\^76 - 5138\ M\^78 + 8841\ M\^80 - 8464\ M\^82 - 
                    1813\ M\^84 + 18770\ M\^86 - 24656\ M\^88 + 190\ M\^90 + 
                    41291\ M\^92 - 44137\ M\^94 + 6126\ M\^96 + 5106\ M\^98 - 
                    56925\ M\^100 + 170952\ M\^102 - 9404\ M\^104 - 
                    347970\ M\^106 + 184331\ M\^108 + 311187\ M\^110 - 
                    293835\ M\^112 - 66744\ M\^114 + 218017\ M\^116 - 
                    97534\ M\^118 - 47122\ M\^120 + 75146\ M\^122 - 
                    33928\ M\^124 - 1694\ M\^126 + 12027\ M\^128 - 
                    8577\ M\^130 + 3536\ M\^132 - 958\ M\^134 + 172\ M\^136 - 
                    19\ M\^138 + M\^140)\) + 
              L\^12\ \((\(-3\)\ M\^10 + 91\ M\^12 - 1276\ M\^14 + 
                    10974\ M\^16 - 64429\ M\^18 + 270118\ M\^20 - 
                    813097\ M\^22 + 1678610\ M\^24 - 1937052\ M\^26 - 
                    460712\ M\^28 + 5893205\ M\^30 - 7400572\ M\^32 - 
                    7630001\ M\^34 + 34647378\ M\^36 - 26439418\ M\^38 - 
                    63940147\ M\^40 + 147927322\ M\^42 + 10867680\ M\^44 - 
                    384286318\ M\^46 + 300848657\ M\^48 + 631924360\ M\^50 - 
                    1031881916\ M\^52 - 613398151\ M\^54 + 
                    2094249644\ M\^56 + 59987245\ M\^58 - 3023394860\ M\^60 + 
                    949502298\ M\^62 + 3195269504\ M\^64 - 
                    1850162935\ M\^66 - 2311158904\ M\^68 + 
                    1954528442\ M\^70 + 676190082\ M\^72 - 
                    1031710033\ M\^74 + 1046155768\ M\^76 - 
                    467044149\ M\^78 - 2287461571\ M\^80 + 
                    1804727430\ M\^82 + 2757905224\ M\^84 - 
                    2530161200\ M\^86 - 2455552446\ M\^88 + 
                    2585883096\ M\^90 + 1611250425\ M\^92 - 
                    2110666204\ M\^94 - 629578362\ M\^96 + 
                    1345378213\ M\^98 - 61485816\ M\^100 - 
                    591049487\ M\^102 + 285039833\ M\^104 + 
                    95537323\ M\^106 - 193991006\ M\^108 + 84908753\ M\^110 + 
                    46600400\ M\^112 - 71329361\ M\^114 + 19276682\ M\^116 + 
                    18845392\ M\^118 - 17445368\ M\^120 + 3323959\ M\^122 + 
                    3615313\ M\^124 - 3452185\ M\^126 + 1609864\ M\^128 - 
                    493707\ M\^130 + 106178\ M\^132 - 16061\ M\^134 + 
                    1644\ M\^136 - 103\ M\^138 + 3\ M\^140)\) + 
              L\^8\ \((\(-M\^30\) + 24\ M\^32 - 270\ M\^34 + 1880\ M\^36 - 
                    9015\ M\^38 + 31294\ M\^40 - 79993\ M\^42 + 
                    147948\ M\^44 - 181909\ M\^46 + 105275\ M\^48 + 
                    41578\ M\^50 + 39457\ M\^52 - 583526\ M\^54 + 
                    780451\ M\^56 + 1054840\ M\^58 - 3327332\ M\^60 - 
                    1070907\ M\^62 + 10672720\ M\^64 - 3335811\ M\^66 - 
                    23702929\ M\^68 + 17397642\ M\^70 + 36043830\ M\^72 - 
                    40580750\ M\^74 - 28265409\ M\^76 + 50940156\ M\^78 - 
                    19037474\ M\^80 - 8405708\ M\^82 + 98821911\ M\^84 - 
                    105532994\ M\^86 - 168212279\ M\^88 + 245308159\ M\^90 + 
                    180593769\ M\^92 - 318399710\ M\^94 - 135197077\ M\^96 + 
                    271822103\ M\^98 + 74011071\ M\^100 - 151540493\ M\^102 - 
                    37107627\ M\^104 + 52715215\ M\^106 + 27011248\ M\^108 - 
                    15676725\ M\^110 - 19192636\ M\^112 + 12770607\ M\^114 + 
                    4669593\ M\^116 - 9499066\ M\^118 + 4688452\ M\^120 + 
                    1737646\ M\^122 - 3923089\ M\^124 + 1981412\ M\^126 + 
                    252626\ M\^128 - 872738\ M\^130 + 548702\ M\^132 - 
                    195252\ M\^134 + 44236\ M\^136 - 6364\ M\^138 + 
                    534\ M\^140 - 20\ M\^142)\) + 
              L\^6\ \((3\ M\^48 - 61\ M\^50 + 562\ M\^52 - 3061\ M\^54 + 
                    10799\ M\^56 - 25423\ M\^58 + 39575\ M\^60 - 
                    38855\ M\^62 + 22735\ M\^64 - 9661\ M\^66 - 7459\ M\^68 + 
                    60324\ M\^70 - 47195\ M\^72 - 236428\ M\^74 + 
                    458344\ M\^76 + 203008\ M\^78 - 1194051\ M\^80 + 
                    709880\ M\^82 + 1115173\ M\^84 - 2985583\ M\^86 + 
                    2808216\ M\^88 + 4671816\ M\^90 - 12158397\ M\^92 - 
                    2125010\ M\^94 + 22730409\ M\^96 - 5250737\ M\^98 - 
                    26203046\ M\^100 + 12246855\ M\^102 + 18598199\ M\^104 - 
                    12669396\ M\^106 - 6400190\ M\^108 + 6955290\ M\^110 - 
                    664352\ M\^112 - 734536\ M\^114 + 931103\ M\^116 - 
                    1439434\ M\^118 + 634291\ M\^120 + 552714\ M\^122 - 
                    747982\ M\^124 + 283032\ M\^126 + 75466\ M\^128 - 
                    146713\ M\^130 + 90155\ M\^132 - 34290\ M\^134 + 
                    8703\ M\^136 - 1435\ M\^138 + 139\ M\^140 - 
                    6\ M\^142)\) + 
              L\^10\ \((\(-M\^18\) + 30\ M\^20 - 424\ M\^22 + 3668\ M\^24 - 
                    21436\ M\^26 + 88313\ M\^28 - 257725\ M\^30 + 
                    508987\ M\^32 - 564944\ M\^34 + 9702\ M\^36 + 
                    706056\ M\^38 + 524617\ M\^40 - 4184798\ M\^42 + 
                    2873154\ M\^44 + 11318021\ M\^46 - 18328864\ M\^48 - 
                    22079526\ M\^50 + 70639750\ M\^52 + 13556525\ M\^54 - 
                    186554798\ M\^56 + 68574544\ M\^58 + 362916225\ M\^60 - 
                    289624488\ M\^62 - 519689172\ M\^64 + 639119841\ M\^66 + 
                    508712996\ M\^68 - 938221828\ M\^70 - 224462217\ M\^72 + 
                    885544976\ M\^74 - 260251826\ M\^76 - 300158584\ M\^78 + 
                    701701478\ M\^80 - 626169656\ M\^82 - 880103774\ M\^84 + 
                    1406582508\ M\^86 + 778485916\ M\^88 - 
                    1654085637\ M\^90 - 561243061\ M\^92 + 
                    1376355056\ M\^94 + 359363107\ M\^96 - 882683219\ M\^98 - 
                    183975160\ M\^100 + 463225587\ M\^102 + 
                    33233706\ M\^104 - 196526283\ M\^106 + 57913769\ M\^108 + 
                    42482223\ M\^110 - 67581590\ M\^112 + 28993211\ M\^114 + 
                    28350776\ M\^116 - 35203519\ M\^118 + 3758586\ M\^120 + 
                    14277817\ M\^122 - 9184478\ M\^124 + 143539\ M\^126 + 
                    2794848\ M\^128 - 1901184\ M\^130 + 711334\ M\^132 - 
                    174147\ M\^134 + 28788\ M\^136 - 3127\ M\^138 + 
                    203\ M\^140 - 6\ M\^142)\) + 
              L\^11\ \((3\ M\^14 - 86\ M\^16 + 1139\ M\^18 - 9232\ M\^20 + 
                    50902\ M\^22 - 199558\ M\^24 + 558331\ M\^26 - 
                    1056557\ M\^28 + 1054195\ M\^30 + 504420\ M\^32 - 
                    3031925\ M\^34 + 1568059\ M\^36 + 8930219\ M\^38 - 
                    17402053\ M\^40 - 6650967\ M\^42 + 59102578\ M\^44 - 
                    38460019\ M\^46 - 129046397\ M\^48 + 204234154\ M\^50 + 
                    166515781\ M\^52 - 567209127\ M\^54 - 24281031\ M\^56 + 
                    1094444211\ M\^58 - 446484099\ M\^60 - 
                    1552954115\ M\^62 + 1190755693\ M\^64 + 
                    1589791805\ M\^66 - 1810027040\ M\^68 - 
                    1013055713\ M\^70 + 1756273169\ M\^72 + 16767180\ M\^74 - 
                    801703351\ M\^76 + 936199735\ M\^78 - 653879339\ M\^80 - 
                    1476079436\ M\^82 + 1864858394\ M\^84 + 
                    1530780776\ M\^86 - 2347798000\ M\^88 - 
                    1255411116\ M\^90 + 2134579752\ M\^92 + 
                    813365558\ M\^94 - 1550060749\ M\^96 - 328318776\ M\^98 + 
                    907064742\ M\^100 - 40926022\ M\^102 - 
                    380878213\ M\^104 + 181356717\ M\^106 + 
                    52267838\ M\^108 - 133112226\ M\^110 + 71219962\ M\^112 + 
                    33813242\ M\^114 - 60271289\ M\^116 + 16306529\ M\^118 + 
                    17547073\ M\^120 - 15653244\ M\^122 + 2600134\ M\^124 + 
                    3414384\ M\^126 - 3010280\ M\^128 + 1316440\ M\^130 - 
                    374749\ M\^132 + 73526\ M\^134 - 9907\ M\^136 + 
                    874\ M\^138 - 45\ M\^140 + M\^142)\) + 
              L\^5\ \((3\ M\^58 - 54\ M\^60 + 443\ M\^62 - 2158\ M\^64 + 
                    6844\ M\^66 - 14532\ M\^68 + 19932\ M\^70 - 
                    14175\ M\^72 - 2452\ M\^74 + 11450\ M\^76 - 3613\ M\^78 + 
                    11886\ M\^80 - 40751\ M\^82 - 6262\ M\^84 + 
                    171420\ M\^86 - 275128\ M\^88 - 64127\ M\^90 + 
                    920967\ M\^92 - 799691\ M\^94 - 1609643\ M\^96 + 
                    2578583\ M\^98 + 1578367\ M\^100 - 4291466\ M\^102 - 
                    444076\ M\^104 + 4331951\ M\^106 - 1022001\ M\^108 - 
                    2491848\ M\^110 + 1615831\ M\^112 + 422963\ M\^114 - 
                    944591\ M\^116 + 445186\ M\^118 + 60124\ M\^120 - 
                    254252\ M\^122 + 192226\ M\^124 - 46666\ M\^126 - 
                    37437\ M\^128 + 43422\ M\^130 - 22598\ M\^132 + 
                    7529\ M\^134 - 1714\ M\^136 + 261\ M\^138 - 24\ M\^140 + 
                    M\^142)\) + 
              L\^9\ \((\(-2\)\ M\^24 + 50\ M\^26 - 590\ M\^28 + 4287\ M\^30 - 
                    21048\ M\^32 + 72156\ M\^34 - 171746\ M\^36 + 
                    270350\ M\^38 - 252762\ M\^40 + 155803\ M\^42 - 
                    312949\ M\^44 + 565384\ M\^46 + 488367\ M\^48 - 
                    2305146\ M\^50 - 1377393\ M\^52 + 10531764\ M\^54 - 
                    1986509\ M\^56 - 31635835\ M\^58 + 21249502\ M\^60 + 
                    70743166\ M\^62 - 77780221\ M\^64 - 113852339\ M\^66 + 
                    180188900\ M\^68 + 118469429\ M\^70 - 285697115\ M\^72 - 
                    33412547\ M\^74 + 284222013\ M\^76 - 144099165\ M\^78 - 
                    68433692\ M\^80 + 331696631\ M\^82 - 331934479\ M\^84 - 
                    414678061\ M\^86 + 711068813\ M\^88 + 353261790\ M\^90 - 
                    838385917\ M\^92 - 226435701\ M\^94 + 670060875\ M\^96 + 
                    129042342\ M\^98 - 378671215\ M\^100 - 79586084\ M\^102 + 
                    163360128\ M\^104 + 45549523\ M\^106 - 68857103\ M\^108 - 
                    7041720\ M\^110 + 30560685\ M\^112 - 19446762\ M\^114 - 
                    3300800\ M\^116 + 18508699\ M\^118 - 10314617\ M\^120 - 
                    4752065\ M\^122 + 7871146\ M\^124 - 2617867\ M\^126 - 
                    1225589\ M\^128 + 1600187\ M\^130 - 790133\ M\^132 + 
                    236667\ M\^134 - 46421\ M\^136 + 5888\ M\^138 - 
                    442\ M\^140 + 15\ M\^142)\) + 
              L\^7\ \((M\^38 - 24\ M\^40 + 263\ M\^42 - 1731\ M\^44 + 
                    7606\ M\^46 - 23567\ M\^48 + 53172\ M\^50 - 
                    88185\ M\^52 + 101224\ M\^54 - 50825\ M\^56 - 
                    70448\ M\^58 + 118177\ M\^60 + 201294\ M\^62 - 
                    791162\ M\^64 + 453791\ M\^66 + 1795865\ M\^68 - 
                    2948470\ M\^70 - 1995658\ M\^72 + 7727205\ M\^74 - 
                    1101892\ M\^76 - 10689100\ M\^78 + 6875396\ M\^80 + 
                    3462227\ M\^82 - 8227093\ M\^84 + 20790770\ M\^86 - 
                    5949309\ M\^88 - 55952381\ M\^90 + 37767348\ M\^92 + 
                    80412417\ M\^94 - 69101721\ M\^96 - 75487088\ M\^98 + 
                    74247169\ M\^100 + 45637735\ M\^102 - 48543679\ M\^104 - 
                    16429402\ M\^106 + 15987127\ M\^108 + 5814762\ M\^110 + 
                    743576\ M\^112 - 6562770\ M\^114 - 968332\ M\^116 + 
                    5367348\ M\^118 - 2270178\ M\^120 - 1275579\ M\^122 + 
                    1895649\ M\^124 - 851735\ M\^126 - 99600\ M\^128 + 
                    378551\ M\^130 - 258879\ M\^132 + 101498\ M\^134 - 
                    25401\ M\^136 + 4022\ M\^138 - 369\ M\^140 + 
                    15\ M\^142)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 26]\), \(L\^23\ M\^24 + M\^118 + 
              L\^22\ \((4\ M\^18 - 16\ M\^20 + 26\ M\^22 - 15\ M\^24 - 
                    28\ M\^26 + 72\ M\^28 - 4\ M\^30 - 35\ M\^32 + 
                    25\ M\^34 - 6\ M\^36)\) + 
              L\^21\ \((6\ M\^12 - 44\ M\^14 + 132\ M\^16 - 192\ M\^18 + 
                    61\ M\^20 + 202\ M\^22 - 235\ M\^24 + 236\ M\^26 - 
                    556\ M\^28 - 31\ M\^30 + 1579\ M\^32 - 870\ M\^34 - 
                    806\ M\^36 + 1221\ M\^38 - 301\ M\^40 - 404\ M\^42 + 
                    358\ M\^44 - 118\ M\^46 + 15\ M\^48)\) + 
              L\^20\ \((4\ M\^6 - 40\ M\^8 + 170\ M\^10 - 392\ M\^12 + 
                    508\ M\^14 - 444\ M\^16 + 639\ M\^18 - 419\ M\^20 - 
                    2886\ M\^22 + 6719\ M\^24 - 1435\ M\^26 - 10699\ M\^28 + 
                    12595\ M\^30 - 114\ M\^32 - 13547\ M\^34 + 14590\ M\^36 + 
                    746\ M\^38 - 9570\ M\^40 + 9567\ M\^42 - 4173\ M\^44 - 
                    4068\ M\^46 + 7090\ M\^48 - 2924\ M\^50 - 1523\ M\^52 + 
                    2180\ M\^54 - 1005\ M\^56 + 222\ M\^58 - 20\ M\^60)\) + 
              L\^19\ \((1 - 12\ M\^2 + 64\ M\^4 - 204\ M\^6 + 496\ M\^8 - 
                    1244\ M\^10 + 3028\ M\^12 - 4378\ M\^14 - 1096\ M\^16 + 
                    14997\ M\^18 - 15969\ M\^20 - 16683\ M\^22 + 
                    41491\ M\^24 + 5935\ M\^26 - 64842\ M\^28 - 3047\ M\^30 + 
                    114004\ M\^32 - 12547\ M\^34 - 189986\ M\^36 + 
                    84858\ M\^38 + 209819\ M\^40 - 176102\ M\^42 - 
                    97123\ M\^44 + 195486\ M\^46 - 40327\ M\^48 - 
                    101362\ M\^50 + 79812\ M\^52 - 1091\ M\^54 - 
                    31289\ M\^56 + 23779\ M\^58 - 7167\ M\^60 - 4289\ M\^62 + 
                    6697\ M\^64 - 3888\ M\^66 + 1227\ M\^68 - 208\ M\^70 + 
                    15\ M\^72)\) + 
              L\^18\ \((\(-5\) + 76\ M\^2 - 513\ M\^4 + 1951\ M\^6 - 
                    4305\ M\^8 + 4759\ M\^10 - 658\ M\^12 - 1561\ M\^14 - 
                    3481\ M\^16 - 12457\ M\^18 + 66757\ M\^20 - 
                    36646\ M\^22 - 200651\ M\^24 + 307795\ M\^26 + 
                    222102\ M\^28 - 791023\ M\^30 + 94465\ M\^32 + 
                    1186677\ M\^34 - 717829\ M\^36 - 1136738\ M\^38 + 
                    1275976\ M\^40 + 659202\ M\^42 - 1467015\ M\^44 - 
                    91640\ M\^46 + 1366853\ M\^48 - 243214\ M\^50 - 
                    1052191\ M\^52 + 427609\ M\^54 + 618371\ M\^56 - 
                    474330\ M\^58 - 188508\ M\^60 + 333312\ M\^62 - 
                    63914\ M\^64 - 107555\ M\^66 + 80246\ M\^68 - 
                    11416\ M\^70 - 16700\ M\^72 + 15785\ M\^74 - 
                    8295\ M\^76 + 2966\ M\^78 - 699\ M\^80 + 97\ M\^82 - 
                    6\ M\^84)\) + 
              L\^17\ \((10 - 182\ M\^2 + 1473\ M\^4 - 6724\ M\^6 + 
                    17657\ M\^8 - 20452\ M\^10 - 20400\ M\^12 + 
                    96529\ M\^14 - 57324\ M\^16 - 220081\ M\^18 + 
                    357673\ M\^20 + 268549\ M\^22 - 1004488\ M\^24 - 
                    37679\ M\^26 + 2139808\ M\^28 - 930175\ M\^30 - 
                    3607879\ M\^32 + 3175619\ M\^34 + 4698927\ M\^36 - 
                    6832254\ M\^38 - 4188938\ M\^40 + 10730746\ M\^42 + 
                    1439263\ M\^44 - 12593306\ M\^46 + 2478935\ M\^48 + 
                    10849115\ M\^50 - 5091906\ M\^52 - 6300552\ M\^54 + 
                    5196783\ M\^56 + 2074801\ M\^58 - 3441149\ M\^60 - 
                    81037\ M\^62 + 1625630\ M\^64 - 240705\ M\^66 - 
                    692034\ M\^68 + 166074\ M\^70 + 342891\ M\^72 - 
                    178894\ M\^74 - 103270\ M\^76 + 125936\ M\^78 - 
                    23473\ M\^80 - 31462\ M\^82 + 28631\ M\^84 - 
                    12936\ M\^86 + 3969\ M\^88 - 912\ M\^90 + 157\ M\^92 - 
                    18\ M\^94 + M\^96)\) + 
              L\^16\ \((\(-10\) + 210\ M\^2 - 1967\ M\^4 + 10526\ M\^6 - 
                    33622\ M\^8 + 55453\ M\^10 + 7631\ M\^12 - 
                    230309\ M\^14 + 346901\ M\^16 + 299210\ M\^18 - 
                    1484340\ M\^20 + 806935\ M\^22 + 3127268\ M\^24 - 
                    4698508\ M\^26 - 3412491\ M\^28 + 12248499\ M\^30 - 
                    1138086\ M\^32 - 21287549\ M\^34 + 12730577\ M\^36 + 
                    26722216\ M\^38 - 28971213\ M\^40 - 23841009\ M\^42 + 
                    42580354\ M\^44 + 12674847\ M\^46 - 46524147\ M\^48 + 
                    1788171\ M\^50 + 39400244\ M\^52 - 13299248\ M\^54 - 
                    25626767\ M\^56 + 18696877\ M\^58 + 12135720\ M\^60 - 
                    17089575\ M\^62 - 2680637\ M\^64 + 11031839\ M\^66 - 
                    1768736\ M\^68 - 4636367\ M\^70 + 2133496\ M\^72 + 
                    867868\ M\^74 - 865745\ M\^76 + 149250\ M\^78 - 
                    1814\ M\^80 + 24427\ M\^82 + 49290\ M\^84 - 
                    95504\ M\^86 + 49393\ M\^88 + 15575\ M\^90 - 
                    38757\ M\^92 + 27936\ M\^94 - 11943\ M\^96 + 
                    3324\ M\^98 - 596\ M\^100 + 63\ M\^102 - 3\ M\^104)\) + 
              L\^15\ \((5 - 118\ M\^2 + 1252\ M\^4 - 7754\ M\^6 + 
                    30155\ M\^8 - 71020\ M\^10 + 70947\ M\^12 + 
                    114902\ M\^14 - 475261\ M\^16 + 386658\ M\^18 + 
                    965613\ M\^20 - 2444763\ M\^22 + 321759\ M\^24 + 
                    5702771\ M\^26 - 6151400\ M\^28 - 6853237\ M\^30 + 
                    17511079\ M\^32 + 23475\ M\^34 - 29692301\ M\^36 + 
                    17939647\ M\^38 + 33492190\ M\^40 - 42929609\ M\^42 - 
                    21397113\ M\^44 + 64408258\ M\^46 - 5124678\ M\^48 - 
                    73271392\ M\^50 + 35255399\ M\^52 + 68114970\ M\^54 - 
                    55644803\ M\^56 - 54701815\ M\^58 + 60010833\ M\^60 + 
                    40832829\ M\^62 - 50394478\ M\^64 - 28319034\ M\^66 + 
                    36418411\ M\^68 + 17025437\ M\^70 - 24602623\ M\^72 - 
                    7460476\ M\^74 + 15460802\ M\^76 + 900555\ M\^78 - 
                    7927188\ M\^80 + 1631942\ M\^82 + 2711676\ M\^84 - 
                    1306249\ M\^86 - 356514\ M\^88 + 294806\ M\^90 + 
                    7663\ M\^92 + 39492\ M\^94 - 71634\ M\^96 - 4067\ M\^98 + 
                    52270\ M\^100 - 41568\ M\^102 + 17502\ M\^104 - 
                    4569\ M\^106 + 748\ M\^108 - 71\ M\^110 + 3\ M\^112)\) + 
              L\^14\ \((\(-1\) + 26\ M\^2 - 313\ M\^4 + 2305\ M\^6 - 
                    11485\ M\^8 + 39851\ M\^10 - 92320\ M\^12 + 
                    110744\ M\^14 + 82649\ M\^16 - 600044\ M\^18 + 
                    855439\ M\^20 + 633705\ M\^22 - 3866299\ M\^24 + 
                    3546340\ M\^26 + 6593360\ M\^28 - 16783219\ M\^30 + 
                    87870\ M\^32 + 39207079\ M\^34 - 31006540\ M\^36 - 
                    58514181\ M\^38 + 95588481\ M\^40 + 52226615\ M\^42 - 
                    185603181\ M\^44 - 1245850\ M\^46 + 271652810\ M\^48 - 
                    92093272\ M\^50 - 316895359\ M\^52 + 199393487\ M\^54 + 
                    297928957\ M\^56 - 280213863\ M\^58 - 219906869\ M\^60 + 
                    305893398\ M\^62 + 114550868\ M\^64 - 272303075\ M\^66 - 
                    19103664\ M\^68 + 202760614\ M\^70 - 38089723\ M\^72 - 
                    126458993\ M\^74 + 52807343\ M\^76 + 65142141\ M\^78 - 
                    41399123\ M\^80 - 27295831\ M\^82 + 23732989\ M\^84 + 
                    9471411\ M\^86 - 11173281\ M\^88 - 2661739\ M\^90 + 
                    4795921\ M\^92 + 282267\ M\^94 - 1936595\ M\^96 + 
                    505782\ M\^98 + 462891\ M\^100 - 361636\ M\^102 + 
                    35913\ M\^104 + 94381\ M\^106 - 79238\ M\^108 + 
                    36367\ M\^110 - 11172\ M\^112 + 2353\ M\^114 - 
                    327\ M\^116 + 27\ M\^118 - M\^120)\) + 
              L\ \((\(-6\)\ M\^106 + 25\ M\^108 - 35\ M\^110 - 4\ M\^112 + 
                    72\ M\^114 - 28\ M\^116 - 15\ M\^118 + 26\ M\^120 - 
                    16\ M\^122 + 4\ M\^124)\) + 
              L\^13\ \((4\ M\^4 - 102\ M\^6 + 1167\ M\^8 - 7867\ M\^10 + 
                    34121\ M\^12 - 95169\ M\^14 + 146863\ M\^16 + 
                    1336\ M\^18 - 527204\ M\^20 + 932115\ M\^22 + 
                    216347\ M\^24 - 3105449\ M\^26 + 3483548\ M\^28 + 
                    3593384\ M\^30 - 11590438\ M\^32 + 3307544\ M\^34 + 
                    19521777\ M\^36 - 21396839\ M\^38 - 15484461\ M\^40 + 
                    43399940\ M\^42 - 11594421\ M\^44 - 49575154\ M\^46 + 
                    62880584\ M\^48 + 16125786\ M\^50 - 124644468\ M\^52 + 
                    67526932\ M\^54 + 174465438\ M\^56 - 187353301\ M\^58 - 
                    192054238\ M\^60 + 306111440\ M\^62 + 169282525\ M\^64 - 
                    381608794\ M\^66 - 112641112\ M\^68 + 391271597\ M\^70 + 
                    44970892\ M\^72 - 335576432\ M\^74 + 13729841\ M\^76 + 
                    241071736\ M\^78 - 49531315\ M\^80 - 142845073\ M\^82 + 
                    58506118\ M\^84 + 65847857\ M\^86 - 47266639\ M\^88 - 
                    19269906\ M\^90 + 28293885\ M\^92 - 1121503\ M\^94 - 
                    11771355\ M\^96 + 5247248\ M\^98 + 2471989\ M\^100 - 
                    3203634\ M\^102 + 659274\ M\^104 + 770691\ M\^106 - 
                    602275\ M\^108 + 78769\ M\^110 + 136556\ M\^112 - 
                    113219\ M\^114 + 48414\ M\^116 - 13592\ M\^118 + 
                    2622\ M\^120 - 340\ M\^122 + 27\ M\^124 - M\^126)\) + 
              L\^12\ \((\(-6\)\ M\^8 + 151\ M\^10 - 1701\ M\^12 + 
                    11265\ M\^14 - 47790\ M\^16 + 128770\ M\^18 - 
                    181522\ M\^20 - 69790\ M\^22 + 827412\ M\^24 - 
                    1203816\ M\^26 - 976365\ M\^28 + 5248363\ M\^30 - 
                    3582101\ M\^32 - 10450318\ M\^34 + 18563595\ M\^36 + 
                    10137073\ M\^38 - 47831901\ M\^40 + 6784121\ M\^42 + 
                    87971299\ M\^44 - 46903737\ M\^46 - 129920946\ M\^48 + 
                    102971494\ M\^50 + 165281033\ M\^52 - 148876541\ M\^54 - 
                    195787058\ M\^56 + 151450456\ M\^58 + 232156666\ M\^60 - 
                    92921907\ M\^62 - 283219316\ M\^64 - 10722591\ M\^66 + 
                    342373624\ M\^68 + 114086307\ M\^70 - 388084136\ M\^72 - 
                    165750417\ M\^74 + 402843943\ M\^76 + 149505848\ M\^78 - 
                    373122408\ M\^80 - 84411141\ M\^82 + 300370810\ M\^84 + 
                    10257595\ M\^86 - 203661260\ M\^88 + 38518830\ M\^90 + 
                    110268242\ M\^92 - 50756161\ M\^94 - 42507688\ M\^96 + 
                    38084823\ M\^98 + 6800185\ M\^100 - 18852724\ M\^102 + 
                    4278495\ M\^104 + 5534163\ M\^106 - 3873897\ M\^108 - 
                    111556\ M\^110 + 1230590\ M\^112 - 552801\ M\^114 - 
                    55941\ M\^116 + 179936\ M\^118 - 103874\ M\^120 + 
                    35224\ M\^122 - 7883\ M\^124 + 1155\ M\^126 - 
                    101\ M\^128 + 4\ M\^130)\) + 
              L\^2\ \((15\ M\^94 - 118\ M\^96 + 358\ M\^98 - 404\ M\^100 - 
                    301\ M\^102 + 1221\ M\^104 - 806\ M\^106 - 870\ M\^108 + 
                    1579\ M\^110 - 31\ M\^112 - 556\ M\^114 + 236\ M\^116 - 
                    235\ M\^118 + 202\ M\^120 + 61\ M\^122 - 192\ M\^124 + 
                    132\ M\^126 - 44\ M\^128 + 6\ M\^130)\) + 
              L\^11\ \((4\ M\^12 - 101\ M\^14 + 1155\ M\^16 - 7883\ M\^18 + 
                    35224\ M\^20 - 103874\ M\^22 + 179936\ M\^24 - 
                    55941\ M\^26 - 552801\ M\^28 + 1230590\ M\^30 - 
                    111556\ M\^32 - 3873897\ M\^34 + 5534163\ M\^36 + 
                    4278495\ M\^38 - 18852724\ M\^40 + 6800185\ M\^42 + 
                    38084823\ M\^44 - 42507688\ M\^46 - 50756161\ M\^48 + 
                    110268242\ M\^50 + 38518830\ M\^52 - 203661260\ M\^54 + 
                    10257595\ M\^56 + 300370810\ M\^58 - 84411141\ M\^60 - 
                    373122408\ M\^62 + 149505848\ M\^64 + 402843943\ M\^66 - 
                    165750417\ M\^68 - 388084136\ M\^70 + 114086307\ M\^72 + 
                    342373624\ M\^74 - 10722591\ M\^76 - 283219316\ M\^78 - 
                    92921907\ M\^80 + 232156666\ M\^82 + 151450456\ M\^84 - 
                    195787058\ M\^86 - 148876541\ M\^88 + 165281033\ M\^90 + 
                    102971494\ M\^92 - 129920946\ M\^94 - 46903737\ M\^96 + 
                    87971299\ M\^98 + 6784121\ M\^100 - 47831901\ M\^102 + 
                    10137073\ M\^104 + 18563595\ M\^106 - 10450318\ M\^108 - 
                    3582101\ M\^110 + 5248363\ M\^112 - 976365\ M\^114 - 
                    1203816\ M\^116 + 827412\ M\^118 - 69790\ M\^120 - 
                    181522\ M\^122 + 128770\ M\^124 - 47790\ M\^126 + 
                    11265\ M\^128 - 1701\ M\^130 + 151\ M\^132 - 
                    6\ M\^134)\) + 
              L\^3\ \((\(-20\)\ M\^82 + 222\ M\^84 - 1005\ M\^86 + 
                    2180\ M\^88 - 1523\ M\^90 - 2924\ M\^92 + 7090\ M\^94 - 
                    4068\ M\^96 - 4173\ M\^98 + 9567\ M\^100 - 9570\ M\^102 + 
                    746\ M\^104 + 14590\ M\^106 - 13547\ M\^108 - 
                    114\ M\^110 + 12595\ M\^112 - 10699\ M\^114 - 
                    1435\ M\^116 + 6719\ M\^118 - 2886\ M\^120 - 
                    419\ M\^122 + 639\ M\^124 - 444\ M\^126 + 508\ M\^128 - 
                    392\ M\^130 + 170\ M\^132 - 40\ M\^134 + 4\ M\^136)\) + 
              L\^10\ \((\(-M\^16\) + 27\ M\^18 - 340\ M\^20 + 2622\ M\^22 - 
                    13592\ M\^24 + 48414\ M\^26 - 113219\ M\^28 + 
                    136556\ M\^30 + 78769\ M\^32 - 602275\ M\^34 + 
                    770691\ M\^36 + 659274\ M\^38 - 3203634\ M\^40 + 
                    2471989\ M\^42 + 5247248\ M\^44 - 11771355\ M\^46 - 
                    1121503\ M\^48 + 28293885\ M\^50 - 19269906\ M\^52 - 
                    47266639\ M\^54 + 65847857\ M\^56 + 58506118\ M\^58 - 
                    142845073\ M\^60 - 49531315\ M\^62 + 241071736\ M\^64 + 
                    13729841\ M\^66 - 335576432\ M\^68 + 44970892\ M\^70 + 
                    391271597\ M\^72 - 112641112\ M\^74 - 381608794\ M\^76 + 
                    169282525\ M\^78 + 306111440\ M\^80 - 192054238\ M\^82 - 
                    187353301\ M\^84 + 174465438\ M\^86 + 67526932\ M\^88 - 
                    124644468\ M\^90 + 16125786\ M\^92 + 62880584\ M\^94 - 
                    49575154\ M\^96 - 11594421\ M\^98 + 43399940\ M\^100 - 
                    15484461\ M\^102 - 21396839\ M\^104 + 19521777\ M\^106 + 
                    3307544\ M\^108 - 11590438\ M\^110 + 3593384\ M\^112 + 
                    3483548\ M\^114 - 3105449\ M\^116 + 216347\ M\^118 + 
                    932115\ M\^120 - 527204\ M\^122 + 1336\ M\^124 + 
                    146863\ M\^126 - 95169\ M\^128 + 34121\ M\^130 - 
                    7867\ M\^132 + 1167\ M\^134 - 102\ M\^136 + 
                    4\ M\^138)\) + 
              L\^7\ \((\(-3\)\ M\^38 + 63\ M\^40 - 596\ M\^42 + 3324\ M\^44 - 
                    11943\ M\^46 + 27936\ M\^48 - 38757\ M\^50 + 
                    15575\ M\^52 + 49393\ M\^54 - 95504\ M\^56 + 
                    49290\ M\^58 + 24427\ M\^60 - 1814\ M\^62 + 
                    149250\ M\^64 - 865745\ M\^66 + 867868\ M\^68 + 
                    2133496\ M\^70 - 4636367\ M\^72 - 1768736\ M\^74 + 
                    11031839\ M\^76 - 2680637\ M\^78 - 17089575\ M\^80 + 
                    12135720\ M\^82 + 18696877\ M\^84 - 25626767\ M\^86 - 
                    13299248\ M\^88 + 39400244\ M\^90 + 1788171\ M\^92 - 
                    46524147\ M\^94 + 12674847\ M\^96 + 42580354\ M\^98 - 
                    23841009\ M\^100 - 28971213\ M\^102 + 26722216\ M\^104 + 
                    12730577\ M\^106 - 21287549\ M\^108 - 1138086\ M\^110 + 
                    12248499\ M\^112 - 3412491\ M\^114 - 4698508\ M\^116 + 
                    3127268\ M\^118 + 806935\ M\^120 - 1484340\ M\^122 + 
                    299210\ M\^124 + 346901\ M\^126 - 230309\ M\^128 + 
                    7631\ M\^130 + 55453\ M\^132 - 33622\ M\^134 + 
                    10526\ M\^136 - 1967\ M\^138 + 210\ M\^140 - 
                    10\ M\^142)\) + 
              L\^5\ \((\(-6\)\ M\^58 + 97\ M\^60 - 699\ M\^62 + 2966\ M\^64 - 
                    8295\ M\^66 + 15785\ M\^68 - 16700\ M\^70 - 
                    11416\ M\^72 + 80246\ M\^74 - 107555\ M\^76 - 
                    63914\ M\^78 + 333312\ M\^80 - 188508\ M\^82 - 
                    474330\ M\^84 + 618371\ M\^86 + 427609\ M\^88 - 
                    1052191\ M\^90 - 243214\ M\^92 + 1366853\ M\^94 - 
                    91640\ M\^96 - 1467015\ M\^98 + 659202\ M\^100 + 
                    1275976\ M\^102 - 1136738\ M\^104 - 717829\ M\^106 + 
                    1186677\ M\^108 + 94465\ M\^110 - 791023\ M\^112 + 
                    222102\ M\^114 + 307795\ M\^116 - 200651\ M\^118 - 
                    36646\ M\^120 + 66757\ M\^122 - 12457\ M\^124 - 
                    3481\ M\^126 - 1561\ M\^128 - 658\ M\^130 + 
                    4759\ M\^132 - 4305\ M\^134 + 1951\ M\^136 - 
                    513\ M\^138 + 76\ M\^140 - 5\ M\^142)\) + 
              L\^9\ \((\(-M\^22\) + 27\ M\^24 - 327\ M\^26 + 2353\ M\^28 - 
                    11172\ M\^30 + 36367\ M\^32 - 79238\ M\^34 + 
                    94381\ M\^36 + 35913\ M\^38 - 361636\ M\^40 + 
                    462891\ M\^42 + 505782\ M\^44 - 1936595\ M\^46 + 
                    282267\ M\^48 + 4795921\ M\^50 - 2661739\ M\^52 - 
                    11173281\ M\^54 + 9471411\ M\^56 + 23732989\ M\^58 - 
                    27295831\ M\^60 - 41399123\ M\^62 + 65142141\ M\^64 + 
                    52807343\ M\^66 - 126458993\ M\^68 - 38089723\ M\^70 + 
                    202760614\ M\^72 - 19103664\ M\^74 - 272303075\ M\^76 + 
                    114550868\ M\^78 + 305893398\ M\^80 - 219906869\ M\^82 - 
                    280213863\ M\^84 + 297928957\ M\^86 + 199393487\ M\^88 - 
                    316895359\ M\^90 - 92093272\ M\^92 + 271652810\ M\^94 - 
                    1245850\ M\^96 - 185603181\ M\^98 + 52226615\ M\^100 + 
                    95588481\ M\^102 - 58514181\ M\^104 - 31006540\ M\^106 + 
                    39207079\ M\^108 + 87870\ M\^110 - 16783219\ M\^112 + 
                    6593360\ M\^114 + 3546340\ M\^116 - 3866299\ M\^118 + 
                    633705\ M\^120 + 855439\ M\^122 - 600044\ M\^124 + 
                    82649\ M\^126 + 110744\ M\^128 - 92320\ M\^130 + 
                    39851\ M\^132 - 11485\ M\^134 + 2305\ M\^136 - 
                    313\ M\^138 + 26\ M\^140 - M\^142)\) + 
              L\^4\ \((15\ M\^70 - 208\ M\^72 + 1227\ M\^74 - 3888\ M\^76 + 
                    6697\ M\^78 - 4289\ M\^80 - 7167\ M\^82 + 23779\ M\^84 - 
                    31289\ M\^86 - 1091\ M\^88 + 79812\ M\^90 - 
                    101362\ M\^92 - 40327\ M\^94 + 195486\ M\^96 - 
                    97123\ M\^98 - 176102\ M\^100 + 209819\ M\^102 + 
                    84858\ M\^104 - 189986\ M\^106 - 12547\ M\^108 + 
                    114004\ M\^110 - 3047\ M\^112 - 64842\ M\^114 + 
                    5935\ M\^116 + 41491\ M\^118 - 16683\ M\^120 - 
                    15969\ M\^122 + 14997\ M\^124 - 1096\ M\^126 - 
                    4378\ M\^128 + 3028\ M\^130 - 1244\ M\^132 + 
                    496\ M\^134 - 204\ M\^136 + 64\ M\^138 - 12\ M\^140 + 
                    M\^142)\) + 
              L\^8\ \((3\ M\^30 - 71\ M\^32 + 748\ M\^34 - 4569\ M\^36 + 
                    17502\ M\^38 - 41568\ M\^40 + 52270\ M\^42 - 
                    4067\ M\^44 - 71634\ M\^46 + 39492\ M\^48 + 7663\ M\^50 + 
                    294806\ M\^52 - 356514\ M\^54 - 1306249\ M\^56 + 
                    2711676\ M\^58 + 1631942\ M\^60 - 7927188\ M\^62 + 
                    900555\ M\^64 + 15460802\ M\^66 - 7460476\ M\^68 - 
                    24602623\ M\^70 + 17025437\ M\^72 + 36418411\ M\^74 - 
                    28319034\ M\^76 - 50394478\ M\^78 + 40832829\ M\^80 + 
                    60010833\ M\^82 - 54701815\ M\^84 - 55644803\ M\^86 + 
                    68114970\ M\^88 + 35255399\ M\^90 - 73271392\ M\^92 - 
                    5124678\ M\^94 + 64408258\ M\^96 - 21397113\ M\^98 - 
                    42929609\ M\^100 + 33492190\ M\^102 + 17939647\ M\^104 - 
                    29692301\ M\^106 + 23475\ M\^108 + 17511079\ M\^110 - 
                    6853237\ M\^112 - 6151400\ M\^114 + 5702771\ M\^116 + 
                    321759\ M\^118 - 2444763\ M\^120 + 965613\ M\^122 + 
                    386658\ M\^124 - 475261\ M\^126 + 114902\ M\^128 + 
                    70947\ M\^130 - 71020\ M\^132 + 30155\ M\^134 - 
                    7754\ M\^136 + 1252\ M\^138 - 118\ M\^140 + 
                    5\ M\^142)\) + 
              L\^6\ \((M\^46 - 18\ M\^48 + 157\ M\^50 - 912\ M\^52 + 
                    3969\ M\^54 - 12936\ M\^56 + 28631\ M\^58 - 
                    31462\ M\^60 - 23473\ M\^62 + 125936\ M\^64 - 
                    103270\ M\^66 - 178894\ M\^68 + 342891\ M\^70 + 
                    166074\ M\^72 - 692034\ M\^74 - 240705\ M\^76 + 
                    1625630\ M\^78 - 81037\ M\^80 - 3441149\ M\^82 + 
                    2074801\ M\^84 + 5196783\ M\^86 - 6300552\ M\^88 - 
                    5091906\ M\^90 + 10849115\ M\^92 + 2478935\ M\^94 - 
                    12593306\ M\^96 + 1439263\ M\^98 + 10730746\ M\^100 - 
                    4188938\ M\^102 - 6832254\ M\^104 + 4698927\ M\^106 + 
                    3175619\ M\^108 - 3607879\ M\^110 - 930175\ M\^112 + 
                    2139808\ M\^114 - 37679\ M\^116 - 1004488\ M\^118 + 
                    268549\ M\^120 + 357673\ M\^122 - 220081\ M\^124 - 
                    57324\ M\^126 + 96529\ M\^128 - 20400\ M\^130 - 
                    20452\ M\^132 + 17657\ M\^134 - 6724\ M\^136 + 
                    1473\ M\^138 - 182\ M\^140 + 10\ M\^142)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 27]\), \(M\^64 + L\^24\ M\^64 + 
              L\ \((4\ M\^54 - 18\ M\^56 + 31\ M\^58 - 15\ M\^60 - 
                    36\ M\^62 + 78\ M\^64 - M\^66 - 39\ M\^68 + 26\ M\^70 - 
                    6\ M\^72)\) + 
              L\^23\ \((\(-6\)\ M\^56 + 26\ M\^58 - 39\ M\^60 - M\^62 + 
                    78\ M\^64 - 36\ M\^66 - 15\ M\^68 + 31\ M\^70 - 
                    18\ M\^72 + 4\ M\^74)\) + 
              L\^2\ \((6\ M\^44 - 52\ M\^46 + 193\ M\^48 - 332\ M\^50 + 
                    67\ M\^52 + 607\ M\^54 - 663\ M\^56 - 67\ M\^58 + 
                    187\ M\^60 - 179\ M\^62 + 1160\ M\^64 - 587\ M\^66 - 
                    801\ M\^68 + 1178\ M\^70 - 269\ M\^72 - 446\ M\^74 + 
                    381\ M\^76 - 122\ M\^78 + 15\ M\^80)\) + 
              L\^22\ \((15\ M\^48 - 122\ M\^50 + 381\ M\^52 - 446\ M\^54 - 
                    269\ M\^56 + 1178\ M\^58 - 801\ M\^60 - 587\ M\^62 + 
                    1160\ M\^64 - 179\ M\^66 + 187\ M\^68 - 67\ M\^70 - 
                    663\ M\^72 + 607\ M\^74 + 67\ M\^76 - 332\ M\^78 + 
                    193\ M\^80 - 52\ M\^82 + 6\ M\^84)\) + 
              L\^3\ \((4\ M\^34 - 50\ M\^36 + 278\ M\^38 - 846\ M\^40 + 
                    1359\ M\^42 - 553\ M\^44 - 2019\ M\^46 + 3963\ M\^48 - 
                    2302\ M\^50 - 2351\ M\^52 + 5606\ M\^54 - 4034\ M\^56 + 
                    1273\ M\^58 - 64\ M\^60 - 6972\ M\^62 + 13658\ M\^64 + 
                    145\ M\^66 - 12113\ M\^68 + 10233\ M\^70 - 1532\ M\^72 - 
                    5896\ M\^74 + 6938\ M\^76 - 2280\ M\^78 - 1939\ M\^80 + 
                    2367\ M\^82 - 1057\ M\^84 + 228\ M\^86 - 20\ M\^88)\) + 
              L\^21\ \((\(-20\)\ M\^40 + 228\ M\^42 - 1057\ M\^44 + 
                    2367\ M\^46 - 1939\ M\^48 - 2280\ M\^50 + 6938\ M\^52 - 
                    5896\ M\^54 - 1532\ M\^56 + 10233\ M\^58 - 12113\ M\^60 + 
                    145\ M\^62 + 13658\ M\^64 - 6972\ M\^66 - 64\ M\^68 + 
                    1273\ M\^70 - 4034\ M\^72 + 5606\ M\^74 - 2351\ M\^76 - 
                    2302\ M\^78 + 3963\ M\^80 - 2019\ M\^82 - 553\ M\^84 + 
                    1359\ M\^86 - 846\ M\^88 + 278\ M\^90 - 50\ M\^92 + 
                    4\ M\^94)\) + 
              L\^4\ \((M\^24 - 16\ M\^26 + 120\ M\^28 - 552\ M\^30 + 
                    1718\ M\^32 - 3684\ M\^34 + 4746\ M\^36 + 58\ M\^38 - 
                    15114\ M\^40 + 28458\ M\^42 - 3248\ M\^44 - 
                    69098\ M\^46 + 84651\ M\^48 + 66191\ M\^50 - 
                    201690\ M\^52 + 653\ M\^54 + 301856\ M\^56 - 
                    101545\ M\^58 - 349859\ M\^60 + 201055\ M\^62 + 
                    315743\ M\^64 - 263945\ M\^66 - 162285\ M\^68 + 
                    256893\ M\^70 - 11281\ M\^72 - 144727\ M\^74 + 
                    83473\ M\^76 + 16717\ M\^78 - 43547\ M\^80 + 
                    24932\ M\^82 - 3775\ M\^84 - 6986\ M\^86 + 7824\ M\^88 - 
                    4192\ M\^90 + 1278\ M\^92 - 212\ M\^94 + 15\ M\^96)\) + 
              L\^5\ \((3\ M\^18 - 55\ M\^20 + 451\ M\^22 - 2140\ M\^24 + 
                    6256\ M\^26 - 10484\ M\^28 + 4785\ M\^30 + 19350\ M\^32 - 
                    40563\ M\^34 + 7373\ M\^36 + 74443\ M\^38 - 
                    74523\ M\^40 - 71359\ M\^42 + 116563\ M\^44 + 
                    125631\ M\^46 - 198170\ M\^48 - 287523\ M\^50 + 
                    497468\ M\^52 + 365458\ M\^54 - 1016184\ M\^56 - 
                    88391\ M\^58 + 1404123\ M\^60 - 412490\ M\^62 - 
                    1409064\ M\^64 + 738182\ M\^66 + 1249518\ M\^68 - 
                    667466\ M\^70 - 1037967\ M\^72 + 553325\ M\^74 + 
                    738223\ M\^76 - 528412\ M\^78 - 322948\ M\^80 + 
                    412665\ M\^82 - 15398\ M\^84 - 166744\ M\^86 + 
                    91676\ M\^88 + 1014\ M\^90 - 27296\ M\^92 + 
                    20004\ M\^94 - 9342\ M\^96 + 3137\ M\^98 - 717\ M\^100 + 
                    98\ M\^102 - 6\ M\^104)\) + 
              L\^20\ \((15\ M\^32 - 212\ M\^34 + 1278\ M\^36 - 4192\ M\^38 + 
                    7824\ M\^40 - 6986\ M\^42 - 3775\ M\^44 + 24932\ M\^46 - 
                    43547\ M\^48 + 16717\ M\^50 + 83473\ M\^52 - 
                    144727\ M\^54 - 11281\ M\^56 + 256893\ M\^58 - 
                    162285\ M\^60 - 263945\ M\^62 + 315743\ M\^64 + 
                    201055\ M\^66 - 349859\ M\^68 - 101545\ M\^70 + 
                    301856\ M\^72 + 653\ M\^74 - 201690\ M\^76 + 
                    66191\ M\^78 + 84651\ M\^80 - 69098\ M\^82 - 
                    3248\ M\^84 + 28458\ M\^86 - 15114\ M\^88 + 58\ M\^90 + 
                    4746\ M\^92 - 3684\ M\^94 + 1718\ M\^96 - 552\ M\^98 + 
                    120\ M\^100 - 16\ M\^102 + M\^104)\) + 
              L\^19\ \((\(-6\)\ M\^24 + 98\ M\^26 - 717\ M\^28 + 
                    3137\ M\^30 - 9342\ M\^32 + 20004\ M\^34 - 27296\ M\^36 + 
                    1014\ M\^38 + 91676\ M\^40 - 166744\ M\^42 - 
                    15398\ M\^44 + 412665\ M\^46 - 322948\ M\^48 - 
                    528412\ M\^50 + 738223\ M\^52 + 553325\ M\^54 - 
                    1037967\ M\^56 - 667466\ M\^58 + 1249518\ M\^60 + 
                    738182\ M\^62 - 1409064\ M\^64 - 412490\ M\^66 + 
                    1404123\ M\^68 - 88391\ M\^70 - 1016184\ M\^72 + 
                    365458\ M\^74 + 497468\ M\^76 - 287523\ M\^78 - 
                    198170\ M\^80 + 125631\ M\^82 + 116563\ M\^84 - 
                    71359\ M\^86 - 74523\ M\^88 + 74443\ M\^90 + 
                    7373\ M\^92 - 40563\ M\^94 + 19350\ M\^96 + 4785\ M\^98 - 
                    10484\ M\^100 + 6256\ M\^102 - 2140\ M\^104 + 
                    451\ M\^106 - 55\ M\^108 + 3\ M\^110)\) + 
              L\^6\ \((3\ M\^12 - 62\ M\^14 + 574\ M\^16 - 3095\ M\^18 + 
                    10443\ M\^20 - 21203\ M\^22 + 17864\ M\^24 + 
                    26878\ M\^26 - 93387\ M\^28 + 69011\ M\^30 + 
                    130602\ M\^32 - 331088\ M\^34 + 166146\ M\^36 + 
                    452298\ M\^38 - 962745\ M\^40 + 263173\ M\^42 + 
                    1973861\ M\^44 - 2756254\ M\^46 - 1953545\ M\^48 + 
                    7169743\ M\^50 - 833280\ M\^52 - 11786831\ M\^54 + 
                    6961663\ M\^56 + 13273680\ M\^58 - 14011011\ M\^60 - 
                    9514584\ M\^62 + 17626876\ M\^64 + 2289587\ M\^66 - 
                    15114274\ M\^68 + 4065545\ M\^70 + 8761384\ M\^72 - 
                    5892152\ M\^74 - 2865598\ M\^76 + 3842194\ M\^78 + 
                    87908\ M\^80 - 1224122\ M\^82 + 143920\ M\^84 + 
                    73232\ M\^86 + 224798\ M\^88 - 66996\ M\^90 - 
                    180639\ M\^92 + 141590\ M\^94 + 216\ M\^96 - 
                    54318\ M\^98 + 38379\ M\^100 - 15278\ M\^102 + 
                    4280\ M\^104 - 930\ M\^106 + 157\ M\^108 - 18\ M\^110 + 
                    M\^112)\) + 
              L\^7\ \((M\^6 - 24\ M\^8 + 262\ M\^10 - 1712\ M\^12 + 
                    7368\ M\^14 - 21361\ M\^16 + 39168\ M\^18 - 
                    29177\ M\^20 - 58783\ M\^22 + 193876\ M\^24 - 
                    130555\ M\^26 - 381960\ M\^28 + 825368\ M\^30 + 
                    112467\ M\^32 - 1991745\ M\^34 + 951908\ M\^36 + 
                    3878761\ M\^38 - 3524451\ M\^40 - 6996810\ M\^42 + 
                    9424211\ M\^44 + 10958825\ M\^46 - 21294683\ M\^48 - 
                    12178126\ M\^50 + 38428162\ M\^52 + 5859762\ M\^54 - 
                    53844079\ M\^56 + 8078629\ M\^58 + 58097290\ M\^60 - 
                    21987977\ M\^62 - 48440941\ M\^64 + 27127890\ M\^66 + 
                    32248544\ M\^68 - 21591421\ M\^70 - 18029744\ M\^72 + 
                    13067962\ M\^74 + 9685195\ M\^76 - 7934207\ M\^78 - 
                    5177971\ M\^80 + 6035211\ M\^82 + 1681336\ M\^84 - 
                    4307715\ M\^86 + 716699\ M\^88 + 1946650\ M\^90 - 
                    1245839\ M\^92 - 166848\ M\^94 + 534994\ M\^96 - 
                    257923\ M\^98 + 2283\ M\^100 + 80846\ M\^102 - 
                    70213\ M\^104 + 37756\ M\^106 - 14172\ M\^108 + 
                    3690\ M\^110 - 635\ M\^112 + 65\ M\^114 - 3\ M\^116)\) + 
              L\^18\ \((M\^16 - 18\ M\^18 + 157\ M\^20 - 930\ M\^22 + 
                    4280\ M\^24 - 15278\ M\^26 + 38379\ M\^28 - 
                    54318\ M\^30 + 216\ M\^32 + 141590\ M\^34 - 
                    180639\ M\^36 - 66996\ M\^38 + 224798\ M\^40 + 
                    73232\ M\^42 + 143920\ M\^44 - 1224122\ M\^46 + 
                    87908\ M\^48 + 3842194\ M\^50 - 2865598\ M\^52 - 
                    5892152\ M\^54 + 8761384\ M\^56 + 4065545\ M\^58 - 
                    15114274\ M\^60 + 2289587\ M\^62 + 17626876\ M\^64 - 
                    9514584\ M\^66 - 14011011\ M\^68 + 13273680\ M\^70 + 
                    6961663\ M\^72 - 11786831\ M\^74 - 833280\ M\^76 + 
                    7169743\ M\^78 - 1953545\ M\^80 - 2756254\ M\^82 + 
                    1973861\ M\^84 + 263173\ M\^86 - 962745\ M\^88 + 
                    452298\ M\^90 + 166146\ M\^92 - 331088\ M\^94 + 
                    130602\ M\^96 + 69011\ M\^98 - 93387\ M\^100 + 
                    26878\ M\^102 + 17864\ M\^104 - 21203\ M\^106 + 
                    10443\ M\^108 - 3095\ M\^110 + 574\ M\^112 - 62\ M\^114 + 
                    3\ M\^116)\) + 
              L\^8\ \((\(-M\^2\) + 24\ M\^4 - 265\ M\^6 + 1776\ M\^8 - 
                    7971\ M\^10 + 24687\ M\^12 - 51264\ M\^14 + 
                    60262\ M\^16 + 3437\ M\^18 - 157944\ M\^20 + 
                    300692\ M\^22 - 202866\ M\^24 - 475839\ M\^26 + 
                    1764380\ M\^28 - 1738951\ M\^30 - 2953142\ M\^32 + 
                    8560794\ M\^34 - 932782\ M\^36 - 18652631\ M\^38 + 
                    15023410\ M\^40 + 24528887\ M\^42 - 37716763\ M\^44 - 
                    17880337\ M\^46 + 59252330\ M\^48 - 2902746\ M\^50 - 
                    68559882\ M\^52 + 30383141\ M\^54 + 63204865\ M\^56 - 
                    55783345\ M\^58 - 49226998\ M\^60 + 75458787\ M\^62 + 
                    32512826\ M\^64 - 88654982\ M\^66 - 12458713\ M\^68 + 
                    93125220\ M\^70 - 8846578\ M\^72 - 81206998\ M\^74 + 
                    26611326\ M\^76 + 53741933\ M\^78 - 33317280\ M\^80 - 
                    23136974\ M\^82 + 26715732\ M\^84 + 2503407\ M\^86 - 
                    13536250\ M\^88 + 4230636\ M\^90 + 3312636\ M\^92 - 
                    2825696\ M\^94 + 519945\ M\^96 + 327657\ M\^98 - 
                    404767\ M\^100 + 308392\ M\^102 - 94232\ M\^104 - 
                    81778\ M\^106 + 114316\ M\^108 - 67424\ M\^110 + 
                    24290\ M\^112 - 5705\ M\^114 + 860\ M\^116 - 76\ M\^118 + 
                    3\ M\^120)\) + 
              L\^17\ \((\(-3\)\ M\^12 + 65\ M\^14 - 635\ M\^16 + 
                    3690\ M\^18 - 14172\ M\^20 + 37756\ M\^22 - 
                    70213\ M\^24 + 80846\ M\^26 + 2283\ M\^28 - 
                    257923\ M\^30 + 534994\ M\^32 - 166848\ M\^34 - 
                    1245839\ M\^36 + 1946650\ M\^38 + 716699\ M\^40 - 
                    4307715\ M\^42 + 1681336\ M\^44 + 6035211\ M\^46 - 
                    5177971\ M\^48 - 7934207\ M\^50 + 9685195\ M\^52 + 
                    13067962\ M\^54 - 18029744\ M\^56 - 21591421\ M\^58 + 
                    32248544\ M\^60 + 27127890\ M\^62 - 48440941\ M\^64 - 
                    21987977\ M\^66 + 58097290\ M\^68 + 8078629\ M\^70 - 
                    53844079\ M\^72 + 5859762\ M\^74 + 38428162\ M\^76 - 
                    12178126\ M\^78 - 21294683\ M\^80 + 10958825\ M\^82 + 
                    9424211\ M\^84 - 6996810\ M\^86 - 3524451\ M\^88 + 
                    3878761\ M\^90 + 951908\ M\^92 - 1991745\ M\^94 + 
                    112467\ M\^96 + 825368\ M\^98 - 381960\ M\^100 - 
                    130555\ M\^102 + 193876\ M\^104 - 58783\ M\^106 - 
                    29177\ M\^108 + 39168\ M\^110 - 21361\ M\^112 + 
                    7368\ M\^114 - 1712\ M\^116 + 262\ M\^118 - 24\ M\^120 + 
                    M\^122)\) + 
              L\^9\ \((\(-1\) + 27\ M\^2 - 345\ M\^4 + 2699\ M\^6 - 
                    14028\ M\^8 + 48984\ M\^10 - 107443\ M\^12 + 
                    102196\ M\^14 + 149236\ M\^16 - 561823\ M\^18 + 
                    293540\ M\^20 + 1233070\ M\^22 - 1768218\ M\^24 - 
                    1599475\ M\^26 + 3717152\ M\^28 + 3668747\ M\^30 - 
                    7535006\ M\^32 - 11679410\ M\^34 + 19694072\ M\^36 + 
                    27383292\ M\^38 - 50834436\ M\^40 - 40067756\ M\^42 + 
                    101173950\ M\^44 + 27696102\ M\^46 - 146013994\ M\^48 + 
                    27012766\ M\^50 + 139082707\ M\^52 - 113248586\ M\^54 - 
                    45828773\ M\^56 + 188647721\ M\^58 - 119994038\ M\^60 - 
                    204589097\ M\^62 + 291312909\ M\^64 + 144426498\ M\^66 - 
                    387427774\ M\^68 - 35589679\ M\^70 + 374680812\ M\^72 - 
                    61332008\ M\^74 - 276616441\ M\^76 + 103573086\ M\^78 + 
                    153249067\ M\^80 - 90646548\ M\^82 - 59151946\ M\^84 + 
                    52237187\ M\^86 + 12859055\ M\^88 - 18810509\ M\^90 - 
                    1106023\ M\^92 + 3109536\ M\^94 + 1566996\ M\^96 - 
                    106207\ M\^98 - 1997603\ M\^100 + 850898\ M\^102 + 
                    691166\ M\^104 - 720966\ M\^106 + 111054\ M\^108 + 
                    191806\ M\^110 - 165587\ M\^112 + 71311\ M\^114 - 
                    19625\ M\^116 + 3611\ M\^118 - 434\ M\^120 + 31\ M\^122 - 
                    M\^124)\) + 
              L\^10\ \((4 - 106\ M\^2 + 1296\ M\^4 - 9519\ M\^6 + 
                    45528\ M\^8 - 141026\ M\^10 + 244119\ M\^12 - 
                    19492\ M\^14 - 947647\ M\^16 + 1700081\ M\^18 + 
                    1007272\ M\^20 - 7124031\ M\^22 + 4532118\ M\^24 + 
                    16291198\ M\^26 - 23765338\ M\^28 - 25565283\ M\^30 + 
                    66056354\ M\^32 + 27108789\ M\^34 - 133885857\ M\^36 - 
                    18725483\ M\^38 + 218378762\ M\^40 + 16434461\ M\^42 - 
                    300785632\ M\^44 - 60389204\ M\^46 + 367056314\ M\^48 + 
                    190323132\ M\^50 - 416155798\ M\^52 - 403434133\ M\^54 + 
                    452907593\ M\^56 + 630199673\ M\^58 - 469505182\ M\^60 - 
                    767864455\ M\^62 + 443563297\ M\^64 + 749860796\ M\^66 - 
                    357290905\ M\^68 - 588012035\ M\^70 + 231097000\ M\^72 + 
                    371405173\ M\^74 - 101376263\ M\^76 - 190390849\ M\^78 + 
                    2977407\ M\^80 + 84261943\ M\^82 + 46979795\ M\^84 - 
                    42313197\ M\^86 - 53171919\ M\^88 + 31796803\ M\^90 + 
                    35227534\ M\^92 - 26778964\ M\^94 - 13459809\ M\^96 + 
                    17633969\ M\^98 + 452304\ M\^100 - 7707437\ M\^102 + 
                    2990982\ M\^104 + 1390638\ M\^106 - 1600390\ M\^108 + 
                    367399\ M\^110 + 262736\ M\^112 - 264058\ M\^114 + 
                    119033\ M\^116 - 33971\ M\^118 + 6488\ M\^120 - 
                    810\ M\^122 + 60\ M\^124 - 2\ M\^126)\) + 
              L\^16\ \((3\ M\^8 - 76\ M\^10 + 860\ M\^12 - 5705\ M\^14 + 
                    24290\ M\^16 - 67424\ M\^18 + 114316\ M\^20 - 
                    81778\ M\^22 - 94232\ M\^24 + 308392\ M\^26 - 
                    404767\ M\^28 + 327657\ M\^30 + 519945\ M\^32 - 
                    2825696\ M\^34 + 3312636\ M\^36 + 4230636\ M\^38 - 
                    13536250\ M\^40 + 2503407\ M\^42 + 26715732\ M\^44 - 
                    23136974\ M\^46 - 33317280\ M\^48 + 53741933\ M\^50 + 
                    26611326\ M\^52 - 81206998\ M\^54 - 8846578\ M\^56 + 
                    93125220\ M\^58 - 12458713\ M\^60 - 88654982\ M\^62 + 
                    32512826\ M\^64 + 75458787\ M\^66 - 49226998\ M\^68 - 
                    55783345\ M\^70 + 63204865\ M\^72 + 30383141\ M\^74 - 
                    68559882\ M\^76 - 2902746\ M\^78 + 59252330\ M\^80 - 
                    17880337\ M\^82 - 37716763\ M\^84 + 24528887\ M\^86 + 
                    15023410\ M\^88 - 18652631\ M\^90 - 932782\ M\^92 + 
                    8560794\ M\^94 - 2953142\ M\^96 - 1738951\ M\^98 + 
                    1764380\ M\^100 - 475839\ M\^102 - 202866\ M\^104 + 
                    300692\ M\^106 - 157944\ M\^108 + 3437\ M\^110 + 
                    60262\ M\^112 - 51264\ M\^114 + 24687\ M\^116 - 
                    7971\ M\^118 + 1776\ M\^120 - 265\ M\^122 + 24\ M\^124 - 
                    M\^126)\) + 
              L\^13\ \((\(-1\) + 35\ M\^2 - 547\ M\^4 + 5045\ M\^6 - 
                    30314\ M\^8 + 121756\ M\^10 - 311581\ M\^12 + 
                    382536\ M\^14 + 382240\ M\^16 - 2264506\ M\^18 + 
                    2363041\ M\^20 + 4518502\ M\^22 - 13270776\ M\^24 + 
                    844902\ M\^26 + 34835144\ M\^28 - 27409789\ M\^30 - 
                    63875274\ M\^32 + 89534526\ M\^34 + 92684079\ M\^36 - 
                    191313808\ M\^38 - 122683594\ M\^40 + 322180098\ M\^42 + 
                    173662649\ M\^44 - 451488692\ M\^46 - 284215258\ M\^48 + 
                    529368869\ M\^50 + 492954499\ M\^52 - 496503508\ M\^54 - 
                    813514837\ M\^56 + 318190626\ M\^58 + 1211208754\ M\^60 - 
                    22457611\ M\^62 - 1604569902\ M\^64 - 296112401\ M\^66 + 
                    1891888818\ M\^68 + 531213838\ M\^70 - 
                    1976144624\ M\^72 - 602561157\ M\^74 + 
                    1828279014\ M\^76 + 507153194\ M\^78 - 
                    1498325044\ M\^80 - 313672048\ M\^82 + 
                    1083458291\ M\^84 + 116577128\ M\^86 - 683673285\ M\^88 + 
                    16609599\ M\^90 + 367060348\ M\^92 - 68442866\ M\^94 - 
                    159829715\ M\^96 + 63251578\ M\^98 + 50311170\ M\^100 - 
                    36716935\ M\^102 - 7360274\ M\^104 + 14258081\ M\^106 - 
                    2647654\ M\^108 - 2942062\ M\^110 + 1822204\ M\^112 - 
                    75227\ M\^114 - 380437\ M\^116 + 230682\ M\^118 - 
                    75040\ M\^120 + 15593\ M\^122 - 2088\ M\^124 + 
                    166\ M\^126 - 6\ M\^128)\) + 
              L\^15\ \((\(-M\^4\) + 31\ M\^6 - 434\ M\^8 + 3611\ M\^10 - 
                    19625\ M\^12 + 71311\ M\^14 - 165587\ M\^16 + 
                    191806\ M\^18 + 111054\ M\^20 - 720966\ M\^22 + 
                    691166\ M\^24 + 850898\ M\^26 - 1997603\ M\^28 - 
                    106207\ M\^30 + 1566996\ M\^32 + 3109536\ M\^34 - 
                    1106023\ M\^36 - 18810509\ M\^38 + 12859055\ M\^40 + 
                    52237187\ M\^42 - 59151946\ M\^44 - 90646548\ M\^46 + 
                    153249067\ M\^48 + 103573086\ M\^50 - 276616441\ M\^52 - 
                    61332008\ M\^54 + 374680812\ M\^56 - 35589679\ M\^58 - 
                    387427774\ M\^60 + 144426498\ M\^62 + 291312909\ M\^64 - 
                    204589097\ M\^66 - 119994038\ M\^68 + 188647721\ M\^70 - 
                    45828773\ M\^72 - 113248586\ M\^74 + 139082707\ M\^76 + 
                    27012766\ M\^78 - 146013994\ M\^80 + 27696102\ M\^82 + 
                    101173950\ M\^84 - 40067756\ M\^86 - 50834436\ M\^88 + 
                    27383292\ M\^90 + 19694072\ M\^92 - 11679410\ M\^94 - 
                    7535006\ M\^96 + 3668747\ M\^98 + 3717152\ M\^100 - 
                    1599475\ M\^102 - 1768218\ M\^104 + 1233070\ M\^106 + 
                    293540\ M\^108 - 561823\ M\^110 + 149236\ M\^112 + 
                    102196\ M\^114 - 107443\ M\^116 + 48984\ M\^118 - 
                    14028\ M\^120 + 2699\ M\^122 - 345\ M\^124 + 27\ M\^126 - 
                    M\^128)\) + 
              L\^11\ \((\(-6\) + 166\ M\^2 - 2088\ M\^4 + 15593\ M\^6 - 
                    75040\ M\^8 + 230682\ M\^10 - 380437\ M\^12 - 
                    75227\ M\^14 + 1822204\ M\^16 - 2942062\ M\^18 - 
                    2647654\ M\^20 + 14258081\ M\^22 - 7360274\ M\^24 - 
                    36716935\ M\^26 + 50311170\ M\^28 + 63251578\ M\^30 - 
                    159829715\ M\^32 - 68442866\ M\^34 + 367060348\ M\^36 + 
                    16609599\ M\^38 - 683673285\ M\^40 + 116577128\ M\^42 + 
                    1083458291\ M\^44 - 313672048\ M\^46 - 
                    1498325044\ M\^48 + 507153194\ M\^50 + 
                    1828279014\ M\^52 - 602561157\ M\^54 - 
                    1976144624\ M\^56 + 531213838\ M\^58 + 
                    1891888818\ M\^60 - 296112401\ M\^62 - 
                    1604569902\ M\^64 - 22457611\ M\^66 + 1211208754\ M\^68 + 
                    318190626\ M\^70 - 813514837\ M\^72 - 496503508\ M\^74 + 
                    492954499\ M\^76 + 529368869\ M\^78 - 284215258\ M\^80 - 
                    451488692\ M\^82 + 173662649\ M\^84 + 322180098\ M\^86 - 
                    122683594\ M\^88 - 191313808\ M\^90 + 92684079\ M\^92 + 
                    89534526\ M\^94 - 63875274\ M\^96 - 27409789\ M\^98 + 
                    34835144\ M\^100 + 844902\ M\^102 - 13270776\ M\^104 + 
                    4518502\ M\^106 + 2363041\ M\^108 - 2264506\ M\^110 + 
                    382240\ M\^112 + 382536\ M\^114 - 311581\ M\^116 + 
                    121756\ M\^118 - 30314\ M\^120 + 5045\ M\^122 - 
                    547\ M\^124 + 35\ M\^126 - M\^128)\) + 
              L\^12\ \((4 - 119\ M\^2 + 1600\ M\^4 - 12775\ M\^6 + 
                    66126\ M\^8 - 222610\ M\^10 + 431587\ M\^12 - 
                    141254\ M\^14 - 1572947\ M\^16 + 3484857\ M\^18 + 
                    670418\ M\^20 - 13558981\ M\^22 + 13436243\ M\^24 + 
                    28265688\ M\^26 - 60320596\ M\^28 - 34334666\ M\^30 + 
                    165970983\ M\^32 + 5871231\ M\^34 - 356576348\ M\^36 + 
                    80894971\ M\^38 + 661492367\ M\^40 - 230241934\ M\^42 - 
                    1111496325\ M\^44 + 410888430\ M\^46 + 
                    1724270343\ M\^48 - 558416255\ M\^50 - 
                    2465408960\ M\^52 + 593745631\ M\^54 + 
                    3218024858\ M\^56 - 463170826\ M\^58 - 
                    3795704280\ M\^60 + 176948612\ M\^62 + 
                    4016134010\ M\^64 + 176948612\ M\^66 - 
                    3795704280\ M\^68 - 463170826\ M\^70 + 
                    3218024858\ M\^72 + 593745631\ M\^74 - 
                    2465408960\ M\^76 - 558416255\ M\^78 + 
                    1724270343\ M\^80 + 410888430\ M\^82 - 
                    1111496325\ M\^84 - 230241934\ M\^86 + 661492367\ M\^88 + 
                    80894971\ M\^90 - 356576348\ M\^92 + 5871231\ M\^94 + 
                    165970983\ M\^96 - 34334666\ M\^98 - 60320596\ M\^100 + 
                    28265688\ M\^102 + 13436243\ M\^104 - 13558981\ M\^106 + 
                    670418\ M\^108 + 3484857\ M\^110 - 1572947\ M\^112 - 
                    141254\ M\^114 + 431587\ M\^116 - 222610\ M\^118 + 
                    66126\ M\^120 - 12775\ M\^122 + 1600\ M\^124 - 
                    119\ M\^126 + 4\ M\^128)\) + 
              L\^14\ \((\(-2\)\ M\^2 + 60\ M\^4 - 810\ M\^6 + 6488\ M\^8 - 
                    33971\ M\^10 + 119033\ M\^12 - 264058\ M\^14 + 
                    262736\ M\^16 + 367399\ M\^18 - 1600390\ M\^20 + 
                    1390638\ M\^22 + 2990982\ M\^24 - 7707437\ M\^26 + 
                    452304\ M\^28 + 17633969\ M\^30 - 13459809\ M\^32 - 
                    26778964\ M\^34 + 35227534\ M\^36 + 31796803\ M\^38 - 
                    53171919\ M\^40 - 42313197\ M\^42 + 46979795\ M\^44 + 
                    84261943\ M\^46 + 2977407\ M\^48 - 190390849\ M\^50 - 
                    101376263\ M\^52 + 371405173\ M\^54 + 231097000\ M\^56 - 
                    588012035\ M\^58 - 357290905\ M\^60 + 749860796\ M\^62 + 
                    443563297\ M\^64 - 767864455\ M\^66 - 469505182\ M\^68 + 
                    630199673\ M\^70 + 452907593\ M\^72 - 403434133\ M\^74 - 
                    416155798\ M\^76 + 190323132\ M\^78 + 367056314\ M\^80 - 
                    60389204\ M\^82 - 300785632\ M\^84 + 16434461\ M\^86 + 
                    218378762\ M\^88 - 18725483\ M\^90 - 133885857\ M\^92 + 
                    27108789\ M\^94 + 66056354\ M\^96 - 25565283\ M\^98 - 
                    23765338\ M\^100 + 16291198\ M\^102 + 4532118\ M\^104 - 
                    7124031\ M\^106 + 1007272\ M\^108 + 1700081\ M\^110 - 
                    947647\ M\^112 - 19492\ M\^114 + 244119\ M\^116 - 
                    141026\ M\^118 + 45528\ M\^120 - 9519\ M\^122 + 
                    1296\ M\^124 - 106\ M\^126 + 4\ M\^128)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 28]\), \(L\^18\ M\^28 + M\^88 + 
              L\^17\ \((3\ M\^22 - 15\ M\^24 + 27\ M\^26 - 5\ M\^28 - 
                    56\ M\^30 + 81\ M\^32 + 18\ M\^34 - 55\ M\^36 + 
                    24\ M\^38 - 4\ M\^40)\) + 
              L\^16\ \((3\ M\^16 - 30\ M\^18 + 126\ M\^20 - 249\ M\^22 + 
                    116\ M\^24 + 387\ M\^26 - 699\ M\^28 + 538\ M\^30 - 
                    122\ M\^32 - 1006\ M\^34 + 1802\ M\^36 + 69\ M\^38 - 
                    1620\ M\^40 + 911\ M\^42 + 237\ M\^44 - 524\ M\^46 + 
                    272\ M\^48 - 64\ M\^50 + 6\ M\^52)\) + 
              L\^15\ \((M\^10 - 15\ M\^12 + 101\ M\^14 - 398\ M\^16 + 
                    980\ M\^18 - 1311\ M\^20 - 264\ M\^22 + 4559\ M\^24 - 
                    6457\ M\^26 - 1598\ M\^28 + 11342\ M\^30 - 4635\ M\^32 - 
                    6556\ M\^34 + 222\ M\^36 + 437\ M\^38 + 14239\ M\^40 - 
                    4219\ M\^42 - 16329\ M\^44 + 11479\ M\^46 + 4709\ M\^48 - 
                    8825\ M\^50 + 3199\ M\^52 + 1401\ M\^54 - 2064\ M\^56 + 
                    1100\ M\^58 - 334\ M\^60 + 56\ M\^62 - 4\ M\^64)\) + 
              L\^14\ \((\(-M\^6\) + 18\ M\^8 - 145\ M\^10 + 675\ M\^12 - 
                    1962\ M\^14 + 3485\ M\^16 - 2564\ M\^18 - 5263\ M\^20 + 
                    20284\ M\^22 - 22957\ M\^24 - 23629\ M\^26 + 
                    91325\ M\^28 - 33256\ M\^30 - 163844\ M\^32 + 
                    153425\ M\^34 + 192743\ M\^36 - 269071\ M\^38 - 
                    183041\ M\^40 + 318704\ M\^42 + 150769\ M\^44 - 
                    289101\ M\^46 - 82533\ M\^48 + 215362\ M\^50 + 
                    557\ M\^52 - 119292\ M\^54 + 41866\ M\^56 + 
                    34354\ M\^58 - 30971\ M\^60 + 3857\ M\^62 + 6886\ M\^64 - 
                    5241\ M\^66 + 2104\ M\^68 - 588\ M\^70 + 120\ M\^72 - 
                    16\ M\^74 + M\^76)\) + 
              L\^13\ \((\(-3\)\ M\^4 + 55\ M\^6 - 447\ M\^8 + 2048\ M\^10 - 
                    5431\ M\^12 + 6675\ M\^14 + 4430\ M\^16 - 27063\ M\^18 + 
                    26129\ M\^20 + 28285\ M\^22 - 77687\ M\^24 + 
                    34016\ M\^26 + 61406\ M\^28 - 124076\ M\^30 + 
                    135878\ M\^32 + 80960\ M\^34 - 498612\ M\^36 + 
                    216962\ M\^38 + 881293\ M\^40 - 696332\ M\^42 - 
                    1106504\ M\^44 + 1089007\ M\^46 + 1067444\ M\^48 - 
                    1173471\ M\^50 - 717537\ M\^52 + 967834\ M\^54 + 
                    249233\ M\^56 - 604532\ M\^58 + 66762\ M\^60 + 
                    243870\ M\^62 - 132909\ M\^64 - 26060\ M\^66 + 
                    61488\ M\^68 - 26939\ M\^70 - 3944\ M\^72 + 
                    11284\ M\^74 - 6857\ M\^76 + 2347\ M\^78 - 489\ M\^80 + 
                    58\ M\^82 - 3\ M\^84)\) + 
              L\^12\ \((\(-3\)\ M\^2 + 60\ M\^4 - 528\ M\^6 + 2600\ M\^8 - 
                    7319\ M\^10 + 9097\ M\^12 + 8871\ M\^14 - 44776\ M\^16 + 
                    26366\ M\^18 + 105915\ M\^20 - 160064\ M\^22 - 
                    144085\ M\^24 + 419883\ M\^26 + 125467\ M\^28 - 
                    805397\ M\^30 - 106082\ M\^32 + 1378000\ M\^34 + 
                    188281\ M\^36 - 2268220\ M\^38 - 425605\ M\^40 + 
                    3539213\ M\^42 + 730732\ M\^44 - 4881717\ M\^46 - 
                    988306\ M\^48 + 5664086\ M\^50 + 989350\ M\^52 - 
                    5332583\ M\^54 - 503548\ M\^56 + 4090968\ M\^58 - 
                    244293\ M\^60 - 2516067\ M\^62 + 729961\ M\^64 + 
                    1138031\ M\^66 - 718793\ M\^68 - 273889\ M\^70 + 
                    409461\ M\^72 - 57931\ M\^74 - 119892\ M\^76 + 
                    70930\ M\^78 + 2066\ M\^80 - 20318\ M\^82 + 
                    11577\ M\^84 - 3510\ M\^86 + 638\ M\^88 - 66\ M\^90 + 
                    3\ M\^92)\) + 
              L\ \((\(-4\)\ M\^76 + 24\ M\^78 - 55\ M\^80 + 18\ M\^82 + 
                    81\ M\^84 - 56\ M\^86 - 5\ M\^88 + 27\ M\^90 - 
                    15\ M\^92 + 3\ M\^94)\) + 
              L\^11\ \((\(-1\) + 22\ M\^2 - 213\ M\^4 + 1174\ M\^6 - 
                    3921\ M\^8 + 7416\ M\^10 - 4215\ M\^12 - 13589\ M\^14 + 
                    29445\ M\^16 + 900\ M\^18 - 77940\ M\^20 + 93886\ M\^22 + 
                    61349\ M\^24 - 267551\ M\^26 + 134210\ M\^28 + 
                    436030\ M\^30 - 526873\ M\^32 - 552477\ M\^34 + 
                    919828\ M\^36 + 879132\ M\^38 - 984385\ M\^40 - 
                    2088086\ M\^42 + 542447\ M\^44 + 4729988\ M\^46 + 
                    431142\ M\^48 - 8534659\ M\^50 - 1745322\ M\^52 + 
                    11950195\ M\^54 + 2859252\ M\^56 - 13091886\ M\^58 - 
                    2771152\ M\^60 + 11532800\ M\^62 + 1433716\ M\^64 - 
                    8321163\ M\^66 + 141299\ M\^68 + 4844364\ M\^70 - 
                    984276\ M\^72 - 2139941\ M\^74 + 956167\ M\^76 + 
                    605008\ M\^78 - 532733\ M\^80 - 24236\ M\^82 + 
                    165725\ M\^84 - 55729\ M\^86 - 18167\ M\^88 + 
                    22043\ M\^90 - 9055\ M\^92 + 2141\ M\^94 - 310\ M\^96 + 
                    26\ M\^98 - M\^100)\) + 
              L\^2\ \((6\ M\^64 - 64\ M\^66 + 272\ M\^68 - 524\ M\^70 + 
                    237\ M\^72 + 911\ M\^74 - 1620\ M\^76 + 69\ M\^78 + 
                    1802\ M\^80 - 1006\ M\^82 - 122\ M\^84 + 538\ M\^86 - 
                    699\ M\^88 + 387\ M\^90 + 116\ M\^92 - 249\ M\^94 + 
                    126\ M\^96 - 30\ M\^98 + 3\ M\^100)\) + 
              L\^10\ \((\(-2\)\ M\^2 + 44\ M\^4 - 426\ M\^6 + 2319\ M\^8 - 
                    7321\ M\^10 + 10929\ M\^12 + 6469\ M\^14 - 51515\ M\^16 + 
                    36552\ M\^18 + 157830\ M\^20 - 288176\ M\^22 - 
                    236006\ M\^24 + 974873\ M\^26 - 10148\ M\^28 - 
                    2264094\ M\^30 + 1130056\ M\^32 + 3890109\ M\^34 - 
                    3519099\ M\^36 - 5098129\ M\^38 + 6929279\ M\^40 + 
                    5120627\ M\^42 - 10416948\ M\^44 - 3888652\ M\^46 + 
                    12722302\ M\^48 + 2455792\ M\^50 - 12800364\ M\^52 - 
                    2219650\ M\^54 + 9984076\ M\^56 + 3507477\ M\^58 - 
                    4886974\ M\^60 - 4891898\ M\^62 - 30832\ M\^64 + 
                    5311613\ M\^66 + 2574551\ M\^68 - 4719159\ M\^70 - 
                    2653215\ M\^72 + 3560049\ M\^74 + 1550492\ M\^76 - 
                    2263028\ M\^78 - 496569\ M\^80 + 1161518\ M\^82 - 
                    25627\ M\^84 - 447926\ M\^86 + 133931\ M\^88 + 
                    98095\ M\^90 - 68980\ M\^92 - 1385\ M\^94 + 
                    18480\ M\^96 - 9970\ M\^98 + 2820\ M\^100 - 476\ M\^102 + 
                    46\ M\^104 - 2\ M\^106)\) + 
              L\^3\ \((\(-4\)\ M\^52 + 56\ M\^54 - 334\ M\^56 + 1100\ M\^58 - 
                    2064\ M\^60 + 1401\ M\^62 + 3199\ M\^64 - 8825\ M\^66 + 
                    4709\ M\^68 + 11479\ M\^70 - 16329\ M\^72 - 4219\ M\^74 + 
                    14239\ M\^76 + 437\ M\^78 + 222\ M\^80 - 6556\ M\^82 - 
                    4635\ M\^84 + 11342\ M\^86 - 1598\ M\^88 - 6457\ M\^90 + 
                    4559\ M\^92 - 264\ M\^94 - 1311\ M\^96 + 980\ M\^98 - 
                    398\ M\^100 + 101\ M\^102 - 15\ M\^104 + M\^106)\) + 
              L\^4\ \((M\^40 - 16\ M\^42 + 120\ M\^44 - 588\ M\^46 + 
                    2104\ M\^48 - 5241\ M\^50 + 6886\ M\^52 + 3857\ M\^54 - 
                    30971\ M\^56 + 34354\ M\^58 + 41866\ M\^60 - 
                    119292\ M\^62 + 557\ M\^64 + 215362\ M\^66 - 
                    82533\ M\^68 - 289101\ M\^70 + 150769\ M\^72 + 
                    318704\ M\^74 - 183041\ M\^76 - 269071\ M\^78 + 
                    192743\ M\^80 + 153425\ M\^82 - 163844\ M\^84 - 
                    33256\ M\^86 + 91325\ M\^88 - 23629\ M\^90 - 
                    22957\ M\^92 + 20284\ M\^94 - 5263\ M\^96 - 2564\ M\^98 + 
                    3485\ M\^100 - 1962\ M\^102 + 675\ M\^104 - 145\ M\^106 + 
                    18\ M\^108 - M\^110)\) + 
              L\^5\ \((\(-3\)\ M\^32 + 58\ M\^34 - 489\ M\^36 + 2347\ M\^38 - 
                    6857\ M\^40 + 11284\ M\^42 - 3944\ M\^44 - 26939\ M\^46 + 
                    61488\ M\^48 - 26060\ M\^50 - 132909\ M\^52 + 
                    243870\ M\^54 + 66762\ M\^56 - 604532\ M\^58 + 
                    249233\ M\^60 + 967834\ M\^62 - 717537\ M\^64 - 
                    1173471\ M\^66 + 1067444\ M\^68 + 1089007\ M\^70 - 
                    1106504\ M\^72 - 696332\ M\^74 + 881293\ M\^76 + 
                    216962\ M\^78 - 498612\ M\^80 + 80960\ M\^82 + 
                    135878\ M\^84 - 124076\ M\^86 + 61406\ M\^88 + 
                    34016\ M\^90 - 77687\ M\^92 + 28285\ M\^94 + 
                    26129\ M\^96 - 27063\ M\^98 + 4430\ M\^100 + 
                    6675\ M\^102 - 5431\ M\^104 + 2048\ M\^106 - 
                    447\ M\^108 + 55\ M\^110 - 3\ M\^112)\) + 
              L\^9\ \((\(-M\^4\) + 22\ M\^6 - 217\ M\^8 + 1236\ M\^10 - 
                    4327\ M\^12 + 8822\ M\^14 - 6734\ M\^16 - 12237\ M\^18 + 
                    32167\ M\^20 - 4293\ M\^22 - 73379\ M\^24 + 
                    101271\ M\^26 + 902\ M\^28 - 204075\ M\^30 + 
                    338116\ M\^32 + 108173\ M\^34 - 1083237\ M\^36 + 
                    472234\ M\^38 + 2213143\ M\^40 - 1547543\ M\^42 - 
                    3524837\ M\^44 + 2632043\ M\^46 + 4502709\ M\^48 - 
                    3048902\ M\^50 - 4149222\ M\^52 + 2622072\ M\^54 + 
                    1767787\ M\^56 - 2234766\ M\^58 + 1767787\ M\^60 + 
                    2622072\ M\^62 - 4149222\ M\^64 - 3048902\ M\^66 + 
                    4502709\ M\^68 + 2632043\ M\^70 - 3524837\ M\^72 - 
                    1547543\ M\^74 + 2213143\ M\^76 + 472234\ M\^78 - 
                    1083237\ M\^80 + 108173\ M\^82 + 338116\ M\^84 - 
                    204075\ M\^86 + 902\ M\^88 + 101271\ M\^90 - 
                    73379\ M\^92 - 4293\ M\^94 + 32167\ M\^96 - 
                    12237\ M\^98 - 6734\ M\^100 + 8822\ M\^102 - 
                    4327\ M\^104 + 1236\ M\^106 - 217\ M\^108 + 22\ M\^110 - 
                    M\^112)\) + 
              L\^6\ \((3\ M\^24 - 66\ M\^26 + 638\ M\^28 - 3510\ M\^30 + 
                    11577\ M\^32 - 20318\ M\^34 + 2066\ M\^36 + 
                    70930\ M\^38 - 119892\ M\^40 - 57931\ M\^42 + 
                    409461\ M\^44 - 273889\ M\^46 - 718793\ M\^48 + 
                    1138031\ M\^50 + 729961\ M\^52 - 2516067\ M\^54 - 
                    244293\ M\^56 + 4090968\ M\^58 - 503548\ M\^60 - 
                    5332583\ M\^62 + 989350\ M\^64 + 5664086\ M\^66 - 
                    988306\ M\^68 - 4881717\ M\^70 + 730732\ M\^72 + 
                    3539213\ M\^74 - 425605\ M\^76 - 2268220\ M\^78 + 
                    188281\ M\^80 + 1378000\ M\^82 - 106082\ M\^84 - 
                    805397\ M\^86 + 125467\ M\^88 + 419883\ M\^90 - 
                    144085\ M\^92 - 160064\ M\^94 + 105915\ M\^96 + 
                    26366\ M\^98 - 44776\ M\^100 + 8871\ M\^102 + 
                    9097\ M\^104 - 7319\ M\^106 + 2600\ M\^108 - 
                    528\ M\^110 + 60\ M\^112 - 3\ M\^114)\) + 
              L\^8\ \((\(-2\)\ M\^10 + 46\ M\^12 - 476\ M\^14 + 2820\ M\^16 - 
                    9970\ M\^18 + 18480\ M\^20 - 1385\ M\^22 - 68980\ M\^24 + 
                    98095\ M\^26 + 133931\ M\^28 - 447926\ M\^30 - 
                    25627\ M\^32 + 1161518\ M\^34 - 496569\ M\^36 - 
                    2263028\ M\^38 + 1550492\ M\^40 + 3560049\ M\^42 - 
                    2653215\ M\^44 - 4719159\ M\^46 + 2574551\ M\^48 + 
                    5311613\ M\^50 - 30832\ M\^52 - 4891898\ M\^54 - 
                    4886974\ M\^56 + 3507477\ M\^58 + 9984076\ M\^60 - 
                    2219650\ M\^62 - 12800364\ M\^64 + 2455792\ M\^66 + 
                    12722302\ M\^68 - 3888652\ M\^70 - 10416948\ M\^72 + 
                    5120627\ M\^74 + 6929279\ M\^76 - 5098129\ M\^78 - 
                    3519099\ M\^80 + 3890109\ M\^82 + 1130056\ M\^84 - 
                    2264094\ M\^86 - 10148\ M\^88 + 974873\ M\^90 - 
                    236006\ M\^92 - 288176\ M\^94 + 157830\ M\^96 + 
                    36552\ M\^98 - 51515\ M\^100 + 6469\ M\^102 + 
                    10929\ M\^104 - 7321\ M\^106 + 2319\ M\^108 - 
                    426\ M\^110 + 44\ M\^112 - 2\ M\^114)\) + 
              L\^7\ \((\(-M\^16\) + 26\ M\^18 - 310\ M\^20 + 2141\ M\^22 - 
                    9055\ M\^24 + 22043\ M\^26 - 18167\ M\^28 - 
                    55729\ M\^30 + 165725\ M\^32 - 24236\ M\^34 - 
                    532733\ M\^36 + 605008\ M\^38 + 956167\ M\^40 - 
                    2139941\ M\^42 - 984276\ M\^44 + 4844364\ M\^46 + 
                    141299\ M\^48 - 8321163\ M\^50 + 1433716\ M\^52 + 
                    11532800\ M\^54 - 2771152\ M\^56 - 13091886\ M\^58 + 
                    2859252\ M\^60 + 11950195\ M\^62 - 1745322\ M\^64 - 
                    8534659\ M\^66 + 431142\ M\^68 + 4729988\ M\^70 + 
                    542447\ M\^72 - 2088086\ M\^74 - 984385\ M\^76 + 
                    879132\ M\^78 + 919828\ M\^80 - 552477\ M\^82 - 
                    526873\ M\^84 + 436030\ M\^86 + 134210\ M\^88 - 
                    267551\ M\^90 + 61349\ M\^92 + 93886\ M\^94 - 
                    77940\ M\^96 + 900\ M\^98 + 29445\ M\^100 - 
                    13589\ M\^102 - 4215\ M\^104 + 7416\ M\^106 - 
                    3921\ M\^108 + 1174\ M\^110 - 213\ M\^112 + 22\ M\^114 - 
                    M\^116)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 29]\), \(\(-M\^120\) - 
              2\ M\^122 - M\^124 + L\^19\ \((M\^24 + 2\ M\^26 + M\^28)\) + 
              L\^18\ \((\(-2\)\ M\^16 + 5\ M\^18 - 5\ M\^20 - 6\ M\^22 + 
                    53\ M\^24 - 32\ M\^26 - 86\ M\^28 + 84\ M\^30 + 
                    81\ M\^32 - 21\ M\^34 - 11\ M\^36 + 4\ M\^38 + 4\ M\^40 + 
                    M\^42 - M\^44)\) + 
              L\^17\ \((M\^8 - 7\ M\^10 + 27\ M\^12 - 52\ M\^14 + 31\ M\^16 + 
                    86\ M\^18 - 253\ M\^20 + 367\ M\^22 - 10\ M\^24 - 
                    1201\ M\^26 + 1318\ M\^28 + 1454\ M\^30 - 2509\ M\^32 - 
                    737\ M\^34 + 2255\ M\^36 + 325\ M\^38 - 839\ M\^40 - 
                    28\ M\^42 + 410\ M\^44 + 30\ M\^46 - 173\ M\^48 + 
                    24\ M\^50 + 29\ M\^52 - 8\ M\^54)\) + 
              L\^16\ \((\(-2\)\ M\^4 + 21\ M\^6 - 103\ M\^8 + 265\ M\^10 - 
                    261\ M\^12 - 420\ M\^14 + 1472\ M\^16 - 980\ M\^18 - 
                    1707\ M\^20 + 3227\ M\^22 - 1431\ M\^24 - 1055\ M\^26 + 
                    5094\ M\^28 - 8989\ M\^30 - 1829\ M\^32 + 19328\ M\^34 - 
                    9039\ M\^36 - 18680\ M\^38 + 18704\ M\^40 + 9135\ M\^42 - 
                    18372\ M\^44 + 464\ M\^46 + 13590\ M\^48 - 2128\ M\^50 - 
                    6664\ M\^52 + 2123\ M\^54 + 1910\ M\^56 - 1025\ M\^58 - 
                    144\ M\^60 + 176\ M\^62 - 28\ M\^64)\) + 
              L\^15\ \((1 - 14\ M\^2 + 91\ M\^4 - 334\ M\^6 + 656\ M\^8 - 
                    232\ M\^10 - 1994\ M\^12 + 3916\ M\^14 + 1118\ M\^16 - 
                    12444\ M\^18 + 8785\ M\^20 + 22100\ M\^22 - 
                    31999\ M\^24 - 29348\ M\^26 + 69788\ M\^28 + 
                    37777\ M\^30 - 129057\ M\^32 - 43885\ M\^34 + 
                    214063\ M\^36 + 23872\ M\^38 - 303143\ M\^40 + 
                    39112\ M\^42 + 341552\ M\^44 - 114546\ M\^46 - 
                    285771\ M\^48 + 151832\ M\^50 + 173664\ M\^52 - 
                    121691\ M\^54 - 68098\ M\^56 + 70088\ M\^58 + 
                    17059\ M\^60 - 29416\ M\^62 - 573\ M\^64 + 8241\ M\^66 - 
                    1478\ M\^68 - 1164\ M\^70 + 504\ M\^72 - 56\ M\^74)\) + 
              L\^14\ \((\(-1\) + 16\ M\^2 - 122\ M\^4 + 546\ M\^6 - 
                    1438\ M\^8 + 1605\ M\^10 + 2464\ M\^12 - 10729\ M\^14 + 
                    6899\ M\^16 + 28065\ M\^18 - 52835\ M\^20 - 
                    27893\ M\^22 + 153396\ M\^24 - 38559\ M\^26 - 
                    290074\ M\^28 + 209693\ M\^30 + 396489\ M\^32 - 
                    477105\ M\^34 - 396184\ M\^36 + 795377\ M\^38 + 
                    272298\ M\^40 - 1125409\ M\^42 - 74808\ M\^44 + 
                    1402992\ M\^46 - 149848\ M\^48 - 1518049\ M\^50 + 
                    363228\ M\^52 + 1383358\ M\^54 - 477477\ M\^56 - 
                    998353\ M\^58 + 454308\ M\^60 + 554327\ M\^62 - 
                    296047\ M\^64 - 218556\ M\^66 + 133228\ M\^68 + 
                    64740\ M\^70 - 44269\ M\^72 - 15451\ M\^74 + 
                    12619\ M\^76 + 1955\ M\^78 - 3120\ M\^80 + 826\ M\^82 - 
                    70\ M\^84)\) + 
              L\^13\ \((\(-2\)\ M\^2 + 32\ M\^4 - 234\ M\^6 + 969\ M\^8 - 
                    2226\ M\^10 + 1643\ M\^12 + 4913\ M\^14 - 11699\ M\^16 - 
                    5334\ M\^18 + 44770\ M\^20 - 22642\ M\^22 - 
                    102551\ M\^24 + 116138\ M\^26 + 160175\ M\^28 - 
                    307419\ M\^30 - 133060\ M\^32 + 557828\ M\^34 - 
                    83906\ M\^36 - 736985\ M\^38 + 492820\ M\^40 + 
                    687523\ M\^42 - 947848\ M\^44 - 315148\ M\^46 + 
                    1303124\ M\^48 - 320881\ M\^50 - 1615289\ M\^52 + 
                    931579\ M\^54 + 1940740\ M\^56 - 1235663\ M\^58 - 
                    2097372\ M\^60 + 1220508\ M\^62 + 1905108\ M\^64 - 
                    958277\ M\^66 - 1333859\ M\^68 + 634945\ M\^70 + 
                    702547\ M\^72 - 318011\ M\^74 - 258250\ M\^76 + 
                    96642\ M\^78 + 75664\ M\^80 - 12525\ M\^82 - 
                    26880\ M\^84 + 3487\ M\^86 + 8476\ M\^88 - 4356\ M\^90 + 
                    826\ M\^92 - 56\ M\^94)\) + 
              L\^12\ \((\(-M\^4\) + 16\ M\^6 - 118\ M\^8 + 513\ M\^10 - 
                    1394\ M\^12 + 2290\ M\^14 - 1812\ M\^16 - 640\ M\^18 + 
                    3586\ M\^20 - 4174\ M\^22 - 6370\ M\^24 + 31254\ M\^26 - 
                    17639\ M\^28 - 80134\ M\^30 + 92343\ M\^32 + 
                    136827\ M\^34 - 223311\ M\^36 - 165093\ M\^38 + 
                    383919\ M\^40 + 150688\ M\^42 - 575539\ M\^44 - 
                    158767\ M\^46 + 803909\ M\^48 + 232116\ M\^50 - 
                    970107\ M\^52 - 245029\ M\^54 + 920688\ M\^56 - 
                    10989\ M\^58 - 799039\ M\^60 + 294470\ M\^62 + 
                    763998\ M\^64 - 145546\ M\^66 - 697787\ M\^68 - 
                    316116\ M\^70 + 596580\ M\^72 + 686773\ M\^74 - 
                    411446\ M\^76 - 629544\ M\^78 + 242106\ M\^80 + 
                    348982\ M\^82 - 103134\ M\^84 - 103869\ M\^86 + 
                    1794\ M\^88 + 16781\ M\^90 + 27865\ M\^92 - 
                    13184\ M\^94 - 11463\ M\^96 + 10767\ M\^98 - 
                    3456\ M\^100 + 504\ M\^102 - 28\ M\^104)\) + 
              L\^11\ \((\(-2\)\ M\^10 + 35\ M\^12 - 275\ M\^14 + 
                    1249\ M\^16 - 3375\ M\^18 + 4573\ M\^20 + 510\ M\^22 - 
                    9358\ M\^24 + 225\ M\^26 + 32443\ M\^28 - 29025\ M\^30 - 
                    54051\ M\^32 + 111300\ M\^34 + 16344\ M\^36 - 
                    277385\ M\^38 + 235035\ M\^40 + 485464\ M\^42 - 
                    911128\ M\^44 - 578499\ M\^46 + 2183041\ M\^48 + 
                    365602\ M\^50 - 4042838\ M\^52 + 39975\ M\^54 + 
                    6040423\ M\^56 - 264468\ M\^58 - 7281313\ M\^60 + 
                    201270\ M\^62 + 7007445\ M\^64 - 469443\ M\^66 - 
                    5636965\ M\^68 + 1250828\ M\^70 + 4083677\ M\^72 - 
                    1828551\ M\^74 - 2611796\ M\^76 + 1790770\ M\^78 + 
                    1351721\ M\^80 - 1206005\ M\^82 - 421391\ M\^84 + 
                    575125\ M\^86 + 18272\ M\^88 - 178027\ M\^90 + 
                    37185\ M\^92 + 22594\ M\^94 + 12679\ M\^96 - 
                    7259\ M\^98 - 27114\ M\^100 + 18590\ M\^102 + 
                    8387\ M\^104 - 13987\ M\^106 + 6585\ M\^108 - 
                    1524\ M\^110 + 176\ M\^112 - 8\ M\^114)\) + 
              L\^10\ \((\(-M\^16\) + 20\ M\^18 - 189\ M\^20 + 1085\ M\^22 - 
                    3910\ M\^24 + 8029\ M\^26 - 5578\ M\^28 - 10569\ M\^30 + 
                    16996\ M\^32 + 21123\ M\^34 - 50664\ M\^36 - 
                    20119\ M\^38 + 87043\ M\^40 + 13162\ M\^42 - 
                    94948\ M\^44 - 34042\ M\^46 + 72304\ M\^48 + 
                    118228\ M\^50 - 164063\ M\^52 - 183725\ M\^54 + 
                    725900\ M\^56 + 101887\ M\^58 - 2307298\ M\^60 - 
                    148056\ M\^62 + 4841449\ M\^64 + 960852\ M\^66 - 
                    7140908\ M\^68 - 2697591\ M\^70 + 7495676\ M\^72 + 
                    4044197\ M\^74 - 5935872\ M\^76 - 3899947\ M\^78 + 
                    4019270\ M\^80 + 2749903\ M\^82 - 2579869\ M\^84 - 
                    1452065\ M\^86 + 1624041\ M\^88 + 559277\ M\^90 - 
                    872562\ M\^92 - 74512\ M\^94 + 328490\ M\^96 - 
                    69943\ M\^98 - 48582\ M\^100 + 50264\ M\^102 - 
                    33384\ M\^104 - 4478\ M\^106 + 25918\ M\^108 - 
                    12386\ M\^110 - 7188\ M\^112 + 11514\ M\^114 - 
                    6332\ M\^116 + 1897\ M\^118 - 324\ M\^120 + 29\ M\^122 - 
                    M\^124)\) + 
              L\^9\ \((M\^24 - 29\ M\^26 + 324\ M\^28 - 1897\ M\^30 + 
                    6332\ M\^32 - 11514\ M\^34 + 7188\ M\^36 + 12386\ M\^38 - 
                    25918\ M\^40 + 4478\ M\^42 + 33384\ M\^44 - 
                    50264\ M\^46 + 48582\ M\^48 + 69943\ M\^50 - 
                    328490\ M\^52 + 74512\ M\^54 + 872562\ M\^56 - 
                    559277\ M\^58 - 1624041\ M\^60 + 1452065\ M\^62 + 
                    2579869\ M\^64 - 2749903\ M\^66 - 4019270\ M\^68 + 
                    3899947\ M\^70 + 5935872\ M\^72 - 4044197\ M\^74 - 
                    7495676\ M\^76 + 2697591\ M\^78 + 7140908\ M\^80 - 
                    960852\ M\^82 - 4841449\ M\^84 + 148056\ M\^86 + 
                    2307298\ M\^88 - 101887\ M\^90 - 725900\ M\^92 + 
                    183725\ M\^94 + 164063\ M\^96 - 118228\ M\^98 - 
                    72304\ M\^100 + 34042\ M\^102 + 94948\ M\^104 - 
                    13162\ M\^106 - 87043\ M\^108 + 20119\ M\^110 + 
                    50664\ M\^112 - 21123\ M\^114 - 16996\ M\^116 + 
                    10569\ M\^118 + 5578\ M\^120 - 8029\ M\^122 + 
                    3910\ M\^124 - 1085\ M\^126 + 189\ M\^128 - 20\ M\^130 + 
                    M\^132)\) + 
              L\ \((M\^104 - M\^106 - 4\ M\^108 - 4\ M\^110 + 11\ M\^112 + 
                    21\ M\^114 - 81\ M\^116 - 84\ M\^118 + 86\ M\^120 + 
                    32\ M\^122 - 53\ M\^124 + 6\ M\^126 + 5\ M\^128 - 
                    5\ M\^130 + 2\ M\^132)\) + 
              L\^8\ \((8\ M\^34 - 176\ M\^36 + 1524\ M\^38 - 6585\ M\^40 + 
                    13987\ M\^42 - 8387\ M\^44 - 18590\ M\^46 + 
                    27114\ M\^48 + 7259\ M\^50 - 12679\ M\^52 - 
                    22594\ M\^54 - 37185\ M\^56 + 178027\ M\^58 - 
                    18272\ M\^60 - 575125\ M\^62 + 421391\ M\^64 + 
                    1206005\ M\^66 - 1351721\ M\^68 - 1790770\ M\^70 + 
                    2611796\ M\^72 + 1828551\ M\^74 - 4083677\ M\^76 - 
                    1250828\ M\^78 + 5636965\ M\^80 + 469443\ M\^82 - 
                    7007445\ M\^84 - 201270\ M\^86 + 7281313\ M\^88 + 
                    264468\ M\^90 - 6040423\ M\^92 - 39975\ M\^94 + 
                    4042838\ M\^96 - 365602\ M\^98 - 2183041\ M\^100 + 
                    578499\ M\^102 + 911128\ M\^104 - 485464\ M\^106 - 
                    235035\ M\^108 + 277385\ M\^110 - 16344\ M\^112 - 
                    111300\ M\^114 + 54051\ M\^116 + 29025\ M\^118 - 
                    32443\ M\^120 - 225\ M\^122 + 9358\ M\^124 - 
                    510\ M\^126 - 4573\ M\^128 + 3375\ M\^130 - 
                    1249\ M\^132 + 275\ M\^134 - 35\ M\^136 + 2\ M\^138)\) + 
              L\^2\ \((8\ M\^94 - 29\ M\^96 - 24\ M\^98 + 173\ M\^100 - 
                    30\ M\^102 - 410\ M\^104 + 28\ M\^106 + 839\ M\^108 - 
                    325\ M\^110 - 2255\ M\^112 + 737\ M\^114 + 2509\ M\^116 - 
                    1454\ M\^118 - 1318\ M\^120 + 1201\ M\^122 + 10\ M\^124 - 
                    367\ M\^126 + 253\ M\^128 - 86\ M\^130 - 31\ M\^132 + 
                    52\ M\^134 - 27\ M\^136 + 7\ M\^138 - M\^140)\) + 
              L\^7\ \((28\ M\^44 - 504\ M\^46 + 3456\ M\^48 - 10767\ M\^50 + 
                    11463\ M\^52 + 13184\ M\^54 - 27865\ M\^56 - 
                    16781\ M\^58 - 1794\ M\^60 + 103869\ M\^62 + 
                    103134\ M\^64 - 348982\ M\^66 - 242106\ M\^68 + 
                    629544\ M\^70 + 411446\ M\^72 - 686773\ M\^74 - 
                    596580\ M\^76 + 316116\ M\^78 + 697787\ M\^80 + 
                    145546\ M\^82 - 763998\ M\^84 - 294470\ M\^86 + 
                    799039\ M\^88 + 10989\ M\^90 - 920688\ M\^92 + 
                    245029\ M\^94 + 970107\ M\^96 - 232116\ M\^98 - 
                    803909\ M\^100 + 158767\ M\^102 + 575539\ M\^104 - 
                    150688\ M\^106 - 383919\ M\^108 + 165093\ M\^110 + 
                    223311\ M\^112 - 136827\ M\^114 - 92343\ M\^116 + 
                    80134\ M\^118 + 17639\ M\^120 - 31254\ M\^122 + 
                    6370\ M\^124 + 4174\ M\^126 - 3586\ M\^128 + 
                    640\ M\^130 + 1812\ M\^132 - 2290\ M\^134 + 
                    1394\ M\^136 - 513\ M\^138 + 118\ M\^140 - 16\ M\^142 + 
                    M\^144)\) + 
              L\^3\ \((28\ M\^84 - 176\ M\^86 + 144\ M\^88 + 1025\ M\^90 - 
                    1910\ M\^92 - 2123\ M\^94 + 6664\ M\^96 + 2128\ M\^98 - 
                    13590\ M\^100 - 464\ M\^102 + 18372\ M\^104 - 
                    9135\ M\^106 - 18704\ M\^108 + 18680\ M\^110 + 
                    9039\ M\^112 - 19328\ M\^114 + 1829\ M\^116 + 
                    8989\ M\^118 - 5094\ M\^120 + 1055\ M\^122 + 
                    1431\ M\^124 - 3227\ M\^126 + 1707\ M\^128 + 
                    980\ M\^130 - 1472\ M\^132 + 420\ M\^134 + 261\ M\^136 - 
                    265\ M\^138 + 103\ M\^140 - 21\ M\^142 + 2\ M\^144)\) + 
              L\^6\ \((56\ M\^54 - 826\ M\^56 + 4356\ M\^58 - 8476\ M\^60 - 
                    3487\ M\^62 + 26880\ M\^64 + 12525\ M\^66 - 
                    75664\ M\^68 - 96642\ M\^70 + 258250\ M\^72 + 
                    318011\ M\^74 - 702547\ M\^76 - 634945\ M\^78 + 
                    1333859\ M\^80 + 958277\ M\^82 - 1905108\ M\^84 - 
                    1220508\ M\^86 + 2097372\ M\^88 + 1235663\ M\^90 - 
                    1940740\ M\^92 - 931579\ M\^94 + 1615289\ M\^96 + 
                    320881\ M\^98 - 1303124\ M\^100 + 315148\ M\^102 + 
                    947848\ M\^104 - 687523\ M\^106 - 492820\ M\^108 + 
                    736985\ M\^110 + 83906\ M\^112 - 557828\ M\^114 + 
                    133060\ M\^116 + 307419\ M\^118 - 160175\ M\^120 - 
                    116138\ M\^122 + 102551\ M\^124 + 22642\ M\^126 - 
                    44770\ M\^128 + 5334\ M\^130 + 11699\ M\^132 - 
                    4913\ M\^134 - 1643\ M\^136 + 2226\ M\^138 - 
                    969\ M\^140 + 234\ M\^142 - 32\ M\^144 + 2\ M\^146)\) + 
              L\^4\ \((56\ M\^74 - 504\ M\^76 + 1164\ M\^78 + 1478\ M\^80 - 
                    8241\ M\^82 + 573\ M\^84 + 29416\ M\^86 - 17059\ M\^88 - 
                    70088\ M\^90 + 68098\ M\^92 + 121691\ M\^94 - 
                    173664\ M\^96 - 151832\ M\^98 + 285771\ M\^100 + 
                    114546\ M\^102 - 341552\ M\^104 - 39112\ M\^106 + 
                    303143\ M\^108 - 23872\ M\^110 - 214063\ M\^112 + 
                    43885\ M\^114 + 129057\ M\^116 - 37777\ M\^118 - 
                    69788\ M\^120 + 29348\ M\^122 + 31999\ M\^124 - 
                    22100\ M\^126 - 8785\ M\^128 + 12444\ M\^130 - 
                    1118\ M\^132 - 3916\ M\^134 + 1994\ M\^136 + 
                    232\ M\^138 - 656\ M\^140 + 334\ M\^142 - 91\ M\^144 + 
                    14\ M\^146 - M\^148)\) + 
              L\^5\ \((70\ M\^64 - 826\ M\^66 + 3120\ M\^68 - 1955\ M\^70 - 
                    12619\ M\^72 + 15451\ M\^74 + 44269\ M\^76 - 
                    64740\ M\^78 - 133228\ M\^80 + 218556\ M\^82 + 
                    296047\ M\^84 - 554327\ M\^86 - 454308\ M\^88 + 
                    998353\ M\^90 + 477477\ M\^92 - 1383358\ M\^94 - 
                    363228\ M\^96 + 1518049\ M\^98 + 149848\ M\^100 - 
                    1402992\ M\^102 + 74808\ M\^104 + 1125409\ M\^106 - 
                    272298\ M\^108 - 795377\ M\^110 + 396184\ M\^112 + 
                    477105\ M\^114 - 396489\ M\^116 - 209693\ M\^118 + 
                    290074\ M\^120 + 38559\ M\^122 - 153396\ M\^124 + 
                    27893\ M\^126 + 52835\ M\^128 - 28065\ M\^130 - 
                    6899\ M\^132 + 10729\ M\^134 - 2464\ M\^136 - 
                    1605\ M\^138 + 1438\ M\^140 - 546\ M\^142 + 122\ M\^144 - 
                    16\ M\^146 + M\^148)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 30]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[9, 31]\), \(L\^17\ M\^22 + M\^96 + 
              L\^16\ \((3\ M\^16 - 17\ M\^18 + 39\ M\^20 - 24\ M\^22 - 
                    51\ M\^24 + 93\ M\^26 - 3\ M\^28 - 46\ M\^30 + 
                    29\ M\^32 - 6\ M\^34)\) + 
              L\^15\ \((3\ M\^10 - 30\ M\^12 + 123\ M\^14 - 227\ M\^16 + 
                    86\ M\^18 + 275\ M\^20 - 277\ M\^22 + 150\ M\^24 - 
                    592\ M\^26 + 262\ M\^28 + 1471\ M\^30 - 1482\ M\^32 - 
                    527\ M\^34 + 1641\ M\^36 - 608\ M\^38 - 455\ M\^40 + 
                    442\ M\^42 - 134\ M\^44 + 15\ M\^46)\) + 
              L\^14\ \((M\^4 - 13\ M\^6 + 74\ M\^8 - 243\ M\^10 + 
                    528\ M\^12 - 813\ M\^14 + 579\ M\^16 + 1490\ M\^18 - 
                    5137\ M\^20 + 3730\ M\^22 + 8042\ M\^24 - 14545\ M\^26 - 
                    3411\ M\^28 + 21323\ M\^30 - 3333\ M\^32 - 20344\ M\^34 + 
                    4977\ M\^36 + 18742\ M\^38 - 4513\ M\^40 - 15624\ M\^42 + 
                    7100\ M\^44 + 7703\ M\^46 - 6827\ M\^48 - 216\ M\^50 + 
                    2354\ M\^52 - 1170\ M\^54 + 246\ M\^56 - 20\ M\^58)\) + 
              L\^13\ \((2\ M\^2 - 33\ M\^4 + 235\ M\^6 - 910\ M\^8 + 
                    1872\ M\^10 - 989\ M\^12 - 4223\ M\^14 + 7855\ M\^16 + 
                    4784\ M\^18 - 23494\ M\^20 - 2851\ M\^22 + 62013\ M\^24 - 
                    14107\ M\^26 - 135264\ M\^28 + 78747\ M\^30 + 
                    204880\ M\^32 - 188296\ M\^34 - 200565\ M\^36 + 
                    266045\ M\^38 + 117992\ M\^40 - 241337\ M\^42 - 
                    30651\ M\^44 + 150644\ M\^46 - 573\ M\^48 - 
                    72788\ M\^50 - 1281\ M\^52 + 36447\ M\^54 - 1419\ M\^56 - 
                    17919\ M\^58 + 6479\ M\^60 + 3773\ M\^62 - 3816\ M\^64 + 
                    1337\ M\^66 - 224\ M\^68 + 15\ M\^70)\) + 
              L\^12\ \((1 - 19\ M\^2 + 157\ M\^4 - 729\ M\^6 + 2032\ M\^8 - 
                    3237\ M\^10 + 1784\ M\^12 + 4377\ M\^14 - 14701\ M\^16 + 
                    22376\ M\^18 + 4186\ M\^20 - 94824\ M\^22 + 
                    119742\ M\^24 + 156406\ M\^26 - 431711\ M\^28 - 
                    91020\ M\^30 + 908009\ M\^32 - 190318\ M\^34 - 
                    1400092\ M\^36 + 665413\ M\^38 + 1678528\ M\^40 - 
                    1170811\ M\^42 - 1549609\ M\^44 + 1453259\ M\^46 + 
                    1044108\ M\^48 - 1319379\ M\^50 - 434800\ M\^52 + 
                    865542\ M\^54 + 41781\ M\^56 - 387082\ M\^58 + 
                    62202\ M\^60 + 104034\ M\^62 - 27829\ M\^64 - 
                    16213\ M\^66 - 1190\ M\^68 + 7308\ M\^70 + 972\ M\^72 - 
                    4375\ M\^74 + 2517\ M\^76 - 702\ M\^78 + 101\ M\^80 - 
                    6\ M\^82)\) + 
              L\^11\ \((3\ M\^2 - 61\ M\^4 + 533\ M\^6 - 2546\ M\^8 + 
                    6739\ M\^10 - 7067\ M\^12 - 11081\ M\^14 + 41474\ M\^16 - 
                    22395\ M\^18 - 79127\ M\^20 + 117169\ M\^22 + 
                    47970\ M\^24 - 167190\ M\^26 + 17364\ M\^28 - 
                    21214\ M\^30 + 155903\ M\^32 + 507340\ M\^34 - 
                    947348\ M\^36 - 1024805\ M\^38 + 2428791\ M\^40 + 
                    1128037\ M\^42 - 4109135\ M\^44 - 572728\ M\^46 + 
                    5181375\ M\^48 - 434493\ M\^50 - 5074466\ M\^52 + 
                    1395515\ M\^54 + 3907203\ M\^56 - 1815733\ M\^58 - 
                    2294399\ M\^60 + 1590106\ M\^62 + 930324\ M\^64 - 
                    1000897\ M\^66 - 172367\ M\^68 + 438039\ M\^70 - 
                    61692\ M\^72 - 112493\ M\^74 + 51823\ M\^76 + 
                    5351\ M\^78 - 10809\ M\^80 + 4931\ M\^82 - 2601\ M\^84 + 
                    1502\ M\^86 - 591\ M\^88 + 139\ M\^90 - 18\ M\^92 + 
                    M\^94)\) + 
              L\^10\ \((3\ M\^4 - 65\ M\^6 + 611\ M\^8 - 3178\ M\^10 + 
                    9355\ M\^12 - 11822\ M\^14 - 14756\ M\^16 + 
                    74126\ M\^18 - 55866\ M\^20 - 176931\ M\^22 + 
                    355009\ M\^24 + 160533\ M\^26 - 983256\ M\^28 + 
                    245405\ M\^30 + 1879060\ M\^32 - 1245229\ M\^34 - 
                    2818360\ M\^36 + 2739931\ M\^38 + 3617217\ M\^40 - 
                    4338818\ M\^42 - 4260933\ M\^44 + 5606078\ M\^46 + 
                    4773927\ M\^48 - 6289343\ M\^50 - 4982161\ M\^52 + 
                    6315526\ M\^54 + 4594209\ M\^56 - 5627332\ M\^58 - 
                    3497351\ M\^60 + 4365132\ M\^62 + 2107878\ M\^64 - 
                    2858216\ M\^66 - 984119\ M\^68 + 1540836\ M\^70 + 
                    361531\ M\^72 - 687634\ M\^74 - 111726\ M\^76 + 
                    268894\ M\^78 + 27460\ M\^80 - 101922\ M\^82 + 
                    5225\ M\^84 + 33208\ M\^86 - 10542\ M\^88 - 6645\ M\^90 + 
                    6510\ M\^92 - 2474\ M\^94 + 520\ M\^96 - 60\ M\^98 + 
                    3\ M\^100)\) + 
              L\ \((\(-6\)\ M\^84 + 29\ M\^86 - 46\ M\^88 - 3\ M\^90 + 
                    93\ M\^92 - 51\ M\^94 - 24\ M\^96 + 39\ M\^98 - 
                    17\ M\^100 + 3\ M\^102)\) + 
              L\^9\ \((M\^6 - 23\ M\^8 + 236\ M\^10 - 1395\ M\^12 + 
                    5041\ M\^14 - 10280\ M\^16 + 5209\ M\^18 + 28796\ M\^20 - 
                    66314\ M\^22 - 2621\ M\^24 + 207104\ M\^26 - 
                    212969\ M\^28 - 328225\ M\^30 + 737429\ M\^32 + 
                    227157\ M\^34 - 1566263\ M\^36 + 260705\ M\^38 + 
                    2585912\ M\^40 - 1085658\ M\^42 - 3781847\ M\^44 + 
                    2068033\ M\^46 + 5269686\ M\^48 - 3090501\ M\^50 - 
                    7045607\ M\^52 + 4161233\ M\^54 + 8708059\ M\^56 - 
                    5295544\ M\^58 - 9472248\ M\^60 + 6381043\ M\^62 + 
                    8732600\ M\^64 - 6911720\ M\^66 - 6473484\ M\^68 + 
                    6399693\ M\^70 + 3555326\ M\^72 - 4853408\ M\^74 - 
                    1158767\ M\^76 + 2895477\ M\^78 - 87086\ M\^80 - 
                    1285653\ M\^82 + 366744\ M\^84 + 379718\ M\^86 - 
                    235646\ M\^88 - 42318\ M\^90 + 78849\ M\^92 - 
                    17202\ M\^94 - 11869\ M\^96 + 9585\ M\^98 - 
                    3231\ M\^100 + 615\ M\^102 - 65\ M\^104 + 3\ M\^106)\) + 
              L\^2\ \((15\ M\^72 - 134\ M\^74 + 442\ M\^76 - 455\ M\^78 - 
                    608\ M\^80 + 1641\ M\^82 - 527\ M\^84 - 1482\ M\^86 + 
                    1471\ M\^88 + 262\ M\^90 - 592\ M\^92 + 150\ M\^94 - 
                    277\ M\^96 + 275\ M\^98 + 86\ M\^100 - 227\ M\^102 + 
                    123\ M\^104 - 30\ M\^106 + 3\ M\^108)\) + 
              L\^8\ \((3\ M\^12 - 65\ M\^14 + 615\ M\^16 - 3231\ M\^18 + 
                    9585\ M\^20 - 11869\ M\^22 - 17202\ M\^24 + 
                    78849\ M\^26 - 42318\ M\^28 - 235646\ M\^30 + 
                    379718\ M\^32 + 366744\ M\^34 - 1285653\ M\^36 - 
                    87086\ M\^38 + 2895477\ M\^40 - 1158767\ M\^42 - 
                    4853408\ M\^44 + 3555326\ M\^46 + 6399693\ M\^48 - 
                    6473484\ M\^50 - 6911720\ M\^52 + 8732600\ M\^54 + 
                    6381043\ M\^56 - 9472248\ M\^58 - 5295544\ M\^60 + 
                    8708059\ M\^62 + 4161233\ M\^64 - 7045607\ M\^66 - 
                    3090501\ M\^68 + 5269686\ M\^70 + 2068033\ M\^72 - 
                    3781847\ M\^74 - 1085658\ M\^76 + 2585912\ M\^78 + 
                    260705\ M\^80 - 1566263\ M\^82 + 227157\ M\^84 + 
                    737429\ M\^86 - 328225\ M\^88 - 212969\ M\^90 + 
                    207104\ M\^92 - 2621\ M\^94 - 66314\ M\^96 + 
                    28796\ M\^98 + 5209\ M\^100 - 10280\ M\^102 + 
                    5041\ M\^104 - 1395\ M\^106 + 236\ M\^108 - 23\ M\^110 + 
                    M\^112)\) + 
              L\^3\ \((\(-20\)\ M\^60 + 246\ M\^62 - 1170\ M\^64 + 
                    2354\ M\^66 - 216\ M\^68 - 6827\ M\^70 + 7703\ M\^72 + 
                    7100\ M\^74 - 15624\ M\^76 - 4513\ M\^78 + 18742\ M\^80 + 
                    4977\ M\^82 - 20344\ M\^84 - 3333\ M\^86 + 21323\ M\^88 - 
                    3411\ M\^90 - 14545\ M\^92 + 8042\ M\^94 + 3730\ M\^96 - 
                    5137\ M\^98 + 1490\ M\^100 + 579\ M\^102 - 813\ M\^104 + 
                    528\ M\^106 - 243\ M\^108 + 74\ M\^110 - 13\ M\^112 + 
                    M\^114)\) + 
              L\^7\ \((3\ M\^18 - 60\ M\^20 + 520\ M\^22 - 2474\ M\^24 + 
                    6510\ M\^26 - 6645\ M\^28 - 10542\ M\^30 + 33208\ M\^32 + 
                    5225\ M\^34 - 101922\ M\^36 + 27460\ M\^38 + 
                    268894\ M\^40 - 111726\ M\^42 - 687634\ M\^44 + 
                    361531\ M\^46 + 1540836\ M\^48 - 984119\ M\^50 - 
                    2858216\ M\^52 + 2107878\ M\^54 + 4365132\ M\^56 - 
                    3497351\ M\^58 - 5627332\ M\^60 + 4594209\ M\^62 + 
                    6315526\ M\^64 - 4982161\ M\^66 - 6289343\ M\^68 + 
                    4773927\ M\^70 + 5606078\ M\^72 - 4260933\ M\^74 - 
                    4338818\ M\^76 + 3617217\ M\^78 + 2739931\ M\^80 - 
                    2818360\ M\^82 - 1245229\ M\^84 + 1879060\ M\^86 + 
                    245405\ M\^88 - 983256\ M\^90 + 160533\ M\^92 + 
                    355009\ M\^94 - 176931\ M\^96 - 55866\ M\^98 + 
                    74126\ M\^100 - 14756\ M\^102 - 11822\ M\^104 + 
                    9355\ M\^106 - 3178\ M\^108 + 611\ M\^110 - 65\ M\^112 + 
                    3\ M\^114)\) + 
              L\^4\ \((15\ M\^48 - 224\ M\^50 + 1337\ M\^52 - 3816\ M\^54 + 
                    3773\ M\^56 + 6479\ M\^58 - 17919\ M\^60 - 1419\ M\^62 + 
                    36447\ M\^64 - 1281\ M\^66 - 72788\ M\^68 - 573\ M\^70 + 
                    150644\ M\^72 - 30651\ M\^74 - 241337\ M\^76 + 
                    117992\ M\^78 + 266045\ M\^80 - 200565\ M\^82 - 
                    188296\ M\^84 + 204880\ M\^86 + 78747\ M\^88 - 
                    135264\ M\^90 - 14107\ M\^92 + 62013\ M\^94 - 
                    2851\ M\^96 - 23494\ M\^98 + 4784\ M\^100 + 
                    7855\ M\^102 - 4223\ M\^104 - 989\ M\^106 + 
                    1872\ M\^108 - 910\ M\^110 + 235\ M\^112 - 33\ M\^114 + 
                    2\ M\^116)\) + 
              L\^6\ \((M\^24 - 18\ M\^26 + 139\ M\^28 - 591\ M\^30 + 
                    1502\ M\^32 - 2601\ M\^34 + 4931\ M\^36 - 10809\ M\^38 + 
                    5351\ M\^40 + 51823\ M\^42 - 112493\ M\^44 - 
                    61692\ M\^46 + 438039\ M\^48 - 172367\ M\^50 - 
                    1000897\ M\^52 + 930324\ M\^54 + 1590106\ M\^56 - 
                    2294399\ M\^58 - 1815733\ M\^60 + 3907203\ M\^62 + 
                    1395515\ M\^64 - 5074466\ M\^66 - 434493\ M\^68 + 
                    5181375\ M\^70 - 572728\ M\^72 - 4109135\ M\^74 + 
                    1128037\ M\^76 + 2428791\ M\^78 - 1024805\ M\^80 - 
                    947348\ M\^82 + 507340\ M\^84 + 155903\ M\^86 - 
                    21214\ M\^88 + 17364\ M\^90 - 167190\ M\^92 + 
                    47970\ M\^94 + 117169\ M\^96 - 79127\ M\^98 - 
                    22395\ M\^100 + 41474\ M\^102 - 11081\ M\^104 - 
                    7067\ M\^106 + 6739\ M\^108 - 2546\ M\^110 + 
                    533\ M\^112 - 61\ M\^114 + 3\ M\^116)\) + 
              L\^5\ \((\(-6\)\ M\^36 + 101\ M\^38 - 702\ M\^40 + 
                    2517\ M\^42 - 4375\ M\^44 + 972\ M\^46 + 7308\ M\^48 - 
                    1190\ M\^50 - 16213\ M\^52 - 27829\ M\^54 + 
                    104034\ M\^56 + 62202\ M\^58 - 387082\ M\^60 + 
                    41781\ M\^62 + 865542\ M\^64 - 434800\ M\^66 - 
                    1319379\ M\^68 + 1044108\ M\^70 + 1453259\ M\^72 - 
                    1549609\ M\^74 - 1170811\ M\^76 + 1678528\ M\^78 + 
                    665413\ M\^80 - 1400092\ M\^82 - 190318\ M\^84 + 
                    908009\ M\^86 - 91020\ M\^88 - 431711\ M\^90 + 
                    156406\ M\^92 + 119742\ M\^94 - 94824\ M\^96 + 
                    4186\ M\^98 + 22376\ M\^100 - 14701\ M\^102 + 
                    4377\ M\^104 + 1784\ M\^106 - 3237\ M\^108 + 
                    2032\ M\^110 - 729\ M\^112 + 157\ M\^114 - 19\ M\^116 + 
                    M\^118)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 32]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[9, 33]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[9, 34]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[9, 35]\), \(1 + L\^7\ M\^34 + 
              L\ \((\(-4\) + 6\ M\^2 + 15\ M\^4 - 19\ M\^6 + 6\ M\^8 + 
                    10\ M\^10 - 14\ M\^12 + 9\ M\^14 - 3\ M\^16 + M\^18)\) + 
              L\^2\ \((6 - 18\ M\^2 - 13\ M\^4 + 73\ M\^6 - 18\ M\^8 - 
                    50\ M\^10 + 56\ M\^12 - 12\ M\^14 - 33\ M\^16 + 
                    39\ M\^18 - 3\ M\^20 - 19\ M\^22 + 22\ M\^24 - 
                    12\ M\^26 + 4\ M\^28 - M\^30)\) + 
              L\^6\ \((M\^16 - 3\ M\^18 + 9\ M\^20 - 14\ M\^22 + 10\ M\^24 + 
                    6\ M\^26 - 19\ M\^28 + 15\ M\^30 + 6\ M\^32 - 
                    4\ M\^34)\) + 
              L\^4\ \((1 - 6\ M\^2 + 19\ M\^4 - 39\ M\^6 + 27\ M\^8 + 
                    35\ M\^10 - 72\ M\^12 + 39\ M\^14 + 92\ M\^16 - 
                    85\ M\^18 - 66\ M\^20 + 62\ M\^22 + 35\ M\^24 + 
                    18\ M\^26 - 19\ M\^28 - 20\ M\^30 + 18\ M\^32 - 
                    4\ M\^34)\) + 
              L\^3\ \((\(-4\) + 18\ M\^2 - 20\ M\^4 - 19\ M\^6 + 18\ M\^8 + 
                    35\ M\^10 + 62\ M\^12 - 66\ M\^14 - 85\ M\^16 + 
                    92\ M\^18 + 39\ M\^20 - 72\ M\^22 + 35\ M\^24 + 
                    27\ M\^26 - 39\ M\^28 + 19\ M\^30 - 6\ M\^32 + M\^34)\) + 
              L\^5\ \((\(-M\^4\) + 4\ M\^6 - 12\ M\^8 + 22\ M\^10 - 
                    19\ M\^12 - 3\ M\^14 + 39\ M\^16 - 33\ M\^18 - 
                    12\ M\^20 + 56\ M\^22 - 50\ M\^24 - 18\ M\^26 + 
                    73\ M\^28 - 13\ M\^30 - 18\ M\^32 + 6\ M\^34)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 36]\), \(M\^8 + L\^22\ M\^184 + 
              L\ \((\(-2\)\ M\^4 + 8\ M\^6 - 19\ M\^8 + 29\ M\^10 - 
                    9\ M\^12 - 30\ M\^14 + 47\ M\^16 + 5\ M\^18 - 17\ M\^20 + 
                    6\ M\^22)\) + 
              L\^2\ \((1 - 8\ M\^2 + 32\ M\^4 - 80\ M\^6 + 125\ M\^8 - 
                    110\ M\^10 + 31\ M\^12 + 88\ M\^14 - 261\ M\^16 + 
                    283\ M\^18 + 192\ M\^20 - 626\ M\^22 + 383\ M\^24 + 
                    473\ M\^26 - 521\ M\^28 + 64\ M\^30 + 156\ M\^32 - 
                    86\ M\^34 + 15\ M\^36)\) + 
              L\^3\ \((\(-1\) + 11\ M\^2 - 59\ M\^4 + 200\ M\^6 - 458\ M\^8 + 
                    672\ M\^10 - 459\ M\^12 - 316\ M\^14 + 1109\ M\^16 - 
                    1050\ M\^18 - 427\ M\^20 + 2719\ M\^22 - 2800\ M\^24 - 
                    1264\ M\^26 + 4970\ M\^28 - 2421\ M\^30 - 2408\ M\^32 + 
                    4912\ M\^34 - 1511\ M\^36 - 2742\ M\^38 + 2973\ M\^40 - 
                    521\ M\^42 - 790\ M\^44 + 596\ M\^46 - 175\ M\^48 + 
                    20\ M\^50)\) + 
              L\^4\ \((\(-2\)\ M\^2 + 24\ M\^4 - 138\ M\^6 + 495\ M\^8 - 
                    1173\ M\^10 + 1756\ M\^12 - 1310\ M\^14 - 184\ M\^16 + 
                    907\ M\^18 - 159\ M\^20 - 188\ M\^22 + 890\ M\^24 - 
                    3122\ M\^26 + 921\ M\^28 + 7049\ M\^30 - 4649\ M\^32 - 
                    11708\ M\^34 + 14321\ M\^36 + 8090\ M\^38 - 
                    19713\ M\^40 + 10192\ M\^42 + 11512\ M\^44 - 
                    19820\ M\^46 + 4908\ M\^48 + 11859\ M\^50 - 
                    11038\ M\^52 + 1737\ M\^54 + 3038\ M\^56 - 2499\ M\^58 + 
                    924\ M\^60 - 180\ M\^62 + 15\ M\^64)\) + 
              L\^5\ \((\(-M\^4\) + 15\ M\^6 - 105\ M\^8 + 453\ M\^10 - 
                    1311\ M\^12 + 2655\ M\^14 - 3963\ M\^16 + 4590\ M\^18 - 
                    3039\ M\^20 - 3493\ M\^22 + 9771\ M\^24 + 2815\ M\^26 - 
                    26595\ M\^28 + 7124\ M\^30 + 45523\ M\^32 - 
                    18837\ M\^34 - 67726\ M\^36 + 39480\ M\^38 + 
                    76311\ M\^40 - 65995\ M\^42 - 55661\ M\^44 + 
                    87817\ M\^46 + 8425\ M\^48 - 65833\ M\^50 + 
                    46787\ M\^52 + 1247\ M\^54 - 44935\ M\^56 + 
                    44694\ M\^58 + 1147\ M\^60 - 33707\ M\^62 + 
                    25412\ M\^64 - 2136\ M\^66 - 8750\ M\^68 + 7010\ M\^70 - 
                    2827\ M\^72 + 681\ M\^74 - 95\ M\^76 + 6\ M\^78)\) + 
              L\^6\ \((M\^8 - 14\ M\^10 + 104\ M\^12 - 510\ M\^14 + 
                    1774\ M\^16 - 4455\ M\^18 + 7861\ M\^20 - 8962\ M\^22 + 
                    5766\ M\^24 - 3304\ M\^26 + 4004\ M\^28 + 11036\ M\^30 - 
                    39119\ M\^32 - 9149\ M\^34 + 147304\ M\^36 - 
                    89003\ M\^38 - 281248\ M\^40 + 312040\ M\^42 + 
                    344522\ M\^44 - 570513\ M\^46 - 249800\ M\^48 + 
                    670012\ M\^50 + 53717\ M\^52 - 507126\ M\^54 + 
                    96604\ M\^56 + 224647\ M\^58 - 79821\ M\^60 - 
                    16337\ M\^62 - 48526\ M\^64 + 8962\ M\^66 + 
                    98974\ M\^68 - 78565\ M\^70 - 30408\ M\^72 + 
                    77507\ M\^74 - 39312\ M\^76 - 7276\ M\^78 + 
                    20749\ M\^80 - 13728\ M\^82 + 5346\ M\^84 - 1360\ M\^86 + 
                    224\ M\^88 - 22\ M\^90 + M\^92)\) + 
              L\^7\ \((M\^14 - 18\ M\^16 + 156\ M\^18 - 865\ M\^20 + 
                    3345\ M\^22 - 9227\ M\^24 + 17660\ M\^26 - 20253\ M\^28 + 
                    4246\ M\^30 + 20842\ M\^32 - 2998\ M\^34 - 70071\ M\^36 + 
                    48668\ M\^38 + 164894\ M\^40 - 183968\ M\^42 - 
                    354752\ M\^44 + 595068\ M\^46 + 488332\ M\^48 - 
                    1324449\ M\^50 - 328118\ M\^52 + 2070026\ M\^54 - 
                    124862\ M\^56 - 2317650\ M\^58 + 493637\ M\^60 + 
                    1917504\ M\^62 - 415081\ M\^64 - 1244281\ M\^66 + 
                    90045\ M\^68 + 747150\ M\^70 + 84459\ M\^72 - 
                    507712\ M\^74 + 43024\ M\^76 + 270827\ M\^78 - 
                    176881\ M\^80 - 10425\ M\^82 + 116665\ M\^84 - 
                    102353\ M\^86 + 23813\ M\^88 + 30844\ M\^90 - 
                    35882\ M\^92 + 19456\ M\^94 - 6606\ M\^96 + 1503\ M\^98 - 
                    230\ M\^100 + 22\ M\^102 - M\^104)\) + 
              L\^8\ \((\(-3\)\ M\^22 + 57\ M\^24 - 507\ M\^26 + 2780\ M\^28 - 
                    10255\ M\^30 + 25877\ M\^32 - 42132\ M\^34 + 
                    30854\ M\^36 + 34708\ M\^38 - 100303\ M\^40 - 
                    2195\ M\^42 + 293967\ M\^44 - 277936\ M\^46 - 
                    482134\ M\^48 + 934486\ M\^50 + 471247\ M\^52 - 
                    2007692\ M\^54 + 70013\ M\^56 + 3131714\ M\^58 - 
                    1193939\ M\^60 - 3699302\ M\^62 + 2307805\ M\^64 + 
                    3569487\ M\^66 - 2564280\ M\^68 - 3217647\ M\^70 + 
                    1901897\ M\^72 + 3053175\ M\^74 - 1140510\ M\^76 - 
                    2707573\ M\^78 + 785696\ M\^80 + 1800972\ M\^82 - 
                    636579\ M\^84 - 651530\ M\^86 + 290527\ M\^88 - 
                    4263\ M\^90 + 89975\ M\^92 + 42404\ M\^94 - 
                    183293\ M\^96 + 102344\ M\^98 + 28332\ M\^100 - 
                    66946\ M\^102 + 43941\ M\^104 - 17548\ M\^106 + 
                    4789\ M\^108 - 884\ M\^110 + 99\ M\^112 - 5\ M\^114)\) + 
              L\^9\ \((3\ M\^30 - 68\ M\^32 + 702\ M\^34 - 4301\ M\^36 + 
                    17117\ M\^38 - 45230\ M\^40 + 74943\ M\^42 - 
                    51780\ M\^44 - 80776\ M\^46 + 233365\ M\^48 - 
                    75911\ M\^50 - 544627\ M\^52 + 793015\ M\^54 + 
                    525906\ M\^56 - 2114065\ M\^58 + 356856\ M\^60 + 
                    3583698\ M\^62 - 2160028\ M\^64 - 4460477\ M\^66 + 
                    4064747\ M\^68 + 4478493\ M\^70 - 5007994\ M\^72 - 
                    4253627\ M\^74 + 5038413\ M\^76 + 4447380\ M\^78 - 
                    5008564\ M\^80 - 4639354\ M\^82 + 5101851\ M\^84 + 
                    3769439\ M\^86 - 4463331\ M\^88 - 2103045\ M\^90 + 
                    2754979\ M\^92 + 811484\ M\^94 - 919397\ M\^96 - 
                    464348\ M\^98 + 58318\ M\^100 + 486973\ M\^102 - 
                    57083\ M\^104 - 322229\ M\^106 + 196616\ M\^108 + 
                    27273\ M\^110 - 99576\ M\^112 + 71216\ M\^114 - 
                    30958\ M\^116 + 9085\ M\^118 - 1752\ M\^120 + 
                    199\ M\^122 - 10\ M\^124)\) + 
              L\^10\ \((\(-M\^38\) + 33\ M\^40 - 446\ M\^42 + 3371\ M\^44 - 
                    15959\ M\^46 + 49314\ M\^48 - 97451\ M\^50 + 
                    100470\ M\^52 + 42141\ M\^54 - 313878\ M\^56 + 
                    339320\ M\^58 + 390110\ M\^60 - 1298199\ M\^62 + 
                    359376\ M\^64 + 2433956\ M\^66 - 2293009\ M\^68 - 
                    3002744\ M\^70 + 4838864\ M\^72 + 2953148\ M\^74 - 
                    7006167\ M\^76 - 3150804\ M\^78 + 8406005\ M\^80 + 
                    4160012\ M\^82 - 9389611\ M\^84 - 4847902\ M\^86 + 
                    9962966\ M\^88 + 3834032\ M\^90 - 9382710\ M\^92 - 
                    1730631\ M\^94 + 7001984\ M\^96 + 383056\ M\^98 - 
                    4002650\ M\^100 - 402710\ M\^102 + 2013722\ M\^104 + 
                    767165\ M\^106 - 1290384\ M\^108 - 525127\ M\^110 + 
                    929318\ M\^112 - 848\ M\^114 - 453963\ M\^116 + 
                    236006\ M\^118 + 39547\ M\^120 - 123259\ M\^122 + 
                    88538\ M\^124 - 38280\ M\^126 + 10865\ M\^128 - 
                    1981\ M\^130 + 211\ M\^132 - 10\ M\^134)\) + 
              L\^11\ \((\(-5\)\ M\^48 + 120\ M\^50 - 1286\ M\^52 + 
                    8009\ M\^54 - 31676\ M\^56 + 81308\ M\^58 - 
                    127451\ M\^60 + 72585\ M\^62 + 177542\ M\^64 - 
                    465677\ M\^66 + 193382\ M\^68 + 905220\ M\^70 - 
                    1267466\ M\^72 - 880714\ M\^74 + 2734848\ M\^76 + 
                    326282\ M\^78 - 4086558\ M\^80 - 25672\ M\^82 + 
                    5745606\ M\^84 + 447039\ M\^86 - 8521369\ M\^88 - 
                    889702\ M\^90 + 11883497\ M\^92 + 453026\ M\^94 - 
                    13495368\ M\^96 + 453026\ M\^98 + 11883497\ M\^100 - 
                    889702\ M\^102 - 8521369\ M\^104 + 447039\ M\^106 + 
                    5745606\ M\^108 - 25672\ M\^110 - 4086558\ M\^112 + 
                    326282\ M\^114 + 2734848\ M\^116 - 880714\ M\^118 - 
                    1267466\ M\^120 + 905220\ M\^122 + 193382\ M\^124 - 
                    465677\ M\^126 + 177542\ M\^128 + 72585\ M\^130 - 
                    127451\ M\^132 + 81308\ M\^134 - 31676\ M\^136 + 
                    8009\ M\^138 - 1286\ M\^140 + 120\ M\^142 - 
                    5\ M\^144)\) + 
              L\^12\ \((\(-10\)\ M\^58 + 211\ M\^60 - 1981\ M\^62 + 
                    10865\ M\^64 - 38280\ M\^66 + 88538\ M\^68 - 
                    123259\ M\^70 + 39547\ M\^72 + 236006\ M\^74 - 
                    453963\ M\^76 - 848\ M\^78 + 929318\ M\^80 - 
                    525127\ M\^82 - 1290384\ M\^84 + 767165\ M\^86 + 
                    2013722\ M\^88 - 402710\ M\^90 - 4002650\ M\^92 + 
                    383056\ M\^94 + 7001984\ M\^96 - 1730631\ M\^98 - 
                    9382710\ M\^100 + 3834032\ M\^102 + 9962966\ M\^104 - 
                    4847902\ M\^106 - 9389611\ M\^108 + 4160012\ M\^110 + 
                    8406005\ M\^112 - 3150804\ M\^114 - 7006167\ M\^116 + 
                    2953148\ M\^118 + 4838864\ M\^120 - 3002744\ M\^122 - 
                    2293009\ M\^124 + 2433956\ M\^126 + 359376\ M\^128 - 
                    1298199\ M\^130 + 390110\ M\^132 + 339320\ M\^134 - 
                    313878\ M\^136 + 42141\ M\^138 + 100470\ M\^140 - 
                    97451\ M\^142 + 49314\ M\^144 - 15959\ M\^146 + 
                    3371\ M\^148 - 446\ M\^150 + 33\ M\^152 - M\^154)\) + 
              L\^13\ \((\(-10\)\ M\^68 + 199\ M\^70 - 1752\ M\^72 + 
                    9085\ M\^74 - 30958\ M\^76 + 71216\ M\^78 - 
                    99576\ M\^80 + 27273\ M\^82 + 196616\ M\^84 - 
                    322229\ M\^86 - 57083\ M\^88 + 486973\ M\^90 + 
                    58318\ M\^92 - 464348\ M\^94 - 919397\ M\^96 + 
                    811484\ M\^98 + 2754979\ M\^100 - 2103045\ M\^102 - 
                    4463331\ M\^104 + 3769439\ M\^106 + 5101851\ M\^108 - 
                    4639354\ M\^110 - 5008564\ M\^112 + 4447380\ M\^114 + 
                    5038413\ M\^116 - 4253627\ M\^118 - 5007994\ M\^120 + 
                    4478493\ M\^122 + 4064747\ M\^124 - 4460477\ M\^126 - 
                    2160028\ M\^128 + 3583698\ M\^130 + 356856\ M\^132 - 
                    2114065\ M\^134 + 525906\ M\^136 + 793015\ M\^138 - 
                    544627\ M\^140 - 75911\ M\^142 + 233365\ M\^144 - 
                    80776\ M\^146 - 51780\ M\^148 + 74943\ M\^150 - 
                    45230\ M\^152 + 17117\ M\^154 - 4301\ M\^156 + 
                    702\ M\^158 - 68\ M\^160 + 3\ M\^162)\) + 
              L\^14\ \((\(-5\)\ M\^78 + 99\ M\^80 - 884\ M\^82 + 
                    4789\ M\^84 - 17548\ M\^86 + 43941\ M\^88 - 
                    66946\ M\^90 + 28332\ M\^92 + 102344\ M\^94 - 
                    183293\ M\^96 + 42404\ M\^98 + 89975\ M\^100 - 
                    4263\ M\^102 + 290527\ M\^104 - 651530\ M\^106 - 
                    636579\ M\^108 + 1800972\ M\^110 + 785696\ M\^112 - 
                    2707573\ M\^114 - 1140510\ M\^116 + 3053175\ M\^118 + 
                    1901897\ M\^120 - 3217647\ M\^122 - 2564280\ M\^124 + 
                    3569487\ M\^126 + 2307805\ M\^128 - 3699302\ M\^130 - 
                    1193939\ M\^132 + 3131714\ M\^134 + 70013\ M\^136 - 
                    2007692\ M\^138 + 471247\ M\^140 + 934486\ M\^142 - 
                    482134\ M\^144 - 277936\ M\^146 + 293967\ M\^148 - 
                    2195\ M\^150 - 100303\ M\^152 + 34708\ M\^154 + 
                    30854\ M\^156 - 42132\ M\^158 + 25877\ M\^160 - 
                    10255\ M\^162 + 2780\ M\^164 - 507\ M\^166 + 57\ M\^168 - 
                    3\ M\^170)\) + 
              L\^15\ \((\(-M\^88\) + 22\ M\^90 - 230\ M\^92 + 1503\ M\^94 - 
                    6606\ M\^96 + 19456\ M\^98 - 35882\ M\^100 + 
                    30844\ M\^102 + 23813\ M\^104 - 102353\ M\^106 + 
                    116665\ M\^108 - 10425\ M\^110 - 176881\ M\^112 + 
                    270827\ M\^114 + 43024\ M\^116 - 507712\ M\^118 + 
                    84459\ M\^120 + 747150\ M\^122 + 90045\ M\^124 - 
                    1244281\ M\^126 - 415081\ M\^128 + 1917504\ M\^130 + 
                    493637\ M\^132 - 2317650\ M\^134 - 124862\ M\^136 + 
                    2070026\ M\^138 - 328118\ M\^140 - 1324449\ M\^142 + 
                    488332\ M\^144 + 595068\ M\^146 - 354752\ M\^148 - 
                    183968\ M\^150 + 164894\ M\^152 + 48668\ M\^154 - 
                    70071\ M\^156 - 2998\ M\^158 + 20842\ M\^160 + 
                    4246\ M\^162 - 20253\ M\^164 + 17660\ M\^166 - 
                    9227\ M\^168 + 3345\ M\^170 - 865\ M\^172 + 156\ M\^174 - 
                    18\ M\^176 + M\^178)\) + 
              L\^16\ \((M\^100 - 22\ M\^102 + 224\ M\^104 - 1360\ M\^106 + 
                    5346\ M\^108 - 13728\ M\^110 + 20749\ M\^112 - 
                    7276\ M\^114 - 39312\ M\^116 + 77507\ M\^118 - 
                    30408\ M\^120 - 78565\ M\^122 + 98974\ M\^124 + 
                    8962\ M\^126 - 48526\ M\^128 - 16337\ M\^130 - 
                    79821\ M\^132 + 224647\ M\^134 + 96604\ M\^136 - 
                    507126\ M\^138 + 53717\ M\^140 + 670012\ M\^142 - 
                    249800\ M\^144 - 570513\ M\^146 + 344522\ M\^148 + 
                    312040\ M\^150 - 281248\ M\^152 - 89003\ M\^154 + 
                    147304\ M\^156 - 9149\ M\^158 - 39119\ M\^160 + 
                    11036\ M\^162 + 4004\ M\^164 - 3304\ M\^166 + 
                    5766\ M\^168 - 8962\ M\^170 + 7861\ M\^172 - 
                    4455\ M\^174 + 1774\ M\^176 - 510\ M\^178 + 104\ M\^180 - 
                    14\ M\^182 + M\^184)\) + 
              L\^21\ \((6\ M\^170 - 17\ M\^172 + 5\ M\^174 + 47\ M\^176 - 
                    30\ M\^178 - 9\ M\^180 + 29\ M\^182 - 19\ M\^184 + 
                    8\ M\^186 - 2\ M\^188)\) + 
              L\^17\ \((6\ M\^114 - 95\ M\^116 + 681\ M\^118 - 2827\ M\^120 + 
                    7010\ M\^122 - 8750\ M\^124 - 2136\ M\^126 + 
                    25412\ M\^128 - 33707\ M\^130 + 1147\ M\^132 + 
                    44694\ M\^134 - 44935\ M\^136 + 1247\ M\^138 + 
                    46787\ M\^140 - 65833\ M\^142 + 8425\ M\^144 + 
                    87817\ M\^146 - 55661\ M\^148 - 65995\ M\^150 + 
                    76311\ M\^152 + 39480\ M\^154 - 67726\ M\^156 - 
                    18837\ M\^158 + 45523\ M\^160 + 7124\ M\^162 - 
                    26595\ M\^164 + 2815\ M\^166 + 9771\ M\^168 - 
                    3493\ M\^170 - 3039\ M\^172 + 4590\ M\^174 - 
                    3963\ M\^176 + 2655\ M\^178 - 1311\ M\^180 + 
                    453\ M\^182 - 105\ M\^184 + 15\ M\^186 - M\^188)\) + 
              L\^18\ \((15\ M\^128 - 180\ M\^130 + 924\ M\^132 - 
                    2499\ M\^134 + 3038\ M\^136 + 1737\ M\^138 - 
                    11038\ M\^140 + 11859\ M\^142 + 4908\ M\^144 - 
                    19820\ M\^146 + 11512\ M\^148 + 10192\ M\^150 - 
                    19713\ M\^152 + 8090\ M\^154 + 14321\ M\^156 - 
                    11708\ M\^158 - 4649\ M\^160 + 7049\ M\^162 + 
                    921\ M\^164 - 3122\ M\^166 + 890\ M\^168 - 188\ M\^170 - 
                    159\ M\^172 + 907\ M\^174 - 184\ M\^176 - 1310\ M\^178 + 
                    1756\ M\^180 - 1173\ M\^182 + 495\ M\^184 - 138\ M\^186 + 
                    24\ M\^188 - 2\ M\^190)\) + 
              L\^19\ \((20\ M\^142 - 175\ M\^144 + 596\ M\^146 - 
                    790\ M\^148 - 521\ M\^150 + 2973\ M\^152 - 2742\ M\^154 - 
                    1511\ M\^156 + 4912\ M\^158 - 2408\ M\^160 - 
                    2421\ M\^162 + 4970\ M\^164 - 1264\ M\^166 - 
                    2800\ M\^168 + 2719\ M\^170 - 427\ M\^172 - 
                    1050\ M\^174 + 1109\ M\^176 - 316\ M\^178 - 459\ M\^180 + 
                    672\ M\^182 - 458\ M\^184 + 200\ M\^186 - 59\ M\^188 + 
                    11\ M\^190 - M\^192)\) + 
              L\^20\ \((15\ M\^156 - 86\ M\^158 + 156\ M\^160 + 64\ M\^162 - 
                    521\ M\^164 + 473\ M\^166 + 383\ M\^168 - 626\ M\^170 + 
                    192\ M\^172 + 283\ M\^174 - 261\ M\^176 + 88\ M\^178 + 
                    31\ M\^180 - 110\ M\^182 + 125\ M\^184 - 80\ M\^186 + 
                    32\ M\^188 - 8\ M\^190 + M\^192)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 37]\), \(\(-M\^42\) + 
              L\^14\ M\^44 + 
              L\ \((\(-3\)\ M\^32 + 14\ M\^34 - 33\ M\^36 + 41\ M\^38 + 
                    10\ M\^40 - 73\ M\^42 + 28\ M\^44 + 15\ M\^46 - 
                    15\ M\^48 + 4\ M\^50)\) + 
              L\^13\ \((\(-4\)\ M\^36 + 15\ M\^38 - 15\ M\^40 - 28\ M\^42 + 
                    73\ M\^44 - 10\ M\^46 - 41\ M\^48 + 33\ M\^50 - 
                    14\ M\^52 + 3\ M\^54)\) + 
              L\^2\ \((\(-3\)\ M\^22 + 24\ M\^24 - 86\ M\^26 + 181\ M\^28 - 
                    198\ M\^30 + 57\ M\^32 + 148\ M\^34 - 523\ M\^36 + 
                    787\ M\^38 + 159\ M\^40 - 1321\ M\^42 + 649\ M\^44 + 
                    491\ M\^46 - 683\ M\^48 + 235\ M\^50 + 104\ M\^52 - 
                    125\ M\^54 + 45\ M\^56 - 6\ M\^58)\) + 
              L\^12\ \((6\ M\^28 - 45\ M\^30 + 125\ M\^32 - 104\ M\^34 - 
                    235\ M\^36 + 683\ M\^38 - 491\ M\^40 - 649\ M\^42 + 
                    1321\ M\^44 - 159\ M\^46 - 787\ M\^48 + 523\ M\^50 - 
                    148\ M\^52 - 57\ M\^54 + 198\ M\^56 - 181\ M\^58 + 
                    86\ M\^60 - 24\ M\^62 + 3\ M\^64)\) + 
              L\^3\ \((\(-M\^12\) + 10\ M\^14 - 45\ M\^16 + 123\ M\^18 - 
                    244\ M\^20 + 470\ M\^22 - 930\ M\^24 + 976\ M\^26 + 
                    995\ M\^28 - 4536\ M\^30 + 5189\ M\^32 + 2250\ M\^34 - 
                    13688\ M\^36 + 10221\ M\^38 + 11805\ M\^40 - 
                    19320\ M\^42 - 51\ M\^44 + 13131\ M\^46 - 6980\ M\^48 - 
                    1815\ M\^50 + 3874\ M\^52 - 2146\ M\^54 + 320\ M\^56 + 
                    502\ M\^58 - 485\ M\^60 + 208\ M\^62 - 45\ M\^64 + 
                    4\ M\^66)\) + 
              L\^4\ \((3\ M\^8 - 40\ M\^10 + 240\ M\^12 - 799\ M\^14 + 
                    1445\ M\^16 - 904\ M\^18 - 1598\ M\^20 + 3784\ M\^22 - 
                    1710\ M\^24 - 3500\ M\^26 + 3410\ M\^28 - 130\ M\^30 + 
                    6554\ M\^32 - 2910\ M\^34 - 25147\ M\^36 + 18139\ M\^38 + 
                    32483\ M\^40 - 33654\ M\^42 - 11934\ M\^44 + 
                    27677\ M\^46 - 17904\ M\^48 - 1540\ M\^50 + 
                    21265\ M\^52 - 16016\ M\^54 - 4164\ M\^56 + 
                    11029\ M\^58 - 4851\ M\^60 - 676\ M\^62 + 1774\ M\^64 - 
                    1073\ M\^66 + 405\ M\^68 - 101\ M\^70 + 15\ M\^72 - 
                    M\^74)\) + 
              L\^11\ \((\(-4\)\ M\^20 + 45\ M\^22 - 208\ M\^24 + 485\ M\^26 - 
                    502\ M\^28 - 320\ M\^30 + 2146\ M\^32 - 3874\ M\^34 + 
                    1815\ M\^36 + 6980\ M\^38 - 13131\ M\^40 + 51\ M\^42 + 
                    19320\ M\^44 - 11805\ M\^46 - 10221\ M\^48 + 
                    13688\ M\^50 - 2250\ M\^52 - 5189\ M\^54 + 4536\ M\^56 - 
                    995\ M\^58 - 976\ M\^60 + 930\ M\^62 - 470\ M\^64 + 
                    244\ M\^66 - 123\ M\^68 + 45\ M\^70 - 10\ M\^72 + 
                    M\^74)\) + 
              L\^10\ \((M\^12 - 15\ M\^14 + 101\ M\^16 - 405\ M\^18 + 
                    1073\ M\^20 - 1774\ M\^22 + 676\ M\^24 + 4851\ M\^26 - 
                    11029\ M\^28 + 4164\ M\^30 + 16016\ M\^32 - 
                    21265\ M\^34 + 1540\ M\^36 + 17904\ M\^38 - 
                    27677\ M\^40 + 11934\ M\^42 + 33654\ M\^44 - 
                    32483\ M\^46 - 18139\ M\^48 + 25147\ M\^50 + 
                    2910\ M\^52 - 6554\ M\^54 + 130\ M\^56 - 3410\ M\^58 + 
                    3500\ M\^60 + 1710\ M\^62 - 3784\ M\^64 + 1598\ M\^66 + 
                    904\ M\^68 - 1445\ M\^70 + 799\ M\^72 - 240\ M\^74 + 
                    40\ M\^76 - 3\ M\^78)\) + 
              L\^5\ \((\(-3\)\ M\^4 + 50\ M\^6 - 355\ M\^8 + 1337\ M\^10 - 
                    2616\ M\^12 + 1342\ M\^14 + 5121\ M\^16 - 10298\ M\^18 - 
                    668\ M\^20 + 23262\ M\^22 - 17328\ M\^24 - 26187\ M\^26 + 
                    37127\ M\^28 + 7085\ M\^30 - 33273\ M\^32 + 
                    25725\ M\^34 + 1608\ M\^36 - 44115\ M\^38 + 
                    17503\ M\^40 + 39820\ M\^42 + 5442\ M\^44 - 
                    29328\ M\^46 - 41376\ M\^48 + 25870\ M\^50 + 
                    45938\ M\^52 - 30182\ M\^54 - 17556\ M\^56 + 
                    24241\ M\^58 - 8914\ M\^60 - 5986\ M\^62 + 11195\ M\^64 - 
                    5254\ M\^66 - 1605\ M\^68 + 2949\ M\^70 - 1475\ M\^72 + 
                    381\ M\^74 - 52\ M\^76 + 3\ M\^78)\) + 
              L\^6\ \((1 - 20\ M\^2 + 161\ M\^4 - 676\ M\^6 + 1531\ M\^8 - 
                    1435\ M\^10 - 1144\ M\^12 + 3320\ M\^14 + 1620\ M\^16 - 
                    8362\ M\^18 - 2870\ M\^20 + 19952\ M\^22 - 10\ M\^24 - 
                    26634\ M\^26 + 8431\ M\^28 - 534\ M\^30 - 8328\ M\^32 + 
                    58609\ M\^34 - 4463\ M\^36 - 91970\ M\^38 + 
                    13301\ M\^40 + 52596\ M\^42 - 1341\ M\^44 + 
                    27434\ M\^46 - 16570\ M\^48 - 68211\ M\^50 + 
                    9830\ M\^52 + 49683\ M\^54 + 12257\ M\^56 - 
                    16632\ M\^58 - 26790\ M\^60 + 3153\ M\^62 + 
                    25922\ M\^64 - 6111\ M\^66 - 13886\ M\^68 + 8249\ M\^70 + 
                    1825\ M\^72 - 3697\ M\^74 + 1755\ M\^76 - 427\ M\^78 + 
                    55\ M\^80 - 3\ M\^82)\) + 
              L\^9\ \((\(-3\)\ M\^8 + 52\ M\^10 - 381\ M\^12 + 1475\ M\^14 - 
                    2949\ M\^16 + 1605\ M\^18 + 5254\ M\^20 - 11195\ M\^22 + 
                    5986\ M\^24 + 8914\ M\^26 - 24241\ M\^28 + 17556\ M\^30 + 
                    30182\ M\^32 - 45938\ M\^34 - 25870\ M\^36 + 
                    41376\ M\^38 + 29328\ M\^40 - 5442\ M\^42 - 
                    39820\ M\^44 - 17503\ M\^46 + 44115\ M\^48 - 
                    1608\ M\^50 - 25725\ M\^52 + 33273\ M\^54 - 7085\ M\^56 - 
                    37127\ M\^58 + 26187\ M\^60 + 17328\ M\^62 - 
                    23262\ M\^64 + 668\ M\^66 + 10298\ M\^68 - 5121\ M\^70 - 
                    1342\ M\^72 + 2616\ M\^74 - 1337\ M\^76 + 355\ M\^78 - 
                    50\ M\^80 + 3\ M\^82)\) + 
              L\^8\ \((3\ M\^4 - 55\ M\^6 + 427\ M\^8 - 1755\ M\^10 + 
                    3697\ M\^12 - 1825\ M\^14 - 8249\ M\^16 + 13886\ M\^18 + 
                    6111\ M\^20 - 25922\ M\^22 - 3153\ M\^24 + 26790\ M\^26 + 
                    16632\ M\^28 - 12257\ M\^30 - 49683\ M\^32 - 
                    9830\ M\^34 + 68211\ M\^36 + 16570\ M\^38 - 
                    27434\ M\^40 + 1341\ M\^42 - 52596\ M\^44 - 
                    13301\ M\^46 + 91970\ M\^48 + 4463\ M\^50 - 
                    58609\ M\^52 + 8328\ M\^54 + 534\ M\^56 - 8431\ M\^58 + 
                    26634\ M\^60 + 10\ M\^62 - 19952\ M\^64 + 2870\ M\^66 + 
                    8362\ M\^68 - 1620\ M\^70 - 3320\ M\^72 + 1144\ M\^74 + 
                    1435\ M\^76 - 1531\ M\^78 + 676\ M\^80 - 161\ M\^82 + 
                    20\ M\^84 - M\^86)\) + 
              L\^7\ \((\(-1\) + 18\ M\^2 - 140\ M\^4 + 613\ M\^6 - 
                    1630\ M\^8 + 2457\ M\^10 - 688\ M\^12 - 5297\ M\^14 + 
                    8418\ M\^16 + 5174\ M\^18 - 21540\ M\^20 - 1817\ M\^22 + 
                    34387\ M\^24 - 173\ M\^26 - 27231\ M\^28 - 9396\ M\^30 - 
                    17921\ M\^32 + 31530\ M\^34 + 68494\ M\^36 - 
                    32198\ M\^38 - 66024\ M\^40 - 13071\ M\^42 + 
                    13071\ M\^44 + 66024\ M\^46 + 32198\ M\^48 - 
                    68494\ M\^50 - 31530\ M\^52 + 17921\ M\^54 + 
                    9396\ M\^56 + 27231\ M\^58 + 173\ M\^60 - 34387\ M\^62 + 
                    1817\ M\^64 + 21540\ M\^66 - 5174\ M\^68 - 8418\ M\^70 + 
                    5297\ M\^72 + 688\ M\^74 - 2457\ M\^76 + 1630\ M\^78 - 
                    613\ M\^80 + 140\ M\^82 - 18\ M\^84 + M\^86)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 38]\), \(\(-M\^176\) + 
              2\ M\^180 - M\^184 + L\^21\ \((1 - 2\ M\^4 + M\^8)\) + 
              L\^20\ \((\(-5\) + 18\ M\^2 - 21\ M\^4 - 72\ M\^6 + 204\ M\^8 + 
                    16\ M\^10 - 346\ M\^12 + 144\ M\^14 + 176\ M\^16 - 
                    133\ M\^18 - 5\ M\^20 + 36\ M\^22 - 7\ M\^24 - 8\ M\^26 + 
                    M\^28 + 3\ M\^30 - M\^32)\) + 
              L\^19\ \((10 - 73\ M\^2 + 230\ M\^4 - 228\ M\^6 - 524\ M\^8 + 
                    2022\ M\^10 - 1920\ M\^12 - 4067\ M\^14 + 9416\ M\^16 + 
                    1833\ M\^18 - 15717\ M\^20 + 4926\ M\^22 + 12233\ M\^24 - 
                    8716\ M\^26 - 3199\ M\^28 + 5775\ M\^30 - 1381\ M\^32 - 
                    1735\ M\^34 + 1371\ M\^36 + 32\ M\^38 - 493\ M\^40 + 
                    217\ M\^42 + 28\ M\^44 - 59\ M\^46 + 22\ M\^48 - 
                    3\ M\^50)\) + 
              L\^18\ \((\(-10\) + 112\ M\^2 - 566\ M\^4 + 1551\ M\^6 - 
                    2103\ M\^8 - 757\ M\^10 + 8394\ M\^12 - 12561\ M\^14 - 
                    2529\ M\^16 + 43017\ M\^18 - 51976\ M\^20 - 
                    74121\ M\^22 + 177461\ M\^24 + 62317\ M\^26 - 
                    307054\ M\^28 + 15846\ M\^30 + 320132\ M\^32 - 
                    106102\ M\^34 - 198683\ M\^36 + 123108\ M\^38 + 
                    58747\ M\^40 - 66445\ M\^42 + 2298\ M\^44 + 
                    13406\ M\^46 - 4319\ M\^48 + 2994\ M\^50 - 2426\ M\^52 - 
                    1274\ M\^54 + 2498\ M\^56 - 886\ M\^58 - 369\ M\^60 + 
                    444\ M\^62 - 176\ M\^64 + 35\ M\^66 - 3\ M\^68)\) + 
              L\^17\ \((5 - 78\ M\^2 + 558\ M\^4 - 2361\ M\^6 + 6321\ M\^8 - 
                    9726\ M\^10 + 2471\ M\^12 + 23786\ M\^14 - 41923\ M\^16 - 
                    9552\ M\^18 + 90126\ M\^20 - 45653\ M\^22 - 
                    80313\ M\^24 + 191588\ M\^26 - 192466\ M\^28 - 
                    493957\ M\^30 + 921843\ M\^32 + 914574\ M\^34 - 
                    1897005\ M\^36 - 1210780\ M\^38 + 2526980\ M\^40 + 
                    1122386\ M\^42 - 2416054\ M\^44 - 731602\ M\^46 + 
                    1745826\ M\^48 + 357117\ M\^50 - 1013501\ M\^52 - 
                    141929\ M\^54 + 527160\ M\^56 + 12304\ M\^58 - 
                    244652\ M\^60 + 50195\ M\^62 + 80199\ M\^64 - 
                    44828\ M\^66 - 7898\ M\^68 + 16965\ M\^70 - 6668\ M\^72 - 
                    761\ M\^74 + 2328\ M\^76 - 1444\ M\^78 + 523\ M\^80 - 
                    119\ M\^82 + 16\ M\^84 - M\^86)\) + 
              L\^16\ \((\(-1\) + 22\ M\^2 - 221\ M\^4 + 1345\ M\^6 - 
                    5460\ M\^8 + 14996\ M\^10 - 25783\ M\^12 + 16802\ M\^14 + 
                    36555\ M\^16 - 98898\ M\^18 + 26717\ M\^20 + 
                    231732\ M\^22 - 171948\ M\^24 - 612219\ M\^26 + 
                    444878\ M\^28 + 1564994\ M\^30 - 613026\ M\^32 - 
                    3204878\ M\^34 - 87605\ M\^36 + 4698332\ M\^38 + 
                    2428696\ M\^40 - 4605974\ M\^42 - 5863240\ M\^44 + 
                    2847538\ M\^46 + 8332883\ M\^48 - 1111078\ M\^50 - 
                    8273009\ M\^52 + 709076\ M\^54 + 6042613\ M\^56 - 
                    1270456\ M\^58 - 3144933\ M\^60 + 1712629\ M\^62 + 
                    858874\ M\^64 - 1508171\ M\^66 + 318244\ M\^68 + 
                    931440\ M\^70 - 642243\ M\^72 - 349693\ M\^74 + 
                    504989\ M\^76 - 6138\ M\^78 - 228935\ M\^80 + 
                    90548\ M\^82 + 46221\ M\^84 - 48023\ M\^86 + 
                    7595\ M\^88 + 9124\ M\^90 - 6873\ M\^92 + 2397\ M\^94 - 
                    489\ M\^96 + 57\ M\^98 - 3\ M\^100)\) + 
              L\^15\ \((\(-M\^2\) + 22\ M\^4 - 239\ M\^6 + 1653\ M\^8 - 
                    7710\ M\^10 + 23584\ M\^12 - 41645\ M\^14 + 
                    22347\ M\^16 + 46930\ M\^18 - 37403\ M\^20 - 
                    85988\ M\^22 - 83320\ M\^24 + 331041\ M\^26 + 
                    561462\ M\^28 - 854818\ M\^30 - 1873391\ M\^32 + 
                    1057289\ M\^34 + 4216755\ M\^36 + 1255287\ M\^38 - 
                    6746720\ M\^40 - 7703458\ M\^42 + 6581005\ M\^44 + 
                    15069595\ M\^46 - 1093839\ M\^48 - 15884272\ M\^50 - 
                    7152595\ M\^52 + 6457518\ M\^54 + 11907280\ M\^56 + 
                    7365800\ M\^58 - 10236707\ M\^60 - 16355262\ M\^62 + 
                    4726216\ M\^64 + 17141660\ M\^66 + 419623\ M\^68 - 
                    12706641\ M\^70 - 3129871\ M\^72 + 7374437\ M\^74 + 
                    3797967\ M\^76 - 3766640\ M\^78 - 3246283\ M\^80 + 
                    2065219\ M\^82 + 2164458\ M\^84 - 1394580\ M\^86 - 
                    1033062\ M\^88 + 953828\ M\^90 + 220926\ M\^92 - 
                    472101\ M\^94 + 86376\ M\^96 + 122488\ M\^98 - 
                    73910\ M\^100 - 2211\ M\^102 + 20466\ M\^104 - 
                    11420\ M\^106 + 3409\ M\^108 - 615\ M\^110 + 64\ M\^112 - 
                    3\ M\^114)\) + 
              L\^14\ \((4\ M\^6 - 92\ M\^8 + 940\ M\^10 - 5571\ M\^12 + 
                    20729\ M\^14 - 47128\ M\^16 + 49370\ M\^18 + 
                    37109\ M\^20 - 133170\ M\^22 - 109665\ M\^24 + 
                    570037\ M\^26 + 132028\ M\^28 - 1749661\ M\^30 - 
                    203298\ M\^32 + 4332949\ M\^34 + 875732\ M\^36 - 
                    8482907\ M\^38 - 2456949\ M\^40 + 10610970\ M\^42 + 
                    3401979\ M\^44 - 2688072\ M\^46 + 4193679\ M\^48 - 
                    20904346\ M\^50 - 33395939\ M\^52 + 50465560\ M\^54 + 
                    87180794\ M\^56 - 61336283\ M\^58 - 144461415\ M\^60 + 
                    39774195\ M\^62 + 172221235\ M\^64 - 1110407\ M\^66 - 
                    156471387\ M\^68 - 27254372\ M\^70 + 112468494\ M\^72 + 
                    32573053\ M\^74 - 65088836\ M\^76 - 21756134\ M\^78 + 
                    29486602\ M\^80 + 8163765\ M\^82 - 8499228\ M\^84 + 
                    336754\ M\^86 - 1617634\ M\^88 - 2500716\ M\^90 + 
                    4832284\ M\^92 + 1160276\ M\^94 - 4199836\ M\^96 + 
                    482266\ M\^98 + 2234101\ M\^100 - 1015629\ M\^102 - 
                    608788\ M\^104 + 646432\ M\^106 - 64123\ M\^108 - 
                    170990\ M\^110 + 93125\ M\^112 + 4141\ M\^114 - 
                    28990\ M\^116 + 18190\ M\^118 - 6566\ M\^120 + 
                    1561\ M\^122 - 244\ M\^124 + 23\ M\^126 - M\^128)\) + 
              L\^13\ \((\(-6\)\ M\^10 + 146\ M\^12 - 1523\ M\^14 + 
                    8824\ M\^16 - 30063\ M\^18 + 54708\ M\^20 - 
                    19066\ M\^22 - 108689\ M\^24 + 72499\ M\^26 + 
                    451253\ M\^28 - 663084\ M\^30 - 887741\ M\^32 + 
                    1935351\ M\^34 + 1957702\ M\^36 - 3897547\ M\^38 - 
                    5837263\ M\^40 + 5359309\ M\^42 + 17485947\ M\^44 - 
                    1176930\ M\^46 - 43153729\ M\^48 - 19285582\ M\^50 + 
                    79469515\ M\^52 + 68376169\ M\^54 - 102506939\ M\^56 - 
                    140532847\ M\^58 + 78022126\ M\^60 + 192402019\ M\^62 - 
                    1391535\ M\^64 - 166568363\ M\^66 - 73987195\ M\^68 + 
                    53638183\ M\^70 + 79840041\ M\^72 + 83640211\ M\^74 - 
                    6534907\ M\^76 - 168016454\ M\^78 - 90252463\ M\^80 + 
                    175127810\ M\^82 + 149515069\ M\^84 - 133399295\ M\^86 - 
                    152513808\ M\^88 + 81824746\ M\^90 + 118976429\ M\^92 - 
                    42769829\ M\^94 - 76802085\ M\^96 + 19982020\ M\^98 + 
                    43279222\ M\^100 - 9295599\ M\^102 - 22063948\ M\^104 + 
                    5194861\ M\^106 + 10054638\ M\^108 - 3427512\ M\^110 - 
                    3784769\ M\^112 + 2132272\ M\^114 + 913755\ M\^116 - 
                    1013948\ M\^118 + 51394\ M\^120 + 259412\ M\^122 - 
                    100410\ M\^124 - 29389\ M\^126 + 41358\ M\^128 - 
                    18599\ M\^130 + 4834\ M\^132 - 775\ M\^134 + 72\ M\^136 - 
                    3\ M\^138)\) + 
              L\^12\ \((4\ M\^14 - 99\ M\^16 + 1057\ M\^18 - 6293\ M\^20 + 
                    22091\ M\^22 - 41741\ M\^24 + 18117\ M\^26 + 
                    72519\ M\^28 - 31936\ M\^30 - 372309\ M\^32 + 
                    489306\ M\^34 + 685726\ M\^36 - 1270523\ M\^38 - 
                    1527472\ M\^40 + 2311944\ M\^42 + 3673919\ M\^44 - 
                    1472476\ M\^46 - 8726848\ M\^48 - 8446341\ M\^50 + 
                    17183159\ M\^52 + 41612587\ M\^54 - 18675480\ M\^56 - 
                    115741875\ M\^58 - 19312720\ M\^60 + 229515653\ M\^62 + 
                    147670452\ M\^64 - 330121859\ M\^66 - 390433967\ M\^68 + 
                    317255455\ M\^70 + 676511042\ M\^72 - 122595202\ M\^74 - 
                    846815702\ M\^76 - 188613049\ M\^78 + 780447623\ M\^80 + 
                    439361990\ M\^82 - 516386925\ M\^84 - 495997539\ M\^86 + 
                    215361589\ M\^88 + 374372362\ M\^90 - 11792587\ M\^92 - 
                    189953033\ M\^94 - 67089309\ M\^96 + 44118003\ M\^98 + 
                    67339887\ M\^100 + 30403874\ M\^102 - 43530309\ M\^104 - 
                    49252186\ M\^106 + 23965401\ M\^108 + 40761713\ M\^110 - 
                    13873492\ M\^112 - 25229178\ M\^114 + 9217172\ M\^116 + 
                    12105907\ M\^118 - 5973954\ M\^120 - 4249720\ M\^122 + 
                    3221683\ M\^124 + 831323\ M\^126 - 1326397\ M\^128 + 
                    164263\ M\^130 + 288550\ M\^132 - 122166\ M\^134 - 
                    33215\ M\^136 + 48274\ M\^138 - 21389\ M\^140 + 
                    5413\ M\^142 - 839\ M\^144 + 75\ M\^146 - 3\ M\^148)\) + 
              L\^11\ \((\(-M\^18\) + 24\ M\^20 - 258\ M\^22 + 1622\ M\^24 - 
                    6533\ M\^26 + 17370\ M\^28 - 30251\ M\^30 + 
                    31907\ M\^32 - 7183\ M\^34 - 53739\ M\^36 + 
                    76486\ M\^38 + 274431\ M\^40 - 1007895\ M\^42 + 
                    188186\ M\^44 + 3482690\ M\^46 - 2567060\ M\^48 - 
                    9511350\ M\^50 + 9358286\ M\^52 + 22708737\ M\^54 - 
                    22312157\ M\^56 - 49225903\ M\^58 + 36791156\ M\^60 + 
                    96786432\ M\^62 - 35426139\ M\^64 - 162477144\ M\^66 - 
                    13819096\ M\^68 + 212368995\ M\^70 + 134988748\ M\^72 - 
                    172623725\ M\^74 - 293273409\ M\^76 - 29950622\ M\^78 + 
                    368352925\ M\^80 + 370787532\ M\^82 - 224071374\ M\^84 - 
                    681373570\ M\^86 - 140959166\ M\^88 + 761975453\ M\^90 + 
                    541528711\ M\^92 - 569068879\ M\^94 - 758042306\ M\^96 + 
                    252558451\ M\^98 + 721454098\ M\^100 - 165071\ M\^102 - 
                    526974143\ M\^104 - 109212374\ M\^106 + 
                    310900590\ M\^108 + 108760916\ M\^110 - 
                    151091529\ M\^112 - 67597731\ M\^114 + 59906374\ M\^116 + 
                    30802037\ M\^118 - 17861774\ M\^120 - 10827679\ M\^122 + 
                    2400380\ M\^124 + 3422699\ M\^126 + 1256340\ M\^128 - 
                    1352711\ M\^130 - 1122886\ M\^132 + 725837\ M\^134 + 
                    486645\ M\^136 - 449612\ M\^138 - 12283\ M\^140 + 
                    112092\ M\^142 - 11858\ M\^144 - 38709\ M\^146 + 
                    28287\ M\^148 - 10339\ M\^150 + 2324\ M\^152 - 
                    328\ M\^154 + 27\ M\^156 - M\^158)\) + 
              L\^10\ \((M\^26 - 27\ M\^28 + 328\ M\^30 - 2324\ M\^32 + 
                    10339\ M\^34 - 28287\ M\^36 + 38709\ M\^38 + 
                    11858\ M\^40 - 112092\ M\^42 + 12283\ M\^44 + 
                    449612\ M\^46 - 486645\ M\^48 - 725837\ M\^50 + 
                    1122886\ M\^52 + 1352711\ M\^54 - 1256340\ M\^56 - 
                    3422699\ M\^58 - 2400380\ M\^60 + 10827679\ M\^62 + 
                    17861774\ M\^64 - 30802037\ M\^66 - 59906374\ M\^68 + 
                    67597731\ M\^70 + 151091529\ M\^72 - 108760916\ M\^74 - 
                    310900590\ M\^76 + 109212374\ M\^78 + 526974143\ M\^80 + 
                    165071\ M\^82 - 721454098\ M\^84 - 252558451\ M\^86 + 
                    758042306\ M\^88 + 569068879\ M\^90 - 541528711\ M\^92 - 
                    761975453\ M\^94 + 140959166\ M\^96 + 681373570\ M\^98 + 
                    224071374\ M\^100 - 370787532\ M\^102 - 
                    368352925\ M\^104 + 29950622\ M\^106 + 
                    293273409\ M\^108 + 172623725\ M\^110 - 
                    134988748\ M\^112 - 212368995\ M\^114 + 
                    13819096\ M\^116 + 162477144\ M\^118 + 35426139\ M\^120 - 
                    96786432\ M\^122 - 36791156\ M\^124 + 49225903\ M\^126 + 
                    22312157\ M\^128 - 22708737\ M\^130 - 9358286\ M\^132 + 
                    9511350\ M\^134 + 2567060\ M\^136 - 3482690\ M\^138 - 
                    188186\ M\^140 + 1007895\ M\^142 - 274431\ M\^144 - 
                    76486\ M\^146 + 53739\ M\^148 + 7183\ M\^150 - 
                    31907\ M\^152 + 30251\ M\^154 - 17370\ M\^156 + 
                    6533\ M\^158 - 1622\ M\^160 + 258\ M\^162 - 24\ M\^164 + 
                    M\^166)\) + 
              L\^9\ \((3\ M\^36 - 75\ M\^38 + 839\ M\^40 - 5413\ M\^42 + 
                    21389\ M\^44 - 48274\ M\^46 + 33215\ M\^48 + 
                    122166\ M\^50 - 288550\ M\^52 - 164263\ M\^54 + 
                    1326397\ M\^56 - 831323\ M\^58 - 3221683\ M\^60 + 
                    4249720\ M\^62 + 5973954\ M\^64 - 12105907\ M\^66 - 
                    9217172\ M\^68 + 25229178\ M\^70 + 13873492\ M\^72 - 
                    40761713\ M\^74 - 23965401\ M\^76 + 49252186\ M\^78 + 
                    43530309\ M\^80 - 30403874\ M\^82 - 67339887\ M\^84 - 
                    44118003\ M\^86 + 67089309\ M\^88 + 189953033\ M\^90 + 
                    11792587\ M\^92 - 374372362\ M\^94 - 215361589\ M\^96 + 
                    495997539\ M\^98 + 516386925\ M\^100 - 
                    439361990\ M\^102 - 780447623\ M\^104 + 
                    188613049\ M\^106 + 846815702\ M\^108 + 
                    122595202\ M\^110 - 676511042\ M\^112 - 
                    317255455\ M\^114 + 390433967\ M\^116 + 
                    330121859\ M\^118 - 147670452\ M\^120 - 
                    229515653\ M\^122 + 19312720\ M\^124 + 
                    115741875\ M\^126 + 18675480\ M\^128 - 41612587\ M\^130 - 
                    17183159\ M\^132 + 8446341\ M\^134 + 8726848\ M\^136 + 
                    1472476\ M\^138 - 3673919\ M\^140 - 2311944\ M\^142 + 
                    1527472\ M\^144 + 1270523\ M\^146 - 685726\ M\^148 - 
                    489306\ M\^150 + 372309\ M\^152 + 31936\ M\^154 - 
                    72519\ M\^156 - 18117\ M\^158 + 41741\ M\^160 - 
                    22091\ M\^162 + 6293\ M\^164 - 1057\ M\^166 + 
                    99\ M\^168 - 4\ M\^170)\) + 
              L\^8\ \((3\ M\^46 - 72\ M\^48 + 775\ M\^50 - 4834\ M\^52 + 
                    18599\ M\^54 - 41358\ M\^56 + 29389\ M\^58 + 
                    100410\ M\^60 - 259412\ M\^62 - 51394\ M\^64 + 
                    1013948\ M\^66 - 913755\ M\^68 - 2132272\ M\^70 + 
                    3784769\ M\^72 + 3427512\ M\^74 - 10054638\ M\^76 - 
                    5194861\ M\^78 + 22063948\ M\^80 + 9295599\ M\^82 - 
                    43279222\ M\^84 - 19982020\ M\^86 + 76802085\ M\^88 + 
                    42769829\ M\^90 - 118976429\ M\^92 - 81824746\ M\^94 + 
                    152513808\ M\^96 + 133399295\ M\^98 - 149515069\ M\^100 - 
                    175127810\ M\^102 + 90252463\ M\^104 + 
                    168016454\ M\^106 + 6534907\ M\^108 - 83640211\ M\^110 - 
                    79840041\ M\^112 - 53638183\ M\^114 + 73987195\ M\^116 + 
                    166568363\ M\^118 + 1391535\ M\^120 - 192402019\ M\^122 - 
                    78022126\ M\^124 + 140532847\ M\^126 + 
                    102506939\ M\^128 - 68376169\ M\^130 - 79469515\ M\^132 + 
                    19285582\ M\^134 + 43153729\ M\^136 + 1176930\ M\^138 - 
                    17485947\ M\^140 - 5359309\ M\^142 + 5837263\ M\^144 + 
                    3897547\ M\^146 - 1957702\ M\^148 - 1935351\ M\^150 + 
                    887741\ M\^152 + 663084\ M\^154 - 451253\ M\^156 - 
                    72499\ M\^158 + 108689\ M\^160 + 19066\ M\^162 - 
                    54708\ M\^164 + 30063\ M\^166 - 8824\ M\^168 + 
                    1523\ M\^170 - 146\ M\^172 + 6\ M\^174)\) + 
              L\^7\ \((M\^56 - 23\ M\^58 + 244\ M\^60 - 1561\ M\^62 + 
                    6566\ M\^64 - 18190\ M\^66 + 28990\ M\^68 - 4141\ M\^70 - 
                    93125\ M\^72 + 170990\ M\^74 + 64123\ M\^76 - 
                    646432\ M\^78 + 608788\ M\^80 + 1015629\ M\^82 - 
                    2234101\ M\^84 - 482266\ M\^86 + 4199836\ M\^88 - 
                    1160276\ M\^90 - 4832284\ M\^92 + 2500716\ M\^94 + 
                    1617634\ M\^96 - 336754\ M\^98 + 8499228\ M\^100 - 
                    8163765\ M\^102 - 29486602\ M\^104 + 21756134\ M\^106 + 
                    65088836\ M\^108 - 32573053\ M\^110 - 112468494\ M\^112 + 
                    27254372\ M\^114 + 156471387\ M\^116 + 1110407\ M\^118 - 
                    172221235\ M\^120 - 39774195\ M\^122 + 
                    144461415\ M\^124 + 61336283\ M\^126 - 87180794\ M\^128 - 
                    50465560\ M\^130 + 33395939\ M\^132 + 20904346\ M\^134 - 
                    4193679\ M\^136 + 2688072\ M\^138 - 3401979\ M\^140 - 
                    10610970\ M\^142 + 2456949\ M\^144 + 8482907\ M\^146 - 
                    875732\ M\^148 - 4332949\ M\^150 + 203298\ M\^152 + 
                    1749661\ M\^154 - 132028\ M\^156 - 570037\ M\^158 + 
                    109665\ M\^160 + 133170\ M\^162 - 37109\ M\^164 - 
                    49370\ M\^166 + 47128\ M\^168 - 20729\ M\^170 + 
                    5571\ M\^172 - 940\ M\^174 + 92\ M\^176 - 4\ M\^178)\) + 
              L\^6\ \((3\ M\^70 - 64\ M\^72 + 615\ M\^74 - 3409\ M\^76 + 
                    11420\ M\^78 - 20466\ M\^80 + 2211\ M\^82 + 
                    73910\ M\^84 - 122488\ M\^86 - 86376\ M\^88 + 
                    472101\ M\^90 - 220926\ M\^92 - 953828\ M\^94 + 
                    1033062\ M\^96 + 1394580\ M\^98 - 2164458\ M\^100 - 
                    2065219\ M\^102 + 3246283\ M\^104 + 3766640\ M\^106 - 
                    3797967\ M\^108 - 7374437\ M\^110 + 3129871\ M\^112 + 
                    12706641\ M\^114 - 419623\ M\^116 - 17141660\ M\^118 - 
                    4726216\ M\^120 + 16355262\ M\^122 + 10236707\ M\^124 - 
                    7365800\ M\^126 - 11907280\ M\^128 - 6457518\ M\^130 + 
                    7152595\ M\^132 + 15884272\ M\^134 + 1093839\ M\^136 - 
                    15069595\ M\^138 - 6581005\ M\^140 + 7703458\ M\^142 + 
                    6746720\ M\^144 - 1255287\ M\^146 - 4216755\ M\^148 - 
                    1057289\ M\^150 + 1873391\ M\^152 + 854818\ M\^154 - 
                    561462\ M\^156 - 331041\ M\^158 + 83320\ M\^160 + 
                    85988\ M\^162 + 37403\ M\^164 - 46930\ M\^166 - 
                    22347\ M\^168 + 41645\ M\^170 - 23584\ M\^172 + 
                    7710\ M\^174 - 1653\ M\^176 + 239\ M\^178 - 22\ M\^180 + 
                    M\^182)\) + 
              L\^2\ \((3\ M\^134 - 22\ M\^136 + 59\ M\^138 - 28\ M\^140 - 
                    217\ M\^142 + 493\ M\^144 - 32\ M\^146 - 1371\ M\^148 + 
                    1735\ M\^150 + 1381\ M\^152 - 5775\ M\^154 + 
                    3199\ M\^156 + 8716\ M\^158 - 12233\ M\^160 - 
                    4926\ M\^162 + 15717\ M\^164 - 1833\ M\^166 - 
                    9416\ M\^168 + 4067\ M\^170 + 1920\ M\^172 - 
                    2022\ M\^174 + 524\ M\^176 + 228\ M\^178 - 230\ M\^180 + 
                    73\ M\^182 - 10\ M\^184)\) + 
              L\^4\ \((M\^98 - 16\ M\^100 + 119\ M\^102 - 523\ M\^104 + 
                    1444\ M\^106 - 2328\ M\^108 + 761\ M\^110 + 
                    6668\ M\^112 - 16965\ M\^114 + 7898\ M\^116 + 
                    44828\ M\^118 - 80199\ M\^120 - 50195\ M\^122 + 
                    244652\ M\^124 - 12304\ M\^126 - 527160\ M\^128 + 
                    141929\ M\^130 + 1013501\ M\^132 - 357117\ M\^134 - 
                    1745826\ M\^136 + 731602\ M\^138 + 2416054\ M\^140 - 
                    1122386\ M\^142 - 2526980\ M\^144 + 1210780\ M\^146 + 
                    1897005\ M\^148 - 914574\ M\^150 - 921843\ M\^152 + 
                    493957\ M\^154 + 192466\ M\^156 - 191588\ M\^158 + 
                    80313\ M\^160 + 45653\ M\^162 - 90126\ M\^164 + 
                    9552\ M\^166 + 41923\ M\^168 - 23786\ M\^170 - 
                    2471\ M\^172 + 9726\ M\^174 - 6321\ M\^176 + 
                    2361\ M\^178 - 558\ M\^180 + 78\ M\^182 - 5\ M\^184)\) + 
              L\^5\ \((3\ M\^84 - 57\ M\^86 + 489\ M\^88 - 2397\ M\^90 + 
                    6873\ M\^92 - 9124\ M\^94 - 7595\ M\^96 + 48023\ M\^98 - 
                    46221\ M\^100 - 90548\ M\^102 + 228935\ M\^104 + 
                    6138\ M\^106 - 504989\ M\^108 + 349693\ M\^110 + 
                    642243\ M\^112 - 931440\ M\^114 - 318244\ M\^116 + 
                    1508171\ M\^118 - 858874\ M\^120 - 1712629\ M\^122 + 
                    3144933\ M\^124 + 1270456\ M\^126 - 6042613\ M\^128 - 
                    709076\ M\^130 + 8273009\ M\^132 + 1111078\ M\^134 - 
                    8332883\ M\^136 - 2847538\ M\^138 + 5863240\ M\^140 + 
                    4605974\ M\^142 - 2428696\ M\^144 - 4698332\ M\^146 + 
                    87605\ M\^148 + 3204878\ M\^150 + 613026\ M\^152 - 
                    1564994\ M\^154 - 444878\ M\^156 + 612219\ M\^158 + 
                    171948\ M\^160 - 231732\ M\^162 - 26717\ M\^164 + 
                    98898\ M\^166 - 36555\ M\^168 - 16802\ M\^170 + 
                    25783\ M\^172 - 14996\ M\^174 + 5460\ M\^176 - 
                    1345\ M\^178 + 221\ M\^180 - 22\ M\^182 + M\^184)\) + 
              L\ \((M\^152 - 3\ M\^154 - M\^156 + 8\ M\^158 + 7\ M\^160 - 
                    36\ M\^162 + 5\ M\^164 + 133\ M\^166 - 176\ M\^168 - 
                    144\ M\^170 + 346\ M\^172 - 16\ M\^174 - 204\ M\^176 + 
                    72\ M\^178 + 21\ M\^180 - 18\ M\^182 + 5\ M\^184)\) + 
              L\^3\ \((3\ M\^116 - 35\ M\^118 + 176\ M\^120 - 444\ M\^122 + 
                    369\ M\^124 + 886\ M\^126 - 2498\ M\^128 + 1274\ M\^130 + 
                    2426\ M\^132 - 2994\ M\^134 + 4319\ M\^136 - 
                    13406\ M\^138 - 2298\ M\^140 + 66445\ M\^142 - 
                    58747\ M\^144 - 123108\ M\^146 + 198683\ M\^148 + 
                    106102\ M\^150 - 320132\ M\^152 - 15846\ M\^154 + 
                    307054\ M\^156 - 62317\ M\^158 - 177461\ M\^160 + 
                    74121\ M\^162 + 51976\ M\^164 - 43017\ M\^166 + 
                    2529\ M\^168 + 12561\ M\^170 - 8394\ M\^172 + 
                    757\ M\^174 + 2103\ M\^176 - 1551\ M\^178 + 566\ M\^180 - 
                    112\ M\^182 + 10\ M\^184)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 39]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[9, 40]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[9, 41]\), \(M\^38 + L\^15\ M\^52 + 
              L\ \((\(-2\)\ M\^26 + 9\ M\^28 - 22\ M\^30 + 29\ M\^32 + 
                    9\ M\^34 - 61\ M\^36 + 54\ M\^38 - 11\ M\^42 + 3\ M\^48 - 
                    M\^50)\) + 
              L\^2\ \((M\^14 - 9\ M\^16 + 44\ M\^18 - 132\ M\^20 + 
                    246\ M\^22 - 264\ M\^24 + 77\ M\^26 + 209\ M\^28 - 
                    186\ M\^30 - 286\ M\^32 + 545\ M\^34 - 139\ M\^36 - 
                    239\ M\^38 + 148\ M\^40 + 155\ M\^42 - 264\ M\^44 + 
                    M\^46 + 232\ M\^48 - 124\ M\^50 - 24\ M\^52 + 33\ M\^54 - 
                    8\ M\^56 + M\^58)\) + 
              L\^14\ \((\(-M\^40\) + 3\ M\^42 - 11\ M\^48 + 54\ M\^52 - 
                    61\ M\^54 + 9\ M\^56 + 29\ M\^58 - 22\ M\^60 + 9\ M\^62 - 
                    2\ M\^64)\) + 
              L\^3\ \((\(-M\^6\) + 12\ M\^8 - 69\ M\^10 + 239\ M\^12 - 
                    516\ M\^14 + 647\ M\^16 - 332\ M\^18 - 87\ M\^20 - 
                    273\ M\^22 + 1245\ M\^24 - 950\ M\^26 - 709\ M\^28 - 
                    616\ M\^30 + 4990\ M\^32 - 2686\ M\^34 - 5826\ M\^36 + 
                    3868\ M\^38 + 7676\ M\^40 - 7252\ M\^42 - 4298\ M\^44 + 
                    5879\ M\^46 + 2165\ M\^48 - 4390\ M\^50 + 447\ M\^52 + 
                    1397\ M\^54 - 510\ M\^56 - 145\ M\^58 + 137\ M\^60 - 
                    40\ M\^62 + 5\ M\^64)\) + 
              L\^4\ \((M\^2 - 13\ M\^4 + 89\ M\^6 - 390\ M\^8 + 1139\ M\^10 - 
                    2060\ M\^12 + 1494\ M\^14 + 2148\ M\^16 - 4947\ M\^18 - 
                    2096\ M\^20 + 16233\ M\^22 - 13994\ M\^24 - 
                    14306\ M\^26 + 27056\ M\^28 + 14521\ M\^30 - 
                    49663\ M\^32 - 1030\ M\^34 + 52691\ M\^36 + 6829\ M\^38 - 
                    60070\ M\^40 - 6166\ M\^42 + 46774\ M\^44 + 
                    16065\ M\^46 - 39236\ M\^48 - 8822\ M\^50 + 
                    21065\ M\^52 + 5333\ M\^54 - 11570\ M\^56 + 1267\ M\^58 + 
                    2353\ M\^60 - 451\ M\^62 - 581\ M\^64 + 384\ M\^66 - 
                    100\ M\^68 + 10\ M\^70)\) + 
              L\^13\ \((M\^32 - 8\ M\^34 + 33\ M\^36 - 24\ M\^38 - 
                    124\ M\^40 + 232\ M\^42 + M\^44 - 264\ M\^46 + 
                    155\ M\^48 + 148\ M\^50 - 239\ M\^52 - 139\ M\^54 + 
                    545\ M\^56 - 286\ M\^58 - 186\ M\^60 + 209\ M\^62 + 
                    77\ M\^64 - 264\ M\^66 + 246\ M\^68 - 132\ M\^70 + 
                    44\ M\^72 - 9\ M\^74 + M\^76)\) + 
              L\^5\ \((1 - 17\ M\^2 + 133\ M\^4 - 625\ M\^6 + 1854\ M\^8 - 
                    3196\ M\^10 + 1502\ M\^12 + 5840\ M\^14 - 9544\ M\^16 - 
                    10613\ M\^18 + 43763\ M\^20 - 25073\ M\^22 - 
                    62387\ M\^24 + 87490\ M\^26 + 61364\ M\^28 - 
                    200077\ M\^30 + 27219\ M\^32 + 266193\ M\^34 - 
                    115415\ M\^36 - 303542\ M\^38 + 195138\ M\^40 + 
                    241603\ M\^42 - 162238\ M\^44 - 173585\ M\^46 + 
                    106498\ M\^48 + 80365\ M\^50 - 27610\ M\^52 - 
                    41465\ M\^54 + 5636\ M\^56 + 10651\ M\^58 + 6845\ M\^60 - 
                    7907\ M\^62 - 51\ M\^64 + 868\ M\^66 + 1252\ M\^68 - 
                    1448\ M\^70 + 616\ M\^72 - 125\ M\^74 + 10\ M\^76)\) + 
              L\^6\ \((\(-2\) + 35\ M\^2 - 266\ M\^4 + 1114\ M\^6 - 
                    2627\ M\^8 + 2757\ M\^10 + 1587\ M\^12 - 5893\ M\^14 - 
                    4938\ M\^16 + 28577\ M\^18 - 21142\ M\^20 - 
                    35589\ M\^22 + 54050\ M\^24 + 45062\ M\^26 - 
                    151983\ M\^28 + 50814\ M\^30 + 225232\ M\^32 - 
                    193958\ M\^34 - 326061\ M\^36 + 426459\ M\^38 + 
                    317562\ M\^40 - 527165\ M\^42 - 296314\ M\^44 + 
                    533658\ M\^46 + 164600\ M\^48 - 362660\ M\^50 - 
                    79778\ M\^52 + 215317\ M\^54 - 11087\ M\^56 - 
                    73896\ M\^58 + 16209\ M\^60 + 26435\ M\^62 - 
                    22575\ M\^64 + 4705\ M\^66 + 4093\ M\^68 - 1815\ M\^70 - 
                    2350\ M\^72 + 3044\ M\^74 - 1633\ M\^76 + 482\ M\^78 - 
                    76\ M\^80 + 5\ M\^82)\) + 
              L\^12\ \((5\ M\^26 - 40\ M\^28 + 137\ M\^30 - 145\ M\^32 - 
                    510\ M\^34 + 1397\ M\^36 + 447\ M\^38 - 4390\ M\^40 + 
                    2165\ M\^42 + 5879\ M\^44 - 4298\ M\^46 - 7252\ M\^48 + 
                    7676\ M\^50 + 3868\ M\^52 - 5826\ M\^54 - 2686\ M\^56 + 
                    4990\ M\^58 - 616\ M\^60 - 709\ M\^62 - 950\ M\^64 + 
                    1245\ M\^66 - 273\ M\^68 - 87\ M\^70 - 332\ M\^72 + 
                    647\ M\^74 - 516\ M\^76 + 239\ M\^78 - 69\ M\^80 + 
                    12\ M\^82 - M\^84)\) + 
              L\^7\ \((1 - 20\ M\^2 + 173\ M\^4 - 838\ M\^6 + 2411\ M\^8 - 
                    3713\ M\^10 + 635\ M\^12 + 8134\ M\^14 - 8096\ M\^16 - 
                    22841\ M\^18 + 60307\ M\^20 - 15106\ M\^22 - 
                    117354\ M\^24 + 104878\ M\^26 + 201580\ M\^28 - 
                    333147\ M\^30 - 183831\ M\^32 + 550598\ M\^34 + 
                    152729\ M\^36 - 754320\ M\^38 - 13068\ M\^40 + 
                    710497\ M\^42 - 45050\ M\^44 - 567723\ M\^46 + 
                    133149\ M\^48 + 279075\ M\^50 - 90040\ M\^52 - 
                    107964\ M\^54 + 73965\ M\^56 - 23756\ M\^58 + 
                    2994\ M\^60 + 20096\ M\^62 - 11009\ M\^64 - 
                    19419\ M\^66 + 25100\ M\^68 - 7388\ M\^70 - 4516\ M\^72 + 
                    2738\ M\^74 + 2081\ M\^76 - 3206\ M\^78 + 1901\ M\^80 - 
                    665\ M\^82 + 144\ M\^84 - 18\ M\^86 + M\^88)\) + 
              L\^11\ \((10\ M\^20 - 100\ M\^22 + 384\ M\^24 - 581\ M\^26 - 
                    451\ M\^28 + 2353\ M\^30 + 1267\ M\^32 - 11570\ M\^34 + 
                    5333\ M\^36 + 21065\ M\^38 - 8822\ M\^40 - 39236\ M\^42 + 
                    16065\ M\^44 + 46774\ M\^46 - 6166\ M\^48 - 
                    60070\ M\^50 + 6829\ M\^52 + 52691\ M\^54 - 1030\ M\^56 - 
                    49663\ M\^58 + 14521\ M\^60 + 27056\ M\^62 - 
                    14306\ M\^64 - 13994\ M\^66 + 16233\ M\^68 - 
                    2096\ M\^70 - 4947\ M\^72 + 2148\ M\^74 + 1494\ M\^76 - 
                    2060\ M\^78 + 1139\ M\^80 - 390\ M\^82 + 89\ M\^84 - 
                    13\ M\^86 + M\^88)\) + 
              L\^9\ \((5\ M\^8 - 76\ M\^10 + 482\ M\^12 - 1633\ M\^14 + 
                    3044\ M\^16 - 2350\ M\^18 - 1815\ M\^20 + 4093\ M\^22 + 
                    4705\ M\^24 - 22575\ M\^26 + 26435\ M\^28 + 
                    16209\ M\^30 - 73896\ M\^32 - 11087\ M\^34 + 
                    215317\ M\^36 - 79778\ M\^38 - 362660\ M\^40 + 
                    164600\ M\^42 + 533658\ M\^44 - 296314\ M\^46 - 
                    527165\ M\^48 + 317562\ M\^50 + 426459\ M\^52 - 
                    326061\ M\^54 - 193958\ M\^56 + 225232\ M\^58 + 
                    50814\ M\^60 - 151983\ M\^62 + 45062\ M\^64 + 
                    54050\ M\^66 - 35589\ M\^68 - 21142\ M\^70 + 
                    28577\ M\^72 - 4938\ M\^74 - 5893\ M\^76 + 1587\ M\^78 + 
                    2757\ M\^80 - 2627\ M\^82 + 1114\ M\^84 - 266\ M\^86 + 
                    35\ M\^88 - 2\ M\^90)\) + 
              L\^8\ \((M\^2 - 18\ M\^4 + 144\ M\^6 - 665\ M\^8 + 
                    1901\ M\^10 - 3206\ M\^12 + 2081\ M\^14 + 2738\ M\^16 - 
                    4516\ M\^18 - 7388\ M\^20 + 25100\ M\^22 - 19419\ M\^24 - 
                    11009\ M\^26 + 20096\ M\^28 + 2994\ M\^30 - 
                    23756\ M\^32 + 73965\ M\^34 - 107964\ M\^36 - 
                    90040\ M\^38 + 279075\ M\^40 + 133149\ M\^42 - 
                    567723\ M\^44 - 45050\ M\^46 + 710497\ M\^48 - 
                    13068\ M\^50 - 754320\ M\^52 + 152729\ M\^54 + 
                    550598\ M\^56 - 183831\ M\^58 - 333147\ M\^60 + 
                    201580\ M\^62 + 104878\ M\^64 - 117354\ M\^66 - 
                    15106\ M\^68 + 60307\ M\^70 - 22841\ M\^72 - 
                    8096\ M\^74 + 8134\ M\^76 + 635\ M\^78 - 3713\ M\^80 + 
                    2411\ M\^82 - 838\ M\^84 + 173\ M\^86 - 20\ M\^88 + 
                    M\^90)\) + 
              L\^10\ \((10\ M\^14 - 125\ M\^16 + 616\ M\^18 - 1448\ M\^20 + 
                    1252\ M\^22 + 868\ M\^24 - 51\ M\^26 - 7907\ M\^28 + 
                    6845\ M\^30 + 10651\ M\^32 + 5636\ M\^34 - 41465\ M\^36 - 
                    27610\ M\^38 + 80365\ M\^40 + 106498\ M\^42 - 
                    173585\ M\^44 - 162238\ M\^46 + 241603\ M\^48 + 
                    195138\ M\^50 - 303542\ M\^52 - 115415\ M\^54 + 
                    266193\ M\^56 + 27219\ M\^58 - 200077\ M\^60 + 
                    61364\ M\^62 + 87490\ M\^64 - 62387\ M\^66 - 
                    25073\ M\^68 + 43763\ M\^70 - 10613\ M\^72 - 
                    9544\ M\^74 + 5840\ M\^76 + 1502\ M\^78 - 3196\ M\^80 + 
                    1854\ M\^82 - 625\ M\^84 + 133\ M\^86 - 17\ M\^88 + 
                    M\^90)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 42]\), \(M\^8 + L\^7\ M\^26 + 
              L\ \((\(-1\) + M\^2 + 3\ M\^6 - 3\ M\^8 + 4\ M\^12 - M\^14)\) + 
              L\^2\ \((\(-M\^4\) - M\^6 + 10\ M\^8 - 3\ M\^10 - 9\ M\^12 + 
                    M\^14 + 4\ M\^16 + M\^18 - M\^20)\) + 
              L\^3\ \((\(-M\^6\) - 4\ M\^8 + 14\ M\^10 + 9\ M\^12 - 
                    18\ M\^14 - 18\ M\^16 + M\^18 + 13\ M\^20 + 4\ M\^22 - 
                    6\ M\^24 + M\^26)\) + 
              L\^4\ \((M\^8 - 6\ M\^10 + 4\ M\^12 + 13\ M\^14 + M\^16 - 
                    18\ M\^18 - 18\ M\^20 + 9\ M\^22 + 14\ M\^24 - 4\ M\^26 - 
                    M\^28)\) + 
              L\^5\ \((\(-M\^14\) + M\^16 + 4\ M\^18 + M\^20 - 9\ M\^22 - 
                    3\ M\^24 + 10\ M\^26 - M\^28 - M\^30)\) + 
              L\^6\ \((\(-M\^20\) + 4\ M\^22 - 3\ M\^26 + 3\ M\^28 + M\^32 - 
                    M\^34)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 43]\), \(M\^4 + L\^10\ M\^84 + 
              L\ \((\(-1\) + 2\ M\^2 - 3\ M\^4 + 7\ M\^6 - 3\ M\^8 - 
                    4\ M\^10 + 7\ M\^12 + M\^14)\) + 
              L\^2\ \((\(-2\)\ M\^6 + 9\ M\^8 - 17\ M\^10 + 14\ M\^12 + 
                    2\ M\^14 + 2\ M\^16 + 7\ M\^18 - 14\ M\^20 + 8\ M\^22 + 
                    4\ M\^24)\) + 
              L\^3\ \((\(-M\^12\) + 8\ M\^14 - 23\ M\^16 + 26\ M\^18 - 
                    2\ M\^24 - 2\ M\^26 + 19\ M\^28 - 28\ M\^30 - 3\ M\^32 + 
                    16\ M\^34 - 2\ M\^36)\) + 
              L\^4\ \((M\^20 - 6\ M\^22 + 18\ M\^24 - 28\ M\^26 + 13\ M\^28 + 
                    33\ M\^30 + 14\ M\^32 - 38\ M\^34 - 59\ M\^36 + 
                    31\ M\^38 + 17\ M\^40 - 22\ M\^42 + 12\ M\^44)\) + 
              L\^5\ \((\(-M\^32\) + 15\ M\^34 - 37\ M\^36 + 22\ M\^38 + 
                    69\ M\^40 - 31\ M\^42 - 102\ M\^44 - 31\ M\^46 + 
                    69\ M\^48 + 22\ M\^50 - 37\ M\^52 + 15\ M\^54 - 
                    M\^56)\) + 
              L\^6\ \((12\ M\^44 - 22\ M\^46 + 17\ M\^48 + 31\ M\^50 - 
                    59\ M\^52 - 38\ M\^54 + 14\ M\^56 + 33\ M\^58 + 
                    13\ M\^60 - 28\ M\^62 + 18\ M\^64 - 6\ M\^66 + M\^68)\) + 
              L\^7\ \((\(-2\)\ M\^52 + 16\ M\^54 - 3\ M\^56 - 28\ M\^58 + 
                    19\ M\^60 - 2\ M\^62 - 2\ M\^64 + 26\ M\^70 - 23\ M\^72 + 
                    8\ M\^74 - M\^76)\) + 
              L\^8\ \((4\ M\^64 + 8\ M\^66 - 14\ M\^68 + 7\ M\^70 + 
                    2\ M\^72 + 2\ M\^74 + 14\ M\^76 - 17\ M\^78 + 9\ M\^80 - 
                    2\ M\^82)\) + 
              L\^9\ \((M\^74 + 7\ M\^76 - 4\ M\^78 - 3\ M\^80 + 7\ M\^82 - 
                    3\ M\^84 + 2\ M\^86 - M\^88)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 44]\), \(L\^12\ M\^20 + M\^24 + 
              L\ \((2\ M\^14 - 4\ M\^16 + 6\ M\^18 - 11\ M\^20 + 2\ M\^22 + 
                    16\ M\^24 - 2\ M\^26 - M\^28)\) + 
              L\^11\ \((\(-M\^16\) - 2\ M\^18 + 16\ M\^20 + 2\ M\^22 - 
                    11\ M\^24 + 6\ M\^26 - 4\ M\^28 + 2\ M\^30)\) + 
              L\^2\ \((M\^4 - 4\ M\^6 + 10\ M\^8 - 14\ M\^10 - M\^12 + 
                    16\ M\^14 + 31\ M\^16 - 60\ M\^18 - 15\ M\^20 + 
                    25\ M\^22 + 11\ M\^24 + 46\ M\^26 - 13\ M\^28 - 
                    11\ M\^30 + 4\ M\^32)\) + 
              L\^3\ \((2\ M\^2 - 15\ M\^4 + 41\ M\^6 - 39\ M\^8 - M\^10 + 
                    4\ M\^12 + 16\ M\^14 + 19\ M\^16 + 42\ M\^18 - 
                    117\ M\^20 - 79\ M\^22 + 88\ M\^24 + 62\ M\^26 + 
                    86\ M\^28 - 63\ M\^30 - 32\ M\^32 + 35\ M\^34 - 
                    10\ M\^36 + M\^38)\) + 
              L\^10\ \((4\ M\^12 - 11\ M\^14 - 13\ M\^16 + 46\ M\^18 + 
                    11\ M\^20 + 25\ M\^22 - 15\ M\^24 - 60\ M\^26 + 
                    31\ M\^28 + 16\ M\^30 - M\^32 - 14\ M\^34 + 10\ M\^36 - 
                    4\ M\^38 + M\^40)\) + 
              L\^4\ \((1 - 12\ M\^2 + 49\ M\^4 - 88\ M\^6 + 64\ M\^8 + 
                    18\ M\^10 - 40\ M\^12 + 7\ M\^14 + 5\ M\^16 + 27\ M\^18 + 
                    76\ M\^20 - 199\ M\^22 - 195\ M\^24 + 202\ M\^26 + 
                    117\ M\^28 + 77\ M\^30 - 79\ M\^32 - 67\ M\^34 + 
                    71\ M\^36 - 21\ M\^38 + 2\ M\^40)\) + 
              L\^6\ \((\(-3\)\ M\^2 + 35\ M\^4 - 135\ M\^6 + 175\ M\^8 + 
                    73\ M\^10 - 249\ M\^12 - 101\ M\^14 + 262\ M\^16 + 
                    502\ M\^18 - 195\ M\^20 - 812\ M\^22 - 195\ M\^24 + 
                    502\ M\^26 + 262\ M\^28 - 101\ M\^30 - 249\ M\^32 + 
                    73\ M\^34 + 175\ M\^36 - 135\ M\^38 + 35\ M\^40 - 
                    3\ M\^42)\) + 
              L\^5\ \((\(-1\) + 14\ M\^2 - 68\ M\^4 + 131\ M\^6 - 66\ M\^8 - 
                    48\ M\^10 + 7\ M\^12 - 24\ M\^14 + 220\ M\^16 + 
                    117\ M\^18 - 312\ M\^20 - 275\ M\^22 + 21\ M\^24 + 
                    148\ M\^26 + 151\ M\^28 - 133\ M\^30 + 71\ M\^32 + 
                    72\ M\^34 - 122\ M\^36 + 61\ M\^38 - 13\ M\^40 + 
                    M\^42)\) + 
              L\^9\ \((M\^6 - 10\ M\^8 + 35\ M\^10 - 32\ M\^12 - 63\ M\^14 + 
                    86\ M\^16 + 62\ M\^18 + 88\ M\^20 - 79\ M\^22 - 
                    117\ M\^24 + 42\ M\^26 + 19\ M\^28 + 16\ M\^30 + 
                    4\ M\^32 - M\^34 - 39\ M\^36 + 41\ M\^38 - 15\ M\^40 + 
                    2\ M\^42)\) + 
              L\^7\ \((M\^2 - 13\ M\^4 + 61\ M\^6 - 122\ M\^8 + 72\ M\^10 + 
                    71\ M\^12 - 133\ M\^14 + 151\ M\^16 + 148\ M\^18 + 
                    21\ M\^20 - 275\ M\^22 - 312\ M\^24 + 117\ M\^26 + 
                    220\ M\^28 - 24\ M\^30 + 7\ M\^32 - 48\ M\^34 - 
                    66\ M\^36 + 131\ M\^38 - 68\ M\^40 + 14\ M\^42 - 
                    M\^44)\) + 
              L\^8\ \((2\ M\^4 - 21\ M\^6 + 71\ M\^8 - 67\ M\^10 - 
                    79\ M\^12 + 77\ M\^14 + 117\ M\^16 + 202\ M\^18 - 
                    195\ M\^20 - 199\ M\^22 + 76\ M\^24 + 27\ M\^26 + 
                    5\ M\^28 + 7\ M\^30 - 40\ M\^32 + 18\ M\^34 + 64\ M\^36 - 
                    88\ M\^38 + 49\ M\^40 - 12\ M\^42 + M\^44)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 45]\), \(L\^15\ M\^4 + M\^62 + 
              L\^14\ \((\(-4\)\ M\^4 + 2\ M\^6 + 28\ M\^8 - 6\ M\^10 - 
                    18\ M\^12 + 13\ M\^14 - 6\ M\^16 + 2\ M\^18)\) + 
              L\^13\ \((\(-2\)\ M\^2 + 22\ M\^4 - 53\ M\^6 - 61\ M\^8 + 
                    216\ M\^10 + 80\ M\^12 - 123\ M\^14 - 75\ M\^16 + 
                    12\ M\^18 + 51\ M\^20 - 16\ M\^22 + 18\ M\^24 - 
                    30\ M\^26 + 19\ M\^28 - 6\ M\^30 + M\^32)\) + 
              L\^12\ \((4\ M\^2 - 42\ M\^4 + 158\ M\^6 - 96\ M\^8 - 
                    566\ M\^10 + 501\ M\^12 + 1050\ M\^14 - 240\ M\^16 - 
                    972\ M\^18 + 62\ M\^20 + 216\ M\^22 + 88\ M\^24 + 
                    168\ M\^26 - 238\ M\^28 + 2\ M\^30 + 28\ M\^32 + 
                    70\ M\^34 - 77\ M\^36 + 34\ M\^38 - 8\ M\^40 + M\^42)\) + 
              L\^11\ \((1 - 15\ M\^2 + 91\ M\^4 - 297\ M\^6 + 392\ M\^8 + 
                    642\ M\^10 - 2261\ M\^12 - 387\ M\^14 + 4669\ M\^16 + 
                    534\ M\^18 - 3937\ M\^20 - 496\ M\^22 + 552\ M\^24 + 
                    962\ M\^26 + 698\ M\^28 - 1035\ M\^30 - 392\ M\^32 + 
                    496\ M\^34 + 248\ M\^36 - 362\ M\^38 + 173\ M\^40 - 
                    110\ M\^42 + 87\ M\^44 - 41\ M\^46 + 10\ M\^48 - 
                    M\^50)\) + 
              L\^10\ \((\(-1\) + 17\ M\^2 - 125\ M\^4 + 493\ M\^6 - 
                    998\ M\^8 + 211\ M\^10 + 3517\ M\^12 - 5012\ M\^14 - 
                    4718\ M\^16 + 10971\ M\^18 + 4600\ M\^20 - 8609\ M\^22 - 
                    4121\ M\^24 + 763\ M\^26 + 4070\ M\^28 + 2406\ M\^30 - 
                    3343\ M\^32 - 2464\ M\^34 + 2219\ M\^36 + 1437\ M\^38 - 
                    1073\ M\^40 - 383\ M\^42 + 70\ M\^44 + 501\ M\^46 - 
                    402\ M\^48 + 140\ M\^50 - 25\ M\^52 + 2\ M\^54)\) + 
              L\^9\ \((\(-4\)\ M\^2 + 61\ M\^4 - 383\ M\^6 + 1265\ M\^8 - 
                    1969\ M\^10 - 709\ M\^12 + 7696\ M\^14 - 6759\ M\^16 - 
                    11407\ M\^18 + 15236\ M\^20 + 12484\ M\^22 - 
                    11453\ M\^24 - 12457\ M\^26 + 485\ M\^28 + 10892\ M\^30 + 
                    5369\ M\^32 - 9331\ M\^34 - 7056\ M\^36 + 7268\ M\^38 + 
                    5265\ M\^40 - 4179\ M\^42 - 2658\ M\^44 + 2300\ M\^46 + 
                    792\ M\^48 - 1414\ M\^50 + 660\ M\^52 - 155\ M\^54 + 
                    19\ M\^56 - M\^58)\) + 
              L\^8\ \((\(-6\)\ M\^4 + 90\ M\^6 - 560\ M\^8 + 1807\ M\^10 - 
                    2658\ M\^12 - 1114\ M\^14 + 9426\ M\^16 - 6487\ M\^18 - 
                    14349\ M\^20 + 14127\ M\^22 + 17978\ M\^24 - 
                    11305\ M\^26 - 19637\ M\^28 + 1666\ M\^30 + 
                    18134\ M\^32 + 5994\ M\^34 - 17708\ M\^36 - 
                    11035\ M\^38 + 15101\ M\^40 + 10494\ M\^42 - 
                    10461\ M\^44 - 5229\ M\^46 + 6730\ M\^48 - 51\ M\^50 - 
                    2493\ M\^52 + 1476\ M\^54 - 417\ M\^56 + 62\ M\^58 - 
                    4\ M\^60)\) + 
              L\^7\ \((\(-4\)\ M\^6 + 62\ M\^8 - 417\ M\^10 + 1476\ M\^12 - 
                    2493\ M\^14 - 51\ M\^16 + 6730\ M\^18 - 5229\ M\^20 - 
                    10461\ M\^22 + 10494\ M\^24 + 15101\ M\^26 - 
                    11035\ M\^28 - 17708\ M\^30 + 5994\ M\^32 + 
                    18134\ M\^34 + 1666\ M\^36 - 19637\ M\^38 - 
                    11305\ M\^40 + 17978\ M\^42 + 14127\ M\^44 - 
                    14349\ M\^46 - 6487\ M\^48 + 9426\ M\^50 - 1114\ M\^52 - 
                    2658\ M\^54 + 1807\ M\^56 - 560\ M\^58 + 90\ M\^60 - 
                    6\ M\^62)\) + 
              L\ \((2\ M\^48 - 6\ M\^50 + 13\ M\^52 - 18\ M\^54 - 6\ M\^56 + 
                    28\ M\^58 + 2\ M\^60 - 4\ M\^62)\) + 
              L\^6\ \((\(-M\^8\) + 19\ M\^10 - 155\ M\^12 + 660\ M\^14 - 
                    1414\ M\^16 + 792\ M\^18 + 2300\ M\^20 - 2658\ M\^22 - 
                    4179\ M\^24 + 5265\ M\^26 + 7268\ M\^28 - 7056\ M\^30 - 
                    9331\ M\^32 + 5369\ M\^34 + 10892\ M\^36 + 485\ M\^38 - 
                    12457\ M\^40 - 11453\ M\^42 + 12484\ M\^44 + 
                    15236\ M\^46 - 11407\ M\^48 - 6759\ M\^50 + 7696\ M\^52 - 
                    709\ M\^54 - 1969\ M\^56 + 1265\ M\^58 - 383\ M\^60 + 
                    61\ M\^62 - 4\ M\^64)\) + 
              L\^2\ \((M\^34 - 6\ M\^36 + 19\ M\^38 - 30\ M\^40 + 18\ M\^42 - 
                    16\ M\^44 + 51\ M\^46 + 12\ M\^48 - 75\ M\^50 - 
                    123\ M\^52 + 80\ M\^54 + 216\ M\^56 - 61\ M\^58 - 
                    53\ M\^60 + 22\ M\^62 - 2\ M\^64)\) + 
              L\^3\ \((M\^24 - 8\ M\^26 + 34\ M\^28 - 77\ M\^30 + 70\ M\^32 + 
                    28\ M\^34 + 2\ M\^36 - 238\ M\^38 + 168\ M\^40 + 
                    88\ M\^42 + 216\ M\^44 + 62\ M\^46 - 972\ M\^48 - 
                    240\ M\^50 + 1050\ M\^52 + 501\ M\^54 - 566\ M\^56 - 
                    96\ M\^58 + 158\ M\^60 - 42\ M\^62 + 4\ M\^64)\) + 
              L\^5\ \((2\ M\^12 - 25\ M\^14 + 140\ M\^16 - 402\ M\^18 + 
                    501\ M\^20 + 70\ M\^22 - 383\ M\^24 - 1073\ M\^26 + 
                    1437\ M\^28 + 2219\ M\^30 - 2464\ M\^32 - 3343\ M\^34 + 
                    2406\ M\^36 + 4070\ M\^38 + 763\ M\^40 - 4121\ M\^42 - 
                    8609\ M\^44 + 4600\ M\^46 + 10971\ M\^48 - 4718\ M\^50 - 
                    5012\ M\^52 + 3517\ M\^54 + 211\ M\^56 - 998\ M\^58 + 
                    493\ M\^60 - 125\ M\^62 + 17\ M\^64 - M\^66)\) + 
              L\^4\ \((\(-M\^16\) + 10\ M\^18 - 41\ M\^20 + 87\ M\^22 - 
                    110\ M\^24 + 173\ M\^26 - 362\ M\^28 + 248\ M\^30 + 
                    496\ M\^32 - 392\ M\^34 - 1035\ M\^36 + 698\ M\^38 + 
                    962\ M\^40 + 552\ M\^42 - 496\ M\^44 - 3937\ M\^46 + 
                    534\ M\^48 + 4669\ M\^50 - 387\ M\^52 - 2261\ M\^54 + 
                    642\ M\^56 + 392\ M\^58 - 297\ M\^60 + 91\ M\^62 - 
                    15\ M\^64 + M\^66)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 46]\), \(L\^5\ M\^6 + M\^12 + 
              L\ \((\(-1\) + M\^2 - 2\ M\^4 + 5\ M\^6 - 2\ M\^8 + 5\ M\^12 - 
                    M\^14)\) + 
              L\^2\ \((1 - 6\ M\^2 + 4\ M\^4 + 3\ M\^6 + 3\ M\^8 + 3\ M\^10 + 
                    3\ M\^12 - M\^16)\) + 
              L\^4\ \((\(-M\^4\) + 5\ M\^6 - 2\ M\^10 + 5\ M\^12 - 2\ M\^14 + 
                    M\^16 - M\^18)\) + 
              L\^3\ \((\(-M\^2\) + 3\ M\^6 + 3\ M\^8 + 3\ M\^10 + 3\ M\^12 + 
                    4\ M\^14 - 6\ M\^16 + M\^18)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 47]\), \(\(-M\^12\) + 
              L\^11\ M\^68 + 
              L\ \((\(-2\)\ M\^6 + 6\ M\^8 - 12\ M\^10 + 12\ M\^12 + 
                    18\ M\^14 - 31\ M\^16 + 4\ M\^18 - M\^26 + M\^28)\) + 
              L\^2\ \((\(-1\) + 6\ M\^2 - 21\ M\^4 + 49\ M\^6 - 66\ M\^8 + 
                    55\ M\^10 - 36\ M\^12 - 18\ M\^14 - 15\ M\^16 + 
                    113\ M\^18 + 42\ M\^20 - 102\ M\^22 - 72\ M\^24 + 
                    87\ M\^26 - 12\ M\^28 - 52\ M\^30 + 41\ M\^32 - M\^34 - 
                    4\ M\^36)\) + 
              L\^3\ \((1 - 12\ M\^2 + 56\ M\^4 - 115\ M\^6 + 90\ M\^8 - 
                    53\ M\^10 + 241\ M\^12 - 360\ M\^14 + 94\ M\^16 + 
                    110\ M\^18 - 38\ M\^20 - 509\ M\^22 + 795\ M\^24 + 
                    87\ M\^26 - 435\ M\^28 - 43\ M\^30 + 335\ M\^32 - 
                    414\ M\^34 + 56\ M\^36 + 202\ M\^38 - 48\ M\^40 - 
                    59\ M\^42 + 26\ M\^44 - 2\ M\^46)\) + 
              L\^4\ \((\(-2\)\ M\^4 + 27\ M\^6 - 142\ M\^8 + 338\ M\^10 - 
                    271\ M\^12 - 196\ M\^14 + 54\ M\^16 + 986\ M\^18 - 
                    972\ M\^20 - 24\ M\^22 + 3\ M\^24 + 446\ M\^26 - 
                    916\ M\^28 + 677\ M\^30 - 10\ M\^32 + 846\ M\^34 - 
                    587\ M\^36 - 363\ M\^38 - 532\ M\^40 + 796\ M\^42 + 
                    84\ M\^44 - 161\ M\^46 - 212\ M\^48 + 207\ M\^50 - 
                    59\ M\^52 + 5\ M\^54)\) + 
              L\^5\ \((M\^8 - 15\ M\^10 + 89\ M\^12 - 253\ M\^14 + 
                    317\ M\^16 - 65\ M\^18 + 57\ M\^20 - 772\ M\^22 + 
                    779\ M\^24 + 41\ M\^26 + 880\ M\^28 - 1446\ M\^30 - 
                    394\ M\^32 - 298\ M\^34 + 2178\ M\^36 - 1163\ M\^38 + 
                    240\ M\^40 - 146\ M\^42 + 260\ M\^44 - 1106\ M\^46 + 
                    1291\ M\^48 - 270\ M\^50 - 310\ M\^52 - 102\ M\^54 + 
                    455\ M\^56 - 323\ M\^58 + 104\ M\^60 - 16\ M\^62 + 
                    M\^64)\) + 
              L\^6\ \((\(-M\^16\) + 16\ M\^18 - 104\ M\^20 + 323\ M\^22 - 
                    455\ M\^24 + 102\ M\^26 + 310\ M\^28 + 270\ M\^30 - 
                    1291\ M\^32 + 1106\ M\^34 - 260\ M\^36 + 146\ M\^38 - 
                    240\ M\^40 + 1163\ M\^42 - 2178\ M\^44 + 298\ M\^46 + 
                    394\ M\^48 + 1446\ M\^50 - 880\ M\^52 - 41\ M\^54 - 
                    779\ M\^56 + 772\ M\^58 - 57\ M\^60 + 65\ M\^62 - 
                    317\ M\^64 + 253\ M\^66 - 89\ M\^68 + 15\ M\^70 - 
                    M\^72)\) + 
              L\^10\ \((\(-M\^52\) + M\^54 - 4\ M\^62 + 31\ M\^64 - 
                    18\ M\^66 - 12\ M\^68 + 12\ M\^70 - 6\ M\^72 + 
                    2\ M\^74)\) + 
              L\^7\ \((\(-5\)\ M\^26 + 59\ M\^28 - 207\ M\^30 + 212\ M\^32 + 
                    161\ M\^34 - 84\ M\^36 - 796\ M\^38 + 532\ M\^40 + 
                    363\ M\^42 + 587\ M\^44 - 846\ M\^46 + 10\ M\^48 - 
                    677\ M\^50 + 916\ M\^52 - 446\ M\^54 - 3\ M\^56 + 
                    24\ M\^58 + 972\ M\^60 - 986\ M\^62 - 54\ M\^64 + 
                    196\ M\^66 + 271\ M\^68 - 338\ M\^70 + 142\ M\^72 - 
                    27\ M\^74 + 2\ M\^76)\) + 
              L\^8\ \((2\ M\^34 - 26\ M\^36 + 59\ M\^38 + 48\ M\^40 - 
                    202\ M\^42 - 56\ M\^44 + 414\ M\^46 - 335\ M\^48 + 
                    43\ M\^50 + 435\ M\^52 - 87\ M\^54 - 795\ M\^56 + 
                    509\ M\^58 + 38\ M\^60 - 110\ M\^62 - 94\ M\^64 + 
                    360\ M\^66 - 241\ M\^68 + 53\ M\^70 - 90\ M\^72 + 
                    115\ M\^74 - 56\ M\^76 + 12\ M\^78 - M\^80)\) + 
              L\^9\ \((4\ M\^44 + M\^46 - 41\ M\^48 + 52\ M\^50 + 12\ M\^52 - 
                    87\ M\^54 + 72\ M\^56 + 102\ M\^58 - 42\ M\^60 - 
                    113\ M\^62 + 15\ M\^64 + 18\ M\^66 + 36\ M\^68 - 
                    55\ M\^70 + 66\ M\^72 - 49\ M\^74 + 21\ M\^76 - 
                    6\ M\^78 + M\^80)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 48]\), \(\(-M\^12\) + 
              L\^8\ M\^42 + 
              L\ \((3\ M\^8 - 10\ M\^10 + 24\ M\^12 - 12\ M\^14 - 27\ M\^16 + 
                    18\ M\^18 + M\^20 - 3\ M\^22)\) + 
              L\^2\ \((\(-3\)\ M\^4 + 20\ M\^6 - 57\ M\^8 + 48\ M\^10 + 
                    41\ M\^12 - 84\ M\^14 + 65\ M\^16 + 56\ M\^18 - 
                    168\ M\^20 + 11\ M\^22 + 130\ M\^24 - 75\ M\^26 - 
                    20\ M\^28 + 30\ M\^30 - 9\ M\^32 + M\^34)\) + 
              L\^3\ \((1 - 10\ M\^2 + 33\ M\^4 - 36\ M\^6 - 13\ M\^8 + 
                    54\ M\^10 - 16\ M\^12 - 125\ M\^14 + 138\ M\^16 + 
                    113\ M\^18 - 110\ M\^20 + 30\ M\^22 - 107\ M\^24 - 
                    93\ M\^26 + 240\ M\^28 - 2\ M\^30 - 243\ M\^32 + 
                    111\ M\^34 + 77\ M\^36 - 75\ M\^38 + 21\ M\^40 - 
                    2\ M\^42)\) + 
              L\^7\ \((3\ M\^32 - M\^34 - 18\ M\^36 + 27\ M\^38 + 12\ M\^40 - 
                    24\ M\^42 + 10\ M\^44 - 3\ M\^46)\) + 
              L\^4\ \((\(-M\^4\) + 12\ M\^6 - 50\ M\^8 + 72\ M\^10 + 
                    30\ M\^12 - 124\ M\^14 - 22\ M\^16 + 66\ M\^18 + 
                    52\ M\^20 + 140\ M\^22 - 120\ M\^24 - 242\ M\^26 + 
                    242\ M\^28 + 120\ M\^30 - 140\ M\^32 - 52\ M\^34 - 
                    66\ M\^36 + 22\ M\^38 + 124\ M\^40 - 30\ M\^42 - 
                    72\ M\^44 + 50\ M\^46 - 12\ M\^48 + M\^50)\) + 
              L\^6\ \((\(-M\^20\) + 9\ M\^22 - 30\ M\^24 + 20\ M\^26 + 
                    75\ M\^28 - 130\ M\^30 - 11\ M\^32 + 168\ M\^34 - 
                    56\ M\^36 - 65\ M\^38 + 84\ M\^40 - 41\ M\^42 - 
                    48\ M\^44 + 57\ M\^46 - 20\ M\^48 + 3\ M\^50)\) + 
              L\^5\ \((2\ M\^12 - 21\ M\^14 + 75\ M\^16 - 77\ M\^18 - 
                    111\ M\^20 + 243\ M\^22 + 2\ M\^24 - 240\ M\^26 + 
                    93\ M\^28 + 107\ M\^30 - 30\ M\^32 + 110\ M\^34 - 
                    113\ M\^36 - 138\ M\^38 + 125\ M\^40 + 16\ M\^42 - 
                    54\ M\^44 + 13\ M\^46 + 36\ M\^48 - 33\ M\^50 + 
                    10\ M\^52 - M\^54)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[9, 49]\), \(\(-1\) + L\^8\ M\^58 + 
              L\ \((2 - 6\ M\^2 + 13\ M\^4 + 4\ M\^6 - 29\ M\^8 + 14\ M\^10 + 
                    3\ M\^12 - 3\ M\^14)\) + 
              L\^2\ \((\(-1\) + 6\ M\^2 - 23\ M\^4 + 41\ M\^6 - 44\ M\^8 - 
                    3\ M\^10 + 79\ M\^12 - 14\ M\^14 - 76\ M\^16 + 5\ M\^18 + 
                    64\ M\^20 - 16\ M\^22 - 41\ M\^24 + 33\ M\^26 - 
                    9\ M\^28 + M\^30)\) + 
              L\^3\ \((M\^4 - 9\ M\^6 + 40\ M\^8 - 73\ M\^10 + 49\ M\^12 + 
                    3\ M\^14 - 17\ M\^16 - 55\ M\^18 + 92\ M\^20 + 
                    46\ M\^22 - 114\ M\^24 - 8\ M\^26 + 118\ M\^28 - 
                    65\ M\^30 - 21\ M\^32 - 4\ M\^34 + 51\ M\^36 - 
                    37\ M\^38 + 10\ M\^40 - M\^42)\) + 
              L\^4\ \((2\ M\^10 - 18\ M\^12 + 50\ M\^14 - 54\ M\^16 + 
                    26\ M\^18 - 51\ M\^20 + 86\ M\^22 - M\^24 - 61\ M\^26 - 
                    60\ M\^28 + 60\ M\^30 + 61\ M\^32 + M\^34 - 86\ M\^36 + 
                    51\ M\^38 - 26\ M\^40 + 54\ M\^42 - 50\ M\^44 + 
                    18\ M\^46 - 2\ M\^48)\) + 
              L\^5\ \((M\^16 - 10\ M\^18 + 37\ M\^20 - 51\ M\^22 + 4\ M\^24 + 
                    21\ M\^26 + 65\ M\^28 - 118\ M\^30 + 8\ M\^32 + 
                    114\ M\^34 - 46\ M\^36 - 92\ M\^38 + 55\ M\^40 + 
                    17\ M\^42 - 3\ M\^44 - 49\ M\^46 + 73\ M\^48 - 
                    40\ M\^50 + 9\ M\^52 - M\^54)\) + 
              L\^7\ \((3\ M\^44 - 3\ M\^46 - 14\ M\^48 + 29\ M\^50 - 
                    4\ M\^52 - 13\ M\^54 + 6\ M\^56 - 2\ M\^58)\) + 
              L\^6\ \((\(-M\^28\) + 9\ M\^30 - 33\ M\^32 + 41\ M\^34 + 
                    16\ M\^36 - 64\ M\^38 - 5\ M\^40 + 76\ M\^42 + 
                    14\ M\^44 - 79\ M\^46 + 3\ M\^48 + 44\ M\^50 - 
                    41\ M\^52 + 23\ M\^54 - 6\ M\^56 + M\^58)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 1]\), \(M\^16 + L\^8\ M\^16 + 
              L\ \((\(-1\) + M\^2 + 5\ M\^16 + 7\ M\^18 - 4\ M\^20)\) + 
              L\^2\ \((3 - 9\ M\^2 + M\^4 + 5\ M\^6 + 15\ M\^16 + 25\ M\^18 + 
                    3\ M\^20 - 21\ M\^22 + 6\ M\^24)\) + 
              L\^3\ \((\(-3\) + 15\ M\^2 - 16\ M\^4 - 15\ M\^6 + 4\ M\^8 + 
                    15\ M\^10 + 35\ M\^16 + 45\ M\^18 + 8\ M\^20 - 
                    25\ M\^22 - 24\ M\^24 + 21\ M\^26 - 4\ M\^28)\) + 
              L\^5\ \((\(-4\)\ M\^4 + 21\ M\^6 - 24\ M\^8 - 25\ M\^10 + 
                    8\ M\^12 + 45\ M\^14 + 35\ M\^16 + 15\ M\^22 + 4\ M\^24 - 
                    15\ M\^26 - 16\ M\^28 + 15\ M\^30 - 3\ M\^32)\) + 
              L\^7\ \((\(-4\)\ M\^12 + 7\ M\^14 + 5\ M\^16 + M\^30 - 
                    M\^32)\) + 
              L\^4\ \((1 - 7\ M\^2 + 19\ M\^4 - 15\ M\^6 - 9\ M\^8 - 
                    13\ M\^10 - 11\ M\^12 + 35\ M\^14 + 70\ M\^16 + 
                    35\ M\^18 - 11\ M\^20 - 13\ M\^22 - 9\ M\^24 - 
                    15\ M\^26 + 19\ M\^28 - 7\ M\^30 + M\^32)\) + 
              L\^6\ \((6\ M\^8 - 21\ M\^10 + 3\ M\^12 + 25\ M\^14 + 
                    15\ M\^16 + 5\ M\^26 + M\^28 - 9\ M\^30 + 3\ M\^32)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 2]\), \(L\^11\ M\^4 + M\^142 + 
              L\^10\ \((\(-1\) + 2\ M\^2 - M\^4 - 4\ M\^14 + 9\ M\^16 + 
                    13\ M\^18 - 7\ M\^20)\) + 
              L\^9\ \((\(-3\)\ M\^10 + 11\ M\^12 - 15\ M\^14 + 16\ M\^18 - 
                    9\ M\^20 + 10\ M\^24 - 45\ M\^26 + 13\ M\^28 + 
                    119\ M\^30 + 9\ M\^32 - 72\ M\^34 + 21\ M\^36)\) + 
              L\^8\ \((\(-3\)\ M\^20 + 16\ M\^22 - 33\ M\^24 + 19\ M\^26 + 
                    22\ M\^28 - 18\ M\^30 - 61\ M\^32 + 83\ M\^34 + 
                    90\ M\^36 - 250\ M\^38 - 107\ M\^40 + 477\ M\^42 + 
                    262\ M\^44 - 327\ M\^46 - 135\ M\^48 + 165\ M\^50 - 
                    35\ M\^52)\) + 
              L\^7\ \((\(-M\^30\) + 7\ M\^32 - 19\ M\^34 + 27\ M\^36 - 
                    42\ M\^38 + 91\ M\^40 - 87\ M\^42 - 145\ M\^44 + 
                    306\ M\^46 + 233\ M\^48 - 729\ M\^50 - 452\ M\^52 + 
                    1134\ M\^54 + 949\ M\^56 - 677\ M\^58 - 695\ M\^60 + 
                    315\ M\^62 + 280\ M\^64 - 200\ M\^66 + 35\ M\^68)\) + 
              L\^6\ \((7\ M\^46 - 48\ M\^48 + 111\ M\^50 - 47\ M\^52 - 
                    133\ M\^54 - 93\ M\^56 + 533\ M\^58 + 284\ M\^60 - 
                    1239\ M\^62 - 809\ M\^64 + 1728\ M\^66 + 1668\ M\^68 - 
                    891\ M\^70 - 1259\ M\^72 + 291\ M\^74 + 580\ M\^76 - 
                    80\ M\^78 - 255\ M\^80 + 135\ M\^82 - 21\ M\^84)\) + 
              L\^5\ \((\(-21\)\ M\^62 + 135\ M\^64 - 255\ M\^66 - 80\ M\^68 + 
                    580\ M\^70 + 291\ M\^72 - 1259\ M\^74 - 891\ M\^76 + 
                    1668\ M\^78 + 1728\ M\^80 - 809\ M\^82 - 1239\ M\^84 + 
                    284\ M\^86 + 533\ M\^88 - 93\ M\^90 - 133\ M\^92 - 
                    47\ M\^94 + 111\ M\^96 - 48\ M\^98 + 7\ M\^100)\) + 
              L\^4\ \((35\ M\^78 - 200\ M\^80 + 280\ M\^82 + 315\ M\^84 - 
                    695\ M\^86 - 677\ M\^88 + 949\ M\^90 + 1134\ M\^92 - 
                    452\ M\^94 - 729\ M\^96 + 233\ M\^98 + 306\ M\^100 - 
                    145\ M\^102 - 87\ M\^104 + 91\ M\^106 - 42\ M\^108 + 
                    27\ M\^110 - 19\ M\^112 + 7\ M\^114 - M\^116)\) + 
              L\^3\ \((\(-35\)\ M\^94 + 165\ M\^96 - 135\ M\^98 - 
                    327\ M\^100 + 262\ M\^102 + 477\ M\^104 - 107\ M\^106 - 
                    250\ M\^108 + 90\ M\^110 + 83\ M\^112 - 61\ M\^114 - 
                    18\ M\^116 + 22\ M\^118 + 19\ M\^120 - 33\ M\^122 + 
                    16\ M\^124 - 3\ M\^126)\) + 
              L\^2\ \((21\ M\^110 - 72\ M\^112 + 9\ M\^114 + 119\ M\^116 + 
                    13\ M\^118 - 45\ M\^120 + 10\ M\^122 - 9\ M\^126 + 
                    16\ M\^128 - 15\ M\^132 + 11\ M\^134 - 3\ M\^136)\) + 
              L\ \((\(-7\)\ M\^126 + 13\ M\^128 + 9\ M\^130 - 4\ M\^132 - 
                    M\^142 + 2\ M\^144 - M\^146)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 3]\), \(M\^24 + L\^12\ M\^24 + 
              L\ \((\(-2\)\ M\^12 + 3\ M\^14 - 2\ M\^16 + M\^18 + 11\ M\^24 + 
                    3\ M\^26 - 4\ M\^28 + 5\ M\^30 - 3\ M\^32)\) + 
              L\^11\ \((\(-3\)\ M\^16 + 5\ M\^18 - 4\ M\^20 + 3\ M\^22 + 
                    11\ M\^24 + M\^30 - 2\ M\^32 + 3\ M\^34 - 2\ M\^36)\) + 
              L\^2\ \((1 - 3\ M\^2 + 3\ M\^4 - M\^6 - 4\ M\^12 - 22\ M\^14 + 
                    39\ M\^16 - 42\ M\^20 + 42\ M\^22 + 49\ M\^24 - M\^26 + 
                    15\ M\^28 + 7\ M\^30 - 12\ M\^32 - 12\ M\^34 + 
                    14\ M\^36 - 10\ M\^38 + 3\ M\^40)\) + 
              L\^7\ \((\(-21\) + 133\ M\^2 - 275\ M\^4 + 36\ M\^6 + 
                    601\ M\^8 - 415\ M\^10 - 1020\ M\^12 + 1198\ M\^14 + 
                    829\ M\^16 - 1989\ M\^18 - 386\ M\^20 + 2083\ M\^22 + 
                    342\ M\^24 - 143\ M\^26 + 104\ M\^28 - 843\ M\^30 + 
                    589\ M\^32 + 1060\ M\^34 - 1351\ M\^36 - 441\ M\^38 + 
                    1019\ M\^40 - 102\ M\^42 - 396\ M\^44 + 215\ M\^46 - 
                    35\ M\^48)\) + 
              L\^5\ \((\(-35\) + 215\ M\^2 - 396\ M\^4 - 102\ M\^6 + 
                    1019\ M\^8 - 441\ M\^10 - 1351\ M\^12 + 1060\ M\^14 + 
                    589\ M\^16 - 843\ M\^18 + 104\ M\^20 - 143\ M\^22 + 
                    342\ M\^24 + 2083\ M\^26 - 386\ M\^28 - 1989\ M\^30 + 
                    829\ M\^32 + 1198\ M\^34 - 1020\ M\^36 - 415\ M\^38 + 
                    601\ M\^40 + 36\ M\^42 - 275\ M\^44 + 133\ M\^46 - 
                    21\ M\^48)\) + 
              L\^9\ \((\(-1\) + 5\ M\^2 - 10\ M\^4 + 10\ M\^6 - 6\ M\^8 + 
                    34\ M\^10 - 87\ M\^12 + 35\ M\^14 + 70\ M\^16 - 
                    197\ M\^18 + 28\ M\^20 + 389\ M\^22 - 88\ M\^24 - 
                    151\ M\^26 + 300\ M\^28 + 10\ M\^30 - 187\ M\^32 + 
                    113\ M\^34 - 30\ M\^36 - 51\ M\^38 + 57\ M\^40 - 
                    8\ M\^42 - 39\ M\^44 + 31\ M\^46 - 7\ M\^48)\) + 
              L\^3\ \((\(-7\) + 31\ M\^2 - 39\ M\^4 - 8\ M\^6 + 57\ M\^8 - 
                    51\ M\^10 - 30\ M\^12 + 113\ M\^14 - 187\ M\^16 + 
                    10\ M\^18 + 300\ M\^20 - 151\ M\^22 - 88\ M\^24 + 
                    389\ M\^26 + 28\ M\^28 - 197\ M\^30 + 70\ M\^32 + 
                    35\ M\^34 - 87\ M\^36 + 34\ M\^38 - 6\ M\^40 + 
                    10\ M\^42 - 10\ M\^44 + 5\ M\^46 - M\^48)\) + 
              L\^10\ \((3\ M\^8 - 10\ M\^10 + 14\ M\^12 - 12\ M\^14 - 
                    12\ M\^16 + 7\ M\^18 + 15\ M\^20 - M\^22 + 49\ M\^24 + 
                    42\ M\^26 - 42\ M\^28 + 39\ M\^32 - 22\ M\^34 - 
                    4\ M\^36 - M\^42 + 3\ M\^44 - 3\ M\^46 + M\^48)\) + 
              L\^4\ \((21 - 115\ M\^2 + 182\ M\^4 + 61\ M\^6 - 404\ M\^8 + 
                    181\ M\^10 + 358\ M\^12 - 515\ M\^14 + 151\ M\^16 + 
                    124\ M\^18 - 554\ M\^20 + 748\ M\^22 + 752\ M\^24 - 
                    860\ M\^26 + 94\ M\^28 + 848\ M\^30 - 377\ M\^32 - 
                    477\ M\^34 + 284\ M\^36 + 83\ M\^38 - 103\ M\^40 - 
                    37\ M\^42 + 84\ M\^44 - 41\ M\^46 + 7\ M\^48)\) + 
              L\^8\ \((7 - 41\ M\^2 + 84\ M\^4 - 37\ M\^6 - 103\ M\^8 + 
                    83\ M\^10 + 284\ M\^12 - 477\ M\^14 - 377\ M\^16 + 
                    848\ M\^18 + 94\ M\^20 - 860\ M\^22 + 752\ M\^24 + 
                    748\ M\^26 - 554\ M\^28 + 124\ M\^30 + 151\ M\^32 - 
                    515\ M\^34 + 358\ M\^36 + 181\ M\^38 - 404\ M\^40 + 
                    61\ M\^42 + 182\ M\^44 - 115\ M\^46 + 21\ M\^48)\) + 
              L\^6\ \((35 - 225\ M\^2 + 451\ M\^4 + 41\ M\^6 - 1167\ M\^8 + 
                    613\ M\^10 + 1779\ M\^12 - 1629\ M\^14 - 1631\ M\^16 + 
                    1528\ M\^18 + 808\ M\^20 - 328\ M\^22 + 374\ M\^24 - 
                    328\ M\^26 + 808\ M\^28 + 1528\ M\^30 - 1631\ M\^32 - 
                    1629\ M\^34 + 1779\ M\^36 + 613\ M\^38 - 1167\ M\^40 + 
                    41\ M\^42 + 451\ M\^44 - 225\ M\^46 + 35\ M\^48)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 4]\), \(L\^13\ M\^18 + M\^72 + 
              L\^12\ \((\(-M\^6\) + 2\ M\^8 - 3\ M\^10 + 2\ M\^12 + 
                    4\ M\^18 - 8\ M\^20 + 9\ M\^22 + 17\ M\^24 - 
                    9\ M\^26)\) + 
              L\^11\ \((\(-1\) + 3\ M\^2 - 3\ M\^4 + 3\ M\^6 - 8\ M\^8 + 
                    15\ M\^10 - 10\ M\^12 - 30\ M\^14 + 35\ M\^16 + 
                    3\ M\^18 - 37\ M\^20 + 90\ M\^22 - 52\ M\^24 - 
                    75\ M\^26 + 202\ M\^28 + 33\ M\^30 - 126\ M\^32 + 
                    36\ M\^34)\) + 
              L\^10\ \((2 - 10\ M\^2 + 19\ M\^4 - 16\ M\^6 - 5\ M\^8 + 
                    26\ M\^10 - 11\ M\^12 - 23\ M\^14 + 81\ M\^16 - 
                    153\ M\^18 + 16\ M\^20 + 74\ M\^22 - 215\ M\^24 + 
                    466\ M\^26 + 154\ M\^28 - 941\ M\^30 + 601\ M\^32 + 
                    1043\ M\^34 - 762\ M\^36 - 382\ M\^38 + 406\ M\^40 - 
                    84\ M\^42)\) + 
              L\^9\ \((\(-1\) + 7\ M\^2 - 19\ M\^4 + 21\ M\^6 + 4\ M\^8 - 
                    13\ M\^10 - 80\ M\^12 + 195\ M\^14 - 70\ M\^16 - 
                    299\ M\^18 + 390\ M\^20 + 338\ M\^22 - 801\ M\^24 - 
                    414\ M\^26 + 83\ M\^28 + 1368\ M\^30 + 1125\ M\^32 - 
                    3019\ M\^34 - 455\ M\^36 + 4865\ M\^38 - 579\ M\^40 - 
                    3579\ M\^42 + 1145\ M\^44 + 1119\ M\^46 - 742\ M\^48 + 
                    126\ M\^50)\) + 
              L\^8\ \((9\ M\^8 - 66\ M\^10 + 177\ M\^12 - 135\ M\^14 - 
                    291\ M\^16 + 569\ M\^18 + 377\ M\^20 - 1783\ M\^22 + 
                    491\ M\^24 + 3017\ M\^26 - 1534\ M\^28 - 4105\ M\^30 + 
                    938\ M\^32 + 4235\ M\^34 + 2030\ M\^36 - 4963\ M\^38 - 
                    3703\ M\^40 + 8723\ M\^42 + 4253\ M\^44 - 9172\ M\^46 - 
                    1609\ M\^48 + 5456\ M\^50 - 691\ M\^52 - 1650\ M\^54 + 
                    840\ M\^56 - 126\ M\^58)\) + 
              L\^7\ \((\(-36\)\ M\^16 + 266\ M\^18 - 678\ M\^20 + 
                    288\ M\^22 + 1709\ M\^24 - 2243\ M\^26 - 2615\ M\^28 + 
                    6185\ M\^30 + 2013\ M\^32 - 9238\ M\^34 - 2065\ M\^36 + 
                    8652\ M\^38 + 3647\ M\^40 - 5824\ M\^42 - 5697\ M\^44 + 
                    7490\ M\^46 + 9533\ M\^48 - 8520\ M\^50 - 8485\ M\^52 + 
                    7458\ M\^54 + 3336\ M\^56 - 4273\ M\^58 - 60\ M\^60 + 
                    1391\ M\^62 - 602\ M\^64 + 84\ M\^66)\) + 
              L\^6\ \((84\ M\^24 - 602\ M\^26 + 1391\ M\^28 - 60\ M\^30 - 
                    4273\ M\^32 + 3336\ M\^34 + 7458\ M\^36 - 8485\ M\^38 - 
                    8520\ M\^40 + 9533\ M\^42 + 7490\ M\^44 - 5697\ M\^46 - 
                    5824\ M\^48 + 3647\ M\^50 + 8652\ M\^52 - 2065\ M\^54 - 
                    9238\ M\^56 + 2013\ M\^58 + 6185\ M\^60 - 2615\ M\^62 - 
                    2243\ M\^64 + 1709\ M\^66 + 288\ M\^68 - 678\ M\^70 + 
                    266\ M\^72 - 36\ M\^74)\) + 
              L\^5\ \((\(-126\)\ M\^32 + 840\ M\^34 - 1650\ M\^36 - 
                    691\ M\^38 + 5456\ M\^40 - 1609\ M\^42 - 9172\ M\^44 + 
                    4253\ M\^46 + 8723\ M\^48 - 3703\ M\^50 - 4963\ M\^52 + 
                    2030\ M\^54 + 4235\ M\^56 + 938\ M\^58 - 4105\ M\^60 - 
                    1534\ M\^62 + 3017\ M\^64 + 491\ M\^66 - 1783\ M\^68 + 
                    377\ M\^70 + 569\ M\^72 - 291\ M\^74 - 135\ M\^76 + 
                    177\ M\^78 - 66\ M\^80 + 9\ M\^82)\) + 
              L\ \((\(-9\)\ M\^64 + 17\ M\^66 + 9\ M\^68 - 8\ M\^70 + 
                    4\ M\^72 + 2\ M\^78 - 3\ M\^80 + 2\ M\^82 - M\^84)\) + 
              L\^2\ \((36\ M\^56 - 126\ M\^58 + 33\ M\^60 + 202\ M\^62 - 
                    75\ M\^64 - 52\ M\^66 + 90\ M\^68 - 37\ M\^70 + 
                    3\ M\^72 + 35\ M\^74 - 30\ M\^76 - 10\ M\^78 + 
                    15\ M\^80 - 8\ M\^82 + 3\ M\^84 - 3\ M\^86 + 3\ M\^88 - 
                    M\^90)\) + 
              L\^4\ \((126\ M\^40 - 742\ M\^42 + 1119\ M\^44 + 1145\ M\^46 - 
                    3579\ M\^48 - 579\ M\^50 + 4865\ M\^52 - 455\ M\^54 - 
                    3019\ M\^56 + 1125\ M\^58 + 1368\ M\^60 + 83\ M\^62 - 
                    414\ M\^64 - 801\ M\^66 + 338\ M\^68 + 390\ M\^70 - 
                    299\ M\^72 - 70\ M\^74 + 195\ M\^76 - 80\ M\^78 - 
                    13\ M\^80 + 4\ M\^82 + 21\ M\^84 - 19\ M\^86 + 7\ M\^88 - 
                    M\^90)\) + 
              L\^3\ \((\(-84\)\ M\^48 + 406\ M\^50 - 382\ M\^52 - 
                    762\ M\^54 + 1043\ M\^56 + 601\ M\^58 - 941\ M\^60 + 
                    154\ M\^62 + 466\ M\^64 - 215\ M\^66 + 74\ M\^68 + 
                    16\ M\^70 - 153\ M\^72 + 81\ M\^74 - 23\ M\^76 - 
                    11\ M\^78 + 26\ M\^80 - 5\ M\^82 - 16\ M\^84 + 
                    19\ M\^86 - 10\ M\^88 + 2\ M\^90)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 5]\), \(M\^12 + L\^15\ M\^138 + 
              L\ \((2\ M\^6 - 5\ M\^8 + 4\ M\^10 - M\^12 - 8\ M\^18 + 
                    23\ M\^20 + 11\ M\^22 - 17\ M\^24 + 6\ M\^26)\) + 
              L\^2\ \((1 - 4\ M\^2 + 6\ M\^4 - 4\ M\^6 + M\^8 - 4\ M\^12 - 
                    5\ M\^14 + 55\ M\^16 - 47\ M\^18 - 74\ M\^20 + 
                    126\ M\^22 - 23\ M\^24 - 199\ M\^26 + 272\ M\^28 + 
                    213\ M\^30 - 290\ M\^32 + 146\ M\^36 - 80\ M\^38 + 
                    15\ M\^40)\) + 
              L\^3\ \((4\ M\^6 - 29\ M\^8 + 77\ M\^10 - 71\ M\^12 - 
                    54\ M\^14 + 177\ M\^16 - 149\ M\^18 - 80\ M\^20 + 
                    446\ M\^22 - 463\ M\^24 - 255\ M\^26 + 900\ M\^28 - 
                    4\ M\^30 - 1286\ M\^32 - 235\ M\^34 + 1898\ M\^36 + 
                    870\ M\^38 - 1872\ M\^40 - 208\ M\^42 + 1395\ M\^44 - 
                    486\ M\^46 - 419\ M\^48 + 429\ M\^50 - 150\ M\^52 + 
                    20\ M\^54)\) + 
              L\^4\ \((6\ M\^12 - 63\ M\^14 + 246\ M\^16 - 381\ M\^18 - 
                    111\ M\^20 + 1087\ M\^22 - 714\ M\^24 - 1510\ M\^26 + 
                    1920\ M\^28 + 815\ M\^30 - 1640\ M\^32 - 324\ M\^34 + 
                    304\ M\^36 + 950\ M\^38 + 596\ M\^40 - 3362\ M\^42 + 
                    567\ M\^44 + 6188\ M\^46 - 1684\ M\^48 - 4447\ M\^50 + 
                    3491\ M\^52 + 827\ M\^54 - 2650\ M\^56 + 1320\ M\^58 + 
                    489\ M\^60 - 960\ M\^62 + 530\ M\^64 - 140\ M\^66 + 
                    15\ M\^68)\) + 
              L\^5\ \((4\ M\^18 - 55\ M\^20 + 288\ M\^22 - 663\ M\^24 + 
                    312\ M\^26 + 1374\ M\^28 - 1585\ M\^30 - 2342\ M\^32 + 
                    3920\ M\^34 + 2791\ M\^36 - 6570\ M\^38 - 2315\ M\^40 + 
                    8798\ M\^42 - 495\ M\^44 - 5254\ M\^46 + 1247\ M\^48 - 
                    5140\ M\^50 + 2874\ M\^52 + 16975\ M\^54 - 6796\ M\^56 - 
                    16888\ M\^58 + 11580\ M\^60 + 10285\ M\^62 - 
                    11362\ M\^64 - 1868\ M\^66 + 6189\ M\^68 - 2055\ M\^70 - 
                    1122\ M\^72 + 1470\ M\^74 - 840\ M\^76 + 305\ M\^78 - 
                    65\ M\^80 + 6\ M\^82)\) + 
              L\^6\ \((M\^24 - 17\ M\^26 + 116\ M\^28 - 395\ M\^30 + 
                    655\ M\^32 - 298\ M\^34 - 485\ M\^36 + 188\ M\^38 + 
                    988\ M\^40 + 661\ M\^42 - 4056\ M\^44 - 1785\ M\^46 + 
                    12254\ M\^48 - 1333\ M\^50 - 21497\ M\^52 + 
                    10867\ M\^54 + 28070\ M\^56 - 29318\ M\^58 - 
                    26408\ M\^60 + 47992\ M\^62 + 19718\ M\^64 - 
                    45184\ M\^66 - 3407\ M\^68 + 29660\ M\^70 - 7018\ M\^72 - 
                    12261\ M\^74 + 9330\ M\^76 + 590\ M\^78 - 5377\ M\^80 + 
                    2860\ M\^82 + 1037\ M\^84 - 1899\ M\^86 + 1011\ M\^88 - 
                    316\ M\^90 + 72\ M\^92 - 12\ M\^94 + M\^96)\) + 
              L\^7\ \((3\ M\^34 - 46\ M\^36 + 274\ M\^38 - 779\ M\^40 + 
                    955\ M\^42 + 85\ M\^44 - 1257\ M\^46 + 177\ M\^48 + 
                    2640\ M\^50 - 2352\ M\^52 - 4840\ M\^54 + 9322\ M\^56 + 
                    4008\ M\^58 - 23520\ M\^60 + 11109\ M\^62 + 
                    36627\ M\^64 - 39528\ M\^66 - 42117\ M\^68 + 
                    68222\ M\^70 + 38044\ M\^72 - 69942\ M\^74 - 
                    14171\ M\^76 + 54701\ M\^78 - 6699\ M\^80 - 
                    30790\ M\^82 + 12774\ M\^84 + 12964\ M\^86 - 
                    9994\ M\^88 - 2993\ M\^90 + 4782\ M\^92 - 625\ M\^94 - 
                    1401\ M\^96 + 1303\ M\^98 - 691\ M\^100 + 228\ M\^102 - 
                    41\ M\^104 + 3\ M\^106)\) + 
              L\^8\ \((3\ M\^44 - 41\ M\^46 + 228\ M\^48 - 691\ M\^50 + 
                    1303\ M\^52 - 1401\ M\^54 - 625\ M\^56 + 4782\ M\^58 - 
                    2993\ M\^60 - 9994\ M\^62 + 12964\ M\^64 + 12774\ M\^66 - 
                    30790\ M\^68 - 6699\ M\^70 + 54701\ M\^72 - 
                    14171\ M\^74 - 69942\ M\^76 + 38044\ M\^78 + 
                    68222\ M\^80 - 42117\ M\^82 - 39528\ M\^84 + 
                    36627\ M\^86 + 11109\ M\^88 - 23520\ M\^90 + 
                    4008\ M\^92 + 9322\ M\^94 - 4840\ M\^96 - 2352\ M\^98 + 
                    2640\ M\^100 + 177\ M\^102 - 1257\ M\^104 + 85\ M\^106 + 
                    955\ M\^108 - 779\ M\^110 + 274\ M\^112 - 46\ M\^114 + 
                    3\ M\^116)\) + 
              L\^9\ \((M\^54 - 12\ M\^56 + 72\ M\^58 - 316\ M\^60 + 
                    1011\ M\^62 - 1899\ M\^64 + 1037\ M\^66 + 2860\ M\^68 - 
                    5377\ M\^70 + 590\ M\^72 + 9330\ M\^74 - 12261\ M\^76 - 
                    7018\ M\^78 + 29660\ M\^80 - 3407\ M\^82 - 45184\ M\^84 + 
                    19718\ M\^86 + 47992\ M\^88 - 26408\ M\^90 - 
                    29318\ M\^92 + 28070\ M\^94 + 10867\ M\^96 - 
                    21497\ M\^98 - 1333\ M\^100 + 12254\ M\^102 - 
                    1785\ M\^104 - 4056\ M\^106 + 661\ M\^108 + 988\ M\^110 + 
                    188\ M\^112 - 485\ M\^114 - 298\ M\^116 + 655\ M\^118 - 
                    395\ M\^120 + 116\ M\^122 - 17\ M\^124 + M\^126)\) + 
              L\^10\ \((6\ M\^68 - 65\ M\^70 + 305\ M\^72 - 840\ M\^74 + 
                    1470\ M\^76 - 1122\ M\^78 - 2055\ M\^80 + 6189\ M\^82 - 
                    1868\ M\^84 - 11362\ M\^86 + 10285\ M\^88 + 
                    11580\ M\^90 - 16888\ M\^92 - 6796\ M\^94 + 
                    16975\ M\^96 + 2874\ M\^98 - 5140\ M\^100 + 
                    1247\ M\^102 - 5254\ M\^104 - 495\ M\^106 + 
                    8798\ M\^108 - 2315\ M\^110 - 6570\ M\^112 + 
                    2791\ M\^114 + 3920\ M\^116 - 2342\ M\^118 - 
                    1585\ M\^120 + 1374\ M\^122 + 312\ M\^124 - 663\ M\^126 + 
                    288\ M\^128 - 55\ M\^130 + 4\ M\^132)\) + 
              L\^11\ \((15\ M\^82 - 140\ M\^84 + 530\ M\^86 - 960\ M\^88 + 
                    489\ M\^90 + 1320\ M\^92 - 2650\ M\^94 + 827\ M\^96 + 
                    3491\ M\^98 - 4447\ M\^100 - 1684\ M\^102 + 
                    6188\ M\^104 + 567\ M\^106 - 3362\ M\^108 + 596\ M\^110 + 
                    950\ M\^112 + 304\ M\^114 - 324\ M\^116 - 1640\ M\^118 + 
                    815\ M\^120 + 1920\ M\^122 - 1510\ M\^124 - 714\ M\^126 + 
                    1087\ M\^128 - 111\ M\^130 - 381\ M\^132 + 246\ M\^134 - 
                    63\ M\^136 + 6\ M\^138)\) + 
              L\^14\ \((6\ M\^124 - 17\ M\^126 + 11\ M\^128 + 23\ M\^130 - 
                    8\ M\^132 - M\^138 + 4\ M\^140 - 5\ M\^142 + 
                    2\ M\^144)\) + 
              L\^12\ \((20\ M\^96 - 150\ M\^98 + 429\ M\^100 - 419\ M\^102 - 
                    486\ M\^104 + 1395\ M\^106 - 208\ M\^108 - 1872\ M\^110 + 
                    870\ M\^112 + 1898\ M\^114 - 235\ M\^116 - 1286\ M\^118 - 
                    4\ M\^120 + 900\ M\^122 - 255\ M\^124 - 463\ M\^126 + 
                    446\ M\^128 - 80\ M\^130 - 149\ M\^132 + 177\ M\^134 - 
                    54\ M\^136 - 71\ M\^138 + 77\ M\^140 - 29\ M\^142 + 
                    4\ M\^144)\) + 
              L\^13\ \((15\ M\^110 - 80\ M\^112 + 146\ M\^114 - 290\ M\^118 + 
                    213\ M\^120 + 272\ M\^122 - 199\ M\^124 - 23\ M\^126 + 
                    126\ M\^128 - 74\ M\^130 - 47\ M\^132 + 55\ M\^134 - 
                    5\ M\^136 - 4\ M\^138 + M\^142 - 4\ M\^144 + 6\ M\^146 - 
                    4\ M\^148 + M\^150)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 6]\), \(L\^18\ M\^8 + M\^156 + 
              L\^17\ \((\(-2\)\ M\^4 + 6\ M\^6 - 8\ M\^8 + 6\ M\^10 - 
                    2\ M\^12 - 9\ M\^14 + 31\ M\^16 + 2\ M\^18 - 14\ M\^20 + 
                    13\ M\^22 - 5\ M\^24)\) + 
              L\^16\ \((1 - 6\ M\^2 + 13\ M\^4 - 14\ M\^6 + 12\ M\^8 - 
                    11\ M\^10 + 13\ M\^12 - 79\ M\^14 + 103\ M\^16 + 
                    149\ M\^18 - 389\ M\^20 + 45\ M\^22 + 544\ M\^24 - 
                    248\ M\^26 - 140\ M\^28 + 320\ M\^30 - 134\ M\^32 - 
                    104\ M\^34 + 120\ M\^36 - 52\ M\^38 + 10\ M\^40)\) + 
              L\^15\ \((\(-1\) + 6\ M\^2 - 15\ M\^4 + 26\ M\^6 - 66\ M\^8 + 
                    148\ M\^10 - 134\ M\^12 - 111\ M\^14 + 338\ M\^16 - 
                    45\ M\^18 - 790\ M\^20 + 829\ M\^22 + 826\ M\^24 - 
                    2037\ M\^26 + 161\ M\^28 + 2669\ M\^30 - 1821\ M\^32 - 
                    1253\ M\^34 + 3606\ M\^36 + 99\ M\^38 - 2935\ M\^40 + 
                    969\ M\^42 + 1148\ M\^44 - 1014\ M\^46 - 31\ M\^48 + 
                    452\ M\^50 - 276\ M\^52 + 78\ M\^54 - 10\ M\^56)\) + 
              L\^14\ \((\(-6\)\ M\^6 + 51\ M\^8 - 179\ M\^10 + 316\ M\^12 - 
                    283\ M\^14 + 133\ M\^16 + 69\ M\^18 - 901\ M\^20 + 
                    1567\ M\^22 + 1701\ M\^24 - 6767\ M\^26 + 1339\ M\^28 + 
                    11446\ M\^30 - 8300\ M\^32 - 9017\ M\^34 + 13020\ M\^36 - 
                    2496\ M\^38 - 7643\ M\^40 + 14335\ M\^42 - 987\ M\^44 - 
                    11707\ M\^46 + 9598\ M\^48 + 4151\ M\^50 - 9525\ M\^52 - 
                    84\ M\^54 + 5158\ M\^56 - 1032\ M\^58 - 1710\ M\^60 + 
                    659\ M\^62 + 557\ M\^64 - 604\ M\^66 + 248\ M\^68 - 
                    52\ M\^70 + 5\ M\^72)\) + 
              L\^13\ \((\(-15\)\ M\^12 + 159\ M\^14 - 692\ M\^16 + 
                    1476\ M\^18 - 1181\ M\^20 - 1124\ M\^22 + 3280\ M\^24 - 
                    3300\ M\^26 + 540\ M\^28 + 10994\ M\^30 - 22241\ M\^32 - 
                    4306\ M\^34 + 48938\ M\^36 - 22871\ M\^38 - 
                    51806\ M\^40 + 49971\ M\^42 + 17937\ M\^44 - 
                    45816\ M\^46 + 19434\ M\^48 + 15632\ M\^50 - 
                    18422\ M\^52 + 24675\ M\^54 + 5438\ M\^56 - 
                    31506\ M\^58 + 6201\ M\^60 + 16274\ M\^62 - 
                    11040\ M\^64 - 1868\ M\^66 + 6486\ M\^68 - 1837\ M\^70 - 
                    1928\ M\^72 + 1291\ M\^74 - 370\ M\^76 + 453\ M\^78 - 
                    480\ M\^80 + 258\ M\^82 - 78\ M\^84 + 13\ M\^86 - 
                    M\^88)\) + 
              L\^12\ \((\(-20\)\ M\^18 + 246\ M\^20 - 1247\ M\^22 + 
                    3195\ M\^24 - 3639\ M\^26 - 1153\ M\^28 + 9119\ M\^30 - 
                    12273\ M\^32 + 1540\ M\^34 + 30261\ M\^36 - 
                    49117\ M\^38 - 14474\ M\^40 + 94805\ M\^42 - 
                    28670\ M\^44 - 85447\ M\^46 + 37215\ M\^48 + 
                    23337\ M\^50 + 39965\ M\^52 + 13265\ M\^54 - 
                    154014\ M\^56 + 11224\ M\^58 + 217910\ M\^60 - 
                    19859\ M\^62 - 143398\ M\^64 + 16470\ M\^66 + 
                    34076\ M\^68 - 1139\ M\^70 + 18921\ M\^72 - 
                    17715\ M\^74 - 15795\ M\^76 + 17839\ M\^78 + 
                    4237\ M\^80 - 10014\ M\^82 + 1152\ M\^84 + 2575\ M\^86 + 
                    14\ M\^88 - 1628\ M\^90 + 1092\ M\^92 - 345\ M\^94 + 
                    57\ M\^96 - 4\ M\^98)\) + 
              L\^11\ \((\(-15\)\ M\^24 + 204\ M\^26 - 1162\ M\^28 + 
                    3552\ M\^30 - 6034\ M\^32 + 4163\ M\^34 + 6525\ M\^36 - 
                    23944\ M\^38 + 23703\ M\^40 + 31702\ M\^42 - 
                    91216\ M\^44 + 15858\ M\^46 + 117147\ M\^48 - 
                    55004\ M\^50 - 57922\ M\^52 - 27892\ M\^54 - 
                    32103\ M\^56 + 245320\ M\^58 + 55728\ M\^60 - 
                    411618\ M\^62 - 40236\ M\^64 + 328720\ M\^66 + 
                    79716\ M\^68 + 17748\ M\^70 - 95497\ M\^72 - 
                    262465\ M\^74 + 102553\ M\^76 + 258165\ M\^78 - 
                    99228\ M\^80 - 128729\ M\^82 + 64174\ M\^84 + 
                    32751\ M\^86 - 27545\ M\^88 + 3989\ M\^90 + 2804\ M\^92 - 
                    4461\ M\^94 + 2318\ M\^96 + 1834\ M\^98 - 3142\ M\^100 + 
                    1841\ M\^102 - 562\ M\^104 + 90\ M\^106 - 6\ M\^108)\) + 
              L\^10\ \((\(-6\)\ M\^30 + 87\ M\^32 - 550\ M\^34 + 
                    2059\ M\^36 - 5205\ M\^38 + 8652\ M\^40 - 4719\ M\^42 - 
                    17493\ M\^44 + 39615\ M\^46 - 3639\ M\^48 - 
                    73951\ M\^50 + 61756\ M\^52 + 22614\ M\^54 - 
                    37279\ M\^56 + 104005\ M\^58 - 154949\ M\^60 - 
                    226489\ M\^62 + 405670\ M\^64 + 288948\ M\^66 - 
                    405024\ M\^68 - 343536\ M\^70 - 3416\ M\^72 + 
                    379025\ M\^74 + 551531\ M\^76 - 262941\ M\^78 - 
                    649276\ M\^80 + 136702\ M\^82 + 345636\ M\^84 - 
                    45772\ M\^86 - 24210\ M\^88 - 31847\ M\^90 - 
                    90867\ M\^92 + 68628\ M\^94 + 63404\ M\^96 - 
                    62977\ M\^98 - 7714\ M\^100 + 26794\ M\^102 - 
                    8685\ M\^104 - 4273\ M\^106 + 5956\ M\^108 - 
                    3662\ M\^110 + 1498\ M\^112 - 399\ M\^114 + 61\ M\^116 - 
                    4\ M\^118)\) + 
              L\^9\ \((\(-M\^36\) + 15\ M\^38 - 111\ M\^40 + 578\ M\^42 - 
                    2246\ M\^44 + 5865\ M\^46 - 8303\ M\^48 + 890\ M\^50 + 
                    16522\ M\^52 - 23914\ M\^54 + 7930\ M\^56 + 
                    33656\ M\^58 - 102865\ M\^60 + 62343\ M\^62 + 
                    211107\ M\^64 - 240776\ M\^66 - 276749\ M\^68 + 
                    312142\ M\^70 + 281289\ M\^72 - 25976\ M\^74 - 
                    251170\ M\^76 - 538213\ M\^78 + 124597\ M\^80 + 
                    875400\ M\^82 + 124597\ M\^84 - 538213\ M\^86 - 
                    251170\ M\^88 - 25976\ M\^90 + 281289\ M\^92 + 
                    312142\ M\^94 - 276749\ M\^96 - 240776\ M\^98 + 
                    211107\ M\^100 + 62343\ M\^102 - 102865\ M\^104 + 
                    33656\ M\^106 + 7930\ M\^108 - 23914\ M\^110 + 
                    16522\ M\^112 + 890\ M\^114 - 8303\ M\^116 + 
                    5865\ M\^118 - 2246\ M\^120 + 578\ M\^122 - 111\ M\^124 + 
                    15\ M\^126 - M\^128)\) + 
              L\^8\ \((\(-4\)\ M\^46 + 61\ M\^48 - 399\ M\^50 + 1498\ M\^52 - 
                    3662\ M\^54 + 5956\ M\^56 - 4273\ M\^58 - 8685\ M\^60 + 
                    26794\ M\^62 - 7714\ M\^64 - 62977\ M\^66 + 
                    63404\ M\^68 + 68628\ M\^70 - 90867\ M\^72 - 
                    31847\ M\^74 - 24210\ M\^76 - 45772\ M\^78 + 
                    345636\ M\^80 + 136702\ M\^82 - 649276\ M\^84 - 
                    262941\ M\^86 + 551531\ M\^88 + 379025\ M\^90 - 
                    3416\ M\^92 - 343536\ M\^94 - 405024\ M\^96 + 
                    288948\ M\^98 + 405670\ M\^100 - 226489\ M\^102 - 
                    154949\ M\^104 + 104005\ M\^106 - 37279\ M\^108 + 
                    22614\ M\^110 + 61756\ M\^112 - 73951\ M\^114 - 
                    3639\ M\^116 + 39615\ M\^118 - 17493\ M\^120 - 
                    4719\ M\^122 + 8652\ M\^124 - 5205\ M\^126 + 
                    2059\ M\^128 - 550\ M\^130 + 87\ M\^132 - 6\ M\^134)\) + 
              L\^7\ \((\(-6\)\ M\^56 + 90\ M\^58 - 562\ M\^60 + 1841\ M\^62 - 
                    3142\ M\^64 + 1834\ M\^66 + 2318\ M\^68 - 4461\ M\^70 + 
                    2804\ M\^72 + 3989\ M\^74 - 27545\ M\^76 + 32751\ M\^78 + 
                    64174\ M\^80 - 128729\ M\^82 - 99228\ M\^84 + 
                    258165\ M\^86 + 102553\ M\^88 - 262465\ M\^90 - 
                    95497\ M\^92 + 17748\ M\^94 + 79716\ M\^96 + 
                    328720\ M\^98 - 40236\ M\^100 - 411618\ M\^102 + 
                    55728\ M\^104 + 245320\ M\^106 - 32103\ M\^108 - 
                    27892\ M\^110 - 57922\ M\^112 - 55004\ M\^114 + 
                    117147\ M\^116 + 15858\ M\^118 - 91216\ M\^120 + 
                    31702\ M\^122 + 23703\ M\^124 - 23944\ M\^126 + 
                    6525\ M\^128 + 4163\ M\^130 - 6034\ M\^132 + 
                    3552\ M\^134 - 1162\ M\^136 + 204\ M\^138 - 
                    15\ M\^140)\) + 
              L\^6\ \((\(-4\)\ M\^66 + 57\ M\^68 - 345\ M\^70 + 1092\ M\^72 - 
                    1628\ M\^74 + 14\ M\^76 + 2575\ M\^78 + 1152\ M\^80 - 
                    10014\ M\^82 + 4237\ M\^84 + 17839\ M\^86 - 
                    15795\ M\^88 - 17715\ M\^90 + 18921\ M\^92 - 
                    1139\ M\^94 + 34076\ M\^96 + 16470\ M\^98 - 
                    143398\ M\^100 - 19859\ M\^102 + 217910\ M\^104 + 
                    11224\ M\^106 - 154014\ M\^108 + 13265\ M\^110 + 
                    39965\ M\^112 + 23337\ M\^114 + 37215\ M\^116 - 
                    85447\ M\^118 - 28670\ M\^120 + 94805\ M\^122 - 
                    14474\ M\^124 - 49117\ M\^126 + 30261\ M\^128 + 
                    1540\ M\^130 - 12273\ M\^132 + 9119\ M\^134 - 
                    1153\ M\^136 - 3639\ M\^138 + 3195\ M\^140 - 
                    1247\ M\^142 + 246\ M\^144 - 20\ M\^146)\) + 
              L\^5\ \((\(-M\^76\) + 13\ M\^78 - 78\ M\^80 + 258\ M\^82 - 
                    480\ M\^84 + 453\ M\^86 - 370\ M\^88 + 1291\ M\^90 - 
                    1928\ M\^92 - 1837\ M\^94 + 6486\ M\^96 - 1868\ M\^98 - 
                    11040\ M\^100 + 16274\ M\^102 + 6201\ M\^104 - 
                    31506\ M\^106 + 5438\ M\^108 + 24675\ M\^110 - 
                    18422\ M\^112 + 15632\ M\^114 + 19434\ M\^116 - 
                    45816\ M\^118 + 17937\ M\^120 + 49971\ M\^122 - 
                    51806\ M\^124 - 22871\ M\^126 + 48938\ M\^128 - 
                    4306\ M\^130 - 22241\ M\^132 + 10994\ M\^134 + 
                    540\ M\^136 - 3300\ M\^138 + 3280\ M\^140 - 
                    1124\ M\^142 - 1181\ M\^144 + 1476\ M\^146 - 
                    692\ M\^148 + 159\ M\^150 - 15\ M\^152)\) + 
              L\^4\ \((5\ M\^92 - 52\ M\^94 + 248\ M\^96 - 604\ M\^98 + 
                    557\ M\^100 + 659\ M\^102 - 1710\ M\^104 - 1032\ M\^106 + 
                    5158\ M\^108 - 84\ M\^110 - 9525\ M\^112 + 4151\ M\^114 + 
                    9598\ M\^116 - 11707\ M\^118 - 987\ M\^120 + 
                    14335\ M\^122 - 7643\ M\^124 - 2496\ M\^126 + 
                    13020\ M\^128 - 9017\ M\^130 - 8300\ M\^132 + 
                    11446\ M\^134 + 1339\ M\^136 - 6767\ M\^138 + 
                    1701\ M\^140 + 1567\ M\^142 - 901\ M\^144 + 69\ M\^146 + 
                    133\ M\^148 - 283\ M\^150 + 316\ M\^152 - 179\ M\^154 + 
                    51\ M\^156 - 6\ M\^158)\) + 
              L\ \((\(-5\)\ M\^140 + 13\ M\^142 - 14\ M\^144 + 2\ M\^146 + 
                    31\ M\^148 - 9\ M\^150 - 2\ M\^152 + 6\ M\^154 - 
                    8\ M\^156 + 6\ M\^158 - 2\ M\^160)\) + 
              L\^3\ \((\(-10\)\ M\^108 + 78\ M\^110 - 276\ M\^112 + 
                    452\ M\^114 - 31\ M\^116 - 1014\ M\^118 + 1148\ M\^120 + 
                    969\ M\^122 - 2935\ M\^124 + 99\ M\^126 + 3606\ M\^128 - 
                    1253\ M\^130 - 1821\ M\^132 + 2669\ M\^134 + 
                    161\ M\^136 - 2037\ M\^138 + 826\ M\^140 + 829\ M\^142 - 
                    790\ M\^144 - 45\ M\^146 + 338\ M\^148 - 111\ M\^150 - 
                    134\ M\^152 + 148\ M\^154 - 66\ M\^156 + 26\ M\^158 - 
                    15\ M\^160 + 6\ M\^162 - M\^164)\) + 
              L\^2\ \((10\ M\^124 - 52\ M\^126 + 120\ M\^128 - 104\ M\^130 - 
                    134\ M\^132 + 320\ M\^134 - 140\ M\^136 - 248\ M\^138 + 
                    544\ M\^140 + 45\ M\^142 - 389\ M\^144 + 149\ M\^146 + 
                    103\ M\^148 - 79\ M\^150 + 13\ M\^152 - 11\ M\^154 + 
                    12\ M\^156 - 14\ M\^158 + 13\ M\^160 - 6\ M\^162 + 
                    M\^164)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 7]\), \(L\^21\ M\^20 + M\^106 + 
              L\^20\ \((\(-5\)\ M\^16 + 18\ M\^18 - 23\ M\^20 - M\^22 + 
                    44\ M\^24 - 12\ M\^26 - 2\ M\^28 + 7\ M\^30 - 11\ M\^32 + 
                    9\ M\^34 - 3\ M\^36)\) + 
              L\^19\ \((10\ M\^12 - 72\ M\^14 + 206\ M\^16 - 231\ M\^18 - 
                    145\ M\^20 + 651\ M\^22 - 483\ M\^24 - 482\ M\^26 + 
                    1053\ M\^28 + 38\ M\^30 - 613\ M\^32 + 227\ M\^34 + 
                    141\ M\^36 - 73\ M\^38 - 23\ M\^40 - 6\ M\^42 + 
                    28\ M\^44 - 36\ M\^46 + 33\ M\^48 - 16\ M\^50 + 
                    3\ M\^52)\) + 
              L\^18\ \((\(-10\)\ M\^8 + 108\ M\^10 - 484\ M\^12 + 
                    1056\ M\^14 - 745\ M\^16 - 1511\ M\^18 + 3515\ M\^20 - 
                    906\ M\^22 - 4354\ M\^24 + 3993\ M\^26 + 1192\ M\^28 - 
                    3134\ M\^30 + 1604\ M\^32 + 1665\ M\^34 + 486\ M\^36 - 
                    1369\ M\^38 - 1192\ M\^40 + 2633\ M\^42 - 638\ M\^44 - 
                    1735\ M\^46 + 1480\ M\^48 + 26\ M\^50 - 668\ M\^52 + 
                    421\ M\^54 - 126\ M\^56 + 41\ M\^58 - 38\ M\^60 + 
                    35\ M\^62 - 21\ M\^64 + 7\ M\^66 - M\^68)\) + 
              L\^17\ \((5\ M\^4 - 72\ M\^6 + 446\ M\^8 - 1488\ M\^10 + 
                    2644\ M\^12 - 1462\ M\^14 - 3453\ M\^16 + 7655\ M\^18 - 
                    4972\ M\^20 - 4896\ M\^22 + 18017\ M\^24 - 19700\ M\^26 - 
                    9728\ M\^28 + 44678\ M\^30 - 27963\ M\^32 - 
                    35385\ M\^34 + 61155\ M\^36 - 3070\ M\^38 - 
                    39865\ M\^40 + 36299\ M\^42 - 5637\ M\^44 - 
                    26739\ M\^46 + 28704\ M\^48 - 370\ M\^50 - 18554\ M\^52 + 
                    9791\ M\^54 + 3848\ M\^56 - 5608\ M\^58 + 1189\ M\^60 + 
                    1192\ M\^62 - 625\ M\^64 - 442\ M\^66 + 654\ M\^68 - 
                    353\ M\^70 + 113\ M\^72 - 29\ M\^74 + 7\ M\^76 - 
                    M\^78)\) + 
              L\^16\ \((\(-1\) + 18\ M\^2 - 149\ M\^4 + 737\ M\^6 - 
                    2378\ M\^8 + 4988\ M\^10 - 5456\ M\^12 - 3204\ M\^14 + 
                    21839\ M\^16 - 24120\ M\^18 - 27199\ M\^20 + 
                    89698\ M\^22 - 32876\ M\^24 - 135874\ M\^26 + 
                    166339\ M\^28 + 65898\ M\^30 - 277844\ M\^32 + 
                    153923\ M\^34 + 199412\ M\^36 - 383218\ M\^38 + 
                    87763\ M\^40 + 428009\ M\^42 - 346476\ M\^44 - 
                    200145\ M\^46 + 400088\ M\^48 - 71656\ M\^50 - 
                    239254\ M\^52 + 183077\ M\^54 + 40701\ M\^56 - 
                    131022\ M\^58 + 49860\ M\^60 + 43721\ M\^62 - 
                    47252\ M\^64 + 3649\ M\^66 + 16914\ M\^68 - 8834\ M\^70 - 
                    1422\ M\^72 + 3628\ M\^74 - 2255\ M\^76 + 998\ M\^78 - 
                    346\ M\^80 + 78\ M\^82 - 8\ M\^84)\) + 
              L\^15\ \((4 - 73\ M\^2 + 586\ M\^4 - 2635\ M\^6 + 6852\ M\^8 - 
                    8519\ M\^10 - 3249\ M\^12 + 25105\ M\^14 - 19293\ M\^16 - 
                    39624\ M\^18 + 89654\ M\^20 - 27621\ M\^22 - 
                    155216\ M\^24 + 264753\ M\^26 + 63537\ M\^28 - 
                    689650\ M\^30 + 419259\ M\^32 + 1043625\ M\^34 - 
                    1286289\ M\^36 - 904398\ M\^38 + 2111064\ M\^40 + 
                    23825\ M\^42 - 2262943\ M\^44 + 1239980\ M\^46 + 
                    1558446\ M\^48 - 1945456\ M\^50 - 263088\ M\^52 + 
                    1707049\ M\^54 - 780898\ M\^56 - 839466\ M\^58 + 
                    1039098\ M\^60 + 22111\ M\^62 - 668717\ M\^64 + 
                    300805\ M\^66 + 207274\ M\^68 - 228496\ M\^70 + 
                    24150\ M\^72 + 63783\ M\^74 - 39375\ M\^76 + 
                    5522\ M\^78 + 7961\ M\^80 - 8825\ M\^82 + 5107\ M\^84 - 
                    1767\ M\^86 + 340\ M\^88 - 28\ M\^90)\) + 
              L\^14\ \((\(-6\) + 111\ M\^2 - 906\ M\^4 + 4123\ M\^6 - 
                    10560\ M\^8 + 11078\ M\^10 + 13861\ M\^12 - 
                    46696\ M\^14 - 9840\ M\^16 + 166615\ M\^18 - 
                    110668\ M\^20 - 353799\ M\^22 + 491891\ M\^24 + 
                    453610\ M\^26 - 1249437\ M\^28 + 3278\ M\^30 + 
                    2074451\ M\^32 - 1437587\ M\^34 - 2290528\ M\^36 + 
                    3758748\ M\^38 + 1212649\ M\^40 - 5907225\ M\^42 + 
                    1161850\ M\^44 + 6343491\ M\^46 - 4102123\ M\^48 - 
                    4233261\ M\^50 + 6439284\ M\^52 + 674028\ M\^54 - 
                    6599533\ M\^56 + 2693343\ M\^58 + 4533634\ M\^60 - 
                    4232090\ M\^62 - 1657080\ M\^64 + 3586639\ M\^66 - 
                    382835\ M\^68 - 1830289\ M\^70 + 914012\ M\^72 + 
                    405516\ M\^74 - 531312\ M\^76 + 138267\ M\^78 + 
                    85282\ M\^80 - 105337\ M\^82 + 46825\ M\^84 + 
                    8033\ M\^86 - 24127\ M\^88 + 14890\ M\^90 - 4728\ M\^92 + 
                    794\ M\^94 - 56\ M\^96)\) + 
              L\^13\ \((4 - 75\ M\^2 + 625\ M\^4 - 2932\ M\^6 + 7968\ M\^8 - 
                    10602\ M\^10 - 608\ M\^12 + 13224\ M\^14 + 25007\ M\^16 - 
                    76399\ M\^18 - 72782\ M\^20 + 311794\ M\^22 + 
                    125857\ M\^24 - 1014166\ M\^26 + 146518\ M\^28 + 
                    2229330\ M\^30 - 1202737\ M\^32 - 3384855\ M\^34 + 
                    3363425\ M\^36 + 2878636\ M\^38 - 5994798\ M\^40 + 
                    786903\ M\^42 + 7616899\ M\^44 - 7334069\ M\^46 - 
                    6603294\ M\^48 + 13961554\ M\^50 + 2313464\ M\^52 - 
                    17454615\ M\^54 + 4271370\ M\^56 + 16934092\ M\^58 - 
                    9719906\ M\^60 - 12636912\ M\^62 + 11761942\ M\^64 + 
                    6531893\ M\^66 - 10080210\ M\^68 - 1317585\ M\^70 + 
                    6208493\ M\^72 - 1336026\ M\^74 - 2378833\ M\^76 + 
                    1524218\ M\^78 + 173912\ M\^80 - 680174\ M\^82 + 
                    399998\ M\^84 + 13864\ M\^86 - 193974\ M\^88 + 
                    114563\ M\^90 + 8454\ M\^92 - 44227\ M\^94 + 
                    25586\ M\^96 - 7364\ M\^98 + 1110\ M\^100 - 
                    70\ M\^102)\) + 
              L\^12\ \((\(-1\) + 19\ M\^2 - 161\ M\^4 + 787\ M\^6 - 
                    2471\ M\^8 + 5671\ M\^10 - 12122\ M\^12 + 25268\ M\^14 - 
                    27997\ M\^16 - 42443\ M\^18 + 163127\ M\^20 - 
                    11398\ M\^22 - 536537\ M\^24 + 419586\ M\^26 + 
                    1123427\ M\^28 - 1432441\ M\^30 - 1796549\ M\^32 + 
                    3226902\ M\^34 + 1836640\ M\^36 - 5073187\ M\^38 - 
                    75655\ M\^40 + 5154701\ M\^42 - 4918694\ M\^44 - 
                    1096007\ M\^46 + 13325535\ M\^48 - 7618718\ M\^50 - 
                    22743231\ M\^52 + 18257567\ M\^54 + 28282171\ M\^56 - 
                    26425744\ M\^58 - 26197230\ M\^60 + 29755583\ M\^62 + 
                    18640626\ M\^64 - 26869280\ M\^66 - 9706072\ M\^68 + 
                    19576047\ M\^70 + 2722943\ M\^72 - 11339960\ M\^74 + 
                    967081\ M\^76 + 4904924\ M\^78 - 1920889\ M\^80 - 
                    1113885\ M\^82 + 1343987\ M\^84 - 375502\ M\^86 - 
                    480165\ M\^88 + 540092\ M\^90 - 56646\ M\^92 - 
                    220757\ M\^94 + 127271\ M\^96 + 18663\ M\^98 - 
                    51386\ M\^100 + 26480\ M\^102 - 6946\ M\^104 + 
                    962\ M\^106 - 56\ M\^108)\) + 
              L\ \((\(-3\)\ M\^90 + 9\ M\^92 - 11\ M\^94 + 7\ M\^96 - 
                    2\ M\^98 - 12\ M\^100 + 44\ M\^102 - M\^104 - 
                    23\ M\^106 + 18\ M\^108 - 5\ M\^110)\) + 
              L\^11\ \((\(-8\)\ M\^6 + 150\ M\^8 - 1219\ M\^10 + 
                    5473\ M\^12 - 14044\ M\^14 + 17944\ M\^16 - 3477\ M\^18 - 
                    4522\ M\^20 - 36461\ M\^22 + 6666\ M\^24 + 
                    247117\ M\^26 - 203964\ M\^28 - 775842\ M\^30 + 
                    1139858\ M\^32 + 1134556\ M\^34 - 2806905\ M\^36 - 
                    426697\ M\^38 + 3916228\ M\^40 - 2243678\ M\^42 - 
                    1349508\ M\^44 + 6551680\ M\^46 - 8013734\ M\^48 - 
                    11244317\ M\^50 + 24611871\ M\^52 + 14883509\ M\^54 - 
                    43661544\ M\^56 - 16533490\ M\^58 + 56450703\ M\^60 + 
                    14962396\ M\^62 - 56038598\ M\^64 - 8993458\ M\^66 + 
                    44374163\ M\^68 + 2005304\ M\^70 - 27693521\ M\^72 + 
                    2234915\ M\^74 + 12833519\ M\^76 - 2912838\ M\^78 - 
                    3646278\ M\^80 + 1707053\ M\^82 - 4000\ M\^84 - 
                    609227\ M\^86 + 685570\ M\^88 + 137464\ M\^90 - 
                    488232\ M\^92 - 36241\ M\^94 + 301252\ M\^96 - 
                    51749\ M\^98 - 122829\ M\^100 + 58732\ M\^102 + 
                    26187\ M\^104 - 36033\ M\^106 + 16259\ M\^108 - 
                    3919\ M\^110 + 508\ M\^112 - 28\ M\^114)\) + 
              L\^2\ \((3\ M\^74 - 16\ M\^76 + 33\ M\^78 - 36\ M\^80 + 
                    28\ M\^82 - 6\ M\^84 - 23\ M\^86 - 73\ M\^88 + 
                    141\ M\^90 + 227\ M\^92 - 613\ M\^94 + 38\ M\^96 + 
                    1053\ M\^98 - 482\ M\^100 - 483\ M\^102 + 651\ M\^104 - 
                    145\ M\^106 - 231\ M\^108 + 206\ M\^110 - 72\ M\^112 + 
                    10\ M\^114)\) + 
              L\^3\ \((\(-M\^58\) + 7\ M\^60 - 21\ M\^62 + 35\ M\^64 - 
                    38\ M\^66 + 41\ M\^68 - 126\ M\^70 + 421\ M\^72 - 
                    668\ M\^74 + 26\ M\^76 + 1480\ M\^78 - 1735\ M\^80 - 
                    638\ M\^82 + 2633\ M\^84 - 1192\ M\^86 - 1369\ M\^88 + 
                    486\ M\^90 + 1665\ M\^92 + 1604\ M\^94 - 3134\ M\^96 + 
                    1192\ M\^98 + 3993\ M\^100 - 4354\ M\^102 - 906\ M\^104 + 
                    3515\ M\^106 - 1511\ M\^108 - 745\ M\^110 + 
                    1056\ M\^112 - 484\ M\^114 + 108\ M\^116 - 
                    10\ M\^118)\) + 
              L\^10\ \((\(-28\)\ M\^12 + 508\ M\^14 - 3919\ M\^16 + 
                    16259\ M\^18 - 36033\ M\^20 + 26187\ M\^22 + 
                    58732\ M\^24 - 122829\ M\^26 - 51749\ M\^28 + 
                    301252\ M\^30 - 36241\ M\^32 - 488232\ M\^34 + 
                    137464\ M\^36 + 685570\ M\^38 - 609227\ M\^40 - 
                    4000\ M\^42 + 1707053\ M\^44 - 3646278\ M\^46 - 
                    2912838\ M\^48 + 12833519\ M\^50 + 2234915\ M\^52 - 
                    27693521\ M\^54 + 2005304\ M\^56 + 44374163\ M\^58 - 
                    8993458\ M\^60 - 56038598\ M\^62 + 14962396\ M\^64 + 
                    56450703\ M\^66 - 16533490\ M\^68 - 43661544\ M\^70 + 
                    14883509\ M\^72 + 24611871\ M\^74 - 11244317\ M\^76 - 
                    8013734\ M\^78 + 6551680\ M\^80 - 1349508\ M\^82 - 
                    2243678\ M\^84 + 3916228\ M\^86 - 426697\ M\^88 - 
                    2806905\ M\^90 + 1134556\ M\^92 + 1139858\ M\^94 - 
                    775842\ M\^96 - 203964\ M\^98 + 247117\ M\^100 + 
                    6666\ M\^102 - 36461\ M\^104 - 4522\ M\^106 - 
                    3477\ M\^108 + 17944\ M\^110 - 14044\ M\^112 + 
                    5473\ M\^114 - 1219\ M\^116 + 150\ M\^118 - 
                    8\ M\^120)\) + 
              L\^4\ \((\(-M\^48\) + 7\ M\^50 - 29\ M\^52 + 113\ M\^54 - 
                    353\ M\^56 + 654\ M\^58 - 442\ M\^60 - 625\ M\^62 + 
                    1192\ M\^64 + 1189\ M\^66 - 5608\ M\^68 + 3848\ M\^70 + 
                    9791\ M\^72 - 18554\ M\^74 - 370\ M\^76 + 28704\ M\^78 - 
                    26739\ M\^80 - 5637\ M\^82 + 36299\ M\^84 - 
                    39865\ M\^86 - 3070\ M\^88 + 61155\ M\^90 - 
                    35385\ M\^92 - 27963\ M\^94 + 44678\ M\^96 - 
                    9728\ M\^98 - 19700\ M\^100 + 18017\ M\^102 - 
                    4896\ M\^104 - 4972\ M\^106 + 7655\ M\^108 - 
                    3453\ M\^110 - 1462\ M\^112 + 2644\ M\^114 - 
                    1488\ M\^116 + 446\ M\^118 - 72\ M\^120 + 5\ M\^122)\) + 
              L\^7\ \((\(-56\)\ M\^30 + 794\ M\^32 - 4728\ M\^34 + 
                    14890\ M\^36 - 24127\ M\^38 + 8033\ M\^40 + 
                    46825\ M\^42 - 105337\ M\^44 + 85282\ M\^46 + 
                    138267\ M\^48 - 531312\ M\^50 + 405516\ M\^52 + 
                    914012\ M\^54 - 1830289\ M\^56 - 382835\ M\^58 + 
                    3586639\ M\^60 - 1657080\ M\^62 - 4232090\ M\^64 + 
                    4533634\ M\^66 + 2693343\ M\^68 - 6599533\ M\^70 + 
                    674028\ M\^72 + 6439284\ M\^74 - 4233261\ M\^76 - 
                    4102123\ M\^78 + 6343491\ M\^80 + 1161850\ M\^82 - 
                    5907225\ M\^84 + 1212649\ M\^86 + 3758748\ M\^88 - 
                    2290528\ M\^90 - 1437587\ M\^92 + 2074451\ M\^94 + 
                    3278\ M\^96 - 1249437\ M\^98 + 453610\ M\^100 + 
                    491891\ M\^102 - 353799\ M\^104 - 110668\ M\^106 + 
                    166615\ M\^108 - 9840\ M\^110 - 46696\ M\^112 + 
                    13861\ M\^114 + 11078\ M\^116 - 10560\ M\^118 + 
                    4123\ M\^120 - 906\ M\^122 + 111\ M\^124 - 6\ M\^126)\) + 
              L\^5\ \((\(-8\)\ M\^42 + 78\ M\^44 - 346\ M\^46 + 998\ M\^48 - 
                    2255\ M\^50 + 3628\ M\^52 - 1422\ M\^54 - 8834\ M\^56 + 
                    16914\ M\^58 + 3649\ M\^60 - 47252\ M\^62 + 
                    43721\ M\^64 + 49860\ M\^66 - 131022\ M\^68 + 
                    40701\ M\^70 + 183077\ M\^72 - 239254\ M\^74 - 
                    71656\ M\^76 + 400088\ M\^78 - 200145\ M\^80 - 
                    346476\ M\^82 + 428009\ M\^84 + 87763\ M\^86 - 
                    383218\ M\^88 + 199412\ M\^90 + 153923\ M\^92 - 
                    277844\ M\^94 + 65898\ M\^96 + 166339\ M\^98 - 
                    135874\ M\^100 - 32876\ M\^102 + 89698\ M\^104 - 
                    27199\ M\^106 - 24120\ M\^108 + 21839\ M\^110 - 
                    3204\ M\^112 - 5456\ M\^114 + 4988\ M\^116 - 
                    2378\ M\^118 + 737\ M\^120 - 149\ M\^122 + 18\ M\^124 - 
                    M\^126)\) + 
              L\^9\ \((\(-56\)\ M\^18 + 962\ M\^20 - 6946\ M\^22 + 
                    26480\ M\^24 - 51386\ M\^26 + 18663\ M\^28 + 
                    127271\ M\^30 - 220757\ M\^32 - 56646\ M\^34 + 
                    540092\ M\^36 - 480165\ M\^38 - 375502\ M\^40 + 
                    1343987\ M\^42 - 1113885\ M\^44 - 1920889\ M\^46 + 
                    4904924\ M\^48 + 967081\ M\^50 - 11339960\ M\^52 + 
                    2722943\ M\^54 + 19576047\ M\^56 - 9706072\ M\^58 - 
                    26869280\ M\^60 + 18640626\ M\^62 + 29755583\ M\^64 - 
                    26197230\ M\^66 - 26425744\ M\^68 + 28282171\ M\^70 + 
                    18257567\ M\^72 - 22743231\ M\^74 - 7618718\ M\^76 + 
                    13325535\ M\^78 - 1096007\ M\^80 - 4918694\ M\^82 + 
                    5154701\ M\^84 - 75655\ M\^86 - 5073187\ M\^88 + 
                    1836640\ M\^90 + 3226902\ M\^92 - 1796549\ M\^94 - 
                    1432441\ M\^96 + 1123427\ M\^98 + 419586\ M\^100 - 
                    536537\ M\^102 - 11398\ M\^104 + 163127\ M\^106 - 
                    42443\ M\^108 - 27997\ M\^110 + 25268\ M\^112 - 
                    12122\ M\^114 + 5671\ M\^116 - 2471\ M\^118 + 
                    787\ M\^120 - 161\ M\^122 + 19\ M\^124 - M\^126)\) + 
              L\^8\ \((\(-70\)\ M\^24 + 1110\ M\^26 - 7364\ M\^28 + 
                    25586\ M\^30 - 44227\ M\^32 + 8454\ M\^34 + 
                    114563\ M\^36 - 193974\ M\^38 + 13864\ M\^40 + 
                    399998\ M\^42 - 680174\ M\^44 + 173912\ M\^46 + 
                    1524218\ M\^48 - 2378833\ M\^50 - 1336026\ M\^52 + 
                    6208493\ M\^54 - 1317585\ M\^56 - 10080210\ M\^58 + 
                    6531893\ M\^60 + 11761942\ M\^62 - 12636912\ M\^64 - 
                    9719906\ M\^66 + 16934092\ M\^68 + 4271370\ M\^70 - 
                    17454615\ M\^72 + 2313464\ M\^74 + 13961554\ M\^76 - 
                    6603294\ M\^78 - 7334069\ M\^80 + 7616899\ M\^82 + 
                    786903\ M\^84 - 5994798\ M\^86 + 2878636\ M\^88 + 
                    3363425\ M\^90 - 3384855\ M\^92 - 1202737\ M\^94 + 
                    2229330\ M\^96 + 146518\ M\^98 - 1014166\ M\^100 + 
                    125857\ M\^102 + 311794\ M\^104 - 72782\ M\^106 - 
                    76399\ M\^108 + 25007\ M\^110 + 13224\ M\^112 - 
                    608\ M\^114 - 10602\ M\^116 + 7968\ M\^118 - 
                    2932\ M\^120 + 625\ M\^122 - 75\ M\^124 + 4\ M\^126)\) + 
              L\^6\ \((\(-28\)\ M\^36 + 340\ M\^38 - 1767\ M\^40 + 
                    5107\ M\^42 - 8825\ M\^44 + 7961\ M\^46 + 5522\ M\^48 - 
                    39375\ M\^50 + 63783\ M\^52 + 24150\ M\^54 - 
                    228496\ M\^56 + 207274\ M\^58 + 300805\ M\^60 - 
                    668717\ M\^62 + 22111\ M\^64 + 1039098\ M\^66 - 
                    839466\ M\^68 - 780898\ M\^70 + 1707049\ M\^72 - 
                    263088\ M\^74 - 1945456\ M\^76 + 1558446\ M\^78 + 
                    1239980\ M\^80 - 2262943\ M\^82 + 23825\ M\^84 + 
                    2111064\ M\^86 - 904398\ M\^88 - 1286289\ M\^90 + 
                    1043625\ M\^92 + 419259\ M\^94 - 689650\ M\^96 + 
                    63537\ M\^98 + 264753\ M\^100 - 155216\ M\^102 - 
                    27621\ M\^104 + 89654\ M\^106 - 39624\ M\^108 - 
                    19293\ M\^110 + 25105\ M\^112 - 3249\ M\^114 - 
                    8519\ M\^116 + 6852\ M\^118 - 2635\ M\^120 + 
                    586\ M\^122 - 73\ M\^124 + 4\ M\^126)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 8]\), \(L\^14\ M\^8 + M\^124 + 
              L\^13\ \((\(-1\) + 2\ M\^2 - 3\ M\^4 + 4\ M\^6 - 2\ M\^8 - 
                    3\ M\^10 + 7\ M\^12 - 8\ M\^14 + 9\ M\^16 + 19\ M\^18 - 
                    10\ M\^20)\) + 
              L\^12\ \((1 - 6\ M\^2 + 13\ M\^4 - 16\ M\^6 + 15\ M\^8 + 
                    M\^10 - 55\ M\^12 + 66\ M\^14 + 24\ M\^16 - 119\ M\^18 + 
                    114\ M\^20 - 7\ M\^22 - 117\ M\^24 + 240\ M\^26 + 
                    51\ M\^28 - 159\ M\^30 + 45\ M\^32)\) + 
              L\^11\ \((2\ M\^2 - 12\ M\^4 + 29\ M\^6 - 33\ M\^8 + 6\ M\^10 + 
                    46\ M\^12 - 102\ M\^14 + 42\ M\^16 + 268\ M\^18 - 
                    487\ M\^20 + 42\ M\^22 + 605\ M\^24 - 750\ M\^26 + 
                    195\ M\^28 + 816\ M\^30 - 1247\ M\^32 + 486\ M\^34 + 
                    1602\ M\^36 - 1021\ M\^38 - 591\ M\^40 + 588\ M\^42 - 
                    120\ M\^44)\) + 
              L\^10\ \((M\^4 - 7\ M\^6 + 19\ M\^8 - 21\ M\^10 - 6\ M\^12 + 
                    37\ M\^14 + 13\ M\^16 - 218\ M\^18 + 428\ M\^20 - 
                    149\ M\^22 - 793\ M\^24 + 1142\ M\^26 + 375\ M\^28 - 
                    1861\ M\^30 + 1076\ M\^32 - 267\ M\^34 - 853\ M\^36 + 
                    3503\ M\^38 - 2808\ M\^40 - 2876\ M\^42 + 7633\ M\^44 + 
                    245\ M\^46 - 6284\ M\^48 + 1732\ M\^50 + 1990\ M\^52 - 
                    1260\ M\^54 + 210\ M\^56)\) + 
              L\^9\ \((\(-10\)\ M\^16 + 75\ M\^18 - 210\ M\^20 + 194\ M\^22 + 
                    279\ M\^24 - 801\ M\^26 + 126\ M\^28 + 1996\ M\^30 - 
                    2757\ M\^32 - 1193\ M\^34 + 5677\ M\^36 - 1608\ M\^38 - 
                    5481\ M\^40 + 2179\ M\^42 + 1200\ M\^44 + 3624\ M\^46 - 
                    1439\ M\^48 - 10161\ M\^50 + 12007\ M\^52 + 
                    12478\ M\^54 - 17187\ M\^56 - 5375\ M\^58 + 
                    11566\ M\^60 - 1128\ M\^62 - 3519\ M\^64 + 1722\ M\^66 - 
                    252\ M\^68)\) + 
              L\^8\ \((45\ M\^28 - 348\ M\^30 + 966\ M\^32 - 678\ M\^34 - 
                    2053\ M\^36 + 4138\ M\^38 + 1146\ M\^40 - 10021\ M\^42 + 
                    5559\ M\^44 + 11682\ M\^46 - 12064\ M\^48 - 7602\ M\^50 + 
                    9216\ M\^52 + 3848\ M\^54 - 1140\ M\^56 - 9588\ M\^58 + 
                    5273\ M\^60 + 24750\ M\^62 - 13620\ M\^64 - 
                    25054\ M\^66 + 17397\ M\^68 + 10788\ M\^70 - 
                    11658\ M\^72 - 361\ M\^74 + 3726\ M\^76 - 1554\ M\^78 + 
                    210\ M\^80)\) + 
              L\^7\ \((\(-120\)\ M\^40 + 924\ M\^42 - 2441\ M\^44 + 
                    1051\ M\^46 + 6664\ M\^48 - 9510\ M\^50 - 7980\ M\^52 + 
                    22979\ M\^54 + 1099\ M\^56 - 27392\ M\^58 + 4604\ M\^60 + 
                    16366\ M\^62 - 1826\ M\^64 - 5404\ M\^66 - 1826\ M\^68 + 
                    16366\ M\^70 + 4604\ M\^72 - 27392\ M\^74 + 1099\ M\^76 + 
                    22979\ M\^78 - 7980\ M\^80 - 9510\ M\^82 + 6664\ M\^84 + 
                    1051\ M\^86 - 2441\ M\^88 + 924\ M\^90 - 120\ M\^92)\) + 
              L\^6\ \((210\ M\^52 - 1554\ M\^54 + 3726\ M\^56 - 361\ M\^58 - 
                    11658\ M\^60 + 10788\ M\^62 + 17397\ M\^64 - 
                    25054\ M\^66 - 13620\ M\^68 + 24750\ M\^70 + 
                    5273\ M\^72 - 9588\ M\^74 - 1140\ M\^76 + 3848\ M\^78 + 
                    9216\ M\^80 - 7602\ M\^82 - 12064\ M\^84 + 11682\ M\^86 + 
                    5559\ M\^88 - 10021\ M\^90 + 1146\ M\^92 + 4138\ M\^94 - 
                    2053\ M\^96 - 678\ M\^98 + 966\ M\^100 - 348\ M\^102 + 
                    45\ M\^104)\) + 
              L\^5\ \((\(-252\)\ M\^64 + 1722\ M\^66 - 3519\ M\^68 - 
                    1128\ M\^70 + 11566\ M\^72 - 5375\ M\^74 - 17187\ M\^76 + 
                    12478\ M\^78 + 12007\ M\^80 - 10161\ M\^82 - 
                    1439\ M\^84 + 3624\ M\^86 + 1200\ M\^88 + 2179\ M\^90 - 
                    5481\ M\^92 - 1608\ M\^94 + 5677\ M\^96 - 1193\ M\^98 - 
                    2757\ M\^100 + 1996\ M\^102 + 126\ M\^104 - 801\ M\^106 + 
                    279\ M\^108 + 194\ M\^110 - 210\ M\^112 + 75\ M\^114 - 
                    10\ M\^116)\) + 
              L\^4\ \((210\ M\^76 - 1260\ M\^78 + 1990\ M\^80 + 1732\ M\^82 - 
                    6284\ M\^84 + 245\ M\^86 + 7633\ M\^88 - 2876\ M\^90 - 
                    2808\ M\^92 + 3503\ M\^94 - 853\ M\^96 - 267\ M\^98 + 
                    1076\ M\^100 - 1861\ M\^102 + 375\ M\^104 + 
                    1142\ M\^106 - 793\ M\^108 - 149\ M\^110 + 428\ M\^112 - 
                    218\ M\^114 + 13\ M\^116 + 37\ M\^118 - 6\ M\^120 - 
                    21\ M\^122 + 19\ M\^124 - 7\ M\^126 + M\^128)\) + 
              L\^3\ \((\(-120\)\ M\^88 + 588\ M\^90 - 591\ M\^92 - 
                    1021\ M\^94 + 1602\ M\^96 + 486\ M\^98 - 1247\ M\^100 + 
                    816\ M\^102 + 195\ M\^104 - 750\ M\^106 + 605\ M\^108 + 
                    42\ M\^110 - 487\ M\^112 + 268\ M\^114 + 42\ M\^116 - 
                    102\ M\^118 + 46\ M\^120 + 6\ M\^122 - 33\ M\^124 + 
                    29\ M\^126 - 12\ M\^128 + 2\ M\^130)\) + 
              L\ \((\(-10\)\ M\^112 + 19\ M\^114 + 9\ M\^116 - 8\ M\^118 + 
                    7\ M\^120 - 3\ M\^122 - 2\ M\^124 + 4\ M\^126 - 
                    3\ M\^128 + 2\ M\^130 - M\^132)\) + 
              L\^2\ \((45\ M\^100 - 159\ M\^102 + 51\ M\^104 + 240\ M\^106 - 
                    117\ M\^108 - 7\ M\^110 + 114\ M\^112 - 119\ M\^114 + 
                    24\ M\^116 + 66\ M\^118 - 55\ M\^120 + M\^122 + 
                    15\ M\^124 - 16\ M\^126 + 13\ M\^128 - 6\ M\^130 + 
                    M\^132)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 9]\), \(M\^24 + L\^18\ M\^96 + 
              L\ \((\(-3\)\ M\^16 + 8\ M\^18 - 8\ M\^20 + 4\ M\^22 - M\^24 - 
                    9\ M\^26 + 33\ M\^28 + M\^30 - 16\ M\^32 + 14\ M\^34 - 
                    5\ M\^36)\) + 
              L\^2\ \((3\ M\^8 - 14\ M\^10 + 26\ M\^12 - 24\ M\^14 + 
                    11\ M\^16 + 10\ M\^18 - 39\ M\^20 - 24\ M\^22 + 
                    116\ M\^24 + 52\ M\^26 - 297\ M\^28 + 42\ M\^30 + 
                    504\ M\^32 - 168\ M\^34 - 215\ M\^36 + 290\ M\^38 - 
                    83\ M\^40 - 112\ M\^42 + 117\ M\^44 - 52\ M\^46 + 
                    10\ M\^48)\) + 
              L\^3\ \((\(-1\) + 6\ M\^2 - 15\ M\^4 + 20\ M\^6 - 15\ M\^8 + 
                    9\ M\^10 - 58\ M\^12 + 255\ M\^14 - 414\ M\^16 - 
                    11\ M\^18 + 950\ M\^20 - 1172\ M\^22 - 260\ M\^24 + 
                    2093\ M\^26 - 1982\ M\^28 - 581\ M\^30 + 2654\ M\^32 - 
                    1748\ M\^34 - 359\ M\^36 + 2522\ M\^38 - 665\ M\^40 - 
                    526\ M\^42 + 104\ M\^44 - 194\ M\^46 + 481\ M\^48 - 
                    295\ M\^50 - 182\ M\^52 + 366\ M\^54 - 228\ M\^56 + 
                    72\ M\^58 - 10\ M\^60)\) + 
              L\^4\ \((\(-6\)\ M\^2 + 53\ M\^4 - 192\ M\^6 + 335\ M\^8 - 
                    166\ M\^10 - 460\ M\^12 + 852\ M\^14 + 395\ M\^16 - 
                    3104\ M\^18 + 2789\ M\^20 + 4332\ M\^22 - 9787\ M\^24 + 
                    168\ M\^26 + 15350\ M\^28 - 12158\ M\^30 - 8172\ M\^32 + 
                    19578\ M\^34 - 11223\ M\^36 - 12186\ M\^38 + 
                    28101\ M\^40 - 2528\ M\^42 - 23834\ M\^44 + 
                    16252\ M\^46 + 9765\ M\^48 - 16286\ M\^50 + 1699\ M\^52 + 
                    6876\ M\^54 - 3690\ M\^56 - 364\ M\^58 + 1062\ M\^60 - 
                    652\ M\^62 + 500\ M\^64 - 372\ M\^66 + 172\ M\^68 - 
                    44\ M\^70 + 5\ M\^72)\) + 
              L\^5\ \((\(-15\)\ M\^4 + 175\ M\^6 - 829\ M\^8 + 1892\ M\^10 - 
                    1323\ M\^12 - 3425\ M\^14 + 8304\ M\^16 - 1726\ M\^18 - 
                    17164\ M\^20 + 22957\ M\^22 + 9265\ M\^24 - 
                    49763\ M\^26 + 23886\ M\^28 + 62811\ M\^30 - 
                    69601\ M\^32 - 52185\ M\^34 + 99837\ M\^36 + 
                    33846\ M\^38 - 108758\ M\^40 - 24562\ M\^42 + 
                    100347\ M\^44 + 31601\ M\^46 - 63232\ M\^48 - 
                    24373\ M\^50 + 28339\ M\^52 + 16044\ M\^54 - 
                    6920\ M\^56 - 12262\ M\^58 - 2109\ M\^60 + 11278\ M\^62 + 
                    228\ M\^64 - 6940\ M\^66 + 2264\ M\^68 + 2128\ M\^70 - 
                    2156\ M\^72 + 937\ M\^74 - 317\ M\^76 + 125\ M\^78 - 
                    45\ M\^80 + 10\ M\^82 - M\^84)\) + 
              L\^6\ \((\(-20\)\ M\^6 + 290\ M\^8 - 1700\ M\^10 + 
                    4878\ M\^12 - 5210\ M\^14 - 7762\ M\^16 + 27830\ M\^18 - 
                    13431\ M\^20 - 53172\ M\^22 + 82911\ M\^24 + 
                    26050\ M\^26 - 164677\ M\^28 + 75390\ M\^30 + 
                    196703\ M\^32 - 204764\ M\^34 - 154861\ M\^36 + 
                    278494\ M\^38 + 97232\ M\^40 - 279742\ M\^42 - 
                    82000\ M\^44 + 242104\ M\^46 + 91964\ M\^48 - 
                    199384\ M\^50 - 33746\ M\^52 + 204230\ M\^54 - 
                    39860\ M\^56 - 188770\ M\^58 + 83851\ M\^60 + 
                    125244\ M\^62 - 81456\ M\^64 - 56346\ M\^66 + 
                    53591\ M\^68 + 11410\ M\^70 - 21586\ M\^72 + 
                    1992\ M\^74 + 4398\ M\^76 - 1974\ M\^78 + 1203\ M\^80 - 
                    1362\ M\^82 + 869\ M\^84 - 296\ M\^86 + 53\ M\^88 - 
                    4\ M\^90)\) + 
              L\^7\ \((\(-15\)\ M\^8 + 260\ M\^10 - 1825\ M\^12 + 
                    6411\ M\^14 - 9877\ M\^16 - 4760\ M\^18 + 40381\ M\^20 - 
                    35568\ M\^22 - 68663\ M\^24 + 143888\ M\^26 + 
                    23559\ M\^28 - 269571\ M\^30 + 123321\ M\^32 + 
                    323809\ M\^34 - 330602\ M\^36 - 228597\ M\^38 + 
                    474687\ M\^40 + 18310\ M\^42 - 452053\ M\^44 + 
                    193584\ M\^46 + 246768\ M\^48 - 317998\ M\^50 - 
                    2297\ M\^52 + 362245\ M\^54 - 83095\ M\^56 - 
                    265947\ M\^58 + 100699\ M\^60 + 154685\ M\^62 - 
                    111739\ M\^64 - 92723\ M\^66 + 107208\ M\^68 + 
                    60203\ M\^70 - 85741\ M\^72 - 27657\ M\^74 + 
                    48454\ M\^76 + 9640\ M\^78 - 22019\ M\^80 - 1184\ M\^82 + 
                    7388\ M\^84 + 357\ M\^86 - 3861\ M\^88 + 2338\ M\^90 - 
                    672\ M\^92 + 99\ M\^94 - 6\ M\^96)\) + 
              L\^8\ \((\(-6\)\ M\^10 + 121\ M\^12 - 1000\ M\^14 + 
                    4280\ M\^16 - 9306\ M\^18 + 4947\ M\^20 + 21444\ M\^22 - 
                    39928\ M\^24 - 15436\ M\^26 + 104503\ M\^28 - 
                    48072\ M\^30 - 150763\ M\^32 + 161214\ M\^34 + 
                    138484\ M\^36 - 304718\ M\^38 - 19268\ M\^40 + 
                    409140\ M\^42 - 202263\ M\^44 - 401240\ M\^46 + 
                    498804\ M\^48 + 232000\ M\^50 - 791944\ M\^52 + 
                    11954\ M\^54 + 917454\ M\^56 - 207216\ M\^58 - 
                    672715\ M\^60 + 416836\ M\^62 + 301066\ M\^64 - 
                    493810\ M\^66 + 2040\ M\^68 + 376910\ M\^70 - 
                    159129\ M\^72 - 186436\ M\^74 + 180479\ M\^76 + 
                    41000\ M\^78 - 117452\ M\^80 + 12462\ M\^82 + 
                    55696\ M\^84 - 20316\ M\^86 - 16831\ M\^88 + 
                    11116\ M\^90 + 3381\ M\^92 - 5874\ M\^94 + 2717\ M\^96 - 
                    642\ M\^98 + 79\ M\^100 - 4\ M\^102)\) + 
              L\^17\ \((\(-5\)\ M\^84 + 14\ M\^86 - 16\ M\^88 + M\^90 + 
                    33\ M\^92 - 9\ M\^94 - M\^96 + 4\ M\^98 - 8\ M\^100 + 
                    8\ M\^102 - 3\ M\^104)\) + 
              L\^9\ \((\(-M\^12\) + 23\ M\^14 - 227\ M\^16 + 1249\ M\^18 - 
                    4103\ M\^20 + 7522\ M\^22 - 3722\ M\^24 - 14803\ M\^26 + 
                    29205\ M\^28 + 7051\ M\^30 - 80625\ M\^32 + 
                    53190\ M\^34 + 123449\ M\^36 - 161690\ M\^38 - 
                    125812\ M\^40 + 283369\ M\^42 + 46690\ M\^44 - 
                    324460\ M\^46 + 61772\ M\^48 + 276151\ M\^50 - 
                    97609\ M\^52 - 198536\ M\^54 - 8278\ M\^56 + 
                    95244\ M\^58 + 118522\ M\^60 + 95244\ M\^62 - 
                    8278\ M\^64 - 198536\ M\^66 - 97609\ M\^68 + 
                    276151\ M\^70 + 61772\ M\^72 - 324460\ M\^74 + 
                    46690\ M\^76 + 283369\ M\^78 - 125812\ M\^80 - 
                    161690\ M\^82 + 123449\ M\^84 + 53190\ M\^86 - 
                    80625\ M\^88 + 7051\ M\^90 + 29205\ M\^92 - 
                    14803\ M\^94 - 3722\ M\^96 + 7522\ M\^98 - 4103\ M\^100 + 
                    1249\ M\^102 - 227\ M\^104 + 23\ M\^106 - M\^108)\) + 
              L\^10\ \((\(-4\)\ M\^18 + 79\ M\^20 - 642\ M\^22 + 
                    2717\ M\^24 - 5874\ M\^26 + 3381\ M\^28 + 11116\ M\^30 - 
                    16831\ M\^32 - 20316\ M\^34 + 55696\ M\^36 + 
                    12462\ M\^38 - 117452\ M\^40 + 41000\ M\^42 + 
                    180479\ M\^44 - 186436\ M\^46 - 159129\ M\^48 + 
                    376910\ M\^50 + 2040\ M\^52 - 493810\ M\^54 + 
                    301066\ M\^56 + 416836\ M\^58 - 672715\ M\^60 - 
                    207216\ M\^62 + 917454\ M\^64 + 11954\ M\^66 - 
                    791944\ M\^68 + 232000\ M\^70 + 498804\ M\^72 - 
                    401240\ M\^74 - 202263\ M\^76 + 409140\ M\^78 - 
                    19268\ M\^80 - 304718\ M\^82 + 138484\ M\^84 + 
                    161214\ M\^86 - 150763\ M\^88 - 48072\ M\^90 + 
                    104503\ M\^92 - 15436\ M\^94 - 39928\ M\^96 + 
                    21444\ M\^98 + 4947\ M\^100 - 9306\ M\^102 + 
                    4280\ M\^104 - 1000\ M\^106 + 121\ M\^108 - 
                    6\ M\^110)\) + 
              L\^11\ \((\(-6\)\ M\^24 + 99\ M\^26 - 672\ M\^28 + 
                    2338\ M\^30 - 3861\ M\^32 + 357\ M\^34 + 7388\ M\^36 - 
                    1184\ M\^38 - 22019\ M\^40 + 9640\ M\^42 + 48454\ M\^44 - 
                    27657\ M\^46 - 85741\ M\^48 + 60203\ M\^50 + 
                    107208\ M\^52 - 92723\ M\^54 - 111739\ M\^56 + 
                    154685\ M\^58 + 100699\ M\^60 - 265947\ M\^62 - 
                    83095\ M\^64 + 362245\ M\^66 - 2297\ M\^68 - 
                    317998\ M\^70 + 246768\ M\^72 + 193584\ M\^74 - 
                    452053\ M\^76 + 18310\ M\^78 + 474687\ M\^80 - 
                    228597\ M\^82 - 330602\ M\^84 + 323809\ M\^86 + 
                    123321\ M\^88 - 269571\ M\^90 + 23559\ M\^92 + 
                    143888\ M\^94 - 68663\ M\^96 - 35568\ M\^98 + 
                    40381\ M\^100 - 4760\ M\^102 - 9877\ M\^104 + 
                    6411\ M\^106 - 1825\ M\^108 + 260\ M\^110 - 
                    15\ M\^112)\) + 
              L\^16\ \((10\ M\^72 - 52\ M\^74 + 117\ M\^76 - 112\ M\^78 - 
                    83\ M\^80 + 290\ M\^82 - 215\ M\^84 - 168\ M\^86 + 
                    504\ M\^88 + 42\ M\^90 - 297\ M\^92 + 52\ M\^94 + 
                    116\ M\^96 - 24\ M\^98 - 39\ M\^100 + 10\ M\^102 + 
                    11\ M\^104 - 24\ M\^106 + 26\ M\^108 - 14\ M\^110 + 
                    3\ M\^112)\) + 
              L\^12\ \((\(-4\)\ M\^30 + 53\ M\^32 - 296\ M\^34 + 869\ M\^36 - 
                    1362\ M\^38 + 1203\ M\^40 - 1974\ M\^42 + 4398\ M\^44 + 
                    1992\ M\^46 - 21586\ M\^48 + 11410\ M\^50 + 
                    53591\ M\^52 - 56346\ M\^54 - 81456\ M\^56 + 
                    125244\ M\^58 + 83851\ M\^60 - 188770\ M\^62 - 
                    39860\ M\^64 + 204230\ M\^66 - 33746\ M\^68 - 
                    199384\ M\^70 + 91964\ M\^72 + 242104\ M\^74 - 
                    82000\ M\^76 - 279742\ M\^78 + 97232\ M\^80 + 
                    278494\ M\^82 - 154861\ M\^84 - 204764\ M\^86 + 
                    196703\ M\^88 + 75390\ M\^90 - 164677\ M\^92 + 
                    26050\ M\^94 + 82911\ M\^96 - 53172\ M\^98 - 
                    13431\ M\^100 + 27830\ M\^102 - 7762\ M\^104 - 
                    5210\ M\^106 + 4878\ M\^108 - 1700\ M\^110 + 
                    290\ M\^112 - 20\ M\^114)\) + 
              L\^13\ \((\(-M\^36\) + 10\ M\^38 - 45\ M\^40 + 125\ M\^42 - 
                    317\ M\^44 + 937\ M\^46 - 2156\ M\^48 + 2128\ M\^50 + 
                    2264\ M\^52 - 6940\ M\^54 + 228\ M\^56 + 11278\ M\^58 - 
                    2109\ M\^60 - 12262\ M\^62 - 6920\ M\^64 + 16044\ M\^66 + 
                    28339\ M\^68 - 24373\ M\^70 - 63232\ M\^72 + 
                    31601\ M\^74 + 100347\ M\^76 - 24562\ M\^78 - 
                    108758\ M\^80 + 33846\ M\^82 + 99837\ M\^84 - 
                    52185\ M\^86 - 69601\ M\^88 + 62811\ M\^90 + 
                    23886\ M\^92 - 49763\ M\^94 + 9265\ M\^96 + 
                    22957\ M\^98 - 17164\ M\^100 - 1726\ M\^102 + 
                    8304\ M\^104 - 3425\ M\^106 - 1323\ M\^108 + 
                    1892\ M\^110 - 829\ M\^112 + 175\ M\^114 - 
                    15\ M\^116)\) + 
              L\^14\ \((5\ M\^48 - 44\ M\^50 + 172\ M\^52 - 372\ M\^54 + 
                    500\ M\^56 - 652\ M\^58 + 1062\ M\^60 - 364\ M\^62 - 
                    3690\ M\^64 + 6876\ M\^66 + 1699\ M\^68 - 16286\ M\^70 + 
                    9765\ M\^72 + 16252\ M\^74 - 23834\ M\^76 - 2528\ M\^78 + 
                    28101\ M\^80 - 12186\ M\^82 - 11223\ M\^84 + 
                    19578\ M\^86 - 8172\ M\^88 - 12158\ M\^90 + 
                    15350\ M\^92 + 168\ M\^94 - 9787\ M\^96 + 4332\ M\^98 + 
                    2789\ M\^100 - 3104\ M\^102 + 395\ M\^104 + 852\ M\^106 - 
                    460\ M\^108 - 166\ M\^110 + 335\ M\^112 - 192\ M\^114 + 
                    53\ M\^116 - 6\ M\^118)\) + 
              L\^15\ \((\(-10\)\ M\^60 + 72\ M\^62 - 228\ M\^64 + 
                    366\ M\^66 - 182\ M\^68 - 295\ M\^70 + 481\ M\^72 - 
                    194\ M\^74 + 104\ M\^76 - 526\ M\^78 - 665\ M\^80 + 
                    2522\ M\^82 - 359\ M\^84 - 1748\ M\^86 + 2654\ M\^88 - 
                    581\ M\^90 - 1982\ M\^92 + 2093\ M\^94 - 260\ M\^96 - 
                    1172\ M\^98 + 950\ M\^100 - 11\ M\^102 - 414\ M\^104 + 
                    255\ M\^106 - 58\ M\^108 + 9\ M\^110 - 15\ M\^112 + 
                    20\ M\^114 - 15\ M\^116 + 6\ M\^118 - M\^120)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 10]\), \(M\^60 + L\^22\ M\^60 + 
              L\^21\ \((2\ M\^46 - 7\ M\^48 + 11\ M\^50 - 9\ M\^52 + 
                    3\ M\^54 + 6\ M\^56 - 26\ M\^58 + 43\ M\^60 + 22\ M\^62 - 
                    33\ M\^64 + 10\ M\^66)\) + 
              L\^20\ \((M\^32 - 6\ M\^34 + 15\ M\^36 - 22\ M\^38 + 
                    28\ M\^40 - 34\ M\^42 + 35\ M\^44 - 78\ M\^46 + 
                    134\ M\^48 + 112\ M\^50 - 591\ M\^52 + 538\ M\^54 + 
                    364\ M\^56 - 1342\ M\^58 + 946\ M\^60 + 1134\ M\^62 - 
                    1318\ M\^64 - 18\ M\^66 + 572\ M\^68 - 284\ M\^70 + 
                    45\ M\^72)\) + 
              L\ \((10\ M\^54 - 33\ M\^56 + 22\ M\^58 + 43\ M\^60 - 
                    26\ M\^62 + 6\ M\^64 + 3\ M\^66 - 9\ M\^68 + 11\ M\^70 - 
                    7\ M\^72 + 2\ M\^74)\) + 
              L\^19\ \((M\^22 - 8\ M\^24 + 25\ M\^26 - 45\ M\^28 + 
                    80\ M\^30 - 176\ M\^32 + 257\ M\^34 - 113\ M\^36 - 
                    172\ M\^38 + 289\ M\^40 - 65\ M\^42 - 639\ M\^44 + 
                    316\ M\^46 + 2771\ M\^48 - 3152\ M\^50 - 5335\ M\^52 + 
                    12895\ M\^54 - 1285\ M\^56 - 22437\ M\^58 + 
                    18400\ M\^60 + 19161\ M\^62 - 26273\ M\^64 - 
                    2688\ M\^66 + 16980\ M\^68 - 6319\ M\^70 - 3482\ M\^72 + 
                    3518\ M\^74 - 1084\ M\^76 + 120\ M\^78)\) + 
              L\^18\ \((M\^16 - 11\ M\^18 + 67\ M\^20 - 257\ M\^22 + 
                    610\ M\^24 - 821\ M\^26 + 380\ M\^28 + 632\ M\^30 - 
                    1567\ M\^32 + 1973\ M\^34 - 200\ M\^36 - 6466\ M\^38 + 
                    12104\ M\^40 + 1062\ M\^42 - 28862\ M\^44 + 
                    22718\ M\^46 + 35311\ M\^48 - 68193\ M\^50 - 527\ M\^52 + 
                    120677\ M\^54 - 98396\ M\^56 - 141333\ M\^58 + 
                    227927\ M\^60 + 90804\ M\^62 - 275210\ M\^64 + 
                    23303\ M\^66 + 199126\ M\^68 - 93930\ M\^70 - 
                    68425\ M\^72 + 70788\ M\^74 - 5740\ M\^76 - 
                    18538\ M\^78 + 10506\ M\^80 - 2408\ M\^82 + 
                    210\ M\^84)\) + 
              L\^2\ \((45\ M\^48 - 284\ M\^50 + 572\ M\^52 - 18\ M\^54 - 
                    1318\ M\^56 + 1134\ M\^58 + 946\ M\^60 - 1342\ M\^62 + 
                    364\ M\^64 + 538\ M\^66 - 591\ M\^68 + 112\ M\^70 + 
                    134\ M\^72 - 78\ M\^74 + 35\ M\^76 - 34\ M\^78 + 
                    28\ M\^80 - 22\ M\^82 + 15\ M\^84 - 6\ M\^86 + M\^88)\) + 
              L\^17\ \((\(-4\)\ M\^12 + 48\ M\^14 - 282\ M\^16 + 
                    1058\ M\^18 - 2634\ M\^20 + 3806\ M\^22 - 773\ M\^24 - 
                    8355\ M\^26 + 14800\ M\^28 - 5452\ M\^30 - 17464\ M\^32 + 
                    45224\ M\^34 - 61584\ M\^36 - 16453\ M\^38 + 
                    205009\ M\^40 - 183710\ M\^42 - 252284\ M\^44 + 
                    492412\ M\^46 + 21781\ M\^48 - 647135\ M\^50 + 
                    469080\ M\^52 + 471599\ M\^54 - 1027596\ M\^56 - 
                    42212\ M\^58 + 1399855\ M\^60 - 419513\ M\^62 - 
                    1341930\ M\^64 + 794872\ M\^66 + 898179\ M\^68 - 
                    880233\ M\^70 - 314546\ M\^72 + 629713\ M\^74 - 
                    71176\ M\^76 - 251033\ M\^78 + 130053\ M\^80 + 
                    23028\ M\^82 - 45054\ M\^84 + 18418\ M\^86 - 
                    3430\ M\^88 + 252\ M\^90)\) + 
              L\^16\ \((6\ M\^8 - 88\ M\^10 + 587\ M\^12 - 2319\ M\^14 + 
                    5799\ M\^16 - 8460\ M\^18 + 2671\ M\^20 + 16881\ M\^22 - 
                    36260\ M\^24 + 23345\ M\^26 + 38470\ M\^28 - 
                    133044\ M\^30 + 183075\ M\^32 + 43483\ M\^34 - 
                    609362\ M\^36 + 630440\ M\^38 + 762460\ M\^40 - 
                    1834308\ M\^42 + 44010\ M\^44 + 2585762\ M\^46 - 
                    1759929\ M\^48 - 1707141\ M\^50 + 3333847\ M\^52 - 
                    841189\ M\^54 - 3630582\ M\^56 + 3574023\ M\^58 + 
                    2586722\ M\^60 - 4895830\ M\^62 - 942466\ M\^64 + 
                    4714679\ M\^66 - 309261\ M\^68 - 3611816\ M\^70 + 
                    971289\ M\^72 + 2217900\ M\^74 - 1125906\ M\^76 - 
                    963513\ M\^78 + 868111\ M\^80 + 147743\ M\^82 - 
                    400143\ M\^84 + 109044\ M\^86 + 70903\ M\^88 - 
                    62344\ M\^90 + 20362\ M\^92 - 3248\ M\^94 + 
                    210\ M\^96)\) + 
              L\^3\ \((120\ M\^42 - 1084\ M\^44 + 3518\ M\^46 - 3482\ M\^48 - 
                    6319\ M\^50 + 16980\ M\^52 - 2688\ M\^54 - 26273\ M\^56 + 
                    19161\ M\^58 + 18400\ M\^60 - 22437\ M\^62 - 
                    1285\ M\^64 + 12895\ M\^66 - 5335\ M\^68 - 3152\ M\^70 + 
                    2771\ M\^72 + 316\ M\^74 - 639\ M\^76 - 65\ M\^78 + 
                    289\ M\^80 - 172\ M\^82 - 113\ M\^84 + 257\ M\^86 - 
                    176\ M\^88 + 80\ M\^90 - 45\ M\^92 + 25\ M\^94 - 
                    8\ M\^96 + M\^98)\) + 
              L\^15\ \((\(-4\)\ M\^4 + 71\ M\^6 - 560\ M\^8 + 2540\ M\^10 - 
                    7115\ M\^12 + 11951\ M\^14 - 8980\ M\^16 - 8162\ M\^18 + 
                    35491\ M\^20 - 59826\ M\^22 + 26107\ M\^24 + 
                    179180\ M\^26 - 436798\ M\^28 + 82660\ M\^30 + 
                    1102683\ M\^32 - 1319634\ M\^34 - 1212273\ M\^36 + 
                    3551692\ M\^38 - 583775\ M\^40 - 5138218\ M\^42 + 
                    4365908\ M\^44 + 3590446\ M\^46 - 7648546\ M\^48 + 
                    1925008\ M\^50 + 6739274\ M\^52 - 8605417\ M\^54 - 
                    669216\ M\^56 + 11858045\ M\^58 - 7178230\ M\^60 - 
                    9492385\ M\^62 + 12172387\ M\^64 + 4356093\ M\^66 - 
                    11860603\ M\^68 - 187836\ M\^70 + 8237759\ M\^72 - 
                    1350247\ M\^74 - 4545539\ M\^76 + 1224506\ M\^78 + 
                    2211622\ M\^80 - 805171\ M\^82 - 988097\ M\^84 + 
                    562052\ M\^86 + 285845\ M\^88 - 309244\ M\^90 + 
                    17054\ M\^92 + 87878\ M\^94 - 52350\ M\^96 + 
                    14442\ M\^98 - 2044\ M\^100 + 120\ M\^102)\) + 
              L\^4\ \((210\ M\^36 - 2408\ M\^38 + 10506\ M\^40 - 
                    18538\ M\^42 - 5740\ M\^44 + 70788\ M\^46 - 
                    68425\ M\^48 - 93930\ M\^50 + 199126\ M\^52 + 
                    23303\ M\^54 - 275210\ M\^56 + 90804\ M\^58 + 
                    227927\ M\^60 - 141333\ M\^62 - 98396\ M\^64 + 
                    120677\ M\^66 - 527\ M\^68 - 68193\ M\^70 + 
                    35311\ M\^72 + 22718\ M\^74 - 28862\ M\^76 + 
                    1062\ M\^78 + 12104\ M\^80 - 6466\ M\^82 - 200\ M\^84 + 
                    1973\ M\^86 - 1567\ M\^88 + 632\ M\^90 + 380\ M\^92 - 
                    821\ M\^94 + 610\ M\^96 - 257\ M\^98 + 67\ M\^100 - 
                    11\ M\^102 + M\^104)\) + 
              L\^5\ \((252\ M\^30 - 3430\ M\^32 + 18418\ M\^34 - 
                    45054\ M\^36 + 23028\ M\^38 + 130053\ M\^40 - 
                    251033\ M\^42 - 71176\ M\^44 + 629713\ M\^46 - 
                    314546\ M\^48 - 880233\ M\^50 + 898179\ M\^52 + 
                    794872\ M\^54 - 1341930\ M\^56 - 419513\ M\^58 + 
                    1399855\ M\^60 - 42212\ M\^62 - 1027596\ M\^64 + 
                    471599\ M\^66 + 469080\ M\^68 - 647135\ M\^70 + 
                    21781\ M\^72 + 492412\ M\^74 - 252284\ M\^76 - 
                    183710\ M\^78 + 205009\ M\^80 - 16453\ M\^82 - 
                    61584\ M\^84 + 45224\ M\^86 - 17464\ M\^88 - 
                    5452\ M\^90 + 14800\ M\^92 - 8355\ M\^94 - 773\ M\^96 + 
                    3806\ M\^98 - 2634\ M\^100 + 1058\ M\^102 - 282\ M\^104 + 
                    48\ M\^106 - 4\ M\^108)\) + 
              L\^14\ \((1 - 21\ M\^2 + 200\ M\^4 - 1130\ M\^6 + 4185\ M\^8 - 
                    10581\ M\^10 + 17862\ M\^12 - 14933\ M\^14 - 
                    18279\ M\^16 + 82618\ M\^18 - 82682\ M\^20 - 
                    152089\ M\^22 + 484758\ M\^24 - 196495\ M\^26 - 
                    974292\ M\^28 + 1425464\ M\^30 + 632922\ M\^32 - 
                    3392672\ M\^34 + 2170330\ M\^36 + 4032649\ M\^38 - 
                    7631187\ M\^40 - 480374\ M\^42 + 12725750\ M\^44 - 
                    7831215\ M\^46 - 11983397\ M\^48 + 16409439\ M\^50 + 
                    2546964\ M\^52 - 18033401\ M\^54 + 11650980\ M\^56 + 
                    8664696\ M\^58 - 21806065\ M\^60 + 7944914\ M\^62 + 
                    22256770\ M\^64 - 21438255\ M\^66 - 13610793\ M\^68 + 
                    25003281\ M\^70 + 3030421\ M\^72 - 19323411\ M\^74 + 
                    3179840\ M\^76 + 10448100\ M\^78 - 3962957\ M\^80 - 
                    3912609\ M\^82 + 2243379\ M\^84 + 999268\ M\^86 - 
                    707995\ M\^88 - 295091\ M\^90 + 177342\ M\^92 + 
                    138902\ M\^94 - 88692\ M\^96 - 39700\ M\^98 + 
                    57962\ M\^100 - 26530\ M\^102 + 6398\ M\^104 - 
                    824\ M\^106 + 45\ M\^108)\) + 
              L\^6\ \((210\ M\^24 - 3248\ M\^26 + 20362\ M\^28 - 
                    62344\ M\^30 + 70903\ M\^32 + 109044\ M\^34 - 
                    400143\ M\^36 + 147743\ M\^38 + 868111\ M\^40 - 
                    963513\ M\^42 - 1125906\ M\^44 + 2217900\ M\^46 + 
                    971289\ M\^48 - 3611816\ M\^50 - 309261\ M\^52 + 
                    4714679\ M\^54 - 942466\ M\^56 - 4895830\ M\^58 + 
                    2586722\ M\^60 + 3574023\ M\^62 - 3630582\ M\^64 - 
                    841189\ M\^66 + 3333847\ M\^68 - 1707141\ M\^70 - 
                    1759929\ M\^72 + 2585762\ M\^74 + 44010\ M\^76 - 
                    1834308\ M\^78 + 762460\ M\^80 + 630440\ M\^82 - 
                    609362\ M\^84 + 43483\ M\^86 + 183075\ M\^88 - 
                    133044\ M\^90 + 38470\ M\^92 + 23345\ M\^94 - 
                    36260\ M\^96 + 16881\ M\^98 + 2671\ M\^100 - 
                    8460\ M\^102 + 5799\ M\^104 - 2319\ M\^106 + 
                    587\ M\^108 - 88\ M\^110 + 6\ M\^112)\) + 
              L\^13\ \((\(-4\) + 83\ M\^2 - 762\ M\^4 + 3986\ M\^6 - 
                    12568\ M\^8 + 22001\ M\^10 - 9907\ M\^12 - 37609\ M\^14 + 
                    56509\ M\^16 + 56267\ M\^18 - 212213\ M\^20 + 
                    117866\ M\^22 + 258246\ M\^24 - 527989\ M\^26 + 
                    306854\ M\^28 + 784426\ M\^30 - 2315650\ M\^32 + 
                    946067\ M\^34 + 5195778\ M\^36 - 6751091\ M\^38 - 
                    6178321\ M\^40 + 16693620\ M\^42 + 1056178\ M\^44 - 
                    26096511\ M\^46 + 11176709\ M\^48 + 27783866\ M\^50 - 
                    24813974\ M\^52 - 17914628\ M\^54 + 29584093\ M\^56 + 
                    335673\ M\^58 - 19676104\ M\^60 + 16060543\ M\^62 + 
                    424906\ M\^64 - 22056704\ M\^66 + 18276309\ M\^68 + 
                    17351028\ M\^70 - 27473156\ M\^72 - 7477904\ M\^74 + 
                    25451976\ M\^76 - 928751\ M\^78 - 16980209\ M\^80 + 
                    4577700\ M\^82 + 8113595\ M\^84 - 4086233\ M\^86 - 
                    2531363\ M\^88 + 2174005\ M\^90 + 265738\ M\^92 - 
                    655162\ M\^94 + 110124\ M\^96 + 76633\ M\^98 - 
                    34567\ M\^100 + 24286\ M\^102 - 30519\ M\^104 + 
                    20326\ M\^106 - 7505\ M\^108 + 1616\ M\^110 - 
                    193\ M\^112 + 10\ M\^114)\) + 
              L\^7\ \((120\ M\^18 - 2044\ M\^20 + 14442\ M\^22 - 
                    52350\ M\^24 + 87878\ M\^26 + 17054\ M\^28 - 
                    309244\ M\^30 + 285845\ M\^32 + 562052\ M\^34 - 
                    988097\ M\^36 - 805171\ M\^38 + 2211622\ M\^40 + 
                    1224506\ M\^42 - 4545539\ M\^44 - 1350247\ M\^46 + 
                    8237759\ M\^48 - 187836\ M\^50 - 11860603\ M\^52 + 
                    4356093\ M\^54 + 12172387\ M\^56 - 9492385\ M\^58 - 
                    7178230\ M\^60 + 11858045\ M\^62 - 669216\ M\^64 - 
                    8605417\ M\^66 + 6739274\ M\^68 + 1925008\ M\^70 - 
                    7648546\ M\^72 + 3590446\ M\^74 + 4365908\ M\^76 - 
                    5138218\ M\^78 - 583775\ M\^80 + 3551692\ M\^82 - 
                    1212273\ M\^84 - 1319634\ M\^86 + 1102683\ M\^88 + 
                    82660\ M\^90 - 436798\ M\^92 + 179180\ M\^94 + 
                    26107\ M\^96 - 59826\ M\^98 + 35491\ M\^100 - 
                    8162\ M\^102 - 8980\ M\^104 + 11951\ M\^106 - 
                    7115\ M\^108 + 2540\ M\^110 - 560\ M\^112 + 71\ M\^114 - 
                    4\ M\^116)\) + 
              L\^11\ \((\(-4\) + 81\ M\^2 - 730\ M\^4 + 3748\ M\^6 - 
                    11557\ M\^8 + 20067\ M\^10 - 14135\ M\^12 + 875\ M\^14 - 
                    40273\ M\^16 + 86608\ M\^18 + 151707\ M\^20 - 
                    490600\ M\^22 - 339367\ M\^24 + 1980171\ M\^26 - 
                    184348\ M\^28 - 5033324\ M\^30 + 2827315\ M\^32 + 
                    8840310\ M\^34 - 8796225\ M\^36 - 9453696\ M\^38 + 
                    15140442\ M\^40 + 2434074\ M\^42 - 13691648\ M\^44 + 
                    12222711\ M\^46 - 4708646\ M\^48 - 26028665\ M\^50 + 
                    39673988\ M\^52 + 27123720\ M\^54 - 76154898\ M\^56 - 
                    11353364\ M\^58 + 92296758\ M\^60 - 11353364\ M\^62 - 
                    76154898\ M\^64 + 27123720\ M\^66 + 39673988\ M\^68 - 
                    26028665\ M\^70 - 4708646\ M\^72 + 12222711\ M\^74 - 
                    13691648\ M\^76 + 2434074\ M\^78 + 15140442\ M\^80 - 
                    9453696\ M\^82 - 8796225\ M\^84 + 8840310\ M\^86 + 
                    2827315\ M\^88 - 5033324\ M\^90 - 184348\ M\^92 + 
                    1980171\ M\^94 - 339367\ M\^96 - 490600\ M\^98 + 
                    151707\ M\^100 + 86608\ M\^102 - 40273\ M\^104 + 
                    875\ M\^106 - 14135\ M\^108 + 20067\ M\^110 - 
                    11557\ M\^112 + 3748\ M\^114 - 730\ M\^116 + 81\ M\^118 - 
                    4\ M\^120)\) + 
              L\^9\ \((10\ M\^6 - 193\ M\^8 + 1616\ M\^10 - 7505\ M\^12 + 
                    20326\ M\^14 - 30519\ M\^16 + 24286\ M\^18 - 
                    34567\ M\^20 + 76633\ M\^22 + 110124\ M\^24 - 
                    655162\ M\^26 + 265738\ M\^28 + 2174005\ M\^30 - 
                    2531363\ M\^32 - 4086233\ M\^34 + 8113595\ M\^36 + 
                    4577700\ M\^38 - 16980209\ M\^40 - 928751\ M\^42 + 
                    25451976\ M\^44 - 7477904\ M\^46 - 27473156\ M\^48 + 
                    17351028\ M\^50 + 18276309\ M\^52 - 22056704\ M\^54 + 
                    424906\ M\^56 + 16060543\ M\^58 - 19676104\ M\^60 + 
                    335673\ M\^62 + 29584093\ M\^64 - 17914628\ M\^66 - 
                    24813974\ M\^68 + 27783866\ M\^70 + 11176709\ M\^72 - 
                    26096511\ M\^74 + 1056178\ M\^76 + 16693620\ M\^78 - 
                    6178321\ M\^80 - 6751091\ M\^82 + 5195778\ M\^84 + 
                    946067\ M\^86 - 2315650\ M\^88 + 784426\ M\^90 + 
                    306854\ M\^92 - 527989\ M\^94 + 258246\ M\^96 + 
                    117866\ M\^98 - 212213\ M\^100 + 56267\ M\^102 + 
                    56509\ M\^104 - 37609\ M\^106 - 9907\ M\^108 + 
                    22001\ M\^110 - 12568\ M\^112 + 3986\ M\^114 - 
                    762\ M\^116 + 83\ M\^118 - 4\ M\^120)\) + 
              L\^8\ \((45\ M\^12 - 824\ M\^14 + 6398\ M\^16 - 26530\ M\^18 + 
                    57962\ M\^20 - 39700\ M\^22 - 88692\ M\^24 + 
                    138902\ M\^26 + 177342\ M\^28 - 295091\ M\^30 - 
                    707995\ M\^32 + 999268\ M\^34 + 2243379\ M\^36 - 
                    3912609\ M\^38 - 3962957\ M\^40 + 10448100\ M\^42 + 
                    3179840\ M\^44 - 19323411\ M\^46 + 3030421\ M\^48 + 
                    25003281\ M\^50 - 13610793\ M\^52 - 21438255\ M\^54 + 
                    22256770\ M\^56 + 7944914\ M\^58 - 21806065\ M\^60 + 
                    8664696\ M\^62 + 11650980\ M\^64 - 18033401\ M\^66 + 
                    2546964\ M\^68 + 16409439\ M\^70 - 11983397\ M\^72 - 
                    7831215\ M\^74 + 12725750\ M\^76 - 480374\ M\^78 - 
                    7631187\ M\^80 + 4032649\ M\^82 + 2170330\ M\^84 - 
                    3392672\ M\^86 + 632922\ M\^88 + 1425464\ M\^90 - 
                    974292\ M\^92 - 196495\ M\^94 + 484758\ M\^96 - 
                    152089\ M\^98 - 82682\ M\^100 + 82618\ M\^102 - 
                    18279\ M\^104 - 14933\ M\^106 + 17862\ M\^108 - 
                    10581\ M\^110 + 4185\ M\^112 - 1130\ M\^114 + 
                    200\ M\^116 - 21\ M\^118 + M\^120)\) + 
              L\^12\ \((6 - 123\ M\^2 + 1118\ M\^4 - 5763\ M\^6 + 
                    17529\ M\^8 - 27156\ M\^10 - 1228\ M\^12 + 71870\ M\^14 - 
                    32231\ M\^16 - 263171\ M\^18 + 361199\ M\^20 + 
                    474765\ M\^22 - 1256456\ M\^24 - 365355\ M\^26 + 
                    2864505\ M\^28 - 964251\ M\^30 - 3847387\ M\^32 + 
                    3462217\ M\^34 + 1497314\ M\^36 - 4912990\ M\^38 + 
                    7198466\ M\^40 - 7089\ M\^42 - 21230115\ M\^44 + 
                    15397753\ M\^46 + 33732481\ M\^48 - 38174985\ M\^50 - 
                    35244999\ M\^52 + 56219007\ M\^54 + 22279727\ M\^56 - 
                    56942336\ M\^58 - 2226214\ M\^60 + 38343544\ M\^62 - 
                    10789875\ M\^64 - 9339625\ M\^66 + 11649252\ M\^68 - 
                    14071146\ M\^70 - 4153949\ M\^72 + 23173854\ M\^74 - 
                    3511351\ M\^76 - 20334413\ M\^78 + 6691768\ M\^80 + 
                    12855066\ M\^82 - 5883959\ M\^84 - 6404178\ M\^86 + 
                    3803769\ M\^88 + 2482729\ M\^90 - 1945920\ M\^92 - 
                    719329\ M\^94 + 856399\ M\^96 + 57815\ M\^98 - 
                    269391\ M\^100 + 51605\ M\^102 + 61459\ M\^104 - 
                    48951\ M\^106 + 21389\ M\^108 - 8422\ M\^110 + 
                    3161\ M\^112 - 922\ M\^114 + 178\ M\^116 - 20\ M\^118 + 
                    M\^120)\) + 
              L\^10\ \((1 - 20\ M\^2 + 178\ M\^4 - 922\ M\^6 + 3161\ M\^8 - 
                    8422\ M\^10 + 21389\ M\^12 - 48951\ M\^14 + 
                    61459\ M\^16 + 51605\ M\^18 - 269391\ M\^20 + 
                    57815\ M\^22 + 856399\ M\^24 - 719329\ M\^26 - 
                    1945920\ M\^28 + 2482729\ M\^30 + 3803769\ M\^32 - 
                    6404178\ M\^34 - 5883959\ M\^36 + 12855066\ M\^38 + 
                    6691768\ M\^40 - 20334413\ M\^42 - 3511351\ M\^44 + 
                    23173854\ M\^46 - 4153949\ M\^48 - 14071146\ M\^50 + 
                    11649252\ M\^52 - 9339625\ M\^54 - 10789875\ M\^56 + 
                    38343544\ M\^58 - 2226214\ M\^60 - 56942336\ M\^62 + 
                    22279727\ M\^64 + 56219007\ M\^66 - 35244999\ M\^68 - 
                    38174985\ M\^70 + 33732481\ M\^72 + 15397753\ M\^74 - 
                    21230115\ M\^76 - 7089\ M\^78 + 7198466\ M\^80 - 
                    4912990\ M\^82 + 1497314\ M\^84 + 3462217\ M\^86 - 
                    3847387\ M\^88 - 964251\ M\^90 + 2864505\ M\^92 - 
                    365355\ M\^94 - 1256456\ M\^96 + 474765\ M\^98 + 
                    361199\ M\^100 - 263171\ M\^102 - 32231\ M\^104 + 
                    71870\ M\^106 - 1228\ M\^108 - 27156\ M\^110 + 
                    17529\ M\^112 - 5763\ M\^114 + 1118\ M\^116 - 
                    123\ M\^118 + 6\ M\^120)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 11]\), \(L\^21\ M\^22 + 
              M\^108 + 
              L\^20\ \((\(-2\)\ M\^14 + 6\ M\^16 - 13\ M\^18 + 16\ M\^20 - 
                    2\ M\^22 - 23\ M\^24 + 45\ M\^26 + 5\ M\^28 - 22\ M\^30 + 
                    17\ M\^32 - 6\ M\^34)\) + 
              L\^19\ \((M\^6 - 6\ M\^8 + 17\ M\^10 - 29\ M\^12 + 34\ M\^14 - 
                    48\ M\^16 + 106\ M\^18 - 186\ M\^20 - 46\ M\^22 + 
                    629\ M\^24 - 688\ M\^26 - 337\ M\^28 + 1330\ M\^30 - 
                    442\ M\^32 - 609\ M\^34 + 763\ M\^36 - 174\ M\^38 - 
                    259\ M\^40 + 224\ M\^42 - 85\ M\^44 + 15\ M\^46)\) + 
              L\^18\ \((3\ M\^4 - 27\ M\^6 + 110\ M\^8 - 270\ M\^10 + 
                    378\ M\^12 - 83\ M\^14 - 581\ M\^16 + 408\ M\^18 + 
                    1253\ M\^20 - 1673\ M\^22 - 2304\ M\^24 + 5716\ M\^26 - 
                    274\ M\^28 - 10038\ M\^30 + 8788\ M\^32 + 7411\ M\^34 - 
                    14014\ M\^36 + 3433\ M\^38 + 10669\ M\^40 - 9146\ M\^42 - 
                    1639\ M\^44 + 5718\ M\^46 - 2602\ M\^48 - 753\ M\^50 + 
                    1373\ M\^52 - 676\ M\^54 + 170\ M\^56 - 20\ M\^58)\) + 
              L\^17\ \((3\ M\^2 - 38\ M\^4 + 224\ M\^6 - 802\ M\^8 + 
                    1759\ M\^10 - 1789\ M\^12 - 1519\ M\^14 + 7022\ M\^16 - 
                    5564\ M\^18 - 9958\ M\^20 + 23472\ M\^22 - 4050\ M\^24 - 
                    40041\ M\^26 + 41186\ M\^28 + 30749\ M\^30 - 
                    84922\ M\^32 + 15973\ M\^34 + 95171\ M\^36 - 
                    70005\ M\^38 - 50192\ M\^40 + 96283\ M\^42 - 
                    7546\ M\^44 - 67732\ M\^46 + 41850\ M\^48 + 
                    19319\ M\^50 - 38017\ M\^52 + 10559\ M\^54 + 
                    13821\ M\^56 - 12028\ M\^58 + 1055\ M\^60 + 3614\ M\^62 - 
                    2621\ M\^64 + 904\ M\^66 - 170\ M\^68 + 15\ M\^70)\) + 
              L\^16\ \((1 - 21\ M\^2 + 188\ M\^4 - 957\ M\^6 + 3015\ M\^8 - 
                    5593\ M\^10 + 3494\ M\^12 + 9760\ M\^14 - 25239\ M\^16 + 
                    10974\ M\^18 + 45679\ M\^20 - 81290\ M\^22 + 
                    3470\ M\^24 + 146935\ M\^26 - 150273\ M\^28 - 
                    105215\ M\^30 + 310794\ M\^32 - 75891\ M\^34 - 
                    345117\ M\^36 + 254948\ M\^38 + 228755\ M\^40 - 
                    280633\ M\^42 - 59592\ M\^44 + 193893\ M\^46 - 
                    12639\ M\^48 - 65047\ M\^50 + 26114\ M\^52 - 
                    15949\ M\^54 - 15857\ M\^56 + 36832\ M\^58 - 
                    3844\ M\^60 - 29028\ M\^62 + 19910\ M\^64 + 6951\ M\^66 - 
                    14856\ M\^68 + 4508\ M\^70 + 4121\ M\^72 - 4605\ M\^74 + 
                    2140\ M\^76 - 566\ M\^78 + 85\ M\^80 - 6\ M\^82)\) + 
              L\^15\ \((\(-4\) + 64\ M\^2 - 483\ M\^4 + 2185\ M\^6 - 
                    6189\ M\^8 + 9953\ M\^10 - 3405\ M\^12 - 20710\ M\^14 + 
                    38918\ M\^16 - 780\ M\^18 - 87732\ M\^20 + 
                    102514\ M\^22 + 55285\ M\^24 - 223141\ M\^26 + 
                    100193\ M\^28 + 247481\ M\^30 - 333771\ M\^32 - 
                    68508\ M\^34 + 517571\ M\^36 - 299609\ M\^38 - 
                    583831\ M\^40 + 738793\ M\^42 + 447990\ M\^44 - 
                    1068985\ M\^46 - 36083\ M\^48 + 1194894\ M\^50 - 
                    416422\ M\^52 - 890813\ M\^54 + 747940\ M\^56 + 
                    366810\ M\^58 - 759197\ M\^60 + 56163\ M\^62 + 
                    471148\ M\^64 - 207751\ M\^66 - 158431\ M\^68 + 
                    155919\ M\^70 - 7101\ M\^72 - 48039\ M\^74 + 
                    25076\ M\^76 - 590\ M\^78 - 6652\ M\^80 + 6297\ M\^82 - 
                    4159\ M\^84 + 1980\ M\^86 - 644\ M\^88 + 136\ M\^90 - 
                    17\ M\^92 + M\^94)\) + 
              L\^14\ \((6 - 92\ M\^2 + 650\ M\^4 - 2727\ M\^6 + 7156\ M\^8 - 
                    10784\ M\^10 + 4587\ M\^12 + 13393\ M\^14 - 
                    19934\ M\^16 - 7006\ M\^18 + 31497\ M\^20 - 3595\ M\^22 - 
                    23619\ M\^24 - 36720\ M\^26 + 105639\ M\^28 + 
                    6888\ M\^30 - 245006\ M\^32 + 168540\ M\^34 + 
                    257402\ M\^36 - 283820\ M\^38 + 10958\ M\^40 - 
                    150457\ M\^42 - 306010\ M\^44 + 1111913\ M\^46 + 
                    177013\ M\^48 - 1928267\ M\^50 + 421811\ M\^52 + 
                    2070388\ M\^54 - 663330\ M\^56 - 1294661\ M\^58 + 
                    494427\ M\^60 + 422507\ M\^62 - 126988\ M\^64 - 
                    43204\ M\^66 - 157629\ M\^68 + 5382\ M\^70 + 
                    248847\ M\^72 - 62166\ M\^74 - 184864\ M\^76 + 
                    77205\ M\^78 + 98592\ M\^80 - 72716\ M\^82 - 
                    20753\ M\^84 + 36188\ M\^86 - 4868\ M\^88 - 
                    12304\ M\^90 + 9845\ M\^92 - 3778\ M\^94 + 845\ M\^96 - 
                    107\ M\^98 + 6\ M\^100)\) + 
              L\^13\ \((\(-4\) + 61\ M\^2 - 423\ M\^4 + 1730\ M\^6 - 
                    4520\ M\^8 + 7761\ M\^10 - 9492\ M\^12 + 11274\ M\^14 - 
                    11978\ M\^16 - 14465\ M\^18 + 83713\ M\^20 - 
                    73296\ M\^22 - 184135\ M\^24 + 390866\ M\^26 + 
                    97927\ M\^28 - 830468\ M\^30 + 233293\ M\^32 + 
                    1308798\ M\^34 - 871871\ M\^36 - 1790209\ M\^38 + 
                    1921748\ M\^40 + 2412231\ M\^42 - 3632596\ M\^44 - 
                    3144472\ M\^46 + 5792133\ M\^48 + 3764695\ M\^50 - 
                    7817277\ M\^52 - 4030922\ M\^54 + 9008239\ M\^56 + 
                    3857267\ M\^58 - 8582236\ M\^60 - 2237774\ M\^62 + 
                    7326110\ M\^64 - 199478\ M\^66 - 5588913\ M\^68 + 
                    2152908\ M\^70 + 3382004\ M\^72 - 2792826\ M\^74 - 
                    1297347\ M\^76 + 2250515\ M\^78 - 38298\ M\^80 - 
                    1208375\ M\^82 + 439503\ M\^84 + 406099\ M\^86 - 
                    337079\ M\^88 - 24330\ M\^90 + 121380\ M\^92 - 
                    38579\ M\^94 - 20352\ M\^96 + 22324\ M\^98 - 
                    9256\ M\^100 + 2140\ M\^102 - 273\ M\^104 + 
                    15\ M\^106)\) + 
              L\^12\ \((1 - 15\ M\^2 + 103\ M\^4 - 431\ M\^6 + 1312\ M\^8 - 
                    3576\ M\^10 + 9475\ M\^12 - 19716\ M\^14 + 20168\ M\^16 + 
                    17815\ M\^18 - 70419\ M\^20 + 7650\ M\^22 + 
                    187596\ M\^24 - 117529\ M\^26 - 441772\ M\^28 + 
                    409164\ M\^30 + 1140029\ M\^32 - 1515734\ M\^34 - 
                    2149239\ M\^36 + 4074939\ M\^38 + 2792093\ M\^40 - 
                    8196703\ M\^42 - 1979544\ M\^44 + 13007145\ M\^46 - 
                    817830\ M\^48 - 17657494\ M\^50 + 5438738\ M\^52 + 
                    21611062\ M\^54 - 11127524\ M\^56 - 24395894\ M\^58 + 
                    16311185\ M\^60 + 24819874\ M\^62 - 18297162\ M\^64 - 
                    20562438\ M\^66 + 17267283\ M\^68 + 13162638\ M\^70 - 
                    14127541\ M\^72 - 6183015\ M\^74 + 10013020\ M\^76 + 
                    1680184\ M\^78 - 6098391\ M\^80 + 323412\ M\^82 + 
                    3189706\ M\^84 - 874273\ M\^86 - 1335115\ M\^88 + 
                    730020\ M\^90 + 386280\ M\^92 - 422117\ M\^94 + 
                    2035\ M\^96 + 139888\ M\^98 - 50988\ M\^100 - 
                    22617\ M\^102 + 27605\ M\^104 - 11877\ M\^106 + 
                    2806\ M\^108 - 362\ M\^110 + 20\ M\^112)\) + 
              L\ \((\(-6\)\ M\^96 + 17\ M\^98 - 22\ M\^100 + 5\ M\^102 + 
                    45\ M\^104 - 23\ M\^106 - 2\ M\^108 + 16\ M\^110 - 
                    13\ M\^112 + 6\ M\^114 - 2\ M\^116)\) + 
              L\^11\ \((6\ M\^6 - 99\ M\^8 + 723\ M\^10 - 2994\ M\^12 + 
                    7457\ M\^14 - 11017\ M\^16 + 10499\ M\^18 - 
                    15345\ M\^20 + 27306\ M\^22 + 23421\ M\^24 - 
                    184447\ M\^26 + 146959\ M\^28 + 517451\ M\^30 - 
                    989995\ M\^32 - 617259\ M\^34 + 2907918\ M\^36 - 
                    115856\ M\^38 - 6393658\ M\^40 + 2627109\ M\^42 + 
                    11566051\ M\^44 - 7673181\ M\^46 - 17823480\ M\^48 + 
                    15554080\ M\^50 + 22873815\ M\^52 - 25247425\ M\^54 - 
                    23779770\ M\^56 + 34756204\ M\^58 + 19064892\ M\^60 - 
                    42176666\ M\^62 - 10900938\ M\^64 + 45923154\ M\^66 + 
                    3884109\ M\^68 - 42843069\ M\^70 + 1309418\ M\^72 + 
                    33668950\ M\^74 - 4941172\ M\^76 - 21938285\ M\^78 + 
                    6427287\ M\^80 + 11461971\ M\^82 - 5626988\ M\^84 - 
                    4455737\ M\^86 + 3484710\ M\^88 + 1137916\ M\^90 - 
                    1569036\ M\^92 - 110460\ M\^94 + 512790\ M\^96 - 
                    4766\ M\^98 - 176292\ M\^100 + 24378\ M\^102 + 
                    56854\ M\^104 - 18789\ M\^106 - 18850\ M\^108 + 
                    19442\ M\^110 - 8327\ M\^112 + 1995\ M\^114 - 
                    263\ M\^116 + 15\ M\^118)\) + 
              L\^2\ \((15\ M\^84 - 85\ M\^86 + 224\ M\^88 - 259\ M\^90 - 
                    174\ M\^92 + 763\ M\^94 - 609\ M\^96 - 442\ M\^98 + 
                    1330\ M\^100 - 337\ M\^102 - 688\ M\^104 + 629\ M\^106 - 
                    46\ M\^108 - 186\ M\^110 + 106\ M\^112 - 48\ M\^114 + 
                    34\ M\^116 - 29\ M\^118 + 17\ M\^120 - 6\ M\^122 + 
                    M\^124)\) + 
              L\^10\ \((15\ M\^12 - 263\ M\^14 + 1995\ M\^16 - 8327\ M\^18 + 
                    19442\ M\^20 - 18850\ M\^22 - 18789\ M\^24 + 
                    56854\ M\^26 + 24378\ M\^28 - 176292\ M\^30 - 
                    4766\ M\^32 + 512790\ M\^34 - 110460\ M\^36 - 
                    1569036\ M\^38 + 1137916\ M\^40 + 3484710\ M\^42 - 
                    4455737\ M\^44 - 5626988\ M\^46 + 11461971\ M\^48 + 
                    6427287\ M\^50 - 21938285\ M\^52 - 4941172\ M\^54 + 
                    33668950\ M\^56 + 1309418\ M\^58 - 42843069\ M\^60 + 
                    3884109\ M\^62 + 45923154\ M\^64 - 10900938\ M\^66 - 
                    42176666\ M\^68 + 19064892\ M\^70 + 34756204\ M\^72 - 
                    23779770\ M\^74 - 25247425\ M\^76 + 22873815\ M\^78 + 
                    15554080\ M\^80 - 17823480\ M\^82 - 7673181\ M\^84 + 
                    11566051\ M\^86 + 2627109\ M\^88 - 6393658\ M\^90 - 
                    115856\ M\^92 + 2907918\ M\^94 - 617259\ M\^96 - 
                    989995\ M\^98 + 517451\ M\^100 + 146959\ M\^102 - 
                    184447\ M\^104 + 23421\ M\^106 + 27306\ M\^108 - 
                    15345\ M\^110 + 10499\ M\^112 - 11017\ M\^114 + 
                    7457\ M\^116 - 2994\ M\^118 + 723\ M\^120 - 99\ M\^122 + 
                    6\ M\^124)\) + 
              L\^3\ \((\(-20\)\ M\^72 + 170\ M\^74 - 676\ M\^76 + 
                    1373\ M\^78 - 753\ M\^80 - 2602\ M\^82 + 5718\ M\^84 - 
                    1639\ M\^86 - 9146\ M\^88 + 10669\ M\^90 + 3433\ M\^92 - 
                    14014\ M\^94 + 7411\ M\^96 + 8788\ M\^98 - 
                    10038\ M\^100 - 274\ M\^102 + 5716\ M\^104 - 
                    2304\ M\^106 - 1673\ M\^108 + 1253\ M\^110 + 
                    408\ M\^112 - 581\ M\^114 - 83\ M\^116 + 378\ M\^118 - 
                    270\ M\^120 + 110\ M\^122 - 27\ M\^124 + 3\ M\^126)\) + 
              L\^4\ \((15\ M\^60 - 170\ M\^62 + 904\ M\^64 - 2621\ M\^66 + 
                    3614\ M\^68 + 1055\ M\^70 - 12028\ M\^72 + 13821\ M\^74 + 
                    10559\ M\^76 - 38017\ M\^78 + 19319\ M\^80 + 
                    41850\ M\^82 - 67732\ M\^84 - 7546\ M\^86 + 
                    96283\ M\^88 - 50192\ M\^90 - 70005\ M\^92 + 
                    95171\ M\^94 + 15973\ M\^96 - 84922\ M\^98 + 
                    30749\ M\^100 + 41186\ M\^102 - 40041\ M\^104 - 
                    4050\ M\^106 + 23472\ M\^108 - 9958\ M\^110 - 
                    5564\ M\^112 + 7022\ M\^114 - 1519\ M\^116 - 
                    1789\ M\^118 + 1759\ M\^120 - 802\ M\^122 + 224\ M\^124 - 
                    38\ M\^126 + 3\ M\^128)\) + 
              L\^8\ \((15\ M\^24 - 273\ M\^26 + 2140\ M\^28 - 9256\ M\^30 + 
                    22324\ M\^32 - 20352\ M\^34 - 38579\ M\^36 + 
                    121380\ M\^38 - 24330\ M\^40 - 337079\ M\^42 + 
                    406099\ M\^44 + 439503\ M\^46 - 1208375\ M\^48 - 
                    38298\ M\^50 + 2250515\ M\^52 - 1297347\ M\^54 - 
                    2792826\ M\^56 + 3382004\ M\^58 + 2152908\ M\^60 - 
                    5588913\ M\^62 - 199478\ M\^64 + 7326110\ M\^66 - 
                    2237774\ M\^68 - 8582236\ M\^70 + 3857267\ M\^72 + 
                    9008239\ M\^74 - 4030922\ M\^76 - 7817277\ M\^78 + 
                    3764695\ M\^80 + 5792133\ M\^82 - 3144472\ M\^84 - 
                    3632596\ M\^86 + 2412231\ M\^88 + 1921748\ M\^90 - 
                    1790209\ M\^92 - 871871\ M\^94 + 1308798\ M\^96 + 
                    233293\ M\^98 - 830468\ M\^100 + 97927\ M\^102 + 
                    390866\ M\^104 - 184135\ M\^106 - 73296\ M\^108 + 
                    83713\ M\^110 - 14465\ M\^112 - 11978\ M\^114 + 
                    11274\ M\^116 - 9492\ M\^118 + 7761\ M\^120 - 
                    4520\ M\^122 + 1730\ M\^124 - 423\ M\^126 + 61\ M\^128 - 
                    4\ M\^130)\) + 
              L\^6\ \((M\^36 - 17\ M\^38 + 136\ M\^40 - 644\ M\^42 + 
                    1980\ M\^44 - 4159\ M\^46 + 6297\ M\^48 - 6652\ M\^50 - 
                    590\ M\^52 + 25076\ M\^54 - 48039\ M\^56 - 7101\ M\^58 + 
                    155919\ M\^60 - 158431\ M\^62 - 207751\ M\^64 + 
                    471148\ M\^66 + 56163\ M\^68 - 759197\ M\^70 + 
                    366810\ M\^72 + 747940\ M\^74 - 890813\ M\^76 - 
                    416422\ M\^78 + 1194894\ M\^80 - 36083\ M\^82 - 
                    1068985\ M\^84 + 447990\ M\^86 + 738793\ M\^88 - 
                    583831\ M\^90 - 299609\ M\^92 + 517571\ M\^94 - 
                    68508\ M\^96 - 333771\ M\^98 + 247481\ M\^100 + 
                    100193\ M\^102 - 223141\ M\^104 + 55285\ M\^106 + 
                    102514\ M\^108 - 87732\ M\^110 - 780\ M\^112 + 
                    38918\ M\^114 - 20710\ M\^116 - 3405\ M\^118 + 
                    9953\ M\^120 - 6189\ M\^122 + 2185\ M\^124 - 
                    483\ M\^126 + 64\ M\^128 - 4\ M\^130)\) + 
              L\^5\ \((\(-6\)\ M\^48 + 85\ M\^50 - 566\ M\^52 + 2140\ M\^54 - 
                    4605\ M\^56 + 4121\ M\^58 + 4508\ M\^60 - 14856\ M\^62 + 
                    6951\ M\^64 + 19910\ M\^66 - 29028\ M\^68 - 3844\ M\^70 + 
                    36832\ M\^72 - 15857\ M\^74 - 15949\ M\^76 + 
                    26114\ M\^78 - 65047\ M\^80 - 12639\ M\^82 + 
                    193893\ M\^84 - 59592\ M\^86 - 280633\ M\^88 + 
                    228755\ M\^90 + 254948\ M\^92 - 345117\ M\^94 - 
                    75891\ M\^96 + 310794\ M\^98 - 105215\ M\^100 - 
                    150273\ M\^102 + 146935\ M\^104 + 3470\ M\^106 - 
                    81290\ M\^108 + 45679\ M\^110 + 10974\ M\^112 - 
                    25239\ M\^114 + 9760\ M\^116 + 3494\ M\^118 - 
                    5593\ M\^120 + 3015\ M\^122 - 957\ M\^124 + 188\ M\^126 - 
                    21\ M\^128 + M\^130)\) + 
              L\^9\ \((20\ M\^18 - 362\ M\^20 + 2806\ M\^22 - 11877\ M\^24 + 
                    27605\ M\^26 - 22617\ M\^28 - 50988\ M\^30 + 
                    139888\ M\^32 + 2035\ M\^34 - 422117\ M\^36 + 
                    386280\ M\^38 + 730020\ M\^40 - 1335115\ M\^42 - 
                    874273\ M\^44 + 3189706\ M\^46 + 323412\ M\^48 - 
                    6098391\ M\^50 + 1680184\ M\^52 + 10013020\ M\^54 - 
                    6183015\ M\^56 - 14127541\ M\^58 + 13162638\ M\^60 + 
                    17267283\ M\^62 - 20562438\ M\^64 - 18297162\ M\^66 + 
                    24819874\ M\^68 + 16311185\ M\^70 - 24395894\ M\^72 - 
                    11127524\ M\^74 + 21611062\ M\^76 + 5438738\ M\^78 - 
                    17657494\ M\^80 - 817830\ M\^82 + 13007145\ M\^84 - 
                    1979544\ M\^86 - 8196703\ M\^88 + 2792093\ M\^90 + 
                    4074939\ M\^92 - 2149239\ M\^94 - 1515734\ M\^96 + 
                    1140029\ M\^98 + 409164\ M\^100 - 441772\ M\^102 - 
                    117529\ M\^104 + 187596\ M\^106 + 7650\ M\^108 - 
                    70419\ M\^110 + 17815\ M\^112 + 20168\ M\^114 - 
                    19716\ M\^116 + 9475\ M\^118 - 3576\ M\^120 + 
                    1312\ M\^122 - 431\ M\^124 + 103\ M\^126 - 15\ M\^128 + 
                    M\^130)\) + 
              L\^7\ \((6\ M\^30 - 107\ M\^32 + 845\ M\^34 - 3778\ M\^36 + 
                    9845\ M\^38 - 12304\ M\^40 - 4868\ M\^42 + 36188\ M\^44 - 
                    20753\ M\^46 - 72716\ M\^48 + 98592\ M\^50 + 
                    77205\ M\^52 - 184864\ M\^54 - 62166\ M\^56 + 
                    248847\ M\^58 + 5382\ M\^60 - 157629\ M\^62 - 
                    43204\ M\^64 - 126988\ M\^66 + 422507\ M\^68 + 
                    494427\ M\^70 - 1294661\ M\^72 - 663330\ M\^74 + 
                    2070388\ M\^76 + 421811\ M\^78 - 1928267\ M\^80 + 
                    177013\ M\^82 + 1111913\ M\^84 - 306010\ M\^86 - 
                    150457\ M\^88 + 10958\ M\^90 - 283820\ M\^92 + 
                    257402\ M\^94 + 168540\ M\^96 - 245006\ M\^98 + 
                    6888\ M\^100 + 105639\ M\^102 - 36720\ M\^104 - 
                    23619\ M\^106 - 3595\ M\^108 + 31497\ M\^110 - 
                    7006\ M\^112 - 19934\ M\^114 + 13393\ M\^116 + 
                    4587\ M\^118 - 10784\ M\^120 + 7156\ M\^122 - 
                    2727\ M\^124 + 650\ M\^126 - 92\ M\^128 + 
                    6\ M\^130)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 12]\), \(M\^24 + 
              L\^23\ M\^118 + 
              L\ \((4\ M\^18 - 13\ M\^20 + 18\ M\^22 - 15\ M\^24 - 7\ M\^26 + 
                    55\ M\^28 - 18\ M\^30 - 13\ M\^32 + 22\ M\^34 - 
                    14\ M\^36 + 4\ M\^38)\) + 
              L\^2\ \((6\ M\^12 - 36\ M\^14 + 90\ M\^16 - 120\ M\^18 + 
                    60\ M\^20 + 98\ M\^22 - 96\ M\^24 - 109\ M\^26 - 
                    182\ M\^28 + 619\ M\^30 + 135\ M\^32 - 689\ M\^34 + 
                    668\ M\^36 + 115\ M\^38 - 672\ M\^40 + 404\ M\^42 + 
                    99\ M\^44 - 240\ M\^46 + 139\ M\^48 - 42\ M\^50 + 
                    6\ M\^52)\) + 
              L\^3\ \((4\ M\^6 - 33\ M\^8 + 116\ M\^10 - 229\ M\^12 + 
                    264\ M\^14 - 211\ M\^16 + 484\ M\^18 - 1112\ M\^20 + 
                    125\ M\^22 + 3268\ M\^24 - 3838\ M\^26 - 1532\ M\^28 + 
                    5994\ M\^30 - 3615\ M\^32 - 3283\ M\^34 + 5664\ M\^36 + 
                    3612\ M\^38 - 2514\ M\^40 - 5564\ M\^42 + 3034\ M\^44 + 
                    4975\ M\^46 - 4990\ M\^48 - 435\ M\^50 + 3204\ M\^52 - 
                    1881\ M\^54 - 263\ M\^56 + 980\ M\^58 - 629\ M\^60 + 
                    214\ M\^62 - 42\ M\^64 + 4\ M\^66)\) + 
              L\^4\ \((1 - 10\ M\^2 + 45\ M\^4 - 118\ M\^6 + 211\ M\^8 - 
                    432\ M\^10 + 1339\ M\^12 - 3182\ M\^14 + 3305\ M\^16 + 
                    2740\ M\^18 - 10721\ M\^20 + 5328\ M\^22 + 14246\ M\^24 - 
                    19456\ M\^26 - 5941\ M\^28 + 29998\ M\^30 - 
                    20201\ M\^32 - 10763\ M\^34 + 51195\ M\^36 - 
                    71000\ M\^38 - 30256\ M\^40 + 176884\ M\^42 - 
                    45653\ M\^44 - 186225\ M\^46 + 118972\ M\^48 + 
                    94963\ M\^50 - 111917\ M\^52 - 2209\ M\^54 + 
                    49776\ M\^56 - 25716\ M\^58 - 1170\ M\^60 + 
                    10333\ M\^62 - 7803\ M\^64 + 1634\ M\^66 + 2068\ M\^68 - 
                    2319\ M\^70 + 1259\ M\^72 - 436\ M\^74 + 99\ M\^76 - 
                    14\ M\^78 + M\^80)\) + 
              L\^5\ \((\(-1\) + 19\ M\^2 - 158\ M\^4 + 742\ M\^6 - 
                    2068\ M\^8 + 3142\ M\^10 - 1569\ M\^12 - 374\ M\^14 - 
                    8535\ M\^16 + 24753\ M\^18 - 217\ M\^20 - 88948\ M\^22 + 
                    110417\ M\^24 + 102482\ M\^26 - 353449\ M\^28 + 
                    146246\ M\^30 + 497957\ M\^32 - 718030\ M\^34 - 
                    138160\ M\^36 + 1277567\ M\^38 - 860674\ M\^40 - 
                    1244559\ M\^42 + 1919607\ M\^44 + 495073\ M\^46 - 
                    2125490\ M\^48 + 458891\ M\^50 + 1386753\ M\^52 - 
                    860858\ M\^54 - 392175\ M\^56 + 618680\ M\^58 - 
                    174897\ M\^60 - 179567\ M\^62 + 230922\ M\^64 - 
                    62978\ M\^66 - 93624\ M\^68 + 87869\ M\^70 - 
                    3521\ M\^72 - 33309\ M\^74 + 17285\ M\^76 + 2450\ M\^78 - 
                    7026\ M\^80 + 4080\ M\^82 - 1343\ M\^84 + 276\ M\^86 - 
                    34\ M\^88 + 2\ M\^90)\) + 
              L\^6\ \((\(-9\)\ M\^2 + 149\ M\^4 - 1152\ M\^6 + 5272\ M\^8 - 
                    14604\ M\^10 + 20497\ M\^12 + 4547\ M\^14 - 
                    62257\ M\^16 + 44434\ M\^18 + 157668\ M\^20 - 
                    269858\ M\^22 - 237930\ M\^24 + 928262\ M\^26 - 
                    124824\ M\^28 - 2087628\ M\^30 + 1836052\ M\^32 + 
                    2830144\ M\^34 - 5275705\ M\^36 - 1512141\ M\^38 + 
                    9207505\ M\^40 - 2586246\ M\^42 - 11110828\ M\^44 + 
                    7665877\ M\^46 + 9445465\ M\^48 - 10533730\ M\^50 - 
                    5026339\ M\^52 + 9841532\ M\^54 + 622106\ M\^56 - 
                    6548744\ M\^58 + 1890761\ M\^60 + 2837500\ M\^62 - 
                    2269222\ M\^64 - 300338\ M\^66 + 1440485\ M\^68 - 
                    699225\ M\^70 - 459699\ M\^72 + 642664\ M\^74 - 
                    62097\ M\^76 - 285011\ M\^78 + 155208\ M\^80 + 
                    40101\ M\^82 - 69846\ M\^84 + 17985\ M\^86 + 
                    12704\ M\^88 - 12875\ M\^90 + 5581\ M\^92 - 1462\ M\^94 + 
                    240\ M\^96 - 23\ M\^98 + M\^100)\) + 
              L\^7\ \((\(-36\)\ M\^4 + 614\ M\^6 - 4750\ M\^8 + 
                    21379\ M\^10 - 57648\ M\^12 + 74138\ M\^14 + 
                    56198\ M\^16 - 364872\ M\^18 + 344473\ M\^20 + 
                    732214\ M\^22 - 1824336\ M\^24 - 235912\ M\^26 + 
                    4764699\ M\^28 - 2958685\ M\^30 - 8438704\ M\^32 + 
                    11030077\ M\^34 + 10148624\ M\^36 - 24307291\ M\^38 - 
                    6665329\ M\^40 + 39827534\ M\^42 - 3085200\ M\^44 - 
                    51925483\ M\^46 + 16196082\ M\^48 + 55305587\ M\^50 - 
                    27496927\ M\^52 - 47833934\ M\^54 + 33040217\ M\^56 + 
                    32723699\ M\^58 - 30676961\ M\^60 - 15776252\ M\^62 + 
                    22075951\ M\^64 + 3237783\ M\^66 - 11648547\ M\^68 + 
                    2305571\ M\^70 + 3820111\ M\^72 - 2679678\ M\^74 - 
                    196301\ M\^76 + 1410399\ M\^78 - 609470\ M\^80 - 
                    455531\ M\^82 + 427386\ M\^84 + 94879\ M\^86 - 
                    215161\ M\^88 + 28605\ M\^90 + 68886\ M\^92 - 
                    32261\ M\^94 - 12649\ M\^96 + 19508\ M\^98 - 
                    10045\ M\^100 + 3008\ M\^102 - 560\ M\^104 + 61\ M\^106 - 
                    3\ M\^108)\) + 
              L\^8\ \((\(-84\)\ M\^6 + 1512\ M\^8 - 12031\ M\^10 + 
                    54385\ M\^12 - 143884\ M\^14 + 169889\ M\^16 + 
                    205597\ M\^18 - 1059430\ M\^20 + 988003\ M\^22 + 
                    2169017\ M\^24 - 5667917\ M\^26 - 176764\ M\^28 + 
                    14667701\ M\^30 - 10711295\ M\^32 - 25049275\ M\^34 + 
                    35776314\ M\^36 + 29925103\ M\^38 - 74926430\ M\^40 - 
                    23271138\ M\^42 + 121644555\ M\^44 + 3921025\ M\^46 - 
                    164631526\ M\^48 + 23924016\ M\^50 + 190924186\ M\^52 - 
                    53289771\ M\^54 - 189769801\ M\^56 + 76024305\ M\^58 + 
                    159873609\ M\^60 - 83240681\ M\^62 - 110531958\ M\^64 + 
                    73093617\ M\^66 + 59999823\ M\^68 - 51608141\ M\^70 - 
                    22906368\ M\^72 + 28849482\ M\^74 + 3394522\ M\^76 - 
                    12057318\ M\^78 + 2999472\ M\^80 + 2823838\ M\^82 - 
                    2813175\ M\^84 + 519368\ M\^86 + 1110842\ M\^88 - 
                    829536\ M\^90 - 67477\ M\^92 + 284862\ M\^94 - 
                    75917\ M\^96 - 28102\ M\^98 + 15279\ M\^100 - 
                    9742\ M\^102 + 18086\ M\^104 - 16589\ M\^106 + 
                    8442\ M\^108 - 2649\ M\^110 + 519\ M\^112 - 59\ M\^114 + 
                    3\ M\^116)\) + 
              L\^9\ \((\(-126\)\ M\^8 + 2380\ M\^10 - 19588\ M\^12 + 
                    90166\ M\^14 - 238269\ M\^16 + 267598\ M\^18 + 
                    396019\ M\^20 - 1835033\ M\^22 + 1592605\ M\^24 + 
                    3989394\ M\^26 - 9949572\ M\^28 - 954293\ M\^30 + 
                    27142444\ M\^32 - 18179836\ M\^34 - 49845081\ M\^36 + 
                    65336860\ M\^38 + 67614588\ M\^40 - 145451909\ M\^42 - 
                    68550953\ M\^44 + 252110760\ M\^46 + 44752660\ M\^48 - 
                    365681280\ M\^50 + 3612692\ M\^52 + 456518212\ M\^54 - 
                    67075586\ M\^56 - 494512837\ M\^58 + 127599894\ M\^60 + 
                    465024672\ M\^62 - 164458742\ M\^64 - 375643025\ M\^66 + 
                    168680284\ M\^68 + 257209954\ M\^70 - 143263092\ M\^72 - 
                    144541207\ M\^74 + 101255668\ M\^76 + 62274749\ M\^78 - 
                    58345160\ M\^80 - 17019554\ M\^82 + 26128731\ M\^84 + 
                    264400\ M\^86 - 8568866\ M\^88 + 2533574\ M\^90 + 
                    1834449\ M\^92 - 1604913\ M\^94 - 141680\ M\^96 + 
                    790875\ M\^98 - 246473\ M\^100 - 231880\ M\^102 + 
                    183690\ M\^104 + 4596\ M\^106 - 67075\ M\^108 + 
                    45855\ M\^110 - 20375\ M\^112 + 8058\ M\^114 - 
                    2922\ M\^116 + 835\ M\^118 - 163\ M\^120 + 19\ M\^122 - 
                    M\^124)\) + 
              L\^22\ \((4\ M\^104 - 14\ M\^106 + 22\ M\^108 - 13\ M\^110 - 
                    18\ M\^112 + 55\ M\^114 - 7\ M\^116 - 15\ M\^118 + 
                    18\ M\^120 - 13\ M\^122 + 4\ M\^124)\) + 
              L\^10\ \((\(-126\)\ M\^10 + 2464\ M\^12 - 20893\ M\^14 + 
                    98477\ M\^16 - 264471\ M\^18 + 301101\ M\^20 + 
                    425215\ M\^22 - 1964691\ M\^24 + 1594069\ M\^26 + 
                    4449279\ M\^28 - 10337311\ M\^30 - 2379865\ M\^32 + 
                    30081139\ M\^34 - 15780208\ M\^36 - 62266686\ M\^38 + 
                    67933552\ M\^40 + 101038249\ M\^42 - 172061324\ M\^44 - 
                    132114439\ M\^46 + 335465081\ M\^48 + 134958472\ M\^50 - 
                    540470834\ M\^52 - 94292302\ M\^54 + 741991487\ M\^56 + 
                    11882950\ M\^58 - 882442389\ M\^60 + 90537624\ M\^62 + 
                    917794854\ M\^64 - 181871337\ M\^66 - 834467404\ M\^68 + 
                    238983181\ M\^70 + 660259094\ M\^72 - 248141529\ M\^74 - 
                    447267949\ M\^76 + 213286307\ M\^78 + 253066640\ M\^80 - 
                    153224488\ M\^82 - 114612704\ M\^84 + 91751099\ M\^86 + 
                    37581850\ M\^88 - 45440991\ M\^90 - 5159903\ M\^92 + 
                    17735686\ M\^94 - 3321313\ M\^96 - 4643806\ M\^98 + 
                    2844302\ M\^100 + 164431\ M\^102 - 834756\ M\^104 + 
                    295762\ M\^106 + 39819\ M\^108 - 44495\ M\^110 + 
                    26894\ M\^112 - 45891\ M\^114 + 44591\ M\^116 - 
                    23864\ M\^118 + 7744\ M\^120 - 1546\ M\^122 + 
                    177\ M\^124 - 9\ M\^126)\) + 
              L\^11\ \((\(-84\)\ M\^12 + 1674\ M\^14 - 14502\ M\^16 + 
                    69982\ M\^18 - 193720\ M\^20 + 240932\ M\^22 + 
                    225812\ M\^24 - 1238367\ M\^26 + 984822\ M\^28 + 
                    2812769\ M\^30 - 6028566\ M\^32 - 2708340\ M\^34 + 
                    19032529\ M\^36 - 5274808\ M\^38 - 47020861\ M\^40 + 
                    37506373\ M\^42 + 94272951\ M\^44 - 120470199\ M\^46 - 
                    154195037\ M\^48 + 279601376\ M\^50 + 204541169\ M\^52 - 
                    516694798\ M\^54 - 217941774\ M\^56 + 797011825\ M\^58 + 
                    176967179\ M\^60 - 1054959228\ M\^62 - 84195537\ M\^64 + 
                    1217543825\ M\^66 - 38879900\ M\^68 - 1231868207\ M\^70 + 
                    160213868\ M\^72 + 1093971487\ M\^74 - 242886110\ M\^76 - 
                    846165910\ M\^78 + 268037477\ M\^80 + 563330911\ M\^82 - 
                    239296676\ M\^84 - 315331316\ M\^86 + 177639831\ M\^88 + 
                    141558128\ M\^90 - 109277260\ M\^92 - 45409040\ M\^94 + 
                    53981729\ M\^96 + 6675801\ M\^98 - 20522029\ M\^100 + 
                    2306653\ M\^102 + 5758885\ M\^104 - 1788684\ M\^106 - 
                    1417915\ M\^108 + 830421\ M\^110 + 296340\ M\^112 - 
                    364071\ M\^114 + 13339\ M\^116 + 131168\ M\^118 - 
                    89629\ M\^120 + 31002\ M\^122 - 6295\ M\^124 + 
                    719\ M\^126 - 36\ M\^128)\) + 
              L\^12\ \((\(-36\)\ M\^14 + 719\ M\^16 - 6295\ M\^18 + 
                    31002\ M\^20 - 89629\ M\^22 + 131168\ M\^24 + 
                    13339\ M\^26 - 364071\ M\^28 + 296340\ M\^30 + 
                    830421\ M\^32 - 1417915\ M\^34 - 1788684\ M\^36 + 
                    5758885\ M\^38 + 2306653\ M\^40 - 20522029\ M\^42 + 
                    6675801\ M\^44 + 53981729\ M\^46 - 45409040\ M\^48 - 
                    109277260\ M\^50 + 141558128\ M\^52 + 177639831\ M\^54 - 
                    315331316\ M\^56 - 239296676\ M\^58 + 563330911\ M\^60 + 
                    268037477\ M\^62 - 846165910\ M\^64 - 242886110\ M\^66 + 
                    1093971487\ M\^68 + 160213868\ M\^70 - 
                    1231868207\ M\^72 - 38879900\ M\^74 + 1217543825\ M\^76 - 
                    84195537\ M\^78 - 1054959228\ M\^80 + 176967179\ M\^82 + 
                    797011825\ M\^84 - 217941774\ M\^86 - 516694798\ M\^88 + 
                    204541169\ M\^90 + 279601376\ M\^92 - 154195037\ M\^94 - 
                    120470199\ M\^96 + 94272951\ M\^98 + 37506373\ M\^100 - 
                    47020861\ M\^102 - 5274808\ M\^104 + 19032529\ M\^106 - 
                    2708340\ M\^108 - 6028566\ M\^110 + 2812769\ M\^112 + 
                    984822\ M\^114 - 1238367\ M\^116 + 225812\ M\^118 + 
                    240932\ M\^120 - 193720\ M\^122 + 69982\ M\^124 - 
                    14502\ M\^126 + 1674\ M\^128 - 84\ M\^130)\) + 
              L\^21\ \((6\ M\^90 - 42\ M\^92 + 139\ M\^94 - 240\ M\^96 + 
                    99\ M\^98 + 404\ M\^100 - 672\ M\^102 + 115\ M\^104 + 
                    668\ M\^106 - 689\ M\^108 + 135\ M\^110 + 619\ M\^112 - 
                    182\ M\^114 - 109\ M\^116 - 96\ M\^118 + 98\ M\^120 + 
                    60\ M\^122 - 120\ M\^124 + 90\ M\^126 - 36\ M\^128 + 
                    6\ M\^130)\) + 
              L\^13\ \((\(-9\)\ M\^16 + 177\ M\^18 - 1546\ M\^20 + 
                    7744\ M\^22 - 23864\ M\^24 + 44591\ M\^26 - 
                    45891\ M\^28 + 26894\ M\^30 - 44495\ M\^32 + 
                    39819\ M\^34 + 295762\ M\^36 - 834756\ M\^38 + 
                    164431\ M\^40 + 2844302\ M\^42 - 4643806\ M\^44 - 
                    3321313\ M\^46 + 17735686\ M\^48 - 5159903\ M\^50 - 
                    45440991\ M\^52 + 37581850\ M\^54 + 91751099\ M\^56 - 
                    114612704\ M\^58 - 153224488\ M\^60 + 253066640\ M\^62 + 
                    213286307\ M\^64 - 447267949\ M\^66 - 248141529\ M\^68 + 
                    660259094\ M\^70 + 238983181\ M\^72 - 834467404\ M\^74 - 
                    181871337\ M\^76 + 917794854\ M\^78 + 90537624\ M\^80 - 
                    882442389\ M\^82 + 11882950\ M\^84 + 741991487\ M\^86 - 
                    94292302\ M\^88 - 540470834\ M\^90 + 134958472\ M\^92 + 
                    335465081\ M\^94 - 132114439\ M\^96 - 172061324\ M\^98 + 
                    101038249\ M\^100 + 67933552\ M\^102 - 62266686\ M\^104 - 
                    15780208\ M\^106 + 30081139\ M\^108 - 2379865\ M\^110 - 
                    10337311\ M\^112 + 4449279\ M\^114 + 1594069\ M\^116 - 
                    1964691\ M\^118 + 425215\ M\^120 + 301101\ M\^122 - 
                    264471\ M\^124 + 98477\ M\^126 - 20893\ M\^128 + 
                    2464\ M\^130 - 126\ M\^132)\) + 
              L\^14\ \((\(-M\^18\) + 19\ M\^20 - 163\ M\^22 + 835\ M\^24 - 
                    2922\ M\^26 + 8058\ M\^28 - 20375\ M\^30 + 45855\ M\^32 - 
                    67075\ M\^34 + 4596\ M\^36 + 183690\ M\^38 - 
                    231880\ M\^40 - 246473\ M\^42 + 790875\ M\^44 - 
                    141680\ M\^46 - 1604913\ M\^48 + 1834449\ M\^50 + 
                    2533574\ M\^52 - 8568866\ M\^54 + 264400\ M\^56 + 
                    26128731\ M\^58 - 17019554\ M\^60 - 58345160\ M\^62 + 
                    62274749\ M\^64 + 101255668\ M\^66 - 144541207\ M\^68 - 
                    143263092\ M\^70 + 257209954\ M\^72 + 168680284\ M\^74 - 
                    375643025\ M\^76 - 164458742\ M\^78 + 465024672\ M\^80 + 
                    127599894\ M\^82 - 494512837\ M\^84 - 67075586\ M\^86 + 
                    456518212\ M\^88 + 3612692\ M\^90 - 365681280\ M\^92 + 
                    44752660\ M\^94 + 252110760\ M\^96 - 68550953\ M\^98 - 
                    145451909\ M\^100 + 67614588\ M\^102 + 65336860\ M\^104 - 
                    49845081\ M\^106 - 18179836\ M\^108 + 27142444\ M\^110 - 
                    954293\ M\^112 - 9949572\ M\^114 + 3989394\ M\^116 + 
                    1592605\ M\^118 - 1835033\ M\^120 + 396019\ M\^122 + 
                    267598\ M\^124 - 238269\ M\^126 + 90166\ M\^128 - 
                    19588\ M\^130 + 2380\ M\^132 - 126\ M\^134)\) + 
              L\^15\ \((3\ M\^26 - 59\ M\^28 + 519\ M\^30 - 2649\ M\^32 + 
                    8442\ M\^34 - 16589\ M\^36 + 18086\ M\^38 - 9742\ M\^40 + 
                    15279\ M\^42 - 28102\ M\^44 - 75917\ M\^46 + 
                    284862\ M\^48 - 67477\ M\^50 - 829536\ M\^52 + 
                    1110842\ M\^54 + 519368\ M\^56 - 2813175\ M\^58 + 
                    2823838\ M\^60 + 2999472\ M\^62 - 12057318\ M\^64 + 
                    3394522\ M\^66 + 28849482\ M\^68 - 22906368\ M\^70 - 
                    51608141\ M\^72 + 59999823\ M\^74 + 73093617\ M\^76 - 
                    110531958\ M\^78 - 83240681\ M\^80 + 159873609\ M\^82 + 
                    76024305\ M\^84 - 189769801\ M\^86 - 53289771\ M\^88 + 
                    190924186\ M\^90 + 23924016\ M\^92 - 164631526\ M\^94 + 
                    3921025\ M\^96 + 121644555\ M\^98 - 23271138\ M\^100 - 
                    74926430\ M\^102 + 29925103\ M\^104 + 35776314\ M\^106 - 
                    25049275\ M\^108 - 10711295\ M\^110 + 14667701\ M\^112 - 
                    176764\ M\^114 - 5667917\ M\^116 + 2169017\ M\^118 + 
                    988003\ M\^120 - 1059430\ M\^122 + 205597\ M\^124 + 
                    169889\ M\^126 - 143884\ M\^128 + 54385\ M\^130 - 
                    12031\ M\^132 + 1512\ M\^134 - 84\ M\^136)\) + 
              L\^20\ \((4\ M\^76 - 42\ M\^78 + 214\ M\^80 - 629\ M\^82 + 
                    980\ M\^84 - 263\ M\^86 - 1881\ M\^88 + 3204\ M\^90 - 
                    435\ M\^92 - 4990\ M\^94 + 4975\ M\^96 + 3034\ M\^98 - 
                    5564\ M\^100 - 2514\ M\^102 + 3612\ M\^104 + 
                    5664\ M\^106 - 3283\ M\^108 - 3615\ M\^110 + 
                    5994\ M\^112 - 1532\ M\^114 - 3838\ M\^116 + 
                    3268\ M\^118 + 125\ M\^120 - 1112\ M\^122 + 484\ M\^124 - 
                    211\ M\^126 + 264\ M\^128 - 229\ M\^130 + 116\ M\^132 - 
                    33\ M\^134 + 4\ M\^136)\) + 
              L\^16\ \((\(-3\)\ M\^34 + 61\ M\^36 - 560\ M\^38 + 
                    3008\ M\^40 - 10045\ M\^42 + 19508\ M\^44 - 
                    12649\ M\^46 - 32261\ M\^48 + 68886\ M\^50 + 
                    28605\ M\^52 - 215161\ M\^54 + 94879\ M\^56 + 
                    427386\ M\^58 - 455531\ M\^60 - 609470\ M\^62 + 
                    1410399\ M\^64 - 196301\ M\^66 - 2679678\ M\^68 + 
                    3820111\ M\^70 + 2305571\ M\^72 - 11648547\ M\^74 + 
                    3237783\ M\^76 + 22075951\ M\^78 - 15776252\ M\^80 - 
                    30676961\ M\^82 + 32723699\ M\^84 + 33040217\ M\^86 - 
                    47833934\ M\^88 - 27496927\ M\^90 + 55305587\ M\^92 + 
                    16196082\ M\^94 - 51925483\ M\^96 - 3085200\ M\^98 + 
                    39827534\ M\^100 - 6665329\ M\^102 - 24307291\ M\^104 + 
                    10148624\ M\^106 + 11030077\ M\^108 - 8438704\ M\^110 - 
                    2958685\ M\^112 + 4764699\ M\^114 - 235912\ M\^116 - 
                    1824336\ M\^118 + 732214\ M\^120 + 344473\ M\^122 - 
                    364872\ M\^124 + 56198\ M\^126 + 74138\ M\^128 - 
                    57648\ M\^130 + 21379\ M\^132 - 4750\ M\^134 + 
                    614\ M\^136 - 36\ M\^138)\) + 
              L\^17\ \((M\^42 - 23\ M\^44 + 240\ M\^46 - 1462\ M\^48 + 
                    5581\ M\^50 - 12875\ M\^52 + 12704\ M\^54 + 
                    17985\ M\^56 - 69846\ M\^58 + 40101\ M\^60 + 
                    155208\ M\^62 - 285011\ M\^64 - 62097\ M\^66 + 
                    642664\ M\^68 - 459699\ M\^70 - 699225\ M\^72 + 
                    1440485\ M\^74 - 300338\ M\^76 - 2269222\ M\^78 + 
                    2837500\ M\^80 + 1890761\ M\^82 - 6548744\ M\^84 + 
                    622106\ M\^86 + 9841532\ M\^88 - 5026339\ M\^90 - 
                    10533730\ M\^92 + 9445465\ M\^94 + 7665877\ M\^96 - 
                    11110828\ M\^98 - 2586246\ M\^100 + 9207505\ M\^102 - 
                    1512141\ M\^104 - 5275705\ M\^106 + 2830144\ M\^108 + 
                    1836052\ M\^110 - 2087628\ M\^112 - 124824\ M\^114 + 
                    928262\ M\^116 - 237930\ M\^118 - 269858\ M\^120 + 
                    157668\ M\^122 + 44434\ M\^124 - 62257\ M\^126 + 
                    4547\ M\^128 + 20497\ M\^130 - 14604\ M\^132 + 
                    5272\ M\^134 - 1152\ M\^136 + 149\ M\^138 - 
                    9\ M\^140)\) + 
              L\^18\ \((2\ M\^52 - 34\ M\^54 + 276\ M\^56 - 1343\ M\^58 + 
                    4080\ M\^60 - 7026\ M\^62 + 2450\ M\^64 + 17285\ M\^66 - 
                    33309\ M\^68 - 3521\ M\^70 + 87869\ M\^72 - 
                    93624\ M\^74 - 62978\ M\^76 + 230922\ M\^78 - 
                    179567\ M\^80 - 174897\ M\^82 + 618680\ M\^84 - 
                    392175\ M\^86 - 860858\ M\^88 + 1386753\ M\^90 + 
                    458891\ M\^92 - 2125490\ M\^94 + 495073\ M\^96 + 
                    1919607\ M\^98 - 1244559\ M\^100 - 860674\ M\^102 + 
                    1277567\ M\^104 - 138160\ M\^106 - 718030\ M\^108 + 
                    497957\ M\^110 + 146246\ M\^112 - 353449\ M\^114 + 
                    102482\ M\^116 + 110417\ M\^118 - 88948\ M\^120 - 
                    217\ M\^122 + 24753\ M\^124 - 8535\ M\^126 - 
                    374\ M\^128 - 1569\ M\^130 + 3142\ M\^132 - 
                    2068\ M\^134 + 742\ M\^136 - 158\ M\^138 + 19\ M\^140 - 
                    M\^142)\) + 
              L\^19\ \((M\^62 - 14\ M\^64 + 99\ M\^66 - 436\ M\^68 + 
                    1259\ M\^70 - 2319\ M\^72 + 2068\ M\^74 + 1634\ M\^76 - 
                    7803\ M\^78 + 10333\ M\^80 - 1170\ M\^82 - 25716\ M\^84 + 
                    49776\ M\^86 - 2209\ M\^88 - 111917\ M\^90 + 
                    94963\ M\^92 + 118972\ M\^94 - 186225\ M\^96 - 
                    45653\ M\^98 + 176884\ M\^100 - 30256\ M\^102 - 
                    71000\ M\^104 + 51195\ M\^106 - 10763\ M\^108 - 
                    20201\ M\^110 + 29998\ M\^112 - 5941\ M\^114 - 
                    19456\ M\^116 + 14246\ M\^118 + 5328\ M\^120 - 
                    10721\ M\^122 + 2740\ M\^124 + 3305\ M\^126 - 
                    3182\ M\^128 + 1339\ M\^130 - 432\ M\^132 + 211\ M\^134 - 
                    118\ M\^136 + 45\ M\^138 - 10\ M\^140 + M\^142)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 13]\), \(M\^76 + L\^26\ M\^76 + 
              L\ \((\(-3\)\ M\^64 + 11\ M\^66 - 21\ M\^68 + 22\ M\^70 - 
                    2\ M\^72 - 39\ M\^74 + 73\ M\^76 + 7\ M\^78 - 41\ M\^80 + 
                    25\ M\^82 - 6\ M\^84)\) + 
              L\^25\ \((\(-6\)\ M\^68 + 25\ M\^70 - 41\ M\^72 + 7\ M\^74 + 
                    73\ M\^76 - 39\ M\^78 - 2\ M\^80 + 22\ M\^82 - 
                    21\ M\^84 + 11\ M\^86 - 3\ M\^88)\) + 
              L\^2\ \((3\ M\^52 - 22\ M\^54 + 70\ M\^56 - 123\ M\^58 + 
                    126\ M\^60 - 39\ M\^62 - 63\ M\^64 - 43\ M\^66 - 
                    74\ M\^68 + 962\ M\^70 - 1148\ M\^72 - 1024\ M\^74 + 
                    2918\ M\^76 - 497\ M\^78 - 1937\ M\^80 + 1579\ M\^82 - 
                    19\ M\^84 - 668\ M\^86 + 434\ M\^88 - 125\ M\^90 + 
                    15\ M\^92)\) + 
              L\^3\ \((\(-M\^40\) + 11\ M\^42 - 50\ M\^44 + 129\ M\^46 - 
                    223\ M\^48 + 322\ M\^50 - 634\ M\^52 + 1467\ M\^54 - 
                    1829\ M\^56 - 1017\ M\^58 + 6669\ M\^60 - 7153\ M\^62 - 
                    2238\ M\^64 + 9826\ M\^66 - 7504\ M\^68 + 5408\ M\^70 - 
                    4341\ M\^72 - 11332\ M\^74 + 29214\ M\^76 - 9748\ M\^78 - 
                    22018\ M\^80 + 29254\ M\^82 - 6148\ M\^84 - 
                    16419\ M\^86 + 15031\ M\^88 - 2078\ M\^90 - 4575\ M\^92 + 
                    3680\ M\^94 - 1333\ M\^96 + 250\ M\^98 - 20\ M\^100)\) + 
              L\^24\ \((15\ M\^60 - 125\ M\^62 + 434\ M\^64 - 668\ M\^66 - 
                    19\ M\^68 + 1579\ M\^70 - 1937\ M\^72 - 497\ M\^74 + 
                    2918\ M\^76 - 1024\ M\^78 - 1148\ M\^80 + 962\ M\^82 - 
                    74\ M\^84 - 43\ M\^86 - 63\ M\^88 - 39\ M\^90 + 
                    126\ M\^92 - 123\ M\^94 + 70\ M\^96 - 22\ M\^98 + 
                    3\ M\^100)\) + 
              L\^4\ \((M\^32 - 11\ M\^34 + 59\ M\^36 - 237\ M\^38 + 
                    799\ M\^40 - 2028\ M\^42 + 3319\ M\^44 - 2392\ M\^46 - 
                    2038\ M\^48 + 4056\ M\^50 + 4722\ M\^52 - 10129\ M\^54 - 
                    20017\ M\^56 + 61545\ M\^58 - 20291\ M\^60 - 
                    110036\ M\^62 + 172725\ M\^64 - 40305\ M\^66 - 
                    221125\ M\^68 + 340758\ M\^70 - 77254\ M\^72 - 
                    376018\ M\^74 + 463973\ M\^76 + 9347\ M\^78 - 
                    419616\ M\^80 + 332935\ M\^82 + 46009\ M\^84 - 
                    261289\ M\^86 + 169856\ M\^88 + 12071\ M\^90 - 
                    94230\ M\^92 + 67553\ M\^94 - 10592\ M\^96 - 
                    18291\ M\^98 + 16823\ M\^100 - 7279\ M\^102 + 
                    1812\ M\^104 - 250\ M\^106 + 15\ M\^108)\) + 
              L\^23\ \((\(-20\)\ M\^52 + 250\ M\^54 - 1333\ M\^56 + 
                    3680\ M\^58 - 4575\ M\^60 - 2078\ M\^62 + 15031\ M\^64 - 
                    16419\ M\^66 - 6148\ M\^68 + 29254\ M\^70 - 
                    22018\ M\^72 - 9748\ M\^74 + 29214\ M\^76 - 
                    11332\ M\^78 - 4341\ M\^80 + 5408\ M\^82 - 7504\ M\^84 + 
                    9826\ M\^86 - 2238\ M\^88 - 7153\ M\^90 + 6669\ M\^92 - 
                    1017\ M\^94 - 1829\ M\^96 + 1467\ M\^98 - 634\ M\^100 + 
                    322\ M\^102 - 223\ M\^104 + 129\ M\^106 - 50\ M\^108 + 
                    11\ M\^110 - M\^112)\) + 
              L\^5\ \((\(-8\)\ M\^28 + 107\ M\^30 - 656\ M\^32 + 
                    2500\ M\^34 - 6735\ M\^36 + 13215\ M\^38 - 16264\ M\^40 + 
                    350\ M\^42 + 44363\ M\^44 - 74203\ M\^46 - 4167\ M\^48 + 
                    167261\ M\^50 - 157000\ M\^52 - 186036\ M\^54 + 
                    453892\ M\^56 - 8114\ M\^58 - 922172\ M\^60 + 
                    898017\ M\^62 + 979928\ M\^64 - 2590110\ M\^66 + 
                    530621\ M\^68 + 3554436\ M\^70 - 3487902\ M\^72 - 
                    1735480\ M\^74 + 5223848\ M\^76 - 2033357\ M\^78 - 
                    3324545\ M\^80 + 4268484\ M\^82 - 397309\ M\^84 - 
                    2965162\ M\^86 + 2295989\ M\^88 + 327084\ M\^90 - 
                    1521650\ M\^92 + 821158\ M\^94 + 170649\ M\^96 - 
                    445208\ M\^98 + 228159\ M\^100 - 6289\ M\^102 - 
                    65693\ M\^104 + 50670\ M\^106 - 22179\ M\^108 + 
                    6332\ M\^110 - 1163\ M\^112 + 125\ M\^114 - 
                    6\ M\^116)\) + 
              L\^22\ \((15\ M\^44 - 250\ M\^46 + 1812\ M\^48 - 7279\ M\^50 + 
                    16823\ M\^52 - 18291\ M\^54 - 10592\ M\^56 + 
                    67553\ M\^58 - 94230\ M\^60 + 12071\ M\^62 + 
                    169856\ M\^64 - 261289\ M\^66 + 46009\ M\^68 + 
                    332935\ M\^70 - 419616\ M\^72 + 9347\ M\^74 + 
                    463973\ M\^76 - 376018\ M\^78 - 77254\ M\^80 + 
                    340758\ M\^82 - 221125\ M\^84 - 40305\ M\^86 + 
                    172725\ M\^88 - 110036\ M\^90 - 20291\ M\^92 + 
                    61545\ M\^94 - 20017\ M\^96 - 10129\ M\^98 + 
                    4722\ M\^100 + 4056\ M\^102 - 2038\ M\^104 - 
                    2392\ M\^106 + 3319\ M\^108 - 2028\ M\^110 + 
                    799\ M\^112 - 237\ M\^114 + 59\ M\^116 - 11\ M\^118 + 
                    M\^120)\) + 
              L\^21\ \((\(-6\)\ M\^36 + 125\ M\^38 - 1163\ M\^40 + 
                    6332\ M\^42 - 22179\ M\^44 + 50670\ M\^46 - 
                    65693\ M\^48 - 6289\ M\^50 + 228159\ M\^52 - 
                    445208\ M\^54 + 170649\ M\^56 + 821158\ M\^58 - 
                    1521650\ M\^60 + 327084\ M\^62 + 2295989\ M\^64 - 
                    2965162\ M\^66 - 397309\ M\^68 + 4268484\ M\^70 - 
                    3324545\ M\^72 - 2033357\ M\^74 + 5223848\ M\^76 - 
                    1735480\ M\^78 - 3487902\ M\^80 + 3554436\ M\^82 + 
                    530621\ M\^84 - 2590110\ M\^86 + 979928\ M\^88 + 
                    898017\ M\^90 - 922172\ M\^92 - 8114\ M\^94 + 
                    453892\ M\^96 - 186036\ M\^98 - 157000\ M\^100 + 
                    167261\ M\^102 - 4167\ M\^104 - 74203\ M\^106 + 
                    44363\ M\^108 + 350\ M\^110 - 16264\ M\^112 + 
                    13215\ M\^114 - 6735\ M\^116 + 2500\ M\^118 - 
                    656\ M\^120 + 107\ M\^122 - 8\ M\^124)\) + 
              L\^6\ \((28\ M\^24 - 436\ M\^26 + 3089\ M\^28 - 12962\ M\^30 + 
                    34605\ M\^32 - 56026\ M\^34 + 32180\ M\^36 + 
                    90060\ M\^38 - 274386\ M\^40 + 270749\ M\^42 + 
                    277521\ M\^44 - 1134321\ M\^46 + 919414\ M\^48 + 
                    1295432\ M\^50 - 3258172\ M\^52 + 1017109\ M\^54 + 
                    4802320\ M\^56 - 6449551\ M\^58 - 2434230\ M\^60 + 
                    13075377\ M\^62 - 6023251\ M\^64 - 16191301\ M\^66 + 
                    19381110\ M\^68 + 9489931\ M\^70 - 30409461\ M\^72 + 
                    8304350\ M\^74 + 28526152\ M\^76 - 26840070\ M\^78 - 
                    10594753\ M\^80 + 31492606\ M\^82 - 10380430\ M\^84 - 
                    18549323\ M\^86 + 18493488\ M\^88 + 1142225\ M\^90 - 
                    11876569\ M\^92 + 6655753\ M\^94 + 2033181\ M\^96 - 
                    4536839\ M\^98 + 2002549\ M\^100 + 622840\ M\^102 - 
                    1186975\ M\^104 + 542248\ M\^106 + 33662\ M\^108 - 
                    180745\ M\^110 + 116992\ M\^112 - 44832\ M\^114 + 
                    11867\ M\^116 - 2249\ M\^118 + 298\ M\^120 - 25\ M\^122 + 
                    M\^124)\) + 
              L\^7\ \((\(-56\)\ M\^20 + 985\ M\^22 - 7873\ M\^24 + 
                    36818\ M\^26 - 106304\ M\^28 + 171701\ M\^30 - 
                    41110\ M\^32 - 479096\ M\^34 + 1050746\ M\^36 - 
                    502872\ M\^38 - 2034581\ M\^40 + 4582653\ M\^42 - 
                    1812471\ M\^44 - 8170933\ M\^46 + 14203786\ M\^48 + 
                    276264\ M\^50 - 27701055\ M\^52 + 27833182\ M\^54 + 
                    21329609\ M\^56 - 65750542\ M\^58 + 21609747\ M\^60 + 
                    82773117\ M\^62 - 91264495\ M\^64 - 48719302\ M\^66 + 
                    147699617\ M\^68 - 37685739\ M\^70 - 146035727\ M\^72 + 
                    136260036\ M\^74 + 71801529\ M\^76 - 184998434\ M\^78 + 
                    39769472\ M\^80 + 150724035\ M\^82 - 119253925\ M\^84 - 
                    58615652\ M\^86 + 122987872\ M\^88 - 24666085\ M\^90 - 
                    68966618\ M\^92 + 52859814\ M\^94 + 11640394\ M\^96 - 
                    34988006\ M\^98 + 13111407\ M\^100 + 9072407\ M\^102 - 
                    10591249\ M\^104 + 2406804\ M\^106 + 2394946\ M\^108 - 
                    2195810\ M\^110 + 607140\ M\^112 + 242801\ M\^114 - 
                    329272\ M\^116 + 175949\ M\^118 - 59511\ M\^120 + 
                    13528\ M\^122 - 2011\ M\^124 + 177\ M\^126 - 
                    7\ M\^128)\) + 
              L\^20\ \((M\^28 - 25\ M\^30 + 298\ M\^32 - 2249\ M\^34 + 
                    11867\ M\^36 - 44832\ M\^38 + 116992\ M\^40 - 
                    180745\ M\^42 + 33662\ M\^44 + 542248\ M\^46 - 
                    1186975\ M\^48 + 622840\ M\^50 + 2002549\ M\^52 - 
                    4536839\ M\^54 + 2033181\ M\^56 + 6655753\ M\^58 - 
                    11876569\ M\^60 + 1142225\ M\^62 + 18493488\ M\^64 - 
                    18549323\ M\^66 - 10380430\ M\^68 + 31492606\ M\^70 - 
                    10594753\ M\^72 - 26840070\ M\^74 + 28526152\ M\^76 + 
                    8304350\ M\^78 - 30409461\ M\^80 + 9489931\ M\^82 + 
                    19381110\ M\^84 - 16191301\ M\^86 - 6023251\ M\^88 + 
                    13075377\ M\^90 - 2434230\ M\^92 - 6449551\ M\^94 + 
                    4802320\ M\^96 + 1017109\ M\^98 - 3258172\ M\^100 + 
                    1295432\ M\^102 + 919414\ M\^104 - 1134321\ M\^106 + 
                    277521\ M\^108 + 270749\ M\^110 - 274386\ M\^112 + 
                    90060\ M\^114 + 32180\ M\^116 - 56026\ M\^118 + 
                    34605\ M\^120 - 12962\ M\^122 + 3089\ M\^124 - 
                    436\ M\^126 + 28\ M\^128)\) + 
              L\^19\ \((\(-7\)\ M\^24 + 177\ M\^26 - 2011\ M\^28 + 
                    13528\ M\^30 - 59511\ M\^32 + 175949\ M\^34 - 
                    329272\ M\^36 + 242801\ M\^38 + 607140\ M\^40 - 
                    2195810\ M\^42 + 2394946\ M\^44 + 2406804\ M\^46 - 
                    10591249\ M\^48 + 9072407\ M\^50 + 13111407\ M\^52 - 
                    34988006\ M\^54 + 11640394\ M\^56 + 52859814\ M\^58 - 
                    68966618\ M\^60 - 24666085\ M\^62 + 122987872\ M\^64 - 
                    58615652\ M\^66 - 119253925\ M\^68 + 150724035\ M\^70 + 
                    39769472\ M\^72 - 184998434\ M\^74 + 71801529\ M\^76 + 
                    136260036\ M\^78 - 146035727\ M\^80 - 37685739\ M\^82 + 
                    147699617\ M\^84 - 48719302\ M\^86 - 91264495\ M\^88 + 
                    82773117\ M\^90 + 21609747\ M\^92 - 65750542\ M\^94 + 
                    21329609\ M\^96 + 27833182\ M\^98 - 27701055\ M\^100 + 
                    276264\ M\^102 + 14203786\ M\^104 - 8170933\ M\^106 - 
                    1812471\ M\^108 + 4582653\ M\^110 - 2034581\ M\^112 - 
                    502872\ M\^114 + 1050746\ M\^116 - 479096\ M\^118 - 
                    41110\ M\^120 + 171701\ M\^122 - 106304\ M\^124 + 
                    36818\ M\^126 - 7873\ M\^128 + 985\ M\^130 - 
                    56\ M\^132)\) + 
              L\^8\ \((70\ M\^16 - 1360\ M\^18 + 11995\ M\^20 - 
                    61810\ M\^22 + 196715\ M\^24 - 353144\ M\^26 + 
                    120670\ M\^28 + 1049983\ M\^30 - 2476435\ M\^32 + 
                    1049336\ M\^34 + 5702469\ M\^36 - 11851103\ M\^38 + 
                    2326669\ M\^40 + 26047701\ M\^42 - 38314351\ M\^44 - 
                    12225186\ M\^46 + 93513205\ M\^48 - 73080818\ M\^50 - 
                    103476241\ M\^52 + 224629183\ M\^54 - 19626014\ M\^56 - 
                    342308496\ M\^58 + 290949736\ M\^60 + 276590490\ M\^62 - 
                    583184042\ M\^64 + 39142469\ M\^66 + 688384120\ M\^68 - 
                    493631678\ M\^70 - 481126403\ M\^72 + 841798307\ M\^74 + 
                    35844394\ M\^76 - 875189570\ M\^78 + 413235958\ M\^80 + 
                    587419436\ M\^82 - 626693549\ M\^84 - 169012574\ M\^86 + 
                    538883038\ M\^88 - 138683379\ M\^90 - 281168701\ M\^92 + 
                    222766102\ M\^94 + 49217973\ M\^96 - 146974294\ M\^98 + 
                    52889206\ M\^100 + 43888440\ M\^102 - 49714959\ M\^104 + 
                    8259783\ M\^106 + 16980971\ M\^108 - 13229856\ M\^110 + 
                    1018847\ M\^112 + 4168568\ M\^114 - 2698023\ M\^116 + 
                    277174\ M\^118 + 589354\ M\^120 - 442652\ M\^122 + 
                    169466\ M\^124 - 40515\ M\^126 + 6116\ M\^128 - 
                    537\ M\^130 + 21\ M\^132)\) + 
              L\^9\ \((\(-56\)\ M\^12 + 1181\ M\^14 - 11291\ M\^16 + 
                    63233\ M\^18 - 221291\ M\^20 + 456019\ M\^22 - 
                    310338\ M\^24 - 1065263\ M\^26 + 3253824\ M\^28 - 
                    2157957\ M\^30 - 7060774\ M\^32 + 17560565\ M\^34 - 
                    5141910\ M\^36 - 39884950\ M\^38 + 61993244\ M\^40 + 
                    23120606\ M\^42 - 164575658\ M\^44 + 120116282\ M\^46 + 
                    216804219\ M\^48 - 431906217\ M\^50 - 24268157\ M\^52 + 
                    778872175\ M\^54 - 559227729\ M\^56 - 828148332\ M\^58 + 
                    1403638153\ M\^60 + 263637341\ M\^62 - 
                    2050810954\ M\^64 + 885637912\ M\^66 + 
                    1992380338\ M\^68 - 2120663554\ M\^70 - 
                    1100425691\ M\^72 + 2749723560\ M\^74 - 
                    205382771\ M\^76 - 2394267647\ M\^78 + 
                    1235130401\ M\^80 + 1300710432\ M\^82 - 
                    1499366121\ M\^84 - 120586098\ M\^86 + 
                    1042976426\ M\^88 - 570888559\ M\^90 - 324451609\ M\^92 + 
                    639171937\ M\^94 - 189588342\ M\^96 - 351992474\ M\^98 + 
                    331816932\ M\^100 + 58905020\ M\^102 - 
                    229606743\ M\^104 + 72491681\ M\^106 + 85460606\ M\^108 - 
                    71087663\ M\^110 - 6282549\ M\^112 + 30821532\ M\^114 - 
                    11279466\ M\^116 - 5032013\ M\^118 + 5631506\ M\^120 - 
                    1282112\ M\^122 - 773746\ M\^124 + 727393\ M\^126 - 
                    289992\ M\^128 + 69615\ M\^130 - 10426\ M\^132 + 
                    905\ M\^134 - 35\ M\^136)\) + 
              L\^18\ \((21\ M\^20 - 537\ M\^22 + 6116\ M\^24 - 40515\ M\^26 + 
                    169466\ M\^28 - 442652\ M\^30 + 589354\ M\^32 + 
                    277174\ M\^34 - 2698023\ M\^36 + 4168568\ M\^38 + 
                    1018847\ M\^40 - 13229856\ M\^42 + 16980971\ M\^44 + 
                    8259783\ M\^46 - 49714959\ M\^48 + 43888440\ M\^50 + 
                    52889206\ M\^52 - 146974294\ M\^54 + 49217973\ M\^56 + 
                    222766102\ M\^58 - 281168701\ M\^60 - 138683379\ M\^62 + 
                    538883038\ M\^64 - 169012574\ M\^66 - 626693549\ M\^68 + 
                    587419436\ M\^70 + 413235958\ M\^72 - 875189570\ M\^74 + 
                    35844394\ M\^76 + 841798307\ M\^78 - 481126403\ M\^80 - 
                    493631678\ M\^82 + 688384120\ M\^84 + 39142469\ M\^86 - 
                    583184042\ M\^88 + 276590490\ M\^90 + 290949736\ M\^92 - 
                    342308496\ M\^94 - 19626014\ M\^96 + 224629183\ M\^98 - 
                    103476241\ M\^100 - 73080818\ M\^102 + 93513205\ M\^104 - 
                    12225186\ M\^106 - 38314351\ M\^108 + 26047701\ M\^110 + 
                    2326669\ M\^112 - 11851103\ M\^114 + 5702469\ M\^116 + 
                    1049336\ M\^118 - 2476435\ M\^120 + 1049983\ M\^122 + 
                    120670\ M\^124 - 353144\ M\^126 + 196715\ M\^128 - 
                    61810\ M\^130 + 11995\ M\^132 - 1360\ M\^134 + 
                    70\ M\^136)\) + 
              L\^17\ \((\(-35\)\ M\^16 + 905\ M\^18 - 10426\ M\^20 + 
                    69615\ M\^22 - 289992\ M\^24 + 727393\ M\^26 - 
                    773746\ M\^28 - 1282112\ M\^30 + 5631506\ M\^32 - 
                    5032013\ M\^34 - 11279466\ M\^36 + 30821532\ M\^38 - 
                    6282549\ M\^40 - 71087663\ M\^42 + 85460606\ M\^44 + 
                    72491681\ M\^46 - 229606743\ M\^48 + 58905020\ M\^50 + 
                    331816932\ M\^52 - 351992474\ M\^54 - 189588342\ M\^56 + 
                    639171937\ M\^58 - 324451609\ M\^60 - 570888559\ M\^62 + 
                    1042976426\ M\^64 - 120586098\ M\^66 - 
                    1499366121\ M\^68 + 1300710432\ M\^70 + 
                    1235130401\ M\^72 - 2394267647\ M\^74 - 
                    205382771\ M\^76 + 2749723560\ M\^78 - 
                    1100425691\ M\^80 - 2120663554\ M\^82 + 
                    1992380338\ M\^84 + 885637912\ M\^86 - 
                    2050810954\ M\^88 + 263637341\ M\^90 + 
                    1403638153\ M\^92 - 828148332\ M\^94 - 559227729\ M\^96 + 
                    778872175\ M\^98 - 24268157\ M\^100 - 431906217\ M\^102 + 
                    216804219\ M\^104 + 120116282\ M\^106 - 
                    164575658\ M\^108 + 23120606\ M\^110 + 61993244\ M\^112 - 
                    39884950\ M\^114 - 5141910\ M\^116 + 17560565\ M\^118 - 
                    7060774\ M\^120 - 2157957\ M\^122 + 3253824\ M\^124 - 
                    1065263\ M\^126 - 310338\ M\^128 + 456019\ M\^130 - 
                    221291\ M\^132 + 63233\ M\^134 - 11291\ M\^136 + 
                    1181\ M\^138 - 56\ M\^140)\) + 
              L\^10\ \((28\ M\^8 - 632\ M\^10 + 6459\ M\^12 - 38872\ M\^14 + 
                    148816\ M\^16 - 354525\ M\^18 + 403788\ M\^20 + 
                    367662\ M\^22 - 2110416\ M\^24 + 2315178\ M\^26 + 
                    3546139\ M\^28 - 12888608\ M\^30 + 7016692\ M\^32 + 
                    28074389\ M\^34 - 51614260\ M\^36 - 17114782\ M\^38 + 
                    145261934\ M\^40 - 100344940\ M\^42 - 232395717\ M\^44 + 
                    414311779\ M\^46 + 130074333\ M\^48 - 897082674\ M\^50 + 
                    419997107\ M\^52 + 1258328949\ M\^54 - 
                    1504548209\ M\^56 - 1011822918\ M\^58 + 
                    2774500145\ M\^60 - 150873348\ M\^62 - 
                    3495263249\ M\^64 + 1917060334\ M\^66 + 
                    3054058500\ M\^68 - 3332549696\ M\^70 - 
                    1535029247\ M\^72 + 3392143279\ M\^74 - 
                    220472396\ M\^76 - 1857003573\ M\^78 + 
                    1196076569\ M\^80 - 443691855\ M\^82 - 902071904\ M\^84 + 
                    2246051444\ M\^86 - 268659220\ M\^88 - 
                    2750304688\ M\^90 + 1425102381\ M\^92 + 
                    2075255489\ M\^94 - 1888363491\ M\^96 - 
                    962750731\ M\^98 + 1599142782\ M\^100 + 
                    109711846\ M\^102 - 958895814\ M\^104 + 
                    240617124\ M\^106 + 397089402\ M\^108 - 
                    237454920\ M\^110 - 90227220\ M\^112 + 
                    123342292\ M\^114 - 9488691\ M\^116 - 38437547\ M\^118 + 
                    17265657\ M\^120 + 4417856\ M\^122 - 6345117\ M\^124 + 
                    1558533\ M\^126 + 790862\ M\^128 - 768750\ M\^130 + 
                    304386\ M\^132 - 72180\ M\^134 + 10671\ M\^136 - 
                    915\ M\^138 + 35\ M\^140)\) + 
              L\^11\ \((\(-8\)\ M\^4 + 191\ M\^6 - 2066\ M\^8 + 
                    13291\ M\^10 - 55840\ M\^12 + 156021\ M\^14 - 
                    268766\ M\^16 + 167761\ M\^18 + 394339\ M\^20 - 
                    919048\ M\^22 - 295575\ M\^24 + 3598537\ M\^26 - 
                    3140559\ M\^28 - 8566176\ M\^30 + 18567731\ M\^32 + 
                    10343179\ M\^34 - 68751096\ M\^36 + 33379675\ M\^38 + 
                    162861762\ M\^40 - 221344295\ M\^42 - 231583148\ M\^44 + 
                    671813576\ M\^46 + 59587479\ M\^48 - 1374055126\ M\^50 + 
                    635442128\ M\^52 + 2059299692\ M\^54 - 
                    1949256475\ M\^56 - 2293011243\ M\^58 + 
                    3541858464\ M\^60 + 1880481458\ M\^62 - 
                    4744081543\ M\^64 - 1259151297\ M\^66 + 
                    5051766786\ M\^68 + 1393271024\ M\^70 - 
                    4599107615\ M\^72 - 3024355168\ M\^74 + 
                    4088129746\ M\^76 + 5834005509\ M\^78 - 
                    4214398947\ M\^80 - 8386905530\ M\^82 + 
                    5125161324\ M\^84 + 9202328879\ M\^86 - 
                    6201541092\ M\^88 - 7825062500\ M\^90 + 
                    6548960670\ M\^92 + 5063531821\ M\^94 - 
                    5733954697\ M\^96 - 2275360924\ M\^98 + 
                    4075523407\ M\^100 + 422941670\ M\^102 - 
                    2306791155\ M\^104 + 339413920\ M\^106 + 
                    1001935277\ M\^108 - 405131131\ M\^110 - 
                    305044031\ M\^112 + 238189899\ M\^114 + 
                    43374121\ M\^116 - 92426215\ M\^118 + 13200782\ M\^120 + 
                    23692000\ M\^122 - 11334383\ M\^124 - 2428121\ M\^126 + 
                    3636979\ M\^128 - 734930\ M\^130 - 594765\ M\^132 + 
                    504372\ M\^134 - 192670\ M\^136 + 44834\ M\^138 - 
                    6543\ M\^140 + 555\ M\^142 - 21\ M\^144)\) + 
              L\^16\ \((35\ M\^12 - 915\ M\^14 + 10671\ M\^16 - 
                    72180\ M\^18 + 304386\ M\^20 - 768750\ M\^22 + 
                    790862\ M\^24 + 1558533\ M\^26 - 6345117\ M\^28 + 
                    4417856\ M\^30 + 17265657\ M\^32 - 38437547\ M\^34 - 
                    9488691\ M\^36 + 123342292\ M\^38 - 90227220\ M\^40 - 
                    237454920\ M\^42 + 397089402\ M\^44 + 240617124\ M\^46 - 
                    958895814\ M\^48 + 109711846\ M\^50 + 1599142782\ M\^52 - 
                    962750731\ M\^54 - 1888363491\ M\^56 + 
                    2075255489\ M\^58 + 1425102381\ M\^60 - 
                    2750304688\ M\^62 - 268659220\ M\^64 + 
                    2246051444\ M\^66 - 902071904\ M\^68 - 443691855\ M\^70 + 
                    1196076569\ M\^72 - 1857003573\ M\^74 - 
                    220472396\ M\^76 + 3392143279\ M\^78 - 
                    1535029247\ M\^80 - 3332549696\ M\^82 + 
                    3054058500\ M\^84 + 1917060334\ M\^86 - 
                    3495263249\ M\^88 - 150873348\ M\^90 + 
                    2774500145\ M\^92 - 1011822918\ M\^94 - 
                    1504548209\ M\^96 + 1258328949\ M\^98 + 
                    419997107\ M\^100 - 897082674\ M\^102 + 
                    130074333\ M\^104 + 414311779\ M\^106 - 
                    232395717\ M\^108 - 100344940\ M\^110 + 
                    145261934\ M\^112 - 17114782\ M\^114 - 51614260\ M\^116 + 
                    28074389\ M\^118 + 7016692\ M\^120 - 12888608\ M\^122 + 
                    3546139\ M\^124 + 2315178\ M\^126 - 2110416\ M\^128 + 
                    367662\ M\^130 + 403788\ M\^132 - 354525\ M\^134 + 
                    148816\ M\^136 - 38872\ M\^138 + 6459\ M\^140 - 
                    632\ M\^142 + 28\ M\^144)\) + 
              L\^15\ \((\(-21\)\ M\^8 + 555\ M\^10 - 6543\ M\^12 + 
                    44834\ M\^14 - 192670\ M\^16 + 504372\ M\^18 - 
                    594765\ M\^20 - 734930\ M\^22 + 3636979\ M\^24 - 
                    2428121\ M\^26 - 11334383\ M\^28 + 23692000\ M\^30 + 
                    13200782\ M\^32 - 92426215\ M\^34 + 43374121\ M\^36 + 
                    238189899\ M\^38 - 305044031\ M\^40 - 405131131\ M\^42 + 
                    1001935277\ M\^44 + 339413920\ M\^46 - 
                    2306791155\ M\^48 + 422941670\ M\^50 + 
                    4075523407\ M\^52 - 2275360924\ M\^54 - 
                    5733954697\ M\^56 + 5063531821\ M\^58 + 
                    6548960670\ M\^60 - 7825062500\ M\^62 - 
                    6201541092\ M\^64 + 9202328879\ M\^66 + 
                    5125161324\ M\^68 - 8386905530\ M\^70 - 
                    4214398947\ M\^72 + 5834005509\ M\^74 + 
                    4088129746\ M\^76 - 3024355168\ M\^78 - 
                    4599107615\ M\^80 + 1393271024\ M\^82 + 
                    5051766786\ M\^84 - 1259151297\ M\^86 - 
                    4744081543\ M\^88 + 1880481458\ M\^90 + 
                    3541858464\ M\^92 - 2293011243\ M\^94 - 
                    1949256475\ M\^96 + 2059299692\ M\^98 + 
                    635442128\ M\^100 - 1374055126\ M\^102 + 
                    59587479\ M\^104 + 671813576\ M\^106 - 
                    231583148\ M\^108 - 221344295\ M\^110 + 
                    162861762\ M\^112 + 33379675\ M\^114 - 68751096\ M\^116 + 
                    10343179\ M\^118 + 18567731\ M\^120 - 8566176\ M\^122 - 
                    3140559\ M\^124 + 3598537\ M\^126 - 295575\ M\^128 - 
                    919048\ M\^130 + 394339\ M\^132 + 167761\ M\^134 - 
                    268766\ M\^136 + 156021\ M\^138 - 55840\ M\^140 + 
                    13291\ M\^142 - 2066\ M\^144 + 191\ M\^146 - 
                    8\ M\^148)\) + 
              L\^12\ \((1 - 25\ M\^2 + 286\ M\^4 - 1999\ M\^6 + 9611\ M\^8 - 
                    33699\ M\^10 + 87673\ M\^12 - 164443\ M\^14 + 
                    200167\ M\^16 - 114233\ M\^18 + 13514\ M\^20 - 
                    196979\ M\^22 + 189383\ M\^24 + 1889853\ M\^26 - 
                    4636417\ M\^28 - 3528017\ M\^30 + 27183878\ M\^32 - 
                    20509740\ M\^34 - 73310117\ M\^36 + 135620650\ M\^38 + 
                    100470484\ M\^40 - 446142654\ M\^42 + 53221251\ M\^44 + 
                    1016457711\ M\^46 - 678826302\ M\^48 - 
                    1759249887\ M\^50 + 2083446873\ M\^52 + 
                    2393586806\ M\^54 - 4346148254\ M\^56 - 
                    2647995291\ M\^58 + 7163421750\ M\^60 + 
                    2628512590\ M\^62 - 9963353012\ M\^64 - 
                    3020612491\ M\^66 + 12245804822\ M\^68 + 
                    4716063285\ M\^70 - 13821231486\ M\^72 - 
                    7966414072\ M\^74 + 14753720747\ M\^76 + 
                    11745897456\ M\^78 - 15141963016\ M\^80 - 
                    14137565679\ M\^82 + 14997315133\ M\^84 + 
                    13759051161\ M\^86 - 14069816045\ M\^88 - 
                    10731061071\ M\^90 + 12086434165\ M\^92 + 
                    6489091123\ M\^94 - 9167628197\ M\^96 - 
                    2732129827\ M\^98 + 5941203351\ M\^100 + 
                    420913324\ M\^102 - 3182534760\ M\^104 + 
                    465895882\ M\^106 + 1349209383\ M\^108 - 
                    511423151\ M\^110 - 415112695\ M\^112 + 
                    292752224\ M\^114 + 70240067\ M\^116 - 
                    112647724\ M\^118 + 8229978\ M\^120 + 28828661\ M\^122 - 
                    9073204\ M\^124 - 4841646\ M\^126 + 3206166\ M\^128 + 
                    309077\ M\^130 - 680728\ M\^132 - 30444\ M\^134 + 
                    292745\ M\^136 - 191081\ M\^138 + 68272\ M\^140 - 
                    15463\ M\^142 + 2226\ M\^144 - 187\ M\^146 + 
                    7\ M\^148)\) + 
              L\^13\ \((\(-1\) + 27\ M\^2 - 326\ M\^4 + 2330\ M\^6 - 
                    10956\ M\^8 + 35518\ M\^10 - 81386\ M\^12 + 
                    139590\ M\^14 - 220818\ M\^16 + 390513\ M\^18 - 
                    443061\ M\^20 - 699156\ M\^22 + 3411486\ M\^24 - 
                    2151017\ M\^26 - 11622943\ M\^28 + 22799825\ M\^30 + 
                    16590371\ M\^32 - 90484455\ M\^34 + 22385662\ M\^36 + 
                    252053223\ M\^38 - 235654075\ M\^40 - 518366810\ M\^42 + 
                    879771868\ M\^44 + 765757573\ M\^46 - 2297040051\ M\^48 - 
                    647856065\ M\^50 + 4721750639\ M\^52 - 279965319\ M\^54 - 
                    8066284057\ M\^56 + 2174620224\ M\^58 + 
                    11873376623\ M\^60 - 4468369784\ M\^62 - 
                    15534211297\ M\^64 + 5884338447\ M\^66 + 
                    18535885948\ M\^68 - 5154720313\ M\^70 - 
                    20526923870\ M\^72 + 2067675949\ M\^74 + 
                    21238640488\ M\^76 + 2067675949\ M\^78 - 
                    20526923870\ M\^80 - 5154720313\ M\^82 + 
                    18535885948\ M\^84 + 5884338447\ M\^86 - 
                    15534211297\ M\^88 - 4468369784\ M\^90 + 
                    11873376623\ M\^92 + 2174620224\ M\^94 - 
                    8066284057\ M\^96 - 279965319\ M\^98 + 
                    4721750639\ M\^100 - 647856065\ M\^102 - 
                    2297040051\ M\^104 + 765757573\ M\^106 + 
                    879771868\ M\^108 - 518366810\ M\^110 - 
                    235654075\ M\^112 + 252053223\ M\^114 + 
                    22385662\ M\^116 - 90484455\ M\^118 + 16590371\ M\^120 + 
                    22799825\ M\^122 - 11622943\ M\^124 - 2151017\ M\^126 + 
                    3411486\ M\^128 - 699156\ M\^130 - 443061\ M\^132 + 
                    390513\ M\^134 - 220818\ M\^136 + 139590\ M\^138 - 
                    81386\ M\^140 + 35518\ M\^142 - 10956\ M\^144 + 
                    2330\ M\^146 - 326\ M\^148 + 27\ M\^150 - M\^152)\) + 
              L\^14\ \((7\ M\^4 - 187\ M\^6 + 2226\ M\^8 - 15463\ M\^10 + 
                    68272\ M\^12 - 191081\ M\^14 + 292745\ M\^16 - 
                    30444\ M\^18 - 680728\ M\^20 + 309077\ M\^22 + 
                    3206166\ M\^24 - 4841646\ M\^26 - 9073204\ M\^28 + 
                    28828661\ M\^30 + 8229978\ M\^32 - 112647724\ M\^34 + 
                    70240067\ M\^36 + 292752224\ M\^38 - 415112695\ M\^40 - 
                    511423151\ M\^42 + 1349209383\ M\^44 + 465895882\ M\^46 - 
                    3182534760\ M\^48 + 420913324\ M\^50 + 
                    5941203351\ M\^52 - 2732129827\ M\^54 - 
                    9167628197\ M\^56 + 6489091123\ M\^58 + 
                    12086434165\ M\^60 - 10731061071\ M\^62 - 
                    14069816045\ M\^64 + 13759051161\ M\^66 + 
                    14997315133\ M\^68 - 14137565679\ M\^70 - 
                    15141963016\ M\^72 + 11745897456\ M\^74 + 
                    14753720747\ M\^76 - 7966414072\ M\^78 - 
                    13821231486\ M\^80 + 4716063285\ M\^82 + 
                    12245804822\ M\^84 - 3020612491\ M\^86 - 
                    9963353012\ M\^88 + 2628512590\ M\^90 + 
                    7163421750\ M\^92 - 2647995291\ M\^94 - 
                    4346148254\ M\^96 + 2393586806\ M\^98 + 
                    2083446873\ M\^100 - 1759249887\ M\^102 - 
                    678826302\ M\^104 + 1016457711\ M\^106 + 
                    53221251\ M\^108 - 446142654\ M\^110 + 
                    100470484\ M\^112 + 135620650\ M\^114 - 
                    73310117\ M\^116 - 20509740\ M\^118 + 27183878\ M\^120 - 
                    3528017\ M\^122 - 4636417\ M\^124 + 1889853\ M\^126 + 
                    189383\ M\^128 - 196979\ M\^130 + 13514\ M\^132 - 
                    114233\ M\^134 + 200167\ M\^136 - 164443\ M\^138 + 
                    87673\ M\^140 - 33699\ M\^142 + 9611\ M\^144 - 
                    1999\ M\^146 + 286\ M\^148 - 25\ M\^150 + M\^152)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 14]\), \(L\^28\ M\^12 + 
              M\^240 + 
              L\^27\ \((\(-3\)\ M\^8 + 12\ M\^10 - 23\ M\^12 + 26\ M\^14 - 
                    6\ M\^16 - 43\ M\^18 + 85\ M\^20 + M\^22 - 45\ M\^24 + 
                    32\ M\^26 - 8\ M\^28)\) + 
              L\^26\ \((3\ M\^4 - 22\ M\^6 + 69\ M\^8 - 120\ M\^10 + 
                    127\ M\^12 - 94\ M\^14 + 148\ M\^16 - 348\ M\^18 - 
                    107\ M\^20 + 1830\ M\^22 - 2279\ M\^24 - 1234\ M\^26 + 
                    4642\ M\^28 - 1658\ M\^30 - 2535\ M\^32 + 2798\ M\^34 - 
                    426\ M\^36 - 932\ M\^38 + 708\ M\^40 - 220\ M\^42 + 
                    28\ M\^44)\) + 
              L\^25\ \((\(-1\) + 10\ M\^2 - 44\ M\^4 + 120\ M\^6 - 
                    254\ M\^8 + 516\ M\^10 - 1113\ M\^12 + 1956\ M\^14 - 
                    1252\ M\^16 - 3169\ M\^18 + 6782\ M\^20 + 207\ M\^22 - 
                    10726\ M\^24 - 1665\ M\^26 + 26245\ M\^28 - 2974\ M\^30 - 
                    58633\ M\^32 + 44494\ M\^34 + 60012\ M\^36 - 
                    92044\ M\^38 + 795\ M\^40 + 82458\ M\^42 - 51790\ M\^44 - 
                    21840\ M\^46 + 40079\ M\^48 - 13391\ M\^50 - 
                    6834\ M\^52 + 7950\ M\^54 - 3210\ M\^56 + 648\ M\^58 - 
                    56\ M\^60)\) + 
              L\^24\ \((1 - 14\ M\^2 + 89\ M\^4 - 359\ M\^6 + 1049\ M\^8 - 
                    2232\ M\^10 + 3218\ M\^12 - 3060\ M\^14 + 3561\ M\^16 - 
                    7972\ M\^18 + 5473\ M\^20 + 32779\ M\^22 - 86077\ M\^24 + 
                    21854\ M\^26 + 203315\ M\^28 - 245149\ M\^30 - 
                    208534\ M\^32 + 559253\ M\^34 + 42469\ M\^36 - 
                    842849\ M\^38 + 271605\ M\^40 + 993116\ M\^42 - 
                    721454\ M\^44 - 800536\ M\^46 + 1146731\ M\^48 + 
                    215402\ M\^50 - 1096613\ M\^52 + 402427\ M\^54 + 
                    537565\ M\^56 - 538294\ M\^58 + 5631\ M\^60 + 
                    249969\ M\^62 - 140038\ M\^64 - 8229\ M\^66 + 
                    45400\ M\^68 - 25046\ M\^70 + 7014\ M\^72 - 1060\ M\^74 + 
                    70\ M\^76)\) + 
              L\^23\ \((\(-8\)\ M\^4 + 131\ M\^6 - 969\ M\^8 + 4266\ M\^10 - 
                    12126\ M\^12 + 21317\ M\^14 - 14784\ M\^16 - 
                    27190\ M\^18 + 70801\ M\^20 + 858\ M\^22 - 
                    204881\ M\^24 + 172591\ M\^26 + 481418\ M\^28 - 
                    965866\ M\^30 - 421577\ M\^32 + 2805608\ M\^34 - 
                    1463618\ M\^36 - 4453442\ M\^38 + 5782222\ M\^40 + 
                    3392864\ M\^42 - 10225229\ M\^44 + 932043\ M\^46 + 
                    11427751\ M\^48 - 5775845\ M\^50 - 8926375\ M\^52 + 
                    8458948\ M\^54 + 5047980\ M\^56 - 8573239\ M\^58 - 
                    1226913\ M\^60 + 7055136\ M\^62 - 1818724\ M\^64 - 
                    4228120\ M\^66 + 3102909\ M\^68 + 1111355\ M\^70 - 
                    2285096\ M\^72 + 623712\ M\^74 + 694772\ M\^76 - 
                    626004\ M\^78 + 106410\ M\^80 + 128647\ M\^82 - 
                    105363\ M\^84 + 39470\ M\^86 - 8544\ M\^88 + 
                    1040\ M\^90 - 56\ M\^92)\) + 
              L\^22\ \((28\ M\^8 - 518\ M\^10 + 4357\ M\^12 - 21635\ M\^14 + 
                    67843\ M\^16 - 126434\ M\^18 + 74824\ M\^20 + 
                    276172\ M\^22 - 784397\ M\^24 + 487185\ M\^26 + 
                    1608279\ M\^28 - 3691975\ M\^30 + 452813\ M\^32 + 
                    8699805\ M\^34 - 10158974\ M\^36 - 9310124\ M\^38 + 
                    28910464\ M\^40 - 5905334\ M\^42 - 46245412\ M\^44 + 
                    41711473\ M\^46 + 42063035\ M\^48 - 81974562\ M\^50 - 
                    7048735\ M\^52 + 96374390\ M\^54 - 39231266\ M\^56 - 
                    71705527\ M\^58 + 63403406\ M\^60 + 29077815\ M\^62 - 
                    53255719\ M\^64 + 446227\ M\^66 + 29087005\ M\^68 - 
                    6918284\ M\^70 - 13285530\ M\^72 + 4098248\ M\^74 + 
                    8498387\ M\^76 - 4085654\ M\^78 - 5589866\ M\^80 + 
                    5466708\ M\^82 + 1136717\ M\^84 - 4019278\ M\^86 + 
                    1651655\ M\^88 + 1005114\ M\^90 - 1302925\ M\^92 + 
                    396032\ M\^94 + 196510\ M\^96 - 245476\ M\^98 + 
                    118237\ M\^100 - 33756\ M\^102 + 5976\ M\^104 - 
                    612\ M\^106 + 28\ M\^108)\) + 
              L\^21\ \((\(-56\)\ M\^12 + 1141\ M\^14 - 10608\ M\^16 + 
                    58432\ M\^18 - 204997\ M\^20 + 440864\ M\^22 - 
                    392199\ M\^24 - 777787\ M\^26 + 3186538\ M\^28 - 
                    3688712\ M\^30 - 3718870\ M\^32 + 17632914\ M\^34 - 
                    16145632\ M\^36 - 24836467\ M\^38 + 70734485\ M\^40 - 
                    23005909\ M\^42 - 126509104\ M\^44 + 166395260\ M\^46 + 
                    85055959\ M\^48 - 355172705\ M\^50 + 135684953\ M\^52 + 
                    432633483\ M\^54 - 474172459\ M\^56 - 262491734\ M\^58 + 
                    714734098\ M\^60 - 104330741\ M\^62 - 666043843\ M\^64 + 
                    432632977\ M\^66 + 354350302\ M\^68 - 509510457\ M\^70 - 
                    5419191\ M\^72 + 339275327\ M\^74 - 166460229\ M\^76 - 
                    102179627\ M\^78 + 139772139\ M\^80 - 33956332\ M\^82 - 
                    40713545\ M\^84 + 45391535\ M\^86 - 17499205\ M\^88 - 
                    9036695\ M\^90 + 18181663\ M\^92 - 10527257\ M\^94 - 
                    1557542\ M\^96 + 7050285\ M\^98 - 4797123\ M\^100 + 
                    241348\ M\^102 + 2035129\ M\^104 - 1601790\ M\^106 + 
                    353166\ M\^108 + 333474\ M\^110 - 374649\ M\^112 + 
                    197807\ M\^114 - 66910\ M\^116 + 15206\ M\^118 - 
                    2262\ M\^120 + 200\ M\^122 - 8\ M\^124)\) + 
              L\^20\ \((70\ M\^16 - 1540\ M\^18 + 15506\ M\^20 - 
                    93241\ M\^22 + 363752\ M\^24 - 913201\ M\^26 + 
                    1212766\ M\^28 + 566813\ M\^30 - 6208140\ M\^32 + 
                    11610373\ M\^34 - 1477944\ M\^36 - 35991728\ M\^38 + 
                    64879112\ M\^40 + 6014380\ M\^42 - 181569651\ M\^44 + 
                    208545016\ M\^46 + 181353207\ M\^48 - 630218641\ M\^50 + 
                    246570790\ M\^52 + 941892279\ M\^54 - 1217432842\ M\^56 - 
                    572598261\ M\^58 + 2289472774\ M\^60 - 768595240\ M\^62 - 
                    2604251436\ M\^64 + 2589316971\ M\^66 + 
                    1610721290\ M\^68 - 3810755433\ M\^70 + 
                    315002475\ M\^72 + 3624138266\ M\^74 - 
                    2071150598\ M\^76 - 2186951532\ M\^78 + 
                    2718417348\ M\^80 + 458093100\ M\^82 - 
                    2162454707\ M\^84 + 643202198\ M\^86 + 
                    1072671957\ M\^88 - 853465140\ M\^90 - 198014057\ M\^92 + 
                    526199150\ M\^94 - 160672029\ M\^96 - 152245541\ M\^98 + 
                    150638007\ M\^100 - 27079071\ M\^102 - 43640373\ M\^104 + 
                    42163164\ M\^106 - 12113148\ M\^108 - 10488568\ M\^110 + 
                    13701129\ M\^112 - 5260455\ M\^114 - 2179313\ M\^116 + 
                    3610510\ M\^118 - 1645293\ M\^120 - 138148\ M\^122 + 
                    662240\ M\^124 - 483023\ M\^126 + 217943\ M\^128 - 
                    70451\ M\^130 + 16903\ M\^132 - 2978\ M\^134 + 
                    366\ M\^136 - 28\ M\^138 + M\^140)\) + 
              L\^19\ \((\(-56\)\ M\^20 + 1309\ M\^22 - 14072\ M\^24 + 
                    91393\ M\^26 - 394199\ M\^28 + 1151340\ M\^30 - 
                    2102345\ M\^32 + 1226646\ M\^34 + 5466251\ M\^36 - 
                    18047708\ M\^38 + 18863364\ M\^40 + 27311632\ M\^42 - 
                    112597233\ M\^44 + 95781935\ M\^46 + 187800292\ M\^48 - 
                    497638504\ M\^50 + 113626583\ M\^52 + 1019478029\ M\^54 - 
                    1230195848\ M\^56 - 912618349\ M\^58 + 
                    3038772175\ M\^60 - 807975734\ M\^62 - 
                    4369795872\ M\^64 + 4340797328\ M\^66 + 
                    3492939731\ M\^68 - 8305413763\ M\^70 + 
                    441375271\ M\^72 + 10270312820\ M\^74 - 
                    6249990131\ M\^76 - 8489509235\ M\^78 + 
                    11129604405\ M\^80 + 3412003416\ M\^82 - 
                    12541630183\ M\^84 + 2460146387\ M\^86 + 
                    10001085338\ M\^88 - 6300987117\ M\^90 - 
                    5244403245\ M\^92 + 6870543735\ M\^94 + 
                    853438852\ M\^96 - 4950398880\ M\^98 + 
                    1522718766\ M\^100 + 2342580627\ M\^102 - 
                    1863136623\ M\^104 - 471290901\ M\^106 + 
                    1163643387\ M\^108 - 284458332\ M\^110 - 
                    411139295\ M\^112 + 314290428\ M\^114 + 
                    23192731\ M\^116 - 135033474\ M\^118 + 58482111\ M\^120 + 
                    17676465\ M\^122 - 29448663\ M\^124 + 11320653\ M\^126 + 
                    2302510\ M\^128 - 5068369\ M\^130 + 2777434\ M\^132 - 
                    307853\ M\^134 - 768310\ M\^136 + 769401\ M\^138 - 
                    427740\ M\^140 + 163000\ M\^142 - 44289\ M\^144 + 
                    8490\ M\^146 - 1094\ M\^148 + 85\ M\^150 - 3\ M\^152)\) + 
              L\^18\ \((28\ M\^24 - 686\ M\^26 + 7801\ M\^28 - 54507\ M\^30 + 
                    260351\ M\^32 - 886282\ M\^34 + 2114010\ M\^36 - 
                    2993246\ M\^38 - 335058\ M\^40 + 13289249\ M\^42 - 
                    29708849\ M\^44 + 11492879\ M\^46 + 84924298\ M\^48 - 
                    179206153\ M\^50 + 2931513\ M\^52 + 503520905\ M\^54 - 
                    589463740\ M\^56 - 615567590\ M\^58 + 1909425668\ M\^60 - 
                    327743902\ M\^62 - 3502289451\ M\^64 + 
                    3141563002\ M\^66 + 3960373765\ M\^68 - 
                    7712925416\ M\^70 - 1491451001\ M\^72 + 
                    12404941881\ M\^74 - 4781190370\ M\^76 - 
                    14526492767\ M\^78 + 13660841019\ M\^80 + 
                    11854862212\ M\^82 - 21877265499\ M\^84 - 
                    4333187774\ M\^86 + 25669675022\ M\^88 - 
                    5391024253\ M\^90 - 23107287479\ M\^92 + 
                    13252142460\ M\^94 + 15440541374\ M\^96 - 
                    16181095006\ M\^98 - 6261734036\ M\^100 + 
                    13959071240\ M\^102 - 713910391\ M\^104 - 
                    8850695341\ M\^106 + 3789554294\ M\^108 + 
                    3774989767\ M\^110 - 3648878673\ M\^112 - 
                    547984737\ M\^114 + 2122863106\ M\^116 - 
                    654292261\ M\^118 - 720775919\ M\^120 + 
                    642264010\ M\^122 + 11003620\ M\^124 - 
                    290190690\ M\^126 + 148694430\ M\^128 + 
                    44569892\ M\^130 - 88024795\ M\^132 + 29956264\ M\^134 + 
                    19801345\ M\^136 - 23231965\ M\^138 + 5620169\ M\^140 + 
                    5180359\ M\^142 - 4840271\ M\^144 + 1051710\ M\^146 + 
                    1002915\ M\^148 - 1103100\ M\^150 + 589244\ M\^152 - 
                    210411\ M\^154 + 53485\ M\^156 - 9655\ M\^158 + 
                    1182\ M\^160 - 88\ M\^162 + 3\ M\^164)\) + 
              L\^17\ \((\(-8\)\ M\^28 + 203\ M\^30 - 2435\ M\^32 + 
                    18455\ M\^34 - 99449\ M\^36 + 402499\ M\^38 - 
                    1232661\ M\^40 + 2694337\ M\^42 - 3227938\ M\^44 - 
                    2266630\ M\^46 + 18225687\ M\^48 - 29686054\ M\^50 - 
                    10028255\ M\^52 + 115221013\ M\^54 - 135853277\ M\^56 - 
                    154024910\ M\^58 + 525627928\ M\^60 - 128331941\ M\^62 - 
                    1099631803\ M\^64 + 1098261504\ M\^66 + 
                    1517039528\ M\^68 - 3011902714\ M\^70 - 
                    1213244727\ M\^72 + 5893757511\ M\^74 - 
                    529747045\ M\^76 - 9412405632\ M\^78 + 
                    4453737870\ M\^80 + 12706498051\ M\^82 - 
                    10946573476\ M\^84 - 14307435211\ M\^86 + 
                    19298506259\ M\^88 + 12651065119\ M\^90 - 
                    27248235260\ M\^92 - 7156009251\ M\^94 + 
                    31634342115\ M\^96 - 865501417\ M\^98 - 
                    30164442946\ M\^100 + 8451428909\ M\^102 + 
                    23014688719\ M\^104 - 12546669470\ M\^106 - 
                    12893985063\ M\^108 + 12003276434\ M\^110 + 
                    3695771003\ M\^112 - 8002872162\ M\^114 + 
                    1815305298\ M\^116 + 3093751670\ M\^118 - 
                    3288687524\ M\^120 + 471325710\ M\^122 + 
                    2269324563\ M\^124 - 1870319936\ M\^126 - 
                    702031278\ M\^128 + 1661747282\ M\^130 - 
                    269800760\ M\^132 - 885081326\ M\^134 + 
                    495961067\ M\^136 + 266199772\ M\^138 - 
                    339420758\ M\^140 + 10012173\ M\^142 + 
                    138425006\ M\^144 - 57518071\ M\^146 - 29053426\ M\^148 + 
                    32687620\ M\^150 - 4656508\ M\^152 - 8082723\ M\^154 + 
                    5114807\ M\^156 - 87007\ M\^158 - 1585109\ M\^160 + 
                    1168296\ M\^162 - 498186\ M\^164 + 147219\ M\^166 - 
                    31449\ M\^168 + 4825\ M\^170 - 508\ M\^172 + 33\ M\^174 - 
                    M\^176)\) + 
              L\^16\ \((M\^32 - 26\ M\^34 + 335\ M\^36 - 2900\ M\^38 + 
                    18970\ M\^40 - 97495\ M\^42 + 391378\ M\^44 - 
                    1184225\ M\^46 + 2489079\ M\^48 - 2747674\ M\^50 - 
                    1988562\ M\^52 + 13205277\ M\^54 - 17831436\ M\^56 - 
                    10291768\ M\^58 + 64103237\ M\^60 - 56371586\ M\^62 - 
                    82104627\ M\^64 + 187936447\ M\^66 + 16100963\ M\^68 - 
                    318952046\ M\^70 + 69709827\ M\^72 + 436705778\ M\^74 + 
                    86241603\ M\^76 - 862996825\ M\^78 - 726137432\ M\^80 + 
                    2324461302\ M\^82 + 1745076069\ M\^84 - 
                    5632895475\ M\^86 - 2524230423\ M\^88 + 
                    10983638995\ M\^90 + 2249469707\ M\^92 - 
                    17360511838\ M\^94 - 537079223\ M\^96 + 
                    22496255543\ M\^98 - 1987813885\ M\^100 - 
                    23818017981\ M\^102 + 3738975124\ M\^104 + 
                    20091737080\ M\^106 - 3014648106\ M\^108 - 
                    12657823768\ M\^110 - 802741402\ M\^112 + 
                    4877039354\ M\^114 + 6634874413\ M\^116 + 
                    218983284\ M\^118 - 11877195051\ M\^120 - 
                    1521198326\ M\^122 + 14187574051\ M\^124 + 
                    99042514\ M\^126 - 12983986827\ M\^128 + 
                    1971566855\ M\^130 + 9475771105\ M\^132 - 
                    3138311374\ M\^134 - 5517418713\ M\^136 + 
                    3036128488\ M\^138 + 2463397181\ M\^140 - 
                    2170786715\ M\^142 - 717632595\ M\^144 + 
                    1200355390\ M\^146 + 10405971\ M\^148 - 
                    510286861\ M\^150 + 142067415\ M\^152 + 
                    150832429\ M\^154 - 99001559\ M\^156 - 19305530\ M\^158 + 
                    40496788\ M\^160 - 10105807\ M\^162 - 7926398\ M\^164 + 
                    6432161\ M\^166 - 681112\ M\^168 - 1705211\ M\^170 + 
                    1451284\ M\^172 - 676795\ M\^174 + 216365\ M\^176 - 
                    49752\ M\^178 + 8177\ M\^180 - 917\ M\^182 + 63\ M\^184 - 
                    2\ M\^186)\) + 
              L\^15\ \((\(-3\)\ M\^40 + 86\ M\^42 - 1153\ M\^44 + 
                    9595\ M\^46 - 55340\ M\^48 + 232899\ M\^50 - 
                    726967\ M\^52 + 1644080\ M\^54 - 2394241\ M\^56 + 
                    957575\ M\^58 + 4925247\ M\^60 - 12141895\ M\^62 + 
                    6855795\ M\^64 + 22896477\ M\^66 - 51219774\ M\^68 + 
                    12076649\ M\^70 + 96675159\ M\^72 - 116183157\ M\^74 - 
                    86661997\ M\^76 + 274109724\ M\^78 - 28047067\ M\^80 - 
                    467255241\ M\^82 + 307496276\ M\^84 + 792781676\ M\^86 - 
                    1119370212\ M\^88 - 1305810038\ M\^90 + 
                    3395264206\ M\^92 + 1592225335\ M\^94 - 
                    8388438149\ M\^96 - 474266583\ M\^98 + 
                    16612946710\ M\^100 - 3343384457\ M\^102 - 
                    26730732444\ M\^104 + 9871144263\ M\^106 + 
                    35480973062\ M\^108 - 16887423281\ M\^110 - 
                    39266196462\ M\^112 + 20845230001\ M\^114 + 
                    36491110152\ M\^116 - 19277977634\ M\^118 - 
                    28890654996\ M\^120 + 12673052172\ M\^122 + 
                    20464522377\ M\^124 - 3956702125\ M\^126 - 
                    14291643240\ M\^128 - 3100972683\ M\^130 + 
                    10939266756\ M\^132 + 6208636668\ M\^134 - 
                    9092456238\ M\^136 - 5641404105\ M\^138 + 
                    7281828934\ M\^140 + 3372804127\ M\^142 - 
                    5070397968\ M\^144 - 1297494711\ M\^146 + 
                    2908979447\ M\^148 + 199728665\ M\^150 - 
                    1351969723\ M\^152 + 116664434\ M\^154 + 
                    516220169\ M\^156 - 108870512\ M\^158 - 
                    178212190\ M\^160 + 62332169\ M\^162 + 59162903\ M\^164 - 
                    35076816\ M\^166 - 16301776\ M\^168 + 20570137\ M\^170 - 
                    2264688\ M\^172 - 6540197\ M\^174 + 3930255\ M\^176 + 
                    352220\ M\^178 - 1797890\ M\^180 + 1316289\ M\^182 - 
                    576646\ M\^184 + 174581\ M\^186 - 37697\ M\^188 + 
                    5737\ M\^190 - 586\ M\^192 + 36\ M\^194 - M\^196)\) + 
              L\^14\ \((3\ M\^48 - 94\ M\^50 + 1351\ M\^52 - 11767\ M\^54 + 
                    68890\ M\^56 - 282791\ M\^58 + 816396\ M\^60 - 
                    1581180\ M\^62 + 1688854\ M\^64 + 216535\ M\^66 - 
                    3588190\ M\^68 + 4650549\ M\^70 - 1614371\ M\^72 - 
                    1852916\ M\^74 + 12404584\ M\^76 - 48377618\ M\^78 + 
                    57179603\ M\^80 + 117939206\ M\^82 - 387297125\ M\^84 + 
                    51050426\ M\^86 + 1151841116\ M\^88 - 1161767540\ M\^90 - 
                    2142376526\ M\^92 + 4287822345\ M\^94 + 
                    2384945495\ M\^96 - 10386881720\ M\^98 - 
                    106591045\ M\^100 + 19386747092\ M\^102 - 
                    6349179688\ M\^104 - 29537895816\ M\^106 + 
                    16813375372\ M\^108 + 37789099026\ M\^110 - 
                    27819698758\ M\^112 - 41553888140\ M\^114 + 
                    33239708901\ M\^116 + 40686300692\ M\^118 - 
                    27680786515\ M\^120 - 37838049084\ M\^122 + 
                    10849159953\ M\^124 + 36413525590\ M\^126 + 
                    10849159953\ M\^128 - 37838049084\ M\^130 - 
                    27680786515\ M\^132 + 40686300692\ M\^134 + 
                    33239708901\ M\^136 - 41553888140\ M\^138 - 
                    27819698758\ M\^140 + 37789099026\ M\^142 + 
                    16813375372\ M\^144 - 29537895816\ M\^146 - 
                    6349179688\ M\^148 + 19386747092\ M\^150 - 
                    106591045\ M\^152 - 10386881720\ M\^154 + 
                    2384945495\ M\^156 + 4287822345\ M\^158 - 
                    2142376526\ M\^160 - 1161767540\ M\^162 + 
                    1151841116\ M\^164 + 51050426\ M\^166 - 
                    387297125\ M\^168 + 117939206\ M\^170 + 
                    57179603\ M\^172 - 48377618\ M\^174 + 12404584\ M\^176 - 
                    1852916\ M\^178 - 1614371\ M\^180 + 4650549\ M\^182 - 
                    3588190\ M\^184 + 216535\ M\^186 + 1688854\ M\^188 - 
                    1581180\ M\^190 + 816396\ M\^192 - 282791\ M\^194 + 
                    68890\ M\^196 - 11767\ M\^198 + 1351\ M\^200 - 
                    94\ M\^202 + 3\ M\^204)\) + 
              L\^13\ \((\(-M\^56\) + 36\ M\^58 - 586\ M\^60 + 5737\ M\^62 - 
                    37697\ M\^64 + 174581\ M\^66 - 576646\ M\^68 + 
                    1316289\ M\^70 - 1797890\ M\^72 + 352220\ M\^74 + 
                    3930255\ M\^76 - 6540197\ M\^78 - 2264688\ M\^80 + 
                    20570137\ M\^82 - 16301776\ M\^84 - 35076816\ M\^86 + 
                    59162903\ M\^88 + 62332169\ M\^90 - 178212190\ M\^92 - 
                    108870512\ M\^94 + 516220169\ M\^96 + 116664434\ M\^98 - 
                    1351969723\ M\^100 + 199728665\ M\^102 + 
                    2908979447\ M\^104 - 1297494711\ M\^106 - 
                    5070397968\ M\^108 + 3372804127\ M\^110 + 
                    7281828934\ M\^112 - 5641404105\ M\^114 - 
                    9092456238\ M\^116 + 6208636668\ M\^118 + 
                    10939266756\ M\^120 - 3100972683\ M\^122 - 
                    14291643240\ M\^124 - 3956702125\ M\^126 + 
                    20464522377\ M\^128 + 12673052172\ M\^130 - 
                    28890654996\ M\^132 - 19277977634\ M\^134 + 
                    36491110152\ M\^136 + 20845230001\ M\^138 - 
                    39266196462\ M\^140 - 16887423281\ M\^142 + 
                    35480973062\ M\^144 + 9871144263\ M\^146 - 
                    26730732444\ M\^148 - 3343384457\ M\^150 + 
                    16612946710\ M\^152 - 474266583\ M\^154 - 
                    8388438149\ M\^156 + 1592225335\ M\^158 + 
                    3395264206\ M\^160 - 1305810038\ M\^162 - 
                    1119370212\ M\^164 + 792781676\ M\^166 + 
                    307496276\ M\^168 - 467255241\ M\^170 - 
                    28047067\ M\^172 + 274109724\ M\^174 - 86661997\ M\^176 - 
                    116183157\ M\^178 + 96675159\ M\^180 + 12076649\ M\^182 - 
                    51219774\ M\^184 + 22896477\ M\^186 + 6855795\ M\^188 - 
                    12141895\ M\^190 + 4925247\ M\^192 + 957575\ M\^194 - 
                    2394241\ M\^196 + 1644080\ M\^198 - 726967\ M\^200 + 
                    232899\ M\^202 - 55340\ M\^204 + 9595\ M\^206 - 
                    1153\ M\^208 + 86\ M\^210 - 3\ M\^212)\) + 
              L\^12\ \((\(-2\)\ M\^66 + 63\ M\^68 - 917\ M\^70 + 
                    8177\ M\^72 - 49752\ M\^74 + 216365\ M\^76 - 
                    676795\ M\^78 + 1451284\ M\^80 - 1705211\ M\^82 - 
                    681112\ M\^84 + 6432161\ M\^86 - 7926398\ M\^88 - 
                    10105807\ M\^90 + 40496788\ M\^92 - 19305530\ M\^94 - 
                    99001559\ M\^96 + 150832429\ M\^98 + 142067415\ M\^100 - 
                    510286861\ M\^102 + 10405971\ M\^104 + 
                    1200355390\ M\^106 - 717632595\ M\^108 - 
                    2170786715\ M\^110 + 2463397181\ M\^112 + 
                    3036128488\ M\^114 - 5517418713\ M\^116 - 
                    3138311374\ M\^118 + 9475771105\ M\^120 + 
                    1971566855\ M\^122 - 12983986827\ M\^124 + 
                    99042514\ M\^126 + 14187574051\ M\^128 - 
                    1521198326\ M\^130 - 11877195051\ M\^132 + 
                    218983284\ M\^134 + 6634874413\ M\^136 + 
                    4877039354\ M\^138 - 802741402\ M\^140 - 
                    12657823768\ M\^142 - 3014648106\ M\^144 + 
                    20091737080\ M\^146 + 3738975124\ M\^148 - 
                    23818017981\ M\^150 - 1987813885\ M\^152 + 
                    22496255543\ M\^154 - 537079223\ M\^156 - 
                    17360511838\ M\^158 + 2249469707\ M\^160 + 
                    10983638995\ M\^162 - 2524230423\ M\^164 - 
                    5632895475\ M\^166 + 1745076069\ M\^168 + 
                    2324461302\ M\^170 - 726137432\ M\^172 - 
                    862996825\ M\^174 + 86241603\ M\^176 + 
                    436705778\ M\^178 + 69709827\ M\^180 - 
                    318952046\ M\^182 + 16100963\ M\^184 + 
                    187936447\ M\^186 - 82104627\ M\^188 - 56371586\ M\^190 + 
                    64103237\ M\^192 - 10291768\ M\^194 - 17831436\ M\^196 + 
                    13205277\ M\^198 - 1988562\ M\^200 - 2747674\ M\^202 + 
                    2489079\ M\^204 - 1184225\ M\^206 + 391378\ M\^208 - 
                    97495\ M\^210 + 18970\ M\^212 - 2900\ M\^214 + 
                    335\ M\^216 - 26\ M\^218 + M\^220)\) + 
              L\^11\ \((\(-M\^76\) + 33\ M\^78 - 508\ M\^80 + 4825\ M\^82 - 
                    31449\ M\^84 + 147219\ M\^86 - 498186\ M\^88 + 
                    1168296\ M\^90 - 1585109\ M\^92 - 87007\ M\^94 + 
                    5114807\ M\^96 - 8082723\ M\^98 - 4656508\ M\^100 + 
                    32687620\ M\^102 - 29053426\ M\^104 - 57518071\ M\^106 + 
                    138425006\ M\^108 + 10012173\ M\^110 - 
                    339420758\ M\^112 + 266199772\ M\^114 + 
                    495961067\ M\^116 - 885081326\ M\^118 - 
                    269800760\ M\^120 + 1661747282\ M\^122 - 
                    702031278\ M\^124 - 1870319936\ M\^126 + 
                    2269324563\ M\^128 + 471325710\ M\^130 - 
                    3288687524\ M\^132 + 3093751670\ M\^134 + 
                    1815305298\ M\^136 - 8002872162\ M\^138 + 
                    3695771003\ M\^140 + 12003276434\ M\^142 - 
                    12893985063\ M\^144 - 12546669470\ M\^146 + 
                    23014688719\ M\^148 + 8451428909\ M\^150 - 
                    30164442946\ M\^152 - 865501417\ M\^154 + 
                    31634342115\ M\^156 - 7156009251\ M\^158 - 
                    27248235260\ M\^160 + 12651065119\ M\^162 + 
                    19298506259\ M\^164 - 14307435211\ M\^166 - 
                    10946573476\ M\^168 + 12706498051\ M\^170 + 
                    4453737870\ M\^172 - 9412405632\ M\^174 - 
                    529747045\ M\^176 + 5893757511\ M\^178 - 
                    1213244727\ M\^180 - 3011902714\ M\^182 + 
                    1517039528\ M\^184 + 1098261504\ M\^186 - 
                    1099631803\ M\^188 - 128331941\ M\^190 + 
                    525627928\ M\^192 - 154024910\ M\^194 - 
                    135853277\ M\^196 + 115221013\ M\^198 - 
                    10028255\ M\^200 - 29686054\ M\^202 + 18225687\ M\^204 - 
                    2266630\ M\^206 - 3227938\ M\^208 + 2694337\ M\^210 - 
                    1232661\ M\^212 + 402499\ M\^214 - 99449\ M\^216 + 
                    18455\ M\^218 - 2435\ M\^220 + 203\ M\^222 - 
                    8\ M\^224)\) + 
              L\^10\ \((3\ M\^88 - 88\ M\^90 + 1182\ M\^92 - 9655\ M\^94 + 
                    53485\ M\^96 - 210411\ M\^98 + 589244\ M\^100 - 
                    1103100\ M\^102 + 1002915\ M\^104 + 1051710\ M\^106 - 
                    4840271\ M\^108 + 5180359\ M\^110 + 5620169\ M\^112 - 
                    23231965\ M\^114 + 19801345\ M\^116 + 29956264\ M\^118 - 
                    88024795\ M\^120 + 44569892\ M\^122 + 148694430\ M\^124 - 
                    290190690\ M\^126 + 11003620\ M\^128 + 
                    642264010\ M\^130 - 720775919\ M\^132 - 
                    654292261\ M\^134 + 2122863106\ M\^136 - 
                    547984737\ M\^138 - 3648878673\ M\^140 + 
                    3774989767\ M\^142 + 3789554294\ M\^144 - 
                    8850695341\ M\^146 - 713910391\ M\^148 + 
                    13959071240\ M\^150 - 6261734036\ M\^152 - 
                    16181095006\ M\^154 + 15440541374\ M\^156 + 
                    13252142460\ M\^158 - 23107287479\ M\^160 - 
                    5391024253\ M\^162 + 25669675022\ M\^164 - 
                    4333187774\ M\^166 - 21877265499\ M\^168 + 
                    11854862212\ M\^170 + 13660841019\ M\^172 - 
                    14526492767\ M\^174 - 4781190370\ M\^176 + 
                    12404941881\ M\^178 - 1491451001\ M\^180 - 
                    7712925416\ M\^182 + 3960373765\ M\^184 + 
                    3141563002\ M\^186 - 3502289451\ M\^188 - 
                    327743902\ M\^190 + 1909425668\ M\^192 - 
                    615567590\ M\^194 - 589463740\ M\^196 + 
                    503520905\ M\^198 + 2931513\ M\^200 - 179206153\ M\^202 + 
                    84924298\ M\^204 + 11492879\ M\^206 - 29708849\ M\^208 + 
                    13289249\ M\^210 - 335058\ M\^212 - 2993246\ M\^214 + 
                    2114010\ M\^216 - 886282\ M\^218 + 260351\ M\^220 - 
                    54507\ M\^222 + 7801\ M\^224 - 686\ M\^226 + 
                    28\ M\^228)\) + 
              L\^9\ \((\(-3\)\ M\^100 + 85\ M\^102 - 1094\ M\^104 + 
                    8490\ M\^106 - 44289\ M\^108 + 163000\ M\^110 - 
                    427740\ M\^112 + 769401\ M\^114 - 768310\ M\^116 - 
                    307853\ M\^118 + 2777434\ M\^120 - 5068369\ M\^122 + 
                    2302510\ M\^124 + 11320653\ M\^126 - 29448663\ M\^128 + 
                    17676465\ M\^130 + 58482111\ M\^132 - 135033474\ M\^134 + 
                    23192731\ M\^136 + 314290428\ M\^138 - 
                    411139295\ M\^140 - 284458332\ M\^142 + 
                    1163643387\ M\^144 - 471290901\ M\^146 - 
                    1863136623\ M\^148 + 2342580627\ M\^150 + 
                    1522718766\ M\^152 - 4950398880\ M\^154 + 
                    853438852\ M\^156 + 6870543735\ M\^158 - 
                    5244403245\ M\^160 - 6300987117\ M\^162 + 
                    10001085338\ M\^164 + 2460146387\ M\^166 - 
                    12541630183\ M\^168 + 3412003416\ M\^170 + 
                    11129604405\ M\^172 - 8489509235\ M\^174 - 
                    6249990131\ M\^176 + 10270312820\ M\^178 + 
                    441375271\ M\^180 - 8305413763\ M\^182 + 
                    3492939731\ M\^184 + 4340797328\ M\^186 - 
                    4369795872\ M\^188 - 807975734\ M\^190 + 
                    3038772175\ M\^192 - 912618349\ M\^194 - 
                    1230195848\ M\^196 + 1019478029\ M\^198 + 
                    113626583\ M\^200 - 497638504\ M\^202 + 
                    187800292\ M\^204 + 95781935\ M\^206 - 
                    112597233\ M\^208 + 27311632\ M\^210 + 18863364\ M\^212 - 
                    18047708\ M\^214 + 5466251\ M\^216 + 1226646\ M\^218 - 
                    2102345\ M\^220 + 1151340\ M\^222 - 394199\ M\^224 + 
                    91393\ M\^226 - 14072\ M\^228 + 1309\ M\^230 - 
                    56\ M\^232)\) + 
              L\^8\ \((M\^112 - 28\ M\^114 + 366\ M\^116 - 2978\ M\^118 + 
                    16903\ M\^120 - 70451\ M\^122 + 217943\ M\^124 - 
                    483023\ M\^126 + 662240\ M\^128 - 138148\ M\^130 - 
                    1645293\ M\^132 + 3610510\ M\^134 - 2179313\ M\^136 - 
                    5260455\ M\^138 + 13701129\ M\^140 - 10488568\ M\^142 - 
                    12113148\ M\^144 + 42163164\ M\^146 - 43640373\ M\^148 - 
                    27079071\ M\^150 + 150638007\ M\^152 - 
                    152245541\ M\^154 - 160672029\ M\^156 + 
                    526199150\ M\^158 - 198014057\ M\^160 - 
                    853465140\ M\^162 + 1072671957\ M\^164 + 
                    643202198\ M\^166 - 2162454707\ M\^168 + 
                    458093100\ M\^170 + 2718417348\ M\^172 - 
                    2186951532\ M\^174 - 2071150598\ M\^176 + 
                    3624138266\ M\^178 + 315002475\ M\^180 - 
                    3810755433\ M\^182 + 1610721290\ M\^184 + 
                    2589316971\ M\^186 - 2604251436\ M\^188 - 
                    768595240\ M\^190 + 2289472774\ M\^192 - 
                    572598261\ M\^194 - 1217432842\ M\^196 + 
                    941892279\ M\^198 + 246570790\ M\^200 - 
                    630218641\ M\^202 + 181353207\ M\^204 + 
                    208545016\ M\^206 - 181569651\ M\^208 + 6014380\ M\^210 + 
                    64879112\ M\^212 - 35991728\ M\^214 - 1477944\ M\^216 + 
                    11610373\ M\^218 - 6208140\ M\^220 + 566813\ M\^222 + 
                    1212766\ M\^224 - 913201\ M\^226 + 363752\ M\^228 - 
                    93241\ M\^230 + 15506\ M\^232 - 1540\ M\^234 + 
                    70\ M\^236)\) + 
              L\^7\ \((\(-8\)\ M\^128 + 200\ M\^130 - 2262\ M\^132 + 
                    15206\ M\^134 - 66910\ M\^136 + 197807\ M\^138 - 
                    374649\ M\^140 + 333474\ M\^142 + 353166\ M\^144 - 
                    1601790\ M\^146 + 2035129\ M\^148 + 241348\ M\^150 - 
                    4797123\ M\^152 + 7050285\ M\^154 - 1557542\ M\^156 - 
                    10527257\ M\^158 + 18181663\ M\^160 - 9036695\ M\^162 - 
                    17499205\ M\^164 + 45391535\ M\^166 - 40713545\ M\^168 - 
                    33956332\ M\^170 + 139772139\ M\^172 - 
                    102179627\ M\^174 - 166460229\ M\^176 + 
                    339275327\ M\^178 - 5419191\ M\^180 - 509510457\ M\^182 + 
                    354350302\ M\^184 + 432632977\ M\^186 - 
                    666043843\ M\^188 - 104330741\ M\^190 + 
                    714734098\ M\^192 - 262491734\ M\^194 - 
                    474172459\ M\^196 + 432633483\ M\^198 + 
                    135684953\ M\^200 - 355172705\ M\^202 + 
                    85055959\ M\^204 + 166395260\ M\^206 - 
                    126509104\ M\^208 - 23005909\ M\^210 + 70734485\ M\^212 - 
                    24836467\ M\^214 - 16145632\ M\^216 + 17632914\ M\^218 - 
                    3718870\ M\^220 - 3688712\ M\^222 + 3186538\ M\^224 - 
                    777787\ M\^226 - 392199\ M\^228 + 440864\ M\^230 - 
                    204997\ M\^232 + 58432\ M\^234 - 10608\ M\^236 + 
                    1141\ M\^238 - 56\ M\^240)\) + 
              L\ \((\(-8\)\ M\^224 + 32\ M\^226 - 45\ M\^228 + M\^230 + 
                    85\ M\^232 - 43\ M\^234 - 6\ M\^236 + 26\ M\^238 - 
                    23\ M\^240 + 12\ M\^242 - 3\ M\^244)\) + 
              L\^6\ \((28\ M\^144 - 612\ M\^146 + 5976\ M\^148 - 
                    33756\ M\^150 + 118237\ M\^152 - 245476\ M\^154 + 
                    196510\ M\^156 + 396032\ M\^158 - 1302925\ M\^160 + 
                    1005114\ M\^162 + 1651655\ M\^164 - 4019278\ M\^166 + 
                    1136717\ M\^168 + 5466708\ M\^170 - 5589866\ M\^172 - 
                    4085654\ M\^174 + 8498387\ M\^176 + 4098248\ M\^178 - 
                    13285530\ M\^180 - 6918284\ M\^182 + 29087005\ M\^184 + 
                    446227\ M\^186 - 53255719\ M\^188 + 29077815\ M\^190 + 
                    63403406\ M\^192 - 71705527\ M\^194 - 39231266\ M\^196 + 
                    96374390\ M\^198 - 7048735\ M\^200 - 81974562\ M\^202 + 
                    42063035\ M\^204 + 41711473\ M\^206 - 46245412\ M\^208 - 
                    5905334\ M\^210 + 28910464\ M\^212 - 9310124\ M\^214 - 
                    10158974\ M\^216 + 8699805\ M\^218 + 452813\ M\^220 - 
                    3691975\ M\^222 + 1608279\ M\^224 + 487185\ M\^226 - 
                    784397\ M\^228 + 276172\ M\^230 + 74824\ M\^232 - 
                    126434\ M\^234 + 67843\ M\^236 - 21635\ M\^238 + 
                    4357\ M\^240 - 518\ M\^242 + 28\ M\^244)\) + 
              L\^5\ \((\(-56\)\ M\^160 + 1040\ M\^162 - 8544\ M\^164 + 
                    39470\ M\^166 - 105363\ M\^168 + 128647\ M\^170 + 
                    106410\ M\^172 - 626004\ M\^174 + 694772\ M\^176 + 
                    623712\ M\^178 - 2285096\ M\^180 + 1111355\ M\^182 + 
                    3102909\ M\^184 - 4228120\ M\^186 - 1818724\ M\^188 + 
                    7055136\ M\^190 - 1226913\ M\^192 - 8573239\ M\^194 + 
                    5047980\ M\^196 + 8458948\ M\^198 - 8926375\ M\^200 - 
                    5775845\ M\^202 + 11427751\ M\^204 + 932043\ M\^206 - 
                    10225229\ M\^208 + 3392864\ M\^210 + 5782222\ M\^212 - 
                    4453442\ M\^214 - 1463618\ M\^216 + 2805608\ M\^218 - 
                    421577\ M\^220 - 965866\ M\^222 + 481418\ M\^224 + 
                    172591\ M\^226 - 204881\ M\^228 + 858\ M\^230 + 
                    70801\ M\^232 - 27190\ M\^234 - 14784\ M\^236 + 
                    21317\ M\^238 - 12126\ M\^240 + 4266\ M\^242 - 
                    969\ M\^244 + 131\ M\^246 - 8\ M\^248)\) + 
              L\^2\ \((28\ M\^208 - 220\ M\^210 + 708\ M\^212 - 932\ M\^214 - 
                    426\ M\^216 + 2798\ M\^218 - 2535\ M\^220 - 
                    1658\ M\^222 + 4642\ M\^224 - 1234\ M\^226 - 
                    2279\ M\^228 + 1830\ M\^230 - 107\ M\^232 - 348\ M\^234 + 
                    148\ M\^236 - 94\ M\^238 + 127\ M\^240 - 120\ M\^242 + 
                    69\ M\^244 - 22\ M\^246 + 3\ M\^248)\) + 
              L\^3\ \((\(-56\)\ M\^192 + 648\ M\^194 - 3210\ M\^196 + 
                    7950\ M\^198 - 6834\ M\^200 - 13391\ M\^202 + 
                    40079\ M\^204 - 21840\ M\^206 - 51790\ M\^208 + 
                    82458\ M\^210 + 795\ M\^212 - 92044\ M\^214 + 
                    60012\ M\^216 + 44494\ M\^218 - 58633\ M\^220 - 
                    2974\ M\^222 + 26245\ M\^224 - 1665\ M\^226 - 
                    10726\ M\^228 + 207\ M\^230 + 6782\ M\^232 - 
                    3169\ M\^234 - 1252\ M\^236 + 1956\ M\^238 - 
                    1113\ M\^240 + 516\ M\^242 - 254\ M\^244 + 120\ M\^246 - 
                    44\ M\^248 + 10\ M\^250 - M\^252)\) + 
              L\^4\ \((70\ M\^176 - 1060\ M\^178 + 7014\ M\^180 - 
                    25046\ M\^182 + 45400\ M\^184 - 8229\ M\^186 - 
                    140038\ M\^188 + 249969\ M\^190 + 5631\ M\^192 - 
                    538294\ M\^194 + 537565\ M\^196 + 402427\ M\^198 - 
                    1096613\ M\^200 + 215402\ M\^202 + 1146731\ M\^204 - 
                    800536\ M\^206 - 721454\ M\^208 + 993116\ M\^210 + 
                    271605\ M\^212 - 842849\ M\^214 + 42469\ M\^216 + 
                    559253\ M\^218 - 208534\ M\^220 - 245149\ M\^222 + 
                    203315\ M\^224 + 21854\ M\^226 - 86077\ M\^228 + 
                    32779\ M\^230 + 5473\ M\^232 - 7972\ M\^234 + 
                    3561\ M\^236 - 3060\ M\^238 + 3218\ M\^240 - 
                    2232\ M\^242 + 1049\ M\^244 - 359\ M\^246 + 89\ M\^248 - 
                    14\ M\^250 + M\^252)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 15]\), \(M\^30 + 
              L\^21\ M\^116 + 
              L\ \((2\ M\^20 - 6\ M\^22 + 10\ M\^24 - 12\ M\^26 + 3\ M\^28 + 
                    15\ M\^30 - 26\ M\^32 + 30\ M\^34 + 24\ M\^36 - 
                    27\ M\^38 + 8\ M\^40)\) + 
              L\^2\ \((M\^10 - 6\ M\^12 + 19\ M\^14 - 34\ M\^16 + 33\ M\^18 - 
                    14\ M\^20 - 26\ M\^22 + 31\ M\^24 + 85\ M\^26 - 
                    15\ M\^28 - 348\ M\^30 + 251\ M\^32 + 433\ M\^34 - 
                    784\ M\^36 + 292\ M\^38 + 910\ M\^40 - 714\ M\^42 - 
                    159\ M\^44 + 407\ M\^46 - 180\ M\^48 + 28\ M\^50)\) + 
              L\^3\ \((\(-M\^6\) + 11\ M\^8 - 50\ M\^10 + 143\ M\^12 - 
                    280\ M\^14 + 290\ M\^16 + 113\ M\^18 - 829\ M\^20 + 
                    1106\ M\^22 + 32\ M\^24 - 2225\ M\^26 + 2041\ M\^28 + 
                    1370\ M\^30 - 1710\ M\^32 - 1387\ M\^34 + 999\ M\^36 + 
                    2344\ M\^38 - 4697\ M\^40 + 1257\ M\^42 + 9486\ M\^44 - 
                    6644\ M\^46 - 5178\ M\^48 + 7415\ M\^50 - 1120\ M\^52 - 
                    2505\ M\^54 + 1806\ M\^56 - 513\ M\^58 + 56\ M\^60)\) + 
              L\^4\ \((\(-3\)\ M\^4 + 35\ M\^6 - 173\ M\^8 + 495\ M\^10 - 
                    907\ M\^12 + 868\ M\^14 + 690\ M\^16 - 3687\ M\^18 + 
                    3945\ M\^20 + 3436\ M\^22 - 12107\ M\^24 + 4732\ M\^26 + 
                    14801\ M\^28 - 7975\ M\^30 - 21869\ M\^32 - 498\ M\^34 + 
                    50654\ M\^36 + 2246\ M\^38 - 81875\ M\^40 + 
                    20068\ M\^42 + 60875\ M\^44 - 32163\ M\^46 + 
                    9212\ M\^48 + 7592\ M\^50 - 39387\ M\^52 + 32160\ M\^54 + 
                    12707\ M\^56 - 31409\ M\^58 + 12782\ M\^60 + 
                    6228\ M\^62 - 8535\ M\^64 + 3787\ M\^66 - 810\ M\^68 + 
                    70\ M\^70)\) + 
              L\^5\ \((\(-3\)\ M\^2 + 42\ M\^4 - 250\ M\^6 + 836\ M\^8 - 
                    1653\ M\^10 + 1447\ M\^12 + 1653\ M\^14 - 5864\ M\^16 + 
                    2685\ M\^18 + 10150\ M\^20 - 14202\ M\^22 - 1944\ M\^24 + 
                    13659\ M\^26 - 3905\ M\^28 - 4662\ M\^30 - 13273\ M\^32 + 
                    40476\ M\^34 + 20706\ M\^36 - 173688\ M\^38 + 
                    98052\ M\^40 + 280821\ M\^42 - 317202\ M\^44 - 
                    163423\ M\^46 + 376299\ M\^48 - 114679\ M\^50 - 
                    107628\ M\^52 + 247021\ M\^54 - 188211\ M\^56 - 
                    78209\ M\^58 + 220454\ M\^60 - 105551\ M\^62 - 
                    54742\ M\^64 + 93549\ M\^66 - 41677\ M\^68 - 
                    9213\ M\^70 + 22359\ M\^72 - 13566\ M\^74 + 4394\ M\^76 - 
                    765\ M\^78 + 56\ M\^80)\) + 
              L\^6\ \((\(-1\) + 18\ M\^2 - 136\ M\^4 + 575\ M\^6 - 
                    1453\ M\^8 + 1926\ M\^10 - 21\ M\^12 - 3547\ M\^14 + 
                    1100\ M\^16 + 10554\ M\^18 - 11644\ M\^20 - 
                    10483\ M\^22 + 12290\ M\^24 + 26598\ M\^26 - 
                    4059\ M\^28 - 108535\ M\^30 + 67194\ M\^32 + 
                    271096\ M\^34 - 434355\ M\^36 - 229658\ M\^38 + 
                    1106211\ M\^40 - 413762\ M\^42 - 1515720\ M\^44 + 
                    1493278\ M\^46 + 981324\ M\^48 - 1962370\ M\^50 + 
                    138563\ M\^52 + 1135447\ M\^54 - 577493\ M\^56 + 
                    214017\ M\^58 + 1009\ M\^60 - 600990\ M\^62 + 
                    684016\ M\^64 + 56557\ M\^66 - 606711\ M\^68 + 
                    369046\ M\^70 + 110161\ M\^72 - 245771\ M\^74 + 
                    108467\ M\^76 + 17446\ M\^78 - 46685\ M\^80 + 
                    29869\ M\^82 - 11625\ M\^84 + 2925\ M\^86 - 432\ M\^88 + 
                    28\ M\^90)\) + 
              L\^7\ \((1 - 15\ M\^2 + 101\ M\^4 - 401\ M\^6 + 1020\ M\^8 - 
                    1673\ M\^10 + 1754\ M\^12 - 1499\ M\^14 + 1882\ M\^16 - 
                    2251\ M\^18 + 5513\ M\^20 - 18090\ M\^22 + 2234\ M\^24 + 
                    99282\ M\^26 - 108490\ M\^28 - 282967\ M\^30 + 
                    586659\ M\^32 + 263307\ M\^34 - 1492507\ M\^36 + 
                    601675\ M\^38 + 2097879\ M\^40 - 2549158\ M\^42 - 
                    1056916\ M\^44 + 4190755\ M\^46 - 1558992\ M\^48 - 
                    3541805\ M\^50 + 3286029\ M\^52 + 879770\ M\^54 - 
                    1909042\ M\^56 + 772968\ M\^58 - 966773\ M\^60 + 
                    485847\ M\^62 + 1986779\ M\^64 - 2313130\ M\^66 - 
                    302917\ M\^68 + 2102596\ M\^70 - 1318406\ M\^72 - 
                    419130\ M\^74 + 1123217\ M\^76 - 585442\ M\^78 - 
                    162175\ M\^80 + 390561\ M\^82 - 198487\ M\^84 - 
                    14490\ M\^86 + 76879\ M\^88 - 51133\ M\^90 + 
                    19964\ M\^92 - 5442\ M\^94 + 1066\ M\^96 - 135\ M\^98 + 
                    8\ M\^100)\) + 
              L\^8\ \((6\ M\^6 - 99\ M\^8 + 718\ M\^10 - 2915\ M\^12 + 
                    6795\ M\^14 - 7187\ M\^16 - 4817\ M\^18 + 23181\ M\^20 - 
                    14746\ M\^22 - 45947\ M\^24 + 129237\ M\^26 - 
                    120961\ M\^28 - 181227\ M\^30 + 674984\ M\^32 - 
                    354990\ M\^34 - 1376272\ M\^36 + 2103041\ M\^38 + 
                    915188\ M\^40 - 4171141\ M\^42 + 1691178\ M\^44 + 
                    3918819\ M\^46 - 4808787\ M\^48 - 30302\ M\^50 + 
                    4607178\ M\^52 - 3810980\ M\^54 - 375809\ M\^56 + 
                    2453103\ M\^58 - 3094003\ M\^60 + 3092868\ M\^62 + 
                    1627936\ M\^64 - 5592288\ M\^66 + 2426737\ M\^68 + 
                    2327405\ M\^70 - 3049404\ M\^72 + 2084673\ M\^74 - 
                    72452\ M\^76 - 2622578\ M\^78 + 2398420\ M\^80 + 
                    450236\ M\^82 - 1844548\ M\^84 + 856319\ M\^86 + 
                    335910\ M\^88 - 538515\ M\^90 + 215911\ M\^92 + 
                    33776\ M\^94 - 92210\ M\^96 + 61173\ M\^98 - 
                    24736\ M\^100 + 6727\ M\^102 - 1275\ M\^104 + 
                    177\ M\^106 - 18\ M\^108 + M\^110)\) + 
              L\^9\ \((15\ M\^12 - 264\ M\^14 + 2016\ M\^16 - 8491\ M\^18 + 
                    20161\ M\^20 - 20825\ M\^22 - 18270\ M\^24 + 
                    82667\ M\^26 - 80802\ M\^28 - 76331\ M\^30 + 
                    375388\ M\^32 - 508124\ M\^34 - 265844\ M\^36 + 
                    1774077\ M\^38 - 1332804\ M\^40 - 2442740\ M\^42 + 
                    4359311\ M\^44 + 375998\ M\^46 - 5557062\ M\^48 + 
                    3419619\ M\^50 + 1655811\ M\^52 - 3854060\ M\^54 + 
                    4585327\ M\^56 - 1875881\ M\^58 - 5210320\ M\^60 + 
                    7906649\ M\^62 - 1762989\ M\^64 - 6411223\ M\^66 + 
                    7844615\ M\^68 - 242210\ M\^70 - 4822289\ M\^72 + 
                    3194677\ M\^74 - 2402974\ M\^76 + 809205\ M\^78 + 
                    4747068\ M\^80 - 5245545\ M\^82 - 1312083\ M\^84 + 
                    4675196\ M\^86 - 1842740\ M\^88 - 1349534\ M\^90 + 
                    1740619\ M\^92 - 582363\ M\^94 - 357368\ M\^96 + 
                    491755\ M\^98 - 186674\ M\^100 - 52568\ M\^102 + 
                    95176\ M\^104 - 54465\ M\^106 + 19689\ M\^108 - 
                    5108\ M\^110 + 947\ M\^112 - 111\ M\^114 + 6\ M\^116)\) + 
              L\^10\ \((20\ M\^18 - 366\ M\^20 + 2899\ M\^22 - 12793\ M\^24 + 
                    33224\ M\^26 - 45796\ M\^28 + 3811\ M\^30 + 
                    119007\ M\^32 - 244854\ M\^34 + 110879\ M\^36 + 
                    575200\ M\^38 - 1241996\ M\^40 + 173371\ M\^42 + 
                    2473485\ M\^44 - 2679976\ M\^46 - 1370763\ M\^48 + 
                    4623583\ M\^50 - 2827103\ M\^52 - 1537017\ M\^54 + 
                    5129555\ M\^56 - 5920763\ M\^58 - 142038\ M\^60 + 
                    9670445\ M\^62 - 8164320\ M\^64 - 4274097\ M\^66 + 
                    8894542\ M\^68 - 3230946\ M\^70 - 42353\ M\^72 + 
                    2784413\ M\^74 - 6405217\ M\^76 + 4835854\ M\^78 + 
                    2871771\ M\^80 - 8428485\ M\^82 + 4525254\ M\^84 + 
                    3942807\ M\^86 - 6286932\ M\^88 + 1951259\ M\^90 + 
                    2320668\ M\^92 - 3175817\ M\^94 + 1145478\ M\^96 + 
                    1191320\ M\^98 - 1472693\ M\^100 + 301875\ M\^102 + 
                    439236\ M\^104 - 347723\ M\^106 + 58918\ M\^108 + 
                    67163\ M\^110 - 65359\ M\^112 + 32857\ M\^114 - 
                    10785\ M\^116 + 2278\ M\^118 - 279\ M\^120 + 
                    15\ M\^122)\) + 
              L\^20\ \((8\ M\^106 - 27\ M\^108 + 24\ M\^110 + 30\ M\^112 - 
                    26\ M\^114 + 15\ M\^116 + 3\ M\^118 - 12\ M\^120 + 
                    10\ M\^122 - 6\ M\^124 + 2\ M\^126)\) + 
              L\^11\ \((15\ M\^24 - 279\ M\^26 + 2278\ M\^28 - 10785\ M\^30 + 
                    32857\ M\^32 - 65359\ M\^34 + 67163\ M\^36 + 
                    58918\ M\^38 - 347723\ M\^40 + 439236\ M\^42 + 
                    301875\ M\^44 - 1472693\ M\^46 + 1191320\ M\^48 + 
                    1145478\ M\^50 - 3175817\ M\^52 + 2320668\ M\^54 + 
                    1951259\ M\^56 - 6286932\ M\^58 + 3942807\ M\^60 + 
                    4525254\ M\^62 - 8428485\ M\^64 + 2871771\ M\^66 + 
                    4835854\ M\^68 - 6405217\ M\^70 + 2784413\ M\^72 - 
                    42353\ M\^74 - 3230946\ M\^76 + 8894542\ M\^78 - 
                    4274097\ M\^80 - 8164320\ M\^82 + 9670445\ M\^84 - 
                    142038\ M\^86 - 5920763\ M\^88 + 5129555\ M\^90 - 
                    1537017\ M\^92 - 2827103\ M\^94 + 4623583\ M\^96 - 
                    1370763\ M\^98 - 2679976\ M\^100 + 2473485\ M\^102 + 
                    173371\ M\^104 - 1241996\ M\^106 + 575200\ M\^108 + 
                    110879\ M\^110 - 244854\ M\^112 + 119007\ M\^114 + 
                    3811\ M\^116 - 45796\ M\^118 + 33224\ M\^120 - 
                    12793\ M\^122 + 2899\ M\^124 - 366\ M\^126 + 
                    20\ M\^128)\) + 
              L\^12\ \((6\ M\^30 - 111\ M\^32 + 947\ M\^34 - 5108\ M\^36 + 
                    19689\ M\^38 - 54465\ M\^40 + 95176\ M\^42 - 
                    52568\ M\^44 - 186674\ M\^46 + 491755\ M\^48 - 
                    357368\ M\^50 - 582363\ M\^52 + 1740619\ M\^54 - 
                    1349534\ M\^56 - 1842740\ M\^58 + 4675196\ M\^60 - 
                    1312083\ M\^62 - 5245545\ M\^64 + 4747068\ M\^66 + 
                    809205\ M\^68 - 2402974\ M\^70 + 3194677\ M\^72 - 
                    4822289\ M\^74 - 242210\ M\^76 + 7844615\ M\^78 - 
                    6411223\ M\^80 - 1762989\ M\^82 + 7906649\ M\^84 - 
                    5210320\ M\^86 - 1875881\ M\^88 + 4585327\ M\^90 - 
                    3854060\ M\^92 + 1655811\ M\^94 + 3419619\ M\^96 - 
                    5557062\ M\^98 + 375998\ M\^100 + 4359311\ M\^102 - 
                    2442740\ M\^104 - 1332804\ M\^106 + 1774077\ M\^108 - 
                    265844\ M\^110 - 508124\ M\^112 + 375388\ M\^114 - 
                    76331\ M\^116 - 80802\ M\^118 + 82667\ M\^120 - 
                    18270\ M\^122 - 20825\ M\^124 + 20161\ M\^126 - 
                    8491\ M\^128 + 2016\ M\^130 - 264\ M\^132 + 
                    15\ M\^134)\) + 
              L\^19\ \((28\ M\^96 - 180\ M\^98 + 407\ M\^100 - 159\ M\^102 - 
                    714\ M\^104 + 910\ M\^106 + 292\ M\^108 - 784\ M\^110 + 
                    433\ M\^112 + 251\ M\^114 - 348\ M\^116 - 15\ M\^118 + 
                    85\ M\^120 + 31\ M\^122 - 26\ M\^124 - 14\ M\^126 + 
                    33\ M\^128 - 34\ M\^130 + 19\ M\^132 - 6\ M\^134 + 
                    M\^136)\) + 
              L\^18\ \((56\ M\^86 - 513\ M\^88 + 1806\ M\^90 - 2505\ M\^92 - 
                    1120\ M\^94 + 7415\ M\^96 - 5178\ M\^98 - 6644\ M\^100 + 
                    9486\ M\^102 + 1257\ M\^104 - 4697\ M\^106 + 
                    2344\ M\^108 + 999\ M\^110 - 1387\ M\^112 - 
                    1710\ M\^114 + 1370\ M\^116 + 2041\ M\^118 - 
                    2225\ M\^120 + 32\ M\^122 + 1106\ M\^124 - 829\ M\^126 + 
                    113\ M\^128 + 290\ M\^130 - 280\ M\^132 + 143\ M\^134 - 
                    50\ M\^136 + 11\ M\^138 - M\^140)\) + 
              L\^13\ \((M\^36 - 18\ M\^38 + 177\ M\^40 - 1275\ M\^42 + 
                    6727\ M\^44 - 24736\ M\^46 + 61173\ M\^48 - 
                    92210\ M\^50 + 33776\ M\^52 + 215911\ M\^54 - 
                    538515\ M\^56 + 335910\ M\^58 + 856319\ M\^60 - 
                    1844548\ M\^62 + 450236\ M\^64 + 2398420\ M\^66 - 
                    2622578\ M\^68 - 72452\ M\^70 + 2084673\ M\^72 - 
                    3049404\ M\^74 + 2327405\ M\^76 + 2426737\ M\^78 - 
                    5592288\ M\^80 + 1627936\ M\^82 + 3092868\ M\^84 - 
                    3094003\ M\^86 + 2453103\ M\^88 - 375809\ M\^90 - 
                    3810980\ M\^92 + 4607178\ M\^94 - 30302\ M\^96 - 
                    4808787\ M\^98 + 3918819\ M\^100 + 1691178\ M\^102 - 
                    4171141\ M\^104 + 915188\ M\^106 + 2103041\ M\^108 - 
                    1376272\ M\^110 - 354990\ M\^112 + 674984\ M\^114 - 
                    181227\ M\^116 - 120961\ M\^118 + 129237\ M\^120 - 
                    45947\ M\^122 - 14746\ M\^124 + 23181\ M\^126 - 
                    4817\ M\^128 - 7187\ M\^130 + 6795\ M\^132 - 
                    2915\ M\^134 + 718\ M\^136 - 99\ M\^138 + 6\ M\^140)\) + 
              L\^17\ \((70\ M\^76 - 810\ M\^78 + 3787\ M\^80 - 8535\ M\^82 + 
                    6228\ M\^84 + 12782\ M\^86 - 31409\ M\^88 + 
                    12707\ M\^90 + 32160\ M\^92 - 39387\ M\^94 + 
                    7592\ M\^96 + 9212\ M\^98 - 32163\ M\^100 + 
                    60875\ M\^102 + 20068\ M\^104 - 81875\ M\^106 + 
                    2246\ M\^108 + 50654\ M\^110 - 498\ M\^112 - 
                    21869\ M\^114 - 7975\ M\^116 + 14801\ M\^118 + 
                    4732\ M\^120 - 12107\ M\^122 + 3436\ M\^124 + 
                    3945\ M\^126 - 3687\ M\^128 + 690\ M\^130 + 868\ M\^132 - 
                    907\ M\^134 + 495\ M\^136 - 173\ M\^138 + 35\ M\^140 - 
                    3\ M\^142)\) + 
              L\^16\ \((56\ M\^66 - 765\ M\^68 + 4394\ M\^70 - 13566\ M\^72 + 
                    22359\ M\^74 - 9213\ M\^76 - 41677\ M\^78 + 
                    93549\ M\^80 - 54742\ M\^82 - 105551\ M\^84 + 
                    220454\ M\^86 - 78209\ M\^88 - 188211\ M\^90 + 
                    247021\ M\^92 - 107628\ M\^94 - 114679\ M\^96 + 
                    376299\ M\^98 - 163423\ M\^100 - 317202\ M\^102 + 
                    280821\ M\^104 + 98052\ M\^106 - 173688\ M\^108 + 
                    20706\ M\^110 + 40476\ M\^112 - 13273\ M\^114 - 
                    4662\ M\^116 - 3905\ M\^118 + 13659\ M\^120 - 
                    1944\ M\^122 - 14202\ M\^124 + 10150\ M\^126 + 
                    2685\ M\^128 - 5864\ M\^130 + 1653\ M\^132 + 
                    1447\ M\^134 - 1653\ M\^136 + 836\ M\^138 - 250\ M\^140 + 
                    42\ M\^142 - 3\ M\^144)\) + 
              L\^15\ \((28\ M\^56 - 432\ M\^58 + 2925\ M\^60 - 11625\ M\^62 + 
                    29869\ M\^64 - 46685\ M\^66 + 17446\ M\^68 + 
                    108467\ M\^70 - 245771\ M\^72 + 110161\ M\^74 + 
                    369046\ M\^76 - 606711\ M\^78 + 56557\ M\^80 + 
                    684016\ M\^82 - 600990\ M\^84 + 1009\ M\^86 + 
                    214017\ M\^88 - 577493\ M\^90 + 1135447\ M\^92 + 
                    138563\ M\^94 - 1962370\ M\^96 + 981324\ M\^98 + 
                    1493278\ M\^100 - 1515720\ M\^102 - 413762\ M\^104 + 
                    1106211\ M\^106 - 229658\ M\^108 - 434355\ M\^110 + 
                    271096\ M\^112 + 67194\ M\^114 - 108535\ M\^116 - 
                    4059\ M\^118 + 26598\ M\^120 + 12290\ M\^122 - 
                    10483\ M\^124 - 11644\ M\^126 + 10554\ M\^128 + 
                    1100\ M\^130 - 3547\ M\^132 - 21\ M\^134 + 1926\ M\^136 - 
                    1453\ M\^138 + 575\ M\^140 - 136\ M\^142 + 18\ M\^144 - 
                    M\^146)\) + 
              L\^14\ \((8\ M\^46 - 135\ M\^48 + 1066\ M\^50 - 5442\ M\^52 + 
                    19964\ M\^54 - 51133\ M\^56 + 76879\ M\^58 - 
                    14490\ M\^60 - 198487\ M\^62 + 390561\ M\^64 - 
                    162175\ M\^66 - 585442\ M\^68 + 1123217\ M\^70 - 
                    419130\ M\^72 - 1318406\ M\^74 + 2102596\ M\^76 - 
                    302917\ M\^78 - 2313130\ M\^80 + 1986779\ M\^82 + 
                    485847\ M\^84 - 966773\ M\^86 + 772968\ M\^88 - 
                    1909042\ M\^90 + 879770\ M\^92 + 3286029\ M\^94 - 
                    3541805\ M\^96 - 1558992\ M\^98 + 4190755\ M\^100 - 
                    1056916\ M\^102 - 2549158\ M\^104 + 2097879\ M\^106 + 
                    601675\ M\^108 - 1492507\ M\^110 + 263307\ M\^112 + 
                    586659\ M\^114 - 282967\ M\^116 - 108490\ M\^118 + 
                    99282\ M\^120 + 2234\ M\^122 - 18090\ M\^124 + 
                    5513\ M\^126 - 2251\ M\^128 + 1882\ M\^130 - 
                    1499\ M\^132 + 1754\ M\^134 - 1673\ M\^136 + 
                    1020\ M\^138 - 401\ M\^140 + 101\ M\^142 - 15\ M\^144 + 
                    M\^146)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 16]\), \(M\^28 + 
              L\^23\ M\^122 + 
              L\ \((\(-3\)\ M\^20 + 10\ M\^22 - 16\ M\^24 + 13\ M\^26 + 
                    2\ M\^28 - 26\ M\^30 + 50\ M\^32 + 7\ M\^34 - 27\ M\^36 + 
                    19\ M\^38 - 6\ M\^40)\) + 
              L\^2\ \((3\ M\^12 - 20\ M\^14 + 57\ M\^16 - 89\ M\^18 + 
                    72\ M\^20 + 28\ M\^22 - 154\ M\^24 + 75\ M\^26 + 
                    68\ M\^28 + 140\ M\^30 - 298\ M\^32 - 312\ M\^34 + 
                    986\ M\^36 - 26\ M\^38 - 677\ M\^40 + 579\ M\^42 - 
                    52\ M\^44 - 286\ M\^46 + 233\ M\^48 - 89\ M\^50 + 
                    15\ M\^52)\) + 
              L\^3\ \((\(-M\^4\) + 10\ M\^6 - 42\ M\^8 + 101\ M\^10 - 
                    154\ M\^12 + 153\ M\^14 - 170\ M\^16 + 431\ M\^18 - 
                    709\ M\^20 - 222\ M\^22 + 2790\ M\^24 - 3643\ M\^26 - 
                    1112\ M\^28 + 7708\ M\^30 - 7656\ M\^32 - 2030\ M\^34 + 
                    11970\ M\^36 - 9110\ M\^38 - 3220\ M\^40 + 11631\ M\^42 - 
                    3959\ M\^44 - 2113\ M\^46 + 2034\ M\^48 - 2080\ M\^50 + 
                    1880\ M\^52 - 424\ M\^54 - 1020\ M\^56 + 1193\ M\^58 - 
                    611\ M\^60 + 166\ M\^62 - 20\ M\^64)\) + 
              L\^4\ \((1 - 9\ M\^2 + 40\ M\^4 - 134\ M\^6 + 401\ M\^8 - 
                    974\ M\^10 + 1559\ M\^12 - 889\ M\^14 - 2245\ M\^16 + 
                    5436\ M\^18 - 1821\ M\^20 - 9585\ M\^22 + 13156\ M\^24 + 
                    4783\ M\^26 - 24081\ M\^28 + 9382\ M\^30 + 32767\ M\^32 - 
                    39839\ M\^34 - 29917\ M\^36 + 82728\ M\^38 - 
                    5656\ M\^40 - 107817\ M\^42 + 76960\ M\^44 + 
                    77025\ M\^46 - 111507\ M\^48 + 10601\ M\^50 + 
                    77941\ M\^52 - 58113\ M\^54 - 11712\ M\^56 + 
                    33686\ M\^58 - 14403\ M\^60 - 914\ M\^62 + 3710\ M\^64 - 
                    3369\ M\^66 + 2938\ M\^68 - 1844\ M\^70 + 709\ M\^72 - 
                    154\ M\^74 + 15\ M\^76)\) + 
              L\^5\ \((\(-4\) + 49\ M\^2 - 275\ M\^4 + 936\ M\^6 - 
                    2173\ M\^8 + 3622\ M\^10 - 3897\ M\^12 - 314\ M\^14 + 
                    12390\ M\^16 - 24180\ M\^18 + 6189\ M\^20 + 
                    57342\ M\^22 - 96429\ M\^24 - 5846\ M\^26 + 
                    191544\ M\^28 - 164480\ M\^30 - 173427\ M\^32 + 
                    344624\ M\^34 + 54398\ M\^36 - 400195\ M\^38 - 
                    27430\ M\^40 + 400384\ M\^42 + 110665\ M\^44 - 
                    465081\ M\^46 - 96362\ M\^48 + 519234\ M\^50 - 
                    25835\ M\^52 - 320113\ M\^54 + 114825\ M\^56 + 
                    49584\ M\^58 - 25029\ M\^60 + 41632\ M\^62 - 
                    86999\ M\^64 + 25894\ M\^66 + 59967\ M\^68 - 
                    55097\ M\^70 + 5039\ M\^72 + 17807\ M\^74 - 
                    13853\ M\^76 + 6544\ M\^78 - 2916\ M\^80 + 1232\ M\^82 - 
                    382\ M\^84 + 71\ M\^86 - 6\ M\^88)\) + 
              L\^6\ \((6 - 94\ M\^2 + 668\ M\^4 - 2760\ M\^6 + 6982\ M\^8 - 
                    9825\ M\^10 + 2476\ M\^12 + 17712\ M\^14 - 31810\ M\^16 + 
                    9539\ M\^18 + 55842\ M\^20 - 113037\ M\^22 + 
                    43044\ M\^24 + 189148\ M\^26 - 273754\ M\^28 - 
                    123063\ M\^30 + 540687\ M\^32 - 154771\ M\^34 - 
                    667484\ M\^36 + 546246\ M\^38 + 646495\ M\^40 - 
                    951888\ M\^42 - 519060\ M\^44 + 1262746\ M\^46 + 
                    139168\ M\^48 - 1419323\ M\^50 + 726318\ M\^52 + 
                    1196781\ M\^54 - 1623232\ M\^56 - 142982\ M\^58 + 
                    1986748\ M\^60 - 1046497\ M\^62 - 1321379\ M\^64 + 
                    1433643\ M\^66 + 254469\ M\^68 - 916316\ M\^70 + 
                    274377\ M\^72 + 220929\ M\^74 - 166844\ M\^76 + 
                    34998\ M\^78 - 8725\ M\^80 - 3499\ M\^82 + 22541\ M\^84 - 
                    23240\ M\^86 + 12212\ M\^88 - 4124\ M\^90 + 1123\ M\^92 - 
                    310\ M\^94 + 78\ M\^96 - 13\ M\^98 + M\^100)\) + 
              L\^7\ \((\(-4\) + 78\ M\^2 - 687\ M\^4 + 3506\ M\^6 - 
                    10953\ M\^8 + 19309\ M\^10 - 9024\ M\^12 - 35310\ M\^14 + 
                    62476\ M\^16 + 29209\ M\^18 - 184001\ M\^20 + 
                    102093\ M\^22 + 302526\ M\^24 - 446923\ M\^26 - 
                    295009\ M\^28 + 1239911\ M\^30 - 431281\ M\^32 - 
                    2297512\ M\^34 + 2597367\ M\^36 + 2867795\ M\^38 - 
                    6627899\ M\^40 - 1783805\ M\^42 + 11968010\ M\^44 - 
                    1559940\ M\^46 - 17067554\ M\^48 + 6912826\ M\^50 + 
                    19270946\ M\^52 - 13086685\ M\^54 - 16508778\ M\^56 + 
                    17717284\ M\^58 + 9727462\ M\^60 - 17319436\ M\^62 - 
                    1964105\ M\^64 + 12209817\ M\^66 - 2443511\ M\^68 - 
                    5839950\ M\^70 + 2444292\ M\^72 + 1801238\ M\^74 - 
                    768546\ M\^76 - 630334\ M\^78 - 92284\ M\^80 + 
                    582843\ M\^82 - 41945\ M\^84 - 330582\ M\^86 + 
                    170182\ M\^88 + 37238\ M\^90 - 74685\ M\^92 + 
                    42789\ M\^94 - 19568\ M\^96 + 8859\ M\^98 - 
                    3305\ M\^100 + 831\ M\^102 - 122\ M\^104 + 8\ M\^106)\) + 
              L\^8\ \((1 - 25\ M\^2 + 276\ M\^4 - 1770\ M\^6 + 7187\ M\^8 - 
                    18521\ M\^10 + 27126\ M\^12 - 9915\ M\^14 - 
                    34169\ M\^16 + 42222\ M\^18 + 22131\ M\^20 - 
                    34351\ M\^22 - 42999\ M\^24 - 163809\ M\^26 + 
                    609927\ M\^28 + 108899\ M\^30 - 2182994\ M\^32 + 
                    1259062\ M\^34 + 5031804\ M\^36 - 6155848\ M\^38 - 
                    7943628\ M\^40 + 16956127\ M\^42 + 7801767\ M\^44 - 
                    34661450\ M\^46 - 159027\ M\^48 + 55842606\ M\^50 - 
                    16561410\ M\^52 - 71876404\ M\^54 + 37096464\ M\^56 + 
                    72643136\ M\^58 - 51069859\ M\^60 - 56245543\ M\^62 + 
                    51200707\ M\^64 + 34146019\ M\^66 - 37255603\ M\^68 - 
                    17941503\ M\^70 + 19785327\ M\^72 + 11163917\ M\^74 - 
                    8801254\ M\^76 - 8959573\ M\^78 + 5365502\ M\^80 + 
                    5947788\ M\^82 - 4433567\ M\^84 - 2178051\ M\^86 + 
                    2782199\ M\^88 - 94271\ M\^90 - 857426\ M\^92 + 
                    353470\ M\^94 + 5792\ M\^96 - 47672\ M\^98 + 
                    53465\ M\^100 - 58761\ M\^102 + 38938\ M\^104 - 
                    15299\ M\^106 + 3609\ M\^108 - 480\ M\^110 + 
                    28\ M\^112)\) + 
              L\^9\ \((M\^2 - 21\ M\^4 + 212\ M\^6 - 1346\ M\^8 + 
                    5962\ M\^10 - 19175\ M\^12 + 43705\ M\^14 - 
                    59012\ M\^16 - 2976\ M\^18 + 185683\ M\^20 - 
                    251914\ M\^22 - 301578\ M\^24 + 1106759\ M\^26 - 
                    94302\ M\^28 - 2896824\ M\^30 + 1697052\ M\^32 + 
                    6677022\ M\^34 - 6803813\ M\^36 - 13028472\ M\^38 + 
                    19463987\ M\^40 + 20336944\ M\^42 - 44791290\ M\^44 - 
                    21176882\ M\^46 + 83522941\ M\^48 + 3915418\ M\^50 - 
                    127329491\ M\^52 + 38425598\ M\^54 + 159125504\ M\^56 - 
                    94528282\ M\^58 - 163230116\ M\^60 + 134425745\ M\^62 + 
                    136537442\ M\^64 - 132647512\ M\^66 - 93511912\ M\^68 + 
                    95254431\ M\^70 + 59656457\ M\^72 - 48414034\ M\^74 - 
                    43630811\ M\^76 + 17656285\ M\^78 + 35667767\ M\^80 - 
                    7910172\ M\^82 - 25253132\ M\^84 + 7524784\ M\^86 + 
                    12387328\ M\^88 - 5914449\ M\^90 - 3321302\ M\^92 + 
                    2417713\ M\^94 + 128168\ M\^96 - 51133\ M\^98 - 
                    149576\ M\^100 - 239901\ M\^102 + 294644\ M\^104 - 
                    20338\ M\^106 - 130584\ M\^108 + 100950\ M\^110 - 
                    38466\ M\^112 + 8494\ M\^114 - 1046\ M\^116 + 
                    56\ M\^118)\) + 
              L\^10\ \((8\ M\^8 - 170\ M\^10 + 1646\ M\^12 - 9446\ M\^14 + 
                    34759\ M\^16 - 81008\ M\^18 + 97903\ M\^20 + 
                    39733\ M\^22 - 371051\ M\^24 + 498207\ M\^26 + 
                    396912\ M\^28 - 2229007\ M\^30 + 1950352\ M\^32 + 
                    4260353\ M\^34 - 10295247\ M\^36 - 2310744\ M\^38 + 
                    29157859\ M\^40 - 14606109\ M\^42 - 58059229\ M\^44 + 
                    61212743\ M\^46 + 85047224\ M\^48 - 145476697\ M\^50 - 
                    83992390\ M\^52 + 251797644\ M\^54 + 29040095\ M\^56 - 
                    339996227\ M\^58 + 79754277\ M\^60 + 369906146\ M\^62 - 
                    200571634\ M\^64 - 334737079\ M\^66 + 270857794\ M\^68 + 
                    262193827\ M\^70 - 257658484\ M\^72 - 184501521\ M\^74 + 
                    190778326\ M\^76 + 123352465\ M\^78 - 119059002\ M\^80 - 
                    78132861\ M\^82 + 68885925\ M\^84 + 40962655\ M\^86 - 
                    38022180\ M\^88 - 12683508\ M\^90 + 16865473\ M\^92 - 
                    1898469\ M\^94 - 2651748\ M\^96 + 3801702\ M\^98 - 
                    3171609\ M\^100 - 849129\ M\^102 + 2897931\ M\^104 - 
                    1067446\ M\^106 - 840805\ M\^108 + 828260\ M\^110 - 
                    102055\ M\^112 - 215246\ M\^114 + 160962\ M\^116 - 
                    57678\ M\^118 + 11984\ M\^120 - 1390\ M\^122 + 
                    70\ M\^124)\) + 
              L\^22\ \((\(-6\)\ M\^110 + 19\ M\^112 - 27\ M\^114 + 
                    7\ M\^116 + 50\ M\^118 - 26\ M\^120 + 2\ M\^122 + 
                    13\ M\^124 - 16\ M\^126 + 10\ M\^128 - 3\ M\^130)\) + 
              L\^11\ \((28\ M\^14 - 592\ M\^16 + 5549\ M\^18 - 29891\ M\^20 + 
                    98364\ M\^22 - 181963\ M\^24 + 74697\ M\^26 + 
                    480662\ M\^28 - 1056423\ M\^30 + 119218\ M\^32 + 
                    2939446\ M\^34 - 4456874\ M\^36 - 1589945\ M\^38 + 
                    13351695\ M\^40 - 12705330\ M\^42 - 19024634\ M\^44 + 
                    51723611\ M\^46 - 1456041\ M\^48 - 114965842\ M\^50 + 
                    76263718\ M\^52 + 178935621\ M\^54 - 213754219\ M\^56 - 
                    202843036\ M\^58 + 379555989\ M\^60 + 156810805\ M\^62 - 
                    511128430\ M\^64 - 46421388\ M\^66 + 564391606\ M\^68 - 
                    88258721\ M\^70 - 546381178\ M\^72 + 197897270\ M\^74 + 
                    489499287\ M\^76 - 251152508\ M\^78 - 402565189\ M\^80 + 
                    255585556\ M\^82 + 286558523\ M\^84 - 226113779\ M\^86 - 
                    158288178\ M\^88 + 171008456\ M\^90 + 50986979\ M\^92 - 
                    102442616\ M\^94 + 9024172\ M\^96 + 41802266\ M\^98 - 
                    23399366\ M\^100 - 5515470\ M\^102 + 14294549\ M\^104 - 
                    6004430\ M\^106 - 3578839\ M\^108 + 4743431\ M\^110 - 
                    920532\ M\^112 - 1327849\ M\^114 + 939924\ M\^116 - 
                    46634\ M\^118 - 245962\ M\^120 + 160433\ M\^122 - 
                    53276\ M\^124 + 10460\ M\^126 - 1158\ M\^128 + 
                    56\ M\^130)\) + 
              L\^12\ \((56\ M\^20 - 1158\ M\^22 + 10460\ M\^24 - 
                    53276\ M\^26 + 160433\ M\^28 - 245962\ M\^30 - 
                    46634\ M\^32 + 939924\ M\^34 - 1327849\ M\^36 - 
                    920532\ M\^38 + 4743431\ M\^40 - 3578839\ M\^42 - 
                    6004430\ M\^44 + 14294549\ M\^46 - 5515470\ M\^48 - 
                    23399366\ M\^50 + 41802266\ M\^52 + 9024172\ M\^54 - 
                    102442616\ M\^56 + 50986979\ M\^58 + 171008456\ M\^60 - 
                    158288178\ M\^62 - 226113779\ M\^64 + 286558523\ M\^66 + 
                    255585556\ M\^68 - 402565189\ M\^70 - 251152508\ M\^72 + 
                    489499287\ M\^74 + 197897270\ M\^76 - 546381178\ M\^78 - 
                    88258721\ M\^80 + 564391606\ M\^82 - 46421388\ M\^84 - 
                    511128430\ M\^86 + 156810805\ M\^88 + 379555989\ M\^90 - 
                    202843036\ M\^92 - 213754219\ M\^94 + 178935621\ M\^96 + 
                    76263718\ M\^98 - 114965842\ M\^100 - 1456041\ M\^102 + 
                    51723611\ M\^104 - 19024634\ M\^106 - 12705330\ M\^108 + 
                    13351695\ M\^110 - 1589945\ M\^112 - 4456874\ M\^114 + 
                    2939446\ M\^116 + 119218\ M\^118 - 1056423\ M\^120 + 
                    480662\ M\^122 + 74697\ M\^124 - 181963\ M\^126 + 
                    98364\ M\^128 - 29891\ M\^130 + 5549\ M\^132 - 
                    592\ M\^134 + 28\ M\^136)\) + 
              L\^21\ \((15\ M\^98 - 89\ M\^100 + 233\ M\^102 - 286\ M\^104 - 
                    52\ M\^106 + 579\ M\^108 - 677\ M\^110 - 26\ M\^112 + 
                    986\ M\^114 - 312\ M\^116 - 298\ M\^118 + 140\ M\^120 + 
                    68\ M\^122 + 75\ M\^124 - 154\ M\^126 + 28\ M\^128 + 
                    72\ M\^130 - 89\ M\^132 + 57\ M\^134 - 20\ M\^136 + 
                    3\ M\^138)\) + 
              L\^13\ \((70\ M\^26 - 1390\ M\^28 + 11984\ M\^30 - 
                    57678\ M\^32 + 160962\ M\^34 - 215246\ M\^36 - 
                    102055\ M\^38 + 828260\ M\^40 - 840805\ M\^42 - 
                    1067446\ M\^44 + 2897931\ M\^46 - 849129\ M\^48 - 
                    3171609\ M\^50 + 3801702\ M\^52 - 2651748\ M\^54 - 
                    1898469\ M\^56 + 16865473\ M\^58 - 12683508\ M\^60 - 
                    38022180\ M\^62 + 40962655\ M\^64 + 68885925\ M\^66 - 
                    78132861\ M\^68 - 119059002\ M\^70 + 123352465\ M\^72 + 
                    190778326\ M\^74 - 184501521\ M\^76 - 257658484\ M\^78 + 
                    262193827\ M\^80 + 270857794\ M\^82 - 334737079\ M\^84 - 
                    200571634\ M\^86 + 369906146\ M\^88 + 79754277\ M\^90 - 
                    339996227\ M\^92 + 29040095\ M\^94 + 251797644\ M\^96 - 
                    83992390\ M\^98 - 145476697\ M\^100 + 85047224\ M\^102 + 
                    61212743\ M\^104 - 58059229\ M\^106 - 14606109\ M\^108 + 
                    29157859\ M\^110 - 2310744\ M\^112 - 10295247\ M\^114 + 
                    4260353\ M\^116 + 1950352\ M\^118 - 2229007\ M\^120 + 
                    396912\ M\^122 + 498207\ M\^124 - 371051\ M\^126 + 
                    39733\ M\^128 + 97903\ M\^130 - 81008\ M\^132 + 
                    34759\ M\^134 - 9446\ M\^136 + 1646\ M\^138 - 
                    170\ M\^140 + 8\ M\^142)\) + 
              L\^20\ \((\(-20\)\ M\^86 + 166\ M\^88 - 611\ M\^90 + 
                    1193\ M\^92 - 1020\ M\^94 - 424\ M\^96 + 1880\ M\^98 - 
                    2080\ M\^100 + 2034\ M\^102 - 2113\ M\^104 - 
                    3959\ M\^106 + 11631\ M\^108 - 3220\ M\^110 - 
                    9110\ M\^112 + 11970\ M\^114 - 2030\ M\^116 - 
                    7656\ M\^118 + 7708\ M\^120 - 1112\ M\^122 - 
                    3643\ M\^124 + 2790\ M\^126 - 222\ M\^128 - 709\ M\^130 + 
                    431\ M\^132 - 170\ M\^134 + 153\ M\^136 - 154\ M\^138 + 
                    101\ M\^140 - 42\ M\^142 + 10\ M\^144 - M\^146)\) + 
              L\^14\ \((56\ M\^32 - 1046\ M\^34 + 8494\ M\^36 - 
                    38466\ M\^38 + 100950\ M\^40 - 130584\ M\^42 - 
                    20338\ M\^44 + 294644\ M\^46 - 239901\ M\^48 - 
                    149576\ M\^50 - 51133\ M\^52 + 128168\ M\^54 + 
                    2417713\ M\^56 - 3321302\ M\^58 - 5914449\ M\^60 + 
                    12387328\ M\^62 + 7524784\ M\^64 - 25253132\ M\^66 - 
                    7910172\ M\^68 + 35667767\ M\^70 + 17656285\ M\^72 - 
                    43630811\ M\^74 - 48414034\ M\^76 + 59656457\ M\^78 + 
                    95254431\ M\^80 - 93511912\ M\^82 - 132647512\ M\^84 + 
                    136537442\ M\^86 + 134425745\ M\^88 - 163230116\ M\^90 - 
                    94528282\ M\^92 + 159125504\ M\^94 + 38425598\ M\^96 - 
                    127329491\ M\^98 + 3915418\ M\^100 + 83522941\ M\^102 - 
                    21176882\ M\^104 - 44791290\ M\^106 + 20336944\ M\^108 + 
                    19463987\ M\^110 - 13028472\ M\^112 - 6803813\ M\^114 + 
                    6677022\ M\^116 + 1697052\ M\^118 - 2896824\ M\^120 - 
                    94302\ M\^122 + 1106759\ M\^124 - 301578\ M\^126 - 
                    251914\ M\^128 + 185683\ M\^130 - 2976\ M\^132 - 
                    59012\ M\^134 + 43705\ M\^136 - 19175\ M\^138 + 
                    5962\ M\^140 - 1346\ M\^142 + 212\ M\^144 - 21\ M\^146 + 
                    M\^148)\) + 
              L\^18\ \((\(-6\)\ M\^62 + 71\ M\^64 - 382\ M\^66 + 
                    1232\ M\^68 - 2916\ M\^70 + 6544\ M\^72 - 13853\ M\^74 + 
                    17807\ M\^76 + 5039\ M\^78 - 55097\ M\^80 + 
                    59967\ M\^82 + 25894\ M\^84 - 86999\ M\^86 + 
                    41632\ M\^88 - 25029\ M\^90 + 49584\ M\^92 + 
                    114825\ M\^94 - 320113\ M\^96 - 25835\ M\^98 + 
                    519234\ M\^100 - 96362\ M\^102 - 465081\ M\^104 + 
                    110665\ M\^106 + 400384\ M\^108 - 27430\ M\^110 - 
                    400195\ M\^112 + 54398\ M\^114 + 344624\ M\^116 - 
                    173427\ M\^118 - 164480\ M\^120 + 191544\ M\^122 - 
                    5846\ M\^124 - 96429\ M\^126 + 57342\ M\^128 + 
                    6189\ M\^130 - 24180\ M\^132 + 12390\ M\^134 - 
                    314\ M\^136 - 3897\ M\^138 + 3622\ M\^140 - 
                    2173\ M\^142 + 936\ M\^144 - 275\ M\^146 + 49\ M\^148 - 
                    4\ M\^150)\) + 
              L\^16\ \((8\ M\^44 - 122\ M\^46 + 831\ M\^48 - 3305\ M\^50 + 
                    8859\ M\^52 - 19568\ M\^54 + 42789\ M\^56 - 
                    74685\ M\^58 + 37238\ M\^60 + 170182\ M\^62 - 
                    330582\ M\^64 - 41945\ M\^66 + 582843\ M\^68 - 
                    92284\ M\^70 - 630334\ M\^72 - 768546\ M\^74 + 
                    1801238\ M\^76 + 2444292\ M\^78 - 5839950\ M\^80 - 
                    2443511\ M\^82 + 12209817\ M\^84 - 1964105\ M\^86 - 
                    17319436\ M\^88 + 9727462\ M\^90 + 17717284\ M\^92 - 
                    16508778\ M\^94 - 13086685\ M\^96 + 19270946\ M\^98 + 
                    6912826\ M\^100 - 17067554\ M\^102 - 1559940\ M\^104 + 
                    11968010\ M\^106 - 1783805\ M\^108 - 6627899\ M\^110 + 
                    2867795\ M\^112 + 2597367\ M\^114 - 2297512\ M\^116 - 
                    431281\ M\^118 + 1239911\ M\^120 - 295009\ M\^122 - 
                    446923\ M\^124 + 302526\ M\^126 + 102093\ M\^128 - 
                    184001\ M\^130 + 29209\ M\^132 + 62476\ M\^134 - 
                    35310\ M\^136 - 9024\ M\^138 + 19309\ M\^140 - 
                    10953\ M\^142 + 3506\ M\^144 - 687\ M\^146 + 78\ M\^148 - 
                    4\ M\^150)\) + 
              L\^15\ \((28\ M\^38 - 480\ M\^40 + 3609\ M\^42 - 15299\ M\^44 + 
                    38938\ M\^46 - 58761\ M\^48 + 53465\ M\^50 - 
                    47672\ M\^52 + 5792\ M\^54 + 353470\ M\^56 - 
                    857426\ M\^58 - 94271\ M\^60 + 2782199\ M\^62 - 
                    2178051\ M\^64 - 4433567\ M\^66 + 5947788\ M\^68 + 
                    5365502\ M\^70 - 8959573\ M\^72 - 8801254\ M\^74 + 
                    11163917\ M\^76 + 19785327\ M\^78 - 17941503\ M\^80 - 
                    37255603\ M\^82 + 34146019\ M\^84 + 51200707\ M\^86 - 
                    56245543\ M\^88 - 51069859\ M\^90 + 72643136\ M\^92 + 
                    37096464\ M\^94 - 71876404\ M\^96 - 16561410\ M\^98 + 
                    55842606\ M\^100 - 159027\ M\^102 - 34661450\ M\^104 + 
                    7801767\ M\^106 + 16956127\ M\^108 - 7943628\ M\^110 - 
                    6155848\ M\^112 + 5031804\ M\^114 + 1259062\ M\^116 - 
                    2182994\ M\^118 + 108899\ M\^120 + 609927\ M\^122 - 
                    163809\ M\^124 - 42999\ M\^126 - 34351\ M\^128 + 
                    22131\ M\^130 + 42222\ M\^132 - 34169\ M\^134 - 
                    9915\ M\^136 + 27126\ M\^138 - 18521\ M\^140 + 
                    7187\ M\^142 - 1770\ M\^144 + 276\ M\^146 - 25\ M\^148 + 
                    M\^150)\) + 
              L\^19\ \((15\ M\^74 - 154\ M\^76 + 709\ M\^78 - 1844\ M\^80 + 
                    2938\ M\^82 - 3369\ M\^84 + 3710\ M\^86 - 914\ M\^88 - 
                    14403\ M\^90 + 33686\ M\^92 - 11712\ M\^94 - 
                    58113\ M\^96 + 77941\ M\^98 + 10601\ M\^100 - 
                    111507\ M\^102 + 77025\ M\^104 + 76960\ M\^106 - 
                    107817\ M\^108 - 5656\ M\^110 + 82728\ M\^112 - 
                    29917\ M\^114 - 39839\ M\^116 + 32767\ M\^118 + 
                    9382\ M\^120 - 24081\ M\^122 + 4783\ M\^124 + 
                    13156\ M\^126 - 9585\ M\^128 - 1821\ M\^130 + 
                    5436\ M\^132 - 2245\ M\^134 - 889\ M\^136 + 
                    1559\ M\^138 - 974\ M\^140 + 401\ M\^142 - 134\ M\^144 + 
                    40\ M\^146 - 9\ M\^148 + M\^150)\) + 
              L\^17\ \((M\^50 - 13\ M\^52 + 78\ M\^54 - 310\ M\^56 + 
                    1123\ M\^58 - 4124\ M\^60 + 12212\ M\^62 - 23240\ M\^64 + 
                    22541\ M\^66 - 3499\ M\^68 - 8725\ M\^70 + 34998\ M\^72 - 
                    166844\ M\^74 + 220929\ M\^76 + 274377\ M\^78 - 
                    916316\ M\^80 + 254469\ M\^82 + 1433643\ M\^84 - 
                    1321379\ M\^86 - 1046497\ M\^88 + 1986748\ M\^90 - 
                    142982\ M\^92 - 1623232\ M\^94 + 1196781\ M\^96 + 
                    726318\ M\^98 - 1419323\ M\^100 + 139168\ M\^102 + 
                    1262746\ M\^104 - 519060\ M\^106 - 951888\ M\^108 + 
                    646495\ M\^110 + 546246\ M\^112 - 667484\ M\^114 - 
                    154771\ M\^116 + 540687\ M\^118 - 123063\ M\^120 - 
                    273754\ M\^122 + 189148\ M\^124 + 43044\ M\^126 - 
                    113037\ M\^128 + 55842\ M\^130 + 9539\ M\^132 - 
                    31810\ M\^134 + 17712\ M\^136 + 2476\ M\^138 - 
                    9825\ M\^140 + 6982\ M\^142 - 2760\ M\^144 + 
                    668\ M\^146 - 94\ M\^148 + 6\ M\^150)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 17]\), \(M\^52 + L\^20\ M\^52 + 
              L\ \((4\ M\^42 - 11\ M\^44 + 12\ M\^46 - 8\ M\^48 - 6\ M\^50 + 
                    38\ M\^52 - 6\ M\^54 - 8\ M\^56 + 12\ M\^58 - 11\ M\^60 + 
                    4\ M\^62)\) + 
              L\^19\ \((4\ M\^42 - 11\ M\^44 + 12\ M\^46 - 8\ M\^48 - 
                    6\ M\^50 + 38\ M\^52 - 6\ M\^54 - 8\ M\^56 + 12\ M\^58 - 
                    11\ M\^60 + 4\ M\^62)\) + 
              L\^2\ \((6\ M\^32 - 30\ M\^34 + 63\ M\^36 - 72\ M\^38 + 
                    24\ M\^40 + 97\ M\^42 - 122\ M\^44 - 27\ M\^46 + 
                    23\ M\^48 + 32\ M\^50 + 202\ M\^52 + 32\ M\^54 + 
                    23\ M\^56 - 27\ M\^58 - 122\ M\^60 + 97\ M\^62 + 
                    24\ M\^64 - 72\ M\^66 + 63\ M\^68 - 30\ M\^70 + 
                    6\ M\^72)\) + 
              L\^18\ \((6\ M\^32 - 30\ M\^34 + 63\ M\^36 - 72\ M\^38 + 
                    24\ M\^40 + 97\ M\^42 - 122\ M\^44 - 27\ M\^46 + 
                    23\ M\^48 + 32\ M\^50 + 202\ M\^52 + 32\ M\^54 + 
                    23\ M\^56 - 27\ M\^58 - 122\ M\^60 + 97\ M\^62 + 
                    24\ M\^64 - 72\ M\^66 + 63\ M\^68 - 30\ M\^70 + 
                    6\ M\^72)\) + 
              L\^3\ \((4\ M\^22 - 27\ M\^24 + 78\ M\^26 - 125\ M\^28 + 
                    108\ M\^30 - 45\ M\^32 + 190\ M\^34 - 576\ M\^36 + 
                    180\ M\^38 + 1487\ M\^40 - 2242\ M\^42 + 62\ M\^44 + 
                    4168\ M\^46 - 5423\ M\^48 - 1916\ M\^50 + 9294\ M\^52 - 
                    1916\ M\^54 - 5423\ M\^56 + 4168\ M\^58 + 62\ M\^60 - 
                    2242\ M\^62 + 1487\ M\^64 + 180\ M\^66 - 576\ M\^68 + 
                    190\ M\^70 - 45\ M\^72 + 108\ M\^74 - 125\ M\^76 + 
                    78\ M\^78 - 27\ M\^80 + 4\ M\^82)\) + 
              L\^17\ \((4\ M\^22 - 27\ M\^24 + 78\ M\^26 - 125\ M\^28 + 
                    108\ M\^30 - 45\ M\^32 + 190\ M\^34 - 576\ M\^36 + 
                    180\ M\^38 + 1487\ M\^40 - 2242\ M\^42 + 62\ M\^44 + 
                    4168\ M\^46 - 5423\ M\^48 - 1916\ M\^50 + 9294\ M\^52 - 
                    1916\ M\^54 - 5423\ M\^56 + 4168\ M\^58 + 62\ M\^60 - 
                    2242\ M\^62 + 1487\ M\^64 + 180\ M\^66 - 576\ M\^68 + 
                    190\ M\^70 - 45\ M\^72 + 108\ M\^74 - 125\ M\^76 + 
                    78\ M\^78 - 27\ M\^80 + 4\ M\^82)\) + 
              L\^4\ \((M\^12 - 8\ M\^14 + 28\ M\^16 - 56\ M\^18 + 78\ M\^20 - 
                    173\ M\^22 + 623\ M\^24 - 1390\ M\^26 + 1072\ M\^28 + 
                    1834\ M\^30 - 4574\ M\^32 + 749\ M\^34 + 8499\ M\^36 - 
                    8311\ M\^38 - 8896\ M\^40 + 17092\ M\^42 + 6509\ M\^44 - 
                    21216\ M\^46 - 2877\ M\^48 + 11479\ M\^50 + 3919\ M\^52 + 
                    11479\ M\^54 - 2877\ M\^56 - 21216\ M\^58 + 6509\ M\^60 + 
                    17092\ M\^62 - 8896\ M\^64 - 8311\ M\^66 + 8499\ M\^68 + 
                    749\ M\^70 - 4574\ M\^72 + 1834\ M\^74 + 1072\ M\^76 - 
                    1390\ M\^78 + 623\ M\^80 - 173\ M\^82 + 78\ M\^84 - 
                    56\ M\^86 + 28\ M\^88 - 8\ M\^90 + M\^92)\) + 
              L\^16\ \((M\^12 - 8\ M\^14 + 28\ M\^16 - 56\ M\^18 + 
                    78\ M\^20 - 173\ M\^22 + 623\ M\^24 - 1390\ M\^26 + 
                    1072\ M\^28 + 1834\ M\^30 - 4574\ M\^32 + 749\ M\^34 + 
                    8499\ M\^36 - 8311\ M\^38 - 8896\ M\^40 + 17092\ M\^42 + 
                    6509\ M\^44 - 21216\ M\^46 - 2877\ M\^48 + 11479\ M\^50 + 
                    3919\ M\^52 + 11479\ M\^54 - 2877\ M\^56 - 21216\ M\^58 + 
                    6509\ M\^60 + 17092\ M\^62 - 8896\ M\^64 - 8311\ M\^66 + 
                    8499\ M\^68 + 749\ M\^70 - 4574\ M\^72 + 1834\ M\^74 + 
                    1072\ M\^76 - 1390\ M\^78 + 623\ M\^80 - 173\ M\^82 + 
                    78\ M\^84 - 56\ M\^86 + 28\ M\^88 - 8\ M\^90 + M\^92)\) + 
              L\^5\ \((6\ M\^10 - 65\ M\^12 + 300\ M\^14 - 732\ M\^16 + 
                    844\ M\^18 + 18\ M\^20 - 326\ M\^22 - 3626\ M\^24 + 
                    8508\ M\^26 + 1763\ M\^28 - 27962\ M\^30 + 22779\ M\^32 + 
                    45010\ M\^34 - 80908\ M\^36 - 21608\ M\^38 + 
                    127453\ M\^40 - 37242\ M\^42 - 103930\ M\^44 + 
                    73246\ M\^46 + 13422\ M\^48 - 33024\ M\^50 + 
                    47652\ M\^52 - 33024\ M\^54 + 13422\ M\^56 + 
                    73246\ M\^58 - 103930\ M\^60 - 37242\ M\^62 + 
                    127453\ M\^64 - 21608\ M\^66 - 80908\ M\^68 + 
                    45010\ M\^70 + 22779\ M\^72 - 27962\ M\^74 + 
                    1763\ M\^76 + 8508\ M\^78 - 3626\ M\^80 - 326\ M\^82 + 
                    18\ M\^84 + 844\ M\^86 - 732\ M\^88 + 300\ M\^90 - 
                    65\ M\^92 + 6\ M\^94)\) + 
              L\^15\ \((6\ M\^10 - 65\ M\^12 + 300\ M\^14 - 732\ M\^16 + 
                    844\ M\^18 + 18\ M\^20 - 326\ M\^22 - 3626\ M\^24 + 
                    8508\ M\^26 + 1763\ M\^28 - 27962\ M\^30 + 22779\ M\^32 + 
                    45010\ M\^34 - 80908\ M\^36 - 21608\ M\^38 + 
                    127453\ M\^40 - 37242\ M\^42 - 103930\ M\^44 + 
                    73246\ M\^46 + 13422\ M\^48 - 33024\ M\^50 + 
                    47652\ M\^52 - 33024\ M\^54 + 13422\ M\^56 + 
                    73246\ M\^58 - 103930\ M\^60 - 37242\ M\^62 + 
                    127453\ M\^64 - 21608\ M\^66 - 80908\ M\^68 + 
                    45010\ M\^70 + 22779\ M\^72 - 27962\ M\^74 + 
                    1763\ M\^76 + 8508\ M\^78 - 3626\ M\^80 - 326\ M\^82 + 
                    18\ M\^84 + 844\ M\^86 - 732\ M\^88 + 300\ M\^90 - 
                    65\ M\^92 + 6\ M\^94)\) + 
              L\^6\ \((15\ M\^8 - 205\ M\^10 + 1176\ M\^12 - 3509\ M\^14 + 
                    4830\ M\^16 + 1496\ M\^18 - 13982\ M\^20 + 6765\ M\^22 + 
                    36811\ M\^24 - 49589\ M\^26 - 53470\ M\^28 + 
                    143024\ M\^30 + 27709\ M\^32 - 274950\ M\^34 + 
                    90780\ M\^36 + 345368\ M\^38 - 257724\ M\^40 - 
                    287533\ M\^42 + 390422\ M\^44 + 139036\ M\^46 - 
                    449309\ M\^48 - 19903\ M\^50 + 484244\ M\^52 - 
                    19903\ M\^54 - 449309\ M\^56 + 139036\ M\^58 + 
                    390422\ M\^60 - 287533\ M\^62 - 257724\ M\^64 + 
                    345368\ M\^66 + 90780\ M\^68 - 274950\ M\^70 + 
                    27709\ M\^72 + 143024\ M\^74 - 53470\ M\^76 - 
                    49589\ M\^78 + 36811\ M\^80 + 6765\ M\^82 - 
                    13982\ M\^84 + 1496\ M\^86 + 4830\ M\^88 - 3509\ M\^90 + 
                    1176\ M\^92 - 205\ M\^94 + 15\ M\^96)\) + 
              L\^14\ \((15\ M\^8 - 205\ M\^10 + 1176\ M\^12 - 3509\ M\^14 + 
                    4830\ M\^16 + 1496\ M\^18 - 13982\ M\^20 + 6765\ M\^22 + 
                    36811\ M\^24 - 49589\ M\^26 - 53470\ M\^28 + 
                    143024\ M\^30 + 27709\ M\^32 - 274950\ M\^34 + 
                    90780\ M\^36 + 345368\ M\^38 - 257724\ M\^40 - 
                    287533\ M\^42 + 390422\ M\^44 + 139036\ M\^46 - 
                    449309\ M\^48 - 19903\ M\^50 + 484244\ M\^52 - 
                    19903\ M\^54 - 449309\ M\^56 + 139036\ M\^58 + 
                    390422\ M\^60 - 287533\ M\^62 - 257724\ M\^64 + 
                    345368\ M\^66 + 90780\ M\^68 - 274950\ M\^70 + 
                    27709\ M\^72 + 143024\ M\^74 - 53470\ M\^76 - 
                    49589\ M\^78 + 36811\ M\^80 + 6765\ M\^82 - 
                    13982\ M\^84 + 1496\ M\^86 + 4830\ M\^88 - 3509\ M\^90 + 
                    1176\ M\^92 - 205\ M\^94 + 15\ M\^96)\) + 
              L\^7\ \((20\ M\^6 - 330\ M\^8 + 2270\ M\^10 - 8150\ M\^12 + 
                    14308\ M\^14 - 1087\ M\^16 - 42944\ M\^18 + 
                    54349\ M\^20 + 64060\ M\^22 - 198259\ M\^24 + 
                    23118\ M\^26 + 368084\ M\^28 - 267514\ M\^30 - 
                    424575\ M\^32 + 601086\ M\^34 + 215772\ M\^36 - 
                    773872\ M\^38 + 153929\ M\^40 + 666566\ M\^42 - 
                    430718\ M\^44 - 393924\ M\^46 + 462514\ M\^48 + 
                    145586\ M\^50 - 383058\ M\^52 + 145586\ M\^54 + 
                    462514\ M\^56 - 393924\ M\^58 - 430718\ M\^60 + 
                    666566\ M\^62 + 153929\ M\^64 - 773872\ M\^66 + 
                    215772\ M\^68 + 601086\ M\^70 - 424575\ M\^72 - 
                    267514\ M\^74 + 368084\ M\^76 + 23118\ M\^78 - 
                    198259\ M\^80 + 64060\ M\^82 + 54349\ M\^84 - 
                    42944\ M\^86 - 1087\ M\^88 + 14308\ M\^90 - 8150\ M\^92 + 
                    2270\ M\^94 - 330\ M\^96 + 20\ M\^98)\) + 
              L\^13\ \((20\ M\^6 - 330\ M\^8 + 2270\ M\^10 - 8150\ M\^12 + 
                    14308\ M\^14 - 1087\ M\^16 - 42944\ M\^18 + 
                    54349\ M\^20 + 64060\ M\^22 - 198259\ M\^24 + 
                    23118\ M\^26 + 368084\ M\^28 - 267514\ M\^30 - 
                    424575\ M\^32 + 601086\ M\^34 + 215772\ M\^36 - 
                    773872\ M\^38 + 153929\ M\^40 + 666566\ M\^42 - 
                    430718\ M\^44 - 393924\ M\^46 + 462514\ M\^48 + 
                    145586\ M\^50 - 383058\ M\^52 + 145586\ M\^54 + 
                    462514\ M\^56 - 393924\ M\^58 - 430718\ M\^60 + 
                    666566\ M\^62 + 153929\ M\^64 - 773872\ M\^66 + 
                    215772\ M\^68 + 601086\ M\^70 - 424575\ M\^72 - 
                    267514\ M\^74 + 368084\ M\^76 + 23118\ M\^78 - 
                    198259\ M\^80 + 64060\ M\^82 + 54349\ M\^84 - 
                    42944\ M\^86 - 1087\ M\^88 + 14308\ M\^90 - 8150\ M\^92 + 
                    2270\ M\^94 - 330\ M\^96 + 20\ M\^98)\) + 
              L\^8\ \((15\ M\^4 - 290\ M\^6 + 2340\ M\^8 - 10009\ M\^10 + 
                    22627\ M\^12 - 15886\ M\^14 - 45353\ M\^16 + 
                    108092\ M\^18 - 1677\ M\^20 - 261461\ M\^22 + 
                    246937\ M\^24 + 277756\ M\^26 - 605641\ M\^28 + 
                    9355\ M\^30 + 807741\ M\^32 - 599487\ M\^34 - 
                    497315\ M\^36 + 1139343\ M\^38 - 332727\ M\^40 - 
                    1282788\ M\^42 + 1391668\ M\^44 + 920652\ M\^46 - 
                    2265418\ M\^48 - 285277\ M\^50 + 2679576\ M\^52 - 
                    285277\ M\^54 - 2265418\ M\^56 + 920652\ M\^58 + 
                    1391668\ M\^60 - 1282788\ M\^62 - 332727\ M\^64 + 
                    1139343\ M\^66 - 497315\ M\^68 - 599487\ M\^70 + 
                    807741\ M\^72 + 9355\ M\^74 - 605641\ M\^76 + 
                    277756\ M\^78 + 246937\ M\^80 - 261461\ M\^82 - 
                    1677\ M\^84 + 108092\ M\^86 - 45353\ M\^88 - 
                    15886\ M\^90 + 22627\ M\^92 - 10009\ M\^94 + 
                    2340\ M\^96 - 290\ M\^98 + 15\ M\^100)\) + 
              L\^12\ \((15\ M\^4 - 290\ M\^6 + 2340\ M\^8 - 10009\ M\^10 + 
                    22627\ M\^12 - 15886\ M\^14 - 45353\ M\^16 + 
                    108092\ M\^18 - 1677\ M\^20 - 261461\ M\^22 + 
                    246937\ M\^24 + 277756\ M\^26 - 605641\ M\^28 + 
                    9355\ M\^30 + 807741\ M\^32 - 599487\ M\^34 - 
                    497315\ M\^36 + 1139343\ M\^38 - 332727\ M\^40 - 
                    1282788\ M\^42 + 1391668\ M\^44 + 920652\ M\^46 - 
                    2265418\ M\^48 - 285277\ M\^50 + 2679576\ M\^52 - 
                    285277\ M\^54 - 2265418\ M\^56 + 920652\ M\^58 + 
                    1391668\ M\^60 - 1282788\ M\^62 - 332727\ M\^64 + 
                    1139343\ M\^66 - 497315\ M\^68 - 599487\ M\^70 + 
                    807741\ M\^72 + 9355\ M\^74 - 605641\ M\^76 + 
                    277756\ M\^78 + 246937\ M\^80 - 261461\ M\^82 - 
                    1677\ M\^84 + 108092\ M\^86 - 45353\ M\^88 - 
                    15886\ M\^90 + 22627\ M\^92 - 10009\ M\^94 + 
                    2340\ M\^96 - 290\ M\^98 + 15\ M\^100)\) + 
              L\^9\ \((6\ M\^2 - 133\ M\^4 + 1248\ M\^6 - 6413\ M\^8 + 
                    19088\ M\^10 - 29328\ M\^12 + 1492\ M\^14 + 
                    81055\ M\^16 - 122190\ M\^18 - 38926\ M\^20 + 
                    329554\ M\^22 - 308515\ M\^24 - 290006\ M\^26 + 
                    844472\ M\^28 - 275474\ M\^30 - 1192055\ M\^32 + 
                    1317238\ M\^34 + 946939\ M\^36 - 2257934\ M\^38 - 
                    238378\ M\^40 + 2560392\ M\^42 - 481526\ M\^44 - 
                    2017812\ M\^46 + 816418\ M\^48 + 818378\ M\^50 - 
                    787220\ M\^52 + 818378\ M\^54 + 816418\ M\^56 - 
                    2017812\ M\^58 - 481526\ M\^60 + 2560392\ M\^62 - 
                    238378\ M\^64 - 2257934\ M\^66 + 946939\ M\^68 + 
                    1317238\ M\^70 - 1192055\ M\^72 - 275474\ M\^74 + 
                    844472\ M\^76 - 290006\ M\^78 - 308515\ M\^80 + 
                    329554\ M\^82 - 38926\ M\^84 - 122190\ M\^86 + 
                    81055\ M\^88 + 1492\ M\^90 - 29328\ M\^92 + 
                    19088\ M\^94 - 6413\ M\^96 + 1248\ M\^98 - 133\ M\^100 + 
                    6\ M\^102)\) + 
              L\^11\ \((6\ M\^2 - 133\ M\^4 + 1248\ M\^6 - 6413\ M\^8 + 
                    19088\ M\^10 - 29328\ M\^12 + 1492\ M\^14 + 
                    81055\ M\^16 - 122190\ M\^18 - 38926\ M\^20 + 
                    329554\ M\^22 - 308515\ M\^24 - 290006\ M\^26 + 
                    844472\ M\^28 - 275474\ M\^30 - 1192055\ M\^32 + 
                    1317238\ M\^34 + 946939\ M\^36 - 2257934\ M\^38 - 
                    238378\ M\^40 + 2560392\ M\^42 - 481526\ M\^44 - 
                    2017812\ M\^46 + 816418\ M\^48 + 818378\ M\^50 - 
                    787220\ M\^52 + 818378\ M\^54 + 816418\ M\^56 - 
                    2017812\ M\^58 - 481526\ M\^60 + 2560392\ M\^62 - 
                    238378\ M\^64 - 2257934\ M\^66 + 946939\ M\^68 + 
                    1317238\ M\^70 - 1192055\ M\^72 - 275474\ M\^74 + 
                    844472\ M\^76 - 290006\ M\^78 - 308515\ M\^80 + 
                    329554\ M\^82 - 38926\ M\^84 - 122190\ M\^86 + 
                    81055\ M\^88 + 1492\ M\^90 - 29328\ M\^92 + 
                    19088\ M\^94 - 6413\ M\^96 + 1248\ M\^98 - 133\ M\^100 + 
                    6\ M\^102)\) + 
              L\^10\ \((1 - 25\ M\^2 + 280\ M\^4 - 1854\ M\^6 + 7942\ M\^8 - 
                    21984\ M\^10 + 33781\ M\^12 - 1463\ M\^14 - 
                    104626\ M\^16 + 162775\ M\^18 + 72435\ M\^20 - 
                    484346\ M\^22 + 339699\ M\^24 + 619349\ M\^26 - 
                    1040969\ M\^28 - 249956\ M\^30 + 1559358\ M\^32 - 
                    626263\ M\^34 - 1232962\ M\^36 + 1467806\ M\^38 - 
                    108479\ M\^40 - 1755174\ M\^42 + 2087224\ M\^44 + 
                    1301264\ M\^46 - 3884905\ M\^48 - 410129\ M\^50 + 
                    4727198\ M\^52 - 410129\ M\^54 - 3884905\ M\^56 + 
                    1301264\ M\^58 + 2087224\ M\^60 - 1755174\ M\^62 - 
                    108479\ M\^64 + 1467806\ M\^66 - 1232962\ M\^68 - 
                    626263\ M\^70 + 1559358\ M\^72 - 249956\ M\^74 - 
                    1040969\ M\^76 + 619349\ M\^78 + 339699\ M\^80 - 
                    484346\ M\^82 + 72435\ M\^84 + 162775\ M\^86 - 
                    104626\ M\^88 - 1463\ M\^90 + 33781\ M\^92 - 
                    21984\ M\^94 + 7942\ M\^96 - 1854\ M\^98 + 280\ M\^100 - 
                    25\ M\^102 + M\^104)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 18]\), \(L\^27\ M\^32 + 
              M\^142 + 
              L\^26\ \((\(-3\)\ M\^24 + 12\ M\^26 - 22\ M\^28 + 21\ M\^30 - 
                    3\ M\^32 - 35\ M\^34 + 76\ M\^36 - M\^38 - 40\ M\^40 + 
                    30\ M\^42 - 8\ M\^44)\) + 
              L\^25\ \((3\ M\^16 - 22\ M\^18 + 68\ M\^20 - 118\ M\^22 + 
                    124\ M\^24 - 63\ M\^26 + 34\ M\^28 - 182\ M\^30 - 
                    40\ M\^32 + 1210\ M\^34 - 1689\ M\^36 - 715\ M\^38 + 
                    3491\ M\^40 - 1427\ M\^42 - 1813\ M\^44 + 2245\ M\^46 - 
                    466\ M\^48 - 724\ M\^50 + 611\ M\^52 - 204\ M\^54 + 
                    28\ M\^56)\) + 
              L\^24\ \((\(-M\^8\) + 10\ M\^10 - 45\ M\^12 + 128\ M\^14 - 
                    270\ M\^16 + 479\ M\^18 - 864\ M\^20 + 1442\ M\^22 - 
                    1062\ M\^24 - 2189\ M\^26 + 6275\ M\^28 - 3236\ M\^30 - 
                    7964\ M\^32 + 9043\ M\^34 + 8988\ M\^36 - 13357\ M\^38 - 
                    18464\ M\^40 + 31495\ M\^42 + 18068\ M\^44 - 
                    51917\ M\^46 + 13734\ M\^48 + 44412\ M\^50 - 
                    38866\ M\^52 - 8377\ M\^54 + 27105\ M\^56 - 
                    11620\ M\^58 - 3876\ M\^60 + 6018\ M\^62 - 2702\ M\^64 + 
                    594\ M\^66 - 56\ M\^68)\) + 
              L\^23\ \((2\ M\^4 - 23\ M\^6 + 119\ M\^8 - 389\ M\^10 + 
                    956\ M\^12 - 1869\ M\^14 + 2765\ M\^16 - 2857\ M\^18 + 
                    2132\ M\^20 - 1768\ M\^22 - 1112\ M\^24 + 18309\ M\^26 - 
                    45820\ M\^28 + 21157\ M\^30 + 108281\ M\^32 - 
                    192313\ M\^34 - 26077\ M\^36 + 374173\ M\^38 - 
                    252355\ M\^40 - 350126\ M\^42 + 520709\ M\^44 + 
                    114988\ M\^46 - 577292\ M\^48 + 144281\ M\^50 + 
                    495025\ M\^52 - 291374\ M\^54 - 325109\ M\^56 + 
                    356602\ M\^58 + 98157\ M\^60 - 293246\ M\^62 + 
                    81172\ M\^64 + 116663\ M\^66 - 97434\ M\^68 + 
                    7090\ M\^70 + 27603\ M\^72 - 18338\ M\^74 + 5758\ M\^76 - 
                    960\ M\^78 + 70\ M\^80)\) + 
              L\^22\ \((\(-1\) + 12\ M\^2 - 74\ M\^4 + 344\ M\^6 - 
                    1352\ M\^8 + 4225\ M\^10 - 9476\ M\^12 + 13218\ M\^14 - 
                    5323\ M\^16 - 19986\ M\^18 + 40081\ M\^20 - 368\ M\^22 - 
                    105490\ M\^24 + 109387\ M\^26 + 198931\ M\^28 - 
                    509274\ M\^30 - 73527\ M\^32 + 1407090\ M\^34 - 
                    1113453\ M\^36 - 2006326\ M\^38 + 3696306\ M\^40 + 
                    558817\ M\^42 - 5923390\ M\^44 + 3300912\ M\^46 + 
                    4993863\ M\^48 - 6872442\ M\^50 - 772278\ M\^52 + 
                    7066845\ M\^54 - 3274347\ M\^56 - 3965420\ M\^58 + 
                    4433454\ M\^60 + 772663\ M\^62 - 3185636\ M\^64 + 
                    616289\ M\^66 + 1611189\ M\^68 - 761413\ M\^70 - 
                    644843\ M\^72 + 595471\ M\^74 + 95151\ M\^76 - 
                    308261\ M\^78 + 109956\ M\^80 + 50141\ M\^82 - 
                    62851\ M\^84 + 27876\ M\^86 - 6834\ M\^88 + 930\ M\^90 - 
                    56\ M\^92)\) + 
              L\^21\ \((4 - 67\ M\^2 + 519\ M\^4 - 2501\ M\^6 + 8512\ M\^8 - 
                    21631\ M\^10 + 39809\ M\^12 - 41126\ M\^14 - 
                    22765\ M\^16 + 165893\ M\^18 - 236333\ M\^20 - 
                    66446\ M\^22 + 759159\ M\^24 - 1005178\ M\^26 - 
                    487301\ M\^28 + 3159189\ M\^30 - 2658072\ M\^32 - 
                    4388835\ M\^34 + 10135067\ M\^36 - 203929\ M\^38 - 
                    18561455\ M\^40 + 13933704\ M\^42 + 19023163\ M\^44 - 
                    31486317\ M\^46 - 4711951\ M\^48 + 38984720\ M\^50 - 
                    18938566\ M\^52 - 26671467\ M\^54 + 35878285\ M\^56 + 
                    643389\ M\^58 - 33590320\ M\^60 + 22099577\ M\^62 + 
                    16473570\ M\^64 - 27880749\ M\^66 + 1146348\ M\^68 + 
                    18844733\ M\^70 - 8617328\ M\^72 - 7117975\ M\^74 + 
                    6868587\ M\^76 + 885539\ M\^78 - 2962063\ M\^80 + 
                    362092\ M\^82 + 990252\ M\^84 - 307138\ M\^86 - 
                    338915\ M\^88 + 247753\ M\^90 + 24942\ M\^92 - 
                    109746\ M\^94 + 68183\ M\^96 - 22944\ M\^98 + 
                    4651\ M\^100 - 540\ M\^102 + 28\ M\^104)\) + 
              L\^20\ \((\(-6\) + 130\ M\^2 - 1269\ M\^4 + 7358\ M\^6 - 
                    27969\ M\^8 + 71529\ M\^10 - 115666\ M\^12 + 
                    71798\ M\^14 + 174751\ M\^16 - 555993\ M\^18 + 
                    554445\ M\^20 + 555114\ M\^22 - 2432590\ M\^24 + 
                    2548619\ M\^26 + 1902326\ M\^28 - 8705811\ M\^30 + 
                    7816353\ M\^32 + 9489008\ M\^34 - 29055001\ M\^36 + 
                    11805199\ M\^38 + 47058382\ M\^40 - 62807258\ M\^42 - 
                    33289400\ M\^44 + 121291448\ M\^46 - 26735663\ M\^48 - 
                    139759973\ M\^50 + 103751237\ M\^52 + 88351727\ M\^54 - 
                    138870390\ M\^56 + 6300151\ M\^58 + 98437434\ M\^60 - 
                    78633644\ M\^62 - 9797873\ M\^64 + 88158123\ M\^66 - 
                    62576458\ M\^68 - 46194205\ M\^70 + 82568626\ M\^72 - 
                    3107482\ M\^74 - 60256252\ M\^76 + 27815779\ M\^78 + 
                    26212809\ M\^80 - 25923719\ M\^82 - 3480691\ M\^84 + 
                    13080146\ M\^86 - 3425170\ M\^88 - 3262413\ M\^90 + 
                    2014720\ M\^92 + 230734\ M\^94 - 390007\ M\^96 - 
                    92374\ M\^98 + 131027\ M\^100 + 53727\ M\^102 - 
                    128863\ M\^104 + 90214\ M\^106 - 37124\ M\^108 + 
                    9924\ M\^110 - 1710\ M\^112 + 174\ M\^114 - 
                    8\ M\^116)\) + 
              L\^19\ \((4 - 108\ M\^2 + 1303\ M\^4 - 9274\ M\^6 + 
                    43018\ M\^8 - 133938\ M\^10 + 268529\ M\^12 - 
                    261007\ M\^14 - 248566\ M\^16 + 1303814\ M\^18 - 
                    1786298\ M\^20 - 395919\ M\^22 + 5480161\ M\^24 - 
                    8186072\ M\^26 + 5886\ M\^28 + 18820650\ M\^30 - 
                    28169128\ M\^32 - 1702830\ M\^34 + 66432217\ M\^36 - 
                    82142766\ M\^38 - 49183685\ M\^40 + 222005248\ M\^42 - 
                    105693102\ M\^44 - 317927766\ M\^46 + 399161269\ M\^48 + 
                    234719999\ M\^50 - 684292364\ M\^52 + 58150254\ M\^54 + 
                    748105661\ M\^56 - 407958013\ M\^58 - 502118214\ M\^60 + 
                    577782682\ M\^62 + 87459340\ M\^64 - 454876471\ M\^66 + 
                    239602762\ M\^68 + 152660623\ M\^70 - 316601719\ M\^72 + 
                    120139505\ M\^74 + 196890112\ M\^76 - 230473273\ M\^78 - 
                    39818991\ M\^80 + 193747426\ M\^82 - 48524919\ M\^84 - 
                    105113210\ M\^86 + 62350463\ M\^88 + 35516161\ M\^90 - 
                    41227441\ M\^92 - 2515516\ M\^94 + 17628914\ M\^96 - 
                    5379984\ M\^98 - 3653273\ M\^100 + 2827268\ M\^102 - 
                    251915\ M\^104 - 329131\ M\^106 + 105194\ M\^108 - 
                    84876\ M\^110 + 162228\ M\^112 - 149480\ M\^114 + 
                    83324\ M\^116 - 32110\ M\^118 + 9036\ M\^120 - 
                    1862\ M\^122 + 268\ M\^124 - 24\ M\^126 + M\^128)\) + 
              L\^18\ \((\(-1\) + 34\ M\^2 - 517\ M\^4 + 4687\ M\^6 - 
                    28286\ M\^8 + 119094\ M\^10 - 350503\ M\^12 + 
                    670299\ M\^14 - 537452\ M\^16 - 1137108\ M\^18 + 
                    4390514\ M\^20 - 4970067\ M\^22 - 4714669\ M\^24 + 
                    22530037\ M\^26 - 22467456\ M\^28 - 23554347\ M\^30 + 
                    83688020\ M\^32 - 55549561\ M\^34 - 105199135\ M\^36 + 
                    235680466\ M\^38 - 69241214\ M\^40 - 365999659\ M\^42 + 
                    513002800\ M\^44 + 147224614\ M\^46 - 1009189847\ M\^48 + 
                    588128392\ M\^50 + 1087113135\ M\^52 - 
                    1549875657\ M\^54 - 432459134\ M\^56 + 
                    2060817440\ M\^58 - 692356551\ M\^60 - 
                    1608287673\ M\^62 + 1530415861\ M\^64 + 
                    372605781\ M\^66 - 1458884537\ M\^68 + 864959603\ M\^70 + 
                    542675142\ M\^72 - 1374799645\ M\^74 + 574072644\ M\^76 + 
                    1072034318\ M\^78 - 1217900779\ M\^80 - 
                    404828235\ M\^82 + 1194347535\ M\^84 - 127822516\ M\^86 - 
                    779525525\ M\^88 + 322919331\ M\^90 + 349042573\ M\^92 - 
                    269242008\ M\^94 - 94253652\ M\^96 + 146087309\ M\^98 + 
                    1594015\ M\^100 - 57742428\ M\^102 + 13486023\ M\^104 + 
                    17413977\ M\^106 - 9523723\ M\^108 - 2647921\ M\^110 + 
                    3769794\ M\^112 - 785105\ M\^114 - 555631\ M\^116 + 
                    497282\ M\^118 - 280638\ M\^120 + 180203\ M\^122 - 
                    111044\ M\^124 + 51933\ M\^126 - 17179\ M\^128 + 
                    3904\ M\^130 - 581\ M\^132 + 51\ M\^134 - 2\ M\^136)\) + 
              L\^17\ \((\(-M\^2\) + 36\ M\^4 - 598\ M\^6 + 6021\ M\^8 - 
                    40627\ M\^10 + 191041\ M\^12 - 623207\ M\^14 + 
                    1306112\ M\^16 - 1154479\ M\^18 - 2326850\ M\^20 + 
                    9396381\ M\^22 - 10281032\ M\^24 - 11943456\ M\^26 + 
                    49654956\ M\^28 - 42006818\ M\^30 - 64183100\ M\^32 + 
                    178072540\ M\^34 - 82000062\ M\^36 - 259225822\ M\^38 + 
                    458656216\ M\^40 - 45342797\ M\^42 - 788428180\ M\^44 + 
                    940069823\ M\^46 + 450358068\ M\^48 - 2050317762\ M\^50 + 
                    968101422\ M\^52 + 2439238106\ M\^54 - 
                    2963147845\ M\^56 - 1418754046\ M\^58 + 
                    4169373425\ M\^60 - 539006308\ M\^62 - 
                    3442011665\ M\^64 + 1879262556\ M\^66 + 
                    1029130445\ M\^68 - 1241874436\ M\^70 + 
                    1458691928\ M\^72 - 1259318678\ M\^74 - 
                    2385354965\ M\^76 + 4069176987\ M\^78 + 
                    1453617056\ M\^80 - 5449338340\ M\^82 + 
                    339502745\ M\^84 + 4874594627\ M\^86 - 
                    1707334285\ M\^88 - 3143171537\ M\^90 + 
                    2041702719\ M\^92 + 1413707335\ M\^94 - 
                    1575546669\ M\^96 - 339143321\ M\^98 + 
                    882405181\ M\^100 - 76402558\ M\^102 - 
                    358781288\ M\^104 + 124067136\ M\^106 + 
                    96562296\ M\^108 - 65375555\ M\^110 - 10963040\ M\^112 + 
                    18770655\ M\^114 - 1248079\ M\^116 - 3525947\ M\^118 + 
                    425532\ M\^120 + 950565\ M\^122 - 257186\ M\^124 - 
                    398991\ M\^126 + 471112\ M\^128 - 293246\ M\^130 + 
                    131640\ M\^132 - 45730\ M\^134 + 12250\ M\^136 - 
                    2424\ M\^138 + 329\ M\^140 - 27\ M\^142 + M\^144)\) + 
              L\^16\ \((7\ M\^6 - 222\ M\^8 + 3220\ M\^10 - 28177\ M\^12 + 
                    164348\ M\^14 - 659050\ M\^16 + 1759982\ M\^18 - 
                    2565525\ M\^20 - 870037\ M\^22 + 12923379\ M\^24 - 
                    23509335\ M\^26 - 3979939\ M\^28 + 86278722\ M\^30 - 
                    115476550\ M\^32 - 92176263\ M\^34 + 410751251\ M\^36 - 
                    217142568\ M\^38 - 706279982\ M\^40 + 1080281368\ M\^42 + 
                    471729750\ M\^44 - 2294277377\ M\^46 + 821616841\ M\^48 + 
                    3073039084\ M\^50 - 3064872799\ M\^52 - 
                    2607253725\ M\^54 + 5326282134\ M\^56 + 
                    992188991\ M\^58 - 6661960018\ M\^60 + 600424505\ M\^62 + 
                    7223180637\ M\^64 - 995715531\ M\^66 - 
                    8321780412\ M\^68 + 360749440\ M\^70 + 
                    11029840107\ M\^72 - 214281297\ M\^74 - 
                    14697812475\ M\^76 + 1892206481\ M\^78 + 
                    17088542657\ M\^80 - 5078201249\ M\^82 - 
                    16187561811\ M\^84 + 8079779910\ M\^86 + 
                    12094610113\ M\^88 - 9180615990\ M\^90 - 
                    6812298993\ M\^92 + 7958295827\ M\^94 + 
                    2516591636\ M\^96 - 5386207626\ M\^98 - 
                    176557316\ M\^100 + 2848178613\ M\^102 - 
                    534316784\ M\^104 - 1154676113\ M\^106 + 
                    453308900\ M\^108 + 345127197\ M\^110 - 
                    222085001\ M\^112 - 68143211\ M\^114 + 77147110\ M\^116 + 
                    4067990\ M\^118 - 20370004\ M\^120 + 4593868\ M\^122 + 
                    2001888\ M\^124 - 1440380\ M\^126 + 938653\ M\^128 - 
                    512882\ M\^130 - 250487\ M\^132 + 652250\ M\^134 - 
                    527233\ M\^136 + 252472\ M\^138 - 79916\ M\^140 + 
                    17012\ M\^142 - 2355\ M\^144 + 192\ M\^146 - 
                    7\ M\^148)\) + 
              L\ \((\(-8\)\ M\^130 + 30\ M\^132 - 40\ M\^134 - M\^136 + 
                    76\ M\^138 - 35\ M\^140 - 3\ M\^142 + 21\ M\^144 - 
                    22\ M\^146 + 12\ M\^148 - 3\ M\^150)\) + 
              L\^15\ \((\(-21\)\ M\^10 + 632\ M\^12 - 8639\ M\^14 + 
                    70524\ M\^16 - 377837\ M\^18 + 1355005\ M\^20 - 
                    3041878\ M\^22 + 2732534\ M\^24 + 6652547\ M\^26 - 
                    26532495\ M\^28 + 26491140\ M\^30 + 50829674\ M\^32 - 
                    172667490\ M\^34 + 88208385\ M\^36 + 395303325\ M\^38 - 
                    705951361\ M\^40 - 276826336\ M\^42 + 1983355523\ M\^44 - 
                    1092448401\ M\^46 - 3403876019\ M\^48 + 
                    4627830857\ M\^50 + 3527221665\ M\^52 - 
                    10337798781\ M\^54 - 662647851\ M\^56 + 
                    16816255590\ M\^58 - 5773219921\ M\^60 - 
                    21983977014\ M\^62 + 14494785861\ M\^64 + 
                    24598866328\ M\^66 - 23066827073\ M\^68 - 
                    25060789666\ M\^70 + 29514470803\ M\^72 + 
                    24562175838\ M\^74 - 33214310334\ M\^76 - 
                    23538743467\ M\^78 + 34434427039\ M\^80 + 
                    21195593699\ M\^82 - 33435266358\ M\^84 - 
                    16708225274\ M\^86 + 30263791540\ M\^88 + 
                    10730416647\ M\^90 - 24986711154\ M\^92 - 
                    5021767796\ M\^94 + 18344742391\ M\^96 + 
                    1068502393\ M\^98 - 11749853448\ M\^100 + 
                    781303155\ M\^102 + 6507688916\ M\^104 - 
                    1141596546\ M\^106 - 3119344226\ M\^108 + 
                    857464253\ M\^110 + 1307579106\ M\^112 - 
                    491080296\ M\^114 - 484556728\ M\^116 + 
                    237966895\ M\^118 + 161377814\ M\^120 - 
                    105302070\ M\^122 - 47397188\ M\^124 + 44697044\ M\^126 + 
                    10873188\ M\^128 - 19620415\ M\^130 + 1756878\ M\^132 + 
                    6265070\ M\^134 - 3058340\ M\^136 - 675901\ M\^138 + 
                    1406171\ M\^140 - 782356\ M\^142 + 255348\ M\^144 - 
                    54011\ M\^146 + 7338\ M\^148 - 586\ M\^150 + 
                    21\ M\^152)\) + 
              L\^14\ \((35\ M\^14 - 1020\ M\^16 + 13447\ M\^18 - 
                    105095\ M\^20 + 532298\ M\^22 - 1762476\ M\^24 + 
                    3435171\ M\^26 - 1542474\ M\^28 - 11221451\ M\^30 + 
                    29475711\ M\^32 - 11115445\ M\^34 - 86691273\ M\^36 + 
                    165417242\ M\^38 + 44681741\ M\^40 - 528614189\ M\^42 + 
                    457832237\ M\^44 + 895612573\ M\^46 - 1925326598\ M\^48 - 
                    459704229\ M\^50 + 4533537337\ M\^52 - 
                    2117119243\ M\^54 - 7578373478\ M\^56 + 
                    7998695170\ M\^58 + 9445917414\ M\^60 - 
                    17105570518\ M\^62 - 8572444560\ M\^64 + 
                    27635139787\ M\^66 + 4790060896\ M\^68 - 
                    36741689569\ M\^70 + 175754811\ M\^72 + 
                    42018776573\ M\^74 - 3811944366\ M\^76 - 
                    42663275167\ M\^78 + 4635310787\ M\^80 + 
                    39632070385\ M\^82 - 3059974837\ M\^84 - 
                    34850346915\ M\^86 + 594976239\ M\^88 + 
                    30007120232\ M\^90 + 1489879874\ M\^92 - 
                    25576873508\ M\^94 - 2421093559\ M\^96 + 
                    21263994758\ M\^98 + 1995587836\ M\^100 - 
                    16765949289\ M\^102 - 636121528\ M\^104 + 
                    12151688896\ M\^106 - 846515736\ M\^108 - 
                    7822219780\ M\^110 + 1741158016\ M\^112 + 
                    4283742620\ M\^114 - 1831391990\ M\^116 - 
                    1856359514\ M\^118 + 1367153879\ M\^120 + 
                    527248469\ M\^122 - 766562197\ M\^124 + 3047843\ M\^126 + 
                    311154828\ M\^128 - 105728694\ M\^130 - 
                    76329895\ M\^132 + 65838366\ M\^134 - 2173686\ M\^136 - 
                    18744794\ M\^138 + 8989519\ M\^140 + 644929\ M\^142 - 
                    2633685\ M\^144 + 1474203\ M\^146 - 469863\ M\^148 + 
                    96612\ M\^150 - 12774\ M\^152 + 996\ M\^154 - 
                    35\ M\^156)\) + 
              L\^2\ \((28\ M\^118 - 204\ M\^120 + 611\ M\^122 - 724\ M\^124 - 
                    466\ M\^126 + 2245\ M\^128 - 1813\ M\^130 - 
                    1427\ M\^132 + 3491\ M\^134 - 715\ M\^136 - 
                    1689\ M\^138 + 1210\ M\^140 - 40\ M\^142 - 182\ M\^144 + 
                    34\ M\^146 - 63\ M\^148 + 124\ M\^150 - 118\ M\^152 + 
                    68\ M\^154 - 22\ M\^156 + 3\ M\^158)\) + 
              L\^13\ \((\(-35\)\ M\^18 + 996\ M\^20 - 12774\ M\^22 + 
                    96612\ M\^24 - 469863\ M\^26 + 1474203\ M\^28 - 
                    2633685\ M\^30 + 644929\ M\^32 + 8989519\ M\^34 - 
                    18744794\ M\^36 - 2173686\ M\^38 + 65838366\ M\^40 - 
                    76329895\ M\^42 - 105728694\ M\^44 + 311154828\ M\^46 + 
                    3047843\ M\^48 - 766562197\ M\^50 + 527248469\ M\^52 + 
                    1367153879\ M\^54 - 1856359514\ M\^56 - 
                    1831391990\ M\^58 + 4283742620\ M\^60 + 
                    1741158016\ M\^62 - 7822219780\ M\^64 - 
                    846515736\ M\^66 + 12151688896\ M\^68 - 
                    636121528\ M\^70 - 16765949289\ M\^72 + 
                    1995587836\ M\^74 + 21263994758\ M\^76 - 
                    2421093559\ M\^78 - 25576873508\ M\^80 + 
                    1489879874\ M\^82 + 30007120232\ M\^84 + 
                    594976239\ M\^86 - 34850346915\ M\^88 - 
                    3059974837\ M\^90 + 39632070385\ M\^92 + 
                    4635310787\ M\^94 - 42663275167\ M\^96 - 
                    3811944366\ M\^98 + 42018776573\ M\^100 + 
                    175754811\ M\^102 - 36741689569\ M\^104 + 
                    4790060896\ M\^106 + 27635139787\ M\^108 - 
                    8572444560\ M\^110 - 17105570518\ M\^112 + 
                    9445917414\ M\^114 + 7998695170\ M\^116 - 
                    7578373478\ M\^118 - 2117119243\ M\^120 + 
                    4533537337\ M\^122 - 459704229\ M\^124 - 
                    1925326598\ M\^126 + 895612573\ M\^128 + 
                    457832237\ M\^130 - 528614189\ M\^132 + 
                    44681741\ M\^134 + 165417242\ M\^136 - 86691273\ M\^138 - 
                    11115445\ M\^140 + 29475711\ M\^142 - 11221451\ M\^144 - 
                    1542474\ M\^146 + 3435171\ M\^148 - 1762476\ M\^150 + 
                    532298\ M\^152 - 105095\ M\^154 + 13447\ M\^156 - 
                    1020\ M\^158 + 35\ M\^160)\) + 
              L\^12\ \((21\ M\^22 - 586\ M\^24 + 7338\ M\^26 - 54011\ M\^28 + 
                    255348\ M\^30 - 782356\ M\^32 + 1406171\ M\^34 - 
                    675901\ M\^36 - 3058340\ M\^38 + 6265070\ M\^40 + 
                    1756878\ M\^42 - 19620415\ M\^44 + 10873188\ M\^46 + 
                    44697044\ M\^48 - 47397188\ M\^50 - 105302070\ M\^52 + 
                    161377814\ M\^54 + 237966895\ M\^56 - 484556728\ M\^58 - 
                    491080296\ M\^60 + 1307579106\ M\^62 + 857464253\ M\^64 - 
                    3119344226\ M\^66 - 1141596546\ M\^68 + 
                    6507688916\ M\^70 + 781303155\ M\^72 - 
                    11749853448\ M\^74 + 1068502393\ M\^76 + 
                    18344742391\ M\^78 - 5021767796\ M\^80 - 
                    24986711154\ M\^82 + 10730416647\ M\^84 + 
                    30263791540\ M\^86 - 16708225274\ M\^88 - 
                    33435266358\ M\^90 + 21195593699\ M\^92 + 
                    34434427039\ M\^94 - 23538743467\ M\^96 - 
                    33214310334\ M\^98 + 24562175838\ M\^100 + 
                    29514470803\ M\^102 - 25060789666\ M\^104 - 
                    23066827073\ M\^106 + 24598866328\ M\^108 + 
                    14494785861\ M\^110 - 21983977014\ M\^112 - 
                    5773219921\ M\^114 + 16816255590\ M\^116 - 
                    662647851\ M\^118 - 10337798781\ M\^120 + 
                    3527221665\ M\^122 + 4627830857\ M\^124 - 
                    3403876019\ M\^126 - 1092448401\ M\^128 + 
                    1983355523\ M\^130 - 276826336\ M\^132 - 
                    705951361\ M\^134 + 395303325\ M\^136 + 
                    88208385\ M\^138 - 172667490\ M\^140 + 50829674\ M\^142 + 
                    26491140\ M\^144 - 26532495\ M\^146 + 6652547\ M\^148 + 
                    2732534\ M\^150 - 3041878\ M\^152 + 1355005\ M\^154 - 
                    377837\ M\^156 + 70524\ M\^158 - 8639\ M\^160 + 
                    632\ M\^162 - 21\ M\^164)\) + 
              L\^3\ \((\(-56\)\ M\^106 + 594\ M\^108 - 2702\ M\^110 + 
                    6018\ M\^112 - 3876\ M\^114 - 11620\ M\^116 + 
                    27105\ M\^118 - 8377\ M\^120 - 38866\ M\^122 + 
                    44412\ M\^124 + 13734\ M\^126 - 51917\ M\^128 + 
                    18068\ M\^130 + 31495\ M\^132 - 18464\ M\^134 - 
                    13357\ M\^136 + 8988\ M\^138 + 9043\ M\^140 - 
                    7964\ M\^142 - 3236\ M\^144 + 6275\ M\^146 - 
                    2189\ M\^148 - 1062\ M\^150 + 1442\ M\^152 - 
                    864\ M\^154 + 479\ M\^156 - 270\ M\^158 + 128\ M\^160 - 
                    45\ M\^162 + 10\ M\^164 - M\^166)\) + 
              L\^11\ \((\(-7\)\ M\^26 + 192\ M\^28 - 2355\ M\^30 + 
                    17012\ M\^32 - 79916\ M\^34 + 252472\ M\^36 - 
                    527233\ M\^38 + 652250\ M\^40 - 250487\ M\^42 - 
                    512882\ M\^44 + 938653\ M\^46 - 1440380\ M\^48 + 
                    2001888\ M\^50 + 4593868\ M\^52 - 20370004\ M\^54 + 
                    4067990\ M\^56 + 77147110\ M\^58 - 68143211\ M\^60 - 
                    222085001\ M\^62 + 345127197\ M\^64 + 453308900\ M\^66 - 
                    1154676113\ M\^68 - 534316784\ M\^70 + 
                    2848178613\ M\^72 - 176557316\ M\^74 - 
                    5386207626\ M\^76 + 2516591636\ M\^78 + 
                    7958295827\ M\^80 - 6812298993\ M\^82 - 
                    9180615990\ M\^84 + 12094610113\ M\^86 + 
                    8079779910\ M\^88 - 16187561811\ M\^90 - 
                    5078201249\ M\^92 + 17088542657\ M\^94 + 
                    1892206481\ M\^96 - 14697812475\ M\^98 - 
                    214281297\ M\^100 + 11029840107\ M\^102 + 
                    360749440\ M\^104 - 8321780412\ M\^106 - 
                    995715531\ M\^108 + 7223180637\ M\^110 + 
                    600424505\ M\^112 - 6661960018\ M\^114 + 
                    992188991\ M\^116 + 5326282134\ M\^118 - 
                    2607253725\ M\^120 - 3064872799\ M\^122 + 
                    3073039084\ M\^124 + 821616841\ M\^126 - 
                    2294277377\ M\^128 + 471729750\ M\^130 + 
                    1080281368\ M\^132 - 706279982\ M\^134 - 
                    217142568\ M\^136 + 410751251\ M\^138 - 
                    92176263\ M\^140 - 115476550\ M\^142 + 86278722\ M\^144 - 
                    3979939\ M\^146 - 23509335\ M\^148 + 12923379\ M\^150 - 
                    870037\ M\^152 - 2565525\ M\^154 + 1759982\ M\^156 - 
                    659050\ M\^158 + 164348\ M\^160 - 28177\ M\^162 + 
                    3220\ M\^164 - 222\ M\^166 + 7\ M\^168)\) + 
              L\^4\ \((70\ M\^94 - 960\ M\^96 + 5758\ M\^98 - 18338\ M\^100 + 
                    27603\ M\^102 + 7090\ M\^104 - 97434\ M\^106 + 
                    116663\ M\^108 + 81172\ M\^110 - 293246\ M\^112 + 
                    98157\ M\^114 + 356602\ M\^116 - 325109\ M\^118 - 
                    291374\ M\^120 + 495025\ M\^122 + 144281\ M\^124 - 
                    577292\ M\^126 + 114988\ M\^128 + 520709\ M\^130 - 
                    350126\ M\^132 - 252355\ M\^134 + 374173\ M\^136 - 
                    26077\ M\^138 - 192313\ M\^140 + 108281\ M\^142 + 
                    21157\ M\^144 - 45820\ M\^146 + 18309\ M\^148 - 
                    1112\ M\^150 - 1768\ M\^152 + 2132\ M\^154 - 
                    2857\ M\^156 + 2765\ M\^158 - 1869\ M\^160 + 
                    956\ M\^162 - 389\ M\^164 + 119\ M\^166 - 23\ M\^168 + 
                    2\ M\^170)\) + 
              L\^10\ \((M\^30 - 27\ M\^32 + 329\ M\^34 - 2424\ M\^36 + 
                    12250\ M\^38 - 45730\ M\^40 + 131640\ M\^42 - 
                    293246\ M\^44 + 471112\ M\^46 - 398991\ M\^48 - 
                    257186\ M\^50 + 950565\ M\^52 + 425532\ M\^54 - 
                    3525947\ M\^56 - 1248079\ M\^58 + 18770655\ M\^60 - 
                    10963040\ M\^62 - 65375555\ M\^64 + 96562296\ M\^66 + 
                    124067136\ M\^68 - 358781288\ M\^70 - 76402558\ M\^72 + 
                    882405181\ M\^74 - 339143321\ M\^76 - 1575546669\ M\^78 + 
                    1413707335\ M\^80 + 2041702719\ M\^82 - 
                    3143171537\ M\^84 - 1707334285\ M\^86 + 
                    4874594627\ M\^88 + 339502745\ M\^90 - 
                    5449338340\ M\^92 + 1453617056\ M\^94 + 
                    4069176987\ M\^96 - 2385354965\ M\^98 - 
                    1259318678\ M\^100 + 1458691928\ M\^102 - 
                    1241874436\ M\^104 + 1029130445\ M\^106 + 
                    1879262556\ M\^108 - 3442011665\ M\^110 - 
                    539006308\ M\^112 + 4169373425\ M\^114 - 
                    1418754046\ M\^116 - 2963147845\ M\^118 + 
                    2439238106\ M\^120 + 968101422\ M\^122 - 
                    2050317762\ M\^124 + 450358068\ M\^126 + 
                    940069823\ M\^128 - 788428180\ M\^130 - 
                    45342797\ M\^132 + 458656216\ M\^134 - 
                    259225822\ M\^136 - 82000062\ M\^138 + 
                    178072540\ M\^140 - 64183100\ M\^142 - 42006818\ M\^144 + 
                    49654956\ M\^146 - 11943456\ M\^148 - 10281032\ M\^150 + 
                    9396381\ M\^152 - 2326850\ M\^154 - 1154479\ M\^156 + 
                    1306112\ M\^158 - 623207\ M\^160 + 191041\ M\^162 - 
                    40627\ M\^164 + 6021\ M\^166 - 598\ M\^168 + 36\ M\^170 - 
                    M\^172)\) + 
              L\^7\ \((\(-8\)\ M\^58 + 174\ M\^60 - 1710\ M\^62 + 
                    9924\ M\^64 - 37124\ M\^66 + 90214\ M\^68 - 
                    128863\ M\^70 + 53727\ M\^72 + 131027\ M\^74 - 
                    92374\ M\^76 - 390007\ M\^78 + 230734\ M\^80 + 
                    2014720\ M\^82 - 3262413\ M\^84 - 3425170\ M\^86 + 
                    13080146\ M\^88 - 3480691\ M\^90 - 25923719\ M\^92 + 
                    26212809\ M\^94 + 27815779\ M\^96 - 60256252\ M\^98 - 
                    3107482\ M\^100 + 82568626\ M\^102 - 46194205\ M\^104 - 
                    62576458\ M\^106 + 88158123\ M\^108 - 9797873\ M\^110 - 
                    78633644\ M\^112 + 98437434\ M\^114 + 6300151\ M\^116 - 
                    138870390\ M\^118 + 88351727\ M\^120 + 
                    103751237\ M\^122 - 139759973\ M\^124 - 
                    26735663\ M\^126 + 121291448\ M\^128 - 33289400\ M\^130 - 
                    62807258\ M\^132 + 47058382\ M\^134 + 11805199\ M\^136 - 
                    29055001\ M\^138 + 9489008\ M\^140 + 7816353\ M\^142 - 
                    8705811\ M\^144 + 1902326\ M\^146 + 2548619\ M\^148 - 
                    2432590\ M\^150 + 555114\ M\^152 + 554445\ M\^154 - 
                    555993\ M\^156 + 174751\ M\^158 + 71798\ M\^160 - 
                    115666\ M\^162 + 71529\ M\^164 - 27969\ M\^166 + 
                    7358\ M\^168 - 1269\ M\^170 + 130\ M\^172 - 
                    6\ M\^174)\) + 
              L\^5\ \((\(-56\)\ M\^82 + 930\ M\^84 - 6834\ M\^86 + 
                    27876\ M\^88 - 62851\ M\^90 + 50141\ M\^92 + 
                    109956\ M\^94 - 308261\ M\^96 + 95151\ M\^98 + 
                    595471\ M\^100 - 644843\ M\^102 - 761413\ M\^104 + 
                    1611189\ M\^106 + 616289\ M\^108 - 3185636\ M\^110 + 
                    772663\ M\^112 + 4433454\ M\^114 - 3965420\ M\^116 - 
                    3274347\ M\^118 + 7066845\ M\^120 - 772278\ M\^122 - 
                    6872442\ M\^124 + 4993863\ M\^126 + 3300912\ M\^128 - 
                    5923390\ M\^130 + 558817\ M\^132 + 3696306\ M\^134 - 
                    2006326\ M\^136 - 1113453\ M\^138 + 1407090\ M\^140 - 
                    73527\ M\^142 - 509274\ M\^144 + 198931\ M\^146 + 
                    109387\ M\^148 - 105490\ M\^150 - 368\ M\^152 + 
                    40081\ M\^154 - 19986\ M\^156 - 5323\ M\^158 + 
                    13218\ M\^160 - 9476\ M\^162 + 4225\ M\^164 - 
                    1352\ M\^166 + 344\ M\^168 - 74\ M\^170 + 12\ M\^172 - 
                    M\^174)\) + 
              L\^9\ \((\(-2\)\ M\^38 + 51\ M\^40 - 581\ M\^42 + 3904\ M\^44 - 
                    17179\ M\^46 + 51933\ M\^48 - 111044\ M\^50 + 
                    180203\ M\^52 - 280638\ M\^54 + 497282\ M\^56 - 
                    555631\ M\^58 - 785105\ M\^60 + 3769794\ M\^62 - 
                    2647921\ M\^64 - 9523723\ M\^66 + 17413977\ M\^68 + 
                    13486023\ M\^70 - 57742428\ M\^72 + 1594015\ M\^74 + 
                    146087309\ M\^76 - 94253652\ M\^78 - 269242008\ M\^80 + 
                    349042573\ M\^82 + 322919331\ M\^84 - 779525525\ M\^86 - 
                    127822516\ M\^88 + 1194347535\ M\^90 - 404828235\ M\^92 - 
                    1217900779\ M\^94 + 1072034318\ M\^96 + 
                    574072644\ M\^98 - 1374799645\ M\^100 + 
                    542675142\ M\^102 + 864959603\ M\^104 - 
                    1458884537\ M\^106 + 372605781\ M\^108 + 
                    1530415861\ M\^110 - 1608287673\ M\^112 - 
                    692356551\ M\^114 + 2060817440\ M\^116 - 
                    432459134\ M\^118 - 1549875657\ M\^120 + 
                    1087113135\ M\^122 + 588128392\ M\^124 - 
                    1009189847\ M\^126 + 147224614\ M\^128 + 
                    513002800\ M\^130 - 365999659\ M\^132 - 
                    69241214\ M\^134 + 235680466\ M\^136 - 
                    105199135\ M\^138 - 55549561\ M\^140 + 83688020\ M\^142 - 
                    23554347\ M\^144 - 22467456\ M\^146 + 22530037\ M\^148 - 
                    4714669\ M\^150 - 4970067\ M\^152 + 4390514\ M\^154 - 
                    1137108\ M\^156 - 537452\ M\^158 + 670299\ M\^160 - 
                    350503\ M\^162 + 119094\ M\^164 - 28286\ M\^166 + 
                    4687\ M\^168 - 517\ M\^170 + 34\ M\^172 - M\^174)\) + 
              L\^8\ \((M\^46 - 24\ M\^48 + 268\ M\^50 - 1862\ M\^52 + 
                    9036\ M\^54 - 32110\ M\^56 + 83324\ M\^58 - 
                    149480\ M\^60 + 162228\ M\^62 - 84876\ M\^64 + 
                    105194\ M\^66 - 329131\ M\^68 - 251915\ M\^70 + 
                    2827268\ M\^72 - 3653273\ M\^74 - 5379984\ M\^76 + 
                    17628914\ M\^78 - 2515516\ M\^80 - 41227441\ M\^82 + 
                    35516161\ M\^84 + 62350463\ M\^86 - 105113210\ M\^88 - 
                    48524919\ M\^90 + 193747426\ M\^92 - 39818991\ M\^94 - 
                    230473273\ M\^96 + 196890112\ M\^98 + 120139505\ M\^100 - 
                    316601719\ M\^102 + 152660623\ M\^104 + 
                    239602762\ M\^106 - 454876471\ M\^108 + 
                    87459340\ M\^110 + 577782682\ M\^112 - 
                    502118214\ M\^114 - 407958013\ M\^116 + 
                    748105661\ M\^118 + 58150254\ M\^120 - 
                    684292364\ M\^122 + 234719999\ M\^124 + 
                    399161269\ M\^126 - 317927766\ M\^128 - 
                    105693102\ M\^130 + 222005248\ M\^132 - 
                    49183685\ M\^134 - 82142766\ M\^136 + 66432217\ M\^138 - 
                    1702830\ M\^140 - 28169128\ M\^142 + 18820650\ M\^144 + 
                    5886\ M\^146 - 8186072\ M\^148 + 5480161\ M\^150 - 
                    395919\ M\^152 - 1786298\ M\^154 + 1303814\ M\^156 - 
                    248566\ M\^158 - 261007\ M\^160 + 268529\ M\^162 - 
                    133938\ M\^164 + 43018\ M\^166 - 9274\ M\^168 + 
                    1303\ M\^170 - 108\ M\^172 + 4\ M\^174)\) + 
              L\^6\ \((28\ M\^70 - 540\ M\^72 + 4651\ M\^74 - 22944\ M\^76 + 
                    68183\ M\^78 - 109746\ M\^80 + 24942\ M\^82 + 
                    247753\ M\^84 - 338915\ M\^86 - 307138\ M\^88 + 
                    990252\ M\^90 + 362092\ M\^92 - 2962063\ M\^94 + 
                    885539\ M\^96 + 6868587\ M\^98 - 7117975\ M\^100 - 
                    8617328\ M\^102 + 18844733\ M\^104 + 1146348\ M\^106 - 
                    27880749\ M\^108 + 16473570\ M\^110 + 22099577\ M\^112 - 
                    33590320\ M\^114 + 643389\ M\^116 + 35878285\ M\^118 - 
                    26671467\ M\^120 - 18938566\ M\^122 + 38984720\ M\^124 - 
                    4711951\ M\^126 - 31486317\ M\^128 + 19023163\ M\^130 + 
                    13933704\ M\^132 - 18561455\ M\^134 - 203929\ M\^136 + 
                    10135067\ M\^138 - 4388835\ M\^140 - 2658072\ M\^142 + 
                    3159189\ M\^144 - 487301\ M\^146 - 1005178\ M\^148 + 
                    759159\ M\^150 - 66446\ M\^152 - 236333\ M\^154 + 
                    165893\ M\^156 - 22765\ M\^158 - 41126\ M\^160 + 
                    39809\ M\^162 - 21631\ M\^164 + 8512\ M\^166 - 
                    2501\ M\^168 + 519\ M\^170 - 67\ M\^172 + 
                    4\ M\^174)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 19]\), \(M\^34 + 
              L\^25\ M\^136 + 
              L\ \((2\ M\^24 - 7\ M\^26 + 14\ M\^28 - 19\ M\^30 + 9\ M\^32 + 
                    18\ M\^34 - 44\ M\^36 + 49\ M\^38 + 32\ M\^40 - 
                    41\ M\^42 + 12\ M\^44)\) + 
              L\^2\ \((M\^14 - 6\ M\^16 + 17\ M\^18 - 30\ M\^20 + 40\ M\^22 - 
                    66\ M\^24 + 151\ M\^26 - 271\ M\^28 + 171\ M\^30 + 
                    579\ M\^32 - 1465\ M\^34 + 788\ M\^36 + 1511\ M\^38 - 
                    2769\ M\^40 + 943\ M\^42 + 2433\ M\^44 - 2091\ M\^46 - 
                    224\ M\^48 + 956\ M\^50 - 434\ M\^52 + 66\ M\^54)\) + 
              L\^3\ \((\(-2\)\ M\^10 + 21\ M\^12 - 99\ M\^14 + 276\ M\^16 - 
                    481\ M\^18 + 408\ M\^20 + 258\ M\^22 - 816\ M\^24 - 
                    595\ M\^26 + 3207\ M\^28 - 758\ M\^30 - 9951\ M\^32 + 
                    15165\ M\^34 + 6217\ M\^36 - 39692\ M\^38 + 
                    30211\ M\^40 + 34542\ M\^42 - 73879\ M\^44 + 
                    16122\ M\^46 + 70511\ M\^48 - 54069\ M\^50 - 
                    20802\ M\^52 + 40656\ M\^54 - 10661\ M\^56 - 
                    8962\ M\^58 + 7338\ M\^60 - 2085\ M\^62 + 220\ M\^64)\) + 
              L\^4\ \((M\^6 - 18\ M\^8 + 137\ M\^10 - 617\ M\^12 + 
                    1804\ M\^14 - 3258\ M\^16 + 2285\ M\^18 + 4736\ M\^20 - 
                    14311\ M\^22 + 9111\ M\^24 + 23926\ M\^26 - 
                    52363\ M\^28 + 7034\ M\^30 + 105458\ M\^32 - 
                    126004\ M\^34 - 85791\ M\^36 + 338125\ M\^38 - 
                    164670\ M\^40 - 459725\ M\^42 + 672837\ M\^44 + 
                    184484\ M\^46 - 1092279\ M\^48 + 469649\ M\^50 + 
                    961778\ M\^52 - 931829\ M\^54 - 329143\ M\^56 + 
                    790104\ M\^58 - 171910\ M\^60 - 305210\ M\^62 + 
                    210245\ M\^64 + 3293\ M\^66 - 58116\ M\^68 + 
                    28392\ M\^70 - 6000\ M\^72 + 495\ M\^74)\) + 
              L\^5\ \((3\ M\^4 - 53\ M\^6 + 414\ M\^8 - 1919\ M\^10 + 
                    5768\ M\^12 - 10821\ M\^14 + 8314\ M\^16 + 15932\ M\^18 - 
                    55713\ M\^20 + 53769\ M\^22 + 57910\ M\^24 - 
                    224419\ M\^26 + 208817\ M\^28 + 185363\ M\^30 - 
                    708425\ M\^32 + 569377\ M\^34 + 745257\ M\^36 - 
                    2048621\ M\^38 + 789699\ M\^40 + 3038022\ M\^42 - 
                    4157339\ M\^44 - 1654582\ M\^46 + 7770432\ M\^48 - 
                    2785014\ M\^50 - 8854007\ M\^52 + 8150763\ M\^54 + 
                    5937624\ M\^56 - 10584334\ M\^58 - 686889\ M\^60 + 
                    8403479\ M\^62 - 2984932\ M\^64 - 3776181\ M\^66 + 
                    3136582\ M\^68 + 373305\ M\^70 - 1383516\ M\^72 + 
                    494296\ M\^74 + 155211\ M\^76 - 186872\ M\^78 + 
                    67128\ M\^80 - 11490\ M\^82 + 792\ M\^84)\) + 
              L\^6\ \((3\ M\^2 - 61\ M\^4 + 542\ M\^6 - 2809\ M\^8 + 
                    9355\ M\^10 - 19988\ M\^12 + 22489\ M\^14 + 9222\ M\^16 - 
                    82945\ M\^18 + 132200\ M\^20 - 26193\ M\^22 - 
                    284754\ M\^24 + 587580\ M\^26 - 384279\ M\^28 - 
                    768473\ M\^30 + 2297268\ M\^32 - 1698263\ M\^34 - 
                    3218367\ M\^36 + 8056368\ M\^38 - 2217441\ M\^40 - 
                    13923106\ M\^42 + 17234375\ M\^44 + 9412315\ M\^46 - 
                    36001184\ M\^48 + 12378170\ M\^50 + 44395067\ M\^52 - 
                    44938723\ M\^54 - 32015298\ M\^56 + 69492795\ M\^58 + 
                    3538406\ M\^60 - 69869737\ M\^62 + 24242086\ M\^64 + 
                    47658279\ M\^66 - 35535804\ M\^68 - 18838705\ M\^70 + 
                    28466001\ M\^72 - 408456\ M\^74 - 13755610\ M\^76 + 
                    5620653\ M\^78 + 2975859\ M\^80 - 3139126\ M\^82 + 
                    524291\ M\^84 + 526018\ M\^86 - 363808\ M\^88 + 
                    105336\ M\^90 - 15372\ M\^92 + 924\ M\^94)\) + 
              L\^7\ \((1 - 25\ M\^2 + 267\ M\^4 - 1654\ M\^6 + 6691\ M\^8 - 
                    18547\ M\^10 + 34933\ M\^12 - 38604\ M\^14 - 
                    3963\ M\^16 + 122857\ M\^18 - 281103\ M\^20 + 
                    239059\ M\^22 + 446965\ M\^24 - 1694763\ M\^26 + 
                    1701240\ M\^28 + 1993226\ M\^30 - 7393836\ M\^32 + 
                    5253003\ M\^34 + 10483296\ M\^36 - 24229288\ M\^38 + 
                    5087575\ M\^40 + 43018237\ M\^42 - 51162723\ M\^44 - 
                    30749929\ M\^46 + 112559256\ M\^48 - 40326734\ M\^50 - 
                    142092755\ M\^52 + 159688902\ M\^54 + 93728231\ M\^56 - 
                    268936025\ M\^58 + 29875807\ M\^60 + 298994222\ M\^62 - 
                    166257372\ M\^64 - 227940300\ M\^66 + 240828572\ M\^68 + 
                    103507616\ M\^70 - 223404683\ M\^72 + 2465647\ M\^74 + 
                    145945321\ M\^76 - 50097430\ M\^78 - 64175762\ M\^80 + 
                    47215165\ M\^82 + 13356484\ M\^84 - 24719861\ M\^86 + 
                    4213393\ M\^88 + 6732988\ M\^90 - 4045917\ M\^92 - 
                    195\ M\^94 + 916170\ M\^96 - 462909\ M\^98 + 
                    113820\ M\^100 - 14658\ M\^102 + 792\ M\^104)\) + 
              L\^8\ \((\(-1\) + 22\ M\^2 - 226\ M\^4 + 1448\ M\^6 - 
                    6495\ M\^8 + 21463\ M\^10 - 52155\ M\^12 + 84369\ M\^14 - 
                    46282\ M\^16 - 180407\ M\^18 + 545945\ M\^20 - 
                    478024\ M\^22 - 843392\ M\^24 + 2870994\ M\^26 - 
                    2454639\ M\^28 - 3353183\ M\^30 + 10979690\ M\^32 - 
                    8411536\ M\^34 - 13638091\ M\^36 + 39009202\ M\^38 - 
                    19884468\ M\^40 - 63971375\ M\^42 + 113745517\ M\^44 + 
                    17452593\ M\^46 - 235372905\ M\^48 + 164189318\ M\^50 + 
                    272961288\ M\^52 - 457523280\ M\^54 - 103058273\ M\^56 + 
                    717983006\ M\^58 - 285022063\ M\^60 - 763493627\ M\^62 + 
                    730927301\ M\^64 + 522964553\ M\^66 - 1000038949\ M\^68 - 
                    108615810\ M\^70 + 962852644\ M\^72 - 257840767\ M\^74 - 
                    678565105\ M\^76 + 418341604\ M\^78 + 331766040\ M\^80 - 
                    371824986\ M\^82 - 81971471\ M\^84 + 229202708\ M\^86 - 
                    27462367\ M\^88 - 99764028\ M\^90 + 42606868\ M\^92 + 
                    26846454\ M\^94 - 24743051\ M\^96 - 500943\ M\^98 + 
                    7840661\ M\^100 - 2988128\ M\^102 - 660231\ M\^104 + 
                    973952\ M\^106 - 395830\ M\^108 + 85488\ M\^110 - 
                    9960\ M\^112 + 495\ M\^114)\) + 
              L\^9\ \((\(-2\)\ M\^2 + 41\ M\^4 - 404\ M\^6 + 2553\ M\^8 - 
                    11502\ M\^10 + 38230\ M\^12 - 91675\ M\^14 + 
                    138786\ M\^16 - 40400\ M\^18 - 397460\ M\^20 + 
                    996059\ M\^22 - 565347\ M\^24 - 2433639\ M\^26 + 
                    6167712\ M\^28 - 2397622\ M\^30 - 14268380\ M\^32 + 
                    25491361\ M\^34 + 4766248\ M\^36 - 66617339\ M\^38 + 
                    62910048\ M\^40 + 76419986\ M\^42 - 206644351\ M\^44 + 
                    50001968\ M\^46 + 347825294\ M\^48 - 398297369\ M\^50 - 
                    284283462\ M\^52 + 894852971\ M\^54 - 193481777\ M\^56 - 
                    1247603377\ M\^58 + 1085185873\ M\^60 + 
                    1097543209\ M\^62 - 2055668380\ M\^64 - 
                    319716719\ M\^66 + 2591513579\ M\^68 - 802005230\ M\^70 - 
                    2371485268\ M\^72 + 1730391092\ M\^74 + 
                    1530104855\ M\^76 - 2043933835\ M\^78 - 
                    531274248\ M\^80 + 1723441199\ M\^82 - 178218376\ M\^84 - 
                    1085058628\ M\^86 + 440399316\ M\^88 + 494107667\ M\^90 - 
                    379957365\ M\^92 - 137983841\ M\^94 + 214643692\ M\^96 + 
                    128608\ M\^98 - 85713912\ M\^100 + 23097155\ M\^102 + 
                    23562572\ M\^104 - 14284686\ M\^106 - 2872978\ M\^108 + 
                    5043430\ M\^110 - 1087154\ M\^112 - 807186\ M\^114 + 
                    658422\ M\^116 - 226526\ M\^118 + 43992\ M\^120 - 
                    4725\ M\^122 + 220\ M\^124)\) + 
              L\^10\ \((\(-M\^4\) + 23\ M\^6 - 265\ M\^8 + 1968\ M\^10 - 
                    10237\ M\^12 + 38303\ M\^14 - 101364\ M\^16 + 
                    170596\ M\^18 - 88833\ M\^20 - 382513\ M\^22 + 
                    1051207\ M\^24 - 530470\ M\^26 - 2870283\ M\^28 + 
                    6598792\ M\^30 - 747962\ M\^32 - 19300226\ M\^34 + 
                    26830145\ M\^36 + 22917432\ M\^38 - 97946220\ M\^40 + 
                    42602020\ M\^42 + 196142655\ M\^44 - 273538337\ M\^46 - 
                    194383831\ M\^48 + 711981647\ M\^50 - 156380452\ M\^52 - 
                    1184383697\ M\^54 + 1066552204\ M\^56 + 
                    1245891641\ M\^58 - 2425440660\ M\^60 - 
                    409079935\ M\^62 + 3655177981\ M\^64 - 
                    1405751652\ M\^66 - 3969152153\ M\^68 + 
                    3603907594\ M\^70 + 2936956211\ M\^72 - 
                    5181931941\ M\^74 - 898560469\ M\^76 + 
                    5393713297\ M\^78 - 1191038802\ M\^80 - 
                    4243795848\ M\^82 + 2435375027\ M\^84 + 
                    2446972720\ M\^86 - 2547418306\ M\^88 - 
                    861506039\ M\^90 + 1885589137\ M\^92 - 48453742\ M\^94 - 
                    1041729881\ M\^96 + 325515264\ M\^98 + 
                    417412119\ M\^100 - 265189983\ M\^102 - 
                    103606444\ M\^104 + 132104303\ M\^106 + 2168445\ M\^108 - 
                    44214365\ M\^110 + 10123833\ M\^112 + 10145383\ M\^114 - 
                    4931490\ M\^116 - 1527920\ M\^118 + 1592351\ M\^120 + 
                    1394\ M\^122 - 478680\ M\^124 + 278157\ M\^126 - 
                    83492\ M\^128 + 14828\ M\^130 - 1490\ M\^132 + 
                    66\ M\^134)\) + 
              L\^11\ \((3\ M\^8 - 64\ M\^10 + 660\ M\^12 - 4322\ M\^14 + 
                    19664\ M\^16 - 63146\ M\^18 + 135396\ M\^20 - 
                    148311\ M\^22 - 96643\ M\^24 + 564132\ M\^26 - 
                    328658\ M\^28 - 1800843\ M\^30 + 3714240\ M\^32 + 
                    1700883\ M\^34 - 14944224\ M\^36 + 11689114\ M\^38 + 
                    35116910\ M\^40 - 74392633\ M\^42 - 24098496\ M\^44 + 
                    225614741\ M\^46 - 142777429\ M\^48 - 424252792\ M\^50 + 
                    658946685\ M\^52 + 422237772\ M\^54 - 1613944015\ M\^56 + 
                    239923185\ M\^58 + 2730797478\ M\^60 - 
                    1964761738\ M\^62 - 3267551499\ M\^64 + 
                    4618597483\ M\^66 + 2368210286\ M\^68 - 
                    7268555065\ M\^70 + 247372814\ M\^72 + 
                    8557410188\ M\^74 - 3810177604\ M\^76 - 
                    7616402960\ M\^78 + 6796847517\ M\^80 + 
                    4768062898\ M\^82 - 7884946115\ M\^84 - 
                    1331634126\ M\^86 + 6848982991\ M\^88 - 
                    1251204734\ M\^90 - 4555240556\ M\^92 + 
                    2298999810\ M\^94 + 2226712455\ M\^96 - 
                    2075756892\ M\^98 - 657576763\ M\^100 + 
                    1311353251\ M\^102 - 42998528\ M\^104 - 
                    611981654\ M\^106 + 189656236\ M\^108 + 
                    203570312\ M\^110 - 129822498\ M\^112 - 
                    38801740\ M\^114 + 54747844\ M\^116 - 3276995\ M\^118 - 
                    14435635\ M\^120 + 4760363\ M\^122 + 1900248\ M\^124 - 
                    1336343\ M\^126 + 13650\ M\^128 + 59072\ M\^130 + 
                    150449\ M\^132 - 150588\ M\^134 + 67571\ M\^136 - 
                    17977\ M\^138 + 2954\ M\^140 - 281\ M\^142 + 
                    12\ M\^144)\) + 
              L\^24\ \((12\ M\^126 - 41\ M\^128 + 32\ M\^130 + 49\ M\^132 - 
                    44\ M\^134 + 18\ M\^136 + 9\ M\^138 - 19\ M\^140 + 
                    14\ M\^142 - 7\ M\^144 + 2\ M\^146)\) + 
              L\^12\ \((\(-3\)\ M\^12 + 67\ M\^14 - 699\ M\^16 + 
                    4445\ M\^18 - 18788\ M\^20 + 53513\ M\^22 - 
                    96842\ M\^24 + 87507\ M\^26 - 2635\ M\^28 + 
                    106667\ M\^30 - 739818\ M\^32 + 805662\ M\^34 + 
                    2631640\ M\^36 - 7592806\ M\^38 - 757122\ M\^40 + 
                    29592224\ M\^42 - 31563818\ M\^44 - 59673608\ M\^46 + 
                    149756846\ M\^48 + 30299446\ M\^50 - 412714740\ M\^52 + 
                    248952010\ M\^54 + 764847896\ M\^56 - 1066324911\ M\^58 - 
                    872329060\ M\^60 + 2592397300\ M\^62 + 95621177\ M\^64 - 
                    4511339094\ M\^66 + 2158940490\ M\^68 + 
                    5858098553\ M\^70 - 5813386178\ M\^72 - 
                    5425699751\ M\^74 + 9740778284\ M\^76 + 
                    2658783271\ M\^78 - 12181347629\ M\^80 + 
                    1676521127\ M\^82 + 11830261566\ M\^84 - 
                    5761356181\ M\^86 - 8774256413\ M\^88 + 
                    7890597240\ M\^90 + 4500551831\ M\^92 - 
                    7512737618\ M\^94 - 840186216\ M\^96 + 
                    5412306383\ M\^98 - 1159557471\ M\^100 - 
                    2948440036\ M\^102 + 1578563099\ M\^104 + 
                    1123740714\ M\^106 - 1150563915\ M\^108 - 
                    197975697\ M\^110 + 580814720\ M\^112 - 
                    85323263\ M\^114 - 206966928\ M\^116 + 90159602\ M\^118 + 
                    46956811\ M\^120 - 41693139\ M\^122 - 3541494\ M\^124 + 
                    13043835\ M\^126 - 2550533\ M\^128 - 2656971\ M\^130 + 
                    1412767\ M\^132 + 177869\ M\^134 - 435115\ M\^136 + 
                    230629\ M\^138 - 98200\ M\^140 + 47639\ M\^142 - 
                    21756\ M\^144 + 7407\ M\^146 - 1730\ M\^148 + 
                    264\ M\^150 - 24\ M\^152 + M\^154)\) + 
              L\^23\ \((66\ M\^116 - 434\ M\^118 + 956\ M\^120 - 
                    224\ M\^122 - 2091\ M\^124 + 2433\ M\^126 + 943\ M\^128 - 
                    2769\ M\^130 + 1511\ M\^132 + 788\ M\^134 - 
                    1465\ M\^136 + 579\ M\^138 + 171\ M\^140 - 271\ M\^142 + 
                    151\ M\^144 - 66\ M\^146 + 40\ M\^148 - 30\ M\^150 + 
                    17\ M\^152 - 6\ M\^154 + M\^156)\) + 
              L\^13\ \((M\^16 - 24\ M\^18 + 264\ M\^20 - 1730\ M\^22 + 
                    7407\ M\^24 - 21756\ M\^26 + 47639\ M\^28 - 
                    98200\ M\^30 + 230629\ M\^32 - 435115\ M\^34 + 
                    177869\ M\^36 + 1412767\ M\^38 - 2656971\ M\^40 - 
                    2550533\ M\^42 + 13043835\ M\^44 - 3541494\ M\^46 - 
                    41693139\ M\^48 + 46956811\ M\^50 + 90159602\ M\^52 - 
                    206966928\ M\^54 - 85323263\ M\^56 + 580814720\ M\^58 - 
                    197975697\ M\^60 - 1150563915\ M\^62 + 
                    1123740714\ M\^64 + 1578563099\ M\^66 - 
                    2948440036\ M\^68 - 1159557471\ M\^70 + 
                    5412306383\ M\^72 - 840186216\ M\^74 - 
                    7512737618\ M\^76 + 4500551831\ M\^78 + 
                    7890597240\ M\^80 - 8774256413\ M\^82 - 
                    5761356181\ M\^84 + 11830261566\ M\^86 + 
                    1676521127\ M\^88 - 12181347629\ M\^90 + 
                    2658783271\ M\^92 + 9740778284\ M\^94 - 
                    5425699751\ M\^96 - 5813386178\ M\^98 + 
                    5858098553\ M\^100 + 2158940490\ M\^102 - 
                    4511339094\ M\^104 + 95621177\ M\^106 + 
                    2592397300\ M\^108 - 872329060\ M\^110 - 
                    1066324911\ M\^112 + 764847896\ M\^114 + 
                    248952010\ M\^116 - 412714740\ M\^118 + 
                    30299446\ M\^120 + 149756846\ M\^122 - 59673608\ M\^124 - 
                    31563818\ M\^126 + 29592224\ M\^128 - 757122\ M\^130 - 
                    7592806\ M\^132 + 2631640\ M\^134 + 805662\ M\^136 - 
                    739818\ M\^138 + 106667\ M\^140 - 2635\ M\^142 + 
                    87507\ M\^144 - 96842\ M\^146 + 53513\ M\^148 - 
                    18788\ M\^150 + 4445\ M\^152 - 699\ M\^154 + 67\ M\^156 - 
                    3\ M\^158)\) + 
              L\^22\ \((220\ M\^106 - 2085\ M\^108 + 7338\ M\^110 - 
                    8962\ M\^112 - 10661\ M\^114 + 40656\ M\^116 - 
                    20802\ M\^118 - 54069\ M\^120 + 70511\ M\^122 + 
                    16122\ M\^124 - 73879\ M\^126 + 34542\ M\^128 + 
                    30211\ M\^130 - 39692\ M\^132 + 6217\ M\^134 + 
                    15165\ M\^136 - 9951\ M\^138 - 758\ M\^140 + 
                    3207\ M\^142 - 595\ M\^144 - 816\ M\^146 + 258\ M\^148 + 
                    408\ M\^150 - 481\ M\^152 + 276\ M\^154 - 99\ M\^156 + 
                    21\ M\^158 - 2\ M\^160)\) + 
              L\^14\ \((12\ M\^26 - 281\ M\^28 + 2954\ M\^30 - 17977\ M\^32 + 
                    67571\ M\^34 - 150588\ M\^36 + 150449\ M\^38 + 
                    59072\ M\^40 + 13650\ M\^42 - 1336343\ M\^44 + 
                    1900248\ M\^46 + 4760363\ M\^48 - 14435635\ M\^50 - 
                    3276995\ M\^52 + 54747844\ M\^54 - 38801740\ M\^56 - 
                    129822498\ M\^58 + 203570312\ M\^60 + 189656236\ M\^62 - 
                    611981654\ M\^64 - 42998528\ M\^66 + 1311353251\ M\^68 - 
                    657576763\ M\^70 - 2075756892\ M\^72 + 
                    2226712455\ M\^74 + 2298999810\ M\^76 - 
                    4555240556\ M\^78 - 1251204734\ M\^80 + 
                    6848982991\ M\^82 - 1331634126\ M\^84 - 
                    7884946115\ M\^86 + 4768062898\ M\^88 + 
                    6796847517\ M\^90 - 7616402960\ M\^92 - 
                    3810177604\ M\^94 + 8557410188\ M\^96 + 
                    247372814\ M\^98 - 7268555065\ M\^100 + 
                    2368210286\ M\^102 + 4618597483\ M\^104 - 
                    3267551499\ M\^106 - 1964761738\ M\^108 + 
                    2730797478\ M\^110 + 239923185\ M\^112 - 
                    1613944015\ M\^114 + 422237772\ M\^116 + 
                    658946685\ M\^118 - 424252792\ M\^120 - 
                    142777429\ M\^122 + 225614741\ M\^124 - 
                    24098496\ M\^126 - 74392633\ M\^128 + 35116910\ M\^130 + 
                    11689114\ M\^132 - 14944224\ M\^134 + 1700883\ M\^136 + 
                    3714240\ M\^138 - 1800843\ M\^140 - 328658\ M\^142 + 
                    564132\ M\^144 - 96643\ M\^146 - 148311\ M\^148 + 
                    135396\ M\^150 - 63146\ M\^152 + 19664\ M\^154 - 
                    4322\ M\^156 + 660\ M\^158 - 64\ M\^160 + 3\ M\^162)\) + 
              L\^21\ \((495\ M\^96 - 6000\ M\^98 + 28392\ M\^100 - 
                    58116\ M\^102 + 3293\ M\^104 + 210245\ M\^106 - 
                    305210\ M\^108 - 171910\ M\^110 + 790104\ M\^112 - 
                    329143\ M\^114 - 931829\ M\^116 + 961778\ M\^118 + 
                    469649\ M\^120 - 1092279\ M\^122 + 184484\ M\^124 + 
                    672837\ M\^126 - 459725\ M\^128 - 164670\ M\^130 + 
                    338125\ M\^132 - 85791\ M\^134 - 126004\ M\^136 + 
                    105458\ M\^138 + 7034\ M\^140 - 52363\ M\^142 + 
                    23926\ M\^144 + 9111\ M\^146 - 14311\ M\^148 + 
                    4736\ M\^150 + 2285\ M\^152 - 3258\ M\^154 + 
                    1804\ M\^156 - 617\ M\^158 + 137\ M\^160 - 18\ M\^162 + 
                    M\^164)\) + 
              L\^15\ \((66\ M\^36 - 1490\ M\^38 + 14828\ M\^40 - 
                    83492\ M\^42 + 278157\ M\^44 - 478680\ M\^46 + 
                    1394\ M\^48 + 1592351\ M\^50 - 1527920\ M\^52 - 
                    4931490\ M\^54 + 10145383\ M\^56 + 10123833\ M\^58 - 
                    44214365\ M\^60 + 2168445\ M\^62 + 132104303\ M\^64 - 
                    103606444\ M\^66 - 265189983\ M\^68 + 417412119\ M\^70 + 
                    325515264\ M\^72 - 1041729881\ M\^74 - 48453742\ M\^76 + 
                    1885589137\ M\^78 - 861506039\ M\^80 - 
                    2547418306\ M\^82 + 2446972720\ M\^84 + 
                    2435375027\ M\^86 - 4243795848\ M\^88 - 
                    1191038802\ M\^90 + 5393713297\ M\^92 - 
                    898560469\ M\^94 - 5181931941\ M\^96 + 
                    2936956211\ M\^98 + 3603907594\ M\^100 - 
                    3969152153\ M\^102 - 1405751652\ M\^104 + 
                    3655177981\ M\^106 - 409079935\ M\^108 - 
                    2425440660\ M\^110 + 1245891641\ M\^112 + 
                    1066552204\ M\^114 - 1184383697\ M\^116 - 
                    156380452\ M\^118 + 711981647\ M\^120 - 
                    194383831\ M\^122 - 273538337\ M\^124 + 
                    196142655\ M\^126 + 42602020\ M\^128 - 97946220\ M\^130 + 
                    22917432\ M\^132 + 26830145\ M\^134 - 19300226\ M\^136 - 
                    747962\ M\^138 + 6598792\ M\^140 - 2870283\ M\^142 - 
                    530470\ M\^144 + 1051207\ M\^146 - 382513\ M\^148 - 
                    88833\ M\^150 + 170596\ M\^152 - 101364\ M\^154 + 
                    38303\ M\^156 - 10237\ M\^158 + 1968\ M\^160 - 
                    265\ M\^162 + 23\ M\^164 - M\^166)\) + 
              L\^20\ \((792\ M\^86 - 11490\ M\^88 + 67128\ M\^90 - 
                    186872\ M\^92 + 155211\ M\^94 + 494296\ M\^96 - 
                    1383516\ M\^98 + 373305\ M\^100 + 3136582\ M\^102 - 
                    3776181\ M\^104 - 2984932\ M\^106 + 8403479\ M\^108 - 
                    686889\ M\^110 - 10584334\ M\^112 + 5937624\ M\^114 + 
                    8150763\ M\^116 - 8854007\ M\^118 - 2785014\ M\^120 + 
                    7770432\ M\^122 - 1654582\ M\^124 - 4157339\ M\^126 + 
                    3038022\ M\^128 + 789699\ M\^130 - 2048621\ M\^132 + 
                    745257\ M\^134 + 569377\ M\^136 - 708425\ M\^138 + 
                    185363\ M\^140 + 208817\ M\^142 - 224419\ M\^144 + 
                    57910\ M\^146 + 53769\ M\^148 - 55713\ M\^150 + 
                    15932\ M\^152 + 8314\ M\^154 - 10821\ M\^156 + 
                    5768\ M\^158 - 1919\ M\^160 + 414\ M\^162 - 53\ M\^164 + 
                    3\ M\^166)\) + 
              L\^16\ \((220\ M\^46 - 4725\ M\^48 + 43992\ M\^50 - 
                    226526\ M\^52 + 658422\ M\^54 - 807186\ M\^56 - 
                    1087154\ M\^58 + 5043430\ M\^60 - 2872978\ M\^62 - 
                    14284686\ M\^64 + 23562572\ M\^66 + 23097155\ M\^68 - 
                    85713912\ M\^70 + 128608\ M\^72 + 214643692\ M\^74 - 
                    137983841\ M\^76 - 379957365\ M\^78 + 494107667\ M\^80 + 
                    440399316\ M\^82 - 1085058628\ M\^84 - 178218376\ M\^86 + 
                    1723441199\ M\^88 - 531274248\ M\^90 - 
                    2043933835\ M\^92 + 1530104855\ M\^94 + 
                    1730391092\ M\^96 - 2371485268\ M\^98 - 
                    802005230\ M\^100 + 2591513579\ M\^102 - 
                    319716719\ M\^104 - 2055668380\ M\^106 + 
                    1097543209\ M\^108 + 1085185873\ M\^110 - 
                    1247603377\ M\^112 - 193481777\ M\^114 + 
                    894852971\ M\^116 - 284283462\ M\^118 - 
                    398297369\ M\^120 + 347825294\ M\^122 + 
                    50001968\ M\^124 - 206644351\ M\^126 + 76419986\ M\^128 + 
                    62910048\ M\^130 - 66617339\ M\^132 + 4766248\ M\^134 + 
                    25491361\ M\^136 - 14268380\ M\^138 - 2397622\ M\^140 + 
                    6167712\ M\^142 - 2433639\ M\^144 - 565347\ M\^146 + 
                    996059\ M\^148 - 397460\ M\^150 - 40400\ M\^152 + 
                    138786\ M\^154 - 91675\ M\^156 + 38230\ M\^158 - 
                    11502\ M\^160 + 2553\ M\^162 - 404\ M\^164 + 41\ M\^166 - 
                    2\ M\^168)\) + 
              L\^19\ \((924\ M\^76 - 15372\ M\^78 + 105336\ M\^80 - 
                    363808\ M\^82 + 526018\ M\^84 + 524291\ M\^86 - 
                    3139126\ M\^88 + 2975859\ M\^90 + 5620653\ M\^92 - 
                    13755610\ M\^94 - 408456\ M\^96 + 28466001\ M\^98 - 
                    18838705\ M\^100 - 35535804\ M\^102 + 47658279\ M\^104 + 
                    24242086\ M\^106 - 69869737\ M\^108 + 3538406\ M\^110 + 
                    69492795\ M\^112 - 32015298\ M\^114 - 44938723\ M\^116 + 
                    44395067\ M\^118 + 12378170\ M\^120 - 36001184\ M\^122 + 
                    9412315\ M\^124 + 17234375\ M\^126 - 13923106\ M\^128 - 
                    2217441\ M\^130 + 8056368\ M\^132 - 3218367\ M\^134 - 
                    1698263\ M\^136 + 2297268\ M\^138 - 768473\ M\^140 - 
                    384279\ M\^142 + 587580\ M\^144 - 284754\ M\^146 - 
                    26193\ M\^148 + 132200\ M\^150 - 82945\ M\^152 + 
                    9222\ M\^154 + 22489\ M\^156 - 19988\ M\^158 + 
                    9355\ M\^160 - 2809\ M\^162 + 542\ M\^164 - 61\ M\^166 + 
                    3\ M\^168)\) + 
              L\^17\ \((495\ M\^56 - 9960\ M\^58 + 85488\ M\^60 - 
                    395830\ M\^62 + 973952\ M\^64 - 660231\ M\^66 - 
                    2988128\ M\^68 + 7840661\ M\^70 - 500943\ M\^72 - 
                    24743051\ M\^74 + 26846454\ M\^76 + 42606868\ M\^78 - 
                    99764028\ M\^80 - 27462367\ M\^82 + 229202708\ M\^84 - 
                    81971471\ M\^86 - 371824986\ M\^88 + 331766040\ M\^90 + 
                    418341604\ M\^92 - 678565105\ M\^94 - 257840767\ M\^96 + 
                    962852644\ M\^98 - 108615810\ M\^100 - 
                    1000038949\ M\^102 + 522964553\ M\^104 + 
                    730927301\ M\^106 - 763493627\ M\^108 - 
                    285022063\ M\^110 + 717983006\ M\^112 - 
                    103058273\ M\^114 - 457523280\ M\^116 + 
                    272961288\ M\^118 + 164189318\ M\^120 - 
                    235372905\ M\^122 + 17452593\ M\^124 + 
                    113745517\ M\^126 - 63971375\ M\^128 - 19884468\ M\^130 + 
                    39009202\ M\^132 - 13638091\ M\^134 - 8411536\ M\^136 + 
                    10979690\ M\^138 - 3353183\ M\^140 - 2454639\ M\^142 + 
                    2870994\ M\^144 - 843392\ M\^146 - 478024\ M\^148 + 
                    545945\ M\^150 - 180407\ M\^152 - 46282\ M\^154 + 
                    84369\ M\^156 - 52155\ M\^158 + 21463\ M\^160 - 
                    6495\ M\^162 + 1448\ M\^164 - 226\ M\^166 + 22\ M\^168 - 
                    M\^170)\) + 
              L\^18\ \((792\ M\^66 - 14658\ M\^68 + 113820\ M\^70 - 
                    462909\ M\^72 + 916170\ M\^74 - 195\ M\^76 - 
                    4045917\ M\^78 + 6732988\ M\^80 + 4213393\ M\^82 - 
                    24719861\ M\^84 + 13356484\ M\^86 + 47215165\ M\^88 - 
                    64175762\ M\^90 - 50097430\ M\^92 + 145945321\ M\^94 + 
                    2465647\ M\^96 - 223404683\ M\^98 + 103507616\ M\^100 + 
                    240828572\ M\^102 - 227940300\ M\^104 - 
                    166257372\ M\^106 + 298994222\ M\^108 + 
                    29875807\ M\^110 - 268936025\ M\^112 + 93728231\ M\^114 + 
                    159688902\ M\^116 - 142092755\ M\^118 - 
                    40326734\ M\^120 + 112559256\ M\^122 - 30749929\ M\^124 - 
                    51162723\ M\^126 + 43018237\ M\^128 + 5087575\ M\^130 - 
                    24229288\ M\^132 + 10483296\ M\^134 + 5253003\ M\^136 - 
                    7393836\ M\^138 + 1993226\ M\^140 + 1701240\ M\^142 - 
                    1694763\ M\^144 + 446965\ M\^146 + 239059\ M\^148 - 
                    281103\ M\^150 + 122857\ M\^152 - 3963\ M\^154 - 
                    38604\ M\^156 + 34933\ M\^158 - 18547\ M\^160 + 
                    6691\ M\^162 - 1654\ M\^164 + 267\ M\^166 - 25\ M\^168 + 
                    M\^170)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 20]\), \(L\^17\ M\^12 + M\^82 + 
              L\^16\ \((\(-3\)\ M\^8 + 10\ M\^10 - 15\ M\^12 + 4\ M\^14 + 
                    27\ M\^16 - 6\ M\^18 + 2\ M\^22 - 6\ M\^24 + 7\ M\^26 - 
                    3\ M\^28)\) + 
              L\^15\ \((3\ M\^4 - 20\ M\^6 + 56\ M\^8 - 75\ M\^10 + 
                    4\ M\^12 + 103\ M\^14 - 91\ M\^16 - 125\ M\^18 + 
                    236\ M\^20 + 338\ M\^22 - 306\ M\^24 - 264\ M\^26 + 
                    371\ M\^28 + 33\ M\^30 - 184\ M\^32 + 52\ M\^34 + 
                    17\ M\^36 - 28\ M\^38 + 27\ M\^40 - 14\ M\^42 + 
                    3\ M\^44)\) + 
              L\^14\ \((\(-1\) + 10\ M\^2 - 43\ M\^4 + 104\ M\^6 - 
                    178\ M\^8 + 275\ M\^10 - 277\ M\^12 - 334\ M\^14 + 
                    1366\ M\^16 - 530\ M\^18 - 2993\ M\^20 + 2350\ M\^22 + 
                    4492\ M\^24 - 3547\ M\^26 - 3435\ M\^28 + 3705\ M\^30 + 
                    1499\ M\^32 - 1605\ M\^34 - 460\ M\^36 - 158\ M\^38 + 
                    383\ M\^40 + 447\ M\^42 - 458\ M\^44 - 52\ M\^46 + 
                    162\ M\^48 - 27\ M\^50 - 35\ M\^52 + 35\ M\^54 - 
                    21\ M\^56 + 7\ M\^58 - M\^60)\) + 
              L\^13\ \((2 - 25\ M\^2 + 144\ M\^4 - 448\ M\^6 + 674\ M\^8 - 
                    148\ M\^10 - 558\ M\^12 - 611\ M\^14 + 962\ M\^16 + 
                    4535\ M\^18 - 3778\ M\^20 - 13350\ M\^22 + 9207\ M\^24 + 
                    23696\ M\^26 - 12939\ M\^28 - 26538\ M\^30 + 
                    13390\ M\^32 + 23567\ M\^34 - 6094\ M\^36 - 
                    16387\ M\^38 + 1081\ M\^40 + 8931\ M\^42 - 697\ M\^44 - 
                    3180\ M\^46 + 885\ M\^48 - 333\ M\^50 + 315\ M\^52 + 
                    522\ M\^54 - 534\ M\^56 - 73\ M\^58 + 281\ M\^60 - 
                    154\ M\^62 + 39\ M\^64 - 4\ M\^66)\) + 
              L\^12\ \((\(-1\) + 20\ M\^2 - 173\ M\^4 + 792\ M\^6 - 
                    1928\ M\^8 + 1614\ M\^10 + 2966\ M\^12 - 5701\ M\^14 - 
                    6294\ M\^16 + 14847\ M\^18 + 15178\ M\^20 - 
                    28698\ M\^22 - 36214\ M\^24 + 39388\ M\^26 + 
                    66295\ M\^28 - 33541\ M\^30 - 89218\ M\^32 + 
                    13442\ M\^34 + 95147\ M\^36 + 15699\ M\^38 - 
                    71961\ M\^40 - 20430\ M\^42 + 39838\ M\^44 + 
                    9216\ M\^46 - 15944\ M\^48 - 2535\ M\^50 + 4317\ M\^52 + 
                    2507\ M\^54 - 3096\ M\^56 - 435\ M\^58 + 2049\ M\^60 - 
                    860\ M\^62 - 608\ M\^64 + 791\ M\^66 - 347\ M\^68 + 
                    72\ M\^70 - 6\ M\^72)\) + 
              L\^11\ \((\(-5\)\ M\^2 + 79\ M\^4 - 570\ M\^6 + 2234\ M\^8 - 
                    4408\ M\^10 + 1130\ M\^12 + 11760\ M\^14 - 11742\ M\^16 - 
                    27685\ M\^18 + 38469\ M\^20 + 56972\ M\^22 - 
                    76207\ M\^24 - 109753\ M\^26 + 96778\ M\^28 + 
                    175786\ M\^30 - 60083\ M\^32 - 218716\ M\^34 - 
                    26395\ M\^36 + 199201\ M\^38 + 112461\ M\^40 - 
                    107944\ M\^42 - 109776\ M\^44 + 31708\ M\^46 + 
                    60663\ M\^48 - 4245\ M\^50 - 24099\ M\^52 + 3099\ M\^54 + 
                    7580\ M\^56 - 6757\ M\^58 + 3309\ M\^60 + 1772\ M\^62 - 
                    4113\ M\^64 + 1778\ M\^66 + 1060\ M\^68 - 1674\ M\^70 + 
                    973\ M\^72 - 315\ M\^74 + 55\ M\^76 - 4\ M\^78)\) + 
              L\^10\ \((\(-10\)\ M\^4 + 152\ M\^6 - 1014\ M\^8 + 
                    3596\ M\^10 - 6040\ M\^12 - 716\ M\^14 + 18090\ M\^16 - 
                    9782\ M\^18 - 44710\ M\^20 + 34251\ M\^22 + 
                    95677\ M\^24 - 56224\ M\^26 - 184933\ M\^28 + 
                    32244\ M\^30 + 277824\ M\^32 + 86582\ M\^34 - 
                    289341\ M\^36 - 252435\ M\^38 + 147406\ M\^40 + 
                    335057\ M\^42 + 82661\ M\^44 - 225779\ M\^46 - 
                    185950\ M\^48 + 76494\ M\^50 + 149697\ M\^52 - 
                    971\ M\^54 - 79745\ M\^56 - 12666\ M\^58 + 34712\ M\^60 + 
                    11153\ M\^62 - 20657\ M\^64 - 1487\ M\^66 + 
                    10626\ M\^68 - 3698\ M\^70 - 2598\ M\^72 + 3165\ M\^74 - 
                    1587\ M\^76 + 497\ M\^78 - 107\ M\^80 + 15\ M\^82 - 
                    M\^84)\) + 
              L\^9\ \((\(-10\)\ M\^6 + 152\ M\^8 - 987\ M\^10 + 3380\ M\^12 - 
                    5592\ M\^14 + 376\ M\^16 + 12392\ M\^18 - 7199\ M\^20 - 
                    23478\ M\^22 + 10871\ M\^24 + 44263\ M\^26 + 
                    17552\ M\^28 - 82749\ M\^30 - 112709\ M\^32 + 
                    83954\ M\^34 + 286792\ M\^36 + 44055\ M\^38 - 
                    422349\ M\^40 - 326638\ M\^42 + 349212\ M\^44 + 
                    575873\ M\^46 - 55755\ M\^48 - 534226\ M\^50 - 
                    154090\ M\^52 + 321271\ M\^54 + 171245\ M\^56 - 
                    142869\ M\^58 - 92357\ M\^60 + 55952\ M\^62 + 
                    28523\ M\^64 - 26661\ M\^66 + 3704\ M\^68 + 7828\ M\^70 - 
                    8377\ M\^72 + 1839\ M\^74 + 3462\ M\^76 - 3715\ M\^78 + 
                    1801\ M\^80 - 497\ M\^82 + 76\ M\^84 - 5\ M\^86)\) + 
              L\ \((\(-3\)\ M\^66 + 7\ M\^68 - 6\ M\^70 + 2\ M\^72 - 
                    6\ M\^76 + 27\ M\^78 + 4\ M\^80 - 15\ M\^82 + 10\ M\^84 - 
                    3\ M\^86)\) + 
              L\^8\ \((\(-5\)\ M\^8 + 76\ M\^10 - 497\ M\^12 + 1801\ M\^14 - 
                    3715\ M\^16 + 3462\ M\^18 + 1839\ M\^20 - 8377\ M\^22 + 
                    7828\ M\^24 + 3704\ M\^26 - 26661\ M\^28 + 28523\ M\^30 + 
                    55952\ M\^32 - 92357\ M\^34 - 142869\ M\^36 + 
                    171245\ M\^38 + 321271\ M\^40 - 154090\ M\^42 - 
                    534226\ M\^44 - 55755\ M\^46 + 575873\ M\^48 + 
                    349212\ M\^50 - 326638\ M\^52 - 422349\ M\^54 + 
                    44055\ M\^56 + 286792\ M\^58 + 83954\ M\^60 - 
                    112709\ M\^62 - 82749\ M\^64 + 17552\ M\^66 + 
                    44263\ M\^68 + 10871\ M\^70 - 23478\ M\^72 - 
                    7199\ M\^74 + 12392\ M\^76 + 376\ M\^78 - 5592\ M\^80 + 
                    3380\ M\^82 - 987\ M\^84 + 152\ M\^86 - 10\ M\^88)\) + 
              L\^7\ \((\(-M\^10\) + 15\ M\^12 - 107\ M\^14 + 497\ M\^16 - 
                    1587\ M\^18 + 3165\ M\^20 - 2598\ M\^22 - 3698\ M\^24 + 
                    10626\ M\^26 - 1487\ M\^28 - 20657\ M\^30 + 
                    11153\ M\^32 + 34712\ M\^34 - 12666\ M\^36 - 
                    79745\ M\^38 - 971\ M\^40 + 149697\ M\^42 + 
                    76494\ M\^44 - 185950\ M\^46 - 225779\ M\^48 + 
                    82661\ M\^50 + 335057\ M\^52 + 147406\ M\^54 - 
                    252435\ M\^56 - 289341\ M\^58 + 86582\ M\^60 + 
                    277824\ M\^62 + 32244\ M\^64 - 184933\ M\^66 - 
                    56224\ M\^68 + 95677\ M\^70 + 34251\ M\^72 - 
                    44710\ M\^74 - 9782\ M\^76 + 18090\ M\^78 - 716\ M\^80 - 
                    6040\ M\^82 + 3596\ M\^84 - 1014\ M\^86 + 152\ M\^88 - 
                    10\ M\^90)\) + 
              L\^2\ \((3\ M\^50 - 14\ M\^52 + 27\ M\^54 - 28\ M\^56 + 
                    17\ M\^58 + 52\ M\^60 - 184\ M\^62 + 33\ M\^64 + 
                    371\ M\^66 - 264\ M\^68 - 306\ M\^70 + 338\ M\^72 + 
                    236\ M\^74 - 125\ M\^76 - 91\ M\^78 + 103\ M\^80 + 
                    4\ M\^82 - 75\ M\^84 + 56\ M\^86 - 20\ M\^88 + 
                    3\ M\^90)\) + 
              L\^6\ \((\(-4\)\ M\^16 + 55\ M\^18 - 315\ M\^20 + 973\ M\^22 - 
                    1674\ M\^24 + 1060\ M\^26 + 1778\ M\^28 - 4113\ M\^30 + 
                    1772\ M\^32 + 3309\ M\^34 - 6757\ M\^36 + 7580\ M\^38 + 
                    3099\ M\^40 - 24099\ M\^42 - 4245\ M\^44 + 60663\ M\^46 + 
                    31708\ M\^48 - 109776\ M\^50 - 107944\ M\^52 + 
                    112461\ M\^54 + 199201\ M\^56 - 26395\ M\^58 - 
                    218716\ M\^60 - 60083\ M\^62 + 175786\ M\^64 + 
                    96778\ M\^66 - 109753\ M\^68 - 76207\ M\^70 + 
                    56972\ M\^72 + 38469\ M\^74 - 27685\ M\^76 - 
                    11742\ M\^78 + 11760\ M\^80 + 1130\ M\^82 - 4408\ M\^84 + 
                    2234\ M\^86 - 570\ M\^88 + 79\ M\^90 - 5\ M\^92)\) + 
              L\^3\ \((\(-M\^34\) + 7\ M\^36 - 21\ M\^38 + 35\ M\^40 - 
                    35\ M\^42 - 27\ M\^44 + 162\ M\^46 - 52\ M\^48 - 
                    458\ M\^50 + 447\ M\^52 + 383\ M\^54 - 158\ M\^56 - 
                    460\ M\^58 - 1605\ M\^60 + 1499\ M\^62 + 3705\ M\^64 - 
                    3435\ M\^66 - 3547\ M\^68 + 4492\ M\^70 + 2350\ M\^72 - 
                    2993\ M\^74 - 530\ M\^76 + 1366\ M\^78 - 334\ M\^80 - 
                    277\ M\^82 + 275\ M\^84 - 178\ M\^86 + 104\ M\^88 - 
                    43\ M\^90 + 10\ M\^92 - M\^94)\) + 
              L\^5\ \((\(-6\)\ M\^22 + 72\ M\^24 - 347\ M\^26 + 791\ M\^28 - 
                    608\ M\^30 - 860\ M\^32 + 2049\ M\^34 - 435\ M\^36 - 
                    3096\ M\^38 + 2507\ M\^40 + 4317\ M\^42 - 2535\ M\^44 - 
                    15944\ M\^46 + 9216\ M\^48 + 39838\ M\^50 - 
                    20430\ M\^52 - 71961\ M\^54 + 15699\ M\^56 + 
                    95147\ M\^58 + 13442\ M\^60 - 89218\ M\^62 - 
                    33541\ M\^64 + 66295\ M\^66 + 39388\ M\^68 - 
                    36214\ M\^70 - 28698\ M\^72 + 15178\ M\^74 + 
                    14847\ M\^76 - 6294\ M\^78 - 5701\ M\^80 + 2966\ M\^82 + 
                    1614\ M\^84 - 1928\ M\^86 + 792\ M\^88 - 173\ M\^90 + 
                    20\ M\^92 - M\^94)\) + 
              L\^4\ \((\(-4\)\ M\^28 + 39\ M\^30 - 154\ M\^32 + 281\ M\^34 - 
                    73\ M\^36 - 534\ M\^38 + 522\ M\^40 + 315\ M\^42 - 
                    333\ M\^44 + 885\ M\^46 - 3180\ M\^48 - 697\ M\^50 + 
                    8931\ M\^52 + 1081\ M\^54 - 16387\ M\^56 - 6094\ M\^58 + 
                    23567\ M\^60 + 13390\ M\^62 - 26538\ M\^64 - 
                    12939\ M\^66 + 23696\ M\^68 + 9207\ M\^70 - 
                    13350\ M\^72 - 3778\ M\^74 + 4535\ M\^76 + 962\ M\^78 - 
                    611\ M\^80 - 558\ M\^82 - 148\ M\^84 + 674\ M\^86 - 
                    448\ M\^88 + 144\ M\^90 - 25\ M\^92 + 2\ M\^94)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 21]\), \(L\^16\ M\^12 + 
              M\^132 + 
              L\^15\ \((\(-3\)\ M\^8 + 10\ M\^10 - 14\ M\^12 + 10\ M\^14 + 
                    2\ M\^16 - 23\ M\^18 + 45\ M\^20 + 2\ M\^22 - 25\ M\^24 + 
                    17\ M\^26 - 5\ M\^28)\) + 
              L\^14\ \((3\ M\^4 - 20\ M\^6 + 61\ M\^8 - 112\ M\^10 + 
                    97\ M\^12 + 133\ M\^14 - 454\ M\^16 + 196\ M\^18 + 
                    566\ M\^20 - 444\ M\^22 - 535\ M\^24 + 431\ M\^26 + 
                    715\ M\^28 - 490\ M\^30 - 385\ M\^32 + 565\ M\^34 - 
                    135\ M\^36 - 195\ M\^38 + 177\ M\^40 - 64\ M\^42 + 
                    10\ M\^44)\) + 
              L\^13\ \((\(-1\) + 10\ M\^2 - 48\ M\^4 + 146\ M\^6 - 
                    280\ M\^8 + 235\ M\^10 + 226\ M\^12 - 788\ M\^14 + 
                    914\ M\^16 - 451\ M\^18 - 1443\ M\^20 + 3943\ M\^22 - 
                    1077\ M\^24 - 6647\ M\^26 + 4905\ M\^28 + 6896\ M\^30 - 
                    7501\ M\^32 - 4781\ M\^34 + 8193\ M\^36 + 1374\ M\^38 - 
                    4805\ M\^40 + 1764\ M\^42 - 64\ M\^44 - 1071\ M\^46 + 
                    1738\ M\^48 - 840\ M\^50 - 393\ M\^52 + 680\ M\^54 - 
                    354\ M\^56 + 90\ M\^58 - 10\ M\^60)\) + 
              L\^12\ \((1 - 12\ M\^2 + 69\ M\^4 - 257\ M\^6 + 662\ M\^8 - 
                    1012\ M\^10 + 127\ M\^12 + 3042\ M\^14 - 6043\ M\^16 + 
                    2744\ M\^18 + 9912\ M\^20 - 23815\ M\^22 + 14588\ M\^24 + 
                    36859\ M\^26 - 77952\ M\^28 + 4504\ M\^30 + 
                    133802\ M\^32 - 104137\ M\^34 - 110992\ M\^36 + 
                    183130\ M\^38 + 10352\ M\^40 - 162344\ M\^42 + 
                    80309\ M\^44 + 67322\ M\^46 - 86195\ M\^48 + 
                    13486\ M\^50 + 36533\ M\^52 - 30141\ M\^54 + 
                    1546\ M\^56 + 10707\ M\^58 - 6162\ M\^60 + 1418\ M\^62 - 
                    113\ M\^64 - 683\ M\^66 + 1091\ M\^68 - 755\ M\^70 + 
                    280\ M\^72 - 56\ M\^74 + 5\ M\^76)\) + 
              L\^11\ \((\(-3\)\ M\^4 + 43\ M\^6 - 278\ M\^8 + 1026\ M\^10 - 
                    2198\ M\^12 + 1999\ M\^14 + 2814\ M\^16 - 12379\ M\^18 + 
                    17272\ M\^20 + 2771\ M\^22 - 56100\ M\^24 + 
                    80999\ M\^26 + 32544\ M\^28 - 212922\ M\^30 + 
                    137262\ M\^32 + 239476\ M\^34 - 371567\ M\^36 - 
                    49077\ M\^38 + 435390\ M\^40 - 219378\ M\^42 - 
                    219694\ M\^44 + 286648\ M\^46 - 91264\ M\^48 - 
                    69991\ M\^50 + 215568\ M\^52 - 200145\ M\^54 - 
                    78887\ M\^56 + 283212\ M\^58 - 114202\ M\^60 - 
                    166201\ M\^62 + 171368\ M\^64 + 17744\ M\^66 - 
                    103197\ M\^68 + 41029\ M\^70 + 23745\ M\^72 - 
                    26035\ M\^74 + 5362\ M\^76 + 3786\ M\^78 - 3190\ M\^80 + 
                    1477\ M\^82 - 666\ M\^84 + 273\ M\^86 - 78\ M\^88 + 
                    13\ M\^90 - M\^92)\) + 
              L\^10\ \((3\ M\^8 - 48\ M\^10 + 339\ M\^12 - 1344\ M\^14 + 
                    3044\ M\^16 - 2871\ M\^18 - 4017\ M\^20 + 16545\ M\^22 - 
                    18233\ M\^24 - 12863\ M\^26 + 63721\ M\^28 - 
                    53543\ M\^30 - 69119\ M\^32 + 136785\ M\^34 + 
                    17856\ M\^36 - 77929\ M\^38 - 103622\ M\^40 - 
                    217486\ M\^42 + 671936\ M\^44 + 551298\ M\^46 - 
                    1825760\ M\^48 - 512176\ M\^50 + 3123816\ M\^52 - 
                    116376\ M\^54 - 3824603\ M\^56 + 1031725\ M\^58 + 
                    3511799\ M\^60 - 1626787\ M\^62 - 2401243\ M\^64 + 
                    1592963\ M\^66 + 1169041\ M\^68 - 1089660\ M\^70 - 
                    356972\ M\^72 + 512994\ M\^74 + 45831\ M\^76 - 
                    149398\ M\^78 - 5848\ M\^80 + 20544\ M\^82 + 
                    20030\ M\^84 - 6719\ M\^86 - 13496\ M\^88 + 8577\ M\^90 + 
                    1556\ M\^92 - 3739\ M\^94 + 1881\ M\^96 - 488\ M\^98 + 
                    68\ M\^100 - 4\ M\^102)\) + 
              L\^9\ \((\(-M\^12\) + 17\ M\^14 - 130\ M\^16 + 569\ M\^18 - 
                    1506\ M\^20 + 2313\ M\^22 - 1672\ M\^24 - 409\ M\^26 + 
                    5507\ M\^28 - 23878\ M\^30 + 41881\ M\^32 + 
                    39265\ M\^34 - 244657\ M\^36 + 145738\ M\^38 + 
                    637940\ M\^40 - 865120\ M\^42 - 1078868\ M\^44 + 
                    2386508\ M\^46 + 1312971\ M\^48 - 4585143\ M\^50 - 
                    1277965\ M\^52 + 6829080\ M\^54 + 1276343\ M\^56 - 
                    8243874\ M\^58 - 1716308\ M\^60 + 8271634\ M\^62 + 
                    2633581\ M\^64 - 7104971\ M\^66 - 3590563\ M\^68 + 
                    5537217\ M\^70 + 4061547\ M\^72 - 4181062\ M\^74 - 
                    3708614\ M\^76 + 3157796\ M\^78 + 2645477\ M\^80 - 
                    2291546\ M\^82 - 1388195\ M\^84 + 1465513\ M\^86 + 
                    454172\ M\^88 - 752039\ M\^90 - 20417\ M\^92 + 
                    280956\ M\^94 - 69900\ M\^96 - 61512\ M\^98 + 
                    38553\ M\^100 + 348\ M\^102 - 8250\ M\^104 + 
                    3926\ M\^106 - 920\ M\^108 + 114\ M\^110 - 6\ M\^112)\) + 
              L\^8\ \((\(-4\)\ M\^22 + 76\ M\^24 - 624\ M\^26 + 2783\ M\^28 - 
                    6483\ M\^30 + 2927\ M\^32 + 24292\ M\^34 - 51557\ M\^36 - 
                    27789\ M\^38 + 210582\ M\^40 - 85108\ M\^42 - 
                    558993\ M\^44 + 507427\ M\^46 + 1180021\ M\^48 - 
                    1470608\ M\^50 - 2172201\ M\^52 + 3036204\ M\^54 + 
                    3703662\ M\^56 - 4829766\ M\^58 - 5908421\ M\^60 + 
                    5890919\ M\^62 + 8591608\ M\^64 - 5072807\ M\^66 - 
                    10973768\ M\^68 + 2034347\ M\^70 + 11959432\ M\^72 + 
                    2034347\ M\^74 - 10973768\ M\^76 - 5072807\ M\^78 + 
                    8591608\ M\^80 + 5890919\ M\^82 - 5908421\ M\^84 - 
                    4829766\ M\^86 + 3703662\ M\^88 + 3036204\ M\^90 - 
                    2172201\ M\^92 - 1470608\ M\^94 + 1180021\ M\^96 + 
                    507427\ M\^98 - 558993\ M\^100 - 85108\ M\^102 + 
                    210582\ M\^104 - 27789\ M\^106 - 51557\ M\^108 + 
                    24292\ M\^110 + 2927\ M\^112 - 6483\ M\^114 + 
                    2783\ M\^116 - 624\ M\^118 + 76\ M\^120 - 4\ M\^122)\) + 
              L\^7\ \((\(-6\)\ M\^32 + 114\ M\^34 - 920\ M\^36 + 
                    3926\ M\^38 - 8250\ M\^40 + 348\ M\^42 + 38553\ M\^44 - 
                    61512\ M\^46 - 69900\ M\^48 + 280956\ M\^50 - 
                    20417\ M\^52 - 752039\ M\^54 + 454172\ M\^56 + 
                    1465513\ M\^58 - 1388195\ M\^60 - 2291546\ M\^62 + 
                    2645477\ M\^64 + 3157796\ M\^66 - 3708614\ M\^68 - 
                    4181062\ M\^70 + 4061547\ M\^72 + 5537217\ M\^74 - 
                    3590563\ M\^76 - 7104971\ M\^78 + 2633581\ M\^80 + 
                    8271634\ M\^82 - 1716308\ M\^84 - 8243874\ M\^86 + 
                    1276343\ M\^88 + 6829080\ M\^90 - 1277965\ M\^92 - 
                    4585143\ M\^94 + 1312971\ M\^96 + 2386508\ M\^98 - 
                    1078868\ M\^100 - 865120\ M\^102 + 637940\ M\^104 + 
                    145738\ M\^106 - 244657\ M\^108 + 39265\ M\^110 + 
                    41881\ M\^112 - 23878\ M\^114 + 5507\ M\^116 - 
                    409\ M\^118 - 1672\ M\^120 + 2313\ M\^122 - 
                    1506\ M\^124 + 569\ M\^126 - 130\ M\^128 + 17\ M\^130 - 
                    M\^132)\) + 
              L\ \((\(-5\)\ M\^116 + 17\ M\^118 - 25\ M\^120 + 2\ M\^122 + 
                    45\ M\^124 - 23\ M\^126 + 2\ M\^128 + 10\ M\^130 - 
                    14\ M\^132 + 10\ M\^134 - 3\ M\^136)\) + 
              L\^6\ \((\(-4\)\ M\^42 + 68\ M\^44 - 488\ M\^46 + 1881\ M\^48 - 
                    3739\ M\^50 + 1556\ M\^52 + 8577\ M\^54 - 13496\ M\^56 - 
                    6719\ M\^58 + 20030\ M\^60 + 20544\ M\^62 - 5848\ M\^64 - 
                    149398\ M\^66 + 45831\ M\^68 + 512994\ M\^70 - 
                    356972\ M\^72 - 1089660\ M\^74 + 1169041\ M\^76 + 
                    1592963\ M\^78 - 2401243\ M\^80 - 1626787\ M\^82 + 
                    3511799\ M\^84 + 1031725\ M\^86 - 3824603\ M\^88 - 
                    116376\ M\^90 + 3123816\ M\^92 - 512176\ M\^94 - 
                    1825760\ M\^96 + 551298\ M\^98 + 671936\ M\^100 - 
                    217486\ M\^102 - 103622\ M\^104 - 77929\ M\^106 + 
                    17856\ M\^108 + 136785\ M\^110 - 69119\ M\^112 - 
                    53543\ M\^114 + 63721\ M\^116 - 12863\ M\^118 - 
                    18233\ M\^120 + 16545\ M\^122 - 4017\ M\^124 - 
                    2871\ M\^126 + 3044\ M\^128 - 1344\ M\^130 + 
                    339\ M\^132 - 48\ M\^134 + 3\ M\^136)\) + 
              L\^5\ \((\(-M\^52\) + 13\ M\^54 - 78\ M\^56 + 273\ M\^58 - 
                    666\ M\^60 + 1477\ M\^62 - 3190\ M\^64 + 3786\ M\^66 + 
                    5362\ M\^68 - 26035\ M\^70 + 23745\ M\^72 + 
                    41029\ M\^74 - 103197\ M\^76 + 17744\ M\^78 + 
                    171368\ M\^80 - 166201\ M\^82 - 114202\ M\^84 + 
                    283212\ M\^86 - 78887\ M\^88 - 200145\ M\^90 + 
                    215568\ M\^92 - 69991\ M\^94 - 91264\ M\^96 + 
                    286648\ M\^98 - 219694\ M\^100 - 219378\ M\^102 + 
                    435390\ M\^104 - 49077\ M\^106 - 371567\ M\^108 + 
                    239476\ M\^110 + 137262\ M\^112 - 212922\ M\^114 + 
                    32544\ M\^116 + 80999\ M\^118 - 56100\ M\^120 + 
                    2771\ M\^122 + 17272\ M\^124 - 12379\ M\^126 + 
                    2814\ M\^128 + 1999\ M\^130 - 2198\ M\^132 + 
                    1026\ M\^134 - 278\ M\^136 + 43\ M\^138 - 3\ M\^140)\) + 
              L\^2\ \((10\ M\^100 - 64\ M\^102 + 177\ M\^104 - 195\ M\^106 - 
                    135\ M\^108 + 565\ M\^110 - 385\ M\^112 - 490\ M\^114 + 
                    715\ M\^116 + 431\ M\^118 - 535\ M\^120 - 444\ M\^122 + 
                    566\ M\^124 + 196\ M\^126 - 454\ M\^128 + 133\ M\^130 + 
                    97\ M\^132 - 112\ M\^134 + 61\ M\^136 - 20\ M\^138 + 
                    3\ M\^140)\) + 
              L\^3\ \((\(-10\)\ M\^84 + 90\ M\^86 - 354\ M\^88 + 680\ M\^90 - 
                    393\ M\^92 - 840\ M\^94 + 1738\ M\^96 - 1071\ M\^98 - 
                    64\ M\^100 + 1764\ M\^102 - 4805\ M\^104 + 1374\ M\^106 + 
                    8193\ M\^108 - 4781\ M\^110 - 7501\ M\^112 + 
                    6896\ M\^114 + 4905\ M\^116 - 6647\ M\^118 - 
                    1077\ M\^120 + 3943\ M\^122 - 1443\ M\^124 - 
                    451\ M\^126 + 914\ M\^128 - 788\ M\^130 + 226\ M\^132 + 
                    235\ M\^134 - 280\ M\^136 + 146\ M\^138 - 48\ M\^140 + 
                    10\ M\^142 - M\^144)\) + 
              L\^4\ \((5\ M\^68 - 56\ M\^70 + 280\ M\^72 - 755\ M\^74 + 
                    1091\ M\^76 - 683\ M\^78 - 113\ M\^80 + 1418\ M\^82 - 
                    6162\ M\^84 + 10707\ M\^86 + 1546\ M\^88 - 30141\ M\^90 + 
                    36533\ M\^92 + 13486\ M\^94 - 86195\ M\^96 + 
                    67322\ M\^98 + 80309\ M\^100 - 162344\ M\^102 + 
                    10352\ M\^104 + 183130\ M\^106 - 110992\ M\^108 - 
                    104137\ M\^110 + 133802\ M\^112 + 4504\ M\^114 - 
                    77952\ M\^116 + 36859\ M\^118 + 14588\ M\^120 - 
                    23815\ M\^122 + 9912\ M\^124 + 2744\ M\^126 - 
                    6043\ M\^128 + 3042\ M\^130 + 127\ M\^132 - 
                    1012\ M\^134 + 662\ M\^136 - 257\ M\^138 + 69\ M\^140 - 
                    12\ M\^142 + M\^144)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 22]\), \(M\^66 + L\^24\ M\^66 + 
              L\^23\ \((\(-3\)\ M\^54 + 10\ M\^56 - 19\ M\^58 + 18\ M\^60 + 
                    5\ M\^62 - 34\ M\^64 + 53\ M\^66 + 10\ M\^68 - 
                    30\ M\^70 + 20\ M\^72 - 6\ M\^74)\) + 
              L\ \((\(-6\)\ M\^58 + 20\ M\^60 - 30\ M\^62 + 10\ M\^64 + 
                    53\ M\^66 - 34\ M\^68 + 5\ M\^70 + 18\ M\^72 - 
                    19\ M\^74 + 10\ M\^76 - 3\ M\^78)\) + 
              L\^22\ \((3\ M\^42 - 20\ M\^44 + 70\ M\^46 - 143\ M\^48 + 
                    140\ M\^50 + 71\ M\^52 - 369\ M\^54 + 244\ M\^56 + 
                    198\ M\^58 - 173\ M\^60 - 158\ M\^62 - 145\ M\^64 + 
                    855\ M\^66 - 47\ M\^68 - 695\ M\^70 + 642\ M\^72 - 
                    43\ M\^74 - 335\ M\^76 + 260\ M\^78 - 94\ M\^80 + 
                    15\ M\^82)\) + 
              L\^21\ \((\(-M\^30\) + 10\ M\^32 - 51\ M\^34 + 162\ M\^36 - 
                    345\ M\^38 + 472\ M\^40 - 316\ M\^42 - 196\ M\^44 + 
                    923\ M\^46 - 1439\ M\^48 + 914\ M\^50 + 1065\ M\^52 - 
                    3089\ M\^54 + 2856\ M\^56 - 1561\ M\^58 - 1454\ M\^60 + 
                    6863\ M\^62 - 6546\ M\^64 - 1293\ M\^66 + 9904\ M\^68 - 
                    4648\ M\^70 - 1262\ M\^72 + 2366\ M\^74 - 2808\ M\^76 + 
                    2292\ M\^78 - 343\ M\^80 - 1347\ M\^82 + 1427\ M\^84 - 
                    687\ M\^86 + 176\ M\^88 - 20\ M\^90)\) + 
              L\^2\ \((15\ M\^50 - 94\ M\^52 + 260\ M\^54 - 335\ M\^56 - 
                    43\ M\^58 + 642\ M\^60 - 695\ M\^62 - 47\ M\^64 + 
                    855\ M\^66 - 145\ M\^68 - 158\ M\^70 - 173\ M\^72 + 
                    198\ M\^74 + 244\ M\^76 - 369\ M\^78 + 71\ M\^80 + 
                    140\ M\^82 - 143\ M\^84 + 70\ M\^86 - 20\ M\^88 + 
                    3\ M\^90)\) + 
              L\^20\ \((\(-5\)\ M\^24 + 60\ M\^26 - 341\ M\^28 + 
                    1169\ M\^30 - 2477\ M\^32 + 2562\ M\^34 + 1865\ M\^36 - 
                    10718\ M\^38 + 13358\ M\^40 + 5529\ M\^42 - 
                    39754\ M\^44 + 44476\ M\^46 + 24172\ M\^48 - 
                    115487\ M\^50 + 76688\ M\^52 + 126904\ M\^54 - 
                    221612\ M\^56 - 31752\ M\^58 + 296662\ M\^60 - 
                    111805\ M\^62 - 246279\ M\^64 + 215851\ M\^66 + 
                    119911\ M\^68 - 209475\ M\^70 + 19403\ M\^72 + 
                    121824\ M\^74 - 77209\ M\^76 - 20445\ M\^78 + 
                    43890\ M\^80 - 17589\ M\^82 - 1242\ M\^84 + 4849\ M\^86 - 
                    4622\ M\^88 + 3853\ M\^90 - 2244\ M\^92 + 805\ M\^94 - 
                    164\ M\^96 + 15\ M\^98)\) + 
              L\^3\ \((\(-20\)\ M\^42 + 176\ M\^44 - 687\ M\^46 + 
                    1427\ M\^48 - 1347\ M\^50 - 343\ M\^52 + 2292\ M\^54 - 
                    2808\ M\^56 + 2366\ M\^58 - 1262\ M\^60 - 4648\ M\^62 + 
                    9904\ M\^64 - 1293\ M\^66 - 6546\ M\^68 + 6863\ M\^70 - 
                    1454\ M\^72 - 1561\ M\^74 + 2856\ M\^76 - 3089\ M\^78 + 
                    1065\ M\^80 + 914\ M\^82 - 1439\ M\^84 + 923\ M\^86 - 
                    196\ M\^88 - 316\ M\^90 + 472\ M\^92 - 345\ M\^94 + 
                    162\ M\^96 - 51\ M\^98 + 10\ M\^100 - M\^102)\) + 
              L\^19\ \((\(-10\)\ M\^18 + 139\ M\^20 - 889\ M\^22 + 
                    3384\ M\^24 - 7933\ M\^26 + 9139\ M\^28 + 6826\ M\^30 - 
                    44354\ M\^32 + 58716\ M\^34 + 34217\ M\^36 - 
                    214847\ M\^38 + 218164\ M\^40 + 224913\ M\^42 - 
                    747390\ M\^44 + 321563\ M\^46 + 1146194\ M\^48 - 
                    1525705\ M\^50 - 763712\ M\^52 + 2758940\ M\^54 - 
                    495816\ M\^56 - 3159270\ M\^58 + 1869266\ M\^60 + 
                    2438229\ M\^62 - 2424203\ M\^64 - 1141660\ M\^66 + 
                    1989167\ M\^68 + 163904\ M\^70 - 999122\ M\^72 + 
                    175060\ M\^74 + 212650\ M\^76 - 41667\ M\^78 + 
                    45542\ M\^80 - 128654\ M\^82 + 27283\ M\^84 + 
                    92716\ M\^86 - 73680\ M\^88 + 2645\ M\^90 + 
                    25273\ M\^92 - 18596\ M\^94 + 8763\ M\^96 - 3836\ M\^98 + 
                    1524\ M\^100 - 439\ M\^102 + 76\ M\^104 - 6\ M\^106)\) + 
              L\^4\ \((15\ M\^34 - 164\ M\^36 + 805\ M\^38 - 2244\ M\^40 + 
                    3853\ M\^42 - 4622\ M\^44 + 4849\ M\^46 - 1242\ M\^48 - 
                    17589\ M\^50 + 43890\ M\^52 - 20445\ M\^54 - 
                    77209\ M\^56 + 121824\ M\^58 + 19403\ M\^60 - 
                    209475\ M\^62 + 119911\ M\^64 + 215851\ M\^66 - 
                    246279\ M\^68 - 111805\ M\^70 + 296662\ M\^72 - 
                    31752\ M\^74 - 221612\ M\^76 + 126904\ M\^78 + 
                    76688\ M\^80 - 115487\ M\^82 + 24172\ M\^84 + 
                    44476\ M\^86 - 39754\ M\^88 + 5529\ M\^90 + 
                    13358\ M\^92 - 10718\ M\^94 + 1865\ M\^96 + 2562\ M\^98 - 
                    2477\ M\^100 + 1169\ M\^102 - 341\ M\^104 + 60\ M\^106 - 
                    5\ M\^108)\) + 
              L\^5\ \((\(-6\)\ M\^26 + 76\ M\^28 - 439\ M\^30 + 1524\ M\^32 - 
                    3836\ M\^34 + 8763\ M\^36 - 18596\ M\^38 + 25273\ M\^40 + 
                    2645\ M\^42 - 73680\ M\^44 + 92716\ M\^46 + 
                    27283\ M\^48 - 128654\ M\^50 + 45542\ M\^52 - 
                    41667\ M\^54 + 212650\ M\^56 + 175060\ M\^58 - 
                    999122\ M\^60 + 163904\ M\^62 + 1989167\ M\^64 - 
                    1141660\ M\^66 - 2424203\ M\^68 + 2438229\ M\^70 + 
                    1869266\ M\^72 - 3159270\ M\^74 - 495816\ M\^76 + 
                    2758940\ M\^78 - 763712\ M\^80 - 1525705\ M\^82 + 
                    1146194\ M\^84 + 321563\ M\^86 - 747390\ M\^88 + 
                    224913\ M\^90 + 218164\ M\^92 - 214847\ M\^94 + 
                    34217\ M\^96 + 58716\ M\^98 - 44354\ M\^100 + 
                    6826\ M\^102 + 9139\ M\^104 - 7933\ M\^106 + 
                    3384\ M\^108 - 889\ M\^110 + 139\ M\^112 - 
                    10\ M\^114)\) + 
              L\^18\ \((\(-10\)\ M\^12 + 157\ M\^14 - 1124\ M\^16 + 
                    4761\ M\^18 - 12487\ M\^20 + 17135\ M\^22 + 5271\ M\^24 - 
                    68213\ M\^26 + 94440\ M\^28 + 80126\ M\^30 - 
                    404740\ M\^32 + 294166\ M\^34 + 746534\ M\^36 - 
                    1553177\ M\^38 - 304617\ M\^40 + 3860513\ M\^42 - 
                    2493117\ M\^44 - 5877338\ M\^46 + 8433500\ M\^48 + 
                    4955134\ M\^50 - 15512828\ M\^52 + 392253\ M\^54 + 
                    19301714\ M\^56 - 7685418\ M\^58 - 17096658\ M\^60 + 
                    11817500\ M\^62 + 10759621\ M\^64 - 10003320\ M\^66 - 
                    4749419\ M\^68 + 4863263\ M\^70 + 2228720\ M\^72 - 
                    595780\ M\^74 - 2178826\ M\^76 - 625659\ M\^78 + 
                    2175249\ M\^80 - 7079\ M\^82 - 1296406\ M\^84 + 
                    485186\ M\^86 + 278764\ M\^88 - 256347\ M\^90 + 
                    79832\ M\^92 - 20133\ M\^94 - 13596\ M\^96 + 
                    38683\ M\^98 - 33952\ M\^100 + 16616\ M\^102 - 
                    5460\ M\^104 + 1475\ M\^106 - 391\ M\^108 + 91\ M\^110 - 
                    14\ M\^112 + M\^114)\) + 
              L\^17\ \((\(-5\)\ M\^6 + 87\ M\^8 - 690\ M\^10 + 3240\ M\^12 - 
                    9522\ M\^14 + 15798\ M\^16 - 4834\ M\^18 - 34132\ M\^20 + 
                    37638\ M\^22 + 117295\ M\^24 - 305144\ M\^26 - 
                    85484\ M\^28 + 1204236\ M\^30 - 1059132\ M\^32 - 
                    2653320\ M\^34 + 5716189\ M\^36 + 1758698\ M\^38 - 
                    15243692\ M\^40 + 7423599\ M\^42 + 26023323\ M\^44 - 
                    29575747\ M\^46 - 28422152\ M\^48 + 60355564\ M\^50 + 
                    14373569\ M\^52 - 85256177\ M\^54 + 12757295\ M\^56 + 
                    89745215\ M\^58 - 37545733\ M\^60 - 73006624\ M\^62 + 
                    44988168\ M\^64 + 47977612\ M\^66 - 33712696\ M\^68 - 
                    27591266\ M\^70 + 17037548\ M\^72 + 16407519\ M\^74 - 
                    6307675\ M\^76 - 10047782\ M\^78 + 2888972\ M\^80 + 
                    4531216\ M\^82 - 1891594\ M\^84 - 487081\ M\^86 + 
                    471038\ M\^88 - 774753\ M\^90 + 632809\ M\^92 + 
                    209086\ M\^94 - 531966\ M\^96 + 213093\ M\^98 + 
                    73605\ M\^100 - 114382\ M\^102 + 66855\ M\^104 - 
                    31399\ M\^106 + 13494\ M\^108 - 4604\ M\^110 + 
                    1066\ M\^112 - 146\ M\^114 + 9\ M\^116)\) + 
              L\^16\ \((\(-1\) + 19\ M\^2 - 165\ M\^4 + 854\ M\^6 - 
                    2835\ M\^8 + 5990\ M\^10 - 7931\ M\^12 + 11446\ M\^14 - 
                    41203\ M\^16 + 106857\ M\^18 - 53543\ M\^20 - 
                    395339\ M\^22 + 840407\ M\^24 + 391065\ M\^26 - 
                    3499441\ M\^28 + 2635334\ M\^30 + 7900668\ M\^32 - 
                    14424192\ M\^34 - 8870626\ M\^36 + 41436330\ M\^38 - 
                    8175332\ M\^40 - 80516525\ M\^42 + 60621300\ M\^44 + 
                    110886752\ M\^46 - 150342139\ M\^48 - 102728578\ M\^50 + 
                    248992132\ M\^52 + 44760870\ M\^54 - 311564805\ M\^56 + 
                    39261708\ M\^58 + 310541405\ M\^60 - 107377568\ M\^62 - 
                    257596850\ M\^64 + 132957176\ M\^66 + 186431846\ M\^68 - 
                    118693507\ M\^70 - 118673766\ M\^72 + 88472634\ M\^74 + 
                    62281212\ M\^76 - 58390037\ M\^78 - 20301516\ M\^80 + 
                    32152510\ M\^82 - 3286710\ M\^84 - 11308373\ M\^86 + 
                    8679426\ M\^88 - 390093\ M\^90 - 4509088\ M\^92 + 
                    2896456\ M\^94 + 323555\ M\^96 - 1118130\ M\^98 + 
                    499677\ M\^100 - 82615\ M\^102 - 46223\ M\^104 + 
                    112341\ M\^106 - 115033\ M\^108 + 66849\ M\^110 - 
                    23770\ M\^112 + 5201\ M\^114 - 651\ M\^116 + 
                    36\ M\^118)\) + 
              L\^6\ \((M\^18 - 14\ M\^20 + 91\ M\^22 - 391\ M\^24 + 
                    1475\ M\^26 - 5460\ M\^28 + 16616\ M\^30 - 33952\ M\^32 + 
                    38683\ M\^34 - 13596\ M\^36 - 20133\ M\^38 + 
                    79832\ M\^40 - 256347\ M\^42 + 278764\ M\^44 + 
                    485186\ M\^46 - 1296406\ M\^48 - 7079\ M\^50 + 
                    2175249\ M\^52 - 625659\ M\^54 - 2178826\ M\^56 - 
                    595780\ M\^58 + 2228720\ M\^60 + 4863263\ M\^62 - 
                    4749419\ M\^64 - 10003320\ M\^66 + 10759621\ M\^68 + 
                    11817500\ M\^70 - 17096658\ M\^72 - 7685418\ M\^74 + 
                    19301714\ M\^76 + 392253\ M\^78 - 15512828\ M\^80 + 
                    4955134\ M\^82 + 8433500\ M\^84 - 5877338\ M\^86 - 
                    2493117\ M\^88 + 3860513\ M\^90 - 304617\ M\^92 - 
                    1553177\ M\^94 + 746534\ M\^96 + 294166\ M\^98 - 
                    404740\ M\^100 + 80126\ M\^102 + 94440\ M\^104 - 
                    68213\ M\^106 + 5271\ M\^108 + 17135\ M\^110 - 
                    12487\ M\^112 + 4761\ M\^114 - 1124\ M\^116 + 
                    157\ M\^118 - 10\ M\^120)\) + 
              L\^15\ \((1 - 28\ M\^2 + 345\ M\^4 - 2472\ M\^6 + 11189\ M\^8 - 
                    31621\ M\^10 + 46598\ M\^12 + 7283\ M\^14 - 
                    140410\ M\^16 + 47415\ M\^18 + 686344\ M\^20 - 
                    1122589\ M\^22 - 1474112\ M\^24 + 5947104\ M\^26 - 
                    1091886\ M\^28 - 18146633\ M\^30 + 19645814\ M\^32 + 
                    32921907\ M\^34 - 73375338\ M\^36 - 26121884\ M\^38 + 
                    172246747\ M\^40 - 45235507\ M\^42 - 291797835\ M\^44 + 
                    212371506\ M\^46 + 371710791\ M\^48 - 455284534\ M\^50 - 
                    353490021\ M\^52 + 696006900\ M\^54 + 231656503\ M\^56 - 
                    845982905\ M\^58 - 59365679\ M\^60 + 862544581\ M\^62 - 
                    98190343\ M\^64 - 765229172\ M\^66 + 204811715\ M\^68 + 
                    602987318\ M\^70 - 247058158\ M\^72 - 411325848\ M\^74 + 
                    231608889\ M\^76 + 225773462\ M\^78 - 173912866\ M\^80 - 
                    85998741\ M\^82 + 100153758\ M\^84 + 13203491\ M\^86 - 
                    40352503\ M\^88 + 5710617\ M\^90 + 10100337\ M\^92 - 
                    2944248\ M\^94 - 2218604\ M\^96 - 20707\ M\^98 + 
                    1836074\ M\^100 - 435056\ M\^102 - 979178\ M\^104 + 
                    729358\ M\^106 + 28044\ M\^108 - 298917\ M\^110 + 
                    197085\ M\^112 - 68434\ M\^114 + 14119\ M\^116 - 
                    1646\ M\^118 + 84\ M\^120)\) + 
              L\^14\ \((9\ M\^2 - 216\ M\^4 + 2359\ M\^6 - 15176\ M\^8 + 
                    61483\ M\^10 - 149662\ M\^12 + 145566\ M\^14 + 
                    297173\ M\^16 - 1089411\ M\^18 + 404464\ M\^20 + 
                    3784515\ M\^22 - 6328835\ M\^24 - 5711698\ M\^26 + 
                    25989016\ M\^28 - 7117235\ M\^30 - 65967705\ M\^32 + 
                    69119149\ M\^34 + 108850939\ M\^36 - 221100375\ M\^38 - 
                    97890957\ M\^40 + 472020614\ M\^42 - 44073143\ M\^44 - 
                    763520290\ M\^46 + 352582966\ M\^48 + 990499011\ M\^50 - 
                    776706564\ M\^52 - 1068070790\ M\^54 + 
                    1200310200\ M\^56 + 983623759\ M\^58 - 
                    1510346836\ M\^60 - 780708145\ M\^62 + 
                    1637997343\ M\^64 + 509991016\ M\^66 - 
                    1564717859\ M\^68 - 217770239\ M\^70 + 
                    1322694194\ M\^72 - 21489501\ M\^74 - 973260752\ M\^76 + 
                    148718627\ M\^78 + 609963520\ M\^80 - 163388815\ M\^82 - 
                    323803113\ M\^84 + 118439931\ M\^86 + 150035917\ M\^88 - 
                    70149249\ M\^90 - 64043830\ M\^92 + 40919260\ M\^94 + 
                    23320752\ M\^96 - 24065010\ M\^98 - 3881738\ M\^100 + 
                    11622301\ M\^102 - 2860801\ M\^104 - 3044082\ M\^106 + 
                    2283471\ M\^108 - 152991\ M\^110 - 558453\ M\^112 + 
                    367591\ M\^114 - 121697\ M\^116 + 23770\ M\^118 - 
                    2618\ M\^120 + 126\ M\^122)\) + 
              L\^13\ \((36\ M\^4 - 826\ M\^6 + 8521\ M\^8 - 51060\ M\^10 + 
                    188846\ M\^12 - 397979\ M\^14 + 202122\ M\^16 + 
                    1304521\ M\^18 - 3281260\ M\^20 + 122452\ M\^22 + 
                    12062523\ M\^24 - 15826261\ M\^26 - 20137732\ M\^28 + 
                    65589406\ M\^30 - 2898223\ M\^32 - 161060982\ M\^34 + 
                    117551371\ M\^36 + 273252144\ M\^38 - 382365342\ M\^40 - 
                    325221923\ M\^42 + 803178572\ M\^44 + 231724240\ M\^46 - 
                    1308499399\ M\^48 + 35969373\ M\^50 + 1786770877\ M\^52 - 
                    434547172\ M\^54 - 2138482749\ M\^56 + 888120224\ M\^58 + 
                    2297558186\ M\^60 - 1320692667\ M\^62 - 
                    2226202273\ M\^64 + 1639296607\ M\^66 + 
                    1922183911\ M\^68 - 1743960746\ M\^70 - 
                    1441721373\ M\^72 + 1610985048\ M\^74 + 
                    926136024\ M\^76 - 1305366324\ M\^78 - 507230005\ M\^80 + 
                    941859714\ M\^82 + 227769701\ M\^84 - 616208834\ M\^86 - 
                    60371443\ M\^88 + 364718451\ M\^90 - 28397048\ M\^92 - 
                    185456113\ M\^94 + 57358040\ M\^96 + 70603801\ M\^98 - 
                    45839626\ M\^100 - 13343196\ M\^102 + 21863886\ M\^104 - 
                    3882363\ M\^106 - 5225698\ M\^108 + 3348311\ M\^110 - 
                    158594\ M\^112 - 735196\ M\^114 + 448202\ M\^116 - 
                    139920\ M\^118 + 25965\ M\^120 - 2730\ M\^122 + 
                    126\ M\^124)\) + 
              L\^7\ \((9\ M\^16 - 146\ M\^18 + 1066\ M\^20 - 4604\ M\^22 + 
                    13494\ M\^24 - 31399\ M\^26 + 66855\ M\^28 - 
                    114382\ M\^30 + 73605\ M\^32 + 213093\ M\^34 - 
                    531966\ M\^36 + 209086\ M\^38 + 632809\ M\^40 - 
                    774753\ M\^42 + 471038\ M\^44 - 487081\ M\^46 - 
                    1891594\ M\^48 + 4531216\ M\^50 + 2888972\ M\^52 - 
                    10047782\ M\^54 - 6307675\ M\^56 + 16407519\ M\^58 + 
                    17037548\ M\^60 - 27591266\ M\^62 - 33712696\ M\^64 + 
                    47977612\ M\^66 + 44988168\ M\^68 - 73006624\ M\^70 - 
                    37545733\ M\^72 + 89745215\ M\^74 + 12757295\ M\^76 - 
                    85256177\ M\^78 + 14373569\ M\^80 + 60355564\ M\^82 - 
                    28422152\ M\^84 - 29575747\ M\^86 + 26023323\ M\^88 + 
                    7423599\ M\^90 - 15243692\ M\^92 + 1758698\ M\^94 + 
                    5716189\ M\^96 - 2653320\ M\^98 - 1059132\ M\^100 + 
                    1204236\ M\^102 - 85484\ M\^104 - 305144\ M\^106 + 
                    117295\ M\^108 + 37638\ M\^110 - 34132\ M\^112 - 
                    4834\ M\^114 + 15798\ M\^116 - 9522\ M\^118 + 
                    3240\ M\^120 - 690\ M\^122 + 87\ M\^124 - 5\ M\^126)\) + 
              L\^12\ \((84\ M\^6 - 1876\ M\^8 + 18576\ M\^10 - 
                    105297\ M\^12 + 360640\ M\^14 - 664113\ M\^16 + 
                    59100\ M\^18 + 2778191\ M\^20 - 5375788\ M\^22 - 
                    1948884\ M\^24 + 21578964\ M\^26 - 20240068\ M\^28 - 
                    41623361\ M\^30 + 92644279\ M\^32 + 29806317\ M\^34 - 
                    233793463\ M\^36 + 76273109\ M\^38 + 424090250\ M\^40 - 
                    328416923\ M\^42 - 612644430\ M\^44 + 736141931\ M\^46 + 
                    749420657\ M\^48 - 1268054944\ M\^50 - 805839799\ M\^52 + 
                    1871417713\ M\^54 + 754549653\ M\^56 - 
                    2459531091\ M\^58 - 558618136\ M\^60 + 
                    2899751880\ M\^62 + 210373036\ M\^64 - 
                    3062108258\ M\^66 + 210373036\ M\^68 + 
                    2899751880\ M\^70 - 558618136\ M\^72 - 
                    2459531091\ M\^74 + 754549653\ M\^76 + 
                    1871417713\ M\^78 - 805839799\ M\^80 - 
                    1268054944\ M\^82 + 749420657\ M\^84 + 736141931\ M\^86 - 
                    612644430\ M\^88 - 328416923\ M\^90 + 424090250\ M\^92 + 
                    76273109\ M\^94 - 233793463\ M\^96 + 29806317\ M\^98 + 
                    92644279\ M\^100 - 41623361\ M\^102 - 20240068\ M\^104 + 
                    21578964\ M\^106 - 1948884\ M\^108 - 5375788\ M\^110 + 
                    2778191\ M\^112 + 59100\ M\^114 - 664113\ M\^116 + 
                    360640\ M\^118 - 105297\ M\^120 + 18576\ M\^122 - 
                    1876\ M\^124 + 84\ M\^126)\) + 
              L\^11\ \((126\ M\^8 - 2730\ M\^10 + 25965\ M\^12 - 
                    139920\ M\^14 + 448202\ M\^16 - 735196\ M\^18 - 
                    158594\ M\^20 + 3348311\ M\^22 - 5225698\ M\^24 - 
                    3882363\ M\^26 + 21863886\ M\^28 - 13343196\ M\^30 - 
                    45839626\ M\^32 + 70603801\ M\^34 + 57358040\ M\^36 - 
                    185456113\ M\^38 - 28397048\ M\^40 + 364718451\ M\^42 - 
                    60371443\ M\^44 - 616208834\ M\^46 + 227769701\ M\^48 + 
                    941859714\ M\^50 - 507230005\ M\^52 - 1305366324\ M\^54 + 
                    926136024\ M\^56 + 1610985048\ M\^58 - 
                    1441721373\ M\^60 - 1743960746\ M\^62 + 
                    1922183911\ M\^64 + 1639296607\ M\^66 - 
                    2226202273\ M\^68 - 1320692667\ M\^70 + 
                    2297558186\ M\^72 + 888120224\ M\^74 - 
                    2138482749\ M\^76 - 434547172\ M\^78 + 
                    1786770877\ M\^80 + 35969373\ M\^82 - 1308499399\ M\^84 + 
                    231724240\ M\^86 + 803178572\ M\^88 - 325221923\ M\^90 - 
                    382365342\ M\^92 + 273252144\ M\^94 + 117551371\ M\^96 - 
                    161060982\ M\^98 - 2898223\ M\^100 + 65589406\ M\^102 - 
                    20137732\ M\^104 - 15826261\ M\^106 + 12062523\ M\^108 + 
                    122452\ M\^110 - 3281260\ M\^112 + 1304521\ M\^114 + 
                    202122\ M\^116 - 397979\ M\^118 + 188846\ M\^120 - 
                    51060\ M\^122 + 8521\ M\^124 - 826\ M\^126 + 
                    36\ M\^128)\) + 
              L\^10\ \((126\ M\^10 - 2618\ M\^12 + 23770\ M\^14 - 
                    121697\ M\^16 + 367591\ M\^18 - 558453\ M\^20 - 
                    152991\ M\^22 + 2283471\ M\^24 - 3044082\ M\^26 - 
                    2860801\ M\^28 + 11622301\ M\^30 - 3881738\ M\^32 - 
                    24065010\ M\^34 + 23320752\ M\^36 + 40919260\ M\^38 - 
                    64043830\ M\^40 - 70149249\ M\^42 + 150035917\ M\^44 + 
                    118439931\ M\^46 - 323803113\ M\^48 - 163388815\ M\^50 + 
                    609963520\ M\^52 + 148718627\ M\^54 - 973260752\ M\^56 - 
                    21489501\ M\^58 + 1322694194\ M\^60 - 217770239\ M\^62 - 
                    1564717859\ M\^64 + 509991016\ M\^66 + 
                    1637997343\ M\^68 - 780708145\ M\^70 - 
                    1510346836\ M\^72 + 983623759\ M\^74 + 
                    1200310200\ M\^76 - 1068070790\ M\^78 - 
                    776706564\ M\^80 + 990499011\ M\^82 + 352582966\ M\^84 - 
                    763520290\ M\^86 - 44073143\ M\^88 + 472020614\ M\^90 - 
                    97890957\ M\^92 - 221100375\ M\^94 + 108850939\ M\^96 + 
                    69119149\ M\^98 - 65967705\ M\^100 - 7117235\ M\^102 + 
                    25989016\ M\^104 - 5711698\ M\^106 - 6328835\ M\^108 + 
                    3784515\ M\^110 + 404464\ M\^112 - 1089411\ M\^114 + 
                    297173\ M\^116 + 145566\ M\^118 - 149662\ M\^120 + 
                    61483\ M\^122 - 15176\ M\^124 + 2359\ M\^126 - 
                    216\ M\^128 + 9\ M\^130)\) + 
              L\^8\ \((36\ M\^14 - 651\ M\^16 + 5201\ M\^18 - 23770\ M\^20 + 
                    66849\ M\^22 - 115033\ M\^24 + 112341\ M\^26 - 
                    46223\ M\^28 - 82615\ M\^30 + 499677\ M\^32 - 
                    1118130\ M\^34 + 323555\ M\^36 + 2896456\ M\^38 - 
                    4509088\ M\^40 - 390093\ M\^42 + 8679426\ M\^44 - 
                    11308373\ M\^46 - 3286710\ M\^48 + 32152510\ M\^50 - 
                    20301516\ M\^52 - 58390037\ M\^54 + 62281212\ M\^56 + 
                    88472634\ M\^58 - 118673766\ M\^60 - 118693507\ M\^62 + 
                    186431846\ M\^64 + 132957176\ M\^66 - 257596850\ M\^68 - 
                    107377568\ M\^70 + 310541405\ M\^72 + 39261708\ M\^74 - 
                    311564805\ M\^76 + 44760870\ M\^78 + 248992132\ M\^80 - 
                    102728578\ M\^82 - 150342139\ M\^84 + 110886752\ M\^86 + 
                    60621300\ M\^88 - 80516525\ M\^90 - 8175332\ M\^92 + 
                    41436330\ M\^94 - 8870626\ M\^96 - 14424192\ M\^98 + 
                    7900668\ M\^100 + 2635334\ M\^102 - 3499441\ M\^104 + 
                    391065\ M\^106 + 840407\ M\^108 - 395339\ M\^110 - 
                    53543\ M\^112 + 106857\ M\^114 - 41203\ M\^116 + 
                    11446\ M\^118 - 7931\ M\^120 + 5990\ M\^122 - 
                    2835\ M\^124 + 854\ M\^126 - 165\ M\^128 + 19\ M\^130 - 
                    M\^132)\) + 
              L\^9\ \((84\ M\^12 - 1646\ M\^14 + 14119\ M\^16 - 
                    68434\ M\^18 + 197085\ M\^20 - 298917\ M\^22 + 
                    28044\ M\^24 + 729358\ M\^26 - 979178\ M\^28 - 
                    435056\ M\^30 + 1836074\ M\^32 - 20707\ M\^34 - 
                    2218604\ M\^36 - 2944248\ M\^38 + 10100337\ M\^40 + 
                    5710617\ M\^42 - 40352503\ M\^44 + 13203491\ M\^46 + 
                    100153758\ M\^48 - 85998741\ M\^50 - 173912866\ M\^52 + 
                    225773462\ M\^54 + 231608889\ M\^56 - 411325848\ M\^58 - 
                    247058158\ M\^60 + 602987318\ M\^62 + 204811715\ M\^64 - 
                    765229172\ M\^66 - 98190343\ M\^68 + 862544581\ M\^70 - 
                    59365679\ M\^72 - 845982905\ M\^74 + 231656503\ M\^76 + 
                    696006900\ M\^78 - 353490021\ M\^80 - 455284534\ M\^82 + 
                    371710791\ M\^84 + 212371506\ M\^86 - 291797835\ M\^88 - 
                    45235507\ M\^90 + 172246747\ M\^92 - 26121884\ M\^94 - 
                    73375338\ M\^96 + 32921907\ M\^98 + 19645814\ M\^100 - 
                    18146633\ M\^102 - 1091886\ M\^104 + 5947104\ M\^106 - 
                    1474112\ M\^108 - 1122589\ M\^110 + 686344\ M\^112 + 
                    47415\ M\^114 - 140410\ M\^116 + 7283\ M\^118 + 
                    46598\ M\^120 - 31621\ M\^122 + 11189\ M\^124 - 
                    2472\ M\^126 + 345\ M\^128 - 28\ M\^130 + M\^132)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 23]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 24]\), \(M\^24 + 
              L\^27\ M\^134 + 
              L\ \((\(-6\)\ M\^20 + 26\ M\^22 - 44\ M\^24 + 11\ M\^26 + 
                    76\ M\^28 - 49\ M\^30 + M\^32 + 28\ M\^34 - 24\ M\^36 + 
                    11\ M\^38 - 3\ M\^40)\) + 
              L\^2\ \((15\ M\^16 - 130\ M\^18 + 467\ M\^20 - 748\ M\^22 + 
                    16\ M\^24 + 1711\ M\^26 - 2033\ M\^28 - 571\ M\^30 + 
                    2843\ M\^32 - 792\ M\^34 - 1027\ M\^36 + 521\ M\^38 + 
                    114\ M\^40 + 229\ M\^42 - 359\ M\^44 - 10\ M\^46 + 
                    229\ M\^48 - 188\ M\^50 + 83\ M\^52 - 22\ M\^54 + 
                    3\ M\^56)\) + 
              L\^3\ \((\(-20\)\ M\^12 + 260\ M\^14 - 1435\ M\^16 + 
                    4083\ M\^18 - 5257\ M\^20 - 1861\ M\^22 + 15877\ M\^24 - 
                    17784\ M\^26 - 5380\ M\^28 + 29687\ M\^30 - 
                    21496\ M\^32 - 11392\ M\^34 + 25522\ M\^36 - 
                    2219\ M\^38 - 5241\ M\^40 - 7025\ M\^42 + 3294\ M\^44 + 
                    14356\ M\^46 - 13678\ M\^48 - 3170\ M\^50 + 
                    10240\ M\^52 - 4650\ M\^54 - 869\ M\^56 + 1769\ M\^58 - 
                    1084\ M\^60 + 669\ M\^62 - 413\ M\^64 + 191\ M\^66 - 
                    59\ M\^68 + 11\ M\^70 - M\^72)\) + 
              L\^4\ \((15\ M\^8 - 260\ M\^10 + 1950\ M\^12 - 8056\ M\^14 + 
                    19085\ M\^16 - 21737\ M\^18 - 8700\ M\^20 + 
                    69363\ M\^22 - 98035\ M\^24 + 22917\ M\^26 + 
                    136662\ M\^28 - 239654\ M\^30 + 124201\ M\^32 + 
                    208261\ M\^34 - 453966\ M\^36 + 184463\ M\^38 + 
                    408837\ M\^40 - 456918\ M\^42 - 51654\ M\^44 + 
                    319798\ M\^46 - 108020\ M\^48 - 48778\ M\^50 + 
                    8419\ M\^52 - 24130\ M\^54 + 63819\ M\^56 - 
                    19427\ M\^58 - 26635\ M\^60 + 15804\ M\^62 + 807\ M\^64 + 
                    2958\ M\^66 - 5729\ M\^68 - 50\ M\^70 + 4568\ M\^72 - 
                    4042\ M\^74 + 1845\ M\^76 - 506\ M\^78 + 81\ M\^80 - 
                    6\ M\^82)\) + 
              L\^5\ \((\(-6\)\ M\^4 + 130\ M\^6 - 1250\ M\^8 + 6985\ M\^10 - 
                    24986\ M\^12 + 58358\ M\^14 - 79688\ M\^16 + 
                    10233\ M\^18 + 221675\ M\^20 - 476044\ M\^22 + 
                    263898\ M\^24 + 709173\ M\^26 - 1483126\ M\^28 + 
                    419843\ M\^30 + 1827564\ M\^32 - 2053598\ M\^34 - 
                    250631\ M\^36 + 1674622\ M\^38 - 1401545\ M\^40 + 
                    941185\ M\^42 + 850297\ M\^44 - 2817122\ M\^46 + 
                    1101992\ M\^48 + 2358623\ M\^50 - 1816135\ M\^52 - 
                    862873\ M\^54 + 1043219\ M\^56 + 84637\ M\^58 - 
                    228696\ M\^60 + 10936\ M\^62 - 42568\ M\^64 - 
                    35447\ M\^66 + 105933\ M\^68 + 49883\ M\^70 - 
                    148302\ M\^72 + 33699\ M\^74 + 80656\ M\^76 - 
                    64661\ M\^78 + 1832\ M\^80 + 24739\ M\^82 - 
                    18410\ M\^84 + 7192\ M\^86 - 1708\ M\^88 + 237\ M\^90 - 
                    15\ M\^92)\) + 
              L\^6\ \((1 - 26\ M\^2 + 319\ M\^4 - 2460\ M\^6 + 13230\ M\^8 - 
                    51062\ M\^10 + 137291\ M\^12 - 224954\ M\^14 + 
                    86881\ M\^16 + 531604\ M\^18 - 1228439\ M\^20 + 
                    698513\ M\^22 + 1547764\ M\^24 - 3385053\ M\^26 + 
                    2045907\ M\^28 + 2637209\ M\^30 - 7463384\ M\^32 + 
                    5027272\ M\^34 + 8147586\ M\^36 - 14545623\ M\^38 - 
                    3137260\ M\^40 + 18260736\ M\^42 - 1020389\ M\^44 - 
                    16713978\ M\^46 + 1827339\ M\^48 + 14504994\ M\^50 - 
                    3466004\ M\^52 - 9994445\ M\^54 + 7472633\ M\^56 + 
                    1639564\ M\^58 - 7708892\ M\^60 + 5812458\ M\^62 + 
                    2453829\ M\^64 - 6640374\ M\^66 + 2242400\ M\^68 + 
                    2610836\ M\^70 - 2264275\ M\^72 + 285632\ M\^74 + 
                    221857\ M\^76 - 313800\ M\^78 + 581247\ M\^80 - 
                    370231\ M\^82 - 145408\ M\^84 + 318095\ M\^86 - 
                    137806\ M\^88 - 39287\ M\^90 + 75112\ M\^92 - 
                    42675\ M\^94 + 14113\ M\^96 - 2925\ M\^98 + 358\ M\^100 - 
                    20\ M\^102)\) + 
              L\^7\ \((\(-7\) + 185\ M\^2 - 2185\ M\^4 + 15204\ M\^6 - 
                    68859\ M\^8 + 209214\ M\^10 - 408773\ M\^12 + 
                    375549\ M\^14 + 460605\ M\^16 - 2193165\ M\^18 + 
                    2856645\ M\^20 + 1375251\ M\^22 - 9459772\ M\^24 + 
                    8240502\ M\^26 + 11419842\ M\^28 - 23800244\ M\^30 - 
                    3459453\ M\^32 + 34563724\ M\^34 - 9257779\ M\^36 - 
                    33339761\ M\^38 + 21358703\ M\^40 + 14843716\ M\^42 - 
                    32223404\ M\^44 + 31872556\ M\^46 + 25198362\ M\^48 - 
                    99138047\ M\^50 + 27912374\ M\^52 + 136438501\ M\^54 - 
                    114811590\ M\^56 - 91104207\ M\^58 + 169560067\ M\^60 - 
                    12019410\ M\^62 - 137222466\ M\^64 + 88778746\ M\^66 + 
                    46740362\ M\^68 - 88035579\ M\^70 + 23940659\ M\^72 + 
                    38617747\ M\^74 - 37189352\ M\^76 + 1700784\ M\^78 + 
                    17098164\ M\^80 - 10658707\ M\^82 - 450731\ M\^84 + 
                    4054576\ M\^86 - 2649185\ M\^88 + 700558\ M\^90 + 
                    451082\ M\^92 - 735842\ M\^94 + 414220\ M\^96 - 
                    9992\ M\^98 - 152008\ M\^100 + 121044\ M\^102 - 
                    52798\ M\^104 + 14830\ M\^106 - 2708\ M\^108 + 
                    297\ M\^110 - 15\ M\^112)\) + 
              L\^8\ \((21 - 564\ M\^2 + 6722\ M\^4 - 46484\ M\^6 + 
                    202784\ M\^8 - 555518\ M\^10 + 816943\ M\^12 + 
                    36156\ M\^14 - 2603287\ M\^16 + 4185874\ M\^18 + 
                    246123\ M\^20 - 9311709\ M\^22 + 11168420\ M\^24 + 
                    3410441\ M\^26 - 25213531\ M\^28 + 22188342\ M\^30 + 
                    33295570\ M\^32 - 74976423\ M\^34 - 31048492\ M\^36 + 
                    170547925\ M\^38 - 1030710\ M\^40 - 301666861\ M\^42 + 
                    98474120\ M\^44 + 382291578\ M\^46 - 233315646\ M\^48 - 
                    306152093\ M\^50 + 267929807\ M\^52 + 110782924\ M\^54 - 
                    88987249\ M\^56 - 4960032\ M\^58 - 202210601\ M\^60 + 
                    141967255\ M\^62 + 353252528\ M\^64 - 398832478\ M\^66 - 
                    222824617\ M\^68 + 515376062\ M\^70 - 54336965\ M\^72 - 
                    385434863\ M\^74 + 234227347\ M\^76 + 139067200\ M\^78 - 
                    220565543\ M\^80 + 35661960\ M\^82 + 104558836\ M\^84 - 
                    76171621\ M\^86 - 11035860\ M\^88 + 42067976\ M\^90 - 
                    17866257\ M\^92 - 7635602\ M\^94 + 10998281\ M\^96 - 
                    3516635\ M\^98 - 1497032\ M\^100 + 1955017\ M\^102 - 
                    802854\ M\^104 - 30539\ M\^106 + 271227\ M\^108 - 
                    211847\ M\^110 + 101784\ M\^112 - 34135\ M\^114 + 
                    8077\ M\^116 - 1300\ M\^118 + 129\ M\^120 - 
                    6\ M\^122)\) + 
              L\^9\ \((\(-35\) + 955\ M\^2 - 11582\ M\^4 + 81380\ M\^6 - 
                    358054\ M\^8 + 966211\ M\^10 - 1261433\ M\^12 - 
                    847263\ M\^14 + 5865336\ M\^16 - 5800609\ M\^18 - 
                    9905510\ M\^20 + 24310598\ M\^22 + 5193974\ M\^24 - 
                    56118840\ M\^26 + 14637372\ M\^28 + 107964764\ M\^30 - 
                    75619948\ M\^32 - 156081158\ M\^34 + 215844766\ M\^36 + 
                    82994804\ M\^38 - 390076601\ M\^40 + 278702581\ M\^42 + 
                    378506047\ M\^44 - 908292099\ M\^46 + 50988088\ M\^48 + 
                    1452397978\ M\^50 - 805885902\ M\^52 - 
                    1496060027\ M\^54 + 1411080602\ M\^56 + 
                    1036159840\ M\^58 - 1449989813\ M\^60 - 
                    529229127\ M\^62 + 1008024961\ M\^64 + 396709010\ M\^66 - 
                    539136799\ M\^68 - 562780253\ M\^70 + 394570870\ M\^72 + 
                    660764723\ M\^74 - 507889530\ M\^76 - 485755540\ M\^78 + 
                    584988907\ M\^80 + 155539302\ M\^82 - 469135879\ M\^84 + 
                    96726282\ M\^86 + 238845801\ M\^88 - 165041446\ M\^90 - 
                    49086959\ M\^92 + 109748161\ M\^94 - 31898990\ M\^96 - 
                    35863631\ M\^98 + 33373211\ M\^100 - 1730440\ M\^102 - 
                    12999685\ M\^104 + 7619377\ M\^106 + 562398\ M\^108 - 
                    2800520\ M\^110 + 1347445\ M\^112 + 58386\ M\^114 - 
                    433192\ M\^116 + 300323\ M\^118 - 129502\ M\^120 + 
                    41183\ M\^122 - 10104\ M\^124 + 1897\ M\^126 - 
                    259\ M\^128 + 23\ M\^130 - M\^132)\) + 
              L\^26\ \((\(-3\)\ M\^118 + 11\ M\^120 - 24\ M\^122 + 
                    28\ M\^124 + M\^126 - 49\ M\^128 + 76\ M\^130 + 
                    11\ M\^132 - 44\ M\^134 + 26\ M\^136 - 6\ M\^138)\) + 
              L\^10\ \((35 - 970\ M\^2 + 11981\ M\^4 - 85995\ M\^6 + 
                    388084\ M\^8 - 1082714\ M\^10 + 1509190\ M\^12 + 
                    714666\ M\^14 - 6429526\ M\^16 + 6478796\ M\^18 + 
                    12109431\ M\^20 - 27925132\ M\^22 - 14430757\ M\^24 + 
                    76479020\ M\^26 + 14053728\ M\^28 - 201124129\ M\^30 + 
                    30443608\ M\^32 + 453250084\ M\^34 - 231783173\ M\^36 - 
                    763001937\ M\^38 + 659876089\ M\^40 + 867749260\ M\^42 - 
                    1096102255\ M\^44 - 555251130\ M\^46 + 
                    1059349976\ M\^48 + 44675921\ M\^50 - 306694003\ M\^52 + 
                    29981214\ M\^54 - 680017948\ M\^56 + 746016194\ M\^58 + 
                    979190320\ M\^60 - 1976673334\ M\^62 - 132268221\ M\^64 + 
                    2671233110\ M\^66 - 1338372427\ M\^68 - 
                    2150014144\ M\^70 + 2417410449\ M\^72 + 
                    737421895\ M\^74 - 2409971064\ M\^76 + 634409109\ M\^78 + 
                    1468168466\ M\^80 - 1251060995\ M\^82 - 
                    322148578\ M\^84 + 1063659519\ M\^86 - 392737170\ M\^88 - 
                    515893090\ M\^90 + 536869526\ M\^92 + 63648005\ M\^94 - 
                    351735337\ M\^96 + 119689173\ M\^98 + 131454209\ M\^100 - 
                    112701481\ M\^102 - 10858167\ M\^104 + 50750262\ M\^106 - 
                    18076941\ M\^108 - 9937803\ M\^110 + 11170786\ M\^112 - 
                    2787698\ M\^114 - 1897696\ M\^116 + 1969182\ M\^118 - 
                    593266\ M\^120 - 285894\ M\^122 + 433156\ M\^124 - 
                    267637\ M\^126 + 107574\ M\^128 - 30477\ M\^130 + 
                    6133\ M\^132 - 847\ M\^134 + 73\ M\^136 - 3\ M\^138)\) + 
              L\^25\ \((3\ M\^102 - 22\ M\^104 + 83\ M\^106 - 188\ M\^108 + 
                    229\ M\^110 - 10\ M\^112 - 359\ M\^114 + 229\ M\^116 + 
                    114\ M\^118 + 521\ M\^120 - 1027\ M\^122 - 792\ M\^124 + 
                    2843\ M\^126 - 571\ M\^128 - 2033\ M\^130 + 
                    1711\ M\^132 + 16\ M\^134 - 748\ M\^136 + 467\ M\^138 - 
                    130\ M\^140 + 15\ M\^142)\) + 
              L\^11\ \((\(-21\) + 591\ M\^2 - 7429\ M\^4 + 54560\ M\^6 - 
                    254958\ M\^8 + 759534\ M\^10 - 1280673\ M\^12 + 
                    424797\ M\^14 + 2798852\ M\^16 - 4319781\ M\^18 - 
                    2381236\ M\^20 + 7627231\ M\^22 + 9464202\ M\^24 - 
                    14052998\ M\^26 - 66878637\ M\^28 + 105164659\ M\^30 + 
                    184722679\ M\^32 - 466233318\ M\^34 - 209077211\ M\^36 + 
                    1262608015\ M\^38 - 236316811\ M\^40 - 
                    2364908466\ M\^42 + 1523182250\ M\^44 + 
                    3244773544\ M\^46 - 3697456543\ M\^48 - 
                    3204477790\ M\^50 + 6419445642\ M\^52 + 
                    1776179077\ M\^54 - 9249433984\ M\^56 + 
                    1071858177\ M\^58 + 11806268519\ M\^60 - 
                    4952647631\ M\^62 - 13584687447\ M\^64 + 
                    9177855733\ M\^66 + 13837790840\ M\^68 - 
                    12879749943\ M\^70 - 12004852364\ M\^72 + 
                    15162414339\ M\^74 + 8357565550\ M\^76 - 
                    15255821954\ M\^78 - 3940047275\ M\^80 + 
                    13064626799\ M\^82 + 112291165\ M\^84 - 
                    9367474014\ M\^86 + 2131878919\ M\^88 + 
                    5440116464\ M\^90 - 2650366897\ M\^92 - 
                    2376340939\ M\^94 + 2047767473\ M\^96 + 
                    611335072\ M\^98 - 1140509272\ M\^100 + 
                    77074389\ M\^102 + 454005160\ M\^104 - 
                    182431720\ M\^106 - 111079194\ M\^108 + 
                    109091436\ M\^110 - 1007341\ M\^112 - 38809798\ M\^114 + 
                    15582539\ M\^116 + 7703546\ M\^118 - 8849501\ M\^120 + 
                    1291835\ M\^122 + 2241687\ M\^124 - 1354685\ M\^126 - 
                    109682\ M\^128 + 547740\ M\^130 - 368786\ M\^132 + 
                    145922\ M\^134 - 39259\ M\^136 + 7395\ M\^138 - 
                    953\ M\^140 + 77\ M\^142 - 3\ M\^144)\) + 
              L\^24\ \((\(-M\^86\) + 11\ M\^88 - 59\ M\^90 + 191\ M\^92 - 
                    413\ M\^94 + 669\ M\^96 - 1084\ M\^98 + 1769\ M\^100 - 
                    869\ M\^102 - 4650\ M\^104 + 10240\ M\^106 - 
                    3170\ M\^108 - 13678\ M\^110 + 14356\ M\^112 + 
                    3294\ M\^114 - 7025\ M\^116 - 5241\ M\^118 - 
                    2219\ M\^120 + 25522\ M\^122 - 11392\ M\^124 - 
                    21496\ M\^126 + 29687\ M\^128 - 5380\ M\^130 - 
                    17784\ M\^132 + 15877\ M\^134 - 1861\ M\^136 - 
                    5257\ M\^138 + 4083\ M\^140 - 1435\ M\^142 + 
                    260\ M\^144 - 20\ M\^146)\) + 
              L\^12\ \((7 - 200\ M\^2 + 2560\ M\^4 - 19334\ M\^6 + 
                    95075\ M\^8 - 315113\ M\^10 + 698039\ M\^12 - 
                    967660\ M\^14 + 629803\ M\^16 + 557626\ M\^18 - 
                    3281421\ M\^20 + 8842154\ M\^22 - 8710637\ M\^24 - 
                    24355974\ M\^26 + 81251628\ M\^28 - 31258073\ M\^30 - 
                    216611869\ M\^32 + 317173973\ M\^34 + 243545429\ M\^36 - 
                    938683907\ M\^38 + 256581177\ M\^40 + 1641192650\ M\^42 - 
                    1733863742\ M\^44 - 1646539571\ M\^46 + 
                    4275107852\ M\^48 - 137766983\ M\^50 - 
                    7323268616\ M\^52 + 4519864007\ M\^54 + 
                    9807122448\ M\^56 - 11356551294\ M\^58 - 
                    10564137004\ M\^60 + 19093391998\ M\^62 + 
                    8871478887\ M\^64 - 24987973080\ M\^66 - 
                    4889485011\ M\^68 + 26360590587\ M\^70 - 
                    140054455\ M\^72 - 22381720997\ M\^74 + 
                    4253407384\ M\^76 + 14828619431\ M\^78 - 
                    5762258492\ M\^80 - 6830147975\ M\^82 + 
                    4556039748\ M\^84 + 1068288987\ M\^86 - 
                    2001506277\ M\^88 + 1504737772\ M\^90 - 
                    241292838\ M\^92 - 1654865000\ M\^94 + 
                    1295822486\ M\^96 + 821558338\ M\^98 - 
                    1268905839\ M\^100 - 81741097\ M\^102 + 
                    765031581\ M\^104 - 218985542\ M\^106 - 
                    294580314\ M\^108 + 206380570\ M\^110 + 
                    46984117\ M\^112 - 104929330\ M\^114 + 26898014\ M\^116 + 
                    29978283\ M\^118 - 24659138\ M\^120 - 861156\ M\^122 + 
                    10580136\ M\^124 - 4906093\ M\^126 - 1410931\ M\^128 + 
                    2273487\ M\^130 - 553197\ M\^132 - 541427\ M\^134 + 
                    585482\ M\^136 - 297006\ M\^138 + 98071\ M\^140 - 
                    22638\ M\^142 + 3688\ M\^144 - 412\ M\^146 + 29\ M\^148 - 
                    M\^150)\) + 
              L\^23\ \((\(-6\)\ M\^76 + 81\ M\^78 - 506\ M\^80 + 
                    1845\ M\^82 - 4042\ M\^84 + 4568\ M\^86 - 50\ M\^88 - 
                    5729\ M\^90 + 2958\ M\^92 + 807\ M\^94 + 15804\ M\^96 - 
                    26635\ M\^98 - 19427\ M\^100 + 63819\ M\^102 - 
                    24130\ M\^104 + 8419\ M\^106 - 48778\ M\^108 - 
                    108020\ M\^110 + 319798\ M\^112 - 51654\ M\^114 - 
                    456918\ M\^116 + 408837\ M\^118 + 184463\ M\^120 - 
                    453966\ M\^122 + 208261\ M\^124 + 124201\ M\^126 - 
                    239654\ M\^128 + 136662\ M\^130 + 22917\ M\^132 - 
                    98035\ M\^134 + 69363\ M\^136 - 8700\ M\^138 - 
                    21737\ M\^140 + 19085\ M\^142 - 8056\ M\^144 + 
                    1950\ M\^146 - 260\ M\^148 + 15\ M\^150)\) + 
              L\^22\ \((\(-15\)\ M\^66 + 237\ M\^68 - 1708\ M\^70 + 
                    7192\ M\^72 - 18410\ M\^74 + 24739\ M\^76 + 1832\ M\^78 - 
                    64661\ M\^80 + 80656\ M\^82 + 33699\ M\^84 - 
                    148302\ M\^86 + 49883\ M\^88 + 105933\ M\^90 - 
                    35447\ M\^92 - 42568\ M\^94 + 10936\ M\^96 - 
                    228696\ M\^98 + 84637\ M\^100 + 1043219\ M\^102 - 
                    862873\ M\^104 - 1816135\ M\^106 + 2358623\ M\^108 + 
                    1101992\ M\^110 - 2817122\ M\^112 + 850297\ M\^114 + 
                    941185\ M\^116 - 1401545\ M\^118 + 1674622\ M\^120 - 
                    250631\ M\^122 - 2053598\ M\^124 + 1827564\ M\^126 + 
                    419843\ M\^128 - 1483126\ M\^130 + 709173\ M\^132 + 
                    263898\ M\^134 - 476044\ M\^136 + 221675\ M\^138 + 
                    10233\ M\^140 - 79688\ M\^142 + 58358\ M\^144 - 
                    24986\ M\^146 + 6985\ M\^148 - 1250\ M\^150 + 
                    130\ M\^152 - 6\ M\^154)\) + 
              L\^13\ \((\(-1\) + 29\ M\^2 - 382\ M\^4 + 3054\ M\^6 - 
                    16774\ M\^8 + 68295\ M\^10 - 217215\ M\^12 + 
                    549189\ M\^14 - 1037932\ M\^16 + 1078759\ M\^18 + 
                    893435\ M\^20 - 5465283\ M\^22 + 6959883\ M\^24 + 
                    5647021\ M\^26 - 28455778\ M\^28 + 26323131\ M\^30 + 
                    27152542\ M\^32 - 88064738\ M\^34 + 72333972\ M\^36 + 
                    65558929\ M\^38 - 286678375\ M\^40 + 294674251\ M\^42 + 
                    365350157\ M\^44 - 1135650184\ M\^46 + 203040926\ M\^48 + 
                    2099270034\ M\^50 - 1768267714\ M\^52 - 
                    2182146102\ M\^54 + 3882536831\ M\^56 + 
                    201491540\ M\^58 - 5198107090\ M\^60 + 
                    4513927260\ M\^62 + 4162114090\ M\^64 - 
                    11777929483\ M\^66 - 2963718\ M\^68 + 
                    20595081918\ M\^70 - 6834487347\ M\^72 - 
                    29069547874\ M\^74 + 14953228430\ M\^76 + 
                    34387286105\ M\^78 - 22549697051\ M\^80 - 
                    33843652087\ M\^82 + 27787646463\ M\^84 + 
                    26946153083\ M\^86 - 28894371097\ M\^88 - 
                    16133767089\ M\^90 + 25170702006\ M\^92 + 
                    5528592725\ M\^94 - 17917423072\ M\^96 + 
                    1485313455\ M\^98 + 9922858407\ M\^100 - 
                    3995062723\ M\^102 - 3775395802\ M\^104 + 
                    3393790765\ M\^106 + 479911961\ M\^108 - 
                    1785605564\ M\^110 + 538896428\ M\^112 + 
                    557615873\ M\^114 - 468689695\ M\^116 - 
                    25544112\ M\^118 + 188721900\ M\^120 - 70127812\ M\^122 - 
                    32257184\ M\^124 + 35515098\ M\^126 - 7698371\ M\^128 - 
                    5012065\ M\^130 + 4442293\ M\^132 - 1415692\ M\^134 - 
                    296276\ M\^136 + 763597\ M\^138 - 608467\ M\^140 + 
                    314188\ M\^142 - 114727\ M\^144 + 30099\ M\^146 - 
                    5601\ M\^148 + 708\ M\^150 - 55\ M\^152 + 2\ M\^154)\) + 
              L\^18\ \((\(-M\^26\) + 23\ M\^28 - 259\ M\^30 + 1897\ M\^32 - 
                    10104\ M\^34 + 41183\ M\^36 - 129502\ M\^38 + 
                    300323\ M\^40 - 433192\ M\^42 + 58386\ M\^44 + 
                    1347445\ M\^46 - 2800520\ M\^48 + 562398\ M\^50 + 
                    7619377\ M\^52 - 12999685\ M\^54 - 1730440\ M\^56 + 
                    33373211\ M\^58 - 35863631\ M\^60 - 31898990\ M\^62 + 
                    109748161\ M\^64 - 49086959\ M\^66 - 165041446\ M\^68 + 
                    238845801\ M\^70 + 96726282\ M\^72 - 469135879\ M\^74 + 
                    155539302\ M\^76 + 584988907\ M\^78 - 485755540\ M\^80 - 
                    507889530\ M\^82 + 660764723\ M\^84 + 394570870\ M\^86 - 
                    562780253\ M\^88 - 539136799\ M\^90 + 396709010\ M\^92 + 
                    1008024961\ M\^94 - 529229127\ M\^96 - 
                    1449989813\ M\^98 + 1036159840\ M\^100 + 
                    1411080602\ M\^102 - 1496060027\ M\^104 - 
                    805885902\ M\^106 + 1452397978\ M\^108 + 
                    50988088\ M\^110 - 908292099\ M\^112 + 
                    378506047\ M\^114 + 278702581\ M\^116 - 
                    390076601\ M\^118 + 82994804\ M\^120 + 
                    215844766\ M\^122 - 156081158\ M\^124 - 
                    75619948\ M\^126 + 107964764\ M\^128 + 14637372\ M\^130 - 
                    56118840\ M\^132 + 5193974\ M\^134 + 24310598\ M\^136 - 
                    9905510\ M\^138 - 5800609\ M\^140 + 5865336\ M\^142 - 
                    847263\ M\^144 - 1261433\ M\^146 + 966211\ M\^148 - 
                    358054\ M\^150 + 81380\ M\^152 - 11582\ M\^154 + 
                    955\ M\^156 - 35\ M\^158)\) + 
              L\^16\ \((\(-3\)\ M\^14 + 77\ M\^16 - 953\ M\^18 + 
                    7395\ M\^20 - 39259\ M\^22 + 145922\ M\^24 - 
                    368786\ M\^26 + 547740\ M\^28 - 109682\ M\^30 - 
                    1354685\ M\^32 + 2241687\ M\^34 + 1291835\ M\^36 - 
                    8849501\ M\^38 + 7703546\ M\^40 + 15582539\ M\^42 - 
                    38809798\ M\^44 - 1007341\ M\^46 + 109091436\ M\^48 - 
                    111079194\ M\^50 - 182431720\ M\^52 + 454005160\ M\^54 + 
                    77074389\ M\^56 - 1140509272\ M\^58 + 611335072\ M\^60 + 
                    2047767473\ M\^62 - 2376340939\ M\^64 - 
                    2650366897\ M\^66 + 5440116464\ M\^68 + 
                    2131878919\ M\^70 - 9367474014\ M\^72 + 
                    112291165\ M\^74 + 13064626799\ M\^76 - 
                    3940047275\ M\^78 - 15255821954\ M\^80 + 
                    8357565550\ M\^82 + 15162414339\ M\^84 - 
                    12004852364\ M\^86 - 12879749943\ M\^88 + 
                    13837790840\ M\^90 + 9177855733\ M\^92 - 
                    13584687447\ M\^94 - 4952647631\ M\^96 + 
                    11806268519\ M\^98 + 1071858177\ M\^100 - 
                    9249433984\ M\^102 + 1776179077\ M\^104 + 
                    6419445642\ M\^106 - 3204477790\ M\^108 - 
                    3697456543\ M\^110 + 3244773544\ M\^112 + 
                    1523182250\ M\^114 - 2364908466\ M\^116 - 
                    236316811\ M\^118 + 1262608015\ M\^120 - 
                    209077211\ M\^122 - 466233318\ M\^124 + 
                    184722679\ M\^126 + 105164659\ M\^128 - 
                    66878637\ M\^130 - 14052998\ M\^132 + 9464202\ M\^134 + 
                    7627231\ M\^136 - 2381236\ M\^138 - 4319781\ M\^140 + 
                    2798852\ M\^142 + 424797\ M\^144 - 1280673\ M\^146 + 
                    759534\ M\^148 - 254958\ M\^150 + 54560\ M\^152 - 
                    7429\ M\^154 + 591\ M\^156 - 21\ M\^158)\) + 
              L\^20\ \((\(-15\)\ M\^46 + 297\ M\^48 - 2708\ M\^50 + 
                    14830\ M\^52 - 52798\ M\^54 + 121044\ M\^56 - 
                    152008\ M\^58 - 9992\ M\^60 + 414220\ M\^62 - 
                    735842\ M\^64 + 451082\ M\^66 + 700558\ M\^68 - 
                    2649185\ M\^70 + 4054576\ M\^72 - 450731\ M\^74 - 
                    10658707\ M\^76 + 17098164\ M\^78 + 1700784\ M\^80 - 
                    37189352\ M\^82 + 38617747\ M\^84 + 23940659\ M\^86 - 
                    88035579\ M\^88 + 46740362\ M\^90 + 88778746\ M\^92 - 
                    137222466\ M\^94 - 12019410\ M\^96 + 169560067\ M\^98 - 
                    91104207\ M\^100 - 114811590\ M\^102 + 
                    136438501\ M\^104 + 27912374\ M\^106 - 99138047\ M\^108 + 
                    25198362\ M\^110 + 31872556\ M\^112 - 32223404\ M\^114 + 
                    14843716\ M\^116 + 21358703\ M\^118 - 33339761\ M\^120 - 
                    9257779\ M\^122 + 34563724\ M\^124 - 3459453\ M\^126 - 
                    23800244\ M\^128 + 11419842\ M\^130 + 8240502\ M\^132 - 
                    9459772\ M\^134 + 1375251\ M\^136 + 2856645\ M\^138 - 
                    2193165\ M\^140 + 460605\ M\^142 + 375549\ M\^144 - 
                    408773\ M\^146 + 209214\ M\^148 - 68859\ M\^150 + 
                    15204\ M\^152 - 2185\ M\^154 + 185\ M\^156 - 
                    7\ M\^158)\) + 
              L\^14\ \((2\ M\^4 - 55\ M\^6 + 708\ M\^8 - 5601\ M\^10 + 
                    30099\ M\^12 - 114727\ M\^14 + 314188\ M\^16 - 
                    608467\ M\^18 + 763597\ M\^20 - 296276\ M\^22 - 
                    1415692\ M\^24 + 4442293\ M\^26 - 5012065\ M\^28 - 
                    7698371\ M\^30 + 35515098\ M\^32 - 32257184\ M\^34 - 
                    70127812\ M\^36 + 188721900\ M\^38 - 25544112\ M\^40 - 
                    468689695\ M\^42 + 557615873\ M\^44 + 538896428\ M\^46 - 
                    1785605564\ M\^48 + 479911961\ M\^50 + 
                    3393790765\ M\^52 - 3775395802\ M\^54 - 
                    3995062723\ M\^56 + 9922858407\ M\^58 + 
                    1485313455\ M\^60 - 17917423072\ M\^62 + 
                    5528592725\ M\^64 + 25170702006\ M\^66 - 
                    16133767089\ M\^68 - 28894371097\ M\^70 + 
                    26946153083\ M\^72 + 27787646463\ M\^74 - 
                    33843652087\ M\^76 - 22549697051\ M\^78 + 
                    34387286105\ M\^80 + 14953228430\ M\^82 - 
                    29069547874\ M\^84 - 6834487347\ M\^86 + 
                    20595081918\ M\^88 - 2963718\ M\^90 - 
                    11777929483\ M\^92 + 4162114090\ M\^94 + 
                    4513927260\ M\^96 - 5198107090\ M\^98 + 
                    201491540\ M\^100 + 3882536831\ M\^102 - 
                    2182146102\ M\^104 - 1768267714\ M\^106 + 
                    2099270034\ M\^108 + 203040926\ M\^110 - 
                    1135650184\ M\^112 + 365350157\ M\^114 + 
                    294674251\ M\^116 - 286678375\ M\^118 + 
                    65558929\ M\^120 + 72333972\ M\^122 - 88064738\ M\^124 + 
                    27152542\ M\^126 + 26323131\ M\^128 - 28455778\ M\^130 + 
                    5647021\ M\^132 + 6959883\ M\^134 - 5465283\ M\^136 + 
                    893435\ M\^138 + 1078759\ M\^140 - 1037932\ M\^142 + 
                    549189\ M\^144 - 217215\ M\^146 + 68295\ M\^148 - 
                    16774\ M\^150 + 3054\ M\^152 - 382\ M\^154 + 29\ M\^156 - 
                    M\^158)\) + 
              L\^21\ \((\(-20\)\ M\^56 + 358\ M\^58 - 2925\ M\^60 + 
                    14113\ M\^62 - 42675\ M\^64 + 75112\ M\^66 - 
                    39287\ M\^68 - 137806\ M\^70 + 318095\ M\^72 - 
                    145408\ M\^74 - 370231\ M\^76 + 581247\ M\^78 - 
                    313800\ M\^80 + 221857\ M\^82 + 285632\ M\^84 - 
                    2264275\ M\^86 + 2610836\ M\^88 + 2242400\ M\^90 - 
                    6640374\ M\^92 + 2453829\ M\^94 + 5812458\ M\^96 - 
                    7708892\ M\^98 + 1639564\ M\^100 + 7472633\ M\^102 - 
                    9994445\ M\^104 - 3466004\ M\^106 + 14504994\ M\^108 + 
                    1827339\ M\^110 - 16713978\ M\^112 - 1020389\ M\^114 + 
                    18260736\ M\^116 - 3137260\ M\^118 - 14545623\ M\^120 + 
                    8147586\ M\^122 + 5027272\ M\^124 - 7463384\ M\^126 + 
                    2637209\ M\^128 + 2045907\ M\^130 - 3385053\ M\^132 + 
                    1547764\ M\^134 + 698513\ M\^136 - 1228439\ M\^138 + 
                    531604\ M\^140 + 86881\ M\^142 - 224954\ M\^144 + 
                    137291\ M\^146 - 51062\ M\^148 + 13230\ M\^150 - 
                    2460\ M\^152 + 319\ M\^154 - 26\ M\^156 + M\^158)\) + 
              L\^15\ \((\(-M\^8\) + 29\ M\^10 - 412\ M\^12 + 3688\ M\^14 - 
                    22638\ M\^16 + 98071\ M\^18 - 297006\ M\^20 + 
                    585482\ M\^22 - 541427\ M\^24 - 553197\ M\^26 + 
                    2273487\ M\^28 - 1410931\ M\^30 - 4906093\ M\^32 + 
                    10580136\ M\^34 - 861156\ M\^36 - 24659138\ M\^38 + 
                    29978283\ M\^40 + 26898014\ M\^42 - 104929330\ M\^44 + 
                    46984117\ M\^46 + 206380570\ M\^48 - 294580314\ M\^50 - 
                    218985542\ M\^52 + 765031581\ M\^54 - 81741097\ M\^56 - 
                    1268905839\ M\^58 + 821558338\ M\^60 + 
                    1295822486\ M\^62 - 1654865000\ M\^64 - 
                    241292838\ M\^66 + 1504737772\ M\^68 - 
                    2001506277\ M\^70 + 1068288987\ M\^72 + 
                    4556039748\ M\^74 - 6830147975\ M\^76 - 
                    5762258492\ M\^78 + 14828619431\ M\^80 + 
                    4253407384\ M\^82 - 22381720997\ M\^84 - 
                    140054455\ M\^86 + 26360590587\ M\^88 - 
                    4889485011\ M\^90 - 24987973080\ M\^92 + 
                    8871478887\ M\^94 + 19093391998\ M\^96 - 
                    10564137004\ M\^98 - 11356551294\ M\^100 + 
                    9807122448\ M\^102 + 4519864007\ M\^104 - 
                    7323268616\ M\^106 - 137766983\ M\^108 + 
                    4275107852\ M\^110 - 1646539571\ M\^112 - 
                    1733863742\ M\^114 + 1641192650\ M\^116 + 
                    256581177\ M\^118 - 938683907\ M\^120 + 
                    243545429\ M\^122 + 317173973\ M\^124 - 
                    216611869\ M\^126 - 31258073\ M\^128 + 81251628\ M\^130 - 
                    24355974\ M\^132 - 8710637\ M\^134 + 8842154\ M\^136 - 
                    3281421\ M\^138 + 557626\ M\^140 + 629803\ M\^142 - 
                    967660\ M\^144 + 698039\ M\^146 - 315113\ M\^148 + 
                    95075\ M\^150 - 19334\ M\^152 + 2560\ M\^154 - 
                    200\ M\^156 + 7\ M\^158)\) + 
              L\^19\ \((\(-6\)\ M\^36 + 129\ M\^38 - 1300\ M\^40 + 
                    8077\ M\^42 - 34135\ M\^44 + 101784\ M\^46 - 
                    211847\ M\^48 + 271227\ M\^50 - 30539\ M\^52 - 
                    802854\ M\^54 + 1955017\ M\^56 - 1497032\ M\^58 - 
                    3516635\ M\^60 + 10998281\ M\^62 - 7635602\ M\^64 - 
                    17866257\ M\^66 + 42067976\ M\^68 - 11035860\ M\^70 - 
                    76171621\ M\^72 + 104558836\ M\^74 + 35661960\ M\^76 - 
                    220565543\ M\^78 + 139067200\ M\^80 + 234227347\ M\^82 - 
                    385434863\ M\^84 - 54336965\ M\^86 + 515376062\ M\^88 - 
                    222824617\ M\^90 - 398832478\ M\^92 + 353252528\ M\^94 + 
                    141967255\ M\^96 - 202210601\ M\^98 - 4960032\ M\^100 - 
                    88987249\ M\^102 + 110782924\ M\^104 + 
                    267929807\ M\^106 - 306152093\ M\^108 - 
                    233315646\ M\^110 + 382291578\ M\^112 + 
                    98474120\ M\^114 - 301666861\ M\^116 - 1030710\ M\^118 + 
                    170547925\ M\^120 - 31048492\ M\^122 - 74976423\ M\^124 + 
                    33295570\ M\^126 + 22188342\ M\^128 - 25213531\ M\^130 + 
                    3410441\ M\^132 + 11168420\ M\^134 - 9311709\ M\^136 + 
                    246123\ M\^138 + 4185874\ M\^140 - 2603287\ M\^142 + 
                    36156\ M\^144 + 816943\ M\^146 - 555518\ M\^148 + 
                    202784\ M\^150 - 46484\ M\^152 + 6722\ M\^154 - 
                    564\ M\^156 + 21\ M\^158)\) + 
              L\^17\ \((\(-3\)\ M\^20 + 73\ M\^22 - 847\ M\^24 + 
                    6133\ M\^26 - 30477\ M\^28 + 107574\ M\^30 - 
                    267637\ M\^32 + 433156\ M\^34 - 285894\ M\^36 - 
                    593266\ M\^38 + 1969182\ M\^40 - 1897696\ M\^42 - 
                    2787698\ M\^44 + 11170786\ M\^46 - 9937803\ M\^48 - 
                    18076941\ M\^50 + 50750262\ M\^52 - 10858167\ M\^54 - 
                    112701481\ M\^56 + 131454209\ M\^58 + 119689173\ M\^60 - 
                    351735337\ M\^62 + 63648005\ M\^64 + 536869526\ M\^66 - 
                    515893090\ M\^68 - 392737170\ M\^70 + 1063659519\ M\^72 - 
                    322148578\ M\^74 - 1251060995\ M\^76 + 
                    1468168466\ M\^78 + 634409109\ M\^80 - 
                    2409971064\ M\^82 + 737421895\ M\^84 + 
                    2417410449\ M\^86 - 2150014144\ M\^88 - 
                    1338372427\ M\^90 + 2671233110\ M\^92 - 
                    132268221\ M\^94 - 1976673334\ M\^96 + 979190320\ M\^98 + 
                    746016194\ M\^100 - 680017948\ M\^102 + 
                    29981214\ M\^104 - 306694003\ M\^106 + 44675921\ M\^108 + 
                    1059349976\ M\^110 - 555251130\ M\^112 - 
                    1096102255\ M\^114 + 867749260\ M\^116 + 
                    659876089\ M\^118 - 763001937\ M\^120 - 
                    231783173\ M\^122 + 453250084\ M\^124 + 
                    30443608\ M\^126 - 201124129\ M\^128 + 14053728\ M\^130 + 
                    76479020\ M\^132 - 14430757\ M\^134 - 27925132\ M\^136 + 
                    12109431\ M\^138 + 6478796\ M\^140 - 6429526\ M\^142 + 
                    714666\ M\^144 + 1509190\ M\^146 - 1082714\ M\^148 + 
                    388084\ M\^150 - 85995\ M\^152 + 11981\ M\^154 - 
                    970\ M\^156 + 35\ M\^158)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 25]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 26]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 27]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 28]\), \(M\^72 + L\^26\ M\^72 + 
              L\^25\ \((2\ M\^58 - 8\ M\^60 + 17\ M\^62 - 20\ M\^64 + 
                    7\ M\^66 + 21\ M\^68 - 48\ M\^70 + 50\ M\^72 + 
                    36\ M\^74 - 43\ M\^76 + 12\ M\^78)\) + 
              L\^24\ \((M\^44 - 8\ M\^46 + 34\ M\^48 - 88\ M\^50 + 
                    126\ M\^52 - 43\ M\^54 - 147\ M\^56 + 99\ M\^58 + 
                    311\ M\^60 - 120\ M\^62 - 1057\ M\^64 + 1238\ M\^66 + 
                    960\ M\^68 - 2903\ M\^70 + 1228\ M\^72 + 2563\ M\^74 - 
                    2228\ M\^76 - 286\ M\^78 + 1031\ M\^80 - 452\ M\^82 + 
                    66\ M\^84)\) + 
              L\ \((12\ M\^66 - 43\ M\^68 + 36\ M\^70 + 50\ M\^72 - 
                    48\ M\^74 + 21\ M\^76 + 7\ M\^78 - 20\ M\^80 + 
                    17\ M\^82 - 8\ M\^84 + 2\ M\^86)\) + 
              L\^23\ \((M\^34 - 13\ M\^36 + 83\ M\^38 - 323\ M\^40 + 
                    788\ M\^42 - 1044\ M\^44 - 89\ M\^46 + 3063\ M\^48 - 
                    4636\ M\^50 - 653\ M\^52 + 10937\ M\^54 - 11726\ M\^56 - 
                    8245\ M\^58 + 29561\ M\^60 - 12109\ M\^62 - 
                    38293\ M\^64 + 49698\ M\^66 + 17562\ M\^68 - 
                    79790\ M\^70 + 30444\ M\^72 + 69151\ M\^74 - 
                    59181\ M\^76 - 20460\ M\^78 + 42614\ M\^80 - 
                    10733\ M\^82 - 9856\ M\^84 + 7784\ M\^86 - 2155\ M\^88 + 
                    220\ M\^90)\) + 
              L\^22\ \((\(-3\)\ M\^26 + 41\ M\^28 - 271\ M\^30 + 
                    1110\ M\^32 - 2951\ M\^34 + 4620\ M\^36 - 1408\ M\^38 - 
                    11863\ M\^40 + 27710\ M\^42 - 16104\ M\^44 - 
                    46656\ M\^46 + 108885\ M\^48 - 47259\ M\^50 - 
                    151570\ M\^52 + 253647\ M\^54 - 37886\ M\^56 - 
                    319285\ M\^58 + 394945\ M\^60 - 33026\ M\^62 - 
                    506991\ M\^64 + 665322\ M\^66 + 84020\ M\^68 - 
                    1064242\ M\^70 + 618060\ M\^72 + 878908\ M\^74 - 
                    994305\ M\^76 - 276567\ M\^78 + 796943\ M\^80 - 
                    177806\ M\^82 - 313838\ M\^84 + 212465\ M\^86 + 
                    8689\ M\^88 - 62428\ M\^90 + 29699\ M\^92 - 6150\ M\^94 + 
                    495\ M\^96)\) + 
              L\^2\ \((66\ M\^60 - 452\ M\^62 + 1031\ M\^64 - 286\ M\^66 - 
                    2228\ M\^68 + 2563\ M\^70 + 1228\ M\^72 - 2903\ M\^74 + 
                    960\ M\^76 + 1238\ M\^78 - 1057\ M\^80 - 120\ M\^82 + 
                    311\ M\^84 + 99\ M\^86 - 147\ M\^88 - 43\ M\^90 + 
                    126\ M\^92 - 88\ M\^94 + 34\ M\^96 - 8\ M\^98 + 
                    M\^100)\) + 
              L\^21\ \((3\ M\^18 - 47\ M\^20 + 347\ M\^22 - 1578\ M\^24 + 
                    4763\ M\^26 - 9237\ M\^28 + 8381\ M\^30 + 9822\ M\^32 - 
                    43808\ M\^34 + 49862\ M\^36 + 31556\ M\^38 - 
                    157507\ M\^40 + 145159\ M\^42 + 82429\ M\^44 - 
                    289245\ M\^46 + 237937\ M\^48 - 17975\ M\^50 - 
                    281488\ M\^52 + 655883\ M\^54 - 611232\ M\^56 - 
                    364048\ M\^58 + 1734120\ M\^60 - 1997062\ M\^62 - 
                    867805\ M\^64 + 5400024\ M\^66 - 3186389\ M\^68 - 
                    6816448\ M\^70 + 8197689\ M\^72 + 4504617\ M\^74 - 
                    10196966\ M\^76 - 63967\ M\^78 + 7931629\ M\^80 - 
                    2959330\ M\^82 - 3626474\ M\^84 + 2996422\ M\^86 + 
                    444236\ M\^88 - 1381307\ M\^90 + 459752\ M\^92 + 
                    181815\ M\^94 - 197083\ M\^96 + 69208\ M\^98 - 
                    11670\ M\^100 + 792\ M\^102)\) + 
              L\^20\ \((\(-M\^10\) + 18\ M\^12 - 152\ M\^14 + 797\ M\^16 - 
                    2869\ M\^18 + 7321\ M\^20 - 13009\ M\^22 + 14556\ M\^24 - 
                    5243\ M\^26 - 17680\ M\^28 + 58113\ M\^30 - 
                    108914\ M\^32 + 65603\ M\^34 + 262656\ M\^36 - 
                    706988\ M\^38 + 388396\ M\^40 + 1150398\ M\^42 - 
                    2170573\ M\^44 + 179710\ M\^46 + 3455453\ M\^48 - 
                    3338647\ M\^50 - 1950640\ M\^52 + 4880377\ M\^54 - 
                    590926\ M\^56 - 1291297\ M\^58 - 2503463\ M\^60 - 
                    5786473\ M\^62 + 16514583\ M\^64 + 9377115\ M\^66 - 
                    37595867\ M\^68 - 3350758\ M\^70 + 53570095\ M\^72 - 
                    11266275\ M\^74 - 53179837\ M\^76 + 26236505\ M\^78 + 
                    36933924\ M\^80 - 31408311\ M\^82 - 15352644\ M\^84 + 
                    24587292\ M\^86 + 137201\ M\^88 - 12441784\ M\^90 + 
                    4659795\ M\^92 + 3099787\ M\^94 - 2935017\ M\^96 + 
                    368543\ M\^98 + 583210\ M\^100 - 376180\ M\^102 + 
                    106862\ M\^104 - 15456\ M\^106 + 924\ M\^108)\) + 
              L\^3\ \((220\ M\^54 - 2155\ M\^56 + 7784\ M\^58 - 9856\ M\^60 - 
                    10733\ M\^62 + 42614\ M\^64 - 20460\ M\^66 - 
                    59181\ M\^68 + 69151\ M\^70 + 30444\ M\^72 - 
                    79790\ M\^74 + 17562\ M\^76 + 49698\ M\^78 - 
                    38293\ M\^80 - 12109\ M\^82 + 29561\ M\^84 - 
                    8245\ M\^86 - 11726\ M\^88 + 10937\ M\^90 - 653\ M\^92 - 
                    4636\ M\^94 + 3063\ M\^96 - 89\ M\^98 - 1044\ M\^100 + 
                    788\ M\^102 - 323\ M\^104 + 83\ M\^106 - 13\ M\^108 + 
                    M\^110)\) + 
              L\^19\ \((M\^6 - 22\ M\^8 + 226\ M\^10 - 1433\ M\^12 + 
                    6199\ M\^14 - 18790\ M\^16 + 38071\ M\^18 - 
                    39635\ M\^20 - 25425\ M\^22 + 150873\ M\^24 - 
                    155976\ M\^26 - 168802\ M\^28 + 535059\ M\^30 - 
                    155903\ M\^32 - 815512\ M\^34 + 459516\ M\^36 + 
                    1661380\ M\^38 - 1152233\ M\^40 - 4202247\ M\^42 + 
                    4356794\ M\^44 + 8095739\ M\^46 - 12426734\ M\^48 - 
                    9340512\ M\^50 + 21029892\ M\^52 + 5231030\ M\^54 - 
                    16442741\ M\^56 - 2637166\ M\^58 - 14479722\ M\^60 + 
                    18394011\ M\^62 + 65166438\ M\^64 - 62597898\ M\^66 - 
                    106216533\ M\^68 + 122358520\ M\^70 + 109218098\ M\^72 - 
                    167268152\ M\^74 - 72068299\ M\^76 + 175629825\ M\^78 + 
                    20087493\ M\^80 - 145579540\ M\^82 + 20735144\ M\^84 + 
                    94657954\ M\^86 - 37883717\ M\^88 - 45154569\ M\^90 + 
                    33634207\ M\^92 + 12012673\ M\^94 - 19327506\ M\^96 + 
                    2055021\ M\^98 + 6330328\ M\^100 - 3334338\ M\^102 - 
                    304514\ M\^104 + 979290\ M\^106 - 467625\ M\^108 + 
                    113344\ M\^110 - 14574\ M\^112 + 792\ M\^114)\) + 
              L\^4\ \((495\ M\^48 - 6150\ M\^50 + 29699\ M\^52 - 
                    62428\ M\^54 + 8689\ M\^56 + 212465\ M\^58 - 
                    313838\ M\^60 - 177806\ M\^62 + 796943\ M\^64 - 
                    276567\ M\^66 - 994305\ M\^68 + 878908\ M\^70 + 
                    618060\ M\^72 - 1064242\ M\^74 + 84020\ M\^76 + 
                    665322\ M\^78 - 506991\ M\^80 - 33026\ M\^82 + 
                    394945\ M\^84 - 319285\ M\^86 - 37886\ M\^88 + 
                    253647\ M\^90 - 151570\ M\^92 - 47259\ M\^94 + 
                    108885\ M\^96 - 46656\ M\^98 - 16104\ M\^100 + 
                    27710\ M\^102 - 11863\ M\^104 - 1408\ M\^106 + 
                    4620\ M\^108 - 2951\ M\^110 + 1110\ M\^112 - 
                    271\ M\^114 + 41\ M\^116 - 3\ M\^118)\) + 
              L\^18\ \((3\ M\^4 - 67\ M\^6 + 697\ M\^8 - 4460\ M\^10 + 
                    19332\ M\^12 - 57767\ M\^14 + 110527\ M\^16 - 
                    85680\ M\^18 - 180557\ M\^20 + 612928\ M\^22 - 
                    463669\ M\^24 - 1095139\ M\^26 + 2798726\ M\^28 - 
                    834078\ M\^30 - 5344616\ M\^32 + 6889322\ M\^34 + 
                    4313404\ M\^36 - 15165266\ M\^38 + 3365600\ M\^40 + 
                    18445678\ M\^42 - 13792282\ M\^44 - 10718697\ M\^46 + 
                    20892911\ M\^48 - 11803544\ M\^50 - 25088050\ M\^52 + 
                    56899524\ M\^54 + 31137762\ M\^56 - 138139043\ M\^58 - 
                    28506831\ M\^60 + 253443306\ M\^62 - 12220426\ M\^64 - 
                    365770522\ M\^66 + 110421819\ M\^68 + 417636374\ M\^70 - 
                    239448715\ M\^72 - 374500748\ M\^74 + 337017963\ M\^76 + 
                    259286015\ M\^78 - 349325543\ M\^80 - 128203018\ M\^82 + 
                    283372576\ M\^84 + 32904367\ M\^86 - 187164801\ M\^88 + 
                    12597711\ M\^90 + 104030765\ M\^92 - 24598388\ M\^94 - 
                    48839993\ M\^96 + 21885994\ M\^98 + 17619707\ M\^100 - 
                    14485730\ M\^102 - 2432866\ M\^104 + 6249143\ M\^106 - 
                    1818425\ M\^108 - 989765\ M\^110 + 1003468\ M\^112 - 
                    388670\ M\^114 + 83294\ M\^116 - 9780\ M\^118 + 
                    495\ M\^120)\) + 
              L\^5\ \((792\ M\^42 - 11670\ M\^44 + 69208\ M\^46 - 
                    197083\ M\^48 + 181815\ M\^50 + 459752\ M\^52 - 
                    1381307\ M\^54 + 444236\ M\^56 + 2996422\ M\^58 - 
                    3626474\ M\^60 - 2959330\ M\^62 + 7931629\ M\^64 - 
                    63967\ M\^66 - 10196966\ M\^68 + 4504617\ M\^70 + 
                    8197689\ M\^72 - 6816448\ M\^74 - 3186389\ M\^76 + 
                    5400024\ M\^78 - 867805\ M\^80 - 1997062\ M\^82 + 
                    1734120\ M\^84 - 364048\ M\^86 - 611232\ M\^88 + 
                    655883\ M\^90 - 281488\ M\^92 - 17975\ M\^94 + 
                    237937\ M\^96 - 289245\ M\^98 + 82429\ M\^100 + 
                    145159\ M\^102 - 157507\ M\^104 + 31556\ M\^106 + 
                    49862\ M\^108 - 43808\ M\^110 + 9822\ M\^112 + 
                    8381\ M\^114 - 9237\ M\^116 + 4763\ M\^118 - 
                    1578\ M\^120 + 347\ M\^122 - 47\ M\^124 + 3\ M\^126)\) + 
              L\^17\ \((3\ M\^2 - 74\ M\^4 + 858\ M\^6 - 6155\ M\^8 + 
                    29893\ M\^10 - 99680\ M\^12 + 213905\ M\^14 - 
                    207659\ M\^16 - 288421\ M\^18 + 1253704\ M\^20 - 
                    1180370\ M\^22 - 2074585\ M\^24 + 6407922\ M\^26 - 
                    2963353\ M\^28 - 12047943\ M\^30 + 20066369\ M\^32 + 
                    5122374\ M\^34 - 46666978\ M\^36 + 35367949\ M\^38 + 
                    57877925\ M\^40 - 119899192\ M\^42 - 13891649\ M\^44 + 
                    235626736\ M\^46 - 116777183\ M\^48 - 346799161\ M\^50 + 
                    330818479\ M\^52 + 411888583\ M\^54 - 595177469\ M\^56 - 
                    389948479\ M\^58 + 868816929\ M\^60 + 243661935\ M\^62 - 
                    1109322266\ M\^64 + 42103340\ M\^66 + 1250073158\ M\^68 - 
                    417498167\ M\^70 - 1208889601\ M\^72 + 760773222\ M\^74 + 
                    962206790\ M\^76 - 929705778\ M\^78 - 585255193\ M\^80 + 
                    873187466\ M\^82 + 228145146\ M\^84 - 649352519\ M\^86 - 
                    3252635\ M\^88 + 381418141\ M\^90 - 75094223\ M\^92 - 
                    173576328\ M\^94 + 63893449\ M\^96 + 60716403\ M\^98 - 
                    30756057\ M\^100 - 18217880\ M\^102 + 9995249\ M\^104 + 
                    7286309\ M\^106 - 4246132\ M\^108 - 2863078\ M\^110 + 
                    2766343\ M\^112 + 4387\ M\^114 - 1013311\ M\^116 + 
                    650456\ M\^118 - 214763\ M\^120 + 41764\ M\^122 - 
                    4575\ M\^124 + 220\ M\^126)\) + 
              L\^16\ \((1 - 32\ M\^2 + 468\ M\^4 - 4111\ M\^6 + 23911\ M\^8 - 
                    94950\ M\^10 + 250594\ M\^12 - 373442\ M\^14 + 
                    3768\ M\^16 + 1236459\ M\^18 - 2066531\ M\^20 - 
                    878044\ M\^22 + 7625983\ M\^24 - 7479288\ M\^26 - 
                    11440563\ M\^28 + 31726170\ M\^30 - 4593203\ M\^32 - 
                    72138115\ M\^34 + 77497965\ M\^36 + 97215384\ M\^38 - 
                    241463782\ M\^40 - 36809748\ M\^42 + 489747949\ M\^44 - 
                    194222578\ M\^46 - 738670115\ M\^48 + 633223576\ M\^50 + 
                    858160241\ M\^52 - 1238315717\ M\^54 - 750676250\ M\^56 + 
                    1908980745\ M\^58 + 412436761\ M\^60 - 
                    2528272420\ M\^62 + 93512712\ M\^64 + 2966172648\ M\^66 - 
                    692385745\ M\^68 - 3086265175\ M\^70 + 
                    1301031731\ M\^72 + 2805451397\ M\^74 - 
                    1773965439\ M\^76 - 2155788996\ M\^78 + 
                    1958151930\ M\^80 + 1333232675\ M\^82 - 
                    1777527022\ M\^84 - 587793042\ M\^86 + 
                    1328272367\ M\^88 + 94635067\ M\^90 - 811353514\ M\^92 + 
                    119211816\ M\^94 + 396584593\ M\^96 - 144295253\ M\^98 - 
                    145454694\ M\^100 + 93084096\ M\^102 + 32463075\ M\^104 - 
                    39889837\ M\^106 + 1285290\ M\^108 + 9252912\ M\^110 - 
                    2810161\ M\^112 - 69606\ M\^114 - 206956\ M\^116 - 
                    105491\ M\^118 + 601204\ M\^120 - 549902\ M\^122 + 
                    261458\ M\^124 - 75778\ M\^126 + 13643\ M\^128 - 
                    1420\ M\^130 + 66\ M\^132)\) + 
              L\^6\ \((924\ M\^36 - 15456\ M\^38 + 106862\ M\^40 - 
                    376180\ M\^42 + 583210\ M\^44 + 368543\ M\^46 - 
                    2935017\ M\^48 + 3099787\ M\^50 + 4659795\ M\^52 - 
                    12441784\ M\^54 + 137201\ M\^56 + 24587292\ M\^58 - 
                    15352644\ M\^60 - 31408311\ M\^62 + 36933924\ M\^64 + 
                    26236505\ M\^66 - 53179837\ M\^68 - 11266275\ M\^70 + 
                    53570095\ M\^72 - 3350758\ M\^74 - 37595867\ M\^76 + 
                    9377115\ M\^78 + 16514583\ M\^80 - 5786473\ M\^82 - 
                    2503463\ M\^84 - 1291297\ M\^86 - 590926\ M\^88 + 
                    4880377\ M\^90 - 1950640\ M\^92 - 3338647\ M\^94 + 
                    3455453\ M\^96 + 179710\ M\^98 - 2170573\ M\^100 + 
                    1150398\ M\^102 + 388396\ M\^104 - 706988\ M\^106 + 
                    262656\ M\^108 + 65603\ M\^110 - 108914\ M\^112 + 
                    58113\ M\^114 - 17680\ M\^116 - 5243\ M\^118 + 
                    14556\ M\^120 - 13009\ M\^122 + 7321\ M\^124 - 
                    2869\ M\^126 + 797\ M\^128 - 152\ M\^130 + 18\ M\^132 - 
                    M\^134)\) + 
              L\^7\ \((792\ M\^30 - 14574\ M\^32 + 113344\ M\^34 - 
                    467625\ M\^36 + 979290\ M\^38 - 304514\ M\^40 - 
                    3334338\ M\^42 + 6330328\ M\^44 + 2055021\ M\^46 - 
                    19327506\ M\^48 + 12012673\ M\^50 + 33634207\ M\^52 - 
                    45154569\ M\^54 - 37883717\ M\^56 + 94657954\ M\^58 + 
                    20735144\ M\^60 - 145579540\ M\^62 + 20087493\ M\^64 + 
                    175629825\ M\^66 - 72068299\ M\^68 - 167268152\ M\^70 + 
                    109218098\ M\^72 + 122358520\ M\^74 - 106216533\ M\^76 - 
                    62597898\ M\^78 + 65166438\ M\^80 + 18394011\ M\^82 - 
                    14479722\ M\^84 - 2637166\ M\^86 - 16442741\ M\^88 + 
                    5231030\ M\^90 + 21029892\ M\^92 - 9340512\ M\^94 - 
                    12426734\ M\^96 + 8095739\ M\^98 + 4356794\ M\^100 - 
                    4202247\ M\^102 - 1152233\ M\^104 + 1661380\ M\^106 + 
                    459516\ M\^108 - 815512\ M\^110 - 155903\ M\^112 + 
                    535059\ M\^114 - 168802\ M\^116 - 155976\ M\^118 + 
                    150873\ M\^120 - 25425\ M\^122 - 39635\ M\^124 + 
                    38071\ M\^126 - 18790\ M\^128 + 6199\ M\^130 - 
                    1433\ M\^132 + 226\ M\^134 - 22\ M\^136 + M\^138)\) + 
              L\^15\ \((\(-4\) + 104\ M\^2 - 1271\ M\^4 + 9502\ M\^6 - 
                    47387\ M\^8 + 160357\ M\^10 - 349139\ M\^12 + 
                    370092\ M\^14 + 279501\ M\^16 - 1417222\ M\^18 + 
                    860313\ M\^20 + 3331652\ M\^22 - 5964228\ M\^24 - 
                    3611844\ M\^26 + 18800991\ M\^28 - 4932249\ M\^30 - 
                    45013776\ M\^32 + 47260380\ M\^34 + 70734159\ M\^36 - 
                    159172265\ M\^38 - 40180146\ M\^40 + 362071215\ M\^42 - 
                    159219189\ M\^44 - 608412016\ M\^46 + 645704341\ M\^48 + 
                    769645474\ M\^50 - 1450302960\ M\^52 - 686461659\ M\^54 + 
                    2434391786\ M\^56 + 270257496\ M\^58 - 
                    3334966174\ M\^60 + 444978225\ M\^62 + 
                    3913582862\ M\^64 - 1323822634\ M\^66 - 
                    4076152947\ M\^68 + 2188140623\ M\^70 + 
                    3841044211\ M\^72 - 2844524750\ M\^74 - 
                    3239648254\ M\^76 + 3148743217\ M\^78 + 
                    2366521570\ M\^80 - 3011614443\ M\^82 - 
                    1398365912\ M\^84 + 2486412699\ M\^86 + 
                    574573949\ M\^88 - 1752969729\ M\^90 - 59713050\ M\^92 + 
                    1039869268\ M\^94 - 146077151\ M\^96 - 510371943\ M\^98 + 
                    156778804\ M\^100 + 204631074\ M\^102 - 
                    100898165\ M\^104 - 65208368\ M\^106 + 49551124\ M\^108 + 
                    15066101\ M\^110 - 20975106\ M\^112 + 415289\ M\^114 + 
                    6453109\ M\^116 - 2511686\ M\^118 - 601421\ M\^120 + 
                    993359\ M\^122 - 621957\ M\^124 + 338238\ M\^126 - 
                    162765\ M\^128 + 59975\ M\^130 - 15455\ M\^132 + 
                    2616\ M\^134 - 263\ M\^136 + 12\ M\^138)\) + 
              L\^8\ \((495\ M\^24 - 9780\ M\^26 + 83294\ M\^28 - 
                    388670\ M\^30 + 1003468\ M\^32 - 989765\ M\^34 - 
                    1818425\ M\^36 + 6249143\ M\^38 - 2432866\ M\^40 - 
                    14485730\ M\^42 + 17619707\ M\^44 + 21885994\ M\^46 - 
                    48839993\ M\^48 - 24598388\ M\^50 + 104030765\ M\^52 + 
                    12597711\ M\^54 - 187164801\ M\^56 + 32904367\ M\^58 + 
                    283372576\ M\^60 - 128203018\ M\^62 - 349325543\ M\^64 + 
                    259286015\ M\^66 + 337017963\ M\^68 - 374500748\ M\^70 - 
                    239448715\ M\^72 + 417636374\ M\^74 + 110421819\ M\^76 - 
                    365770522\ M\^78 - 12220426\ M\^80 + 253443306\ M\^82 - 
                    28506831\ M\^84 - 138139043\ M\^86 + 31137762\ M\^88 + 
                    56899524\ M\^90 - 25088050\ M\^92 - 11803544\ M\^94 + 
                    20892911\ M\^96 - 10718697\ M\^98 - 13792282\ M\^100 + 
                    18445678\ M\^102 + 3365600\ M\^104 - 15165266\ M\^106 + 
                    4313404\ M\^108 + 6889322\ M\^110 - 5344616\ M\^112 - 
                    834078\ M\^114 + 2798726\ M\^116 - 1095139\ M\^118 - 
                    463669\ M\^120 + 612928\ M\^122 - 180557\ M\^124 - 
                    85680\ M\^126 + 110527\ M\^128 - 57767\ M\^130 + 
                    19332\ M\^132 - 4460\ M\^134 + 697\ M\^136 - 67\ M\^138 + 
                    3\ M\^140)\) + 
              L\^9\ \((220\ M\^18 - 4575\ M\^20 + 41764\ M\^22 - 
                    214763\ M\^24 + 650456\ M\^26 - 1013311\ M\^28 + 
                    4387\ M\^30 + 2766343\ M\^32 - 2863078\ M\^34 - 
                    4246132\ M\^36 + 7286309\ M\^38 + 9995249\ M\^40 - 
                    18217880\ M\^42 - 30756057\ M\^44 + 60716403\ M\^46 + 
                    63893449\ M\^48 - 173576328\ M\^50 - 75094223\ M\^52 + 
                    381418141\ M\^54 - 3252635\ M\^56 - 649352519\ M\^58 + 
                    228145146\ M\^60 + 873187466\ M\^62 - 585255193\ M\^64 - 
                    929705778\ M\^66 + 962206790\ M\^68 + 760773222\ M\^70 - 
                    1208889601\ M\^72 - 417498167\ M\^74 + 
                    1250073158\ M\^76 + 42103340\ M\^78 - 1109322266\ M\^80 + 
                    243661935\ M\^82 + 868816929\ M\^84 - 389948479\ M\^86 - 
                    595177469\ M\^88 + 411888583\ M\^90 + 330818479\ M\^92 - 
                    346799161\ M\^94 - 116777183\ M\^96 + 235626736\ M\^98 - 
                    13891649\ M\^100 - 119899192\ M\^102 + 57877925\ M\^104 + 
                    35367949\ M\^106 - 46666978\ M\^108 + 5122374\ M\^110 + 
                    20066369\ M\^112 - 12047943\ M\^114 - 2963353\ M\^116 + 
                    6407922\ M\^118 - 2074585\ M\^120 - 1180370\ M\^122 + 
                    1253704\ M\^124 - 288421\ M\^126 - 207659\ M\^128 + 
                    213905\ M\^130 - 99680\ M\^132 + 29893\ M\^134 - 
                    6155\ M\^136 + 858\ M\^138 - 74\ M\^140 + 3\ M\^142)\) + 
              L\^13\ \((\(-4\) + 93\ M\^2 - 991\ M\^4 + 6361\ M\^6 - 
                    27235\ M\^8 + 82073\ M\^10 - 184023\ M\^12 + 
                    341002\ M\^14 - 580270\ M\^16 + 731811\ M\^18 + 
                    193348\ M\^20 - 3090667\ M\^22 + 3919836\ M\^24 + 
                    6078596\ M\^26 - 18781807\ M\^28 - 3137549\ M\^30 + 
                    55045155\ M\^32 - 21015937\ M\^34 - 133364118\ M\^36 + 
                    112764667\ M\^38 + 258572752\ M\^40 - 343448250\ M\^42 - 
                    392690386\ M\^44 + 777733412\ M\^46 + 429418649\ M\^48 - 
                    1392183492\ M\^50 - 242180905\ M\^52 + 
                    2030712537\ M\^54 - 220892834\ M\^56 - 
                    2434647758\ M\^58 + 851619890\ M\^60 + 
                    2365995562\ M\^62 - 1439844718\ M\^64 - 
                    1730198226\ M\^66 + 1818290140\ M\^68 + 
                    638476065\ M\^70 - 1937024958\ M\^72 + 638476065\ M\^74 + 
                    1818290140\ M\^76 - 1730198226\ M\^78 - 
                    1439844718\ M\^80 + 2365995562\ M\^82 + 
                    851619890\ M\^84 - 2434647758\ M\^86 - 220892834\ M\^88 + 
                    2030712537\ M\^90 - 242180905\ M\^92 - 
                    1392183492\ M\^94 + 429418649\ M\^96 + 777733412\ M\^98 - 
                    392690386\ M\^100 - 343448250\ M\^102 + 
                    258572752\ M\^104 + 112764667\ M\^106 - 
                    133364118\ M\^108 - 21015937\ M\^110 + 55045155\ M\^112 - 
                    3137549\ M\^114 - 18781807\ M\^116 + 6078596\ M\^118 + 
                    3919836\ M\^120 - 3090667\ M\^122 + 193348\ M\^124 + 
                    731811\ M\^126 - 580270\ M\^128 + 341002\ M\^130 - 
                    184023\ M\^132 + 82073\ M\^134 - 27235\ M\^136 + 
                    6361\ M\^138 - 991\ M\^140 + 93\ M\^142 - 4\ M\^144)\) + 
              L\^11\ \((12\ M\^6 - 263\ M\^8 + 2616\ M\^10 - 15455\ M\^12 + 
                    59975\ M\^14 - 162765\ M\^16 + 338238\ M\^18 - 
                    621957\ M\^20 + 993359\ M\^22 - 601421\ M\^24 - 
                    2511686\ M\^26 + 6453109\ M\^28 + 415289\ M\^30 - 
                    20975106\ M\^32 + 15066101\ M\^34 + 49551124\ M\^36 - 
                    65208368\ M\^38 - 100898165\ M\^40 + 204631074\ M\^42 + 
                    156778804\ M\^44 - 510371943\ M\^46 - 146077151\ M\^48 + 
                    1039869268\ M\^50 - 59713050\ M\^52 - 1752969729\ M\^54 + 
                    574573949\ M\^56 + 2486412699\ M\^58 - 
                    1398365912\ M\^60 - 3011614443\ M\^62 + 
                    2366521570\ M\^64 + 3148743217\ M\^66 - 
                    3239648254\ M\^68 - 2844524750\ M\^70 + 
                    3841044211\ M\^72 + 2188140623\ M\^74 - 
                    4076152947\ M\^76 - 1323822634\ M\^78 + 
                    3913582862\ M\^80 + 444978225\ M\^82 - 
                    3334966174\ M\^84 + 270257496\ M\^86 + 
                    2434391786\ M\^88 - 686461659\ M\^90 - 
                    1450302960\ M\^92 + 769645474\ M\^94 + 645704341\ M\^96 - 
                    608412016\ M\^98 - 159219189\ M\^100 + 
                    362071215\ M\^102 - 40180146\ M\^104 - 
                    159172265\ M\^106 + 70734159\ M\^108 + 47260380\ M\^110 - 
                    45013776\ M\^112 - 4932249\ M\^114 + 18800991\ M\^116 - 
                    3611844\ M\^118 - 5964228\ M\^120 + 3331652\ M\^122 + 
                    860313\ M\^124 - 1417222\ M\^126 + 279501\ M\^128 + 
                    370092\ M\^130 - 349139\ M\^132 + 160357\ M\^134 - 
                    47387\ M\^136 + 9502\ M\^138 - 1271\ M\^140 + 
                    104\ M\^142 - 4\ M\^144)\) + 
              L\^10\ \((66\ M\^12 - 1420\ M\^14 + 13643\ M\^16 - 
                    75778\ M\^18 + 261458\ M\^20 - 549902\ M\^22 + 
                    601204\ M\^24 - 105491\ M\^26 - 206956\ M\^28 - 
                    69606\ M\^30 - 2810161\ M\^32 + 9252912\ M\^34 + 
                    1285290\ M\^36 - 39889837\ M\^38 + 32463075\ M\^40 + 
                    93084096\ M\^42 - 145454694\ M\^44 - 144295253\ M\^46 + 
                    396584593\ M\^48 + 119211816\ M\^50 - 811353514\ M\^52 + 
                    94635067\ M\^54 + 1328272367\ M\^56 - 587793042\ M\^58 - 
                    1777527022\ M\^60 + 1333232675\ M\^62 + 
                    1958151930\ M\^64 - 2155788996\ M\^66 - 
                    1773965439\ M\^68 + 2805451397\ M\^70 + 
                    1301031731\ M\^72 - 3086265175\ M\^74 - 
                    692385745\ M\^76 + 2966172648\ M\^78 + 93512712\ M\^80 - 
                    2528272420\ M\^82 + 412436761\ M\^84 + 
                    1908980745\ M\^86 - 750676250\ M\^88 - 
                    1238315717\ M\^90 + 858160241\ M\^92 + 633223576\ M\^94 - 
                    738670115\ M\^96 - 194222578\ M\^98 + 489747949\ M\^100 - 
                    36809748\ M\^102 - 241463782\ M\^104 + 97215384\ M\^106 + 
                    77497965\ M\^108 - 72138115\ M\^110 - 4593203\ M\^112 + 
                    31726170\ M\^114 - 11440563\ M\^116 - 7479288\ M\^118 + 
                    7625983\ M\^120 - 878044\ M\^122 - 2066531\ M\^124 + 
                    1236459\ M\^126 + 3768\ M\^128 - 373442\ M\^130 + 
                    250594\ M\^132 - 94950\ M\^134 + 23911\ M\^136 - 
                    4111\ M\^138 + 468\ M\^140 - 32\ M\^142 + M\^144)\) + 
              L\^14\ \((6 - 146\ M\^2 + 1642\ M\^4 - 11166\ M\^6 + 
                    50248\ M\^8 - 152775\ M\^10 + 301537\ M\^12 - 
                    325502\ M\^14 + 46547\ M\^16 + 174083\ M\^18 + 
                    298051\ M\^20 - 162116\ M\^22 - 3154088\ M\^24 + 
                    4294498\ M\^26 + 9938770\ M\^28 - 26327567\ M\^30 - 
                    6974642\ M\^32 + 77895027\ M\^34 - 38285248\ M\^36 - 
                    153330567\ M\^38 + 177237200\ M\^40 + 192893610\ M\^42 - 
                    429981318\ M\^44 - 85320499\ M\^46 + 721807940\ M\^48 - 
                    280756253\ M\^50 - 846416216\ M\^52 + 917179888\ M\^54 + 
                    558621846\ M\^56 - 1703358604\ M\^58 + 251465320\ M\^60 + 
                    2434795160\ M\^62 - 1437361561\ M\^64 - 
                    2915678142\ M\^66 + 2672259060\ M\^68 + 
                    3013727320\ M\^70 - 3637872915\ M\^72 - 
                    2684648051\ M\^74 + 4171192885\ M\^76 + 
                    2025211547\ M\^78 - 4192079490\ M\^80 - 
                    1199794775\ M\^82 + 3723556474\ M\^84 + 
                    421502966\ M\^86 - 2888776710\ M\^88 + 132661330\ M\^90 + 
                    1920456955\ M\^92 - 389970078\ M\^94 - 
                    1066933130\ M\^96 + 396359771\ M\^98 + 
                    477424787\ M\^100 - 275962213\ M\^102 - 
                    158529426\ M\^104 + 142272450\ M\^106 + 
                    31050159\ M\^108 - 53939610\ M\^110 + 1481334\ M\^112 + 
                    13106399\ M\^114 - 1620921\ M\^116 - 2408138\ M\^118 - 
                    231670\ M\^120 + 998420\ M\^122 + 86581\ M\^124 - 
                    671697\ M\^126 + 513487\ M\^128 - 230056\ M\^130 + 
                    77731\ M\^132 - 23098\ M\^134 + 6251\ M\^136 - 
                    1394\ M\^138 + 223\ M\^140 - 22\ M\^142 + M\^144)\) + 
              L\^12\ \((1 - 22\ M\^2 + 223\ M\^4 - 1394\ M\^6 + 6251\ M\^8 - 
                    23098\ M\^10 + 77731\ M\^12 - 230056\ M\^14 + 
                    513487\ M\^16 - 671697\ M\^18 + 86581\ M\^20 + 
                    998420\ M\^22 - 231670\ M\^24 - 2408138\ M\^26 - 
                    1620921\ M\^28 + 13106399\ M\^30 + 1481334\ M\^32 - 
                    53939610\ M\^34 + 31050159\ M\^36 + 142272450\ M\^38 - 
                    158529426\ M\^40 - 275962213\ M\^42 + 477424787\ M\^44 + 
                    396359771\ M\^46 - 1066933130\ M\^48 - 389970078\ M\^50 + 
                    1920456955\ M\^52 + 132661330\ M\^54 - 
                    2888776710\ M\^56 + 421502966\ M\^58 + 
                    3723556474\ M\^60 - 1199794775\ M\^62 - 
                    4192079490\ M\^64 + 2025211547\ M\^66 + 
                    4171192885\ M\^68 - 2684648051\ M\^70 - 
                    3637872915\ M\^72 + 3013727320\ M\^74 + 
                    2672259060\ M\^76 - 2915678142\ M\^78 - 
                    1437361561\ M\^80 + 2434795160\ M\^82 + 
                    251465320\ M\^84 - 1703358604\ M\^86 + 558621846\ M\^88 + 
                    917179888\ M\^90 - 846416216\ M\^92 - 280756253\ M\^94 + 
                    721807940\ M\^96 - 85320499\ M\^98 - 429981318\ M\^100 + 
                    192893610\ M\^102 + 177237200\ M\^104 - 
                    153330567\ M\^106 - 38285248\ M\^108 + 77895027\ M\^110 - 
                    6974642\ M\^112 - 26327567\ M\^114 + 9938770\ M\^116 + 
                    4294498\ M\^118 - 3154088\ M\^120 - 162116\ M\^122 + 
                    298051\ M\^124 + 174083\ M\^126 + 46547\ M\^128 - 
                    325502\ M\^130 + 301537\ M\^132 - 152775\ M\^134 + 
                    50248\ M\^136 - 11166\ M\^138 + 1642\ M\^140 - 
                    146\ M\^142 + 6\ M\^144)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 29]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 30]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 31]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 32]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 33]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 34]\), \(M\^46 + L\^18\ M\^46 + 
              L\^17\ \((2\ M\^32 - 6\ M\^34 + 7\ M\^36 - 3\ M\^38 + 
                    3\ M\^42 - 14\ M\^44 + 27\ M\^46 + 17\ M\^48 - 
                    21\ M\^50 + 6\ M\^52)\) + 
              L\^16\ \((M\^18 - 6\ M\^20 + 17\ M\^22 - 28\ M\^24 + 
                    27\ M\^26 + 12\ M\^28 - 123\ M\^30 + 132\ M\^32 + 
                    197\ M\^34 - 384\ M\^36 - 66\ M\^38 + 478\ M\^40 - 
                    153\ M\^42 - 446\ M\^44 + 449\ M\^46 + 406\ M\^48 - 
                    435\ M\^50 - 64\ M\^52 + 224\ M\^54 - 100\ M\^56 + 
                    15\ M\^58)\) + 
              L\ \((6\ M\^40 - 21\ M\^42 + 17\ M\^44 + 27\ M\^46 - 
                    14\ M\^48 + 3\ M\^50 - 3\ M\^54 + 7\ M\^56 - 6\ M\^58 + 
                    2\ M\^60)\) + 
              L\^15\ \((M\^8 - 7\ M\^10 + 21\ M\^12 - 37\ M\^14 + 53\ M\^16 - 
                    62\ M\^18 - 39\ M\^20 + 283\ M\^22 - 266\ M\^24 - 
                    410\ M\^26 + 1265\ M\^28 - 734\ M\^30 - 1986\ M\^32 + 
                    2791\ M\^34 + 1581\ M\^36 - 3606\ M\^38 + 46\ M\^40 + 
                    1333\ M\^42 - 1418\ M\^44 + 2171\ M\^46 + 2380\ M\^48 - 
                    3173\ M\^50 - 1069\ M\^52 + 2577\ M\^54 - 452\ M\^56 - 
                    936\ M\^58 + 679\ M\^60 - 190\ M\^62 + 20\ M\^64)\) + 
              L\^14\ \((\(-2\)\ M\^4 + 20\ M\^6 - 85\ M\^8 + 186\ M\^10 - 
                    150\ M\^12 - 218\ M\^14 + 435\ M\^16 + 545\ M\^18 - 
                    1467\ M\^20 - 1001\ M\^22 + 3167\ M\^24 + 1720\ M\^26 - 
                    4576\ M\^28 - 625\ M\^30 - 330\ M\^32 - 628\ M\^34 + 
                    10570\ M\^36 + 1745\ M\^38 - 15101\ M\^40 - 3949\ M\^42 + 
                    6168\ M\^44 + 9410\ M\^46 + 7935\ M\^48 - 10422\ M\^50 - 
                    9371\ M\^52 + 10560\ M\^54 + 3109\ M\^56 - 7226\ M\^58 + 
                    1992\ M\^60 + 2054\ M\^62 - 2114\ M\^64 + 874\ M\^66 - 
                    180\ M\^68 + 15\ M\^70)\) + 
              L\^2\ \((15\ M\^34 - 100\ M\^36 + 224\ M\^38 - 64\ M\^40 - 
                    435\ M\^42 + 406\ M\^44 + 449\ M\^46 - 446\ M\^48 - 
                    153\ M\^50 + 478\ M\^52 - 66\ M\^54 - 384\ M\^56 + 
                    197\ M\^58 + 132\ M\^60 - 123\ M\^62 + 12\ M\^64 + 
                    27\ M\^66 - 28\ M\^68 + 17\ M\^70 - 6\ M\^72 + M\^74)\) + 
              L\^13\ \((1 - 13\ M\^2 + 70\ M\^4 - 195\ M\^6 + 297\ M\^8 - 
                    354\ M\^10 + 777\ M\^12 - 840\ M\^14 - 2724\ M\^16 + 
                    6057\ M\^18 + 6070\ M\^20 - 19304\ M\^22 - 10372\ M\^24 + 
                    36312\ M\^26 + 21916\ M\^28 - 47415\ M\^30 - 
                    38321\ M\^32 + 31200\ M\^34 + 52058\ M\^36 + 
                    10074\ M\^38 - 43775\ M\^40 - 43185\ M\^42 + 
                    8571\ M\^44 + 45501\ M\^46 + 38097\ M\^48 - 
                    25471\ M\^50 - 48622\ M\^52 + 21346\ M\^54 + 
                    33244\ M\^56 - 23250\ M\^58 - 10105\ M\^60 + 
                    16038\ M\^62 - 3398\ M\^64 - 4507\ M\^66 + 4251\ M\^68 - 
                    1909\ M\^70 + 527\ M\^72 - 85\ M\^74 + 6\ M\^76)\) + 
              L\^12\ \((\(-2\) + 31\ M\^2 - 214\ M\^4 + 800\ M\^6 - 
                    1499\ M\^8 + 482\ M\^10 + 2415\ M\^12 + 334\ M\^14 - 
                    8977\ M\^16 - 4904\ M\^18 + 31076\ M\^20 + 22153\ M\^22 - 
                    84182\ M\^24 - 62530\ M\^26 + 152750\ M\^28 + 
                    150543\ M\^30 - 194801\ M\^32 - 271179\ M\^34 + 
                    146570\ M\^36 + 366771\ M\^38 - 14111\ M\^40 - 
                    358449\ M\^42 - 110525\ M\^44 + 249367\ M\^46 + 
                    163953\ M\^48 - 102383\ M\^50 - 134560\ M\^52 + 
                    36540\ M\^54 + 93824\ M\^56 - 30717\ M\^58 - 
                    55996\ M\^60 + 35897\ M\^62 + 17777\ M\^64 - 
                    24997\ M\^66 + 3966\ M\^68 + 7897\ M\^70 - 6726\ M\^72 + 
                    2777\ M\^74 - 722\ M\^76 + 130\ M\^78 - 16\ M\^80 + 
                    M\^82)\) + 
              L\^3\ \((20\ M\^28 - 190\ M\^30 + 679\ M\^32 - 936\ M\^34 - 
                    452\ M\^36 + 2577\ M\^38 - 1069\ M\^40 - 3173\ M\^42 + 
                    2380\ M\^44 + 2171\ M\^46 - 1418\ M\^48 + 1333\ M\^50 + 
                    46\ M\^52 - 3606\ M\^54 + 1581\ M\^56 + 2791\ M\^58 - 
                    1986\ M\^60 - 734\ M\^62 + 1265\ M\^64 - 410\ M\^66 - 
                    266\ M\^68 + 283\ M\^70 - 39\ M\^72 - 62\ M\^74 + 
                    53\ M\^76 - 37\ M\^78 + 21\ M\^80 - 7\ M\^82 + M\^84)\) + 
              L\^11\ \((1 - 23\ M\^2 + 228\ M\^4 - 1221\ M\^6 + 3598\ M\^8 - 
                    4398\ M\^10 - 4604\ M\^12 + 17086\ M\^14 + 6598\ M\^16 - 
                    54116\ M\^18 - 10595\ M\^20 + 133602\ M\^22 + 
                    46574\ M\^24 - 269996\ M\^26 - 164496\ M\^28 + 
                    401402\ M\^30 + 423859\ M\^32 - 422112\ M\^34 - 
                    758741\ M\^36 + 239767\ M\^38 + 990751\ M\^40 + 
                    71982\ M\^42 - 952087\ M\^44 - 295681\ M\^46 + 
                    711669\ M\^48 + 344701\ M\^50 - 414632\ M\^52 - 
                    235526\ M\^54 + 241190\ M\^56 + 109355\ M\^58 - 
                    154949\ M\^60 - 26688\ M\^62 + 93089\ M\^64 - 
                    18043\ M\^66 - 35710\ M\^68 + 23092\ M\^70 + 
                    2530\ M\^72 - 10499\ M\^74 + 6937\ M\^76 - 2601\ M\^78 + 
                    609\ M\^80 - 83\ M\^82 + 5\ M\^84)\) + 
              L\^10\ \((5\ M\^2 - 92\ M\^4 + 760\ M\^6 - 3460\ M\^8 + 
                    8410\ M\^10 - 6038\ M\^12 - 19234\ M\^14 + 36831\ M\^16 + 
                    37814\ M\^18 - 122821\ M\^20 - 67038\ M\^22 + 
                    286499\ M\^24 + 158352\ M\^26 - 517621\ M\^28 - 
                    380247\ M\^30 + 676969\ M\^32 + 759039\ M\^34 - 
                    601873\ M\^36 - 1110589\ M\^38 + 247477\ M\^40 + 
                    1174779\ M\^42 + 178499\ M\^44 - 848050\ M\^46 - 
                    348671\ M\^48 + 420648\ M\^50 + 265191\ M\^52 - 
                    116513\ M\^54 - 65753\ M\^56 + 34050\ M\^58 - 
                    66413\ M\^60 - 32587\ M\^62 + 92101\ M\^64 + 
                    21193\ M\^66 - 72481\ M\^68 + 6111\ M\^70 + 
                    35155\ M\^72 - 15270\ M\^74 - 8333\ M\^76 + 
                    10867\ M\^78 - 4996\ M\^80 + 1248\ M\^82 - 170\ M\^84 + 
                    10\ M\^86)\) + 
              L\^4\ \((15\ M\^22 - 180\ M\^24 + 874\ M\^26 - 2114\ M\^28 + 
                    2054\ M\^30 + 1992\ M\^32 - 7226\ M\^34 + 3109\ M\^36 + 
                    10560\ M\^38 - 9371\ M\^40 - 10422\ M\^42 + 7935\ M\^44 + 
                    9410\ M\^46 + 6168\ M\^48 - 3949\ M\^50 - 15101\ M\^52 + 
                    1745\ M\^54 + 10570\ M\^56 - 628\ M\^58 - 330\ M\^60 - 
                    625\ M\^62 - 4576\ M\^64 + 1720\ M\^66 + 3167\ M\^68 - 
                    1001\ M\^70 - 1467\ M\^72 + 545\ M\^74 + 435\ M\^76 - 
                    218\ M\^78 - 150\ M\^80 + 186\ M\^82 - 85\ M\^84 + 
                    20\ M\^86 - 2\ M\^88)\) + 
              L\^9\ \((10\ M\^4 - 174\ M\^6 + 1325\ M\^8 - 5473\ M\^10 + 
                    11737\ M\^12 - 5759\ M\^14 - 27725\ M\^16 + 
                    40888\ M\^18 + 52570\ M\^20 - 126040\ M\^22 - 
                    91312\ M\^24 + 261860\ M\^26 + 197932\ M\^28 - 
                    414449\ M\^30 - 403187\ M\^32 + 442318\ M\^34 + 
                    634124\ M\^36 - 255259\ M\^38 - 640862\ M\^40 - 
                    52465\ M\^42 + 289698\ M\^44 + 229106\ M\^46 + 
                    289698\ M\^48 - 52465\ M\^50 - 640862\ M\^52 - 
                    255259\ M\^54 + 634124\ M\^56 + 442318\ M\^58 - 
                    403187\ M\^60 - 414449\ M\^62 + 197932\ M\^64 + 
                    261860\ M\^66 - 91312\ M\^68 - 126040\ M\^70 + 
                    52570\ M\^72 + 40888\ M\^74 - 27725\ M\^76 - 
                    5759\ M\^78 + 11737\ M\^80 - 5473\ M\^82 + 1325\ M\^84 - 
                    174\ M\^86 + 10\ M\^88)\) + 
              L\^8\ \((10\ M\^6 - 170\ M\^8 + 1248\ M\^10 - 4996\ M\^12 + 
                    10867\ M\^14 - 8333\ M\^16 - 15270\ M\^18 + 
                    35155\ M\^20 + 6111\ M\^22 - 72481\ M\^24 + 
                    21193\ M\^26 + 92101\ M\^28 - 32587\ M\^30 - 
                    66413\ M\^32 + 34050\ M\^34 - 65753\ M\^36 - 
                    116513\ M\^38 + 265191\ M\^40 + 420648\ M\^42 - 
                    348671\ M\^44 - 848050\ M\^46 + 178499\ M\^48 + 
                    1174779\ M\^50 + 247477\ M\^52 - 1110589\ M\^54 - 
                    601873\ M\^56 + 759039\ M\^58 + 676969\ M\^60 - 
                    380247\ M\^62 - 517621\ M\^64 + 158352\ M\^66 + 
                    286499\ M\^68 - 67038\ M\^70 - 122821\ M\^72 + 
                    37814\ M\^74 + 36831\ M\^76 - 19234\ M\^78 - 
                    6038\ M\^80 + 8410\ M\^82 - 3460\ M\^84 + 760\ M\^86 - 
                    92\ M\^88 + 5\ M\^90)\) + 
              L\^6\ \((M\^10 - 16\ M\^12 + 130\ M\^14 - 722\ M\^16 + 
                    2777\ M\^18 - 6726\ M\^20 + 7897\ M\^22 + 3966\ M\^24 - 
                    24997\ M\^26 + 17777\ M\^28 + 35897\ M\^30 - 
                    55996\ M\^32 - 30717\ M\^34 + 93824\ M\^36 + 
                    36540\ M\^38 - 134560\ M\^40 - 102383\ M\^42 + 
                    163953\ M\^44 + 249367\ M\^46 - 110525\ M\^48 - 
                    358449\ M\^50 - 14111\ M\^52 + 366771\ M\^54 + 
                    146570\ M\^56 - 271179\ M\^58 - 194801\ M\^60 + 
                    150543\ M\^62 + 152750\ M\^64 - 62530\ M\^66 - 
                    84182\ M\^68 + 22153\ M\^70 + 31076\ M\^72 - 
                    4904\ M\^74 - 8977\ M\^76 + 334\ M\^78 + 2415\ M\^80 + 
                    482\ M\^82 - 1499\ M\^84 + 800\ M\^86 - 214\ M\^88 + 
                    31\ M\^90 - 2\ M\^92)\) + 
              L\^7\ \((5\ M\^8 - 83\ M\^10 + 609\ M\^12 - 2601\ M\^14 + 
                    6937\ M\^16 - 10499\ M\^18 + 2530\ M\^20 + 23092\ M\^22 - 
                    35710\ M\^24 - 18043\ M\^26 + 93089\ M\^28 - 
                    26688\ M\^30 - 154949\ M\^32 + 109355\ M\^34 + 
                    241190\ M\^36 - 235526\ M\^38 - 414632\ M\^40 + 
                    344701\ M\^42 + 711669\ M\^44 - 295681\ M\^46 - 
                    952087\ M\^48 + 71982\ M\^50 + 990751\ M\^52 + 
                    239767\ M\^54 - 758741\ M\^56 - 422112\ M\^58 + 
                    423859\ M\^60 + 401402\ M\^62 - 164496\ M\^64 - 
                    269996\ M\^66 + 46574\ M\^68 + 133602\ M\^70 - 
                    10595\ M\^72 - 54116\ M\^74 + 6598\ M\^76 + 
                    17086\ M\^78 - 4604\ M\^80 - 4398\ M\^82 + 3598\ M\^84 - 
                    1221\ M\^86 + 228\ M\^88 - 23\ M\^90 + M\^92)\) + 
              L\^5\ \((6\ M\^16 - 85\ M\^18 + 527\ M\^20 - 1909\ M\^22 + 
                    4251\ M\^24 - 4507\ M\^26 - 3398\ M\^28 + 16038\ M\^30 - 
                    10105\ M\^32 - 23250\ M\^34 + 33244\ M\^36 + 
                    21346\ M\^38 - 48622\ M\^40 - 25471\ M\^42 + 
                    38097\ M\^44 + 45501\ M\^46 + 8571\ M\^48 - 
                    43185\ M\^50 - 43775\ M\^52 + 10074\ M\^54 + 
                    52058\ M\^56 + 31200\ M\^58 - 38321\ M\^60 - 
                    47415\ M\^62 + 21916\ M\^64 + 36312\ M\^66 - 
                    10372\ M\^68 - 19304\ M\^70 + 6070\ M\^72 + 6057\ M\^74 - 
                    2724\ M\^76 - 840\ M\^78 + 777\ M\^80 - 354\ M\^82 + 
                    297\ M\^84 - 195\ M\^86 + 70\ M\^88 - 13\ M\^90 + 
                    M\^92)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 35]\), \(M\^64 + L\^24\ M\^64 + 
              L\^23\ \((\(-3\)\ M\^52 + 11\ M\^54 - 18\ M\^56 + 14\ M\^58 + 
                    M\^60 - 28\ M\^62 + 59\ M\^64 + 6\ M\^66 - 34\ M\^68 + 
                    21\ M\^70 - 5\ M\^72)\) + 
              L\ \((\(-5\)\ M\^56 + 21\ M\^58 - 34\ M\^60 + 6\ M\^62 + 
                    59\ M\^64 - 28\ M\^66 + M\^68 + 14\ M\^70 - 18\ M\^72 + 
                    11\ M\^74 - 3\ M\^76)\) + 
              L\^22\ \((3\ M\^40 - 22\ M\^42 + 73\ M\^44 - 148\ M\^46 + 
                    164\ M\^48 + 103\ M\^50 - 572\ M\^52 + 284\ M\^54 + 
                    763\ M\^56 - 530\ M\^58 - 965\ M\^60 + 540\ M\^62 + 
                    1480\ M\^64 - 689\ M\^66 - 916\ M\^68 + 968\ M\^70 - 
                    52\ M\^72 - 422\ M\^74 + 288\ M\^76 - 84\ M\^78 + 
                    10\ M\^80)\) + 
              L\^21\ \((\(-M\^28\) + 11\ M\^30 - 57\ M\^32 + 183\ M\^34 - 
                    378\ M\^36 + 429\ M\^38 - 143\ M\^40 - 189\ M\^42 + 
                    762\ M\^44 - 2213\ M\^46 + 1392\ M\^48 + 5077\ M\^50 - 
                    7271\ M\^52 - 4963\ M\^54 + 11480\ M\^56 + 4312\ M\^58 - 
                    13330\ M\^60 - 3558\ M\^62 + 14863\ M\^64 + 634\ M\^66 - 
                    8970\ M\^68 + 5926\ M\^70 - 680\ M\^72 - 4759\ M\^74 + 
                    5090\ M\^76 - 770\ M\^78 - 2081\ M\^80 + 1778\ M\^82 - 
                    666\ M\^84 + 126\ M\^86 - 10\ M\^88)\) + 
              L\^2\ \((10\ M\^48 - 84\ M\^50 + 288\ M\^52 - 422\ M\^54 - 
                    52\ M\^56 + 968\ M\^58 - 916\ M\^60 - 689\ M\^62 + 
                    1480\ M\^64 + 540\ M\^66 - 965\ M\^68 - 530\ M\^70 + 
                    763\ M\^72 + 284\ M\^74 - 572\ M\^76 + 103\ M\^78 + 
                    164\ M\^80 - 148\ M\^82 + 73\ M\^84 - 22\ M\^86 + 
                    3\ M\^88)\) + 
              L\^20\ \((2\ M\^20 - 24\ M\^22 + 134\ M\^24 - 475\ M\^26 + 
                    1158\ M\^28 - 1727\ M\^30 + 605\ M\^32 + 3079\ M\^34 - 
                    5524\ M\^36 + 540\ M\^38 + 10182\ M\^40 - 15490\ M\^42 + 
                    2912\ M\^44 + 31069\ M\^46 - 47458\ M\^48 - 
                    19803\ M\^50 + 116356\ M\^52 - 54836\ M\^54 - 
                    150632\ M\^56 + 162866\ M\^58 + 93339\ M\^60 - 
                    206303\ M\^62 + 23826\ M\^64 + 135903\ M\^66 - 
                    71459\ M\^68 - 5061\ M\^70 + 31369\ M\^72 - 
                    49859\ M\^74 + 24741\ M\^76 + 17199\ M\^78 - 
                    27058\ M\^80 + 12847\ M\^82 + 1769\ M\^84 - 7397\ M\^86 + 
                    5749\ M\^88 - 2444\ M\^90 + 610\ M\^92 - 84\ M\^94 + 
                    5\ M\^96)\) + 
              L\^3\ \((\(-10\)\ M\^40 + 126\ M\^42 - 666\ M\^44 + 
                    1778\ M\^46 - 2081\ M\^48 - 770\ M\^50 + 5090\ M\^52 - 
                    4759\ M\^54 - 680\ M\^56 + 5926\ M\^58 - 8970\ M\^60 + 
                    634\ M\^62 + 14863\ M\^64 - 3558\ M\^66 - 13330\ M\^68 + 
                    4312\ M\^70 + 11480\ M\^72 - 4963\ M\^74 - 7271\ M\^76 + 
                    5077\ M\^78 + 1392\ M\^80 - 2213\ M\^82 + 762\ M\^84 - 
                    189\ M\^86 - 143\ M\^88 + 429\ M\^90 - 378\ M\^92 + 
                    183\ M\^94 - 57\ M\^96 + 11\ M\^98 - M\^100)\) + 
              L\^19\ \((\(-M\^12\) + 13\ M\^14 - 84\ M\^16 + 378\ M\^18 - 
                    1281\ M\^20 + 3072\ M\^22 - 4340\ M\^24 + 948\ M\^26 + 
                    8104\ M\^28 - 11115\ M\^30 - 4640\ M\^32 + 24221\ M\^34 - 
                    19108\ M\^36 - 12849\ M\^38 + 57914\ M\^40 - 
                    55049\ M\^42 - 124292\ M\^44 + 304513\ M\^46 + 
                    95379\ M\^48 - 788339\ M\^50 + 294530\ M\^52 + 
                    1210406\ M\^54 - 1095982\ M\^56 - 1051438\ M\^58 + 
                    1704365\ M\^60 + 323498\ M\^62 - 1468358\ M\^64 + 
                    305134\ M\^66 + 705473\ M\^68 - 260326\ M\^70 - 
                    137727\ M\^72 - 29582\ M\^74 + 62884\ M\^76 + 
                    70987\ M\^78 - 123864\ M\^80 + 58648\ M\^82 + 
                    48291\ M\^84 - 80880\ M\^86 + 31459\ M\^88 + 
                    15495\ M\^90 - 23637\ M\^92 + 13536\ M\^94 - 
                    4880\ M\^96 + 1212\ M\^98 - 204\ M\^100 + 21\ M\^102 - 
                    M\^104)\) + 
              L\^4\ \((5\ M\^32 - 84\ M\^34 + 610\ M\^36 - 2444\ M\^38 + 
                    5749\ M\^40 - 7397\ M\^42 + 1769\ M\^44 + 12847\ M\^46 - 
                    27058\ M\^48 + 17199\ M\^50 + 24741\ M\^52 - 
                    49859\ M\^54 + 31369\ M\^56 - 5061\ M\^58 - 
                    71459\ M\^60 + 135903\ M\^62 + 23826\ M\^64 - 
                    206303\ M\^66 + 93339\ M\^68 + 162866\ M\^70 - 
                    150632\ M\^72 - 54836\ M\^74 + 116356\ M\^76 - 
                    19803\ M\^78 - 47458\ M\^80 + 31069\ M\^82 + 
                    2912\ M\^84 - 15490\ M\^86 + 10182\ M\^88 + 540\ M\^90 - 
                    5524\ M\^92 + 3079\ M\^94 + 605\ M\^96 - 1727\ M\^98 + 
                    1158\ M\^100 - 475\ M\^102 + 134\ M\^104 - 24\ M\^106 + 
                    2\ M\^108)\) + 
              L\^18\ \((3\ M\^8 - 49\ M\^10 + 365\ M\^12 - 1620\ M\^14 + 
                    4615\ M\^16 - 8114\ M\^18 + 6124\ M\^20 + 6771\ M\^22 - 
                    16400\ M\^24 - 8701\ M\^26 + 43879\ M\^28 + 
                    15770\ M\^30 - 137578\ M\^32 + 39909\ M\^34 + 
                    340611\ M\^36 - 494392\ M\^38 - 146501\ M\^40 + 
                    1330983\ M\^42 - 1327499\ M\^44 - 1549329\ M\^46 + 
                    4202729\ M\^48 - 345564\ M\^50 - 6680722\ M\^52 + 
                    4206181\ M\^54 + 6484736\ M\^56 - 7591117\ M\^58 - 
                    3704513\ M\^60 + 7917880\ M\^62 + 1074976\ M\^64 - 
                    5393662\ M\^66 - 296431\ M\^68 + 2964925\ M\^70 + 
                    745359\ M\^72 - 1871028\ M\^74 - 588952\ M\^76 + 
                    1266784\ M\^78 - 151600\ M\^80 - 459482\ M\^82 + 
                    406847\ M\^84 - 151840\ M\^86 - 91455\ M\^88 + 
                    170582\ M\^90 - 101286\ M\^92 + 3769\ M\^94 + 
                    45064\ M\^96 - 42067\ M\^98 + 20999\ M\^100 - 
                    6459\ M\^102 + 1220\ M\^104 - 130\ M\^106 + 
                    6\ M\^108)\) + 
              L\^17\ \((\(-3\)\ M\^4 + 57\ M\^6 - 482\ M\^8 + 2378\ M\^10 - 
                    7504\ M\^12 + 15295\ M\^14 - 17718\ M\^16 + 2045\ M\^18 + 
                    26490\ M\^20 - 28893\ M\^22 - 5493\ M\^24 + 1168\ M\^26 + 
                    8665\ M\^28 + 217283\ M\^30 - 466762\ M\^32 - 
                    174562\ M\^34 + 1543713\ M\^36 - 1065893\ M\^38 - 
                    2763865\ M\^40 + 4610346\ M\^42 + 2723155\ M\^44 - 
                    10476377\ M\^46 - 502699\ M\^48 + 17493170\ M\^50 - 
                    3274819\ M\^52 - 23841913\ M\^54 + 6919330\ M\^56 + 
                    28214631\ M\^58 - 10420332\ M\^60 - 29905119\ M\^62 + 
                    14896044\ M\^64 + 28123469\ M\^66 - 18934818\ M\^68 - 
                    20853023\ M\^70 + 20109810\ M\^72 + 10357246\ M\^74 - 
                    15983516\ M\^76 - 1590191\ M\^78 + 8407400\ M\^80 - 
                    1902965\ M\^82 - 2302739\ M\^84 + 1288144\ M\^86 + 
                    32002\ M\^88 - 55776\ M\^90 - 105543\ M\^92 - 
                    86204\ M\^94 + 234773\ M\^96 - 111825\ M\^98 - 
                    57619\ M\^100 + 96135\ M\^102 - 54241\ M\^104 + 
                    17143\ M\^106 - 3214\ M\^108 + 335\ M\^110 - 
                    15\ M\^112)\) + 
              L\^5\ \((\(-M\^24\) + 21\ M\^26 - 204\ M\^28 + 1212\ M\^30 - 
                    4880\ M\^32 + 13536\ M\^34 - 23637\ M\^36 + 
                    15495\ M\^38 + 31459\ M\^40 - 80880\ M\^42 + 
                    48291\ M\^44 + 58648\ M\^46 - 123864\ M\^48 + 
                    70987\ M\^50 + 62884\ M\^52 - 29582\ M\^54 - 
                    137727\ M\^56 - 260326\ M\^58 + 705473\ M\^60 + 
                    305134\ M\^62 - 1468358\ M\^64 + 323498\ M\^66 + 
                    1704365\ M\^68 - 1051438\ M\^70 - 1095982\ M\^72 + 
                    1210406\ M\^74 + 294530\ M\^76 - 788339\ M\^78 + 
                    95379\ M\^80 + 304513\ M\^82 - 124292\ M\^84 - 
                    55049\ M\^86 + 57914\ M\^88 - 12849\ M\^90 - 
                    19108\ M\^92 + 24221\ M\^94 - 4640\ M\^96 - 
                    11115\ M\^98 + 8104\ M\^100 + 948\ M\^102 - 
                    4340\ M\^104 + 3072\ M\^106 - 1281\ M\^108 + 
                    378\ M\^110 - 84\ M\^112 + 13\ M\^114 - M\^116)\) + 
              L\^16\ \((1 - 21\ M\^2 + 203\ M\^4 - 1219\ M\^6 + 5163\ M\^8 - 
                    16162\ M\^10 + 36047\ M\^12 - 47428\ M\^14 + 
                    1149\ M\^16 + 123959\ M\^18 - 189156\ M\^20 - 
                    65837\ M\^22 + 591671\ M\^24 - 612025\ M\^26 - 
                    672526\ M\^28 + 2014736\ M\^30 - 195993\ M\^32 - 
                    4104129\ M\^34 + 3033998\ M\^36 + 6208315\ M\^38 - 
                    8834818\ M\^40 - 7679974\ M\^42 + 19979099\ M\^44 + 
                    6611649\ M\^46 - 39703568\ M\^48 + 336681\ M\^50 + 
                    69073556\ M\^52 - 16440369\ M\^54 - 100664562\ M\^56 + 
                    40473581\ M\^58 + 118520312\ M\^60 - 66196587\ M\^62 - 
                    109934011\ M\^64 + 83641294\ M\^66 + 79399806\ M\^68 - 
                    81244328\ M\^70 - 41396705\ M\^72 + 60243085\ M\^74 + 
                    13101174\ M\^76 - 32647607\ M\^78 - 1112505\ M\^80 + 
                    11723416\ M\^82 + 94006\ M\^84 - 2709246\ M\^86 - 
                    1245877\ M\^88 + 1202888\ M\^90 + 747028\ M\^92 - 
                    1172313\ M\^94 + 411384\ M\^96 + 367236\ M\^98 - 
                    520353\ M\^100 + 179236\ M\^102 + 121576\ M\^104 - 
                    162985\ M\^106 + 84786\ M\^108 - 25386\ M\^110 + 
                    4566\ M\^112 - 460\ M\^114 + 20\ M\^116)\) + 
              L\^15\ \((\(-5\) + 112\ M\^2 - 1115\ M\^4 + 6450\ M\^6 - 
                    23666\ M\^8 + 55457\ M\^10 - 73735\ M\^12 + 
                    13152\ M\^14 + 146512\ M\^16 - 243228\ M\^18 - 
                    12492\ M\^20 + 572830\ M\^22 - 702437\ M\^24 - 
                    95463\ M\^26 + 934267\ M\^28 - 1277583\ M\^30 + 
                    1639278\ M\^32 + 1652485\ M\^34 - 10266966\ M\^36 + 
                    5098219\ M\^38 + 24500387\ M\^40 - 25944559\ M\^42 - 
                    37811944\ M\^44 + 64353035\ M\^46 + 37729333\ M\^48 - 
                    116444581\ M\^50 - 14845732\ M\^52 + 173492598\ M\^54 - 
                    27030685\ M\^56 - 222692859\ M\^58 + 69521945\ M\^60 + 
                    247526564\ M\^62 - 97574379\ M\^64 - 236867202\ M\^66 + 
                    112069721\ M\^68 + 200936604\ M\^70 - 113015296\ M\^72 - 
                    152966607\ M\^74 + 102547054\ M\^76 + 104251335\ M\^78 - 
                    84054352\ M\^80 - 61289883\ M\^82 + 61397036\ M\^84 + 
                    28018501\ M\^86 - 38978961\ M\^88 - 6725496\ M\^90 + 
                    19961216\ M\^92 - 2496727\ M\^94 - 7053526\ M\^96 + 
                    3545623\ M\^98 + 714002\ M\^100 - 1424469\ M\^102 + 
                    616739\ M\^104 + 45299\ M\^106 - 248929\ M\^108 + 
                    185944\ M\^110 - 79591\ M\^112 + 21598\ M\^114 - 
                    3664\ M\^116 + 355\ M\^118 - 15\ M\^120)\) + 
              L\^6\ \((6\ M\^20 - 130\ M\^22 + 1220\ M\^24 - 6459\ M\^26 + 
                    20999\ M\^28 - 42067\ M\^30 + 45064\ M\^32 + 
                    3769\ M\^34 - 101286\ M\^36 + 170582\ M\^38 - 
                    91455\ M\^40 - 151840\ M\^42 + 406847\ M\^44 - 
                    459482\ M\^46 - 151600\ M\^48 + 1266784\ M\^50 - 
                    588952\ M\^52 - 1871028\ M\^54 + 745359\ M\^56 + 
                    2964925\ M\^58 - 296431\ M\^60 - 5393662\ M\^62 + 
                    1074976\ M\^64 + 7917880\ M\^66 - 3704513\ M\^68 - 
                    7591117\ M\^70 + 6484736\ M\^72 + 4206181\ M\^74 - 
                    6680722\ M\^76 - 345564\ M\^78 + 4202729\ M\^80 - 
                    1549329\ M\^82 - 1327499\ M\^84 + 1330983\ M\^86 - 
                    146501\ M\^88 - 494392\ M\^90 + 340611\ M\^92 + 
                    39909\ M\^94 - 137578\ M\^96 + 15770\ M\^98 + 
                    43879\ M\^100 - 8701\ M\^102 - 16400\ M\^104 + 
                    6771\ M\^106 + 6124\ M\^108 - 8114\ M\^110 + 
                    4615\ M\^112 - 1620\ M\^114 + 365\ M\^116 - 49\ M\^118 + 
                    3\ M\^120)\) + 
              L\^7\ \((\(-15\)\ M\^16 + 335\ M\^18 - 3214\ M\^20 + 
                    17143\ M\^22 - 54241\ M\^24 + 96135\ M\^26 - 
                    57619\ M\^28 - 111825\ M\^30 + 234773\ M\^32 - 
                    86204\ M\^34 - 105543\ M\^36 - 55776\ M\^38 + 
                    32002\ M\^40 + 1288144\ M\^42 - 2302739\ M\^44 - 
                    1902965\ M\^46 + 8407400\ M\^48 - 1590191\ M\^50 - 
                    15983516\ M\^52 + 10357246\ M\^54 + 20109810\ M\^56 - 
                    20853023\ M\^58 - 18934818\ M\^60 + 28123469\ M\^62 + 
                    14896044\ M\^64 - 29905119\ M\^66 - 10420332\ M\^68 + 
                    28214631\ M\^70 + 6919330\ M\^72 - 23841913\ M\^74 - 
                    3274819\ M\^76 + 17493170\ M\^78 - 502699\ M\^80 - 
                    10476377\ M\^82 + 2723155\ M\^84 + 4610346\ M\^86 - 
                    2763865\ M\^88 - 1065893\ M\^90 + 1543713\ M\^92 - 
                    174562\ M\^94 - 466762\ M\^96 + 217283\ M\^98 + 
                    8665\ M\^100 + 1168\ M\^102 - 5493\ M\^104 - 
                    28893\ M\^106 + 26490\ M\^108 + 2045\ M\^110 - 
                    17718\ M\^112 + 15295\ M\^114 - 7504\ M\^116 + 
                    2378\ M\^118 - 482\ M\^120 + 57\ M\^122 - 3\ M\^124)\) + 
              L\^14\ \((10 - 235\ M\^2 + 2448\ M\^4 - 14666\ M\^6 + 
                    54219\ M\^8 - 118827\ M\^10 + 107808\ M\^12 + 
                    146602\ M\^14 - 472759\ M\^16 + 74091\ M\^18 + 
                    1228184\ M\^20 - 1199268\ M\^22 - 2125001\ M\^24 + 
                    4027402\ M\^26 + 3085574\ M\^28 - 11904737\ M\^30 - 
                    156311\ M\^32 + 28851881\ M\^34 - 17132579\ M\^36 - 
                    55855220\ M\^38 + 66184609\ M\^40 + 82335022\ M\^42 - 
                    158806199\ M\^44 - 87087639\ M\^46 + 289923585\ M\^48 + 
                    47365083\ M\^50 - 437870612\ M\^52 + 42256256\ M\^54 + 
                    578652941\ M\^56 - 156990551\ M\^58 - 687062195\ M\^60 + 
                    249024952\ M\^62 + 725698355\ M\^64 - 287739337\ M\^66 - 
                    658946437\ M\^68 + 289578447\ M\^70 + 505511329\ M\^72 - 
                    273489000\ M\^74 - 319820415\ M\^76 + 242778394\ M\^78 + 
                    156615045\ M\^80 - 193980217\ M\^82 - 45082719\ M\^84 + 
                    131669683\ M\^86 - 11817544\ M\^88 - 71441693\ M\^90 + 
                    27750997\ M\^92 + 27472745\ M\^94 - 21500940\ M\^96 - 
                    4792671\ M\^98 + 10151812\ M\^100 - 2141333\ M\^102 - 
                    2366689\ M\^104 + 1659997\ M\^106 - 170352\ M\^108 - 
                    343933\ M\^110 + 280259\ M\^112 - 130959\ M\^114 + 
                    43247\ M\^116 - 10123\ M\^118 + 1580\ M\^120 - 
                    146\ M\^122 + 6\ M\^124)\) + 
              L\^11\ \((\(-1\) + 25\ M\^2 - 294\ M\^4 + 2223\ M\^6 - 
                    12294\ M\^8 + 51662\ M\^10 - 159542\ M\^12 + 
                    323230\ M\^14 - 273922\ M\^16 - 509424\ M\^18 + 
                    1786594\ M\^20 - 1249953\ M\^22 - 3332731\ M\^24 + 
                    7961068\ M\^26 - 1969974\ M\^28 - 17456628\ M\^30 + 
                    25550322\ M\^32 + 17523786\ M\^34 - 81090674\ M\^36 + 
                    17494649\ M\^38 + 178016738\ M\^40 - 114825555\ M\^42 - 
                    324643922\ M\^44 + 284092288\ M\^46 + 530242525\ M\^48 - 
                    492399331\ M\^50 - 795911937\ M\^52 + 659313784\ M\^54 + 
                    1079613563\ M\^56 - 691487294\ M\^58 - 
                    1278605854\ M\^60 + 556760797\ M\^62 + 
                    1278086334\ M\^64 - 329611413\ M\^66 - 
                    1052649490\ M\^68 + 141226612\ M\^70 + 715435249\ M\^72 - 
                    46458509\ M\^74 - 403699615\ M\^76 + 17738229\ M\^78 + 
                    187068599\ M\^80 - 9662295\ M\^82 - 66710730\ M\^84 + 
                    2156225\ M\^86 + 13194778\ M\^88 + 4944366\ M\^90 + 
                    3373819\ M\^92 - 7997652\ M\^94 - 4725667\ M\^96 + 
                    7609018\ M\^98 + 1777090\ M\^100 - 5074854\ M\^102 + 
                    385801\ M\^104 + 2508898\ M\^106 - 1150549\ M\^108 - 
                    556747\ M\^110 + 661686\ M\^112 - 95298\ M\^114 - 
                    188333\ M\^116 + 157498\ M\^118 - 64930\ M\^120 + 
                    16496\ M\^122 - 2629\ M\^124 + 243\ M\^126 - 
                    10\ M\^128)\) + 
              L\^9\ \((\(-15\)\ M\^8 + 355\ M\^10 - 3664\ M\^12 + 
                    21598\ M\^14 - 79591\ M\^16 + 185944\ M\^18 - 
                    248929\ M\^20 + 45299\ M\^22 + 616739\ M\^24 - 
                    1424469\ M\^26 + 714002\ M\^28 + 3545623\ M\^30 - 
                    7053526\ M\^32 - 2496727\ M\^34 + 19961216\ M\^36 - 
                    6725496\ M\^38 - 38978961\ M\^40 + 28018501\ M\^42 + 
                    61397036\ M\^44 - 61289883\ M\^46 - 84054352\ M\^48 + 
                    104251335\ M\^50 + 102547054\ M\^52 - 152966607\ M\^54 - 
                    113015296\ M\^56 + 200936604\ M\^58 + 112069721\ M\^60 - 
                    236867202\ M\^62 - 97574379\ M\^64 + 247526564\ M\^66 + 
                    69521945\ M\^68 - 222692859\ M\^70 - 27030685\ M\^72 + 
                    173492598\ M\^74 - 14845732\ M\^76 - 116444581\ M\^78 + 
                    37729333\ M\^80 + 64353035\ M\^82 - 37811944\ M\^84 - 
                    25944559\ M\^86 + 24500387\ M\^88 + 5098219\ M\^90 - 
                    10266966\ M\^92 + 1652485\ M\^94 + 1639278\ M\^96 - 
                    1277583\ M\^98 + 934267\ M\^100 - 95463\ M\^102 - 
                    702437\ M\^104 + 572830\ M\^106 - 12492\ M\^108 - 
                    243228\ M\^110 + 146512\ M\^112 + 13152\ M\^114 - 
                    73735\ M\^116 + 55457\ M\^118 - 23666\ M\^120 + 
                    6450\ M\^122 - 1115\ M\^124 + 112\ M\^126 - 
                    5\ M\^128)\) + 
              L\^13\ \((\(-10\) + 243\ M\^2 - 2629\ M\^4 + 16496\ M\^6 - 
                    64930\ M\^8 + 157498\ M\^10 - 188333\ M\^12 - 
                    95298\ M\^14 + 661686\ M\^16 - 556747\ M\^18 - 
                    1150549\ M\^20 + 2508898\ M\^22 + 385801\ M\^24 - 
                    5074854\ M\^26 + 1777090\ M\^28 + 7609018\ M\^30 - 
                    4725667\ M\^32 - 7997652\ M\^34 + 3373819\ M\^36 + 
                    4944366\ M\^38 + 13194778\ M\^40 + 2156225\ M\^42 - 
                    66710730\ M\^44 - 9662295\ M\^46 + 187068599\ M\^48 + 
                    17738229\ M\^50 - 403699615\ M\^52 - 46458509\ M\^54 + 
                    715435249\ M\^56 + 141226612\ M\^58 - 1052649490\ M\^60 - 
                    329611413\ M\^62 + 1278086334\ M\^64 + 556760797\ M\^66 - 
                    1278605854\ M\^68 - 691487294\ M\^70 + 
                    1079613563\ M\^72 + 659313784\ M\^74 - 795911937\ M\^76 - 
                    492399331\ M\^78 + 530242525\ M\^80 + 284092288\ M\^82 - 
                    324643922\ M\^84 - 114825555\ M\^86 + 178016738\ M\^88 + 
                    17494649\ M\^90 - 81090674\ M\^92 + 17523786\ M\^94 + 
                    25550322\ M\^96 - 17456628\ M\^98 - 1969974\ M\^100 + 
                    7961068\ M\^102 - 3332731\ M\^104 - 1249953\ M\^106 + 
                    1786594\ M\^108 - 509424\ M\^110 - 273922\ M\^112 + 
                    323230\ M\^114 - 159542\ M\^116 + 51662\ M\^118 - 
                    12294\ M\^120 + 2223\ M\^122 - 294\ M\^124 + 25\ M\^126 - 
                    M\^128)\) + 
              L\^8\ \((20\ M\^12 - 460\ M\^14 + 4566\ M\^16 - 25386\ M\^18 + 
                    84786\ M\^20 - 162985\ M\^22 + 121576\ M\^24 + 
                    179236\ M\^26 - 520353\ M\^28 + 367236\ M\^30 + 
                    411384\ M\^32 - 1172313\ M\^34 + 747028\ M\^36 + 
                    1202888\ M\^38 - 1245877\ M\^40 - 2709246\ M\^42 + 
                    94006\ M\^44 + 11723416\ M\^46 - 1112505\ M\^48 - 
                    32647607\ M\^50 + 13101174\ M\^52 + 60243085\ M\^54 - 
                    41396705\ M\^56 - 81244328\ M\^58 + 79399806\ M\^60 + 
                    83641294\ M\^62 - 109934011\ M\^64 - 66196587\ M\^66 + 
                    118520312\ M\^68 + 40473581\ M\^70 - 100664562\ M\^72 - 
                    16440369\ M\^74 + 69073556\ M\^76 + 336681\ M\^78 - 
                    39703568\ M\^80 + 6611649\ M\^82 + 19979099\ M\^84 - 
                    7679974\ M\^86 - 8834818\ M\^88 + 6208315\ M\^90 + 
                    3033998\ M\^92 - 4104129\ M\^94 - 195993\ M\^96 + 
                    2014736\ M\^98 - 672526\ M\^100 - 612025\ M\^102 + 
                    591671\ M\^104 - 65837\ M\^106 - 189156\ M\^108 + 
                    123959\ M\^110 + 1149\ M\^112 - 47428\ M\^114 + 
                    36047\ M\^116 - 16162\ M\^118 + 5163\ M\^120 - 
                    1219\ M\^122 + 203\ M\^124 - 21\ M\^126 + M\^128)\) + 
              L\^12\ \((5 - 124\ M\^2 + 1384\ M\^4 - 9191\ M\^6 + 
                    40354\ M\^8 - 121829\ M\^10 + 245438\ M\^12 - 
                    251634\ M\^14 - 239506\ M\^16 + 1363606\ M\^18 - 
                    1589906\ M\^20 - 2384872\ M\^22 + 8753897\ M\^24 - 
                    2481014\ M\^26 - 23533106\ M\^28 + 24428648\ M\^30 + 
                    46385560\ M\^32 - 81672790\ M\^34 - 74217308\ M\^36 + 
                    193428724\ M\^38 + 108709363\ M\^40 - 372432716\ M\^42 - 
                    165829651\ M\^44 + 605386736\ M\^46 + 284212174\ M\^48 - 
                    830800536\ M\^50 - 504896861\ M\^52 + 926857535\ M\^54 + 
                    819946608\ M\^56 - 760039417\ M\^58 - 1124889046\ M\^60 + 
                    298728874\ M\^62 + 1256505358\ M\^64 + 298728874\ M\^66 - 
                    1124889046\ M\^68 - 760039417\ M\^70 + 819946608\ M\^72 + 
                    926857535\ M\^74 - 504896861\ M\^76 - 830800536\ M\^78 + 
                    284212174\ M\^80 + 605386736\ M\^82 - 165829651\ M\^84 - 
                    372432716\ M\^86 + 108709363\ M\^88 + 193428724\ M\^90 - 
                    74217308\ M\^92 - 81672790\ M\^94 + 46385560\ M\^96 + 
                    24428648\ M\^98 - 23533106\ M\^100 - 2481014\ M\^102 + 
                    8753897\ M\^104 - 2384872\ M\^106 - 1589906\ M\^108 + 
                    1363606\ M\^110 - 239506\ M\^112 - 251634\ M\^114 + 
                    245438\ M\^116 - 121829\ M\^118 + 40354\ M\^120 - 
                    9191\ M\^122 + 1384\ M\^124 - 124\ M\^126 + 
                    5\ M\^128)\) + 
              L\^10\ \((6\ M\^4 - 146\ M\^6 + 1580\ M\^8 - 10123\ M\^10 + 
                    43247\ M\^12 - 130959\ M\^14 + 280259\ M\^16 - 
                    343933\ M\^18 - 170352\ M\^20 + 1659997\ M\^22 - 
                    2366689\ M\^24 - 2141333\ M\^26 + 10151812\ M\^28 - 
                    4792671\ M\^30 - 21500940\ M\^32 + 27472745\ M\^34 + 
                    27750997\ M\^36 - 71441693\ M\^38 - 11817544\ M\^40 + 
                    131669683\ M\^42 - 45082719\ M\^44 - 193980217\ M\^46 + 
                    156615045\ M\^48 + 242778394\ M\^50 - 319820415\ M\^52 - 
                    273489000\ M\^54 + 505511329\ M\^56 + 289578447\ M\^58 - 
                    658946437\ M\^60 - 287739337\ M\^62 + 725698355\ M\^64 + 
                    249024952\ M\^66 - 687062195\ M\^68 - 156990551\ M\^70 + 
                    578652941\ M\^72 + 42256256\ M\^74 - 437870612\ M\^76 + 
                    47365083\ M\^78 + 289923585\ M\^80 - 87087639\ M\^82 - 
                    158806199\ M\^84 + 82335022\ M\^86 + 66184609\ M\^88 - 
                    55855220\ M\^90 - 17132579\ M\^92 + 28851881\ M\^94 - 
                    156311\ M\^96 - 11904737\ M\^98 + 3085574\ M\^100 + 
                    4027402\ M\^102 - 2125001\ M\^104 - 1199268\ M\^106 + 
                    1228184\ M\^108 + 74091\ M\^110 - 472759\ M\^112 + 
                    146602\ M\^114 + 107808\ M\^116 - 118827\ M\^118 + 
                    54219\ M\^120 - 14666\ M\^122 + 2448\ M\^124 - 
                    235\ M\^126 + 10\ M\^128)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 36]\), \(L\^25\ M\^20 + 
              M\^122 + 
              L\^24\ \((\(-5\)\ M\^16 + 22\ M\^18 - 36\ M\^20 + 7\ M\^22 + 
                    61\ M\^24 - 29\ M\^26 + 2\ M\^28 + 11\ M\^30 - 
                    18\ M\^32 + 14\ M\^34 - 4\ M\^36)\) + 
              L\^23\ \((10\ M\^12 - 84\ M\^14 + 285\ M\^16 - 430\ M\^18 + 
                    19\ M\^20 + 861\ M\^22 - 915\ M\^24 - 543\ M\^26 + 
                    1312\ M\^28 + 744\ M\^30 - 994\ M\^32 - 923\ M\^34 + 
                    1132\ M\^36 + 379\ M\^38 - 829\ M\^40 + 226\ M\^42 + 
                    159\ M\^44 - 190\ M\^46 + 115\ M\^48 - 40\ M\^50 + 
                    6\ M\^52)\) + 
              L\^22\ \((\(-10\)\ M\^8 + 120\ M\^10 - 606\ M\^12 + 
                    1617\ M\^14 - 2259\ M\^16 + 904\ M\^18 + 2620\ M\^20 - 
                    6986\ M\^22 + 8560\ M\^24 + 3099\ M\^26 - 24728\ M\^28 + 
                    14090\ M\^30 + 31147\ M\^32 - 26173\ M\^34 - 
                    21455\ M\^36 + 25981\ M\^38 + 8530\ M\^40 - 
                    14933\ M\^42 - 49\ M\^44 + 3695\ M\^46 - 2090\ M\^48 + 
                    2122\ M\^50 + 28\ M\^52 - 1977\ M\^54 + 1126\ M\^56 + 
                    289\ M\^58 - 648\ M\^60 + 414\ M\^62 - 162\ M\^64 + 
                    38\ M\^66 - 4\ M\^68)\) + 
              L\^21\ \((5\ M\^4 - 76\ M\^6 + 510\ M\^8 - 2025\ M\^10 + 
                    5451\ M\^12 - 10305\ M\^14 + 10373\ M\^16 + 9259\ M\^18 - 
                    45852\ M\^20 + 40593\ M\^22 + 38701\ M\^24 - 
                    81986\ M\^26 + 34927\ M\^28 - 6841\ M\^30 - 
                    81489\ M\^32 + 208163\ M\^34 + 31531\ M\^36 - 
                    337357\ M\^38 + 63106\ M\^40 + 319841\ M\^42 - 
                    77348\ M\^44 - 217990\ M\^46 + 42401\ M\^48 + 
                    123753\ M\^50 - 22962\ M\^52 - 55286\ M\^54 + 
                    18368\ M\^56 + 9998\ M\^58 - 6403\ M\^60 + 2767\ M\^62 - 
                    662\ M\^64 - 1287\ M\^66 + 881\ M\^68 - 255\ M\^70 + 
                    536\ M\^72 - 730\ M\^74 + 509\ M\^76 - 224\ M\^78 + 
                    66\ M\^80 - 12\ M\^82 + M\^84)\) + 
              L\^20\ \((\(-1\) + 18\ M\^2 - 159\ M\^4 + 956\ M\^6 - 
                    4308\ M\^8 + 14029\ M\^10 - 29650\ M\^12 + 31692\ M\^14 + 
                    9334\ M\^16 - 84746\ M\^18 + 144214\ M\^20 - 
                    94540\ M\^22 - 233660\ M\^24 + 583042\ M\^26 - 
                    47757\ M\^28 - 908893\ M\^30 + 386010\ M\^32 + 
                    595820\ M\^34 - 20997\ M\^36 + 84828\ M\^38 - 
                    1094465\ M\^40 - 399610\ M\^42 + 2053483\ M\^44 + 
                    332897\ M\^46 - 1972899\ M\^48 - 248985\ M\^50 + 
                    1220798\ M\^52 + 381416\ M\^54 - 612773\ M\^56 - 
                    482324\ M\^58 + 393933\ M\^60 + 342880\ M\^62 - 
                    313962\ M\^64 - 97294\ M\^66 + 178758\ M\^68 - 
                    27152\ M\^70 - 58997\ M\^72 + 37606\ M\^74 + 
                    3211\ M\^76 - 13948\ M\^78 + 4489\ M\^80 + 3323\ M\^82 - 
                    3972\ M\^84 + 1989\ M\^86 - 614\ M\^88 + 125\ M\^90 - 
                    16\ M\^92 + M\^94)\) + 
              L\^19\ \((6 - 121\ M\^2 + 1087\ M\^4 - 5771\ M\^6 + 
                    20415\ M\^8 - 51104\ M\^10 + 87131\ M\^12 - 
                    62083\ M\^14 - 139777\ M\^16 + 426994\ M\^18 - 
                    260530\ M\^20 - 545700\ M\^22 + 954872\ M\^24 - 
                    350292\ M\^26 - 653717\ M\^28 + 2043571\ M\^30 - 
                    1354960\ M\^32 - 3822091\ M\^34 + 3717079\ M\^36 + 
                    5697781\ M\^38 - 4220813\ M\^40 - 8036536\ M\^42 + 
                    2303803\ M\^44 + 10155809\ M\^46 + 972622\ M\^48 - 
                    10384433\ M\^50 - 3425933\ M\^52 + 8844452\ M\^54 + 
                    4192778\ M\^56 - 6499331\ M\^58 - 3364795\ M\^60 + 
                    4129973\ M\^62 + 1845217\ M\^64 - 2172900\ M\^66 - 
                    658564\ M\^68 + 847549\ M\^70 + 200363\ M\^72 - 
                    250373\ M\^74 - 102557\ M\^76 + 87987\ M\^78 + 
                    68923\ M\^80 - 74035\ M\^82 - 6493\ M\^84 + 
                    37138\ M\^86 - 13693\ M\^88 - 9427\ M\^90 + 
                    12462\ M\^92 - 6621\ M\^94 + 2121\ M\^96 - 433\ M\^98 + 
                    53\ M\^100 - 3\ M\^102)\) + 
              L\^18\ \((\(-15\) + 332\ M\^2 - 3239\ M\^4 + 18081\ M\^6 - 
                    62420\ M\^8 + 132129\ M\^10 - 145567\ M\^12 - 
                    21840\ M\^14 + 371539\ M\^16 - 675540\ M\^18 + 
                    342861\ M\^20 + 1290900\ M\^22 - 2650924\ M\^24 - 
                    67684\ M\^26 + 4667232\ M\^28 - 3085173\ M\^30 - 
                    2903735\ M\^32 + 6617722\ M\^34 - 6070005\ M\^36 - 
                    9727843\ M\^38 + 23512827\ M\^40 + 12951767\ M\^42 - 
                    45810329\ M\^44 - 15482108\ M\^46 + 62828004\ M\^48 + 
                    15812822\ M\^50 - 63494563\ M\^52 - 13550758\ M\^54 + 
                    47854847\ M\^56 + 12780477\ M\^58 - 25558638\ M\^60 - 
                    14569512\ M\^62 + 8711949\ M\^64 + 15996699\ M\^66 - 
                    2418005\ M\^68 - 13959646\ M\^70 + 2837488\ M\^72 + 
                    8762908\ M\^74 - 3930935\ M\^76 - 3275229\ M\^78 + 
                    3001125\ M\^80 + 177142\ M\^82 - 1234275\ M\^84 + 
                    523123\ M\^86 + 112322\ M\^88 - 193920\ M\^90 + 
                    86790\ M\^92 - 14551\ M\^94 - 18933\ M\^96 + 
                    29508\ M\^98 - 22279\ M\^100 + 10319\ M\^102 - 
                    3060\ M\^104 + 572\ M\^106 - 62\ M\^108 + 3\ M\^110)\) + 
              L\^17\ \((20 - 478\ M\^2 + 5065\ M\^4 - 30834\ M\^6 + 
                    115619\ M\^8 - 257155\ M\^10 + 245279\ M\^12 + 
                    266641\ M\^14 - 951572\ M\^16 + 445716\ M\^18 + 
                    1281980\ M\^20 - 1242461\ M\^22 - 1085259\ M\^24 + 
                    362939\ M\^26 + 3423082\ M\^28 - 300497\ M\^30 - 
                    10610403\ M\^32 + 8624866\ M\^34 + 21428064\ M\^36 - 
                    32574417\ M\^38 - 34289572\ M\^40 + 69401764\ M\^42 + 
                    53986560\ M\^44 - 104451700\ M\^46 - 88727032\ M\^48 + 
                    120062282\ M\^50 + 139838692\ M\^52 - 111241695\ M\^54 - 
                    192337359\ M\^56 + 92581638\ M\^58 + 223641008\ M\^60 - 
                    77557672\ M\^62 - 211069336\ M\^64 + 69737144\ M\^66 + 
                    156073313\ M\^68 - 61166291\ M\^70 - 86158848\ M\^72 + 
                    44132609\ M\^74 + 32376965\ M\^76 - 22607142\ M\^78 - 
                    6566905\ M\^80 + 6418954\ M\^82 + 867315\ M\^84 - 
                    203624\ M\^86 - 1388359\ M\^88 - 12941\ M\^90 + 
                    1270328\ M\^92 - 700542\ M\^94 - 256609\ M\^96 + 
                    437928\ M\^98 - 146588\ M\^100 - 62996\ M\^102 + 
                    91733\ M\^104 - 54622\ M\^106 + 22718\ M\^108 - 
                    7155\ M\^110 + 1652\ M\^112 - 258\ M\^114 + 24\ M\^116 - 
                    M\^118)\) + 
              L\^16\ \((\(-15\) + 382\ M\^2 - 4358\ M\^4 + 29058\ M\^6 - 
                    122877\ M\^8 + 326856\ M\^10 - 461115\ M\^12 - 
                    59734\ M\^14 + 1449047\ M\^16 - 1786451\ M\^18 - 
                    1727114\ M\^20 + 5664712\ M\^22 - 2193\ M\^24 - 
                    11330744\ M\^26 + 2246826\ M\^28 + 24733790\ M\^30 - 
                    7471457\ M\^32 - 58970414\ M\^34 + 28653108\ M\^36 + 
                    131654917\ M\^38 - 89687078\ M\^40 - 256191234\ M\^42 + 
                    208391571\ M\^44 + 431986123\ M\^46 - 375960479\ M\^48 - 
                    630677451\ M\^50 + 546340084\ M\^52 + 796651128\ M\^54 - 
                    652210031\ M\^56 - 869231726\ M\^58 + 641264610\ M\^60 + 
                    821850306\ M\^62 - 505387897\ M\^64 - 669423814\ M\^66 + 
                    306951950\ M\^68 + 475445387\ M\^70 - 131877367\ M\^72 - 
                    306440508\ M\^74 + 31477899\ M\^76 + 188794076\ M\^78 - 
                    3407172\ M\^80 - 111003289\ M\^82 + 10741346\ M\^84 + 
                    56524761\ M\^86 - 18817570\ M\^88 - 19858799\ M\^90 + 
                    15119760\ M\^92 + 1811376\ M\^94 - 6420379\ M\^96 + 
                    2414110\ M\^98 + 696420\ M\^100 - 986864\ M\^102 + 
                    434834\ M\^104 - 64120\ M\^106 - 107211\ M\^108 + 
                    141725\ M\^110 - 93174\ M\^112 + 37815\ M\^114 - 
                    9820\ M\^116 + 1595\ M\^118 - 148\ M\^120 + 
                    6\ M\^122)\) + 
              L\^15\ \((6 - 161\ M\^2 + 1967\ M\^4 - 14459\ M\^6 + 
                    70853\ M\^8 - 239266\ M\^10 + 536772\ M\^12 - 
                    622314\ M\^14 - 447948\ M\^16 + 3052222\ M\^18 - 
                    3625828\ M\^20 - 4811231\ M\^22 + 17153240\ M\^24 - 
                    3500641\ M\^26 - 43154610\ M\^28 + 36509122\ M\^30 + 
                    85091647\ M\^32 - 117446357\ M\^34 - 147150230\ M\^36 + 
                    274490389\ M\^38 + 240825981\ M\^40 - 530641973\ M\^42 - 
                    389197908\ M\^44 + 876544138\ M\^46 + 629442194\ M\^48 - 
                    1250162320\ M\^50 - 992790146\ M\^52 + 
                    1539355211\ M\^54 + 1467563027\ M\^56 - 
                    1626401440\ M\^58 - 1959461542\ M\^60 + 
                    1456832265\ M\^62 + 2300955117\ M\^64 - 
                    1089360783\ M\^66 - 2325308474\ M\^68 + 
                    692552158\ M\^70 + 2009438325\ M\^72 - 406925191\ M\^74 - 
                    1477158394\ M\^76 + 258955804\ M\^78 + 909797505\ M\^80 - 
                    190135465\ M\^82 - 452031707\ M\^84 + 135918606\ M\^86 + 
                    166903898\ M\^88 - 77204791\ M\^90 - 36304890\ M\^92 + 
                    28448596\ M\^94 + 286482\ M\^96 - 4101373\ M\^98 + 
                    806359\ M\^100 - 1018890\ M\^102 + 1733291\ M\^104 - 
                    537146\ M\^106 - 857206\ M\^108 + 854832\ M\^110 - 
                    120319\ M\^112 - 281805\ M\^114 + 244333\ M\^116 - 
                    103512\ M\^118 + 26585\ M\^120 - 4210\ M\^122 + 
                    380\ M\^124 - 15\ M\^126)\) + 
              L\ \((\(-4\)\ M\^106 + 14\ M\^108 - 18\ M\^110 + 11\ M\^112 + 
                    2\ M\^114 - 29\ M\^116 + 61\ M\^118 + 7\ M\^120 - 
                    36\ M\^122 + 22\ M\^124 - 5\ M\^126)\) + 
              L\^2\ \((6\ M\^90 - 40\ M\^92 + 115\ M\^94 - 190\ M\^96 + 
                    159\ M\^98 + 226\ M\^100 - 829\ M\^102 + 379\ M\^104 + 
                    1132\ M\^106 - 923\ M\^108 - 994\ M\^110 + 744\ M\^112 + 
                    1312\ M\^114 - 543\ M\^116 - 915\ M\^118 + 861\ M\^120 + 
                    19\ M\^122 - 430\ M\^124 + 285\ M\^126 - 84\ M\^128 + 
                    10\ M\^130)\) + 
              L\^14\ \((\(-1\) + 28\ M\^2 - 374\ M\^4 + 3209\ M\^6 - 
                    19808\ M\^8 + 91206\ M\^10 - 307961\ M\^12 + 
                    702510\ M\^14 - 786929\ M\^16 - 838938\ M\^18 + 
                    4581972\ M\^20 - 4725875\ M\^22 - 8711731\ M\^24 + 
                    26035185\ M\^26 - 2021623\ M\^28 - 70054953\ M\^30 + 
                    56133203\ M\^32 + 139729668\ M\^34 - 198387050\ M\^36 - 
                    236210023\ M\^38 + 485347804\ M\^40 + 380902862\ M\^42 - 
                    984864314\ M\^44 - 634450230\ M\^46 + 1752733897\ M\^48 + 
                    1097454590\ M\^50 - 2778400417\ M\^52 - 
                    1857670995\ M\^54 + 3902104295\ M\^56 + 
                    2897495247\ M\^58 - 4795762657\ M\^60 - 
                    4003493709\ M\^62 + 5094237738\ M\^64 + 
                    4793391941\ M\^66 - 4638800395\ M\^68 - 
                    4914616468\ M\^70 + 3623260358\ M\^72 + 
                    4319650206\ M\^74 - 2425989720\ M\^76 - 
                    3263422592\ M\^78 + 1388986302\ M\^80 + 
                    2123487205\ M\^82 - 680408496\ M\^84 - 
                    1193943740\ M\^86 + 289065107\ M\^88 + 586657247\ M\^90 - 
                    113959089\ M\^92 - 257280057\ M\^94 + 50140878\ M\^96 + 
                    102813263\ M\^98 - 29957171\ M\^100 - 35002160\ M\^102 + 
                    19461014\ M\^104 + 7081590\ M\^106 - 10018451\ M\^108 + 
                    1992424\ M\^110 + 2569248\ M\^112 - 2005683\ M\^114 + 
                    220557\ M\^116 + 517800\ M\^118 - 406515\ M\^120 + 
                    160671\ M\^122 - 39151\ M\^124 + 5951\ M\^126 - 
                    520\ M\^128 + 20\ M\^130)\) + 
              L\^13\ \((6\ M\^4 - 164\ M\^6 + 2029\ M\^8 - 15078\ M\^10 + 
                    75089\ M\^12 - 262044\ M\^14 + 630546\ M\^16 - 
                    874488\ M\^18 - 201420\ M\^20 + 3935229\ M\^22 - 
                    6942171\ M\^24 - 3418512\ M\^26 + 30387945\ M\^28 - 
                    25355601\ M\^30 - 70526445\ M\^32 + 132026123\ M\^34 + 
                    107251679\ M\^36 - 398843060\ M\^38 - 89156134\ M\^40 + 
                    935407329\ M\^42 - 44796590\ M\^44 - 1870799282\ M\^46 + 
                    314675858\ M\^48 + 3310373392\ M\^50 - 622726601\ M\^52 - 
                    5246126639\ M\^54 + 722890834\ M\^56 + 
                    7427222511\ M\^58 - 307192316\ M\^60 - 
                    9322499056\ M\^62 - 763825621\ M\^64 + 
                    10304423044\ M\^66 + 2262260896\ M\^68 - 
                    10008419248\ M\^70 - 3652450879\ M\^72 + 
                    8570574632\ M\^74 + 4423968277\ M\^76 - 
                    6496354966\ M\^78 - 4338375951\ M\^80 + 
                    4401176156\ M\^82 + 3542196062\ M\^84 - 
                    2721691071\ M\^86 - 2432060893\ M\^88 + 
                    1584098742\ M\^90 + 1400648303\ M\^92 - 
                    889363474\ M\^94 - 660469100\ M\^96 + 477529183\ M\^98 + 
                    236822973\ M\^100 - 233782381\ M\^102 - 
                    47353342\ M\^104 + 95126993\ M\^106 - 9708604\ M\^108 - 
                    27313246\ M\^110 + 13167545\ M\^112 + 2273100\ M\^114 - 
                    4763539\ M\^116 + 1824339\ M\^118 + 318078\ M\^120 - 
                    722085\ M\^122 + 420560\ M\^124 - 145621\ M\^126 + 
                    32826\ M\^128 - 4742\ M\^130 + 400\ M\^132 - 
                    15\ M\^134)\) + 
              L\^3\ \((\(-4\)\ M\^74 + 38\ M\^76 - 162\ M\^78 + 414\ M\^80 - 
                    648\ M\^82 + 289\ M\^84 + 1126\ M\^86 - 1977\ M\^88 + 
                    28\ M\^90 + 2122\ M\^92 - 2090\ M\^94 + 3695\ M\^96 - 
                    49\ M\^98 - 14933\ M\^100 + 8530\ M\^102 + 
                    25981\ M\^104 - 21455\ M\^106 - 26173\ M\^108 + 
                    31147\ M\^110 + 14090\ M\^112 - 24728\ M\^114 + 
                    3099\ M\^116 + 8560\ M\^118 - 6986\ M\^120 + 
                    2620\ M\^122 + 904\ M\^124 - 2259\ M\^126 + 
                    1617\ M\^128 - 606\ M\^130 + 120\ M\^132 - 
                    10\ M\^134)\) + 
              L\^4\ \((M\^58 - 12\ M\^60 + 66\ M\^62 - 224\ M\^64 + 
                    509\ M\^66 - 730\ M\^68 + 536\ M\^70 - 255\ M\^72 + 
                    881\ M\^74 - 1287\ M\^76 - 662\ M\^78 + 2767\ M\^80 - 
                    6403\ M\^82 + 9998\ M\^84 + 18368\ M\^86 - 55286\ M\^88 - 
                    22962\ M\^90 + 123753\ M\^92 + 42401\ M\^94 - 
                    217990\ M\^96 - 77348\ M\^98 + 319841\ M\^100 + 
                    63106\ M\^102 - 337357\ M\^104 + 31531\ M\^106 + 
                    208163\ M\^108 - 81489\ M\^110 - 6841\ M\^112 + 
                    34927\ M\^114 - 81986\ M\^116 + 38701\ M\^118 + 
                    40593\ M\^120 - 45852\ M\^122 + 9259\ M\^124 + 
                    10373\ M\^126 - 10305\ M\^128 + 5451\ M\^130 - 
                    2025\ M\^132 + 510\ M\^134 - 76\ M\^136 + 5\ M\^138)\) + 
              L\^12\ \((\(-15\)\ M\^8 + 400\ M\^10 - 4742\ M\^12 + 
                    32826\ M\^14 - 145621\ M\^16 + 420560\ M\^18 - 
                    722085\ M\^20 + 318078\ M\^22 + 1824339\ M\^24 - 
                    4763539\ M\^26 + 2273100\ M\^28 + 13167545\ M\^30 - 
                    27313246\ M\^32 - 9708604\ M\^34 + 95126993\ M\^36 - 
                    47353342\ M\^38 - 233782381\ M\^40 + 236822973\ M\^42 + 
                    477529183\ M\^44 - 660469100\ M\^46 - 889363474\ M\^48 + 
                    1400648303\ M\^50 + 1584098742\ M\^52 - 
                    2432060893\ M\^54 - 2721691071\ M\^56 + 
                    3542196062\ M\^58 + 4401176156\ M\^60 - 
                    4338375951\ M\^62 - 6496354966\ M\^64 + 
                    4423968277\ M\^66 + 8570574632\ M\^68 - 
                    3652450879\ M\^70 - 10008419248\ M\^72 + 
                    2262260896\ M\^74 + 10304423044\ M\^76 - 
                    763825621\ M\^78 - 9322499056\ M\^80 - 307192316\ M\^82 + 
                    7427222511\ M\^84 + 722890834\ M\^86 - 
                    5246126639\ M\^88 - 622726601\ M\^90 + 
                    3310373392\ M\^92 + 314675858\ M\^94 - 
                    1870799282\ M\^96 - 44796590\ M\^98 + 935407329\ M\^100 - 
                    89156134\ M\^102 - 398843060\ M\^104 + 
                    107251679\ M\^106 + 132026123\ M\^108 - 
                    70526445\ M\^110 - 25355601\ M\^112 + 30387945\ M\^114 - 
                    3418512\ M\^116 - 6942171\ M\^118 + 3935229\ M\^120 - 
                    201420\ M\^122 - 874488\ M\^124 + 630546\ M\^126 - 
                    262044\ M\^128 + 75089\ M\^130 - 15078\ M\^132 + 
                    2029\ M\^134 - 164\ M\^136 + 6\ M\^138)\) + 
              L\^7\ \((3\ M\^32 - 62\ M\^34 + 572\ M\^36 - 3060\ M\^38 + 
                    10319\ M\^40 - 22279\ M\^42 + 29508\ M\^44 - 
                    18933\ M\^46 - 14551\ M\^48 + 86790\ M\^50 - 
                    193920\ M\^52 + 112322\ M\^54 + 523123\ M\^56 - 
                    1234275\ M\^58 + 177142\ M\^60 + 3001125\ M\^62 - 
                    3275229\ M\^64 - 3930935\ M\^66 + 8762908\ M\^68 + 
                    2837488\ M\^70 - 13959646\ M\^72 - 2418005\ M\^74 + 
                    15996699\ M\^76 + 8711949\ M\^78 - 14569512\ M\^80 - 
                    25558638\ M\^82 + 12780477\ M\^84 + 47854847\ M\^86 - 
                    13550758\ M\^88 - 63494563\ M\^90 + 15812822\ M\^92 + 
                    62828004\ M\^94 - 15482108\ M\^96 - 45810329\ M\^98 + 
                    12951767\ M\^100 + 23512827\ M\^102 - 9727843\ M\^104 - 
                    6070005\ M\^106 + 6617722\ M\^108 - 2903735\ M\^110 - 
                    3085173\ M\^112 + 4667232\ M\^114 - 67684\ M\^116 - 
                    2650924\ M\^118 + 1290900\ M\^120 + 342861\ M\^122 - 
                    675540\ M\^124 + 371539\ M\^126 - 21840\ M\^128 - 
                    145567\ M\^130 + 132129\ M\^132 - 62420\ M\^134 + 
                    18081\ M\^136 - 3239\ M\^138 + 332\ M\^140 - 
                    15\ M\^142)\) + 
              L\^9\ \((6\ M\^20 - 148\ M\^22 + 1595\ M\^24 - 9820\ M\^26 + 
                    37815\ M\^28 - 93174\ M\^30 + 141725\ M\^32 - 
                    107211\ M\^34 - 64120\ M\^36 + 434834\ M\^38 - 
                    986864\ M\^40 + 696420\ M\^42 + 2414110\ M\^44 - 
                    6420379\ M\^46 + 1811376\ M\^48 + 15119760\ M\^50 - 
                    19858799\ M\^52 - 18817570\ M\^54 + 56524761\ M\^56 + 
                    10741346\ M\^58 - 111003289\ M\^60 - 3407172\ M\^62 + 
                    188794076\ M\^64 + 31477899\ M\^66 - 306440508\ M\^68 - 
                    131877367\ M\^70 + 475445387\ M\^72 + 306951950\ M\^74 - 
                    669423814\ M\^76 - 505387897\ M\^78 + 821850306\ M\^80 + 
                    641264610\ M\^82 - 869231726\ M\^84 - 652210031\ M\^86 + 
                    796651128\ M\^88 + 546340084\ M\^90 - 630677451\ M\^92 - 
                    375960479\ M\^94 + 431986123\ M\^96 + 208391571\ M\^98 - 
                    256191234\ M\^100 - 89687078\ M\^102 + 
                    131654917\ M\^104 + 28653108\ M\^106 - 58970414\ M\^108 - 
                    7471457\ M\^110 + 24733790\ M\^112 + 2246826\ M\^114 - 
                    11330744\ M\^116 - 2193\ M\^118 + 5664712\ M\^120 - 
                    1727114\ M\^122 - 1786451\ M\^124 + 1449047\ M\^126 - 
                    59734\ M\^128 - 461115\ M\^130 + 326856\ M\^132 - 
                    122877\ M\^134 + 29058\ M\^136 - 4358\ M\^138 + 
                    382\ M\^140 - 15\ M\^142)\) + 
              L\^5\ \((M\^48 - 16\ M\^50 + 125\ M\^52 - 614\ M\^54 + 
                    1989\ M\^56 - 3972\ M\^58 + 3323\ M\^60 + 4489\ M\^62 - 
                    13948\ M\^64 + 3211\ M\^66 + 37606\ M\^68 - 
                    58997\ M\^70 - 27152\ M\^72 + 178758\ M\^74 - 
                    97294\ M\^76 - 313962\ M\^78 + 342880\ M\^80 + 
                    393933\ M\^82 - 482324\ M\^84 - 612773\ M\^86 + 
                    381416\ M\^88 + 1220798\ M\^90 - 248985\ M\^92 - 
                    1972899\ M\^94 + 332897\ M\^96 + 2053483\ M\^98 - 
                    399610\ M\^100 - 1094465\ M\^102 + 84828\ M\^104 - 
                    20997\ M\^106 + 595820\ M\^108 + 386010\ M\^110 - 
                    908893\ M\^112 - 47757\ M\^114 + 583042\ M\^116 - 
                    233660\ M\^118 - 94540\ M\^120 + 144214\ M\^122 - 
                    84746\ M\^124 + 9334\ M\^126 + 31692\ M\^128 - 
                    29650\ M\^130 + 14029\ M\^132 - 4308\ M\^134 + 
                    956\ M\^136 - 159\ M\^138 + 18\ M\^140 - M\^142)\) + 
              L\^11\ \((20\ M\^12 - 520\ M\^14 + 5951\ M\^16 - 39151\ M\^18 + 
                    160671\ M\^20 - 406515\ M\^22 + 517800\ M\^24 + 
                    220557\ M\^26 - 2005683\ M\^28 + 2569248\ M\^30 + 
                    1992424\ M\^32 - 10018451\ M\^34 + 7081590\ M\^36 + 
                    19461014\ M\^38 - 35002160\ M\^40 - 29957171\ M\^42 + 
                    102813263\ M\^44 + 50140878\ M\^46 - 257280057\ M\^48 - 
                    113959089\ M\^50 + 586657247\ M\^52 + 289065107\ M\^54 - 
                    1193943740\ M\^56 - 680408496\ M\^58 + 
                    2123487205\ M\^60 + 1388986302\ M\^62 - 
                    3263422592\ M\^64 - 2425989720\ M\^66 + 
                    4319650206\ M\^68 + 3623260358\ M\^70 - 
                    4914616468\ M\^72 - 4638800395\ M\^74 + 
                    4793391941\ M\^76 + 5094237738\ M\^78 - 
                    4003493709\ M\^80 - 4795762657\ M\^82 + 
                    2897495247\ M\^84 + 3902104295\ M\^86 - 
                    1857670995\ M\^88 - 2778400417\ M\^90 + 
                    1097454590\ M\^92 + 1752733897\ M\^94 - 
                    634450230\ M\^96 - 984864314\ M\^98 + 380902862\ M\^100 + 
                    485347804\ M\^102 - 236210023\ M\^104 - 
                    198387050\ M\^106 + 139729668\ M\^108 + 
                    56133203\ M\^110 - 70054953\ M\^112 - 2021623\ M\^114 + 
                    26035185\ M\^116 - 8711731\ M\^118 - 4725875\ M\^120 + 
                    4581972\ M\^122 - 838938\ M\^124 - 786929\ M\^126 + 
                    702510\ M\^128 - 307961\ M\^130 + 91206\ M\^132 - 
                    19808\ M\^134 + 3209\ M\^136 - 374\ M\^138 + 28\ M\^140 - 
                    M\^142)\) + 
              L\^10\ \((\(-15\)\ M\^16 + 380\ M\^18 - 4210\ M\^20 + 
                    26585\ M\^22 - 103512\ M\^24 + 244333\ M\^26 - 
                    281805\ M\^28 - 120319\ M\^30 + 854832\ M\^32 - 
                    857206\ M\^34 - 537146\ M\^36 + 1733291\ M\^38 - 
                    1018890\ M\^40 + 806359\ M\^42 - 4101373\ M\^44 + 
                    286482\ M\^46 + 28448596\ M\^48 - 36304890\ M\^50 - 
                    77204791\ M\^52 + 166903898\ M\^54 + 135918606\ M\^56 - 
                    452031707\ M\^58 - 190135465\ M\^60 + 909797505\ M\^62 + 
                    258955804\ M\^64 - 1477158394\ M\^66 - 406925191\ M\^68 + 
                    2009438325\ M\^70 + 692552158\ M\^72 - 
                    2325308474\ M\^74 - 1089360783\ M\^76 + 
                    2300955117\ M\^78 + 1456832265\ M\^80 - 
                    1959461542\ M\^82 - 1626401440\ M\^84 + 
                    1467563027\ M\^86 + 1539355211\ M\^88 - 
                    992790146\ M\^90 - 1250162320\ M\^92 + 629442194\ M\^94 + 
                    876544138\ M\^96 - 389197908\ M\^98 - 530641973\ M\^100 + 
                    240825981\ M\^102 + 274490389\ M\^104 - 
                    147150230\ M\^106 - 117446357\ M\^108 + 
                    85091647\ M\^110 + 36509122\ M\^112 - 43154610\ M\^114 - 
                    3500641\ M\^116 + 17153240\ M\^118 - 4811231\ M\^120 - 
                    3625828\ M\^122 + 3052222\ M\^124 - 447948\ M\^126 - 
                    622314\ M\^128 + 536772\ M\^130 - 239266\ M\^132 + 
                    70853\ M\^134 - 14459\ M\^136 + 1967\ M\^138 - 
                    161\ M\^140 + 6\ M\^142)\) + 
              L\^6\ \((\(-3\)\ M\^40 + 53\ M\^42 - 433\ M\^44 + 2121\ M\^46 - 
                    6621\ M\^48 + 12462\ M\^50 - 9427\ M\^52 - 13693\ M\^54 + 
                    37138\ M\^56 - 6493\ M\^58 - 74035\ M\^60 + 
                    68923\ M\^62 + 87987\ M\^64 - 102557\ M\^66 - 
                    250373\ M\^68 + 200363\ M\^70 + 847549\ M\^72 - 
                    658564\ M\^74 - 2172900\ M\^76 + 1845217\ M\^78 + 
                    4129973\ M\^80 - 3364795\ M\^82 - 6499331\ M\^84 + 
                    4192778\ M\^86 + 8844452\ M\^88 - 3425933\ M\^90 - 
                    10384433\ M\^92 + 972622\ M\^94 + 10155809\ M\^96 + 
                    2303803\ M\^98 - 8036536\ M\^100 - 4220813\ M\^102 + 
                    5697781\ M\^104 + 3717079\ M\^106 - 3822091\ M\^108 - 
                    1354960\ M\^110 + 2043571\ M\^112 - 653717\ M\^114 - 
                    350292\ M\^116 + 954872\ M\^118 - 545700\ M\^120 - 
                    260530\ M\^122 + 426994\ M\^124 - 139777\ M\^126 - 
                    62083\ M\^128 + 87131\ M\^130 - 51104\ M\^132 + 
                    20415\ M\^134 - 5771\ M\^136 + 1087\ M\^138 - 
                    121\ M\^140 + 6\ M\^142)\) + 
              L\^8\ \((\(-M\^24\) + 24\ M\^26 - 258\ M\^28 + 1652\ M\^30 - 
                    7155\ M\^32 + 22718\ M\^34 - 54622\ M\^36 + 
                    91733\ M\^38 - 62996\ M\^40 - 146588\ M\^42 + 
                    437928\ M\^44 - 256609\ M\^46 - 700542\ M\^48 + 
                    1270328\ M\^50 - 12941\ M\^52 - 1388359\ M\^54 - 
                    203624\ M\^56 + 867315\ M\^58 + 6418954\ M\^60 - 
                    6566905\ M\^62 - 22607142\ M\^64 + 32376965\ M\^66 + 
                    44132609\ M\^68 - 86158848\ M\^70 - 61166291\ M\^72 + 
                    156073313\ M\^74 + 69737144\ M\^76 - 211069336\ M\^78 - 
                    77557672\ M\^80 + 223641008\ M\^82 + 92581638\ M\^84 - 
                    192337359\ M\^86 - 111241695\ M\^88 + 139838692\ M\^90 + 
                    120062282\ M\^92 - 88727032\ M\^94 - 104451700\ M\^96 + 
                    53986560\ M\^98 + 69401764\ M\^100 - 34289572\ M\^102 - 
                    32574417\ M\^104 + 21428064\ M\^106 + 8624866\ M\^108 - 
                    10610403\ M\^110 - 300497\ M\^112 + 3423082\ M\^114 + 
                    362939\ M\^116 - 1085259\ M\^118 - 1242461\ M\^120 + 
                    1281980\ M\^122 + 445716\ M\^124 - 951572\ M\^126 + 
                    266641\ M\^128 + 245279\ M\^130 - 257155\ M\^132 + 
                    115619\ M\^134 - 30834\ M\^136 + 5065\ M\^138 - 
                    478\ M\^140 + 20\ M\^142)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 37]\), \(M\^70 + L\^26\ M\^70 + 
              L\ \((4\ M\^60 - 16\ M\^62 + 29\ M\^64 - 23\ M\^66 - 
                    20\ M\^68 + 78\ M\^70 - 20\ M\^72 - 23\ M\^74 + 
                    29\ M\^76 - 16\ M\^78 + 4\ M\^80)\) + 
              L\^25\ \((4\ M\^60 - 16\ M\^62 + 29\ M\^64 - 23\ M\^66 - 
                    20\ M\^68 + 78\ M\^70 - 20\ M\^72 - 23\ M\^74 + 
                    29\ M\^76 - 16\ M\^78 + 4\ M\^80)\) + 
              L\^2\ \((6\ M\^50 - 48\ M\^52 + 176\ M\^54 - 350\ M\^56 + 
                    244\ M\^58 + 453\ M\^60 - 951\ M\^62 + 177\ M\^64 + 
                    688\ M\^66 - 232\ M\^68 - M\^70 - 232\ M\^72 + 
                    688\ M\^74 + 177\ M\^76 - 951\ M\^78 + 453\ M\^80 + 
                    244\ M\^82 - 350\ M\^84 + 176\ M\^86 - 48\ M\^88 + 
                    6\ M\^90)\) + 
              L\^24\ \((6\ M\^50 - 48\ M\^52 + 176\ M\^54 - 350\ M\^56 + 
                    244\ M\^58 + 453\ M\^60 - 951\ M\^62 + 177\ M\^64 + 
                    688\ M\^66 - 232\ M\^68 - M\^70 - 232\ M\^72 + 
                    688\ M\^74 + 177\ M\^76 - 951\ M\^78 + 453\ M\^80 + 
                    244\ M\^82 - 350\ M\^84 + 176\ M\^86 - 48\ M\^88 + 
                    6\ M\^90)\) + 
              L\^3\ \((4\ M\^40 - 48\ M\^42 + 268\ M\^44 - 861\ M\^46 + 
                    1530\ M\^48 - 1035\ M\^50 - 1029\ M\^52 + 2668\ M\^54 - 
                    2900\ M\^56 + 2407\ M\^58 + 1070\ M\^60 - 6697\ M\^62 + 
                    7392\ M\^64 - 2352\ M\^66 - 5035\ M\^68 + 11836\ M\^70 - 
                    5035\ M\^72 - 2352\ M\^74 + 7392\ M\^76 - 6697\ M\^78 + 
                    1070\ M\^80 + 2407\ M\^82 - 2900\ M\^84 + 2668\ M\^86 - 
                    1029\ M\^88 - 1035\ M\^90 + 1530\ M\^92 - 861\ M\^94 + 
                    268\ M\^96 - 48\ M\^98 + 4\ M\^100)\) + 
              L\^23\ \((4\ M\^40 - 48\ M\^42 + 268\ M\^44 - 861\ M\^46 + 
                    1530\ M\^48 - 1035\ M\^50 - 1029\ M\^52 + 2668\ M\^54 - 
                    2900\ M\^56 + 2407\ M\^58 + 1070\ M\^60 - 6697\ M\^62 + 
                    7392\ M\^64 - 2352\ M\^66 - 5035\ M\^68 + 11836\ M\^70 - 
                    5035\ M\^72 - 2352\ M\^74 + 7392\ M\^76 - 6697\ M\^78 + 
                    1070\ M\^80 + 2407\ M\^82 - 2900\ M\^84 + 2668\ M\^86 - 
                    1029\ M\^88 - 1035\ M\^90 + 1530\ M\^92 - 861\ M\^94 + 
                    268\ M\^96 - 48\ M\^98 + 4\ M\^100)\) + 
              L\^4\ \((M\^30 - 16\ M\^32 + 124\ M\^34 - 588\ M\^36 + 
                    1844\ M\^38 - 3994\ M\^40 + 5860\ M\^42 - 4253\ M\^44 - 
                    5092\ M\^46 + 21874\ M\^48 - 28200\ M\^50 - 9976\ M\^52 + 
                    72931\ M\^54 - 39599\ M\^56 - 91499\ M\^58 + 
                    54852\ M\^60 + 136503\ M\^62 - 20323\ M\^64 - 
                    243859\ M\^66 + 2023\ M\^68 + 317724\ M\^70 + 
                    2023\ M\^72 - 243859\ M\^74 - 20323\ M\^76 + 
                    136503\ M\^78 + 54852\ M\^80 - 91499\ M\^82 - 
                    39599\ M\^84 + 72931\ M\^86 - 9976\ M\^88 - 
                    28200\ M\^90 + 21874\ M\^92 - 5092\ M\^94 - 4253\ M\^96 + 
                    5860\ M\^98 - 3994\ M\^100 + 1844\ M\^102 - 588\ M\^104 + 
                    124\ M\^106 - 16\ M\^108 + M\^110)\) + 
              L\^22\ \((M\^30 - 16\ M\^32 + 124\ M\^34 - 588\ M\^36 + 
                    1844\ M\^38 - 3994\ M\^40 + 5860\ M\^42 - 4253\ M\^44 - 
                    5092\ M\^46 + 21874\ M\^48 - 28200\ M\^50 - 9976\ M\^52 + 
                    72931\ M\^54 - 39599\ M\^56 - 91499\ M\^58 + 
                    54852\ M\^60 + 136503\ M\^62 - 20323\ M\^64 - 
                    243859\ M\^66 + 2023\ M\^68 + 317724\ M\^70 + 
                    2023\ M\^72 - 243859\ M\^74 - 20323\ M\^76 + 
                    136503\ M\^78 + 54852\ M\^80 - 91499\ M\^82 - 
                    39599\ M\^84 + 72931\ M\^86 - 9976\ M\^88 - 
                    28200\ M\^90 + 21874\ M\^92 - 5092\ M\^94 - 4253\ M\^96 + 
                    5860\ M\^98 - 3994\ M\^100 + 1844\ M\^102 - 588\ M\^104 + 
                    124\ M\^106 - 16\ M\^108 + M\^110)\) + 
              L\^5\ \((3\ M\^24 - 54\ M\^26 + 460\ M\^28 - 2348\ M\^30 + 
                    7567\ M\^32 - 14745\ M\^34 + 12995\ M\^36 + 8348\ M\^38 - 
                    26625\ M\^40 + 2403\ M\^42 + 10406\ M\^44 + 
                    77889\ M\^46 - 69496\ M\^48 - 243298\ M\^50 + 
                    263400\ M\^52 + 339782\ M\^54 - 121603\ M\^56 - 
                    815801\ M\^58 - 697259\ M\^60 + 2581589\ M\^62 + 
                    1371798\ M\^64 - 5283015\ M\^66 - 718756\ M\^68 + 
                    6698500\ M\^70 - 718756\ M\^72 - 5283015\ M\^74 + 
                    1371798\ M\^76 + 2581589\ M\^78 - 697259\ M\^80 - 
                    815801\ M\^82 - 121603\ M\^84 + 339782\ M\^86 + 
                    263400\ M\^88 - 243298\ M\^90 - 69496\ M\^92 + 
                    77889\ M\^94 + 10406\ M\^96 + 2403\ M\^98 - 
                    26625\ M\^100 + 8348\ M\^102 + 12995\ M\^104 - 
                    14745\ M\^106 + 7567\ M\^108 - 2348\ M\^110 + 
                    460\ M\^112 - 54\ M\^114 + 3\ M\^116)\) + 
              L\^21\ \((3\ M\^24 - 54\ M\^26 + 460\ M\^28 - 2348\ M\^30 + 
                    7567\ M\^32 - 14745\ M\^34 + 12995\ M\^36 + 8348\ M\^38 - 
                    26625\ M\^40 + 2403\ M\^42 + 10406\ M\^44 + 
                    77889\ M\^46 - 69496\ M\^48 - 243298\ M\^50 + 
                    263400\ M\^52 + 339782\ M\^54 - 121603\ M\^56 - 
                    815801\ M\^58 - 697259\ M\^60 + 2581589\ M\^62 + 
                    1371798\ M\^64 - 5283015\ M\^66 - 718756\ M\^68 + 
                    6698500\ M\^70 - 718756\ M\^72 - 5283015\ M\^74 + 
                    1371798\ M\^76 + 2581589\ M\^78 - 697259\ M\^80 - 
                    815801\ M\^82 - 121603\ M\^84 + 339782\ M\^86 + 
                    263400\ M\^88 - 243298\ M\^90 - 69496\ M\^92 + 
                    77889\ M\^94 + 10406\ M\^96 + 2403\ M\^98 - 
                    26625\ M\^100 + 8348\ M\^102 + 12995\ M\^104 - 
                    14745\ M\^106 + 7567\ M\^108 - 2348\ M\^110 + 
                    460\ M\^112 - 54\ M\^114 + 3\ M\^116)\) + 
              L\^6\ \((3\ M\^18 - 63\ M\^20 + 635\ M\^22 - 3866\ M\^24 + 
                    15069\ M\^26 - 36557\ M\^28 + 44583\ M\^30 + 
                    14701\ M\^32 - 121158\ M\^34 + 67905\ M\^36 + 
                    225643\ M\^38 - 193917\ M\^40 - 540351\ M\^42 + 
                    571141\ M\^44 + 1098916\ M\^46 - 1401050\ M\^48 - 
                    1167924\ M\^50 + 803640\ M\^52 + 1769618\ M\^54 + 
                    4239402\ M\^56 - 7950431\ M\^58 - 12362999\ M\^60 + 
                    24021741\ M\^62 + 15383306\ M\^64 - 44110134\ M\^66 - 
                    7081643\ M\^68 + 53657810\ M\^70 - 7081643\ M\^72 - 
                    44110134\ M\^74 + 15383306\ M\^76 + 24021741\ M\^78 - 
                    12362999\ M\^80 - 7950431\ M\^82 + 4239402\ M\^84 + 
                    1769618\ M\^86 + 803640\ M\^88 - 1167924\ M\^90 - 
                    1401050\ M\^92 + 1098916\ M\^94 + 571141\ M\^96 - 
                    540351\ M\^98 - 193917\ M\^100 + 225643\ M\^102 + 
                    67905\ M\^104 - 121158\ M\^106 + 14701\ M\^108 + 
                    44583\ M\^110 - 36557\ M\^112 + 15069\ M\^114 - 
                    3866\ M\^116 + 635\ M\^118 - 63\ M\^120 + 3\ M\^122)\) + 
              L\^20\ \((3\ M\^18 - 63\ M\^20 + 635\ M\^22 - 3866\ M\^24 + 
                    15069\ M\^26 - 36557\ M\^28 + 44583\ M\^30 + 
                    14701\ M\^32 - 121158\ M\^34 + 67905\ M\^36 + 
                    225643\ M\^38 - 193917\ M\^40 - 540351\ M\^42 + 
                    571141\ M\^44 + 1098916\ M\^46 - 1401050\ M\^48 - 
                    1167924\ M\^50 + 803640\ M\^52 + 1769618\ M\^54 + 
                    4239402\ M\^56 - 7950431\ M\^58 - 12362999\ M\^60 + 
                    24021741\ M\^62 + 15383306\ M\^64 - 44110134\ M\^66 - 
                    7081643\ M\^68 + 53657810\ M\^70 - 7081643\ M\^72 - 
                    44110134\ M\^74 + 15383306\ M\^76 + 24021741\ M\^78 - 
                    12362999\ M\^80 - 7950431\ M\^82 + 4239402\ M\^84 + 
                    1769618\ M\^86 + 803640\ M\^88 - 1167924\ M\^90 - 
                    1401050\ M\^92 + 1098916\ M\^94 + 571141\ M\^96 - 
                    540351\ M\^98 - 193917\ M\^100 + 225643\ M\^102 + 
                    67905\ M\^104 - 121158\ M\^106 + 14701\ M\^108 + 
                    44583\ M\^110 - 36557\ M\^112 + 15069\ M\^114 - 
                    3866\ M\^116 + 635\ M\^118 - 63\ M\^120 + 3\ M\^122)\) + 
              L\^7\ \((M\^12 - 28\ M\^14 + 366\ M\^16 - 2836\ M\^18 + 
                    14135\ M\^20 - 45834\ M\^22 + 90001\ M\^24 - 
                    71969\ M\^26 - 94373\ M\^28 + 244338\ M\^30 + 
                    4651\ M\^32 - 321560\ M\^34 - 364432\ M\^36 + 
                    843684\ M\^38 + 2082766\ M\^40 - 4161318\ M\^42 - 
                    4218355\ M\^44 + 11827865\ M\^46 + 5459127\ M\^48 - 
                    23348298\ M\^50 - 6982344\ M\^52 + 38330882\ M\^54 + 
                    11951967\ M\^56 - 59690885\ M\^58 - 18673784\ M\^60 + 
                    88607779\ M\^62 + 19234899\ M\^64 - 116998322\ M\^66 - 
                    8175725\ M\^68 + 129573004\ M\^70 - 8175725\ M\^72 - 
                    116998322\ M\^74 + 19234899\ M\^76 + 88607779\ M\^78 - 
                    18673784\ M\^80 - 59690885\ M\^82 + 11951967\ M\^84 + 
                    38330882\ M\^86 - 6982344\ M\^88 - 23348298\ M\^90 + 
                    5459127\ M\^92 + 11827865\ M\^94 - 4218355\ M\^96 - 
                    4161318\ M\^98 + 2082766\ M\^100 + 843684\ M\^102 - 
                    364432\ M\^104 - 321560\ M\^106 + 4651\ M\^108 + 
                    244338\ M\^110 - 94373\ M\^112 - 71969\ M\^114 + 
                    90001\ M\^116 - 45834\ M\^118 + 14135\ M\^120 - 
                    2836\ M\^122 + 366\ M\^124 - 28\ M\^126 + M\^128)\) + 
              L\^19\ \((M\^12 - 28\ M\^14 + 366\ M\^16 - 2836\ M\^18 + 
                    14135\ M\^20 - 45834\ M\^22 + 90001\ M\^24 - 
                    71969\ M\^26 - 94373\ M\^28 + 244338\ M\^30 + 
                    4651\ M\^32 - 321560\ M\^34 - 364432\ M\^36 + 
                    843684\ M\^38 + 2082766\ M\^40 - 4161318\ M\^42 - 
                    4218355\ M\^44 + 11827865\ M\^46 + 5459127\ M\^48 - 
                    23348298\ M\^50 - 6982344\ M\^52 + 38330882\ M\^54 + 
                    11951967\ M\^56 - 59690885\ M\^58 - 18673784\ M\^60 + 
                    88607779\ M\^62 + 19234899\ M\^64 - 116998322\ M\^66 - 
                    8175725\ M\^68 + 129573004\ M\^70 - 8175725\ M\^72 - 
                    116998322\ M\^74 + 19234899\ M\^76 + 88607779\ M\^78 - 
                    18673784\ M\^80 - 59690885\ M\^82 + 11951967\ M\^84 + 
                    38330882\ M\^86 - 6982344\ M\^88 - 23348298\ M\^90 + 
                    5459127\ M\^92 + 11827865\ M\^94 - 4218355\ M\^96 - 
                    4161318\ M\^98 + 2082766\ M\^100 + 843684\ M\^102 - 
                    364432\ M\^104 - 321560\ M\^106 + 4651\ M\^108 + 
                    244338\ M\^110 - 94373\ M\^112 - 71969\ M\^114 + 
                    90001\ M\^116 - 45834\ M\^118 + 14135\ M\^120 - 
                    2836\ M\^122 + 366\ M\^124 - 28\ M\^126 + M\^128)\) + 
              L\^8\ \((\(-3\)\ M\^8 + 74\ M\^10 - 850\ M\^12 + 5858\ M\^14 - 
                    26227\ M\^16 + 77204\ M\^18 - 141827\ M\^20 + 
                    130742\ M\^22 + 17965\ M\^24 - 175994\ M\^26 + 
                    389116\ M\^28 - 1182801\ M\^30 + 1202784\ M\^32 + 
                    4117482\ M\^34 - 11095090\ M\^36 - 1076782\ M\^38 + 
                    32024349\ M\^40 - 22003269\ M\^42 - 53650264\ M\^44 + 
                    73504967\ M\^46 + 49968890\ M\^48 - 135418887\ M\^50 + 
                    1404664\ M\^52 + 160783463\ M\^54 - 91710884\ M\^56 - 
                    102106905\ M\^58 + 169146810\ M\^60 - 39101251\ M\^62 - 
                    169587205\ M\^64 + 191724889\ M\^66 + 72057772\ M\^68 - 
                    256995305\ M\^70 + 72057772\ M\^72 + 191724889\ M\^74 - 
                    169587205\ M\^76 - 39101251\ M\^78 + 169146810\ M\^80 - 
                    102106905\ M\^82 - 91710884\ M\^84 + 160783463\ M\^86 + 
                    1404664\ M\^88 - 135418887\ M\^90 + 49968890\ M\^92 + 
                    73504967\ M\^94 - 53650264\ M\^96 - 22003269\ M\^98 + 
                    32024349\ M\^100 - 1076782\ M\^102 - 11095090\ M\^104 + 
                    4117482\ M\^106 + 1202784\ M\^108 - 1182801\ M\^110 + 
                    389116\ M\^112 - 175994\ M\^114 + 17965\ M\^116 + 
                    130742\ M\^118 - 141827\ M\^120 + 77204\ M\^122 - 
                    26227\ M\^124 + 5858\ M\^126 - 850\ M\^128 + 74\ M\^130 - 
                    3\ M\^132)\) + 
              L\^18\ \((\(-3\)\ M\^8 + 74\ M\^10 - 850\ M\^12 + 5858\ M\^14 - 
                    26227\ M\^16 + 77204\ M\^18 - 141827\ M\^20 + 
                    130742\ M\^22 + 17965\ M\^24 - 175994\ M\^26 + 
                    389116\ M\^28 - 1182801\ M\^30 + 1202784\ M\^32 + 
                    4117482\ M\^34 - 11095090\ M\^36 - 1076782\ M\^38 + 
                    32024349\ M\^40 - 22003269\ M\^42 - 53650264\ M\^44 + 
                    73504967\ M\^46 + 49968890\ M\^48 - 135418887\ M\^50 + 
                    1404664\ M\^52 + 160783463\ M\^54 - 91710884\ M\^56 - 
                    102106905\ M\^58 + 169146810\ M\^60 - 39101251\ M\^62 - 
                    169587205\ M\^64 + 191724889\ M\^66 + 72057772\ M\^68 - 
                    256995305\ M\^70 + 72057772\ M\^72 + 191724889\ M\^74 - 
                    169587205\ M\^76 - 39101251\ M\^78 + 169146810\ M\^80 - 
                    102106905\ M\^82 - 91710884\ M\^84 + 160783463\ M\^86 + 
                    1404664\ M\^88 - 135418887\ M\^90 + 49968890\ M\^92 + 
                    73504967\ M\^94 - 53650264\ M\^96 - 22003269\ M\^98 + 
                    32024349\ M\^100 - 1076782\ M\^102 - 11095090\ M\^104 + 
                    4117482\ M\^106 + 1202784\ M\^108 - 1182801\ M\^110 + 
                    389116\ M\^112 - 175994\ M\^114 + 17965\ M\^116 + 
                    130742\ M\^118 - 141827\ M\^120 + 77204\ M\^122 - 
                    26227\ M\^124 + 5858\ M\^126 - 850\ M\^128 + 74\ M\^130 - 
                    3\ M\^132)\) + 
              L\^9\ \((3\ M\^4 - 74\ M\^6 + 835\ M\^8 - 5653\ M\^10 + 
                    25346\ M\^12 - 78756\ M\^14 + 174063\ M\^16 - 
                    276706\ M\^18 + 275360\ M\^20 + 144314\ M\^22 - 
                    1461616\ M\^24 + 2733793\ M\^26 + 522205\ M\^28 - 
                    10085720\ M\^30 + 10893345\ M\^32 + 14757597\ M\^34 - 
                    36142661\ M\^36 - 4556170\ M\^38 + 63727389\ M\^40 - 
                    22949921\ M\^42 - 66722225\ M\^44 + 40697516\ M\^46 + 
                    25230722\ M\^48 - 418440\ M\^50 + 55858984\ M\^52 - 
                    138471416\ M\^54 - 143373336\ M\^56 + 374203479\ M\^58 + 
                    192477308\ M\^60 - 652063536\ M\^62 - 167687677\ M\^64 + 
                    878597778\ M\^66 + 67764230\ M\^68 - 964456170\ M\^70 + 
                    67764230\ M\^72 + 878597778\ M\^74 - 167687677\ M\^76 - 
                    652063536\ M\^78 + 192477308\ M\^80 + 374203479\ M\^82 - 
                    143373336\ M\^84 - 138471416\ M\^86 + 55858984\ M\^88 - 
                    418440\ M\^90 + 25230722\ M\^92 + 40697516\ M\^94 - 
                    66722225\ M\^96 - 22949921\ M\^98 + 63727389\ M\^100 - 
                    4556170\ M\^102 - 36142661\ M\^104 + 14757597\ M\^106 + 
                    10893345\ M\^108 - 10085720\ M\^110 + 522205\ M\^112 + 
                    2733793\ M\^114 - 1461616\ M\^116 + 144314\ M\^118 + 
                    275360\ M\^120 - 276706\ M\^122 + 174063\ M\^124 - 
                    78756\ M\^126 + 25346\ M\^128 - 5653\ M\^130 + 
                    835\ M\^132 - 74\ M\^134 + 3\ M\^136)\) + 
              L\^17\ \((3\ M\^4 - 74\ M\^6 + 835\ M\^8 - 5653\ M\^10 + 
                    25346\ M\^12 - 78756\ M\^14 + 174063\ M\^16 - 
                    276706\ M\^18 + 275360\ M\^20 + 144314\ M\^22 - 
                    1461616\ M\^24 + 2733793\ M\^26 + 522205\ M\^28 - 
                    10085720\ M\^30 + 10893345\ M\^32 + 14757597\ M\^34 - 
                    36142661\ M\^36 - 4556170\ M\^38 + 63727389\ M\^40 - 
                    22949921\ M\^42 - 66722225\ M\^44 + 40697516\ M\^46 + 
                    25230722\ M\^48 - 418440\ M\^50 + 55858984\ M\^52 - 
                    138471416\ M\^54 - 143373336\ M\^56 + 374203479\ M\^58 + 
                    192477308\ M\^60 - 652063536\ M\^62 - 167687677\ M\^64 + 
                    878597778\ M\^66 + 67764230\ M\^68 - 964456170\ M\^70 + 
                    67764230\ M\^72 + 878597778\ M\^74 - 167687677\ M\^76 - 
                    652063536\ M\^78 + 192477308\ M\^80 + 374203479\ M\^82 - 
                    143373336\ M\^84 - 138471416\ M\^86 + 55858984\ M\^88 - 
                    418440\ M\^90 + 25230722\ M\^92 + 40697516\ M\^94 - 
                    66722225\ M\^96 - 22949921\ M\^98 + 63727389\ M\^100 - 
                    4556170\ M\^102 - 36142661\ M\^104 + 14757597\ M\^106 + 
                    10893345\ M\^108 - 10085720\ M\^110 + 522205\ M\^112 + 
                    2733793\ M\^114 - 1461616\ M\^116 + 144314\ M\^118 + 
                    275360\ M\^120 - 276706\ M\^122 + 174063\ M\^124 - 
                    78756\ M\^126 + 25346\ M\^128 - 5653\ M\^130 + 
                    835\ M\^132 - 74\ M\^134 + 3\ M\^136)\) + 
              L\^12\ \((\(-15\) + 395\ M\^2 - 4682\ M\^4 + 32688\ M\^6 - 
                    146423\ M\^8 + 421326\ M\^10 - 686341\ M\^12 + 
                    167917\ M\^14 + 1743322\ M\^16 - 2877804\ M\^18 - 
                    1390251\ M\^20 + 8617360\ M\^22 - 4431795\ M\^24 - 
                    12472359\ M\^26 + 11285461\ M\^28 + 14023268\ M\^30 - 
                    7223345\ M\^32 - 22836335\ M\^34 - 27062913\ M\^36 + 
                    69706662\ M\^38 + 94346460\ M\^40 - 183140535\ M\^42 - 
                    171745470\ M\^44 + 355400408\ M\^46 + 209616904\ M\^48 - 
                    500126155\ M\^50 - 171916050\ M\^52 + 480609754\ M\^54 + 
                    67867398\ M\^56 - 203583268\ M\^58 + 39793389\ M\^60 - 
                    269911557\ M\^62 - 84119493\ M\^64 + 722163331\ M\^66 + 
                    44073844\ M\^68 - 902732492\ M\^70 + 44073844\ M\^72 + 
                    722163331\ M\^74 - 84119493\ M\^76 - 269911557\ M\^78 + 
                    39793389\ M\^80 - 203583268\ M\^82 + 67867398\ M\^84 + 
                    480609754\ M\^86 - 171916050\ M\^88 - 500126155\ M\^90 + 
                    209616904\ M\^92 + 355400408\ M\^94 - 171745470\ M\^96 - 
                    183140535\ M\^98 + 94346460\ M\^100 + 69706662\ M\^102 - 
                    27062913\ M\^104 - 22836335\ M\^106 - 7223345\ M\^108 + 
                    14023268\ M\^110 + 11285461\ M\^112 - 12472359\ M\^114 - 
                    4431795\ M\^116 + 8617360\ M\^118 - 1390251\ M\^120 - 
                    2877804\ M\^122 + 1743322\ M\^124 + 167917\ M\^126 - 
                    686341\ M\^128 + 421326\ M\^130 - 146423\ M\^132 + 
                    32688\ M\^134 - 4682\ M\^136 + 395\ M\^138 - 
                    15\ M\^140)\) + 
              L\^14\ \((\(-15\) + 395\ M\^2 - 4682\ M\^4 + 32688\ M\^6 - 
                    146423\ M\^8 + 421326\ M\^10 - 686341\ M\^12 + 
                    167917\ M\^14 + 1743322\ M\^16 - 2877804\ M\^18 - 
                    1390251\ M\^20 + 8617360\ M\^22 - 4431795\ M\^24 - 
                    12472359\ M\^26 + 11285461\ M\^28 + 14023268\ M\^30 - 
                    7223345\ M\^32 - 22836335\ M\^34 - 27062913\ M\^36 + 
                    69706662\ M\^38 + 94346460\ M\^40 - 183140535\ M\^42 - 
                    171745470\ M\^44 + 355400408\ M\^46 + 209616904\ M\^48 - 
                    500126155\ M\^50 - 171916050\ M\^52 + 480609754\ M\^54 + 
                    67867398\ M\^56 - 203583268\ M\^58 + 39793389\ M\^60 - 
                    269911557\ M\^62 - 84119493\ M\^64 + 722163331\ M\^66 + 
                    44073844\ M\^68 - 902732492\ M\^70 + 44073844\ M\^72 + 
                    722163331\ M\^74 - 84119493\ M\^76 - 269911557\ M\^78 + 
                    39793389\ M\^80 - 203583268\ M\^82 + 67867398\ M\^84 + 
                    480609754\ M\^86 - 171916050\ M\^88 - 500126155\ M\^90 + 
                    209616904\ M\^92 + 355400408\ M\^94 - 171745470\ M\^96 - 
                    183140535\ M\^98 + 94346460\ M\^100 + 69706662\ M\^102 - 
                    27062913\ M\^104 - 22836335\ M\^106 - 7223345\ M\^108 + 
                    14023268\ M\^110 + 11285461\ M\^112 - 12472359\ M\^114 - 
                    4431795\ M\^116 + 8617360\ M\^118 - 1390251\ M\^120 - 
                    2877804\ M\^122 + 1743322\ M\^124 + 167917\ M\^126 - 
                    686341\ M\^128 + 421326\ M\^130 - 146423\ M\^132 + 
                    32688\ M\^134 - 4682\ M\^136 + 395\ M\^138 - 
                    15\ M\^140)\) + 
              L\^10\ \((\(-1\) + 25\ M\^2 - 294\ M\^4 + 2186\ M\^6 - 
                    11641\ M\^8 + 47237\ M\^10 - 148050\ M\^12 + 
                    341574\ M\^14 - 483457\ M\^16 + 53830\ M\^18 + 
                    1351657\ M\^20 - 2314262\ M\^22 - 310643\ M\^24 + 
                    5379423\ M\^26 - 4746120\ M\^28 - 1810436\ M\^30 + 
                    1747145\ M\^32 - 5633720\ M\^34 + 33123197\ M\^36 - 
                    14564500\ M\^38 - 109397306\ M\^40 + 119177789\ M\^42 + 
                    178260150\ M\^44 - 324927727\ M\^46 - 146574955\ M\^48 + 
                    545464930\ M\^50 - 41893856\ M\^52 - 609569487\ M\^54 + 
                    318317870\ M\^56 + 393940145\ M\^58 - 511888574\ M\^60 + 
                    49653103\ M\^62 + 472136450\ M\^64 - 493828423\ M\^66 - 
                    189481572\ M\^68 + 682488361\ M\^70 - 189481572\ M\^72 - 
                    493828423\ M\^74 + 472136450\ M\^76 + 49653103\ M\^78 - 
                    511888574\ M\^80 + 393940145\ M\^82 + 318317870\ M\^84 - 
                    609569487\ M\^86 - 41893856\ M\^88 + 545464930\ M\^90 - 
                    146574955\ M\^92 - 324927727\ M\^94 + 178260150\ M\^96 + 
                    119177789\ M\^98 - 109397306\ M\^100 - 14564500\ M\^102 + 
                    33123197\ M\^104 - 5633720\ M\^106 + 1747145\ M\^108 - 
                    1810436\ M\^110 - 4746120\ M\^112 + 5379423\ M\^114 - 
                    310643\ M\^116 - 2314262\ M\^118 + 1351657\ M\^120 + 
                    53830\ M\^122 - 483457\ M\^124 + 341574\ M\^126 - 
                    148050\ M\^128 + 47237\ M\^130 - 11641\ M\^132 + 
                    2186\ M\^134 - 294\ M\^136 + 25\ M\^138 - M\^140)\) + 
              L\^16\ \((\(-1\) + 25\ M\^2 - 294\ M\^4 + 2186\ M\^6 - 
                    11641\ M\^8 + 47237\ M\^10 - 148050\ M\^12 + 
                    341574\ M\^14 - 483457\ M\^16 + 53830\ M\^18 + 
                    1351657\ M\^20 - 2314262\ M\^22 - 310643\ M\^24 + 
                    5379423\ M\^26 - 4746120\ M\^28 - 1810436\ M\^30 + 
                    1747145\ M\^32 - 5633720\ M\^34 + 33123197\ M\^36 - 
                    14564500\ M\^38 - 109397306\ M\^40 + 119177789\ M\^42 + 
                    178260150\ M\^44 - 324927727\ M\^46 - 146574955\ M\^48 + 
                    545464930\ M\^50 - 41893856\ M\^52 - 609569487\ M\^54 + 
                    318317870\ M\^56 + 393940145\ M\^58 - 511888574\ M\^60 + 
                    49653103\ M\^62 + 472136450\ M\^64 - 493828423\ M\^66 - 
                    189481572\ M\^68 + 682488361\ M\^70 - 189481572\ M\^72 - 
                    493828423\ M\^74 + 472136450\ M\^76 + 49653103\ M\^78 - 
                    511888574\ M\^80 + 393940145\ M\^82 + 318317870\ M\^84 - 
                    609569487\ M\^86 - 41893856\ M\^88 + 545464930\ M\^90 - 
                    146574955\ M\^92 - 324927727\ M\^94 + 178260150\ M\^96 + 
                    119177789\ M\^98 - 109397306\ M\^100 - 14564500\ M\^102 + 
                    33123197\ M\^104 - 5633720\ M\^106 + 1747145\ M\^108 - 
                    1810436\ M\^110 - 4746120\ M\^112 + 5379423\ M\^114 - 
                    310643\ M\^116 - 2314262\ M\^118 + 1351657\ M\^120 + 
                    53830\ M\^122 - 483457\ M\^124 + 341574\ M\^126 - 
                    148050\ M\^128 + 47237\ M\^130 - 11641\ M\^132 + 
                    2186\ M\^134 - 294\ M\^136 + 25\ M\^138 - M\^140)\) + 
              L\^11\ \((6 - 155\ M\^2 + 1819\ M\^4 - 12780\ M\^6 + 
                    59418\ M\^8 - 189532\ M\^10 + 409471\ M\^12 - 
                    532211\ M\^14 + 111065\ M\^16 + 1247387\ M\^18 - 
                    3006165\ M\^20 + 2119015\ M\^22 + 6558223\ M\^24 - 
                    18806056\ M\^26 + 6140221\ M\^28 + 47373242\ M\^30 - 
                    60899878\ M\^32 - 65287340\ M\^34 + 173014810\ M\^36 + 
                    28978398\ M\^38 - 305082437\ M\^40 + 79248794\ M\^42 + 
                    357733096\ M\^44 - 198775940\ M\^46 - 208489103\ M\^48 + 
                    168384984\ M\^50 - 181679604\ M\^52 + 195533089\ M\^54 + 
                    683038224\ M\^56 - 960420427\ M\^58 - 1015895423\ M\^60 + 
                    1959534257\ M\^62 + 922025897\ M\^64 - 
                    2824199935\ M\^66 - 370176560\ M\^68 + 
                    3171610420\ M\^70 - 370176560\ M\^72 - 
                    2824199935\ M\^74 + 922025897\ M\^76 + 
                    1959534257\ M\^78 - 1015895423\ M\^80 - 
                    960420427\ M\^82 + 683038224\ M\^84 + 195533089\ M\^86 - 
                    181679604\ M\^88 + 168384984\ M\^90 - 208489103\ M\^92 - 
                    198775940\ M\^94 + 357733096\ M\^96 + 79248794\ M\^98 - 
                    305082437\ M\^100 + 28978398\ M\^102 + 
                    173014810\ M\^104 - 65287340\ M\^106 - 60899878\ M\^108 + 
                    47373242\ M\^110 + 6140221\ M\^112 - 18806056\ M\^114 + 
                    6558223\ M\^116 + 2119015\ M\^118 - 3006165\ M\^120 + 
                    1247387\ M\^122 + 111065\ M\^124 - 532211\ M\^126 + 
                    409471\ M\^128 - 189532\ M\^130 + 59418\ M\^132 - 
                    12780\ M\^134 + 1819\ M\^136 - 155\ M\^138 + 
                    6\ M\^140)\) + 
              L\^15\ \((6 - 155\ M\^2 + 1819\ M\^4 - 12780\ M\^6 + 
                    59418\ M\^8 - 189532\ M\^10 + 409471\ M\^12 - 
                    532211\ M\^14 + 111065\ M\^16 + 1247387\ M\^18 - 
                    3006165\ M\^20 + 2119015\ M\^22 + 6558223\ M\^24 - 
                    18806056\ M\^26 + 6140221\ M\^28 + 47373242\ M\^30 - 
                    60899878\ M\^32 - 65287340\ M\^34 + 173014810\ M\^36 + 
                    28978398\ M\^38 - 305082437\ M\^40 + 79248794\ M\^42 + 
                    357733096\ M\^44 - 198775940\ M\^46 - 208489103\ M\^48 + 
                    168384984\ M\^50 - 181679604\ M\^52 + 195533089\ M\^54 + 
                    683038224\ M\^56 - 960420427\ M\^58 - 1015895423\ M\^60 + 
                    1959534257\ M\^62 + 922025897\ M\^64 - 
                    2824199935\ M\^66 - 370176560\ M\^68 + 
                    3171610420\ M\^70 - 370176560\ M\^72 - 
                    2824199935\ M\^74 + 922025897\ M\^76 + 
                    1959534257\ M\^78 - 1015895423\ M\^80 - 
                    960420427\ M\^82 + 683038224\ M\^84 + 195533089\ M\^86 - 
                    181679604\ M\^88 + 168384984\ M\^90 - 208489103\ M\^92 - 
                    198775940\ M\^94 + 357733096\ M\^96 + 79248794\ M\^98 - 
                    305082437\ M\^100 + 28978398\ M\^102 + 
                    173014810\ M\^104 - 65287340\ M\^106 - 60899878\ M\^108 + 
                    47373242\ M\^110 + 6140221\ M\^112 - 18806056\ M\^114 + 
                    6558223\ M\^116 + 2119015\ M\^118 - 3006165\ M\^120 + 
                    1247387\ M\^122 + 111065\ M\^124 - 532211\ M\^126 + 
                    409471\ M\^128 - 189532\ M\^130 + 59418\ M\^132 - 
                    12780\ M\^134 + 1819\ M\^136 - 155\ M\^138 + 
                    6\ M\^140)\) + 
              L\^13\ \((20 - 530\ M\^2 + 6308\ M\^4 - 44040\ M\^6 + 
                    195632\ M\^8 - 546925\ M\^10 + 800743\ M\^12 + 
                    191810\ M\^14 - 3036795\ M\^16 + 3552276\ M\^18 + 
                    5780473\ M\^20 - 17270407\ M\^22 - 829196\ M\^24 + 
                    46676021\ M\^26 - 27290559\ M\^28 - 96967577\ M\^30 + 
                    109532074\ M\^32 + 150505878\ M\^34 - 264805046\ M\^36 - 
                    159204592\ M\^38 + 446169754\ M\^40 + 67560001\ M\^42 - 
                    479696671\ M\^44 + 88604995\ M\^46 + 132771908\ M\^48 - 
                    118953834\ M\^50 + 680530310\ M\^52 - 248392085\ M\^54 - 
                    1687993455\ M\^56 + 1133339403\ M\^58 + 
                    2313849072\ M\^60 - 2331407104\ M\^62 - 
                    2038173960\ M\^64 + 3368783226\ M\^66 + 
                    817389688\ M\^68 - 3772853032\ M\^70 + 817389688\ M\^72 + 
                    3368783226\ M\^74 - 2038173960\ M\^76 - 
                    2331407104\ M\^78 + 2313849072\ M\^80 + 
                    1133339403\ M\^82 - 1687993455\ M\^84 - 
                    248392085\ M\^86 + 680530310\ M\^88 - 118953834\ M\^90 + 
                    132771908\ M\^92 + 88604995\ M\^94 - 479696671\ M\^96 + 
                    67560001\ M\^98 + 446169754\ M\^100 - 159204592\ M\^102 - 
                    264805046\ M\^104 + 150505878\ M\^106 + 
                    109532074\ M\^108 - 96967577\ M\^110 - 27290559\ M\^112 + 
                    46676021\ M\^114 - 829196\ M\^116 - 17270407\ M\^118 + 
                    5780473\ M\^120 + 3552276\ M\^122 - 3036795\ M\^124 + 
                    191810\ M\^126 + 800743\ M\^128 - 546925\ M\^130 + 
                    195632\ M\^132 - 44040\ M\^134 + 6308\ M\^136 - 
                    530\ M\^138 + 20\ M\^140)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 38]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 39]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 40]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 41]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 42]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 43]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 44]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 45]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 46]\), \(M\^4 + L\^19\ M\^234 + 
              L\ \((\(-1\) + 3\ M\^2 - 7\ M\^4 + 11\ M\^6 - 10\ M\^8 + 
                    8\ M\^10 + 7\ M\^12 - 21\ M\^14 + 21\ M\^16 + 14\ M\^18 - 
                    10\ M\^20)\) + 
              L\^2\ \((\(-M\^2\) + 4\ M\^4 - 12\ M\^6 + 24\ M\^8 - 
                    48\ M\^10 + 82\ M\^12 - 81\ M\^14 + 116\ M\^18 - 
                    142\ M\^20 + 24\ M\^22 + 256\ M\^24 - 241\ M\^26 - 
                    91\ M\^28 + 369\ M\^30 - 73\ M\^32 - 128\ M\^34 + 
                    45\ M\^36)\) + 
              L\^3\ \((\(-2\)\ M\^8 + 12\ M\^10 - 44\ M\^12 + 103\ M\^14 - 
                    174\ M\^16 + 191\ M\^18 - 89\ M\^20 - 89\ M\^22 + 
                    49\ M\^24 + 271\ M\^26 - 449\ M\^28 + 400\ M\^30 + 
                    8\ M\^32 - 864\ M\^34 + 1599\ M\^36 + 156\ M\^38 - 
                    2135\ M\^40 + 1388\ M\^42 + 1525\ M\^44 - 1635\ M\^46 - 
                    198\ M\^48 + 522\ M\^50 - 120\ M\^52)\) + 
              L\^4\ \((\(-M\^14\) + 9\ M\^16 - 42\ M\^18 + 123\ M\^20 - 
                    253\ M\^22 + 364\ M\^24 - 389\ M\^26 + 373\ M\^28 - 
                    540\ M\^30 + 606\ M\^32 + 94\ M\^34 - 925\ M\^36 - 
                    180\ M\^38 + 2166\ M\^40 - 915\ M\^42 - 1239\ M\^44 + 
                    707\ M\^46 + 547\ M\^48 + 3672\ M\^50 - 3667\ M\^52 - 
                    3577\ M\^54 + 8024\ M\^56 - 250\ M\^58 - 6918\ M\^60 + 
                    2712\ M\^62 + 1693\ M\^64 - 1248\ M\^66 + 210\ M\^68)\) + 
              L\^5\ \((M\^22 - 8\ M\^24 + 36\ M\^26 - 114\ M\^28 + 
                    275\ M\^30 - 551\ M\^32 + 956\ M\^34 - 1410\ M\^36 + 
                    1424\ M\^38 - 466\ M\^40 - 697\ M\^42 - 457\ M\^44 + 
                    2820\ M\^46 - 928\ M\^48 - 4242\ M\^50 + 2510\ M\^52 + 
                    6906\ M\^54 - 6373\ M\^56 - 3662\ M\^58 + 9540\ M\^60 - 
                    1218\ M\^62 + 453\ M\^64 - 4447\ M\^66 - 1786\ M\^68 + 
                    14167\ M\^70 - 9519\ M\^72 - 11457\ M\^74 + 
                    13037\ M\^76 + 224\ M\^78 - 4586\ M\^80 + 1932\ M\^82 - 
                    252\ M\^84)\) + 
              L\^6\ \((\(-2\)\ M\^34 + 20\ M\^36 - 105\ M\^38 + 378\ M\^40 - 
                    973\ M\^42 + 1806\ M\^44 - 2403\ M\^46 + 2294\ M\^48 - 
                    1873\ M\^50 + 1343\ M\^52 + 19\ M\^54 - 1387\ M\^56 - 
                    1146\ M\^58 + 4185\ M\^60 + 1010\ M\^62 - 9301\ M\^64 + 
                    4758\ M\^66 + 15846\ M\^68 - 16682\ M\^70 - 4881\ M\^72 + 
                    27405\ M\^74 - 14280\ M\^76 - 5367\ M\^78 + 5460\ M\^80 - 
                    8173\ M\^82 + 15677\ M\^84 - 15814\ M\^86 - 
                    10668\ M\^88 + 26584\ M\^90 - 8550\ M\^92 - 7952\ M\^94 + 
                    6958\ M\^96 - 2016\ M\^98 + 210\ M\^100)\) + 
              L\^7\ \((M\^46 - 14\ M\^48 + 104\ M\^50 - 502\ M\^52 + 
                    1647\ M\^54 - 3724\ M\^56 + 5677\ M\^58 - 5090\ M\^60 + 
                    1037\ M\^62 + 2837\ M\^64 - 2939\ M\^66 - 504\ M\^68 + 
                    4611\ M\^70 - 5500\ M\^72 - 1317\ M\^74 + 9741\ M\^76 - 
                    9323\ M\^78 + 6878\ M\^80 + 16757\ M\^82 - 32443\ M\^84 + 
                    14579\ M\^86 + 42354\ M\^88 - 57106\ M\^90 - 472\ M\^92 + 
                    31122\ M\^94 - 31655\ M\^96 + 20990\ M\^98 - 
                    9345\ M\^100 - 13851\ M\^102 + 35684\ M\^104 - 
                    20387\ M\^106 - 7523\ M\^108 + 13860\ M\^110 - 
                    6608\ M\^112 + 1428\ M\^114 - 120\ M\^116)\) + 
              L\^8\ \((2\ M\^60 - 34\ M\^62 + 263\ M\^64 - 1242\ M\^66 + 
                    3943\ M\^68 - 8485\ M\^70 + 11471\ M\^72 - 6767\ M\^74 - 
                    5783\ M\^76 + 15989\ M\^78 - 14658\ M\^80 - 2061\ M\^82 + 
                    24461\ M\^84 - 23572\ M\^86 - 10263\ M\^88 + 
                    34925\ M\^90 - 14996\ M\^92 + 238\ M\^94 + 22083\ M\^96 - 
                    37589\ M\^98 + 28071\ M\^100 + 21502\ M\^102 - 
                    88564\ M\^104 + 38796\ M\^106 + 43774\ M\^108 - 
                    64991\ M\^110 + 39054\ M\^112 - 120\ M\^114 - 
                    27580\ M\^116 + 41591\ M\^118 - 26023\ M\^120 - 
                    5268\ M\^122 + 18482\ M\^124 - 12250\ M\^126 + 
                    4018\ M\^128 - 672\ M\^130 + 45\ M\^132)\) + 
              L\^9\ \((M\^74 - 32\ M\^76 + 331\ M\^78 - 1809\ M\^80 + 
                    6093\ M\^82 - 13123\ M\^84 + 16731\ M\^86 - 6820\ M\^88 - 
                    15959\ M\^90 + 33779\ M\^92 - 29824\ M\^94 - 
                    2768\ M\^96 + 48298\ M\^98 - 55296\ M\^100 - 
                    5350\ M\^102 + 68642\ M\^104 - 34956\ M\^106 - 
                    10116\ M\^108 + 23516\ M\^110 - 38418\ M\^112 + 
                    30774\ M\^114 - 9042\ M\^116 - 76684\ M\^118 + 
                    78906\ M\^120 + 26802\ M\^122 - 79219\ M\^124 + 
                    52278\ M\^126 + 1062\ M\^128 - 35775\ M\^130 + 
                    43237\ M\^132 - 24004\ M\^134 - 5398\ M\^136 + 
                    19285\ M\^138 - 14922\ M\^140 + 6253\ M\^142 - 
                    1523\ M\^144 + 198\ M\^146 - 10\ M\^148)\) + 
              L\^10\ \((\(-10\)\ M\^90 + 198\ M\^92 - 1523\ M\^94 + 
                    6253\ M\^96 - 14922\ M\^98 + 19285\ M\^100 - 
                    5398\ M\^102 - 24004\ M\^104 + 43237\ M\^106 - 
                    35775\ M\^108 + 1062\ M\^110 + 52278\ M\^112 - 
                    79219\ M\^114 + 26802\ M\^116 + 78906\ M\^118 - 
                    76684\ M\^120 - 9042\ M\^122 + 30774\ M\^124 - 
                    38418\ M\^126 + 23516\ M\^128 - 10116\ M\^130 - 
                    34956\ M\^132 + 68642\ M\^134 - 5350\ M\^136 - 
                    55296\ M\^138 + 48298\ M\^140 - 2768\ M\^142 - 
                    29824\ M\^144 + 33779\ M\^146 - 15959\ M\^148 - 
                    6820\ M\^150 + 16731\ M\^152 - 13123\ M\^154 + 
                    6093\ M\^156 - 1809\ M\^158 + 331\ M\^160 - 32\ M\^162 + 
                    M\^164)\) + 
              L\^11\ \((45\ M\^106 - 672\ M\^108 + 4018\ M\^110 - 
                    12250\ M\^112 + 18482\ M\^114 - 5268\ M\^116 - 
                    26023\ M\^118 + 41591\ M\^120 - 27580\ M\^122 - 
                    120\ M\^124 + 39054\ M\^126 - 64991\ M\^128 + 
                    43774\ M\^130 + 38796\ M\^132 - 88564\ M\^134 + 
                    21502\ M\^136 + 28071\ M\^138 - 37589\ M\^140 + 
                    22083\ M\^142 + 238\ M\^144 - 14996\ M\^146 + 
                    34925\ M\^148 - 10263\ M\^150 - 23572\ M\^152 + 
                    24461\ M\^154 - 2061\ M\^156 - 14658\ M\^158 + 
                    15989\ M\^160 - 5783\ M\^162 - 6767\ M\^164 + 
                    11471\ M\^166 - 8485\ M\^168 + 3943\ M\^170 - 
                    1242\ M\^172 + 263\ M\^174 - 34\ M\^176 + 2\ M\^178)\) + 
              L\^12\ \((\(-120\)\ M\^122 + 1428\ M\^124 - 6608\ M\^126 + 
                    13860\ M\^128 - 7523\ M\^130 - 20387\ M\^132 + 
                    35684\ M\^134 - 13851\ M\^136 - 9345\ M\^138 + 
                    20990\ M\^140 - 31655\ M\^142 + 31122\ M\^144 - 
                    472\ M\^146 - 57106\ M\^148 + 42354\ M\^150 + 
                    14579\ M\^152 - 32443\ M\^154 + 16757\ M\^156 + 
                    6878\ M\^158 - 9323\ M\^160 + 9741\ M\^162 - 
                    1317\ M\^164 - 5500\ M\^166 + 4611\ M\^168 - 
                    504\ M\^170 - 2939\ M\^172 + 2837\ M\^174 + 
                    1037\ M\^176 - 5090\ M\^178 + 5677\ M\^180 - 
                    3724\ M\^182 + 1647\ M\^184 - 502\ M\^186 + 104\ M\^188 - 
                    14\ M\^190 + M\^192)\) + 
              L\^13\ \((210\ M\^138 - 2016\ M\^140 + 6958\ M\^142 - 
                    7952\ M\^144 - 8550\ M\^146 + 26584\ M\^148 - 
                    10668\ M\^150 - 15814\ M\^152 + 15677\ M\^154 - 
                    8173\ M\^156 + 5460\ M\^158 - 5367\ M\^160 - 
                    14280\ M\^162 + 27405\ M\^164 - 4881\ M\^166 - 
                    16682\ M\^168 + 15846\ M\^170 + 4758\ M\^172 - 
                    9301\ M\^174 + 1010\ M\^176 + 4185\ M\^178 - 
                    1146\ M\^180 - 1387\ M\^182 + 19\ M\^184 + 1343\ M\^186 - 
                    1873\ M\^188 + 2294\ M\^190 - 2403\ M\^192 + 
                    1806\ M\^194 - 973\ M\^196 + 378\ M\^198 - 105\ M\^200 + 
                    20\ M\^202 - 2\ M\^204)\) + 
              L\^14\ \((\(-252\)\ M\^154 + 1932\ M\^156 - 4586\ M\^158 + 
                    224\ M\^160 + 13037\ M\^162 - 11457\ M\^164 - 
                    9519\ M\^166 + 14167\ M\^168 - 1786\ M\^170 - 
                    4447\ M\^172 + 453\ M\^174 - 1218\ M\^176 + 
                    9540\ M\^178 - 3662\ M\^180 - 6373\ M\^182 + 
                    6906\ M\^184 + 2510\ M\^186 - 4242\ M\^188 - 
                    928\ M\^190 + 2820\ M\^192 - 457\ M\^194 - 697\ M\^196 - 
                    466\ M\^198 + 1424\ M\^200 - 1410\ M\^202 + 956\ M\^204 - 
                    551\ M\^206 + 275\ M\^208 - 114\ M\^210 + 36\ M\^212 - 
                    8\ M\^214 + M\^216)\) + 
              L\^15\ \((210\ M\^170 - 1248\ M\^172 + 1693\ M\^174 + 
                    2712\ M\^176 - 6918\ M\^178 - 250\ M\^180 + 
                    8024\ M\^182 - 3577\ M\^184 - 3667\ M\^186 + 
                    3672\ M\^188 + 547\ M\^190 + 707\ M\^192 - 1239\ M\^194 - 
                    915\ M\^196 + 2166\ M\^198 - 180\ M\^200 - 925\ M\^202 + 
                    94\ M\^204 + 606\ M\^206 - 540\ M\^208 + 373\ M\^210 - 
                    389\ M\^212 + 364\ M\^214 - 253\ M\^216 + 123\ M\^218 - 
                    42\ M\^220 + 9\ M\^222 - M\^224)\) + 
              L\^16\ \((\(-120\)\ M\^186 + 522\ M\^188 - 198\ M\^190 - 
                    1635\ M\^192 + 1525\ M\^194 + 1388\ M\^196 - 
                    2135\ M\^198 + 156\ M\^200 + 1599\ M\^202 - 864\ M\^204 + 
                    8\ M\^206 + 400\ M\^208 - 449\ M\^210 + 271\ M\^212 + 
                    49\ M\^214 - 89\ M\^216 - 89\ M\^218 + 191\ M\^220 - 
                    174\ M\^222 + 103\ M\^224 - 44\ M\^226 + 12\ M\^228 - 
                    2\ M\^230)\) + 
              L\^17\ \((45\ M\^202 - 128\ M\^204 - 73\ M\^206 + 369\ M\^208 - 
                    91\ M\^210 - 241\ M\^212 + 256\ M\^214 + 24\ M\^216 - 
                    142\ M\^218 + 116\ M\^220 - 81\ M\^224 + 82\ M\^226 - 
                    48\ M\^228 + 24\ M\^230 - 12\ M\^232 + 4\ M\^234 - 
                    M\^236)\) + 
              L\^18\ \((\(-10\)\ M\^218 + 14\ M\^220 + 21\ M\^222 - 
                    21\ M\^224 + 7\ M\^226 + 8\ M\^228 - 10\ M\^230 + 
                    11\ M\^232 - 7\ M\^234 + 3\ M\^236 - M\^238)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 47]\), \(M\^14 + 
              L\^24\ M\^214 + 
              L\ \((2\ M\^8 - 7\ M\^10 + 15\ M\^12 - 16\ M\^14 + 14\ M\^16 - 
                    9\ M\^18 - 27\ M\^20 + 50\ M\^22 + 9\ M\^24 - 18\ M\^26 + 
                    7\ M\^28)\) + 
              L\^2\ \((M\^2 - 6\ M\^4 + 19\ M\^6 - 38\ M\^8 + 55\ M\^10 - 
                    57\ M\^12 + 39\ M\^14 - 51\ M\^16 + 120\ M\^18 - 
                    126\ M\^20 - 77\ M\^22 + 516\ M\^24 - 285\ M\^26 - 
                    853\ M\^28 + 752\ M\^30 + 663\ M\^32 - 661\ M\^34 + 
                    60\ M\^36 + 206\ M\^38 - 110\ M\^40 + 21\ M\^42)\) + 
              L\^3\ \((1 - 7\ M\^2 + 21\ M\^4 - 42\ M\^6 + 83\ M\^8 - 
                    196\ M\^10 + 409\ M\^12 - 625\ M\^14 + 559\ M\^16 + 
                    181\ M\^18 - 1256\ M\^20 + 1259\ M\^22 + 731\ M\^24 - 
                    3137\ M\^26 + 2504\ M\^28 - 422\ M\^30 + 453\ M\^32 + 
                    4653\ M\^34 - 10066\ M\^36 - 1380\ M\^38 + 12863\ M\^40 - 
                    4101\ M\^42 - 5287\ M\^44 + 5312\ M\^46 - 848\ M\^48 - 
                    1253\ M\^50 + 938\ M\^52 - 282\ M\^54 + 35\ M\^56)\) + 
              L\^4\ \((\(-4\)\ M\^4 + 39\ M\^6 - 177\ M\^8 + 490\ M\^10 - 
                    872\ M\^12 + 977\ M\^14 - 692\ M\^16 + 710\ M\^18 - 
                    1435\ M\^20 + 971\ M\^22 + 1090\ M\^24 + 13\ M\^26 - 
                    5354\ M\^28 + 5806\ M\^30 + 5918\ M\^32 - 12441\ M\^34 - 
                    3522\ M\^36 + 7393\ M\^38 - 1640\ M\^40 + 35960\ M\^42 - 
                    19434\ M\^44 - 80824\ M\^46 + 69951\ M\^48 + 
                    50519\ M\^50 - 76555\ M\^52 + 15122\ M\^54 + 
                    32951\ M\^56 - 27805\ M\^58 + 4840\ M\^60 + 5965\ M\^62 - 
                    5055\ M\^64 + 1916\ M\^66 - 390\ M\^68 + 35\ M\^70)\) + 
              L\^5\ \((6\ M\^8 - 76\ M\^10 + 445\ M\^12 - 1557\ M\^14 + 
                    3418\ M\^16 - 4359\ M\^18 + 1932\ M\^20 + 2098\ M\^22 - 
                    966\ M\^24 - 5618\ M\^26 + 10450\ M\^28 - 14047\ M\^30 + 
                    12789\ M\^32 + 11499\ M\^34 - 50248\ M\^36 + 
                    28851\ M\^38 + 77504\ M\^40 - 94066\ M\^42 - 
                    25917\ M\^44 + 60330\ M\^46 - 102392\ M\^48 + 
                    141175\ M\^50 + 222238\ M\^52 - 420141\ M\^54 - 
                    169714\ M\^56 + 551835\ M\^58 - 114338\ M\^60 - 
                    321515\ M\^62 + 285728\ M\^64 - 6411\ M\^66 - 
                    142354\ M\^68 + 101313\ M\^70 - 14948\ M\^72 - 
                    23009\ M\^74 + 19780\ M\^76 - 8272\ M\^78 + 2090\ M\^80 - 
                    310\ M\^82 + 21\ M\^84)\) + 
              L\^6\ \((\(-4\)\ M\^12 + 64\ M\^14 - 460\ M\^16 + 1956\ M\^18 - 
                    5356\ M\^20 + 9609\ M\^22 - 11098\ M\^24 + 8019\ M\^26 - 
                    2844\ M\^28 - 3616\ M\^30 + 4055\ M\^32 + 18699\ M\^34 - 
                    39193\ M\^36 + 12228\ M\^38 + 11983\ M\^40 + 
                    38121\ M\^42 - 48702\ M\^44 - 164103\ M\^46 + 
                    239596\ M\^48 + 237183\ M\^50 - 377706\ M\^52 - 
                    85624\ M\^54 - 91738\ M\^56 - 102889\ M\^58 + 
                    1177191\ M\^60 - 121035\ M\^62 - 1920088\ M\^64 + 
                    802702\ M\^66 + 1502903\ M\^68 - 1413068\ M\^70 - 
                    151809\ M\^72 + 1097534\ M\^74 - 700339\ M\^76 - 
                    130432\ M\^78 + 449567\ M\^80 - 265427\ M\^82 + 
                    19376\ M\^84 + 72914\ M\^86 - 57988\ M\^88 + 
                    24854\ M\^90 - 6898\ M\^92 + 1254\ M\^94 - 138\ M\^96 + 
                    7\ M\^98)\) + 
              L\^7\ \((M\^16 - 21\ M\^18 + 188\ M\^20 - 987\ M\^22 + 
                    3479\ M\^24 - 9036\ M\^26 + 18687\ M\^28 - 31159\ M\^30 + 
                    36058\ M\^32 - 14027\ M\^34 - 27397\ M\^36 + 
                    31604\ M\^38 + 27411\ M\^40 - 97135\ M\^42 + 
                    99989\ M\^44 + 38806\ M\^46 - 192572\ M\^48 + 
                    29099\ M\^50 + 434370\ M\^52 - 371397\ M\^54 - 
                    904001\ M\^56 + 1122056\ M\^58 + 1357433\ M\^60 - 
                    1361955\ M\^62 - 1572317\ M\^64 - 238377\ M\^66 + 
                    2001619\ M\^68 + 2811767\ M\^70 - 3271314\ M\^72 - 
                    3255242\ M\^74 + 4380428\ M\^76 + 672016\ M\^78 - 
                    3775143\ M\^80 + 2345585\ M\^82 + 1215062\ M\^84 - 
                    2597768\ M\^86 + 1094245\ M\^88 + 653369\ M\^90 - 
                    1021356\ M\^92 + 480696\ M\^94 + 29281\ M\^96 - 
                    182244\ M\^98 + 128988\ M\^100 - 53671\ M\^102 + 
                    15134\ M\^104 - 2957\ M\^106 + 386\ M\^108 - 30\ M\^110 + 
                    M\^112)\) + 
              L\^8\ \((M\^22 - 17\ M\^24 + 139\ M\^26 - 752\ M\^28 + 
                    3106\ M\^30 - 10260\ M\^32 + 26411\ M\^34 - 
                    50206\ M\^36 + 67053\ M\^38 - 60387\ M\^40 + 
                    30523\ M\^42 + 17249\ M\^44 - 67849\ M\^46 + 
                    76655\ M\^48 + 3848\ M\^50 - 267614\ M\^52 + 
                    534522\ M\^54 - 49724\ M\^56 - 1109157\ M\^58 + 
                    1090579\ M\^60 + 1489978\ M\^62 - 3000618\ M\^64 - 
                    2041751\ M\^66 + 5143478\ M\^68 + 3160854\ M\^70 - 
                    5036319\ M\^72 - 4953540\ M\^74 + 1245660\ M\^76 + 
                    7118768\ M\^78 + 2593873\ M\^80 - 8901457\ M\^82 - 
                    1039085\ M\^84 + 7809267\ M\^86 - 4356365\ M\^88 - 
                    3075066\ M\^90 + 6919952\ M\^92 - 2300414\ M\^94 - 
                    3740021\ M\^96 + 4109890\ M\^98 - 771451\ M\^100 - 
                    1645939\ M\^102 + 1659035\ M\^104 - 564787\ M\^106 - 
                    203890\ M\^108 + 348601\ M\^110 - 214089\ M\^112 + 
                    83435\ M\^114 - 22630\ M\^116 + 4331\ M\^118 - 
                    571\ M\^120 + 48\ M\^122 - 2\ M\^124)\) + 
              L\^9\ \((7\ M\^32 - 130\ M\^34 + 1087\ M\^36 - 5471\ M\^38 + 
                    18849\ M\^40 - 48116\ M\^42 + 94160\ M\^44 - 
                    134875\ M\^46 + 115948\ M\^48 - 12232\ M\^50 - 
                    81311\ M\^52 + 40674\ M\^54 + 59982\ M\^56 + 
                    81784\ M\^58 - 417522\ M\^60 - 126625\ M\^62 + 
                    1691223\ M\^64 - 1372133\ M\^66 - 2556775\ M\^68 + 
                    5136550\ M\^70 + 1751393\ M\^72 - 10024864\ M\^74 - 
                    1297866\ M\^76 + 13273899\ M\^78 + 4271543\ M\^80 - 
                    11330563\ M\^82 - 10131443\ M\^84 + 4996181\ M\^86 + 
                    13937791\ M\^88 - 1383097\ M\^90 - 12148376\ M\^92 + 
                    5211107\ M\^94 + 4557223\ M\^96 - 10273530\ M\^98 + 
                    4456315\ M\^100 + 8283259\ M\^102 - 8716639\ M\^104 - 
                    760760\ M\^106 + 6176724\ M\^108 - 4088248\ M\^110 - 
                    505517\ M\^112 + 2676631\ M\^114 - 1865540\ M\^116 + 
                    295943\ M\^118 + 489125\ M\^120 - 495297\ M\^122 + 
                    260104\ M\^124 - 91182\ M\^126 + 22521\ M\^128 - 
                    3913\ M\^130 + 457\ M\^132 - 32\ M\^134 + M\^136)\) + 
              L\^10\ \((21\ M\^42 - 417\ M\^44 + 3635\ M\^46 - 18412\ M\^48 + 
                    60601\ M\^50 - 137449\ M\^52 + 218442\ M\^54 - 
                    221007\ M\^56 + 54993\ M\^58 + 217383\ M\^60 - 
                    278665\ M\^62 - 103083\ M\^64 + 520185\ M\^66 + 
                    40366\ M\^68 - 1478969\ M\^70 + 835723\ M\^72 + 
                    3176253\ M\^74 - 4796368\ M\^76 - 2628645\ M\^78 + 
                    12318609\ M\^80 - 1760838\ M\^82 - 19711459\ M\^84 + 
                    3592016\ M\^86 + 22169544\ M\^88 + 3755302\ M\^90 - 
                    17355534\ M\^92 - 14796177\ M\^94 + 9320056\ M\^96 + 
                    16246811\ M\^98 - 5875809\ M\^100 - 5506510\ M\^102 + 
                    8104159\ M\^104 - 8124051\ M\^106 - 8297098\ M\^108 + 
                    14793797\ M\^110 + 1939721\ M\^112 - 11831678\ M\^114 + 
                    5224429\ M\^116 + 4025962\ M\^118 - 6317242\ M\^120 + 
                    2222621\ M\^122 + 2089556\ M\^124 - 2860424\ M\^126 + 
                    1262200\ M\^128 + 244868\ M\^130 - 704010\ M\^132 + 
                    505147\ M\^134 - 221517\ M\^136 + 66799\ M\^138 - 
                    14106\ M\^140 + 2013\ M\^142 - 175\ M\^144 + 
                    7\ M\^146)\) + 
              L\^11\ \((35\ M\^52 - 732\ M\^54 + 6678\ M\^56 - 35112\ M\^58 + 
                    118420\ M\^60 - 267723\ M\^62 + 396104\ M\^64 - 
                    293375\ M\^66 - 190847\ M\^68 + 749048\ M\^70 - 
                    605719\ M\^72 - 533879\ M\^74 + 1539374\ M\^76 - 
                    340567\ M\^78 - 2902452\ M\^80 + 3106427\ M\^82 + 
                    3197200\ M\^84 - 9399780\ M\^86 + 1596735\ M\^88 + 
                    18155709\ M\^90 - 10374444\ M\^92 - 24216756\ M\^94 + 
                    11071841\ M\^96 + 23386732\ M\^98 + 2920800\ M\^100 - 
                    15908475\ M\^102 - 16710124\ M\^104 + 6572915\ M\^106 + 
                    11360770\ M\^108 - 977887\ M\^110 + 7499285\ M\^112 - 
                    1309434\ M\^114 - 19791672\ M\^116 + 4043169\ M\^118 + 
                    17490929\ M\^120 - 7631943\ M\^122 - 7611695\ M\^124 + 
                    8429359\ M\^126 - 652311\ M\^128 - 4997123\ M\^130 + 
                    3850377\ M\^132 + 322353\ M\^134 - 2508495\ M\^136 + 
                    1836368\ M\^138 - 232546\ M\^140 - 647252\ M\^142 + 
                    648507\ M\^144 - 347075\ M\^146 + 123749\ M\^148 - 
                    30405\ M\^150 + 4960\ M\^152 - 482\ M\^154 + 
                    21\ M\^156)\) + 
              L\^12\ \((35\ M\^62 - 765\ M\^64 + 7349\ M\^66 - 41111\ M\^68 + 
                    149269\ M\^70 - 366043\ M\^72 + 582758\ M\^74 - 
                    437127\ M\^76 - 410949\ M\^78 + 1495634\ M\^80 - 
                    1386172\ M\^82 - 717426\ M\^84 + 3104956\ M\^86 - 
                    2046789\ M\^88 - 3305051\ M\^90 + 6576759\ M\^92 - 
                    586718\ M\^94 - 11836683\ M\^96 + 10373970\ M\^98 + 
                    15999702\ M\^100 - 19539352\ M\^102 - 16786039\ M\^104 + 
                    14481451\ M\^106 + 13004523\ M\^108 + 5060995\ M\^110 - 
                    5029391\ M\^112 - 16813142\ M\^114 - 5029391\ M\^116 + 
                    5060995\ M\^118 + 13004523\ M\^120 + 14481451\ M\^122 - 
                    16786039\ M\^124 - 19539352\ M\^126 + 15999702\ M\^128 + 
                    10373970\ M\^130 - 11836683\ M\^132 - 586718\ M\^134 + 
                    6576759\ M\^136 - 3305051\ M\^138 - 2046789\ M\^140 + 
                    3104956\ M\^142 - 717426\ M\^144 - 1386172\ M\^146 + 
                    1495634\ M\^148 - 410949\ M\^150 - 437127\ M\^152 + 
                    582758\ M\^154 - 366043\ M\^156 + 149269\ M\^158 - 
                    41111\ M\^160 + 7349\ M\^162 - 765\ M\^164 + 
                    35\ M\^166)\) + 
              L\^13\ \((21\ M\^72 - 482\ M\^74 + 4960\ M\^76 - 30405\ M\^78 + 
                    123749\ M\^80 - 347075\ M\^82 + 648507\ M\^84 - 
                    647252\ M\^86 - 232546\ M\^88 + 1836368\ M\^90 - 
                    2508495\ M\^92 + 322353\ M\^94 + 3850377\ M\^96 - 
                    4997123\ M\^98 - 652311\ M\^100 + 8429359\ M\^102 - 
                    7611695\ M\^104 - 7631943\ M\^106 + 17490929\ M\^108 + 
                    4043169\ M\^110 - 19791672\ M\^112 - 1309434\ M\^114 + 
                    7499285\ M\^116 - 977887\ M\^118 + 11360770\ M\^120 + 
                    6572915\ M\^122 - 16710124\ M\^124 - 15908475\ M\^126 + 
                    2920800\ M\^128 + 23386732\ M\^130 + 11071841\ M\^132 - 
                    24216756\ M\^134 - 10374444\ M\^136 + 18155709\ M\^138 + 
                    1596735\ M\^140 - 9399780\ M\^142 + 3197200\ M\^144 + 
                    3106427\ M\^146 - 2902452\ M\^148 - 340567\ M\^150 + 
                    1539374\ M\^152 - 533879\ M\^154 - 605719\ M\^156 + 
                    749048\ M\^158 - 190847\ M\^160 - 293375\ M\^162 + 
                    396104\ M\^164 - 267723\ M\^166 + 118420\ M\^168 - 
                    35112\ M\^170 + 6678\ M\^172 - 732\ M\^174 + 
                    35\ M\^176)\) + 
              L\^14\ \((7\ M\^82 - 175\ M\^84 + 2013\ M\^86 - 14106\ M\^88 + 
                    66799\ M\^90 - 221517\ M\^92 + 505147\ M\^94 - 
                    704010\ M\^96 + 244868\ M\^98 + 1262200\ M\^100 - 
                    2860424\ M\^102 + 2089556\ M\^104 + 2222621\ M\^106 - 
                    6317242\ M\^108 + 4025962\ M\^110 + 5224429\ M\^112 - 
                    11831678\ M\^114 + 1939721\ M\^116 + 14793797\ M\^118 - 
                    8297098\ M\^120 - 8124051\ M\^122 + 8104159\ M\^124 - 
                    5506510\ M\^126 - 5875809\ M\^128 + 16246811\ M\^130 + 
                    9320056\ M\^132 - 14796177\ M\^134 - 17355534\ M\^136 + 
                    3755302\ M\^138 + 22169544\ M\^140 + 3592016\ M\^142 - 
                    19711459\ M\^144 - 1760838\ M\^146 + 12318609\ M\^148 - 
                    2628645\ M\^150 - 4796368\ M\^152 + 3176253\ M\^154 + 
                    835723\ M\^156 - 1478969\ M\^158 + 40366\ M\^160 + 
                    520185\ M\^162 - 103083\ M\^164 - 278665\ M\^166 + 
                    217383\ M\^168 + 54993\ M\^170 - 221007\ M\^172 + 
                    218442\ M\^174 - 137449\ M\^176 + 60601\ M\^178 - 
                    18412\ M\^180 + 3635\ M\^182 - 417\ M\^184 + 
                    21\ M\^186)\) + 
              L\^15\ \((M\^92 - 32\ M\^94 + 457\ M\^96 - 3913\ M\^98 + 
                    22521\ M\^100 - 91182\ M\^102 + 260104\ M\^104 - 
                    495297\ M\^106 + 489125\ M\^108 + 295943\ M\^110 - 
                    1865540\ M\^112 + 2676631\ M\^114 - 505517\ M\^116 - 
                    4088248\ M\^118 + 6176724\ M\^120 - 760760\ M\^122 - 
                    8716639\ M\^124 + 8283259\ M\^126 + 4456315\ M\^128 - 
                    10273530\ M\^130 + 4557223\ M\^132 + 5211107\ M\^134 - 
                    12148376\ M\^136 - 1383097\ M\^138 + 13937791\ M\^140 + 
                    4996181\ M\^142 - 10131443\ M\^144 - 11330563\ M\^146 + 
                    4271543\ M\^148 + 13273899\ M\^150 - 1297866\ M\^152 - 
                    10024864\ M\^154 + 1751393\ M\^156 + 5136550\ M\^158 - 
                    2556775\ M\^160 - 1372133\ M\^162 + 1691223\ M\^164 - 
                    126625\ M\^166 - 417522\ M\^168 + 81784\ M\^170 + 
                    59982\ M\^172 + 40674\ M\^174 - 81311\ M\^176 - 
                    12232\ M\^178 + 115948\ M\^180 - 134875\ M\^182 + 
                    94160\ M\^184 - 48116\ M\^186 + 18849\ M\^188 - 
                    5471\ M\^190 + 1087\ M\^192 - 130\ M\^194 + 
                    7\ M\^196)\) + 
              L\^16\ \((\(-2\)\ M\^104 + 48\ M\^106 - 571\ M\^108 + 
                    4331\ M\^110 - 22630\ M\^112 + 83435\ M\^114 - 
                    214089\ M\^116 + 348601\ M\^118 - 203890\ M\^120 - 
                    564787\ M\^122 + 1659035\ M\^124 - 1645939\ M\^126 - 
                    771451\ M\^128 + 4109890\ M\^130 - 3740021\ M\^132 - 
                    2300414\ M\^134 + 6919952\ M\^136 - 3075066\ M\^138 - 
                    4356365\ M\^140 + 7809267\ M\^142 - 1039085\ M\^144 - 
                    8901457\ M\^146 + 2593873\ M\^148 + 7118768\ M\^150 + 
                    1245660\ M\^152 - 4953540\ M\^154 - 5036319\ M\^156 + 
                    3160854\ M\^158 + 5143478\ M\^160 - 2041751\ M\^162 - 
                    3000618\ M\^164 + 1489978\ M\^166 + 1090579\ M\^168 - 
                    1109157\ M\^170 - 49724\ M\^172 + 534522\ M\^174 - 
                    267614\ M\^176 + 3848\ M\^178 + 76655\ M\^180 - 
                    67849\ M\^182 + 17249\ M\^184 + 30523\ M\^186 - 
                    60387\ M\^188 + 67053\ M\^190 - 50206\ M\^192 + 
                    26411\ M\^194 - 10260\ M\^196 + 3106\ M\^198 - 
                    752\ M\^200 + 139\ M\^202 - 17\ M\^204 + M\^206)\) + 
              L\^17\ \((M\^116 - 30\ M\^118 + 386\ M\^120 - 2957\ M\^122 + 
                    15134\ M\^124 - 53671\ M\^126 + 128988\ M\^128 - 
                    182244\ M\^130 + 29281\ M\^132 + 480696\ M\^134 - 
                    1021356\ M\^136 + 653369\ M\^138 + 1094245\ M\^140 - 
                    2597768\ M\^142 + 1215062\ M\^144 + 2345585\ M\^146 - 
                    3775143\ M\^148 + 672016\ M\^150 + 4380428\ M\^152 - 
                    3255242\ M\^154 - 3271314\ M\^156 + 2811767\ M\^158 + 
                    2001619\ M\^160 - 238377\ M\^162 - 1572317\ M\^164 - 
                    1361955\ M\^166 + 1357433\ M\^168 + 1122056\ M\^170 - 
                    904001\ M\^172 - 371397\ M\^174 + 434370\ M\^176 + 
                    29099\ M\^178 - 192572\ M\^180 + 38806\ M\^182 + 
                    99989\ M\^184 - 97135\ M\^186 + 27411\ M\^188 + 
                    31604\ M\^190 - 27397\ M\^192 - 14027\ M\^194 + 
                    36058\ M\^196 - 31159\ M\^198 + 18687\ M\^200 - 
                    9036\ M\^202 + 3479\ M\^204 - 987\ M\^206 + 188\ M\^208 - 
                    21\ M\^210 + M\^212)\) + 
              L\^18\ \((7\ M\^130 - 138\ M\^132 + 1254\ M\^134 - 
                    6898\ M\^136 + 24854\ M\^138 - 57988\ M\^140 + 
                    72914\ M\^142 + 19376\ M\^144 - 265427\ M\^146 + 
                    449567\ M\^148 - 130432\ M\^150 - 700339\ M\^152 + 
                    1097534\ M\^154 - 151809\ M\^156 - 1413068\ M\^158 + 
                    1502903\ M\^160 + 802702\ M\^162 - 1920088\ M\^164 - 
                    121035\ M\^166 + 1177191\ M\^168 - 102889\ M\^170 - 
                    91738\ M\^172 - 85624\ M\^174 - 377706\ M\^176 + 
                    237183\ M\^178 + 239596\ M\^180 - 164103\ M\^182 - 
                    48702\ M\^184 + 38121\ M\^186 + 11983\ M\^188 + 
                    12228\ M\^190 - 39193\ M\^192 + 18699\ M\^194 + 
                    4055\ M\^196 - 3616\ M\^198 - 2844\ M\^200 + 
                    8019\ M\^202 - 11098\ M\^204 + 9609\ M\^206 - 
                    5356\ M\^208 + 1956\ M\^210 - 460\ M\^212 + 64\ M\^214 - 
                    4\ M\^216)\) + 
              L\^23\ \((7\ M\^200 - 18\ M\^202 + 9\ M\^204 + 50\ M\^206 - 
                    27\ M\^208 - 9\ M\^210 + 14\ M\^212 - 16\ M\^214 + 
                    15\ M\^216 - 7\ M\^218 + 2\ M\^220)\) + 
              L\^19\ \((21\ M\^144 - 310\ M\^146 + 2090\ M\^148 - 
                    8272\ M\^150 + 19780\ M\^152 - 23009\ M\^154 - 
                    14948\ M\^156 + 101313\ M\^158 - 142354\ M\^160 - 
                    6411\ M\^162 + 285728\ M\^164 - 321515\ M\^166 - 
                    114338\ M\^168 + 551835\ M\^170 - 169714\ M\^172 - 
                    420141\ M\^174 + 222238\ M\^176 + 141175\ M\^178 - 
                    102392\ M\^180 + 60330\ M\^182 - 25917\ M\^184 - 
                    94066\ M\^186 + 77504\ M\^188 + 28851\ M\^190 - 
                    50248\ M\^192 + 11499\ M\^194 + 12789\ M\^196 - 
                    14047\ M\^198 + 10450\ M\^200 - 5618\ M\^202 - 
                    966\ M\^204 + 2098\ M\^206 + 1932\ M\^208 - 
                    4359\ M\^210 + 3418\ M\^212 - 1557\ M\^214 + 
                    445\ M\^216 - 76\ M\^218 + 6\ M\^220)\) + 
              L\^20\ \((35\ M\^158 - 390\ M\^160 + 1916\ M\^162 - 
                    5055\ M\^164 + 5965\ M\^166 + 4840\ M\^168 - 
                    27805\ M\^170 + 32951\ M\^172 + 15122\ M\^174 - 
                    76555\ M\^176 + 50519\ M\^178 + 69951\ M\^180 - 
                    80824\ M\^182 - 19434\ M\^184 + 35960\ M\^186 - 
                    1640\ M\^188 + 7393\ M\^190 - 3522\ M\^192 - 
                    12441\ M\^194 + 5918\ M\^196 + 5806\ M\^198 - 
                    5354\ M\^200 + 13\ M\^202 + 1090\ M\^204 + 971\ M\^206 - 
                    1435\ M\^208 + 710\ M\^210 - 692\ M\^212 + 977\ M\^214 - 
                    872\ M\^216 + 490\ M\^218 - 177\ M\^220 + 39\ M\^222 - 
                    4\ M\^224)\) + 
              L\^22\ \((21\ M\^186 - 110\ M\^188 + 206\ M\^190 + 60\ M\^192 - 
                    661\ M\^194 + 663\ M\^196 + 752\ M\^198 - 853\ M\^200 - 
                    285\ M\^202 + 516\ M\^204 - 77\ M\^206 - 126\ M\^208 + 
                    120\ M\^210 - 51\ M\^212 + 39\ M\^214 - 57\ M\^216 + 
                    55\ M\^218 - 38\ M\^220 + 19\ M\^222 - 6\ M\^224 + 
                    M\^226)\) + 
              L\^21\ \((35\ M\^172 - 282\ M\^174 + 938\ M\^176 - 
                    1253\ M\^178 - 848\ M\^180 + 5312\ M\^182 - 
                    5287\ M\^184 - 4101\ M\^186 + 12863\ M\^188 - 
                    1380\ M\^190 - 10066\ M\^192 + 4653\ M\^194 + 
                    453\ M\^196 - 422\ M\^198 + 2504\ M\^200 - 3137\ M\^202 + 
                    731\ M\^204 + 1259\ M\^206 - 1256\ M\^208 + 181\ M\^210 + 
                    559\ M\^212 - 625\ M\^214 + 409\ M\^216 - 196\ M\^218 + 
                    83\ M\^220 - 42\ M\^222 + 21\ M\^224 - 7\ M\^226 + 
                    M\^228)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 48]\), \(M\^70 + L\^28\ M\^74 + 
              L\ \((4\ M\^60 - 15\ M\^62 + 36\ M\^64 - 41\ M\^66 - 7\ M\^68 + 
                    64\ M\^70 - 24\ M\^72 + 3\ M\^74 + 11\ M\^76 - 
                    11\ M\^78 + 4\ M\^80)\) + 
              L\^27\ \((4\ M\^64 - 11\ M\^66 + 11\ M\^68 + 3\ M\^70 - 
                    24\ M\^72 + 64\ M\^74 - 7\ M\^76 - 41\ M\^78 + 
                    36\ M\^80 - 15\ M\^82 + 4\ M\^84)\) + 
              L\^2\ \((6\ M\^50 - 42\ M\^52 + 161\ M\^54 - 381\ M\^56 + 
                    527\ M\^58 - 247\ M\^60 - 572\ M\^62 + 1358\ M\^64 - 
                    763\ M\^66 - 1076\ M\^68 + 1598\ M\^70 - 105\ M\^72 - 
                    468\ M\^74 + 489\ M\^76 - 354\ M\^78 + 146\ M\^80 + 
                    52\ M\^82 - 110\ M\^84 + 83\ M\^86 - 34\ M\^88 + 
                    6\ M\^90)\) + 
              L\^26\ \((6\ M\^54 - 34\ M\^56 + 83\ M\^58 - 110\ M\^60 + 
                    52\ M\^62 + 146\ M\^64 - 354\ M\^66 + 489\ M\^68 - 
                    468\ M\^70 - 105\ M\^72 + 1598\ M\^74 - 1076\ M\^76 - 
                    763\ M\^78 + 1358\ M\^80 - 572\ M\^82 - 247\ M\^84 + 
                    527\ M\^86 - 381\ M\^88 + 161\ M\^90 - 42\ M\^92 + 
                    6\ M\^94)\) + 
              L\^3\ \((4\ M\^40 - 39\ M\^42 + 198\ M\^44 - 654\ M\^46 + 
                    1502\ M\^48 - 2473\ M\^50 + 2764\ M\^52 - 1119\ M\^54 - 
                    3358\ M\^56 + 9274\ M\^58 - 9707\ M\^60 - 4778\ M\^62 + 
                    24870\ M\^64 - 17411\ M\^66 - 17230\ M\^68 + 
                    29565\ M\^70 - 5795\ M\^72 - 13271\ M\^74 + 
                    15136\ M\^76 - 4736\ M\^78 - 4573\ M\^80 + 6220\ M\^82 - 
                    2985\ M\^84 + 429\ M\^86 + 477\ M\^88 - 670\ M\^90 + 
                    572\ M\^92 - 349\ M\^94 + 145\ M\^96 - 36\ M\^98 + 
                    4\ M\^100)\) + 
              L\^25\ \((4\ M\^44 - 36\ M\^46 + 145\ M\^48 - 349\ M\^50 + 
                    572\ M\^52 - 670\ M\^54 + 477\ M\^56 + 429\ M\^58 - 
                    2985\ M\^60 + 6220\ M\^62 - 4573\ M\^64 - 4736\ M\^66 + 
                    15136\ M\^68 - 13271\ M\^70 - 5795\ M\^72 + 
                    29565\ M\^74 - 17230\ M\^76 - 17411\ M\^78 + 
                    24870\ M\^80 - 4778\ M\^82 - 9707\ M\^84 + 9274\ M\^86 - 
                    3358\ M\^88 - 1119\ M\^90 + 2764\ M\^92 - 2473\ M\^94 + 
                    1502\ M\^96 - 654\ M\^98 + 198\ M\^100 - 39\ M\^102 + 
                    4\ M\^104)\) + 
              L\^4\ \((M\^30 - 12\ M\^32 + 76\ M\^34 - 335\ M\^36 + 
                    1141\ M\^38 - 3186\ M\^40 + 7240\ M\^42 - 12384\ M\^44 + 
                    12977\ M\^46 + 257\ M\^48 - 27228\ M\^50 + 42942\ M\^52 - 
                    13808\ M\^54 - 58267\ M\^56 + 115950\ M\^58 - 
                    58518\ M\^60 - 146083\ M\^62 + 260562\ M\^64 - 
                    9198\ M\^66 - 321351\ M\^68 + 218443\ M\^70 + 
                    139037\ M\^72 - 229975\ M\^74 + 75527\ M\^76 + 
                    66534\ M\^78 - 83820\ M\^80 + 41629\ M\^82 - 
                    1409\ M\^84 - 19528\ M\^86 + 24006\ M\^88 - 
                    14968\ M\^90 + 2385\ M\^92 + 3901\ M\^94 - 4177\ M\^96 + 
                    2686\ M\^98 - 1491\ M\^100 + 751\ M\^102 - 304\ M\^104 + 
                    85\ M\^106 - 14\ M\^108 + M\^110)\) + 
              L\^24\ \((M\^34 - 14\ M\^36 + 85\ M\^38 - 304\ M\^40 + 
                    751\ M\^42 - 1491\ M\^44 + 2686\ M\^46 - 4177\ M\^48 + 
                    3901\ M\^50 + 2385\ M\^52 - 14968\ M\^54 + 24006\ M\^56 - 
                    19528\ M\^58 - 1409\ M\^60 + 41629\ M\^62 - 
                    83820\ M\^64 + 66534\ M\^66 + 75527\ M\^68 - 
                    229975\ M\^70 + 139037\ M\^72 + 218443\ M\^74 - 
                    321351\ M\^76 - 9198\ M\^78 + 260562\ M\^80 - 
                    146083\ M\^82 - 58518\ M\^84 + 115950\ M\^86 - 
                    58267\ M\^88 - 13808\ M\^90 + 42942\ M\^92 - 
                    27228\ M\^94 + 257\ M\^96 + 12977\ M\^98 - 
                    12384\ M\^100 + 7240\ M\^102 - 3186\ M\^104 + 
                    1141\ M\^106 - 335\ M\^108 + 76\ M\^110 - 12\ M\^112 + 
                    M\^114)\) + 
              L\^5\ \((2\ M\^24 - 31\ M\^26 + 231\ M\^28 - 1116\ M\^30 + 
                    3872\ M\^32 - 10157\ M\^34 + 20843\ M\^36 - 
                    33567\ M\^38 + 38583\ M\^40 - 14368\ M\^42 - 
                    59628\ M\^44 + 140828\ M\^46 - 73661\ M\^48 - 
                    238753\ M\^50 + 502643\ M\^52 - 146642\ M\^54 - 
                    837014\ M\^56 + 1221106\ M\^58 + 289619\ M\^60 - 
                    2292868\ M\^62 + 1309132\ M\^64 + 2175788\ M\^66 - 
                    2783020\ M\^68 - 679556\ M\^70 + 2694013\ M\^72 - 
                    841205\ M\^74 - 1167945\ M\^76 + 1104210\ M\^78 - 
                    151160\ M\^80 - 342416\ M\^82 + 332882\ M\^84 - 
                    138016\ M\^86 - 18182\ M\^88 + 68998\ M\^90 - 
                    56813\ M\^92 + 37235\ M\^94 - 24172\ M\^96 + 
                    7481\ M\^98 + 8860\ M\^100 - 14640\ M\^102 + 
                    10990\ M\^104 - 5502\ M\^106 + 2180\ M\^108 - 
                    787\ M\^110 + 261\ M\^112 - 69\ M\^114 + 12\ M\^116 - 
                    M\^118)\) + 
              L\^23\ \((\(-M\^26\) + 12\ M\^28 - 69\ M\^30 + 261\ M\^32 - 
                    787\ M\^34 + 2180\ M\^36 - 5502\ M\^38 + 10990\ M\^40 - 
                    14640\ M\^42 + 8860\ M\^44 + 7481\ M\^46 - 24172\ M\^48 + 
                    37235\ M\^50 - 56813\ M\^52 + 68998\ M\^54 - 
                    18182\ M\^56 - 138016\ M\^58 + 332882\ M\^60 - 
                    342416\ M\^62 - 151160\ M\^64 + 1104210\ M\^66 - 
                    1167945\ M\^68 - 841205\ M\^70 + 2694013\ M\^72 - 
                    679556\ M\^74 - 2783020\ M\^76 + 2175788\ M\^78 + 
                    1309132\ M\^80 - 2292868\ M\^82 + 289619\ M\^84 + 
                    1221106\ M\^86 - 837014\ M\^88 - 146642\ M\^90 + 
                    502643\ M\^92 - 238753\ M\^94 - 73661\ M\^96 + 
                    140828\ M\^98 - 59628\ M\^100 - 14368\ M\^102 + 
                    38583\ M\^104 - 33567\ M\^106 + 20843\ M\^108 - 
                    10157\ M\^110 + 3872\ M\^112 - 1116\ M\^114 + 
                    231\ M\^116 - 31\ M\^118 + 2\ M\^120)\) + 
              L\^6\ \((M\^18 - 23\ M\^20 + 237\ M\^22 - 1503\ M\^24 + 
                    6551\ M\^26 - 20426\ M\^28 + 45884\ M\^30 - 
                    72172\ M\^32 + 70998\ M\^34 - 21921\ M\^36 - 
                    50360\ M\^38 + 83325\ M\^40 - 8111\ M\^42 - 
                    218476\ M\^44 + 431748\ M\^46 - 115306\ M\^48 - 
                    959745\ M\^50 + 1509299\ M\^52 + 676105\ M\^54 - 
                    4220723\ M\^56 + 2046792\ M\^58 + 6965480\ M\^60 - 
                    7938613\ M\^62 - 6810751\ M\^64 + 14386365\ M\^66 + 
                    2100603\ M\^68 - 16352908\ M\^70 + 4362675\ M\^72 + 
                    11682543\ M\^74 - 7402602\ M\^76 - 4267079\ M\^78 + 
                    5661044\ M\^80 - 237738\ M\^82 - 2211651\ M\^84 + 
                    847207\ M\^86 + 561403\ M\^88 - 394285\ M\^90 - 
                    280624\ M\^92 + 358964\ M\^94 + 16425\ M\^96 - 
                    254207\ M\^98 + 187815\ M\^100 - 32112\ M\^102 - 
                    45302\ M\^104 + 50306\ M\^106 - 35325\ M\^108 + 
                    21529\ M\^110 - 11080\ M\^112 + 4459\ M\^114 - 
                    1368\ M\^116 + 330\ M\^118 - 67\ M\^120 + 11\ M\^122 - 
                    M\^124)\) + 
              L\^7\ \((\(-4\)\ M\^14 + 83\ M\^16 - 807\ M\^18 + 4881\ M\^20 - 
                    20202\ M\^22 + 58932\ M\^24 - 119360\ M\^26 + 
                    153803\ M\^28 - 86302\ M\^30 - 53599\ M\^32 + 
                    37194\ M\^34 + 292553\ M\^36 - 574918\ M\^38 + 
                    135216\ M\^40 + 998530\ M\^42 - 1494456\ M\^44 - 
                    149735\ M\^46 + 3387460\ M\^48 - 4487588\ M\^50 - 
                    1750460\ M\^52 + 12907479\ M\^54 - 9170772\ M\^56 - 
                    19932489\ M\^58 + 30515992\ M\^60 + 18233317\ M\^62 - 
                    54881546\ M\^64 - 4702776\ M\^66 + 69254903\ M\^68 - 
                    15296023\ M\^70 - 64676320\ M\^72 + 30461039\ M\^74 + 
                    44211583\ M\^76 - 32597776\ M\^78 - 20758533\ M\^80 + 
                    24621952\ M\^82 + 5290328\ M\^84 - 14094177\ M\^86 + 
                    1155335\ M\^88 + 6706535\ M\^90 - 2886587\ M\^92 - 
                    2152785\ M\^94 + 2345729\ M\^96 - 269977\ M\^98 - 
                    754289\ M\^100 + 565060\ M\^102 - 216386\ M\^104 + 
                    59514\ M\^106 + 36112\ M\^108 - 105343\ M\^110 + 
                    110502\ M\^112 - 72072\ M\^114 + 33832\ M\^116 - 
                    12429\ M\^118 + 3699\ M\^120 - 855\ M\^122 + 
                    133\ M\^124 - 10\ M\^126)\) + 
              L\^22\ \((\(-M\^20\) + 11\ M\^22 - 67\ M\^24 + 330\ M\^26 - 
                    1368\ M\^28 + 4459\ M\^30 - 11080\ M\^32 + 21529\ M\^34 - 
                    35325\ M\^36 + 50306\ M\^38 - 45302\ M\^40 - 
                    32112\ M\^42 + 187815\ M\^44 - 254207\ M\^46 + 
                    16425\ M\^48 + 358964\ M\^50 - 280624\ M\^52 - 
                    394285\ M\^54 + 561403\ M\^56 + 847207\ M\^58 - 
                    2211651\ M\^60 - 237738\ M\^62 + 5661044\ M\^64 - 
                    4267079\ M\^66 - 7402602\ M\^68 + 11682543\ M\^70 + 
                    4362675\ M\^72 - 16352908\ M\^74 + 2100603\ M\^76 + 
                    14386365\ M\^78 - 6810751\ M\^80 - 7938613\ M\^82 + 
                    6965480\ M\^84 + 2046792\ M\^86 - 4220723\ M\^88 + 
                    676105\ M\^90 + 1509299\ M\^92 - 959745\ M\^94 - 
                    115306\ M\^96 + 431748\ M\^98 - 218476\ M\^100 - 
                    8111\ M\^102 + 83325\ M\^104 - 50360\ M\^106 - 
                    21921\ M\^108 + 70998\ M\^110 - 72172\ M\^112 + 
                    45884\ M\^114 - 20426\ M\^116 + 6551\ M\^118 - 
                    1503\ M\^120 + 237\ M\^122 - 23\ M\^124 + M\^126)\) + 
              L\^8\ \((6\ M\^10 - 132\ M\^12 + 1342\ M\^14 - 8328\ M\^16 + 
                    34739\ M\^18 - 100540\ M\^20 + 199518\ M\^22 - 
                    253739\ M\^24 + 173991\ M\^26 - 75012\ M\^28 + 
                    256823\ M\^30 - 583206\ M\^32 + 228895\ M\^34 + 
                    889588\ M\^36 - 899098\ M\^38 - 1443354\ M\^40 + 
                    3090067\ M\^42 - 46311\ M\^44 - 5296888\ M\^46 + 
                    6271331\ M\^48 + 2126775\ M\^50 - 19438153\ M\^52 + 
                    18716213\ M\^54 + 33824775\ M\^56 - 68569295\ M\^58 - 
                    31651703\ M\^60 + 137656837\ M\^62 - 342213\ M\^64 - 
                    192900358\ M\^66 + 50989857\ M\^68 + 204246787\ M\^70 - 
                    91533279\ M\^72 - 171530634\ M\^74 + 102679339\ M\^76 + 
                    117940413\ M\^78 - 87733286\ M\^80 - 66851646\ M\^82 + 
                    63973940\ M\^84 + 29080497\ M\^86 - 40393008\ M\^88 - 
                    6242776\ M\^90 + 20952908\ M\^92 - 3180228\ M\^94 - 
                    7270027\ M\^96 + 3205441\ M\^98 + 1059456\ M\^100 - 
                    578989\ M\^102 - 361948\ M\^104 - 201270\ M\^106 + 
                    751475\ M\^108 - 457553\ M\^110 - 77696\ M\^112 + 
                    314785\ M\^114 - 276576\ M\^116 + 159046\ M\^118 - 
                    68345\ M\^120 + 21962\ M\^122 - 4960\ M\^124 + 
                    695\ M\^126 - 45\ M\^128)\) + 
              L\^9\ \((\(-4\)\ M\^6 + 98\ M\^8 - 1087\ M\^10 + 7246\ M\^12 - 
                    32263\ M\^14 + 100691\ M\^16 - 225720\ M\^18 + 
                    372063\ M\^20 - 476556\ M\^22 + 516697\ M\^24 - 
                    416322\ M\^26 - 36141\ M\^28 + 501392\ M\^30 + 
                    403768\ M\^32 - 2628600\ M\^34 + 2039712\ M\^36 + 
                    3817418\ M\^38 - 6882614\ M\^40 - 1612364\ M\^42 + 
                    11760104\ M\^44 - 9349074\ M\^46 - 2080167\ M\^48 + 
                    25903310\ M\^50 - 45380155\ M\^52 - 28463138\ M\^54 + 
                    144228481\ M\^56 - 10037986\ M\^58 - 279644097\ M\^60 + 
                    105058543\ M\^62 + 397701794\ M\^64 - 232576628\ M\^66 - 
                    435085177\ M\^68 + 326906551\ M\^70 + 374666514\ M\^72 - 
                    331762563\ M\^74 - 266270150\ M\^76 + 254388959\ M\^78 + 
                    171272183\ M\^80 - 152547638\ M\^82 - 110961984\ M\^84 + 
                    80371169\ M\^86 + 72267322\ M\^88 - 43374130\ M\^90 - 
                    42424114\ M\^92 + 26134573\ M\^94 + 20948638\ M\^96 - 
                    16379540\ M\^98 - 8463397\ M\^100 + 10434719\ M\^102 + 
                    1524519\ M\^104 - 5952943\ M\^106 + 1850870\ M\^108 + 
                    1981879\ M\^110 - 2172636\ M\^112 + 657621\ M\^114 + 
                    481447\ M\^116 - 750489\ M\^118 + 528503\ M\^120 - 
                    244696\ M\^122 + 77901\ M\^124 - 16481\ M\^126 + 
                    2089\ M\^128 - 120\ M\^130)\) + 
              L\^21\ \((\(-10\)\ M\^18 + 133\ M\^20 - 855\ M\^22 + 
                    3699\ M\^24 - 12429\ M\^26 + 33832\ M\^28 - 
                    72072\ M\^30 + 110502\ M\^32 - 105343\ M\^34 + 
                    36112\ M\^36 + 59514\ M\^38 - 216386\ M\^40 + 
                    565060\ M\^42 - 754289\ M\^44 - 269977\ M\^46 + 
                    2345729\ M\^48 - 2152785\ M\^50 - 2886587\ M\^52 + 
                    6706535\ M\^54 + 1155335\ M\^56 - 14094177\ M\^58 + 
                    5290328\ M\^60 + 24621952\ M\^62 - 20758533\ M\^64 - 
                    32597776\ M\^66 + 44211583\ M\^68 + 30461039\ M\^70 - 
                    64676320\ M\^72 - 15296023\ M\^74 + 69254903\ M\^76 - 
                    4702776\ M\^78 - 54881546\ M\^80 + 18233317\ M\^82 + 
                    30515992\ M\^84 - 19932489\ M\^86 - 9170772\ M\^88 + 
                    12907479\ M\^90 - 1750460\ M\^92 - 4487588\ M\^94 + 
                    3387460\ M\^96 - 149735\ M\^98 - 1494456\ M\^100 + 
                    998530\ M\^102 + 135216\ M\^104 - 574918\ M\^106 + 
                    292553\ M\^108 + 37194\ M\^110 - 53599\ M\^112 - 
                    86302\ M\^114 + 153803\ M\^116 - 119360\ M\^118 + 
                    58932\ M\^120 - 20202\ M\^122 + 4881\ M\^124 - 
                    807\ M\^126 + 83\ M\^128 - 4\ M\^130)\) + 
              L\^10\ \((M\^2 - 29\ M\^4 + 370\ M\^6 - 2810\ M\^8 + 
                    14434\ M\^10 - 54232\ M\^12 + 158266\ M\^14 - 
                    373084\ M\^16 + 703300\ M\^18 - 959709\ M\^20 + 
                    695365\ M\^22 + 133859\ M\^24 - 163862\ M\^26 - 
                    1754763\ M\^28 + 2446793\ M\^30 + 2422157\ M\^32 - 
                    6099127\ M\^34 - 4358984\ M\^36 + 17279998\ M\^38 - 
                    1147657\ M\^40 - 30682452\ M\^42 + 22312433\ M\^44 + 
                    22785974\ M\^46 - 55115898\ M\^48 + 54153889\ M\^50 + 
                    59432851\ M\^52 - 244760259\ M\^54 + 45987903\ M\^56 + 
                    522082857\ M\^58 - 315163986\ M\^60 - 766576785\ M\^62 + 
                    692451768\ M\^64 + 835776152\ M\^66 - 1008623339\ M\^68 - 
                    681733838\ M\^70 + 1087516865\ M\^72 + 396391289\ M\^74 - 
                    884022466\ M\^76 - 146472442\ M\^78 + 518497604\ M\^80 + 
                    36717098\ M\^82 - 175597279\ M\^84 - 53129920\ M\^86 - 
                    14660300\ M\^88 + 109405969\ M\^90 + 50126919\ M\^92 - 
                    129746498\ M\^94 - 7778633\ M\^96 + 97567468\ M\^98 - 
                    32032369\ M\^100 - 43305092\ M\^102 + 37167795\ M\^104 + 
                    2965363\ M\^106 - 19209929\ M\^108 + 9779985\ M\^110 + 
                    2557452\ M\^112 - 6019559\ M\^114 + 3051742\ M\^116 + 
                    400361\ M\^118 - 1653439\ M\^120 + 1279734\ M\^122 - 
                    586464\ M\^124 + 176994\ M\^126 - 34710\ M\^128 + 
                    4026\ M\^130 - 210\ M\^132)\) + 
              L\^11\ \((1 - 25\ M\^2 + 296\ M\^4 - 2249\ M\^6 + 12482\ M\^8 - 
                    53667\ M\^10 + 180897\ M\^12 - 470318\ M\^14 + 
                    917981\ M\^16 - 1310111\ M\^18 + 1330067\ M\^20 - 
                    800262\ M\^22 - 454209\ M\^24 + 1921633\ M\^26 - 
                    101293\ M\^28 - 7557190\ M\^30 + 9586605\ M\^32 + 
                    10513881\ M\^34 - 29241930\ M\^36 - 6284128\ M\^38 + 
                    62926062\ M\^40 - 23380769\ M\^42 - 84618776\ M\^44 + 
                    85329071\ M\^46 + 18916007\ M\^48 - 125680412\ M\^50 + 
                    225304070\ M\^52 + 13062464\ M\^54 - 669867596\ M\^56 + 
                    405546395\ M\^58 + 1168417709\ M\^60 - 
                    1146538237\ M\^62 - 1442134977\ M\^64 + 
                    1976225902\ M\^66 + 1277408946\ M\^68 - 
                    2499760134\ M\^70 - 723691882\ M\^72 + 
                    2422716836\ M\^74 + 70201488\ M\^76 - 1771503081\ M\^78 + 
                    351883170\ M\^80 + 869406965\ M\^82 - 399034477\ M\^84 - 
                    108218332\ M\^86 + 165842251\ M\^88 - 264041361\ M\^90 + 
                    120897522\ M\^92 + 261254224\ M\^94 - 264975634\ M\^96 - 
                    84202381\ M\^98 + 224622223\ M\^100 - 56161144\ M\^102 - 
                    98577768\ M\^104 + 82365810\ M\^106 + 2471890\ M\^108 - 
                    40779773\ M\^110 + 23796312\ M\^112 + 3296519\ M\^114 - 
                    12279596\ M\^116 + 6525602\ M\^118 + 598188\ M\^120 - 
                    3031441\ M\^122 + 2233062\ M\^124 - 956391\ M\^126 + 
                    267935\ M\^128 - 48652\ M\^130 + 5222\ M\^132 - 
                    252\ M\^134)\) + 
              L\^20\ \((\(-45\)\ M\^16 + 695\ M\^18 - 4960\ M\^20 + 
                    21962\ M\^22 - 68345\ M\^24 + 159046\ M\^26 - 
                    276576\ M\^28 + 314785\ M\^30 - 77696\ M\^32 - 
                    457553\ M\^34 + 751475\ M\^36 - 201270\ M\^38 - 
                    361948\ M\^40 - 578989\ M\^42 + 1059456\ M\^44 + 
                    3205441\ M\^46 - 7270027\ M\^48 - 3180228\ M\^50 + 
                    20952908\ M\^52 - 6242776\ M\^54 - 40393008\ M\^56 + 
                    29080497\ M\^58 + 63973940\ M\^60 - 66851646\ M\^62 - 
                    87733286\ M\^64 + 117940413\ M\^66 + 102679339\ M\^68 - 
                    171530634\ M\^70 - 91533279\ M\^72 + 204246787\ M\^74 + 
                    50989857\ M\^76 - 192900358\ M\^78 - 342213\ M\^80 + 
                    137656837\ M\^82 - 31651703\ M\^84 - 68569295\ M\^86 + 
                    33824775\ M\^88 + 18716213\ M\^90 - 19438153\ M\^92 + 
                    2126775\ M\^94 + 6271331\ M\^96 - 5296888\ M\^98 - 
                    46311\ M\^100 + 3090067\ M\^102 - 1443354\ M\^104 - 
                    899098\ M\^106 + 889588\ M\^108 + 228895\ M\^110 - 
                    583206\ M\^112 + 256823\ M\^114 - 75012\ M\^116 + 
                    173991\ M\^118 - 253739\ M\^120 + 199518\ M\^122 - 
                    100540\ M\^124 + 34739\ M\^126 - 8328\ M\^128 + 
                    1342\ M\^130 - 132\ M\^132 + 6\ M\^134)\) + 
              L\^12\ \((\(-1\) + 31\ M\^2 - 460\ M\^4 + 4185\ M\^6 - 
                    25812\ M\^8 + 114332\ M\^10 - 376437\ M\^12 + 
                    931119\ M\^14 - 1683303\ M\^16 + 1995435\ M\^18 - 
                    991863\ M\^20 - 751108\ M\^22 + 567784\ M\^24 + 
                    1488781\ M\^26 + 253388\ M\^28 - 3293340\ M\^30 - 
                    9151547\ M\^32 + 25099158\ M\^34 + 10310549\ M\^36 - 
                    72564309\ M\^38 + 18391229\ M\^40 + 127946080\ M\^42 - 
                    93461850\ M\^44 - 119280938\ M\^46 + 178055898\ M\^48 - 
                    85090570\ M\^50 - 113781025\ M\^52 + 573970705\ M\^54 - 
                    333631210\ M\^56 - 1257213199\ M\^58 + 
                    1301850142\ M\^60 + 1821718989\ M\^62 - 
                    2627994078\ M\^64 - 1895648352\ M\^66 + 
                    3832292583\ M\^68 + 1324335712\ M\^70 - 
                    4357243397\ M\^72 - 330988499\ M\^74 + 
                    3903962430\ M\^76 - 617700300\ M\^78 - 
                    2643281768\ M\^80 + 1085297816\ M\^82 + 
                    1128509356\ M\^84 - 943961886\ M\^86 + 42796039\ M\^88 + 
                    414803863\ M\^90 - 547569185\ M\^92 + 121345698\ M\^94 + 
                    473093516\ M\^96 - 380448714\ M\^98 - 162648234\ M\^100 + 
                    333558152\ M\^102 - 67126280\ M\^104 - 
                    151497602\ M\^106 + 116531191\ M\^108 + 
                    10119815\ M\^110 - 61409101\ M\^112 + 31735825\ M\^114 + 
                    7664162\ M\^116 - 17599551\ M\^118 + 7598541\ M\^120 + 
                    1974769\ M\^122 - 4307943\ M\^124 + 2757859\ M\^126 - 
                    1069356\ M\^128 + 275037\ M\^130 - 46260\ M\^132 + 
                    4634\ M\^134 - 210\ M\^136)\) + 
              L\^13\ \((\(-10\)\ M\^2 + 261\ M\^4 - 3269\ M\^6 + 
                    25534\ M\^8 - 136220\ M\^10 + 514542\ M\^12 - 
                    1382453\ M\^14 + 2548764\ M\^16 - 2755904\ M\^18 + 
                    219389\ M\^20 + 4157782\ M\^22 - 3709125\ M\^24 - 
                    7223589\ M\^26 + 15465622\ M\^28 + 4950746\ M\^30 - 
                    35072547\ M\^32 - 930\ M\^34 + 82090118\ M\^36 - 
                    33125558\ M\^38 - 153043144\ M\^40 + 127684148\ M\^42 + 
                    180136595\ M\^44 - 254968492\ M\^46 - 36708967\ M\^48 + 
                    260731150\ M\^50 - 396813208\ M\^52 + 133347420\ M\^54 + 
                    1086525027\ M\^56 - 1153010347\ M\^58 - 
                    1766023059\ M\^60 + 2735895693\ M\^62 + 
                    2039863708\ M\^64 - 4433015471\ M\^66 - 
                    1639875400\ M\^68 + 5591480423\ M\^70 + 
                    638515015\ M\^72 - 5699253546\ M\^74 + 545256595\ M\^76 + 
                    4659804359\ M\^78 - 1402711774\ M\^80 - 
                    2872826338\ M\^82 + 1595736177\ M\^84 + 
                    1046892114\ M\^86 - 1156612740\ M\^88 + 
                    195649082\ M\^90 + 440336325\ M\^92 - 638068569\ M\^94 + 
                    143376735\ M\^96 + 501653401\ M\^98 - 382910784\ M\^100 - 
                    180418651\ M\^102 + 326481245\ M\^104 - 
                    41036974\ M\^106 - 158759414\ M\^108 + 98372578\ M\^110 + 
                    27069025\ M\^112 - 60272906\ M\^114 + 21410584\ M\^116 + 
                    13147676\ M\^118 - 15876810\ M\^120 + 3985634\ M\^122 + 
                    3789786\ M\^124 - 4398890\ M\^126 + 2365584\ M\^128 - 
                    816532\ M\^130 + 191199\ M\^132 - 29725\ M\^134 + 
                    2790\ M\^136 - 120\ M\^138)\) + 
              L\^19\ \((\(-120\)\ M\^14 + 2089\ M\^16 - 16481\ M\^18 + 
                    77901\ M\^20 - 244696\ M\^22 + 528503\ M\^24 - 
                    750489\ M\^26 + 481447\ M\^28 + 657621\ M\^30 - 
                    2172636\ M\^32 + 1981879\ M\^34 + 1850870\ M\^36 - 
                    5952943\ M\^38 + 1524519\ M\^40 + 10434719\ M\^42 - 
                    8463397\ M\^44 - 16379540\ M\^46 + 20948638\ M\^48 + 
                    26134573\ M\^50 - 42424114\ M\^52 - 43374130\ M\^54 + 
                    72267322\ M\^56 + 80371169\ M\^58 - 110961984\ M\^60 - 
                    152547638\ M\^62 + 171272183\ M\^64 + 254388959\ M\^66 - 
                    266270150\ M\^68 - 331762563\ M\^70 + 374666514\ M\^72 + 
                    326906551\ M\^74 - 435085177\ M\^76 - 232576628\ M\^78 + 
                    397701794\ M\^80 + 105058543\ M\^82 - 279644097\ M\^84 - 
                    10037986\ M\^86 + 144228481\ M\^88 - 28463138\ M\^90 - 
                    45380155\ M\^92 + 25903310\ M\^94 - 2080167\ M\^96 - 
                    9349074\ M\^98 + 11760104\ M\^100 - 1612364\ M\^102 - 
                    6882614\ M\^104 + 3817418\ M\^106 + 2039712\ M\^108 - 
                    2628600\ M\^110 + 403768\ M\^112 + 501392\ M\^114 - 
                    36141\ M\^116 - 416322\ M\^118 + 516697\ M\^120 - 
                    476556\ M\^122 + 372063\ M\^124 - 225720\ M\^126 + 
                    100691\ M\^128 - 32263\ M\^130 + 7246\ M\^132 - 
                    1087\ M\^134 + 98\ M\^136 - 4\ M\^138)\) + 
              L\^14\ \((\(-45\)\ M\^4 + 1099\ M\^6 - 12551\ M\^8 + 
                    88067\ M\^10 - 417212\ M\^12 + 1372465\ M\^14 - 
                    3061504\ M\^16 + 4054781\ M\^18 - 968720\ M\^20 - 
                    7265445\ M\^22 + 11725101\ M\^24 + 3333646\ M\^26 - 
                    33485364\ M\^28 + 31194307\ M\^30 + 42270234\ M\^32 - 
                    103641989\ M\^34 - 5806136\ M\^36 + 205549832\ M\^38 - 
                    117905000\ M\^40 - 276472238\ M\^42 + 329901206\ M\^44 + 
                    181248032\ M\^46 - 484479593\ M\^48 + 206527995\ M\^50 + 
                    274924768\ M\^52 - 867971613\ M\^54 + 613395313\ M\^56 + 
                    1550070142\ M\^58 - 2216185378\ M\^60 - 
                    1864763981\ M\^62 + 4143279311\ M\^64 + 
                    1535840784\ M\^66 - 5716942827\ M\^68 - 
                    597204628\ M\^70 + 6322128482\ M\^72 - 597204628\ M\^74 - 
                    5716942827\ M\^76 + 1535840784\ M\^78 + 
                    4143279311\ M\^80 - 1864763981\ M\^82 - 
                    2216185378\ M\^84 + 1550070142\ M\^86 + 
                    613395313\ M\^88 - 867971613\ M\^90 + 274924768\ M\^92 + 
                    206527995\ M\^94 - 484479593\ M\^96 + 181248032\ M\^98 + 
                    329901206\ M\^100 - 276472238\ M\^102 - 
                    117905000\ M\^104 + 205549832\ M\^106 - 5806136\ M\^108 - 
                    103641989\ M\^110 + 42270234\ M\^112 + 31194307\ M\^114 - 
                    33485364\ M\^116 + 3333646\ M\^118 + 11725101\ M\^120 - 
                    7265445\ M\^122 - 968720\ M\^124 + 4054781\ M\^126 - 
                    3061504\ M\^128 + 1372465\ M\^130 - 417212\ M\^132 + 
                    88067\ M\^134 - 12551\ M\^136 + 1099\ M\^138 - 
                    45\ M\^140)\) + 
              L\^15\ \((\(-120\)\ M\^6 + 2790\ M\^8 - 29725\ M\^10 + 
                    191199\ M\^12 - 816532\ M\^14 + 2365584\ M\^16 - 
                    4398890\ M\^18 + 3789786\ M\^20 + 3985634\ M\^22 - 
                    15876810\ M\^24 + 13147676\ M\^26 + 21410584\ M\^28 - 
                    60272906\ M\^30 + 27069025\ M\^32 + 98372578\ M\^34 - 
                    158759414\ M\^36 - 41036974\ M\^38 + 326481245\ M\^40 - 
                    180418651\ M\^42 - 382910784\ M\^44 + 501653401\ M\^46 + 
                    143376735\ M\^48 - 638068569\ M\^50 + 440336325\ M\^52 + 
                    195649082\ M\^54 - 1156612740\ M\^56 + 
                    1046892114\ M\^58 + 1595736177\ M\^60 - 
                    2872826338\ M\^62 - 1402711774\ M\^64 + 
                    4659804359\ M\^66 + 545256595\ M\^68 - 
                    5699253546\ M\^70 + 638515015\ M\^72 + 
                    5591480423\ M\^74 - 1639875400\ M\^76 - 
                    4433015471\ M\^78 + 2039863708\ M\^80 + 
                    2735895693\ M\^82 - 1766023059\ M\^84 - 
                    1153010347\ M\^86 + 1086525027\ M\^88 + 
                    133347420\ M\^90 - 396813208\ M\^92 + 260731150\ M\^94 - 
                    36708967\ M\^96 - 254968492\ M\^98 + 180136595\ M\^100 + 
                    127684148\ M\^102 - 153043144\ M\^104 - 
                    33125558\ M\^106 + 82090118\ M\^108 - 930\ M\^110 - 
                    35072547\ M\^112 + 4950746\ M\^114 + 15465622\ M\^116 - 
                    7223589\ M\^118 - 3709125\ M\^120 + 4157782\ M\^122 + 
                    219389\ M\^124 - 2755904\ M\^126 + 2548764\ M\^128 - 
                    1382453\ M\^130 + 514542\ M\^132 - 136220\ M\^134 + 
                    25534\ M\^136 - 3269\ M\^138 + 261\ M\^140 - 
                    10\ M\^142)\) + 
              L\^18\ \((\(-210\)\ M\^12 + 4026\ M\^14 - 34710\ M\^16 + 
                    176994\ M\^18 - 586464\ M\^20 + 1279734\ M\^22 - 
                    1653439\ M\^24 + 400361\ M\^26 + 3051742\ M\^28 - 
                    6019559\ M\^30 + 2557452\ M\^32 + 9779985\ M\^34 - 
                    19209929\ M\^36 + 2965363\ M\^38 + 37167795\ M\^40 - 
                    43305092\ M\^42 - 32032369\ M\^44 + 97567468\ M\^46 - 
                    7778633\ M\^48 - 129746498\ M\^50 + 50126919\ M\^52 + 
                    109405969\ M\^54 - 14660300\ M\^56 - 53129920\ M\^58 - 
                    175597279\ M\^60 + 36717098\ M\^62 + 518497604\ M\^64 - 
                    146472442\ M\^66 - 884022466\ M\^68 + 396391289\ M\^70 + 
                    1087516865\ M\^72 - 681733838\ M\^74 - 
                    1008623339\ M\^76 + 835776152\ M\^78 + 692451768\ M\^80 - 
                    766576785\ M\^82 - 315163986\ M\^84 + 522082857\ M\^86 + 
                    45987903\ M\^88 - 244760259\ M\^90 + 59432851\ M\^92 + 
                    54153889\ M\^94 - 55115898\ M\^96 + 22785974\ M\^98 + 
                    22312433\ M\^100 - 30682452\ M\^102 - 1147657\ M\^104 + 
                    17279998\ M\^106 - 4358984\ M\^108 - 6099127\ M\^110 + 
                    2422157\ M\^112 + 2446793\ M\^114 - 1754763\ M\^116 - 
                    163862\ M\^118 + 133859\ M\^120 + 695365\ M\^122 - 
                    959709\ M\^124 + 703300\ M\^126 - 373084\ M\^128 + 
                    158266\ M\^130 - 54232\ M\^132 + 14434\ M\^134 - 
                    2810\ M\^136 + 370\ M\^138 - 29\ M\^140 + M\^142)\) + 
              L\^16\ \((\(-210\)\ M\^8 + 4634\ M\^10 - 46260\ M\^12 + 
                    275037\ M\^14 - 1069356\ M\^16 + 2757859\ M\^18 - 
                    4307943\ M\^20 + 1974769\ M\^22 + 7598541\ M\^24 - 
                    17599551\ M\^26 + 7664162\ M\^28 + 31735825\ M\^30 - 
                    61409101\ M\^32 + 10119815\ M\^34 + 116531191\ M\^36 - 
                    151497602\ M\^38 - 67126280\ M\^40 + 333558152\ M\^42 - 
                    162648234\ M\^44 - 380448714\ M\^46 + 473093516\ M\^48 + 
                    121345698\ M\^50 - 547569185\ M\^52 + 414803863\ M\^54 + 
                    42796039\ M\^56 - 943961886\ M\^58 + 1128509356\ M\^60 + 
                    1085297816\ M\^62 - 2643281768\ M\^64 - 
                    617700300\ M\^66 + 3903962430\ M\^68 - 330988499\ M\^70 - 
                    4357243397\ M\^72 + 1324335712\ M\^74 + 
                    3832292583\ M\^76 - 1895648352\ M\^78 - 
                    2627994078\ M\^80 + 1821718989\ M\^82 + 
                    1301850142\ M\^84 - 1257213199\ M\^86 - 
                    333631210\ M\^88 + 573970705\ M\^90 - 113781025\ M\^92 - 
                    85090570\ M\^94 + 178055898\ M\^96 - 119280938\ M\^98 - 
                    93461850\ M\^100 + 127946080\ M\^102 + 18391229\ M\^104 - 
                    72564309\ M\^106 + 10310549\ M\^108 + 25099158\ M\^110 - 
                    9151547\ M\^112 - 3293340\ M\^114 + 253388\ M\^116 + 
                    1488781\ M\^118 + 567784\ M\^120 - 751108\ M\^122 - 
                    991863\ M\^124 + 1995435\ M\^126 - 1683303\ M\^128 + 
                    931119\ M\^130 - 376437\ M\^132 + 114332\ M\^134 - 
                    25812\ M\^136 + 4185\ M\^138 - 460\ M\^140 + 31\ M\^142 - 
                    M\^144)\) + 
              L\^17\ \((\(-252\)\ M\^10 + 5222\ M\^12 - 48652\ M\^14 + 
                    267935\ M\^16 - 956391\ M\^18 + 2233062\ M\^20 - 
                    3031441\ M\^22 + 598188\ M\^24 + 6525602\ M\^26 - 
                    12279596\ M\^28 + 3296519\ M\^30 + 23796312\ M\^32 - 
                    40779773\ M\^34 + 2471890\ M\^36 + 82365810\ M\^38 - 
                    98577768\ M\^40 - 56161144\ M\^42 + 224622223\ M\^44 - 
                    84202381\ M\^46 - 264975634\ M\^48 + 261254224\ M\^50 + 
                    120897522\ M\^52 - 264041361\ M\^54 + 165842251\ M\^56 - 
                    108218332\ M\^58 - 399034477\ M\^60 + 869406965\ M\^62 + 
                    351883170\ M\^64 - 1771503081\ M\^66 + 70201488\ M\^68 + 
                    2422716836\ M\^70 - 723691882\ M\^72 - 
                    2499760134\ M\^74 + 1277408946\ M\^76 + 
                    1976225902\ M\^78 - 1442134977\ M\^80 - 
                    1146538237\ M\^82 + 1168417709\ M\^84 + 
                    405546395\ M\^86 - 669867596\ M\^88 + 13062464\ M\^90 + 
                    225304070\ M\^92 - 125680412\ M\^94 + 18916007\ M\^96 + 
                    85329071\ M\^98 - 84618776\ M\^100 - 23380769\ M\^102 + 
                    62926062\ M\^104 - 6284128\ M\^106 - 29241930\ M\^108 + 
                    10513881\ M\^110 + 9586605\ M\^112 - 7557190\ M\^114 - 
                    101293\ M\^116 + 1921633\ M\^118 - 454209\ M\^120 - 
                    800262\ M\^122 + 1330067\ M\^124 - 1310111\ M\^126 + 
                    917981\ M\^128 - 470318\ M\^130 + 180897\ M\^132 - 
                    53667\ M\^134 + 12482\ M\^136 - 2249\ M\^138 + 
                    296\ M\^140 - 25\ M\^142 + M\^144)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 49]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 50]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 51]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 52]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 53]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 54]\), \(\(-M\^38\) + 
              L\^29\ M\^154 + 
              L\ \((\(-2\)\ M\^28 + 8\ M\^30 - 21\ M\^32 + 35\ M\^34 - 
                    32\ M\^36 + 45\ M\^40 - 65\ M\^42 - 7\ M\^44 + 
                    24\ M\^46 - 8\ M\^48)\) + 
              L\^2\ \((\(-M\^18\) + 8\ M\^20 - 38\ M\^22 + 118\ M\^24 - 
                    242\ M\^26 + 313\ M\^28 - 177\ M\^30 - 109\ M\^32 + 
                    239\ M\^34 - 263\ M\^36 + 384\ M\^38 - 342\ M\^40 - 
                    319\ M\^42 + 1253\ M\^44 - 984\ M\^46 - 947\ M\^48 + 
                    1189\ M\^50 - 154\ M\^52 - 320\ M\^54 + 170\ M\^56 - 
                    28\ M\^58)\) + 
              L\^3\ \((\(-M\^12\) + 12\ M\^14 - 74\ M\^16 + 298\ M\^18 - 
                    830\ M\^20 + 1600\ M\^22 - 1953\ M\^24 + 968\ M\^26 + 
                    1079\ M\^28 - 2225\ M\^30 + 1268\ M\^32 + 156\ M\^34 - 
                    114\ M\^36 + 1199\ M\^38 - 5054\ M\^40 + 4233\ M\^42 + 
                    2505\ M\^44 - 7838\ M\^46 + 10028\ M\^48 - 1465\ M\^50 - 
                    17023\ M\^52 + 11829\ M\^54 + 7169\ M\^56 - 
                    11062\ M\^58 + 2654\ M\^60 + 2306\ M\^62 - 1829\ M\^64 + 
                    518\ M\^66 - 56\ M\^68)\) + 
              L\^4\ \((3\ M\^8 - 36\ M\^10 + 220\ M\^12 - 879\ M\^14 + 
                    2462\ M\^16 - 4870\ M\^18 + 6435\ M\^20 - 4573\ M\^22 - 
                    184\ M\^24 + 1437\ M\^26 + 5298\ M\^28 - 8166\ M\^30 - 
                    11559\ M\^32 + 36093\ M\^34 - 7760\ M\^36 - 
                    73163\ M\^38 + 82846\ M\^40 + 61041\ M\^42 - 
                    167383\ M\^44 + 19239\ M\^46 + 164914\ M\^48 - 
                    103283\ M\^50 - 42526\ M\^52 + 103952\ M\^54 - 
                    111207\ M\^56 - 15040\ M\^58 + 136053\ M\^60 - 
                    75358\ M\^62 - 47423\ M\^64 + 68572\ M\^66 - 
                    19577\ M\^68 - 12141\ M\^70 + 12299\ M\^72 - 
                    4647\ M\^74 + 882\ M\^76 - 70\ M\^78)\) + 
              L\^5\ \((\(-3\)\ M\^4 + 40\ M\^6 - 263\ M\^8 + 1118\ M\^10 - 
                    3359\ M\^12 + 7360\ M\^14 - 11857\ M\^16 + 14665\ M\^18 - 
                    17603\ M\^20 + 26913\ M\^22 - 33218\ M\^24 - 
                    6659\ M\^26 + 104241\ M\^28 - 117470\ M\^30 - 
                    131951\ M\^32 + 423652\ M\^34 - 151531\ M\^36 - 
                    645221\ M\^38 + 783081\ M\^40 + 330255\ M\^42 - 
                    1257866\ M\^44 + 566037\ M\^46 + 935796\ M\^48 - 
                    1334595\ M\^50 + 54171\ M\^52 + 1149430\ M\^54 - 
                    674004\ M\^56 - 211541\ M\^58 + 253277\ M\^60 - 
                    444499\ M\^62 + 542902\ M\^64 + 176117\ M\^66 - 
                    717570\ M\^68 + 328460\ M\^70 + 246827\ M\^72 - 
                    322391\ M\^74 + 96677\ M\^76 + 50751\ M\^78 - 
                    59022\ M\^80 + 26105\ M\^82 - 6511\ M\^84 + 910\ M\^86 - 
                    56\ M\^88)\) + 
              L\^6\ \((1 - 15\ M\^2 + 110\ M\^4 - 525\ M\^6 + 1831\ M\^8 - 
                    5003\ M\^10 + 11473\ M\^12 - 23809\ M\^14 + 
                    45411\ M\^16 - 70146\ M\^18 + 61388\ M\^20 + 
                    25405\ M\^22 - 115535\ M\^24 - 35030\ M\^26 + 
                    440662\ M\^28 - 410358\ M\^30 - 660956\ M\^32 + 
                    1534389\ M\^34 + 33650\ M\^36 - 3013854\ M\^38 + 
                    2536264\ M\^40 + 2845576\ M\^42 - 6239369\ M\^44 + 
                    912426\ M\^46 + 7134617\ M\^48 - 5649936\ M\^50 - 
                    2907953\ M\^52 + 5550259\ M\^54 - 2146365\ M\^56 - 
                    186088\ M\^58 + 2069251\ M\^60 - 3102331\ M\^62 + 
                    1137062\ M\^64 + 400087\ M\^66 - 1335076\ M\^68 + 
                    2749287\ M\^70 - 1655487\ M\^72 - 1672732\ M\^74 + 
                    2807117\ M\^76 - 854588\ M\^78 - 1054969\ M\^80 + 
                    1135994\ M\^82 - 327712\ M\^84 - 173096\ M\^86 + 
                    210365\ M\^88 - 101371\ M\^90 + 29340\ M\^92 - 
                    5347\ M\^94 + 574\ M\^96 - 28\ M\^98)\) + 
              L\^7\ \((\(-1\) + 18\ M\^2 - 168\ M\^4 + 1028\ M\^6 - 
                    4487\ M\^8 + 14470\ M\^10 - 34542\ M\^12 + 59619\ M\^14 - 
                    73400\ M\^16 + 80951\ M\^18 - 149813\ M\^20 + 
                    299570\ M\^22 - 232028\ M\^24 - 462729\ M\^26 + 
                    1162734\ M\^28 + 54068\ M\^30 - 3270054\ M\^32 + 
                    3652214\ M\^34 + 3191562\ M\^36 - 10592762\ M\^38 + 
                    4432792\ M\^40 + 13685605\ M\^42 - 18492250\ M\^44 - 
                    3586334\ M\^46 + 25624630\ M\^48 - 15919556\ M\^50 - 
                    14148627\ M\^52 + 24741929\ M\^54 - 2515413\ M\^56 - 
                    14793587\ M\^58 + 342254\ M\^60 + 5982900\ M\^62 + 
                    15519187\ M\^64 - 13567934\ M\^66 - 16309379\ M\^68 + 
                    19162717\ M\^70 + 15367\ M\^72 - 6375171\ M\^74 + 
                    6371519\ M\^76 - 8318003\ M\^78 + 2722429\ M\^80 + 
                    7032430\ M\^82 - 8170879\ M\^84 + 1371006\ M\^86 + 
                    3550731\ M\^88 - 3112490\ M\^90 + 754674\ M\^92 + 
                    520241\ M\^94 - 578147\ M\^96 + 283704\ M\^98 - 
                    88178\ M\^100 + 18416\ M\^102 - 2535\ M\^104 + 
                    210\ M\^106 - 8\ M\^108)\) + 
              L\^8\ \((\(-3\)\ M\^2 + 56\ M\^4 - 527\ M\^6 + 3274\ M\^8 - 
                    14729\ M\^10 + 49515\ M\^12 - 123398\ M\^14 + 
                    214923\ M\^16 - 214824\ M\^18 + 7637\ M\^20 + 
                    184570\ M\^22 + 348516\ M\^24 - 1742374\ M\^26 + 
                    1880773\ M\^28 + 2009390\ M\^30 - 7007507\ M\^32 + 
                    3097438\ M\^34 + 12132897\ M\^36 - 18771630\ M\^38 - 
                    4209193\ M\^40 + 35761693\ M\^42 - 25258907\ M\^44 - 
                    29380679\ M\^46 + 56595689\ M\^48 - 7239919\ M\^50 - 
                    55944155\ M\^52 + 41874560\ M\^54 + 21719598\ M\^56 - 
                    41909544\ M\^58 + 7619460\ M\^60 + 30429297\ M\^62 - 
                    19064179\ M\^64 - 43084044\ M\^66 + 41733142\ M\^68 + 
                    54452596\ M\^70 - 72161461\ M\^72 - 24545537\ M\^74 + 
                    58467356\ M\^76 - 12313539\ M\^78 - 3839128\ M\^80 + 
                    2176356\ M\^82 - 21532448\ M\^84 + 26346467\ M\^86 + 
                    359045\ M\^88 - 22717181\ M\^90 + 17234495\ M\^92 + 
                    537736\ M\^94 - 9156101\ M\^96 + 6358890\ M\^98 - 
                    1058913\ M\^100 - 1329867\ M\^102 + 1252202\ M\^104 - 
                    591340\ M\^106 + 184551\ M\^108 - 40359\ M\^110 + 
                    6171\ M\^112 - 629\ M\^114 + 38\ M\^116 - M\^118)\) + 
              L\^9\ \((\(-3\)\ M\^4 + 63\ M\^6 - 655\ M\^8 + 4489\ M\^10 - 
                    22414\ M\^12 + 84547\ M\^14 - 241245\ M\^16 + 
                    502611\ M\^18 - 678137\ M\^20 + 301375\ M\^22 + 
                    829422\ M\^24 - 1423107\ M\^26 - 1200603\ M\^28 + 
                    6797008\ M\^30 - 7185785\ M\^32 - 6770147\ M\^34 + 
                    24552570\ M\^36 - 14198573\ M\^38 - 32211840\ M\^40 + 
                    58754588\ M\^42 - 4002813\ M\^44 - 86188109\ M\^46 + 
                    81380008\ M\^48 + 41919566\ M\^50 - 126656488\ M\^52 + 
                    51135913\ M\^54 + 84318730\ M\^56 - 99742525\ M\^58 - 
                    3842602\ M\^60 + 63007460\ M\^62 - 28382961\ M\^64 - 
                    3782335\ M\^66 + 26424563\ M\^68 - 33079499\ M\^70 - 
                    46292290\ M\^72 + 69189794\ M\^74 + 60455649\ M\^76 - 
                    90194804\ M\^78 - 31487211\ M\^80 + 44408391\ M\^82 + 
                    15396485\ M\^84 + 22247637\ M\^86 - 53541391\ M\^88 - 
                    11013132\ M\^90 + 68946706\ M\^92 - 44102231\ M\^94 - 
                    19228056\ M\^96 + 47134210\ M\^98 - 24449529\ M\^100 - 
                    8119704\ M\^102 + 17897260\ M\^104 - 9477730\ M\^106 + 
                    269973\ M\^108 + 2802207\ M\^110 - 2123362\ M\^112 + 
                    920728\ M\^114 - 273899\ M\^116 + 58341\ M\^118 - 
                    8883\ M\^120 + 934\ M\^122 - 62\ M\^124 + 2\ M\^126)\) + 
              L\^10\ \((\(-M\^6\) + 25\ M\^8 - 314\ M\^10 + 2610\ M\^12 - 
                    15778\ M\^14 + 72047\ M\^16 - 251847\ M\^18 + 
                    669559\ M\^20 - 1299554\ M\^22 + 1593795\ M\^24 - 
                    371828\ M\^26 - 2831128\ M\^28 + 4888182\ M\^30 + 
                    823264\ M\^32 - 15040200\ M\^34 + 20861364\ M\^36 + 
                    4835348\ M\^38 - 49940265\ M\^40 + 51215793\ M\^42 + 
                    31537847\ M\^44 - 118044622\ M\^46 + 68985652\ M\^48 + 
                    100255655\ M\^50 - 176337000\ M\^52 + 29603609\ M\^54 + 
                    160681856\ M\^56 - 133684300\ M\^58 - 50827938\ M\^60 + 
                    101249109\ M\^62 + 12271602\ M\^64 - 27176939\ M\^66 - 
                    77775734\ M\^68 + 38291402\ M\^70 + 96091386\ M\^72 - 
                    41222179\ M\^74 - 29703786\ M\^76 - 85856752\ M\^78 + 
                    12015636\ M\^80 + 227430168\ M\^82 - 68457680\ M\^84 - 
                    233584917\ M\^86 + 86062962\ M\^88 + 160217028\ M\^90 - 
                    58183417\ M\^92 - 110471091\ M\^94 + 78664111\ M\^96 + 
                    44803466\ M\^98 - 103005297\ M\^100 + 45427660\ M\^102 + 
                    49213440\ M\^104 - 69838447\ M\^106 + 19615274\ M\^108 + 
                    23401047\ M\^110 - 25650885\ M\^112 + 8967514\ M\^114 + 
                    2496545\ M\^116 - 4583325\ M\^118 + 2762059\ M\^120 - 
                    1054271\ M\^122 + 283417\ M\^124 - 54968\ M\^126 + 
                    7592\ M\^128 - 710\ M\^130 + 40\ M\^132 - M\^134)\) + 
              L\^11\ \((M\^10 - 33\ M\^12 + 485\ M\^14 - 4357\ M\^16 + 
                    27282\ M\^18 - 127041\ M\^20 + 452398\ M\^22 - 
                    1232024\ M\^24 + 2477725\ M\^26 - 3250224\ M\^28 + 
                    1223594\ M\^30 + 5292210\ M\^32 - 11873536\ M\^34 + 
                    5530845\ M\^36 + 21542619\ M\^38 - 45838188\ M\^40 + 
                    17612724\ M\^42 + 69332322\ M\^44 - 117977814\ M\^46 + 
                    14879483\ M\^48 + 173761915\ M\^50 - 189864929\ M\^52 - 
                    72314719\ M\^54 + 293461032\ M\^56 - 121539354\ M\^58 - 
                    231662329\ M\^60 + 237031144\ M\^62 + 148958406\ M\^64 - 
                    316068278\ M\^66 - 139915529\ M\^68 + 515041421\ M\^70 + 
                    66888486\ M\^72 - 719867619\ M\^74 + 101813282\ M\^76 + 
                    672096212\ M\^78 - 87232855\ M\^80 - 478619061\ M\^82 - 
                    198542450\ M\^84 + 453502240\ M\^86 + 377279013\ M\^88 - 
                    536899400\ M\^90 - 215796386\ M\^92 + 450916237\ M\^94 - 
                    37910757\ M\^96 - 217357622\ M\^98 + 122994855\ M\^100 + 
                    78795028\ M\^102 - 120027428\ M\^104 - 40300932\ M\^106 + 
                    128059323\ M\^108 - 21029999\ M\^110 - 87211760\ M\^112 + 
                    66584190\ M\^114 + 6615097\ M\^116 - 38094688\ M\^118 + 
                    24203139\ M\^120 - 2710087\ M\^122 - 6306764\ M\^124 + 
                    5557220\ M\^126 - 2638440\ M\^128 + 848881\ M\^130 - 
                    195260\ M\^132 + 32094\ M\^134 - 3612\ M\^136 + 
                    250\ M\^138 - 8\ M\^140)\) + 
              L\^12\ \((8\ M\^16 - 229\ M\^18 + 2983\ M\^20 - 23701\ M\^22 + 
                    129535\ M\^24 - 517908\ M\^26 + 1555820\ M\^28 - 
                    3478370\ M\^30 + 5355417\ M\^32 - 3884416\ M\^34 - 
                    4990169\ M\^36 + 18475114\ M\^38 - 19532151\ M\^40 - 
                    11292350\ M\^42 + 59659720\ M\^44 - 62466405\ M\^46 - 
                    27309668\ M\^48 + 143659015\ M\^50 - 132469363\ M\^52 - 
                    62704270\ M\^54 + 258601256\ M\^56 - 180573866\ M\^58 - 
                    150237595\ M\^60 + 327881198\ M\^62 - 120972370\ M\^64 - 
                    203855462\ M\^66 + 331325045\ M\^68 - 27521470\ M\^70 - 
                    550971336\ M\^72 + 256110078\ M\^74 + 937222689\ M\^76 - 
                    595430008\ M\^78 - 1245319101\ M\^80 + 932889200\ M\^82 + 
                    1158453589\ M\^84 - 896528977\ M\^86 - 898931618\ M\^88 + 
                    552199933\ M\^90 + 809965736\ M\^92 - 443521103\ M\^94 - 
                    676883321\ M\^96 + 643130931\ M\^98 + 249011343\ M\^100 - 
                    682786848\ M\^102 + 190155354\ M\^104 + 
                    404321041\ M\^106 - 274964700\ M\^108 - 
                    145959300\ M\^110 + 145594990\ M\^112 + 
                    78967072\ M\^114 - 89983064\ M\^116 - 43738851\ M\^118 + 
                    85690018\ M\^120 - 21670851\ M\^122 - 35773087\ M\^124 + 
                    36707476\ M\^126 - 10943485\ M\^128 - 6003843\ M\^130 + 
                    8206889\ M\^132 - 4671410\ M\^134 + 1725551\ M\^136 - 
                    449679\ M\^138 + 83141\ M\^140 - 10443\ M\^142 + 
                    799\ M\^144 - 28\ M\^146)\) + 
              L\^13\ \((28\ M\^22 - 765\ M\^24 + 9472\ M\^26 - 70744\ M\^28 + 
                    357541\ M\^30 - 1295556\ M\^32 + 3440799\ M\^34 - 
                    6532990\ M\^36 + 7608279\ M\^38 - 551701\ M\^40 - 
                    17452142\ M\^42 + 32913832\ M\^44 - 16357627\ M\^46 - 
                    42041520\ M\^48 + 91203989\ M\^50 - 49703950\ M\^52 - 
                    84859362\ M\^54 + 180868056\ M\^56 - 90314287\ M\^58 - 
                    143983894\ M\^60 + 240516865\ M\^62 - 18611679\ M\^64 - 
                    176063743\ M\^66 - 41669385\ M\^68 + 100410277\ M\^70 + 
                    449753051\ M\^72 - 285077177\ M\^74 - 874370311\ M\^76 + 
                    469309408\ M\^78 + 1160107374\ M\^80 - 315657940\ M\^82 - 
                    1480433346\ M\^84 + 113004704\ M\^86 + 
                    1699578646\ M\^88 - 250994330\ M\^90 - 
                    1461005495\ M\^92 + 494865632\ M\^94 + 954257316\ M\^96 - 
                    523079150\ M\^98 - 737992021\ M\^100 + 
                    594883124\ M\^102 + 725024274\ M\^104 - 
                    844549942\ M\^106 - 451237405\ M\^108 + 
                    887539652\ M\^110 + 5751074\ M\^112 - 531607018\ M\^114 + 
                    168426655\ M\^116 + 150555191\ M\^118 - 
                    61309303\ M\^120 - 46799458\ M\^122 - 6618799\ M\^124 + 
                    64598558\ M\^126 - 29969871\ M\^128 - 28730912\ M\^130 + 
                    39942998\ M\^132 - 15542326\ M\^134 - 5204252\ M\^136 + 
                    9763553\ M\^138 - 6228716\ M\^140 + 2522193\ M\^142 - 
                    714370\ M\^144 + 142255\ M\^146 - 19031\ M\^148 + 
                    1533\ M\^150 - 56\ M\^152)\) + 
              L\^14\ \((56\ M\^28 - 1505\ M\^30 + 18255\ M\^32 - 
                    132590\ M\^34 + 644550\ M\^36 - 2210786\ M\^38 + 
                    5419031\ M\^40 - 9024404\ M\^42 + 7625348\ M\^44 + 
                    6215568\ M\^46 - 29635725\ M\^48 + 37909063\ M\^50 - 
                    4151830\ M\^52 - 56386414\ M\^54 + 87459243\ M\^56 - 
                    49711306\ M\^58 - 57361345\ M\^60 + 192681068\ M\^62 - 
                    190796788\ M\^64 - 129587819\ M\^66 + 442394420\ M\^68 - 
                    65868661\ M\^70 - 557510720\ M\^72 + 41680586\ M\^74 + 
                    831965797\ M\^76 + 198795322\ M\^78 - 1443313667\ M\^80 - 
                    253074659\ M\^82 + 1883405983\ M\^84 + 118745607\ M\^86 - 
                    1749165168\ M\^88 - 127042616\ M\^90 + 
                    1481519885\ M\^92 + 92407813\ M\^94 - 1510968458\ M\^96 + 
                    339092866\ M\^98 + 1536882133\ M\^100 - 
                    839767640\ M\^102 - 1260577155\ M\^104 + 
                    838263962\ M\^106 + 1012156571\ M\^108 - 
                    555809043\ M\^110 - 930791995\ M\^112 + 
                    498154226\ M\^114 + 680478332\ M\^116 - 
                    542620157\ M\^118 - 218159910\ M\^120 + 
                    352425811\ M\^122 - 68775923\ M\^124 - 58206094\ M\^126 + 
                    35192439\ M\^128 - 39288359\ M\^130 + 51457445\ M\^132 - 
                    14288319\ M\^134 - 31850796\ M\^136 + 36845424\ M\^138 - 
                    13060404\ M\^140 - 6021785\ M\^142 + 10073563\ M\^144 - 
                    6526790\ M\^146 + 2735709\ M\^148 - 805054\ M\^150 + 
                    166013\ M\^152 - 22857\ M\^154 + 1883\ M\^156 - 
                    70\ M\^158)\) + 
              L\^15\ \((70\ M\^34 - 1883\ M\^36 + 22857\ M\^38 - 
                    166013\ M\^40 + 805054\ M\^42 - 2735709\ M\^44 + 
                    6526790\ M\^46 - 10073563\ M\^48 + 6021785\ M\^50 + 
                    13060404\ M\^52 - 36845424\ M\^54 + 31850796\ M\^56 + 
                    14288319\ M\^58 - 51457445\ M\^60 + 39288359\ M\^62 - 
                    35192439\ M\^64 + 58206094\ M\^66 + 68775923\ M\^68 - 
                    352425811\ M\^70 + 218159910\ M\^72 + 542620157\ M\^74 - 
                    680478332\ M\^76 - 498154226\ M\^78 + 930791995\ M\^80 + 
                    555809043\ M\^82 - 1012156571\ M\^84 - 838263962\ M\^86 + 
                    1260577155\ M\^88 + 839767640\ M\^90 - 
                    1536882133\ M\^92 - 339092866\ M\^94 + 
                    1510968458\ M\^96 - 92407813\ M\^98 - 
                    1481519885\ M\^100 + 127042616\ M\^102 + 
                    1749165168\ M\^104 - 118745607\ M\^106 - 
                    1883405983\ M\^108 + 253074659\ M\^110 + 
                    1443313667\ M\^112 - 198795322\ M\^114 - 
                    831965797\ M\^116 - 41680586\ M\^118 + 
                    557510720\ M\^120 + 65868661\ M\^122 - 
                    442394420\ M\^124 + 129587819\ M\^126 + 
                    190796788\ M\^128 - 192681068\ M\^130 + 
                    57361345\ M\^132 + 49711306\ M\^134 - 87459243\ M\^136 + 
                    56386414\ M\^138 + 4151830\ M\^140 - 37909063\ M\^142 + 
                    29635725\ M\^144 - 6215568\ M\^146 - 7625348\ M\^148 + 
                    9024404\ M\^150 - 5419031\ M\^152 + 2210786\ M\^154 - 
                    644550\ M\^156 + 132590\ M\^158 - 18255\ M\^160 + 
                    1505\ M\^162 - 56\ M\^164)\) + 
              L\^28\ \((8\ M\^144 - 24\ M\^146 + 7\ M\^148 + 65\ M\^150 - 
                    45\ M\^152 + 32\ M\^156 - 35\ M\^158 + 21\ M\^160 - 
                    8\ M\^162 + 2\ M\^164)\) + 
              L\^16\ \((56\ M\^40 - 1533\ M\^42 + 19031\ M\^44 - 
                    142255\ M\^46 + 714370\ M\^48 - 2522193\ M\^50 + 
                    6228716\ M\^52 - 9763553\ M\^54 + 5204252\ M\^56 + 
                    15542326\ M\^58 - 39942998\ M\^60 + 28730912\ M\^62 + 
                    29969871\ M\^64 - 64598558\ M\^66 + 6618799\ M\^68 + 
                    46799458\ M\^70 + 61309303\ M\^72 - 150555191\ M\^74 - 
                    168426655\ M\^76 + 531607018\ M\^78 - 5751074\ M\^80 - 
                    887539652\ M\^82 + 451237405\ M\^84 + 844549942\ M\^86 - 
                    725024274\ M\^88 - 594883124\ M\^90 + 737992021\ M\^92 + 
                    523079150\ M\^94 - 954257316\ M\^96 - 494865632\ M\^98 + 
                    1461005495\ M\^100 + 250994330\ M\^102 - 
                    1699578646\ M\^104 - 113004704\ M\^106 + 
                    1480433346\ M\^108 + 315657940\ M\^110 - 
                    1160107374\ M\^112 - 469309408\ M\^114 + 
                    874370311\ M\^116 + 285077177\ M\^118 - 
                    449753051\ M\^120 - 100410277\ M\^122 + 
                    41669385\ M\^124 + 176063743\ M\^126 + 18611679\ M\^128 - 
                    240516865\ M\^130 + 143983894\ M\^132 + 
                    90314287\ M\^134 - 180868056\ M\^136 + 84859362\ M\^138 + 
                    49703950\ M\^140 - 91203989\ M\^142 + 42041520\ M\^144 + 
                    16357627\ M\^146 - 32913832\ M\^148 + 17452142\ M\^150 + 
                    551701\ M\^152 - 7608279\ M\^154 + 6532990\ M\^156 - 
                    3440799\ M\^158 + 1295556\ M\^160 - 357541\ M\^162 + 
                    70744\ M\^164 - 9472\ M\^166 + 765\ M\^168 - 
                    28\ M\^170)\) + 
              L\^27\ \((28\ M\^134 - 170\ M\^136 + 320\ M\^138 + 
                    154\ M\^140 - 1189\ M\^142 + 947\ M\^144 + 984\ M\^146 - 
                    1253\ M\^148 + 319\ M\^150 + 342\ M\^152 - 384\ M\^154 + 
                    263\ M\^156 - 239\ M\^158 + 109\ M\^160 + 177\ M\^162 - 
                    313\ M\^164 + 242\ M\^166 - 118\ M\^168 + 38\ M\^170 - 
                    8\ M\^172 + M\^174)\) + 
              L\^17\ \((28\ M\^46 - 799\ M\^48 + 10443\ M\^50 - 
                    83141\ M\^52 + 449679\ M\^54 - 1725551\ M\^56 + 
                    4671410\ M\^58 - 8206889\ M\^60 + 6003843\ M\^62 + 
                    10943485\ M\^64 - 36707476\ M\^66 + 35773087\ M\^68 + 
                    21670851\ M\^70 - 85690018\ M\^72 + 43738851\ M\^74 + 
                    89983064\ M\^76 - 78967072\ M\^78 - 145594990\ M\^80 + 
                    145959300\ M\^82 + 274964700\ M\^84 - 404321041\ M\^86 - 
                    190155354\ M\^88 + 682786848\ M\^90 - 249011343\ M\^92 - 
                    643130931\ M\^94 + 676883321\ M\^96 + 443521103\ M\^98 - 
                    809965736\ M\^100 - 552199933\ M\^102 + 
                    898931618\ M\^104 + 896528977\ M\^106 - 
                    1158453589\ M\^108 - 932889200\ M\^110 + 
                    1245319101\ M\^112 + 595430008\ M\^114 - 
                    937222689\ M\^116 - 256110078\ M\^118 + 
                    550971336\ M\^120 + 27521470\ M\^122 - 
                    331325045\ M\^124 + 203855462\ M\^126 + 
                    120972370\ M\^128 - 327881198\ M\^130 + 
                    150237595\ M\^132 + 180573866\ M\^134 - 
                    258601256\ M\^136 + 62704270\ M\^138 + 
                    132469363\ M\^140 - 143659015\ M\^142 + 
                    27309668\ M\^144 + 62466405\ M\^146 - 59659720\ M\^148 + 
                    11292350\ M\^150 + 19532151\ M\^152 - 18475114\ M\^154 + 
                    4990169\ M\^156 + 3884416\ M\^158 - 5355417\ M\^160 + 
                    3478370\ M\^162 - 1555820\ M\^164 + 517908\ M\^166 - 
                    129535\ M\^168 + 23701\ M\^170 - 2983\ M\^172 + 
                    229\ M\^174 - 8\ M\^176)\) + 
              L\^26\ \((56\ M\^124 - 518\ M\^126 + 1829\ M\^128 - 
                    2306\ M\^130 - 2654\ M\^132 + 11062\ M\^134 - 
                    7169\ M\^136 - 11829\ M\^138 + 17023\ M\^140 + 
                    1465\ M\^142 - 10028\ M\^144 + 7838\ M\^146 - 
                    2505\ M\^148 - 4233\ M\^150 + 5054\ M\^152 - 
                    1199\ M\^154 + 114\ M\^156 - 156\ M\^158 - 1268\ M\^160 + 
                    2225\ M\^162 - 1079\ M\^164 - 968\ M\^166 + 
                    1953\ M\^168 - 1600\ M\^170 + 830\ M\^172 - 298\ M\^174 + 
                    74\ M\^176 - 12\ M\^178 + M\^180)\) + 
              L\^18\ \((8\ M\^52 - 250\ M\^54 + 3612\ M\^56 - 32094\ M\^58 + 
                    195260\ M\^60 - 848881\ M\^62 + 2638440\ M\^64 - 
                    5557220\ M\^66 + 6306764\ M\^68 + 2710087\ M\^70 - 
                    24203139\ M\^72 + 38094688\ M\^74 - 6615097\ M\^76 - 
                    66584190\ M\^78 + 87211760\ M\^80 + 21029999\ M\^82 - 
                    128059323\ M\^84 + 40300932\ M\^86 + 120027428\ M\^88 - 
                    78795028\ M\^90 - 122994855\ M\^92 + 217357622\ M\^94 + 
                    37910757\ M\^96 - 450916237\ M\^98 + 215796386\ M\^100 + 
                    536899400\ M\^102 - 377279013\ M\^104 - 
                    453502240\ M\^106 + 198542450\ M\^108 + 
                    478619061\ M\^110 + 87232855\ M\^112 - 
                    672096212\ M\^114 - 101813282\ M\^116 + 
                    719867619\ M\^118 - 66888486\ M\^120 - 
                    515041421\ M\^122 + 139915529\ M\^124 + 
                    316068278\ M\^126 - 148958406\ M\^128 - 
                    237031144\ M\^130 + 231662329\ M\^132 + 
                    121539354\ M\^134 - 293461032\ M\^136 + 
                    72314719\ M\^138 + 189864929\ M\^140 - 
                    173761915\ M\^142 - 14879483\ M\^144 + 
                    117977814\ M\^146 - 69332322\ M\^148 - 17612724\ M\^150 + 
                    45838188\ M\^152 - 21542619\ M\^154 - 5530845\ M\^156 + 
                    11873536\ M\^158 - 5292210\ M\^160 - 1223594\ M\^162 + 
                    3250224\ M\^164 - 2477725\ M\^166 + 1232024\ M\^168 - 
                    452398\ M\^170 + 127041\ M\^172 - 27282\ M\^174 + 
                    4357\ M\^176 - 485\ M\^178 + 33\ M\^180 - M\^182)\) + 
              L\^25\ \((70\ M\^114 - 882\ M\^116 + 4647\ M\^118 - 
                    12299\ M\^120 + 12141\ M\^122 + 19577\ M\^124 - 
                    68572\ M\^126 + 47423\ M\^128 + 75358\ M\^130 - 
                    136053\ M\^132 + 15040\ M\^134 + 111207\ M\^136 - 
                    103952\ M\^138 + 42526\ M\^140 + 103283\ M\^142 - 
                    164914\ M\^144 - 19239\ M\^146 + 167383\ M\^148 - 
                    61041\ M\^150 - 82846\ M\^152 + 73163\ M\^154 + 
                    7760\ M\^156 - 36093\ M\^158 + 11559\ M\^160 + 
                    8166\ M\^162 - 5298\ M\^164 - 1437\ M\^166 + 
                    184\ M\^168 + 4573\ M\^170 - 6435\ M\^172 + 
                    4870\ M\^174 - 2462\ M\^176 + 879\ M\^178 - 220\ M\^180 + 
                    36\ M\^182 - 3\ M\^184)\) + 
              L\^19\ \((M\^58 - 40\ M\^60 + 710\ M\^62 - 7592\ M\^64 + 
                    54968\ M\^66 - 283417\ M\^68 + 1054271\ M\^70 - 
                    2762059\ M\^72 + 4583325\ M\^74 - 2496545\ M\^76 - 
                    8967514\ M\^78 + 25650885\ M\^80 - 23401047\ M\^82 - 
                    19615274\ M\^84 + 69838447\ M\^86 - 49213440\ M\^88 - 
                    45427660\ M\^90 + 103005297\ M\^92 - 44803466\ M\^94 - 
                    78664111\ M\^96 + 110471091\ M\^98 + 58183417\ M\^100 - 
                    160217028\ M\^102 - 86062962\ M\^104 + 
                    233584917\ M\^106 + 68457680\ M\^108 - 
                    227430168\ M\^110 - 12015636\ M\^112 + 85856752\ M\^114 + 
                    29703786\ M\^116 + 41222179\ M\^118 - 96091386\ M\^120 - 
                    38291402\ M\^122 + 77775734\ M\^124 + 27176939\ M\^126 - 
                    12271602\ M\^128 - 101249109\ M\^130 + 50827938\ M\^132 + 
                    133684300\ M\^134 - 160681856\ M\^136 - 
                    29603609\ M\^138 + 176337000\ M\^140 - 
                    100255655\ M\^142 - 68985652\ M\^144 + 
                    118044622\ M\^146 - 31537847\ M\^148 - 51215793\ M\^150 + 
                    49940265\ M\^152 - 4835348\ M\^154 - 20861364\ M\^156 + 
                    15040200\ M\^158 - 823264\ M\^160 - 4888182\ M\^162 + 
                    2831128\ M\^164 + 371828\ M\^166 - 1593795\ M\^168 + 
                    1299554\ M\^170 - 669559\ M\^172 + 251847\ M\^174 - 
                    72047\ M\^176 + 15778\ M\^178 - 2610\ M\^180 + 
                    314\ M\^182 - 25\ M\^184 + M\^186)\) + 
              L\^20\ \((\(-2\)\ M\^66 + 62\ M\^68 - 934\ M\^70 + 
                    8883\ M\^72 - 58341\ M\^74 + 273899\ M\^76 - 
                    920728\ M\^78 + 2123362\ M\^80 - 2802207\ M\^82 - 
                    269973\ M\^84 + 9477730\ M\^86 - 17897260\ M\^88 + 
                    8119704\ M\^90 + 24449529\ M\^92 - 47134210\ M\^94 + 
                    19228056\ M\^96 + 44102231\ M\^98 - 68946706\ M\^100 + 
                    11013132\ M\^102 + 53541391\ M\^104 - 22247637\ M\^106 - 
                    15396485\ M\^108 - 44408391\ M\^110 + 31487211\ M\^112 + 
                    90194804\ M\^114 - 60455649\ M\^116 - 69189794\ M\^118 + 
                    46292290\ M\^120 + 33079499\ M\^122 - 26424563\ M\^124 + 
                    3782335\ M\^126 + 28382961\ M\^128 - 63007460\ M\^130 + 
                    3842602\ M\^132 + 99742525\ M\^134 - 84318730\ M\^136 - 
                    51135913\ M\^138 + 126656488\ M\^140 - 41919566\ M\^142 - 
                    81380008\ M\^144 + 86188109\ M\^146 + 4002813\ M\^148 - 
                    58754588\ M\^150 + 32211840\ M\^152 + 14198573\ M\^154 - 
                    24552570\ M\^156 + 6770147\ M\^158 + 7185785\ M\^160 - 
                    6797008\ M\^162 + 1200603\ M\^164 + 1423107\ M\^166 - 
                    829422\ M\^168 - 301375\ M\^170 + 678137\ M\^172 - 
                    502611\ M\^174 + 241245\ M\^176 - 84547\ M\^178 + 
                    22414\ M\^180 - 4489\ M\^182 + 655\ M\^184 - 63\ M\^186 + 
                    3\ M\^188)\) + 
              L\^24\ \((56\ M\^104 - 910\ M\^106 + 6511\ M\^108 - 
                    26105\ M\^110 + 59022\ M\^112 - 50751\ M\^114 - 
                    96677\ M\^116 + 322391\ M\^118 - 246827\ M\^120 - 
                    328460\ M\^122 + 717570\ M\^124 - 176117\ M\^126 - 
                    542902\ M\^128 + 444499\ M\^130 - 253277\ M\^132 + 
                    211541\ M\^134 + 674004\ M\^136 - 1149430\ M\^138 - 
                    54171\ M\^140 + 1334595\ M\^142 - 935796\ M\^144 - 
                    566037\ M\^146 + 1257866\ M\^148 - 330255\ M\^150 - 
                    783081\ M\^152 + 645221\ M\^154 + 151531\ M\^156 - 
                    423652\ M\^158 + 131951\ M\^160 + 117470\ M\^162 - 
                    104241\ M\^164 + 6659\ M\^166 + 33218\ M\^168 - 
                    26913\ M\^170 + 17603\ M\^172 - 14665\ M\^174 + 
                    11857\ M\^176 - 7360\ M\^178 + 3359\ M\^180 - 
                    1118\ M\^182 + 263\ M\^184 - 40\ M\^186 + 3\ M\^188)\) + 
              L\^21\ \((M\^74 - 38\ M\^76 + 629\ M\^78 - 6171\ M\^80 + 
                    40359\ M\^82 - 184551\ M\^84 + 591340\ M\^86 - 
                    1252202\ M\^88 + 1329867\ M\^90 + 1058913\ M\^92 - 
                    6358890\ M\^94 + 9156101\ M\^96 - 537736\ M\^98 - 
                    17234495\ M\^100 + 22717181\ M\^102 - 359045\ M\^104 - 
                    26346467\ M\^106 + 21532448\ M\^108 - 2176356\ M\^110 + 
                    3839128\ M\^112 + 12313539\ M\^114 - 58467356\ M\^116 + 
                    24545537\ M\^118 + 72161461\ M\^120 - 54452596\ M\^122 - 
                    41733142\ M\^124 + 43084044\ M\^126 + 19064179\ M\^128 - 
                    30429297\ M\^130 - 7619460\ M\^132 + 41909544\ M\^134 - 
                    21719598\ M\^136 - 41874560\ M\^138 + 55944155\ M\^140 + 
                    7239919\ M\^142 - 56595689\ M\^144 + 29380679\ M\^146 + 
                    25258907\ M\^148 - 35761693\ M\^150 + 4209193\ M\^152 + 
                    18771630\ M\^154 - 12132897\ M\^156 - 3097438\ M\^158 + 
                    7007507\ M\^160 - 2009390\ M\^162 - 1880773\ M\^164 + 
                    1742374\ M\^166 - 348516\ M\^168 - 184570\ M\^170 - 
                    7637\ M\^172 + 214824\ M\^174 - 214923\ M\^176 + 
                    123398\ M\^178 - 49515\ M\^180 + 14729\ M\^182 - 
                    3274\ M\^184 + 527\ M\^186 - 56\ M\^188 + 3\ M\^190)\) + 
              L\^23\ \((28\ M\^94 - 574\ M\^96 + 5347\ M\^98 - 
                    29340\ M\^100 + 101371\ M\^102 - 210365\ M\^104 + 
                    173096\ M\^106 + 327712\ M\^108 - 1135994\ M\^110 + 
                    1054969\ M\^112 + 854588\ M\^114 - 2807117\ M\^116 + 
                    1672732\ M\^118 + 1655487\ M\^120 - 2749287\ M\^122 + 
                    1335076\ M\^124 - 400087\ M\^126 - 1137062\ M\^128 + 
                    3102331\ M\^130 - 2069251\ M\^132 + 186088\ M\^134 + 
                    2146365\ M\^136 - 5550259\ M\^138 + 2907953\ M\^140 + 
                    5649936\ M\^142 - 7134617\ M\^144 - 912426\ M\^146 + 
                    6239369\ M\^148 - 2845576\ M\^150 - 2536264\ M\^152 + 
                    3013854\ M\^154 - 33650\ M\^156 - 1534389\ M\^158 + 
                    660956\ M\^160 + 410358\ M\^162 - 440662\ M\^164 + 
                    35030\ M\^166 + 115535\ M\^168 - 25405\ M\^170 - 
                    61388\ M\^172 + 70146\ M\^174 - 45411\ M\^176 + 
                    23809\ M\^178 - 11473\ M\^180 + 5003\ M\^182 - 
                    1831\ M\^184 + 525\ M\^186 - 110\ M\^188 + 15\ M\^190 - 
                    M\^192)\) + 
              L\^22\ \((8\ M\^84 - 210\ M\^86 + 2535\ M\^88 - 18416\ M\^90 + 
                    88178\ M\^92 - 283704\ M\^94 + 578147\ M\^96 - 
                    520241\ M\^98 - 754674\ M\^100 + 3112490\ M\^102 - 
                    3550731\ M\^104 - 1371006\ M\^106 + 8170879\ M\^108 - 
                    7032430\ M\^110 - 2722429\ M\^112 + 8318003\ M\^114 - 
                    6371519\ M\^116 + 6375171\ M\^118 - 15367\ M\^120 - 
                    19162717\ M\^122 + 16309379\ M\^124 + 13567934\ M\^126 - 
                    15519187\ M\^128 - 5982900\ M\^130 - 342254\ M\^132 + 
                    14793587\ M\^134 + 2515413\ M\^136 - 24741929\ M\^138 + 
                    14148627\ M\^140 + 15919556\ M\^142 - 25624630\ M\^144 + 
                    3586334\ M\^146 + 18492250\ M\^148 - 13685605\ M\^150 - 
                    4432792\ M\^152 + 10592762\ M\^154 - 3191562\ M\^156 - 
                    3652214\ M\^158 + 3270054\ M\^160 - 54068\ M\^162 - 
                    1162734\ M\^164 + 462729\ M\^166 + 232028\ M\^168 - 
                    299570\ M\^170 + 149813\ M\^172 - 80951\ M\^174 + 
                    73400\ M\^176 - 59619\ M\^178 + 34542\ M\^180 - 
                    14470\ M\^182 + 4487\ M\^184 - 1028\ M\^186 + 
                    168\ M\^188 - 18\ M\^190 + M\^192)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 55]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 56]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 57]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 58]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 59]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 60]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 61]\), \(\(-M\^10\) + 
              L\^15\ M\^130 + 
              L\ \((M\^2 - 3\ M\^4 + 9\ M\^6 - 17\ M\^8 + 20\ M\^10 - 
                    13\ M\^12 - 10\ M\^14 + 26\ M\^16 - 24\ M\^18 - 
                    11\ M\^20 + 9\ M\^22)\) + 
              L\^2\ \((1 - 5\ M\^2 + 17\ M\^4 - 43\ M\^6 + 80\ M\^8 - 
                    101\ M\^10 + 63\ M\^12 + 74\ M\^14 - 199\ M\^16 + 
                    155\ M\^18 + 80\ M\^20 - 366\ M\^22 + 291\ M\^24 + 
                    74\ M\^26 - 363\ M\^28 + 113\ M\^30 + 88\ M\^32 - 
                    36\ M\^34)\) + 
              L\^3\ \((\(-1\) + 6\ M\^2 - 23\ M\^4 + 63\ M\^6 - 120\ M\^8 + 
                    152\ M\^10 - 86\ M\^12 - 71\ M\^14 + 160\ M\^16 + 
                    7\ M\^18 - 391\ M\^20 + 702\ M\^22 - 470\ M\^24 - 
                    235\ M\^26 + 1083\ M\^28 - 1446\ M\^30 - 187\ M\^32 + 
                    2060\ M\^34 - 1703\ M\^36 - 890\ M\^38 + 1458\ M\^40 - 
                    117\ M\^42 - 308\ M\^44 + 84\ M\^46)\) + 
              L\^4\ \((\(-M\^6\) + 10\ M\^8 - 47\ M\^10 + 149\ M\^12 - 
                    350\ M\^14 + 549\ M\^16 - 466\ M\^18 - 72\ M\^20 + 
                    829\ M\^22 - 956\ M\^24 - 181\ M\^26 + 1854\ M\^28 - 
                    2226\ M\^30 + 452\ M\^32 + 1941\ M\^34 - 1113\ M\^36 - 
                    1796\ M\^38 - 731\ M\^40 + 3767\ M\^42 - 241\ M\^44 - 
                    6057\ M\^46 + 3112\ M\^48 + 3337\ M\^50 - 2561\ M\^52 - 
                    329\ M\^54 + 616\ M\^56 - 126\ M\^58)\) + 
              L\^5\ \((2\ M\^14 - 20\ M\^16 + 100\ M\^18 - 331\ M\^20 + 
                    765\ M\^22 - 1149\ M\^24 + 881\ M\^26 + 114\ M\^28 - 
                    1073\ M\^30 + 1372\ M\^32 - 657\ M\^34 - 29\ M\^36 + 
                    697\ M\^38 - 4296\ M\^40 + 7724\ M\^42 + 226\ M\^44 - 
                    12300\ M\^46 + 4992\ M\^48 + 7919\ M\^50 - 5001\ M\^52 - 
                    5777\ M\^54 + 646\ M\^56 + 8137\ M\^58 - 1096\ M\^60 - 
                    5130\ M\^62 + 1800\ M\^64 + 1127\ M\^66 - 770\ M\^68 + 
                    126\ M\^70)\) + 
              L\^6\ \((\(-M\^22\) + 11\ M\^24 - 69\ M\^26 + 294\ M\^28 - 
                    838\ M\^30 + 1537\ M\^32 - 1686\ M\^34 + 883\ M\^36 + 
                    395\ M\^38 - 2598\ M\^40 + 5003\ M\^42 - 1413\ M\^44 - 
                    6956\ M\^46 + 6839\ M\^48 + 2042\ M\^50 - 382\ M\^52 - 
                    8288\ M\^54 - 3655\ M\^56 + 15359\ M\^58 - 3902\ M\^60 - 
                    14297\ M\^62 + 9101\ M\^64 + 5848\ M\^66 - 1498\ M\^68 - 
                    3272\ M\^70 - 1842\ M\^72 + 2824\ M\^74 + 519\ M\^76 - 
                    1491\ M\^78 + 616\ M\^80 - 84\ M\^82)\) + 
              L\^7\ \((9\ M\^34 - 88\ M\^36 + 403\ M\^38 - 1137\ M\^40 + 
                    2055\ M\^42 - 1988\ M\^44 - 96\ M\^46 + 2469\ M\^48 - 
                    2727\ M\^50 + 2348\ M\^52 - 1555\ M\^54 - 3100\ M\^56 + 
                    10898\ M\^58 - 8180\ M\^60 - 6681\ M\^62 + 6965\ M\^64 - 
                    1373\ M\^66 + 877\ M\^68 - 1335\ M\^70 - 2903\ M\^72 + 
                    10306\ M\^74 - 3860\ M\^76 - 4351\ M\^78 + 3568\ M\^80 - 
                    1586\ M\^82 + 479\ M\^84 + 1111\ M\^86 - 1742\ M\^88 + 
                    1057\ M\^90 - 308\ M\^92 + 36\ M\^94)\) + 
              L\^8\ \((\(-36\)\ M\^46 + 308\ M\^48 - 1057\ M\^50 + 
                    1742\ M\^52 - 1111\ M\^54 - 479\ M\^56 + 1586\ M\^58 - 
                    3568\ M\^60 + 4351\ M\^62 + 3860\ M\^64 - 10306\ M\^66 + 
                    2903\ M\^68 + 1335\ M\^70 - 877\ M\^72 + 1373\ M\^74 - 
                    6965\ M\^76 + 6681\ M\^78 + 8180\ M\^80 - 10898\ M\^82 + 
                    3100\ M\^84 + 1555\ M\^86 - 2348\ M\^88 + 2727\ M\^90 - 
                    2469\ M\^92 + 96\ M\^94 + 1988\ M\^96 - 2055\ M\^98 + 
                    1137\ M\^100 - 403\ M\^102 + 88\ M\^104 - 9\ M\^106)\) + 
              L\^9\ \((84\ M\^58 - 616\ M\^60 + 1491\ M\^62 - 519\ M\^64 - 
                    2824\ M\^66 + 1842\ M\^68 + 3272\ M\^70 + 1498\ M\^72 - 
                    5848\ M\^74 - 9101\ M\^76 + 14297\ M\^78 + 3902\ M\^80 - 
                    15359\ M\^82 + 3655\ M\^84 + 8288\ M\^86 + 382\ M\^88 - 
                    2042\ M\^90 - 6839\ M\^92 + 6956\ M\^94 + 1413\ M\^96 - 
                    5003\ M\^98 + 2598\ M\^100 - 395\ M\^102 - 883\ M\^104 + 
                    1686\ M\^106 - 1537\ M\^108 + 838\ M\^110 - 294\ M\^112 + 
                    69\ M\^114 - 11\ M\^116 + M\^118)\) + 
              L\^10\ \((\(-126\)\ M\^70 + 770\ M\^72 - 1127\ M\^74 - 
                    1800\ M\^76 + 5130\ M\^78 + 1096\ M\^80 - 8137\ M\^82 - 
                    646\ M\^84 + 5777\ M\^86 + 5001\ M\^88 - 7919\ M\^90 - 
                    4992\ M\^92 + 12300\ M\^94 - 226\ M\^96 - 7724\ M\^98 + 
                    4296\ M\^100 - 697\ M\^102 + 29\ M\^104 + 657\ M\^106 - 
                    1372\ M\^108 + 1073\ M\^110 - 114\ M\^112 - 881\ M\^114 + 
                    1149\ M\^116 - 765\ M\^118 + 331\ M\^120 - 100\ M\^122 + 
                    20\ M\^124 - 2\ M\^126)\) + 
              L\^11\ \((126\ M\^82 - 616\ M\^84 + 329\ M\^86 + 2561\ M\^88 - 
                    3337\ M\^90 - 3112\ M\^92 + 6057\ M\^94 + 241\ M\^96 - 
                    3767\ M\^98 + 731\ M\^100 + 1796\ M\^102 + 1113\ M\^104 - 
                    1941\ M\^106 - 452\ M\^108 + 2226\ M\^110 - 
                    1854\ M\^112 + 181\ M\^114 + 956\ M\^116 - 829\ M\^118 + 
                    72\ M\^120 + 466\ M\^122 - 549\ M\^124 + 350\ M\^126 - 
                    149\ M\^128 + 47\ M\^130 - 10\ M\^132 + M\^134)\) + 
              L\^14\ \((\(-9\)\ M\^118 + 11\ M\^120 + 24\ M\^122 - 
                    26\ M\^124 + 10\ M\^126 + 13\ M\^128 - 20\ M\^130 + 
                    17\ M\^132 - 9\ M\^134 + 3\ M\^136 - M\^138)\) + 
              L\^13\ \((36\ M\^106 - 88\ M\^108 - 113\ M\^110 + 363\ M\^112 - 
                    74\ M\^114 - 291\ M\^116 + 366\ M\^118 - 80\ M\^120 - 
                    155\ M\^122 + 199\ M\^124 - 74\ M\^126 - 63\ M\^128 + 
                    101\ M\^130 - 80\ M\^132 + 43\ M\^134 - 17\ M\^136 + 
                    5\ M\^138 - M\^140)\) + 
              L\^12\ \((\(-84\)\ M\^94 + 308\ M\^96 + 117\ M\^98 - 
                    1458\ M\^100 + 890\ M\^102 + 1703\ M\^104 - 
                    2060\ M\^106 + 187\ M\^108 + 1446\ M\^110 - 
                    1083\ M\^112 + 235\ M\^114 + 470\ M\^116 - 702\ M\^118 + 
                    391\ M\^120 - 7\ M\^122 - 160\ M\^124 + 71\ M\^126 + 
                    86\ M\^128 - 152\ M\^130 + 120\ M\^132 - 63\ M\^134 + 
                    23\ M\^136 - 6\ M\^138 + M\^140)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 62]\), \(M\^12 + 
              L\^19\ M\^166 + 
              L\ \((2\ M\^6 - 7\ M\^8 + 18\ M\^10 - 32\ M\^12 + 33\ M\^14 - 
                    5\ M\^16 - 40\ M\^18 + 61\ M\^20 + 2\ M\^22 - 21\ M\^24 + 
                    8\ M\^26)\) + 
              L\^2\ \((1 - 6\ M\^2 + 21\ M\^4 - 52\ M\^6 + 99\ M\^8 - 
                    150\ M\^10 + 185\ M\^12 - 202\ M\^14 + 133\ M\^16 + 
                    283\ M\^18 - 879\ M\^20 + 824\ M\^22 + 331\ M\^24 - 
                    1557\ M\^26 + 1200\ M\^28 + 650\ M\^30 - 1096\ M\^32 + 
                    297\ M\^34 + 208\ M\^36 - 147\ M\^38 + 28\ M\^40)\) + 
              L\^3\ \((\(-1\) + 13\ M\^2 - 70\ M\^4 + 239\ M\^6 - 593\ M\^8 + 
                    1045\ M\^10 - 1131\ M\^12 + 264\ M\^14 + 1477\ M\^16 - 
                    3074\ M\^18 + 3315\ M\^20 - 1005\ M\^22 - 4416\ M\^24 + 
                    9531\ M\^26 - 7591\ M\^28 - 1862\ M\^30 + 12506\ M\^32 - 
                    15358\ M\^34 + 2799\ M\^36 + 15657\ M\^38 - 
                    14074\ M\^40 - 1217\ M\^42 + 8331\ M\^44 - 4106\ M\^46 - 
                    535\ M\^48 + 1210\ M\^50 - 441\ M\^52 + 56\ M\^54)\) + 
              L\^4\ \((\(-3\)\ M\^2 + 40\ M\^4 - 245\ M\^6 + 939\ M\^8 - 
                    2482\ M\^10 + 4397\ M\^12 - 4082\ M\^14 - 2113\ M\^16 + 
                    13660\ M\^18 - 21215\ M\^20 + 9483\ M\^22 + 
                    27969\ M\^24 - 65988\ M\^26 + 48448\ M\^28 + 
                    47187\ M\^30 - 132399\ M\^32 + 81655\ M\^34 + 
                    72162\ M\^36 - 150219\ M\^38 + 83241\ M\^40 + 
                    34373\ M\^42 - 109126\ M\^44 + 101414\ M\^46 - 
                    4071\ M\^48 - 80038\ M\^50 + 69777\ M\^52 - 3470\ M\^54 - 
                    33504\ M\^56 + 22698\ M\^58 - 1863\ M\^60 - 5088\ M\^62 + 
                    3004\ M\^64 - 735\ M\^66 + 70\ M\^68)\) + 
              L\^5\ \((\(-3\)\ M\^4 + 45\ M\^6 - 311\ M\^8 + 1315\ M\^10 - 
                    3732\ M\^12 + 7057\ M\^14 - 7496\ M\^16 - 616\ M\^18 + 
                    16737\ M\^20 - 28224\ M\^22 + 15799\ M\^24 + 
                    32263\ M\^26 - 88073\ M\^28 + 63038\ M\^30 + 
                    86799\ M\^32 - 200980\ M\^34 + 61063\ M\^36 + 
                    215993\ M\^38 - 265355\ M\^40 + 44058\ M\^42 + 
                    207479\ M\^44 - 383962\ M\^46 + 273654\ M\^48 + 
                    314458\ M\^50 - 719977\ M\^52 + 232550\ M\^54 + 
                    549023\ M\^56 - 573059\ M\^58 + 6777\ M\^60 + 
                    348872\ M\^62 - 244887\ M\^64 - 1006\ M\^66 + 
                    107507\ M\^68 - 72385\ M\^70 + 11141\ M\^72 + 
                    14025\ M\^74 - 11290\ M\^76 + 4010\ M\^78 - 735\ M\^80 + 
                    56\ M\^82)\) + 
              L\^6\ \((\(-M\^6\) + 17\ M\^8 - 130\ M\^10 + 596\ M\^12 - 
                    1823\ M\^14 + 3877\ M\^16 - 5909\ M\^18 + 7443\ M\^20 - 
                    10374\ M\^22 + 12001\ M\^24 + 5850\ M\^26 - 
                    52460\ M\^28 + 81842\ M\^30 - 21595\ M\^32 - 
                    136496\ M\^34 + 275708\ M\^36 - 136280\ M\^38 - 
                    392089\ M\^40 + 688826\ M\^42 + 74062\ M\^44 - 
                    1144206\ M\^46 + 580908\ M\^48 + 1229473\ M\^50 - 
                    1176846\ M\^52 - 1090598\ M\^54 + 1603193\ M\^56 + 
                    820710\ M\^58 - 1972197\ M\^60 - 76770\ M\^62 + 
                    1983629\ M\^64 - 1053428\ M\^66 - 1036978\ M\^68 + 
                    1534567\ M\^70 - 249684\ M\^72 - 848974\ M\^74 + 
                    680040\ M\^76 - 3405\ M\^78 - 286903\ M\^80 + 
                    180445\ M\^82 - 23554\ M\^84 - 34751\ M\^86 + 
                    28736\ M\^88 - 11959\ M\^90 + 3032\ M\^92 - 441\ M\^94 + 
                    28\ M\^96)\) + 
              L\^7\ \((M\^14 - 21\ M\^16 + 216\ M\^18 - 1382\ M\^20 + 
                    5814\ M\^22 - 15883\ M\^24 + 25339\ M\^26 - 
                    11394\ M\^28 - 40265\ M\^30 + 81140\ M\^32 - 
                    17048\ M\^34 - 153789\ M\^36 + 232287\ M\^38 + 
                    68129\ M\^40 - 638681\ M\^42 + 535498\ M\^44 + 
                    888492\ M\^46 - 1831886\ M\^48 - 229171\ M\^50 + 
                    3055487\ M\^52 - 1260903\ M\^54 - 3469695\ M\^56 + 
                    2726604\ M\^58 + 3236310\ M\^60 - 3730760\ M\^62 - 
                    2584327\ M\^64 + 4327333\ M\^66 + 1136294\ M\^68 - 
                    3980244\ M\^70 + 1020766\ M\^72 + 2157950\ M\^74 - 
                    2400816\ M\^76 + 535249\ M\^78 + 1665075\ M\^80 - 
                    1983117\ M\^82 + 229695\ M\^84 + 1268734\ M\^86 - 
                    1051035\ M\^88 + 19061\ M\^90 + 489244\ M\^92 - 
                    325079\ M\^94 + 39156\ M\^96 + 69682\ M\^98 - 
                    55436\ M\^100 + 23128\ M\^102 - 6463\ M\^104 + 
                    1238\ M\^106 - 147\ M\^108 + 8\ M\^110)\) + 
              L\^8\ \((4\ M\^24 - 90\ M\^26 + 910\ M\^28 - 5406\ M\^30 + 
                    20405\ M\^32 - 47972\ M\^34 + 56431\ M\^36 + 
                    26553\ M\^38 - 206058\ M\^40 + 269652\ M\^42 + 
                    88402\ M\^44 - 774050\ M\^46 + 856857\ M\^48 + 
                    768855\ M\^50 - 2786656\ M\^52 + 924429\ M\^54 + 
                    4665721\ M\^56 - 4699762\ M\^58 - 4585385\ M\^60 + 
                    8712933\ M\^62 + 2310381\ M\^64 - 10709525\ M\^66 + 
                    477410\ M\^68 + 10047401\ M\^70 - 2211339\ M\^72 - 
                    7088513\ M\^74 + 2036361\ M\^76 + 2961111\ M\^78 + 
                    505435\ M\^80 - 24148\ M\^82 - 3743451\ M\^84 + 
                    85653\ M\^86 + 5043995\ M\^88 - 2066046\ M\^90 - 
                    3277319\ M\^92 + 3125256\ M\^94 + 326467\ M\^96 - 
                    1978194\ M\^98 + 1102569\ M\^100 + 209737\ M\^102 - 
                    629770\ M\^104 + 353947\ M\^106 - 20478\ M\^108 - 
                    97691\ M\^110 + 74036\ M\^112 - 30291\ M\^114 + 
                    8185\ M\^116 - 1570\ M\^118 + 220\ M\^120 - 21\ M\^122 + 
                    M\^124)\) + 
              L\^9\ \((6\ M\^34 - 142\ M\^36 + 1476\ M\^38 - 8758\ M\^40 + 
                    32109\ M\^42 - 71902\ M\^44 + 79699\ M\^46 + 
                    42052\ M\^48 - 309316\ M\^50 + 456647\ M\^52 + 
                    34509\ M\^54 - 1174075\ M\^56 + 1314872\ M\^58 + 
                    1190263\ M\^60 - 3195319\ M\^62 - 222430\ M\^64 + 
                    4206819\ M\^66 - 379569\ M\^68 - 3611110\ M\^70 - 
                    1780933\ M\^72 + 3159784\ M\^74 + 6890808\ M\^76 - 
                    5084917\ M\^78 - 12140857\ M\^80 + 9074442\ M\^82 + 
                    14928062\ M\^84 - 13061289\ M\^86 - 14112845\ M\^88 + 
                    15487525\ M\^90 + 9473496\ M\^92 - 14327208\ M\^94 - 
                    2968057\ M\^96 + 9249485\ M\^98 - 1549877\ M\^100 - 
                    3004092\ M\^102 + 2159509\ M\^104 - 851460\ M\^106 - 
                    483708\ M\^108 + 1369181\ M\^110 - 764302\ M\^112 - 
                    307931\ M\^114 + 582643\ M\^116 - 247979\ M\^118 - 
                    42327\ M\^120 + 104350\ M\^122 - 63498\ M\^124 + 
                    23537\ M\^126 - 5867\ M\^128 + 961\ M\^130 - 93\ M\^132 + 
                    4\ M\^134)\) + 
              L\^10\ \((4\ M\^44 - 93\ M\^46 + 961\ M\^48 - 5867\ M\^50 + 
                    23537\ M\^52 - 63498\ M\^54 + 104350\ M\^56 - 
                    42327\ M\^58 - 247979\ M\^60 + 582643\ M\^62 - 
                    307931\ M\^64 - 764302\ M\^66 + 1369181\ M\^68 - 
                    483708\ M\^70 - 851460\ M\^72 + 2159509\ M\^74 - 
                    3004092\ M\^76 - 1549877\ M\^78 + 9249485\ M\^80 - 
                    2968057\ M\^82 - 14327208\ M\^84 + 9473496\ M\^86 + 
                    15487525\ M\^88 - 14112845\ M\^90 - 13061289\ M\^92 + 
                    14928062\ M\^94 + 9074442\ M\^96 - 12140857\ M\^98 - 
                    5084917\ M\^100 + 6890808\ M\^102 + 3159784\ M\^104 - 
                    1780933\ M\^106 - 3611110\ M\^108 - 379569\ M\^110 + 
                    4206819\ M\^112 - 222430\ M\^114 - 3195319\ M\^116 + 
                    1190263\ M\^118 + 1314872\ M\^120 - 1174075\ M\^122 + 
                    34509\ M\^124 + 456647\ M\^126 - 309316\ M\^128 + 
                    42052\ M\^130 + 79699\ M\^132 - 71902\ M\^134 + 
                    32109\ M\^136 - 8758\ M\^138 + 1476\ M\^140 - 
                    142\ M\^142 + 6\ M\^144)\) + 
              L\^11\ \((M\^54 - 21\ M\^56 + 220\ M\^58 - 1570\ M\^60 + 
                    8185\ M\^62 - 30291\ M\^64 + 74036\ M\^66 - 
                    97691\ M\^68 - 20478\ M\^70 + 353947\ M\^72 - 
                    629770\ M\^74 + 209737\ M\^76 + 1102569\ M\^78 - 
                    1978194\ M\^80 + 326467\ M\^82 + 3125256\ M\^84 - 
                    3277319\ M\^86 - 2066046\ M\^88 + 5043995\ M\^90 + 
                    85653\ M\^92 - 3743451\ M\^94 - 24148\ M\^96 + 
                    505435\ M\^98 + 2961111\ M\^100 + 2036361\ M\^102 - 
                    7088513\ M\^104 - 2211339\ M\^106 + 10047401\ M\^108 + 
                    477410\ M\^110 - 10709525\ M\^112 + 2310381\ M\^114 + 
                    8712933\ M\^116 - 4585385\ M\^118 - 4699762\ M\^120 + 
                    4665721\ M\^122 + 924429\ M\^124 - 2786656\ M\^126 + 
                    768855\ M\^128 + 856857\ M\^130 - 774050\ M\^132 + 
                    88402\ M\^134 + 269652\ M\^136 - 206058\ M\^138 + 
                    26553\ M\^140 + 56431\ M\^142 - 47972\ M\^144 + 
                    20405\ M\^146 - 5406\ M\^148 + 910\ M\^150 - 90\ M\^152 + 
                    4\ M\^154)\) + 
              L\^12\ \((8\ M\^68 - 147\ M\^70 + 1238\ M\^72 - 6463\ M\^74 + 
                    23128\ M\^76 - 55436\ M\^78 + 69682\ M\^80 + 
                    39156\ M\^82 - 325079\ M\^84 + 489244\ M\^86 + 
                    19061\ M\^88 - 1051035\ M\^90 + 1268734\ M\^92 + 
                    229695\ M\^94 - 1983117\ M\^96 + 1665075\ M\^98 + 
                    535249\ M\^100 - 2400816\ M\^102 + 2157950\ M\^104 + 
                    1020766\ M\^106 - 3980244\ M\^108 + 1136294\ M\^110 + 
                    4327333\ M\^112 - 2584327\ M\^114 - 3730760\ M\^116 + 
                    3236310\ M\^118 + 2726604\ M\^120 - 3469695\ M\^122 - 
                    1260903\ M\^124 + 3055487\ M\^126 - 229171\ M\^128 - 
                    1831886\ M\^130 + 888492\ M\^132 + 535498\ M\^134 - 
                    638681\ M\^136 + 68129\ M\^138 + 232287\ M\^140 - 
                    153789\ M\^142 - 17048\ M\^144 + 81140\ M\^146 - 
                    40265\ M\^148 - 11394\ M\^150 + 25339\ M\^152 - 
                    15883\ M\^154 + 5814\ M\^156 - 1382\ M\^158 + 
                    216\ M\^160 - 21\ M\^162 + M\^164)\) + 
              L\^13\ \((28\ M\^82 - 441\ M\^84 + 3032\ M\^86 - 11959\ M\^88 + 
                    28736\ M\^90 - 34751\ M\^92 - 23554\ M\^94 + 
                    180445\ M\^96 - 286903\ M\^98 - 3405\ M\^100 + 
                    680040\ M\^102 - 848974\ M\^104 - 249684\ M\^106 + 
                    1534567\ M\^108 - 1036978\ M\^110 - 1053428\ M\^112 + 
                    1983629\ M\^114 - 76770\ M\^116 - 1972197\ M\^118 + 
                    820710\ M\^120 + 1603193\ M\^122 - 1090598\ M\^124 - 
                    1176846\ M\^126 + 1229473\ M\^128 + 580908\ M\^130 - 
                    1144206\ M\^132 + 74062\ M\^134 + 688826\ M\^136 - 
                    392089\ M\^138 - 136280\ M\^140 + 275708\ M\^142 - 
                    136496\ M\^144 - 21595\ M\^146 + 81842\ M\^148 - 
                    52460\ M\^150 + 5850\ M\^152 + 12001\ M\^154 - 
                    10374\ M\^156 + 7443\ M\^158 - 5909\ M\^160 + 
                    3877\ M\^162 - 1823\ M\^164 + 596\ M\^166 - 130\ M\^168 + 
                    17\ M\^170 - M\^172)\) + 
              L\^18\ \((8\ M\^152 - 21\ M\^154 + 2\ M\^156 + 61\ M\^158 - 
                    40\ M\^160 - 5\ M\^162 + 33\ M\^164 - 32\ M\^166 + 
                    18\ M\^168 - 7\ M\^170 + 2\ M\^172)\) + 
              L\^14\ \((56\ M\^96 - 735\ M\^98 + 4010\ M\^100 - 
                    11290\ M\^102 + 14025\ M\^104 + 11141\ M\^106 - 
                    72385\ M\^108 + 107507\ M\^110 - 1006\ M\^112 - 
                    244887\ M\^114 + 348872\ M\^116 + 6777\ M\^118 - 
                    573059\ M\^120 + 549023\ M\^122 + 232550\ M\^124 - 
                    719977\ M\^126 + 314458\ M\^128 + 273654\ M\^130 - 
                    383962\ M\^132 + 207479\ M\^134 + 44058\ M\^136 - 
                    265355\ M\^138 + 215993\ M\^140 + 61063\ M\^142 - 
                    200980\ M\^144 + 86799\ M\^146 + 63038\ M\^148 - 
                    88073\ M\^150 + 32263\ M\^152 + 15799\ M\^154 - 
                    28224\ M\^156 + 16737\ M\^158 - 616\ M\^160 - 
                    7496\ M\^162 + 7057\ M\^164 - 3732\ M\^166 + 
                    1315\ M\^168 - 311\ M\^170 + 45\ M\^172 - 3\ M\^174)\) + 
              L\^15\ \((70\ M\^110 - 735\ M\^112 + 3004\ M\^114 - 
                    5088\ M\^116 - 1863\ M\^118 + 22698\ M\^120 - 
                    33504\ M\^122 - 3470\ M\^124 + 69777\ M\^126 - 
                    80038\ M\^128 - 4071\ M\^130 + 101414\ M\^132 - 
                    109126\ M\^134 + 34373\ M\^136 + 83241\ M\^138 - 
                    150219\ M\^140 + 72162\ M\^142 + 81655\ M\^144 - 
                    132399\ M\^146 + 47187\ M\^148 + 48448\ M\^150 - 
                    65988\ M\^152 + 27969\ M\^154 + 9483\ M\^156 - 
                    21215\ M\^158 + 13660\ M\^160 - 2113\ M\^162 - 
                    4082\ M\^164 + 4397\ M\^166 - 2482\ M\^168 + 
                    939\ M\^170 - 245\ M\^172 + 40\ M\^174 - 3\ M\^176)\) + 
              L\^16\ \((56\ M\^124 - 441\ M\^126 + 1210\ M\^128 - 
                    535\ M\^130 - 4106\ M\^132 + 8331\ M\^134 - 
                    1217\ M\^136 - 14074\ M\^138 + 15657\ M\^140 + 
                    2799\ M\^142 - 15358\ M\^144 + 12506\ M\^146 - 
                    1862\ M\^148 - 7591\ M\^150 + 9531\ M\^152 - 
                    4416\ M\^154 - 1005\ M\^156 + 3315\ M\^158 - 
                    3074\ M\^160 + 1477\ M\^162 + 264\ M\^164 - 
                    1131\ M\^166 + 1045\ M\^168 - 593\ M\^170 + 239\ M\^172 - 
                    70\ M\^174 + 13\ M\^176 - M\^178)\) + 
              L\^17\ \((28\ M\^138 - 147\ M\^140 + 208\ M\^142 + 
                    297\ M\^144 - 1096\ M\^146 + 650\ M\^148 + 1200\ M\^150 - 
                    1557\ M\^152 + 331\ M\^154 + 824\ M\^156 - 879\ M\^158 + 
                    283\ M\^160 + 133\ M\^162 - 202\ M\^164 + 185\ M\^166 - 
                    150\ M\^168 + 99\ M\^170 - 52\ M\^172 + 21\ M\^174 - 
                    6\ M\^176 + M\^178)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 63]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 64]\), \(M\^30 + 
              L\^20\ M\^106 + 
              L\ \((\(-3\)\ M\^22 + 11\ M\^24 - 30\ M\^26 + 50\ M\^28 - 
                    25\ M\^30 - 33\ M\^32 + 69\ M\^34 - 16\ M\^36 - 
                    11\ M\^38 + 12\ M\^40 - 4\ M\^42)\) + 
              L\^2\ \((3\ M\^14 - 20\ M\^16 + 79\ M\^18 - 209\ M\^20 + 
                    358\ M\^22 - 371\ M\^24 + 49\ M\^26 + 582\ M\^28 - 
                    1049\ M\^30 + 619\ M\^32 + 682\ M\^34 - 1341\ M\^36 + 
                    776\ M\^38 + 386\ M\^40 - 643\ M\^42 + 418\ M\^44 - 
                    97\ M\^46 - 92\ M\^48 + 90\ M\^50 - 36\ M\^52 + 
                    6\ M\^54)\) + 
              L\^3\ \((\(-M\^6\) + 9\ M\^8 - 44\ M\^10 + 147\ M\^12 - 
                    361\ M\^14 + 704\ M\^16 - 1152\ M\^18 + 1542\ M\^20 - 
                    1266\ M\^22 - 812\ M\^24 + 4940\ M\^26 - 7628\ M\^28 + 
                    1652\ M\^30 + 12840\ M\^32 - 18806\ M\^34 + 1044\ M\^36 + 
                    23234\ M\^38 - 20055\ M\^40 - 5274\ M\^42 + 
                    19852\ M\^44 - 9785\ M\^46 - 3518\ M\^48 + 6712\ M\^50 - 
                    3491\ M\^52 + 414\ M\^54 + 590\ M\^56 - 595\ M\^58 + 
                    364\ M\^60 - 148\ M\^62 + 36\ M\^64 - 4\ M\^66)\) + 
              L\^4\ \((\(-3\)\ M\^4 + 34\ M\^6 - 198\ M\^8 + 745\ M\^10 - 
                    1927\ M\^12 + 3373\ M\^14 - 3428\ M\^16 + 5\ M\^18 + 
                    6169\ M\^20 - 8738\ M\^22 - 70\ M\^24 + 18405\ M\^26 - 
                    30085\ M\^28 + 12258\ M\^30 + 43640\ M\^32 - 
                    92698\ M\^34 + 39488\ M\^36 + 111786\ M\^38 - 
                    167603\ M\^40 + 11813\ M\^42 + 160160\ M\^44 - 
                    119339\ M\^46 - 23837\ M\^48 + 78101\ M\^50 - 
                    40984\ M\^52 + 8920\ M\^54 + 4347\ M\^56 - 15635\ M\^58 + 
                    16480\ M\^60 - 6508\ M\^62 - 1967\ M\^64 + 3687\ M\^66 - 
                    2198\ M\^68 + 869\ M\^70 - 276\ M\^72 + 70\ M\^74 - 
                    12\ M\^76 + M\^78)\) + 
              L\^5\ \((\(-3\)\ M\^2 + 41\ M\^4 - 272\ M\^6 + 1137\ M\^8 - 
                    3221\ M\^10 + 6120\ M\^12 - 6880\ M\^14 + 1478\ M\^16 + 
                    8033\ M\^18 - 8187\ M\^20 - 12557\ M\^22 + 35979\ M\^24 - 
                    19235\ M\^26 - 48031\ M\^28 + 102908\ M\^30 - 
                    35929\ M\^32 - 179783\ M\^34 + 279348\ M\^36 + 
                    89008\ M\^38 - 587922\ M\^40 + 295245\ M\^42 + 
                    638269\ M\^44 - 739116\ M\^46 - 252557\ M\^48 + 
                    744073\ M\^50 - 91319\ M\^52 - 367191\ M\^54 + 
                    63939\ M\^56 + 163449\ M\^58 + 30930\ M\^60 - 
                    158381\ M\^62 + 38494\ M\^64 + 74926\ M\^66 - 
                    66516\ M\^68 + 17357\ M\^70 + 9026\ M\^72 - 
                    12554\ M\^74 + 7969\ M\^76 - 3383\ M\^78 + 975\ M\^80 - 
                    182\ M\^82 + 20\ M\^84 - M\^86)\) + 
              L\^6\ \((\(-1\) + 16\ M\^2 - 119\ M\^4 + 552\ M\^6 - 
                    1759\ M\^8 + 4026\ M\^10 - 7011\ M\^12 + 10565\ M\^14 - 
                    15966\ M\^16 + 21156\ M\^18 - 10055\ M\^20 - 
                    36232\ M\^22 + 83005\ M\^24 - 29773\ M\^26 - 
                    137155\ M\^28 + 224927\ M\^30 - 47564\ M\^32 - 
                    318295\ M\^34 + 547638\ M\^36 - 130903\ M\^38 - 
                    971531\ M\^40 + 1235339\ M\^42 + 699553\ M\^44 - 
                    2442944\ M\^46 + 435366\ M\^48 + 2933099\ M\^50 - 
                    1858168\ M\^52 - 2210389\ M\^54 + 2523921\ M\^56 + 
                    1019260\ M\^58 - 2113779\ M\^60 - 191486\ M\^62 + 
                    1386270\ M\^64 - 264254\ M\^66 - 695555\ M\^68 + 
                    453907\ M\^70 + 71727\ M\^72 - 240047\ M\^74 + 
                    148008\ M\^76 - 23381\ M\^78 - 36529\ M\^80 + 
                    37267\ M\^82 - 18813\ M\^84 + 5920\ M\^86 - 1187\ M\^88 + 
                    142\ M\^90 - 8\ M\^92)\) + 
              L\^7\ \((\(-M\^2\) + 15\ M\^4 - 115\ M\^6 + 618\ M\^8 - 
                    2558\ M\^10 + 8104\ M\^12 - 18656\ M\^14 + 28218\ M\^16 - 
                    20149\ M\^18 - 10747\ M\^20 + 20178\ M\^22 + 
                    45758\ M\^24 - 100939\ M\^26 - 56449\ M\^28 + 
                    344040\ M\^30 - 265681\ M\^32 - 387251\ M\^34 + 
                    990170\ M\^36 - 568810\ M\^38 - 1198183\ M\^40 + 
                    2629372\ M\^42 - 507987\ M\^44 - 4335127\ M\^46 + 
                    4240131\ M\^48 + 3692558\ M\^50 - 8001557\ M\^52 - 
                    438876\ M\^54 + 9544969\ M\^56 - 3758992\ M\^58 - 
                    7719459\ M\^60 + 6489957\ M\^62 + 3844654\ M\^64 - 
                    6244539\ M\^66 - 301409\ M\^68 + 3933905\ M\^70 - 
                    1459461\ M\^72 - 1334879\ M\^74 + 1435631\ M\^76 - 
                    299879\ M\^78 - 447056\ M\^80 + 483362\ M\^82 - 
                    178407\ M\^84 - 57522\ M\^86 + 104282\ M\^88 - 
                    60176\ M\^90 + 20171\ M\^92 - 4187\ M\^94 + 507\ M\^96 - 
                    28\ M\^98)\) + 
              L\^8\ \((\(-8\)\ M\^8 + 137\ M\^10 - 1091\ M\^12 + 
                    5352\ M\^14 - 18050\ M\^16 + 43255\ M\^18 - 
                    69523\ M\^20 + 52262\ M\^22 + 56944\ M\^24 - 
                    188449\ M\^26 + 98190\ M\^28 + 303932\ M\^30 - 
                    506857\ M\^32 - 136746\ M\^34 + 1152080\ M\^36 - 
                    1004199\ M\^38 - 950342\ M\^40 + 3031630\ M\^42 - 
                    1918041\ M\^44 - 3589674\ M\^46 + 7225528\ M\^48 - 
                    165674\ M\^50 - 11485363\ M\^52 + 8018406\ M\^54 + 
                    10549437\ M\^56 - 15566005\ M\^58 - 4071070\ M\^60 + 
                    18383948\ M\^62 - 4276121\ M\^64 - 14707094\ M\^66 + 
                    9961410\ M\^68 + 7051045\ M\^70 - 10092405\ M\^72 - 
                    76280\ M\^74 + 5954155\ M\^76 - 2964337\ M\^78 - 
                    1362233\ M\^80 + 2378548\ M\^82 - 1018669\ M\^84 - 
                    478831\ M\^86 + 892294\ M\^88 - 436446\ M\^90 - 
                    54343\ M\^92 + 191680\ M\^94 - 120355\ M\^96 + 
                    41692\ M\^98 - 8743\ M\^100 + 1050\ M\^102 - 
                    56\ M\^104)\) + 
              L\^9\ \((\(-28\)\ M\^14 + 517\ M\^16 - 4257\ M\^18 + 
                    20482\ M\^20 - 62822\ M\^22 + 121803\ M\^24 - 
                    120478\ M\^26 - 56336\ M\^28 + 368686\ M\^30 - 
                    438055\ M\^32 - 148630\ M\^34 + 1023523\ M\^36 - 
                    921640\ M\^38 - 767693\ M\^40 + 2507939\ M\^42 - 
                    1679127\ M\^44 - 2451160\ M\^46 + 6083803\ M\^48 - 
                    2317368\ M\^50 - 8487563\ M\^52 + 11229740\ M\^54 + 
                    4516455\ M\^56 - 18995623\ M\^58 + 5703286\ M\^60 + 
                    19624926\ M\^62 - 16363140\ M\^64 - 11977088\ M\^66 + 
                    21378127\ M\^68 + 750987\ M\^70 - 17974658\ M\^72 + 
                    7825920\ M\^74 + 9019939\ M\^76 - 9614775\ M\^78 - 
                    506419\ M\^80 + 5728171\ M\^82 - 3169073\ M\^84 - 
                    974773\ M\^86 + 2492204\ M\^88 - 1354583\ M\^90 - 
                    414684\ M\^92 + 1050451\ M\^94 - 549033\ M\^96 - 
                    61702\ M\^98 + 242109\ M\^100 - 154084\ M\^102 + 
                    53664\ M\^104 - 11251\ M\^106 + 1341\ M\^108 - 
                    70\ M\^110)\) + 
              L\^19\ \((\(-4\)\ M\^94 + 12\ M\^96 - 11\ M\^98 - 16\ M\^100 + 
                    69\ M\^102 - 33\ M\^104 - 25\ M\^106 + 50\ M\^108 - 
                    30\ M\^110 + 11\ M\^112 - 3\ M\^114)\) + 
              L\^10\ \((\(-56\)\ M\^20 + 1070\ M\^22 - 8958\ M\^24 + 
                    42860\ M\^26 - 125414\ M\^28 + 210677\ M\^30 - 
                    106704\ M\^32 - 346234\ M\^34 + 810490\ M\^36 - 
                    472496\ M\^38 - 861523\ M\^40 + 1897979\ M\^42 - 
                    925410\ M\^44 - 2036228\ M\^46 + 4159775\ M\^48 - 
                    1446084\ M\^50 - 5805945\ M\^52 + 8635362\ M\^54 + 
                    1585710\ M\^56 - 14830046\ M\^58 + 8564448\ M\^60 + 
                    14375499\ M\^62 - 19332494\ M\^64 - 5894221\ M\^66 + 
                    24000642\ M\^68 - 5894221\ M\^70 - 19332494\ M\^72 + 
                    14375499\ M\^74 + 8564448\ M\^76 - 14830046\ M\^78 + 
                    1585710\ M\^80 + 8635362\ M\^82 - 5805945\ M\^84 - 
                    1446084\ M\^86 + 4159775\ M\^88 - 2036228\ M\^90 - 
                    925410\ M\^92 + 1897979\ M\^94 - 861523\ M\^96 - 
                    472496\ M\^98 + 810490\ M\^100 - 346234\ M\^102 - 
                    106704\ M\^104 + 210677\ M\^106 - 125414\ M\^108 + 
                    42860\ M\^110 - 8958\ M\^112 + 1070\ M\^114 - 
                    56\ M\^116)\) + 
              L\^11\ \((\(-70\)\ M\^26 + 1341\ M\^28 - 11251\ M\^30 + 
                    53664\ M\^32 - 154084\ M\^34 + 242109\ M\^36 - 
                    61702\ M\^38 - 549033\ M\^40 + 1050451\ M\^42 - 
                    414684\ M\^44 - 1354583\ M\^46 + 2492204\ M\^48 - 
                    974773\ M\^50 - 3169073\ M\^52 + 5728171\ M\^54 - 
                    506419\ M\^56 - 9614775\ M\^58 + 9019939\ M\^60 + 
                    7825920\ M\^62 - 17974658\ M\^64 + 750987\ M\^66 + 
                    21378127\ M\^68 - 11977088\ M\^70 - 16363140\ M\^72 + 
                    19624926\ M\^74 + 5703286\ M\^76 - 18995623\ M\^78 + 
                    4516455\ M\^80 + 11229740\ M\^82 - 8487563\ M\^84 - 
                    2317368\ M\^86 + 6083803\ M\^88 - 2451160\ M\^90 - 
                    1679127\ M\^92 + 2507939\ M\^94 - 767693\ M\^96 - 
                    921640\ M\^98 + 1023523\ M\^100 - 148630\ M\^102 - 
                    438055\ M\^104 + 368686\ M\^106 - 56336\ M\^108 - 
                    120478\ M\^110 + 121803\ M\^112 - 62822\ M\^114 + 
                    20482\ M\^116 - 4257\ M\^118 + 517\ M\^120 - 
                    28\ M\^122)\) + 
              L\^18\ \((6\ M\^82 - 36\ M\^84 + 90\ M\^86 - 92\ M\^88 - 
                    97\ M\^90 + 418\ M\^92 - 643\ M\^94 + 386\ M\^96 + 
                    776\ M\^98 - 1341\ M\^100 + 682\ M\^102 + 619\ M\^104 - 
                    1049\ M\^106 + 582\ M\^108 + 49\ M\^110 - 371\ M\^112 + 
                    358\ M\^114 - 209\ M\^116 + 79\ M\^118 - 20\ M\^120 + 
                    3\ M\^122)\) + 
              L\^12\ \((\(-56\)\ M\^32 + 1050\ M\^34 - 8743\ M\^36 + 
                    41692\ M\^38 - 120355\ M\^40 + 191680\ M\^42 - 
                    54343\ M\^44 - 436446\ M\^46 + 892294\ M\^48 - 
                    478831\ M\^50 - 1018669\ M\^52 + 2378548\ M\^54 - 
                    1362233\ M\^56 - 2964337\ M\^58 + 5954155\ M\^60 - 
                    76280\ M\^62 - 10092405\ M\^64 + 7051045\ M\^66 + 
                    9961410\ M\^68 - 14707094\ M\^70 - 4276121\ M\^72 + 
                    18383948\ M\^74 - 4071070\ M\^76 - 15566005\ M\^78 + 
                    10549437\ M\^80 + 8018406\ M\^82 - 11485363\ M\^84 - 
                    165674\ M\^86 + 7225528\ M\^88 - 3589674\ M\^90 - 
                    1918041\ M\^92 + 3031630\ M\^94 - 950342\ M\^96 - 
                    1004199\ M\^98 + 1152080\ M\^100 - 136746\ M\^102 - 
                    506857\ M\^104 + 303932\ M\^106 + 98190\ M\^108 - 
                    188449\ M\^110 + 56944\ M\^112 + 52262\ M\^114 - 
                    69523\ M\^116 + 43255\ M\^118 - 18050\ M\^120 + 
                    5352\ M\^122 - 1091\ M\^124 + 137\ M\^126 - 
                    8\ M\^128)\) + 
              L\^17\ \((\(-4\)\ M\^70 + 36\ M\^72 - 148\ M\^74 + 364\ M\^76 - 
                    595\ M\^78 + 590\ M\^80 + 414\ M\^82 - 3491\ M\^84 + 
                    6712\ M\^86 - 3518\ M\^88 - 9785\ M\^90 + 19852\ M\^92 - 
                    5274\ M\^94 - 20055\ M\^96 + 23234\ M\^98 + 
                    1044\ M\^100 - 18806\ M\^102 + 12840\ M\^104 + 
                    1652\ M\^106 - 7628\ M\^108 + 4940\ M\^110 - 
                    812\ M\^112 - 1266\ M\^114 + 1542\ M\^116 - 
                    1152\ M\^118 + 704\ M\^120 - 361\ M\^122 + 147\ M\^124 - 
                    44\ M\^126 + 9\ M\^128 - M\^130)\) + 
              L\^16\ \((M\^58 - 12\ M\^60 + 70\ M\^62 - 276\ M\^64 + 
                    869\ M\^66 - 2198\ M\^68 + 3687\ M\^70 - 1967\ M\^72 - 
                    6508\ M\^74 + 16480\ M\^76 - 15635\ M\^78 + 4347\ M\^80 + 
                    8920\ M\^82 - 40984\ M\^84 + 78101\ M\^86 - 
                    23837\ M\^88 - 119339\ M\^90 + 160160\ M\^92 + 
                    11813\ M\^94 - 167603\ M\^96 + 111786\ M\^98 + 
                    39488\ M\^100 - 92698\ M\^102 + 43640\ M\^104 + 
                    12258\ M\^106 - 30085\ M\^108 + 18405\ M\^110 - 
                    70\ M\^112 - 8738\ M\^114 + 6169\ M\^116 + 5\ M\^118 - 
                    3428\ M\^120 + 3373\ M\^122 - 1927\ M\^124 + 
                    745\ M\^126 - 198\ M\^128 + 34\ M\^130 - 3\ M\^132)\) + 
              L\^15\ \((\(-M\^50\) + 20\ M\^52 - 182\ M\^54 + 975\ M\^56 - 
                    3383\ M\^58 + 7969\ M\^60 - 12554\ M\^62 + 9026\ M\^64 + 
                    17357\ M\^66 - 66516\ M\^68 + 74926\ M\^70 + 
                    38494\ M\^72 - 158381\ M\^74 + 30930\ M\^76 + 
                    163449\ M\^78 + 63939\ M\^80 - 367191\ M\^82 - 
                    91319\ M\^84 + 744073\ M\^86 - 252557\ M\^88 - 
                    739116\ M\^90 + 638269\ M\^92 + 295245\ M\^94 - 
                    587922\ M\^96 + 89008\ M\^98 + 279348\ M\^100 - 
                    179783\ M\^102 - 35929\ M\^104 + 102908\ M\^106 - 
                    48031\ M\^108 - 19235\ M\^110 + 35979\ M\^112 - 
                    12557\ M\^114 - 8187\ M\^116 + 8033\ M\^118 + 
                    1478\ M\^120 - 6880\ M\^122 + 6120\ M\^124 - 
                    3221\ M\^126 + 1137\ M\^128 - 272\ M\^130 + 41\ M\^132 - 
                    3\ M\^134)\) + 
              L\^13\ \((\(-28\)\ M\^38 + 507\ M\^40 - 4187\ M\^42 + 
                    20171\ M\^44 - 60176\ M\^46 + 104282\ M\^48 - 
                    57522\ M\^50 - 178407\ M\^52 + 483362\ M\^54 - 
                    447056\ M\^56 - 299879\ M\^58 + 1435631\ M\^60 - 
                    1334879\ M\^62 - 1459461\ M\^64 + 3933905\ M\^66 - 
                    301409\ M\^68 - 6244539\ M\^70 + 3844654\ M\^72 + 
                    6489957\ M\^74 - 7719459\ M\^76 - 3758992\ M\^78 + 
                    9544969\ M\^80 - 438876\ M\^82 - 8001557\ M\^84 + 
                    3692558\ M\^86 + 4240131\ M\^88 - 4335127\ M\^90 - 
                    507987\ M\^92 + 2629372\ M\^94 - 1198183\ M\^96 - 
                    568810\ M\^98 + 990170\ M\^100 - 387251\ M\^102 - 
                    265681\ M\^104 + 344040\ M\^106 - 56449\ M\^108 - 
                    100939\ M\^110 + 45758\ M\^112 + 20178\ M\^114 - 
                    10747\ M\^116 - 20149\ M\^118 + 28218\ M\^120 - 
                    18656\ M\^122 + 8104\ M\^124 - 2558\ M\^126 + 
                    618\ M\^128 - 115\ M\^130 + 15\ M\^132 - M\^134)\) + 
              L\^14\ \((\(-8\)\ M\^44 + 142\ M\^46 - 1187\ M\^48 + 
                    5920\ M\^50 - 18813\ M\^52 + 37267\ M\^54 - 
                    36529\ M\^56 - 23381\ M\^58 + 148008\ M\^60 - 
                    240047\ M\^62 + 71727\ M\^64 + 453907\ M\^66 - 
                    695555\ M\^68 - 264254\ M\^70 + 1386270\ M\^72 - 
                    191486\ M\^74 - 2113779\ M\^76 + 1019260\ M\^78 + 
                    2523921\ M\^80 - 2210389\ M\^82 - 1858168\ M\^84 + 
                    2933099\ M\^86 + 435366\ M\^88 - 2442944\ M\^90 + 
                    699553\ M\^92 + 1235339\ M\^94 - 971531\ M\^96 - 
                    130903\ M\^98 + 547638\ M\^100 - 318295\ M\^102 - 
                    47564\ M\^104 + 224927\ M\^106 - 137155\ M\^108 - 
                    29773\ M\^110 + 83005\ M\^112 - 36232\ M\^114 - 
                    10055\ M\^116 + 21156\ M\^118 - 15966\ M\^120 + 
                    10565\ M\^122 - 7011\ M\^124 + 4026\ M\^126 - 
                    1759\ M\^128 + 552\ M\^130 - 119\ M\^132 + 16\ M\^134 - 
                    M\^136)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 65]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 66]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 67]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 68]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 69]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 70]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 71]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 72]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 73]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 74]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 75]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 76]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 77]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 78]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 79]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 80]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 81]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 82]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 83]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 84]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 85]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 86]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 87]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 88]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 89]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 90]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 91]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 92]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 93]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 94]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 95]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 96]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 97]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 98]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 99]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 100]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 101]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 102]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 103]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 104]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 105]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 106]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 107]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 108]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 109]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 110]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 111]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 112]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 113]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 114]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 115]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 116]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 117]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 118]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 119]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 120]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 121]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 122]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 123]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 124]\), \(\(-1\) + 
              L\^2\ M\^30\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 125]\), \(M\^10 + L\^9\ M\^44 + 
              L\ \((1 - M\^2 + 2\ M\^4 - 2\ M\^6 + 3\ M\^8 - 3\ M\^10 - 
                    2\ M\^12 + 6\ M\^14 + M\^16)\) + 
              L\^2\ \((\(-3\)\ M\^4 + 10\ M\^6 - 8\ M\^8 + 7\ M\^10 + 
                    3\ M\^12 - 6\ M\^14 - 3\ M\^18 + 6\ M\^20 + 2\ M\^22)\) + 
              L\^3\ \((3\ M\^8 - 17\ M\^10 + 20\ M\^12 + 7\ M\^14 - 
                    19\ M\^16 + 14\ M\^18 + 12\ M\^20 - 19\ M\^22 - 
                    8\ M\^24 + 5\ M\^26 + 2\ M\^28)\) + 
              L\^4\ \((\(-M\^12\) + 8\ M\^14 - 17\ M\^16 - 11\ M\^18 + 
                    28\ M\^20 - 25\ M\^24 + 25\ M\^26 - 3\ M\^28 - 
                    19\ M\^30 - 2\ M\^32 + 3\ M\^34)\) + 
              L\^5\ \((3\ M\^20 - 2\ M\^22 - 19\ M\^24 - 3\ M\^26 + 
                    25\ M\^28 - 25\ M\^30 + 28\ M\^34 - 11\ M\^36 - 
                    17\ M\^38 + 8\ M\^40 - M\^42)\) + 
              L\^6\ \((2\ M\^26 + 5\ M\^28 - 8\ M\^30 - 19\ M\^32 + 
                    12\ M\^34 + 14\ M\^36 - 19\ M\^38 + 7\ M\^40 + 
                    20\ M\^42 - 17\ M\^44 + 3\ M\^46)\) + 
              L\^7\ \((2\ M\^32 + 6\ M\^34 - 3\ M\^36 - 6\ M\^40 + 3\ M\^42 + 
                    7\ M\^44 - 8\ M\^46 + 10\ M\^48 - 3\ M\^50)\) + 
              L\^8\ \((M\^38 + 6\ M\^40 - 2\ M\^42 - 3\ M\^44 + 3\ M\^46 - 
                    2\ M\^48 + 2\ M\^50 - M\^52 + M\^54)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 126]\), \(L\^13\ M\^8 + M\^66 + 
              L\^12\ \((M\^6 + 2\ M\^8 - 6\ M\^10 + 11\ M\^12 + 4\ M\^14 - 
                    10\ M\^16 + 8\ M\^18 - 3\ M\^20 + 2\ M\^22)\) + 
              L\^11\ \((2\ M\^4 - 3\ M\^6 + 3\ M\^8 + 6\ M\^10 - 17\ M\^12 + 
                    29\ M\^14 - 6\ M\^16 + 10\ M\^18 + 4\ M\^20 - 17\ M\^22 + 
                    24\ M\^24 - 9\ M\^26 + 16\ M\^28 - 15\ M\^30 + 9\ M\^32 - 
                    3\ M\^34 + M\^36)\) + 
              L\^10\ \((\(-1\) + 5\ M\^2 - 12\ M\^4 + 20\ M\^6 - 12\ M\^8 - 
                    19\ M\^10 + 54\ M\^12 - 18\ M\^14 - 42\ M\^16 - 
                    11\ M\^18 + 138\ M\^20 - 50\ M\^22 - 101\ M\^24 + 
                    88\ M\^26 + 59\ M\^28 - 50\ M\^30 - 10\ M\^32 + 
                    62\ M\^34 - 72\ M\^36 + 52\ M\^38 - 18\ M\^40 + 
                    5\ M\^42 - M\^44)\) + 
              L\^9\ \((6\ M\^4 - 31\ M\^6 + 62\ M\^8 - 55\ M\^10 - 6\ M\^12 + 
                    100\ M\^14 - 88\ M\^16 - 90\ M\^18 + 128\ M\^20 + 
                    159\ M\^22 - 364\ M\^24 + 98\ M\^26 + 266\ M\^28 - 
                    285\ M\^30 + 92\ M\^32 + 239\ M\^34 - 278\ M\^36 + 
                    11\ M\^38 + 191\ M\^40 - 171\ M\^42 + 107\ M\^44 - 
                    46\ M\^46 + 11\ M\^48 - M\^50)\) + 
              L\^8\ \((\(-15\)\ M\^8 + 81\ M\^10 - 154\ M\^12 + 89\ M\^14 + 
                    83\ M\^16 - 178\ M\^18 + 113\ M\^20 + 123\ M\^22 - 
                    367\ M\^24 + 71\ M\^26 + 568\ M\^28 - 780\ M\^30 - 
                    2\ M\^32 + 856\ M\^34 - 636\ M\^36 - 208\ M\^38 + 
                    588\ M\^40 - 333\ M\^42 - 34\ M\^44 + 260\ M\^46 - 
                    267\ M\^48 + 148\ M\^50 - 45\ M\^52 + 6\ M\^54)\) + 
              L\^7\ \((20\ M\^12 - 115\ M\^14 + 241\ M\^16 - 196\ M\^18 - 
                    48\ M\^20 + 291\ M\^22 - 469\ M\^24 + 72\ M\^26 + 
                    838\ M\^28 - 810\ M\^30 - 515\ M\^32 + 1245\ M\^34 - 
                    726\ M\^36 - 489\ M\^38 + 1053\ M\^40 - 497\ M\^42 - 
                    403\ M\^44 + 663\ M\^46 - 285\ M\^48 - 157\ M\^50 + 
                    319\ M\^52 - 244\ M\^54 + 95\ M\^56 - 15\ M\^58)\) + 
              L\^6\ \((\(-15\)\ M\^16 + 95\ M\^18 - 244\ M\^20 + 319\ M\^22 - 
                    157\ M\^24 - 285\ M\^26 + 663\ M\^28 - 403\ M\^30 - 
                    497\ M\^32 + 1053\ M\^34 - 489\ M\^36 - 726\ M\^38 + 
                    1245\ M\^40 - 515\ M\^42 - 810\ M\^44 + 838\ M\^46 + 
                    72\ M\^48 - 469\ M\^50 + 291\ M\^52 - 48\ M\^54 - 
                    196\ M\^56 + 241\ M\^58 - 115\ M\^60 + 20\ M\^62)\) + 
              L\^5\ \((6\ M\^20 - 45\ M\^22 + 148\ M\^24 - 267\ M\^26 + 
                    260\ M\^28 - 34\ M\^30 - 333\ M\^32 + 588\ M\^34 - 
                    208\ M\^36 - 636\ M\^38 + 856\ M\^40 - 2\ M\^42 - 
                    780\ M\^44 + 568\ M\^46 + 71\ M\^48 - 367\ M\^50 + 
                    123\ M\^52 + 113\ M\^54 - 178\ M\^56 + 83\ M\^58 + 
                    89\ M\^60 - 154\ M\^62 + 81\ M\^64 - 15\ M\^66)\) + 
              L\ \((2\ M\^52 - 3\ M\^54 + 8\ M\^56 - 10\ M\^58 + 4\ M\^60 + 
                    11\ M\^62 - 6\ M\^64 + 2\ M\^66 + M\^68)\) + 
              L\^2\ \((M\^38 - 3\ M\^40 + 9\ M\^42 - 15\ M\^44 + 16\ M\^46 - 
                    9\ M\^48 + 24\ M\^50 - 17\ M\^52 + 4\ M\^54 + 10\ M\^56 - 
                    6\ M\^58 + 29\ M\^60 - 17\ M\^62 + 6\ M\^64 + 3\ M\^66 - 
                    3\ M\^68 + 2\ M\^70)\) + 
              L\^4\ \((\(-M\^24\) + 11\ M\^26 - 46\ M\^28 + 107\ M\^30 - 
                    171\ M\^32 + 191\ M\^34 + 11\ M\^36 - 278\ M\^38 + 
                    239\ M\^40 + 92\ M\^42 - 285\ M\^44 + 266\ M\^46 + 
                    98\ M\^48 - 364\ M\^50 + 159\ M\^52 + 128\ M\^54 - 
                    90\ M\^56 - 88\ M\^58 + 100\ M\^60 - 6\ M\^62 - 
                    55\ M\^64 + 62\ M\^66 - 31\ M\^68 + 6\ M\^70)\) + 
              L\^3\ \((\(-M\^30\) + 5\ M\^32 - 18\ M\^34 + 52\ M\^36 - 
                    72\ M\^38 + 62\ M\^40 - 10\ M\^42 - 50\ M\^44 + 
                    59\ M\^46 + 88\ M\^48 - 101\ M\^50 - 50\ M\^52 + 
                    138\ M\^54 - 11\ M\^56 - 42\ M\^58 - 18\ M\^60 + 
                    54\ M\^62 - 19\ M\^64 - 12\ M\^66 + 20\ M\^68 - 
                    12\ M\^70 + 5\ M\^72 - M\^74)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 127]\), \(L\^18\ M\^2 + 
              M\^146 + 
              L\^17\ \((\(-2\)\ M\^2 + 2\ M\^4 + 6\ M\^6 - 12\ M\^8 + 
                    24\ M\^10 + 14\ M\^12 - 24\ M\^14 + 10\ M\^16 - 
                    4\ M\^18)\) + 
              L\^16\ \((\(-1\) + 7\ M\^2 - 21\ M\^4 + 29\ M\^6 - 18\ M\^8 - 
                    65\ M\^10 + 188\ M\^12 - 146\ M\^14 - 83\ M\^16 + 
                    408\ M\^18 - 3\ M\^20 - 307\ M\^22 + 108\ M\^24 + 
                    64\ M\^26 - 142\ M\^28 + 95\ M\^30 - 30\ M\^32 + 
                    6\ M\^34)\) + 
              L\^15\ \((\(-2\)\ M\^2 + 14\ M\^4 - 47\ M\^6 + 119\ M\^8 - 
                    207\ M\^10 + 87\ M\^12 + 382\ M\^14 - 862\ M\^16 + 
                    423\ M\^18 + 750\ M\^20 - 1017\ M\^22 + 311\ M\^24 + 
                    1676\ M\^26 - 491\ M\^28 - 886\ M\^30 + 36\ M\^32 - 
                    45\ M\^34 - 60\ M\^36 + 130\ M\^38 + 174\ M\^40 - 
                    313\ M\^42 + 256\ M\^44 - 118\ M\^46 + 30\ M\^48 - 
                    4\ M\^50)\) + 
              L\^14\ \((\(-M\^4\) + 7\ M\^6 - 33\ M\^8 + 133\ M\^10 - 
                    352\ M\^12 + 561\ M\^14 - 381\ M\^16 - 553\ M\^18 + 
                    1678\ M\^20 - 1628\ M\^22 - 846\ M\^24 + 2454\ M\^26 - 
                    456\ M\^28 - 1229\ M\^30 + 2909\ M\^32 + 2444\ M\^34 - 
                    748\ M\^36 - 3362\ M\^38 - 1079\ M\^40 + 800\ M\^42 - 
                    34\ M\^44 + 142\ M\^46 + 736\ M\^48 + 51\ M\^50 - 
                    907\ M\^52 + 859\ M\^54 - 526\ M\^56 + 275\ M\^58 - 
                    132\ M\^60 + 47\ M\^62 - 10\ M\^64 + M\^66)\) + 
              L\^13\ \((\(-5\)\ M\^10 + 42\ M\^12 - 175\ M\^14 + 510\ M\^16 - 
                    1053\ M\^18 + 1362\ M\^20 - 523\ M\^22 - 1770\ M\^24 + 
                    3958\ M\^26 - 3226\ M\^28 - 2824\ M\^30 + 5808\ M\^32 - 
                    1180\ M\^34 - 3746\ M\^36 + 6175\ M\^38 + 7682\ M\^40 + 
                    1319\ M\^42 - 8502\ M\^44 - 4369\ M\^46 + 458\ M\^48 - 
                    2976\ M\^50 + 552\ M\^52 + 5631\ M\^54 - 620\ M\^56 - 
                    1839\ M\^58 + 1214\ M\^60 - 1070\ M\^62 + 918\ M\^64 - 
                    1071\ M\^66 + 1082\ M\^68 - 677\ M\^70 + 248\ M\^72 - 
                    49\ M\^74 + 4\ M\^76)\) + 
              L\^12\ \((\(-10\)\ M\^16 + 99\ M\^18 - 445\ M\^20 + 
                    1226\ M\^22 - 2243\ M\^24 + 2522\ M\^26 - 690\ M\^28 - 
                    3177\ M\^30 + 6974\ M\^32 - 6129\ M\^34 - 4766\ M\^36 + 
                    12416\ M\^38 - 5756\ M\^40 - 12819\ M\^42 + 
                    18861\ M\^44 + 19968\ M\^46 - 10212\ M\^48 - 
                    7291\ M\^50 + 8678\ M\^52 - 20813\ M\^54 - 19217\ M\^56 + 
                    19317\ M\^58 + 9654\ M\^60 - 10512\ M\^62 + 7095\ M\^64 + 
                    3894\ M\^66 - 7205\ M\^68 + 436\ M\^70 + 1446\ M\^72 + 
                    906\ M\^74 - 2637\ M\^76 + 2569\ M\^78 - 1445\ M\^80 + 
                    476\ M\^82 - 84\ M\^84 + 6\ M\^86)\) + 
              L\^11\ \((\(-10\)\ M\^22 + 116\ M\^24 - 603\ M\^26 + 
                    1879\ M\^28 - 3823\ M\^30 + 4672\ M\^32 - 1560\ M\^34 - 
                    5255\ M\^36 + 10810\ M\^38 - 7425\ M\^40 - 7761\ M\^42 + 
                    17047\ M\^44 - 9166\ M\^46 - 21993\ M\^48 + 
                    28996\ M\^50 + 27488\ M\^52 - 14731\ M\^54 + 
                    6900\ M\^56 + 9836\ M\^58 - 45384\ M\^60 - 22368\ M\^62 + 
                    20476\ M\^64 - 1245\ M\^66 + 2961\ M\^68 + 23700\ M\^70 - 
                    8486\ M\^72 - 10507\ M\^74 + 8376\ M\^76 - 1247\ M\^78 - 
                    5223\ M\^80 + 2661\ M\^82 + 3658\ M\^84 - 5419\ M\^86 + 
                    3579\ M\^88 - 1505\ M\^90 + 406\ M\^92 - 62\ M\^94 + 
                    4\ M\^96)\) + 
              L\^10\ \((\(-5\)\ M\^28 + 69\ M\^30 - 436\ M\^32 + 
                    1680\ M\^34 - 4266\ M\^36 + 6802\ M\^38 - 5256\ M\^40 - 
                    2809\ M\^42 + 12966\ M\^44 - 13552\ M\^46 - 1685\ M\^48 + 
                    21486\ M\^50 - 26836\ M\^52 - 16899\ M\^54 + 
                    47224\ M\^56 - 1530\ M\^58 - 12910\ M\^60 + 
                    57625\ M\^62 - 8023\ M\^64 - 79748\ M\^66 + 691\ M\^68 + 
                    11492\ M\^70 - 35964\ M\^72 + 28022\ M\^74 + 
                    46202\ M\^76 - 25861\ M\^78 - 4392\ M\^80 + 
                    16881\ M\^82 - 15511\ M\^84 - 4749\ M\^86 + 
                    12520\ M\^88 - 5059\ M\^90 - 3903\ M\^92 + 7449\ M\^94 - 
                    5648\ M\^96 + 2555\ M\^98 - 756\ M\^100 + 149\ M\^102 - 
                    18\ M\^104 + M\^106)\) + 
              L\^9\ \((\(-M\^34\) + 18\ M\^36 - 152\ M\^38 + 781\ M\^40 - 
                    2633\ M\^42 + 5893\ M\^44 - 8283\ M\^46 + 4783\ M\^48 + 
                    6544\ M\^50 - 16335\ M\^52 + 11118\ M\^54 + 
                    13236\ M\^56 - 34399\ M\^58 + 2536\ M\^60 + 
                    33628\ M\^62 - 26979\ M\^64 + 17183\ M\^66 + 
                    62565\ M\^68 - 29629\ M\^70 - 44496\ M\^72 + 
                    6384\ M\^74 - 44496\ M\^76 - 29629\ M\^78 + 
                    62565\ M\^80 + 17183\ M\^82 - 26979\ M\^84 + 
                    33628\ M\^86 + 2536\ M\^88 - 34399\ M\^90 + 
                    13236\ M\^92 + 11118\ M\^94 - 16335\ M\^96 + 
                    6544\ M\^98 + 4783\ M\^100 - 8283\ M\^102 + 
                    5893\ M\^104 - 2633\ M\^106 + 781\ M\^108 - 152\ M\^110 + 
                    18\ M\^112 - M\^114)\) + 
              L\^8\ \((M\^42 - 18\ M\^44 + 149\ M\^46 - 756\ M\^48 + 
                    2555\ M\^50 - 5648\ M\^52 + 7449\ M\^54 - 3903\ M\^56 - 
                    5059\ M\^58 + 12520\ M\^60 - 4749\ M\^62 - 15511\ M\^64 + 
                    16881\ M\^66 - 4392\ M\^68 - 25861\ M\^70 + 
                    46202\ M\^72 + 28022\ M\^74 - 35964\ M\^76 + 
                    11492\ M\^78 + 691\ M\^80 - 79748\ M\^82 - 8023\ M\^84 + 
                    57625\ M\^86 - 12910\ M\^88 - 1530\ M\^90 + 
                    47224\ M\^92 - 16899\ M\^94 - 26836\ M\^96 + 
                    21486\ M\^98 - 1685\ M\^100 - 13552\ M\^102 + 
                    12966\ M\^104 - 2809\ M\^106 - 5256\ M\^108 + 
                    6802\ M\^110 - 4266\ M\^112 + 1680\ M\^114 - 
                    436\ M\^116 + 69\ M\^118 - 5\ M\^120)\) + 
              L\^7\ \((4\ M\^52 - 62\ M\^54 + 406\ M\^56 - 1505\ M\^58 + 
                    3579\ M\^60 - 5419\ M\^62 + 3658\ M\^64 + 2661\ M\^66 - 
                    5223\ M\^68 - 1247\ M\^70 + 8376\ M\^72 - 10507\ M\^74 - 
                    8486\ M\^76 + 23700\ M\^78 + 2961\ M\^80 - 1245\ M\^82 + 
                    20476\ M\^84 - 22368\ M\^86 - 45384\ M\^88 + 
                    9836\ M\^90 + 6900\ M\^92 - 14731\ M\^94 + 27488\ M\^96 + 
                    28996\ M\^98 - 21993\ M\^100 - 9166\ M\^102 + 
                    17047\ M\^104 - 7761\ M\^106 - 7425\ M\^108 + 
                    10810\ M\^110 - 5255\ M\^112 - 1560\ M\^114 + 
                    4672\ M\^116 - 3823\ M\^118 + 1879\ M\^120 - 
                    603\ M\^122 + 116\ M\^124 - 10\ M\^126)\) + 
              L\^6\ \((6\ M\^62 - 84\ M\^64 + 476\ M\^66 - 1445\ M\^68 + 
                    2569\ M\^70 - 2637\ M\^72 + 906\ M\^74 + 1446\ M\^76 + 
                    436\ M\^78 - 7205\ M\^80 + 3894\ M\^82 + 7095\ M\^84 - 
                    10512\ M\^86 + 9654\ M\^88 + 19317\ M\^90 - 
                    19217\ M\^92 - 20813\ M\^94 + 8678\ M\^96 - 7291\ M\^98 - 
                    10212\ M\^100 + 19968\ M\^102 + 18861\ M\^104 - 
                    12819\ M\^106 - 5756\ M\^108 + 12416\ M\^110 - 
                    4766\ M\^112 - 6129\ M\^114 + 6974\ M\^116 - 
                    3177\ M\^118 - 690\ M\^120 + 2522\ M\^122 - 
                    2243\ M\^124 + 1226\ M\^126 - 445\ M\^128 + 99\ M\^130 - 
                    10\ M\^132)\) + 
              L\^5\ \((4\ M\^72 - 49\ M\^74 + 248\ M\^76 - 677\ M\^78 + 
                    1082\ M\^80 - 1071\ M\^82 + 918\ M\^84 - 1070\ M\^86 + 
                    1214\ M\^88 - 1839\ M\^90 - 620\ M\^92 + 5631\ M\^94 + 
                    552\ M\^96 - 2976\ M\^98 + 458\ M\^100 - 4369\ M\^102 - 
                    8502\ M\^104 + 1319\ M\^106 + 7682\ M\^108 + 
                    6175\ M\^110 - 3746\ M\^112 - 1180\ M\^114 + 
                    5808\ M\^116 - 2824\ M\^118 - 3226\ M\^120 + 
                    3958\ M\^122 - 1770\ M\^124 - 523\ M\^126 + 
                    1362\ M\^128 - 1053\ M\^130 + 510\ M\^132 - 175\ M\^134 + 
                    42\ M\^136 - 5\ M\^138)\) + 
              L\^4\ \((M\^82 - 10\ M\^84 + 47\ M\^86 - 132\ M\^88 + 
                    275\ M\^90 - 526\ M\^92 + 859\ M\^94 - 907\ M\^96 + 
                    51\ M\^98 + 736\ M\^100 + 142\ M\^102 - 34\ M\^104 + 
                    800\ M\^106 - 1079\ M\^108 - 3362\ M\^110 - 748\ M\^112 + 
                    2444\ M\^114 + 2909\ M\^116 - 1229\ M\^118 - 
                    456\ M\^120 + 2454\ M\^122 - 846\ M\^124 - 1628\ M\^126 + 
                    1678\ M\^128 - 553\ M\^130 - 381\ M\^132 + 561\ M\^134 - 
                    352\ M\^136 + 133\ M\^138 - 33\ M\^140 + 7\ M\^142 - 
                    M\^144)\) + 
              L\ \((\(-4\)\ M\^130 + 10\ M\^132 - 24\ M\^134 + 14\ M\^136 + 
                    24\ M\^138 - 12\ M\^140 + 6\ M\^142 + 2\ M\^144 - 
                    2\ M\^146)\) + 
              L\^3\ \((\(-4\)\ M\^98 + 30\ M\^100 - 118\ M\^102 + 
                    256\ M\^104 - 313\ M\^106 + 174\ M\^108 + 130\ M\^110 - 
                    60\ M\^112 - 45\ M\^114 + 36\ M\^116 - 886\ M\^118 - 
                    491\ M\^120 + 1676\ M\^122 + 311\ M\^124 - 1017\ M\^126 + 
                    750\ M\^128 + 423\ M\^130 - 862\ M\^132 + 382\ M\^134 + 
                    87\ M\^136 - 207\ M\^138 + 119\ M\^140 - 47\ M\^142 + 
                    14\ M\^144 - 2\ M\^146)\) + 
              L\^2\ \((6\ M\^114 - 30\ M\^116 + 95\ M\^118 - 142\ M\^120 + 
                    64\ M\^122 + 108\ M\^124 - 307\ M\^126 - 3\ M\^128 + 
                    408\ M\^130 - 83\ M\^132 - 146\ M\^134 + 188\ M\^136 - 
                    65\ M\^138 - 18\ M\^140 + 29\ M\^142 - 21\ M\^144 + 
                    7\ M\^146 - M\^148)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 128]\), \(\(-1\) + 
              L\^11\ M\^120 + 
              L\ \((1 - 2\ M\^2 + M\^4 - 6\ M\^6 + 9\ M\^8 - 3\ M\^10 - 
                    8\ M\^12 + 4\ M\^14 - M\^16)\) + 
              L\^2\ \((\(-M\^8\) + 2\ M\^10 - 2\ M\^12 + 11\ M\^14 - 
                    37\ M\^16 + 18\ M\^18 + 32\ M\^20 - 32\ M\^22 - 
                    7\ M\^24 + 13\ M\^26 - 4\ M\^28 - M\^30 + M\^32)\) + 
              L\^3\ \((\(-M\^20\) - 5\ M\^22 + 33\ M\^24 - 64\ M\^26 - 
                    25\ M\^28 + 127\ M\^30 - 7\ M\^32 - 108\ M\^34 + 
                    33\ M\^36 + 48\ M\^38 - 26\ M\^40 - 12\ M\^42 + 
                    17\ M\^44 - 6\ M\^46 + M\^48)\) + 
              L\^4\ \((\(-2\)\ M\^30 - 9\ M\^32 + 48\ M\^34 - 25\ M\^36 - 
                    150\ M\^38 + 123\ M\^40 + 220\ M\^42 - 160\ M\^44 - 
                    154\ M\^46 + 129\ M\^48 + 51\ M\^50 - 69\ M\^52 + 
                    16\ M\^54 + 5\ M\^56 - M\^58)\) + 
              L\^5\ \((2\ M\^40 - 17\ M\^42 + 10\ M\^44 + 78\ M\^46 - 
                    113\ M\^48 - 127\ M\^50 + 267\ M\^52 + 131\ M\^54 - 
                    282\ M\^56 - 53\ M\^58 + 161\ M\^60 - 24\ M\^62 - 
                    34\ M\^64 + 14\ M\^66 + M\^68)\) + 
              L\^6\ \((\(-M\^52\) - 14\ M\^54 + 34\ M\^56 + 24\ M\^58 - 
                    161\ M\^60 + 53\ M\^62 + 282\ M\^64 - 131\ M\^66 - 
                    267\ M\^68 + 127\ M\^70 + 113\ M\^72 - 78\ M\^74 - 
                    10\ M\^76 + 17\ M\^78 - 2\ M\^80)\) + 
              L\^7\ \((M\^62 - 5\ M\^64 - 16\ M\^66 + 69\ M\^68 - 51\ M\^70 - 
                    129\ M\^72 + 154\ M\^74 + 160\ M\^76 - 220\ M\^78 - 
                    123\ M\^80 + 150\ M\^82 + 25\ M\^84 - 48\ M\^86 + 
                    9\ M\^88 + 2\ M\^90)\) + 
              L\^8\ \((\(-M\^72\) + 6\ M\^74 - 17\ M\^76 + 12\ M\^78 + 
                    26\ M\^80 - 48\ M\^82 - 33\ M\^84 + 108\ M\^86 + 
                    7\ M\^88 - 127\ M\^90 + 25\ M\^92 + 64\ M\^94 - 
                    33\ M\^96 + 5\ M\^98 + M\^100)\) + 
              L\^9\ \((\(-M\^88\) + M\^90 + 4\ M\^92 - 13\ M\^94 + 7\ M\^96 + 
                    32\ M\^98 - 32\ M\^100 - 18\ M\^102 + 37\ M\^104 - 
                    11\ M\^106 + 2\ M\^108 - 2\ M\^110 + M\^112)\) + 
              L\^10\ \((M\^104 - 4\ M\^106 + 8\ M\^108 + 3\ M\^110 - 
                    9\ M\^112 + 6\ M\^114 - M\^116 + 2\ M\^118 - 
                    M\^120)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 129]\), \(L\^18\ M\^32 - 
              M\^38 + L\^17\ \((M\^26 + 2\ M\^28 - 13\ M\^30 + 25\ M\^32 + 
                    11\ M\^34 - 22\ M\^36 + 12\ M\^38 - 7\ M\^40 + 
                    3\ M\^42)\) + 
              L\ \((\(-3\)\ M\^28 + 7\ M\^30 - 12\ M\^32 + 22\ M\^34 - 
                    11\ M\^36 - 25\ M\^38 + 13\ M\^40 - 2\ M\^42 - M\^44)\) + 
              L\^2\ \((\(-3\)\ M\^18 + 14\ M\^20 - 39\ M\^22 + 78\ M\^24 - 
                    67\ M\^26 - 22\ M\^28 + 11\ M\^30 + 71\ M\^32 + 
                    26\ M\^34 - 57\ M\^36 - 76\ M\^38 - 50\ M\^40 + 
                    84\ M\^42 - 25\ M\^44 - 20\ M\^46 + 16\ M\^48 - 
                    4\ M\^50)\) + 
              L\^16\ \((4\ M\^20 - 16\ M\^22 + 20\ M\^24 + 25\ M\^26 - 
                    84\ M\^28 + 50\ M\^30 + 76\ M\^32 + 57\ M\^34 - 
                    26\ M\^36 - 71\ M\^38 - 11\ M\^40 + 22\ M\^42 + 
                    67\ M\^44 - 78\ M\^46 + 39\ M\^48 - 14\ M\^50 + 
                    3\ M\^52)\) + 
              L\^3\ \((\(-M\^8\) + 7\ M\^10 - 27\ M\^12 + 65\ M\^14 - 
                    90\ M\^16 + 69\ M\^18 - 101\ M\^20 + 242\ M\^22 - 
                    335\ M\^24 + 410\ M\^26 - 117\ M\^28 - 891\ M\^30 + 
                    926\ M\^32 + 578\ M\^34 - 661\ M\^36 - 366\ M\^38 - 
                    71\ M\^40 + 187\ M\^42 + 185\ M\^44 - 303\ M\^46 + 
                    79\ M\^48 + 100\ M\^50 - 105\ M\^52 + 45\ M\^54 - 
                    10\ M\^56 + M\^58)\) + 
              L\^4\ \((\(-4\)\ M\^6 + 41\ M\^8 - 182\ M\^10 + 433\ M\^12 - 
                    560\ M\^14 + 240\ M\^16 + 318\ M\^18 - 173\ M\^20 - 
                    1283\ M\^22 + 2042\ M\^24 + 497\ M\^26 - 2545\ M\^28 + 
                    652\ M\^30 - 253\ M\^32 + 1675\ M\^34 + 1570\ M\^36 - 
                    3045\ M\^38 - 1346\ M\^40 + 1841\ M\^42 + 566\ M\^44 - 
                    991\ M\^46 - 167\ M\^48 + 609\ M\^50 - 151\ M\^52 - 
                    237\ M\^54 + 227\ M\^56 - 92\ M\^58 + 20\ M\^60 - 
                    2\ M\^62)\) + 
              L\^15\ \((\(-M\^12\) + 10\ M\^14 - 45\ M\^16 + 105\ M\^18 - 
                    100\ M\^20 - 79\ M\^22 + 303\ M\^24 - 185\ M\^26 - 
                    187\ M\^28 + 71\ M\^30 + 366\ M\^32 + 661\ M\^34 - 
                    578\ M\^36 - 926\ M\^38 + 891\ M\^40 + 117\ M\^42 - 
                    410\ M\^44 + 335\ M\^46 - 242\ M\^48 + 101\ M\^50 - 
                    69\ M\^52 + 90\ M\^54 - 65\ M\^56 + 27\ M\^58 - 
                    7\ M\^60 + M\^62)\) + 
              L\^14\ \((2\ M\^8 - 20\ M\^10 + 92\ M\^12 - 227\ M\^14 + 
                    237\ M\^16 + 151\ M\^18 - 609\ M\^20 + 167\ M\^22 + 
                    991\ M\^24 - 566\ M\^26 - 1841\ M\^28 + 1346\ M\^30 + 
                    3045\ M\^32 - 1570\ M\^34 - 1675\ M\^36 + 253\ M\^38 - 
                    652\ M\^40 + 2545\ M\^42 - 497\ M\^44 - 2042\ M\^46 + 
                    1283\ M\^48 + 173\ M\^50 - 318\ M\^52 - 240\ M\^54 + 
                    560\ M\^56 - 433\ M\^58 + 182\ M\^60 - 41\ M\^62 + 
                    4\ M\^64)\) + 
              L\^6\ \((\(-4\)\ M\^2 + 67\ M\^4 - 460\ M\^6 + 1708\ M\^8 - 
                    3550\ M\^10 + 3402\ M\^12 + 1184\ M\^14 - 6371\ M\^16 + 
                    4143\ M\^18 + 4483\ M\^20 - 9227\ M\^22 + 1065\ M\^24 + 
                    13251\ M\^26 - 8012\ M\^28 - 12777\ M\^30 + 6641\ M\^32 + 
                    21048\ M\^34 - 10553\ M\^36 - 21400\ M\^38 + 
                    16692\ M\^40 + 8112\ M\^42 - 11766\ M\^44 - 2748\ M\^46 + 
                    8109\ M\^48 - 256\ M\^50 - 4873\ M\^52 + 2533\ M\^54 + 
                    728\ M\^56 - 1285\ M\^58 + 622\ M\^60 - 169\ M\^62 + 
                    30\ M\^64 - 3\ M\^66)\) + 
              L\^5\ \((\(-6\)\ M\^4 + 80\ M\^6 - 434\ M\^8 + 1240\ M\^10 - 
                    1888\ M\^12 + 959\ M\^14 + 1451\ M\^16 - 2294\ M\^18 - 
                    778\ M\^20 + 3958\ M\^22 - 2278\ M\^24 - 2052\ M\^26 + 
                    5008\ M\^28 - 3186\ M\^30 - 2446\ M\^32 + 2379\ M\^34 + 
                    4546\ M\^36 - 1746\ M\^38 - 6156\ M\^40 + 2808\ M\^42 + 
                    1582\ M\^44 - 969\ M\^46 + 205\ M\^48 - 422\ M\^50 + 
                    330\ M\^52 + 68\ M\^54 - 315\ M\^56 + 311\ M\^58 - 
                    167\ M\^60 + 53\ M\^62 - 10\ M\^64 + M\^66)\) + 
              L\^13\ \((\(-M\^4\) + 10\ M\^6 - 53\ M\^8 + 167\ M\^10 - 
                    311\ M\^12 + 315\ M\^14 - 68\ M\^16 - 330\ M\^18 + 
                    422\ M\^20 - 205\ M\^22 + 969\ M\^24 - 1582\ M\^26 - 
                    2808\ M\^28 + 6156\ M\^30 + 1746\ M\^32 - 4546\ M\^34 - 
                    2379\ M\^36 + 2446\ M\^38 + 3186\ M\^40 - 5008\ M\^42 + 
                    2052\ M\^44 + 2278\ M\^46 - 3958\ M\^48 + 778\ M\^50 + 
                    2294\ M\^52 - 1451\ M\^54 - 959\ M\^56 + 1888\ M\^58 - 
                    1240\ M\^60 + 434\ M\^62 - 80\ M\^64 + 6\ M\^66)\) + 
              L\^7\ \((\(-1\) + 23\ M\^2 - 219\ M\^4 + 1114\ M\^6 - 
                    3320\ M\^8 + 5585\ M\^10 - 3307\ M\^12 - 5425\ M\^14 + 
                    10757\ M\^16 - 2311\ M\^18 - 8652\ M\^20 + 2249\ M\^22 + 
                    5862\ M\^24 + 2936\ M\^26 - 539\ M\^28 - 16621\ M\^30 + 
                    5179\ M\^32 + 20330\ M\^34 - 3290\ M\^36 - 13619\ M\^38 - 
                    6276\ M\^40 + 11565\ M\^42 + 1887\ M\^44 - 1433\ M\^46 - 
                    3565\ M\^48 - 4\ M\^50 + 4918\ M\^52 - 4153\ M\^54 + 
                    897\ M\^56 + 1275\ M\^58 - 1504\ M\^60 + 798\ M\^62 - 
                    233\ M\^64 + 35\ M\^66 - 2\ M\^68)\) + 
              L\^8\ \((2 - 38\ M\^2 + 305\ M\^4 - 1372\ M\^6 + 3765\ M\^8 - 
                    6108\ M\^10 + 3918\ M\^12 + 5791\ M\^14 - 15658\ M\^16 + 
                    11663\ M\^18 + 6380\ M\^20 - 22650\ M\^22 + 
                    11670\ M\^24 + 27878\ M\^26 - 41347\ M\^28 - 
                    15235\ M\^30 + 63726\ M\^32 - 7050\ M\^34 - 
                    61015\ M\^36 + 26637\ M\^38 + 45948\ M\^40 - 
                    46596\ M\^42 - 15255\ M\^44 + 37048\ M\^46 + 
                    1261\ M\^48 - 21261\ M\^50 + 1455\ M\^52 + 14599\ M\^54 - 
                    7958\ M\^56 - 2283\ M\^58 + 4287\ M\^60 - 2118\ M\^62 + 
                    536\ M\^64 - 71\ M\^66 + 4\ M\^68)\) + 
              L\^12\ \((3\ M\^4 - 30\ M\^6 + 169\ M\^8 - 622\ M\^10 + 
                    1285\ M\^12 - 728\ M\^14 - 2533\ M\^16 + 4873\ M\^18 + 
                    256\ M\^20 - 8109\ M\^22 + 2748\ M\^24 + 11766\ M\^26 - 
                    8112\ M\^28 - 16692\ M\^30 + 21400\ M\^32 + 
                    10553\ M\^34 - 21048\ M\^36 - 6641\ M\^38 + 
                    12777\ M\^40 + 8012\ M\^42 - 13251\ M\^44 - 1065\ M\^46 + 
                    9227\ M\^48 - 4483\ M\^50 - 4143\ M\^52 + 6371\ M\^54 - 
                    1184\ M\^56 - 3402\ M\^58 + 3550\ M\^60 - 1708\ M\^62 + 
                    460\ M\^64 - 67\ M\^66 + 4\ M\^68)\) + 
              L\^10\ \((\(-4\)\ M\^2 + 71\ M\^4 - 536\ M\^6 + 2118\ M\^8 - 
                    4287\ M\^10 + 2283\ M\^12 + 7958\ M\^14 - 14599\ M\^16 - 
                    1455\ M\^18 + 21261\ M\^20 - 1261\ M\^22 - 37048\ M\^24 + 
                    15255\ M\^26 + 46596\ M\^28 - 45948\ M\^30 - 
                    26637\ M\^32 + 61015\ M\^34 + 7050\ M\^36 - 
                    63726\ M\^38 + 15235\ M\^40 + 41347\ M\^42 - 
                    27878\ M\^44 - 11670\ M\^46 + 22650\ M\^48 - 
                    6380\ M\^50 - 11663\ M\^52 + 15658\ M\^54 - 5791\ M\^56 - 
                    3918\ M\^58 + 6108\ M\^60 - 3765\ M\^62 + 1372\ M\^64 - 
                    305\ M\^66 + 38\ M\^68 - 2\ M\^70)\) + 
              L\^11\ \((2\ M\^2 - 35\ M\^4 + 233\ M\^6 - 798\ M\^8 + 
                    1504\ M\^10 - 1275\ M\^12 - 897\ M\^14 + 4153\ M\^16 - 
                    4918\ M\^18 + 4\ M\^20 + 3565\ M\^22 + 1433\ M\^24 - 
                    1887\ M\^26 - 11565\ M\^28 + 6276\ M\^30 + 13619\ M\^32 + 
                    3290\ M\^34 - 20330\ M\^36 - 5179\ M\^38 + 16621\ M\^40 + 
                    539\ M\^42 - 2936\ M\^44 - 5862\ M\^46 - 2249\ M\^48 + 
                    8652\ M\^50 + 2311\ M\^52 - 10757\ M\^54 + 5425\ M\^56 + 
                    3307\ M\^58 - 5585\ M\^60 + 3320\ M\^62 - 1114\ M\^64 + 
                    219\ M\^66 - 23\ M\^68 + M\^70)\) + 
              L\^9\ \((\(-1\) + 21\ M\^2 - 185\ M\^4 + 957\ M\^6 - 
                    3165\ M\^8 + 6338\ M\^10 - 4980\ M\^12 - 8607\ M\^14 + 
                    24828\ M\^16 - 10589\ M\^18 - 33446\ M\^20 + 
                    31061\ M\^22 + 38781\ M\^24 - 59477\ M\^26 - 
                    24601\ M\^28 + 75437\ M\^30 - 14782\ M\^32 - 
                    51284\ M\^34 + 51284\ M\^36 + 14782\ M\^38 - 
                    75437\ M\^40 + 24601\ M\^42 + 59477\ M\^44 - 
                    38781\ M\^46 - 31061\ M\^48 + 33446\ M\^50 + 
                    10589\ M\^52 - 24828\ M\^54 + 8607\ M\^56 + 4980\ M\^58 - 
                    6338\ M\^60 + 3165\ M\^62 - 957\ M\^64 + 185\ M\^66 - 
                    21\ M\^68 + M\^70)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 130]\), \(L\^14\ M\^24 - 
              M\^30 + L\ \((\(-M\^16\) + 2\ M\^18 - 5\ M\^20 + 9\ M\^22 - 
                    11\ M\^24 + 2\ M\^26 + 8\ M\^28 - 11\ M\^30 - M\^32)\) + 
              L\^2\ \((M\^8 - 3\ M\^10 + 7\ M\^12 - 21\ M\^14 + 35\ M\^16 - 
                    38\ M\^18 + 16\ M\^20 + 17\ M\^22 - 35\ M\^24 + M\^26 + 
                    7\ M\^28 + 8\ M\^30 - 14\ M\^32 - 6\ M\^34)\) + 
              L\^4\ \((1 - 7\ M\^2 + 23\ M\^4 - 48\ M\^6 + 62\ M\^8 - 
                    54\ M\^10 + 108\ M\^12 - 164\ M\^14 + 12\ M\^16 + 
                    135\ M\^18 - 163\ M\^20 + 211\ M\^22 - 10\ M\^24 - 
                    225\ M\^26 - 51\ M\^28 + 113\ M\^30 + 171\ M\^32 - 
                    71\ M\^34 - 15\ M\^36 - 17\ M\^38)\) + 
              L\^13\ \((M\^22 + 11\ M\^24 - 8\ M\^26 - 2\ M\^28 + 11\ M\^30 - 
                    9\ M\^32 + 5\ M\^34 - 2\ M\^36 + M\^38)\) + 
              L\^3\ \((\(-2\)\ M\^4 + 10\ M\^6 - 25\ M\^8 + 49\ M\^10 - 
                    64\ M\^12 + 50\ M\^14 - 44\ M\^16 + 7\ M\^18 + 
                    35\ M\^20 - 6\ M\^22 + 5\ M\^24 - 112\ M\^26 + 
                    36\ M\^28 + 61\ M\^30 - 10\ M\^32 + 2\ M\^34 - 
                    27\ M\^36 + 3\ M\^38)\) + 
              L\^5\ \((\(-1\) + 7\ M\^2 - 26\ M\^4 + 73\ M\^6 - 154\ M\^8 + 
                    181\ M\^10 + 23\ M\^12 - 182\ M\^14 + 21\ M\^16 + 
                    13\ M\^18 + 34\ M\^20 - 93\ M\^22 + 654\ M\^24 - 
                    372\ M\^26 - 882\ M\^28 + 788\ M\^30 + 194\ M\^32 - 
                    238\ M\^34 + 89\ M\^36 - M\^38 - 46\ M\^40 + 
                    6\ M\^42)\) + 
              L\^6\ \((M\^4 - 18\ M\^6 + 119\ M\^8 - 347\ M\^10 + 
                    378\ M\^12 + 473\ M\^14 - 1232\ M\^16 + 204\ M\^18 + 
                    1020\ M\^20 - 581\ M\^22 - 390\ M\^24 + 1139\ M\^26 - 
                    255\ M\^28 - 1631\ M\^30 + 1171\ M\^32 + 635\ M\^34 - 
                    786\ M\^36 + 200\ M\^38 + 87\ M\^40 - 115\ M\^42 + 
                    30\ M\^44 - 3\ M\^46)\) + 
              L\^12\ \((6\ M\^20 + 14\ M\^22 - 8\ M\^24 - 7\ M\^26 - M\^28 + 
                    35\ M\^30 - 17\ M\^32 - 16\ M\^34 + 38\ M\^36 - 
                    35\ M\^38 + 21\ M\^40 - 7\ M\^42 + 3\ M\^44 - M\^46)\) + 
              L\^7\ \((\(-M\^6\) - 12\ M\^8 + 128\ M\^10 - 337\ M\^12 + 
                    211\ M\^14 + 925\ M\^16 - 1519\ M\^18 - 438\ M\^20 + 
                    1883\ M\^22 - 444\ M\^24 - 944\ M\^26 + 944\ M\^28 + 
                    444\ M\^30 - 1883\ M\^32 + 438\ M\^34 + 1519\ M\^36 - 
                    925\ M\^38 - 211\ M\^40 + 337\ M\^42 - 128\ M\^44 + 
                    12\ M\^46 + M\^48)\) + 
              L\^8\ \((3\ M\^8 - 30\ M\^10 + 115\ M\^12 - 87\ M\^14 - 
                    200\ M\^16 + 786\ M\^18 - 635\ M\^20 - 1171\ M\^22 + 
                    1631\ M\^24 + 255\ M\^26 - 1139\ M\^28 + 390\ M\^30 + 
                    581\ M\^32 - 1020\ M\^34 - 204\ M\^36 + 1232\ M\^38 - 
                    473\ M\^40 - 378\ M\^42 + 347\ M\^44 - 119\ M\^46 + 
                    18\ M\^48 - M\^50)\) + 
              L\^11\ \((\(-3\)\ M\^16 + 27\ M\^18 - 2\ M\^20 + 10\ M\^22 - 
                    61\ M\^24 - 36\ M\^26 + 112\ M\^28 - 5\ M\^30 + 
                    6\ M\^32 - 35\ M\^34 - 7\ M\^36 + 44\ M\^38 - 50\ M\^40 + 
                    64\ M\^42 - 49\ M\^44 + 25\ M\^46 - 10\ M\^48 + 
                    2\ M\^50)\) + 
              L\^10\ \((17\ M\^16 + 15\ M\^18 + 71\ M\^20 - 171\ M\^22 - 
                    113\ M\^24 + 51\ M\^26 + 225\ M\^28 + 10\ M\^30 - 
                    211\ M\^32 + 163\ M\^34 - 135\ M\^36 - 12\ M\^38 + 
                    164\ M\^40 - 108\ M\^42 + 54\ M\^44 - 62\ M\^46 + 
                    48\ M\^48 - 23\ M\^50 + 7\ M\^52 - M\^54)\) + 
              L\^9\ \((\(-6\)\ M\^12 + 46\ M\^14 + M\^16 - 89\ M\^18 + 
                    238\ M\^20 - 194\ M\^22 - 788\ M\^24 + 882\ M\^26 + 
                    372\ M\^28 - 654\ M\^30 + 93\ M\^32 - 34\ M\^34 - 
                    13\ M\^36 - 21\ M\^38 + 182\ M\^40 - 23\ M\^42 - 
                    181\ M\^44 + 154\ M\^46 - 73\ M\^48 + 26\ M\^50 - 
                    7\ M\^52 + M\^54)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 131]\), \(L\^21\ M\^6 - M\^86 + 
              L\^20\ \((\(-6\)\ M\^6 + 3\ M\^8 + 42\ M\^10 - 17\ M\^12 - 
                    20\ M\^14 + 25\ M\^16 - 16\ M\^18 + 6\ M\^20 - 
                    2\ M\^22)\) + 
              L\^19\ \((\(-3\)\ M\^4 + 41\ M\^6 - 100\ M\^8 - 125\ M\^10 + 
                    380\ M\^12 + 232\ M\^14 - 253\ M\^16 - 252\ M\^18 + 
                    229\ M\^20 + 59\ M\^22 - 178\ M\^24 + 98\ M\^26 - 
                    56\ M\^28 + 58\ M\^30 - 44\ M\^32 + 21\ M\^34 - 
                    6\ M\^36 + M\^38)\) + 
              L\^18\ \((12\ M\^4 - 134\ M\^6 + 437\ M\^8 + 25\ M\^10 - 
                    1929\ M\^12 + 389\ M\^14 + 4574\ M\^16 - 400\ M\^18 - 
                    4586\ M\^20 + 1704\ M\^22 + 1191\ M\^24 - 970\ M\^26 + 
                    78\ M\^28 - 299\ M\^30 + 391\ M\^32 + 12\ M\^34 - 
                    17\ M\^36 - 273\ M\^38 + 379\ M\^40 - 254\ M\^42 + 
                    102\ M\^44 - 25\ M\^46 + 3\ M\^48)\) + 
              L\^17\ \((3\ M\^2 - 64\ M\^4 + 484\ M\^6 - 1607\ M\^8 + 
                    1070\ M\^10 + 6713\ M\^12 - 11840\ M\^14 - 11745\ M\^16 + 
                    30725\ M\^18 + 7913\ M\^20 - 29084\ M\^22 + 2557\ M\^24 + 
                    10092\ M\^26 - 1745\ M\^28 - 4238\ M\^30 - 871\ M\^32 + 
                    3044\ M\^34 + 1028\ M\^36 - 1115\ M\^38 - 1271\ M\^40 + 
                    1321\ M\^42 - 380\ M\^44 + 517\ M\^46 - 1029\ M\^48 + 
                    925\ M\^50 - 480\ M\^52 + 159\ M\^54 - 32\ M\^56 + 
                    3\ M\^58)\) + 
              L\^16\ \((\(-8\)\ M\^2 + 148\ M\^4 - 1142\ M\^6 + 4545\ M\^8 - 
                    7627\ M\^10 - 7644\ M\^12 + 48192\ M\^14 - 33998\ M\^16 - 
                    94607\ M\^18 + 119403\ M\^20 + 80142\ M\^22 - 
                    130938\ M\^24 - 24409\ M\^26 + 77875\ M\^28 - 
                    1610\ M\^30 - 48703\ M\^32 + 148\ M\^34 + 27907\ M\^36 + 
                    10685\ M\^38 - 19608\ M\^40 - 7419\ M\^42 + 
                    14641\ M\^44 - 333\ M\^46 - 6177\ M\^48 + 509\ M\^50 + 
                    3997\ M\^52 - 3817\ M\^54 + 2508\ M\^56 - 1541\ M\^58 + 
                    803\ M\^60 - 307\ M\^62 + 80\ M\^64 - 13\ M\^66 + 
                    M\^68)\) + 
              L\^15\ \((\(-1\) + 27\ M\^2 - 301\ M\^4 + 1945\ M\^6 - 
                    8026\ M\^8 + 19436\ M\^10 - 12530\ M\^12 - 68280\ M\^14 + 
                    170861\ M\^16 - 8085\ M\^18 - 397731\ M\^20 + 
                    263662\ M\^22 + 435382\ M\^24 - 412730\ M\^26 - 
                    255463\ M\^28 + 390355\ M\^30 + 50130\ M\^32 - 
                    294537\ M\^34 + 17605\ M\^36 + 163932\ M\^38 + 
                    16975\ M\^40 - 101784\ M\^42 - 12253\ M\^44 + 
                    71186\ M\^46 - 7532\ M\^48 - 41304\ M\^50 + 
                    15920\ M\^52 + 13093\ M\^54 - 11208\ M\^56 + 
                    3616\ M\^58 - 3869\ M\^60 + 5519\ M\^62 - 4377\ M\^64 + 
                    2186\ M\^66 - 747\ M\^68 + 177\ M\^70 - 27\ M\^72 + 
                    2\ M\^74)\) + 
              L\^14\ \((2 - 46\ M\^2 + 473\ M\^4 - 2883\ M\^6 + 11494\ M\^8 - 
                    30246\ M\^10 + 42242\ M\^12 + 28793\ M\^14 - 
                    249787\ M\^16 + 338084\ M\^18 + 276862\ M\^20 - 
                    1033331\ M\^22 + 173575\ M\^24 + 1444977\ M\^26 - 
                    745310\ M\^28 - 1183210\ M\^30 + 1113577\ M\^32 + 
                    506180\ M\^34 - 1076269\ M\^36 - 84631\ M\^38 + 
                    733010\ M\^40 + 3427\ M\^42 - 459035\ M\^44 + 
                    50323\ M\^46 + 295894\ M\^48 - 86332\ M\^50 - 
                    173968\ M\^52 + 94050\ M\^54 + 55478\ M\^56 - 
                    50684\ M\^58 - 7121\ M\^60 + 16500\ M\^62 + 4214\ M\^64 - 
                    16078\ M\^66 + 13402\ M\^68 - 6644\ M\^70 + 2265\ M\^72 - 
                    563\ M\^74 + 105\ M\^76 - 14\ M\^78 + M\^80)\) + 
              L\^13\ \((\(-1\) + 30\ M\^2 - 390\ M\^4 + 2843\ M\^6 - 
                    12909\ M\^8 + 37989\ M\^10 - 68134\ M\^12 + 
                    35447\ M\^14 + 184697\ M\^16 - 528942\ M\^18 + 
                    325837\ M\^20 + 988271\ M\^22 - 1683990\ M\^24 - 
                    662437\ M\^26 + 2988441\ M\^28 - 444970\ M\^30 - 
                    3092865\ M\^32 + 1760057\ M\^34 + 1967461\ M\^36 - 
                    2371734\ M\^38 - 882217\ M\^40 + 2108878\ M\^42 + 
                    242872\ M\^44 - 1553210\ M\^46 + 204949\ M\^48 + 
                    1043878\ M\^50 - 424895\ M\^52 - 576969\ M\^54 + 
                    427586\ M\^56 + 161127\ M\^58 - 247208\ M\^60 + 
                    13563\ M\^62 + 102127\ M\^64 - 52333\ M\^66 - 
                    17922\ M\^68 + 38064\ M\^70 - 26074\ M\^72 + 
                    11288\ M\^74 - 3429\ M\^76 + 724\ M\^78 - 96\ M\^80 + 
                    6\ M\^82)\) + 
              L\^12\ \((\(-6\)\ M\^2 + 140\ M\^4 - 1451\ M\^6 + 8679\ M\^8 - 
                    32635\ M\^10 + 77457\ M\^12 - 97715\ M\^14 - 
                    34879\ M\^16 + 413777\ M\^18 - 705295\ M\^20 - 
                    11833\ M\^22 + 1762132\ M\^24 - 1716695\ M\^26 - 
                    2028249\ M\^28 + 3986355\ M\^30 + 784725\ M\^32 - 
                    5013154\ M\^34 + 1550737\ M\^36 + 4141254\ M\^38 - 
                    3441662\ M\^40 - 2655098\ M\^42 + 3995165\ M\^44 + 
                    1107252\ M\^46 - 3505143\ M\^48 + 327148\ M\^50 + 
                    2583361\ M\^52 - 1173824\ M\^54 - 1405038\ M\^56 + 
                    1237343\ M\^58 + 322212\ M\^60 - 737456\ M\^62 + 
                    159075\ M\^64 + 254807\ M\^66 - 209365\ M\^68 + 
                    26142\ M\^70 + 64183\ M\^72 - 60500\ M\^74 + 
                    30375\ M\^76 - 9929\ M\^78 + 2113\ M\^80 - 266\ M\^82 + 
                    15\ M\^84)\) + 
              L\ \((2\ M\^70 - 6\ M\^72 + 16\ M\^74 - 25\ M\^76 + 20\ M\^78 + 
                    17\ M\^80 - 42\ M\^82 - 3\ M\^84 + 6\ M\^86)\) + 
              L\^11\ \((\(-15\)\ M\^4 + 314\ M\^6 - 2883\ M\^8 + 
                    15244\ M\^10 - 50536\ M\^12 + 103390\ M\^14 - 
                    97875\ M\^16 - 110136\ M\^18 + 524061\ M\^20 - 
                    638217\ M\^22 - 361751\ M\^24 + 1930931\ M\^26 - 
                    1152718\ M\^28 - 2745775\ M\^30 + 3695622\ M\^32 + 
                    1885131\ M\^34 - 5546489\ M\^36 + 697082\ M\^38 + 
                    5637499\ M\^40 - 3529531\ M\^42 - 4561856\ M\^44 + 
                    5180946\ M\^46 + 2479631\ M\^48 - 5344610\ M\^50 + 
                    76113\ M\^52 + 4454839\ M\^54 - 1929759\ M\^56 - 
                    2584468\ M\^58 + 2275919\ M\^60 + 608147\ M\^62 - 
                    1430404\ M\^64 + 370166\ M\^66 + 443879\ M\^68 - 
                    417406\ M\^70 + 98701\ M\^72 + 82423\ M\^74 - 
                    94561\ M\^76 + 49395\ M\^78 - 15998\ M\^80 + 
                    3269\ M\^82 - 386\ M\^84 + 20\ M\^86)\) + 
              L\^3\ \((\(-3\)\ M\^44 + 25\ M\^46 - 102\ M\^48 + 254\ M\^50 - 
                    379\ M\^52 + 273\ M\^54 + 17\ M\^56 - 12\ M\^58 - 
                    391\ M\^60 + 299\ M\^62 - 78\ M\^64 + 970\ M\^66 - 
                    1191\ M\^68 - 1704\ M\^70 + 4586\ M\^72 + 400\ M\^74 - 
                    4574\ M\^76 - 389\ M\^78 + 1929\ M\^80 - 25\ M\^82 - 
                    437\ M\^84 + 134\ M\^86 - 12\ M\^88)\) + 
              L\^2\ \((\(-M\^54\) + 6\ M\^56 - 21\ M\^58 + 44\ M\^60 - 
                    58\ M\^62 + 56\ M\^64 - 98\ M\^66 + 178\ M\^68 - 
                    59\ M\^70 - 229\ M\^72 + 252\ M\^74 + 253\ M\^76 - 
                    232\ M\^78 - 380\ M\^80 + 125\ M\^82 + 100\ M\^84 - 
                    41\ M\^86 + 3\ M\^88)\) + 
              L\^10\ \((\(-20\)\ M\^6 + 386\ M\^8 - 3269\ M\^10 + 
                    15998\ M\^12 - 49395\ M\^14 + 94561\ M\^16 - 
                    82423\ M\^18 - 98701\ M\^20 + 417406\ M\^22 - 
                    443879\ M\^24 - 370166\ M\^26 + 1430404\ M\^28 - 
                    608147\ M\^30 - 2275919\ M\^32 + 2584468\ M\^34 + 
                    1929759\ M\^36 - 4454839\ M\^38 - 76113\ M\^40 + 
                    5344610\ M\^42 - 2479631\ M\^44 - 5180946\ M\^46 + 
                    4561856\ M\^48 + 3529531\ M\^50 - 5637499\ M\^52 - 
                    697082\ M\^54 + 5546489\ M\^56 - 1885131\ M\^58 - 
                    3695622\ M\^60 + 2745775\ M\^62 + 1152718\ M\^64 - 
                    1930931\ M\^66 + 361751\ M\^68 + 638217\ M\^70 - 
                    524061\ M\^72 + 110136\ M\^74 + 97875\ M\^76 - 
                    103390\ M\^78 + 50536\ M\^80 - 15244\ M\^82 + 
                    2883\ M\^84 - 314\ M\^86 + 15\ M\^88)\) + 
              L\^4\ \((\(-3\)\ M\^34 + 32\ M\^36 - 159\ M\^38 + 480\ M\^40 - 
                    925\ M\^42 + 1029\ M\^44 - 517\ M\^46 + 380\ M\^48 - 
                    1321\ M\^50 + 1271\ M\^52 + 1115\ M\^54 - 1028\ M\^56 - 
                    3044\ M\^58 + 871\ M\^60 + 4238\ M\^62 + 1745\ M\^64 - 
                    10092\ M\^66 - 2557\ M\^68 + 29084\ M\^70 - 7913\ M\^72 - 
                    30725\ M\^74 + 11745\ M\^76 + 11840\ M\^78 - 
                    6713\ M\^80 - 1070\ M\^82 + 1607\ M\^84 - 484\ M\^86 + 
                    64\ M\^88 - 3\ M\^90)\) + 
              L\^9\ \((\(-15\)\ M\^8 + 266\ M\^10 - 2113\ M\^12 + 
                    9929\ M\^14 - 30375\ M\^16 + 60500\ M\^18 - 
                    64183\ M\^20 - 26142\ M\^22 + 209365\ M\^24 - 
                    254807\ M\^26 - 159075\ M\^28 + 737456\ M\^30 - 
                    322212\ M\^32 - 1237343\ M\^34 + 1405038\ M\^36 + 
                    1173824\ M\^38 - 2583361\ M\^40 - 327148\ M\^42 + 
                    3505143\ M\^44 - 1107252\ M\^46 - 3995165\ M\^48 + 
                    2655098\ M\^50 + 3441662\ M\^52 - 4141254\ M\^54 - 
                    1550737\ M\^56 + 5013154\ M\^58 - 784725\ M\^60 - 
                    3986355\ M\^62 + 2028249\ M\^64 + 1716695\ M\^66 - 
                    1762132\ M\^68 + 11833\ M\^70 + 705295\ M\^72 - 
                    413777\ M\^74 + 34879\ M\^76 + 97715\ M\^78 - 
                    77457\ M\^80 + 32635\ M\^82 - 8679\ M\^84 + 1451\ M\^86 - 
                    140\ M\^88 + 6\ M\^90)\) + 
              L\^5\ \((\(-M\^24\) + 13\ M\^26 - 80\ M\^28 + 307\ M\^30 - 
                    803\ M\^32 + 1541\ M\^34 - 2508\ M\^36 + 3817\ M\^38 - 
                    3997\ M\^40 - 509\ M\^42 + 6177\ M\^44 + 333\ M\^46 - 
                    14641\ M\^48 + 7419\ M\^50 + 19608\ M\^52 - 
                    10685\ M\^54 - 27907\ M\^56 - 148\ M\^58 + 48703\ M\^60 + 
                    1610\ M\^62 - 77875\ M\^64 + 24409\ M\^66 + 
                    130938\ M\^68 - 80142\ M\^70 - 119403\ M\^72 + 
                    94607\ M\^74 + 33998\ M\^76 - 48192\ M\^78 + 
                    7644\ M\^80 + 7627\ M\^82 - 4545\ M\^84 + 1142\ M\^86 - 
                    148\ M\^88 + 8\ M\^90)\) + 
              L\^7\ \((\(-M\^12\) + 14\ M\^14 - 105\ M\^16 + 563\ M\^18 - 
                    2265\ M\^20 + 6644\ M\^22 - 13402\ M\^24 + 16078\ M\^26 - 
                    4214\ M\^28 - 16500\ M\^30 + 7121\ M\^32 + 50684\ M\^34 - 
                    55478\ M\^36 - 94050\ M\^38 + 173968\ M\^40 + 
                    86332\ M\^42 - 295894\ M\^44 - 50323\ M\^46 + 
                    459035\ M\^48 - 3427\ M\^50 - 733010\ M\^52 + 
                    84631\ M\^54 + 1076269\ M\^56 - 506180\ M\^58 - 
                    1113577\ M\^60 + 1183210\ M\^62 + 745310\ M\^64 - 
                    1444977\ M\^66 - 173575\ M\^68 + 1033331\ M\^70 - 
                    276862\ M\^72 - 338084\ M\^74 + 249787\ M\^76 - 
                    28793\ M\^78 - 42242\ M\^80 + 30246\ M\^82 - 
                    11494\ M\^84 + 2883\ M\^86 - 473\ M\^88 + 46\ M\^90 - 
                    2\ M\^92)\) + 
              L\^8\ \((\(-6\)\ M\^10 + 96\ M\^12 - 724\ M\^14 + 3429\ M\^16 - 
                    11288\ M\^18 + 26074\ M\^20 - 38064\ M\^22 + 
                    17922\ M\^24 + 52333\ M\^26 - 102127\ M\^28 - 
                    13563\ M\^30 + 247208\ M\^32 - 161127\ M\^34 - 
                    427586\ M\^36 + 576969\ M\^38 + 424895\ M\^40 - 
                    1043878\ M\^42 - 204949\ M\^44 + 1553210\ M\^46 - 
                    242872\ M\^48 - 2108878\ M\^50 + 882217\ M\^52 + 
                    2371734\ M\^54 - 1967461\ M\^56 - 1760057\ M\^58 + 
                    3092865\ M\^60 + 444970\ M\^62 - 2988441\ M\^64 + 
                    662437\ M\^66 + 1683990\ M\^68 - 988271\ M\^70 - 
                    325837\ M\^72 + 528942\ M\^74 - 184697\ M\^76 - 
                    35447\ M\^78 + 68134\ M\^80 - 37989\ M\^82 + 
                    12909\ M\^84 - 2843\ M\^86 + 390\ M\^88 - 30\ M\^90 + 
                    M\^92)\) + 
              L\^6\ \((\(-2\)\ M\^18 + 27\ M\^20 - 177\ M\^22 + 747\ M\^24 - 
                    2186\ M\^26 + 4377\ M\^28 - 5519\ M\^30 + 3869\ M\^32 - 
                    3616\ M\^34 + 11208\ M\^36 - 13093\ M\^38 - 
                    15920\ M\^40 + 41304\ M\^42 + 7532\ M\^44 - 
                    71186\ M\^46 + 12253\ M\^48 + 101784\ M\^50 - 
                    16975\ M\^52 - 163932\ M\^54 - 17605\ M\^56 + 
                    294537\ M\^58 - 50130\ M\^60 - 390355\ M\^62 + 
                    255463\ M\^64 + 412730\ M\^66 - 435382\ M\^68 - 
                    263662\ M\^70 + 397731\ M\^72 + 8085\ M\^74 - 
                    170861\ M\^76 + 68280\ M\^78 + 12530\ M\^80 - 
                    19436\ M\^82 + 8026\ M\^84 - 1945\ M\^86 + 301\ M\^88 - 
                    27\ M\^90 + M\^92)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 132]\), \(L\^8\ M\^6 - M\^20 + 
              L\ \((\(-M\^6\) + M\^8 - 4\ M\^14 + 4\ M\^16 + M\^18 - 
                    3\ M\^20)\) + 
              L\^7\ \((3\ M\^6 - M\^8 - 4\ M\^10 + 4\ M\^12 - M\^18 + 
                    M\^20)\) + 
              L\^3\ \((M\^2 - 3\ M\^4 - 2\ M\^6 + 7\ M\^8 + 12\ M\^10 - 
                    13\ M\^12 - 10\ M\^14 + 9\ M\^16 + 5\ M\^18 + 4\ M\^20 - 
                    4\ M\^22)\) + 
              L\^2\ \((M\^4 - 4\ M\^6 + 5\ M\^8 + 5\ M\^10 - 12\ M\^12 - 
                    7\ M\^14 + 9\ M\^16 + 13\ M\^18 - 6\ M\^20 - 
                    2\ M\^22)\) + 
              L\^6\ \((2\ M\^4 + 6\ M\^6 - 13\ M\^8 - 9\ M\^10 + 7\ M\^12 + 
                    12\ M\^14 - 5\ M\^16 - 5\ M\^18 + 4\ M\^20 - M\^22)\) + 
              L\^5\ \((4\ M\^4 - 4\ M\^6 - 5\ M\^8 - 9\ M\^10 + 10\ M\^12 + 
                    13\ M\^14 - 12\ M\^16 - 7\ M\^18 + 2\ M\^20 + 3\ M\^22 - 
                    M\^24)\) + 
              L\^4\ \((\(-1\) + 7\ M\^2 - 7\ M\^4 - 13\ M\^6 + 6\ M\^8 + 
                    18\ M\^10 + 15\ M\^12 - 15\ M\^14 - 18\ M\^16 - 
                    6\ M\^18 + 13\ M\^20 + 7\ M\^22 - 7\ M\^24 + M\^26)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 133]\), \(L\^15\ M\^2 - M\^66 + 
              L\^14\ \((\(-M\^2\) - 4\ M\^4 + 21\ M\^6 + 3\ M\^8 - 
                    14\ M\^10 + 8\ M\^12 - 6\ M\^14 + 4\ M\^16 - 
                    2\ M\^18)\) + 
              L\^13\ \((2\ M\^2 - 7\ M\^4 - 17\ M\^6 + 23\ M\^8 + 57\ M\^10 + 
                    81\ M\^12 - 93\ M\^14 - 78\ M\^16 + 107\ M\^18 - 
                    20\ M\^20 - 38\ M\^22 + 6\ M\^24 + 23\ M\^26 - 
                    20\ M\^28 + 10\ M\^30 - 4\ M\^32 + M\^34)\) + 
              L\^12\ \((\(-5\)\ M\^2 + 32\ M\^4 - 41\ M\^6 - 79\ M\^8 + 
                    41\ M\^10 + 211\ M\^12 + 145\ M\^14 - 201\ M\^16 - 
                    56\ M\^18 + 114\ M\^20 - 81\ M\^22 - 40\ M\^24 + 
                    6\ M\^26 - 17\ M\^28 + 26\ M\^30 + 51\ M\^32 - 
                    95\ M\^34 + 64\ M\^36 - 21\ M\^38 + 3\ M\^40)\) + 
              L\^11\ \((\(-1\) + 14\ M\^2 - 80\ M\^4 + 195\ M\^6 - 21\ M\^8 - 
                    616\ M\^10 + 278\ M\^12 + 1131\ M\^14 - 399\ M\^16 - 
                    1066\ M\^18 + 602\ M\^20 + 735\ M\^22 - 769\ M\^24 - 
                    263\ M\^26 + 394\ M\^28 - 137\ M\^30 + 20\ M\^32 - 
                    34\ M\^34 + 10\ M\^36 + 145\ M\^38 - 206\ M\^40 + 
                    116\ M\^42 - 30\ M\^44 + 3\ M\^46)\) + 
              L\^10\ \((\(-3\)\ M\^2 + 42\ M\^4 - 212\ M\^6 + 393\ M\^8 + 
                    226\ M\^10 - 1380\ M\^12 - 172\ M\^14 + 2712\ M\^16 + 
                    600\ M\^18 - 3119\ M\^20 - 963\ M\^22 + 2278\ M\^24 + 
                    551\ M\^26 - 1074\ M\^28 - 637\ M\^30 + 720\ M\^32 + 
                    334\ M\^34 - 718\ M\^36 + 426\ M\^38 + 8\ M\^40 - 
                    283\ M\^42 + 293\ M\^44 - 174\ M\^46 + 65\ M\^48 - 
                    13\ M\^50 + M\^52)\) + 
              L\^9\ \((\(-3\)\ M\^4 + 43\ M\^6 - 210\ M\^8 + 318\ M\^10 + 
                    417\ M\^12 - 1103\ M\^14 - 1226\ M\^16 + 2045\ M\^18 + 
                    3299\ M\^20 - 1436\ M\^22 - 5216\ M\^24 - 500\ M\^26 + 
                    4704\ M\^28 + 547\ M\^30 - 3386\ M\^32 + 670\ M\^34 + 
                    1567\ M\^36 - 102\ M\^38 - 402\ M\^40 - 772\ M\^42 + 
                    546\ M\^44 + 253\ M\^46 - 378\ M\^48 + 172\ M\^50 - 
                    37\ M\^52 + 3\ M\^54)\) + 
              L\^8\ \((\(-M\^6\) + 16\ M\^8 - 82\ M\^10 + 109\ M\^12 + 
                    200\ M\^14 - 164\ M\^16 - 1028\ M\^18 - 490\ M\^20 + 
                    3469\ M\^22 + 2634\ M\^24 - 5679\ M\^26 - 4263\ M\^28 + 
                    4646\ M\^30 + 2718\ M\^32 - 3461\ M\^34 - 1640\ M\^36 + 
                    3290\ M\^38 + 2769\ M\^40 - 2620\ M\^42 - 2016\ M\^44 + 
                    1450\ M\^46 + 312\ M\^48 - 348\ M\^50 + 112\ M\^52 - 
                    42\ M\^54 + 11\ M\^56 - M\^58)\) + 
              L\^7\ \((M\^10 - 11\ M\^12 + 42\ M\^14 - 112\ M\^16 + 
                    348\ M\^18 - 312\ M\^20 - 1450\ M\^22 + 2016\ M\^24 + 
                    2620\ M\^26 - 2769\ M\^28 - 3290\ M\^30 + 1640\ M\^32 + 
                    3461\ M\^34 - 2718\ M\^36 - 4646\ M\^38 + 4263\ M\^40 + 
                    5679\ M\^42 - 2634\ M\^44 - 3469\ M\^46 + 490\ M\^48 + 
                    1028\ M\^50 + 164\ M\^52 - 200\ M\^54 - 109\ M\^56 + 
                    82\ M\^58 - 16\ M\^60 + M\^62)\) + 
              L\^6\ \((\(-3\)\ M\^14 + 37\ M\^16 - 172\ M\^18 + 378\ M\^20 - 
                    253\ M\^22 - 546\ M\^24 + 772\ M\^26 + 402\ M\^28 + 
                    102\ M\^30 - 1567\ M\^32 - 670\ M\^34 + 3386\ M\^36 - 
                    547\ M\^38 - 4704\ M\^40 + 500\ M\^42 + 5216\ M\^44 + 
                    1436\ M\^46 - 3299\ M\^48 - 2045\ M\^50 + 1226\ M\^52 + 
                    1103\ M\^54 - 417\ M\^56 - 318\ M\^58 + 210\ M\^60 - 
                    43\ M\^62 + 3\ M\^64)\) + 
              L\^2\ \((\(-M\^34\) + 4\ M\^36 - 10\ M\^38 + 20\ M\^40 - 
                    23\ M\^42 - 6\ M\^44 + 38\ M\^46 + 20\ M\^48 - 
                    107\ M\^50 + 78\ M\^52 + 93\ M\^54 - 81\ M\^56 - 
                    57\ M\^58 - 23\ M\^60 + 17\ M\^62 + 7\ M\^64 - 
                    2\ M\^66)\) + 
              L\ \((2\ M\^50 - 4\ M\^52 + 6\ M\^54 - 8\ M\^56 + 14\ M\^58 - 
                    3\ M\^60 - 21\ M\^62 + 4\ M\^64 + M\^66)\) + 
              L\^5\ \((\(-M\^16\) + 13\ M\^18 - 65\ M\^20 + 174\ M\^22 - 
                    293\ M\^24 + 283\ M\^26 - 8\ M\^28 - 426\ M\^30 + 
                    718\ M\^32 - 334\ M\^34 - 720\ M\^36 + 637\ M\^38 + 
                    1074\ M\^40 - 551\ M\^42 - 2278\ M\^44 + 963\ M\^46 + 
                    3119\ M\^48 - 600\ M\^50 - 2712\ M\^52 + 172\ M\^54 + 
                    1380\ M\^56 - 226\ M\^58 - 393\ M\^60 + 212\ M\^62 - 
                    42\ M\^64 + 3\ M\^66)\) + 
              L\^3\ \((\(-3\)\ M\^28 + 21\ M\^30 - 64\ M\^32 + 95\ M\^34 - 
                    51\ M\^36 - 26\ M\^38 + 17\ M\^40 - 6\ M\^42 + 
                    40\ M\^44 + 81\ M\^46 - 114\ M\^48 + 56\ M\^50 + 
                    201\ M\^52 - 145\ M\^54 - 211\ M\^56 - 41\ M\^58 + 
                    79\ M\^60 + 41\ M\^62 - 32\ M\^64 + 5\ M\^66)\) + 
              L\^4\ \((\(-3\)\ M\^22 + 30\ M\^24 - 116\ M\^26 + 206\ M\^28 - 
                    145\ M\^30 - 10\ M\^32 + 34\ M\^34 - 20\ M\^36 + 
                    137\ M\^38 - 394\ M\^40 + 263\ M\^42 + 769\ M\^44 - 
                    735\ M\^46 - 602\ M\^48 + 1066\ M\^50 + 399\ M\^52 - 
                    1131\ M\^54 - 278\ M\^56 + 616\ M\^58 + 21\ M\^60 - 
                    195\ M\^62 + 80\ M\^64 - 14\ M\^66 + M\^68)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 134]\), \(\(-1\) + 
              L\^17\ M\^200 + 
              L\ \((1 - 4\ M\^2 + 9\ M\^4 - 19\ M\^6 + 15\ M\^8 + 7\ M\^10 - 
                    25\ M\^12 + 3\ M\^14 + 2\ M\^16)\) + 
              L\^2\ \((\(-M\^4\) + 5\ M\^6 - 19\ M\^8 + 48\ M\^10 - 
                    93\ M\^12 + 125\ M\^14 - 43\ M\^16 - 140\ M\^18 + 
                    118\ M\^20 + 23\ M\^22 - 48\ M\^24 - 54\ M\^26 + 
                    16\ M\^28 + 16\ M\^30 - 5\ M\^32)\) + 
              L\^3\ \((\(-M\^10\) + 7\ M\^12 - 27\ M\^14 + 86\ M\^16 - 
                    215\ M\^18 + 353\ M\^20 - 339\ M\^22 + 44\ M\^24 + 
                    288\ M\^26 - 187\ M\^28 + 25\ M\^30 - 444\ M\^32 + 
                    145\ M\^34 + 463\ M\^36 - 163\ M\^38 - 344\ M\^40 + 
                    123\ M\^42 + 123\ M\^44 - 87\ M\^46 + 20\ M\^48 - 
                    2\ M\^50)\) + 
              L\^4\ \((M\^18 - 11\ M\^20 + 59\ M\^22 - 196\ M\^24 + 
                    433\ M\^26 - 667\ M\^28 + 794\ M\^30 - 643\ M\^32 - 
                    216\ M\^34 + 1136\ M\^36 - 665\ M\^38 - 164\ M\^40 - 
                    193\ M\^42 + 123\ M\^44 - 539\ M\^46 + 618\ M\^48 + 
                    768\ M\^50 - 863\ M\^52 - 207\ M\^54 + 233\ M\^56 + 
                    126\ M\^58 - 113\ M\^60 + 20\ M\^62 - 4\ M\^64 + 
                    2\ M\^66)\) + 
              L\^5\ \((4\ M\^28 - 40\ M\^30 + 186\ M\^32 - 524\ M\^34 + 
                    998\ M\^36 - 1372\ M\^38 + 1257\ M\^40 - 112\ M\^42 - 
                    1239\ M\^44 + 129\ M\^46 + 1258\ M\^48 + 3254\ M\^50 - 
                    6727\ M\^52 - 1877\ M\^54 + 7954\ M\^56 - 462\ M\^58 - 
                    6090\ M\^60 + 3085\ M\^62 + 3271\ M\^64 - 4325\ M\^66 + 
                    683\ M\^68 + 1438\ M\^70 - 870\ M\^72 + 61\ M\^74 + 
                    98\ M\^76 - 56\ M\^78 + 22\ M\^80 - 4\ M\^82)\) + 
              L\^6\ \((6\ M\^38 - 64\ M\^40 + 297\ M\^42 - 779\ M\^44 + 
                    1330\ M\^46 - 1803\ M\^48 + 2070\ M\^50 - 790\ M\^52 - 
                    2341\ M\^54 + 2942\ M\^56 - 518\ M\^58 + 1776\ M\^60 - 
                    2057\ M\^62 - 2603\ M\^64 - 3370\ M\^66 + 7378\ M\^68 + 
                    8527\ M\^70 - 12004\ M\^72 - 5377\ M\^74 + 10944\ M\^76 + 
                    2229\ M\^78 - 9794\ M\^80 + 3224\ M\^82 + 3533\ M\^84 - 
                    3316\ M\^86 + 966\ M\^88 + 169\ M\^90 - 408\ M\^92 + 
                    278\ M\^94 - 96\ M\^96 + 16\ M\^98 - M\^100)\) + 
              L\^7\ \((4\ M\^48 - 47\ M\^50 + 240\ M\^52 - 710\ M\^54 + 
                    1414\ M\^56 - 2069\ M\^58 + 1881\ M\^60 + 69\ M\^62 - 
                    1551\ M\^64 - 262\ M\^66 - 2010\ M\^68 + 11152\ M\^70 - 
                    5383\ M\^72 - 13198\ M\^74 + 5912\ M\^76 + 11590\ M\^78 - 
                    4639\ M\^80 - 3898\ M\^82 + 6972\ M\^84 - 8422\ M\^86 + 
                    223\ M\^88 + 7455\ M\^90 - 2189\ M\^92 - 6258\ M\^94 + 
                    4030\ M\^96 + 3352\ M\^98 - 4951\ M\^100 + 2036\ M\^102 + 
                    420\ M\^104 - 1131\ M\^106 + 759\ M\^108 - 261\ M\^110 + 
                    45\ M\^112 - 3\ M\^114)\) + 
              L\^8\ \((M\^58 - 13\ M\^60 + 80\ M\^62 - 327\ M\^64 + 
                    971\ M\^66 - 1906\ M\^68 + 1844\ M\^70 + 316\ M\^72 - 
                    1815\ M\^74 + 140\ M\^76 - 836\ M\^78 + 6079\ M\^80 - 
                    5741\ M\^82 + 360\ M\^84 + 831\ M\^86 - 5916\ M\^88 + 
                    7726\ M\^90 + 4953\ M\^92 - 5801\ M\^94 - 3052\ M\^96 + 
                    6889\ M\^98 - 8982\ M\^100 + 1532\ M\^102 + 
                    7759\ M\^104 - 4613\ M\^106 - 2424\ M\^108 + 
                    1192\ M\^110 + 3783\ M\^112 - 3786\ M\^114 + 
                    431\ M\^116 + 1545\ M\^118 - 1514\ M\^120 + 786\ M\^122 - 
                    245\ M\^124 + 42\ M\^126 - 3\ M\^128)\) + 
              L\^9\ \((3\ M\^72 - 42\ M\^74 + 245\ M\^76 - 786\ M\^78 + 
                    1514\ M\^80 - 1545\ M\^82 - 431\ M\^84 + 3786\ M\^86 - 
                    3783\ M\^88 - 1192\ M\^90 + 2424\ M\^92 + 4613\ M\^94 - 
                    7759\ M\^96 - 1532\ M\^98 + 8982\ M\^100 - 6889\ M\^102 + 
                    3052\ M\^104 + 5801\ M\^106 - 4953\ M\^108 - 
                    7726\ M\^110 + 5916\ M\^112 - 831\ M\^114 - 360\ M\^116 + 
                    5741\ M\^118 - 6079\ M\^120 + 836\ M\^122 - 140\ M\^124 + 
                    1815\ M\^126 - 316\ M\^128 - 1844\ M\^130 + 
                    1906\ M\^132 - 971\ M\^134 + 327\ M\^136 - 80\ M\^138 + 
                    13\ M\^140 - M\^142)\) + 
              L\^10\ \((3\ M\^86 - 45\ M\^88 + 261\ M\^90 - 759\ M\^92 + 
                    1131\ M\^94 - 420\ M\^96 - 2036\ M\^98 + 4951\ M\^100 - 
                    3352\ M\^102 - 4030\ M\^104 + 6258\ M\^106 + 
                    2189\ M\^108 - 7455\ M\^110 - 223\ M\^112 + 
                    8422\ M\^114 - 6972\ M\^116 + 3898\ M\^118 + 
                    4639\ M\^120 - 11590\ M\^122 - 5912\ M\^124 + 
                    13198\ M\^126 + 5383\ M\^128 - 11152\ M\^130 + 
                    2010\ M\^132 + 262\ M\^134 + 1551\ M\^136 - 69\ M\^138 - 
                    1881\ M\^140 + 2069\ M\^142 - 1414\ M\^144 + 
                    710\ M\^146 - 240\ M\^148 + 47\ M\^150 - 4\ M\^152)\) + 
              L\^11\ \((M\^100 - 16\ M\^102 + 96\ M\^104 - 278\ M\^106 + 
                    408\ M\^108 - 169\ M\^110 - 966\ M\^112 + 3316\ M\^114 - 
                    3533\ M\^116 - 3224\ M\^118 + 9794\ M\^120 - 
                    2229\ M\^122 - 10944\ M\^124 + 5377\ M\^126 + 
                    12004\ M\^128 - 8527\ M\^130 - 7378\ M\^132 + 
                    3370\ M\^134 + 2603\ M\^136 + 2057\ M\^138 - 
                    1776\ M\^140 + 518\ M\^142 - 2942\ M\^144 + 
                    2341\ M\^146 + 790\ M\^148 - 2070\ M\^150 + 
                    1803\ M\^152 - 1330\ M\^154 + 779\ M\^156 - 297\ M\^158 + 
                    64\ M\^160 - 6\ M\^162)\) + 
              L\^12\ \((4\ M\^118 - 22\ M\^120 + 56\ M\^122 - 98\ M\^124 - 
                    61\ M\^126 + 870\ M\^128 - 1438\ M\^130 - 683\ M\^132 + 
                    4325\ M\^134 - 3271\ M\^136 - 3085\ M\^138 + 
                    6090\ M\^140 + 462\ M\^142 - 7954\ M\^144 + 
                    1877\ M\^146 + 6727\ M\^148 - 3254\ M\^150 - 
                    1258\ M\^152 - 129\ M\^154 + 1239\ M\^156 + 112\ M\^158 - 
                    1257\ M\^160 + 1372\ M\^162 - 998\ M\^164 + 524\ M\^166 - 
                    186\ M\^168 + 40\ M\^170 - 4\ M\^172)\) + 
              L\^13\ \((\(-2\)\ M\^134 + 4\ M\^136 - 20\ M\^138 + 
                    113\ M\^140 - 126\ M\^142 - 233\ M\^144 + 207\ M\^146 + 
                    863\ M\^148 - 768\ M\^150 - 618\ M\^152 + 539\ M\^154 - 
                    123\ M\^156 + 193\ M\^158 + 164\ M\^160 + 665\ M\^162 - 
                    1136\ M\^164 + 216\ M\^166 + 643\ M\^168 - 794\ M\^170 + 
                    667\ M\^172 - 433\ M\^174 + 196\ M\^176 - 59\ M\^178 + 
                    11\ M\^180 - M\^182)\) + 
              L\^14\ \((2\ M\^150 - 20\ M\^152 + 87\ M\^154 - 123\ M\^156 - 
                    123\ M\^158 + 344\ M\^160 + 163\ M\^162 - 463\ M\^164 - 
                    145\ M\^166 + 444\ M\^168 - 25\ M\^170 + 187\ M\^172 - 
                    288\ M\^174 - 44\ M\^176 + 339\ M\^178 - 353\ M\^180 + 
                    215\ M\^182 - 86\ M\^184 + 27\ M\^186 - 7\ M\^188 + 
                    M\^190)\) + 
              L\^15\ \((5\ M\^168 - 16\ M\^170 - 16\ M\^172 + 54\ M\^174 + 
                    48\ M\^176 - 23\ M\^178 - 118\ M\^180 + 140\ M\^182 + 
                    43\ M\^184 - 125\ M\^186 + 93\ M\^188 - 48\ M\^190 + 
                    19\ M\^192 - 5\ M\^194 + M\^196)\) + 
              L\^16\ \((\(-2\)\ M\^184 - 3\ M\^186 + 25\ M\^188 - 7\ M\^190 - 
                    15\ M\^192 + 19\ M\^194 - 9\ M\^196 + 4\ M\^198 - 
                    M\^200)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 135]\), \(\(-M\^56\) + 
              L\^24\ M\^58 + 
              L\ \((\(-3\)\ M\^46 + 11\ M\^48 - 28\ M\^50 + 45\ M\^52 - 
                    3\ M\^54 - 62\ M\^56 + 19\ M\^58 + 8\ M\^60 - 
                    5\ M\^62)\) + 
              L\^23\ \((5\ M\^52 - 8\ M\^54 - 19\ M\^56 + 62\ M\^58 + 
                    3\ M\^60 - 45\ M\^62 + 28\ M\^64 - 11\ M\^66 + 
                    3\ M\^68)\) + 
              L\^2\ \((\(-3\)\ M\^36 + 22\ M\^38 - 85\ M\^40 + 197\ M\^42 - 
                    244\ M\^44 + 165\ M\^46 - 81\ M\^48 - 269\ M\^50 + 
                    698\ M\^52 + 121\ M\^54 - 942\ M\^56 - 97\ M\^58 + 
                    617\ M\^60 - 204\ M\^62 - 121\ M\^64 + 100\ M\^66 - 
                    26\ M\^68 + 2\ M\^70)\) + 
              L\^3\ \((\(-M\^26\) + 11\ M\^28 - 59\ M\^30 + 196\ M\^32 - 
                    486\ M\^34 + 1017\ M\^36 - 1640\ M\^38 + 1184\ M\^40 + 
                    1140\ M\^42 - 2621\ M\^44 + 1247\ M\^46 + 1184\ M\^48 - 
                    5304\ M\^50 + 5475\ M\^52 + 7040\ M\^54 - 11518\ M\^56 - 
                    2903\ M\^58 + 7780\ M\^60 - 1012\ M\^62 - 2892\ M\^64 + 
                    1506\ M\^66 + 232\ M\^68 - 550\ M\^70 + 270\ M\^72 - 
                    64\ M\^74 + 6\ M\^76)\) + 
              L\^22\ \((\(-2\)\ M\^44 + 26\ M\^46 - 100\ M\^48 + 121\ M\^50 + 
                    204\ M\^52 - 617\ M\^54 + 97\ M\^56 + 942\ M\^58 - 
                    121\ M\^60 - 698\ M\^62 + 269\ M\^64 + 81\ M\^66 - 
                    165\ M\^68 + 244\ M\^70 - 197\ M\^72 + 85\ M\^74 - 
                    22\ M\^76 + 3\ M\^78)\) + 
              L\^4\ \((\(-2\)\ M\^20 + 27\ M\^22 - 184\ M\^24 + 769\ M\^26 - 
                    2064\ M\^28 + 3590\ M\^30 - 4027\ M\^32 + 2633\ M\^34 + 
                    882\ M\^36 - 6047\ M\^38 + 4456\ M\^40 + 9900\ M\^42 - 
                    19017\ M\^44 + 6457\ M\^46 + 23751\ M\^48 - 
                    49990\ M\^50 + 5447\ M\^52 + 90044\ M\^54 - 
                    49458\ M\^56 - 85816\ M\^58 + 76152\ M\^60 + 
                    21907\ M\^62 - 52326\ M\^64 + 18596\ M\^66 + 
                    9703\ M\^68 - 12199\ M\^70 + 4272\ M\^72 + 1069\ M\^74 - 
                    1780\ M\^76 + 821\ M\^78 - 191\ M\^80 + 22\ M\^82 - 
                    M\^84)\) + 
              L\^21\ \((\(-6\)\ M\^38 + 64\ M\^40 - 270\ M\^42 + 550\ M\^44 - 
                    232\ M\^46 - 1506\ M\^48 + 2892\ M\^50 + 1012\ M\^52 - 
                    7780\ M\^54 + 2903\ M\^56 + 11518\ M\^58 - 7040\ M\^60 - 
                    5475\ M\^62 + 5304\ M\^64 - 1184\ M\^66 - 1247\ M\^68 + 
                    2621\ M\^70 - 1140\ M\^72 - 1184\ M\^74 + 1640\ M\^76 - 
                    1017\ M\^78 + 486\ M\^80 - 196\ M\^82 + 59\ M\^84 - 
                    11\ M\^86 + M\^88)\) + 
              L\^5\ \((\(-M\^14\) + 19\ M\^16 - 177\ M\^18 + 995\ M\^20 - 
                    3640\ M\^22 + 8804\ M\^24 - 13307\ M\^26 + 9387\ M\^28 + 
                    5184\ M\^30 - 16075\ M\^32 + 4694\ M\^34 + 28884\ M\^36 - 
                    55669\ M\^38 + 18663\ M\^40 + 82320\ M\^42 - 
                    127833\ M\^44 + 5629\ M\^46 + 237449\ M\^48 - 
                    252414\ M\^50 - 260905\ M\^52 + 599644\ M\^54 + 
                    90332\ M\^56 - 727854\ M\^58 + 168698\ M\^60 + 
                    504992\ M\^62 - 339978\ M\^64 - 121974\ M\^66 + 
                    246182\ M\^68 - 101147\ M\^70 - 35060\ M\^72 + 
                    66090\ M\^74 - 33418\ M\^76 + 316\ M\^78 + 9089\ M\^80 - 
                    5485\ M\^82 + 1644\ M\^84 - 276\ M\^86 + 25\ M\^88 - 
                    M\^90)\) + 
              L\^20\ \((M\^30 - 22\ M\^32 + 191\ M\^34 - 821\ M\^36 + 
                    1780\ M\^38 - 1069\ M\^40 - 4272\ M\^42 + 12199\ M\^44 - 
                    9703\ M\^46 - 18596\ M\^48 + 52326\ M\^50 - 
                    21907\ M\^52 - 76152\ M\^54 + 85816\ M\^56 + 
                    49458\ M\^58 - 90044\ M\^60 - 5447\ M\^62 + 
                    49990\ M\^64 - 23751\ M\^66 - 6457\ M\^68 + 
                    19017\ M\^70 - 9900\ M\^72 - 4456\ M\^74 + 6047\ M\^76 - 
                    882\ M\^78 - 2633\ M\^80 + 4027\ M\^82 - 3590\ M\^84 + 
                    2064\ M\^86 - 769\ M\^88 + 184\ M\^90 - 27\ M\^92 + 
                    2\ M\^94)\) + 
              L\^6\ \((3\ M\^10 - 55\ M\^12 + 473\ M\^14 - 2469\ M\^16 + 
                    8527\ M\^18 - 19993\ M\^20 + 30808\ M\^22 - 
                    25408\ M\^24 - 5819\ M\^26 + 40191\ M\^28 - 
                    26596\ M\^30 - 57570\ M\^32 + 143460\ M\^34 - 
                    66116\ M\^36 - 226135\ M\^38 + 324382\ M\^40 + 
                    219099\ M\^42 - 811298\ M\^44 + 64879\ M\^46 + 
                    1635590\ M\^48 - 1066736\ M\^50 - 2261159\ M\^52 + 
                    2531476\ M\^54 + 2128885\ M\^56 - 3567887\ M\^58 - 
                    971476\ M\^60 + 3263884\ M\^62 - 440530\ M\^64 - 
                    1904808\ M\^66 + 1066025\ M\^68 + 463197\ M\^70 - 
                    771456\ M\^72 + 296096\ M\^74 + 127471\ M\^76 - 
                    225391\ M\^78 + 115608\ M\^80 - 1201\ M\^82 - 
                    33330\ M\^84 + 21293\ M\^86 - 6927\ M\^88 + 1299\ M\^90 - 
                    134\ M\^92 + 6\ M\^94)\) + 
              L\^7\ \((\(-3\)\ M\^6 + 60\ M\^8 - 544\ M\^10 + 2968\ M\^12 - 
                    10845\ M\^14 + 27798\ M\^16 - 50411\ M\^18 + 
                    61199\ M\^20 - 34138\ M\^22 - 37277\ M\^24 + 
                    93507\ M\^26 - 15539\ M\^28 - 199120\ M\^30 + 
                    234478\ M\^32 + 201171\ M\^34 - 457376\ M\^36 - 
                    467609\ M\^38 + 1160878\ M\^40 + 1084805\ M\^42 - 
                    3536453\ M\^44 - 800113\ M\^46 + 7693119\ M\^48 - 
                    1690291\ M\^50 - 12158593\ M\^52 + 6780816\ M\^54 + 
                    13554176\ M\^56 - 11640320\ M\^58 - 10153124\ M\^60 + 
                    12962090\ M\^62 + 3971900\ M\^64 - 10079718\ M\^66 + 
                    921932\ M\^68 + 5139409\ M\^70 - 2411312\ M\^72 - 
                    1279346\ M\^74 + 1683706\ M\^76 - 477563\ M\^78 - 
                    381637\ M\^80 + 491337\ M\^82 - 217002\ M\^84 - 
                    22395\ M\^86 + 83760\ M\^88 - 50516\ M\^90 + 
                    16397\ M\^92 - 3126\ M\^94 + 330\ M\^96 - 15\ M\^98)\) + 
              L\^19\ \((M\^24 - 25\ M\^26 + 276\ M\^28 - 1644\ M\^30 + 
                    5485\ M\^32 - 9089\ M\^34 - 316\ M\^36 + 33418\ M\^38 - 
                    66090\ M\^40 + 35060\ M\^42 + 101147\ M\^44 - 
                    246182\ M\^46 + 121974\ M\^48 + 339978\ M\^50 - 
                    504992\ M\^52 - 168698\ M\^54 + 727854\ M\^56 - 
                    90332\ M\^58 - 599644\ M\^60 + 260905\ M\^62 + 
                    252414\ M\^64 - 237449\ M\^66 - 5629\ M\^68 + 
                    127833\ M\^70 - 82320\ M\^72 - 18663\ M\^74 + 
                    55669\ M\^76 - 28884\ M\^78 - 4694\ M\^80 + 
                    16075\ M\^82 - 5184\ M\^84 - 9387\ M\^86 + 13307\ M\^88 - 
                    8804\ M\^90 + 3640\ M\^92 - 995\ M\^94 + 177\ M\^96 - 
                    19\ M\^98 + M\^100)\) + 
              L\^8\ \((M\^2 - 23\ M\^4 + 243\ M\^6 - 1577\ M\^8 + 
                    7053\ M\^10 - 22872\ M\^12 + 54398\ M\^14 - 
                    92142\ M\^16 + 98148\ M\^18 - 28631\ M\^20 - 
                    84675\ M\^22 + 90977\ M\^24 + 89159\ M\^26 - 
                    202894\ M\^28 + 1036\ M\^30 + 25891\ M\^32 + 
                    574869\ M\^34 - 61387\ M\^36 - 2833440\ M\^38 + 
                    2304974\ M\^40 + 6573187\ M\^42 - 9620053\ M\^44 - 
                    9646255\ M\^46 + 24207448\ M\^48 + 6450005\ M\^50 - 
                    41122639\ M\^52 + 5266228\ M\^54 + 50910049\ M\^56 - 
                    20941918\ M\^58 - 46002123\ M\^60 + 31546528\ M\^62 + 
                    28724252\ M\^64 - 30919631\ M\^66 - 10223296\ M\^68 + 
                    21326182\ M\^70 - 1006570\ M\^72 - 9866401\ M\^74 + 
                    3723804\ M\^76 + 2605044\ M\^78 - 2504537\ M\^80 + 
                    261860\ M\^82 + 798901\ M\^84 - 664022\ M\^86 + 
                    190489\ M\^88 + 101837\ M\^90 - 139654\ M\^92 + 
                    74153\ M\^94 - 22912\ M\^96 + 4268\ M\^98 - 445\ M\^100 + 
                    20\ M\^102)\) + 
              L\^18\ \((\(-6\)\ M\^20 + 134\ M\^22 - 1299\ M\^24 + 
                    6927\ M\^26 - 21293\ M\^28 + 33330\ M\^30 + 1201\ M\^32 - 
                    115608\ M\^34 + 225391\ M\^36 - 127471\ M\^38 - 
                    296096\ M\^40 + 771456\ M\^42 - 463197\ M\^44 - 
                    1066025\ M\^46 + 1904808\ M\^48 + 440530\ M\^50 - 
                    3263884\ M\^52 + 971476\ M\^54 + 3567887\ M\^56 - 
                    2128885\ M\^58 - 2531476\ M\^60 + 2261159\ M\^62 + 
                    1066736\ M\^64 - 1635590\ M\^66 - 64879\ M\^68 + 
                    811298\ M\^70 - 219099\ M\^72 - 324382\ M\^74 + 
                    226135\ M\^76 + 66116\ M\^78 - 143460\ M\^80 + 
                    57570\ M\^82 + 26596\ M\^84 - 40191\ M\^86 + 
                    5819\ M\^88 + 25408\ M\^90 - 30808\ M\^92 + 
                    19993\ M\^94 - 8527\ M\^96 + 2469\ M\^98 - 473\ M\^100 + 
                    55\ M\^102 - 3\ M\^104)\) + 
              L\^9\ \((1 - 24\ M\^2 + 270\ M\^4 - 1877\ M\^6 + 8907\ M\^8 - 
                    30005\ M\^10 + 72288\ M\^12 - 121789\ M\^14 + 
                    131662\ M\^16 - 60921\ M\^18 - 52421\ M\^20 + 
                    95999\ M\^22 + 12136\ M\^24 - 277380\ M\^26 + 
                    506783\ M\^28 + 28291\ M\^30 - 1853879\ M\^32 + 
                    2546755\ M\^34 + 2786252\ M\^36 - 10151707\ M\^38 + 
                    1399861\ M\^40 + 23482347\ M\^42 - 19466651\ M\^44 - 
                    34899203\ M\^46 + 53538384\ M\^48 + 33733508\ M\^50 - 
                    94613787\ M\^52 - 12711072\ M\^54 + 122417423\ M\^56 - 
                    21179440\ M\^58 - 119695280\ M\^60 + 50135157\ M\^62 + 
                    88469489\ M\^64 - 60404825\ M\^66 - 46735031\ M\^68 + 
                    49901697\ M\^70 + 14543084\ M\^72 - 29981495\ M\^74 + 
                    1030822\ M\^76 + 12661605\ M\^78 - 3927311\ M\^80 - 
                    3443491\ M\^82 + 2360735\ M\^84 + 379970\ M\^86 - 
                    1024376\ M\^88 + 474502\ M\^90 + 31294\ M\^92 - 
                    193925\ M\^94 + 149244\ M\^96 - 66474\ M\^98 + 
                    18900\ M\^100 - 3364\ M\^102 + 341\ M\^104 - 
                    15\ M\^106)\) + 
              L\^17\ \((15\ M\^16 - 330\ M\^18 + 3126\ M\^20 - 16397\ M\^22 + 
                    50516\ M\^24 - 83760\ M\^26 + 22395\ M\^28 + 
                    217002\ M\^30 - 491337\ M\^32 + 381637\ M\^34 + 
                    477563\ M\^36 - 1683706\ M\^38 + 1279346\ M\^40 + 
                    2411312\ M\^42 - 5139409\ M\^44 - 921932\ M\^46 + 
                    10079718\ M\^48 - 3971900\ M\^50 - 12962090\ M\^52 + 
                    10153124\ M\^54 + 11640320\ M\^56 - 13554176\ M\^58 - 
                    6780816\ M\^60 + 12158593\ M\^62 + 1690291\ M\^64 - 
                    7693119\ M\^66 + 800113\ M\^68 + 3536453\ M\^70 - 
                    1084805\ M\^72 - 1160878\ M\^74 + 467609\ M\^76 + 
                    457376\ M\^78 - 201171\ M\^80 - 234478\ M\^82 + 
                    199120\ M\^84 + 15539\ M\^86 - 93507\ M\^88 + 
                    37277\ M\^90 + 34138\ M\^92 - 61199\ M\^94 + 
                    50411\ M\^96 - 27798\ M\^98 + 10845\ M\^100 - 
                    2968\ M\^102 + 544\ M\^104 - 60\ M\^106 + 3\ M\^108)\) + 
              L\^10\ \((\(-4\) + 90\ M\^2 - 919\ M\^4 + 5615\ M\^6 - 
                    22765\ M\^8 + 63926\ M\^10 - 124613\ M\^12 + 
                    160136\ M\^14 - 111849\ M\^16 + 9086\ M\^18 - 
                    10472\ M\^20 + 194789\ M\^22 - 224976\ M\^24 - 
                    508388\ M\^26 + 1616573\ M\^28 - 597374\ M\^30 - 
                    4054828\ M\^32 + 6039514\ M\^34 + 5056288\ M\^36 - 
                    19710237\ M\^38 + 3880824\ M\^40 + 39605891\ M\^42 - 
                    32293953\ M\^44 - 57347514\ M\^46 + 85081007\ M\^48 + 
                    56168900\ M\^50 - 148115730\ M\^52 - 26702848\ M\^54 + 
                    193093798\ M\^56 - 21576278\ M\^58 - 195729580\ M\^60 + 
                    66593408\ M\^62 + 153784600\ M\^64 - 86226707\ M\^66 - 
                    92494155\ M\^68 + 77042907\ M\^70 + 39355369\ M\^72 - 
                    51235801\ M\^74 - 8714462\ M\^76 + 25840390\ M\^78 - 
                    2137799\ M\^80 - 9581845\ M\^82 + 2949136\ M\^84 + 
                    2334449\ M\^86 - 1177235\ M\^88 - 639797\ M\^90 + 
                    638345\ M\^92 - 37442\ M\^94 - 225411\ M\^96 + 
                    193056\ M\^98 - 97546\ M\^100 + 34695\ M\^102 - 
                    8685\ M\^104 + 1439\ M\^106 - 140\ M\^108 + 
                    6\ M\^110)\) + 
              L\^16\ \((\(-20\)\ M\^12 + 445\ M\^14 - 4268\ M\^16 + 
                    22912\ M\^18 - 74153\ M\^20 + 139654\ M\^22 - 
                    101837\ M\^24 - 190489\ M\^26 + 664022\ M\^28 - 
                    798901\ M\^30 - 261860\ M\^32 + 2504537\ M\^34 - 
                    2605044\ M\^36 - 3723804\ M\^38 + 9866401\ M\^40 + 
                    1006570\ M\^42 - 21326182\ M\^44 + 10223296\ M\^46 + 
                    30919631\ M\^48 - 28724252\ M\^50 - 31546528\ M\^52 + 
                    46002123\ M\^54 + 20941918\ M\^56 - 50910049\ M\^58 - 
                    5266228\ M\^60 + 41122639\ M\^62 - 6450005\ M\^64 - 
                    24207448\ M\^66 + 9646255\ M\^68 + 9620053\ M\^70 - 
                    6573187\ M\^72 - 2304974\ M\^74 + 2833440\ M\^76 + 
                    61387\ M\^78 - 574869\ M\^80 - 25891\ M\^82 - 
                    1036\ M\^84 + 202894\ M\^86 - 89159\ M\^88 - 
                    90977\ M\^90 + 84675\ M\^92 + 28631\ M\^94 - 
                    98148\ M\^96 + 92142\ M\^98 - 54398\ M\^100 + 
                    22872\ M\^102 - 7053\ M\^104 + 1577\ M\^106 - 
                    243\ M\^108 + 23\ M\^110 - M\^112)\) + 
              L\^13\ \((1 - 24\ M\^2 + 266\ M\^4 - 1861\ M\^6 + 9381\ M\^8 - 
                    35714\ M\^10 + 100673\ M\^12 - 193365\ M\^14 + 
                    202353\ M\^16 - 938\ M\^18 - 169327\ M\^20 - 
                    191529\ M\^22 + 350460\ M\^24 + 1596294\ M\^26 - 
                    3345915\ M\^28 - 2474881\ M\^30 + 11091816\ M\^32 + 
                    724891\ M\^34 - 27873161\ M\^36 + 11365576\ M\^38 + 
                    53580023\ M\^40 - 42220630\ M\^42 - 81285399\ M\^44 + 
                    93663738\ M\^46 + 96518899\ M\^48 - 152557058\ M\^50 - 
                    86970160\ M\^52 + 192755590\ M\^54 + 54506856\ M\^56 - 
                    194001919\ M\^58 - 11372697\ M\^60 + 154587197\ M\^62 - 
                    21649526\ M\^64 - 97159910\ M\^66 + 34123281\ M\^68 + 
                    45740566\ M\^70 - 29125563\ M\^72 - 13537877\ M\^74 + 
                    16752195\ M\^76 + 785832\ M\^78 - 7021803\ M\^80 + 
                    2074282\ M\^82 + 1771108\ M\^84 - 1291888\ M\^86 - 
                    222879\ M\^88 + 621101\ M\^90 - 269303\ M\^92 - 
                    29703\ M\^94 + 124136\ M\^96 - 159945\ M\^98 + 
                    176265\ M\^100 - 143438\ M\^102 + 80802\ M\^104 - 
                    31191\ M\^106 + 8134\ M\^108 - 1374\ M\^110 + 
                    136\ M\^112 - 6\ M\^114)\) + 
              L\^15\ \((15\ M\^8 - 341\ M\^10 + 3364\ M\^12 - 18900\ M\^14 + 
                    66474\ M\^16 - 149244\ M\^18 + 193925\ M\^20 - 
                    31294\ M\^22 - 474502\ M\^24 + 1024376\ M\^26 - 
                    379970\ M\^28 - 2360735\ M\^30 + 3443491\ M\^32 + 
                    3927311\ M\^34 - 12661605\ M\^36 - 1030822\ M\^38 + 
                    29981495\ M\^40 - 14543084\ M\^42 - 49901697\ M\^44 + 
                    46735031\ M\^46 + 60404825\ M\^48 - 88469489\ M\^50 - 
                    50135157\ M\^52 + 119695280\ M\^54 + 21179440\ M\^56 - 
                    122417423\ M\^58 + 12711072\ M\^60 + 94613787\ M\^62 - 
                    33733508\ M\^64 - 53538384\ M\^66 + 34899203\ M\^68 + 
                    19466651\ M\^70 - 23482347\ M\^72 - 1399861\ M\^74 + 
                    10151707\ M\^76 - 2786252\ M\^78 - 2546755\ M\^80 + 
                    1853879\ M\^82 - 28291\ M\^84 - 506783\ M\^86 + 
                    277380\ M\^88 - 12136\ M\^90 - 95999\ M\^92 + 
                    52421\ M\^94 + 60921\ M\^96 - 131662\ M\^98 + 
                    121789\ M\^100 - 72288\ M\^102 + 30005\ M\^104 - 
                    8907\ M\^106 + 1877\ M\^108 - 270\ M\^110 + 24\ M\^112 - 
                    M\^114)\) + 
              L\^11\ \((6 - 136\ M\^2 + 1374\ M\^4 - 8134\ M\^6 + 
                    31191\ M\^8 - 80802\ M\^10 + 143438\ M\^12 - 
                    176265\ M\^14 + 159945\ M\^16 - 124136\ M\^18 + 
                    29703\ M\^20 + 269303\ M\^22 - 621101\ M\^24 + 
                    222879\ M\^26 + 1291888\ M\^28 - 1771108\ M\^30 - 
                    2074282\ M\^32 + 7021803\ M\^34 - 785832\ M\^36 - 
                    16752195\ M\^38 + 13537877\ M\^40 + 29125563\ M\^42 - 
                    45740566\ M\^44 - 34123281\ M\^46 + 97159910\ M\^48 + 
                    21649526\ M\^50 - 154587197\ M\^52 + 11372697\ M\^54 + 
                    194001919\ M\^56 - 54506856\ M\^58 - 192755590\ M\^60 + 
                    86970160\ M\^62 + 152557058\ M\^64 - 96518899\ M\^66 - 
                    93663738\ M\^68 + 81285399\ M\^70 + 42220630\ M\^72 - 
                    53580023\ M\^74 - 11365576\ M\^76 + 27873161\ M\^78 - 
                    724891\ M\^80 - 11091816\ M\^82 + 2474881\ M\^84 + 
                    3345915\ M\^86 - 1596294\ M\^88 - 350460\ M\^90 + 
                    191529\ M\^92 + 169327\ M\^94 + 938\ M\^96 - 
                    202353\ M\^98 + 193365\ M\^100 - 100673\ M\^102 + 
                    35714\ M\^104 - 9381\ M\^106 + 1861\ M\^108 - 
                    266\ M\^110 + 24\ M\^112 - M\^114)\) + 
              L\^12\ \((\(-4\) + 93\ M\^2 - 962\ M\^4 + 5877\ M\^6 - 
                    23746\ M\^8 + 67325\ M\^10 - 138361\ M\^12 + 
                    207720\ M\^14 - 215250\ M\^16 + 93106\ M\^18 + 
                    185315\ M\^20 - 453141\ M\^22 + 215348\ M\^24 + 
                    713732\ M\^26 - 771772\ M\^28 - 2162914\ M\^30 + 
                    4392703\ M\^32 + 3146270\ M\^34 - 13673841\ M\^36 - 
                    1032022\ M\^38 + 32826102\ M\^40 - 10749104\ M\^42 - 
                    61712381\ M\^44 + 35654020\ M\^46 + 95043842\ M\^48 - 
                    71765608\ M\^50 - 120877668\ M\^52 + 107805814\ M\^54 + 
                    126179741\ M\^56 - 126179741\ M\^58 - 107805814\ M\^60 + 
                    120877668\ M\^62 + 71765608\ M\^64 - 95043842\ M\^66 - 
                    35654020\ M\^68 + 61712381\ M\^70 + 10749104\ M\^72 - 
                    32826102\ M\^74 + 1032022\ M\^76 + 13673841\ M\^78 - 
                    3146270\ M\^80 - 4392703\ M\^82 + 2162914\ M\^84 + 
                    771772\ M\^86 - 713732\ M\^88 - 215348\ M\^90 + 
                    453141\ M\^92 - 185315\ M\^94 - 93106\ M\^96 + 
                    215250\ M\^98 - 207720\ M\^100 + 138361\ M\^102 - 
                    67325\ M\^104 + 23746\ M\^106 - 5877\ M\^108 + 
                    962\ M\^110 - 93\ M\^112 + 4\ M\^114)\) + 
              L\^14\ \((\(-6\)\ M\^4 + 140\ M\^6 - 1439\ M\^8 + 8685\ M\^10 - 
                    34695\ M\^12 + 97546\ M\^14 - 193056\ M\^16 + 
                    225411\ M\^18 + 37442\ M\^20 - 638345\ M\^22 + 
                    639797\ M\^24 + 1177235\ M\^26 - 2334449\ M\^28 - 
                    2949136\ M\^30 + 9581845\ M\^32 + 2137799\ M\^34 - 
                    25840390\ M\^36 + 8714462\ M\^38 + 51235801\ M\^40 - 
                    39355369\ M\^42 - 77042907\ M\^44 + 92494155\ M\^46 + 
                    86226707\ M\^48 - 153784600\ M\^50 - 66593408\ M\^52 + 
                    195729580\ M\^54 + 21576278\ M\^56 - 193093798\ M\^58 + 
                    26702848\ M\^60 + 148115730\ M\^62 - 56168900\ M\^64 - 
                    85081007\ M\^66 + 57347514\ M\^68 + 32293953\ M\^70 - 
                    39605891\ M\^72 - 3880824\ M\^74 + 19710237\ M\^76 - 
                    5056288\ M\^78 - 6039514\ M\^80 + 4054828\ M\^82 + 
                    597374\ M\^84 - 1616573\ M\^86 + 508388\ M\^88 + 
                    224976\ M\^90 - 194789\ M\^92 + 10472\ M\^94 - 
                    9086\ M\^96 + 111849\ M\^98 - 160136\ M\^100 + 
                    124613\ M\^102 - 63926\ M\^104 + 22765\ M\^106 - 
                    5615\ M\^108 + 919\ M\^110 - 90\ M\^112 + 
                    4\ M\^114)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 136]\), \(\(-M\^8\) + 
              L\^9\ M\^32 + 
              L\ \((1 - 2\ M\^2 + 5\ M\^4 - 14\ M\^6 + 15\ M\^8 - 6\ M\^10 - 
                    11\ M\^12 + 7\ M\^14 - 2\ M\^16)\) + 
              L\^2\ \((\(-1\) + 2\ M\^2 - 13\ M\^4 + 49\ M\^6 - 58\ M\^8 + 
                    2\ M\^10 + 28\ M\^12 - 41\ M\^14 + 14\ M\^16 + 
                    12\ M\^18 - 22\ M\^20 + 9\ M\^22 - M\^24)\) + 
              L\^3\ \((M\^2 + 10\ M\^4 - 60\ M\^6 + 82\ M\^8 + 3\ M\^10 - 
                    57\ M\^12 + 3\ M\^14 - 17\ M\^16 - 16\ M\^18 + 
                    59\ M\^20 - 37\ M\^22 - 5\ M\^24 + 7\ M\^26 - M\^28)\) + 
              L\^4\ \((M\^2 - 11\ M\^4 + 37\ M\^6 - 38\ M\^8 - 11\ M\^10 + 
                    13\ M\^12 + 13\ M\^14 + 22\ M\^16 + 6\ M\^18 - 
                    114\ M\^20 + 56\ M\^22 + 26\ M\^24 - M\^26 - 20\ M\^28 + 
                    8\ M\^30 - M\^32)\) + 
              L\^6\ \((M\^12 - 7\ M\^14 + 5\ M\^16 + 37\ M\^18 - 59\ M\^20 + 
                    16\ M\^22 + 17\ M\^24 - 3\ M\^26 + 57\ M\^28 - 3\ M\^30 - 
                    82\ M\^32 + 60\ M\^34 - 10\ M\^36 - M\^38)\) + 
              L\^5\ \((M\^8 - 8\ M\^10 + 20\ M\^12 + M\^14 - 26\ M\^16 - 
                    56\ M\^18 + 114\ M\^20 - 6\ M\^22 - 22\ M\^24 - 
                    13\ M\^26 - 13\ M\^28 + 11\ M\^30 + 38\ M\^32 - 
                    37\ M\^34 + 11\ M\^36 - M\^38)\) + 
              L\^8\ \((2\ M\^24 - 7\ M\^26 + 11\ M\^28 + 6\ M\^30 - 
                    15\ M\^32 + 14\ M\^34 - 5\ M\^36 + 2\ M\^38 - M\^40)\) + 
              L\^7\ \((M\^16 - 9\ M\^18 + 22\ M\^20 - 12\ M\^22 - 14\ M\^24 + 
                    41\ M\^26 - 28\ M\^28 - 2\ M\^30 + 58\ M\^32 - 
                    49\ M\^34 + 13\ M\^36 - 2\ M\^38 + M\^40)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 137]\), \(L\^18\ M\^32 + 
              M\^40 + L\^17\ \((\(-M\^28\) - 5\ M\^30 + 32\ M\^32 - 
                    3\ M\^34 - 26\ M\^36 + 24\ M\^38 - 11\ M\^40 + 6\ M\^42 - 
                    2\ M\^44)\) + 
              L\ \((\(-2\)\ M\^28 + 6\ M\^30 - 11\ M\^32 + 24\ M\^34 - 
                    26\ M\^36 - 3\ M\^38 + 32\ M\^40 - 5\ M\^42 - M\^44)\) + 
              L\^2\ \((M\^16 - 6\ M\^18 + 21\ M\^20 - 52\ M\^22 + 69\ M\^24 - 
                    40\ M\^26 + 36\ M\^28 - 116\ M\^30 + 139\ M\^32 + 
                    84\ M\^34 - 217\ M\^36 - 23\ M\^38 + 155\ M\^40 + 
                    88\ M\^42 - 33\ M\^44 - 25\ M\^46 + 8\ M\^48)\) + 
              L\^3\ \((\(-M\^8\) + 7\ M\^10 - 27\ M\^12 + 85\ M\^14 - 
                    189\ M\^16 + 263\ M\^18 - 236\ M\^20 + 160\ M\^22 - 
                    207\ M\^24 + 278\ M\^26 + 14\ M\^28 - 313\ M\^30 + 
                    246\ M\^32 + 246\ M\^34 - 142\ M\^36 - 666\ M\^38 - 
                    84\ M\^40 + 1184\ M\^42 + 135\ M\^44 - 508\ M\^46 + 
                    25\ M\^48 + 90\ M\^50 - 26\ M\^52 + 2\ M\^54)\) + 
              L\^16\ \((8\ M\^24 - 25\ M\^26 - 33\ M\^28 + 88\ M\^30 + 
                    155\ M\^32 - 23\ M\^34 - 217\ M\^36 + 84\ M\^38 + 
                    139\ M\^40 - 116\ M\^42 + 36\ M\^44 - 40\ M\^46 + 
                    69\ M\^48 - 52\ M\^50 + 21\ M\^52 - 6\ M\^54 + M\^56)\) + 
              L\^4\ \((2\ M\^4 - 21\ M\^6 + 99\ M\^8 - 278\ M\^10 + 
                    525\ M\^12 - 769\ M\^14 + 1023\ M\^16 - 947\ M\^18 + 
                    27\ M\^20 + 789\ M\^22 - 919\ M\^24 + 551\ M\^26 + 
                    1128\ M\^28 - 1308\ M\^30 - 407\ M\^32 + 1781\ M\^34 + 
                    442\ M\^36 - 4308\ M\^38 - 527\ M\^40 + 4371\ M\^42 + 
                    1682\ M\^44 - 2034\ M\^46 - 588\ M\^48 + 498\ M\^50 + 
                    79\ M\^52 - 83\ M\^54 + 12\ M\^56)\) + 
              L\^5\ \((\(-1\) + 14\ M\^2 - 87\ M\^4 + 309\ M\^6 - 721\ M\^8 + 
                    1367\ M\^10 - 2463\ M\^12 + 3249\ M\^14 - 1430\ M\^16 - 
                    2090\ M\^18 + 3839\ M\^20 - 4891\ M\^22 + 2184\ M\^24 + 
                    4571\ M\^26 - 5581\ M\^28 + 6509\ M\^30 - 2788\ M\^32 - 
                    3854\ M\^34 + 5492\ M\^36 - 8152\ M\^38 - 5577\ M\^40 + 
                    10462\ M\^42 + 5034\ M\^44 - 667\ M\^46 - 4204\ M\^48 - 
                    1200\ M\^50 + 2490\ M\^52 - 66\ M\^54 - 751\ M\^56 + 
                    354\ M\^58 - 68\ M\^60 + 5\ M\^62)\) + 
              L\^15\ \((2\ M\^18 - 26\ M\^20 + 90\ M\^22 + 25\ M\^24 - 
                    508\ M\^26 + 135\ M\^28 + 1184\ M\^30 - 84\ M\^32 - 
                    666\ M\^34 - 142\ M\^36 + 246\ M\^38 + 246\ M\^40 - 
                    313\ M\^42 + 14\ M\^44 + 278\ M\^46 - 207\ M\^48 + 
                    160\ M\^50 - 236\ M\^52 + 263\ M\^54 - 189\ M\^56 + 
                    85\ M\^58 - 27\ M\^60 + 7\ M\^62 - M\^64)\) + 
              L\^7\ \((\(-3\) + 46\ M\^2 - 370\ M\^4 + 1900\ M\^6 - 
                    6166\ M\^8 + 11353\ M\^10 - 6862\ M\^12 - 13300\ M\^14 + 
                    22336\ M\^16 + 3959\ M\^18 - 13878\ M\^20 - 
                    21348\ M\^22 + 10547\ M\^24 + 39046\ M\^26 - 
                    42395\ M\^28 + 17485\ M\^30 + 69964\ M\^32 - 
                    51065\ M\^34 - 87554\ M\^36 + 24770\ M\^38 + 
                    47864\ M\^40 - 16225\ M\^42 + 13238\ M\^44 + 
                    22972\ M\^46 - 37291\ M\^48 + 560\ M\^50 + 18580\ M\^52 - 
                    9203\ M\^54 - 1167\ M\^56 + 3668\ M\^58 - 2427\ M\^60 + 
                    968\ M\^62 - 242\ M\^64 + 34\ M\^66 - 2\ M\^68)\) + 
              L\^6\ \((3 - 44\ M\^2 + 301\ M\^4 - 1219\ M\^6 + 3044\ M\^8 - 
                    4486\ M\^10 + 3579\ M\^12 - 1965\ M\^14 + 668\ M\^16 + 
                    7424\ M\^18 - 16231\ M\^20 + 7047\ M\^22 + 5825\ M\^24 - 
                    18855\ M\^26 + 32250\ M\^28 + 3658\ M\^30 - 
                    30582\ M\^32 + 8929\ M\^34 + 9962\ M\^36 - 22582\ M\^38 - 
                    13189\ M\^40 + 21991\ M\^42 + 22498\ M\^44 - 
                    10720\ M\^46 - 10562\ M\^48 + 1441\ M\^50 + 3635\ M\^52 + 
                    420\ M\^54 - 1422\ M\^56 - 145\ M\^58 + 740\ M\^60 - 
                    423\ M\^62 + 118\ M\^64 - 17\ M\^66 + M\^68)\) + 
              L\^14\ \((12\ M\^16 - 83\ M\^18 + 79\ M\^20 + 498\ M\^22 - 
                    588\ M\^24 - 2034\ M\^26 + 1682\ M\^28 + 4371\ M\^30 - 
                    527\ M\^32 - 4308\ M\^34 + 442\ M\^36 + 1781\ M\^38 - 
                    407\ M\^40 - 1308\ M\^42 + 1128\ M\^44 + 551\ M\^46 - 
                    919\ M\^48 + 789\ M\^50 + 27\ M\^52 - 947\ M\^54 + 
                    1023\ M\^56 - 769\ M\^58 + 525\ M\^60 - 278\ M\^62 + 
                    99\ M\^64 - 21\ M\^66 + 2\ M\^68)\) + 
              L\^8\ \((1 - 18\ M\^2 + 217\ M\^4 - 1584\ M\^6 + 6659\ M\^8 - 
                    15349\ M\^10 + 13453\ M\^12 + 19313\ M\^14 - 
                    55535\ M\^16 + 14560\ M\^18 + 79467\ M\^20 - 
                    42988\ M\^22 - 95389\ M\^24 + 26886\ M\^26 + 
                    114261\ M\^28 - 14960\ M\^30 - 16014\ M\^32 + 
                    23565\ M\^34 - 94629\ M\^36 - 79590\ M\^38 + 
                    67934\ M\^40 + 102268\ M\^42 - 15926\ M\^44 - 
                    61495\ M\^46 + 21628\ M\^48 + 1365\ M\^50 - 3880\ M\^52 + 
                    11004\ M\^54 - 9863\ M\^56 + 234\ M\^58 + 5287\ M\^60 - 
                    4216\ M\^62 + 1658\ M\^64 - 367\ M\^66 + 43\ M\^68 - 
                    2\ M\^70)\) + 
              L\^9\ \((4\ M\^2 - 105\ M\^4 + 965\ M\^6 - 4448\ M\^8 + 
                    10988\ M\^10 - 11385\ M\^12 - 10308\ M\^14 + 
                    42045\ M\^16 - 27509\ M\^18 - 49279\ M\^20 + 
                    79905\ M\^22 + 34090\ M\^24 - 125208\ M\^26 - 
                    63298\ M\^28 + 188585\ M\^30 + 120876\ M\^32 - 
                    107044\ M\^34 - 160608\ M\^36 - 107044\ M\^38 + 
                    120876\ M\^40 + 188585\ M\^42 - 63298\ M\^44 - 
                    125208\ M\^46 + 34090\ M\^48 + 79905\ M\^50 - 
                    49279\ M\^52 - 27509\ M\^54 + 42045\ M\^56 - 
                    10308\ M\^58 - 11385\ M\^60 + 10988\ M\^62 - 
                    4448\ M\^64 + 965\ M\^66 - 105\ M\^68 + 4\ M\^70)\) + 
              L\^11\ \((\(-2\)\ M\^4 + 34\ M\^6 - 242\ M\^8 + 968\ M\^10 - 
                    2427\ M\^12 + 3668\ M\^14 - 1167\ M\^16 - 9203\ M\^18 + 
                    18580\ M\^20 + 560\ M\^22 - 37291\ M\^24 + 22972\ M\^26 + 
                    13238\ M\^28 - 16225\ M\^30 + 47864\ M\^32 + 
                    24770\ M\^34 - 87554\ M\^36 - 51065\ M\^38 + 
                    69964\ M\^40 + 17485\ M\^42 - 42395\ M\^44 + 
                    39046\ M\^46 + 10547\ M\^48 - 21348\ M\^50 - 
                    13878\ M\^52 + 3959\ M\^54 + 22336\ M\^56 - 
                    13300\ M\^58 - 6862\ M\^60 + 11353\ M\^62 - 6166\ M\^64 + 
                    1900\ M\^66 - 370\ M\^68 + 46\ M\^70 - 3\ M\^72)\) + 
              L\^13\ \((5\ M\^10 - 68\ M\^12 + 354\ M\^14 - 751\ M\^16 - 
                    66\ M\^18 + 2490\ M\^20 - 1200\ M\^22 - 4204\ M\^24 - 
                    667\ M\^26 + 5034\ M\^28 + 10462\ M\^30 - 5577\ M\^32 - 
                    8152\ M\^34 + 5492\ M\^36 - 3854\ M\^38 - 2788\ M\^40 + 
                    6509\ M\^42 - 5581\ M\^44 + 4571\ M\^46 + 2184\ M\^48 - 
                    4891\ M\^50 + 3839\ M\^52 - 2090\ M\^54 - 1430\ M\^56 + 
                    3249\ M\^58 - 2463\ M\^60 + 1367\ M\^62 - 721\ M\^64 + 
                    309\ M\^66 - 87\ M\^68 + 14\ M\^70 - M\^72)\) + 
              L\^10\ \((\(-2\)\ M\^2 + 43\ M\^4 - 367\ M\^6 + 1658\ M\^8 - 
                    4216\ M\^10 + 5287\ M\^12 + 234\ M\^14 - 9863\ M\^16 + 
                    11004\ M\^18 - 3880\ M\^20 + 1365\ M\^22 + 21628\ M\^24 - 
                    61495\ M\^26 - 15926\ M\^28 + 102268\ M\^30 + 
                    67934\ M\^32 - 79590\ M\^34 - 94629\ M\^36 + 
                    23565\ M\^38 - 16014\ M\^40 - 14960\ M\^42 + 
                    114261\ M\^44 + 26886\ M\^46 - 95389\ M\^48 - 
                    42988\ M\^50 + 79467\ M\^52 + 14560\ M\^54 - 
                    55535\ M\^56 + 19313\ M\^58 + 13453\ M\^60 - 
                    15349\ M\^62 + 6659\ M\^64 - 1584\ M\^66 + 217\ M\^68 - 
                    18\ M\^70 + M\^72)\) + 
              L\^12\ \((M\^4 - 17\ M\^6 + 118\ M\^8 - 423\ M\^10 + 
                    740\ M\^12 - 145\ M\^14 - 1422\ M\^16 + 420\ M\^18 + 
                    3635\ M\^20 + 1441\ M\^22 - 10562\ M\^24 - 10720\ M\^26 + 
                    22498\ M\^28 + 21991\ M\^30 - 13189\ M\^32 - 
                    22582\ M\^34 + 9962\ M\^36 + 8929\ M\^38 - 30582\ M\^40 + 
                    3658\ M\^42 + 32250\ M\^44 - 18855\ M\^46 + 5825\ M\^48 + 
                    7047\ M\^50 - 16231\ M\^52 + 7424\ M\^54 + 668\ M\^56 - 
                    1965\ M\^58 + 3579\ M\^60 - 4486\ M\^62 + 3044\ M\^64 - 
                    1219\ M\^66 + 301\ M\^68 - 44\ M\^70 + 3\ M\^72)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 138]\), \(\(-M\^14\) + 
              L\^15\ M\^86 + 
              L\ \((M\^6 - 6\ M\^8 + 22\ M\^10 - 41\ M\^12 + 28\ M\^14 + 
                    24\ M\^16 - 49\ M\^18 + 2\ M\^20 + 6\ M\^22)\) + 
              L\^2\ \((\(-M\^2\) + 9\ M\^4 - 45\ M\^6 + 136\ M\^8 - 
                    233\ M\^10 + 156\ M\^12 + 174\ M\^14 - 366\ M\^16 + 
                    449\ M\^20 - 137\ M\^22 - 420\ M\^24 + 133\ M\^26 + 
                    114\ M\^28 - 49\ M\^30 + 3\ M\^32)\) + 
              L\^3\ \((\(-1\) + 11\ M\^2 - 57\ M\^4 + 170\ M\^6 - 303\ M\^8 + 
                    308\ M\^10 - 197\ M\^12 + 377\ M\^14 - 1054\ M\^16 + 
                    1085\ M\^18 + 1191\ M\^20 - 3982\ M\^22 + 2246\ M\^24 + 
                    3445\ M\^26 - 4261\ M\^28 - 930\ M\^30 + 2168\ M\^32 - 
                    223\ M\^34 - 384\ M\^36 + 128\ M\^38 - 10\ M\^40)\) + 
              L\^4\ \((4\ M\^4 - 51\ M\^6 + 294\ M\^8 - 952\ M\^10 + 
                    1747\ M\^12 - 1417\ M\^14 - 699\ M\^16 + 2191\ M\^18 - 
                    16\ M\^20 - 4018\ M\^22 + 4366\ M\^24 + 2565\ M\^26 - 
                    13007\ M\^28 + 11845\ M\^30 + 9135\ M\^32 - 
                    21269\ M\^34 + 3335\ M\^36 + 10714\ M\^38 - 5568\ M\^40 - 
                    934\ M\^42 + 1531\ M\^44 - 498\ M\^46 + 68\ M\^48 - 
                    3\ M\^50)\) + 
              L\^5\ \((\(-6\)\ M\^8 + 84\ M\^10 - 533\ M\^12 + 1922\ M\^14 - 
                    4073\ M\^16 + 4356\ M\^18 + 493\ M\^20 - 7418\ M\^22 + 
                    6529\ M\^24 + 3597\ M\^26 - 9033\ M\^28 + 3657\ M\^30 + 
                    2699\ M\^32 - 13047\ M\^34 + 24879\ M\^36 + 1266\ M\^38 - 
                    43569\ M\^40 + 24881\ M\^42 + 18722\ M\^44 - 
                    22599\ M\^46 + 3866\ M\^48 + 4910\ M\^50 - 3459\ M\^52 + 
                    1007\ M\^54 - 138\ M\^56 + 6\ M\^58)\) + 
              L\^6\ \((4\ M\^12 - 59\ M\^14 + 404\ M\^16 - 1633\ M\^18 + 
                    4228\ M\^20 - 7007\ M\^22 + 5922\ M\^24 + 3443\ M\^26 - 
                    17113\ M\^28 + 17208\ M\^30 + 6342\ M\^32 - 
                    25617\ M\^34 + 10122\ M\^36 + 12208\ M\^38 - 
                    14018\ M\^40 + 21755\ M\^42 - 8580\ M\^44 - 
                    44080\ M\^46 + 44975\ M\^48 + 10353\ M\^50 - 
                    35965\ M\^52 + 17285\ M\^54 + 3491\ M\^56 - 7948\ M\^58 + 
                    4397\ M\^60 - 1324\ M\^62 + 225\ M\^64 - 20\ M\^66 + 
                    M\^68)\) + 
              L\^7\ \((\(-M\^16\) + 15\ M\^18 - 111\ M\^20 + 530\ M\^22 - 
                    1861\ M\^24 + 4946\ M\^26 - 9053\ M\^28 + 8060\ M\^30 + 
                    6535\ M\^32 - 27464\ M\^34 + 25303\ M\^36 + 
                    10725\ M\^38 - 37506\ M\^40 + 16798\ M\^42 + 
                    11530\ M\^44 - 11689\ M\^46 + 22476\ M\^48 - 
                    18137\ M\^50 - 36816\ M\^52 + 52384\ M\^54 - 16\ M\^56 - 
                    37506\ M\^58 + 25847\ M\^60 - 1445\ M\^62 - 8760\ M\^64 + 
                    7488\ M\^66 - 3682\ M\^68 + 1206\ M\^70 - 255\ M\^72 + 
                    32\ M\^74 - 2\ M\^76)\) + 
              L\^8\ \((2\ M\^24 - 32\ M\^26 + 255\ M\^28 - 1206\ M\^30 + 
                    3682\ M\^32 - 7488\ M\^34 + 8760\ M\^36 + 1445\ M\^38 - 
                    25847\ M\^40 + 37506\ M\^42 + 16\ M\^44 - 52384\ M\^46 + 
                    36816\ M\^48 + 18137\ M\^50 - 22476\ M\^52 + 
                    11689\ M\^54 - 11530\ M\^56 - 16798\ M\^58 + 
                    37506\ M\^60 - 10725\ M\^62 - 25303\ M\^64 + 
                    27464\ M\^66 - 6535\ M\^68 - 8060\ M\^70 + 9053\ M\^72 - 
                    4946\ M\^74 + 1861\ M\^76 - 530\ M\^78 + 111\ M\^80 - 
                    15\ M\^82 + M\^84)\) + 
              L\^9\ \((\(-M\^32\) + 20\ M\^34 - 225\ M\^36 + 1324\ M\^38 - 
                    4397\ M\^40 + 7948\ M\^42 - 3491\ M\^44 - 17285\ M\^46 + 
                    35965\ M\^48 - 10353\ M\^50 - 44975\ M\^52 + 
                    44080\ M\^54 + 8580\ M\^56 - 21755\ M\^58 + 
                    14018\ M\^60 - 12208\ M\^62 - 10122\ M\^64 + 
                    25617\ M\^66 - 6342\ M\^68 - 17208\ M\^70 + 
                    17113\ M\^72 - 3443\ M\^74 - 5922\ M\^76 + 7007\ M\^78 - 
                    4228\ M\^80 + 1633\ M\^82 - 404\ M\^84 + 59\ M\^86 - 
                    4\ M\^88)\) + 
              L\^10\ \((\(-6\)\ M\^42 + 138\ M\^44 - 1007\ M\^46 + 
                    3459\ M\^48 - 4910\ M\^50 - 3866\ M\^52 + 22599\ M\^54 - 
                    18722\ M\^56 - 24881\ M\^58 + 43569\ M\^60 - 
                    1266\ M\^62 - 24879\ M\^64 + 13047\ M\^66 - 2699\ M\^68 - 
                    3657\ M\^70 + 9033\ M\^72 - 3597\ M\^74 - 6529\ M\^76 + 
                    7418\ M\^78 - 493\ M\^80 - 4356\ M\^82 + 4073\ M\^84 - 
                    1922\ M\^86 + 533\ M\^88 - 84\ M\^90 + 6\ M\^92)\) + 
              L\^14\ \((\(-6\)\ M\^78 - 2\ M\^80 + 49\ M\^82 - 24\ M\^84 - 
                    28\ M\^86 + 41\ M\^88 - 22\ M\^90 + 6\ M\^92 - M\^94)\) + 
              L\^11\ \((3\ M\^50 - 68\ M\^52 + 498\ M\^54 - 1531\ M\^56 + 
                    934\ M\^58 + 5568\ M\^60 - 10714\ M\^62 - 3335\ M\^64 + 
                    21269\ M\^66 - 9135\ M\^68 - 11845\ M\^70 + 
                    13007\ M\^72 - 2565\ M\^74 - 4366\ M\^76 + 4018\ M\^78 + 
                    16\ M\^80 - 2191\ M\^82 + 699\ M\^84 + 1417\ M\^86 - 
                    1747\ M\^88 + 952\ M\^90 - 294\ M\^92 + 51\ M\^94 - 
                    4\ M\^96)\) + 
              L\^13\ \((\(-3\)\ M\^68 + 49\ M\^70 - 114\ M\^72 - 133\ M\^74 + 
                    420\ M\^76 + 137\ M\^78 - 449\ M\^80 + 366\ M\^84 - 
                    174\ M\^86 - 156\ M\^88 + 233\ M\^90 - 136\ M\^92 + 
                    45\ M\^94 - 9\ M\^96 + M\^98)\) + 
              L\^12\ \((10\ M\^60 - 128\ M\^62 + 384\ M\^64 + 223\ M\^66 - 
                    2168\ M\^68 + 930\ M\^70 + 4261\ M\^72 - 3445\ M\^74 - 
                    2246\ M\^76 + 3982\ M\^78 - 1191\ M\^80 - 1085\ M\^82 + 
                    1054\ M\^84 - 377\ M\^86 + 197\ M\^88 - 308\ M\^90 + 
                    303\ M\^92 - 170\ M\^94 + 57\ M\^96 - 11\ M\^98 + 
                    M\^100)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 139]\), \(\(-1\) + 
              L\^5\ M\^76 + 
              L\ \((M\^12 - 7\ M\^14 + 3\ M\^16 - M\^18 + M\^20)\) + 
              L\^2\ \((\(-6\)\ M\^28 - M\^32 + 7\ M\^34 - 3\ M\^36 + 
                    M\^38)\) + 
              L\^3\ \((\(-M\^38\) + 3\ M\^40 - 7\ M\^42 + M\^44 + 
                    6\ M\^48)\) + 
              L\^4\ \((\(-M\^56\) + M\^58 - 3\ M\^60 + 7\ M\^62 - 
                    M\^64)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 140]\), \(L\^8\ M\^8 - M\^14 + 
              L\ \((\(-1\) + M\^2 - 2\ M\^4 + 5\ M\^6 - 6\ M\^8 + 4\ M\^10 - 
                    2\ M\^12 - 5\ M\^14)\) + 
              L\^2\ \((2 - 10\ M\^2 + 10\ M\^4 - 9\ M\^6 + 11\ M\^8 - 
                    11\ M\^10 + 3\ M\^12 - 7\ M\^14 - 3\ M\^16)\) + 
              L\^4\ \((10\ M\^4 - 18\ M\^6 + 11\ M\^8 - 28\ M\^10 + 
                    28\ M\^12 - 11\ M\^14 + 18\ M\^16 - 10\ M\^18)\) + 
              L\^3\ \((\(-1\) + 9\ M\^2 - 20\ M\^4 + 10\ M\^6 - 12\ M\^8 + 
                    15\ M\^10 - 21\ M\^12 + 17\ M\^14 - 9\ M\^16 - 
                    2\ M\^18)\) + 
              L\^6\ \((3\ M\^6 + 7\ M\^8 - 3\ M\^10 + 11\ M\^12 - 11\ M\^14 + 
                    9\ M\^16 - 10\ M\^18 + 10\ M\^20 - 2\ M\^22)\) + 
              L\^5\ \((2\ M\^4 + 9\ M\^6 - 17\ M\^8 + 21\ M\^10 - 15\ M\^12 + 
                    12\ M\^14 - 10\ M\^16 + 20\ M\^18 - 9\ M\^20 + M\^22)\) + 
              L\^7\ \((5\ M\^8 + 2\ M\^10 - 4\ M\^12 + 6\ M\^14 - 5\ M\^16 + 
                    2\ M\^18 - M\^20 + M\^22)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 141]\), \(\(-M\^20\) + 
              L\^9\ M\^20 + 
              L\ \((M\^8 - 4\ M\^10 + 8\ M\^12 - 12\ M\^14 + 17\ M\^16 - 
                    2\ M\^18 - 23\ M\^20 + 6\ M\^22 + 2\ M\^24)\) + 
              L\^2\ \((\(-M\^4\) + 6\ M\^6 - 25\ M\^8 + 66\ M\^10 - 
                    90\ M\^12 - 6\ M\^14 + 166\ M\^16 - 65\ M\^18 - 
                    126\ M\^20 + 28\ M\^22 + 31\ M\^24 - M\^26 - 
                    3\ M\^28)\) + 
              L\^8\ \((\(-2\)\ M\^16 - 6\ M\^18 + 23\ M\^20 + 2\ M\^22 - 
                    17\ M\^24 + 12\ M\^26 - 8\ M\^28 + 4\ M\^30 - M\^32)\) + 
              L\^3\ \((\(-2\)\ M\^2 + 18\ M\^4 - 67\ M\^6 + 130\ M\^8 - 
                    76\ M\^10 - 162\ M\^12 + 168\ M\^14 + 123\ M\^16 + 
                    39\ M\^18 - 130\ M\^20 - 249\ M\^22 + 127\ M\^24 + 
                    99\ M\^26 - 38\ M\^28 - 12\ M\^30 + 4\ M\^32)\) + 
              L\^7\ \((3\ M\^12 + M\^14 - 31\ M\^16 - 28\ M\^18 + 
                    126\ M\^20 + 65\ M\^22 - 166\ M\^24 + 6\ M\^26 + 
                    90\ M\^28 - 66\ M\^30 + 25\ M\^32 - 6\ M\^34 + M\^36)\) + 
              L\^4\ \((\(-1\) + 13\ M\^2 - 56\ M\^4 + 85\ M\^6 + 6\ M\^8 - 
                    51\ M\^10 - 75\ M\^12 - 174\ M\^14 + 353\ M\^16 + 
                    338\ M\^18 - 340\ M\^20 - 147\ M\^22 - 178\ M\^24 + 
                    100\ M\^26 + 250\ M\^28 - 107\ M\^30 - 76\ M\^32 + 
                    58\ M\^34 - 13\ M\^36 + M\^38)\) + 
              L\^6\ \((\(-4\)\ M\^8 + 12\ M\^10 + 38\ M\^12 - 99\ M\^14 - 
                    127\ M\^16 + 249\ M\^18 + 130\ M\^20 - 39\ M\^22 - 
                    123\ M\^24 - 168\ M\^26 + 162\ M\^28 + 76\ M\^30 - 
                    130\ M\^32 + 67\ M\^34 - 18\ M\^36 + 2\ M\^38)\) + 
              L\^5\ \((\(-M\^2\) + 13\ M\^4 - 58\ M\^6 + 76\ M\^8 + 
                    107\ M\^10 - 250\ M\^12 - 100\ M\^14 + 178\ M\^16 + 
                    147\ M\^18 + 340\ M\^20 - 338\ M\^22 - 353\ M\^24 + 
                    174\ M\^26 + 75\ M\^28 + 51\ M\^30 - 6\ M\^32 - 
                    85\ M\^34 + 56\ M\^36 - 13\ M\^38 + M\^40)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 142]\), \(\(-1\) + 
              L\^9\ M\^104 + 
              L\ \((1 - 2\ M\^2 + 5\ M\^4 - 11\ M\^6 + 9\ M\^8 - 2\ M\^10 - 
                    9\ M\^12 + 3\ M\^14 - M\^16)\) + 
              L\^2\ \((\(-M\^8\) + 3\ M\^10 - 9\ M\^12 + 24\ M\^14 - 
                    20\ M\^16 + M\^18 - 14\ M\^20 + 7\ M\^22 - 4\ M\^24 - 
                    6\ M\^26 - M\^28 - M\^30 + M\^32)\) + 
              L\^3\ \((4\ M\^20 - 15\ M\^22 + 18\ M\^24 - 20\ M\^26 + 
                    44\ M\^28 - 28\ M\^30 - 27\ M\^32 - 20\ M\^34 + 
                    29\ M\^36 - M\^38 - 12\ M\^40 - 10\ M\^42 + 15\ M\^44 - 
                    6\ M\^46 + M\^48)\) + 
              L\^4\ \((\(-6\)\ M\^32 + 27\ M\^34 - 30\ M\^36 - 4\ M\^38 - 
                    23\ M\^40 + 69\ M\^42 - 2\ M\^44 - 41\ M\^46 - 
                    49\ M\^48 + 35\ M\^50 + 20\ M\^52 + 9\ M\^54 - 
                    36\ M\^56 + 21\ M\^58 - 4\ M\^60)\) + 
              L\^5\ \((4\ M\^44 - 21\ M\^46 + 36\ M\^48 - 9\ M\^50 - 
                    20\ M\^52 - 35\ M\^54 + 49\ M\^56 + 41\ M\^58 + 
                    2\ M\^60 - 69\ M\^62 + 23\ M\^64 + 4\ M\^66 + 30\ M\^68 - 
                    27\ M\^70 + 6\ M\^72)\) + 
              L\^6\ \((\(-M\^56\) + 6\ M\^58 - 15\ M\^60 + 10\ M\^62 + 
                    12\ M\^64 + M\^66 - 29\ M\^68 + 20\ M\^70 + 27\ M\^72 + 
                    28\ M\^74 - 44\ M\^76 + 20\ M\^78 - 18\ M\^80 + 
                    15\ M\^82 - 4\ M\^84)\) + 
              L\^7\ \((\(-M\^72\) + M\^74 + M\^76 + 6\ M\^78 + 4\ M\^80 - 
                    7\ M\^82 + 14\ M\^84 - M\^86 + 20\ M\^88 - 24\ M\^90 + 
                    9\ M\^92 - 3\ M\^94 + M\^96)\) + 
              L\^8\ \((M\^88 - 3\ M\^90 + 9\ M\^92 + 2\ M\^94 - 9\ M\^96 + 
                    11\ M\^98 - 5\ M\^100 + 2\ M\^102 - M\^104)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 143]\), \(L\^14\ M\^8 - M\^70 + 
              L\^13\ \((2\ M\^6 - 13\ M\^10 + 28\ M\^12 + 9\ M\^14 - 
                    26\ M\^16 + 16\ M\^18 - 7\ M\^20 + 3\ M\^22)\) + 
              L\^12\ \((3\ M\^4 - 8\ M\^6 - 2\ M\^8 + 74\ M\^10 - 
                    121\ M\^12 - 51\ M\^14 + 295\ M\^16 - 28\ M\^18 - 
                    178\ M\^20 + 67\ M\^22 - 14\ M\^26 + 79\ M\^28 - 
                    79\ M\^30 + 39\ M\^32 - 14\ M\^34 + 3\ M\^36)\) + 
              L\^11\ \((\(-1\) + 10\ M\^2 - 42\ M\^4 + 96\ M\^6 - 49\ M\^8 - 
                    298\ M\^10 + 633\ M\^12 + 179\ M\^14 - 1558\ M\^16 + 
                    485\ M\^18 + 2218\ M\^20 - 874\ M\^22 - 1717\ M\^24 + 
                    963\ M\^26 + 420\ M\^28 - 365\ M\^30 + 239\ M\^32 - 
                    40\ M\^34 - 236\ M\^36 + 205\ M\^38 - 107\ M\^40 + 
                    85\ M\^42 - 55\ M\^44 + 23\ M\^46 - 7\ M\^48 + M\^50)\) + 
              L\^10\ \((1 - 14\ M\^2 + 92\ M\^4 - 308\ M\^6 + 454\ M\^8 + 
                    288\ M\^10 - 2013\ M\^12 + 1530\ M\^14 + 3756\ M\^16 - 
                    5558\ M\^18 - 3977\ M\^20 + 8985\ M\^22 + 3352\ M\^24 - 
                    8189\ M\^26 - 1774\ M\^28 + 3840\ M\^30 + 894\ M\^32 - 
                    188\ M\^34 - 185\ M\^36 - 1308\ M\^38 + 448\ M\^40 + 
                    608\ M\^42 - 321\ M\^44 - 102\ M\^46 + 323\ M\^48 - 
                    340\ M\^50 + 176\ M\^52 - 49\ M\^54 + 9\ M\^56 - 
                    M\^58)\) + 
              L\^9\ \((4\ M\^2 - 51\ M\^4 + 293\ M\^6 - 848\ M\^8 + 
                    877\ M\^10 + 1549\ M\^12 - 4715\ M\^14 + 341\ M\^16 + 
                    10386\ M\^18 - 5748\ M\^20 - 13846\ M\^22 + 8587\ M\^24 + 
                    15180\ M\^26 - 3585\ M\^28 - 14017\ M\^30 - 5185\ M\^32 + 
                    12529\ M\^34 + 7956\ M\^36 - 8404\ M\^38 - 4431\ M\^40 + 
                    3574\ M\^42 - 277\ M\^44 + 743\ M\^46 + 551\ M\^48 - 
                    1303\ M\^50 + 146\ M\^52 + 613\ M\^54 - 599\ M\^56 + 
                    366\ M\^58 - 139\ M\^60 + 27\ M\^62 - 2\ M\^64)\) + 
              L\^8\ \((6\ M\^4 - 79\ M\^6 + 443\ M\^8 - 1224\ M\^10 + 
                    1142\ M\^12 + 2105\ M\^14 - 5129\ M\^16 - 1176\ M\^18 + 
                    9521\ M\^20 + 518\ M\^22 - 10296\ M\^24 - 9140\ M\^26 + 
                    11942\ M\^28 + 25233\ M\^30 - 14480\ M\^32 - 
                    34568\ M\^34 + 14586\ M\^36 + 26342\ M\^38 - 
                    5234\ M\^40 - 11926\ M\^42 - 5280\ M\^44 + 3474\ M\^46 + 
                    8551\ M\^48 - 3185\ M\^50 - 3490\ M\^52 + 1869\ M\^54 - 
                    107\ M\^56 - 224\ M\^58 + 666\ M\^60 - 696\ M\^62 + 
                    348\ M\^64 - 97\ M\^66 + 15\ M\^68 - M\^70)\) + 
              L\^7\ \((4\ M\^6 - 56\ M\^8 + 327\ M\^10 - 963\ M\^12 + 
                    1185\ M\^14 + 572\ M\^16 - 2462\ M\^18 + 32\ M\^20 + 
                    1194\ M\^22 + 3963\ M\^24 + 3312\ M\^26 - 18189\ M\^28 - 
                    4122\ M\^30 + 30983\ M\^32 + 6116\ M\^34 - 29383\ M\^36 - 
                    15199\ M\^38 + 15199\ M\^40 + 29383\ M\^42 - 
                    6116\ M\^44 - 30983\ M\^46 + 4122\ M\^48 + 18189\ M\^50 - 
                    3312\ M\^52 - 3963\ M\^54 - 1194\ M\^56 - 32\ M\^58 + 
                    2462\ M\^60 - 572\ M\^62 - 1185\ M\^64 + 963\ M\^66 - 
                    327\ M\^68 + 56\ M\^70 - 4\ M\^72)\) + 
              L\ \((\(-3\)\ M\^56 + 7\ M\^58 - 16\ M\^60 + 26\ M\^62 - 
                    9\ M\^64 - 28\ M\^66 + 13\ M\^68 - 2\ M\^72)\) + 
              L\^6\ \((M\^8 - 15\ M\^10 + 97\ M\^12 - 348\ M\^14 + 
                    696\ M\^16 - 666\ M\^18 + 224\ M\^20 + 107\ M\^22 - 
                    1869\ M\^24 + 3490\ M\^26 + 3185\ M\^28 - 8551\ M\^30 - 
                    3474\ M\^32 + 5280\ M\^34 + 11926\ M\^36 + 5234\ M\^38 - 
                    26342\ M\^40 - 14586\ M\^42 + 34568\ M\^44 + 
                    14480\ M\^46 - 25233\ M\^48 - 11942\ M\^50 + 
                    9140\ M\^52 + 10296\ M\^54 - 518\ M\^56 - 9521\ M\^58 + 
                    1176\ M\^60 + 5129\ M\^62 - 2105\ M\^64 - 1142\ M\^66 + 
                    1224\ M\^68 - 443\ M\^70 + 79\ M\^72 - 6\ M\^74)\) + 
              L\^2\ \((\(-3\)\ M\^42 + 14\ M\^44 - 39\ M\^46 + 79\ M\^48 - 
                    79\ M\^50 + 14\ M\^52 - 67\ M\^56 + 178\ M\^58 + 
                    28\ M\^60 - 295\ M\^62 + 51\ M\^64 + 121\ M\^66 - 
                    74\ M\^68 + 2\ M\^70 + 8\ M\^72 - 3\ M\^74)\) + 
              L\^5\ \((2\ M\^14 - 27\ M\^16 + 139\ M\^18 - 366\ M\^20 + 
                    599\ M\^22 - 613\ M\^24 - 146\ M\^26 + 1303\ M\^28 - 
                    551\ M\^30 - 743\ M\^32 + 277\ M\^34 - 3574\ M\^36 + 
                    4431\ M\^38 + 8404\ M\^40 - 7956\ M\^42 - 12529\ M\^44 + 
                    5185\ M\^46 + 14017\ M\^48 + 3585\ M\^50 - 15180\ M\^52 - 
                    8587\ M\^54 + 13846\ M\^56 + 5748\ M\^58 - 10386\ M\^60 - 
                    341\ M\^62 + 4715\ M\^64 - 1549\ M\^66 - 877\ M\^68 + 
                    848\ M\^70 - 293\ M\^72 + 51\ M\^74 - 4\ M\^76)\) + 
              L\^4\ \((M\^20 - 9\ M\^22 + 49\ M\^24 - 176\ M\^26 + 
                    340\ M\^28 - 323\ M\^30 + 102\ M\^32 + 321\ M\^34 - 
                    608\ M\^36 - 448\ M\^38 + 1308\ M\^40 + 185\ M\^42 + 
                    188\ M\^44 - 894\ M\^46 - 3840\ M\^48 + 1774\ M\^50 + 
                    8189\ M\^52 - 3352\ M\^54 - 8985\ M\^56 + 3977\ M\^58 + 
                    5558\ M\^60 - 3756\ M\^62 - 1530\ M\^64 + 2013\ M\^66 - 
                    288\ M\^68 - 454\ M\^70 + 308\ M\^72 - 92\ M\^74 + 
                    14\ M\^76 - M\^78)\) + 
              L\^3\ \((\(-M\^28\) + 7\ M\^30 - 23\ M\^32 + 55\ M\^34 - 
                    85\ M\^36 + 107\ M\^38 - 205\ M\^40 + 236\ M\^42 + 
                    40\ M\^44 - 239\ M\^46 + 365\ M\^48 - 420\ M\^50 - 
                    963\ M\^52 + 1717\ M\^54 + 874\ M\^56 - 2218\ M\^58 - 
                    485\ M\^60 + 1558\ M\^62 - 179\ M\^64 - 633\ M\^66 + 
                    298\ M\^68 + 49\ M\^70 - 96\ M\^72 + 42\ M\^74 - 
                    10\ M\^76 + M\^78)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 144]\), \(L\^17\ M\^14 - 
              M\^82 + L\^16\ \((\(-3\)\ M\^10 + 4\ M\^12 + 9\ M\^14 - 
                    33\ M\^16 + 38\ M\^18 + 31\ M\^20 - 44\ M\^22 + 
                    18\ M\^24 - 5\ M\^26)\) + 
              L\^15\ \((\(-M\^4\) + 9\ M\^6 - 36\ M\^8 + 71\ M\^10 - 
                    27\ M\^12 - 235\ M\^14 + 539\ M\^16 - 265\ M\^18 - 
                    798\ M\^20 + 1231\ M\^22 + 255\ M\^24 - 1124\ M\^26 + 
                    517\ M\^28 + 214\ M\^30 - 445\ M\^32 + 261\ M\^34 - 
                    72\ M\^36 + 10\ M\^38)\) + 
              L\^14\ \((\(-2\)\ M\^2 + 24\ M\^4 - 128\ M\^6 + 388\ M\^8 - 
                    567\ M\^10 - 199\ M\^12 + 2200\ M\^14 - 3122\ M\^16 - 
                    226\ M\^18 + 6115\ M\^20 - 6807\ M\^22 - 2866\ M\^24 + 
                    11229\ M\^26 - 3341\ M\^28 - 6129\ M\^30 + 6217\ M\^32 - 
                    1470\ M\^34 - 3382\ M\^36 + 3479\ M\^38 - 352\ M\^40 - 
                    1485\ M\^42 + 1286\ M\^44 - 520\ M\^46 + 108\ M\^48 - 
                    10\ M\^50)\) + 
              L\^13\ \((\(-1\) + 18\ M\^2 - 144\ M\^4 + 652\ M\^6 - 
                    1696\ M\^8 + 1964\ M\^10 + 1719\ M\^12 - 8803\ M\^14 + 
                    9027\ M\^16 + 5351\ M\^18 - 20233\ M\^20 + 14350\ M\^22 + 
                    9436\ M\^24 - 27060\ M\^26 + 12564\ M\^28 + 
                    15204\ M\^30 - 12631\ M\^32 + 9338\ M\^34 - 3274\ M\^36 - 
                    16306\ M\^38 + 11293\ M\^40 + 6989\ M\^42 - 9307\ M\^44 + 
                    946\ M\^46 + 4235\ M\^48 - 2397\ M\^50 - 1294\ M\^52 + 
                    2331\ M\^54 - 1378\ M\^56 + 434\ M\^58 - 72\ M\^60 + 
                    5\ M\^62)\) + 
              L\^12\ \((3 - 51\ M\^2 + 382\ M\^4 - 1617\ M\^6 + 3934\ M\^8 - 
                    4333\ M\^10 - 3094\ M\^12 + 16373\ M\^14 - 17624\ M\^16 - 
                    5475\ M\^18 + 31805\ M\^20 - 28831\ M\^22 - 3063\ M\^24 + 
                    37519\ M\^26 - 37828\ M\^28 - 17511\ M\^30 + 
                    37800\ M\^32 + 9023\ M\^34 + 3403\ M\^36 - 6070\ M\^38 - 
                    19157\ M\^40 + 986\ M\^42 + 1482\ M\^44 - 3809\ M\^46 + 
                    14518\ M\^48 + 1868\ M\^50 - 13225\ M\^52 + 2881\ M\^54 + 
                    4887\ M\^56 - 3348\ M\^58 + 1115\ M\^60 - 1243\ M\^62 + 
                    1661\ M\^64 - 1221\ M\^66 + 523\ M\^68 - 132\ M\^70 + 
                    18\ M\^72 - M\^74)\) + 
              L\^11\ \((\(-3\) + 57\ M\^2 - 463\ M\^4 + 2100\ M\^6 - 
                    5592\ M\^8 + 7691\ M\^10 - 124\ M\^12 - 17762\ M\^14 + 
                    25266\ M\^16 - 3523\ M\^18 - 27162\ M\^20 + 
                    32517\ M\^22 - 14539\ M\^24 - 12795\ M\^26 + 
                    48175\ M\^28 - 33151\ M\^30 - 54181\ M\^32 + 
                    42266\ M\^34 + 31029\ M\^36 + 691\ M\^38 + 23992\ M\^40 - 
                    2159\ M\^42 - 44111\ M\^44 - 37341\ M\^46 + 
                    11072\ M\^48 + 51271\ M\^50 - 1191\ M\^52 - 
                    24956\ M\^54 + 6671\ M\^56 + 928\ M\^58 - 3880\ M\^60 + 
                    3195\ M\^62 - 10\ M\^64 + 795\ M\^66 - 2811\ M\^68 + 
                    3073\ M\^70 - 2282\ M\^72 + 1279\ M\^74 - 499\ M\^76 + 
                    123\ M\^78 - 17\ M\^80 + M\^82)\) + 
              L\^10\ \((1 - 23\ M\^2 + 223\ M\^4 - 1223\ M\^6 + 4169\ M\^8 - 
                    8766\ M\^10 + 9158\ M\^12 + 4033\ M\^14 - 26120\ M\^16 + 
                    27567\ M\^18 + 9402\ M\^20 - 46956\ M\^22 + 
                    32353\ M\^24 + 23130\ M\^26 - 58045\ M\^28 + 
                    26046\ M\^30 + 72425\ M\^32 - 79175\ M\^34 - 
                    97260\ M\^36 + 61461\ M\^38 + 129241\ M\^40 + 
                    22250\ M\^42 - 62569\ M\^44 - 30394\ M\^46 - 
                    36735\ M\^48 - 29288\ M\^50 + 48672\ M\^52 + 
                    23613\ M\^54 - 3254\ M\^56 - 495\ M\^58 - 7493\ M\^60 - 
                    5975\ M\^62 - 5297\ M\^64 + 7591\ M\^66 + 7774\ M\^68 - 
                    6181\ M\^70 - 2150\ M\^72 - 241\ M\^74 + 4802\ M\^76 - 
                    4274\ M\^78 + 1812\ M\^80 - 429\ M\^82 + 55\ M\^84 - 
                    3\ M\^86)\) + 
              L\ \((5\ M\^70 - 18\ M\^72 + 44\ M\^74 - 31\ M\^76 - 
                    38\ M\^78 + 33\ M\^80 - 9\ M\^82 - 4\ M\^84 + 
                    3\ M\^86)\) + 
              L\^9\ \((M\^2 - 21\ M\^4 + 207\ M\^6 - 1218\ M\^8 + 
                    4499\ M\^10 - 9931\ M\^12 + 9853\ M\^14 + 6590\ M\^16 - 
                    25744\ M\^18 + 9406\ M\^20 + 29127\ M\^22 - 
                    17285\ M\^24 - 23538\ M\^26 - 855\ M\^28 + 11507\ M\^30 + 
                    54629\ M\^32 + 19841\ M\^34 - 117594\ M\^36 - 
                    71150\ M\^38 + 82501\ M\^40 + 109500\ M\^42 + 
                    907\ M\^44 - 14842\ M\^46 + 29532\ M\^48 - 84481\ M\^50 - 
                    101209\ M\^52 + 56898\ M\^54 + 59203\ M\^56 - 
                    4453\ M\^58 - 1080\ M\^60 + 20244\ M\^62 - 30312\ M\^64 - 
                    36133\ M\^66 + 36223\ M\^68 + 24860\ M\^70 - 
                    26149\ M\^72 - 9867\ M\^74 + 14986\ M\^76 + 1599\ M\^78 - 
                    9396\ M\^80 + 6461\ M\^82 - 2321\ M\^84 + 490\ M\^86 - 
                    58\ M\^88 + 3\ M\^90)\) + 
              L\^2\ \((\(-10\)\ M\^58 + 72\ M\^60 - 261\ M\^62 + 445\ M\^64 - 
                    214\ M\^66 - 517\ M\^68 + 1124\ M\^70 - 255\ M\^72 - 
                    1231\ M\^74 + 798\ M\^76 + 265\ M\^78 - 539\ M\^80 + 
                    235\ M\^82 + 27\ M\^84 - 71\ M\^86 + 36\ M\^88 - 
                    9\ M\^90 + M\^92)\) + 
              L\^8\ \((\(-3\)\ M\^6 + 58\ M\^8 - 490\ M\^10 + 2321\ M\^12 - 
                    6461\ M\^14 + 9396\ M\^16 - 1599\ M\^18 - 14986\ M\^20 + 
                    9867\ M\^22 + 26149\ M\^24 - 24860\ M\^26 - 
                    36223\ M\^28 + 36133\ M\^30 + 30312\ M\^32 - 
                    20244\ M\^34 + 1080\ M\^36 + 4453\ M\^38 - 59203\ M\^40 - 
                    56898\ M\^42 + 101209\ M\^44 + 84481\ M\^46 - 
                    29532\ M\^48 + 14842\ M\^50 - 907\ M\^52 - 
                    109500\ M\^54 - 82501\ M\^56 + 71150\ M\^58 + 
                    117594\ M\^60 - 19841\ M\^62 - 54629\ M\^64 - 
                    11507\ M\^66 + 855\ M\^68 + 23538\ M\^70 + 17285\ M\^72 - 
                    29127\ M\^74 - 9406\ M\^76 + 25744\ M\^78 - 6590\ M\^80 - 
                    9853\ M\^82 + 9931\ M\^84 - 4499\ M\^86 + 1218\ M\^88 - 
                    207\ M\^90 + 21\ M\^92 - M\^94)\) + 
              L\^3\ \((10\ M\^46 - 108\ M\^48 + 520\ M\^50 - 1286\ M\^52 + 
                    1485\ M\^54 + 352\ M\^56 - 3479\ M\^58 + 3382\ M\^60 + 
                    1470\ M\^62 - 6217\ M\^64 + 6129\ M\^66 + 3341\ M\^68 - 
                    11229\ M\^70 + 2866\ M\^72 + 6807\ M\^74 - 6115\ M\^76 + 
                    226\ M\^78 + 3122\ M\^80 - 2200\ M\^82 + 199\ M\^84 + 
                    567\ M\^86 - 388\ M\^88 + 128\ M\^90 - 24\ M\^92 + 
                    2\ M\^94)\) + 
              L\^5\ \((M\^22 - 18\ M\^24 + 132\ M\^26 - 523\ M\^28 + 
                    1221\ M\^30 - 1661\ M\^32 + 1243\ M\^34 - 1115\ M\^36 + 
                    3348\ M\^38 - 4887\ M\^40 - 2881\ M\^42 + 13225\ M\^44 - 
                    1868\ M\^46 - 14518\ M\^48 + 3809\ M\^50 - 1482\ M\^52 - 
                    986\ M\^54 + 19157\ M\^56 + 6070\ M\^58 - 3403\ M\^60 - 
                    9023\ M\^62 - 37800\ M\^64 + 17511\ M\^66 + 
                    37828\ M\^68 - 37519\ M\^70 + 3063\ M\^72 + 
                    28831\ M\^74 - 31805\ M\^76 + 5475\ M\^78 + 
                    17624\ M\^80 - 16373\ M\^82 + 3094\ M\^84 + 4333\ M\^86 - 
                    3934\ M\^88 + 1617\ M\^90 - 382\ M\^92 + 51\ M\^94 - 
                    3\ M\^96)\) + 
              L\^7\ \((3\ M\^10 - 55\ M\^12 + 429\ M\^14 - 1812\ M\^16 + 
                    4274\ M\^18 - 4802\ M\^20 + 241\ M\^22 + 2150\ M\^24 + 
                    6181\ M\^26 - 7774\ M\^28 - 7591\ M\^30 + 5297\ M\^32 + 
                    5975\ M\^34 + 7493\ M\^36 + 495\ M\^38 + 3254\ M\^40 - 
                    23613\ M\^42 - 48672\ M\^44 + 29288\ M\^46 + 
                    36735\ M\^48 + 30394\ M\^50 + 62569\ M\^52 - 
                    22250\ M\^54 - 129241\ M\^56 - 61461\ M\^58 + 
                    97260\ M\^60 + 79175\ M\^62 - 72425\ M\^64 - 
                    26046\ M\^66 + 58045\ M\^68 - 23130\ M\^70 - 
                    32353\ M\^72 + 46956\ M\^74 - 9402\ M\^76 - 
                    27567\ M\^78 + 26120\ M\^80 - 4033\ M\^82 - 9158\ M\^84 + 
                    8766\ M\^86 - 4169\ M\^88 + 1223\ M\^90 - 223\ M\^92 + 
                    23\ M\^94 - M\^96)\) + 
              L\^4\ \((\(-5\)\ M\^34 + 72\ M\^36 - 434\ M\^38 + 1378\ M\^40 - 
                    2331\ M\^42 + 1294\ M\^44 + 2397\ M\^46 - 4235\ M\^48 - 
                    946\ M\^50 + 9307\ M\^52 - 6989\ M\^54 - 11293\ M\^56 + 
                    16306\ M\^58 + 3274\ M\^60 - 9338\ M\^62 + 12631\ M\^64 - 
                    15204\ M\^66 - 12564\ M\^68 + 27060\ M\^70 - 
                    9436\ M\^72 - 14350\ M\^74 + 20233\ M\^76 - 5351\ M\^78 - 
                    9027\ M\^80 + 8803\ M\^82 - 1719\ M\^84 - 1964\ M\^86 + 
                    1696\ M\^88 - 652\ M\^90 + 144\ M\^92 - 18\ M\^94 + 
                    M\^96)\) + 
              L\^6\ \((\(-M\^14\) + 17\ M\^16 - 123\ M\^18 + 499\ M\^20 - 
                    1279\ M\^22 + 2282\ M\^24 - 3073\ M\^26 + 2811\ M\^28 - 
                    795\ M\^30 + 10\ M\^32 - 3195\ M\^34 + 3880\ M\^36 - 
                    928\ M\^38 - 6671\ M\^40 + 24956\ M\^42 + 1191\ M\^44 - 
                    51271\ M\^46 - 11072\ M\^48 + 37341\ M\^50 + 
                    44111\ M\^52 + 2159\ M\^54 - 23992\ M\^56 - 691\ M\^58 - 
                    31029\ M\^60 - 42266\ M\^62 + 54181\ M\^64 + 
                    33151\ M\^66 - 48175\ M\^68 + 12795\ M\^70 + 
                    14539\ M\^72 - 32517\ M\^74 + 27162\ M\^76 + 
                    3523\ M\^78 - 25266\ M\^80 + 17762\ M\^82 + 124\ M\^84 - 
                    7691\ M\^86 + 5592\ M\^88 - 2100\ M\^90 + 463\ M\^92 - 
                    57\ M\^94 + 3\ M\^96)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 145]\), \(L\^7\ M\^2 + M\^44 + 
              L\^6\ \((1 - M\^2 - 2\ M\^4 - 3\ M\^6 + 8\ M\^8 - 4\ M\^10 + 
                    7\ M\^12 - 3\ M\^14 - M\^18 + M\^20)\) + 
              L\^5\ \((\(-2\)\ M\^4 + 10\ M\^6 - 7\ M\^8 - 16\ M\^12 + 
                    5\ M\^14 + 2\ M\^16 + 13\ M\^18 - 2\ M\^20 - 6\ M\^22 + 
                    5\ M\^24 - M\^26)\) + 
              L\^4\ \((M\^8 - 8\ M\^10 + 15\ M\^12 + 8\ M\^14 - 6\ M\^16 - 
                    23\ M\^18 - 15\ M\^20 + 18\ M\^22 + 4\ M\^24 + M\^26 - 
                    3\ M\^28 + 4\ M\^30 - M\^32)\) + 
              L\^3\ \((\(-M\^14\) + 4\ M\^16 - 3\ M\^18 + M\^20 + 4\ M\^22 + 
                    18\ M\^24 - 15\ M\^26 - 23\ M\^28 - 6\ M\^30 + 8\ M\^32 + 
                    15\ M\^34 - 8\ M\^36 + M\^38)\) + 
              L\^2\ \((\(-M\^20\) + 5\ M\^22 - 6\ M\^24 - 2\ M\^26 + 
                    13\ M\^28 + 2\ M\^30 + 5\ M\^32 - 16\ M\^34 - 7\ M\^38 + 
                    10\ M\^40 - 2\ M\^42)\) + 
              L\ \((M\^26 - M\^28 - 3\ M\^32 + 7\ M\^34 - 4\ M\^36 + 
                    8\ M\^38 - 3\ M\^40 - 2\ M\^42 - M\^44 + M\^46)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 146]\), \(L\^17\ M\^36 + 
              M\^46 + L\^16\ \((2\ M\^30 - 24\ M\^34 + 42\ M\^36 + 
                    21\ M\^38 - 42\ M\^40 + 22\ M\^42 - 12\ M\^44 + 
                    4\ M\^46)\) + 
              L\ \((4\ M\^36 - 12\ M\^38 + 22\ M\^40 - 42\ M\^42 + 
                    21\ M\^44 + 42\ M\^46 - 24\ M\^48 + 2\ M\^52)\) + 
              L\^15\ \((5\ M\^24 - 23\ M\^26 + 12\ M\^28 + 132\ M\^30 - 
                    212\ M\^32 - 202\ M\^34 + 475\ M\^36 + 168\ M\^38 - 
                    194\ M\^40 - 46\ M\^42 - 215\ M\^44 + 104\ M\^46 + 
                    227\ M\^48 - 237\ M\^50 + 112\ M\^52 - 36\ M\^54 + 
                    6\ M\^56)\) + 
              L\^2\ \((6\ M\^26 - 36\ M\^28 + 112\ M\^30 - 237\ M\^32 + 
                    227\ M\^34 + 104\ M\^36 - 215\ M\^38 - 46\ M\^40 - 
                    194\ M\^42 + 168\ M\^44 + 475\ M\^46 - 202\ M\^48 - 
                    212\ M\^50 + 132\ M\^52 + 12\ M\^54 - 23\ M\^56 + 
                    5\ M\^58)\) + 
              L\^3\ \((4\ M\^16 - 36\ M\^18 + 158\ M\^20 - 409\ M\^22 + 
                    530\ M\^24 - 97\ M\^26 - 239\ M\^28 - 324\ M\^30 + 
                    247\ M\^32 + 45\ M\^34 + 1262\ M\^36 + 741\ M\^38 - 
                    3238\ M\^40 - 2375\ M\^42 + 3162\ M\^44 + 3625\ M\^46 - 
                    1499\ M\^48 - 2819\ M\^50 + 1036\ M\^52 + 1120\ M\^54 - 
                    654\ M\^56 - 129\ M\^58 + 217\ M\^60 - 81\ M\^62 + 
                    14\ M\^64 - M\^66)\) + 
              L\^14\ \((\(-M\^16\) + 14\ M\^18 - 81\ M\^20 + 217\ M\^22 - 
                    129\ M\^24 - 654\ M\^26 + 1120\ M\^28 + 1036\ M\^30 - 
                    2819\ M\^32 - 1499\ M\^34 + 3625\ M\^36 + 3162\ M\^38 - 
                    2375\ M\^40 - 3238\ M\^42 + 741\ M\^44 + 1262\ M\^46 + 
                    45\ M\^48 + 247\ M\^50 - 324\ M\^52 - 239\ M\^54 - 
                    97\ M\^56 + 530\ M\^58 - 409\ M\^60 + 158\ M\^62 - 
                    36\ M\^64 + 4\ M\^66)\) + 
              L\^4\ \((M\^6 - 12\ M\^8 + 68\ M\^10 - 215\ M\^12 + 
                    366\ M\^14 - 305\ M\^16 + 296\ M\^18 - 891\ M\^20 + 
                    1505\ M\^22 - 1981\ M\^24 + 1828\ M\^26 + 3866\ M\^28 - 
                    7877\ M\^30 - 5230\ M\^32 + 10074\ M\^34 + 8117\ M\^36 + 
                    496\ M\^38 - 12387\ M\^40 - 18261\ M\^42 + 9336\ M\^44 + 
                    26429\ M\^46 - 578\ M\^48 - 20712\ M\^50 - 2276\ M\^52 + 
                    12600\ M\^54 - 708\ M\^56 - 4898\ M\^58 + 1675\ M\^60 + 
                    756\ M\^62 - 717\ M\^64 + 227\ M\^66 - 34\ M\^68 + 
                    2\ M\^70)\) + 
              L\^5\ \((\(-M\^2\) + 17\ M\^4 - 112\ M\^6 + 383\ M\^8 - 
                    783\ M\^10 + 1241\ M\^12 - 1983\ M\^14 + 1747\ M\^16 + 
                    1859\ M\^18 - 2789\ M\^20 - 6331\ M\^22 + 7799\ M\^24 + 
                    4626\ M\^26 + 785\ M\^28 + 5869\ M\^30 - 40668\ M\^32 - 
                    20291\ M\^34 + 93329\ M\^36 + 37425\ M\^38 - 
                    98237\ M\^40 - 59733\ M\^42 + 34383\ M\^44 + 
                    67330\ M\^46 + 32542\ M\^48 - 51216\ M\^50 - 
                    45909\ M\^52 + 30200\ M\^54 + 29489\ M\^56 - 
                    18373\ M\^58 - 9311\ M\^60 + 8852\ M\^62 - 132\ M\^64 - 
                    2065\ M\^66 + 976\ M\^68 - 212\ M\^70 + 23\ M\^72 - 
                    M\^74)\) + 
              L\^6\ \((\(-1\) + 18\ M\^2 - 144\ M\^4 + 685\ M\^6 - 
                    2086\ M\^8 + 3810\ M\^10 - 3039\ M\^12 - 985\ M\^14 + 
                    1240\ M\^16 + 858\ M\^18 + 10442\ M\^20 - 6258\ M\^22 - 
                    46244\ M\^24 + 26843\ M\^26 + 96774\ M\^28 - 
                    35409\ M\^30 - 133346\ M\^32 - 30510\ M\^34 + 
                    131226\ M\^36 + 185125\ M\^38 - 78573\ M\^40 - 
                    311161\ M\^42 - 7332\ M\^44 + 267908\ M\^46 + 
                    62153\ M\^48 - 105204\ M\^50 - 68373\ M\^52 - 
                    8203\ M\^54 + 54671\ M\^56 + 32808\ M\^58 - 
                    41283\ M\^60 - 12180\ M\^62 + 21865\ M\^64 - 
                    3246\ M\^66 - 4683\ M\^68 + 2817\ M\^70 - 699\ M\^72 + 
                    84\ M\^74 - 4\ M\^76)\) + 
              L\^13\ \((2\ M\^12 - 34\ M\^14 + 227\ M\^16 - 717\ M\^18 + 
                    756\ M\^20 + 1675\ M\^22 - 4898\ M\^24 - 708\ M\^26 + 
                    12600\ M\^28 - 2276\ M\^30 - 20712\ M\^32 - 578\ M\^34 + 
                    26429\ M\^36 + 9336\ M\^38 - 18261\ M\^40 - 
                    12387\ M\^42 + 496\ M\^44 + 8117\ M\^46 + 10074\ M\^48 - 
                    5230\ M\^50 - 7877\ M\^52 + 3866\ M\^54 + 1828\ M\^56 - 
                    1981\ M\^58 + 1505\ M\^60 - 891\ M\^62 + 296\ M\^64 - 
                    305\ M\^66 + 366\ M\^68 - 215\ M\^70 + 68\ M\^72 - 
                    12\ M\^74 + M\^76)\) + 
              L\^12\ \((\(-M\^8\) + 23\ M\^10 - 212\ M\^12 + 976\ M\^14 - 
                    2065\ M\^16 - 132\ M\^18 + 8852\ M\^20 - 9311\ M\^22 - 
                    18373\ M\^24 + 29489\ M\^26 + 30200\ M\^28 - 
                    45909\ M\^30 - 51216\ M\^32 + 32542\ M\^34 + 
                    67330\ M\^36 + 34383\ M\^38 - 59733\ M\^40 - 
                    98237\ M\^42 + 37425\ M\^44 + 93329\ M\^46 - 
                    20291\ M\^48 - 40668\ M\^50 + 5869\ M\^52 + 785\ M\^54 + 
                    4626\ M\^56 + 7799\ M\^58 - 6331\ M\^60 - 2789\ M\^62 + 
                    1859\ M\^64 + 1747\ M\^66 - 1983\ M\^68 + 1241\ M\^70 - 
                    783\ M\^72 + 383\ M\^74 - 112\ M\^76 + 17\ M\^78 - 
                    M\^80)\) + 
              L\^7\ \((3 - 56\ M\^2 + 438\ M\^4 - 1896\ M\^6 + 5013\ M\^8 - 
                    7933\ M\^10 + 4333\ M\^12 + 11311\ M\^14 - 23256\ M\^16 - 
                    4413\ M\^18 + 42704\ M\^20 + 7563\ M\^22 - 60967\ M\^24 - 
                    70890\ M\^26 + 101213\ M\^28 + 212470\ M\^30 - 
                    161232\ M\^32 - 376773\ M\^34 + 162839\ M\^36 + 
                    484976\ M\^38 - 19903\ M\^40 - 463142\ M\^42 - 
                    201098\ M\^44 + 293600\ M\^46 + 294212\ M\^48 - 
                    112881\ M\^50 - 190542\ M\^52 + 12403\ M\^54 + 
                    52164\ M\^56 + 26326\ M\^58 + 11907\ M\^60 - 
                    38462\ M\^62 - 8173\ M\^64 + 27246\ M\^66 - 7156\ M\^68 - 
                    6290\ M\^70 + 5210\ M\^72 - 1712\ M\^74 + 298\ M\^76 - 
                    27\ M\^78 + M\^80)\) + 
              L\^9\ \((1 - 25\ M\^2 + 268\ M\^4 - 1573\ M\^6 + 5291\ M\^8 - 
                    9196\ M\^10 + 2659\ M\^12 + 19119\ M\^14 - 31106\ M\^16 + 
                    7365\ M\^18 + 37517\ M\^20 - 76052\ M\^22 + 
                    23630\ M\^24 + 169576\ M\^26 - 129936\ M\^28 - 
                    289113\ M\^30 + 171682\ M\^32 + 429307\ M\^34 - 
                    69611\ M\^36 - 504329\ M\^38 - 70170\ M\^40 + 
                    443676\ M\^42 + 137993\ M\^44 - 318669\ M\^46 - 
                    157014\ M\^48 + 174960\ M\^50 + 170699\ M\^52 - 
                    76876\ M\^54 - 149383\ M\^56 + 56698\ M\^58 + 
                    82328\ M\^60 - 56635\ M\^62 - 12164\ M\^64 + 
                    32372\ M\^66 - 17766\ M\^68 - 1694\ M\^70 + 9572\ M\^72 - 
                    6936\ M\^74 + 2597\ M\^76 - 552\ M\^78 + 63\ M\^80 - 
                    3\ M\^82)\) + 
              L\^11\ \((\(-4\)\ M\^6 + 84\ M\^8 - 699\ M\^10 + 2817\ M\^12 - 
                    4683\ M\^14 - 3246\ M\^16 + 21865\ M\^18 - 12180\ M\^20 - 
                    41283\ M\^22 + 32808\ M\^24 + 54671\ M\^26 - 
                    8203\ M\^28 - 68373\ M\^30 - 105204\ M\^32 + 
                    62153\ M\^34 + 267908\ M\^36 - 7332\ M\^38 - 
                    311161\ M\^40 - 78573\ M\^42 + 185125\ M\^44 + 
                    131226\ M\^46 - 30510\ M\^48 - 133346\ M\^50 - 
                    35409\ M\^52 + 96774\ M\^54 + 26843\ M\^56 - 
                    46244\ M\^58 - 6258\ M\^60 + 10442\ M\^62 + 858\ M\^64 + 
                    1240\ M\^66 - 985\ M\^68 - 3039\ M\^70 + 3810\ M\^72 - 
                    2086\ M\^74 + 685\ M\^76 - 144\ M\^78 + 18\ M\^80 - 
                    M\^82)\) + 
              L\^8\ \((\(-3\) + 63\ M\^2 - 552\ M\^4 + 2597\ M\^6 - 
                    6936\ M\^8 + 9572\ M\^10 - 1694\ M\^12 - 17766\ M\^14 + 
                    32372\ M\^16 - 12164\ M\^18 - 56635\ M\^20 + 
                    82328\ M\^22 + 56698\ M\^24 - 149383\ M\^26 - 
                    76876\ M\^28 + 170699\ M\^30 + 174960\ M\^32 - 
                    157014\ M\^34 - 318669\ M\^36 + 137993\ M\^38 + 
                    443676\ M\^40 - 70170\ M\^42 - 504329\ M\^44 - 
                    69611\ M\^46 + 429307\ M\^48 + 171682\ M\^50 - 
                    289113\ M\^52 - 129936\ M\^54 + 169576\ M\^56 + 
                    23630\ M\^58 - 76052\ M\^60 + 37517\ M\^62 + 
                    7365\ M\^64 - 31106\ M\^66 + 19119\ M\^68 + 2659\ M\^70 - 
                    9196\ M\^72 + 5291\ M\^74 - 1573\ M\^76 + 268\ M\^78 - 
                    25\ M\^80 + M\^82)\) + 
              L\^10\ \((M\^2 - 27\ M\^4 + 298\ M\^6 - 1712\ M\^8 + 
                    5210\ M\^10 - 6290\ M\^12 - 7156\ M\^14 + 27246\ M\^16 - 
                    8173\ M\^18 - 38462\ M\^20 + 11907\ M\^22 + 
                    26326\ M\^24 + 52164\ M\^26 + 12403\ M\^28 - 
                    190542\ M\^30 - 112881\ M\^32 + 294212\ M\^34 + 
                    293600\ M\^36 - 201098\ M\^38 - 463142\ M\^40 - 
                    19903\ M\^42 + 484976\ M\^44 + 162839\ M\^46 - 
                    376773\ M\^48 - 161232\ M\^50 + 212470\ M\^52 + 
                    101213\ M\^54 - 70890\ M\^56 - 60967\ M\^58 + 
                    7563\ M\^60 + 42704\ M\^62 - 4413\ M\^64 - 23256\ M\^66 + 
                    11311\ M\^68 + 4333\ M\^70 - 7933\ M\^72 + 5013\ M\^74 - 
                    1896\ M\^76 + 438\ M\^78 - 56\ M\^80 + 3\ M\^82)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 147]\), \(M\^22 + 
              L\^16\ M\^78 + 
              L\ \((\(-2\)\ M\^14 + 6\ M\^16 - 13\ M\^18 + 30\ M\^20 - 
                    29\ M\^22 - 8\ M\^24 + 35\ M\^26 - 4\ M\^28 - 
                    3\ M\^30)\) + 
              L\^2\ \((M\^6 - 6\ M\^8 + 19\ M\^10 - 42\ M\^12 + 64\ M\^14 - 
                    66\ M\^16 + 104\ M\^18 - 194\ M\^20 + 43\ M\^22 + 
                    346\ M\^24 - 236\ M\^26 - 281\ M\^28 + 295\ M\^30 + 
                    120\ M\^32 - 105\ M\^34 - 5\ M\^36 + 7\ M\^38)\) + 
              L\^3\ \((\(-M\^2\) + 8\ M\^4 - 33\ M\^6 + 95\ M\^8 - 
                    204\ M\^10 + 279\ M\^12 - 147\ M\^14 - 79\ M\^16 - 
                    121\ M\^18 + 556\ M\^20 + 32\ M\^22 - 853\ M\^24 - 
                    557\ M\^26 + 1761\ M\^28 + 714\ M\^30 - 2200\ M\^32 - 
                    276\ M\^34 + 1663\ M\^36 - 5\ M\^38 - 535\ M\^40 + 
                    17\ M\^42 + 111\ M\^44 - 31\ M\^46 + 2\ M\^48)\) + 
              L\^4\ \((\(-1\) + 9\ M\^2 - 44\ M\^4 + 157\ M\^6 - 397\ M\^8 + 
                    597\ M\^10 - 325\ M\^12 - 270\ M\^14 + 101\ M\^16 + 
                    901\ M\^18 - 1081\ M\^20 - 80\ M\^22 + 232\ M\^24 + 
                    1676\ M\^26 + 89\ M\^28 - 5216\ M\^30 + 407\ M\^32 + 
                    8773\ M\^34 - 1888\ M\^36 - 8544\ M\^38 + 2830\ M\^40 + 
                    5159\ M\^42 - 2085\ M\^44 - 1299\ M\^46 + 582\ M\^48 + 
                    201\ M\^50 - 146\ M\^52 + 28\ M\^54 - 2\ M\^56)\) + 
              L\^5\ \((1 - 14\ M\^2 + 83\ M\^4 - 263\ M\^6 + 455\ M\^8 - 
                    430\ M\^10 + 559\ M\^12 - 1550\ M\^14 + 1654\ M\^16 + 
                    1491\ M\^18 - 3176\ M\^20 - 1041\ M\^22 + 3821\ M\^24 - 
                    2487\ M\^26 + 1101\ M\^28 + 2898\ M\^30 - 1316\ M\^32 - 
                    1608\ M\^34 - 10279\ M\^36 + 3838\ M\^38 + 24676\ M\^40 - 
                    10912\ M\^42 - 25816\ M\^44 + 14651\ M\^46 + 
                    15052\ M\^48 - 11225\ M\^50 - 3145\ M\^52 + 4132\ M\^54 - 
                    381\ M\^56 - 641\ M\^58 + 283\ M\^60 - 51\ M\^62 + 
                    4\ M\^64)\) + 
              L\^6\ \((\(-2\)\ M\^4 + 31\ M\^6 - 221\ M\^8 + 890\ M\^10 - 
                    1976\ M\^12 + 1630\ M\^14 + 2076\ M\^16 - 3772\ M\^18 - 
                    4607\ M\^20 + 9837\ M\^22 + 5952\ M\^24 - 13699\ M\^26 - 
                    7449\ M\^28 + 10458\ M\^30 + 4158\ M\^32 + 3395\ M\^34 + 
                    2852\ M\^36 - 9195\ M\^38 - 10499\ M\^40 - 11096\ M\^42 + 
                    18355\ M\^44 + 41904\ M\^46 - 30904\ M\^48 - 
                    48405\ M\^50 + 36914\ M\^52 + 28490\ M\^54 - 
                    28974\ M\^56 - 4335\ M\^58 + 12016\ M\^60 - 3403\ M\^62 - 
                    1361\ M\^64 + 1367\ M\^66 - 537\ M\^68 + 126\ M\^70 - 
                    17\ M\^72 + M\^74)\) + 
              L\^7\ \((M\^8 - 20\ M\^10 + 190\ M\^12 - 1029\ M\^14 + 
                    3191\ M\^16 - 4639\ M\^18 - 1764\ M\^20 + 15059\ M\^22 - 
                    7760\ M\^24 - 29311\ M\^26 + 24977\ M\^28 + 
                    42137\ M\^30 - 31383\ M\^32 - 52153\ M\^34 + 
                    7937\ M\^36 + 49759\ M\^38 + 40428\ M\^40 - 
                    23567\ M\^42 - 70460\ M\^44 - 18388\ M\^46 + 
                    47649\ M\^48 + 48664\ M\^50 + 5657\ M\^52 - 
                    65005\ M\^54 - 33265\ M\^56 + 64963\ M\^58 + 
                    21724\ M\^60 - 47635\ M\^62 + 2108\ M\^64 + 
                    20411\ M\^66 - 9358\ M\^68 - 1922\ M\^70 + 3506\ M\^72 - 
                    1626\ M\^74 + 403\ M\^76 - 54\ M\^78 + 3\ M\^80)\) + 
              L\^15\ \((\(-3\)\ M\^70 - 4\ M\^72 + 35\ M\^74 - 8\ M\^76 - 
                    29\ M\^78 + 30\ M\^80 - 13\ M\^82 + 6\ M\^84 - 
                    2\ M\^86)\) + 
              L\^8\ \((3\ M\^14 - 55\ M\^16 + 445\ M\^18 - 1980\ M\^20 + 
                    4784\ M\^22 - 3833\ M\^24 - 9691\ M\^26 + 23912\ M\^28 + 
                    2580\ M\^30 - 52663\ M\^32 + 13912\ M\^34 + 
                    77548\ M\^36 - 9605\ M\^38 - 91075\ M\^40 - 
                    41857\ M\^42 + 81466\ M\^44 + 120981\ M\^46 - 
                    35432\ M\^48 - 159738\ M\^50 - 35432\ M\^52 + 
                    120981\ M\^54 + 81466\ M\^56 - 41857\ M\^58 - 
                    91075\ M\^60 - 9605\ M\^62 + 77548\ M\^64 + 
                    13912\ M\^66 - 52663\ M\^68 + 2580\ M\^70 + 
                    23912\ M\^72 - 9691\ M\^74 - 3833\ M\^76 + 4784\ M\^78 - 
                    1980\ M\^80 + 445\ M\^82 - 55\ M\^84 + 3\ M\^86)\) + 
              L\^9\ \((3\ M\^20 - 54\ M\^22 + 403\ M\^24 - 1626\ M\^26 + 
                    3506\ M\^28 - 1922\ M\^30 - 9358\ M\^32 + 20411\ M\^34 + 
                    2108\ M\^36 - 47635\ M\^38 + 21724\ M\^40 + 
                    64963\ M\^42 - 33265\ M\^44 - 65005\ M\^46 + 
                    5657\ M\^48 + 48664\ M\^50 + 47649\ M\^52 - 
                    18388\ M\^54 - 70460\ M\^56 - 23567\ M\^58 + 
                    40428\ M\^60 + 49759\ M\^62 + 7937\ M\^64 - 
                    52153\ M\^66 - 31383\ M\^68 + 42137\ M\^70 + 
                    24977\ M\^72 - 29311\ M\^74 - 7760\ M\^76 + 
                    15059\ M\^78 - 1764\ M\^80 - 4639\ M\^82 + 3191\ M\^84 - 
                    1029\ M\^86 + 190\ M\^88 - 20\ M\^90 + M\^92)\) + 
              L\^14\ \((7\ M\^62 - 5\ M\^64 - 105\ M\^66 + 120\ M\^68 + 
                    295\ M\^70 - 281\ M\^72 - 236\ M\^74 + 346\ M\^76 + 
                    43\ M\^78 - 194\ M\^80 + 104\ M\^82 - 66\ M\^84 + 
                    64\ M\^86 - 42\ M\^88 + 19\ M\^90 - 6\ M\^92 + M\^94)\) + 
              L\^10\ \((M\^26 - 17\ M\^28 + 126\ M\^30 - 537\ M\^32 + 
                    1367\ M\^34 - 1361\ M\^36 - 3403\ M\^38 + 12016\ M\^40 - 
                    4335\ M\^42 - 28974\ M\^44 + 28490\ M\^46 + 
                    36914\ M\^48 - 48405\ M\^50 - 30904\ M\^52 + 
                    41904\ M\^54 + 18355\ M\^56 - 11096\ M\^58 - 
                    10499\ M\^60 - 9195\ M\^62 + 2852\ M\^64 + 3395\ M\^66 + 
                    4158\ M\^68 + 10458\ M\^70 - 7449\ M\^72 - 13699\ M\^74 + 
                    5952\ M\^76 + 9837\ M\^78 - 4607\ M\^80 - 3772\ M\^82 + 
                    2076\ M\^84 + 1630\ M\^86 - 1976\ M\^88 + 890\ M\^90 - 
                    221\ M\^92 + 31\ M\^94 - 2\ M\^96)\) + 
              L\^13\ \((2\ M\^52 - 31\ M\^54 + 111\ M\^56 + 17\ M\^58 - 
                    535\ M\^60 - 5\ M\^62 + 1663\ M\^64 - 276\ M\^66 - 
                    2200\ M\^68 + 714\ M\^70 + 1761\ M\^72 - 557\ M\^74 - 
                    853\ M\^76 + 32\ M\^78 + 556\ M\^80 - 121\ M\^82 - 
                    79\ M\^84 - 147\ M\^86 + 279\ M\^88 - 204\ M\^90 + 
                    95\ M\^92 - 33\ M\^94 + 8\ M\^96 - M\^98)\) + 
              L\^12\ \((\(-2\)\ M\^44 + 28\ M\^46 - 146\ M\^48 + 201\ M\^50 + 
                    582\ M\^52 - 1299\ M\^54 - 2085\ M\^56 + 5159\ M\^58 + 
                    2830\ M\^60 - 8544\ M\^62 - 1888\ M\^64 + 8773\ M\^66 + 
                    407\ M\^68 - 5216\ M\^70 + 89\ M\^72 + 1676\ M\^74 + 
                    232\ M\^76 - 80\ M\^78 - 1081\ M\^80 + 901\ M\^82 + 
                    101\ M\^84 - 270\ M\^86 - 325\ M\^88 + 597\ M\^90 - 
                    397\ M\^92 + 157\ M\^94 - 44\ M\^96 + 9\ M\^98 - 
                    M\^100)\) + 
              L\^11\ \((4\ M\^36 - 51\ M\^38 + 283\ M\^40 - 641\ M\^42 - 
                    381\ M\^44 + 4132\ M\^46 - 3145\ M\^48 - 11225\ M\^50 + 
                    15052\ M\^52 + 14651\ M\^54 - 25816\ M\^56 - 
                    10912\ M\^58 + 24676\ M\^60 + 3838\ M\^62 - 
                    10279\ M\^64 - 1608\ M\^66 - 1316\ M\^68 + 2898\ M\^70 + 
                    1101\ M\^72 - 2487\ M\^74 + 3821\ M\^76 - 1041\ M\^78 - 
                    3176\ M\^80 + 1491\ M\^82 + 1654\ M\^84 - 1550\ M\^86 + 
                    559\ M\^88 - 430\ M\^90 + 455\ M\^92 - 263\ M\^94 + 
                    83\ M\^96 - 14\ M\^98 + M\^100)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 148]\), \(M\^94 - 2\ M\^98 + 
              M\^102 + L\^21\ \((M\^16 - 2\ M\^20 + M\^24)\) + 
              L\^20\ \((\(-M\^8\) + 2\ M\^10 + 3\ M\^12 - 6\ M\^14 - 
                    4\ M\^16 - 6\ M\^18 + 46\ M\^20 + 17\ M\^22 - 
                    113\ M\^24 + 26\ M\^26 + 84\ M\^28 - 48\ M\^30 - 
                    8\ M\^32 + 13\ M\^34 - 7\ M\^36 + 3\ M\^38)\) + 
              L\^19\ \((\(-2\)\ M\^6 + 2\ M\^8 + 27\ M\^10 - 69\ M\^12 - 
                    22\ M\^14 + 255\ M\^16 - 145\ M\^18 - 521\ M\^20 + 
                    544\ M\^22 + 868\ M\^24 - 1108\ M\^26 - 900\ M\^28 + 
                    1456\ M\^30 + 357\ M\^32 - 1151\ M\^34 + 309\ M\^36 + 
                    314\ M\^38 - 360\ M\^40 + 195\ M\^42 + 19\ M\^44 - 
                    92\ M\^46 + 52\ M\^48 - 16\ M\^50 + 3\ M\^52)\) + 
              L\^18\ \((\(-5\)\ M\^4 + 28\ M\^6 - 43\ M\^8 - 69\ M\^10 + 
                    321\ M\^12 - 250\ M\^14 - 718\ M\^16 + 1754\ M\^18 - 
                    25\ M\^20 - 4354\ M\^22 + 2887\ M\^24 + 6871\ M\^26 - 
                    6673\ M\^28 - 7377\ M\^30 + 8559\ M\^32 + 6025\ M\^34 - 
                    7515\ M\^36 - 3618\ M\^38 + 5297\ M\^40 + 604\ M\^42 - 
                    2872\ M\^44 + 1530\ M\^46 + 371\ M\^48 - 1229\ M\^50 + 
                    738\ M\^52 - 49\ M\^54 - 200\ M\^56 + 197\ M\^58 - 
                    115\ M\^60 + 41\ M\^62 - 9\ M\^64 + M\^66)\) + 
              L\^17\ \((1 - 13\ M\^2 + 72\ M\^4 - 206\ M\^6 + 225\ M\^8 + 
                    365\ M\^10 - 1267\ M\^12 + 480\ M\^14 + 1871\ M\^16 - 
                    1634\ M\^18 + 685\ M\^20 - 2655\ M\^22 - 5093\ M\^24 + 
                    16178\ M\^26 + 7352\ M\^28 - 33709\ M\^30 - 5761\ M\^32 + 
                    44005\ M\^34 + 3385\ M\^36 - 39961\ M\^38 - 307\ M\^40 + 
                    25740\ M\^42 - 2779\ M\^44 - 10930\ M\^46 + 3389\ M\^48 + 
                    3533\ M\^50 - 2828\ M\^52 - 2031\ M\^54 + 3366\ M\^56 + 
                    108\ M\^58 - 2398\ M\^60 + 1570\ M\^62 - 54\ M\^64 - 
                    623\ M\^66 + 531\ M\^68 - 236\ M\^70 + 64\ M\^72 - 
                    11\ M\^74 + M\^76)\) + 
              L\^16\ \((\(-2\) + 29\ M\^2 - 193\ M\^4 + 688\ M\^6 - 
                    1087\ M\^8 - 671\ M\^10 + 5103\ M\^12 - 4086\ M\^14 - 
                    9387\ M\^16 + 14070\ M\^18 + 12732\ M\^20 - 
                    22277\ M\^22 - 29614\ M\^24 + 31193\ M\^26 + 
                    76066\ M\^28 - 53447\ M\^30 - 142931\ M\^32 + 
                    87468\ M\^34 + 193032\ M\^36 - 110419\ M\^38 - 
                    192262\ M\^40 + 108147\ M\^42 + 142499\ M\^44 - 
                    84030\ M\^46 - 74800\ M\^48 + 54683\ M\^50 + 
                    23558\ M\^52 - 30797\ M\^54 + 1498\ M\^56 + 
                    12371\ M\^58 - 5986\ M\^60 - 370\ M\^62 - 106\ M\^64 - 
                    514\ M\^66 + 2588\ M\^68 - 2167\ M\^70 + 198\ M\^72 + 
                    978\ M\^74 - 967\ M\^76 + 492\ M\^78 - 147\ M\^80 + 
                    25\ M\^82 - 2\ M\^84)\) + 
              L\^15\ \((1 - 20\ M\^2 + 184\ M\^4 - 910\ M\^6 + 2366\ M\^8 - 
                    1819\ M\^10 - 6477\ M\^12 + 16503\ M\^14 + 599\ M\^16 - 
                    44189\ M\^18 + 28773\ M\^20 + 77413\ M\^22 - 
                    83610\ M\^24 - 121628\ M\^26 + 169603\ M\^28 + 
                    189329\ M\^30 - 289308\ M\^32 - 283312\ M\^34 + 
                    413472\ M\^36 + 381464\ M\^38 - 490295\ M\^40 - 
                    433930\ M\^42 + 476589\ M\^44 + 393809\ M\^46 - 
                    366603\ M\^48 - 271792\ M\^50 + 220469\ M\^52 + 
                    140003\ M\^54 - 103485\ M\^56 - 52919\ M\^58 + 
                    41322\ M\^60 + 15734\ M\^62 - 19648\ M\^64 - 
                    1885\ M\^66 + 12263\ M\^68 - 6664\ M\^70 - 1381\ M\^72 + 
                    4614\ M\^74 - 4010\ M\^76 + 1330\ M\^78 + 1061\ M\^80 - 
                    1621\ M\^82 + 1030\ M\^84 - 405\ M\^86 + 102\ M\^88 - 
                    15\ M\^90 + M\^92)\) + 
              L\^14\ \((4\ M\^2 - 65\ M\^4 + 489\ M\^6 - 2044\ M\^8 + 
                    4457\ M\^10 - 1841\ M\^12 - 14455\ M\^14 + 29044\ M\^16 + 
                    7233\ M\^18 - 83187\ M\^20 + 48592\ M\^22 + 
                    147098\ M\^24 - 171582\ M\^26 - 209533\ M\^28 + 
                    384676\ M\^30 + 261875\ M\^32 - 702286\ M\^34 - 
                    295034\ M\^36 + 1077987\ M\^38 + 305026\ M\^40 - 
                    1374863\ M\^42 - 299294\ M\^44 + 1407127\ M\^46 + 
                    280542\ M\^48 - 1124443\ M\^50 - 247596\ M\^52 + 
                    703508\ M\^54 + 186916\ M\^56 - 343855\ M\^58 - 
                    98560\ M\^60 + 127265\ M\^62 + 25998\ M\^64 - 
                    24384\ M\^66 + 2264\ M\^68 - 11133\ M\^70 + 2319\ M\^72 + 
                    12638\ M\^74 - 13023\ M\^76 + 2053\ M\^78 + 7283\ M\^80 - 
                    7008\ M\^82 + 1625\ M\^84 + 1791\ M\^86 - 2088\ M\^88 + 
                    1193\ M\^90 - 440\ M\^92 + 105\ M\^94 - 15\ M\^96 + 
                    M\^98)\) + 
              L\^13\ \((6\ M\^4 - 95\ M\^6 + 671\ M\^8 - 2700\ M\^10 + 
                    5993\ M\^12 - 3560\ M\^14 - 17176\ M\^16 + 42685\ M\^18 - 
                    5659\ M\^20 - 111729\ M\^22 + 115193\ M\^24 + 
                    164031\ M\^26 - 335001\ M\^28 - 138598\ M\^30 + 
                    655475\ M\^32 - 7151\ M\^34 - 1048694\ M\^36 + 
                    242317\ M\^38 + 1499820\ M\^40 - 462593\ M\^42 - 
                    1956734\ M\^44 + 559805\ M\^46 + 2237149\ M\^48 - 
                    514583\ M\^50 - 2149673\ M\^52 + 375390\ M\^54 + 
                    1673548\ M\^56 - 231214\ M\^58 - 997557\ M\^60 + 
                    115567\ M\^62 + 430219\ M\^64 - 23293\ M\^66 - 
                    131434\ M\^68 - 22476\ M\^70 + 49209\ M\^72 + 
                    15375\ M\^74 - 38953\ M\^76 + 9761\ M\^78 + 
                    26205\ M\^80 - 23556\ M\^82 - 2416\ M\^84 + 
                    13339\ M\^86 - 6849\ M\^88 - 674\ M\^90 + 2818\ M\^92 - 
                    1896\ M\^94 + 711\ M\^96 - 157\ M\^98 + 19\ M\^100 - 
                    M\^102)\) + 
              L\^12\ \((4\ M\^6 - 67\ M\^8 + 498\ M\^10 - 2201\ M\^12 + 
                    5846\ M\^14 - 6294\ M\^16 - 12060\ M\^18 + 46594\ M\^20 - 
                    25228\ M\^22 - 107214\ M\^24 + 160770\ M\^26 + 
                    124655\ M\^28 - 415798\ M\^30 + 4436\ M\^32 + 
                    752816\ M\^34 - 383102\ M\^36 - 1095772\ M\^38 + 
                    1019946\ M\^40 + 1426562\ M\^42 - 1790999\ M\^44 - 
                    1751699\ M\^46 + 2436612\ M\^48 + 1994444\ M\^50 - 
                    2642673\ M\^52 - 2016534\ M\^54 + 2233932\ M\^56 + 
                    1738901\ M\^58 - 1442909\ M\^60 - 1273692\ M\^62 + 
                    695182\ M\^64 + 836957\ M\^66 - 283056\ M\^68 - 
                    504018\ M\^70 + 184996\ M\^72 + 253106\ M\^74 - 
                    186744\ M\^76 - 66052\ M\^78 + 143490\ M\^80 - 
                    28332\ M\^82 - 67811\ M\^84 + 44015\ M\^86 + 
                    9839\ M\^88 - 20730\ M\^90 + 6999\ M\^92 + 2110\ M\^94 - 
                    3145\ M\^96 + 1611\ M\^98 - 471\ M\^100 + 75\ M\^102 - 
                    5\ M\^104)\) + 
              L\^11\ \((M\^8 - 24\ M\^10 + 226\ M\^12 - 1195\ M\^14 + 
                    3838\ M\^16 - 6052\ M\^18 - 3507\ M\^20 + 31465\ M\^22 - 
                    32688\ M\^24 - 58399\ M\^26 + 139813\ M\^28 + 
                    29943\ M\^30 - 323519\ M\^32 + 155131\ M\^34 + 
                    520469\ M\^36 - 615743\ M\^38 - 596918\ M\^40 + 
                    1384750\ M\^42 + 496077\ M\^44 - 2349281\ M\^46 - 
                    330163\ M\^48 + 3225304\ M\^50 + 349659\ M\^52 - 
                    3661363\ M\^54 - 673197\ M\^56 + 3481772\ M\^58 + 
                    1090377\ M\^60 - 2892293\ M\^62 - 1252577\ M\^64 + 
                    2184259\ M\^66 + 996300\ M\^68 - 1553006\ M\^70 - 
                    476950\ M\^72 + 1021615\ M\^74 + 18483\ M\^76 - 
                    568767\ M\^78 + 196527\ M\^80 + 228582\ M\^82 - 
                    188360\ M\^84 - 43763\ M\^86 + 100271\ M\^88 - 
                    17804\ M\^90 - 28937\ M\^92 + 16842\ M\^94 + 355\ M\^96 - 
                    4125\ M\^98 + 2305\ M\^100 - 709\ M\^102 + 123\ M\^104 - 
                    9\ M\^106)\) + 
              L\ \((3\ M\^80 - 7\ M\^82 + 13\ M\^84 - 8\ M\^86 - 48\ M\^88 + 
                    84\ M\^90 + 26\ M\^92 - 113\ M\^94 + 17\ M\^96 + 
                    46\ M\^98 - 6\ M\^100 - 4\ M\^102 - 6\ M\^104 + 
                    3\ M\^106 + 2\ M\^108 - M\^110)\) + 
              L\^10\ \((\(-9\)\ M\^12 + 123\ M\^14 - 709\ M\^16 + 
                    2305\ M\^18 - 4125\ M\^20 + 355\ M\^22 + 16842\ M\^24 - 
                    28937\ M\^26 - 17804\ M\^28 + 100271\ M\^30 - 
                    43763\ M\^32 - 188360\ M\^34 + 228582\ M\^36 + 
                    196527\ M\^38 - 568767\ M\^40 + 18483\ M\^42 + 
                    1021615\ M\^44 - 476950\ M\^46 - 1553006\ M\^48 + 
                    996300\ M\^50 + 2184259\ M\^52 - 1252577\ M\^54 - 
                    2892293\ M\^56 + 1090377\ M\^58 + 3481772\ M\^60 - 
                    673197\ M\^62 - 3661363\ M\^64 + 349659\ M\^66 + 
                    3225304\ M\^68 - 330163\ M\^70 - 2349281\ M\^72 + 
                    496077\ M\^74 + 1384750\ M\^76 - 596918\ M\^78 - 
                    615743\ M\^80 + 520469\ M\^82 + 155131\ M\^84 - 
                    323519\ M\^86 + 29943\ M\^88 + 139813\ M\^90 - 
                    58399\ M\^92 - 32688\ M\^94 + 31465\ M\^96 - 
                    3507\ M\^98 - 6052\ M\^100 + 3838\ M\^102 - 
                    1195\ M\^104 + 226\ M\^106 - 24\ M\^108 + M\^110)\) + 
              L\^2\ \((3\ M\^66 - 16\ M\^68 + 52\ M\^70 - 92\ M\^72 + 
                    19\ M\^74 + 195\ M\^76 - 360\ M\^78 + 314\ M\^80 + 
                    309\ M\^82 - 1151\ M\^84 + 357\ M\^86 + 1456\ M\^88 - 
                    900\ M\^90 - 1108\ M\^92 + 868\ M\^94 + 544\ M\^96 - 
                    521\ M\^98 - 145\ M\^100 + 255\ M\^102 - 22\ M\^104 - 
                    69\ M\^106 + 27\ M\^108 + 2\ M\^110 - 2\ M\^112)\) + 
              L\^9\ \((\(-5\)\ M\^14 + 75\ M\^16 - 471\ M\^18 + 1611\ M\^20 - 
                    3145\ M\^22 + 2110\ M\^24 + 6999\ M\^26 - 20730\ M\^28 + 
                    9839\ M\^30 + 44015\ M\^32 - 67811\ M\^34 - 
                    28332\ M\^36 + 143490\ M\^38 - 66052\ M\^40 - 
                    186744\ M\^42 + 253106\ M\^44 + 184996\ M\^46 - 
                    504018\ M\^48 - 283056\ M\^50 + 836957\ M\^52 + 
                    695182\ M\^54 - 1273692\ M\^56 - 1442909\ M\^58 + 
                    1738901\ M\^60 + 2233932\ M\^62 - 2016534\ M\^64 - 
                    2642673\ M\^66 + 1994444\ M\^68 + 2436612\ M\^70 - 
                    1751699\ M\^72 - 1790999\ M\^74 + 1426562\ M\^76 + 
                    1019946\ M\^78 - 1095772\ M\^80 - 383102\ M\^82 + 
                    752816\ M\^84 + 4436\ M\^86 - 415798\ M\^88 + 
                    124655\ M\^90 + 160770\ M\^92 - 107214\ M\^94 - 
                    25228\ M\^96 + 46594\ M\^98 - 12060\ M\^100 - 
                    6294\ M\^102 + 5846\ M\^104 - 2201\ M\^106 + 
                    498\ M\^108 - 67\ M\^110 + 4\ M\^112)\) + 
              L\^3\ \((M\^52 - 9\ M\^54 + 41\ M\^56 - 115\ M\^58 + 
                    197\ M\^60 - 200\ M\^62 - 49\ M\^64 + 738\ M\^66 - 
                    1229\ M\^68 + 371\ M\^70 + 1530\ M\^72 - 2872\ M\^74 + 
                    604\ M\^76 + 5297\ M\^78 - 3618\ M\^80 - 7515\ M\^82 + 
                    6025\ M\^84 + 8559\ M\^86 - 7377\ M\^88 - 6673\ M\^90 + 
                    6871\ M\^92 + 2887\ M\^94 - 4354\ M\^96 - 25\ M\^98 + 
                    1754\ M\^100 - 718\ M\^102 - 250\ M\^104 + 321\ M\^106 - 
                    69\ M\^108 - 43\ M\^110 + 28\ M\^112 - 5\ M\^114)\) + 
              L\^8\ \((\(-M\^16\) + 19\ M\^18 - 157\ M\^20 + 711\ M\^22 - 
                    1896\ M\^24 + 2818\ M\^26 - 674\ M\^28 - 6849\ M\^30 + 
                    13339\ M\^32 - 2416\ M\^34 - 23556\ M\^36 + 
                    26205\ M\^38 + 9761\ M\^40 - 38953\ M\^42 + 
                    15375\ M\^44 + 49209\ M\^46 - 22476\ M\^48 - 
                    131434\ M\^50 - 23293\ M\^52 + 430219\ M\^54 + 
                    115567\ M\^56 - 997557\ M\^58 - 231214\ M\^60 + 
                    1673548\ M\^62 + 375390\ M\^64 - 2149673\ M\^66 - 
                    514583\ M\^68 + 2237149\ M\^70 + 559805\ M\^72 - 
                    1956734\ M\^74 - 462593\ M\^76 + 1499820\ M\^78 + 
                    242317\ M\^80 - 1048694\ M\^82 - 7151\ M\^84 + 
                    655475\ M\^86 - 138598\ M\^88 - 335001\ M\^90 + 
                    164031\ M\^92 + 115193\ M\^94 - 111729\ M\^96 - 
                    5659\ M\^98 + 42685\ M\^100 - 17176\ M\^102 - 
                    3560\ M\^104 + 5993\ M\^106 - 2700\ M\^108 + 
                    671\ M\^110 - 95\ M\^112 + 6\ M\^114)\) + 
              L\^7\ \((M\^20 - 15\ M\^22 + 105\ M\^24 - 440\ M\^26 + 
                    1193\ M\^28 - 2088\ M\^30 + 1791\ M\^32 + 1625\ M\^34 - 
                    7008\ M\^36 + 7283\ M\^38 + 2053\ M\^40 - 13023\ M\^42 + 
                    12638\ M\^44 + 2319\ M\^46 - 11133\ M\^48 + 2264\ M\^50 - 
                    24384\ M\^52 + 25998\ M\^54 + 127265\ M\^56 - 
                    98560\ M\^58 - 343855\ M\^60 + 186916\ M\^62 + 
                    703508\ M\^64 - 247596\ M\^66 - 1124443\ M\^68 + 
                    280542\ M\^70 + 1407127\ M\^72 - 299294\ M\^74 - 
                    1374863\ M\^76 + 305026\ M\^78 + 1077987\ M\^80 - 
                    295034\ M\^82 - 702286\ M\^84 + 261875\ M\^86 + 
                    384676\ M\^88 - 209533\ M\^90 - 171582\ M\^92 + 
                    147098\ M\^94 + 48592\ M\^96 - 83187\ M\^98 + 
                    7233\ M\^100 + 29044\ M\^102 - 14455\ M\^104 - 
                    1841\ M\^106 + 4457\ M\^108 - 2044\ M\^110 + 
                    489\ M\^112 - 65\ M\^114 + 4\ M\^116)\) + 
              L\^5\ \((\(-2\)\ M\^34 + 25\ M\^36 - 147\ M\^38 + 492\ M\^40 - 
                    967\ M\^42 + 978\ M\^44 + 198\ M\^46 - 2167\ M\^48 + 
                    2588\ M\^50 - 514\ M\^52 - 106\ M\^54 - 370\ M\^56 - 
                    5986\ M\^58 + 12371\ M\^60 + 1498\ M\^62 - 30797\ M\^64 + 
                    23558\ M\^66 + 54683\ M\^68 - 74800\ M\^70 - 
                    84030\ M\^72 + 142499\ M\^74 + 108147\ M\^76 - 
                    192262\ M\^78 - 110419\ M\^80 + 193032\ M\^82 + 
                    87468\ M\^84 - 142931\ M\^86 - 53447\ M\^88 + 
                    76066\ M\^90 + 31193\ M\^92 - 29614\ M\^94 - 
                    22277\ M\^96 + 12732\ M\^98 + 14070\ M\^100 - 
                    9387\ M\^102 - 4086\ M\^104 + 5103\ M\^106 - 
                    671\ M\^108 - 1087\ M\^110 + 688\ M\^112 - 193\ M\^114 + 
                    29\ M\^116 - 2\ M\^118)\) + 
              L\^6\ \((M\^26 - 15\ M\^28 + 102\ M\^30 - 405\ M\^32 + 
                    1030\ M\^34 - 1621\ M\^36 + 1061\ M\^38 + 1330\ M\^40 - 
                    4010\ M\^42 + 4614\ M\^44 - 1381\ M\^46 - 6664\ M\^48 + 
                    12263\ M\^50 - 1885\ M\^52 - 19648\ M\^54 + 
                    15734\ M\^56 + 41322\ M\^58 - 52919\ M\^60 - 
                    103485\ M\^62 + 140003\ M\^64 + 220469\ M\^66 - 
                    271792\ M\^68 - 366603\ M\^70 + 393809\ M\^72 + 
                    476589\ M\^74 - 433930\ M\^76 - 490295\ M\^78 + 
                    381464\ M\^80 + 413472\ M\^82 - 283312\ M\^84 - 
                    289308\ M\^86 + 189329\ M\^88 + 169603\ M\^90 - 
                    121628\ M\^92 - 83610\ M\^94 + 77413\ M\^96 + 
                    28773\ M\^98 - 44189\ M\^100 + 599\ M\^102 + 
                    16503\ M\^104 - 6477\ M\^106 - 1819\ M\^108 + 
                    2366\ M\^110 - 910\ M\^112 + 184\ M\^114 - 20\ M\^116 + 
                    M\^118)\) + 
              L\^4\ \((M\^42 - 11\ M\^44 + 64\ M\^46 - 236\ M\^48 + 
                    531\ M\^50 - 623\ M\^52 - 54\ M\^54 + 1570\ M\^56 - 
                    2398\ M\^58 + 108\ M\^60 + 3366\ M\^62 - 2031\ M\^64 - 
                    2828\ M\^66 + 3533\ M\^68 + 3389\ M\^70 - 10930\ M\^72 - 
                    2779\ M\^74 + 25740\ M\^76 - 307\ M\^78 - 39961\ M\^80 + 
                    3385\ M\^82 + 44005\ M\^84 - 5761\ M\^86 - 33709\ M\^88 + 
                    7352\ M\^90 + 16178\ M\^92 - 5093\ M\^94 - 2655\ M\^96 + 
                    685\ M\^98 - 1634\ M\^100 + 1871\ M\^102 + 480\ M\^104 - 
                    1267\ M\^106 + 365\ M\^108 + 225\ M\^110 - 206\ M\^112 + 
                    72\ M\^114 - 13\ M\^116 + M\^118)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 149]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 150]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 151]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 152]\), \(\(-M\^132\) + 
              2\ M\^136 - M\^140 + L\^9\ \((1 - 2\ M\^4 + M\^8)\) + 
              L\^8\ \((2\ M\^10 - 12\ M\^12 + 25\ M\^14 - 38\ M\^18 + 
                    26\ M\^20 + 8\ M\^22 - 12\ M\^24 + 3\ M\^26 - 2\ M\^28 + 
                    M\^30)\) + 
              L\^7\ \((M\^20 - 13\ M\^22 + 40\ M\^24 - 38\ M\^26 - 
                    29\ M\^28 + 83\ M\^30 - 4\ M\^32 - 78\ M\^34 + 
                    28\ M\^36 + 70\ M\^38 - 78\ M\^40 + 4\ M\^42 + 
                    39\ M\^44 - 26\ M\^46 + 8\ M\^48 - 2\ M\^50)\) + 
              L\^6\ \((\(-2\)\ M\^32 + 14\ M\^34 - 43\ M\^36 + 55\ M\^38 - 
                    20\ M\^40 - 50\ M\^42 + 104\ M\^44 - 37\ M\^46 - 
                    95\ M\^48 + 85\ M\^50 + 69\ M\^52 - 109\ M\^54 + 
                    28\ M\^56 + 16\ M\^58 - 6\ M\^60 + 14\ M\^62 - 
                    30\ M\^64 + 21\ M\^66 - 6\ M\^68 + M\^70)\) + 
              L\^5\ \((\(-M\^42\) + 12\ M\^44 - 40\ M\^46 + 41\ M\^48 + 
                    23\ M\^50 - 63\ M\^52 - 26\ M\^54 + 92\ M\^56 - 
                    34\ M\^58 - 54\ M\^60 + 106\ M\^62 - 113\ M\^64 + 
                    31\ M\^66 + 109\ M\^68 - 100\ M\^70 - 47\ M\^72 + 
                    112\ M\^74 - 24\ M\^76 - 58\ M\^78 + 55\ M\^80 - 
                    18\ M\^82 + 2\ M\^84)\) + 
              L\^4\ \((\(-2\)\ M\^56 + 18\ M\^58 - 55\ M\^60 + 58\ M\^62 + 
                    24\ M\^64 - 112\ M\^66 + 47\ M\^68 + 100\ M\^70 - 
                    109\ M\^72 - 31\ M\^74 + 113\ M\^76 - 106\ M\^78 + 
                    54\ M\^80 + 34\ M\^82 - 92\ M\^84 + 26\ M\^86 + 
                    63\ M\^88 - 23\ M\^90 - 41\ M\^92 + 40\ M\^94 - 
                    12\ M\^96 + M\^98)\) + 
              L\^3\ \((\(-M\^70\) + 6\ M\^72 - 21\ M\^74 + 30\ M\^76 - 
                    14\ M\^78 + 6\ M\^80 - 16\ M\^82 - 28\ M\^84 + 
                    109\ M\^86 - 69\ M\^88 - 85\ M\^90 + 95\ M\^92 + 
                    37\ M\^94 - 104\ M\^96 + 50\ M\^98 + 20\ M\^100 - 
                    55\ M\^102 + 43\ M\^104 - 14\ M\^106 + 2\ M\^108)\) + 
              L\^2\ \((2\ M\^90 - 8\ M\^92 + 26\ M\^94 - 39\ M\^96 - 
                    4\ M\^98 + 78\ M\^100 - 70\ M\^102 - 28\ M\^104 + 
                    78\ M\^106 + 4\ M\^108 - 83\ M\^110 + 29\ M\^112 + 
                    38\ M\^114 - 40\ M\^116 + 13\ M\^118 - M\^120)\) + 
              L\ \((\(-M\^110\) + 2\ M\^112 - 3\ M\^114 + 12\ M\^116 - 
                    8\ M\^118 - 26\ M\^120 + 38\ M\^122 - 25\ M\^126 + 
                    12\ M\^128 - 2\ M\^130)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 153]\), \(\(-M\^36\) + 
              2\ M\^40 - M\^44 + L\^12\ \((M\^10 - 2\ M\^14 + M\^18)\) + 
              L\^11\ \((1 - M\^2 - 3\ M\^6 + 11\ M\^8 - 13\ M\^10 + 
                    4\ M\^12 + 12\ M\^14 - 30\ M\^16 + 19\ M\^18 + 
                    12\ M\^20 - 16\ M\^22 + 2\ M\^26 + 2\ M\^28 - M\^30)\) + 
              L\^10\ \((2\ M\^2 + 10\ M\^4 - 28\ M\^6 - 18\ M\^8 + 
                    97\ M\^10 - 52\ M\^12 - 63\ M\^14 + 52\ M\^16 - 
                    34\ M\^18 + 57\ M\^20 + 47\ M\^22 - 96\ M\^24 - 
                    29\ M\^26 + 82\ M\^28 - 12\ M\^30 - 29\ M\^32 + 
                    19\ M\^34 - 7\ M\^36 + 3\ M\^38 - M\^40)\) + 
              L\^8\ \((17\ M\^6 + 47\ M\^8 - 193\ M\^10 - 68\ M\^12 + 
                    531\ M\^14 - 67\ M\^16 - 667\ M\^18 + 82\ M\^20 + 
                    604\ M\^22 + 110\ M\^24 - 511\ M\^26 - 299\ M\^28 + 
                    447\ M\^30 + 314\ M\^32 - 414\ M\^34 - 79\ M\^36 + 
                    197\ M\^38 - 44\ M\^40 - 7\ M\^42)\) + 
              L\^9\ \((5\ M\^4 + 39\ M\^6 - 102\ M\^8 - 74\ M\^10 + 
                    318\ M\^12 - 42\ M\^14 - 345\ M\^16 + 74\ M\^18 + 
                    179\ M\^20 + 102\ M\^22 - 63\ M\^24 - 246\ M\^26 + 
                    51\ M\^28 + 259\ M\^30 - 126\ M\^32 - 95\ M\^34 + 
                    87\ M\^36 - 14\ M\^38 - M\^40 - M\^42)\) + 
              L\^7\ \((\(-2\)\ M\^6 + 58\ M\^8 - 43\ M\^10 - 247\ M\^12 + 
                    212\ M\^14 + 459\ M\^16 - 397\ M\^18 - 591\ M\^20 + 
                    319\ M\^22 + 752\ M\^24 - 40\ M\^26 - 897\ M\^28 - 
                    172\ M\^30 + 898\ M\^32 + 127\ M\^34 - 681\ M\^36 + 
                    80\ M\^38 + 245\ M\^40 - 88\ M\^42 - 2\ M\^44)\) + 
              L\^6\ \((\(-2\)\ M\^8 + 96\ M\^10 - 183\ M\^12 - 213\ M\^14 + 
                    594\ M\^16 + 173\ M\^18 - 867\ M\^20 - 165\ M\^22 + 
                    761\ M\^24 + 416\ M\^26 - 416\ M\^28 - 761\ M\^30 + 
                    165\ M\^32 + 867\ M\^34 - 173\ M\^36 - 594\ M\^38 + 
                    213\ M\^40 + 183\ M\^42 - 96\ M\^44 + 2\ M\^46)\) + 
              L\^4\ \((7\ M\^12 + 44\ M\^14 - 197\ M\^16 + 79\ M\^18 + 
                    414\ M\^20 - 314\ M\^22 - 447\ M\^24 + 299\ M\^26 + 
                    511\ M\^28 - 110\ M\^30 - 604\ M\^32 - 82\ M\^34 + 
                    667\ M\^36 + 67\ M\^38 - 531\ M\^40 + 68\ M\^42 + 
                    193\ M\^44 - 47\ M\^46 - 17\ M\^48)\) + 
              L\^5\ \((2\ M\^10 + 88\ M\^12 - 245\ M\^14 - 80\ M\^16 + 
                    681\ M\^18 - 127\ M\^20 - 898\ M\^22 + 172\ M\^24 + 
                    897\ M\^26 + 40\ M\^28 - 752\ M\^30 - 319\ M\^32 + 
                    591\ M\^34 + 397\ M\^36 - 459\ M\^38 - 212\ M\^40 + 
                    247\ M\^42 + 43\ M\^44 - 58\ M\^46 + 2\ M\^48)\) + 
              L\^3\ \((M\^12 + M\^14 + 14\ M\^16 - 87\ M\^18 + 95\ M\^20 + 
                    126\ M\^22 - 259\ M\^24 - 51\ M\^26 + 246\ M\^28 + 
                    63\ M\^30 - 102\ M\^32 - 179\ M\^34 - 74\ M\^36 + 
                    345\ M\^38 + 42\ M\^40 - 318\ M\^42 + 74\ M\^44 + 
                    102\ M\^46 - 39\ M\^48 - 5\ M\^50)\) + 
              L\^2\ \((M\^14 - 3\ M\^16 + 7\ M\^18 - 19\ M\^20 + 29\ M\^22 + 
                    12\ M\^24 - 82\ M\^26 + 29\ M\^28 + 96\ M\^30 - 
                    47\ M\^32 - 57\ M\^34 + 34\ M\^36 - 52\ M\^38 + 
                    63\ M\^40 + 52\ M\^42 - 97\ M\^44 + 18\ M\^46 + 
                    28\ M\^48 - 10\ M\^50 - 2\ M\^52)\) + 
              L\ \((M\^24 - 2\ M\^26 - 2\ M\^28 + 16\ M\^32 - 12\ M\^34 - 
                    19\ M\^36 + 30\ M\^38 - 12\ M\^40 - 4\ M\^42 + 
                    13\ M\^44 - 11\ M\^46 + 3\ M\^48 + M\^52 - M\^54)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 154]\), \(\(-M\^2\) + 2\ M\^6 - 
              M\^10 + L\ \((\(-1\) + 2\ M\^2 + 5\ M\^4 - 6\ M\^6 - 12\ M\^8 + 
                    12\ M\^10 - 20\ M\^12 + 22\ M\^14 + 46\ M\^16 - 
                    72\ M\^18 + 2\ M\^20 + 34\ M\^22 - 17\ M\^24 + 6\ M\^26 - 
                    4\ M\^28 + 2\ M\^30)\) + 
              L\^2\ \((\(-3\)\ M\^6 + 22\ M\^8 - 59\ M\^10 + 34\ M\^12 + 
                    117\ M\^14 - 125\ M\^16 - 139\ M\^18 + 176\ M\^20 - 
                    15\ M\^24 + 129\ M\^26 - 344\ M\^28 + 145\ M\^30 + 
                    232\ M\^32 - 235\ M\^34 + 10\ M\^36 + 84\ M\^38 - 
                    20\ M\^40 - 34\ M\^42 + 28\ M\^44 - 12\ M\^46 + 
                    4\ M\^48 - M\^50)\) + 
              L\^3\ \((\(-3\)\ M\^12 + 39\ M\^14 - 177\ M\^16 + 317\ M\^18 - 
                    18\ M\^20 - 649\ M\^22 + 437\ M\^24 + 673\ M\^26 - 
                    356\ M\^28 - 763\ M\^30 - 292\ M\^32 + 1067\ M\^34 + 
                    766\ M\^36 - 1511\ M\^38 - 89\ M\^40 + 950\ M\^42 - 
                    357\ M\^44 - 192\ M\^46 + 273\ M\^48 - 120\ M\^50 - 
                    104\ M\^52 + 172\ M\^54 - 101\ M\^56 + 30\ M\^58 - 
                    4\ M\^60)\) + 
              L\^4\ \((\(-M\^18\) + 20\ M\^20 - 130\ M\^22 + 377\ M\^24 - 
                    410\ M\^26 - 267\ M\^28 + 863\ M\^30 - 115\ M\^32 - 
                    902\ M\^34 + 302\ M\^36 + 1031\ M\^38 - 481\ M\^40 - 
                    1121\ M\^42 + 479\ M\^44 + 1174\ M\^46 - 806\ M\^48 - 
                    178\ M\^50 + 446\ M\^52 - 627\ M\^54 + 144\ M\^56 + 
                    726\ M\^58 - 626\ M\^60 - 182\ M\^62 + 497\ M\^64 - 
                    275\ M\^66 + 66\ M\^68 - 6\ M\^70)\) + 
              L\^5\ \((M\^26 - 17\ M\^28 + 88\ M\^30 - 203\ M\^32 + 
                    254\ M\^34 - 243\ M\^36 + 154\ M\^38 + 518\ M\^40 - 
                    1392\ M\^42 - 90\ M\^44 + 2576\ M\^46 - 496\ M\^48 - 
                    2561\ M\^50 + 902\ M\^52 + 1276\ M\^54 - 808\ M\^56 - 
                    505\ M\^58 + 1355\ M\^60 - 247\ M\^62 - 1613\ M\^64 + 
                    748\ M\^66 + 1244\ M\^68 - 1026\ M\^70 - 276\ M\^72 + 
                    629\ M\^74 - 295\ M\^76 + 58\ M\^78 - 4\ M\^80)\) + 
              L\^6\ \((\(-3\)\ M\^38 + 4\ M\^40 + 7\ M\^42 + 48\ M\^44 - 
                    261\ M\^46 + 564\ M\^48 - 280\ M\^50 - 1177\ M\^52 + 
                    1440\ M\^54 + 1292\ M\^56 - 2450\ M\^58 - 484\ M\^60 + 
                    2878\ M\^62 - 1068\ M\^64 - 2346\ M\^66 + 1617\ M\^68 + 
                    1684\ M\^70 - 1323\ M\^72 - 925\ M\^74 + 758\ M\^76 + 
                    500\ M\^78 - 476\ M\^80 - 138\ M\^82 + 267\ M\^84 - 
                    109\ M\^86 + 18\ M\^88 - M\^90)\) + 
              L\^7\ \((9\ M\^44 - 43\ M\^46 - M\^48 + 279\ M\^50 - 
                    307\ M\^52 - 637\ M\^54 + 1151\ M\^56 + 885\ M\^58 - 
                    2211\ M\^60 - 571\ M\^62 + 2063\ M\^64 - 109\ M\^66 - 
                    910\ M\^68 + 910\ M\^70 + 109\ M\^72 - 2063\ M\^74 + 
                    571\ M\^76 + 2211\ M\^78 - 885\ M\^80 - 1151\ M\^82 + 
                    637\ M\^84 + 307\ M\^86 - 279\ M\^88 + M\^90 + 
                    43\ M\^92 - 9\ M\^94)\) + 
              L\^8\ \((M\^48 - 18\ M\^50 + 109\ M\^52 - 267\ M\^54 + 
                    138\ M\^56 + 476\ M\^58 - 500\ M\^60 - 758\ M\^62 + 
                    925\ M\^64 + 1323\ M\^66 - 1684\ M\^68 - 1617\ M\^70 + 
                    2346\ M\^72 + 1068\ M\^74 - 2878\ M\^76 + 484\ M\^78 + 
                    2450\ M\^80 - 1292\ M\^82 - 1440\ M\^84 + 1177\ M\^86 + 
                    280\ M\^88 - 564\ M\^90 + 261\ M\^92 - 48\ M\^94 - 
                    7\ M\^96 - 4\ M\^98 + 3\ M\^100)\) + 
              L\^9\ \((4\ M\^58 - 58\ M\^60 + 295\ M\^62 - 629\ M\^64 + 
                    276\ M\^66 + 1026\ M\^68 - 1244\ M\^70 - 748\ M\^72 + 
                    1613\ M\^74 + 247\ M\^76 - 1355\ M\^78 + 505\ M\^80 + 
                    808\ M\^82 - 1276\ M\^84 - 902\ M\^86 + 2561\ M\^88 + 
                    496\ M\^90 - 2576\ M\^92 + 90\ M\^94 + 1392\ M\^96 - 
                    518\ M\^98 - 154\ M\^100 + 243\ M\^102 - 254\ M\^104 + 
                    203\ M\^106 - 88\ M\^108 + 17\ M\^110 - M\^112)\) + 
              L\^10\ \((6\ M\^68 - 66\ M\^70 + 275\ M\^72 - 497\ M\^74 + 
                    182\ M\^76 + 626\ M\^78 - 726\ M\^80 - 144\ M\^82 + 
                    627\ M\^84 - 446\ M\^86 + 178\ M\^88 + 806\ M\^90 - 
                    1174\ M\^92 - 479\ M\^94 + 1121\ M\^96 + 481\ M\^98 - 
                    1031\ M\^100 - 302\ M\^102 + 902\ M\^104 + 115\ M\^106 - 
                    863\ M\^108 + 267\ M\^110 + 410\ M\^112 - 377\ M\^114 + 
                    130\ M\^116 - 20\ M\^118 + M\^120)\) + 
              L\^11\ \((4\ M\^78 - 30\ M\^80 + 101\ M\^82 - 172\ M\^84 + 
                    104\ M\^86 + 120\ M\^88 - 273\ M\^90 + 192\ M\^92 + 
                    357\ M\^94 - 950\ M\^96 + 89\ M\^98 + 1511\ M\^100 - 
                    766\ M\^102 - 1067\ M\^104 + 292\ M\^106 + 763\ M\^108 + 
                    356\ M\^110 - 673\ M\^112 - 437\ M\^114 + 649\ M\^116 + 
                    18\ M\^118 - 317\ M\^120 + 177\ M\^122 - 39\ M\^124 + 
                    3\ M\^126)\) + 
              L\^12\ \((M\^88 - 4\ M\^90 + 12\ M\^92 - 28\ M\^94 + 
                    34\ M\^96 + 20\ M\^98 - 84\ M\^100 - 10\ M\^102 + 
                    235\ M\^104 - 232\ M\^106 - 145\ M\^108 + 344\ M\^110 - 
                    129\ M\^112 + 15\ M\^114 - 176\ M\^118 + 139\ M\^120 + 
                    125\ M\^122 - 117\ M\^124 - 34\ M\^126 + 59\ M\^128 - 
                    22\ M\^130 + 3\ M\^132)\) + 
              L\^14\ \((M\^128 - 2\ M\^132 + M\^136)\) + 
              L\^13\ \((\(-2\)\ M\^108 + 4\ M\^110 - 6\ M\^112 + 17\ M\^114 - 
                    34\ M\^116 - 2\ M\^118 + 72\ M\^120 - 46\ M\^122 - 
                    22\ M\^124 + 20\ M\^126 - 12\ M\^128 + 12\ M\^130 + 
                    6\ M\^132 - 5\ M\^134 - 2\ M\^136 + M\^138)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 155]\), \(L\^11\ M\^30 + 
              M\^32 + L\^10\ \((\(-2\)\ M\^18 + 4\ M\^20 - 10\ M\^22 + 
                    15\ M\^24 - 10\ M\^26 - 11\ M\^28 + 24\ M\^30 - 
                    3\ M\^32 - 3\ M\^36 - 2\ M\^38 + M\^40)\) + 
              L\^9\ \((M\^6 - 5\ M\^8 + 18\ M\^10 - 40\ M\^12 + 53\ M\^14 - 
                    32\ M\^16 + 3\ M\^18 - 16\ M\^20 + 34\ M\^22 - 
                    45\ M\^24 + 29\ M\^26 - 33\ M\^28 + 22\ M\^30 + 
                    47\ M\^32 + 32\ M\^34 - 84\ M\^36 - 24\ M\^38 + 
                    46\ M\^40 - 5\ M\^42 - 2\ M\^44)\) + 
              L\ \((M\^22 - 2\ M\^24 - 3\ M\^26 - 3\ M\^30 + 24\ M\^32 - 
                    11\ M\^34 - 10\ M\^36 + 15\ M\^38 - 10\ M\^40 + 
                    4\ M\^42 - 2\ M\^44)\) + 
              L\^8\ \((1 - 7\ M\^2 + 25\ M\^4 - 57\ M\^6 + 62\ M\^8 + M\^10 - 
                    31\ M\^12 - 111\ M\^14 + 305\ M\^16 - 176\ M\^18 - 
                    190\ M\^20 + 96\ M\^22 + 449\ M\^24 - 547\ M\^26 - 
                    215\ M\^28 + 479\ M\^30 + 445\ M\^32 - 535\ M\^34 - 
                    248\ M\^36 + 118\ M\^38 + 193\ M\^40 - 35\ M\^42 - 
                    32\ M\^44 - 6\ M\^46 + 5\ M\^48)\) + 
              L\^7\ \((\(-M\^2\) - 9\ M\^4 + 85\ M\^6 - 196\ M\^8 + 
                    86\ M\^10 + 196\ M\^12 + 7\ M\^14 - 695\ M\^16 + 
                    630\ M\^18 + 531\ M\^20 - 621\ M\^22 - 1215\ M\^24 + 
                    1296\ M\^26 + 733\ M\^28 - 438\ M\^30 - 828\ M\^32 + 
                    500\ M\^34 - 202\ M\^36 - 52\ M\^38 + 144\ M\^40 + 
                    275\ M\^42 - 252\ M\^44 + 2\ M\^46 - 3\ M\^48 + 
                    29\ M\^50 - 8\ M\^52)\) + 
              L\^2\ \((\(-2\)\ M\^18 - 5\ M\^20 + 46\ M\^22 - 24\ M\^24 - 
                    84\ M\^26 + 32\ M\^28 + 47\ M\^30 + 22\ M\^32 - 
                    33\ M\^34 + 29\ M\^36 - 45\ M\^38 + 34\ M\^40 - 
                    16\ M\^42 + 3\ M\^44 - 32\ M\^46 + 53\ M\^48 - 
                    40\ M\^50 + 18\ M\^52 - 5\ M\^54 + M\^56)\) + 
              L\^6\ \((\(-M\^2\) + 14\ M\^4 - 70\ M\^6 + 162\ M\^8 - 
                    130\ M\^10 - 156\ M\^12 + 166\ M\^14 + 680\ M\^16 - 
                    954\ M\^18 - 386\ M\^20 + 619\ M\^22 + 839\ M\^24 - 
                    1195\ M\^26 + 555\ M\^28 + 453\ M\^30 - 531\ M\^32 - 
                    1200\ M\^34 + 1355\ M\^36 + 300\ M\^38 - 532\ M\^40 - 
                    388\ M\^42 + 685\ M\^44 - 193\ M\^46 - 153\ M\^48 + 
                    10\ M\^50 + 126\ M\^52 - 76\ M\^54 + 16\ M\^56 - 
                    M\^58)\) + 
              L\^4\ \((\(-8\)\ M\^10 + 29\ M\^12 - 3\ M\^14 + 2\ M\^16 - 
                    252\ M\^18 + 275\ M\^20 + 144\ M\^22 - 52\ M\^24 - 
                    202\ M\^26 + 500\ M\^28 - 828\ M\^30 - 438\ M\^32 + 
                    733\ M\^34 + 1296\ M\^36 - 1215\ M\^38 - 621\ M\^40 + 
                    531\ M\^42 + 630\ M\^44 - 695\ M\^46 + 7\ M\^48 + 
                    196\ M\^50 + 86\ M\^52 - 196\ M\^54 + 85\ M\^56 - 
                    9\ M\^58 - M\^60)\) + 
              L\^5\ \((\(-M\^4\) + 16\ M\^6 - 76\ M\^8 + 126\ M\^10 + 
                    10\ M\^12 - 153\ M\^14 - 193\ M\^16 + 685\ M\^18 - 
                    388\ M\^20 - 532\ M\^22 + 300\ M\^24 + 1355\ M\^26 - 
                    1200\ M\^28 - 531\ M\^30 + 453\ M\^32 + 555\ M\^34 - 
                    1195\ M\^36 + 839\ M\^38 + 619\ M\^40 - 386\ M\^42 - 
                    954\ M\^44 + 680\ M\^46 + 166\ M\^48 - 156\ M\^50 - 
                    130\ M\^52 + 162\ M\^54 - 70\ M\^56 + 14\ M\^58 - 
                    M\^60)\) + 
              L\^3\ \((5\ M\^14 - 6\ M\^16 - 32\ M\^18 - 35\ M\^20 + 
                    193\ M\^22 + 118\ M\^24 - 248\ M\^26 - 535\ M\^28 + 
                    445\ M\^30 + 479\ M\^32 - 215\ M\^34 - 547\ M\^36 + 
                    449\ M\^38 + 96\ M\^40 - 190\ M\^42 - 176\ M\^44 + 
                    305\ M\^46 - 111\ M\^48 - 31\ M\^50 + M\^52 + 62\ M\^54 - 
                    57\ M\^56 + 25\ M\^58 - 7\ M\^60 + M\^62)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 156]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 157]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 158]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 159]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 160]\), \(\(-M\^8\) + 
              L\^17\ M\^144 + 
              L\ \((2\ M\^4 - 4\ M\^6 + 6\ M\^8 - 19\ M\^10 + 16\ M\^12 + 
                    7\ M\^14 - 16\ M\^16 + 5\ M\^18 - 4\ M\^20 + M\^24 - 
                    M\^26)\) + 
              L\^2\ \((\(-1\) + 5\ M\^2 - 14\ M\^4 + 36\ M\^6 - 57\ M\^8 + 
                    43\ M\^10 - 57\ M\^12 + 115\ M\^14 - 75\ M\^16 - M\^18 - 
                    67\ M\^20 + 69\ M\^22 + 33\ M\^24 - 2\ M\^26 - 
                    70\ M\^28 + 28\ M\^30 + 19\ M\^32 - 26\ M\^34 + 
                    7\ M\^36 - M\^38)\) + 
              L\^3\ \((2\ M\^4 - 15\ M\^6 + 61\ M\^8 - 148\ M\^10 + 
                    190\ M\^12 - 161\ M\^14 + 281\ M\^16 - 427\ M\^18 + 
                    241\ M\^20 - 250\ M\^22 + 521\ M\^24 - 328\ M\^26 + 
                    244\ M\^28 - 564\ M\^30 + 367\ M\^32 + 65\ M\^34 + 
                    46\ M\^36 - 242\ M\^38 + 172\ M\^40 - 43\ M\^42 - 
                    13\ M\^44 + 2\ M\^46 - 2\ M\^48 + M\^50)\) + 
              L\^4\ \((\(-M\^8\) + 10\ M\^10 - 53\ M\^12 + 161\ M\^14 - 
                    296\ M\^16 + 401\ M\^18 - 534\ M\^20 + 636\ M\^22 - 
                    692\ M\^24 + 1136\ M\^26 - 1514\ M\^28 + 955\ M\^30 - 
                    775\ M\^32 + 1579\ M\^34 - 1788\ M\^36 + 1291\ M\^38 - 
                    614\ M\^40 + 568\ M\^42 - 1210\ M\^44 + 1140\ M\^46 - 
                    354\ M\^48 + 94\ M\^50 - 173\ M\^52 + 198\ M\^54 - 
                    159\ M\^56 + 66\ M\^58 - 13\ M\^60 + M\^62)\) + 
              L\^5\ \((M\^16 - 11\ M\^18 + 81\ M\^20 - 349\ M\^22 + 
                    810\ M\^24 - 920\ M\^26 + 518\ M\^28 - 975\ M\^30 + 
                    2408\ M\^32 - 2380\ M\^34 + 1741\ M\^36 - 3103\ M\^38 + 
                    3580\ M\^40 - 1735\ M\^42 + 2311\ M\^44 - 4914\ M\^46 + 
                    3925\ M\^48 - 779\ M\^50 + 1060\ M\^52 - 2924\ M\^54 + 
                    2675\ M\^56 - 1364\ M\^58 + 527\ M\^60 - 244\ M\^62 + 
                    459\ M\^64 - 501\ M\^66 + 229\ M\^68 - 45\ M\^70 + 
                    3\ M\^72)\) + 
              L\^6\ \((\(-28\)\ M\^28 + 274\ M\^30 - 1045\ M\^32 + 
                    1863\ M\^34 - 1336\ M\^36 + 397\ M\^38 - 2201\ M\^40 + 
                    5115\ M\^42 - 4691\ M\^44 + 2892\ M\^46 - 3299\ M\^48 + 
                    5181\ M\^50 - 6734\ M\^52 + 6022\ M\^54 - 4153\ M\^56 + 
                    3970\ M\^58 - 4889\ M\^60 + 4453\ M\^62 - 3091\ M\^64 + 
                    2812\ M\^66 - 2732\ M\^68 + 1585\ M\^70 - 870\ M\^72 + 
                    1022\ M\^74 - 796\ M\^76 + 293\ M\^78 - 49\ M\^80 + 
                    3\ M\^82)\) + 
              L\^7\ \((\(-2\)\ M\^34 + 37\ M\^36 - 292\ M\^38 + 1091\ M\^40 - 
                    1969\ M\^42 + 1499\ M\^44 - 819\ M\^46 + 3370\ M\^48 - 
                    5680\ M\^50 + 3410\ M\^52 - 4696\ M\^54 + 8856\ M\^56 - 
                    3900\ M\^58 + 2111\ M\^60 - 12899\ M\^62 + 12928\ M\^64 - 
                    698\ M\^66 + 283\ M\^68 - 9266\ M\^70 + 10075\ M\^72 - 
                    4418\ M\^74 + 1359\ M\^76 - 1583\ M\^78 + 2320\ M\^80 - 
                    2501\ M\^82 + 1847\ M\^84 - 806\ M\^86 + 187\ M\^88 - 
                    21\ M\^90 + M\^92)\) + 
              L\^8\ \((M\^42 - 23\ M\^44 + 238\ M\^46 - 1156\ M\^48 + 
                    2694\ M\^50 - 2385\ M\^52 - 864\ M\^54 + 12\ M\^56 + 
                    8738\ M\^58 - 10706\ M\^60 + 734\ M\^62 - 1290\ M\^64 + 
                    11825\ M\^66 - 8506\ M\^68 + 2701\ M\^70 - 10826\ M\^72 + 
                    14832\ M\^74 - 5123\ M\^76 - 1701\ M\^78 - 5015\ M\^80 + 
                    12847\ M\^82 - 8905\ M\^84 + 296\ M\^86 + 1176\ M\^88 + 
                    2333\ M\^90 - 3550\ M\^92 + 2005\ M\^94 - 565\ M\^96 + 
                    77\ M\^98 - 4\ M\^100)\) + 
              L\^9\ \((4\ M\^52 - 77\ M\^54 + 565\ M\^56 - 2005\ M\^58 + 
                    3550\ M\^60 - 2333\ M\^62 - 1176\ M\^64 - 296\ M\^66 + 
                    8905\ M\^68 - 12847\ M\^70 + 5015\ M\^72 + 1701\ M\^74 + 
                    5123\ M\^76 - 14832\ M\^78 + 10826\ M\^80 - 2701\ M\^82 + 
                    8506\ M\^84 - 11825\ M\^86 + 1290\ M\^88 - 734\ M\^90 + 
                    10706\ M\^92 - 8738\ M\^94 - 12\ M\^96 + 864\ M\^98 + 
                    2385\ M\^100 - 2694\ M\^102 + 1156\ M\^104 - 
                    238\ M\^106 + 23\ M\^108 - M\^110)\) + 
              L\^10\ \((\(-M\^60\) + 21\ M\^62 - 187\ M\^64 + 806\ M\^66 - 
                    1847\ M\^68 + 2501\ M\^70 - 2320\ M\^72 + 1583\ M\^74 - 
                    1359\ M\^76 + 4418\ M\^78 - 10075\ M\^80 + 9266\ M\^82 - 
                    283\ M\^84 + 698\ M\^86 - 12928\ M\^88 + 12899\ M\^90 - 
                    2111\ M\^92 + 3900\ M\^94 - 8856\ M\^96 + 4696\ M\^98 - 
                    3410\ M\^100 + 5680\ M\^102 - 3370\ M\^104 + 
                    819\ M\^106 - 1499\ M\^108 + 1969\ M\^110 - 
                    1091\ M\^112 + 292\ M\^114 - 37\ M\^116 + 2\ M\^118)\) + 
              L\^11\ \((\(-3\)\ M\^70 + 49\ M\^72 - 293\ M\^74 + 796\ M\^76 - 
                    1022\ M\^78 + 870\ M\^80 - 1585\ M\^82 + 2732\ M\^84 - 
                    2812\ M\^86 + 3091\ M\^88 - 4453\ M\^90 + 4889\ M\^92 - 
                    3970\ M\^94 + 4153\ M\^96 - 6022\ M\^98 + 6734\ M\^100 - 
                    5181\ M\^102 + 3299\ M\^104 - 2892\ M\^106 + 
                    4691\ M\^108 - 5115\ M\^110 + 2201\ M\^112 - 
                    397\ M\^114 + 1336\ M\^116 - 1863\ M\^118 + 
                    1045\ M\^120 - 274\ M\^122 + 28\ M\^124)\) + 
              L\^12\ \((\(-3\)\ M\^80 + 45\ M\^82 - 229\ M\^84 + 501\ M\^86 - 
                    459\ M\^88 + 244\ M\^90 - 527\ M\^92 + 1364\ M\^94 - 
                    2675\ M\^96 + 2924\ M\^98 - 1060\ M\^100 + 779\ M\^102 - 
                    3925\ M\^104 + 4914\ M\^106 - 2311\ M\^108 + 
                    1735\ M\^110 - 3580\ M\^112 + 3103\ M\^114 - 
                    1741\ M\^116 + 2380\ M\^118 - 2408\ M\^120 + 
                    975\ M\^122 - 518\ M\^124 + 920\ M\^126 - 810\ M\^128 + 
                    349\ M\^130 - 81\ M\^132 + 11\ M\^134 - M\^136)\) + 
              L\^13\ \((\(-M\^90\) + 13\ M\^92 - 66\ M\^94 + 159\ M\^96 - 
                    198\ M\^98 + 173\ M\^100 - 94\ M\^102 + 354\ M\^104 - 
                    1140\ M\^106 + 1210\ M\^108 - 568\ M\^110 + 614\ M\^112 - 
                    1291\ M\^114 + 1788\ M\^116 - 1579\ M\^118 + 
                    775\ M\^120 - 955\ M\^122 + 1514\ M\^124 - 1136\ M\^126 + 
                    692\ M\^128 - 636\ M\^130 + 534\ M\^132 - 401\ M\^134 + 
                    296\ M\^136 - 161\ M\^138 + 53\ M\^140 - 10\ M\^142 + 
                    M\^144)\) + 
              L\^16\ \((M\^126 - M\^128 + 4\ M\^132 - 5\ M\^134 + 
                    16\ M\^136 - 7\ M\^138 - 16\ M\^140 + 19\ M\^142 - 
                    6\ M\^144 + 4\ M\^146 - 2\ M\^148)\) + 
              L\^14\ \((\(-M\^102\) + 2\ M\^104 - 2\ M\^106 + 13\ M\^108 + 
                    43\ M\^110 - 172\ M\^112 + 242\ M\^114 - 46\ M\^116 - 
                    65\ M\^118 - 367\ M\^120 + 564\ M\^122 - 244\ M\^124 + 
                    328\ M\^126 - 521\ M\^128 + 250\ M\^130 - 241\ M\^132 + 
                    427\ M\^134 - 281\ M\^136 + 161\ M\^138 - 190\ M\^140 + 
                    148\ M\^142 - 61\ M\^144 + 15\ M\^146 - 2\ M\^148)\) + 
              L\^15\ \((M\^114 - 7\ M\^116 + 26\ M\^118 - 19\ M\^120 - 
                    28\ M\^122 + 70\ M\^124 + 2\ M\^126 - 33\ M\^128 - 
                    69\ M\^130 + 67\ M\^132 + M\^134 + 75\ M\^136 - 
                    115\ M\^138 + 57\ M\^140 - 43\ M\^142 + 57\ M\^144 - 
                    36\ M\^146 + 14\ M\^148 - 5\ M\^150 + M\^152)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 161]\), \(L\^10\ M\^2 + M\^98 + 
              L\^9\ \((1 - M\^2 - 2\ M\^4 - M\^6 + M\^10 + 10\ M\^12 - 
                    6\ M\^14 + 2\ M\^16 - 2\ M\^18 + M\^20 - M\^22)\) + 
              L\^8\ \((\(-3\)\ M\^8 + 15\ M\^10 - 11\ M\^12 - 3\ M\^14 - 
                    17\ M\^16 + 6\ M\^18 - 6\ M\^20 + 40\ M\^22 - 20\ M\^24 + 
                    2\ M\^26 - 5\ M\^28 + 8\ M\^30 - 16\ M\^32 + 11\ M\^34 - 
                    5\ M\^36 + 2\ M\^38 - M\^40)\) + 
              L\^7\ \((3\ M\^16 - 27\ M\^18 + 59\ M\^20 - 10\ M\^22 - 
                    14\ M\^24 - 46\ M\^26 + M\^28 - 24\ M\^30 + 91\ M\^32 - 
                    43\ M\^34 + 15\ M\^36 - 4\ M\^38 + 14\ M\^40 - 
                    41\ M\^42 + 26\ M\^44 - 9\ M\^46 + M\^48)\) + 
              L\^6\ \((\(-M\^24\) + 15\ M\^26 - 64\ M\^28 + 71\ M\^30 + 
                    27\ M\^32 + 9\ M\^34 - 68\ M\^36 + 9\ M\^38 - 
                    103\ M\^40 + 111\ M\^42 - 19\ M\^44 + 31\ M\^46 - 
                    2\ M\^48 + 20\ M\^50 - 64\ M\^52 + 42\ M\^54 - 
                    14\ M\^56 + 2\ M\^58)\) + 
              L\^5\ \((\(-4\)\ M\^34 + 32\ M\^36 - 84\ M\^38 + 51\ M\^40 + 
                    13\ M\^42 + 54\ M\^44 - 19\ M\^46 + 47\ M\^48 - 
                    168\ M\^50 + 47\ M\^52 - 19\ M\^54 + 54\ M\^56 + 
                    13\ M\^58 + 51\ M\^60 - 84\ M\^62 + 32\ M\^64 - 
                    4\ M\^66)\) + 
              L\^4\ \((2\ M\^42 - 14\ M\^44 + 42\ M\^46 - 64\ M\^48 + 
                    20\ M\^50 - 2\ M\^52 + 31\ M\^54 - 19\ M\^56 + 
                    111\ M\^58 - 103\ M\^60 + 9\ M\^62 - 68\ M\^64 + 
                    9\ M\^66 + 27\ M\^68 + 71\ M\^70 - 64\ M\^72 + 
                    15\ M\^74 - M\^76)\) + 
              L\^3\ \((M\^52 - 9\ M\^54 + 26\ M\^56 - 41\ M\^58 + 14\ M\^60 - 
                    4\ M\^62 + 15\ M\^64 - 43\ M\^66 + 91\ M\^68 - 
                    24\ M\^70 + M\^72 - 46\ M\^74 - 14\ M\^76 - 10\ M\^78 + 
                    59\ M\^80 - 27\ M\^82 + 3\ M\^84)\) + 
              L\^2\ \((\(-M\^60\) + 2\ M\^62 - 5\ M\^64 + 11\ M\^66 - 
                    16\ M\^68 + 8\ M\^70 - 5\ M\^72 + 2\ M\^74 - 20\ M\^76 + 
                    40\ M\^78 - 6\ M\^80 + 6\ M\^82 - 17\ M\^84 - 3\ M\^86 - 
                    11\ M\^88 + 15\ M\^90 - 3\ M\^92)\) + 
              L\ \((\(-M\^78\) + M\^80 - 2\ M\^82 + 2\ M\^84 - 6\ M\^86 + 
                    10\ M\^88 + M\^90 - M\^94 - 2\ M\^96 - M\^98 + 
                    M\^100)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 162]\), \(M\^2 + L\^10\ M\^98 + 
              L\ \((1 - M\^2 - 2\ M\^4 - M\^6 + M\^10 + 10\ M\^12 - 
                    6\ M\^14 + 2\ M\^16 - 2\ M\^18 + M\^20 - M\^22)\) + 
              L\^2\ \((\(-3\)\ M\^8 + 15\ M\^10 - 11\ M\^12 - 3\ M\^14 - 
                    17\ M\^16 + 6\ M\^18 - 6\ M\^20 + 40\ M\^22 - 20\ M\^24 + 
                    2\ M\^26 - 5\ M\^28 + 8\ M\^30 - 16\ M\^32 + 11\ M\^34 - 
                    5\ M\^36 + 2\ M\^38 - M\^40)\) + 
              L\^3\ \((3\ M\^16 - 27\ M\^18 + 59\ M\^20 - 10\ M\^22 - 
                    14\ M\^24 - 46\ M\^26 + M\^28 - 24\ M\^30 + 91\ M\^32 - 
                    43\ M\^34 + 15\ M\^36 - 4\ M\^38 + 14\ M\^40 - 
                    41\ M\^42 + 26\ M\^44 - 9\ M\^46 + M\^48)\) + 
              L\^4\ \((\(-M\^24\) + 15\ M\^26 - 64\ M\^28 + 71\ M\^30 + 
                    27\ M\^32 + 9\ M\^34 - 68\ M\^36 + 9\ M\^38 - 
                    103\ M\^40 + 111\ M\^42 - 19\ M\^44 + 31\ M\^46 - 
                    2\ M\^48 + 20\ M\^50 - 64\ M\^52 + 42\ M\^54 - 
                    14\ M\^56 + 2\ M\^58)\) + 
              L\^5\ \((\(-4\)\ M\^34 + 32\ M\^36 - 84\ M\^38 + 51\ M\^40 + 
                    13\ M\^42 + 54\ M\^44 - 19\ M\^46 + 47\ M\^48 - 
                    168\ M\^50 + 47\ M\^52 - 19\ M\^54 + 54\ M\^56 + 
                    13\ M\^58 + 51\ M\^60 - 84\ M\^62 + 32\ M\^64 - 
                    4\ M\^66)\) + 
              L\^6\ \((2\ M\^42 - 14\ M\^44 + 42\ M\^46 - 64\ M\^48 + 
                    20\ M\^50 - 2\ M\^52 + 31\ M\^54 - 19\ M\^56 + 
                    111\ M\^58 - 103\ M\^60 + 9\ M\^62 - 68\ M\^64 + 
                    9\ M\^66 + 27\ M\^68 + 71\ M\^70 - 64\ M\^72 + 
                    15\ M\^74 - M\^76)\) + 
              L\^7\ \((M\^52 - 9\ M\^54 + 26\ M\^56 - 41\ M\^58 + 14\ M\^60 - 
                    4\ M\^62 + 15\ M\^64 - 43\ M\^66 + 91\ M\^68 - 
                    24\ M\^70 + M\^72 - 46\ M\^74 - 14\ M\^76 - 10\ M\^78 + 
                    59\ M\^80 - 27\ M\^82 + 3\ M\^84)\) + 
              L\^8\ \((\(-M\^60\) + 2\ M\^62 - 5\ M\^64 + 11\ M\^66 - 
                    16\ M\^68 + 8\ M\^70 - 5\ M\^72 + 2\ M\^74 - 20\ M\^76 + 
                    40\ M\^78 - 6\ M\^80 + 6\ M\^82 - 17\ M\^84 - 3\ M\^86 - 
                    11\ M\^88 + 15\ M\^90 - 3\ M\^92)\) + 
              L\^9\ \((\(-M\^78\) + M\^80 - 2\ M\^82 + 2\ M\^84 - 6\ M\^86 + 
                    10\ M\^88 + M\^90 - M\^94 - 2\ M\^96 - M\^98 + 
                    M\^100)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 163]\), \(L\^18\ M\^12 - 
              M\^98 + L\^17\ \((\(-3\)\ M\^8 + 6\ M\^10 - 29\ M\^14 + 
                    52\ M\^16 - 2\ M\^18 - 22\ M\^20 + 10\ M\^22 - 
                    4\ M\^24)\) + 
              L\^16\ \((\(-M\^2\) + 9\ M\^4 - 39\ M\^6 + 101\ M\^8 - 
                    131\ M\^10 - 20\ M\^12 + 316\ M\^14 - 337\ M\^16 - 
                    300\ M\^18 + 712\ M\^20 + 31\ M\^22 - 499\ M\^24 + 
                    187\ M\^26 + 67\ M\^28 - 150\ M\^30 + 101\ M\^32 - 
                    30\ M\^34 + 6\ M\^36)\) + 
              L\^15\ \((\(-1\) + 13\ M\^2 - 78\ M\^4 + 283\ M\^6 - 
                    599\ M\^8 + 532\ M\^10 + 521\ M\^12 - 1679\ M\^14 + 
                    627\ M\^16 + 2602\ M\^18 - 3327\ M\^20 - 1485\ M\^22 + 
                    4051\ M\^24 + 1062\ M\^26 - 3550\ M\^28 + 480\ M\^30 + 
                    1166\ M\^32 - 1009\ M\^34 + 172\ M\^36 + 634\ M\^38 - 
                    653\ M\^40 + 365\ M\^42 - 137\ M\^44 + 30\ M\^46 - 
                    4\ M\^48)\) + 
              L\^14\ \((3 - 39\ M\^2 + 229\ M\^4 - 773\ M\^6 + 1462\ M\^8 - 
                    1034\ M\^10 - 1373\ M\^12 + 2926\ M\^14 + 669\ M\^16 - 
                    6126\ M\^18 + 3730\ M\^20 + 6140\ M\^22 - 6860\ M\^24 - 
                    8130\ M\^26 + 12190\ M\^28 + 3967\ M\^30 - 7421\ M\^32 - 
                    1730\ M\^34 + 2824\ M\^36 - 2948\ M\^38 + 1934\ M\^40 + 
                    1727\ M\^42 - 1415\ M\^44 - 857\ M\^46 + 1712\ M\^48 - 
                    1352\ M\^50 + 664\ M\^52 - 229\ M\^54 + 59\ M\^56 - 
                    10\ M\^58 + M\^60)\) + 
              L\^13\ \((\(-3\) + 43\ M\^2 - 274\ M\^4 + 983\ M\^6 - 
                    2024\ M\^8 + 2060\ M\^10 - 61\ M\^12 - 1724\ M\^14 + 
                    43\ M\^16 + 3030\ M\^18 - 2786\ M\^20 + 735\ M\^22 - 
                    718\ M\^24 + 2448\ M\^26 - 2605\ M\^28 + 674\ M\^30 - 
                    4764\ M\^32 + 4417\ M\^34 + 5703\ M\^36 + 6225\ M\^38 - 
                    16682\ M\^40 - 7331\ M\^42 + 11433\ M\^44 + 9429\ M\^46 - 
                    8889\ M\^48 - 1381\ M\^50 + 1929\ M\^52 + 1391\ M\^54 - 
                    3098\ M\^56 + 2646\ M\^58 - 1386\ M\^60 + 495\ M\^62 - 
                    117\ M\^64 + 16\ M\^66 - M\^68)\) + 
              L\^12\ \((1 - 16\ M\^2 + 115\ M\^4 - 484\ M\^6 + 1326\ M\^8 - 
                    2536\ M\^10 + 3354\ M\^12 - 2141\ M\^14 - 1560\ M\^16 + 
                    2474\ M\^18 + 5017\ M\^20 - 11432\ M\^22 + 937\ M\^24 + 
                    12644\ M\^26 + 4002\ M\^28 - 30234\ M\^30 + 7998\ M\^32 + 
                    38238\ M\^34 - 16049\ M\^36 - 62770\ M\^38 + 
                    46967\ M\^40 + 58120\ M\^42 - 38339\ M\^44 - 
                    62915\ M\^46 + 30162\ M\^48 + 38249\ M\^50 - 
                    7600\ M\^52 - 23630\ M\^54 + 8867\ M\^56 + 3066\ M\^58 - 
                    1270\ M\^60 - 3327\ M\^62 + 5335\ M\^64 - 3952\ M\^66 + 
                    1732\ M\^68 - 469\ M\^70 + 73\ M\^72 - 5\ M\^74)\) + 
              L\^11\ \((\(-M\^4\) + 21\ M\^6 - 188\ M\^8 + 915\ M\^10 - 
                    2584\ M\^12 + 4211\ M\^14 - 3638\ M\^16 + 1174\ M\^18 + 
                    405\ M\^20 + 78\ M\^22 - 4997\ M\^24 + 15206\ M\^26 - 
                    17570\ M\^28 - 3839\ M\^30 + 29641\ M\^32 - 9365\ M\^34 - 
                    46996\ M\^36 + 55879\ M\^38 + 18868\ M\^40 - 
                    78426\ M\^42 - 23256\ M\^44 + 126345\ M\^46 + 
                    17994\ M\^48 - 113901\ M\^50 - 50888\ M\^52 + 
                    99731\ M\^54 + 33748\ M\^56 - 49675\ M\^58 - 
                    20059\ M\^60 + 29842\ M\^62 - 7870\ M\^64 - 3187\ M\^66 + 
                    1527\ M\^68 + 5152\ M\^70 - 7225\ M\^72 + 4319\ M\^74 - 
                    1425\ M\^76 + 267\ M\^78 - 26\ M\^80 + M\^82)\) + 
              L\^10\ \((\(-4\)\ M\^10 + 74\ M\^12 - 567\ M\^14 + 
                    2303\ M\^16 - 5219\ M\^18 + 6018\ M\^20 - 1625\ M\^22 - 
                    2831\ M\^24 - 888\ M\^26 + 13007\ M\^28 - 25059\ M\^30 + 
                    15429\ M\^32 + 21376\ M\^34 - 36767\ M\^36 - 
                    9831\ M\^38 + 75667\ M\^40 - 48117\ M\^42 - 
                    54858\ M\^44 + 38771\ M\^46 + 33177\ M\^48 - 
                    19494\ M\^50 + 54897\ M\^52 - 40321\ M\^54 - 
                    77917\ M\^56 + 19278\ M\^58 + 88642\ M\^60 - 
                    17749\ M\^62 - 40217\ M\^64 + 835\ M\^66 + 23037\ M\^68 - 
                    16365\ M\^70 + 4625\ M\^72 + 2204\ M\^74 + 2205\ M\^76 - 
                    7740\ M\^78 + 6430\ M\^80 - 2661\ M\^82 + 613\ M\^84 - 
                    76\ M\^86 + 4\ M\^88)\) + 
              L\^9\ \((\(-6\)\ M\^16 + 108\ M\^18 - 799\ M\^20 + 
                    3101\ M\^22 - 6504\ M\^24 + 6326\ M\^26 - 673\ M\^28 - 
                    342\ M\^30 - 8372\ M\^32 + 15626\ M\^34 - 16958\ M\^36 + 
                    2279\ M\^38 + 23184\ M\^40 - 4947\ M\^42 - 36730\ M\^44 + 
                    49932\ M\^46 - 29449\ M\^48 - 72779\ M\^50 + 
                    77276\ M\^52 + 126431\ M\^54 - 126431\ M\^56 - 
                    77276\ M\^58 + 72779\ M\^60 + 29449\ M\^62 - 
                    49932\ M\^64 + 36730\ M\^66 + 4947\ M\^68 - 
                    23184\ M\^70 - 2279\ M\^72 + 16958\ M\^74 - 
                    15626\ M\^76 + 8372\ M\^78 + 342\ M\^80 + 673\ M\^82 - 
                    6326\ M\^84 + 6504\ M\^86 - 3101\ M\^88 + 799\ M\^90 - 
                    108\ M\^92 + 6\ M\^94)\) + 
              L\^8\ \((\(-4\)\ M\^22 + 76\ M\^24 - 613\ M\^26 + 2661\ M\^28 - 
                    6430\ M\^30 + 7740\ M\^32 - 2205\ M\^34 - 2204\ M\^36 - 
                    4625\ M\^38 + 16365\ M\^40 - 23037\ M\^42 - 835\ M\^44 + 
                    40217\ M\^46 + 17749\ M\^48 - 88642\ M\^50 - 
                    19278\ M\^52 + 77917\ M\^54 + 40321\ M\^56 - 54897\ M\^58 
                    + 19494\ M\^60 - 33177\ M\^62 - 38771\ M\^64 + 
                    54858\ M\^66 + 48117\ M\^68 - 75667\ M\^70 + 
                    9831\ M\^72 + 36767\ M\^74 - 21376\ M\^76 - 
                    15429\ M\^78 + 25059\ M\^80 - 13007\ M\^82 + 888\ M\^84 + 
                    2831\ M\^86 + 1625\ M\^88 - 6018\ M\^90 + 5219\ M\^92 - 
                    2303\ M\^94 + 567\ M\^96 - 74\ M\^98 + 4\ M\^100)\) + 
              L\ \((4\ M\^86 - 10\ M\^88 + 22\ M\^90 + 2\ M\^92 - 52\ M\^94 + 
                    29\ M\^96 - 6\ M\^100 + 3\ M\^102)\) + 
              L\^7\ \((\(-M\^28\) + 26\ M\^30 - 267\ M\^32 + 1425\ M\^34 - 
                    4319\ M\^36 + 7225\ M\^38 - 5152\ M\^40 - 1527\ M\^42 + 
                    3187\ M\^44 + 7870\ M\^46 - 29842\ M\^48 + 20059\ M\^50 + 
                    49675\ M\^52 - 33748\ M\^54 - 99731\ M\^56 + 
                    50888\ M\^58 + 113901\ M\^60 - 17994\ M\^62 - 
                    126345\ M\^64 + 23256\ M\^66 + 78426\ M\^68 - 
                    18868\ M\^70 - 55879\ M\^72 + 46996\ M\^74 + 
                    9365\ M\^76 - 29641\ M\^78 + 3839\ M\^80 + 17570\ M\^82 - 
                    15206\ M\^84 + 4997\ M\^86 - 78\ M\^88 - 405\ M\^90 - 
                    1174\ M\^92 + 3638\ M\^94 - 4211\ M\^96 + 2584\ M\^98 - 
                    915\ M\^100 + 188\ M\^102 - 21\ M\^104 + M\^106)\) + 
              L\^2\ \((\(-6\)\ M\^74 + 30\ M\^76 - 101\ M\^78 + 150\ M\^80 - 
                    67\ M\^82 - 187\ M\^84 + 499\ M\^86 - 31\ M\^88 - 
                    712\ M\^90 + 300\ M\^92 + 337\ M\^94 - 316\ M\^96 + 
                    20\ M\^98 + 131\ M\^100 - 101\ M\^102 + 39\ M\^104 - 
                    9\ M\^106 + M\^108)\) + 
              L\^4\ \((\(-M\^50\) + 10\ M\^52 - 59\ M\^54 + 229\ M\^56 - 
                    664\ M\^58 + 1352\ M\^60 - 1712\ M\^62 + 857\ M\^64 + 
                    1415\ M\^66 - 1727\ M\^68 - 1934\ M\^70 + 2948\ M\^72 - 
                    2824\ M\^74 + 1730\ M\^76 + 7421\ M\^78 - 3967\ M\^80 - 
                    12190\ M\^82 + 8130\ M\^84 + 6860\ M\^86 - 6140\ M\^88 - 
                    3730\ M\^90 + 6126\ M\^92 - 669\ M\^94 - 2926\ M\^96 + 
                    1373\ M\^98 + 1034\ M\^100 - 1462\ M\^102 + 773\ M\^104 - 
                    229\ M\^106 + 39\ M\^108 - 3\ M\^110)\) + 
              L\^6\ \((5\ M\^36 - 73\ M\^38 + 469\ M\^40 - 1732\ M\^42 + 
                    3952\ M\^44 - 5335\ M\^46 + 3327\ M\^48 + 1270\ M\^50 - 
                    3066\ M\^52 - 8867\ M\^54 + 23630\ M\^56 + 7600\ M\^58 - 
                    38249\ M\^60 - 30162\ M\^62 + 62915\ M\^64 + 
                    38339\ M\^66 - 58120\ M\^68 - 46967\ M\^70 + 
                    62770\ M\^72 + 16049\ M\^74 - 38238\ M\^76 - 
                    7998\ M\^78 + 30234\ M\^80 - 4002\ M\^82 - 12644\ M\^84 - 
                    937\ M\^86 + 11432\ M\^88 - 5017\ M\^90 - 2474\ M\^92 + 
                    1560\ M\^94 + 2141\ M\^96 - 3354\ M\^98 + 2536\ M\^100 - 
                    1326\ M\^102 + 484\ M\^104 - 115\ M\^106 + 16\ M\^108 - 
                    M\^110)\) + 
              L\^3\ \((4\ M\^62 - 30\ M\^64 + 137\ M\^66 - 365\ M\^68 + 
                    653\ M\^70 - 634\ M\^72 - 172\ M\^74 + 1009\ M\^76 - 
                    1166\ M\^78 - 480\ M\^80 + 3550\ M\^82 - 1062\ M\^84 - 
                    4051\ M\^86 + 1485\ M\^88 + 3327\ M\^90 - 2602\ M\^92 - 
                    627\ M\^94 + 1679\ M\^96 - 521\ M\^98 - 532\ M\^100 + 
                    599\ M\^102 - 283\ M\^104 + 78\ M\^106 - 13\ M\^108 + 
                    M\^110)\) + 
              L\^5\ \((M\^42 - 16\ M\^44 + 117\ M\^46 - 495\ M\^48 + 
                    1386\ M\^50 - 2646\ M\^52 + 3098\ M\^54 - 1391\ M\^56 - 
                    1929\ M\^58 + 1381\ M\^60 + 8889\ M\^62 - 9429\ M\^64 - 
                    11433\ M\^66 + 7331\ M\^68 + 16682\ M\^70 - 6225\ M\^72 - 
                    5703\ M\^74 - 4417\ M\^76 + 4764\ M\^78 - 674\ M\^80 + 
                    2605\ M\^82 - 2448\ M\^84 + 718\ M\^86 - 735\ M\^88 + 
                    2786\ M\^90 - 3030\ M\^92 - 43\ M\^94 + 1724\ M\^96 + 
                    61\ M\^98 - 2060\ M\^100 + 2024\ M\^102 - 983\ M\^104 + 
                    274\ M\^106 - 43\ M\^108 + 3\ M\^110)\)\)},
          {"\<\"A-polynomial\"\>", \(Knot[10, 164]\), "$Failed"},
          {"\<\"A-polynomial\"\>", \(Knot[10, 165]\), "$Failed"}
          },
        RowSpacings->1,
        ColumnSpacings->3,
        RowAlignments->Baseline,
        ColumnAlignments->{Left}],
      Function[ BoxForm`e$, 
        TableForm[ BoxForm`e$]]]], "Output"]
}, Open  ]],

Cell["\<\
Okay, that's all very well. What can we do with it? Best of all is to write \
the data into the KnotAtlas. This again requires a few steps, but shouldn't \
be too hard. First, we should go to the KnotAtlas wiki, and tell it a little \
about the new invariant we're adding. Go to \
http://katlas.math.toronto.edu/wiki/Invariant_Definition_Table and have a \
quick look. If you're following this tutorial, you can ignore the big warning \
message :-).\
\>", "Text"],

Cell["\<\
Click edit, scroll down, and add a new row to the table; something like\
\>", "Text"],

Cell["\<\
<tr>
<!-- Invariant name --> <td>A-polynomial</td>
<!-- WikiPage =            --> <td>A-polynomial</td>
</tr>\
\>", "Text",
  FontFamily->"Courier New"],

Cell["\<\
That's all! Save your changes, and tell the KnotTheory` package to reload the \
table.\
\>", "Text"],

Cell[BoxData[
    \(\(LoadInvariantRules["\<Invariant_Definition_Table\>"];\)\)], "Input"],

Cell["\<\
To write data to the KnotAtlas, you need to be logged in. If you haven't \
already created an account there, please do so now. If you're going to upload \
lots of data, you might like to create a 'robot' account, like mine, \
\"ScottDataRobot\". This lets us keep track of changes more easily.\
\>", "Text"],

Cell[BoxData[
    \(CreateWikiConnection[\[IndentingNewLine]"\<http://katlas.math.toronto.\
edu/w/index.php\>", \[IndentingNewLine]"\<ScottDataRobot\>", \
\[IndentingNewLine]InputString["\<Enter Your Password\>"]\n]\)], "Input"],

Cell["\<\
You can check that the login worked properly using \"WikiUserName[]\".\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(WikiUserName[]\)], "Input"],

Cell[BoxData[
    \("ScottDataRobot"\)], "Output"]
}, Open  ]],

Cell["\<\
Now upload all the data. StoreInvariants returns a list of failures, so if it \
returns an empty list, all is well.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(StoreInvariants[apolynomialdata, "\<KnotAtlas\>"]\)], "Input"],

Cell[BoxData[
    \({}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(RetrieveInvariants[{"\<A-polynomial\>"}, 
      AllKnots[{3, 9}], "\<KnotAtlas\>"]\)], "Input"],

Cell[BoxData[
    \({{"A-polynomial", Knot[3, 1], L + M\^6}, {"A-polynomial", Knot[4, 1], 
        M\^4 + L\^2\ M\^4 + 
          L\ \((\(-1\) + M\^2 + 2\ M\^4 + M\^6 - M\^8)\)}, {"A-polynomial", 
        Knot[5, 1], L + M\^10}, {"A-polynomial", Knot[5, 2], 
        1 + L\^3\ M\^14 + 
          L\ \((\(-1\) + 2\ M\^2 + 2\ M\^4 - M\^8 + M\^10)\) + 
          L\^2\ \((M\^4 - M\^6 + 2\ M\^10 + 2\ M\^12 - 
                M\^14)\)}, {"A-polynomial", Knot[6, 1], 
        M\^8 + L\^4\ M\^8 + 
          L\^3\ \((\(-1\) + M\^2 + 3\ M\^8 + 3\ M\^10 - 2\ M\^12)\) + 
          L\ \((\(-2\)\ M\^4 + 3\ M\^6 + 3\ M\^8 + M\^14 - M\^16)\) + 
          L\^2\ \((1 - 3\ M\^2 - M\^4 + 3\ M\^6 + 6\ M\^8 + 3\ M\^10 - 
                M\^12 - 3\ M\^14 + M\^16)\)}, {"A-polynomial", Knot[6, 2], 
        M\^4 + L\^5\ M\^26 + 
          L\ \((\(-1\) + 2\ M\^2 - M\^4 - 2\ M\^6 + 5\ M\^8 + 5\ M\^10 - 
                3\ M\^12)\) + 
          L\^2\ \((\(-M\^2\) + 3\ M\^4 - M\^6 - 5\ M\^8 - 3\ M\^10 + 
                12\ M\^12 + 13\ M\^14 - 3\ M\^16 - 8\ M\^18 + 3\ M\^20)\) + 
          L\^3\ \((3\ M\^10 - 8\ M\^12 - 3\ M\^14 + 13\ M\^16 + 12\ M\^18 - 
                3\ M\^20 - 5\ M\^22 - M\^24 + 3\ M\^26 - M\^28)\) + 
          L\^4\ \((\(-3\)\ M\^18 + 5\ M\^20 + 5\ M\^22 - 2\ M\^24 - M\^26 + 
                2\ M\^28 - M\^30)\)}, {"A-polynomial", Knot[6, 3], 
        M\^14 + L\^6\ M\^14 + 
          L\ \((2\ M\^8 - 5\ M\^10 + M\^12 + 10\ M\^14 + M\^16 - 5\ M\^18 + 
                2\ M\^20)\) + 
          L\^5\ \((2\ M\^8 - 5\ M\^10 + M\^12 + 10\ M\^14 + M\^16 - 
                5\ M\^18 + 2\ M\^20)\) + 
          L\^2\ \((M\^2 - 4\ M\^4 + 4\ M\^6 + 2\ M\^8 - 6\ M\^10 + 2\ M\^12 + 
                17\ M\^14 + 2\ M\^16 - 6\ M\^18 + 2\ M\^20 + 4\ M\^22 - 
                4\ M\^24 + M\^26)\) + 
          L\^4\ \((M\^2 - 4\ M\^4 + 4\ M\^6 + 2\ M\^8 - 6\ M\^10 + 2\ M\^12 + 
                17\ M\^14 + 2\ M\^16 - 6\ M\^18 + 2\ M\^20 + 4\ M\^22 - 
                4\ M\^24 + M\^26)\) + 
          L\^3\ \((1 - 5\ M\^2 + 3\ M\^4 + 9\ M\^6 - 2\ M\^8 - 21\ M\^10 + 
                8\ M\^12 + 34\ M\^14 + 8\ M\^16 - 21\ M\^18 - 2\ M\^20 + 
                9\ M\^22 + 3\ M\^24 - 5\ M\^26 + M\^28)\)}, {"A-polynomial", 
        Knot[7, 1], L + M\^14}, {"A-polynomial", Knot[7, 2], 
        L\^5 + M\^22 + 
          L\^4\ \((\(-2\) + 4\ M\^2 + 3\ M\^4 - M\^12 + M\^14)\) + 
          L\^3\ \((1 - 4\ M\^2 + 2\ M\^4 + 5\ M\^6 + 6\ M\^8 - 4\ M\^12 + 
                M\^14 + 5\ M\^16 - 2\ M\^18)\) + 
          L\ \((M\^8 - M\^10 + 3\ M\^18 + 4\ M\^20 - 2\ M\^22)\) + 
          L\^2\ \((\(-2\)\ M\^4 + 5\ M\^6 + M\^8 - 4\ M\^10 + 6\ M\^14 + 
                5\ M\^16 + 2\ M\^18 - 4\ M\^20 + M\^22)\)}, {"A-polynomial", 
        Knot[7, 3], 
        L\^6 + M\^52 + 
          L\^5\ \((\(-1\) + 2\ M\^2 - M\^4 + 5\ M\^8 + 2\ M\^10 - 3\ M\^12 + 
                2\ M\^14)\) + 
          L\^4\ \((3\ M\^8 - 9\ M\^10 + 2\ M\^12 + 14\ M\^14 + 2\ M\^16 - 
                4\ M\^18 + 4\ M\^20 + 2\ M\^22 + 3\ M\^24 - 3\ M\^26 + 
                M\^28)\) + 
          L\^3\ \((\(-3\)\ M\^16 + 10\ M\^18 - 3\ M\^20 - 12\ M\^22 + 
                6\ M\^24 + 24\ M\^26 + 6\ M\^28 - 12\ M\^30 - 3\ M\^32 + 
                10\ M\^34 - 3\ M\^36)\) + 
          L\^2\ \((M\^24 - 3\ M\^26 + 3\ M\^28 + 2\ M\^30 + 4\ M\^32 - 
                4\ M\^34 + 2\ M\^36 + 14\ M\^38 + 2\ M\^40 - 9\ M\^42 + 
                3\ M\^44)\) + 
          L\ \((2\ M\^38 - 3\ M\^40 + 2\ M\^42 + 5\ M\^44 - M\^48 + 
                2\ M\^50 - M\^52)\)}, {"A-polynomial", Knot[7, 4], 
        1 + L\^5\ M\^22 + 
          L\ \((\(-3\) + 7\ M\^2 + 4\ M\^4 - 6\ M\^6 + M\^8 + 3\ M\^10 - 
                2\ M\^12 + M\^14)\) + 
          L\^4\ \((M\^8 - 2\ M\^10 + 3\ M\^12 + M\^14 - 6\ M\^16 + 4\ M\^18 + 
                7\ M\^20 - 3\ M\^22)\) + 
          L\^2\ \((3 - 10\ M\^2 + 3\ M\^4 + 21\ M\^6 - 3\ M\^8 - 17\ M\^10 + 
                6\ M\^12 + 10\ M\^14 - 2\ M\^16 - 3\ M\^18 + 3\ M\^20 - 
                M\^22)\) + 
          L\^3\ \((\(-1\) + 3\ M\^2 - 3\ M\^4 - 2\ M\^6 + 10\ M\^8 + 
                6\ M\^10 - 17\ M\^12 - 3\ M\^14 + 21\ M\^16 + 3\ M\^18 - 
                10\ M\^20 + 3\ M\^22)\)}, {"A-polynomial", Knot[7, 5], 
        L\^8 + M\^68 + 
          L\^7\ \((\(-1\) + 4\ M\^2 - 5\ M\^4 - 2\ M\^6 + 13\ M\^8 + 
                3\ M\^10 - 7\ M\^12 + 3\ M\^14)\) + 
          L\^6\ \((M\^4 - 6\ M\^6 + 17\ M\^8 - 10\ M\^10 - 35\ M\^12 + 
                32\ M\^14 + 56\ M\^16 - 24\ M\^18 - 28\ M\^20 + 22\ M\^22 + 
                14\ M\^24 - 14\ M\^26 + 3\ M\^28)\) + 
          L\^5\ \((2\ M\^10 - 12\ M\^12 + 23\ M\^14 + 6\ M\^16 - 48\ M\^18 - 
                15\ M\^20 + 82\ M\^22 + 28\ M\^24 - 47\ M\^26 + 13\ M\^28 + 
                46\ M\^30 - 15\ M\^32 - 15\ M\^34 + 2\ M\^36 + 12\ M\^38 - 
                7\ M\^40 + M\^42)\) + 
          L\^4\ \((M\^16 - 6\ M\^18 + 11\ M\^20 - 4\ M\^22 - 4\ M\^24 - 
                4\ M\^26 + 11\ M\^28 - 12\ M\^30 + 16\ M\^32 + 52\ M\^34 + 
                16\ M\^36 - 12\ M\^38 + 11\ M\^40 - 4\ M\^42 - 4\ M\^44 - 
                4\ M\^46 + 11\ M\^48 - 6\ M\^50 + M\^52)\) + 
          L\^3\ \((M\^26 - 7\ M\^28 + 12\ M\^30 + 2\ M\^32 - 15\ M\^34 - 
                15\ M\^36 + 46\ M\^38 + 13\ M\^40 - 47\ M\^42 + 28\ M\^44 + 
                82\ M\^46 - 15\ M\^48 - 48\ M\^50 + 6\ M\^52 + 23\ M\^54 - 
                12\ M\^56 + 2\ M\^58)\) + 
          L\^2\ \((3\ M\^40 - 14\ M\^42 + 14\ M\^44 + 22\ M\^46 - 28\ M\^48 - 
                24\ M\^50 + 56\ M\^52 + 32\ M\^54 - 35\ M\^56 - 10\ M\^58 + 
                17\ M\^60 - 6\ M\^62 + M\^64)\) + 
          L\ \((3\ M\^54 - 7\ M\^56 + 3\ M\^58 + 13\ M\^60 - 2\ M\^62 - 
                5\ M\^64 + 4\ M\^66 - M\^68)\)}, {"A-polynomial", Knot[7, 6], 
        L\^9\ M\^8 + M\^46 + 
          L\^8\ \((\(-2\)\ M\^4 + 6\ M\^6 - 5\ M\^8 - 5\ M\^10 + 16\ M\^12 + 
                5\ M\^14 - 9\ M\^16 + 3\ M\^18)\) + 
          L\^7\ \((1 - 6\ M\^2 + 11\ M\^4 + 2\ M\^6 - 16\ M\^8 - 7\ M\^10 + 
                9\ M\^12 + 8\ M\^14 + 42\ M\^16 + 11\ M\^18 - 37\ M\^20 + 
                8\ M\^22 + 23\ M\^24 - 16\ M\^26 + 3\ M\^28)\) + 
          L\^6\ \((\(-1\) + 7\ M\^2 - 16\ M\^4 + M\^6 + 34\ M\^8 + 
                10\ M\^10 - 80\ M\^12 - 9\ M\^14 + 62\ M\^16 + 10\ M\^18 + 
                34\ M\^20 + 83\ M\^22 - 44\ M\^24 - 48\ M\^26 + 47\ M\^28 + 
                10\ M\^30 - 29\ M\^32 + 19\ M\^34 - 7\ M\^36 + M\^38)\) + 
          L\^5\ \((M\^4 - 9\ M\^6 + 32\ M\^8 - 30\ M\^10 - 68\ M\^12 + 
                98\ M\^14 + 164\ M\^16 - 212\ M\^18 - 266\ M\^20 + 
                196\ M\^22 + 377\ M\^24 + 24\ M\^26 - 237\ M\^28 - 
                85\ M\^30 + 158\ M\^32 + 52\ M\^34 - 78\ M\^36 - 18\ M\^38 + 
                41\ M\^40 - 16\ M\^42 + 2\ M\^44)\) + 
          L\ \((3\ M\^36 - 9\ M\^38 + 5\ M\^40 + 16\ M\^42 - 5\ M\^44 - 
                5\ M\^46 + 6\ M\^48 - 2\ M\^50)\) + 
          L\^4\ \((2\ M\^10 - 16\ M\^12 + 41\ M\^14 - 18\ M\^16 - 78\ M\^18 + 
                52\ M\^20 + 158\ M\^22 - 85\ M\^24 - 237\ M\^26 + 24\ M\^28 + 
                377\ M\^30 + 196\ M\^32 - 266\ M\^34 - 212\ M\^36 + 
                164\ M\^38 + 98\ M\^40 - 68\ M\^42 - 30\ M\^44 + 32\ M\^46 - 
                9\ M\^48 + M\^50)\) + 
          L\^3\ \((M\^16 - 7\ M\^18 + 19\ M\^20 - 29\ M\^22 + 10\ M\^24 + 
                47\ M\^26 - 48\ M\^28 - 44\ M\^30 + 83\ M\^32 + 34\ M\^34 + 
                10\ M\^36 + 62\ M\^38 - 9\ M\^40 - 80\ M\^42 + 10\ M\^44 + 
                34\ M\^46 + M\^48 - 16\ M\^50 + 7\ M\^52 - M\^54)\) + 
          L\^2\ \((3\ M\^26 - 16\ M\^28 + 23\ M\^30 + 8\ M\^32 - 37\ M\^34 + 
                11\ M\^36 + 42\ M\^38 + 8\ M\^40 + 9\ M\^42 - 7\ M\^44 - 
                16\ M\^46 + 2\ M\^48 + 11\ M\^50 - 6\ M\^52 + 
                M\^54)\)}, {"A-polynomial", Knot[7, 7], 
        L\^7\ M\^18 + M\^20 + 
          L\^6\ \((3\ M\^12 - 11\ M\^14 + 4\ M\^16 + 20\ M\^18 - 7\ M\^20 - 
                7\ M\^22 + 7\ M\^24 - 2\ M\^26)\) + 
          L\ \((\(-2\)\ M\^12 + 7\ M\^14 - 7\ M\^16 - 7\ M\^18 + 20\ M\^20 + 
                4\ M\^22 - 11\ M\^24 + 3\ M\^26)\) + 
          L\^2\ \((M\^4 - 6\ M\^6 + 11\ M\^8 - M\^10 - 8\ M\^12 + M\^14 - 
                28\ M\^16 + 16\ M\^18 + 84\ M\^20 - 19\ M\^22 - 65\ M\^24 + 
                27\ M\^26 + 23\ M\^28 - 18\ M\^30 + 3\ M\^32)\) + 
          L\^5\ \((3\ M\^6 - 18\ M\^8 + 23\ M\^10 + 27\ M\^12 - 65\ M\^14 - 
                19\ M\^16 + 84\ M\^18 + 16\ M\^20 - 28\ M\^22 + M\^24 - 
                8\ M\^26 - M\^28 + 11\ M\^30 - 6\ M\^32 + M\^34)\) + 
          L\^4\ \((1 - 7\ M\^2 + 15\ M\^4 + M\^6 - 30\ M\^8 - M\^10 + 
                41\ M\^12 + 7\ M\^14 - 29\ M\^16 - 46\ M\^18 + 35\ M\^20 + 
                130\ M\^22 - 2\ M\^24 - 123\ M\^26 + 59\ M\^30 - 4\ M\^32 - 
                19\ M\^34 + 8\ M\^36 - M\^38)\) + 
          L\^3\ \((\(-1\) + 8\ M\^2 - 19\ M\^4 - 4\ M\^6 + 59\ M\^8 - 
                123\ M\^12 - 2\ M\^14 + 130\ M\^16 + 35\ M\^18 - 46\ M\^20 - 
                29\ M\^22 + 7\ M\^24 + 41\ M\^26 - M\^28 - 30\ M\^30 + 
                M\^32 + 15\ M\^34 - 7\ M\^36 + M\^38)\)}, {"A-polynomial", 
        Knot[8, 1], 
        M\^12 + L\^6\ M\^12 + 
          L\^5\ \((\(-1\) + M\^2 + 4\ M\^12 + 5\ M\^14 - 3\ M\^16)\) + 
          L\^4\ \((2 - 6\ M\^2 + 4\ M\^6 + 10\ M\^12 + 12\ M\^14 - 
                10\ M\^18 + 3\ M\^20)\) + 
          L\ \((\(-3\)\ M\^8 + 5\ M\^10 + 4\ M\^12 + M\^22 - M\^24)\) + 
          L\^3\ \((\(-1\) + 5\ M\^2 - 6\ M\^4 - 5\ M\^6 - 3\ M\^8 + 
                10\ M\^10 + 20\ M\^12 + 10\ M\^14 - 3\ M\^16 - 5\ M\^18 - 
                6\ M\^20 + 5\ M\^22 - M\^24)\) + 
          L\^2\ \((3\ M\^4 - 10\ M\^6 + 12\ M\^10 + 10\ M\^12 + 4\ M\^18 - 
                6\ M\^22 + 2\ M\^24)\)}, {"A-polynomial", Knot[8, 2], 
        L\^8\ M\^4 + M\^72 + 
          L\^7\ \((\(-1\) + 2\ M\^2 - M\^4 - 3\ M\^10 + 7\ M\^12 + 9\ M\^14 - 
                5\ M\^16)\) + 
          L\^6\ \((\(-2\)\ M\^6 + 7\ M\^8 - 8\ M\^10 - 4\ M\^12 + 14\ M\^14 - 
                M\^16 - 28\ M\^18 + 17\ M\^20 + 56\ M\^22 - M\^24 - 
                32\ M\^26 + 10\ M\^28)\) + 
          L\^5\ \((\(-M\^12\) + 5\ M\^14 - 8\ M\^16 + M\^18 - 2\ M\^20 + 
                29\ M\^22 - 17\ M\^24 - 77\ M\^26 + 33\ M\^28 + 143\ M\^30 + 
                29\ M\^32 - 87\ M\^34 - 24\ M\^36 + 42\ M\^38 - 
                10\ M\^40)\) + 
          L\^4\ \((5\ M\^24 - 24\ M\^26 + 26\ M\^28 + 36\ M\^30 - 43\ M\^32 - 
                108\ M\^34 + 47\ M\^36 + 192\ M\^38 + 47\ M\^40 - 
                108\ M\^42 - 43\ M\^44 + 36\ M\^46 + 26\ M\^48 - 24\ M\^50 + 
                5\ M\^52)\) + 
          L\^3\ \((\(-10\)\ M\^36 + 42\ M\^38 - 24\ M\^40 - 87\ M\^42 + 
                29\ M\^44 + 143\ M\^46 + 33\ M\^48 - 77\ M\^50 - 17\ M\^52 + 
                29\ M\^54 - 2\ M\^56 + M\^58 - 8\ M\^60 + 5\ M\^62 - 
                M\^64)\) + 
          L\^2\ \((10\ M\^48 - 32\ M\^50 - M\^52 + 56\ M\^54 + 17\ M\^56 - 
                28\ M\^58 - M\^60 + 14\ M\^62 - 4\ M\^64 - 8\ M\^66 + 
                7\ M\^68 - 2\ M\^70)\) + 
          L\ \((\(-5\)\ M\^60 + 9\ M\^62 + 7\ M\^64 - 3\ M\^66 - M\^72 + 
                2\ M\^74 - M\^76)\)}, {"A-polynomial", Knot[8, 3], 
        M\^16 + L\^8\ M\^16 + 
          L\ \((\(-2\)\ M\^8 + 3\ M\^10 - 2\ M\^12 + M\^14 + 8\ M\^16 + 
                M\^18 - 2\ M\^20 + 3\ M\^22 - 2\ M\^24)\) + 
          L\^7\ \((\(-2\)\ M\^8 + 3\ M\^10 - 2\ M\^12 + M\^14 + 8\ M\^16 + 
                M\^18 - 2\ M\^20 + 3\ M\^22 - 2\ M\^24)\) + 
          L\^3\ \((\(-4\) + 16\ M\^2 - 14\ M\^4 - 21\ M\^6 + 24\ M\^8 + 
                16\ M\^10 - 56\ M\^12 + 17\ M\^14 + 100\ M\^16 + 17\ M\^18 - 
                56\ M\^20 + 16\ M\^22 + 24\ M\^24 - 21\ M\^26 - 14\ M\^28 + 
                16\ M\^30 - 4\ M\^32)\) + 
          L\^5\ \((\(-4\) + 16\ M\^2 - 14\ M\^4 - 21\ M\^6 + 24\ M\^8 + 
                16\ M\^10 - 56\ M\^12 + 17\ M\^14 + 100\ M\^16 + 17\ M\^18 - 
                56\ M\^20 + 16\ M\^22 + 24\ M\^24 - 21\ M\^26 - 14\ M\^28 + 
                16\ M\^30 - 4\ M\^32)\) + 
          L\^2\ \((1 - 3\ M\^2 + 3\ M\^4 - M\^6 - 4\ M\^8 - 9\ M\^10 + 
                15\ M\^12 + 13\ M\^14 - 2\ M\^16 + 13\ M\^18 + 15\ M\^20 - 
                9\ M\^22 - 4\ M\^24 - M\^26 + 3\ M\^28 - 3\ M\^30 + 
                M\^32)\) + 
          L\^6\ \((1 - 3\ M\^2 + 3\ M\^4 - M\^6 - 4\ M\^8 - 9\ M\^10 + 
                15\ M\^12 + 13\ M\^14 - 2\ M\^16 + 13\ M\^18 + 15\ M\^20 - 
                9\ M\^22 - 4\ M\^24 - M\^26 + 3\ M\^28 - 3\ M\^30 + 
                M\^32)\) + 
          L\^4\ \((6 - 26\ M\^2 + 22\ M\^4 + 40\ M\^6 - 59\ M\^8 - 
                64\ M\^10 + 82\ M\^12 + 50\ M\^14 - 32\ M\^16 + 50\ M\^18 + 
                82\ M\^20 - 64\ M\^22 - 59\ M\^24 + 40\ M\^26 + 22\ M\^28 - 
                26\ M\^30 + 6\ M\^32)\)}, {"A-polynomial", Knot[8, 4], 
        M\^10 + L\^9\ M\^48 + 
          L\ \((\(-M\^2\) + 2\ M\^4 - 3\ M\^6 + 2\ M\^8 + 3\ M\^10 - 
                6\ M\^12 + 7\ M\^14 + 11\ M\^16 - 6\ M\^18)\) + 
          L\^2\ \((\(-1\) + 4\ M\^2 - 7\ M\^4 + 7\ M\^6 + 2\ M\^8 - 
                27\ M\^10 + 11\ M\^12 + 46\ M\^14 - 44\ M\^16 - 12\ M\^18 + 
                88\ M\^20 + 3\ M\^22 - 49\ M\^24 + 15\ M\^26)\) + 
          L\^3\ \((1 - 5\ M\^2 + 8\ M\^4 - 16\ M\^8 + 9\ M\^10 + 55\ M\^12 - 
                98\ M\^14 - 51\ M\^16 + 178\ M\^18 - 34\ M\^20 - 163\ M\^22 + 
                224\ M\^24 + 154\ M\^26 - 189\ M\^28 - 55\ M\^30 + 
                86\ M\^32 - 20\ M\^34)\) + 
          L\^4\ \((\(-6\)\ M\^8 + 31\ M\^10 - 44\ M\^12 - 37\ M\^14 + 
                136\ M\^16 + 15\ M\^18 - 291\ M\^20 + 105\ M\^22 + 
                245\ M\^24 - 220\ M\^26 + 91\ M\^28 + 399\ M\^30 - 
                201\ M\^32 - 263\ M\^34 + 144\ M\^36 + 81\ M\^38 - 
                74\ M\^40 + 15\ M\^42)\) + 
          L\^5\ \((15\ M\^16 - 74\ M\^18 + 81\ M\^20 + 144\ M\^22 - 
                263\ M\^24 - 201\ M\^26 + 399\ M\^28 + 91\ M\^30 - 
                220\ M\^32 + 245\ M\^34 + 105\ M\^36 - 291\ M\^38 + 
                15\ M\^40 + 136\ M\^42 - 37\ M\^44 - 44\ M\^46 + 31\ M\^48 - 
                6\ M\^50)\) + 
          L\^8\ \((\(-6\)\ M\^40 + 11\ M\^42 + 7\ M\^44 - 6\ M\^46 + 
                3\ M\^48 + 2\ M\^50 - 3\ M\^52 + 2\ M\^54 - M\^56)\) + 
          L\^7\ \((15\ M\^32 - 49\ M\^34 + 3\ M\^36 + 88\ M\^38 - 12\ M\^40 - 
                44\ M\^42 + 46\ M\^44 + 11\ M\^46 - 27\ M\^48 + 2\ M\^50 + 
                7\ M\^52 - 7\ M\^54 + 4\ M\^56 - M\^58)\) + 
          L\^6\ \((\(-20\)\ M\^24 + 86\ M\^26 - 55\ M\^28 - 189\ M\^30 + 
                154\ M\^32 + 224\ M\^34 - 163\ M\^36 - 34\ M\^38 + 
                178\ M\^40 - 51\ M\^42 - 98\ M\^44 + 55\ M\^46 + 9\ M\^48 - 
                16\ M\^50 + 8\ M\^54 - 5\ M\^56 + M\^58)\)}, {"A-polynomial", 
        Knot[8, 5], 
        M\^4 + L\^8\ M\^68 + 
          L\ \((\(-1\) + 3\ M\^2 - 7\ M\^4 + 8\ M\^6 + 3\ M\^8 - 12\ M\^10 + 
                11\ M\^12 + 7\ M\^14 - 4\ M\^16)\) + 
          L\^2\ \((\(-M\^2\) + 4\ M\^4 - 10\ M\^6 + 12\ M\^8 - 16\ M\^12 + 
                33\ M\^16 - 11\ M\^18 - 24\ M\^20 + 56\ M\^22 + M\^24 - 
                22\ M\^26 + 6\ M\^28)\) + 
          L\^3\ \((M\^10 - 6\ M\^12 + 19\ M\^14 - 39\ M\^16 + 18\ M\^18 + 
                54\ M\^20 - 66\ M\^22 - 31\ M\^24 + 93\ M\^26 + 17\ M\^28 - 
                48\ M\^30 + 80\ M\^32 - 15\ M\^34 - 41\ M\^36 + 24\ M\^38 - 
                4\ M\^40)\) + 
          L\^4\ \((M\^20 - 10\ M\^22 + 35\ M\^24 - 47\ M\^26 - 16\ M\^28 + 
                79\ M\^30 - 53\ M\^32 - 2\ M\^34 + 96\ M\^36 - 2\ M\^38 - 
                53\ M\^40 + 79\ M\^42 - 16\ M\^44 - 47\ M\^46 + 35\ M\^48 - 
                10\ M\^50 + M\^52)\) + 
          L\^5\ \((\(-4\)\ M\^32 + 24\ M\^34 - 41\ M\^36 - 15\ M\^38 + 
                80\ M\^40 - 48\ M\^42 + 17\ M\^44 + 93\ M\^46 - 31\ M\^48 - 
                66\ M\^50 + 54\ M\^52 + 18\ M\^54 - 39\ M\^56 + 19\ M\^58 - 
                6\ M\^60 + M\^62)\) + 
          L\^6\ \((6\ M\^44 - 22\ M\^46 + M\^48 + 56\ M\^50 - 24\ M\^52 - 
                11\ M\^54 + 33\ M\^56 - 16\ M\^60 + 12\ M\^64 - 10\ M\^66 + 
                4\ M\^68 - M\^70)\) + 
          L\^7\ \((\(-4\)\ M\^56 + 7\ M\^58 + 11\ M\^60 - 12\ M\^62 + 
                3\ M\^64 + 8\ M\^66 - 7\ M\^68 + 3\ M\^70 - 
                M\^72)\)}, {"A-polynomial", Knot[8, 6], 
        L\^11\ M\^8 + M\^54 + 
          L\^10\ \((\(-2\)\ M\^4 + 6\ M\^6 - 8\ M\^8 + 19\ M\^12 - 2\ M\^14 - 
                6\ M\^16 + 7\ M\^18 - 3\ M\^20)\) + 
          L\^9\ \((1 - 6\ M\^2 + 13\ M\^4 - 11\ M\^6 + 2\ M\^8 - 17\ M\^10 - 
                2\ M\^12 + 70\ M\^14 - 10\ M\^16 - 47\ M\^18 + 85\ M\^20 + 
                29\ M\^22 - 64\ M\^24 - 4\ M\^26 + 27\ M\^28 - 14\ M\^30 + 
                3\ M\^32)\) + 
          L\^8\ \((\(-1\) + 9\ M\^2 - 38\ M\^4 + 63\ M\^6 + 17\ M\^8 - 
                138\ M\^10 - 28\ M\^12 + 250\ M\^14 + 24\ M\^16 - 
                343\ M\^18 + 28\ M\^20 + 491\ M\^22 + 52\ M\^24 - 
                285\ M\^26 + 11\ M\^28 + 120\ M\^30 - 71\ M\^32 - 26\ M\^34 + 
                28\ M\^36 + 17\ M\^38 - 21\ M\^40 + 7\ M\^42 - M\^44)\) + 
          L\^7\ \((\(-3\)\ M\^2 + 26\ M\^4 - 90\ M\^6 + 104\ M\^8 + 
                135\ M\^10 - 336\ M\^12 - 252\ M\^14 + 744\ M\^16 + 
                358\ M\^18 - 1055\ M\^20 - 552\ M\^22 + 1054\ M\^24 + 
                773\ M\^26 - 207\ M\^28 - 340\ M\^30 - 213\ M\^32 + 
                45\ M\^34 + 261\ M\^36 - 78\ M\^38 - 104\ M\^40 + 54\ M\^42 + 
                39\ M\^44 - 48\ M\^46 + 17\ M\^48 - 2\ M\^50)\) + 
          L\^6\ \((\(-3\)\ M\^4 + 27\ M\^6 - 89\ M\^8 + 97\ M\^10 + 
                113\ M\^12 - 237\ M\^14 - 257\ M\^16 + 363\ M\^18 + 
                515\ M\^20 + 21\ M\^22 - 1051\ M\^24 - 998\ M\^26 + 
                1231\ M\^28 + 2288\ M\^30 - 339\ M\^32 - 1945\ M\^34 + 
                2\ M\^36 + 1031\ M\^38 - 141\ M\^40 - 251\ M\^42 + 
                106\ M\^44 - 3\ M\^46 - 54\ M\^48 + 60\ M\^50 - 32\ M\^52 + 
                9\ M\^54 - M\^56)\) + 
          L\^5\ \((\(-M\^6\) + 9\ M\^8 - 32\ M\^10 + 60\ M\^12 - 54\ M\^14 - 
                3\ M\^16 + 106\ M\^18 - 251\ M\^20 - 141\ M\^22 + 
                1031\ M\^24 + 2\ M\^26 - 1945\ M\^28 - 339\ M\^30 + 
                2288\ M\^32 + 1231\ M\^34 - 998\ M\^36 - 1051\ M\^38 + 
                21\ M\^40 + 515\ M\^42 + 363\ M\^44 - 257\ M\^46 - 
                237\ M\^48 + 113\ M\^50 + 97\ M\^52 - 89\ M\^54 + 27\ M\^56 - 
                3\ M\^58)\) + 
          L\ \((\(-3\)\ M\^42 + 7\ M\^44 - 6\ M\^46 - 2\ M\^48 + 19\ M\^50 - 
                8\ M\^54 + 6\ M\^56 - 2\ M\^58)\) + 
          L\^4\ \((\(-2\)\ M\^12 + 17\ M\^14 - 48\ M\^16 + 39\ M\^18 + 
                54\ M\^20 - 104\ M\^22 - 78\ M\^24 + 261\ M\^26 + 45\ M\^28 - 
                213\ M\^30 - 340\ M\^32 - 207\ M\^34 + 773\ M\^36 + 
                1054\ M\^38 - 552\ M\^40 - 1055\ M\^42 + 358\ M\^44 + 
                744\ M\^46 - 252\ M\^48 - 336\ M\^50 + 135\ M\^52 + 
                104\ M\^54 - 90\ M\^56 + 26\ M\^58 - 3\ M\^60)\) + 
          L\^3\ \((\(-M\^18\) + 7\ M\^20 - 21\ M\^22 + 17\ M\^24 + 
                28\ M\^26 - 26\ M\^28 - 71\ M\^30 + 120\ M\^32 + 11\ M\^34 - 
                285\ M\^36 + 52\ M\^38 + 491\ M\^40 + 28\ M\^42 - 
                343\ M\^44 + 24\ M\^46 + 250\ M\^48 - 28\ M\^50 - 
                138\ M\^52 + 17\ M\^54 + 63\ M\^56 - 38\ M\^58 + 9\ M\^60 - 
                M\^62)\) + 
          L\^2\ \((3\ M\^30 - 14\ M\^32 + 27\ M\^34 - 4\ M\^36 - 64\ M\^38 + 
                29\ M\^40 + 85\ M\^42 - 47\ M\^44 - 10\ M\^46 + 70\ M\^48 - 
                2\ M\^50 - 17\ M\^52 + 2\ M\^54 - 11\ M\^56 + 13\ M\^58 - 
                6\ M\^60 + M\^62)\)}, {"A-polynomial", Knot[8, 7], 
        L\^11\ M\^12 + M\^58 + 
          L\^10\ \((2\ M\^6 - 5\ M\^8 + 4\ M\^10 - M\^12 - 5\ M\^14 + 
                17\ M\^16 + 6\ M\^18 - 11\ M\^20 + 4\ M\^22)\) + 
          L\^9\ \((1 - 4\ M\^2 + 6\ M\^4 - 4\ M\^6 - M\^8 - 2\ M\^10 + 
                23\ M\^12 - M\^14 - 67\ M\^16 + M\^18 + 142\ M\^20 + 
                23\ M\^22 - 102\ M\^24 + 17\ M\^26 + 47\ M\^28 - 30\ M\^30 + 
                6\ M\^32)\) + 
          L\^8\ \((3\ M\^2 - 19\ M\^4 + 42\ M\^6 - 17\ M\^8 - 81\ M\^10 + 
                87\ M\^12 + 115\ M\^14 - 192\ M\^16 - 35\ M\^18 + 
                130\ M\^20 - 46\ M\^22 + 104\ M\^24 + 188\ M\^26 - 
                137\ M\^28 - 41\ M\^30 + 124\ M\^32 - 50\ M\^34 - 53\ M\^36 + 
                66\ M\^38 - 27\ M\^40 + 4\ M\^42)\) + 
          L\^7\ \((3\ M\^4 - 26\ M\^6 + 75\ M\^8 - 47\ M\^10 - 142\ M\^12 + 
                163\ M\^14 + 223\ M\^16 - 391\ M\^18 - 77\ M\^20 + 
                512\ M\^22 - 230\ M\^24 - 578\ M\^26 + 711\ M\^28 + 
                748\ M\^30 - 613\ M\^32 - 455\ M\^34 + 521\ M\^36 + 
                150\ M\^38 - 275\ M\^40 - M\^42 + 105\ M\^44 - 67\ M\^46 + 
                28\ M\^48 - 8\ M\^50 + M\^52)\) + 
          L\^6\ \((M\^6 - 11\ M\^8 + 41\ M\^10 - 54\ M\^12 - 4\ M\^14 + 
                30\ M\^16 + 79\ M\^18 - 107\ M\^20 - 179\ M\^22 + 
                276\ M\^24 + 424\ M\^26 - 735\ M\^28 - 527\ M\^30 + 
                1182\ M\^32 + 795\ M\^34 - 791\ M\^36 - 367\ M\^38 + 
                446\ M\^40 + 131\ M\^42 - 229\ M\^44 + 25\ M\^46 + 
                80\ M\^48 - 20\ M\^50 - 68\ M\^52 + 61\ M\^54 - 19\ M\^56 + 
                2\ M\^58)\) + 
          L\^5\ \((2\ M\^12 - 19\ M\^14 + 61\ M\^16 - 68\ M\^18 - 20\ M\^20 + 
                80\ M\^22 + 25\ M\^24 - 229\ M\^26 + 131\ M\^28 + 
                446\ M\^30 - 367\ M\^32 - 791\ M\^34 + 795\ M\^36 + 
                1182\ M\^38 - 527\ M\^40 - 735\ M\^42 + 424\ M\^44 + 
                276\ M\^46 - 179\ M\^48 - 107\ M\^50 + 79\ M\^52 + 
                30\ M\^54 - 4\ M\^56 - 54\ M\^58 + 41\ M\^60 - 11\ M\^62 + 
                M\^64)\) + 
          L\ \((4\ M\^48 - 11\ M\^50 + 6\ M\^52 + 17\ M\^54 - 5\ M\^56 - 
                M\^58 + 4\ M\^60 - 5\ M\^62 + 2\ M\^64)\) + 
          L\^4\ \((M\^18 - 8\ M\^20 + 28\ M\^22 - 67\ M\^24 + 105\ M\^26 - 
                M\^28 - 275\ M\^30 + 150\ M\^32 + 521\ M\^34 - 455\ M\^36 - 
                613\ M\^38 + 748\ M\^40 + 711\ M\^42 - 578\ M\^44 - 
                230\ M\^46 + 512\ M\^48 - 77\ M\^50 - 391\ M\^52 + 
                223\ M\^54 + 163\ M\^56 - 142\ M\^58 - 47\ M\^60 + 
                75\ M\^62 - 26\ M\^64 + 3\ M\^66)\) + 
          L\^3\ \((4\ M\^28 - 27\ M\^30 + 66\ M\^32 - 53\ M\^34 - 50\ M\^36 + 
                124\ M\^38 - 41\ M\^40 - 137\ M\^42 + 188\ M\^44 + 
                104\ M\^46 - 46\ M\^48 + 130\ M\^50 - 35\ M\^52 - 
                192\ M\^54 + 115\ M\^56 + 87\ M\^58 - 81\ M\^60 - 17\ M\^62 + 
                42\ M\^64 - 19\ M\^66 + 3\ M\^68)\) + 
          L\^2\ \((6\ M\^38 - 30\ M\^40 + 47\ M\^42 + 17\ M\^44 - 
                102\ M\^46 + 23\ M\^48 + 142\ M\^50 + M\^52 - 67\ M\^54 - 
                M\^56 + 23\ M\^58 - 2\ M\^60 - M\^62 - 4\ M\^64 + 6\ M\^66 - 
                4\ M\^68 + M\^70)\)}, {"A-polynomial", Knot[8, 8], 
        M\^30 + L\^12\ M\^30 + 
          L\^11\ \((2\ M\^20 - 6\ M\^22 + 7\ M\^24 - M\^26 - 10\ M\^28 + 
                20\ M\^30 + 9\ M\^32 - 13\ M\^34 + 4\ M\^36)\) + 
          L\ \((4\ M\^24 - 13\ M\^26 + 9\ M\^28 + 20\ M\^30 - 10\ M\^32 - 
                M\^34 + 7\ M\^36 - 6\ M\^38 + 2\ M\^40)\) + 
          L\^10\ \((M\^10 - 6\ M\^12 + 17\ M\^14 - 18\ M\^16 - 23\ M\^18 + 
                45\ M\^20 + 44\ M\^22 - 59\ M\^24 - 78\ M\^26 + 23\ M\^28 + 
                169\ M\^30 + 41\ M\^32 - 136\ M\^34 + 10\ M\^36 + 66\ M\^38 - 
                36\ M\^40 + 6\ M\^42)\) + 
          L\^9\ \((M\^4 - 8\ M\^6 + 29\ M\^8 - 46\ M\^10 - 23\ M\^12 + 
                147\ M\^14 - M\^16 - 364\ M\^18 + 119\ M\^20 + 529\ M\^22 - 
                119\ M\^24 - 620\ M\^26 + 10\ M\^28 + 668\ M\^30 + 
                279\ M\^32 - 447\ M\^34 - 125\ M\^36 + 287\ M\^38 - 
                52\ M\^40 - 113\ M\^42 + 98\ M\^44 - 33\ M\^46 + 
                4\ M\^48)\) + 
          L\^2\ \((6\ M\^18 - 36\ M\^20 + 66\ M\^22 + 10\ M\^24 - 
                136\ M\^26 + 41\ M\^28 + 169\ M\^30 + 23\ M\^32 - 78\ M\^34 - 
                59\ M\^36 + 44\ M\^38 + 45\ M\^40 - 23\ M\^42 - 18\ M\^44 + 
                17\ M\^46 - 6\ M\^48 + M\^50)\) + 
          L\^8\ \((\(-1\) + 9\ M\^2 - 29\ M\^4 + 32\ M\^6 + 17\ M\^8 + 
                M\^10 - 175\ M\^12 + 14\ M\^14 + 536\ M\^16 + 7\ M\^18 - 
                1322\ M\^20 + 188\ M\^22 + 2055\ M\^24 - 354\ M\^26 - 
                2323\ M\^28 + 721\ M\^30 + 2283\ M\^32 - 541\ M\^34 - 
                1361\ M\^36 + 649\ M\^38 + 470\ M\^40 - 479\ M\^42 - 
                12\ M\^44 + 202\ M\^46 - 128\ M\^48 + 45\ M\^50 - 10\ M\^52 + 
                M\^54)\) + 
          L\^3\ \((4\ M\^12 - 33\ M\^14 + 98\ M\^16 - 113\ M\^18 - 
                52\ M\^20 + 287\ M\^22 - 125\ M\^24 - 447\ M\^26 + 
                279\ M\^28 + 668\ M\^30 + 10\ M\^32 - 620\ M\^34 - 
                119\ M\^36 + 529\ M\^38 + 119\ M\^40 - 364\ M\^42 - M\^44 + 
                147\ M\^46 - 23\ M\^48 - 46\ M\^50 + 29\ M\^52 - 8\ M\^54 + 
                M\^56)\) + 
          L\^7\ \((1 - 12\ M\^2 + 62\ M\^4 - 142\ M\^6 + 27\ M\^8 + 
                395\ M\^10 - 168\ M\^12 - 1091\ M\^14 + 568\ M\^16 + 
                1932\ M\^18 - 757\ M\^20 - 2792\ M\^22 + 799\ M\^24 + 
                2846\ M\^26 - 447\ M\^28 - 1661\ M\^30 + 881\ M\^32 + 
                765\ M\^34 - 715\ M\^36 + 228\ M\^38 + 677\ M\^40 - 
                673\ M\^42 - 289\ M\^44 + 462\ M\^46 + 20\ M\^48 - 
                225\ M\^50 + 130\ M\^52 - 32\ M\^54 + 3\ M\^56)\) + 
          L\^6\ \((3\ M\^2 - 33\ M\^4 + 143\ M\^6 - 243\ M\^8 - 106\ M\^10 + 
                767\ M\^12 + 2\ M\^14 - 1841\ M\^16 + 357\ M\^18 + 
                2670\ M\^20 - 361\ M\^22 - 2690\ M\^24 + 216\ M\^26 + 
                1370\ M\^28 + 416\ M\^30 + 1370\ M\^32 + 216\ M\^34 - 
                2690\ M\^36 - 361\ M\^38 + 2670\ M\^40 + 357\ M\^42 - 
                1841\ M\^44 + 2\ M\^46 + 767\ M\^48 - 106\ M\^50 - 
                243\ M\^52 + 143\ M\^54 - 33\ M\^56 + 3\ M\^58)\) + 
          L\^4\ \((M\^6 - 10\ M\^8 + 45\ M\^10 - 128\ M\^12 + 202\ M\^14 - 
                12\ M\^16 - 479\ M\^18 + 470\ M\^20 + 649\ M\^22 - 
                1361\ M\^24 - 541\ M\^26 + 2283\ M\^28 + 721\ M\^30 - 
                2323\ M\^32 - 354\ M\^34 + 2055\ M\^36 + 188\ M\^38 - 
                1322\ M\^40 + 7\ M\^42 + 536\ M\^44 + 14\ M\^46 - 
                175\ M\^48 + M\^50 + 17\ M\^52 + 32\ M\^54 - 29\ M\^56 + 
                9\ M\^58 - M\^60)\) + 
          L\^5\ \((3\ M\^4 - 32\ M\^6 + 130\ M\^8 - 225\ M\^10 + 20\ M\^12 + 
                462\ M\^14 - 289\ M\^16 - 673\ M\^18 + 677\ M\^20 + 
                228\ M\^22 - 715\ M\^24 + 765\ M\^26 + 881\ M\^28 - 
                1661\ M\^30 - 447\ M\^32 + 2846\ M\^34 + 799\ M\^36 - 
                2792\ M\^38 - 757\ M\^40 + 1932\ M\^42 + 568\ M\^44 - 
                1091\ M\^46 - 168\ M\^48 + 395\ M\^50 + 27\ M\^52 - 
                142\ M\^54 + 62\ M\^56 - 12\ M\^58 + 
                M\^60)\)}, {"A-polynomial", Knot[8, 9], 
        M\^30 + L\^12\ M\^30 + 
          L\ \((\(-3\)\ M\^22 + 8\ M\^24 - 8\ M\^26 - 2\ M\^28 + 22\ M\^30 - 
                2\ M\^32 - 8\ M\^34 + 8\ M\^36 - 3\ M\^38)\) + 
          L\^11\ \((\(-3\)\ M\^22 + 8\ M\^24 - 8\ M\^26 - 2\ M\^28 + 
                22\ M\^30 - 2\ M\^32 - 8\ M\^34 + 8\ M\^36 - 3\ M\^38)\) + 
          L\^2\ \((3\ M\^14 - 14\ M\^16 + 26\ M\^18 - 15\ M\^20 - 26\ M\^22 + 
                29\ M\^24 - 3\ M\^26 + 66\ M\^30 - 3\ M\^34 + 29\ M\^36 - 
                26\ M\^38 - 15\ M\^40 + 26\ M\^42 - 14\ M\^44 + 3\ M\^46)\) + 
          L\^10\ \((3\ M\^14 - 14\ M\^16 + 26\ M\^18 - 15\ M\^20 - 
                26\ M\^22 + 29\ M\^24 - 3\ M\^26 + 66\ M\^30 - 3\ M\^34 + 
                29\ M\^36 - 26\ M\^38 - 15\ M\^40 + 26\ M\^42 - 14\ M\^44 + 
                3\ M\^46)\) + 
          L\^3\ \((\(-M\^6\) + 6\ M\^8 - 15\ M\^10 + 20\ M\^12 - 32\ M\^14 + 
                78\ M\^16 - 39\ M\^18 - 220\ M\^20 + 274\ M\^22 + 
                282\ M\^24 - 722\ M\^26 - 56\ M\^28 + 1070\ M\^30 - 
                56\ M\^32 - 722\ M\^34 + 282\ M\^36 + 274\ M\^38 - 
                220\ M\^40 - 39\ M\^42 + 78\ M\^44 - 32\ M\^46 + 20\ M\^48 - 
                15\ M\^50 + 6\ M\^52 - M\^54)\) + 
          L\^9\ \((\(-M\^6\) + 6\ M\^8 - 15\ M\^10 + 20\ M\^12 - 32\ M\^14 + 
                78\ M\^16 - 39\ M\^18 - 220\ M\^20 + 274\ M\^22 + 
                282\ M\^24 - 722\ M\^26 - 56\ M\^28 + 1070\ M\^30 - 
                56\ M\^32 - 722\ M\^34 + 282\ M\^36 + 274\ M\^38 - 
                220\ M\^40 - 39\ M\^42 + 78\ M\^44 - 32\ M\^46 + 20\ M\^48 - 
                15\ M\^50 + 6\ M\^52 - M\^54)\) + 
          L\^4\ \((\(-3\)\ M\^4 + 25\ M\^6 - 78\ M\^8 + 79\ M\^10 + 
                79\ M\^12 - 157\ M\^14 - 153\ M\^16 + 303\ M\^18 + 
                180\ M\^20 - 238\ M\^22 - 379\ M\^24 - 38\ M\^26 + 
                354\ M\^28 + 547\ M\^30 + 354\ M\^32 - 38\ M\^34 - 
                379\ M\^36 - 238\ M\^38 + 180\ M\^40 + 303\ M\^42 - 
                153\ M\^44 - 157\ M\^46 + 79\ M\^48 + 79\ M\^50 - 78\ M\^52 + 
                25\ M\^54 - 3\ M\^56)\) + 
          L\^8\ \((\(-3\)\ M\^4 + 25\ M\^6 - 78\ M\^8 + 79\ M\^10 + 
                79\ M\^12 - 157\ M\^14 - 153\ M\^16 + 303\ M\^18 + 
                180\ M\^20 - 238\ M\^22 - 379\ M\^24 - 38\ M\^26 + 
                354\ M\^28 + 547\ M\^30 + 354\ M\^32 - 38\ M\^34 - 
                379\ M\^36 - 238\ M\^38 + 180\ M\^40 + 303\ M\^42 - 
                153\ M\^44 - 157\ M\^46 + 79\ M\^48 + 79\ M\^50 - 78\ M\^52 + 
                25\ M\^54 - 3\ M\^56)\) + 
          L\^5\ \((\(-3\)\ M\^2 + 32\ M\^4 - 121\ M\^6 + 154\ M\^8 + 
                147\ M\^10 - 460\ M\^12 - 183\ M\^14 + 1050\ M\^16 - 
                175\ M\^18 - 1214\ M\^20 + 783\ M\^22 + 748\ M\^24 - 
                2001\ M\^26 + 86\ M\^28 + 3106\ M\^30 + 86\ M\^32 - 
                2001\ M\^34 + 748\ M\^36 + 783\ M\^38 - 1214\ M\^40 - 
                175\ M\^42 + 1050\ M\^44 - 183\ M\^46 - 460\ M\^48 + 
                147\ M\^50 + 154\ M\^52 - 121\ M\^54 + 32\ M\^56 - 
                3\ M\^58)\) + 
          L\^7\ \((\(-3\)\ M\^2 + 32\ M\^4 - 121\ M\^6 + 154\ M\^8 + 
                147\ M\^10 - 460\ M\^12 - 183\ M\^14 + 1050\ M\^16 - 
                175\ M\^18 - 1214\ M\^20 + 783\ M\^22 + 748\ M\^24 - 
                2001\ M\^26 + 86\ M\^28 + 3106\ M\^30 + 86\ M\^32 - 
                2001\ M\^34 + 748\ M\^36 + 783\ M\^38 - 1214\ M\^40 - 
                175\ M\^42 + 1050\ M\^44 - 183\ M\^46 - 460\ M\^48 + 
                147\ M\^50 + 154\ M\^52 - 121\ M\^54 + 32\ M\^56 - 
                3\ M\^58)\) + 
          L\^6\ \((\(-1\) + 13\ M\^2 - 61\ M\^4 + 125\ M\^6 - 68\ M\^8 - 
                134\ M\^10 + 167\ M\^12 + 94\ M\^14 - 395\ M\^16 + 
                313\ M\^18 + 408\ M\^20 - 285\ M\^22 - 715\ M\^24 - 
                558\ M\^26 + 665\ M\^28 + 1788\ M\^30 + 665\ M\^32 - 
                558\ M\^34 - 715\ M\^36 - 285\ M\^38 + 408\ M\^40 + 
                313\ M\^42 - 395\ M\^44 + 94\ M\^46 + 167\ M\^48 - 
                134\ M\^50 - 68\ M\^52 + 125\ M\^54 - 61\ M\^56 + 13\ M\^58 - 
                M\^60)\)}, {"A-polynomial", Knot[8, 10], 
        M\^14 + L\^11\ M\^60 + 
          L\ \((2\ M\^8 - 7\ M\^10 + 15\ M\^12 - 11\ M\^14 - 14\ M\^16 + 
                31\ M\^18 + 2\ M\^20 - 11\ M\^22 + 4\ M\^24)\) + 
          L\^2\ \((M\^2 - 6\ M\^4 + 19\ M\^6 - 34\ M\^8 + 30\ M\^10 - 
                52\ M\^14 + 121\ M\^16 - 33\ M\^18 - 205\ M\^20 + 
                201\ M\^22 + 147\ M\^24 - 192\ M\^26 + 29\ M\^28 + 
                57\ M\^30 - 34\ M\^32 + 6\ M\^34)\) + 
          L\^3\ \((1 - 8\ M\^2 + 31\ M\^4 - 75\ M\^6 + 88\ M\^8 + 33\ M\^10 - 
                253\ M\^12 + 295\ M\^14 + 176\ M\^16 - 903\ M\^18 + 
                626\ M\^20 + 1016\ M\^22 - 1517\ M\^24 - 197\ M\^26 + 
                1520\ M\^28 - 501\ M\^30 - 601\ M\^32 + 607\ M\^34 - 
                103\ M\^36 - 159\ M\^38 + 121\ M\^40 - 36\ M\^42 + 
                4\ M\^44)\) + 
          L\^4\ \((\(-1\) + 11\ M\^2 - 49\ M\^4 + 98\ M\^6 - 23\ M\^8 - 
                246\ M\^10 + 310\ M\^12 + 355\ M\^14 - 1279\ M\^16 + 
                605\ M\^18 + 2265\ M\^20 - 2971\ M\^22 - 1821\ M\^24 + 
                5219\ M\^26 - 466\ M\^28 - 4704\ M\^30 + 2950\ M\^32 + 
                1701\ M\^34 - 2716\ M\^36 + 852\ M\^38 + 1052\ M\^40 - 
                1133\ M\^42 + 190\ M\^44 + 295\ M\^46 - 230\ M\^48 + 
                79\ M\^50 - 14\ M\^52 + M\^54)\) + 
          L\^5\ \((\(-2\)\ M\^6 + 27\ M\^8 - 130\ M\^10 + 253\ M\^12 - 
                41\ M\^14 - 597\ M\^16 + 826\ M\^18 + 260\ M\^20 - 
                2174\ M\^22 + 1852\ M\^24 + 2370\ M\^26 - 4216\ M\^28 - 
                206\ M\^30 + 3447\ M\^32 - 1998\ M\^34 + 947\ M\^36 + 
                2159\ M\^38 - 4549\ M\^40 + 339\ M\^42 + 4196\ M\^44 - 
                1842\ M\^46 - 1610\ M\^48 + 1420\ M\^50 - 55\ M\^52 - 
                393\ M\^54 + 242\ M\^56 - 75\ M\^58 + 13\ M\^60 - M\^62)\) + 
          L\^10\ \((4\ M\^50 - 11\ M\^52 + 2\ M\^54 + 31\ M\^56 - 14\ M\^58 - 
                11\ M\^60 + 15\ M\^62 - 7\ M\^64 + 2\ M\^66)\) + 
          L\^6\ \((\(-M\^12\) + 13\ M\^14 - 75\ M\^16 + 242\ M\^18 - 
                393\ M\^20 - 55\ M\^22 + 1420\ M\^24 - 1610\ M\^26 - 
                1842\ M\^28 + 4196\ M\^30 + 339\ M\^32 - 4549\ M\^34 + 
                2159\ M\^36 + 947\ M\^38 - 1998\ M\^40 + 3447\ M\^42 - 
                206\ M\^44 - 4216\ M\^46 + 2370\ M\^48 + 1852\ M\^50 - 
                2174\ M\^52 + 260\ M\^54 + 826\ M\^56 - 597\ M\^58 - 
                41\ M\^60 + 253\ M\^62 - 130\ M\^64 + 27\ M\^66 - 
                2\ M\^68)\) + 
          L\^9\ \((6\ M\^40 - 34\ M\^42 + 57\ M\^44 + 29\ M\^46 - 
                192\ M\^48 + 147\ M\^50 + 201\ M\^52 - 205\ M\^54 - 
                33\ M\^56 + 121\ M\^58 - 52\ M\^60 + 30\ M\^64 - 34\ M\^66 + 
                19\ M\^68 - 6\ M\^70 + M\^72)\) + 
          L\^7\ \((M\^20 - 14\ M\^22 + 79\ M\^24 - 230\ M\^26 + 295\ M\^28 + 
                190\ M\^30 - 1133\ M\^32 + 1052\ M\^34 + 852\ M\^36 - 
                2716\ M\^38 + 1701\ M\^40 + 2950\ M\^42 - 4704\ M\^44 - 
                466\ M\^46 + 5219\ M\^48 - 1821\ M\^50 - 2971\ M\^52 + 
                2265\ M\^54 + 605\ M\^56 - 1279\ M\^58 + 355\ M\^60 + 
                310\ M\^62 - 246\ M\^64 - 23\ M\^66 + 98\ M\^68 - 49\ M\^70 + 
                11\ M\^72 - M\^74)\) + 
          L\^8\ \((4\ M\^30 - 36\ M\^32 + 121\ M\^34 - 159\ M\^36 - 
                103\ M\^38 + 607\ M\^40 - 601\ M\^42 - 501\ M\^44 + 
                1520\ M\^46 - 197\ M\^48 - 1517\ M\^50 + 1016\ M\^52 + 
                626\ M\^54 - 903\ M\^56 + 176\ M\^58 + 295\ M\^60 - 
                253\ M\^62 + 33\ M\^64 + 88\ M\^66 - 75\ M\^68 + 31\ M\^70 - 
                8\ M\^72 + M\^74)\)}, {"A-polynomial", Knot[8, 11], 
        L\^11\ M\^12 + M\^54 + 
          L\^10\ \((\(-3\)\ M\^8 + 10\ M\^10 - 11\ M\^12 - 5\ M\^14 + 
                28\ M\^16 - 3\ M\^18 - 11\ M\^20 + 9\ M\^22 - 3\ M\^24)\) + 
          L\^9\ \((3\ M\^4 - 20\ M\^6 + 48\ M\^8 - 25\ M\^10 - 79\ M\^12 + 
                103\ M\^14 + 20\ M\^16 - 100\ M\^18 + 93\ M\^20 + 41\ M\^22 - 
                34\ M\^24 + 35\ M\^26 - 32\ M\^28 - 18\ M\^30 + 33\ M\^32 - 
                16\ M\^34 + 3\ M\^36)\) + 
          L\^8\ \((\(-1\) + 10\ M\^2 - 39\ M\^4 + 67\ M\^6 - 27\ M\^8 - 
                81\ M\^10 + 154\ M\^12 - 108\ M\^14 - 124\ M\^16 + 
                432\ M\^18 - 369\ M\^20 - 419\ M\^22 + 977\ M\^24 + 
                172\ M\^26 - 821\ M\^28 + 237\ M\^30 + 318\ M\^32 - 
                242\ M\^34 - 16\ M\^36 + 61\ M\^38 - 30\ M\^40 + 29\ M\^42 - 
                21\ M\^44 + 7\ M\^46 - M\^48)\) + 
          L\^7\ \((2 - 21\ M\^2 + 87\ M\^4 - 141\ M\^6 - 91\ M\^8 + 
                579\ M\^10 - 202\ M\^12 - 1403\ M\^14 + 1171\ M\^16 + 
                2142\ M\^18 - 2370\ M\^20 - 2446\ M\^22 + 2936\ M\^24 + 
                2057\ M\^26 - 2381\ M\^28 - 1001\ M\^30 + 2032\ M\^32 + 
                502\ M\^34 - 1608\ M\^36 - 242\ M\^38 + 1091\ M\^40 - 
                76\ M\^42 - 489\ M\^44 + 156\ M\^46 + 123\ M\^48 - 
                103\ M\^50 + 29\ M\^52 - 3\ M\^54)\) + 
          L\ \((\(-3\)\ M\^42 + 9\ M\^44 - 11\ M\^46 - 3\ M\^48 + 28\ M\^50 - 
                5\ M\^52 - 11\ M\^54 + 10\ M\^56 - 3\ M\^58)\) + 
          L\^6\ \((\(-1\) + 11\ M\^2 - 49\ M\^4 + 90\ M\^6 + 19\ M\^8 - 
                286\ M\^10 + 58\ M\^12 + 903\ M\^14 - 427\ M\^16 - 
                2136\ M\^18 + 855\ M\^20 + 4252\ M\^22 - 783\ M\^24 - 
                6431\ M\^26 - 970\ M\^28 + 7391\ M\^30 + 3909\ M\^32 - 
                6169\ M\^34 - 4818\ M\^36 + 4955\ M\^38 + 3567\ M\^40 - 
                3742\ M\^42 - 1559\ M\^44 + 2296\ M\^46 + 123\ M\^48 - 
                890\ M\^50 + 210\ M\^52 + 185\ M\^54 - 131\ M\^56 + 
                33\ M\^58 - 3\ M\^60)\) + 
          L\^2\ \((3\ M\^30 - 16\ M\^32 + 33\ M\^34 - 18\ M\^36 - 32\ M\^38 + 
                35\ M\^40 - 34\ M\^42 + 41\ M\^44 + 93\ M\^46 - 100\ M\^48 + 
                20\ M\^50 + 103\ M\^52 - 79\ M\^54 - 25\ M\^56 + 48\ M\^58 - 
                20\ M\^60 + 3\ M\^62)\) + 
          L\^3\ \((\(-M\^18\) + 7\ M\^20 - 21\ M\^22 + 29\ M\^24 - 
                30\ M\^26 + 61\ M\^28 - 16\ M\^30 - 242\ M\^32 + 318\ M\^34 + 
                237\ M\^36 - 821\ M\^38 + 172\ M\^40 + 977\ M\^42 - 
                419\ M\^44 - 369\ M\^46 + 432\ M\^48 - 124\ M\^50 - 
                108\ M\^52 + 154\ M\^54 - 81\ M\^56 - 27\ M\^58 + 67\ M\^60 - 
                39\ M\^62 + 10\ M\^64 - M\^66)\) + 
          L\^5\ \((\(-3\)\ M\^6 + 33\ M\^8 - 131\ M\^10 + 185\ M\^12 + 
                210\ M\^14 - 890\ M\^16 + 123\ M\^18 + 2296\ M\^20 - 
                1559\ M\^22 - 3742\ M\^24 + 3567\ M\^26 + 4955\ M\^28 - 
                4818\ M\^30 - 6169\ M\^32 + 3909\ M\^34 + 7391\ M\^36 - 
                970\ M\^38 - 6431\ M\^40 - 783\ M\^42 + 4252\ M\^44 + 
                855\ M\^46 - 2136\ M\^48 - 427\ M\^50 + 903\ M\^52 + 
                58\ M\^54 - 286\ M\^56 + 19\ M\^58 + 90\ M\^60 - 49\ M\^62 + 
                11\ M\^64 - M\^66)\) + 
          L\^4\ \((\(-3\)\ M\^12 + 29\ M\^14 - 103\ M\^16 + 123\ M\^18 + 
                156\ M\^20 - 489\ M\^22 - 76\ M\^24 + 1091\ M\^26 - 
                242\ M\^28 - 1608\ M\^30 + 502\ M\^32 + 2032\ M\^34 - 
                1001\ M\^36 - 2381\ M\^38 + 2057\ M\^40 + 2936\ M\^42 - 
                2446\ M\^44 - 2370\ M\^46 + 2142\ M\^48 + 1171\ M\^50 - 
                1403\ M\^52 - 202\ M\^54 + 579\ M\^56 - 91\ M\^58 - 
                141\ M\^60 + 87\ M\^62 - 21\ M\^64 + 
                2\ M\^66)\)}, {"A-polynomial", Knot[8, 12], 
        M\^36 + L\^14\ M\^36 + 
          L\ \((\(-3\)\ M\^28 + 11\ M\^30 - 15\ M\^32 - 4\ M\^34 + 
                36\ M\^36 - 4\ M\^38 - 15\ M\^40 + 11\ M\^42 - 3\ M\^44)\) + 
          L\^13\ \((\(-3\)\ M\^28 + 11\ M\^30 - 15\ M\^32 - 4\ M\^34 + 
                36\ M\^36 - 4\ M\^38 - 15\ M\^40 + 11\ M\^42 - 3\ M\^44)\) + 
          L\^2\ \((3\ M\^20 - 22\ M\^22 + 60\ M\^24 - 47\ M\^26 - 67\ M\^28 + 
                111\ M\^30 - 35\ M\^32 - 42\ M\^34 + 169\ M\^36 - 42\ M\^38 - 
                35\ M\^40 + 111\ M\^42 - 67\ M\^44 - 47\ M\^46 + 60\ M\^48 - 
                22\ M\^50 + 3\ M\^52)\) + 
          L\^12\ \((3\ M\^20 - 22\ M\^22 + 60\ M\^24 - 47\ M\^26 - 
                67\ M\^28 + 111\ M\^30 - 35\ M\^32 - 42\ M\^34 + 169\ M\^36 - 
                42\ M\^38 - 35\ M\^40 + 111\ M\^42 - 67\ M\^44 - 47\ M\^46 + 
                60\ M\^48 - 22\ M\^50 + 3\ M\^52)\) + 
          L\^3\ \((\(-M\^12\) + 11\ M\^14 - 48\ M\^16 + 103\ M\^18 - 
                120\ M\^20 + 22\ M\^22 + 283\ M\^24 - 440\ M\^26 - 
                176\ M\^28 + 754\ M\^30 - 469\ M\^32 - 268\ M\^34 + 
                1062\ M\^36 - 268\ M\^38 - 469\ M\^40 + 754\ M\^42 - 
                176\ M\^44 - 440\ M\^46 + 283\ M\^48 + 22\ M\^50 - 
                120\ M\^52 + 103\ M\^54 - 48\ M\^56 + 11\ M\^58 - M\^60)\) + 
          L\^11\ \((\(-M\^12\) + 11\ M\^14 - 48\ M\^16 + 103\ M\^18 - 
                120\ M\^20 + 22\ M\^22 + 283\ M\^24 - 440\ M\^26 - 
                176\ M\^28 + 754\ M\^30 - 469\ M\^32 - 268\ M\^34 + 
                1062\ M\^36 - 268\ M\^38 - 469\ M\^40 + 754\ M\^42 - 
                176\ M\^44 - 440\ M\^46 + 283\ M\^48 + 22\ M\^50 - 
                120\ M\^52 + 103\ M\^54 - 48\ M\^56 + 11\ M\^58 - M\^60)\) + 
          L\^4\ \((3\ M\^8 - 35\ M\^10 + 154\ M\^12 - 285\ M\^14 + 
                56\ M\^16 + 552\ M\^18 - 563\ M\^20 - 540\ M\^22 + 
                1380\ M\^24 - 280\ M\^26 - 1538\ M\^28 + 600\ M\^30 + 
                874\ M\^32 - 12\ M\^34 + 269\ M\^36 - 12\ M\^38 + 
                874\ M\^40 + 600\ M\^42 - 1538\ M\^44 - 280\ M\^46 + 
                1380\ M\^48 - 540\ M\^50 - 563\ M\^52 + 552\ M\^54 + 
                56\ M\^56 - 285\ M\^58 + 154\ M\^60 - 35\ M\^62 + 
                3\ M\^64)\) + 
          L\^10\ \((3\ M\^8 - 35\ M\^10 + 154\ M\^12 - 285\ M\^14 + 
                56\ M\^16 + 552\ M\^18 - 563\ M\^20 - 540\ M\^22 + 
                1380\ M\^24 - 280\ M\^26 - 1538\ M\^28 + 600\ M\^30 + 
                874\ M\^32 - 12\ M\^34 + 269\ M\^36 - 12\ M\^38 + 
                874\ M\^40 + 600\ M\^42 - 1538\ M\^44 - 280\ M\^46 + 
                1380\ M\^48 - 540\ M\^50 - 563\ M\^52 + 552\ M\^54 + 
                56\ M\^56 - 285\ M\^58 + 154\ M\^60 - 35\ M\^62 + 
                3\ M\^64)\) + 
          L\^5\ \((\(-3\)\ M\^4 + 37\ M\^6 - 175\ M\^8 + 365\ M\^10 - 
                168\ M\^12 - 609\ M\^14 + 869\ M\^16 + 349\ M\^18 - 
                1793\ M\^20 + 1922\ M\^22 - 6\ M\^24 - 4521\ M\^26 + 
                5689\ M\^28 + 4728\ M\^30 - 14289\ M\^32 - 1270\ M\^34 + 
                19752\ M\^36 - 1270\ M\^38 - 14289\ M\^40 + 4728\ M\^42 + 
                5689\ M\^44 - 4521\ M\^46 - 6\ M\^48 + 1922\ M\^50 - 
                1793\ M\^52 + 349\ M\^54 + 869\ M\^56 - 609\ M\^58 - 
                168\ M\^60 + 365\ M\^62 - 175\ M\^64 + 37\ M\^66 - 
                3\ M\^68)\) + 
          L\^9\ \((\(-3\)\ M\^4 + 37\ M\^6 - 175\ M\^8 + 365\ M\^10 - 
                168\ M\^12 - 609\ M\^14 + 869\ M\^16 + 349\ M\^18 - 
                1793\ M\^20 + 1922\ M\^22 - 6\ M\^24 - 4521\ M\^26 + 
                5689\ M\^28 + 4728\ M\^30 - 14289\ M\^32 - 1270\ M\^34 + 
                19752\ M\^36 - 1270\ M\^38 - 14289\ M\^40 + 4728\ M\^42 + 
                5689\ M\^44 - 4521\ M\^46 - 6\ M\^48 + 1922\ M\^50 - 
                1793\ M\^52 + 349\ M\^54 + 869\ M\^56 - 609\ M\^58 - 
                168\ M\^60 + 365\ M\^62 - 175\ M\^64 + 37\ M\^66 - 
                3\ M\^68)\) + 
          L\^7\ \((\(-2\) + 27\ M\^2 - 143\ M\^4 + 348\ M\^6 - 210\ M\^8 - 
                775\ M\^10 + 1066\ M\^12 + 2476\ M\^14 - 5483\ M\^16 - 
                4936\ M\^18 + 17933\ M\^20 + 8023\ M\^22 - 47011\ M\^24 - 
                7629\ M\^26 + 92880\ M\^28 + 3286\ M\^30 - 141994\ M\^32 + 
                896\ M\^34 + 165928\ M\^36 + 896\ M\^38 - 141994\ M\^40 + 
                3286\ M\^42 + 92880\ M\^44 - 7629\ M\^46 - 47011\ M\^48 + 
                8023\ M\^50 + 17933\ M\^52 - 4936\ M\^54 - 5483\ M\^56 + 
                2476\ M\^58 + 1066\ M\^60 - 775\ M\^62 - 210\ M\^64 + 
                348\ M\^66 - 143\ M\^68 + 27\ M\^70 - 2\ M\^72)\) + 
          L\^6\ \((1 - 13\ M\^2 + 68\ M\^4 - 188\ M\^6 + 311\ M\^8 - 
                279\ M\^10 - 314\ M\^12 + 1647\ M\^14 - 957\ M\^16 - 
                5334\ M\^18 + 8053\ M\^20 + 9216\ M\^22 - 24416\ M\^24 - 
                11061\ M\^26 + 51479\ M\^28 + 7421\ M\^30 - 80653\ M\^32 - 
                1409\ M\^34 + 95859\ M\^36 - 1409\ M\^38 - 80653\ M\^40 + 
                7421\ M\^42 + 51479\ M\^44 - 11061\ M\^46 - 24416\ M\^48 + 
                9216\ M\^50 + 8053\ M\^52 - 5334\ M\^54 - 957\ M\^56 + 
                1647\ M\^58 - 314\ M\^60 - 279\ M\^62 + 311\ M\^64 - 
                188\ M\^66 + 68\ M\^68 - 13\ M\^70 + M\^72)\) + 
          L\^8\ \((1 - 13\ M\^2 + 68\ M\^4 - 188\ M\^6 + 311\ M\^8 - 
                279\ M\^10 - 314\ M\^12 + 1647\ M\^14 - 957\ M\^16 - 
                5334\ M\^18 + 8053\ M\^20 + 9216\ M\^22 - 24416\ M\^24 - 
                11061\ M\^26 + 51479\ M\^28 + 7421\ M\^30 - 80653\ M\^32 - 
                1409\ M\^34 + 95859\ M\^36 - 1409\ M\^38 - 80653\ M\^40 + 
                7421\ M\^42 + 51479\ M\^44 - 11061\ M\^46 - 24416\ M\^48 + 
                9216\ M\^50 + 8053\ M\^52 - 5334\ M\^54 - 957\ M\^56 + 
                1647\ M\^58 - 314\ M\^60 - 279\ M\^62 + 311\ M\^64 - 
                188\ M\^66 + 68\ M\^68 - 13\ M\^70 + 
                M\^72)\)}, {"A-polynomial", Knot[8, 13], 
        M\^36 + L\^14\ M\^36 + 
          L\^13\ \((2\ M\^26 - 7\ M\^28 + 11\ M\^30 - 5\ M\^32 - 15\ M\^34 + 
                31\ M\^36 + 10\ M\^38 - 19\ M\^40 + 6\ M\^42)\) + 
          L\ \((6\ M\^30 - 19\ M\^32 + 10\ M\^34 + 31\ M\^36 - 15\ M\^38 - 
                5\ M\^40 + 11\ M\^42 - 7\ M\^44 + 2\ M\^46)\) + 
          L\^12\ \((M\^16 - 6\ M\^18 + 15\ M\^20 - 20\ M\^22 + 22\ M\^24 - 
                40\ M\^26 + 20\ M\^28 + 166\ M\^30 - 248\ M\^32 - 
                192\ M\^34 + 531\ M\^36 + 94\ M\^38 - 408\ M\^40 + 
                86\ M\^42 + 143\ M\^44 - 88\ M\^46 + 15\ M\^48)\) + 
          L\^11\ \((M\^10 - 10\ M\^12 + 43\ M\^14 - 96\ M\^16 + 69\ M\^18 + 
                131\ M\^20 - 181\ M\^22 - 263\ M\^24 + 237\ M\^26 + 
                820\ M\^28 - 195\ M\^30 - 2157\ M\^32 + 517\ M\^34 + 
                3417\ M\^36 - 889\ M\^38 - 2821\ M\^40 + 1442\ M\^42 + 
                1276\ M\^44 - 1120\ M\^46 - 135\ M\^48 + 420\ M\^50 - 
                162\ M\^52 + 20\ M\^54)\) + 
          L\^2\ \((15\ M\^24 - 88\ M\^26 + 143\ M\^28 + 86\ M\^30 - 
                408\ M\^32 + 94\ M\^34 + 531\ M\^36 - 192\ M\^38 - 
                248\ M\^40 + 166\ M\^42 + 20\ M\^44 - 40\ M\^46 + 22\ M\^48 - 
                20\ M\^50 + 15\ M\^52 - 6\ M\^54 + M\^56)\) + 
          L\^10\ \((\(-2\)\ M\^6 + 24\ M\^8 - 115\ M\^10 + 258\ M\^12 - 
                120\ M\^14 - 641\ M\^16 + 939\ M\^18 + 1095\ M\^20 - 
                3002\ M\^22 - 879\ M\^24 + 5728\ M\^26 + 619\ M\^28 - 
                8552\ M\^30 - 1018\ M\^32 + 10154\ M\^34 + 2413\ M\^36 - 
                9136\ M\^38 - 1750\ M\^40 + 7677\ M\^42 + 509\ M\^44 - 
                5226\ M\^46 + 817\ M\^48 + 2317\ M\^50 - 970\ M\^52 - 
                514\ M\^54 + 509\ M\^56 - 148\ M\^58 + 15\ M\^60)\) + 
          L\^3\ \((20\ M\^18 - 162\ M\^20 + 420\ M\^22 - 135\ M\^24 - 
                1120\ M\^26 + 1276\ M\^28 + 1442\ M\^30 - 2821\ M\^32 - 
                889\ M\^34 + 3417\ M\^36 + 517\ M\^38 - 2157\ M\^40 - 
                195\ M\^42 + 820\ M\^44 + 237\ M\^46 - 263\ M\^48 - 
                181\ M\^50 + 131\ M\^52 + 69\ M\^54 - 96\ M\^56 + 43\ M\^58 - 
                10\ M\^60 + M\^62)\) + 
          L\^4\ \((15\ M\^12 - 148\ M\^14 + 509\ M\^16 - 514\ M\^18 - 
                970\ M\^20 + 2317\ M\^22 + 817\ M\^24 - 5226\ M\^26 + 
                509\ M\^28 + 7677\ M\^30 - 1750\ M\^32 - 9136\ M\^34 + 
                2413\ M\^36 + 10154\ M\^38 - 1018\ M\^40 - 8552\ M\^42 + 
                619\ M\^44 + 5728\ M\^46 - 879\ M\^48 - 3002\ M\^50 + 
                1095\ M\^52 + 939\ M\^54 - 641\ M\^56 - 120\ M\^58 + 
                258\ M\^60 - 115\ M\^62 + 24\ M\^64 - 2\ M\^66)\) + 
          L\^9\ \((M\^2 - 16\ M\^4 + 98\ M\^6 - 288\ M\^8 + 330\ M\^10 + 
                352\ M\^12 - 1391\ M\^14 + 392\ M\^16 + 3608\ M\^18 - 
                4196\ M\^20 - 5382\ M\^22 + 11133\ M\^24 + 6576\ M\^26 - 
                20088\ M\^28 - 6778\ M\^30 + 26483\ M\^32 + 5965\ M\^34 - 
                26328\ M\^36 - 1692\ M\^38 + 22374\ M\^40 - 264\ M\^42 - 
                14320\ M\^44 + 877\ M\^46 + 6971\ M\^48 - 282\ M\^50 - 
                3135\ M\^52 + 211\ M\^54 + 1186\ M\^56 - 161\ M\^58 - 
                453\ M\^60 + 280\ M\^62 - 67\ M\^64 + 6\ M\^66)\) + 
          L\^5\ \((6\ M\^6 - 67\ M\^8 + 280\ M\^10 - 453\ M\^12 - 
                161\ M\^14 + 1186\ M\^16 + 211\ M\^18 - 3135\ M\^20 - 
                282\ M\^22 + 6971\ M\^24 + 877\ M\^26 - 14320\ M\^28 - 
                264\ M\^30 + 22374\ M\^32 - 1692\ M\^34 - 26328\ M\^36 + 
                5965\ M\^38 + 26483\ M\^40 - 6778\ M\^42 - 20088\ M\^44 + 
                6576\ M\^46 + 11133\ M\^48 - 5382\ M\^50 - 4196\ M\^52 + 
                3608\ M\^54 + 392\ M\^56 - 1391\ M\^58 + 352\ M\^60 + 
                330\ M\^62 - 288\ M\^64 + 98\ M\^66 - 16\ M\^68 + M\^70)\) + 
          L\^7\ \((\(-2\) + 26\ M\^2 - 139\ M\^4 + 351\ M\^6 - 232\ M\^8 - 
                768\ M\^10 + 1076\ M\^12 + 2272\ M\^14 - 4643\ M\^16 - 
                4160\ M\^18 + 12366\ M\^20 + 6882\ M\^22 - 27465\ M\^24 - 
                7544\ M\^26 + 47282\ M\^28 + 5444\ M\^30 - 66058\ M\^32 - 
                787\ M\^34 + 75630\ M\^36 - 787\ M\^38 - 66058\ M\^40 + 
                5444\ M\^42 + 47282\ M\^44 - 7544\ M\^46 - 27465\ M\^48 + 
                6882\ M\^50 + 12366\ M\^52 - 4160\ M\^54 - 4643\ M\^56 + 
                2272\ M\^58 + 1076\ M\^60 - 768\ M\^62 - 232\ M\^64 + 
                351\ M\^66 - 139\ M\^68 + 26\ M\^70 - 2\ M\^72)\) + 
          L\^6\ \((1 - 12\ M\^2 + 58\ M\^4 - 130\ M\^6 + 113\ M\^8 - 
                96\ M\^10 + 613\ M\^12 - 806\ M\^14 - 2059\ M\^16 + 
                4317\ M\^18 + 4419\ M\^20 - 13635\ M\^22 - 4701\ M\^24 + 
                27191\ M\^26 + 2353\ M\^28 - 42112\ M\^30 + 2505\ M\^32 + 
                52732\ M\^34 - 6703\ M\^36 - 52559\ M\^38 + 13798\ M\^40 + 
                46426\ M\^42 - 16089\ M\^44 - 33470\ M\^46 + 14529\ M\^48 + 
                18153\ M\^50 - 10122\ M\^52 - 6945\ M\^54 + 5681\ M\^56 + 
                1027\ M\^58 - 1961\ M\^60 + 248\ M\^62 + 472\ M\^64 - 
                314\ M\^66 + 95\ M\^68 - 15\ M\^70 + M\^72)\) + 
          L\^8\ \((1 - 15\ M\^2 + 95\ M\^4 - 314\ M\^6 + 472\ M\^8 + 
                248\ M\^10 - 1961\ M\^12 + 1027\ M\^14 + 5681\ M\^16 - 
                6945\ M\^18 - 10122\ M\^20 + 18153\ M\^22 + 14529\ M\^24 - 
                33470\ M\^26 - 16089\ M\^28 + 46426\ M\^30 + 13798\ M\^32 - 
                52559\ M\^34 - 6703\ M\^36 + 52732\ M\^38 + 2505\ M\^40 - 
                42112\ M\^42 + 2353\ M\^44 + 27191\ M\^46 - 4701\ M\^48 - 
                13635\ M\^50 + 4419\ M\^52 + 4317\ M\^54 - 2059\ M\^56 - 
                806\ M\^58 + 613\ M\^60 - 96\ M\^62 + 113\ M\^64 - 
                130\ M\^66 + 58\ M\^68 - 12\ M\^70 + 
                M\^72)\)}, {"A-polynomial", Knot[8, 14], 
        M\^12 + L\^15\ M\^74 + 
          L\ \((\(-3\)\ M\^8 + 12\ M\^10 - 17\ M\^12 - 3\ M\^14 + 39\ M\^16 - 
                8\ M\^18 - 15\ M\^20 + 14\ M\^22 - 4\ M\^24)\) + 
          L\^2\ \((3\ M\^4 - 22\ M\^6 + 59\ M\^8 - 56\ M\^10 - 21\ M\^12 + 
                69\ M\^14 - 108\ M\^16 + 125\ M\^18 + 155\ M\^20 - 
                249\ M\^22 + 99\ M\^24 + 219\ M\^26 - 187\ M\^28 - 
                46\ M\^30 + 99\ M\^32 - 40\ M\^34 + 6\ M\^36)\) + 
          L\^3\ \((\(-1\) + 10\ M\^2 - 42\ M\^4 + 108\ M\^6 - 212\ M\^8 + 
                203\ M\^10 + 359\ M\^12 - 987\ M\^14 - 104\ M\^16 + 
                1754\ M\^18 - 525\ M\^20 - 1703\ M\^22 + 843\ M\^24 + 
                1449\ M\^26 - 29\ M\^28 - 795\ M\^30 - 395\ M\^32 + 
                655\ M\^34 + 177\ M\^36 - 477\ M\^38 + 73\ M\^40 + 
                200\ M\^42 - 140\ M\^44 + 38\ M\^46 - 4\ M\^48)\) + 
          L\^4\ \((3 - 37\ M\^2 + 177\ M\^4 - 375\ M\^6 + 199\ M\^8 + 
                445\ M\^10 - 530\ M\^12 + 97\ M\^14 - 80\ M\^16 - 
                1740\ M\^18 + 3481\ M\^20 + 2628\ M\^22 - 8635\ M\^24 - 
                1213\ M\^26 + 12078\ M\^28 - 1385\ M\^30 - 8710\ M\^32 + 
                4666\ M\^34 + 3664\ M\^36 - 4281\ M\^38 - 190\ M\^40 + 
                1570\ M\^42 - 353\ M\^44 - 152\ M\^46 + 35\ M\^48 - 
                71\ M\^50 + 167\ M\^52 - 140\ M\^54 + 58\ M\^56 - 12\ M\^58 + 
                M\^60)\) + 
          L\^5\ \((\(-3\) + 43\ M\^2 - 248\ M\^4 + 671\ M\^6 - 526\ M\^8 - 
                1430\ M\^10 + 2705\ M\^12 + 2531\ M\^14 - 8259\ M\^16 - 
                1821\ M\^18 + 15481\ M\^20 - 1339\ M\^22 - 21247\ M\^24 + 
                7720\ M\^26 + 20156\ M\^28 - 13923\ M\^30 - 12925\ M\^32 + 
                16830\ M\^34 + 8661\ M\^36 - 8748\ M\^38 - 4646\ M\^40 + 
                1311\ M\^42 + 2608\ M\^44 + 820\ M\^46 - 2024\ M\^48 + 
                378\ M\^50 + 530\ M\^52 - 475\ M\^54 + 148\ M\^56 + 
                199\ M\^58 - 332\ M\^60 + 222\ M\^62 - 78\ M\^64 + 
                14\ M\^66 - M\^68)\) + 
          L\^6\ \((1 - 16\ M\^2 + 109\ M\^4 - 405\ M\^6 + 819\ M\^8 - 
                483\ M\^10 - 1688\ M\^12 + 3439\ M\^14 + 1706\ M\^16 - 
                11184\ M\^18 + 4654\ M\^20 + 20523\ M\^22 - 18998\ M\^24 - 
                28928\ M\^26 + 40319\ M\^28 + 33615\ M\^30 - 58438\ M\^32 - 
                37681\ M\^34 + 62452\ M\^36 + 40302\ M\^38 - 43222\ M\^40 - 
                26536\ M\^42 + 23612\ M\^44 + 10137\ M\^46 - 8538\ M\^48 - 
                1893\ M\^50 - 188\ M\^52 + 500\ M\^54 + 3287\ M\^56 - 
                2079\ M\^58 - 1479\ M\^60 + 1560\ M\^62 + 151\ M\^64 - 
                751\ M\^66 + 429\ M\^68 - 119\ M\^70 + 17\ M\^72 - M\^74)\) + 
          L\^14\ \((\(-4\)\ M\^62 + 14\ M\^64 - 15\ M\^66 - 8\ M\^68 + 
                39\ M\^70 - 3\ M\^72 - 17\ M\^74 + 12\ M\^76 - 3\ M\^78)\) + 
          L\^7\ \((\(-3\)\ M\^4 + 47\ M\^6 - 303\ M\^8 + 978\ M\^10 - 
                1269\ M\^12 - 1456\ M\^14 + 6376\ M\^16 - 1023\ M\^18 - 
                19764\ M\^20 + 15925\ M\^22 + 40777\ M\^24 - 48584\ M\^26 - 
                70123\ M\^28 + 98204\ M\^30 + 102643\ M\^32 - 143991\ M\^34 - 
                131180\ M\^36 + 160673\ M\^38 + 139227\ M\^40 - 
                139158\ M\^42 - 113292\ M\^44 + 112077\ M\^46 + 
                81848\ M\^48 - 82687\ M\^50 - 53994\ M\^52 + 55829\ M\^54 + 
                28995\ M\^56 - 33079\ M\^58 - 10817\ M\^60 + 17004\ M\^62 + 
                461\ M\^64 - 5859\ M\^66 + 1433\ M\^68 + 1230\ M\^70 - 
                968\ M\^72 + 302\ M\^74 - 47\ M\^76 + 3\ M\^78)\) + 
          L\^8\ \((3\ M\^8 - 47\ M\^10 + 302\ M\^12 - 968\ M\^14 + 
                1230\ M\^16 + 1433\ M\^18 - 5859\ M\^20 + 461\ M\^22 + 
                17004\ M\^24 - 10817\ M\^26 - 33079\ M\^28 + 28995\ M\^30 + 
                55829\ M\^32 - 53994\ M\^34 - 82687\ M\^36 + 81848\ M\^38 + 
                112077\ M\^40 - 113292\ M\^42 - 139158\ M\^44 + 
                139227\ M\^46 + 160673\ M\^48 - 131180\ M\^50 - 
                143991\ M\^52 + 102643\ M\^54 + 98204\ M\^56 - 70123\ M\^58 - 
                48584\ M\^60 + 40777\ M\^62 + 15925\ M\^64 - 19764\ M\^66 - 
                1023\ M\^68 + 6376\ M\^70 - 1456\ M\^72 - 1269\ M\^74 + 
                978\ M\^76 - 303\ M\^78 + 47\ M\^80 - 3\ M\^82)\) + 
          L\^13\ \((6\ M\^50 - 40\ M\^52 + 99\ M\^54 - 46\ M\^56 - 
                187\ M\^58 + 219\ M\^60 + 99\ M\^62 - 249\ M\^64 + 
                155\ M\^66 + 125\ M\^68 - 108\ M\^70 + 69\ M\^72 - 
                21\ M\^74 - 56\ M\^76 + 59\ M\^78 - 22\ M\^80 + 3\ M\^82)\) + 
          L\^10\ \((\(-M\^18\) + 14\ M\^20 - 78\ M\^22 + 222\ M\^24 - 
                332\ M\^26 + 199\ M\^28 + 148\ M\^30 - 475\ M\^32 + 
                530\ M\^34 + 378\ M\^36 - 2024\ M\^38 + 820\ M\^40 + 
                2608\ M\^42 + 1311\ M\^44 - 4646\ M\^46 - 8748\ M\^48 + 
                8661\ M\^50 + 16830\ M\^52 - 12925\ M\^54 - 13923\ M\^56 + 
                20156\ M\^58 + 7720\ M\^60 - 21247\ M\^62 - 1339\ M\^64 + 
                15481\ M\^66 - 1821\ M\^68 - 8259\ M\^70 + 2531\ M\^72 + 
                2705\ M\^74 - 1430\ M\^76 - 526\ M\^78 + 671\ M\^80 - 
                248\ M\^82 + 43\ M\^84 - 3\ M\^86)\) + 
          L\^12\ \((\(-4\)\ M\^38 + 38\ M\^40 - 140\ M\^42 + 200\ M\^44 + 
                73\ M\^46 - 477\ M\^48 + 177\ M\^50 + 655\ M\^52 - 
                395\ M\^54 - 795\ M\^56 - 29\ M\^58 + 1449\ M\^60 + 
                843\ M\^62 - 1703\ M\^64 - 525\ M\^66 + 1754\ M\^68 - 
                104\ M\^70 - 987\ M\^72 + 359\ M\^74 + 203\ M\^76 - 
                212\ M\^78 + 108\ M\^80 - 42\ M\^82 + 10\ M\^84 - M\^86)\) + 
          L\^9\ \((\(-M\^12\) + 17\ M\^14 - 119\ M\^16 + 429\ M\^18 - 
                751\ M\^20 + 151\ M\^22 + 1560\ M\^24 - 1479\ M\^26 - 
                2079\ M\^28 + 3287\ M\^30 + 500\ M\^32 - 188\ M\^34 - 
                1893\ M\^36 - 8538\ M\^38 + 10137\ M\^40 + 23612\ M\^42 - 
                26536\ M\^44 - 43222\ M\^46 + 40302\ M\^48 + 62452\ M\^50 - 
                37681\ M\^52 - 58438\ M\^54 + 33615\ M\^56 + 40319\ M\^58 - 
                28928\ M\^60 - 18998\ M\^62 + 20523\ M\^64 + 4654\ M\^66 - 
                11184\ M\^68 + 1706\ M\^70 + 3439\ M\^72 - 1688\ M\^74 - 
                483\ M\^76 + 819\ M\^78 - 405\ M\^80 + 109\ M\^82 - 
                16\ M\^84 + M\^86)\) + 
          L\^11\ \((M\^26 - 12\ M\^28 + 58\ M\^30 - 140\ M\^32 + 167\ M\^34 - 
                71\ M\^36 + 35\ M\^38 - 152\ M\^40 - 353\ M\^42 + 
                1570\ M\^44 - 190\ M\^46 - 4281\ M\^48 + 3664\ M\^50 + 
                4666\ M\^52 - 8710\ M\^54 - 1385\ M\^56 + 12078\ M\^58 - 
                1213\ M\^60 - 8635\ M\^62 + 2628\ M\^64 + 3481\ M\^66 - 
                1740\ M\^68 - 80\ M\^70 + 97\ M\^72 - 530\ M\^74 + 
                445\ M\^76 + 199\ M\^78 - 375\ M\^80 + 177\ M\^82 - 
                37\ M\^84 + 3\ M\^86)\)}, {"A-polynomial", Knot[8, 15], 
        L\^12 + M\^92 + 
          L\^11\ \((\(-2\) + 10\ M\^2 - 12\ M\^4 - 18\ M\^6 + 47\ M\^8 + 
                4\ M\^10 - 28\ M\^12 + 14\ M\^14 - 3\ M\^16)\) + 
          L\^10\ \((1 - 11\ M\^2 + 47\ M\^4 - 86\ M\^6 + 14\ M\^8 + 
                234\ M\^10 - 363\ M\^12 - 81\ M\^14 + 564\ M\^16 - 
                116\ M\^18 - 296\ M\^20 + 207\ M\^22 - M\^24 - 103\ M\^26 + 
                77\ M\^28 - 24\ M\^30 + 3\ M\^32)\) + 
          L\^9\ \((M\^2 - 14\ M\^4 + 81\ M\^6 - 231\ M\^8 + 262\ M\^10 + 
                231\ M\^12 - 1024\ M\^14 + 892\ M\^16 + 905\ M\^18 - 
                2415\ M\^20 + 169\ M\^22 + 3088\ M\^24 - 1082\ M\^26 - 
                1720\ M\^28 + 1252\ M\^30 + 237\ M\^32 - 678\ M\^34 + 
                234\ M\^36 + 132\ M\^38 - 169\ M\^40 + 102\ M\^42 - 
                42\ M\^44 + 10\ M\^46 - M\^48)\) + 
          L\^8\ \((\(-3\)\ M\^6 + 41\ M\^8 - 221\ M\^10 + 536\ M\^12 - 
                297\ M\^14 - 1150\ M\^16 + 1788\ M\^18 + 788\ M\^20 - 
                2756\ M\^22 + 363\ M\^24 + 1876\ M\^26 - 1883\ M\^28 + 
                470\ M\^30 + 2028\ M\^32 - 724\ M\^34 - 134\ M\^36 + 
                419\ M\^38 - 1327\ M\^40 + 72\ M\^42 + 1209\ M\^44 - 
                427\ M\^46 - 460\ M\^48 + 264\ M\^50 + 160\ M\^52 - 
                206\ M\^54 + 83\ M\^56 - 15\ M\^58 + M\^60)\) + 
          L\^7\ \((3\ M\^10 - 42\ M\^12 + 231\ M\^14 - 570\ M\^16 + 
                317\ M\^18 + 1244\ M\^20 - 1641\ M\^22 - 2006\ M\^24 + 
                3371\ M\^26 + 3007\ M\^28 - 3822\ M\^30 - 4174\ M\^32 + 
                906\ M\^34 + 4711\ M\^36 + 2581\ M\^38 - 2505\ M\^40 - 
                3469\ M\^42 + 2052\ M\^44 + 3346\ M\^46 - 1857\ M\^48 - 
                3096\ M\^50 + 1481\ M\^52 + 1999\ M\^54 - 1065\ M\^56 - 
                677\ M\^58 + 427\ M\^60 + 225\ M\^62 - 267\ M\^64 + 
                97\ M\^66 - 16\ M\^68 + M\^70)\) + 
          L\^6\ \((\(-M\^14\) + 14\ M\^16 - 80\ M\^18 + 231\ M\^20 - 
                312\ M\^22 + 2\ M\^24 + 508\ M\^26 - 224\ M\^28 - 
                852\ M\^30 + 595\ M\^32 + 825\ M\^34 + 3\ M\^36 - 
                359\ M\^38 - 1782\ M\^40 - 761\ M\^42 + 1371\ M\^44 + 
                2568\ M\^46 + 1371\ M\^48 - 761\ M\^50 - 1782\ M\^52 - 
                359\ M\^54 + 3\ M\^56 + 825\ M\^58 + 595\ M\^60 - 
                852\ M\^62 - 224\ M\^64 + 508\ M\^66 + 2\ M\^68 - 
                312\ M\^70 + 231\ M\^72 - 80\ M\^74 + 14\ M\^76 - M\^78)\) + 
          L\^5\ \((M\^22 - 16\ M\^24 + 97\ M\^26 - 267\ M\^28 + 225\ M\^30 + 
                427\ M\^32 - 677\ M\^34 - 1065\ M\^36 + 1999\ M\^38 + 
                1481\ M\^40 - 3096\ M\^42 - 1857\ M\^44 + 3346\ M\^46 + 
                2052\ M\^48 - 3469\ M\^50 - 2505\ M\^52 + 2581\ M\^54 + 
                4711\ M\^56 + 906\ M\^58 - 4174\ M\^60 - 3822\ M\^62 + 
                3007\ M\^64 + 3371\ M\^66 - 2006\ M\^68 - 1641\ M\^70 + 
                1244\ M\^72 + 317\ M\^74 - 570\ M\^76 + 231\ M\^78 - 
                42\ M\^80 + 3\ M\^82)\) + 
          L\^4\ \((M\^32 - 15\ M\^34 + 83\ M\^36 - 206\ M\^38 + 160\ M\^40 + 
                264\ M\^42 - 460\ M\^44 - 427\ M\^46 + 1209\ M\^48 + 
                72\ M\^50 - 1327\ M\^52 + 419\ M\^54 - 134\ M\^56 - 
                724\ M\^58 + 2028\ M\^60 + 470\ M\^62 - 1883\ M\^64 + 
                1876\ M\^66 + 363\ M\^68 - 2756\ M\^70 + 788\ M\^72 + 
                1788\ M\^74 - 1150\ M\^76 - 297\ M\^78 + 536\ M\^80 - 
                221\ M\^82 + 41\ M\^84 - 3\ M\^86)\) + 
          L\^3\ \((\(-M\^44\) + 10\ M\^46 - 42\ M\^48 + 102\ M\^50 - 
                169\ M\^52 + 132\ M\^54 + 234\ M\^56 - 678\ M\^58 + 
                237\ M\^60 + 1252\ M\^62 - 1720\ M\^64 - 1082\ M\^66 + 
                3088\ M\^68 + 169\ M\^70 - 2415\ M\^72 + 905\ M\^74 + 
                892\ M\^76 - 1024\ M\^78 + 231\ M\^80 + 262\ M\^82 - 
                231\ M\^84 + 81\ M\^86 - 14\ M\^88 + M\^90)\) + 
          L\ \((\(-3\)\ M\^76 + 14\ M\^78 - 28\ M\^80 + 4\ M\^82 + 
                47\ M\^84 - 18\ M\^86 - 12\ M\^88 + 10\ M\^90 - 2\ M\^92)\) + 
          L\^2\ \((3\ M\^60 - 24\ M\^62 + 77\ M\^64 - 103\ M\^66 - M\^68 + 
                207\ M\^70 - 296\ M\^72 - 116\ M\^74 + 564\ M\^76 - 
                81\ M\^78 - 363\ M\^80 + 234\ M\^82 + 14\ M\^84 - 86\ M\^86 + 
                47\ M\^88 - 11\ M\^90 + M\^92)\)}, {"A-polynomial", 
        Knot[8, 16], 
        M\^14 - 2\ M\^18 + M\^22 + 
          L\ \((2\ M\^8 - 8\ M\^10 + 11\ M\^12 + 15\ M\^14 - 65\ M\^16 + 
                39\ M\^18 + 91\ M\^20 - 103\ M\^22 - 43\ M\^24 + 70\ M\^26 + 
                7\ M\^28 - 15\ M\^30 - 4\ M\^32 + 3\ M\^34 + 2\ M\^36 - 
                M\^38)\) + 
          L\^2\ \((M\^2 - 8\ M\^4 + 30\ M\^6 - 52\ M\^8 + 4\ M\^10 + 
                144\ M\^12 - 215\ M\^14 - 30\ M\^16 + 405\ M\^18 - 
                260\ M\^20 - 488\ M\^22 + 698\ M\^24 + 357\ M\^26 - 
                1067\ M\^28 + 30\ M\^30 + 901\ M\^32 - 266\ M\^34 - 
                448\ M\^36 + 247\ M\^38 + 133\ M\^40 - 115\ M\^42 - 
                7\ M\^44 + 21\ M\^46 - 4\ M\^48)\) + 
          L\^3\ \((1 - 11\ M\^2 + 50\ M\^4 - 103\ M\^6 + 21\ M\^8 + 
                276\ M\^10 - 240\ M\^12 - 640\ M\^14 + 810\ M\^16 + 
                1410\ M\^18 - 2250\ M\^20 - 2070\ M\^22 + 3801\ M\^24 + 
                2691\ M\^26 - 4354\ M\^28 - 3289\ M\^30 + 3311\ M\^32 + 
                3513\ M\^34 - 1210\ M\^36 - 3502\ M\^38 - 124\ M\^40 + 
                2921\ M\^42 + 320\ M\^44 - 1740\ M\^46 - 3\ M\^48 + 
                664\ M\^50 - 130\ M\^52 - 114\ M\^54 + 52\ M\^56 - 
                6\ M\^58)\) + 
          L\^4\ \((M\^2 - 13\ M\^4 + 77\ M\^6 - 240\ M\^8 + 297\ M\^10 + 
                356\ M\^12 - 1315\ M\^14 - 79\ M\^16 + 3365\ M\^18 - 
                784\ M\^20 - 6791\ M\^22 + 2314\ M\^24 + 10377\ M\^26 - 
                2683\ M\^28 - 12964\ M\^30 + 1113\ M\^32 + 12724\ M\^34 + 
                1337\ M\^36 - 8861\ M\^38 - 2947\ M\^40 + 3409\ M\^42 + 
                2401\ M\^44 + 1388\ M\^46 - 836\ M\^48 - 3325\ M\^50 + 
                238\ M\^52 + 2812\ M\^54 - 491\ M\^56 - 1323\ M\^58 + 
                528\ M\^60 + 244\ M\^62 - 210\ M\^64 + 50\ M\^66 - 
                4\ M\^68)\) + 
          L\^5\ \((\(-2\)\ M\^6 + 27\ M\^8 - 151\ M\^10 + 407\ M\^12 - 
                381\ M\^14 - 577\ M\^16 + 1497\ M\^18 + 32\ M\^20 - 
                2612\ M\^22 + 1375\ M\^24 + 2333\ M\^26 - 3453\ M\^28 + 
                1732\ M\^30 + 4984\ M\^32 - 9174\ M\^34 - 7199\ M\^36 + 
                17268\ M\^38 + 11348\ M\^40 - 21699\ M\^42 - 15395\ M\^44 + 
                19170\ M\^46 + 15968\ M\^48 - 11228\ M\^50 - 11165\ M\^52 + 
                3438\ M\^54 + 5220\ M\^56 + 990\ M\^58 - 1749\ M\^60 - 
                2108\ M\^62 + 880\ M\^64 + 1337\ M\^66 - 748\ M\^68 - 
                288\ M\^70 + 357\ M\^72 - 121\ M\^74 + 18\ M\^76 - M\^78)\) + 
          L\^6\ \((M\^10 - 15\ M\^12 + 94\ M\^14 - 312\ M\^16 + 542\ M\^18 - 
                166\ M\^20 - 1361\ M\^22 + 2200\ M\^24 + 1791\ M\^26 - 
                6862\ M\^28 - 1528\ M\^30 + 15381\ M\^32 + 1276\ M\^34 - 
                27149\ M\^36 - 4559\ M\^38 + 39322\ M\^40 + 13451\ M\^42 - 
                45491\ M\^44 - 24482\ M\^46 + 40138\ M\^48 + 30245\ M\^50 - 
                25687\ M\^52 - 25627\ M\^54 + 12155\ M\^56 + 15682\ M\^58 - 
                5000\ M\^60 - 6708\ M\^62 + 2726\ M\^64 + 1415\ M\^66 - 
                1815\ M\^68 + 696\ M\^70 + 821\ M\^72 - 989\ M\^74 + 
                68\ M\^76 + 428\ M\^78 - 299\ M\^80 + 94\ M\^82 - 15\ M\^84 + 
                M\^86)\) + L\^13\ \((M\^80 - 2\ M\^84 + M\^88)\) + 
          L\^7\ \((M\^16 - 15\ M\^18 + 94\ M\^20 - 299\ M\^22 + 428\ M\^24 + 
                68\ M\^26 - 989\ M\^28 + 821\ M\^30 + 696\ M\^32 - 
                1815\ M\^34 + 1415\ M\^36 + 2726\ M\^38 - 6708\ M\^40 - 
                5000\ M\^42 + 15682\ M\^44 + 12155\ M\^46 - 25627\ M\^48 - 
                25687\ M\^50 + 30245\ M\^52 + 40138\ M\^54 - 24482\ M\^56 - 
                45491\ M\^58 + 13451\ M\^60 + 39322\ M\^62 - 4559\ M\^64 - 
                27149\ M\^66 + 1276\ M\^68 + 15381\ M\^70 - 1528\ M\^72 - 
                6862\ M\^74 + 1791\ M\^76 + 2200\ M\^78 - 1361\ M\^80 - 
                166\ M\^82 + 542\ M\^84 - 312\ M\^86 + 94\ M\^88 - 
                15\ M\^90 + M\^92)\) + 
          L\^12\ \((\(-M\^64\) + 2\ M\^66 + 3\ M\^68 - 4\ M\^70 - 15\ M\^72 + 
                7\ M\^74 + 70\ M\^76 - 43\ M\^78 - 103\ M\^80 + 91\ M\^82 + 
                39\ M\^84 - 65\ M\^86 + 15\ M\^88 + 11\ M\^90 - 8\ M\^92 + 
                2\ M\^94)\) + 
          L\^8\ \((\(-M\^24\) + 18\ M\^26 - 121\ M\^28 + 357\ M\^30 - 
                288\ M\^32 - 748\ M\^34 + 1337\ M\^36 + 880\ M\^38 - 
                2108\ M\^40 - 1749\ M\^42 + 990\ M\^44 + 5220\ M\^46 + 
                3438\ M\^48 - 11165\ M\^50 - 11228\ M\^52 + 15968\ M\^54 + 
                19170\ M\^56 - 15395\ M\^58 - 21699\ M\^60 + 11348\ M\^62 + 
                17268\ M\^64 - 7199\ M\^66 - 9174\ M\^68 + 4984\ M\^70 + 
                1732\ M\^72 - 3453\ M\^74 + 2333\ M\^76 + 1375\ M\^78 - 
                2612\ M\^80 + 32\ M\^82 + 1497\ M\^84 - 577\ M\^86 - 
                381\ M\^88 + 407\ M\^90 - 151\ M\^92 + 27\ M\^94 - 
                2\ M\^96)\) + 
          L\^9\ \((\(-4\)\ M\^34 + 50\ M\^36 - 210\ M\^38 + 244\ M\^40 + 
                528\ M\^42 - 1323\ M\^44 - 491\ M\^46 + 2812\ M\^48 + 
                238\ M\^50 - 3325\ M\^52 - 836\ M\^54 + 1388\ M\^56 + 
                2401\ M\^58 + 3409\ M\^60 - 2947\ M\^62 - 8861\ M\^64 + 
                1337\ M\^66 + 12724\ M\^68 + 1113\ M\^70 - 12964\ M\^72 - 
                2683\ M\^74 + 10377\ M\^76 + 2314\ M\^78 - 6791\ M\^80 - 
                784\ M\^82 + 3365\ M\^84 - 79\ M\^86 - 1315\ M\^88 + 
                356\ M\^90 + 297\ M\^92 - 240\ M\^94 + 77\ M\^96 - 
                13\ M\^98 + M\^100)\) + 
          L\^11\ \((\(-4\)\ M\^54 + 21\ M\^56 - 7\ M\^58 - 115\ M\^60 + 
                133\ M\^62 + 247\ M\^64 - 448\ M\^66 - 266\ M\^68 + 
                901\ M\^70 + 30\ M\^72 - 1067\ M\^74 + 357\ M\^76 + 
                698\ M\^78 - 488\ M\^80 - 260\ M\^82 + 405\ M\^84 - 
                30\ M\^86 - 215\ M\^88 + 144\ M\^90 + 4\ M\^92 - 52\ M\^94 + 
                30\ M\^96 - 8\ M\^98 + M\^100)\) + 
          L\^10\ \((\(-6\)\ M\^44 + 52\ M\^46 - 114\ M\^48 - 130\ M\^50 + 
                664\ M\^52 - 3\ M\^54 - 1740\ M\^56 + 320\ M\^58 + 
                2921\ M\^60 - 124\ M\^62 - 3502\ M\^64 - 1210\ M\^66 + 
                3513\ M\^68 + 3311\ M\^70 - 3289\ M\^72 - 4354\ M\^74 + 
                2691\ M\^76 + 3801\ M\^78 - 2070\ M\^80 - 2250\ M\^82 + 
                1410\ M\^84 + 810\ M\^86 - 640\ M\^88 - 240\ M\^90 + 
                276\ M\^92 + 21\ M\^94 - 103\ M\^96 + 50\ M\^98 - 
                11\ M\^100 + M\^102)\)}, {"A-polynomial", Knot[8, 17], 
        M\^58 - 2\ M\^62 + M\^66 + L\^20\ \((M\^58 - 2\ M\^62 + M\^66)\) + 
          L\ \((M\^44 - 2\ M\^46 - 3\ M\^48 + 5\ M\^50 + 14\ M\^52 - 
                25\ M\^54 - 40\ M\^56 + 111\ M\^58 + 28\ M\^60 - 179\ M\^62 + 
                28\ M\^64 + 111\ M\^66 - 40\ M\^68 - 25\ M\^70 + 14\ M\^72 + 
                5\ M\^74 - 3\ M\^76 - 2\ M\^78 + M\^80)\) + 
          L\^19\ \((M\^44 - 2\ M\^46 - 3\ M\^48 + 5\ M\^50 + 14\ M\^52 - 
                25\ M\^54 - 40\ M\^56 + 111\ M\^58 + 28\ M\^60 - 179\ M\^62 + 
                28\ M\^64 + 111\ M\^66 - 40\ M\^68 - 25\ M\^70 + 14\ M\^72 + 
                5\ M\^74 - 3\ M\^76 - 2\ M\^78 + M\^80)\) + 
          L\^2\ \((\(-3\)\ M\^36 + 18\ M\^38 - 29\ M\^40 - 51\ M\^42 + 
                202\ M\^44 - 24\ M\^46 - 588\ M\^48 + 522\ M\^50 + 
                963\ M\^52 - 1770\ M\^54 - 756\ M\^56 + 3206\ M\^58 + 
                202\ M\^60 - 3802\ M\^62 + 202\ M\^64 + 3206\ M\^66 - 
                756\ M\^68 - 1770\ M\^70 + 963\ M\^72 + 522\ M\^74 - 
                588\ M\^76 - 24\ M\^78 + 202\ M\^80 - 51\ M\^82 - 29\ M\^84 + 
                18\ M\^86 - 3\ M\^88)\) + 
          L\^18\ \((\(-3\)\ M\^36 + 18\ M\^38 - 29\ M\^40 - 51\ M\^42 + 
                202\ M\^44 - 24\ M\^46 - 588\ M\^48 + 522\ M\^50 + 
                963\ M\^52 - 1770\ M\^54 - 756\ M\^56 + 3206\ M\^58 + 
                202\ M\^60 - 3802\ M\^62 + 202\ M\^64 + 3206\ M\^66 - 
                756\ M\^68 - 1770\ M\^70 + 963\ M\^72 + 522\ M\^74 - 
                588\ M\^76 - 24\ M\^78 + 202\ M\^80 - 51\ M\^82 - 29\ M\^84 + 
                18\ M\^86 - 3\ M\^88)\) + 
          L\^3\ \((3\ M\^28 - 31\ M\^30 + 123\ M\^32 - 170\ M\^34 - 
                250\ M\^36 + 1073\ M\^38 - 536\ M\^40 - 2396\ M\^42 + 
                3463\ M\^44 + 1789\ M\^46 - 7683\ M\^48 + 4319\ M\^50 + 
                9243\ M\^52 - 16447\ M\^54 - 6809\ M\^56 + 30667\ M\^58 + 
                2446\ M\^60 - 37761\ M\^62 + 2446\ M\^64 + 30667\ M\^66 - 
                6809\ M\^68 - 16447\ M\^70 + 9243\ M\^72 + 4319\ M\^74 - 
                7683\ M\^76 + 1789\ M\^78 + 3463\ M\^80 - 2396\ M\^82 - 
                536\ M\^84 + 1073\ M\^86 - 250\ M\^88 - 170\ M\^90 + 
                123\ M\^92 - 31\ M\^94 + 3\ M\^96)\) + 
          L\^17\ \((3\ M\^28 - 31\ M\^30 + 123\ M\^32 - 170\ M\^34 - 
                250\ M\^36 + 1073\ M\^38 - 536\ M\^40 - 2396\ M\^42 + 
                3463\ M\^44 + 1789\ M\^46 - 7683\ M\^48 + 4319\ M\^50 + 
                9243\ M\^52 - 16447\ M\^54 - 6809\ M\^56 + 30667\ M\^58 + 
                2446\ M\^60 - 37761\ M\^62 + 2446\ M\^64 + 30667\ M\^66 - 
                6809\ M\^68 - 16447\ M\^70 + 9243\ M\^72 + 4319\ M\^74 - 
                7683\ M\^76 + 1789\ M\^78 + 3463\ M\^80 - 2396\ M\^82 - 
                536\ M\^84 + 1073\ M\^86 - 250\ M\^88 - 170\ M\^90 + 
                123\ M\^92 - 31\ M\^94 + 3\ M\^96)\) + 
          L\^4\ \((\(-M\^20\) + 16\ M\^22 - 108\ M\^24 + 383\ M\^26 - 
                630\ M\^28 - 307\ M\^30 + 3195\ M\^32 - 3456\ M\^34 - 
                6602\ M\^36 + 15714\ M\^38 + 6124\ M\^40 - 36835\ M\^42 + 
                974\ M\^44 + 57615\ M\^46 - 8306\ M\^48 - 60622\ M\^50 + 
                8010\ M\^52 + 28048\ M\^54 + 11\ M\^56 + 25719\ M\^58 - 
                3075\ M\^60 - 52550\ M\^62 - 3075\ M\^64 + 25719\ M\^66 + 
                11\ M\^68 + 28048\ M\^70 + 8010\ M\^72 - 60622\ M\^74 - 
                8306\ M\^76 + 57615\ M\^78 + 974\ M\^80 - 36835\ M\^82 + 
                6124\ M\^84 + 15714\ M\^86 - 6602\ M\^88 - 3456\ M\^90 + 
                3195\ M\^92 - 307\ M\^94 - 630\ M\^96 + 383\ M\^98 - 
                108\ M\^100 + 16\ M\^102 - M\^104)\) + 
          L\^16\ \((\(-M\^20\) + 16\ M\^22 - 108\ M\^24 + 383\ M\^26 - 
                630\ M\^28 - 307\ M\^30 + 3195\ M\^32 - 3456\ M\^34 - 
                6602\ M\^36 + 15714\ M\^38 + 6124\ M\^40 - 36835\ M\^42 + 
                974\ M\^44 + 57615\ M\^46 - 8306\ M\^48 - 60622\ M\^50 + 
                8010\ M\^52 + 28048\ M\^54 + 11\ M\^56 + 25719\ M\^58 - 
                3075\ M\^60 - 52550\ M\^62 - 3075\ M\^64 + 25719\ M\^66 + 
                11\ M\^68 + 28048\ M\^70 + 8010\ M\^72 - 60622\ M\^74 - 
                8306\ M\^76 + 57615\ M\^78 + 974\ M\^80 - 36835\ M\^82 + 
                6124\ M\^84 + 15714\ M\^86 - 6602\ M\^88 - 3456\ M\^90 + 
                3195\ M\^92 - 307\ M\^94 - 630\ M\^96 + 383\ M\^98 - 
                108\ M\^100 + 16\ M\^102 - M\^104)\) + 
          L\^5\ \((\(-M\^14\) + 17\ M\^16 - 134\ M\^18 + 603\ M\^20 - 
                1490\ M\^22 + 1025\ M\^24 + 4598\ M\^26 - 10942\ M\^28 - 
                4735\ M\^30 + 40539\ M\^32 - 12290\ M\^34 - 100363\ M\^36 + 
                58397\ M\^38 + 201005\ M\^40 - 120180\ M\^42 - 
                347559\ M\^44 + 130537\ M\^46 + 517005\ M\^48 + 
                23815\ M\^50 - 634820\ M\^52 - 382738\ M\^54 + 
                563488\ M\^56 + 800903\ M\^58 - 229998\ M\^60 - 
                996424\ M\^62 - 229998\ M\^64 + 800903\ M\^66 + 
                563488\ M\^68 - 382738\ M\^70 - 634820\ M\^72 + 
                23815\ M\^74 + 517005\ M\^76 + 130537\ M\^78 - 
                347559\ M\^80 - 120180\ M\^82 + 201005\ M\^84 + 
                58397\ M\^86 - 100363\ M\^88 - 12290\ M\^90 + 40539\ M\^92 - 
                4735\ M\^94 - 10942\ M\^96 + 4598\ M\^98 + 1025\ M\^100 - 
                1490\ M\^102 + 603\ M\^104 - 134\ M\^106 + 17\ M\^108 - 
                M\^110)\) + 
          L\^15\ \((\(-M\^14\) + 17\ M\^16 - 134\ M\^18 + 603\ M\^20 - 
                1490\ M\^22 + 1025\ M\^24 + 4598\ M\^26 - 10942\ M\^28 - 
                4735\ M\^30 + 40539\ M\^32 - 12290\ M\^34 - 100363\ M\^36 + 
                58397\ M\^38 + 201005\ M\^40 - 120180\ M\^42 - 
                347559\ M\^44 + 130537\ M\^46 + 517005\ M\^48 + 
                23815\ M\^50 - 634820\ M\^52 - 382738\ M\^54 + 
                563488\ M\^56 + 800903\ M\^58 - 229998\ M\^60 - 
                996424\ M\^62 - 229998\ M\^64 + 800903\ M\^66 + 
                563488\ M\^68 - 382738\ M\^70 - 634820\ M\^72 + 
                23815\ M\^74 + 517005\ M\^76 + 130537\ M\^78 - 
                347559\ M\^80 - 120180\ M\^82 + 201005\ M\^84 + 
                58397\ M\^86 - 100363\ M\^88 - 12290\ M\^90 + 40539\ M\^92 - 
                4735\ M\^94 - 10942\ M\^96 + 4598\ M\^98 + 1025\ M\^100 - 
                1490\ M\^102 + 603\ M\^104 - 134\ M\^106 + 17\ M\^108 - 
                M\^110)\) + 
          L\^6\ \((3\ M\^10 - 52\ M\^12 + 385\ M\^14 - 1494\ M\^16 + 
                2751\ M\^18 + 316\ M\^20 - 10616\ M\^22 + 10575\ M\^24 + 
                25053\ M\^26 - 50199\ M\^28 - 31976\ M\^30 + 130086\ M\^32 + 
                8792\ M\^34 - 244141\ M\^36 + 75827\ M\^38 + 347946\ M\^40 - 
                233410\ M\^42 - 385836\ M\^44 + 424868\ M\^46 + 
                356823\ M\^48 - 579188\ M\^50 - 287533\ M\^52 + 
                609105\ M\^54 + 201277\ M\^56 - 530229\ M\^58 - 
                82052\ M\^60 + 477270\ M\^62 - 82052\ M\^64 - 530229\ M\^66 + 
                201277\ M\^68 + 609105\ M\^70 - 287533\ M\^72 - 
                579188\ M\^74 + 356823\ M\^76 + 424868\ M\^78 - 
                385836\ M\^80 - 233410\ M\^82 + 347946\ M\^84 + 
                75827\ M\^86 - 244141\ M\^88 + 8792\ M\^90 + 130086\ M\^92 - 
                31976\ M\^94 - 50199\ M\^96 + 25053\ M\^98 + 10575\ M\^100 - 
                10616\ M\^102 + 316\ M\^104 + 2751\ M\^106 - 1494\ M\^108 + 
                385\ M\^110 - 52\ M\^112 + 3\ M\^114)\) + 
          L\^14\ \((3\ M\^10 - 52\ M\^12 + 385\ M\^14 - 1494\ M\^16 + 
                2751\ M\^18 + 316\ M\^20 - 10616\ M\^22 + 10575\ M\^24 + 
                25053\ M\^26 - 50199\ M\^28 - 31976\ M\^30 + 130086\ M\^32 + 
                8792\ M\^34 - 244141\ M\^36 + 75827\ M\^38 + 347946\ M\^40 - 
                233410\ M\^42 - 385836\ M\^44 + 424868\ M\^46 + 
                356823\ M\^48 - 579188\ M\^50 - 287533\ M\^52 + 
                609105\ M\^54 + 201277\ M\^56 - 530229\ M\^58 - 
                82052\ M\^60 + 477270\ M\^62 - 82052\ M\^64 - 530229\ M\^66 + 
                201277\ M\^68 + 609105\ M\^70 - 287533\ M\^72 - 
                579188\ M\^74 + 356823\ M\^76 + 424868\ M\^78 - 
                385836\ M\^80 - 233410\ M\^82 + 347946\ M\^84 + 
                75827\ M\^86 - 244141\ M\^88 + 8792\ M\^90 + 130086\ M\^92 - 
                31976\ M\^94 - 50199\ M\^96 + 25053\ M\^98 + 10575\ M\^100 - 
                10616\ M\^102 + 316\ M\^104 + 2751\ M\^106 - 1494\ M\^108 + 
                385\ M\^110 - 52\ M\^112 + 3\ M\^114)\) + 
          L\^7\ \((\(-3\)\ M\^6 + 56\ M\^8 - 434\ M\^10 + 1741\ M\^12 - 
                3392\ M\^14 + 567\ M\^16 + 10010\ M\^18 - 10719\ M\^20 - 
                22944\ M\^22 + 47900\ M\^24 + 20915\ M\^26 - 109816\ M\^28 + 
                28577\ M\^30 + 161436\ M\^32 - 196500\ M\^34 - 78535\ M\^36 + 
                565804\ M\^38 - 301270\ M\^40 - 1228349\ M\^42 + 
                1026959\ M\^44 + 2267977\ M\^46 - 1892430\ M\^48 - 
                3626086\ M\^50 + 2382035\ M\^52 + 5075006\ M\^54 - 
                2027581\ M\^56 - 6203868\ M\^58 + 799657\ M\^60 + 
                6608010\ M\^62 + 799657\ M\^64 - 6203868\ M\^66 - 
                2027581\ M\^68 + 5075006\ M\^70 + 2382035\ M\^72 - 
                3626086\ M\^74 - 1892430\ M\^76 + 2267977\ M\^78 + 
                1026959\ M\^80 - 1228349\ M\^82 - 301270\ M\^84 + 
                565804\ M\^86 - 78535\ M\^88 - 196500\ M\^90 + 
                161436\ M\^92 + 28577\ M\^94 - 109816\ M\^96 + 20915\ M\^98 + 
                47900\ M\^100 - 22944\ M\^102 - 10719\ M\^104 + 
                10010\ M\^106 + 567\ M\^108 - 3392\ M\^110 + 1741\ M\^112 - 
                434\ M\^114 + 56\ M\^116 - 3\ M\^118)\) + 
          L\^13\ \((\(-3\)\ M\^6 + 56\ M\^8 - 434\ M\^10 + 1741\ M\^12 - 
                3392\ M\^14 + 567\ M\^16 + 10010\ M\^18 - 10719\ M\^20 - 
                22944\ M\^22 + 47900\ M\^24 + 20915\ M\^26 - 109816\ M\^28 + 
                28577\ M\^30 + 161436\ M\^32 - 196500\ M\^34 - 78535\ M\^36 + 
                565804\ M\^38 - 301270\ M\^40 - 1228349\ M\^42 + 
                1026959\ M\^44 + 2267977\ M\^46 - 1892430\ M\^48 - 
                3626086\ M\^50 + 2382035\ M\^52 + 5075006\ M\^54 - 
                2027581\ M\^56 - 6203868\ M\^58 + 799657\ M\^60 + 
                6608010\ M\^62 + 799657\ M\^64 - 6203868\ M\^66 - 
                2027581\ M\^68 + 5075006\ M\^70 + 2382035\ M\^72 - 
                3626086\ M\^74 - 1892430\ M\^76 + 2267977\ M\^78 + 
                1026959\ M\^80 - 1228349\ M\^82 - 301270\ M\^84 + 
                565804\ M\^86 - 78535\ M\^88 - 196500\ M\^90 + 
                161436\ M\^92 + 28577\ M\^94 - 109816\ M\^96 + 20915\ M\^98 + 
                47900\ M\^100 - 22944\ M\^102 - 10719\ M\^104 + 
                10010\ M\^106 + 567\ M\^108 - 3392\ M\^110 + 1741\ M\^112 - 
                434\ M\^114 + 56\ M\^116 - 3\ M\^118)\) + 
          L\^8\ \((M\^2 - 21\ M\^4 + 188\ M\^6 - 924\ M\^8 + 2586\ M\^10 - 
                3371\ M\^12 - 1717\ M\^14 + 11234\ M\^16 - 3258\ M\^18 - 
                37243\ M\^20 + 47397\ M\^22 + 67532\ M\^24 - 175294\ M\^26 - 
                82388\ M\^28 + 486442\ M\^30 + 10629\ M\^32 - 
                1047727\ M\^34 + 163026\ M\^36 + 1799870\ M\^38 - 
                305426\ M\^40 - 2366813\ M\^42 + 154875\ M\^44 + 
                1989076\ M\^46 + 457144\ M\^48 - 28066\ M\^50 - 
                1179848\ M\^52 - 3341745\ M\^54 + 1353494\ M\^56 + 
                6703365\ M\^58 - 624625\ M\^60 - 8128610\ M\^62 - 
                624625\ M\^64 + 6703365\ M\^66 + 1353494\ M\^68 - 
                3341745\ M\^70 - 1179848\ M\^72 - 28066\ M\^74 + 
                457144\ M\^76 + 1989076\ M\^78 + 154875\ M\^80 - 
                2366813\ M\^82 - 305426\ M\^84 + 1799870\ M\^86 + 
                163026\ M\^88 - 1047727\ M\^90 + 10629\ M\^92 + 
                486442\ M\^94 - 82388\ M\^96 - 175294\ M\^98 + 
                67532\ M\^100 + 47397\ M\^102 - 37243\ M\^104 - 
                3258\ M\^106 + 11234\ M\^108 - 1717\ M\^110 - 3371\ M\^112 + 
                2586\ M\^114 - 924\ M\^116 + 188\ M\^118 - 21\ M\^120 + 
                M\^122)\) + 
          L\^12\ \((M\^2 - 21\ M\^4 + 188\ M\^6 - 924\ M\^8 + 2586\ M\^10 - 
                3371\ M\^12 - 1717\ M\^14 + 11234\ M\^16 - 3258\ M\^18 - 
                37243\ M\^20 + 47397\ M\^22 + 67532\ M\^24 - 175294\ M\^26 - 
                82388\ M\^28 + 486442\ M\^30 + 10629\ M\^32 - 
                1047727\ M\^34 + 163026\ M\^36 + 1799870\ M\^38 - 
                305426\ M\^40 - 2366813\ M\^42 + 154875\ M\^44 + 
                1989076\ M\^46 + 457144\ M\^48 - 28066\ M\^50 - 
                1179848\ M\^52 - 3341745\ M\^54 + 1353494\ M\^56 + 
                6703365\ M\^58 - 624625\ M\^60 - 8128610\ M\^62 - 
                624625\ M\^64 + 6703365\ M\^66 + 1353494\ M\^68 - 
                3341745\ M\^70 - 1179848\ M\^72 - 28066\ M\^74 + 
                457144\ M\^76 + 1989076\ M\^78 + 154875\ M\^80 - 
                2366813\ M\^82 - 305426\ M\^84 + 1799870\ M\^86 + 
                163026\ M\^88 - 1047727\ M\^90 + 10629\ M\^92 + 
                486442\ M\^94 - 82388\ M\^96 - 175294\ M\^98 + 
                67532\ M\^100 + 47397\ M\^102 - 37243\ M\^104 - 
                3258\ M\^106 + 11234\ M\^108 - 1717\ M\^110 - 3371\ M\^112 + 
                2586\ M\^114 - 924\ M\^116 + 188\ M\^118 - 21\ M\^120 + 
                M\^122)\) + 
          L\^10\ \((\(-2\) + 39\ M\^2 - 318\ M\^4 + 1363\ M\^6 - 2959\ M\^8 + 
                1440\ M\^10 + 5839\ M\^12 - 1619\ M\^14 - 33450\ M\^16 + 
                23992\ M\^18 + 119966\ M\^20 - 97513\ M\^22 - 407029\ M\^24 + 
                345289\ M\^26 + 1113641\ M\^28 - 906347\ M\^30 - 
                2538819\ M\^32 + 1957890\ M\^34 + 4659712\ M\^36 - 
                3741520\ M\^38 - 6634081\ M\^40 + 6979626\ M\^42 + 
                6841859\ M\^44 - 13338023\ M\^46 - 4134900\ M\^48 + 
                24483742\ M\^50 - 206284\ M\^52 - 39672954\ M\^54 + 
                2862683\ M\^56 + 53803932\ M\^58 - 1670168\ M\^60 - 
                59678674\ M\^62 - 1670168\ M\^64 + 53803932\ M\^66 + 
                2862683\ M\^68 - 39672954\ M\^70 - 206284\ M\^72 + 
                24483742\ M\^74 - 4134900\ M\^76 - 13338023\ M\^78 + 
                6841859\ M\^80 + 6979626\ M\^82 - 6634081\ M\^84 - 
                3741520\ M\^86 + 4659712\ M\^88 + 1957890\ M\^90 - 
                2538819\ M\^92 - 906347\ M\^94 + 1113641\ M\^96 + 
                345289\ M\^98 - 407029\ M\^100 - 97513\ M\^102 + 
                119966\ M\^104 + 23992\ M\^106 - 33450\ M\^108 - 
                1619\ M\^110 + 5839\ M\^112 + 1440\ M\^114 - 2959\ M\^116 + 
                1363\ M\^118 - 318\ M\^120 + 39\ M\^122 - 2\ M\^124)\) + 
          L\^9\ \((1 - 20\ M\^2 + 175\ M\^4 - 856\ M\^6 + 2404\ M\^8 - 
                3139\ M\^10 - 1361\ M\^12 + 7915\ M\^14 + 4526\ M\^16 - 
                33845\ M\^18 + 7476\ M\^20 + 78684\ M\^22 - 36013\ M\^24 - 
                135728\ M\^26 + 126434\ M\^28 + 16044\ M\^30 - 
                320363\ M\^32 + 756600\ M\^34 + 797993\ M\^36 - 
                2988606\ M\^38 - 1970513\ M\^40 + 7319228\ M\^42 + 
                4279307\ M\^44 - 13302878\ M\^46 - 7500156\ M\^48 + 
                19237876\ M\^50 + 9859600\ M\^52 - 23202163\ M\^54 - 
                8850832\ M\^56 + 24794668\ M\^58 + 3601322\ M\^60 - 
                25131318\ M\^62 + 3601322\ M\^64 + 24794668\ M\^66 - 
                8850832\ M\^68 - 23202163\ M\^70 + 9859600\ M\^72 + 
                19237876\ M\^74 - 7500156\ M\^76 - 13302878\ M\^78 + 
                4279307\ M\^80 + 7319228\ M\^82 - 1970513\ M\^84 - 
                2988606\ M\^86 + 797993\ M\^88 + 756600\ M\^90 - 
                320363\ M\^92 + 16044\ M\^94 + 126434\ M\^96 - 
                135728\ M\^98 - 36013\ M\^100 + 78684\ M\^102 + 
                7476\ M\^104 - 33845\ M\^106 + 4526\ M\^108 + 7915\ M\^110 - 
                1361\ M\^112 - 3139\ M\^114 + 2404\ M\^116 - 856\ M\^118 + 
                175\ M\^120 - 20\ M\^122 + M\^124)\) + 
          L\^11\ \((1 - 20\ M\^2 + 175\ M\^4 - 856\ M\^6 + 2404\ M\^8 - 
                3139\ M\^10 - 1361\ M\^12 + 7915\ M\^14 + 4526\ M\^16 - 
                33845\ M\^18 + 7476\ M\^20 + 78684\ M\^22 - 36013\ M\^24 - 
                135728\ M\^26 + 126434\ M\^28 + 16044\ M\^30 - 
                320363\ M\^32 + 756600\ M\^34 + 797993\ M\^36 - 
                2988606\ M\^38 - 1970513\ M\^40 + 7319228\ M\^42 + 
                4279307\ M\^44 - 13302878\ M\^46 - 7500156\ M\^48 + 
                19237876\ M\^50 + 9859600\ M\^52 - 23202163\ M\^54 - 
                8850832\ M\^56 + 24794668\ M\^58 + 3601322\ M\^60 - 
                25131318\ M\^62 + 3601322\ M\^64 + 24794668\ M\^66 - 
                8850832\ M\^68 - 23202163\ M\^70 + 9859600\ M\^72 + 
                19237876\ M\^74 - 7500156\ M\^76 - 13302878\ M\^78 + 
                4279307\ M\^80 + 7319228\ M\^82 - 1970513\ M\^84 - 
                2988606\ M\^86 + 797993\ M\^88 + 756600\ M\^90 - 
                320363\ M\^92 + 16044\ M\^94 + 126434\ M\^96 - 
                135728\ M\^98 - 36013\ M\^100 + 78684\ M\^102 + 
                7476\ M\^104 - 33845\ M\^106 + 4526\ M\^108 + 7915\ M\^110 - 
                1361\ M\^112 - 3139\ M\^114 + 2404\ M\^116 - 856\ M\^118 + 
                175\ M\^120 - 20\ M\^122 + M\^124)\)}, {"A-polynomial", 
        Knot[8, 18], 
        M\^32 + L\^8\ M\^32 + 
          L\ \((M\^18 - 2\ M\^20 - 3\ M\^22 + 18\ M\^26 - 12\ M\^28 - 
                30\ M\^30 + 64\ M\^32 - 30\ M\^34 - 12\ M\^36 + 18\ M\^38 - 
                3\ M\^42 - 2\ M\^44 + M\^46)\) + 
          L\^7\ \((M\^18 - 2\ M\^20 - 3\ M\^22 + 18\ M\^26 - 12\ M\^28 - 
                30\ M\^30 + 64\ M\^32 - 30\ M\^34 - 12\ M\^36 + 18\ M\^38 - 
                3\ M\^42 - 2\ M\^44 + M\^46)\) + 
          L\^2\ \((\(-2\)\ M\^10 + 16\ M\^12 - 32\ M\^14 - 38\ M\^16 + 
                162\ M\^18 + 13\ M\^20 - 370\ M\^22 + 170\ M\^24 + 
                200\ M\^26 + 20\ M\^28 - 174\ M\^30 + 98\ M\^32 - 
                174\ M\^34 + 20\ M\^36 + 200\ M\^38 + 170\ M\^40 - 
                370\ M\^42 + 13\ M\^44 + 162\ M\^46 - 38\ M\^48 - 32\ M\^50 + 
                16\ M\^52 - 2\ M\^54)\) + 
          L\^6\ \((\(-2\)\ M\^10 + 16\ M\^12 - 32\ M\^14 - 38\ M\^16 + 
                162\ M\^18 + 13\ M\^20 - 370\ M\^22 + 170\ M\^24 + 
                200\ M\^26 + 20\ M\^28 - 174\ M\^30 + 98\ M\^32 - 
                174\ M\^34 + 20\ M\^36 + 200\ M\^38 + 170\ M\^40 - 
                370\ M\^42 + 13\ M\^44 + 162\ M\^46 - 38\ M\^48 - 32\ M\^50 + 
                16\ M\^52 - 2\ M\^54)\) + 
          L\^3\ \((M\^2 - 14\ M\^4 + 75\ M\^6 - 184\ M\^8 + 173\ M\^10 + 
                32\ M\^12 - 67\ M\^14 - 22\ M\^16 + 49\ M\^18 - 890\ M\^20 + 
                849\ M\^22 + 1838\ M\^24 - 704\ M\^26 - 4126\ M\^28 + 
                678\ M\^30 + 4680\ M\^32 + 678\ M\^34 - 4126\ M\^36 - 
                704\ M\^38 + 1838\ M\^40 + 849\ M\^42 - 890\ M\^44 + 
                49\ M\^46 - 22\ M\^48 - 67\ M\^50 + 32\ M\^52 + 173\ M\^54 - 
                184\ M\^56 + 75\ M\^58 - 14\ M\^60 + M\^62)\) + 
          L\^5\ \((M\^2 - 14\ M\^4 + 75\ M\^6 - 184\ M\^8 + 173\ M\^10 + 
                32\ M\^12 - 67\ M\^14 - 22\ M\^16 + 49\ M\^18 - 890\ M\^20 + 
                849\ M\^22 + 1838\ M\^24 - 704\ M\^26 - 4126\ M\^28 + 
                678\ M\^30 + 4680\ M\^32 + 678\ M\^34 - 4126\ M\^36 - 
                704\ M\^38 + 1838\ M\^40 + 849\ M\^42 - 890\ M\^44 + 
                49\ M\^46 - 22\ M\^48 - 67\ M\^50 + 32\ M\^52 + 173\ M\^54 - 
                184\ M\^56 + 75\ M\^58 - 14\ M\^60 + M\^62)\) + 
          L\^4\ \((1 - 16\ M\^2 + 101\ M\^4 - 300\ M\^6 + 324\ M\^8 + 
                284\ M\^10 - 499\ M\^12 - 1264\ M\^14 + 1677\ M\^16 + 
                2184\ M\^18 - 1954\ M\^20 - 4120\ M\^22 + 2202\ M\^24 + 
                3536\ M\^26 - 234\ M\^28 - 1888\ M\^30 + 2\ M\^32 - 
                1888\ M\^34 - 234\ M\^36 + 3536\ M\^38 + 2202\ M\^40 - 
                4120\ M\^42 - 1954\ M\^44 + 2184\ M\^46 + 1677\ M\^48 - 
                1264\ M\^50 - 499\ M\^52 + 284\ M\^54 + 324\ M\^56 - 
                300\ M\^58 + 101\ M\^60 - 16\ M\^62 + 
                M\^64)\)}, {"A-polynomial", Knot[8, 19], 
        1 + L\ M\^12}, {"A-polynomial", Knot[8, 20], 
        L\^5\ M\^8 + M\^10 + 
          L\ \((1 - M\^2 + 2\ M\^4 - 2\ M\^6 - M\^8 + 5\ M\^10 + M\^12)\) + 
          L\^2\ \((\(-1\) + 5\ M\^2 - 3\ M\^6 + 3\ M\^8 + 4\ M\^12 + 
                2\ M\^14)\) + 
          L\^3\ \((2\ M\^4 + 4\ M\^6 + 3\ M\^10 - 3\ M\^12 + 5\ M\^16 - 
                M\^18)\) + 
          L\^4\ \((M\^6 + 5\ M\^8 - M\^10 - 2\ M\^12 + 2\ M\^14 - M\^16 + 
                M\^18)\)}, {"A-polynomial", Knot[8, 21], 
        L\^5\ M\^2 + M\^20 + 
          L\^4\ \((\(-2\)\ M\^2 - 2\ M\^4 + 13\ M\^6 + M\^8 - 8\ M\^10 + 
                4\ M\^12 - M\^14)\) + 
          L\ \((\(-M\^8\) + 4\ M\^10 - 8\ M\^12 + M\^14 + 13\ M\^16 - 
                2\ M\^18 - 2\ M\^20)\) + 
          L\^3\ \((\(-1\) + 7\ M\^2 - 11\ M\^4 - 13\ M\^6 + 25\ M\^8 + 
                9\ M\^10 + 4\ M\^12 - 6\ M\^14 - 10\ M\^16 + 7\ M\^18 - 
                M\^20)\) + 
          L\^2\ \((\(-M\^2\) + 7\ M\^4 - 10\ M\^6 - 6\ M\^8 + 4\ M\^10 + 
                9\ M\^12 + 25\ M\^14 - 13\ M\^16 - 11\ M\^18 + 7\ M\^20 - 
                M\^22)\)}, {"A-polynomial", Knot[9, 1], 
        L + M\^18}, {"A-polynomial", Knot[9, 2], 
        L\^7 + M\^30 + 
          L\^6\ \((\(-3\) + 6\ M\^2 + 4\ M\^4 - M\^16 + M\^18)\) + 
          L\^5\ \((3 - 12\ M\^2 + 5\ M\^4 + 15\ M\^6 + 10\ M\^8 - 5\ M\^16 + 
                8\ M\^20 - 3\ M\^22)\) + 
          L\^4\ \((\(-1\) + 6\ M\^2 - 12\ M\^4 + 2\ M\^6 + 5\ M\^8 + 
                15\ M\^10 + 20\ M\^12 - 15\ M\^16 + 16\ M\^20 + 9\ M\^22 - 
                13\ M\^24 + 3\ M\^26)\) + 
          L\ \((M\^12 - M\^14 + 4\ M\^26 + 6\ M\^28 - 3\ M\^30)\) + 
          L\^3\ \((3\ M\^4 - 13\ M\^6 + 9\ M\^8 + 16\ M\^10 - 15\ M\^14 + 
                20\ M\^18 + 15\ M\^20 + 5\ M\^22 + 2\ M\^24 - 12\ M\^26 + 
                6\ M\^28 - M\^30)\) + 
          L\^2\ \((\(-3\)\ M\^8 + 8\ M\^10 - 5\ M\^14 + 10\ M\^22 + 
                15\ M\^24 + 5\ M\^26 - 12\ M\^28 + 
                3\ M\^30)\)}, {"A-polynomial", Knot[9, 3], 
        1 + L\^9\ M\^114 + 
          L\ \((\(-1\) + 2\ M\^2 - M\^4 + 7\ M\^12 + 4\ M\^14 - 5\ M\^16 + 
                3\ M\^18)\) + 
          L\^2\ \((5\ M\^12 - 17\ M\^14 + 10\ M\^16 + 15\ M\^18 - 17\ M\^20 + 
                4\ M\^22 + 28\ M\^24 + 5\ M\^26 - 7\ M\^28 + 3\ M\^30 + 
                14\ M\^32 - 10\ M\^34 + 3\ M\^36)\) + 
          L\^3\ \((\(-10\)\ M\^24 + 44\ M\^26 - 40\ M\^28 - 61\ M\^30 + 
                77\ M\^32 + 74\ M\^34 - 55\ M\^36 - 15\ M\^38 + 69\ M\^40 + 
                12\ M\^42 - 26\ M\^44 + 19\ M\^46 - 10\ M\^48 + 10\ M\^50 - 
                5\ M\^52 + M\^54)\) + 
          L\^4\ \((10\ M\^36 - 50\ M\^38 + 61\ M\^40 + 60\ M\^42 - 
                110\ M\^44 - 95\ M\^46 + 170\ M\^48 + 165\ M\^50 - 
                90\ M\^52 - 85\ M\^54 + 65\ M\^56 + 50\ M\^58 - 6\ M\^60 - 
                40\ M\^62 + 26\ M\^64 - 5\ M\^66)\) + 
          L\^5\ \((\(-5\)\ M\^48 + 26\ M\^50 - 40\ M\^52 - 6\ M\^54 + 
                50\ M\^56 + 65\ M\^58 - 85\ M\^60 - 90\ M\^62 + 165\ M\^64 + 
                170\ M\^66 - 95\ M\^68 - 110\ M\^70 + 60\ M\^72 + 61\ M\^74 - 
                50\ M\^76 + 10\ M\^78)\) + 
          L\^6\ \((M\^60 - 5\ M\^62 + 10\ M\^64 - 10\ M\^66 + 19\ M\^68 - 
                26\ M\^70 + 12\ M\^72 + 69\ M\^74 - 15\ M\^76 - 55\ M\^78 + 
                74\ M\^80 + 77\ M\^82 - 61\ M\^84 - 40\ M\^86 + 44\ M\^88 - 
                10\ M\^90)\) + 
          L\^7\ \((3\ M\^78 - 10\ M\^80 + 14\ M\^82 + 3\ M\^84 - 7\ M\^86 + 
                5\ M\^88 + 28\ M\^90 + 4\ M\^92 - 17\ M\^94 + 15\ M\^96 + 
                10\ M\^98 - 17\ M\^100 + 5\ M\^102)\) + 
          L\^8\ \((3\ M\^96 - 5\ M\^98 + 4\ M\^100 + 7\ M\^102 - M\^110 + 
                2\ M\^112 - M\^114)\)}, {"A-polynomial", Knot[9, 4], 
        1 + L\^10\ M\^84 + 
          L\ \((\(-2\) + 4\ M\^2 - 3\ M\^4 + 2\ M\^6 + 9\ M\^8 - M\^12 + 
                2\ M\^14 - 3\ M\^16 + 2\ M\^18)\) + 
          L\^2\ \((1 - 4\ M\^2 + 6\ M\^4 - 4\ M\^6 - 3\ M\^8 - 5\ M\^10 + 
                20\ M\^12 + M\^14 + 6\ M\^16 + 41\ M\^18 - 6\ M\^20 - 
                28\ M\^22 + 17\ M\^24 + 3\ M\^26 - M\^30 + 3\ M\^32 - 
                3\ M\^34 + M\^36)\) + 
          L\^3\ \((\(-6\)\ M\^8 + 29\ M\^10 - 43\ M\^12 - 8\ M\^14 + 
                77\ M\^16 - 44\ M\^18 - 114\ M\^20 + 178\ M\^22 + 
                130\ M\^24 - 158\ M\^26 + 5\ M\^28 + 124\ M\^30 - 69\ M\^32 - 
                12\ M\^34 + 54\ M\^36 - 14\ M\^38 - 28\ M\^40 + 25\ M\^42 - 
                6\ M\^44)\) + 
          L\^4\ \((15\ M\^16 - 79\ M\^18 + 116\ M\^20 + 68\ M\^22 - 
                292\ M\^24 + 37\ M\^26 + 402\ M\^28 - 209\ M\^30 - 
                146\ M\^32 + 383\ M\^34 - 73\ M\^36 - 224\ M\^38 + 
                306\ M\^40 + 21\ M\^42 - 228\ M\^44 + 77\ M\^46 + 95\ M\^48 - 
                74\ M\^50 + 15\ M\^52)\) + 
          L\^5\ \((\(-20\)\ M\^24 + 106\ M\^26 - 149\ M\^28 - 118\ M\^30 + 
                395\ M\^32 + 4\ M\^34 - 511\ M\^36 + 274\ M\^38 + 
                285\ M\^40 - 280\ M\^42 + 285\ M\^44 + 274\ M\^46 - 
                511\ M\^48 + 4\ M\^50 + 395\ M\^52 - 118\ M\^54 - 
                149\ M\^56 + 106\ M\^58 - 20\ M\^60)\) + 
          L\^6\ \((15\ M\^32 - 74\ M\^34 + 95\ M\^36 + 77\ M\^38 - 
                228\ M\^40 + 21\ M\^42 + 306\ M\^44 - 224\ M\^46 - 
                73\ M\^48 + 383\ M\^50 - 146\ M\^52 - 209\ M\^54 + 
                402\ M\^56 + 37\ M\^58 - 292\ M\^60 + 68\ M\^62 + 
                116\ M\^64 - 79\ M\^66 + 15\ M\^68)\) + 
          L\^7\ \((\(-6\)\ M\^40 + 25\ M\^42 - 28\ M\^44 - 14\ M\^46 + 
                54\ M\^48 - 12\ M\^50 - 69\ M\^52 + 124\ M\^54 + 5\ M\^56 - 
                158\ M\^58 + 130\ M\^60 + 178\ M\^62 - 114\ M\^64 - 
                44\ M\^66 + 77\ M\^68 - 8\ M\^70 - 43\ M\^72 + 29\ M\^74 - 
                6\ M\^76)\) + 
          L\^9\ \((2\ M\^66 - 3\ M\^68 + 2\ M\^70 - M\^72 + 9\ M\^76 + 
                2\ M\^78 - 3\ M\^80 + 4\ M\^82 - 2\ M\^84)\) + 
          L\^8\ \((M\^48 - 3\ M\^50 + 3\ M\^52 - M\^54 + 3\ M\^58 + 
                17\ M\^60 - 28\ M\^62 - 6\ M\^64 + 41\ M\^66 + 6\ M\^68 + 
                M\^70 + 20\ M\^72 - 5\ M\^74 - 3\ M\^76 - 4\ M\^78 + 
                6\ M\^80 - 4\ M\^82 + M\^84)\)}, {"A-polynomial", Knot[9, 5], 
        1 + L\^11\ M\^46 + 
          L\ \((\(-7\) + 14\ M\^2 + 8\ M\^4 - 7\ M\^6 + 3\ M\^8 - 2\ M\^12 + 
                3\ M\^14 - 2\ M\^16 + M\^18)\) + 
          L\^2\ \((21 - 76\ M\^2 + 19\ M\^4 + 137\ M\^6 - 35\ M\^8 - 
                57\ M\^10 + 64\ M\^12 - 18\ M\^14 - 30\ M\^16 + 32\ M\^18 + 
                5\ M\^20 - 12\ M\^22 + 8\ M\^24 - 5\ M\^26 + 3\ M\^28 - 
                M\^30)\) + 
          L\^3\ \((\(-35\) + 170\ M\^2 - 157\ M\^4 - 354\ M\^6 + 439\ M\^8 + 
                426\ M\^10 - 419\ M\^12 - 72\ M\^14 + 242\ M\^16 - 
                158\ M\^18 - 12\ M\^20 + 164\ M\^22 - 65\ M\^24 - 22\ M\^26 + 
                29\ M\^28 - 7\ M\^30 - 13\ M\^32 + 16\ M\^34 - 9\ M\^36 + 
                2\ M\^38)\) + 
          L\^8\ \((2\ M\^8 - 9\ M\^10 + 16\ M\^12 - 13\ M\^14 - 7\ M\^16 + 
                29\ M\^18 - 22\ M\^20 - 65\ M\^22 + 164\ M\^24 - 12\ M\^26 - 
                158\ M\^28 + 242\ M\^30 - 72\ M\^32 - 419\ M\^34 + 
                426\ M\^36 + 439\ M\^38 - 354\ M\^40 - 157\ M\^42 + 
                170\ M\^44 - 35\ M\^46)\) + 
          L\^6\ \((7 - 44\ M\^2 + 91\ M\^4 - 7\ M\^6 - 222\ M\^8 + 
                130\ M\^10 + 533\ M\^12 - 634\ M\^14 - 733\ M\^16 + 
                985\ M\^18 + 899\ M\^20 - 350\ M\^22 - 413\ M\^24 - 
                894\ M\^26 - 87\ M\^28 + 1903\ M\^30 + 456\ M\^32 - 
                1583\ M\^34 - 88\ M\^36 + 753\ M\^38 - 111\ M\^40 - 
                238\ M\^42 + 130\ M\^44 - 21\ M\^46)\) + 
          L\^10\ \((M\^28 - 2\ M\^30 + 3\ M\^32 - 2\ M\^34 + 3\ M\^38 - 
                7\ M\^40 + 8\ M\^42 + 14\ M\^44 - 7\ M\^46)\) + 
          L\^4\ \((35 - 200\ M\^2 + 290\ M\^4 + 334\ M\^6 - 952\ M\^8 - 
                398\ M\^10 + 1538\ M\^12 + 364\ M\^14 - 976\ M\^16 - 
                10\ M\^18 + 112\ M\^20 - 100\ M\^22 + 438\ M\^24 + 
                49\ M\^26 - 300\ M\^28 + 55\ M\^30 + 129\ M\^32 - 92\ M\^34 + 
                2\ M\^36 + 12\ M\^38 + 8\ M\^40 - 13\ M\^42 + 6\ M\^44 - 
                M\^46)\) + 
          L\^5\ \((\(-21\) + 130\ M\^2 - 238\ M\^4 - 111\ M\^6 + 753\ M\^8 - 
                88\ M\^10 - 1583\ M\^12 + 456\ M\^14 + 1903\ M\^16 - 
                87\ M\^18 - 894\ M\^20 - 413\ M\^22 - 350\ M\^24 + 
                899\ M\^26 + 985\ M\^28 - 733\ M\^30 - 634\ M\^32 + 
                533\ M\^34 + 130\ M\^36 - 222\ M\^38 - 7\ M\^40 + 91\ M\^42 - 
                44\ M\^44 + 7\ M\^46)\) + 
          L\^9\ \((\(-M\^16\) + 3\ M\^18 - 5\ M\^20 + 8\ M\^22 - 12\ M\^24 + 
                5\ M\^26 + 32\ M\^28 - 30\ M\^30 - 18\ M\^32 + 64\ M\^34 - 
                57\ M\^36 - 35\ M\^38 + 137\ M\^40 + 19\ M\^42 - 76\ M\^44 + 
                21\ M\^46)\) + 
          L\^7\ \((\(-1\) + 6\ M\^2 - 13\ M\^4 + 8\ M\^6 + 12\ M\^8 + 
                2\ M\^10 - 92\ M\^12 + 129\ M\^14 + 55\ M\^16 - 300\ M\^18 + 
                49\ M\^20 + 438\ M\^22 - 100\ M\^24 + 112\ M\^26 - 
                10\ M\^28 - 976\ M\^30 + 364\ M\^32 + 1538\ M\^34 - 
                398\ M\^36 - 952\ M\^38 + 334\ M\^40 + 290\ M\^42 - 
                200\ M\^44 + 35\ M\^46)\)}, {"A-polynomial", Knot[9, 6], 
        L\^12 + M\^144 + 
          L\^11\ \((\(-1\) + 4\ M\^2 - 5\ M\^4 + 2\ M\^6 - 6\ M\^10 + 
                19\ M\^12 + 8\ M\^14 - 13\ M\^16 + 4\ M\^18)\) + 
          L\^10\ \((M\^4 - 4\ M\^6 + 6\ M\^8 - 8\ M\^10 + 26\ M\^12 - 
                39\ M\^14 - 14\ M\^16 + 100\ M\^18 - 86\ M\^20 - 97\ M\^22 + 
                226\ M\^24 + 75\ M\^26 - 173\ M\^28 + 12\ M\^30 + 74\ M\^32 - 
                39\ M\^34 + 6\ M\^36)\) + 
          L\^9\ \((4\ M\^14 - 25\ M\^16 + 58\ M\^18 - 45\ M\^20 - 24\ M\^22 + 
                36\ M\^24 + 10\ M\^26 + 110\ M\^28 - 110\ M\^30 - 
                477\ M\^32 + 336\ M\^34 + 891\ M\^36 - 266\ M\^38 - 
                712\ M\^40 + 304\ M\^42 + 369\ M\^44 - 222\ M\^46 - 
                108\ M\^48 + 126\ M\^50 - 39\ M\^52 + 4\ M\^54)\) + 
          L\^8\ \((6\ M\^24 - 48\ M\^26 + 144\ M\^28 - 153\ M\^30 - 
                111\ M\^32 + 316\ M\^34 - 22\ M\^36 - 159\ M\^38 + 
                253\ M\^40 - 758\ M\^42 - 356\ M\^44 + 1634\ M\^46 + 
                572\ M\^48 - 1144\ M\^50 - 120\ M\^52 + 612\ M\^54 + 
                54\ M\^56 - 290\ M\^58 - 44\ M\^60 + 127\ M\^62 + 56\ M\^64 - 
                124\ M\^66 + 62\ M\^68 - 13\ M\^70 + M\^72)\) + 
          L\^7\ \((4\ M\^34 - 37\ M\^36 + 132\ M\^38 - 213\ M\^40 + 
                108\ M\^42 + 162\ M\^44 - 538\ M\^46 + 561\ M\^48 + 
                1114\ M\^50 - 2208\ M\^52 - 1186\ M\^54 + 3106\ M\^56 + 
                742\ M\^58 - 2181\ M\^60 + 656\ M\^62 + 1580\ M\^64 - 
                746\ M\^66 - 750\ M\^68 + 524\ M\^70 + 51\ M\^72 - 
                194\ M\^74 + 121\ M\^76 + 78\ M\^78 - 169\ M\^80 + 
                96\ M\^82 - 23\ M\^84 + 2\ M\^86)\) + 
          L\^6\ \((M\^44 - 10\ M\^46 + 43\ M\^48 - 121\ M\^50 + 256\ M\^52 - 
                240\ M\^54 - 408\ M\^56 + 975\ M\^58 + 364\ M\^60 - 
                1926\ M\^62 + 477\ M\^64 + 1707\ M\^66 - 1739\ M\^68 - 
                385\ M\^70 + 2936\ M\^72 - 385\ M\^74 - 1739\ M\^76 + 
                1707\ M\^78 + 477\ M\^80 - 1926\ M\^82 + 364\ M\^84 + 
                975\ M\^86 - 408\ M\^88 - 240\ M\^90 + 256\ M\^92 - 
                121\ M\^94 + 43\ M\^96 - 10\ M\^98 + M\^100)\) + 
          L\^5\ \((2\ M\^58 - 23\ M\^60 + 96\ M\^62 - 169\ M\^64 + 
                78\ M\^66 + 121\ M\^68 - 194\ M\^70 + 51\ M\^72 + 
                524\ M\^74 - 750\ M\^76 - 746\ M\^78 + 1580\ M\^80 + 
                656\ M\^82 - 2181\ M\^84 + 742\ M\^86 + 3106\ M\^88 - 
                1186\ M\^90 - 2208\ M\^92 + 1114\ M\^94 + 561\ M\^96 - 
                538\ M\^98 + 162\ M\^100 + 108\ M\^102 - 213\ M\^104 + 
                132\ M\^106 - 37\ M\^108 + 4\ M\^110)\) + 
          L\^4\ \((M\^72 - 13\ M\^74 + 62\ M\^76 - 124\ M\^78 + 56\ M\^80 + 
                127\ M\^82 - 44\ M\^84 - 290\ M\^86 + 54\ M\^88 + 
                612\ M\^90 - 120\ M\^92 - 1144\ M\^94 + 572\ M\^96 + 
                1634\ M\^98 - 356\ M\^100 - 758\ M\^102 + 253\ M\^104 - 
                159\ M\^106 - 22\ M\^108 + 316\ M\^110 - 111\ M\^112 - 
                153\ M\^114 + 144\ M\^116 - 48\ M\^118 + 6\ M\^120)\) + 
          L\^3\ \((4\ M\^90 - 39\ M\^92 + 126\ M\^94 - 108\ M\^96 - 
                222\ M\^98 + 369\ M\^100 + 304\ M\^102 - 712\ M\^104 - 
                266\ M\^106 + 891\ M\^108 + 336\ M\^110 - 477\ M\^112 - 
                110\ M\^114 + 110\ M\^116 + 10\ M\^118 + 36\ M\^120 - 
                24\ M\^122 - 45\ M\^124 + 58\ M\^126 - 25\ M\^128 + 
                4\ M\^130)\) + 
          L\^2\ \((6\ M\^108 - 39\ M\^110 + 74\ M\^112 + 12\ M\^114 - 
                173\ M\^116 + 75\ M\^118 + 226\ M\^120 - 97\ M\^122 - 
                86\ M\^124 + 100\ M\^126 - 14\ M\^128 - 39\ M\^130 + 
                26\ M\^132 - 8\ M\^134 + 6\ M\^136 - 4\ M\^138 + M\^140)\) + 
          L\ \((4\ M\^126 - 13\ M\^128 + 8\ M\^130 + 19\ M\^132 - 6\ M\^134 + 
                2\ M\^138 - 5\ M\^140 + 4\ M\^142 - 
                M\^144)\)}, {"A-polynomial", Knot[9, 7], 
        L\^14 + M\^116 + 
          L\^13\ \((\(-2\) + 8\ M\^2 - 12\ M\^4 + 2\ M\^6 + 23\ M\^8 - 
                5\ M\^10 - 2\ M\^12 + 6\ M\^14 - 7\ M\^16 + 3\ M\^18)\) + 
          L\^12\ \((1 - 8\ M\^2 + 26\ M\^4 - 38\ M\^6 + 12\ M\^8 + 
                25\ M\^10 - 9\ M\^12 - 19\ M\^14 - 28\ M\^16 + 177\ M\^18 + 
                109\ M\^20 - 215\ M\^22 - 28\ M\^24 + 120\ M\^26 - 
                22\ M\^28 - 28\ M\^30 + 27\ M\^32 - 14\ M\^34 + 3\ M\^36)\) + 
          L\^11\ \((\(-2\)\ M\^4 + 20\ M\^6 - 91\ M\^8 + 202\ M\^10 - 
                104\ M\^12 - 391\ M\^14 + 500\ M\^16 + 644\ M\^18 - 
                1183\ M\^20 - 898\ M\^22 + 1939\ M\^24 + 1101\ M\^26 - 
                1601\ M\^28 - 411\ M\^30 + 714\ M\^32 - 97\ M\^34 + 
                3\ M\^36 + 199\ M\^38 - 219\ M\^40 - 31\ M\^42 + 86\ M\^44 + 
                4\ M\^46 - 35\ M\^48 + 21\ M\^50 - 7\ M\^52 + M\^54)\) + 
          L\^10\ \((M\^8 - 16\ M\^10 + 112\ M\^12 - 391\ M\^14 + 593\ M\^16 + 
                273\ M\^18 - 2022\ M\^20 + 738\ M\^22 + 4448\ M\^24 - 
                2886\ M\^26 - 7089\ M\^28 + 4441\ M\^30 + 8444\ M\^32 - 
                2871\ M\^34 - 5980\ M\^36 + 565\ M\^38 + 3381\ M\^40 + 
                865\ M\^42 - 1676\ M\^44 - 579\ M\^46 + 898\ M\^48 - 
                255\ M\^50 - 166\ M\^52 + 282\ M\^54 - 47\ M\^56 - 
                138\ M\^58 + 101\ M\^60 - 28\ M\^62 + 3\ M\^64)\) + 
          L\^9\ \((4\ M\^14 - 51\ M\^16 + 289\ M\^18 - 818\ M\^20 + 
                811\ M\^22 + 1396\ M\^24 - 3546\ M\^26 - 1591\ M\^28 + 
                8595\ M\^30 + 1890\ M\^32 - 14218\ M\^34 - 5297\ M\^36 + 
                17199\ M\^38 + 11712\ M\^40 - 13537\ M\^42 - 14043\ M\^44 + 
                9480\ M\^46 + 11536\ M\^48 - 5242\ M\^50 - 5412\ M\^52 + 
                2101\ M\^54 + 893\ M\^56 + 197\ M\^58 - 159\ M\^60 - 
                449\ M\^62 + 247\ M\^64 + 176\ M\^66 - 270\ M\^68 + 
                139\ M\^70 - 33\ M\^72 + 3\ M\^74)\) + 
          L\^8\ \((6\ M\^20 - 74\ M\^22 + 380\ M\^24 - 932\ M\^26 + 
                647\ M\^28 + 1665\ M\^30 - 2467\ M\^32 - 2976\ M\^34 + 
                4676\ M\^36 + 6768\ M\^38 - 4799\ M\^40 - 16223\ M\^42 + 
                1881\ M\^44 + 26298\ M\^46 + 4628\ M\^48 - 27211\ M\^50 - 
                6975\ M\^52 + 20943\ M\^54 + 8308\ M\^56 - 10261\ M\^58 - 
                6851\ M\^60 + 3730\ M\^62 + 4615\ M\^64 - 2475\ M\^66 - 
                1429\ M\^68 + 1497\ M\^70 - 137\ M\^72 - 526\ M\^74 + 
                455\ M\^76 - 211\ M\^78 + 64\ M\^80 - 12\ M\^82 + M\^84)\) + 
          L\^7\ \((4\ M\^26 - 49\ M\^28 + 244\ M\^30 - 600\ M\^32 + 
                586\ M\^34 + 326\ M\^36 - 852\ M\^38 + 9\ M\^40 - 
                809\ M\^42 + 1134\ M\^44 + 6452\ M\^46 - 5749\ M\^48 - 
                13890\ M\^50 + 7885\ M\^52 + 21264\ M\^54 - 2956\ M\^56 - 
                22566\ M\^58 - 2956\ M\^60 + 21264\ M\^62 + 7885\ M\^64 - 
                13890\ M\^66 - 5749\ M\^68 + 6452\ M\^70 + 1134\ M\^72 - 
                809\ M\^74 + 9\ M\^76 - 852\ M\^78 + 326\ M\^80 + 
                586\ M\^82 - 600\ M\^84 + 244\ M\^86 - 49\ M\^88 + 
                4\ M\^90)\) + 
          L\^6\ \((M\^32 - 12\ M\^34 + 64\ M\^36 - 211\ M\^38 + 455\ M\^40 - 
                526\ M\^42 - 137\ M\^44 + 1497\ M\^46 - 1429\ M\^48 - 
                2475\ M\^50 + 4615\ M\^52 + 3730\ M\^54 - 6851\ M\^56 - 
                10261\ M\^58 + 8308\ M\^60 + 20943\ M\^62 - 6975\ M\^64 - 
                27211\ M\^66 + 4628\ M\^68 + 26298\ M\^70 + 1881\ M\^72 - 
                16223\ M\^74 - 4799\ M\^76 + 6768\ M\^78 + 4676\ M\^80 - 
                2976\ M\^82 - 2467\ M\^84 + 1665\ M\^86 + 647\ M\^88 - 
                932\ M\^90 + 380\ M\^92 - 74\ M\^94 + 6\ M\^96)\) + 
          L\^5\ \((3\ M\^42 - 33\ M\^44 + 139\ M\^46 - 270\ M\^48 + 
                176\ M\^50 + 247\ M\^52 - 449\ M\^54 - 159\ M\^56 + 
                197\ M\^58 + 893\ M\^60 + 2101\ M\^62 - 5412\ M\^64 - 
                5242\ M\^66 + 11536\ M\^68 + 9480\ M\^70 - 14043\ M\^72 - 
                13537\ M\^74 + 11712\ M\^76 + 17199\ M\^78 - 5297\ M\^80 - 
                14218\ M\^82 + 1890\ M\^84 + 8595\ M\^86 - 1591\ M\^88 - 
                3546\ M\^90 + 1396\ M\^92 + 811\ M\^94 - 818\ M\^96 + 
                289\ M\^98 - 51\ M\^100 + 4\ M\^102)\) + 
          L\^4\ \((3\ M\^52 - 28\ M\^54 + 101\ M\^56 - 138\ M\^58 - 
                47\ M\^60 + 282\ M\^62 - 166\ M\^64 - 255\ M\^66 + 
                898\ M\^68 - 579\ M\^70 - 1676\ M\^72 + 865\ M\^74 + 
                3381\ M\^76 + 565\ M\^78 - 5980\ M\^80 - 2871\ M\^82 + 
                8444\ M\^84 + 4441\ M\^86 - 7089\ M\^88 - 2886\ M\^90 + 
                4448\ M\^92 + 738\ M\^94 - 2022\ M\^96 + 273\ M\^98 + 
                593\ M\^100 - 391\ M\^102 + 112\ M\^104 - 16\ M\^106 + 
                M\^108)\) + 
          L\^3\ \((M\^62 - 7\ M\^64 + 21\ M\^66 - 35\ M\^68 + 4\ M\^70 + 
                86\ M\^72 - 31\ M\^74 - 219\ M\^76 + 199\ M\^78 + 3\ M\^80 - 
                97\ M\^82 + 714\ M\^84 - 411\ M\^86 - 1601\ M\^88 + 
                1101\ M\^90 + 1939\ M\^92 - 898\ M\^94 - 1183\ M\^96 + 
                644\ M\^98 + 500\ M\^100 - 391\ M\^102 - 104\ M\^104 + 
                202\ M\^106 - 91\ M\^108 + 20\ M\^110 - 2\ M\^112)\) + 
          L\ \((3\ M\^98 - 7\ M\^100 + 6\ M\^102 - 2\ M\^104 - 5\ M\^106 + 
                23\ M\^108 + 2\ M\^110 - 12\ M\^112 + 8\ M\^114 - 
                2\ M\^116)\) + 
          L\^2\ \((3\ M\^80 - 14\ M\^82 + 27\ M\^84 - 28\ M\^86 - 22\ M\^88 + 
                120\ M\^90 - 28\ M\^92 - 215\ M\^94 + 109\ M\^96 + 
                177\ M\^98 - 28\ M\^100 - 19\ M\^102 - 9\ M\^104 + 
                25\ M\^106 + 12\ M\^108 - 38\ M\^110 + 26\ M\^112 - 
                8\ M\^114 + M\^116)\)}, {"A-polynomial", Knot[9, 8], 
        L\^15\ M\^20 + M\^82 + 
          L\^14\ \((\(-2\)\ M\^12 + 6\ M\^14 - 7\ M\^16 + 3\ M\^18 + 
                3\ M\^20 - 12\ M\^22 + 23\ M\^24 + 13\ M\^26 - 17\ M\^28 + 
                5\ M\^30)\) + 
          L\^13\ \((M\^4 - 6\ M\^6 + 17\ M\^8 - 28\ M\^10 + 9\ M\^12 + 
                72\ M\^14 - 88\ M\^16 - 116\ M\^18 + 201\ M\^20 + 
                105\ M\^22 - 245\ M\^24 - 119\ M\^26 + 336\ M\^28 + 
                173\ M\^30 - 269\ M\^32 - 17\ M\^34 + 133\ M\^36 - 
                64\ M\^38 + 10\ M\^40)\) + 
          L\^12\ \((\(-1\) + 7\ M\^2 - 22\ M\^4 + 45\ M\^6 - 54\ M\^8 - 
                45\ M\^10 + 245\ M\^12 - 88\ M\^14 - 616\ M\^16 + 
                694\ M\^18 + 696\ M\^20 - 1401\ M\^22 - 427\ M\^24 + 
                1647\ M\^26 - 216\ M\^28 - 1190\ M\^30 + 995\ M\^32 + 
                1071\ M\^34 - 1044\ M\^36 - 408\ M\^38 + 905\ M\^40 - 
                185\ M\^42 - 371\ M\^44 + 298\ M\^46 - 90\ M\^48 + 
                10\ M\^50)\) + 
          L\^11\ \((1 - 11\ M\^2 + 47\ M\^4 - 85\ M\^6 + 12\ M\^8 + 
                113\ M\^10 + 179\ M\^12 - 605\ M\^14 - 431\ M\^16 + 
                1804\ M\^18 + 422\ M\^20 - 3704\ M\^22 + 1478\ M\^24 + 
                4096\ M\^26 - 4308\ M\^28 - 2743\ M\^30 + 5102\ M\^32 + 
                325\ M\^34 - 1927\ M\^36 + 2915\ M\^38 - 270\ M\^40 - 
                3127\ M\^42 + 2003\ M\^44 + 1492\ M\^46 - 1948\ M\^48 + 
                208\ M\^50 + 744\ M\^52 - 622\ M\^54 + 256\ M\^56 - 
                56\ M\^58 + 5\ M\^60)\) + 
          L\^10\ \((\(-2\)\ M\^4 + 26\ M\^6 - 146\ M\^8 + 408\ M\^10 - 
                373\ M\^12 - 810\ M\^14 + 1734\ M\^16 + 1649\ M\^18 - 
                5292\ M\^20 - 2440\ M\^22 + 10109\ M\^24 + 3740\ M\^26 - 
                13908\ M\^28 - 2656\ M\^30 + 11794\ M\^32 - 913\ M\^34 - 
                7333\ M\^36 + 2445\ M\^38 + 5976\ M\^40 + 4363\ M\^42 - 
                3225\ M\^44 - 7893\ M\^46 + 3152\ M\^48 + 7419\ M\^50 - 
                4151\ M\^52 - 3387\ M\^54 + 3464\ M\^56 - 4\ M\^58 - 
                1439\ M\^60 + 973\ M\^62 - 347\ M\^64 + 82\ M\^66 - 
                13\ M\^68 + M\^70)\) + 
          L\^9\ \((M\^8 - 19\ M\^10 + 155\ M\^12 - 652\ M\^14 + 1300\ M\^16 - 
                68\ M\^18 - 4334\ M\^20 + 3727\ M\^22 + 10183\ M\^24 - 
                12851\ M\^26 - 17902\ M\^28 + 21976\ M\^30 + 28451\ M\^32 - 
                21393\ M\^34 - 34691\ M\^36 + 3322\ M\^38 + 29348\ M\^40 + 
                15552\ M\^42 - 15649\ M\^44 - 13280\ M\^46 + 13708\ M\^48 + 
                4210\ M\^50 - 11902\ M\^52 + 4211\ M\^54 + 7148\ M\^56 - 
                7348\ M\^58 - 409\ M\^60 + 4397\ M\^62 - 2347\ M\^64 - 
                748\ M\^66 + 1624\ M\^68 - 981\ M\^70 + 317\ M\^72 - 
                55\ M\^74 + 4\ M\^76)\) + 
          L\^8\ \((4\ M\^14 - 61\ M\^16 + 404\ M\^18 - 1374\ M\^20 + 
                1971\ M\^22 + 1398\ M\^24 - 7488\ M\^26 + 1252\ M\^28 + 
                17897\ M\^30 - 6873\ M\^32 - 31806\ M\^34 + 4000\ M\^36 + 
                49455\ M\^38 + 19864\ M\^40 - 58202\ M\^42 - 65255\ M\^44 + 
                42831\ M\^46 + 97601\ M\^48 - 5314\ M\^50 - 78858\ M\^52 - 
                13595\ M\^54 + 39698\ M\^56 + 16212\ M\^58 - 10294\ M\^60 - 
                12735\ M\^62 - 1373\ M\^64 + 10117\ M\^66 + 572\ M\^68 - 
                5751\ M\^70 + 1336\ M\^72 + 1914\ M\^74 - 1547\ M\^76 + 
                515\ M\^78 - 86\ M\^80 + 6\ M\^82)\) + 
          L\^7\ \((6\ M\^20 - 86\ M\^22 + 515\ M\^24 - 1547\ M\^26 + 
                1914\ M\^28 + 1336\ M\^30 - 5751\ M\^32 + 572\ M\^34 + 
                10117\ M\^36 - 1373\ M\^38 - 12735\ M\^40 - 10294\ M\^42 + 
                16212\ M\^44 + 39698\ M\^46 - 13595\ M\^48 - 78858\ M\^50 - 
                5314\ M\^52 + 97601\ M\^54 + 42831\ M\^56 - 65255\ M\^58 - 
                58202\ M\^60 + 19864\ M\^62 + 49455\ M\^64 + 4000\ M\^66 - 
                31806\ M\^68 - 6873\ M\^70 + 17897\ M\^72 + 1252\ M\^74 - 
                7488\ M\^76 + 1398\ M\^78 + 1971\ M\^80 - 1374\ M\^82 + 
                404\ M\^84 - 61\ M\^86 + 4\ M\^88)\) + 
          L\ \((5\ M\^72 - 17\ M\^74 + 13\ M\^76 + 23\ M\^78 - 12\ M\^80 + 
                3\ M\^82 + 3\ M\^84 - 7\ M\^86 + 6\ M\^88 - 2\ M\^90)\) + 
          L\^6\ \((4\ M\^26 - 55\ M\^28 + 317\ M\^30 - 981\ M\^32 + 
                1624\ M\^34 - 748\ M\^36 - 2347\ M\^38 + 4397\ M\^40 - 
                409\ M\^42 - 7348\ M\^44 + 7148\ M\^46 + 4211\ M\^48 - 
                11902\ M\^50 + 4210\ M\^52 + 13708\ M\^54 - 13280\ M\^56 - 
                15649\ M\^58 + 15552\ M\^60 + 29348\ M\^62 + 3322\ M\^64 - 
                34691\ M\^66 - 21393\ M\^68 + 28451\ M\^70 + 21976\ M\^72 - 
                17902\ M\^74 - 12851\ M\^76 + 10183\ M\^78 + 3727\ M\^80 - 
                4334\ M\^82 - 68\ M\^84 + 1300\ M\^86 - 652\ M\^88 + 
                155\ M\^90 - 19\ M\^92 + M\^94)\) + 
          L\^5\ \((M\^32 - 13\ M\^34 + 82\ M\^36 - 347\ M\^38 + 973\ M\^40 - 
                1439\ M\^42 - 4\ M\^44 + 3464\ M\^46 - 3387\ M\^48 - 
                4151\ M\^50 + 7419\ M\^52 + 3152\ M\^54 - 7893\ M\^56 - 
                3225\ M\^58 + 4363\ M\^60 + 5976\ M\^62 + 2445\ M\^64 - 
                7333\ M\^66 - 913\ M\^68 + 11794\ M\^70 - 2656\ M\^72 - 
                13908\ M\^74 + 3740\ M\^76 + 10109\ M\^78 - 2440\ M\^80 - 
                5292\ M\^82 + 1649\ M\^84 + 1734\ M\^86 - 810\ M\^88 - 
                373\ M\^90 + 408\ M\^92 - 146\ M\^94 + 26\ M\^96 - 
                2\ M\^98)\) + 
          L\^2\ \((10\ M\^62 - 64\ M\^64 + 133\ M\^66 - 17\ M\^68 - 
                269\ M\^70 + 173\ M\^72 + 336\ M\^74 - 119\ M\^76 - 
                245\ M\^78 + 105\ M\^80 + 201\ M\^82 - 116\ M\^84 - 
                88\ M\^86 + 72\ M\^88 + 9\ M\^90 - 28\ M\^92 + 17\ M\^94 - 
                6\ M\^96 + M\^98)\) + 
          L\^3\ \((10\ M\^52 - 90\ M\^54 + 298\ M\^56 - 371\ M\^58 - 
                185\ M\^60 + 905\ M\^62 - 408\ M\^64 - 1044\ M\^66 + 
                1071\ M\^68 + 995\ M\^70 - 1190\ M\^72 - 216\ M\^74 + 
                1647\ M\^76 - 427\ M\^78 - 1401\ M\^80 + 696\ M\^82 + 
                694\ M\^84 - 616\ M\^86 - 88\ M\^88 + 245\ M\^90 - 
                45\ M\^92 - 54\ M\^94 + 45\ M\^96 - 22\ M\^98 + 7\ M\^100 - 
                M\^102)\) + 
          L\^4\ \((5\ M\^42 - 56\ M\^44 + 256\ M\^46 - 622\ M\^48 + 
                744\ M\^50 + 208\ M\^52 - 1948\ M\^54 + 1492\ M\^56 + 
                2003\ M\^58 - 3127\ M\^60 - 270\ M\^62 + 2915\ M\^64 - 
                1927\ M\^66 + 325\ M\^68 + 5102\ M\^70 - 2743\ M\^72 - 
                4308\ M\^74 + 4096\ M\^76 + 1478\ M\^78 - 3704\ M\^80 + 
                422\ M\^82 + 1804\ M\^84 - 431\ M\^86 - 605\ M\^88 + 
                179\ M\^90 + 113\ M\^92 + 12\ M\^94 - 85\ M\^96 + 47\ M\^98 - 
                11\ M\^100 + M\^102)\)}, {"A-polynomial", Knot[9, 9], 
        L\^15 + M\^186 + 
          L\^14\ \((\(-1\) + 4\ M\^2 - 8\ M\^4 + 8\ M\^6 + 2\ M\^8 - 
                16\ M\^10 + 24\ M\^12 + 13\ M\^14 - 17\ M\^16 + 6\ M\^18)\) + 
          L\^13\ \((\(-2\)\ M\^6 + 14\ M\^8 - 42\ M\^10 + 74\ M\^12 - 
                44\ M\^14 - 136\ M\^16 + 279\ M\^18 - 26\ M\^20 - 
                402\ M\^22 + 366\ M\^24 + 280\ M\^26 - 360\ M\^28 + 
                16\ M\^30 + 158\ M\^32 - 85\ M\^34 + 15\ M\^36)\) + 
          L\^12\ \((\(-M\^12\) + 12\ M\^14 - 63\ M\^16 + 180\ M\^18 - 
                254\ M\^20 - 32\ M\^22 + 756\ M\^24 - 912\ M\^26 - 
                670\ M\^28 + 2374\ M\^30 - 948\ M\^32 - 2740\ M\^34 + 
                3192\ M\^36 + 1664\ M\^38 - 3408\ M\^40 + 338\ M\^42 + 
                1967\ M\^44 - 954\ M\^46 - 401\ M\^48 + 505\ M\^50 - 
                170\ M\^52 + 20\ M\^54)\) + 
          L\^11\ \((3\ M\^20 - 30\ M\^22 + 135\ M\^24 - 332\ M\^26 + 
                363\ M\^28 + 271\ M\^30 - 1237\ M\^32 + 677\ M\^34 + 
                1930\ M\^36 - 3066\ M\^38 + 3\ M\^40 + 4678\ M\^42 - 
                5200\ M\^44 - 3486\ M\^46 + 11298\ M\^48 + 222\ M\^50 - 
                11765\ M\^52 + 4562\ M\^54 + 7854\ M\^56 - 5902\ M\^58 - 
                2243\ M\^60 + 3679\ M\^62 - 491\ M\^64 - 1103\ M\^66 + 
                700\ M\^68 - 170\ M\^70 + 15\ M\^72)\) + 
          L\^10\ \((\(-3\)\ M\^28 + 30\ M\^30 - 129\ M\^32 + 291\ M\^34 - 
                281\ M\^36 - 145\ M\^38 + 408\ M\^40 + 429\ M\^42 - 
                1075\ M\^44 - 503\ M\^46 + 1131\ M\^48 + 1580\ M\^50 - 
                255\ M\^52 - 2703\ M\^54 - 4350\ M\^56 + 3940\ M\^58 + 
                11451\ M\^60 - 3982\ M\^62 - 12657\ M\^64 + 8269\ M\^66 + 
                10067\ M\^68 - 10044\ M\^70 - 3951\ M\^72 + 8427\ M\^74 - 
                1089\ M\^76 - 3815\ M\^78 + 1921\ M\^80 + 768\ M\^82 - 
                1103\ M\^84 + 455\ M\^86 - 85\ M\^88 + 6\ M\^90)\) + 
          L\^9\ \((M\^36 - 10\ M\^38 + 43\ M\^40 - 100\ M\^42 + 150\ M\^44 - 
                275\ M\^46 + 658\ M\^48 - 621\ M\^50 - 1260\ M\^52 + 
                2629\ M\^54 + 2308\ M\^56 - 8000\ M\^58 - 887\ M\^60 + 
                13115\ M\^62 - 347\ M\^64 - 14515\ M\^66 - 1967\ M\^68 + 
                8319\ M\^70 + 10777\ M\^72 + 1410\ M\^74 - 15969\ M\^76 + 
                967\ M\^78 + 19712\ M\^80 - 6615\ M\^82 - 14470\ M\^84 + 
                9563\ M\^86 + 6400\ M\^88 - 7795\ M\^90 - 44\ M\^92 + 
                3105\ M\^94 - 1119\ M\^96 - 723\ M\^98 + 900\ M\^100 - 
                437\ M\^102 + 118\ M\^104 - 17\ M\^106 + M\^108)\) + 
          L\^8\ \((3\ M\^50 - 37\ M\^52 + 185\ M\^54 - 437\ M\^56 + 
                316\ M\^58 + 538\ M\^60 - 446\ M\^62 - 2282\ M\^64 + 
                2812\ M\^66 + 3668\ M\^68 - 5708\ M\^70 - 6290\ M\^72 + 
                9379\ M\^74 + 8608\ M\^76 - 8295\ M\^78 - 13385\ M\^80 + 
                2819\ M\^82 + 18772\ M\^84 + 3939\ M\^86 - 17929\ M\^88 + 
                3111\ M\^90 + 19994\ M\^92 - 9948\ M\^94 - 15764\ M\^96 + 
                13898\ M\^98 + 7344\ M\^100 - 11102\ M\^102 - 1388\ M\^104 + 
                6582\ M\^106 - 1556\ M\^108 - 2023\ M\^110 + 879\ M\^112 + 
                656\ M\^114 - 691\ M\^116 + 254\ M\^118 - 44\ M\^120 + 
                3\ M\^122)\) + 
          L\^7\ \((3\ M\^64 - 44\ M\^66 + 254\ M\^68 - 691\ M\^70 + 
                656\ M\^72 + 879\ M\^74 - 2023\ M\^76 - 1556\ M\^78 + 
                6582\ M\^80 - 1388\ M\^82 - 11102\ M\^84 + 7344\ M\^86 + 
                13898\ M\^88 - 15764\ M\^90 - 9948\ M\^92 + 19994\ M\^94 + 
                3111\ M\^96 - 17929\ M\^98 + 3939\ M\^100 + 18772\ M\^102 + 
                2819\ M\^104 - 13385\ M\^106 - 8295\ M\^108 + 8608\ M\^110 + 
                9379\ M\^112 - 6290\ M\^114 - 5708\ M\^116 + 3668\ M\^118 + 
                2812\ M\^120 - 2282\ M\^122 - 446\ M\^124 + 538\ M\^126 + 
                316\ M\^128 - 437\ M\^130 + 185\ M\^132 - 37\ M\^134 + 
                3\ M\^136)\) + 
          L\^6\ \((M\^78 - 17\ M\^80 + 118\ M\^82 - 437\ M\^84 + 900\ M\^86 - 
                723\ M\^88 - 1119\ M\^90 + 3105\ M\^92 - 44\ M\^94 - 
                7795\ M\^96 + 6400\ M\^98 + 9563\ M\^100 - 14470\ M\^102 - 
                6615\ M\^104 + 19712\ M\^106 + 967\ M\^108 - 15969\ M\^110 + 
                1410\ M\^112 + 10777\ M\^114 + 8319\ M\^116 - 1967\ M\^118 - 
                14515\ M\^120 - 347\ M\^122 + 13115\ M\^124 - 887\ M\^126 - 
                8000\ M\^128 + 2308\ M\^130 + 2629\ M\^132 - 1260\ M\^134 - 
                621\ M\^136 + 658\ M\^138 - 275\ M\^140 + 150\ M\^142 - 
                100\ M\^144 + 43\ M\^146 - 10\ M\^148 + M\^150)\) + 
          L\^5\ \((6\ M\^96 - 85\ M\^98 + 455\ M\^100 - 1103\ M\^102 + 
                768\ M\^104 + 1921\ M\^106 - 3815\ M\^108 - 1089\ M\^110 + 
                8427\ M\^112 - 3951\ M\^114 - 10044\ M\^116 + 10067\ M\^118 + 
                8269\ M\^120 - 12657\ M\^122 - 3982\ M\^124 + 11451\ M\^126 + 
                3940\ M\^128 - 4350\ M\^130 - 2703\ M\^132 - 255\ M\^134 + 
                1580\ M\^136 + 1131\ M\^138 - 503\ M\^140 - 1075\ M\^142 + 
                429\ M\^144 + 408\ M\^146 - 145\ M\^148 - 281\ M\^150 + 
                291\ M\^152 - 129\ M\^154 + 30\ M\^156 - 3\ M\^158)\) + 
          L\^4\ \((15\ M\^114 - 170\ M\^116 + 700\ M\^118 - 1103\ M\^120 - 
                491\ M\^122 + 3679\ M\^124 - 2243\ M\^126 - 5902\ M\^128 + 
                7854\ M\^130 + 4562\ M\^132 - 11765\ M\^134 + 222\ M\^136 + 
                11298\ M\^138 - 3486\ M\^140 - 5200\ M\^142 + 4678\ M\^144 + 
                3\ M\^146 - 3066\ M\^148 + 1930\ M\^150 + 677\ M\^152 - 
                1237\ M\^154 + 271\ M\^156 + 363\ M\^158 - 332\ M\^160 + 
                135\ M\^162 - 30\ M\^164 + 3\ M\^166)\) + 
          L\^3\ \((20\ M\^132 - 170\ M\^134 + 505\ M\^136 - 401\ M\^138 - 
                954\ M\^140 + 1967\ M\^142 + 338\ M\^144 - 3408\ M\^146 + 
                1664\ M\^148 + 3192\ M\^150 - 2740\ M\^152 - 948\ M\^154 + 
                2374\ M\^156 - 670\ M\^158 - 912\ M\^160 + 756\ M\^162 - 
                32\ M\^164 - 254\ M\^166 + 180\ M\^168 - 63\ M\^170 + 
                12\ M\^172 - M\^174)\) + 
          L\^2\ \((15\ M\^150 - 85\ M\^152 + 158\ M\^154 + 16\ M\^156 - 
                360\ M\^158 + 280\ M\^160 + 366\ M\^162 - 402\ M\^164 - 
                26\ M\^166 + 279\ M\^168 - 136\ M\^170 - 44\ M\^172 + 
                74\ M\^174 - 42\ M\^176 + 14\ M\^178 - 2\ M\^180)\) + 
          L\ \((6\ M\^168 - 17\ M\^170 + 13\ M\^172 + 24\ M\^174 - 
                16\ M\^176 + 2\ M\^178 + 8\ M\^180 - 8\ M\^182 + 4\ M\^184 - 
                M\^186)\)}, {"A-polynomial", Knot[9, 10], 
        1 + L\^11\ M\^90 + 
          L\ \((\(-3\) + 10\ M\^2 - 13\ M\^4 + 2\ M\^6 + 29\ M\^8 - 
                16\ M\^10 - 7\ M\^12 + 13\ M\^14 - 6\ M\^16 + 2\ M\^18)\) + 
          L\^2\ \((3 - 20\ M\^2 + 58\ M\^4 - 51\ M\^6 - 97\ M\^8 + 
                184\ M\^10 + 37\ M\^12 - 213\ M\^14 + 110\ M\^16 + 
                131\ M\^18 - 109\ M\^20 - 16\ M\^22 + 35\ M\^24 + 11\ M\^26 - 
                20\ M\^30 + 17\ M\^32 - 6\ M\^34 + M\^36)\) + 
          L\^3\ \((\(-1\) + 10\ M\^2 - 45\ M\^4 + 84\ M\^6 - 5\ M\^8 - 
                201\ M\^10 + 157\ M\^12 + 383\ M\^14 - 589\ M\^16 - 
                495\ M\^18 + 1247\ M\^20 + 407\ M\^22 - 1579\ M\^24 + 
                127\ M\^26 + 1371\ M\^28 - 532\ M\^30 - 689\ M\^32 + 
                581\ M\^34 + 154\ M\^36 - 325\ M\^38 + 58\ M\^40 + 103\ M\^42 
                - 76\ M\^44 + 23\ M\^46 - 3\ M\^48)\) + 
          L\^4\ \((\(-3\)\ M\^6 + 33\ M\^8 - 131\ M\^10 + 214\ M\^12 + 
                78\ M\^14 - 785\ M\^16 + 457\ M\^18 + 1657\ M\^20 - 
                1668\ M\^22 - 2299\ M\^24 + 2689\ M\^26 + 2562\ M\^28 - 
                2559\ M\^30 - 2273\ M\^32 + 1812\ M\^34 + 2002\ M\^36 - 
                1002\ M\^38 - 1476\ M\^40 + 754\ M\^42 + 895\ M\^44 - 
                635\ M\^46 - 286\ M\^48 + 389\ M\^50 - 13\ M\^52 - 
                155\ M\^54 + 96\ M\^56 - 26\ M\^58 + 3\ M\^60)\) + 
          L\^5\ \((\(-3\)\ M\^12 + 32\ M\^14 - 126\ M\^16 + 197\ M\^18 + 
                51\ M\^20 - 509\ M\^22 + 190\ M\^24 + 889\ M\^26 - 
                522\ M\^28 - 1282\ M\^30 + 511\ M\^32 + 2222\ M\^34 - 
                303\ M\^36 - 3338\ M\^38 + 504\ M\^40 + 4183\ M\^42 - 
                757\ M\^44 - 3533\ M\^46 + 1251\ M\^48 + 2003\ M\^50 - 
                1168\ M\^52 - 452\ M\^54 + 561\ M\^56 - 94\ M\^58 - 
                78\ M\^60 + 63\ M\^62 - 74\ M\^64 + 72\ M\^66 - 36\ M\^68 + 
                9\ M\^70 - M\^72)\) + 
          L\^6\ \((\(-M\^18\) + 9\ M\^20 - 36\ M\^22 + 72\ M\^24 - 
                74\ M\^26 + 63\ M\^28 - 78\ M\^30 - 94\ M\^32 + 561\ M\^34 - 
                452\ M\^36 - 1168\ M\^38 + 2003\ M\^40 + 1251\ M\^42 - 
                3533\ M\^44 - 757\ M\^46 + 4183\ M\^48 + 504\ M\^50 - 
                3338\ M\^52 - 303\ M\^54 + 2222\ M\^56 + 511\ M\^58 - 
                1282\ M\^60 - 522\ M\^62 + 889\ M\^64 + 190\ M\^66 - 
                509\ M\^68 + 51\ M\^70 + 197\ M\^72 - 126\ M\^74 + 
                32\ M\^76 - 3\ M\^78)\) + 
          L\^7\ \((3\ M\^30 - 26\ M\^32 + 96\ M\^34 - 155\ M\^36 - 
                13\ M\^38 + 389\ M\^40 - 286\ M\^42 - 635\ M\^44 + 
                895\ M\^46 + 754\ M\^48 - 1476\ M\^50 - 1002\ M\^52 + 
                2002\ M\^54 + 1812\ M\^56 - 2273\ M\^58 - 2559\ M\^60 + 
                2562\ M\^62 + 2689\ M\^64 - 2299\ M\^66 - 1668\ M\^68 + 
                1657\ M\^70 + 457\ M\^72 - 785\ M\^74 + 78\ M\^76 + 
                214\ M\^78 - 131\ M\^80 + 33\ M\^82 - 3\ M\^84)\) + 
          L\^10\ \((2\ M\^72 - 6\ M\^74 + 13\ M\^76 - 7\ M\^78 - 16\ M\^80 + 
                29\ M\^82 + 2\ M\^84 - 13\ M\^86 + 10\ M\^88 - 3\ M\^90)\) + 
          L\^8\ \((\(-3\)\ M\^42 + 23\ M\^44 - 76\ M\^46 + 103\ M\^48 + 
                58\ M\^50 - 325\ M\^52 + 154\ M\^54 + 581\ M\^56 - 
                689\ M\^58 - 532\ M\^60 + 1371\ M\^62 + 127\ M\^64 - 
                1579\ M\^66 + 407\ M\^68 + 1247\ M\^70 - 495\ M\^72 - 
                589\ M\^74 + 383\ M\^76 + 157\ M\^78 - 201\ M\^80 - 
                5\ M\^82 + 84\ M\^84 - 45\ M\^86 + 10\ M\^88 - M\^90)\) + 
          L\^9\ \((M\^54 - 6\ M\^56 + 17\ M\^58 - 20\ M\^60 + 11\ M\^64 + 
                35\ M\^66 - 16\ M\^68 - 109\ M\^70 + 131\ M\^72 + 
                110\ M\^74 - 213\ M\^76 + 37\ M\^78 + 184\ M\^80 - 
                97\ M\^82 - 51\ M\^84 + 58\ M\^86 - 20\ M\^88 + 
                3\ M\^90)\)}, {"A-polynomial", Knot[9, 11], 
        M\^8 + L\^16\ M\^140 + 
          L\ \((\(-2\)\ M\^4 + 6\ M\^6 - 8\ M\^8 + 6\ M\^10 + 3\ M\^12 - 
                17\ M\^14 + 26\ M\^16 + 15\ M\^18 - 19\ M\^20 + 6\ M\^22)\) + 
          L\^2\ \((1 - 6\ M\^2 + 13\ M\^4 - 14\ M\^6 + 10\ M\^8 + 2\ M\^10 + 
                3\ M\^12 - 53\ M\^14 - 20\ M\^16 + 203\ M\^18 - 55\ M\^20 - 
                348\ M\^22 + 335\ M\^24 + 326\ M\^26 - 359\ M\^28 - 
                21\ M\^30 + 177\ M\^32 - 89\ M\^34 + 15\ M\^36)\) + 
          L\^3\ \((\(-4\)\ M\^4 + 31\ M\^6 - 98\ M\^8 + 147\ M\^10 - 
                26\ M\^12 - 320\ M\^14 + 455\ M\^16 + 238\ M\^18 - 
                896\ M\^20 + 200\ M\^22 + 432\ M\^24 + 66\ M\^26 + 
                30\ M\^28 - 1060\ M\^30 + 940\ M\^32 + 1890\ M\^34 - 
                1836\ M\^36 - 661\ M\^38 + 1727\ M\^40 - 502\ M\^42 - 
                558\ M\^44 + 511\ M\^46 - 166\ M\^48 + 20\ M\^50)\) + 
          L\^4\ \((6\ M\^8 - 57\ M\^10 + 231\ M\^12 - 455\ M\^14 + 
                175\ M\^16 + 1057\ M\^18 - 1614\ M\^20 - 938\ M\^22 + 
                3556\ M\^24 - 1198\ M\^26 - 1912\ M\^28 + 2811\ M\^30 - 
                4554\ M\^32 - 1134\ M\^34 + 10937\ M\^36 - 4775\ M\^38 - 
                8939\ M\^40 + 11249\ M\^42 + 2082\ M\^44 - 9246\ M\^46 + 
                4387\ M\^48 + 2649\ M\^50 - 4042\ M\^52 + 1437\ M\^54 + 
                863\ M\^56 - 1246\ M\^58 + 629\ M\^60 - 154\ M\^62 + 
                15\ M\^64)\) + 
          L\^5\ \((\(-4\)\ M\^12 + 45\ M\^14 - 220\ M\^16 + 538\ M\^18 - 
                429\ M\^20 - 861\ M\^22 + 1658\ M\^24 + 1389\ M\^26 - 
                4497\ M\^28 - 193\ M\^30 + 5481\ M\^32 - 2854\ M\^34 + 
                952\ M\^36 + 8076\ M\^38 - 19979\ M\^40 - 7682\ M\^42 + 
                38245\ M\^44 - 5805\ M\^46 - 36498\ M\^48 + 30004\ M\^50 + 
                18482\ M\^52 - 36630\ M\^54 + 5592\ M\^56 + 24148\ M\^58 - 
                15490\ M\^60 - 5571\ M\^62 + 9765\ M\^64 - 2728\ M\^66 - 
                1894\ M\^68 + 2124\ M\^70 - 1093\ M\^72 + 362\ M\^74 - 
                71\ M\^76 + 6\ M\^78)\) + 
          L\^6\ \((M\^16 - 13\ M\^18 + 74\ M\^20 - 220\ M\^22 + 327\ M\^24 - 
                213\ M\^26 + 396\ M\^28 - 1283\ M\^30 + 259\ M\^32 + 
                3520\ M\^34 - 2084\ M\^36 - 4174\ M\^38 + 4292\ M\^40 - 
                9642\ M\^42 + 6385\ M\^44 + 43643\ M\^46 - 37034\ M\^48 - 
                83178\ M\^50 + 77937\ M\^52 + 83838\ M\^54 - 99318\ M\^56 - 
                26750\ M\^58 + 91593\ M\^60 - 29475\ M\^62 - 45471\ M\^64 + 
                47525\ M\^66 + 752\ M\^68 - 28909\ M\^70 + 17165\ M\^72 + 
                4458\ M\^74 - 10832\ M\^76 + 4181\ M\^78 + 2064\ M\^80 - 
                2882\ M\^82 + 1415\ M\^84 - 413\ M\^86 + 86\ M\^88 - 
                13\ M\^90 + M\^92)\) + 
          L\^7\ \((4\ M\^26 - 55\ M\^28 + 311\ M\^30 - 866\ M\^32 + 
                942\ M\^34 + 748\ M\^36 - 2178\ M\^38 - 1355\ M\^40 + 
                6397\ M\^42 - 4483\ M\^44 - 4092\ M\^46 + 21608\ M\^48 - 
                22241\ M\^50 - 53386\ M\^52 + 91893\ M\^54 + 80391\ M\^56 - 
                173391\ M\^58 - 77250\ M\^60 + 194611\ M\^62 + 26888\ M\^64 - 
                112371\ M\^66 + 50021\ M\^68 + 2178\ M\^70 - 71095\ M\^72 + 
                72339\ M\^74 + 36605\ M\^76 - 76648\ M\^78 + 3277\ M\^80 + 
                42768\ M\^82 - 17474\ M\^84 - 10536\ M\^86 + 10132\ M\^88 - 
                1071\ M\^90 - 2601\ M\^92 + 2202\ M\^94 - 1050\ M\^96 + 
                319\ M\^98 - 55\ M\^100 + 4\ M\^102)\) + 
          L\^8\ \((6\ M\^36 - 84\ M\^38 + 477\ M\^40 - 1358\ M\^42 + 
                1779\ M\^44 + 83\ M\^46 - 3558\ M\^48 + 4693\ M\^50 + 
                1807\ M\^52 - 19967\ M\^54 + 23898\ M\^56 + 35906\ M\^58 - 
                85441\ M\^60 - 35563\ M\^62 + 163041\ M\^64 + 6271\ M\^66 - 
                181044\ M\^68 + 27107\ M\^70 + 85470\ M\^72 - 34176\ M\^74 + 
                85470\ M\^76 + 27107\ M\^78 - 181044\ M\^80 + 6271\ M\^82 + 
                163041\ M\^84 - 35563\ M\^86 - 85441\ M\^88 + 35906\ M\^90 + 
                23898\ M\^92 - 19967\ M\^94 + 1807\ M\^96 + 4693\ M\^98 - 
                3558\ M\^100 + 83\ M\^102 + 1779\ M\^104 - 1358\ M\^106 + 
                477\ M\^108 - 84\ M\^110 + 6\ M\^112)\) + 
          L\^9\ \((4\ M\^46 - 55\ M\^48 + 319\ M\^50 - 1050\ M\^52 + 
                2202\ M\^54 - 2601\ M\^56 - 1071\ M\^58 + 10132\ M\^60 - 
                10536\ M\^62 - 17474\ M\^64 + 42768\ M\^66 + 3277\ M\^68 - 
                76648\ M\^70 + 36605\ M\^72 + 72339\ M\^74 - 71095\ M\^76 + 
                2178\ M\^78 + 50021\ M\^80 - 112371\ M\^82 + 26888\ M\^84 + 
                194611\ M\^86 - 77250\ M\^88 - 173391\ M\^90 + 80391\ M\^92 + 
                91893\ M\^94 - 53386\ M\^96 - 22241\ M\^98 + 21608\ M\^100 - 
                4092\ M\^102 - 4483\ M\^104 + 6397\ M\^106 - 1355\ M\^108 - 
                2178\ M\^110 + 748\ M\^112 + 942\ M\^114 - 866\ M\^116 + 
                311\ M\^118 - 55\ M\^120 + 4\ M\^122)\) + 
          L\^10\ \((M\^56 - 13\ M\^58 + 86\ M\^60 - 413\ M\^62 + 
                1415\ M\^64 - 2882\ M\^66 + 2064\ M\^68 + 4181\ M\^70 - 
                10832\ M\^72 + 4458\ M\^74 + 17165\ M\^76 - 28909\ M\^78 + 
                752\ M\^80 + 47525\ M\^82 - 45471\ M\^84 - 29475\ M\^86 + 
                91593\ M\^88 - 26750\ M\^90 - 99318\ M\^92 + 83838\ M\^94 + 
                77937\ M\^96 - 83178\ M\^98 - 37034\ M\^100 + 43643\ M\^102 + 
                6385\ M\^104 - 9642\ M\^106 + 4292\ M\^108 - 4174\ M\^110 - 
                2084\ M\^112 + 3520\ M\^114 + 259\ M\^116 - 1283\ M\^118 + 
                396\ M\^120 - 213\ M\^122 + 327\ M\^124 - 220\ M\^126 + 
                74\ M\^128 - 13\ M\^130 + M\^132)\) + 
          L\^11\ \((6\ M\^70 - 71\ M\^72 + 362\ M\^74 - 1093\ M\^76 + 
                2124\ M\^78 - 1894\ M\^80 - 2728\ M\^82 + 9765\ M\^84 - 
                5571\ M\^86 - 15490\ M\^88 + 24148\ M\^90 + 5592\ M\^92 - 
                36630\ M\^94 + 18482\ M\^96 + 30004\ M\^98 - 36498\ M\^100 - 
                5805\ M\^102 + 38245\ M\^104 - 7682\ M\^106 - 19979\ M\^108 + 
                8076\ M\^110 + 952\ M\^112 - 2854\ M\^114 + 5481\ M\^116 - 
                193\ M\^118 - 4497\ M\^120 + 1389\ M\^122 + 1658\ M\^124 - 
                861\ M\^126 - 429\ M\^128 + 538\ M\^130 - 220\ M\^132 + 
                45\ M\^134 - 4\ M\^136)\) + 
          L\^12\ \((15\ M\^84 - 154\ M\^86 + 629\ M\^88 - 1246\ M\^90 + 
                863\ M\^92 + 1437\ M\^94 - 4042\ M\^96 + 2649\ M\^98 + 
                4387\ M\^100 - 9246\ M\^102 + 2082\ M\^104 + 11249\ M\^106 - 
                8939\ M\^108 - 4775\ M\^110 + 10937\ M\^112 - 1134\ M\^114 - 
                4554\ M\^116 + 2811\ M\^118 - 1912\ M\^120 - 1198\ M\^122 + 
                3556\ M\^124 - 938\ M\^126 - 1614\ M\^128 + 1057\ M\^130 + 
                175\ M\^132 - 455\ M\^134 + 231\ M\^136 - 57\ M\^138 + 
                6\ M\^140)\) + 
          L\^13\ \((20\ M\^98 - 166\ M\^100 + 511\ M\^102 - 558\ M\^104 - 
                502\ M\^106 + 1727\ M\^108 - 661\ M\^110 - 1836\ M\^112 + 
                1890\ M\^114 + 940\ M\^116 - 1060\ M\^118 + 30\ M\^120 + 
                66\ M\^122 + 432\ M\^124 + 200\ M\^126 - 896\ M\^128 + 
                238\ M\^130 + 455\ M\^132 - 320\ M\^134 - 26\ M\^136 + 
                147\ M\^138 - 98\ M\^140 + 31\ M\^142 - 4\ M\^144)\) + 
          L\^15\ \((6\ M\^126 - 19\ M\^128 + 15\ M\^130 + 26\ M\^132 - 
                17\ M\^134 + 3\ M\^136 + 6\ M\^138 - 8\ M\^140 + 6\ M\^142 - 
                2\ M\^144)\) + 
          L\^14\ \((15\ M\^112 - 89\ M\^114 + 177\ M\^116 - 21\ M\^118 - 
                359\ M\^120 + 326\ M\^122 + 335\ M\^124 - 348\ M\^126 - 
                55\ M\^128 + 203\ M\^130 - 20\ M\^132 - 53\ M\^134 + 
                3\ M\^136 + 2\ M\^138 + 10\ M\^140 - 14\ M\^142 + 
                13\ M\^144 - 6\ M\^146 + M\^148)\)}, {"A-polynomial", 
        Knot[9, 12], 
        M\^16 + L\^17\ M\^86 + 
          L\ \((\(-4\)\ M\^12 + 14\ M\^14 - 17\ M\^16 - 3\ M\^18 + 
                37\ M\^20 - 8\ M\^22 - 7\ M\^24 + 11\ M\^26 - 9\ M\^28 + 
                3\ M\^30)\) + 
          L\^2\ \((6\ M\^8 - 42\ M\^10 + 113\ M\^12 - 102\ M\^14 - 
                130\ M\^16 + 341\ M\^18 - 111\ M\^20 - 375\ M\^22 + 
                428\ M\^24 + 243\ M\^26 - 251\ M\^28 - 31\ M\^30 + 
                31\ M\^32 + 18\ M\^34 + 14\ M\^36 - 36\ M\^38 + 33\ M\^40 - 
                16\ M\^42 + 3\ M\^44)\) + 
          L\^3\ \((\(-4\)\ M\^4 + 42\ M\^6 - 178\ M\^8 + 346\ M\^10 - 
                153\ M\^12 - 562\ M\^14 + 857\ M\^16 + 125\ M\^18 - 
                1065\ M\^20 + 655\ M\^22 - 277\ M\^24 - 290\ M\^26 + 
                1746\ M\^28 - 376\ M\^30 - 529\ M\^32 + 1102\ M\^34 - 
                1021\ M\^36 - 409\ M\^38 + 1105\ M\^40 - 292\ M\^42 - 
                348\ M\^44 + 277\ M\^46 - 91\ M\^48 + 40\ M\^50 - 35\ M\^52 + 
                21\ M\^54 - 7\ M\^56 + M\^58)\) + 
          L\^4\ \((1 - 14\ M\^2 + 85\ M\^4 - 284\ M\^6 + 548\ M\^8 - 
                511\ M\^10 - 334\ M\^12 + 2000\ M\^14 - 2344\ M\^16 - 
                2672\ M\^18 + 9079\ M\^20 - 2058\ M\^22 - 14107\ M\^24 + 
                11364\ M\^26 + 9035\ M\^28 - 18740\ M\^30 + 6554\ M\^32 + 
                17956\ M\^34 - 15451\ M\^36 - 6057\ M\^38 + 13832\ M\^40 - 
                3230\ M\^42 - 6194\ M\^44 + 4055\ M\^46 + 1483\ M\^48 - 
                1936\ M\^50 - 180\ M\^52 + 626\ M\^54 + 104\ M\^56 - 
                420\ M\^58 + 256\ M\^60 - 78\ M\^62 + 13\ M\^64 - M\^66)\) + 
          L\^5\ \((\(-3\) + 43\ M\^2 - 260\ M\^4 + 802\ M\^6 - 1061\ M\^8 - 
                672\ M\^10 + 3704\ M\^12 - 1314\ M\^14 - 8977\ M\^16 + 
                11444\ M\^18 + 8516\ M\^20 - 29920\ M\^22 + 9350\ M\^24 + 
                49576\ M\^26 - 50830\ M\^28 - 50264\ M\^30 + 92704\ M\^32 + 
                19221\ M\^34 - 104314\ M\^36 + 35575\ M\^38 + 87004\ M\^40 - 
                66320\ M\^42 - 46567\ M\^44 + 64802\ M\^46 + 10146\ M\^48 - 
                43022\ M\^50 + 7395\ M\^52 + 20568\ M\^54 - 10496\ M\^56 - 
                5143\ M\^58 + 5862\ M\^60 - 461\ M\^62 - 1773\ M\^64 + 
                1201\ M\^66 - 394\ M\^68 + 72\ M\^70 - 6\ M\^72)\) + 
          L\^6\ \((3 - 44\ M\^2 + 276\ M\^4 - 891\ M\^6 + 1245\ M\^8 + 
                733\ M\^10 - 4109\ M\^12 - 594\ M\^14 + 15011\ M\^16 - 
                8202\ M\^18 - 34409\ M\^20 + 34438\ M\^22 + 60987\ M\^24 - 
                94703\ M\^26 - 73977\ M\^28 + 189143\ M\^30 + 56842\ M\^32 - 
                299776\ M\^34 - 3977\ M\^36 + 371998\ M\^38 - 73306\ M\^40 - 
                354214\ M\^42 + 173862\ M\^44 + 275906\ M\^46 - 
                229136\ M\^48 - 158193\ M\^50 + 216275\ M\^52 + 
                45335\ M\^54 - 147639\ M\^56 + 20249\ M\^58 + 68390\ M\^60 - 
                33871\ M\^62 - 14145\ M\^64 + 17592\ M\^66 - 3044\ M\^68 - 
                3833\ M\^70 + 3034\ M\^72 - 1058\ M\^74 + 193\ M\^76 - 
                15\ M\^78)\) + 
          L\^7\ \((\(-1\) + 15\ M\^2 - 97\ M\^4 + 335\ M\^6 - 635\ M\^8 + 
                706\ M\^10 - 1179\ M\^12 + 2950\ M\^14 - 533\ M\^16 - 
                13372\ M\^18 + 15634\ M\^20 + 33586\ M\^22 - 75226\ M\^24 - 
                32552\ M\^26 + 193413\ M\^28 - 33672\ M\^30 - 360739\ M\^32 + 
                200745\ M\^34 + 529141\ M\^36 - 425434\ M\^38 - 
                660931\ M\^40 + 607781\ M\^42 + 711176\ M\^44 - 
                667025\ M\^46 - 616708\ M\^48 + 658997\ M\^50 + 
                433400\ M\^52 - 576197\ M\^54 - 216283\ M\^56 + 
                429427\ M\^58 + 35472\ M\^60 - 250776\ M\^62 + 52819\ M\^64 + 
                101631\ M\^66 - 58765\ M\^68 - 16267\ M\^70 + 26756\ M\^72 - 
                6191\ M\^74 - 5119\ M\^76 + 4484\ M\^78 - 1575\ M\^80 + 
                277\ M\^82 - 20\ M\^84)\) + 
          L\^8\ \((\(-6\)\ M\^6 + 90\ M\^8 - 558\ M\^10 + 1754\ M\^12 - 
                2410\ M\^14 - 1043\ M\^16 + 6672\ M\^18 + 294\ M\^20 - 
                21100\ M\^22 + 9654\ M\^24 + 54596\ M\^26 - 45976\ M\^28 - 
                132158\ M\^30 + 173510\ M\^32 + 241841\ M\^34 - 
                449781\ M\^36 - 346568\ M\^38 + 871505\ M\^40 + 
                400126\ M\^42 - 1303528\ M\^44 - 410715\ M\^46 + 
                1545319\ M\^48 + 393558\ M\^50 - 1448791\ M\^52 - 
                280983\ M\^54 + 1145076\ M\^56 + 129400\ M\^58 - 
                771867\ M\^60 - 1811\ M\^62 + 442789\ M\^64 - 68986\ M\^66 - 
                205888\ M\^68 + 77083\ M\^70 + 70683\ M\^72 - 53476\ M\^74 - 
                8414\ M\^76 + 21305\ M\^78 - 5132\ M\^80 - 4494\ M\^82 + 
                3837\ M\^84 - 1301\ M\^86 + 219\ M\^88 - 15\ M\^90)\) + 
          L\^16\ \((3\ M\^72 - 9\ M\^74 + 11\ M\^76 - 7\ M\^78 - 8\ M\^80 + 
                37\ M\^82 - 3\ M\^84 - 17\ M\^86 + 14\ M\^88 - 4\ M\^90)\) + 
          L\^15\ \((3\ M\^58 - 16\ M\^60 + 33\ M\^62 - 36\ M\^64 + 
                14\ M\^66 + 18\ M\^68 + 31\ M\^70 - 31\ M\^72 - 251\ M\^74 + 
                243\ M\^76 + 428\ M\^78 - 375\ M\^80 - 111\ M\^82 + 
                341\ M\^84 - 130\ M\^86 - 102\ M\^88 + 113\ M\^90 - 
                42\ M\^92 + 6\ M\^94)\) + 
          L\^9\ \((\(-15\)\ M\^12 + 219\ M\^14 - 1301\ M\^16 + 3837\ M\^18 - 
                4494\ M\^20 - 5132\ M\^22 + 21305\ M\^24 - 8414\ M\^26 - 
                53476\ M\^28 + 70683\ M\^30 + 77083\ M\^32 - 205888\ M\^34 - 
                68986\ M\^36 + 442789\ M\^38 - 1811\ M\^40 - 771867\ M\^42 + 
                129400\ M\^44 + 1145076\ M\^46 - 280983\ M\^48 - 
                1448791\ M\^50 + 393558\ M\^52 + 1545319\ M\^54 - 
                410715\ M\^56 - 1303528\ M\^58 + 400126\ M\^60 + 
                871505\ M\^62 - 346568\ M\^64 - 449781\ M\^66 + 
                241841\ M\^68 + 173510\ M\^70 - 132158\ M\^72 - 
                45976\ M\^74 + 54596\ M\^76 + 9654\ M\^78 - 21100\ M\^80 + 
                294\ M\^82 + 6672\ M\^84 - 1043\ M\^86 - 2410\ M\^88 + 
                1754\ M\^90 - 558\ M\^92 + 90\ M\^94 - 6\ M\^96)\) + 
          L\^14\ \((M\^44 - 7\ M\^46 + 21\ M\^48 - 35\ M\^50 + 40\ M\^52 - 
                91\ M\^54 + 277\ M\^56 - 348\ M\^58 - 292\ M\^60 + 
                1105\ M\^62 - 409\ M\^64 - 1021\ M\^66 + 1102\ M\^68 - 
                529\ M\^70 - 376\ M\^72 + 1746\ M\^74 - 290\ M\^76 - 
                277\ M\^78 + 655\ M\^80 - 1065\ M\^82 + 125\ M\^84 + 
                857\ M\^86 - 562\ M\^88 - 153\ M\^90 + 346\ M\^92 - 
                178\ M\^94 + 42\ M\^96 - 4\ M\^98)\) + 
          L\^12\ \((\(-6\)\ M\^30 + 72\ M\^32 - 394\ M\^34 + 1201\ M\^36 - 
                1773\ M\^38 - 461\ M\^40 + 5862\ M\^42 - 5143\ M\^44 - 
                10496\ M\^46 + 20568\ M\^48 + 7395\ M\^50 - 43022\ M\^52 + 
                10146\ M\^54 + 64802\ M\^56 - 46567\ M\^58 - 66320\ M\^60 + 
                87004\ M\^62 + 35575\ M\^64 - 104314\ M\^66 + 19221\ M\^68 + 
                92704\ M\^70 - 50264\ M\^72 - 50830\ M\^74 + 49576\ M\^76 + 
                9350\ M\^78 - 29920\ M\^80 + 8516\ M\^82 + 11444\ M\^84 - 
                8977\ M\^86 - 1314\ M\^88 + 3704\ M\^90 - 672\ M\^92 - 
                1061\ M\^94 + 802\ M\^96 - 260\ M\^98 + 43\ M\^100 - 
                3\ M\^102)\) + 
          L\^10\ \((\(-20\)\ M\^18 + 277\ M\^20 - 1575\ M\^22 + 4484\ M\^24 - 
                5119\ M\^26 - 6191\ M\^28 + 26756\ M\^30 - 16267\ M\^32 - 
                58765\ M\^34 + 101631\ M\^36 + 52819\ M\^38 - 250776\ M\^40 + 
                35472\ M\^42 + 429427\ M\^44 - 216283\ M\^46 - 
                576197\ M\^48 + 433400\ M\^50 + 658997\ M\^52 - 
                616708\ M\^54 - 667025\ M\^56 + 711176\ M\^58 + 
                607781\ M\^60 - 660931\ M\^62 - 425434\ M\^64 + 
                529141\ M\^66 + 200745\ M\^68 - 360739\ M\^70 - 
                33672\ M\^72 + 193413\ M\^74 - 32552\ M\^76 - 75226\ M\^78 + 
                33586\ M\^80 + 15634\ M\^82 - 13372\ M\^84 - 533\ M\^86 + 
                2950\ M\^88 - 1179\ M\^90 + 706\ M\^92 - 635\ M\^94 + 
                335\ M\^96 - 97\ M\^98 + 15\ M\^100 - M\^102)\) + 
          L\^13\ \((\(-M\^36\) + 13\ M\^38 - 78\ M\^40 + 256\ M\^42 - 
                420\ M\^44 + 104\ M\^46 + 626\ M\^48 - 180\ M\^50 - 
                1936\ M\^52 + 1483\ M\^54 + 4055\ M\^56 - 6194\ M\^58 - 
                3230\ M\^60 + 13832\ M\^62 - 6057\ M\^64 - 15451\ M\^66 + 
                17956\ M\^68 + 6554\ M\^70 - 18740\ M\^72 + 9035\ M\^74 + 
                11364\ M\^76 - 14107\ M\^78 - 2058\ M\^80 + 9079\ M\^82 - 
                2672\ M\^84 - 2344\ M\^86 + 2000\ M\^88 - 334\ M\^90 - 
                511\ M\^92 + 548\ M\^94 - 284\ M\^96 + 85\ M\^98 - 
                14\ M\^100 + M\^102)\) + 
          L\^11\ \((\(-15\)\ M\^24 + 193\ M\^26 - 1058\ M\^28 + 3034\ M\^30 - 
                3833\ M\^32 - 3044\ M\^34 + 17592\ M\^36 - 14145\ M\^38 - 
                33871\ M\^40 + 68390\ M\^42 + 20249\ M\^44 - 147639\ M\^46 + 
                45335\ M\^48 + 216275\ M\^50 - 158193\ M\^52 - 
                229136\ M\^54 + 275906\ M\^56 + 173862\ M\^58 - 
                354214\ M\^60 - 73306\ M\^62 + 371998\ M\^64 - 3977\ M\^66 - 
                299776\ M\^68 + 56842\ M\^70 + 189143\ M\^72 - 73977\ M\^74 - 
                94703\ M\^76 + 60987\ M\^78 + 34438\ M\^80 - 34409\ M\^82 - 
                8202\ M\^84 + 15011\ M\^86 - 594\ M\^88 - 4109\ M\^90 + 
                733\ M\^92 + 1245\ M\^94 - 891\ M\^96 + 276\ M\^98 - 
                44\ M\^100 + 3\ M\^102)\)}, {"A-polynomial", Knot[9, 13], 
        1 + L\^18\ M\^148 + 
          L\ \((\(-4\) + 14\ M\^2 - 19\ M\^4 + 2\ M\^6 + 40\ M\^8 - 
                17\ M\^10 - 7\ M\^12 + 16\ M\^14 - 10\ M\^16 + 3\ M\^18)\) + 
          L\^2\ \((6 - 38\ M\^2 + 99\ M\^4 - 108\ M\^6 - 41\ M\^8 + 
                210\ M\^10 - 179\ M\^12 - 19\ M\^14 + 287\ M\^16 - 
                69\ M\^18 + 23\ M\^20 + 3\ M\^22 - 117\ M\^24 + 121\ M\^26 + 
                15\ M\^28 - 80\ M\^30 + 57\ M\^32 - 20\ M\^34 + 3\ M\^36)\) + 
          L\^3\ \((\(-4\) + 34\ M\^2 - 125\ M\^4 + 238\ M\^6 - 242\ M\^8 + 
                183\ M\^10 - 110\ M\^12 - 603\ M\^14 + 1882\ M\^16 - 
                1014\ M\^18 - 3370\ M\^20 + 4628\ M\^22 + 2204\ M\^24 - 
                5364\ M\^26 + 1383\ M\^28 + 3074\ M\^30 - 2560\ M\^32 - 
                104\ M\^34 + 1218\ M\^36 - 571\ M\^38 - 31\ M\^40 + 
                127\ M\^42 - 134\ M\^44 + 145\ M\^46 - 101\ M\^48 + 
                42\ M\^50 - 10\ M\^52 + M\^54)\) + 
          L\^4\ \((1 - 10\ M\^2 + 45\ M\^4 - 126\ M\^6 + 318\ M\^8 - 
                850\ M\^10 + 1539\ M\^12 - 459\ M\^14 - 3697\ M\^16 + 
                5125\ M\^18 + 3038\ M\^20 - 10180\ M\^22 + 1159\ M\^24 + 
                10380\ M\^26 - 5013\ M\^28 - 8726\ M\^30 + 10409\ M\^32 + 
                11240\ M\^34 - 13405\ M\^36 - 9600\ M\^38 + 15542\ M\^40 + 
                3984\ M\^42 - 12236\ M\^44 + 1733\ M\^46 + 6450\ M\^48 - 
                3814\ M\^50 - 1257\ M\^52 + 2199\ M\^54 - 507\ M\^56 - 
                615\ M\^58 + 600\ M\^60 - 269\ M\^62 + 73\ M\^64 - 
                12\ M\^66 + M\^68)\) + 
          L\^5\ \((6\ M\^6 - 74\ M\^8 + 388\ M\^10 - 1074\ M\^12 + 
                1580\ M\^14 - 1127\ M\^16 + 521\ M\^18 + 845\ M\^20 - 
                7898\ M\^22 + 11468\ M\^24 + 13359\ M\^26 - 40842\ M\^28 - 
                319\ M\^30 + 69622\ M\^32 - 25802\ M\^34 - 88556\ M\^36 + 
                42762\ M\^38 + 105799\ M\^40 - 27719\ M\^42 - 113324\ M\^44 + 
                13085\ M\^46 + 112420\ M\^48 - 12385\ M\^50 - 89083\ M\^52 + 
                24120\ M\^54 + 50394\ M\^56 - 26746\ M\^58 - 17301\ M\^60 + 
                18411\ M\^62 - 149\ M\^64 - 6586\ M\^66 + 2691\ M\^68 + 
                1008\ M\^70 - 1490\ M\^72 + 746\ M\^74 - 216\ M\^76 + 
                37\ M\^78 - 3\ M\^80)\) + 
          L\^6\ \((15\ M\^12 - 214\ M\^14 + 1266\ M\^16 - 3828\ M\^18 + 
                5184\ M\^20 + 1877\ M\^22 - 15205\ M\^24 + 11315\ M\^26 + 
                19460\ M\^28 - 35667\ M\^30 + 4219\ M\^32 + 40576\ M\^34 - 
                56470\ M\^36 + 2796\ M\^38 + 137673\ M\^40 - 120476\ M\^42 - 
                239318\ M\^44 + 291284\ M\^46 + 339816\ M\^48 - 
                431600\ M\^50 - 349404\ M\^52 + 493278\ M\^54 + 
                266409\ M\^56 - 430359\ M\^58 - 131179\ M\^60 + 
                288753\ M\^62 + 31560\ M\^64 - 148206\ M\^66 + 8615\ M\^68 + 
                57307\ M\^70 - 10140\ M\^72 - 19717\ M\^74 + 7027\ M\^76 + 
                5051\ M\^78 - 3163\ M\^80 - 1155\ M\^82 + 1927\ M\^84 - 
                973\ M\^86 + 269\ M\^88 - 42\ M\^90 + 3\ M\^92)\) + 
          L\^7\ \((20\ M\^18 - 317\ M\^20 + 2076\ M\^22 - 6989\ M\^24 + 
                10747\ M\^26 + 4111\ M\^28 - 42586\ M\^30 + 42101\ M\^32 + 
                72673\ M\^34 - 177721\ M\^36 - 4701\ M\^38 + 360189\ M\^40 - 
                252336\ M\^42 - 460104\ M\^44 + 717489\ M\^46 + 
                325422\ M\^48 - 1263900\ M\^50 + 31659\ M\^52 + 
                1694766\ M\^54 - 433054\ M\^56 - 1805129\ M\^58 + 
                731767\ M\^60 + 1598631\ M\^62 - 804126\ M\^64 - 
                1144932\ M\^66 + 697698\ M\^68 + 651047\ M\^70 - 
                494591\ M\^72 - 268096\ M\^74 + 280236\ M\^76 + 
                61706\ M\^78 - 121522\ M\^80 + 10096\ M\^82 + 30873\ M\^84 - 
                10460\ M\^86 - 2673\ M\^88 + 2500\ M\^90 - 1649\ M\^92 + 
                1712\ M\^94 - 1192\ M\^96 + 485\ M\^98 - 117\ M\^100 + 
                16\ M\^102 - M\^104)\) + 
          L\^8\ \((15\ M\^24 - 258\ M\^26 + 1860\ M\^28 - 7058\ M\^30 + 
                13227\ M\^32 - 981\ M\^34 - 47548\ M\^36 + 69564\ M\^38 + 
                69789\ M\^40 - 273708\ M\^42 + 63034\ M\^44 + 594049\ M\^46 - 
                520253\ M\^48 - 900017\ M\^50 + 1402529\ M\^52 + 
                981125\ M\^54 - 2577094\ M\^56 - 756438\ M\^58 + 
                3714387\ M\^60 + 304367\ M\^62 - 4384016\ M\^64 + 
                238659\ M\^66 + 4353506\ M\^68 - 655821\ M\^70 - 
                3564021\ M\^72 + 875545\ M\^74 + 2384516\ M\^76 - 
                833314\ M\^78 - 1259196\ M\^80 + 596314\ M\^82 + 
                508080\ M\^84 - 324390\ M\^86 - 148218\ M\^88 + 
                130458\ M\^90 + 34542\ M\^92 - 46249\ M\^94 - 3561\ M\^96 + 
                14068\ M\^98 - 1357\ M\^100 - 4783\ M\^102 + 3325\ M\^104 - 
                1110\ M\^106 + 211\ M\^108 - 22\ M\^110 + M\^112)\) + 
          L\^9\ \((6\ M\^30 - 112\ M\^32 + 902\ M\^34 - 3939\ M\^36 + 
                9210\ M\^38 - 6429\ M\^40 - 22163\ M\^42 + 52290\ M\^44 + 
                16369\ M\^46 - 181843\ M\^48 + 104091\ M\^50 + 
                407315\ M\^52 - 477357\ M\^54 - 713085\ M\^56 + 
                1291165\ M\^58 + 975724\ M\^60 - 2599503\ M\^62 - 
                1046810\ M\^64 + 4192754\ M\^66 + 812818\ M\^68 - 
                5537535\ M\^70 - 295929\ M\^72 + 6092742\ M\^74 - 
                295929\ M\^76 - 5537535\ M\^78 + 812818\ M\^80 + 
                4192754\ M\^82 - 1046810\ M\^84 - 2599503\ M\^86 + 
                975724\ M\^88 + 1291165\ M\^90 - 713085\ M\^92 - 
                477357\ M\^94 + 407315\ M\^96 + 104091\ M\^98 - 
                181843\ M\^100 + 16369\ M\^102 + 52290\ M\^104 - 
                22163\ M\^106 - 6429\ M\^108 + 9210\ M\^110 - 3939\ M\^112 + 
                902\ M\^114 - 112\ M\^116 + 6\ M\^118)\) + 
          L\^10\ \((M\^36 - 22\ M\^38 + 211\ M\^40 - 1110\ M\^42 + 
                3325\ M\^44 - 4783\ M\^46 - 1357\ M\^48 + 14068\ M\^50 - 
                3561\ M\^52 - 46249\ M\^54 + 34542\ M\^56 + 130458\ M\^58 - 
                148218\ M\^60 - 324390\ M\^62 + 508080\ M\^64 + 
                596314\ M\^66 - 1259196\ M\^68 - 833314\ M\^70 + 
                2384516\ M\^72 + 875545\ M\^74 - 3564021\ M\^76 - 
                655821\ M\^78 + 4353506\ M\^80 + 238659\ M\^82 - 
                4384016\ M\^84 + 304367\ M\^86 + 3714387\ M\^88 - 
                756438\ M\^90 - 2577094\ M\^92 + 981125\ M\^94 + 
                1402529\ M\^96 - 900017\ M\^98 - 520253\ M\^100 + 
                594049\ M\^102 + 63034\ M\^104 - 273708\ M\^106 + 
                69789\ M\^108 + 69564\ M\^110 - 47548\ M\^112 - 981\ M\^114 + 
                13227\ M\^116 - 7058\ M\^118 + 1860\ M\^120 - 258\ M\^122 + 
                15\ M\^124)\) + 
          L\^11\ \((\(-M\^44\) + 16\ M\^46 - 117\ M\^48 + 485\ M\^50 - 
                1192\ M\^52 + 1712\ M\^54 - 1649\ M\^56 + 2500\ M\^58 - 
                2673\ M\^60 - 10460\ M\^62 + 30873\ M\^64 + 10096\ M\^66 - 
                121522\ M\^68 + 61706\ M\^70 + 280236\ M\^72 - 
                268096\ M\^74 - 494591\ M\^76 + 651047\ M\^78 + 
                697698\ M\^80 - 1144932\ M\^82 - 804126\ M\^84 + 
                1598631\ M\^86 + 731767\ M\^88 - 1805129\ M\^90 - 
                433054\ M\^92 + 1694766\ M\^94 + 31659\ M\^96 - 
                1263900\ M\^98 + 325422\ M\^100 + 717489\ M\^102 - 
                460104\ M\^104 - 252336\ M\^106 + 360189\ M\^108 - 
                4701\ M\^110 - 177721\ M\^112 + 72673\ M\^114 + 
                42101\ M\^116 - 42586\ M\^118 + 4111\ M\^120 + 
                10747\ M\^122 - 6989\ M\^124 + 2076\ M\^126 - 317\ M\^128 + 
                20\ M\^130)\) + 
          L\^12\ \((3\ M\^56 - 42\ M\^58 + 269\ M\^60 - 973\ M\^62 + 
                1927\ M\^64 - 1155\ M\^66 - 3163\ M\^68 + 5051\ M\^70 + 
                7027\ M\^72 - 19717\ M\^74 - 10140\ M\^76 + 57307\ M\^78 + 
                8615\ M\^80 - 148206\ M\^82 + 31560\ M\^84 + 288753\ M\^86 - 
                131179\ M\^88 - 430359\ M\^90 + 266409\ M\^92 + 
                493278\ M\^94 - 349404\ M\^96 - 431600\ M\^98 + 
                339816\ M\^100 + 291284\ M\^102 - 239318\ M\^104 - 
                120476\ M\^106 + 137673\ M\^108 + 2796\ M\^110 - 
                56470\ M\^112 + 40576\ M\^114 + 4219\ M\^116 - 
                35667\ M\^118 + 19460\ M\^120 + 11315\ M\^122 - 
                15205\ M\^124 + 1877\ M\^126 + 5184\ M\^128 - 3828\ M\^130 + 
                1266\ M\^132 - 214\ M\^134 + 15\ M\^136)\) + 
          L\^13\ \((\(-3\)\ M\^68 + 37\ M\^70 - 216\ M\^72 + 746\ M\^74 - 
                1490\ M\^76 + 1008\ M\^78 + 2691\ M\^80 - 6586\ M\^82 - 
                149\ M\^84 + 18411\ M\^86 - 17301\ M\^88 - 26746\ M\^90 + 
                50394\ M\^92 + 24120\ M\^94 - 89083\ M\^96 - 12385\ M\^98 + 
                112420\ M\^100 + 13085\ M\^102 - 113324\ M\^104 - 
                27719\ M\^106 + 105799\ M\^108 + 42762\ M\^110 - 
                88556\ M\^112 - 25802\ M\^114 + 69622\ M\^116 - 319\ M\^118 - 
                40842\ M\^120 + 13359\ M\^122 + 11468\ M\^124 - 
                7898\ M\^126 + 845\ M\^128 + 521\ M\^130 - 1127\ M\^132 + 
                1580\ M\^134 - 1074\ M\^136 + 388\ M\^138 - 74\ M\^140 + 
                6\ M\^142)\) + 
          L\^17\ \((3\ M\^130 - 10\ M\^132 + 16\ M\^134 - 7\ M\^136 - 
                17\ M\^138 + 40\ M\^140 + 2\ M\^142 - 19\ M\^144 + 
                14\ M\^146 - 4\ M\^148)\) + 
          L\^15\ \((M\^94 - 10\ M\^96 + 42\ M\^98 - 101\ M\^100 + 
                145\ M\^102 - 134\ M\^104 + 127\ M\^106 - 31\ M\^108 - 
                571\ M\^110 + 1218\ M\^112 - 104\ M\^114 - 2560\ M\^116 + 
                3074\ M\^118 + 1383\ M\^120 - 5364\ M\^122 + 2204\ M\^124 + 
                4628\ M\^126 - 3370\ M\^128 - 1014\ M\^130 + 1882\ M\^132 - 
                603\ M\^134 - 110\ M\^136 + 183\ M\^138 - 242\ M\^140 + 
                238\ M\^142 - 125\ M\^144 + 34\ M\^146 - 4\ M\^148)\) + 
          L\^14\ \((M\^80 - 12\ M\^82 + 73\ M\^84 - 269\ M\^86 + 600\ M\^88 - 
                615\ M\^90 - 507\ M\^92 + 2199\ M\^94 - 1257\ M\^96 - 
                3814\ M\^98 + 6450\ M\^100 + 1733\ M\^102 - 12236\ M\^104 + 
                3984\ M\^106 + 15542\ M\^108 - 9600\ M\^110 - 13405\ M\^112 + 
                11240\ M\^114 + 10409\ M\^116 - 8726\ M\^118 - 5013\ M\^120 + 
                10380\ M\^122 + 1159\ M\^124 - 10180\ M\^126 + 3038\ M\^128 + 
                5125\ M\^130 - 3697\ M\^132 - 459\ M\^134 + 1539\ M\^136 - 
                850\ M\^138 + 318\ M\^140 - 126\ M\^142 + 45\ M\^144 - 
                10\ M\^146 + M\^148)\) + 
          L\^16\ \((3\ M\^112 - 20\ M\^114 + 57\ M\^116 - 80\ M\^118 + 
                15\ M\^120 + 121\ M\^122 - 117\ M\^124 + 3\ M\^126 + 
                23\ M\^128 - 69\ M\^130 + 287\ M\^132 - 19\ M\^134 - 
                179\ M\^136 + 210\ M\^138 - 41\ M\^140 - 108\ M\^142 + 
                99\ M\^144 - 38\ M\^146 + 6\ M\^148)\)}, {"A-polynomial", 
        Knot[9, 14], 
        M\^48 + L\^18\ M\^48 + 
          L\^17\ \((\(-2\)\ M\^36 + 7\ M\^38 - 11\ M\^40 + 9\ M\^42 + 
                2\ M\^44 - 22\ M\^46 + 37\ M\^48 + 16\ M\^50 - 26\ M\^52 + 
                8\ M\^54)\) + 
          L\ \((8\ M\^42 - 26\ M\^44 + 16\ M\^46 + 37\ M\^48 - 22\ M\^50 + 
                2\ M\^52 + 9\ M\^54 - 11\ M\^56 + 7\ M\^58 - 2\ M\^60)\) + 
          L\^16\ \((M\^24 - 6\ M\^26 + 15\ M\^28 - 22\ M\^30 + 26\ M\^32 - 
                34\ M\^34 + 66\ M\^36 - 70\ M\^38 - 174\ M\^40 + 488\ M\^42 - 
                141\ M\^44 - 764\ M\^46 + 843\ M\^48 + 516\ M\^50 - 
                832\ M\^52 + 64\ M\^54 + 321\ M\^56 - 172\ M\^58 + 
                28\ M\^60)\) + 
          L\^15\ \((\(-M\^16\) + 8\ M\^18 - 28\ M\^20 + 69\ M\^22 - 
                139\ M\^24 + 159\ M\^26 + 32\ M\^28 - 286\ M\^30 + 
                64\ M\^32 + 594\ M\^34 - 49\ M\^36 - 2094\ M\^38 + 
                916\ M\^40 + 5128\ M\^42 - 5176\ M\^44 - 6306\ M\^46 + 
                11177\ M\^48 + 3630\ M\^50 - 11495\ M\^52 + 1685\ M\^54 + 
                6229\ M\^56 - 3279\ M\^58 - 1044\ M\^60 + 1452\ M\^62 - 
                486\ M\^64 + 56\ M\^66)\) + 
          L\^2\ \((28\ M\^36 - 172\ M\^38 + 321\ M\^40 + 64\ M\^42 - 
                832\ M\^44 + 516\ M\^46 + 843\ M\^48 - 764\ M\^50 - 
                141\ M\^52 + 488\ M\^54 - 174\ M\^56 - 70\ M\^58 + 
                66\ M\^60 - 34\ M\^62 + 26\ M\^64 - 22\ M\^66 + 15\ M\^68 - 
                6\ M\^70 + M\^72)\) + 
          L\^14\ \((4\ M\^12 - 41\ M\^14 + 183\ M\^16 - 451\ M\^18 + 
                579\ M\^20 + 32\ M\^22 - 1538\ M\^24 + 2159\ M\^26 + 
                1165\ M\^28 - 6951\ M\^30 + 4724\ M\^32 + 10954\ M\^34 - 
                16354\ M\^36 - 12228\ M\^38 + 33260\ M\^40 + 7714\ M\^42 - 
                53831\ M\^44 + 3953\ M\^46 + 71203\ M\^48 - 18533\ M\^50 - 
                64627\ M\^52 + 33658\ M\^54 + 37932\ M\^56 - 33975\ M\^58 - 
                8828\ M\^60 + 18825\ M\^62 - 3910\ M\^64 - 4356\ M\^66 + 
                3028\ M\^68 - 760\ M\^70 + 70\ M\^72)\) + 
          L\^13\ \((\(-6\)\ M\^8 + 75\ M\^10 - 408\ M\^12 + 1188\ M\^14 - 
                1656\ M\^16 - 386\ M\^18 + 5355\ M\^20 - 7063\ M\^22 - 
                3637\ M\^24 + 23941\ M\^26 - 24152\ M\^28 - 28971\ M\^30 + 
                81168\ M\^32 - 4155\ M\^34 - 141583\ M\^36 + 83433\ M\^38 + 
                163646\ M\^40 - 178913\ M\^42 - 143490\ M\^44 + 
                244899\ M\^46 + 106755\ M\^48 - 244761\ M\^50 - 
                54529\ M\^52 + 208639\ M\^54 + 8093\ M\^56 - 148841\ M\^58 + 
                24778\ M\^60 + 82330\ M\^62 - 35879\ M\^64 - 26896\ M\^66 + 
                23387\ M\^68 + 541\ M\^70 - 7132\ M\^72 + 3452\ M\^74 - 
                710\ M\^76 + 56\ M\^78)\) + 
          L\^3\ \((56\ M\^30 - 486\ M\^32 + 1452\ M\^34 - 1044\ M\^36 - 
                3279\ M\^38 + 6229\ M\^40 + 1685\ M\^42 - 11495\ M\^44 + 
                3630\ M\^46 + 11177\ M\^48 - 6306\ M\^50 - 5176\ M\^52 + 
                5128\ M\^54 + 916\ M\^56 - 2094\ M\^58 - 49\ M\^60 + 
                594\ M\^62 + 64\ M\^64 - 286\ M\^66 + 32\ M\^68 + 
                159\ M\^70 - 139\ M\^72 + 69\ M\^74 - 28\ M\^76 + 8\ M\^78 - 
                M\^80)\) + 
          L\^4\ \((70\ M\^24 - 760\ M\^26 + 3028\ M\^28 - 4356\ M\^30 - 
                3910\ M\^32 + 18825\ M\^34 - 8828\ M\^36 - 33975\ M\^38 + 
                37932\ M\^40 + 33658\ M\^42 - 64627\ M\^44 - 18533\ M\^46 + 
                71203\ M\^48 + 3953\ M\^50 - 53831\ M\^52 + 7714\ M\^54 + 
                33260\ M\^56 - 12228\ M\^58 - 16354\ M\^60 + 10954\ M\^62 + 
                4724\ M\^64 - 6951\ M\^66 + 1165\ M\^68 + 2159\ M\^70 - 
                1538\ M\^72 + 32\ M\^74 + 579\ M\^76 - 451\ M\^78 + 
                183\ M\^80 - 41\ M\^82 + 4\ M\^84)\) + 
          L\^12\ \((4\ M\^4 - 59\ M\^6 + 377\ M\^8 - 1311\ M\^10 + 
                2447\ M\^12 - 1333\ M\^14 - 4446\ M\^16 + 11122\ M\^18 - 
                6005\ M\^20 - 24001\ M\^22 + 57003\ M\^24 - 5199\ M\^26 - 
                142676\ M\^28 + 125710\ M\^30 + 207112\ M\^32 - 
                347768\ M\^34 - 143606\ M\^36 + 585894\ M\^38 - 
                92425\ M\^40 - 718197\ M\^42 + 398335\ M\^44 + 
                655279\ M\^46 - 585161\ M\^48 - 415140\ M\^50 + 
                599400\ M\^52 + 180433\ M\^54 - 453985\ M\^56 - 
                41783\ M\^58 + 271720\ M\^60 - 6378\ M\^62 - 137110\ M\^64 + 
                12244\ M\^66 + 64757\ M\^68 - 15101\ M\^70 - 24006\ M\^72 + 
                11860\ M\^74 + 4572\ M\^76 - 5876\ M\^78 + 2229\ M\^80 - 
                396\ M\^82 + 28\ M\^84)\) + 
          L\^5\ \((56\ M\^18 - 710\ M\^20 + 3452\ M\^22 - 7132\ M\^24 + 
                541\ M\^26 + 23387\ M\^28 - 26896\ M\^30 - 35879\ M\^32 + 
                82330\ M\^34 + 24778\ M\^36 - 148841\ M\^38 + 8093\ M\^40 + 
                208639\ M\^42 - 54529\ M\^44 - 244761\ M\^46 + 
                106755\ M\^48 + 244899\ M\^50 - 143490\ M\^52 - 
                178913\ M\^54 + 163646\ M\^56 + 83433\ M\^58 - 
                141583\ M\^60 - 4155\ M\^62 + 81168\ M\^64 - 28971\ M\^66 - 
                24152\ M\^68 + 23941\ M\^70 - 3637\ M\^72 - 7063\ M\^74 + 
                5355\ M\^76 - 386\ M\^78 - 1656\ M\^80 + 1188\ M\^82 - 
                408\ M\^84 + 75\ M\^86 - 6\ M\^88)\) + 
          L\^11\ \((\(-1\) + 17\ M\^2 - 128\ M\^4 + 561\ M\^6 - 1586\ M\^8 + 
                2891\ M\^10 - 2220\ M\^12 - 5221\ M\^14 + 17477\ M\^16 - 
                5694\ M\^18 - 55416\ M\^20 + 76452\ M\^22 + 81811\ M\^24 - 
                237024\ M\^26 - 23524\ M\^28 + 465423\ M\^30 - 
                235445\ M\^32 - 608447\ M\^34 + 720990\ M\^36 + 
                491861\ M\^38 - 1286816\ M\^40 - 66437\ M\^42 + 
                1613076\ M\^44 - 524581\ M\^46 - 1480310\ M\^48 + 
                1049449\ M\^50 + 1029162\ M\^52 - 1187556\ M\^54 - 
                474763\ M\^56 + 958242\ M\^58 + 72651\ M\^60 - 
                570049\ M\^62 + 83638\ M\^64 + 247416\ M\^66 - 81630\ M\^68 - 
                71110\ M\^70 + 31806\ M\^72 + 16028\ M\^74 - 6555\ M\^76 - 
                4636\ M\^78 + 343\ M\^80 + 3463\ M\^82 - 2438\ M\^84 + 
                768\ M\^86 - 122\ M\^88 + 8\ M\^90)\) + 
          L\^6\ \((28\ M\^12 - 396\ M\^14 + 2229\ M\^16 - 5876\ M\^18 + 
                4572\ M\^20 + 11860\ M\^22 - 24006\ M\^24 - 15101\ M\^26 + 
                64757\ M\^28 + 12244\ M\^30 - 137110\ M\^32 - 6378\ M\^34 + 
                271720\ M\^36 - 41783\ M\^38 - 453985\ M\^40 + 
                180433\ M\^42 + 599400\ M\^44 - 415140\ M\^46 - 
                585161\ M\^48 + 655279\ M\^50 + 398335\ M\^52 - 
                718197\ M\^54 - 92425\ M\^56 + 585894\ M\^58 - 
                143606\ M\^60 - 347768\ M\^62 + 207112\ M\^64 + 
                125710\ M\^66 - 142676\ M\^68 - 5199\ M\^70 + 57003\ M\^72 - 
                24001\ M\^74 - 6005\ M\^76 + 11122\ M\^78 - 4446\ M\^80 - 
                1333\ M\^82 + 2447\ M\^84 - 1311\ M\^86 + 377\ M\^88 - 
                59\ M\^90 + 4\ M\^92)\) + 
          L\^9\ \((\(-3\) + 49\ M\^2 - 345\ M\^4 + 1298\ M\^6 - 2485\ M\^8 + 
                994\ M\^10 + 4227\ M\^12 - 1371\ M\^14 - 19703\ M\^16 + 
                19501\ M\^18 + 50330\ M\^20 - 83806\ M\^22 - 92134\ M\^24 + 
                265318\ M\^26 + 53028\ M\^28 - 583643\ M\^30 + 
                207978\ M\^32 + 974540\ M\^34 - 847583\ M\^36 - 
                1198605\ M\^38 + 1815346\ M\^40 + 1005105\ M\^42 - 
                2779959\ M\^44 - 375070\ M\^46 + 3222606\ M\^48 - 
                375070\ M\^50 - 2779959\ M\^52 + 1005105\ M\^54 + 
                1815346\ M\^56 - 1198605\ M\^58 - 847583\ M\^60 + 
                974540\ M\^62 + 207978\ M\^64 - 583643\ M\^66 + 
                53028\ M\^68 + 265318\ M\^70 - 92134\ M\^72 - 83806\ M\^74 + 
                50330\ M\^76 + 19501\ M\^78 - 19703\ M\^80 - 1371\ M\^82 + 
                4227\ M\^84 + 994\ M\^86 - 2485\ M\^88 + 1298\ M\^90 - 
                345\ M\^92 + 49\ M\^94 - 3\ M\^96)\) + 
          L\^7\ \((8\ M\^6 - 122\ M\^8 + 768\ M\^10 - 2438\ M\^12 + 
                3463\ M\^14 + 343\ M\^16 - 4636\ M\^18 - 6555\ M\^20 + 
                16028\ M\^22 + 31806\ M\^24 - 71110\ M\^26 - 81630\ M\^28 + 
                247416\ M\^30 + 83638\ M\^32 - 570049\ M\^34 + 72651\ M\^36 + 
                958242\ M\^38 - 474763\ M\^40 - 1187556\ M\^42 + 
                1029162\ M\^44 + 1049449\ M\^46 - 1480310\ M\^48 - 
                524581\ M\^50 + 1613076\ M\^52 - 66437\ M\^54 - 
                1286816\ M\^56 + 491861\ M\^58 + 720990\ M\^60 - 
                608447\ M\^62 - 235445\ M\^64 + 465423\ M\^66 - 
                23524\ M\^68 - 237024\ M\^70 + 81811\ M\^72 + 76452\ M\^74 - 
                55416\ M\^76 - 5694\ M\^78 + 17477\ M\^80 - 5221\ M\^82 - 
                2220\ M\^84 + 2891\ M\^86 - 1586\ M\^88 + 561\ M\^90 - 
                128\ M\^92 + 17\ M\^94 - M\^96)\) + 
          L\^10\ \((3 - 50\ M\^2 + 359\ M\^4 - 1398\ M\^6 + 2910\ M\^8 - 
                1820\ M\^10 - 5278\ M\^12 + 10305\ M\^14 + 7159\ M\^16 - 
                41492\ M\^18 + 27615\ M\^20 + 82406\ M\^22 - 154117\ M\^24 - 
                75611\ M\^26 + 447725\ M\^28 - 144327\ M\^30 - 
                859183\ M\^32 + 701848\ M\^34 + 1195792\ M\^36 - 
                1531175\ M\^38 - 1162329\ M\^40 + 2289145\ M\^42 + 
                598927\ M\^44 - 2626485\ M\^46 + 308907\ M\^48 + 
                2433260\ M\^50 - 1016520\ M\^52 - 1729167\ M\^54 + 
                1300132\ M\^56 + 894011\ M\^58 - 1144615\ M\^60 - 
                268929\ M\^62 + 756635\ M\^64 - 40346\ M\^66 - 
                373194\ M\^68 + 103537\ M\^70 + 135269\ M\^72 - 
                72547\ M\^74 - 26791\ M\^76 + 27474\ M\^78 - 136\ M\^80 - 
                6634\ M\^82 + 3454\ M\^84 - 1577\ M\^86 + 923\ M\^88 - 
                412\ M\^90 + 110\ M\^92 - 16\ M\^94 + M\^96)\) + 
          L\^8\ \((1 - 16\ M\^2 + 110\ M\^4 - 412\ M\^6 + 923\ M\^8 - 
                1577\ M\^10 + 3454\ M\^12 - 6634\ M\^14 - 136\ M\^16 + 
                27474\ M\^18 - 26791\ M\^20 - 72547\ M\^22 + 135269\ M\^24 + 
                103537\ M\^26 - 373194\ M\^28 - 40346\ M\^30 + 
                756635\ M\^32 - 268929\ M\^34 - 1144615\ M\^36 + 
                894011\ M\^38 + 1300132\ M\^40 - 1729167\ M\^42 - 
                1016520\ M\^44 + 2433260\ M\^46 + 308907\ M\^48 - 
                2626485\ M\^50 + 598927\ M\^52 + 2289145\ M\^54 - 
                1162329\ M\^56 - 1531175\ M\^58 + 1195792\ M\^60 + 
                701848\ M\^62 - 859183\ M\^64 - 144327\ M\^66 + 
                447725\ M\^68 - 75611\ M\^70 - 154117\ M\^72 + 82406\ M\^74 + 
                27615\ M\^76 - 41492\ M\^78 + 7159\ M\^80 + 10305\ M\^82 - 
                5278\ M\^84 - 1820\ M\^86 + 2910\ M\^88 - 1398\ M\^90 + 
                359\ M\^92 - 50\ M\^94 + 3\ M\^96)\)}, {"A-polynomial", 
        Knot[9, 15], 
        M\^16 + L\^19\ M\^94 + 
          L\ \((\(-4\)\ M\^12 + 16\ M\^14 - 24\ M\^16 + M\^18 + 49\ M\^20 - 
                19\ M\^22 - 10\ M\^24 + 18\ M\^26 - 11\ M\^28 + 3\ M\^30)\) + 
          L\^2\ \((6\ M\^8 - 48\ M\^10 + 150\ M\^12 - 174\ M\^14 - 
                116\ M\^16 + 436\ M\^18 - 194\ M\^20 - 326\ M\^22 + 
                358\ M\^24 + 213\ M\^26 + 54\ M\^28 - 202\ M\^30 - 
                185\ M\^32 + 247\ M\^34 + 22\ M\^36 - 124\ M\^38 + 
                73\ M\^40 - 22\ M\^42 + 3\ M\^44)\) + 
          L\^3\ \((\(-4\)\ M\^4 + 48\ M\^6 - 232\ M\^8 + 532\ M\^10 - 
                479\ M\^12 - 271\ M\^14 + 1095\ M\^16 - 1273\ M\^18 + 
                265\ M\^20 + 2155\ M\^22 - 2950\ M\^24 - 938\ M\^26 + 
                3478\ M\^28 + 1046\ M\^30 - 302\ M\^32 - 1863\ M\^34 - 
                1104\ M\^36 + 2438\ M\^38 + 79\ M\^40 - 1239\ M\^42 + 
                545\ M\^44 + 16\ M\^46 - 215\ M\^48 + 260\ M\^50 - 
                165\ M\^52 + 57\ M\^54 - 11\ M\^56 + M\^58)\) + 
          L\^4\ \((1 - 16\ M\^2 + 110\ M\^4 - 426\ M\^6 + 1045\ M\^8 - 
                1623\ M\^10 + 899\ M\^12 + 2659\ M\^14 - 6410\ M\^16 + 
                2448\ M\^18 + 8525\ M\^20 - 12719\ M\^22 + 3901\ M\^24 + 
                14812\ M\^26 - 27000\ M\^28 - 4399\ M\^30 + 44351\ M\^32 - 
                5568\ M\^34 - 37087\ M\^36 + 12560\ M\^38 + 20756\ M\^40 - 
                11979\ M\^42 - 7628\ M\^44 + 8270\ M\^46 + 1538\ M\^48 - 
                5438\ M\^50 + 1510\ M\^52 + 2279\ M\^54 - 1728\ M\^56 - 
                298\ M\^58 + 933\ M\^60 - 536\ M\^62 + 158\ M\^64 - 
                26\ M\^66 + 2\ M\^68)\) + 
          L\^5\ \((\(-4\) + 67\ M\^2 - 460\ M\^4 + 1630\ M\^6 - 2948\ M\^8 + 
                1630\ M\^10 + 2937\ M\^12 - 4806\ M\^14 + 1885\ M\^16 - 
                1124\ M\^18 - 601\ M\^20 + 12816\ M\^22 - 31851\ M\^24 + 
                21312\ M\^26 + 80944\ M\^28 - 138537\ M\^30 - 97834\ M\^32 + 
                281746\ M\^34 + 48908\ M\^36 - 320860\ M\^38 + 46889\ M\^40 + 
                244854\ M\^42 - 92414\ M\^44 - 120292\ M\^46 + 70841\ M\^48 + 
                37880\ M\^50 - 31724\ M\^52 - 10277\ M\^54 + 10969\ M\^56 + 
                6593\ M\^58 - 8044\ M\^60 - 1602\ M\^62 + 4880\ M\^64 - 
                1107\ M\^66 - 1737\ M\^68 + 1567\ M\^70 - 619\ M\^72 + 
                137\ M\^74 - 17\ M\^76 + M\^78)\) + 
          L\^6\ \((6 - 105\ M\^2 + 759\ M\^4 - 2862\ M\^6 + 5558\ M\^8 - 
                3117\ M\^10 - 7965\ M\^12 + 13930\ M\^14 + 1443\ M\^16 - 
                20688\ M\^18 + 14011\ M\^20 + 15188\ M\^22 - 17803\ M\^24 - 
                70527\ M\^26 + 65038\ M\^28 + 285670\ M\^30 - 268331\ M\^32 - 
                655961\ M\^34 + 670844\ M\^36 + 943008\ M\^38 - 
                1110681\ M\^40 - 860704\ M\^42 + 1367164\ M\^44 + 
                471001\ M\^46 - 1221325\ M\^48 - 20819\ M\^50 + 
                784591\ M\^52 - 224444\ M\^54 - 328714\ M\^56 + 
                215932\ M\^58 + 61328\ M\^60 - 103392\ M\^62 + 17680\ M\^64 + 
                19558\ M\^66 - 8028\ M\^68 + 608\ M\^70 - 1692\ M\^72 + 
                110\ M\^74 + 2261\ M\^76 - 2145\ M\^78 + 955\ M\^80 - 
                239\ M\^82 + 33\ M\^84 - 2\ M\^86)\) + 
          L\^7\ \((\(-4\) + 73\ M\^2 - 562\ M\^4 + 2367\ M\^6 - 5848\ M\^8 + 
                7970\ M\^10 - 2017\ M\^12 - 17332\ M\^14 + 40700\ M\^16 - 
                12616\ M\^18 - 122212\ M\^20 + 175296\ M\^22 + 
                182132\ M\^24 - 496593\ M\^26 - 189319\ M\^28 + 
                911488\ M\^30 + 206815\ M\^32 - 1244347\ M\^34 - 
                403011\ M\^36 + 1393253\ M\^38 + 794751\ M\^40 - 
                1425545\ M\^42 - 1192496\ M\^44 + 1506240\ M\^46 + 
                1411865\ M\^48 - 1500290\ M\^50 - 1219068\ M\^52 + 
                1365241\ M\^54 + 746676\ M\^56 - 1061716\ M\^58 - 
                272398\ M\^60 + 677249\ M\^62 - 23704\ M\^64 - 
                327603\ M\^66 + 110419\ M\^68 + 104991\ M\^70 - 
                84447\ M\^72 - 3160\ M\^74 + 28159\ M\^76 - 11436\ M\^78 - 
                2362\ M\^80 + 5067\ M\^82 - 3486\ M\^84 + 1662\ M\^86 - 
                566\ M\^88 + 128\ M\^90 - 17\ M\^92 + M\^94)\) + 
          L\^18\ \((3\ M\^80 - 11\ M\^82 + 18\ M\^84 - 10\ M\^86 - 
                19\ M\^88 + 49\ M\^90 + M\^92 - 24\ M\^94 + 16\ M\^96 - 
                4\ M\^98)\) + 
          L\^8\ \((1 - 19\ M\^2 + 161\ M\^4 - 833\ M\^6 + 3033\ M\^8 - 
                8004\ M\^10 + 12885\ M\^12 - 658\ M\^14 - 46726\ M\^16 + 
                69313\ M\^18 + 76337\ M\^20 - 284974\ M\^22 + 28851\ M\^24 + 
                650382\ M\^26 - 403196\ M\^28 - 1104621\ M\^30 + 
                1156770\ M\^32 + 1490127\ M\^34 - 2230105\ M\^36 - 
                1713132\ M\^38 + 3474559\ M\^40 + 1727799\ M\^42 - 
                4620101\ M\^44 - 1581101\ M\^46 + 5269679\ M\^48 + 
                1320176\ M\^50 - 4935663\ M\^52 - 785504\ M\^54 + 
                3837308\ M\^56 + 226766\ M\^58 - 2411054\ M\^60 + 
                122142\ M\^62 + 1195454\ M\^64 - 208830\ M\^66 - 
                446375\ M\^68 + 128962\ M\^70 + 128716\ M\^72 - 
                42127\ M\^74 - 35231\ M\^76 - 2785\ M\^78 + 26486\ M\^80 - 
                1651\ M\^82 - 13751\ M\^84 + 4507\ M\^86 + 5147\ M\^88 - 
                5376\ M\^90 + 2325\ M\^92 - 555\ M\^94 + 72\ M\^96 - 
                4\ M\^98)\) + 
          L\^9\ \((\(-4\)\ M\^4 + 76\ M\^6 - 620\ M\^8 + 2800\ M\^10 - 
                7356\ M\^12 + 9812\ M\^14 + 945\ M\^16 - 26531\ M\^18 + 
                36738\ M\^20 + 19296\ M\^22 - 141410\ M\^24 + 146034\ M\^26 + 
                241883\ M\^28 - 610424\ M\^30 - 225815\ M\^32 + 
                1502992\ M\^34 - 19376\ M\^36 - 2724634\ M\^38 + 
                298882\ M\^40 + 4020560\ M\^42 - 185629\ M\^44 - 
                5050131\ M\^46 - 621061\ M\^48 + 5586790\ M\^50 + 
                1741819\ M\^52 - 5528536\ M\^54 - 2252516\ M\^56 + 
                5133634\ M\^58 + 1927885\ M\^60 - 4267939\ M\^62 - 
                1016481\ M\^64 + 3049877\ M\^66 + 120430\ M\^68 - 
                1790635\ M\^70 + 334506\ M\^72 + 833624\ M\^74 - 
                385188\ M\^76 - 278516\ M\^78 + 245762\ M\^80 + 
                54572\ M\^82 - 119111\ M\^84 + 15969\ M\^86 + 36338\ M\^88 - 
                16026\ M\^90 - 6679\ M\^92 + 8816\ M\^94 - 3830\ M\^96 + 
                892\ M\^98 - 112\ M\^100 + 6\ M\^102)\) + 
          L\^17\ \((3\ M\^66 - 22\ M\^68 + 73\ M\^70 - 124\ M\^72 + 
                22\ M\^74 + 247\ M\^76 - 185\ M\^78 - 202\ M\^80 + 
                54\ M\^82 + 213\ M\^84 + 358\ M\^86 - 326\ M\^88 - 
                194\ M\^90 + 436\ M\^92 - 116\ M\^94 - 174\ M\^96 + 
                150\ M\^98 - 48\ M\^100 + 6\ M\^102)\) + 
          L\^16\ \((M\^52 - 11\ M\^54 + 57\ M\^56 - 165\ M\^58 + 260\ M\^60 - 
                215\ M\^62 + 16\ M\^64 + 545\ M\^66 - 1239\ M\^68 + 
                79\ M\^70 + 2438\ M\^72 - 1104\ M\^74 - 1863\ M\^76 - 
                302\ M\^78 + 1046\ M\^80 + 3478\ M\^82 - 938\ M\^84 - 
                2950\ M\^86 + 2155\ M\^88 + 265\ M\^90 - 1273\ M\^92 + 
                1095\ M\^94 - 271\ M\^96 - 479\ M\^98 + 532\ M\^100 - 
                232\ M\^102 + 48\ M\^104 - 4\ M\^106)\) + 
          L\^10\ \((6\ M\^8 - 112\ M\^10 + 892\ M\^12 - 3830\ M\^14 + 
                8816\ M\^16 - 6679\ M\^18 - 16026\ M\^20 + 36338\ M\^22 + 
                15969\ M\^24 - 119111\ M\^26 + 54572\ M\^28 + 245762\ M\^30 - 
                278516\ M\^32 - 385188\ M\^34 + 833624\ M\^36 + 
                334506\ M\^38 - 1790635\ M\^40 + 120430\ M\^42 + 
                3049877\ M\^44 - 1016481\ M\^46 - 4267939\ M\^48 + 
                1927885\ M\^50 + 5133634\ M\^52 - 2252516\ M\^54 - 
                5528536\ M\^56 + 1741819\ M\^58 + 5586790\ M\^60 - 
                621061\ M\^62 - 5050131\ M\^64 - 185629\ M\^66 + 
                4020560\ M\^68 + 298882\ M\^70 - 2724634\ M\^72 - 
                19376\ M\^74 + 1502992\ M\^76 - 225815\ M\^78 - 
                610424\ M\^80 + 241883\ M\^82 + 146034\ M\^84 - 
                141410\ M\^86 + 19296\ M\^88 + 36738\ M\^90 - 26531\ M\^92 + 
                945\ M\^94 + 9812\ M\^96 - 7356\ M\^98 + 2800\ M\^100 - 
                620\ M\^102 + 76\ M\^104 - 4\ M\^106)\) + 
          L\^14\ \((M\^32 - 17\ M\^34 + 137\ M\^36 - 619\ M\^38 + 
                1567\ M\^40 - 1737\ M\^42 - 1107\ M\^44 + 4880\ M\^46 - 
                1602\ M\^48 - 8044\ M\^50 + 6593\ M\^52 + 10969\ M\^54 - 
                10277\ M\^56 - 31724\ M\^58 + 37880\ M\^60 + 70841\ M\^62 - 
                120292\ M\^64 - 92414\ M\^66 + 244854\ M\^68 + 46889\ M\^70 - 
                320860\ M\^72 + 48908\ M\^74 + 281746\ M\^76 - 97834\ M\^78 - 
                138537\ M\^80 + 80944\ M\^82 + 21312\ M\^84 - 31851\ M\^86 + 
                12816\ M\^88 - 601\ M\^90 - 1124\ M\^92 + 1885\ M\^94 - 
                4806\ M\^96 + 2937\ M\^98 + 1630\ M\^100 - 2948\ M\^102 + 
                1630\ M\^104 - 460\ M\^106 + 67\ M\^108 - 4\ M\^110)\) + 
          L\^12\ \((M\^16 - 17\ M\^18 + 128\ M\^20 - 566\ M\^22 + 1662\ M\^24 
                - 3486\ M\^26 + 5067\ M\^28 - 2362\ M\^30 - 11436\ M\^32 + 
                28159\ M\^34 - 3160\ M\^36 - 84447\ M\^38 + 104991\ M\^40 + 
                110419\ M\^42 - 327603\ M\^44 - 23704\ M\^46 + 
                677249\ M\^48 - 272398\ M\^50 - 1061716\ M\^52 + 
                746676\ M\^54 + 1365241\ M\^56 - 1219068\ M\^58 - 
                1500290\ M\^60 + 1411865\ M\^62 + 1506240\ M\^64 - 
                1192496\ M\^66 - 1425545\ M\^68 + 794751\ M\^70 + 
                1393253\ M\^72 - 403011\ M\^74 - 1244347\ M\^76 + 
                206815\ M\^78 + 911488\ M\^80 - 189319\ M\^82 - 
                496593\ M\^84 + 182132\ M\^86 + 175296\ M\^88 - 
                122212\ M\^90 - 12616\ M\^92 + 40700\ M\^94 - 17332\ M\^96 - 
                2017\ M\^98 + 7970\ M\^100 - 5848\ M\^102 + 2367\ M\^104 - 
                562\ M\^106 + 73\ M\^108 - 4\ M\^110)\) + 
          L\^11\ \((\(-4\)\ M\^12 + 72\ M\^14 - 555\ M\^16 + 2325\ M\^18 - 
                5376\ M\^20 + 5147\ M\^22 + 4507\ M\^24 - 13751\ M\^26 - 
                1651\ M\^28 + 26486\ M\^30 - 2785\ M\^32 - 35231\ M\^34 - 
                42127\ M\^36 + 128716\ M\^38 + 128962\ M\^40 - 
                446375\ M\^42 - 208830\ M\^44 + 1195454\ M\^46 + 
                122142\ M\^48 - 2411054\ M\^50 + 226766\ M\^52 + 
                3837308\ M\^54 - 785504\ M\^56 - 4935663\ M\^58 + 
                1320176\ M\^60 + 5269679\ M\^62 - 1581101\ M\^64 - 
                4620101\ M\^66 + 1727799\ M\^68 + 3474559\ M\^70 - 
                1713132\ M\^72 - 2230105\ M\^74 + 1490127\ M\^76 + 
                1156770\ M\^78 - 1104621\ M\^80 - 403196\ M\^82 + 
                650382\ M\^84 + 28851\ M\^86 - 284974\ M\^88 + 76337\ M\^90 + 
                69313\ M\^92 - 46726\ M\^94 - 658\ M\^96 + 12885\ M\^98 - 
                8004\ M\^100 + 3033\ M\^102 - 833\ M\^104 + 161\ M\^106 - 
                19\ M\^108 + M\^110)\) + 
          L\^15\ \((2\ M\^42 - 26\ M\^44 + 158\ M\^46 - 536\ M\^48 + 
                933\ M\^50 - 298\ M\^52 - 1728\ M\^54 + 2279\ M\^56 + 
                1510\ M\^58 - 5438\ M\^60 + 1538\ M\^62 + 8270\ M\^64 - 
                7628\ M\^66 - 11979\ M\^68 + 20756\ M\^70 + 12560\ M\^72 - 
                37087\ M\^74 - 5568\ M\^76 + 44351\ M\^78 - 4399\ M\^80 - 
                27000\ M\^82 + 14812\ M\^84 + 3901\ M\^86 - 12719\ M\^88 + 
                8525\ M\^90 + 2448\ M\^92 - 6410\ M\^94 + 2659\ M\^96 + 
                899\ M\^98 - 1623\ M\^100 + 1045\ M\^102 - 426\ M\^104 + 
                110\ M\^106 - 16\ M\^108 + M\^110)\) + 
          L\^13\ \((\(-2\)\ M\^24 + 33\ M\^26 - 239\ M\^28 + 955\ M\^30 - 
                2145\ M\^32 + 2261\ M\^34 + 110\ M\^36 - 1692\ M\^38 + 
                608\ M\^40 - 8028\ M\^42 + 19558\ M\^44 + 17680\ M\^46 - 
                103392\ M\^48 + 61328\ M\^50 + 215932\ M\^52 - 
                328714\ M\^54 - 224444\ M\^56 + 784591\ M\^58 - 
                20819\ M\^60 - 1221325\ M\^62 + 471001\ M\^64 + 
                1367164\ M\^66 - 860704\ M\^68 - 1110681\ M\^70 + 
                943008\ M\^72 + 670844\ M\^74 - 655961\ M\^76 - 
                268331\ M\^78 + 285670\ M\^80 + 65038\ M\^82 - 70527\ M\^84 - 
                17803\ M\^86 + 15188\ M\^88 + 14011\ M\^90 - 20688\ M\^92 + 
                1443\ M\^94 + 13930\ M\^96 - 7965\ M\^98 - 3117\ M\^100 + 
                5558\ M\^102 - 2862\ M\^104 + 759\ M\^106 - 105\ M\^108 + 
                6\ M\^110)\)}, {"A-polynomial", Knot[9, 16], 
        1 + L\^14\ M\^176 + 
          L\ \((\(-1\) + 6\ M\^2 - 18\ M\^4 + 26\ M\^6 - M\^8 - 40\ M\^10 + 
                42\ M\^12 + 14\ M\^14 - 20\ M\^16 + 6\ M\^18)\) + 
          L\^2\ \((M\^4 - 9\ M\^6 + 42\ M\^8 - 112\ M\^10 + 155\ M\^12 + 
                2\ M\^14 - 374\ M\^16 + 463\ M\^18 + 178\ M\^20 - 
                810\ M\^22 + 403\ M\^24 + 571\ M\^26 - 551\ M\^28 + 
                17\ M\^30 + 198\ M\^32 - 98\ M\^34 + 15\ M\^36)\) + 
          L\^3\ \((2\ M\^10 - 20\ M\^12 + 96\ M\^14 - 259\ M\^16 + 
                350\ M\^18 + 55\ M\^20 - 968\ M\^22 + 1087\ M\^24 + 
                849\ M\^26 - 2957\ M\^28 + 1591\ M\^30 + 2837\ M\^32 - 
                4778\ M\^34 + 618\ M\^36 + 4994\ M\^38 - 3471\ M\^40 - 
                1669\ M\^42 + 3153\ M\^44 - 889\ M\^46 - 722\ M\^48 + 
                634\ M\^50 - 189\ M\^52 + 20\ M\^54)\) + 
          L\^4\ \((M\^16 - 11\ M\^18 + 57\ M\^20 - 174\ M\^22 + 318\ M\^24 - 
                288\ M\^26 - 80\ M\^28 + 587\ M\^30 - 669\ M\^32 - 
                444\ M\^34 + 2801\ M\^36 - 2877\ M\^38 - 3733\ M\^40 + 
                9805\ M\^42 + 367\ M\^44 - 17167\ M\^46 + 7780\ M\^48 + 
                19272\ M\^50 - 15463\ M\^52 - 10506\ M\^54 + 16051\ M\^56 - 
                136\ M\^58 - 8653\ M\^60 + 4421\ M\^62 + 920\ M\^64 - 
                1862\ M\^66 + 849\ M\^68 - 180\ M\^70 + 15\ M\^72)\) + 
          L\^5\ \((M\^26 - 13\ M\^28 + 78\ M\^30 - 276\ M\^32 + 562\ M\^34 - 
                436\ M\^36 - 653\ M\^38 + 1530\ M\^40 + 641\ M\^42 - 
                4897\ M\^44 + 3472\ M\^46 + 7134\ M\^48 - 12889\ M\^50 - 
                3220\ M\^52 + 25499\ M\^54 - 10682\ M\^56 - 33620\ M\^58 + 
                29933\ M\^60 + 29681\ M\^62 - 40160\ M\^64 - 13985\ M\^66 + 
                39786\ M\^68 - 3036\ M\^70 - 26309\ M\^72 + 11502\ M\^74 + 
                9867\ M\^76 - 9623\ M\^78 + 615\ M\^80 + 2838\ M\^82 - 
                1796\ M\^84 + 538\ M\^86 - 86\ M\^88 + 6\ M\^90)\) + 
          L\^6\ \((\(-M\^38\) + 17\ M\^40 - 115\ M\^42 + 393\ M\^44 - 
                644\ M\^46 + 174\ M\^48 + 660\ M\^50 + 756\ M\^52 - 
                3647\ M\^54 - 851\ M\^56 + 11668\ M\^58 - 2744\ M\^60 - 
                27728\ M\^62 + 17827\ M\^64 + 46512\ M\^66 - 46271\ M\^68 - 
                54800\ M\^70 + 74813\ M\^72 + 45013\ M\^74 - 85231\ M\^76 - 
                16437\ M\^78 + 72605\ M\^80 - 5826\ M\^82 - 46103\ M\^84 + 
                15994\ M\^86 + 20266\ M\^88 - 14322\ M\^90 - 4455\ M\^92 + 
                8199\ M\^94 - 1895\ M\^96 - 2180\ M\^98 + 1962\ M\^100 - 
                744\ M\^102 + 155\ M\^104 - 18\ M\^106 + M\^108)\) + 
          L\^7\ \((\(-M\^52\) + 16\ M\^54 - 111\ M\^56 + 444\ M\^58 - 
                1103\ M\^60 + 1447\ M\^62 + 396\ M\^64 - 4236\ M\^66 + 
                1924\ M\^68 + 11019\ M\^70 - 10445\ M\^72 - 22576\ M\^74 + 
                31447\ M\^76 + 34137\ M\^78 - 65796\ M\^80 - 34800\ M\^82 + 
                101889\ M\^84 + 15341\ M\^86 - 114552\ M\^88 + 15341\ M\^90 + 
                101889\ M\^92 - 34800\ M\^94 - 65796\ M\^96 + 34137\ M\^98 + 
                31447\ M\^100 - 22576\ M\^102 - 10445\ M\^104 + 
                11019\ M\^106 + 1924\ M\^108 - 4236\ M\^110 + 396\ M\^112 + 
                1447\ M\^114 - 1103\ M\^116 + 444\ M\^118 - 111\ M\^120 + 
                16\ M\^122 - M\^124)\) + 
          L\^8\ \((M\^68 - 18\ M\^70 + 155\ M\^72 - 744\ M\^74 + 
                1962\ M\^76 - 2180\ M\^78 - 1895\ M\^80 + 8199\ M\^82 - 
                4455\ M\^84 - 14322\ M\^86 + 20266\ M\^88 + 15994\ M\^90 - 
                46103\ M\^92 - 5826\ M\^94 + 72605\ M\^96 - 16437\ M\^98 - 
                85231\ M\^100 + 45013\ M\^102 + 74813\ M\^104 - 
                54800\ M\^106 - 46271\ M\^108 + 46512\ M\^110 + 
                17827\ M\^112 - 27728\ M\^114 - 2744\ M\^116 + 
                11668\ M\^118 - 851\ M\^120 - 3647\ M\^122 + 756\ M\^124 + 
                660\ M\^126 + 174\ M\^128 - 644\ M\^130 + 393\ M\^132 - 
                115\ M\^134 + 17\ M\^136 - M\^138)\) + 
          L\^9\ \((6\ M\^86 - 86\ M\^88 + 538\ M\^90 - 1796\ M\^92 + 
                2838\ M\^94 + 615\ M\^96 - 9623\ M\^98 + 9867\ M\^100 + 
                11502\ M\^102 - 26309\ M\^104 - 3036\ M\^106 + 
                39786\ M\^108 - 13985\ M\^110 - 40160\ M\^112 + 
                29681\ M\^114 + 29933\ M\^116 - 33620\ M\^118 - 
                10682\ M\^120 + 25499\ M\^122 - 3220\ M\^124 - 
                12889\ M\^126 + 7134\ M\^128 + 3472\ M\^130 - 4897\ M\^132 + 
                641\ M\^134 + 1530\ M\^136 - 653\ M\^138 - 436\ M\^140 + 
                562\ M\^142 - 276\ M\^144 + 78\ M\^146 - 13\ M\^148 + 
                M\^150)\) + 
          L\^10\ \((15\ M\^104 - 180\ M\^106 + 849\ M\^108 - 1862\ M\^110 + 
                920\ M\^112 + 4421\ M\^114 - 8653\ M\^116 - 136\ M\^118 + 
                16051\ M\^120 - 10506\ M\^122 - 15463\ M\^124 + 
                19272\ M\^126 + 7780\ M\^128 - 17167\ M\^130 + 367\ M\^132 + 
                9805\ M\^134 - 3733\ M\^136 - 2877\ M\^138 + 2801\ M\^140 - 
                444\ M\^142 - 669\ M\^144 + 587\ M\^146 - 80\ M\^148 - 
                288\ M\^150 + 318\ M\^152 - 174\ M\^154 + 57\ M\^156 - 
                11\ M\^158 + M\^160)\) + 
          L\^11\ \((20\ M\^122 - 189\ M\^124 + 634\ M\^126 - 722\ M\^128 - 
                889\ M\^130 + 3153\ M\^132 - 1669\ M\^134 - 3471\ M\^136 + 
                4994\ M\^138 + 618\ M\^140 - 4778\ M\^142 + 2837\ M\^144 + 
                1591\ M\^146 - 2957\ M\^148 + 849\ M\^150 + 1087\ M\^152 - 
                968\ M\^154 + 55\ M\^156 + 350\ M\^158 - 259\ M\^160 + 
                96\ M\^162 - 20\ M\^164 + 2\ M\^166)\) + 
          L\^12\ \((15\ M\^140 - 98\ M\^142 + 198\ M\^144 + 17\ M\^146 - 
                551\ M\^148 + 571\ M\^150 + 403\ M\^152 - 810\ M\^154 + 
                178\ M\^156 + 463\ M\^158 - 374\ M\^160 + 2\ M\^162 + 
                155\ M\^164 - 112\ M\^166 + 42\ M\^168 - 9\ M\^170 + 
                M\^172)\) + 
          L\^13\ \((6\ M\^158 - 20\ M\^160 + 14\ M\^162 + 42\ M\^164 - 
                40\ M\^166 - M\^168 + 26\ M\^170 - 18\ M\^172 + 6\ M\^174 - 
                M\^176)\)}, {"A-polynomial", Knot[9, 17], 
        L\^13\ M\^22 + M\^72 + 
          L\^12\ \((\(-2\)\ M\^14 + 8\ M\^16 - 13\ M\^18 + 8\ M\^20 + 
                7\ M\^22 - 27\ M\^24 + 33\ M\^26 + 18\ M\^28 - 25\ M\^30 + 
                6\ M\^32)\) + 
          L\^11\ \((M\^6 - 8\ M\^8 + 30\ M\^10 - 58\ M\^12 + 16\ M\^14 + 
                144\ M\^16 - 147\ M\^18 - 227\ M\^20 + 305\ M\^22 + 
                342\ M\^24 - 514\ M\^26 - 437\ M\^28 + 792\ M\^30 + 
                327\ M\^32 - 664\ M\^34 + 30\ M\^36 + 244\ M\^38 - 
                113\ M\^40 + 15\ M\^42)\) + 
          L\^10\ \((\(-M\^2\) + 11\ M\^4 - 55\ M\^6 + 149\ M\^8 - 
                155\ M\^10 - 275\ M\^12 + 938\ M\^14 - 346\ M\^16 - 
                1913\ M\^18 + 2445\ M\^20 + 1253\ M\^22 - 4605\ M\^24 + 
                1761\ M\^26 + 4592\ M\^28 - 5465\ M\^30 - 2458\ M\^32 + 
                7405\ M\^34 + 347\ M\^36 - 6029\ M\^38 + 1306\ M\^40 + 
                3021\ M\^42 - 1580\ M\^44 - 558\ M\^46 + 680\ M\^48 - 
                202\ M\^50 + 20\ M\^52)\) + 
          L\^9\ \((\(-1\) + 12\ M\^2 - 60\ M\^4 + 145\ M\^6 - 102\ M\^8 - 
                274\ M\^10 + 489\ M\^12 + 428\ M\^14 - 1373\ M\^16 - 
                333\ M\^18 + 2531\ M\^20 + 232\ M\^22 - 4092\ M\^24 - 
                448\ M\^26 + 6236\ M\^28 + 2846\ M\^30 - 10377\ M\^32 - 
                8168\ M\^34 + 15191\ M\^36 + 12940\ M\^38 - 15920\ M\^40 - 
                11550\ M\^42 + 13484\ M\^44 + 6607\ M\^46 - 9372\ M\^48 - 
                1673\ M\^50 + 4775\ M\^52 - 836\ M\^54 - 1231\ M\^56 + 
                772\ M\^58 - 178\ M\^60 + 15\ M\^62)\) + 
          L\^8\ \((3\ M\^4 - 39\ M\^6 + 209\ M\^8 - 526\ M\^10 + 291\ M\^12 + 
                1550\ M\^14 - 2726\ M\^16 - 2481\ M\^18 + 8794\ M\^20 + 
                1899\ M\^22 - 17616\ M\^24 - 1858\ M\^26 + 26521\ M\^28 + 
                6867\ M\^30 - 30272\ M\^32 - 18647\ M\^34 + 25691\ M\^36 + 
                28484\ M\^38 - 16256\ M\^40 - 27751\ M\^42 + 9657\ M\^44 + 
                21514\ M\^46 - 4119\ M\^48 - 12746\ M\^50 - 62\ M\^52 + 
                5475\ M\^54 + 2306\ M\^56 - 2555\ M\^58 - 1937\ M\^60 + 
                1775\ M\^62 + 407\ M\^64 - 884\ M\^66 + 390\ M\^68 - 
                77\ M\^70 + 6\ M\^72)\) + 
          L\ \((6\ M\^62 - 25\ M\^64 + 18\ M\^66 + 33\ M\^68 - 27\ M\^70 + 
                7\ M\^72 + 8\ M\^74 - 13\ M\^76 + 8\ M\^78 - 2\ M\^80)\) + 
          L\^7\ \((\(-3\)\ M\^8 + 40\ M\^10 - 220\ M\^12 + 583\ M\^14 - 
                485\ M\^16 - 1081\ M\^18 + 2233\ M\^20 + 1321\ M\^22 - 
                4988\ M\^24 - 1806\ M\^26 + 7566\ M\^28 + 4896\ M\^30 - 
                9945\ M\^32 - 10065\ M\^34 + 10400\ M\^36 + 16503\ M\^38 - 
                6649\ M\^40 - 24733\ M\^42 - 5571\ M\^44 + 32309\ M\^46 + 
                24991\ M\^48 - 29883\ M\^50 - 33709\ M\^52 + 22721\ M\^54 + 
                27609\ M\^56 - 16039\ M\^58 - 14954\ M\^60 + 10212\ M\^62 + 
                4142\ M\^64 - 4398\ M\^66 + 282\ M\^68 + 707\ M\^70 - 
                475\ M\^72 + 356\ M\^74 - 211\ M\^76 + 72\ M\^78 - 
                13\ M\^80 + M\^82)\) + 
          L\^6\ \((M\^12 - 13\ M\^14 + 72\ M\^16 - 211\ M\^18 + 356\ M\^20 - 
                475\ M\^22 + 707\ M\^24 + 282\ M\^26 - 4398\ M\^28 + 
                4142\ M\^30 + 10212\ M\^32 - 14954\ M\^34 - 16039\ M\^36 + 
                27609\ M\^38 + 22721\ M\^40 - 33709\ M\^42 - 29883\ M\^44 + 
                24991\ M\^46 + 32309\ M\^48 - 5571\ M\^50 - 24733\ M\^52 - 
                6649\ M\^54 + 16503\ M\^56 + 10400\ M\^58 - 10065\ M\^60 - 
                9945\ M\^62 + 4896\ M\^64 + 7566\ M\^66 - 1806\ M\^68 - 
                4988\ M\^70 + 1321\ M\^72 + 2233\ M\^74 - 1081\ M\^76 - 
                485\ M\^78 + 583\ M\^80 - 220\ M\^82 + 40\ M\^84 - 
                3\ M\^86)\) + 
          L\^2\ \((15\ M\^52 - 113\ M\^54 + 244\ M\^56 + 30\ M\^58 - 
                664\ M\^60 + 327\ M\^62 + 792\ M\^64 - 437\ M\^66 - 
                514\ M\^68 + 342\ M\^70 + 305\ M\^72 - 227\ M\^74 - 
                147\ M\^76 + 144\ M\^78 + 16\ M\^80 - 58\ M\^82 + 30\ M\^84 - 
                8\ M\^86 + M\^88)\) + 
          L\^5\ \((6\ M\^22 - 77\ M\^24 + 390\ M\^26 - 884\ M\^28 + 
                407\ M\^30 + 1775\ M\^32 - 1937\ M\^34 - 2555\ M\^36 + 
                2306\ M\^38 + 5475\ M\^40 - 62\ M\^42 - 12746\ M\^44 - 
                4119\ M\^46 + 21514\ M\^48 + 9657\ M\^50 - 27751\ M\^52 - 
                16256\ M\^54 + 28484\ M\^56 + 25691\ M\^58 - 18647\ M\^60 - 
                30272\ M\^62 + 6867\ M\^64 + 26521\ M\^66 - 1858\ M\^68 - 
                17616\ M\^70 + 1899\ M\^72 + 8794\ M\^74 - 2481\ M\^76 - 
                2726\ M\^78 + 1550\ M\^80 + 291\ M\^82 - 526\ M\^84 + 
                209\ M\^86 - 39\ M\^88 + 3\ M\^90)\) + 
          L\^3\ \((20\ M\^42 - 202\ M\^44 + 680\ M\^46 - 558\ M\^48 - 
                1580\ M\^50 + 3021\ M\^52 + 1306\ M\^54 - 6029\ M\^56 + 
                347\ M\^58 + 7405\ M\^60 - 2458\ M\^62 - 5465\ M\^64 + 
                4592\ M\^66 + 1761\ M\^68 - 4605\ M\^70 + 1253\ M\^72 + 
                2445\ M\^74 - 1913\ M\^76 - 346\ M\^78 + 938\ M\^80 - 
                275\ M\^82 - 155\ M\^84 + 149\ M\^86 - 55\ M\^88 + 
                11\ M\^90 - M\^92)\) + 
          L\^4\ \((15\ M\^32 - 178\ M\^34 + 772\ M\^36 - 1231\ M\^38 - 
                836\ M\^40 + 4775\ M\^42 - 1673\ M\^44 - 9372\ M\^46 + 
                6607\ M\^48 + 13484\ M\^50 - 11550\ M\^52 - 15920\ M\^54 + 
                12940\ M\^56 + 15191\ M\^58 - 8168\ M\^60 - 10377\ M\^62 + 
                2846\ M\^64 + 6236\ M\^66 - 448\ M\^68 - 4092\ M\^70 + 
                232\ M\^72 + 2531\ M\^74 - 333\ M\^76 - 1373\ M\^78 + 
                428\ M\^80 + 489\ M\^82 - 274\ M\^84 - 102\ M\^86 + 
                145\ M\^88 - 60\ M\^90 + 12\ M\^92 - 
                M\^94)\)}, {"A-polynomial", Knot[9, 18], 
        L\^20 + M\^164 + 
          L\^19\ \((\(-4\) + 18\ M\^2 - 28\ M\^4 + 57\ M\^8 - 22\ M\^10 - 
                14\ M\^12 + 21\ M\^14 - 11\ M\^16 + 3\ M\^18)\) + 
          L\^18\ \((6 - 54\ M\^2 + 189\ M\^4 - 256\ M\^6 - 110\ M\^8 + 
                681\ M\^10 - 498\ M\^12 - 497\ M\^14 + 963\ M\^16 + 
                34\ M\^18 - 409\ M\^20 + 148\ M\^22 - 78\ M\^24 + 77\ M\^26 + 
                54\ M\^28 - 111\ M\^30 + 70\ M\^32 - 22\ M\^34 + 
                3\ M\^36)\) + 
          L\^17\ \((\(-4\) + 54\ M\^2 - 299\ M\^4 + 820\ M\^6 - 963\ M\^8 - 
                483\ M\^10 + 2899\ M\^12 - 2810\ M\^14 - 1432\ M\^16 + 
                5442\ M\^18 - 4181\ M\^20 - 1745\ M\^22 + 6675\ M\^24 - 
                3369\ M\^26 - 892\ M\^28 + 4159\ M\^30 - 4039\ M\^32 - 
                209\ M\^34 + 3036\ M\^36 - 1611\ M\^38 - 294\ M\^40 + 
                618\ M\^42 - 366\ M\^44 + 223\ M\^46 - 129\ M\^48 + 
                50\ M\^50 - 11\ M\^52 + M\^54)\) + 
          L\^16\ \((1 - 18\ M\^2 + 143\ M\^4 - 656\ M\^6 + 1882\ M\^8 - 
                3145\ M\^10 + 1171\ M\^12 + 7657\ M\^14 - 16158\ M\^16 + 
                2641\ M\^18 + 33116\ M\^20 - 37770\ M\^22 - 21375\ M\^24 + 
                68816\ M\^26 - 23471\ M\^28 - 59069\ M\^30 + 62628\ M\^32 + 
                19382\ M\^34 - 51573\ M\^36 + 14164\ M\^38 + 17272\ M\^40 - 
                14648\ M\^42 + 2373\ M\^44 + 2807\ M\^46 - 762\ M\^48 - 
                314\ M\^50 - 1686\ M\^52 + 1982\ M\^54 + 89\ M\^56 - 
                1371\ M\^58 + 1087\ M\^60 - 443\ M\^62 + 107\ M\^64 - 
                15\ M\^66 + M\^68)\) + 
          L\^15\ \((\(-5\)\ M\^4 + 92\ M\^6 - 716\ M\^8 + 2997\ M\^10 - 
                6772\ M\^12 + 5220\ M\^14 + 12223\ M\^16 - 33675\ M\^18 + 
                12216\ M\^20 + 64121\ M\^22 - 91786\ M\^24 - 34199\ M\^26 + 
                178624\ M\^28 - 71397\ M\^30 - 193432\ M\^32 + 
                184383\ M\^34 + 111639\ M\^36 - 222526\ M\^38 - 5926\ M\^40 + 
                206063\ M\^42 - 29853\ M\^44 - 158705\ M\^46 + 35146\ M\^48 + 
                118094\ M\^50 - 44875\ M\^52 - 67476\ M\^54 + 53447\ M\^56 + 
                14774\ M\^58 - 40741\ M\^60 + 17296\ M\^62 + 11839\ M\^64 - 
                15678\ M\^66 + 3521\ M\^68 + 4309\ M\^70 - 4111\ M\^72 + 
                1749\ M\^74 - 434\ M\^76 + 62\ M\^78 - 4\ M\^80)\) + 
          L\^14\ \((10\ M\^8 - 188\ M\^10 + 1500\ M\^12 - 6393\ M\^14 + 
                14148\ M\^16 - 7122\ M\^18 - 40149\ M\^20 + 81095\ M\^22 + 
                29464\ M\^24 - 252373\ M\^26 + 135520\ M\^28 + 
                450508\ M\^30 - 503428\ M\^32 - 567160\ M\^34 + 
                996030\ M\^36 + 543977\ M\^38 - 1423550\ M\^40 - 
                429569\ M\^42 + 1654334\ M\^44 + 206187\ M\^46 - 
                1641870\ M\^48 + 221356\ M\^50 + 1432462\ M\^52 - 
                692982\ M\^54 - 942366\ M\^56 + 964003\ M\^58 + 
                323938\ M\^60 - 837599\ M\^62 + 146844\ M\^64 + 
                442868\ M\^66 - 281918\ M\^68 - 93532\ M\^70 + 
                181184\ M\^72 - 53199\ M\^74 - 45462\ M\^76 + 44133\ M\^78 - 
                7710\ M\^80 - 10334\ M\^82 + 8951\ M\^84 - 3570\ M\^86 + 
                822\ M\^88 - 106\ M\^90 + 6\ M\^92)\) + 
          L\^13\ \((\(-10\)\ M\^12 + 192\ M\^14 - 1575\ M\^16 + 6964\ M\^18 - 
                16349\ M\^20 + 11024\ M\^22 + 40466\ M\^24 - 91261\ M\^26 - 
                33890\ M\^28 + 302748\ M\^30 - 126103\ M\^32 - 
                672630\ M\^34 + 549396\ M\^36 + 1318362\ M\^38 - 
                1432559\ M\^40 - 2369193\ M\^42 + 3017903\ M\^44 + 
                3765322\ M\^46 - 5428163\ M\^48 - 4846663\ M\^50 + 
                8141993\ M\^52 + 4582475\ M\^54 - 10001820\ M\^56 - 
                2441687\ M\^58 + 10011863\ M\^60 - 565371\ M\^62 - 
                7855110\ M\^64 + 2896571\ M\^66 + 4549138\ M\^68 - 
                3479610\ M\^70 - 1575170\ M\^72 + 2588188\ M\^74 - 
                117236\ M\^76 - 1249265\ M\^78 + 543167\ M\^80 + 
                318878\ M\^82 - 354747\ M\^84 + 41606\ M\^86 + 
                101287\ M\^88 - 61199\ M\^90 - 298\ M\^92 + 18198\ M\^94 - 
                11432\ M\^96 + 3788\ M\^98 - 747\ M\^100 + 83\ M\^102 - 
                4\ M\^104)\) + 
          L\^12\ \((5\ M\^16 - 98\ M\^18 + 824\ M\^20 - 3802\ M\^22 + 
                9896\ M\^24 - 11508\ M\^26 - 7235\ M\^28 + 32233\ M\^30 + 
                5705\ M\^32 - 88510\ M\^34 - 24941\ M\^36 + 299017\ M\^38 + 
                112581\ M\^40 - 1170967\ M\^42 + 95538\ M\^44 + 
                3407215\ M\^46 - 1570936\ M\^48 - 7403543\ M\^50 + 
                5634501\ M\^52 + 12316937\ M\^54 - 12579629\ M\^56 - 
                16002786\ M\^58 + 20623367\ M\^60 + 16010142\ M\^62 - 
                26372146\ M\^64 - 11497193\ M\^66 + 27224054\ M\^68 + 
                4571917\ M\^70 - 22672505\ M\^72 + 1625007\ M\^74 + 
                15120786\ M\^76 - 4764204\ M\^78 - 7737948\ M\^80 + 
                4776979\ M\^82 + 2683426\ M\^84 - 3176311\ M\^86 - 
                261643\ M\^88 + 1484533\ M\^90 - 376982\ M\^92 - 
                456467\ M\^94 + 302795\ M\^96 + 39724\ M\^98 - 
                108161\ M\^100 + 31690\ M\^102 + 16572\ M\^104 - 
                18011\ M\^106 + 7738\ M\^108 - 1967\ M\^110 + 307\ M\^112 - 
                27\ M\^114 + M\^116)\) + 
          L\^11\ \((\(-M\^20\) + 20\ M\^22 - 173\ M\^24 + 857\ M\^26 - 
                2712\ M\^28 + 5886\ M\^30 - 9679\ M\^32 + 13156\ M\^34 - 
                7795\ M\^36 - 35450\ M\^38 + 109262\ M\^40 - 23345\ M\^42 - 
                409321\ M\^44 + 553363\ M\^46 + 768778\ M\^48 - 
                2159473\ M\^50 - 638352\ M\^52 + 5566631\ M\^54 - 
                1261508\ M\^56 - 10773677\ M\^58 + 6270383\ M\^60 + 
                16469119\ M\^62 - 14542620\ M\^64 - 20106222\ M\^66 + 
                23812538\ M\^68 + 19504666\ M\^70 - 30316862\ M\^72 - 
                14448935\ M\^74 + 31424045\ M\^76 + 7473366\ M\^78 - 
                26804586\ M\^80 - 1289728\ M\^82 + 19036005\ M\^84 - 
                2241330\ M\^86 - 11215833\ M\^88 + 3146855\ M\^90 + 
                5374720\ M\^92 - 2488704\ M\^94 - 2025938\ M\^96 + 
                1454103\ M\^98 + 533922\ M\^100 - 653921\ M\^102 - 
                62909\ M\^104 + 245749\ M\^106 - 45339\ M\^108 - 
                61886\ M\^110 + 31651\ M\^112 + 8526\ M\^114 - 
                15293\ M\^116 + 7908\ M\^118 - 2339\ M\^120 + 424\ M\^122 - 
                44\ M\^124 + 2\ M\^126)\) + 
          L\^10\ \((M\^28 - 21\ M\^30 + 200\ M\^32 - 1123\ M\^34 + 
                4013\ M\^36 - 8957\ M\^38 + 10223\ M\^40 + 2082\ M\^42 - 
                19053\ M\^44 - 1493\ M\^46 + 56998\ M\^48 - 9121\ M\^50 - 
                171049\ M\^52 + 43104\ M\^54 + 556309\ M\^56 - 
                371091\ M\^58 - 1217169\ M\^60 + 1250851\ M\^62 + 
                1945570\ M\^64 - 2736771\ M\^66 - 2259065\ M\^68 + 
                4322608\ M\^70 + 1948329\ M\^72 - 5432957\ M\^74 - 
                1203736\ M\^76 + 5858512\ M\^78 + 440807\ M\^80 - 
                5831246\ M\^82 + 440807\ M\^84 + 5858512\ M\^86 - 
                1203736\ M\^88 - 5432957\ M\^90 + 1948329\ M\^92 + 
                4322608\ M\^94 - 2259065\ M\^96 - 2736771\ M\^98 + 
                1945570\ M\^100 + 1250851\ M\^102 - 1217169\ M\^104 - 
                371091\ M\^106 + 556309\ M\^108 + 43104\ M\^110 - 
                171049\ M\^112 - 9121\ M\^114 + 56998\ M\^116 - 
                1493\ M\^118 - 19053\ M\^120 + 2082\ M\^122 + 10223\ M\^124 - 
                8957\ M\^126 + 4013\ M\^128 - 1123\ M\^130 + 200\ M\^132 - 
                21\ M\^134 + M\^136)\) + 
          L\^9\ \((2\ M\^38 - 44\ M\^40 + 424\ M\^42 - 2339\ M\^44 + 
                7908\ M\^46 - 15293\ M\^48 + 8526\ M\^50 + 31651\ M\^52 - 
                61886\ M\^54 - 45339\ M\^56 + 245749\ M\^58 - 62909\ M\^60 - 
                653921\ M\^62 + 533922\ M\^64 + 1454103\ M\^66 - 
                2025938\ M\^68 - 2488704\ M\^70 + 5374720\ M\^72 + 
                3146855\ M\^74 - 11215833\ M\^76 - 2241330\ M\^78 + 
                19036005\ M\^80 - 1289728\ M\^82 - 26804586\ M\^84 + 
                7473366\ M\^86 + 31424045\ M\^88 - 14448935\ M\^90 - 
                30316862\ M\^92 + 19504666\ M\^94 + 23812538\ M\^96 - 
                20106222\ M\^98 - 14542620\ M\^100 + 16469119\ M\^102 + 
                6270383\ M\^104 - 10773677\ M\^106 - 1261508\ M\^108 + 
                5566631\ M\^110 - 638352\ M\^112 - 2159473\ M\^114 + 
                768778\ M\^116 + 553363\ M\^118 - 409321\ M\^120 - 
                23345\ M\^122 + 109262\ M\^124 - 35450\ M\^126 - 
                7795\ M\^128 + 13156\ M\^130 - 9679\ M\^132 + 5886\ M\^134 - 
                2712\ M\^136 + 857\ M\^138 - 173\ M\^140 + 20\ M\^142 - 
                M\^144)\) + 
          L\^8\ \((M\^48 - 27\ M\^50 + 307\ M\^52 - 1967\ M\^54 + 
                7738\ M\^56 - 18011\ M\^58 + 16572\ M\^60 + 31690\ M\^62 - 
                108161\ M\^64 + 39724\ M\^66 + 302795\ M\^68 - 
                456467\ M\^70 - 376982\ M\^72 + 1484533\ M\^74 - 
                261643\ M\^76 - 3176311\ M\^78 + 2683426\ M\^80 + 
                4776979\ M\^82 - 7737948\ M\^84 - 4764204\ M\^86 + 
                15120786\ M\^88 + 1625007\ M\^90 - 22672505\ M\^92 + 
                4571917\ M\^94 + 27224054\ M\^96 - 11497193\ M\^98 - 
                26372146\ M\^100 + 16010142\ M\^102 + 20623367\ M\^104 - 
                16002786\ M\^106 - 12579629\ M\^108 + 12316937\ M\^110 + 
                5634501\ M\^112 - 7403543\ M\^114 - 1570936\ M\^116 + 
                3407215\ M\^118 + 95538\ M\^120 - 1170967\ M\^122 + 
                112581\ M\^124 + 299017\ M\^126 - 24941\ M\^128 - 
                88510\ M\^130 + 5705\ M\^132 + 32233\ M\^134 - 7235\ M\^136 - 
                11508\ M\^138 + 9896\ M\^140 - 3802\ M\^142 + 824\ M\^144 - 
                98\ M\^146 + 5\ M\^148)\) + 
          L\^7\ \((\(-4\)\ M\^60 + 83\ M\^62 - 747\ M\^64 + 3788\ M\^66 - 
                11432\ M\^68 + 18198\ M\^70 - 298\ M\^72 - 61199\ M\^74 + 
                101287\ M\^76 + 41606\ M\^78 - 354747\ M\^80 + 
                318878\ M\^82 + 543167\ M\^84 - 1249265\ M\^86 - 
                117236\ M\^88 + 2588188\ M\^90 - 1575170\ M\^92 - 
                3479610\ M\^94 + 4549138\ M\^96 + 2896571\ M\^98 - 
                7855110\ M\^100 - 565371\ M\^102 + 10011863\ M\^104 - 
                2441687\ M\^106 - 10001820\ M\^108 + 4582475\ M\^110 + 
                8141993\ M\^112 - 4846663\ M\^114 - 5428163\ M\^116 + 
                3765322\ M\^118 + 3017903\ M\^120 - 2369193\ M\^122 - 
                1432559\ M\^124 + 1318362\ M\^126 + 549396\ M\^128 - 
                672630\ M\^130 - 126103\ M\^132 + 302748\ M\^134 - 
                33890\ M\^136 - 91261\ M\^138 + 40466\ M\^140 + 
                11024\ M\^142 - 16349\ M\^144 + 6964\ M\^146 - 1575\ M\^148 + 
                192\ M\^150 - 10\ M\^152)\) + 
          L\^6\ \((6\ M\^72 - 106\ M\^74 + 822\ M\^76 - 3570\ M\^78 + 
                8951\ M\^80 - 10334\ M\^82 - 7710\ M\^84 + 44133\ M\^86 - 
                45462\ M\^88 - 53199\ M\^90 + 181184\ M\^92 - 93532\ M\^94 - 
                281918\ M\^96 + 442868\ M\^98 + 146844\ M\^100 - 
                837599\ M\^102 + 323938\ M\^104 + 964003\ M\^106 - 
                942366\ M\^108 - 692982\ M\^110 + 1432462\ M\^112 + 
                221356\ M\^114 - 1641870\ M\^116 + 206187\ M\^118 + 
                1654334\ M\^120 - 429569\ M\^122 - 1423550\ M\^124 + 
                543977\ M\^126 + 996030\ M\^128 - 567160\ M\^130 - 
                503428\ M\^132 + 450508\ M\^134 + 135520\ M\^136 - 
                252373\ M\^138 + 29464\ M\^140 + 81095\ M\^142 - 
                40149\ M\^144 - 7122\ M\^146 + 14148\ M\^148 - 6393\ M\^150 + 
                1500\ M\^152 - 188\ M\^154 + 10\ M\^156)\) + 
          L\^5\ \((\(-4\)\ M\^84 + 62\ M\^86 - 434\ M\^88 + 1749\ M\^90 - 
                4111\ M\^92 + 4309\ M\^94 + 3521\ M\^96 - 15678\ M\^98 + 
                11839\ M\^100 + 17296\ M\^102 - 40741\ M\^104 + 
                14774\ M\^106 + 53447\ M\^108 - 67476\ M\^110 - 
                44875\ M\^112 + 118094\ M\^114 + 35146\ M\^116 - 
                158705\ M\^118 - 29853\ M\^120 + 206063\ M\^122 - 
                5926\ M\^124 - 222526\ M\^126 + 111639\ M\^128 + 
                184383\ M\^130 - 193432\ M\^132 - 71397\ M\^134 + 
                178624\ M\^136 - 34199\ M\^138 - 91786\ M\^140 + 
                64121\ M\^142 + 12216\ M\^144 - 33675\ M\^146 + 
                12223\ M\^148 + 5220\ M\^150 - 6772\ M\^152 + 2997\ M\^154 - 
                716\ M\^156 + 92\ M\^158 - 5\ M\^160)\) + 
          L\ \((3\ M\^146 - 11\ M\^148 + 21\ M\^150 - 14\ M\^152 - 
                22\ M\^154 + 57\ M\^156 - 28\ M\^160 + 18\ M\^162 - 
                4\ M\^164)\) + 
          L\^3\ \((M\^110 - 11\ M\^112 + 50\ M\^114 - 129\ M\^116 + 
                223\ M\^118 - 366\ M\^120 + 618\ M\^122 - 294\ M\^124 - 
                1611\ M\^126 + 3036\ M\^128 - 209\ M\^130 - 4039\ M\^132 + 
                4159\ M\^134 - 892\ M\^136 - 3369\ M\^138 + 6675\ M\^140 - 
                1745\ M\^142 - 4181\ M\^144 + 5442\ M\^146 - 1432\ M\^148 - 
                2810\ M\^150 + 2899\ M\^152 - 483\ M\^154 - 963\ M\^156 + 
                820\ M\^158 - 299\ M\^160 + 54\ M\^162 - 4\ M\^164)\) + 
          L\^4\ \((M\^96 - 15\ M\^98 + 107\ M\^100 - 443\ M\^102 + 
                1087\ M\^104 - 1371\ M\^106 + 89\ M\^108 + 1982\ M\^110 - 
                1686\ M\^112 - 314\ M\^114 - 762\ M\^116 + 2807\ M\^118 + 
                2373\ M\^120 - 14648\ M\^122 + 17272\ M\^124 + 
                14164\ M\^126 - 51573\ M\^128 + 19382\ M\^130 + 
                62628\ M\^132 - 59069\ M\^134 - 23471\ M\^136 + 
                68816\ M\^138 - 21375\ M\^140 - 37770\ M\^142 + 
                33116\ M\^144 + 2641\ M\^146 - 16158\ M\^148 + 7657\ M\^150 + 
                1171\ M\^152 - 3145\ M\^154 + 1882\ M\^156 - 656\ M\^158 + 
                143\ M\^160 - 18\ M\^162 + M\^164)\) + 
          L\^2\ \((3\ M\^128 - 22\ M\^130 + 70\ M\^132 - 111\ M\^134 + 
                54\ M\^136 + 77\ M\^138 - 78\ M\^140 + 148\ M\^142 - 
                409\ M\^144 + 34\ M\^146 + 963\ M\^148 - 497\ M\^150 - 
                498\ M\^152 + 681\ M\^154 - 110\ M\^156 - 256\ M\^158 + 
                189\ M\^160 - 54\ M\^162 + 6\ M\^164)\)}, {"A-polynomial", 
        Knot[9, 19], 
        M\^52 + L\^20\ M\^52 + 
          L\ \((4\ M\^42 - 14\ M\^44 + 18\ M\^46 - 7\ M\^48 - 21\ M\^50 + 
                52\ M\^52 + 2\ M\^54 - 27\ M\^56 + 17\ M\^58 - 4\ M\^60)\) + 
          L\^19\ \((\(-4\)\ M\^44 + 17\ M\^46 - 27\ M\^48 + 2\ M\^50 + 
                52\ M\^52 - 21\ M\^54 - 7\ M\^56 + 18\ M\^58 - 14\ M\^60 + 
                4\ M\^62)\) + 
          L\^2\ \((6\ M\^32 - 40\ M\^34 + 115\ M\^36 - 162\ M\^38 - 
                39\ M\^40 + 458\ M\^42 - 313\ M\^44 - 440\ M\^46 + 
                367\ M\^48 + 181\ M\^50 + 227\ M\^52 - 119\ M\^54 - 
                296\ M\^56 + 369\ M\^58 - 35\ M\^60 - 199\ M\^62 + 
                152\ M\^64 - 48\ M\^66 + 6\ M\^68)\) + 
          L\^18\ \((6\ M\^36 - 48\ M\^38 + 152\ M\^40 - 199\ M\^42 - 
                35\ M\^44 + 369\ M\^46 - 296\ M\^48 - 119\ M\^50 + 
                227\ M\^52 + 181\ M\^54 + 367\ M\^56 - 440\ M\^58 - 
                313\ M\^60 + 458\ M\^62 - 39\ M\^64 - 162\ M\^66 + 
                115\ M\^68 - 40\ M\^70 + 6\ M\^72)\) + 
          L\^3\ \((4\ M\^22 - 38\ M\^24 + 162\ M\^26 - 372\ M\^28 + 
                336\ M\^30 + 385\ M\^32 - 1124\ M\^34 + 427\ M\^36 + 
                1270\ M\^38 - 2032\ M\^40 - 27\ M\^42 + 4126\ M\^44 - 
                1215\ M\^46 - 8096\ M\^48 + 2368\ M\^50 + 11951\ M\^52 - 
                3215\ M\^54 - 9331\ M\^56 + 5074\ M\^58 + 3158\ M\^60 - 
                3704\ M\^62 + 858\ M\^64 + 626\ M\^66 - 820\ M\^68 + 
                540\ M\^70 - 212\ M\^72 + 45\ M\^74 - 4\ M\^76)\) + 
          L\^17\ \((\(-4\)\ M\^28 + 45\ M\^30 - 212\ M\^32 + 540\ M\^34 - 
                820\ M\^36 + 626\ M\^38 + 858\ M\^40 - 3704\ M\^42 + 
                3158\ M\^44 + 5074\ M\^46 - 9331\ M\^48 - 3215\ M\^50 + 
                11951\ M\^52 + 2368\ M\^54 - 8096\ M\^56 - 1215\ M\^58 + 
                4126\ M\^60 - 27\ M\^62 - 2032\ M\^64 + 1270\ M\^66 + 
                427\ M\^68 - 1124\ M\^70 + 385\ M\^72 + 336\ M\^74 - 
                372\ M\^76 + 162\ M\^78 - 38\ M\^80 + 4\ M\^82)\) + 
          L\^4\ \((M\^12 - 12\ M\^14 + 66\ M\^16 - 208\ M\^18 + 385\ M\^20 - 
                394\ M\^22 + 256\ M\^24 - 303\ M\^26 + 126\ M\^28 + 
                1162\ M\^30 - 2764\ M\^32 + 1350\ M\^34 + 5892\ M\^36 - 
                9007\ M\^38 - 4890\ M\^40 + 7888\ M\^42 + 13108\ M\^44 + 
                8183\ M\^46 - 42623\ M\^48 - 22818\ M\^50 + 73112\ M\^52 + 
                17935\ M\^54 - 60982\ M\^56 + 7246\ M\^58 + 22849\ M\^60 - 
                17258\ M\^62 + 6724\ M\^64 + 6228\ M\^66 - 10545\ M\^68 + 
                3372\ M\^70 + 2680\ M\^72 - 2937\ M\^74 + 1356\ M\^76 - 
                413\ M\^78 + 93\ M\^80 - 14\ M\^82 + M\^84)\) + 
          L\^5\ \((\(-3\)\ M\^8 + 43\ M\^10 - 270\ M\^12 + 913\ M\^14 - 
                1571\ M\^16 + 506\ M\^18 + 2646\ M\^20 - 2433\ M\^22 - 
                5621\ M\^24 + 9614\ M\^26 + 7253\ M\^28 - 24320\ M\^30 - 
                3385\ M\^32 + 43026\ M\^34 - 12258\ M\^36 - 23670\ M\^38 - 
                3712\ M\^40 - 62292\ M\^42 + 105481\ M\^44 + 177088\ M\^46 - 
                273823\ M\^48 - 213512\ M\^50 + 376010\ M\^52 + 
                120087\ M\^54 - 286700\ M\^56 + 58866\ M\^58 + 
                102146\ M\^60 - 142086\ M\^62 + 38406\ M\^64 + 92248\ M\^66 - 
                72122\ M\^68 - 13276\ M\^70 + 35155\ M\^72 - 13791\ M\^74 - 
                2396\ M\^76 + 6353\ M\^78 - 4685\ M\^80 + 2060\ M\^82 - 
                546\ M\^84 + 80\ M\^86 - 5\ M\^88)\) + 
          L\^16\ \((M\^20 - 14\ M\^22 + 93\ M\^24 - 413\ M\^26 + 
                1356\ M\^28 - 2937\ M\^30 + 2680\ M\^32 + 3372\ M\^34 - 
                10545\ M\^36 + 6228\ M\^38 + 6724\ M\^40 - 17258\ M\^42 + 
                22849\ M\^44 + 7246\ M\^46 - 60982\ M\^48 + 17935\ M\^50 + 
                73112\ M\^52 - 22818\ M\^54 - 42623\ M\^56 + 8183\ M\^58 + 
                13108\ M\^60 + 7888\ M\^62 - 4890\ M\^64 - 9007\ M\^66 + 
                5892\ M\^68 + 1350\ M\^70 - 2764\ M\^72 + 1162\ M\^74 + 
                126\ M\^76 - 303\ M\^78 + 256\ M\^80 - 394\ M\^82 + 
                385\ M\^84 - 208\ M\^86 + 66\ M\^88 - 12\ M\^90 + M\^92)\) + 
          L\^6\ \((3\ M\^4 - 50\ M\^6 + 357\ M\^8 - 1381\ M\^10 + 
                2964\ M\^12 - 2825\ M\^14 - 1429\ M\^16 + 6398\ M\^18 - 
                4620\ M\^20 - 4539\ M\^22 + 13229\ M\^24 - 15853\ M\^26 + 
                8117\ M\^28 + 29343\ M\^30 - 63018\ M\^32 - 37901\ M\^34 + 
                142302\ M\^36 + 98548\ M\^38 - 208962\ M\^40 - 
                339720\ M\^42 + 268152\ M\^44 + 738630\ M\^46 - 
                329760\ M\^48 - 1025207\ M\^50 + 377502\ M\^52 + 
                908486\ M\^54 - 301526\ M\^56 - 366487\ M\^58 + 
                182505\ M\^60 - 100457\ M\^62 - 26945\ M\^64 + 
                193984\ M\^66 - 83767\ M\^68 - 66549\ M\^70 + 82654\ M\^72 - 
                38147\ M\^74 - 16628\ M\^76 + 38035\ M\^78 - 15376\ M\^80 - 
                8708\ M\^82 + 11652\ M\^84 - 5422\ M\^86 + 1344\ M\^88 - 
                178\ M\^90 + 10\ M\^92)\) + 
          L\^7\ \((\(-1\) + 19\ M\^2 - 157\ M\^4 + 756\ M\^6 - 2426\ M\^8 + 
                5502\ M\^10 - 7943\ M\^12 + 1605\ M\^14 + 21911\ M\^16 - 
                41142\ M\^18 - 5552\ M\^20 + 126294\ M\^22 - 147332\ M\^24 - 
                149166\ M\^26 + 474548\ M\^28 - 27225\ M\^30 - 
                884533\ M\^32 + 472141\ M\^34 + 1026572\ M\^36 - 
                951913\ M\^38 - 500171\ M\^40 + 1086064\ M\^42 - 
                857492\ M\^44 - 669335\ M\^46 + 2553389\ M\^48 - 
                34672\ M\^50 - 3650688\ M\^52 + 579883\ M\^54 + 
                3622189\ M\^56 - 553174\ M\^58 - 2552588\ M\^60 + 
                328535\ M\^62 + 1372750\ M\^64 - 205806\ M\^66 - 
                689093\ M\^68 + 211447\ M\^70 + 364796\ M\^72 - 
                227376\ M\^74 - 159276\ M\^76 + 185298\ M\^78 + 
                12420\ M\^80 - 84448\ M\^82 + 28364\ M\^84 + 16733\ M\^86 - 
                18062\ M\^88 + 7306\ M\^90 - 1615\ M\^92 + 194\ M\^94 - 
                10\ M\^96)\) + 
          L\^15\ \((\(-5\)\ M\^16 + 80\ M\^18 - 546\ M\^20 + 2060\ M\^22 - 
                4685\ M\^24 + 6353\ M\^26 - 2396\ M\^28 - 13791\ M\^30 + 
                35155\ M\^32 - 13276\ M\^34 - 72122\ M\^36 + 92248\ M\^38 + 
                38406\ M\^40 - 142086\ M\^42 + 102146\ M\^44 + 58866\ M\^46 - 
                286700\ M\^48 + 120087\ M\^50 + 376010\ M\^52 - 
                213512\ M\^54 - 273823\ M\^56 + 177088\ M\^58 + 
                105481\ M\^60 - 62292\ M\^62 - 3712\ M\^64 - 23670\ M\^66 - 
                12258\ M\^68 + 43026\ M\^70 - 3385\ M\^72 - 24320\ M\^74 + 
                7253\ M\^76 + 9614\ M\^78 - 5621\ M\^80 - 2433\ M\^82 + 
                2646\ M\^84 + 506\ M\^86 - 1571\ M\^88 + 913\ M\^90 - 
                270\ M\^92 + 43\ M\^94 - 3\ M\^96)\) + 
          L\^14\ \((10\ M\^12 - 178\ M\^14 + 1344\ M\^16 - 5422\ M\^18 + 
                11652\ M\^20 - 8708\ M\^22 - 15376\ M\^24 + 38035\ M\^26 - 
                16628\ M\^28 - 38147\ M\^30 + 82654\ M\^32 - 66549\ M\^34 - 
                83767\ M\^36 + 193984\ M\^38 - 26945\ M\^40 - 100457\ M\^42 + 
                182505\ M\^44 - 366487\ M\^46 - 301526\ M\^48 + 
                908486\ M\^50 + 377502\ M\^52 - 1025207\ M\^54 - 
                329760\ M\^56 + 738630\ M\^58 + 268152\ M\^60 - 
                339720\ M\^62 - 208962\ M\^64 + 98548\ M\^66 + 
                142302\ M\^68 - 37901\ M\^70 - 63018\ M\^72 + 29343\ M\^74 + 
                8117\ M\^76 - 15853\ M\^78 + 13229\ M\^80 - 4539\ M\^82 - 
                4620\ M\^84 + 6398\ M\^86 - 1429\ M\^88 - 2825\ M\^90 + 
                2964\ M\^92 - 1381\ M\^94 + 357\ M\^96 - 50\ M\^98 + 
                3\ M\^100)\) + 
          L\^8\ \((4 - 79\ M\^2 + 664\ M\^4 - 3073\ M\^6 + 8273\ M\^8 - 
                11466\ M\^10 - 106\ M\^12 + 28384\ M\^14 - 35707\ M\^16 - 
                25612\ M\^18 + 99358\ M\^20 - 42537\ M\^22 - 60064\ M\^24 + 
                9365\ M\^26 - 181286\ M\^28 + 456806\ M\^30 + 698627\ M\^32 - 
                1688435\ M\^34 - 1440577\ M\^36 + 3389843\ M\^38 + 
                2418589\ M\^40 - 4572706\ M\^42 - 3501472\ M\^44 + 
                4100108\ M\^46 + 4213328\ M\^48 - 1890015\ M\^50 - 
                3891886\ M\^52 - 826705\ M\^54 + 2669110\ M\^56 + 
                2907287\ M\^58 - 1023448\ M\^60 - 3509115\ M\^62 + 
                67646\ M\^64 + 2940045\ M\^66 - 78715\ M\^68 - 
                1942178\ M\^70 + 437986\ M\^72 + 952854\ M\^74 - 
                536627\ M\^76 - 265203\ M\^78 + 355469\ M\^80 - 
                46306\ M\^82 - 113511\ M\^84 + 68296\ M\^86 + 4482\ M\^88 - 
                24604\ M\^90 + 14883\ M\^92 - 4850\ M\^94 + 945\ M\^96 - 
                104\ M\^98 + 5\ M\^100)\) + 
          L\^11\ \((\(-1\) + 22\ M\^2 - 221\ M\^4 + 1359\ M\^6 - 5706\ M\^8 + 
                16578\ M\^10 - 29271\ M\^12 + 9034\ M\^14 + 94190\ M\^16 - 
                187396\ M\^18 - 82897\ M\^20 + 709097\ M\^22 - 
                421638\ M\^24 - 1540486\ M\^26 + 1884740\ M\^28 + 
                2525290\ M\^30 - 4601955\ M\^32 - 3569124\ M\^34 + 
                8161481\ M\^36 + 5157495\ M\^38 - 11344602\ M\^40 - 
                7760710\ M\^42 + 12261588\ M\^44 + 10731201\ M\^46 - 
                9681845\ M\^48 - 12184454\ M\^50 + 4079064\ M\^52 + 
                10775355\ M\^54 + 2607011\ M\^56 - 6553719\ M\^58 - 
                7245308\ M\^60 + 2050508\ M\^62 + 8491031\ M\^64 + 
                637488\ M\^66 - 7088999\ M\^68 - 1216809\ M\^70 + 
                4655682\ M\^72 + 713406\ M\^74 - 2530249\ M\^76 - 
                109135\ M\^78 + 1130756\ M\^80 - 156028\ M\^82 - 
                399286\ M\^84 + 173722\ M\^86 + 71919\ M\^88 - 79573\ M\^90 + 
                10226\ M\^92 + 19480\ M\^94 - 14624\ M\^96 + 5236\ M\^98 - 
                1080\ M\^100 + 123\ M\^102 - 6\ M\^104)\) + 
          L\^13\ \((\(-10\)\ M\^8 + 194\ M\^10 - 1615\ M\^12 + 7306\ M\^14 - 
                18062\ M\^16 + 16733\ M\^18 + 28364\ M\^20 - 84448\ M\^22 + 
                12420\ M\^24 + 185298\ M\^26 - 159276\ M\^28 - 
                227376\ M\^30 + 364796\ M\^32 + 211447\ M\^34 - 
                689093\ M\^36 - 205806\ M\^38 + 1372750\ M\^40 + 
                328535\ M\^42 - 2552588\ M\^44 - 553174\ M\^46 + 
                3622189\ M\^48 + 579883\ M\^50 - 3650688\ M\^52 - 
                34672\ M\^54 + 2553389\ M\^56 - 669335\ M\^58 - 
                857492\ M\^60 + 1086064\ M\^62 - 500171\ M\^64 - 
                951913\ M\^66 + 1026572\ M\^68 + 472141\ M\^70 - 
                884533\ M\^72 - 27225\ M\^74 + 474548\ M\^76 - 
                149166\ M\^78 - 147332\ M\^80 + 126294\ M\^82 - 5552\ M\^84 - 
                41142\ M\^86 + 21911\ M\^88 + 1605\ M\^90 - 7943\ M\^92 + 
                5502\ M\^94 - 2426\ M\^96 + 756\ M\^98 - 157\ M\^100 + 
                19\ M\^102 - M\^104)\) + 
          L\^9\ \((\(-6\) + 123\ M\^2 - 1080\ M\^4 + 5236\ M\^6 - 
                14624\ M\^8 + 19480\ M\^10 + 10226\ M\^12 - 79573\ M\^14 + 
                71919\ M\^16 + 173722\ M\^18 - 399286\ M\^20 - 
                156028\ M\^22 + 1130756\ M\^24 - 109135\ M\^26 - 
                2530249\ M\^28 + 713406\ M\^30 + 4655682\ M\^32 - 
                1216809\ M\^34 - 7088999\ M\^36 + 637488\ M\^38 + 
                8491031\ M\^40 + 2050508\ M\^42 - 7245308\ M\^44 - 
                6553719\ M\^46 + 2607011\ M\^48 + 10775355\ M\^50 + 
                4079064\ M\^52 - 12184454\ M\^54 - 9681845\ M\^56 + 
                10731201\ M\^58 + 12261588\ M\^60 - 7760710\ M\^62 - 
                11344602\ M\^64 + 5157495\ M\^66 + 8161481\ M\^68 - 
                3569124\ M\^70 - 4601955\ M\^72 + 2525290\ M\^74 + 
                1884740\ M\^76 - 1540486\ M\^78 - 421638\ M\^80 + 
                709097\ M\^82 - 82897\ M\^84 - 187396\ M\^86 + 94190\ M\^88 + 
                9034\ M\^90 - 29271\ M\^92 + 16578\ M\^94 - 5706\ M\^96 + 
                1359\ M\^98 - 221\ M\^100 + 22\ M\^102 - M\^104)\) + 
          L\^10\ \((4 - 85\ M\^2 + 786\ M\^4 - 4124\ M\^6 + 13194\ M\^8 - 
                24128\ M\^10 + 11219\ M\^12 + 60450\ M\^14 - 139115\ M\^16 - 
                10015\ M\^18 + 483086\ M\^20 - 472805\ M\^22 - 
                975707\ M\^24 + 1774970\ M\^26 + 1581600\ M\^28 - 
                4238332\ M\^30 - 2543966\ M\^32 + 7639657\ M\^34 + 
                4972515\ M\^36 - 10975190\ M\^38 - 10175635\ M\^40 + 
                12346275\ M\^42 + 18046761\ M\^44 - 10087391\ M\^46 - 
                25770109\ M\^48 + 3990718\ M\^50 + 29175490\ M\^52 + 
                3990718\ M\^54 - 25770109\ M\^56 - 10087391\ M\^58 + 
                18046761\ M\^60 + 12346275\ M\^62 - 10175635\ M\^64 - 
                10975190\ M\^66 + 4972515\ M\^68 + 7639657\ M\^70 - 
                2543966\ M\^72 - 4238332\ M\^74 + 1581600\ M\^76 + 
                1774970\ M\^78 - 975707\ M\^80 - 472805\ M\^82 + 
                483086\ M\^84 - 10015\ M\^86 - 139115\ M\^88 + 60450\ M\^90 + 
                11219\ M\^92 - 24128\ M\^94 + 13194\ M\^96 - 4124\ M\^98 + 
                786\ M\^100 - 85\ M\^102 + 4\ M\^104)\) + 
          L\^12\ \((5\ M\^4 - 104\ M\^6 + 945\ M\^8 - 4850\ M\^10 + 
                14883\ M\^12 - 24604\ M\^14 + 4482\ M\^16 + 68296\ M\^18 - 
                113511\ M\^20 - 46306\ M\^22 + 355469\ M\^24 - 
                265203\ M\^26 - 536627\ M\^28 + 952854\ M\^30 + 
                437986\ M\^32 - 1942178\ M\^34 - 78715\ M\^36 + 
                2940045\ M\^38 + 67646\ M\^40 - 3509115\ M\^42 - 
                1023448\ M\^44 + 2907287\ M\^46 + 2669110\ M\^48 - 
                826705\ M\^50 - 3891886\ M\^52 - 1890015\ M\^54 + 
                4213328\ M\^56 + 4100108\ M\^58 - 3501472\ M\^60 - 
                4572706\ M\^62 + 2418589\ M\^64 + 3389843\ M\^66 - 
                1440577\ M\^68 - 1688435\ M\^70 + 698627\ M\^72 + 
                456806\ M\^74 - 181286\ M\^76 + 9365\ M\^78 - 60064\ M\^80 - 
                42537\ M\^82 + 99358\ M\^84 - 25612\ M\^86 - 35707\ M\^88 + 
                28384\ M\^90 - 106\ M\^92 - 11466\ M\^94 + 8273\ M\^96 - 
                3073\ M\^98 + 664\ M\^100 - 79\ M\^102 + 
                4\ M\^104)\)}, {"A-polynomial", Knot[9, 20], 
        L\^20\ M\^8 + M\^172 + 
          L\^19\ \((\(-2\)\ M\^4 + 8\ M\^6 - 15\ M\^8 + 14\ M\^10 + 
                5\ M\^12 - 34\ M\^14 + 43\ M\^16 + 23\ M\^18 - 31\ M\^20 + 
                9\ M\^22)\) + 
          L\^18\ \((1 - 8\ M\^2 + 28\ M\^4 - 50\ M\^6 + 39\ M\^8 + 
                38\ M\^12 - 110\ M\^14 - 204\ M\^16 + 702\ M\^18 - 
                149\ M\^20 - 1230\ M\^22 + 1107\ M\^24 + 916\ M\^26 - 
                1182\ M\^28 + 14\ M\^30 + 476\ M\^32 - 234\ M\^34 + 
                36\ M\^36)\) + 
          L\^17\ \((\(-1\) + 12\ M\^2 - 67\ M\^4 + 214\ M\^6 - 374\ M\^8 + 
                138\ M\^10 + 824\ M\^12 - 1367\ M\^14 - 787\ M\^16 + 
                3939\ M\^18 - 656\ M\^20 - 7875\ M\^22 + 4153\ M\^24 + 
                12938\ M\^26 - 11584\ M\^28 - 14875\ M\^30 + 20413\ M\^32 + 
                9914\ M\^34 - 20622\ M\^36 + 525\ M\^38 + 11706\ M\^40 - 
                4980\ M\^42 - 2235\ M\^44 + 2473\ M\^46 - 770\ M\^48 + 
                84\ M\^50)\) + 
          L\^16\ \((\(-2\)\ M\^2 + 27\ M\^4 - 169\ M\^6 + 615\ M\^8 - 
                1279\ M\^10 + 863\ M\^12 + 2773\ M\^14 - 7402\ M\^16 + 
                1837\ M\^18 + 18679\ M\^20 - 25460\ M\^22 - 16802\ M\^24 + 
                65853\ M\^26 - 19401\ M\^28 - 98552\ M\^30 + 91407\ M\^32 + 
                92222\ M\^34 - 167047\ M\^36 - 46693\ M\^38 + 205177\ M\^40 - 
                9291\ M\^42 - 175913\ M\^44 + 56909\ M\^46 + 104015\ M\^48 - 
                67505\ M\^50 - 30477\ M\^52 + 41018\ M\^54 - 5180\ M\^56 - 
                10219\ M\^58 + 6158\ M\^60 - 1442\ M\^62 + 126\ M\^64)\) + 
          L\^15\ \((\(-M\^4\) + 16\ M\^6 - 122\ M\^8 + 560\ M\^10 - 
                1610\ M\^12 + 2632\ M\^14 - 889\ M\^16 - 6192\ M\^18 + 
                13344\ M\^20 - 3875\ M\^22 - 31487\ M\^24 + 55751\ M\^26 + 
                5755\ M\^28 - 133945\ M\^30 + 118442\ M\^32 + 167136\ M\^34 - 
                352271\ M\^36 - 53136\ M\^38 + 615827\ M\^40 - 
                230912\ M\^42 - 783319\ M\^44 + 559430\ M\^46 + 
                748925\ M\^48 - 726860\ M\^50 - 506203\ M\^52 + 
                689092\ M\^54 + 224235\ M\^56 - 499730\ M\^58 - 
                20526\ M\^60 + 276803\ M\^62 - 69198\ M\^64 - 97742\ M\^66 + 
                61087\ M\^68 + 7605\ M\^70 - 20309\ M\^72 + 8745\ M\^74 - 
                1680\ M\^76 + 126\ M\^78)\) + 
          L\^14\ \((\(-3\)\ M\^10 + 53\ M\^12 - 428\ M\^14 + 2006\ M\^16 - 
                5594\ M\^18 + 7554\ M\^20 + 4316\ M\^22 - 34498\ M\^24 + 
                40647\ M\^26 + 46853\ M\^28 - 172801\ M\^30 + 87344\ M\^32 + 
                291004\ M\^34 - 458050\ M\^36 - 136502\ M\^38 + 
                939950\ M\^40 - 554947\ M\^42 - 1131947\ M\^44 + 
                1703826\ M\^46 + 744235\ M\^48 - 2828102\ M\^50 + 
                66136\ M\^52 + 3318713\ M\^54 - 852102\ M\^56 - 
                2936777\ M\^58 + 1340113\ M\^60 + 2063215\ M\^62 - 
                1338451\ M\^64 - 1133176\ M\^66 + 996813\ M\^68 + 
                475904\ M\^70 - 585220\ M\^72 - 142650\ M\^74 + 
                293850\ M\^76 - 1652\ M\^78 - 116100\ M\^80 + 38076\ M\^82 + 
                22649\ M\^84 - 21823\ M\^86 + 7488\ M\^88 - 1246\ M\^90 + 
                84\ M\^92)\) + 
          L\^13\ \((\(-3\)\ M\^16 + 60\ M\^18 - 548\ M\^20 + 2905\ M\^22 - 
                9283\ M\^24 + 15548\ M\^26 + 629\ M\^28 - 59474\ M\^30 + 
                93852\ M\^32 + 60520\ M\^34 - 354414\ M\^36 + 236365\ M\^38 + 
                620477\ M\^40 - 1068551\ M\^42 - 440914\ M\^44 + 
                2414775\ M\^46 - 796372\ M\^48 - 3746741\ M\^50 + 
                3273148\ M\^52 + 4314079\ M\^54 - 6316982\ M\^56 - 
                3544267\ M\^58 + 8506278\ M\^60 + 1493443\ M\^62 - 
                8765528\ M\^64 + 964955\ M\^66 + 7257511\ M\^68 - 
                2433350\ M\^70 - 4777956\ M\^72 + 2593222\ M\^74 + 
                2461674\ M\^76 - 1903812\ M\^78 - 936862\ M\^80 + 
                1036489\ M\^82 + 226752\ M\^84 - 402828\ M\^86 - 
                46102\ M\^88 + 129166\ M\^90 + 17877\ M\^92 - 50226\ M\^94 + 
                533\ M\^96 + 21375\ M\^98 - 13193\ M\^100 + 3831\ M\^102 - 
                574\ M\^104 + 36\ M\^106)\) + 
          L\^12\ \((\(-M\^22\) + 26\ M\^24 - 298\ M\^26 + 1944\ M\^28 - 
                7723\ M\^30 + 17762\ M\^32 - 14715\ M\^34 - 35851\ M\^36 + 
                106222\ M\^38 - 18310\ M\^40 - 313150\ M\^42 + 
                366851\ M\^44 + 513462\ M\^46 - 1288107\ M\^48 - 
                369859\ M\^50 + 3101750\ M\^52 - 911073\ M\^54 - 
                5617479\ M\^56 + 4100840\ M\^58 + 7792666\ M\^60 - 
                9092187\ M\^62 - 7949728\ M\^64 + 14305771\ M\^66 + 
                5224036\ M\^68 - 17627812\ M\^70 - 711562\ M\^72 + 
                17609396\ M\^74 - 3172667\ M\^76 - 14059318\ M\^78 + 
                5188012\ M\^80 + 8940304\ M\^82 - 5049101\ M\^84 - 
                4436370\ M\^86 + 3604273\ M\^88 + 1626692\ M\^90 - 
                1992730\ M\^92 - 353340\ M\^94 + 874754\ M\^96 - 
                64788\ M\^98 - 258906\ M\^100 + 82360\ M\^102 + 
                30903\ M\^104 - 19552\ M\^106 + 7046\ M\^108 - 
                10486\ M\^110 + 9301\ M\^112 - 4225\ M\^114 + 1078\ M\^116 - 
                150\ M\^118 + 9\ M\^120)\) + 
          L\^11\ \((3\ M\^30 - 62\ M\^32 + 572\ M\^34 - 3028\ M\^36 + 
                9665\ M\^38 - 16695\ M\^40 + 4392\ M\^42 + 38872\ M\^44 - 
                42679\ M\^46 - 99336\ M\^48 + 202581\ M\^50 + 203137\ M\^52 - 
                737770\ M\^54 - 215907\ M\^56 + 2178681\ M\^58 - 
                618850\ M\^60 - 4673586\ M\^62 + 3362901\ M\^64 + 
                7540357\ M\^66 - 8753205\ M\^68 - 9144401\ M\^70 + 
                16199679\ M\^72 + 7997181\ M\^74 - 23450730\ M\^76 - 
                3942666\ M\^78 + 27133213\ M\^80 - 1531150\ M\^82 - 
                25127648\ M\^84 + 6459616\ M\^86 + 19038088\ M\^88 - 
                8550767\ M\^90 - 11723612\ M\^92 + 7635180\ M\^94 + 
                5656085\ M\^96 - 5131998\ M\^98 - 1963451\ M\^100 + 
                2662361\ M\^102 + 378836\ M\^104 - 1052939\ M\^106 + 
                1345\ M\^108 + 356848\ M\^110 - 50705\ M\^112 - 
                106030\ M\^114 + 39814\ M\^116 + 21560\ M\^118 - 
                23895\ M\^120 + 11142\ M\^122 - 4267\ M\^124 + 1678\ M\^126 - 
                562\ M\^128 + 128\ M\^130 - 17\ M\^132 + M\^134)\) + 
          L\^10\ \((\(-3\)\ M\^38 + 56\ M\^40 - 460\ M\^42 + 2141\ M\^44 - 
                5967\ M\^46 + 9407\ M\^48 - 6900\ M\^50 + 6172\ M\^52 - 
                28315\ M\^54 + 23804\ M\^56 + 137684\ M\^58 - 270703\ M\^60 - 
                268494\ M\^62 + 1094369\ M\^64 + 34228\ M\^66 - 
                2788526\ M\^68 + 1215498\ M\^70 + 5522354\ M\^72 - 
                4597428\ M\^74 - 8593732\ M\^76 + 10801040\ M\^78 + 
                10327507\ M\^80 - 18827052\ M\^82 - 8824397\ M\^84 + 
                25626497\ M\^86 + 3583926\ M\^88 - 28160656\ M\^90 + 
                3583926\ M\^92 + 25626497\ M\^94 - 8824397\ M\^96 - 
                18827052\ M\^98 + 10327507\ M\^100 + 10801040\ M\^102 - 
                8593732\ M\^104 - 4597428\ M\^106 + 5522354\ M\^108 + 
                1215498\ M\^110 - 2788526\ M\^112 + 34228\ M\^114 + 
                1094369\ M\^116 - 268494\ M\^118 - 270703\ M\^120 + 
                137684\ M\^122 + 23804\ M\^124 - 28315\ M\^126 + 
                6172\ M\^128 - 6900\ M\^130 + 9407\ M\^132 - 5967\ M\^134 + 
                2141\ M\^136 - 460\ M\^138 + 56\ M\^140 - 3\ M\^142)\) + 
          L\^9\ \((M\^46 - 17\ M\^48 + 128\ M\^50 - 562\ M\^52 + 
                1678\ M\^54 - 4267\ M\^56 + 11142\ M\^58 - 23895\ M\^60 + 
                21560\ M\^62 + 39814\ M\^64 - 106030\ M\^66 - 50705\ M\^68 + 
                356848\ M\^70 + 1345\ M\^72 - 1052939\ M\^74 + 
                378836\ M\^76 + 2662361\ M\^78 - 1963451\ M\^80 - 
                5131998\ M\^82 + 5656085\ M\^84 + 7635180\ M\^86 - 
                11723612\ M\^88 - 8550767\ M\^90 + 19038088\ M\^92 + 
                6459616\ M\^94 - 25127648\ M\^96 - 1531150\ M\^98 + 
                27133213\ M\^100 - 3942666\ M\^102 - 23450730\ M\^104 + 
                7997181\ M\^106 + 16199679\ M\^108 - 9144401\ M\^110 - 
                8753205\ M\^112 + 7540357\ M\^114 + 3362901\ M\^116 - 
                4673586\ M\^118 - 618850\ M\^120 + 2178681\ M\^122 - 
                215907\ M\^124 - 737770\ M\^126 + 203137\ M\^128 + 
                202581\ M\^130 - 99336\ M\^132 - 42679\ M\^134 + 
                38872\ M\^136 + 4392\ M\^138 - 16695\ M\^140 + 9665\ M\^142 - 
                3028\ M\^144 + 572\ M\^146 - 62\ M\^148 + 3\ M\^150)\) + 
          L\^8\ \((9\ M\^60 - 150\ M\^62 + 1078\ M\^64 - 4225\ M\^66 + 
                9301\ M\^68 - 10486\ M\^70 + 7046\ M\^72 - 19552\ M\^74 + 
                30903\ M\^76 + 82360\ M\^78 - 258906\ M\^80 - 64788\ M\^82 + 
                874754\ M\^84 - 353340\ M\^86 - 1992730\ M\^88 + 
                1626692\ M\^90 + 3604273\ M\^92 - 4436370\ M\^94 - 
                5049101\ M\^96 + 8940304\ M\^98 + 5188012\ M\^100 - 
                14059318\ M\^102 - 3172667\ M\^104 + 17609396\ M\^106 - 
                711562\ M\^108 - 17627812\ M\^110 + 5224036\ M\^112 + 
                14305771\ M\^114 - 7949728\ M\^116 - 9092187\ M\^118 + 
                7792666\ M\^120 + 4100840\ M\^122 - 5617479\ M\^124 - 
                911073\ M\^126 + 3101750\ M\^128 - 369859\ M\^130 - 
                1288107\ M\^132 + 513462\ M\^134 + 366851\ M\^136 - 
                313150\ M\^138 - 18310\ M\^140 + 106222\ M\^142 - 
                35851\ M\^144 - 14715\ M\^146 + 17762\ M\^148 - 
                7723\ M\^150 + 1944\ M\^152 - 298\ M\^154 + 26\ M\^156 - 
                M\^158)\) + 
          L\^7\ \((36\ M\^74 - 574\ M\^76 + 3831\ M\^78 - 13193\ M\^80 + 
                21375\ M\^82 + 533\ M\^84 - 50226\ M\^86 + 17877\ M\^88 + 
                129166\ M\^90 - 46102\ M\^92 - 402828\ M\^94 + 
                226752\ M\^96 + 1036489\ M\^98 - 936862\ M\^100 - 
                1903812\ M\^102 + 2461674\ M\^104 + 2593222\ M\^106 - 
                4777956\ M\^108 - 2433350\ M\^110 + 7257511\ M\^112 + 
                964955\ M\^114 - 8765528\ M\^116 + 1493443\ M\^118 + 
                8506278\ M\^120 - 3544267\ M\^122 - 6316982\ M\^124 + 
                4314079\ M\^126 + 3273148\ M\^128 - 3746741\ M\^130 - 
                796372\ M\^132 + 2414775\ M\^134 - 440914\ M\^136 - 
                1068551\ M\^138 + 620477\ M\^140 + 236365\ M\^142 - 
                354414\ M\^144 + 60520\ M\^146 + 93852\ M\^148 - 
                59474\ M\^150 + 629\ M\^152 + 15548\ M\^154 - 9283\ M\^156 + 
                2905\ M\^158 - 548\ M\^160 + 60\ M\^162 - 3\ M\^164)\) + 
          L\^6\ \((84\ M\^88 - 1246\ M\^90 + 7488\ M\^92 - 21823\ M\^94 + 
                22649\ M\^96 + 38076\ M\^98 - 116100\ M\^100 - 1652\ M\^102 + 
                293850\ M\^104 - 142650\ M\^106 - 585220\ M\^108 + 
                475904\ M\^110 + 996813\ M\^112 - 1133176\ M\^114 - 
                1338451\ M\^116 + 2063215\ M\^118 + 1340113\ M\^120 - 
                2936777\ M\^122 - 852102\ M\^124 + 3318713\ M\^126 + 
                66136\ M\^128 - 2828102\ M\^130 + 744235\ M\^132 + 
                1703826\ M\^134 - 1131947\ M\^136 - 554947\ M\^138 + 
                939950\ M\^140 - 136502\ M\^142 - 458050\ M\^144 + 
                291004\ M\^146 + 87344\ M\^148 - 172801\ M\^150 + 
                46853\ M\^152 + 40647\ M\^154 - 34498\ M\^156 + 
                4316\ M\^158 + 7554\ M\^160 - 5594\ M\^162 + 2006\ M\^164 - 
                428\ M\^166 + 53\ M\^168 - 3\ M\^170)\) + 
          L\ \((9\ M\^158 - 31\ M\^160 + 23\ M\^162 + 43\ M\^164 - 
                34\ M\^166 + 5\ M\^168 + 14\ M\^170 - 15\ M\^172 + 
                8\ M\^174 - 2\ M\^176)\) + 
          L\^5\ \((126\ M\^102 - 1680\ M\^104 + 8745\ M\^106 - 
                20309\ M\^108 + 7605\ M\^110 + 61087\ M\^112 - 
                97742\ M\^114 - 69198\ M\^116 + 276803\ M\^118 - 
                20526\ M\^120 - 499730\ M\^122 + 224235\ M\^124 + 
                689092\ M\^126 - 506203\ M\^128 - 726860\ M\^130 + 
                748925\ M\^132 + 559430\ M\^134 - 783319\ M\^136 - 
                230912\ M\^138 + 615827\ M\^140 - 53136\ M\^142 - 
                352271\ M\^144 + 167136\ M\^146 + 118442\ M\^148 - 
                133945\ M\^150 + 5755\ M\^152 + 55751\ M\^154 - 
                31487\ M\^156 - 3875\ M\^158 + 13344\ M\^160 - 6192\ M\^162 - 
                889\ M\^164 + 2632\ M\^166 - 1610\ M\^168 + 560\ M\^170 - 
                122\ M\^172 + 16\ M\^174 - M\^176)\) + 
          L\^4\ \((126\ M\^116 - 1442\ M\^118 + 6158\ M\^120 - 
                10219\ M\^122 - 5180\ M\^124 + 41018\ M\^126 - 
                30477\ M\^128 - 67505\ M\^130 + 104015\ M\^132 + 
                56909\ M\^134 - 175913\ M\^136 - 9291\ M\^138 + 
                205177\ M\^140 - 46693\ M\^142 - 167047\ M\^144 + 
                92222\ M\^146 + 91407\ M\^148 - 98552\ M\^150 - 
                19401\ M\^152 + 65853\ M\^154 - 16802\ M\^156 - 
                25460\ M\^158 + 18679\ M\^160 + 1837\ M\^162 - 7402\ M\^164 + 
                2773\ M\^166 + 863\ M\^168 - 1279\ M\^170 + 615\ M\^172 - 
                169\ M\^174 + 27\ M\^176 - 2\ M\^178)\) + 
          L\^3\ \((84\ M\^130 - 770\ M\^132 + 2473\ M\^134 - 2235\ M\^136 - 
                4980\ M\^138 + 11706\ M\^140 + 525\ M\^142 - 20622\ M\^144 + 
                9914\ M\^146 + 20413\ M\^148 - 14875\ M\^150 - 
                11584\ M\^152 + 12938\ M\^154 + 4153\ M\^156 - 7875\ M\^158 - 
                656\ M\^160 + 3939\ M\^162 - 787\ M\^164 - 1367\ M\^166 + 
                824\ M\^168 + 138\ M\^170 - 374\ M\^172 + 214\ M\^174 - 
                67\ M\^176 + 12\ M\^178 - M\^180)\) + 
          L\^2\ \((36\ M\^144 - 234\ M\^146 + 476\ M\^148 + 14\ M\^150 - 
                1182\ M\^152 + 916\ M\^154 + 1107\ M\^156 - 1230\ M\^158 - 
                149\ M\^160 + 702\ M\^162 - 204\ M\^164 - 110\ M\^166 + 
                38\ M\^168 + 39\ M\^172 - 50\ M\^174 + 28\ M\^176 - 
                8\ M\^178 + M\^180)\)}, {"A-polynomial", Knot[9, 21], 
        M\^24 + L\^21\ M\^110 + 
          L\ \((\(-6\)\ M\^20 + 22\ M\^22 - 27\ M\^24 - 6\ M\^26 + 
                60\ M\^28 - 20\ M\^30 - 15\ M\^32 + 22\ M\^34 - 12\ M\^36 + 
                3\ M\^38)\) + 
          L\^2\ \((15\ M\^16 - 106\ M\^18 + 295\ M\^20 - 264\ M\^22 - 
                417\ M\^24 + 1042\ M\^26 - 307\ M\^28 - 1099\ M\^30 + 
                1203\ M\^32 + 367\ M\^34 - 752\ M\^36 + 198\ M\^38 + 
                27\ M\^40 - 7\ M\^42 + 75\ M\^44 - 109\ M\^46 + 68\ M\^48 - 
                22\ M\^50 + 3\ M\^52)\) + 
          L\^3\ \((\(-20\)\ M\^12 + 204\ M\^14 - 857\ M\^16 + 1586\ M\^18 - 
                142\ M\^20 - 4339\ M\^22 + 5415\ M\^24 + 3393\ M\^26 - 
                11288\ M\^28 + 2146\ M\^30 + 10721\ M\^32 - 4892\ M\^34 - 
                5674\ M\^36 + 3529\ M\^38 + 4944\ M\^40 - 1168\ M\^42 - 
                5381\ M\^44 + 2203\ M\^46 + 3020\ M\^48 - 2517\ M\^50 - 
                13\ M\^52 + 852\ M\^54 - 587\ M\^56 + 287\ M\^58 - 
                128\ M\^60 + 45\ M\^62 - 10\ M\^64 + M\^66)\) + 
          L\^4\ \((15\ M\^8 - 196\ M\^10 + 1076\ M\^12 - 2924\ M\^14 + 
                2909\ M\^16 + 4280\ M\^18 - 13012\ M\^20 + 1846\ M\^22 + 
                25635\ M\^24 - 17248\ M\^26 - 36771\ M\^28 + 41901\ M\^30 + 
                47070\ M\^32 - 87577\ M\^34 - 40037\ M\^36 + 140177\ M\^38 + 
                2220\ M\^40 - 147931\ M\^42 + 55830\ M\^44 + 100958\ M\^46 - 
                77048\ M\^48 - 33013\ M\^50 + 52124\ M\^52 - 5399\ M\^54 - 
                15438\ M\^56 + 7121\ M\^58 - 595\ M\^60 + 34\ M\^62 + 
                642\ M\^64 - 1448\ M\^66 + 1207\ M\^68 - 555\ M\^70 + 
                156\ M\^72 - 26\ M\^74 + 2\ M\^76)\) + 
          L\^5\ \((\(-6\)\ M\^4 + 94\ M\^6 - 627\ M\^8 + 2216\ M\^10 - 
                3968\ M\^12 + 1332\ M\^14 + 7214\ M\^16 - 6205\ M\^18 - 
                18764\ M\^20 + 25673\ M\^22 + 47248\ M\^24 - 105331\ M\^26 - 
                58020\ M\^28 + 287858\ M\^30 - 66150\ M\^32 - 463690\ M\^34 + 
                369208\ M\^36 + 423934\ M\^38 - 716637\ M\^40 - 
                104624\ M\^42 + 875298\ M\^44 - 261094\ M\^46 - 
                712694\ M\^48 + 453103\ M\^50 + 404205\ M\^52 - 
                395995\ M\^54 - 157337\ M\^56 + 236531\ M\^58 + 
                43891\ M\^60 - 112477\ M\^62 - 9593\ M\^64 + 53188\ M\^66 - 
                6861\ M\^68 - 20555\ M\^70 + 10033\ M\^72 + 3356\ M\^74 - 
                5565\ M\^76 + 2874\ M\^78 - 857\ M\^80 + 160\ M\^82 - 
                18\ M\^84 + M\^86)\) + 
          L\^6\ \((1 - 18\ M\^2 + 141\ M\^4 - 628\ M\^6 + 1737\ M\^8 - 
                3010\ M\^10 + 3103\ M\^12 - 2565\ M\^14 + 6566\ M\^16 - 
                11549\ M\^18 - 16673\ M\^20 + 82702\ M\^22 - 39770\ M\^24 - 
                215836\ M\^26 + 293546\ M\^28 + 266072\ M\^30 - 
                770286\ M\^32 - 54006\ M\^34 + 1368935\ M\^36 - 
                538861\ M\^38 - 1771642\ M\^40 + 1390712\ M\^42 + 
                1537611\ M\^44 - 2174474\ M\^46 - 470483\ M\^48 + 
                2506444\ M\^50 - 864056\ M\^52 - 2042808\ M\^54 + 
                1771817\ M\^56 + 1078193\ M\^58 - 1826066\ M\^60 - 
                175904\ M\^62 + 1273749\ M\^64 - 292659\ M\^66 - 
                611227\ M\^68 + 342234\ M\^70 + 171587\ M\^72 - 
                209539\ M\^74 + 13929\ M\^76 + 68982\ M\^78 - 34159\ M\^80 - 
                5652\ M\^82 + 13560\ M\^84 - 7379\ M\^86 + 2291\ M\^88 - 
                448\ M\^90 + 53\ M\^92 - 3\ M\^94)\) + 
          L\^7\ \((\(-1\) + 20\ M\^2 - 170\ M\^4 + 803\ M\^6 - 2301\ M\^8 + 
                4141\ M\^10 - 5303\ M\^12 + 8005\ M\^14 - 13224\ M\^16 - 
                2426\ M\^18 + 63377\ M\^20 - 63931\ M\^22 - 160059\ M\^24 + 
                290364\ M\^26 + 279413\ M\^28 - 772538\ M\^30 - 
                422973\ M\^32 + 1752997\ M\^34 + 230945\ M\^36 - 
                3256298\ M\^38 + 987965\ M\^40 + 4759530\ M\^42 - 
                3610723\ M\^44 - 5157232\ M\^46 + 6668400\ M\^48 + 
                3593438\ M\^50 - 8361157\ M\^52 - 309074\ M\^54 + 
                7824238\ M\^56 - 2832418\ M\^58 - 5305179\ M\^60 + 
                4384949\ M\^62 + 2235189\ M\^64 - 3987872\ M\^66 + 
                50618\ M\^68 + 2475496\ M\^70 - 1013955\ M\^72 - 
                934510\ M\^74 + 910935\ M\^76 + 46177\ M\^78 - 
                432645\ M\^80 + 193767\ M\^82 + 68759\ M\^84 - 
                104259\ M\^86 + 31316\ M\^88 + 16589\ M\^90 - 20455\ M\^92 + 
                10011\ M\^94 - 2950\ M\^96 + 548\ M\^98 - 60\ M\^100 + 
                3\ M\^102)\) + 
          L\^8\ \((6\ M\^4 - 123\ M\^6 + 1074\ M\^8 - 5112\ M\^10 + 
                13596\ M\^12 - 15514\ M\^14 - 13974\ M\^16 + 53844\ M\^18 + 
                19775\ M\^20 - 210554\ M\^22 + 54076\ M\^24 + 680769\ M\^26 - 
                519433\ M\^28 - 1778363\ M\^30 + 2299884\ M\^32 + 
                3256928\ M\^34 - 6374551\ M\^36 - 3963718\ M\^38 + 
                12799911\ M\^40 + 1789006\ M\^42 - 19457097\ M\^44 + 
                4968617\ M\^46 + 22643005\ M\^48 - 15783207\ M\^50 - 
                19192852\ M\^52 + 26549755\ M\^54 + 8978841\ M\^56 - 
                31465075\ M\^58 + 4677775\ M\^60 + 27955543\ M\^62 - 
                15456503\ M\^64 - 17825847\ M\^66 + 19068047\ M\^68 + 
                6444837\ M\^70 - 15578776\ M\^72 + 1291426\ M\^74 + 
                8846522\ M\^76 - 3876103\ M\^78 - 3050828\ M\^80 + 
                2956746\ M\^82 + 139203\ M\^84 - 1223554\ M\^86 + 
                490810\ M\^88 + 172343\ M\^90 - 230238\ M\^92 + 
                68877\ M\^94 + 25008\ M\^96 - 34852\ M\^98 + 18664\ M\^100 - 
                6427\ M\^102 + 1522\ M\^104 - 241\ M\^106 + 23\ M\^108 - 
                M\^110)\) + 
          L\^20\ \((3\ M\^96 - 12\ M\^98 + 22\ M\^100 - 15\ M\^102 - 
                20\ M\^104 + 60\ M\^106 - 6\ M\^108 - 27\ M\^110 + 
                22\ M\^112 - 6\ M\^114)\) + 
          L\^9\ \((\(-15\)\ M\^8 + 315\ M\^10 - 2845\ M\^12 + 14079\ M\^14 - 
                38673\ M\^16 + 41251\ M\^18 + 76353\ M\^20 - 282857\ M\^22 + 
                85566\ M\^24 + 907653\ M\^26 - 1200662\ M\^28 - 
                1667975\ M\^30 + 4553857\ M\^32 + 1376783\ M\^34 - 
                11730899\ M\^36 + 2928793\ M\^38 + 22713136\ M\^40 - 
                14821056\ M\^42 - 33997257\ M\^44 + 35433535\ M\^46 + 
                38232951\ M\^48 - 59626121\ M\^50 - 28435752\ M\^52 + 
                76814845\ M\^54 + 4146514\ M\^56 - 77235077\ M\^58 + 
                25317629\ M\^60 + 59280500\ M\^62 - 45495587\ M\^64 - 
                30103330\ M\^66 + 48212313\ M\^68 + 2820538\ M\^70 - 
                35479509\ M\^72 + 12785887\ M\^74 + 17118122\ M\^76 - 
                15057175\ M\^78 - 2956030\ M\^80 + 9529776\ M\^82 - 
                3135144\ M\^84 - 3182771\ M\^86 + 3166089\ M\^88 - 
                163586\ M\^90 - 1328364\ M\^92 + 764186\ M\^94 + 
                111427\ M\^96 - 320973\ M\^98 + 126553\ M\^100 + 
                29070\ M\^102 - 52051\ M\^104 + 26296\ M\^106 - 
                7587\ M\^108 + 1338\ M\^110 - 135\ M\^112 + 6\ M\^114)\) + 
          L\^10\ \((20\ M\^12 - 430\ M\^14 + 4008\ M\^16 - 20667\ M\^18 + 
                60131\ M\^20 - 72757\ M\^22 - 110531\ M\^24 + 511086\ M\^26 - 
                353095\ M\^28 - 1380503\ M\^30 + 2779538\ M\^32 + 
                1300351\ M\^34 - 8704061\ M\^36 + 3312858\ M\^38 + 
                17977948\ M\^40 - 17501368\ M\^42 - 26441952\ M\^44 + 
                43327853\ M\^46 + 27236993\ M\^48 - 75291675\ M\^50 - 
                15731535\ M\^52 + 99599561\ M\^54 - 4555330\ M\^56 - 
                101499887\ M\^58 + 22266778\ M\^60 + 76038880\ M\^62 - 
                25594626\ M\^64 - 33953695\ M\^66 + 11771608\ M\^68 - 
                3526054\ M\^70 + 12116759\ M\^72 + 21241953\ M\^74 - 
                32193780\ M\^76 - 18194228\ M\^78 + 38524291\ M\^80 + 
                5231249\ M\^82 - 31020879\ M\^84 + 5353914\ M\^86 + 
                17513715\ M\^88 - 8404096\ M\^90 - 6204586\ M\^92 + 
                5963460\ M\^94 + 574052\ M\^96 - 2528477\ M\^98 + 
                760618\ M\^100 + 501221\ M\^102 - 420529\ M\^104 + 
                47942\ M\^106 + 79791\ M\^108 - 53308\ M\^110 + 
                17040\ M\^112 - 3168\ M\^114 + 330\ M\^116 - 15\ M\^118)\) + 
          L\^19\ \((3\ M\^82 - 22\ M\^84 + 68\ M\^86 - 109\ M\^88 + 
                75\ M\^90 - 7\ M\^92 + 27\ M\^94 + 198\ M\^96 - 752\ M\^98 + 
                367\ M\^100 + 1203\ M\^102 - 1099\ M\^104 - 307\ M\^106 + 
                1042\ M\^108 - 417\ M\^110 - 264\ M\^112 + 295\ M\^114 - 
                106\ M\^116 + 15\ M\^118)\) + 
          L\^18\ \((M\^68 - 10\ M\^70 + 45\ M\^72 - 128\ M\^74 + 287\ M\^76 - 
                587\ M\^78 + 852\ M\^80 - 13\ M\^82 - 2517\ M\^84 + 
                3020\ M\^86 + 2203\ M\^88 - 5381\ M\^90 - 1168\ M\^92 + 
                4944\ M\^94 + 3529\ M\^96 - 5674\ M\^98 - 4892\ M\^100 + 
                10721\ M\^102 + 2146\ M\^104 - 11288\ M\^106 + 3393\ M\^108 + 
                5415\ M\^110 - 4339\ M\^112 - 142\ M\^114 + 1586\ M\^116 - 
                857\ M\^118 + 204\ M\^120 - 20\ M\^122)\) + 
          L\^11\ \((\(-15\)\ M\^16 + 330\ M\^18 - 3168\ M\^20 + 
                17040\ M\^22 - 53308\ M\^24 + 79791\ M\^26 + 47942\ M\^28 - 
                420529\ M\^30 + 501221\ M\^32 + 760618\ M\^34 - 
                2528477\ M\^36 + 574052\ M\^38 + 5963460\ M\^40 - 
                6204586\ M\^42 - 8404096\ M\^44 + 17513715\ M\^46 + 
                5353914\ M\^48 - 31020879\ M\^50 + 5231249\ M\^52 + 
                38524291\ M\^54 - 18194228\ M\^56 - 32193780\ M\^58 + 
                21241953\ M\^60 + 12116759\ M\^62 - 3526054\ M\^64 + 
                11771608\ M\^66 - 33953695\ M\^68 - 25594626\ M\^70 + 
                76038880\ M\^72 + 22266778\ M\^74 - 101499887\ M\^76 - 
                4555330\ M\^78 + 99599561\ M\^80 - 15731535\ M\^82 - 
                75291675\ M\^84 + 27236993\ M\^86 + 43327853\ M\^88 - 
                26441952\ M\^90 - 17501368\ M\^92 + 17977948\ M\^94 + 
                3312858\ M\^96 - 8704061\ M\^98 + 1300351\ M\^100 + 
                2779538\ M\^102 - 1380503\ M\^104 - 353095\ M\^106 + 
                511086\ M\^108 - 110531\ M\^110 - 72757\ M\^112 + 
                60131\ M\^114 - 20667\ M\^116 + 4008\ M\^118 - 430\ M\^120 + 
                20\ M\^122)\) + 
          L\^12\ \((6\ M\^20 - 135\ M\^22 + 1338\ M\^24 - 7587\ M\^26 + 
                26296\ M\^28 - 52051\ M\^30 + 29070\ M\^32 + 126553\ M\^34 - 
                320973\ M\^36 + 111427\ M\^38 + 764186\ M\^40 - 
                1328364\ M\^42 - 163586\ M\^44 + 3166089\ M\^46 - 
                3182771\ M\^48 - 3135144\ M\^50 + 9529776\ M\^52 - 
                2956030\ M\^54 - 15057175\ M\^56 + 17118122\ M\^58 + 
                12785887\ M\^60 - 35479509\ M\^62 + 2820538\ M\^64 + 
                48212313\ M\^66 - 30103330\ M\^68 - 45495587\ M\^70 + 
                59280500\ M\^72 + 25317629\ M\^74 - 77235077\ M\^76 + 
                4146514\ M\^78 + 76814845\ M\^80 - 28435752\ M\^82 - 
                59626121\ M\^84 + 38232951\ M\^86 + 35433535\ M\^88 - 
                33997257\ M\^90 - 14821056\ M\^92 + 22713136\ M\^94 + 
                2928793\ M\^96 - 11730899\ M\^98 + 1376783\ M\^100 + 
                4553857\ M\^102 - 1667975\ M\^104 - 1200662\ M\^106 + 
                907653\ M\^108 + 85566\ M\^110 - 282857\ M\^112 + 
                76353\ M\^114 + 41251\ M\^116 - 38673\ M\^118 + 
                14079\ M\^120 - 2845\ M\^122 + 315\ M\^124 - 15\ M\^126)\) + 
          L\^17\ \((2\ M\^58 - 26\ M\^60 + 156\ M\^62 - 555\ M\^64 + 
                1207\ M\^66 - 1448\ M\^68 + 642\ M\^70 + 34\ M\^72 - 
                595\ M\^74 + 7121\ M\^76 - 15438\ M\^78 - 5399\ M\^80 + 
                52124\ M\^82 - 33013\ M\^84 - 77048\ M\^86 + 100958\ M\^88 + 
                55830\ M\^90 - 147931\ M\^92 + 2220\ M\^94 + 140177\ M\^96 - 
                40037\ M\^98 - 87577\ M\^100 + 47070\ M\^102 + 
                41901\ M\^104 - 36771\ M\^106 - 17248\ M\^108 + 
                25635\ M\^110 + 1846\ M\^112 - 13012\ M\^114 + 4280\ M\^116 + 
                2909\ M\^118 - 2924\ M\^120 + 1076\ M\^122 - 196\ M\^124 + 
                15\ M\^126)\) + 
          L\^16\ \((M\^48 - 18\ M\^50 + 160\ M\^52 - 857\ M\^54 + 
                2874\ M\^56 - 5565\ M\^58 + 3356\ M\^60 + 10033\ M\^62 - 
                20555\ M\^64 - 6861\ M\^66 + 53188\ M\^68 - 9593\ M\^70 - 
                112477\ M\^72 + 43891\ M\^74 + 236531\ M\^76 - 
                157337\ M\^78 - 395995\ M\^80 + 404205\ M\^82 + 
                453103\ M\^84 - 712694\ M\^86 - 261094\ M\^88 + 
                875298\ M\^90 - 104624\ M\^92 - 716637\ M\^94 + 
                423934\ M\^96 + 369208\ M\^98 - 463690\ M\^100 - 
                66150\ M\^102 + 287858\ M\^104 - 58020\ M\^106 - 
                105331\ M\^108 + 47248\ M\^110 + 25673\ M\^112 - 
                18764\ M\^114 - 6205\ M\^116 + 7214\ M\^118 + 1332\ M\^120 - 
                3968\ M\^122 + 2216\ M\^124 - 627\ M\^126 + 94\ M\^128 - 
                6\ M\^130)\) + 
          L\^13\ \((\(-M\^24\) + 23\ M\^26 - 241\ M\^28 + 1522\ M\^30 - 
                6427\ M\^32 + 18664\ M\^34 - 34852\ M\^36 + 25008\ M\^38 + 
                68877\ M\^40 - 230238\ M\^42 + 172343\ M\^44 + 
                490810\ M\^46 - 1223554\ M\^48 + 139203\ M\^50 + 
                2956746\ M\^52 - 3050828\ M\^54 - 3876103\ M\^56 + 
                8846522\ M\^58 + 1291426\ M\^60 - 15578776\ M\^62 + 
                6444837\ M\^64 + 19068047\ M\^66 - 17825847\ M\^68 - 
                15456503\ M\^70 + 27955543\ M\^72 + 4677775\ M\^74 - 
                31465075\ M\^76 + 8978841\ M\^78 + 26549755\ M\^80 - 
                19192852\ M\^82 - 15783207\ M\^84 + 22643005\ M\^86 + 
                4968617\ M\^88 - 19457097\ M\^90 + 1789006\ M\^92 + 
                12799911\ M\^94 - 3963718\ M\^96 - 6374551\ M\^98 + 
                3256928\ M\^100 + 2299884\ M\^102 - 1778363\ M\^104 - 
                519433\ M\^106 + 680769\ M\^108 + 54076\ M\^110 - 
                210554\ M\^112 + 19775\ M\^114 + 53844\ M\^116 - 
                13974\ M\^118 - 15514\ M\^120 + 13596\ M\^122 - 
                5112\ M\^124 + 1074\ M\^126 - 123\ M\^128 + 6\ M\^130)\) + 
          L\^14\ \((3\ M\^32 - 60\ M\^34 + 548\ M\^36 - 2950\ M\^38 + 
                10011\ M\^40 - 20455\ M\^42 + 16589\ M\^44 + 31316\ M\^46 - 
                104259\ M\^48 + 68759\ M\^50 + 193767\ M\^52 - 
                432645\ M\^54 + 46177\ M\^56 + 910935\ M\^58 - 
                934510\ M\^60 - 1013955\ M\^62 + 2475496\ M\^64 + 
                50618\ M\^66 - 3987872\ M\^68 + 2235189\ M\^70 + 
                4384949\ M\^72 - 5305179\ M\^74 - 2832418\ M\^76 + 
                7824238\ M\^78 - 309074\ M\^80 - 8361157\ M\^82 + 
                3593438\ M\^84 + 6668400\ M\^86 - 5157232\ M\^88 - 
                3610723\ M\^90 + 4759530\ M\^92 + 987965\ M\^94 - 
                3256298\ M\^96 + 230945\ M\^98 + 1752997\ M\^100 - 
                422973\ M\^102 - 772538\ M\^104 + 279413\ M\^106 + 
                290364\ M\^108 - 160059\ M\^110 - 63931\ M\^112 + 
                63377\ M\^114 - 2426\ M\^116 - 13224\ M\^118 + 8005\ M\^120 - 
                5303\ M\^122 + 4141\ M\^124 - 2301\ M\^126 + 803\ M\^128 - 
                170\ M\^130 + 20\ M\^132 - M\^134)\) + 
          L\^15\ \((\(-3\)\ M\^40 + 53\ M\^42 - 448\ M\^44 + 2291\ M\^46 - 
                7379\ M\^48 + 13560\ M\^50 - 5652\ M\^52 - 34159\ M\^54 + 
                68982\ M\^56 + 13929\ M\^58 - 209539\ M\^60 + 171587\ M\^62 + 
                342234\ M\^64 - 611227\ M\^66 - 292659\ M\^68 + 
                1273749\ M\^70 - 175904\ M\^72 - 1826066\ M\^74 + 
                1078193\ M\^76 + 1771817\ M\^78 - 2042808\ M\^80 - 
                864056\ M\^82 + 2506444\ M\^84 - 470483\ M\^86 - 
                2174474\ M\^88 + 1537611\ M\^90 + 1390712\ M\^92 - 
                1771642\ M\^94 - 538861\ M\^96 + 1368935\ M\^98 - 
                54006\ M\^100 - 770286\ M\^102 + 266072\ M\^104 + 
                293546\ M\^106 - 215836\ M\^108 - 39770\ M\^110 + 
                82702\ M\^112 - 16673\ M\^114 - 11549\ M\^116 + 
                6566\ M\^118 - 2565\ M\^120 + 3103\ M\^122 - 3010\ M\^124 + 
                1737\ M\^126 - 628\ M\^128 + 141\ M\^130 - 18\ M\^132 + 
                M\^134)\)}, {"A-polynomial", Knot[9, 22], 
        M\^28 + L\^25\ M\^134 + 
          L\ \((\(-2\)\ M\^20 + 9\ M\^22 - 24\ M\^24 + 40\ M\^26 - 
                16\ M\^28 - 47\ M\^30 + 68\ M\^32 + 11\ M\^34 - 26\ M\^36 + 
                8\ M\^38)\) + 
          L\^2\ \((M\^12 - 8\ M\^14 + 32\ M\^16 - 82\ M\^18 + 152\ M\^20 - 
                243\ M\^22 + 335\ M\^24 - 126\ M\^26 - 735\ M\^28 + 
                1328\ M\^30 - 1937\ M\^34 + 1319\ M\^36 + 1137\ M\^38 - 
                1283\ M\^40 + 109\ M\^42 + 361\ M\^44 - 180\ M\^46 + 
                28\ M\^48)\) + 
          L\^3\ \((\(-2\)\ M\^8 + 23\ M\^10 - 123\ M\^12 + 399\ M\^14 - 
                828\ M\^16 + 1012\ M\^18 - 490\ M\^20 - 235\ M\^22 + 
                429\ M\^24 - 1487\ M\^26 + 3575\ M\^28 + 178\ M\^30 - 
                11608\ M\^32 + 11946\ M\^34 + 11052\ M\^36 - 26661\ M\^38 + 
                4374\ M\^40 + 25234\ M\^42 - 15909\ M\^44 - 8090\ M\^46 + 
                12496\ M\^48 - 2978\ M\^50 - 2471\ M\^52 + 1933\ M\^54 - 
                537\ M\^56 + 56\ M\^58)\) + 
          L\^4\ \((M\^4 - 17\ M\^6 + 133\ M\^8 - 628\ M\^10 + 1930\ M\^12 - 
                3729\ M\^14 + 3468\ M\^16 + 1944\ M\^18 - 8448\ M\^20 + 
                3683\ M\^22 + 13348\ M\^24 - 20335\ M\^26 + 4366\ M\^28 + 
                11673\ M\^30 - 11744\ M\^32 + 20850\ M\^34 - 37567\ M\^36 - 
                5534\ M\^38 + 107609\ M\^40 - 103034\ M\^42 - 86042\ M\^44 + 
                205199\ M\^46 - 34388\ M\^48 - 157645\ M\^50 + 
                112836\ M\^52 + 33525\ M\^54 - 73821\ M\^56 + 26546\ M\^58 + 
                9521\ M\^60 - 12020\ M\^62 + 4714\ M\^64 - 898\ M\^66 + 
                70\ M\^68)\) + 
          L\^5\ \((2\ M\^2 - 35\ M\^4 + 284\ M\^6 - 1389\ M\^8 + 
                4384\ M\^10 - 8607\ M\^12 + 7975\ M\^14 + 5096\ M\^16 - 
                20341\ M\^18 + 3526\ M\^20 + 53109\ M\^22 - 78424\ M\^24 - 
                6415\ M\^26 + 144019\ M\^28 - 146120\ M\^30 - 88381\ M\^32 + 
                334586\ M\^34 - 184643\ M\^36 - 292295\ M\^38 + 
                426540\ M\^40 - 27221\ M\^42 - 216900\ M\^44 + 
                128228\ M\^46 - 280321\ M\^48 + 329060\ M\^50 + 
                374248\ M\^52 - 797166\ M\^54 + 126996\ M\^56 + 
                619616\ M\^58 - 463796\ M\^60 - 91035\ M\^62 + 
                284777\ M\^64 - 131261\ M\^66 - 20021\ M\^68 + 48491\ M\^70 - 
                24430\ M\^72 + 6433\ M\^74 - 915\ M\^76 + 56\ M\^78)\) + 
          L\^6\ \((1 - 21\ M\^2 + 197\ M\^4 - 1096\ M\^6 + 3994\ M\^8 - 
                9845\ M\^10 + 16180\ M\^12 - 16107\ M\^14 + 4822\ M\^16 + 
                13171\ M\^18 - 21814\ M\^20 - 14260\ M\^22 + 114324\ M\^24 - 
                152073\ M\^26 - 79473\ M\^28 + 361054\ M\^30 - 46003\ M\^32 - 
                622214\ M\^34 + 266403\ M\^36 + 943209\ M\^38 - 
                665902\ M\^40 - 1181652\ M\^42 + 1365891\ M\^44 + 
                790090\ M\^46 - 1775945\ M\^48 + 381259\ M\^50 + 
                696190\ M\^52 - 869770\ M\^54 + 1309474\ M\^56 - 
                456377\ M\^58 - 1756299\ M\^60 + 2139694\ M\^62 + 
                98971\ M\^64 - 1822549\ M\^66 + 1196011\ M\^68 + 
                271340\ M\^70 - 786048\ M\^72 + 398895\ M\^74 + 
                25426\ M\^76 - 137200\ M\^78 + 82810\ M\^80 - 26942\ M\^82 + 
                5236\ M\^84 - 576\ M\^86 + 28\ M\^88)\) + 
          L\^7\ \((\(-1\) + 19\ M\^2 - 175\ M\^4 + 1034\ M\^6 - 4358\ M\^8 + 
                13595\ M\^10 - 30774\ M\^12 + 45432\ M\^14 - 27076\ M\^16 - 
                36847\ M\^18 + 63819\ M\^20 + 65549\ M\^22 - 237827\ M\^24 + 
                118632\ M\^26 + 235667\ M\^28 - 221940\ M\^30 - 
                279766\ M\^32 + 429601\ M\^34 + 194954\ M\^36 - 
                612745\ M\^38 + 73656\ M\^40 + 501976\ M\^42 - 
                412930\ M\^44 + 22119\ M\^46 + 757913\ M\^48 - 
                843005\ M\^50 - 1041384\ M\^52 + 2096735\ M\^54 + 
                595250\ M\^56 - 3977363\ M\^58 + 2199848\ M\^60 + 
                4174879\ M\^62 - 5984919\ M\^64 - 303081\ M\^66 + 
                6241789\ M\^68 - 4323699\ M\^70 - 1731758\ M\^72 + 
                4310191\ M\^74 - 2054069\ M\^76 - 915166\ M\^78 + 
                1671325\ M\^80 - 801407\ M\^82 - 57035\ M\^84 + 
                298097\ M\^86 - 195556\ M\^88 + 72465\ M\^90 - 17076\ M\^92 + 
                2531\ M\^94 - 215\ M\^96 + 8\ M\^98)\) + 
          L\^8\ \((\(-M\^2\) + 23\ M\^4 - 251\ M\^6 + 1695\ M\^8 - 
                7816\ M\^10 + 25704\ M\^12 - 60703\ M\^14 + 97601\ M\^16 - 
                79865\ M\^18 - 55665\ M\^20 + 217384\ M\^22 - 40315\ M\^24 - 
                611340\ M\^26 + 761478\ M\^28 + 594231\ M\^30 - 
                1569362\ M\^32 - 495728\ M\^34 + 2445191\ M\^36 + 
                263577\ M\^38 - 2716786\ M\^40 - 487811\ M\^42 + 
                2501272\ M\^44 + 1398782\ M\^46 - 3262977\ M\^48 - 
                1490473\ M\^50 + 5468623\ M\^52 - 893825\ M\^54 - 
                6929778\ M\^56 + 5576502\ M\^58 + 4753465\ M\^60 - 
                10194397\ M\^62 + 732190\ M\^64 + 12813186\ M\^66 - 
                8036586\ M\^68 - 10653964\ M\^70 + 14405370\ M\^72 + 
                2353531\ M\^74 - 13996581\ M\^76 + 6736215\ M\^78 + 
                5530863\ M\^80 - 7906774\ M\^82 + 2296509\ M\^84 + 
                2409995\ M\^86 - 2807324\ M\^88 + 1077300\ M\^90 + 
                232216\ M\^92 - 528640\ M\^94 + 337512\ M\^96 - 
                131537\ M\^98 + 34458\ M\^100 - 6064\ M\^102 + 678\ M\^104 - 
                42\ M\^106 + M\^108)\) + 
          L\^9\ \((4\ M\^6 - 90\ M\^8 + 943\ M\^10 - 6019\ M\^12 + 
                25728\ M\^14 - 76057\ M\^16 + 153329\ M\^18 - 190139\ M\^20 + 
                69364\ M\^22 + 226206\ M\^24 - 413169\ M\^26 - 74030\ M\^28 + 
                1333130\ M\^30 - 1546574\ M\^32 - 1665696\ M\^34 + 
                4480840\ M\^36 + 1606800\ M\^38 - 9472492\ M\^40 - 
                879640\ M\^42 + 15731138\ M\^44 + 890\ M\^46 - 
                20315751\ M\^48 + 9733\ M\^50 + 19059442\ M\^52 + 
                853574\ M\^54 - 8607788\ M\^56 - 4320907\ M\^58 - 
                7104557\ M\^60 + 13843246\ M\^62 + 17266697\ M\^64 - 
                28715692\ M\^66 - 13599994\ M\^68 + 39532551\ M\^70 - 
                329299\ M\^72 - 38849760\ M\^74 + 15104034\ M\^76 + 
                27686481\ M\^78 - 23909238\ M\^80 - 10420025\ M\^82 + 
                22556768\ M\^84 - 5329553\ M\^86 - 10992940\ M\^88 + 
                10044237\ M\^90 - 837523\ M\^92 - 4324539\ M\^94 + 
                3497345\ M\^96 - 861619\ M\^98 - 611966\ M\^100 + 
                761239\ M\^102 - 435057\ M\^104 + 165100\ M\^106 - 
                44431\ M\^108 + 8446\ M\^110 - 1082\ M\^112 + 84\ M\^114 - 
                3\ M\^116)\) + 
          L\^10\ \((\(-6\)\ M\^10 + 140\ M\^12 - 1518\ M\^14 + 10020\ M\^16 - 
                44244\ M\^18 + 134516\ M\^20 - 275067\ M\^22 + 
                331952\ M\^24 - 81646\ M\^26 - 431191\ M\^28 + 
                611465\ M\^30 + 132926\ M\^32 - 1443817\ M\^34 + 
                1837701\ M\^36 + 635619\ M\^38 - 4807917\ M\^40 + 
                1933057\ M\^42 + 10059940\ M\^44 - 6361612\ M\^46 - 
                19688960\ M\^48 + 10247868\ M\^50 + 36125943\ M\^52 - 
                6937709\ M\^54 - 61464977\ M\^56 - 7617695\ M\^58 + 
                91233746\ M\^60 + 27749632\ M\^62 - 115197666\ M\^64 - 
                35497741\ M\^66 + 123482150\ M\^68 + 16120943\ M\^70 - 
                110444136\ M\^72 + 22566446\ M\^74 + 75229425\ M\^76 - 
                53832076\ M\^78 - 30126120\ M\^80 + 59193186\ M\^82 - 
                5526061\ M\^84 - 43382689\ M\^86 + 23214443\ M\^88 + 
                20498604\ M\^90 - 24945037\ M\^92 - 577785\ M\^94 + 
                14944043\ M\^96 - 8371158\ M\^98 - 2272764\ M\^100 + 
                5408322\ M\^102 - 2879228\ M\^104 + 35165\ M\^106 + 
                1007080\ M\^108 - 833254\ M\^110 + 409689\ M\^112 - 
                142368\ M\^114 + 36421\ M\^116 - 6827\ M\^118 + 899\ M\^120 - 
                75\ M\^122 + 3\ M\^124)\) + 
          L\^11\ \((4\ M\^14 - 99\ M\^16 + 1156\ M\^18 - 8344\ M\^20 + 
                41044\ M\^22 - 142584\ M\^24 + 346915\ M\^26 - 
                546245\ M\^28 + 376027\ M\^30 + 454883\ M\^32 - 
                1301801\ M\^34 + 573889\ M\^36 + 1960928\ M\^38 - 
                3135396\ M\^40 - 518936\ M\^42 + 6303932\ M\^44 - 
                3942332\ M\^46 - 11786556\ M\^48 + 16918160\ M\^50 + 
                20817379\ M\^52 - 47515561\ M\^54 - 32071165\ M\^56 + 
                99264696\ M\^58 + 45065654\ M\^60 - 162541347\ M\^62 - 
                55797325\ M\^64 + 208444229\ M\^66 + 54673735\ M\^68 - 
                203851641\ M\^70 - 31011357\ M\^72 + 141492964\ M\^74 - 
                11603704\ M\^76 - 49543232\ M\^78 + 48259283\ M\^80 - 
                33028758\ M\^82 - 54453428\ M\^84 + 77607765\ M\^86 + 
                31546279\ M\^88 - 78709190\ M\^90 - 3094640\ M\^92 + 
                55166838\ M\^94 - 13535866\ M\^96 - 28897512\ M\^98 + 
                17984847\ M\^100 + 8572318\ M\^102 - 13525093\ M\^104 + 
                2891999\ M\^106 + 4706068\ M\^108 - 4340253\ M\^110 + 
                1094320\ M\^112 + 848567\ M\^114 - 1074811\ M\^116 + 
                643441\ M\^118 - 263168\ M\^120 + 79472\ M\^122 - 
                17991\ M\^124 + 3009\ M\^126 - 356\ M\^128 + 27\ M\^130 - 
                M\^132)\) + 
          L\^12\ \((\(-M\^18\) + 29\ M\^20 - 398\ M\^22 + 3395\ M\^24 - 
                19966\ M\^26 + 84771\ M\^28 - 262834\ M\^30 + 579232\ M\^32 - 
                808479\ M\^34 + 345715\ M\^36 + 1178010\ M\^38 - 
                2480234\ M\^40 + 743258\ M\^42 + 4141222\ M\^44 - 
                5677615\ M\^46 - 2578781\ M\^48 + 12756637\ M\^50 - 
                4345522\ M\^52 - 23111746\ M\^54 + 22777104\ M\^56 + 
                36495502\ M\^58 - 59417213\ M\^60 - 49301882\ M\^62 + 
                106834958\ M\^64 + 64373039\ M\^66 - 141305229\ M\^68 - 
                87285577\ M\^70 + 132037226\ M\^72 + 114012421\ M\^74 - 
                64182042\ M\^76 - 128358410\ M\^78 - 37217197\ M\^80 + 
                119105088\ M\^82 + 119235455\ M\^84 - 96328291\ M\^86 - 
                147736636\ M\^88 + 76880058\ M\^90 + 125484285\ M\^92 - 
                64747803\ M\^94 - 78170598\ M\^96 + 52021562\ M\^98 + 
                34738105\ M\^100 - 35715239\ M\^102 - 9220701\ M\^104 + 
                21300051\ M\^106 - 2401163\ M\^108 - 10206125\ M\^110 + 
                5940929\ M\^112 + 1975964\ M\^114 - 3965007\ M\^116 + 
                1663479\ M\^118 + 493234\ M\^120 - 1042932\ M\^122 + 
                714470\ M\^124 - 317425\ M\^126 + 102961\ M\^128 - 
                25176\ M\^130 + 4602\ M\^132 - 600\ M\^134 + 50\ M\^136 - 
                2\ M\^138)\) + 
          L\^24\ \((8\ M\^124 - 26\ M\^126 + 11\ M\^128 + 68\ M\^130 - 
                47\ M\^132 - 16\ M\^134 + 40\ M\^136 - 24\ M\^138 + 
                9\ M\^140 - 2\ M\^142)\) + 
          L\^13\ \((\(-2\)\ M\^24 + 50\ M\^26 - 600\ M\^28 + 4602\ M\^30 - 
                25176\ M\^32 + 102961\ M\^34 - 317425\ M\^36 + 
                714470\ M\^38 - 1042932\ M\^40 + 493234\ M\^42 + 
                1663479\ M\^44 - 3965007\ M\^46 + 1975964\ M\^48 + 
                5940929\ M\^50 - 10206125\ M\^52 - 2401163\ M\^54 + 
                21300051\ M\^56 - 9220701\ M\^58 - 35715239\ M\^60 + 
                34738105\ M\^62 + 52021562\ M\^64 - 78170598\ M\^66 - 
                64747803\ M\^68 + 125484285\ M\^70 + 76880058\ M\^72 - 
                147736636\ M\^74 - 96328291\ M\^76 + 119235455\ M\^78 + 
                119105088\ M\^80 - 37217197\ M\^82 - 128358410\ M\^84 - 
                64182042\ M\^86 + 114012421\ M\^88 + 132037226\ M\^90 - 
                87285577\ M\^92 - 141305229\ M\^94 + 64373039\ M\^96 + 
                106834958\ M\^98 - 49301882\ M\^100 - 59417213\ M\^102 + 
                36495502\ M\^104 + 22777104\ M\^106 - 23111746\ M\^108 - 
                4345522\ M\^110 + 12756637\ M\^112 - 2578781\ M\^114 - 
                5677615\ M\^116 + 4141222\ M\^118 + 743258\ M\^120 - 
                2480234\ M\^122 + 1178010\ M\^124 + 345715\ M\^126 - 
                808479\ M\^128 + 579232\ M\^130 - 262834\ M\^132 + 
                84771\ M\^134 - 19966\ M\^136 + 3395\ M\^138 - 398\ M\^140 + 
                29\ M\^142 - M\^144)\) + 
          L\^14\ \((\(-M\^30\) + 27\ M\^32 - 356\ M\^34 + 3009\ M\^36 - 
                17991\ M\^38 + 79472\ M\^40 - 263168\ M\^42 + 643441\ M\^44 - 
                1074811\ M\^46 + 848567\ M\^48 + 1094320\ M\^50 - 
                4340253\ M\^52 + 4706068\ M\^54 + 2891999\ M\^56 - 
                13525093\ M\^58 + 8572318\ M\^60 + 17984847\ M\^62 - 
                28897512\ M\^64 - 13535866\ M\^66 + 55166838\ M\^68 - 
                3094640\ M\^70 - 78709190\ M\^72 + 31546279\ M\^74 + 
                77607765\ M\^76 - 54453428\ M\^78 - 33028758\ M\^80 + 
                48259283\ M\^82 - 49543232\ M\^84 - 11603704\ M\^86 + 
                141492964\ M\^88 - 31011357\ M\^90 - 203851641\ M\^92 + 
                54673735\ M\^94 + 208444229\ M\^96 - 55797325\ M\^98 - 
                162541347\ M\^100 + 45065654\ M\^102 + 99264696\ M\^104 - 
                32071165\ M\^106 - 47515561\ M\^108 + 20817379\ M\^110 + 
                16918160\ M\^112 - 11786556\ M\^114 - 3942332\ M\^116 + 
                6303932\ M\^118 - 518936\ M\^120 - 3135396\ M\^122 + 
                1960928\ M\^124 + 573889\ M\^126 - 1301801\ M\^128 + 
                454883\ M\^130 + 376027\ M\^132 - 546245\ M\^134 + 
                346915\ M\^136 - 142584\ M\^138 + 41044\ M\^140 - 
                8344\ M\^142 + 1156\ M\^144 - 99\ M\^146 + 4\ M\^148)\) + 
          L\^23\ \((28\ M\^114 - 180\ M\^116 + 361\ M\^118 + 109\ M\^120 - 
                1283\ M\^122 + 1137\ M\^124 + 1319\ M\^126 - 1937\ M\^128 + 
                1328\ M\^132 - 735\ M\^134 - 126\ M\^136 + 335\ M\^138 - 
                243\ M\^140 + 152\ M\^142 - 82\ M\^144 + 32\ M\^146 - 
                8\ M\^148 + M\^150)\) + 
          L\^15\ \((3\ M\^38 - 75\ M\^40 + 899\ M\^42 - 6827\ M\^44 + 
                36421\ M\^46 - 142368\ M\^48 + 409689\ M\^50 - 
                833254\ M\^52 + 1007080\ M\^54 + 35165\ M\^56 - 
                2879228\ M\^58 + 5408322\ M\^60 - 2272764\ M\^62 - 
                8371158\ M\^64 + 14944043\ M\^66 - 577785\ M\^68 - 
                24945037\ M\^70 + 20498604\ M\^72 + 23214443\ M\^74 - 
                43382689\ M\^76 - 5526061\ M\^78 + 59193186\ M\^80 - 
                30126120\ M\^82 - 53832076\ M\^84 + 75229425\ M\^86 + 
                22566446\ M\^88 - 110444136\ M\^90 + 16120943\ M\^92 + 
                123482150\ M\^94 - 35497741\ M\^96 - 115197666\ M\^98 + 
                27749632\ M\^100 + 91233746\ M\^102 - 7617695\ M\^104 - 
                61464977\ M\^106 - 6937709\ M\^108 + 36125943\ M\^110 + 
                10247868\ M\^112 - 19688960\ M\^114 - 6361612\ M\^116 + 
                10059940\ M\^118 + 1933057\ M\^120 - 4807917\ M\^122 + 
                635619\ M\^124 + 1837701\ M\^126 - 1443817\ M\^128 + 
                132926\ M\^130 + 611465\ M\^132 - 431191\ M\^134 - 
                81646\ M\^136 + 331952\ M\^138 - 275067\ M\^140 + 
                134516\ M\^142 - 44244\ M\^144 + 10020\ M\^146 - 
                1518\ M\^148 + 140\ M\^150 - 6\ M\^152)\) + 
          L\^22\ \((56\ M\^104 - 537\ M\^106 + 1933\ M\^108 - 2471\ M\^110 - 
                2978\ M\^112 + 12496\ M\^114 - 8090\ M\^116 - 15909\ M\^118 + 
                25234\ M\^120 + 4374\ M\^122 - 26661\ M\^124 + 
                11052\ M\^126 + 11946\ M\^128 - 11608\ M\^130 + 178\ M\^132 + 
                3575\ M\^134 - 1487\ M\^136 + 429\ M\^138 - 235\ M\^140 - 
                490\ M\^142 + 1012\ M\^144 - 828\ M\^146 + 399\ M\^148 - 
                123\ M\^150 + 23\ M\^152 - 2\ M\^154)\) + 
          L\^16\ \((\(-3\)\ M\^46 + 84\ M\^48 - 1082\ M\^50 + 8446\ M\^52 - 
                44431\ M\^54 + 165100\ M\^56 - 435057\ M\^58 + 
                761239\ M\^60 - 611966\ M\^62 - 861619\ M\^64 + 
                3497345\ M\^66 - 4324539\ M\^68 - 837523\ M\^70 + 
                10044237\ M\^72 - 10992940\ M\^74 - 5329553\ M\^76 + 
                22556768\ M\^78 - 10420025\ M\^80 - 23909238\ M\^82 + 
                27686481\ M\^84 + 15104034\ M\^86 - 38849760\ M\^88 - 
                329299\ M\^90 + 39532551\ M\^92 - 13599994\ M\^94 - 
                28715692\ M\^96 + 17266697\ M\^98 + 13843246\ M\^100 - 
                7104557\ M\^102 - 4320907\ M\^104 - 8607788\ M\^106 + 
                853574\ M\^108 + 19059442\ M\^110 + 9733\ M\^112 - 
                20315751\ M\^114 + 890\ M\^116 + 15731138\ M\^118 - 
                879640\ M\^120 - 9472492\ M\^122 + 1606800\ M\^124 + 
                4480840\ M\^126 - 1665696\ M\^128 - 1546574\ M\^130 + 
                1333130\ M\^132 - 74030\ M\^134 - 413169\ M\^136 + 
                226206\ M\^138 + 69364\ M\^140 - 190139\ M\^142 + 
                153329\ M\^144 - 76057\ M\^146 + 25728\ M\^148 - 
                6019\ M\^150 + 943\ M\^152 - 90\ M\^154 + 4\ M\^156)\) + 
          L\^21\ \((70\ M\^94 - 898\ M\^96 + 4714\ M\^98 - 12020\ M\^100 + 
                9521\ M\^102 + 26546\ M\^104 - 73821\ M\^106 + 
                33525\ M\^108 + 112836\ M\^110 - 157645\ M\^112 - 
                34388\ M\^114 + 205199\ M\^116 - 86042\ M\^118 - 
                103034\ M\^120 + 107609\ M\^122 - 5534\ M\^124 - 
                37567\ M\^126 + 20850\ M\^128 - 11744\ M\^130 + 
                11673\ M\^132 + 4366\ M\^134 - 20335\ M\^136 + 
                13348\ M\^138 + 3683\ M\^140 - 8448\ M\^142 + 1944\ M\^144 + 
                3468\ M\^146 - 3729\ M\^148 + 1930\ M\^150 - 628\ M\^152 + 
                133\ M\^154 - 17\ M\^156 + M\^158)\) + 
          L\^17\ \((M\^54 - 42\ M\^56 + 678\ M\^58 - 6064\ M\^60 + 
                34458\ M\^62 - 131537\ M\^64 + 337512\ M\^66 - 
                528640\ M\^68 + 232216\ M\^70 + 1077300\ M\^72 - 
                2807324\ M\^74 + 2409995\ M\^76 + 2296509\ M\^78 - 
                7906774\ M\^80 + 5530863\ M\^82 + 6736215\ M\^84 - 
                13996581\ M\^86 + 2353531\ M\^88 + 14405370\ M\^90 - 
                10653964\ M\^92 - 8036586\ M\^94 + 12813186\ M\^96 + 
                732190\ M\^98 - 10194397\ M\^100 + 4753465\ M\^102 + 
                5576502\ M\^104 - 6929778\ M\^106 - 893825\ M\^108 + 
                5468623\ M\^110 - 1490473\ M\^112 - 3262977\ M\^114 + 
                1398782\ M\^116 + 2501272\ M\^118 - 487811\ M\^120 - 
                2716786\ M\^122 + 263577\ M\^124 + 2445191\ M\^126 - 
                495728\ M\^128 - 1569362\ M\^130 + 594231\ M\^132 + 
                761478\ M\^134 - 611340\ M\^136 - 40315\ M\^138 + 
                217384\ M\^140 - 55665\ M\^142 - 79865\ M\^144 + 
                97601\ M\^146 - 60703\ M\^148 + 25704\ M\^150 - 
                7816\ M\^152 + 1695\ M\^154 - 251\ M\^156 + 23\ M\^158 - 
                M\^160)\) + 
          L\^20\ \((56\ M\^84 - 915\ M\^86 + 6433\ M\^88 - 24430\ M\^90 + 
                48491\ M\^92 - 20021\ M\^94 - 131261\ M\^96 + 284777\ M\^98 - 
                91035\ M\^100 - 463796\ M\^102 + 619616\ M\^104 + 
                126996\ M\^106 - 797166\ M\^108 + 374248\ M\^110 + 
                329060\ M\^112 - 280321\ M\^114 + 128228\ M\^116 - 
                216900\ M\^118 - 27221\ M\^120 + 426540\ M\^122 - 
                292295\ M\^124 - 184643\ M\^126 + 334586\ M\^128 - 
                88381\ M\^130 - 146120\ M\^132 + 144019\ M\^134 - 
                6415\ M\^136 - 78424\ M\^138 + 53109\ M\^140 + 3526\ M\^142 - 
                20341\ M\^144 + 5096\ M\^146 + 7975\ M\^148 - 8607\ M\^150 + 
                4384\ M\^152 - 1389\ M\^154 + 284\ M\^156 - 35\ M\^158 + 
                2\ M\^160)\) + 
          L\^18\ \((8\ M\^64 - 215\ M\^66 + 2531\ M\^68 - 17076\ M\^70 + 
                72465\ M\^72 - 195556\ M\^74 + 298097\ M\^76 - 57035\ M\^78 - 
                801407\ M\^80 + 1671325\ M\^82 - 915166\ M\^84 - 
                2054069\ M\^86 + 4310191\ M\^88 - 1731758\ M\^90 - 
                4323699\ M\^92 + 6241789\ M\^94 - 303081\ M\^96 - 
                5984919\ M\^98 + 4174879\ M\^100 + 2199848\ M\^102 - 
                3977363\ M\^104 + 595250\ M\^106 + 2096735\ M\^108 - 
                1041384\ M\^110 - 843005\ M\^112 + 757913\ M\^114 + 
                22119\ M\^116 - 412930\ M\^118 + 501976\ M\^120 + 
                73656\ M\^122 - 612745\ M\^124 + 194954\ M\^126 + 
                429601\ M\^128 - 279766\ M\^130 - 221940\ M\^132 + 
                235667\ M\^134 + 118632\ M\^136 - 237827\ M\^138 + 
                65549\ M\^140 + 63819\ M\^142 - 36847\ M\^144 - 
                27076\ M\^146 + 45432\ M\^148 - 30774\ M\^150 + 
                13595\ M\^152 - 4358\ M\^154 + 1034\ M\^156 - 175\ M\^158 + 
                19\ M\^160 - M\^162)\) + 
          L\^19\ \((28\ M\^74 - 576\ M\^76 + 5236\ M\^78 - 26942\ M\^80 + 
                82810\ M\^82 - 137200\ M\^84 + 25426\ M\^86 + 398895\ M\^88 - 
                786048\ M\^90 + 271340\ M\^92 + 1196011\ M\^94 - 
                1822549\ M\^96 + 98971\ M\^98 + 2139694\ M\^100 - 
                1756299\ M\^102 - 456377\ M\^104 + 1309474\ M\^106 - 
                869770\ M\^108 + 696190\ M\^110 + 381259\ M\^112 - 
                1775945\ M\^114 + 790090\ M\^116 + 1365891\ M\^118 - 
                1181652\ M\^120 - 665902\ M\^122 + 943209\ M\^124 + 
                266403\ M\^126 - 622214\ M\^128 - 46003\ M\^130 + 
                361054\ M\^132 - 79473\ M\^134 - 152073\ M\^136 + 
                114324\ M\^138 - 14260\ M\^140 - 21814\ M\^142 + 
                13171\ M\^144 + 4822\ M\^146 - 16107\ M\^148 + 
                16180\ M\^150 - 9845\ M\^152 + 3994\ M\^154 - 1096\ M\^156 + 
                197\ M\^158 - 21\ M\^160 + M\^162)\)}, {"A-polynomial", 
        Knot[9, 23], 
        L\^14 + M\^116 + 
          L\^13\ \((\(-3\) + 17\ M\^2 - 32\ M\^4 + 3\ M\^6 + 61\ M\^8 - 
                34\ M\^10 - 14\ M\^12 + 25\ M\^14 - 12\ M\^16 + 3\ M\^18)\) + 
          L\^12\ \((3 - 30\ M\^2 + 110\ M\^4 - 153\ M\^6 - 26\ M\^8 + 
                276\ M\^10 - 231\ M\^12 - 18\ M\^14 + 140\ M\^16 - 
                49\ M\^18 + 295\ M\^20 - 192\ M\^22 - 288\ M\^24 + 
                327\ M\^26 + 2\ M\^28 - 137\ M\^30 + 83\ M\^32 - 24\ M\^34 + 
                3\ M\^36)\) + 
          L\^11\ \((\(-1\) + 13\ M\^2 - 69\ M\^4 + 202\ M\^6 - 399\ M\^8 + 
                524\ M\^10 + 150\ M\^12 - 2233\ M\^14 + 2632\ M\^16 + 
                2673\ M\^18 - 6889\ M\^20 - 324\ M\^22 + 8666\ M\^24 - 
                2357\ M\^26 - 5504\ M\^28 + 2885\ M\^30 + 1380\ M\^32 - 
                1251\ M\^34 + 285\ M\^36 + 135\ M\^38 - 182\ M\^40 - 
                69\ M\^42 + 79\ M\^44 + 106\ M\^46 - 136\ M\^48 + 59\ M\^50 - 
                12\ M\^52 + M\^54)\) + 
          L\^10\ \((3\ M\^4 - 45\ M\^6 + 270\ M\^8 - 782\ M\^10 + 
                876\ M\^12 + 752\ M\^14 - 2536\ M\^16 + 481\ M\^18 + 
                2081\ M\^20 + 357\ M\^22 + 1510\ M\^24 - 7235\ M\^26 - 
                4694\ M\^28 + 16147\ M\^30 + 3855\ M\^32 - 16796\ M\^34 + 
                1054\ M\^36 + 9501\ M\^38 - 3760\ M\^40 - 960\ M\^42 + 
                2872\ M\^44 - 2577\ M\^46 - 517\ M\^48 + 1745\ M\^50 - 
                361\ M\^52 - 336\ M\^54 + 53\ M\^56 - 52\ M\^58 + 
                224\ M\^60 - 187\ M\^62 + 70\ M\^64 - 13\ M\^66 + M\^68)\) + 
          L\^9\ \((\(-3\)\ M\^8 + 49\ M\^10 - 327\ M\^12 + 1096\ M\^14 - 
                1599\ M\^16 - 810\ M\^18 + 5745\ M\^20 - 2888\ M\^22 - 
                12102\ M\^24 + 11627\ M\^26 + 19856\ M\^28 - 24497\ M\^30 - 
                26531\ M\^32 + 39932\ M\^34 + 21960\ M\^36 - 53203\ M\^38 + 
                1957\ M\^40 + 57125\ M\^42 - 30850\ M\^44 - 43236\ M\^46 + 
                46874\ M\^48 + 20189\ M\^50 - 41780\ M\^52 - 629\ M\^54 + 
                25679\ M\^56 - 7479\ M\^58 - 10972\ M\^60 + 6968\ M\^62 + 
                2462\ M\^64 - 3471\ M\^66 + 328\ M\^68 + 1031\ M\^70 - 
                667\ M\^72 + 196\ M\^74 - 30\ M\^76 + 2\ M\^78)\) + 
          L\^8\ \((M\^12 - 17\ M\^14 + 126\ M\^16 - 524\ M\^18 + 
                1245\ M\^20 - 1273\ M\^22 - 1093\ M\^24 + 4085\ M\^26 - 
                1827\ M\^28 - 4196\ M\^30 + 4653\ M\^32 - 3851\ M\^34 + 
                3653\ M\^36 + 17747\ M\^38 - 26930\ M\^40 - 33316\ M\^42 + 
                54451\ M\^44 + 46853\ M\^46 - 73567\ M\^48 - 47204\ M\^50 + 
                87241\ M\^52 + 29745\ M\^54 - 90340\ M\^56 + 3594\ M\^58 + 
                76837\ M\^60 - 30883\ M\^62 - 45925\ M\^64 + 36522\ M\^66 + 
                14654\ M\^68 - 24386\ M\^70 + 2014\ M\^72 + 9059\ M\^74 - 
                4118\ M\^76 - 1195\ M\^78 + 1765\ M\^80 - 741\ M\^82 + 
                162\ M\^84 - 19\ M\^86 + M\^88)\) + 
          L\^7\ \((\(-2\)\ M\^20 + 35\ M\^22 - 260\ M\^24 + 1027\ M\^26 - 
                2034\ M\^28 + 554\ M\^30 + 5628\ M\^32 - 7781\ M\^34 - 
                7393\ M\^36 + 22265\ M\^38 + 790\ M\^40 - 36162\ M\^42 + 
                12355\ M\^44 + 42897\ M\^46 - 23790\ M\^48 - 44637\ M\^50 + 
                23963\ M\^52 + 47557\ M\^54 - 9257\ M\^56 - 48078\ M\^58 - 
                9257\ M\^60 + 47557\ M\^62 + 23963\ M\^64 - 44637\ M\^66 - 
                23790\ M\^68 + 42897\ M\^70 + 12355\ M\^72 - 36162\ M\^74 + 
                790\ M\^76 + 22265\ M\^78 - 7393\ M\^80 - 7781\ M\^82 + 
                5628\ M\^84 + 554\ M\^86 - 2034\ M\^88 + 1027\ M\^90 - 
                260\ M\^92 + 35\ M\^94 - 2\ M\^96)\) + 
          L\^6\ \((M\^28 - 19\ M\^30 + 162\ M\^32 - 741\ M\^34 + 
                1765\ M\^36 - 1195\ M\^38 - 4118\ M\^40 + 9059\ M\^42 + 
                2014\ M\^44 - 24386\ M\^46 + 14654\ M\^48 + 36522\ M\^50 - 
                45925\ M\^52 - 30883\ M\^54 + 76837\ M\^56 + 3594\ M\^58 - 
                90340\ M\^60 + 29745\ M\^62 + 87241\ M\^64 - 47204\ M\^66 - 
                73567\ M\^68 + 46853\ M\^70 + 54451\ M\^72 - 33316\ M\^74 - 
                26930\ M\^76 + 17747\ M\^78 + 3653\ M\^80 - 3851\ M\^82 + 
                4653\ M\^84 - 4196\ M\^86 - 1827\ M\^88 + 4085\ M\^90 - 
                1093\ M\^92 - 1273\ M\^94 + 1245\ M\^96 - 524\ M\^98 + 
                126\ M\^100 - 17\ M\^102 + M\^104)\) + 
          L\^5\ \((2\ M\^38 - 30\ M\^40 + 196\ M\^42 - 667\ M\^44 + 
                1031\ M\^46 + 328\ M\^48 - 3471\ M\^50 + 2462\ M\^52 + 
                6968\ M\^54 - 10972\ M\^56 - 7479\ M\^58 + 25679\ M\^60 - 
                629\ M\^62 - 41780\ M\^64 + 20189\ M\^66 + 46874\ M\^68 - 
                43236\ M\^70 - 30850\ M\^72 + 57125\ M\^74 + 1957\ M\^76 - 
                53203\ M\^78 + 21960\ M\^80 + 39932\ M\^82 - 26531\ M\^84 - 
                24497\ M\^86 + 19856\ M\^88 + 11627\ M\^90 - 12102\ M\^92 - 
                2888\ M\^94 + 5745\ M\^96 - 810\ M\^98 - 1599\ M\^100 + 
                1096\ M\^102 - 327\ M\^104 + 49\ M\^106 - 3\ M\^108)\) + 
          L\^4\ \((M\^48 - 13\ M\^50 + 70\ M\^52 - 187\ M\^54 + 224\ M\^56 - 
                52\ M\^58 + 53\ M\^60 - 336\ M\^62 - 361\ M\^64 + 
                1745\ M\^66 - 517\ M\^68 - 2577\ M\^70 + 2872\ M\^72 - 
                960\ M\^74 - 3760\ M\^76 + 9501\ M\^78 + 1054\ M\^80 - 
                16796\ M\^82 + 3855\ M\^84 + 16147\ M\^86 - 4694\ M\^88 - 
                7235\ M\^90 + 1510\ M\^92 + 357\ M\^94 + 2081\ M\^96 + 
                481\ M\^98 - 2536\ M\^100 + 752\ M\^102 + 876\ M\^104 - 
                782\ M\^106 + 270\ M\^108 - 45\ M\^110 + 3\ M\^112)\) + 
          L\ \((3\ M\^98 - 12\ M\^100 + 25\ M\^102 - 14\ M\^104 - 
                34\ M\^106 + 61\ M\^108 + 3\ M\^110 - 32\ M\^112 + 
                17\ M\^114 - 3\ M\^116)\) + 
          L\^3\ \((M\^62 - 12\ M\^64 + 59\ M\^66 - 136\ M\^68 + 106\ M\^70 + 
                79\ M\^72 - 69\ M\^74 - 182\ M\^76 + 135\ M\^78 + 
                285\ M\^80 - 1251\ M\^82 + 1380\ M\^84 + 2885\ M\^86 - 
                5504\ M\^88 - 2357\ M\^90 + 8666\ M\^92 - 324\ M\^94 - 
                6889\ M\^96 + 2673\ M\^98 + 2632\ M\^100 - 2233\ M\^102 + 
                150\ M\^104 + 524\ M\^106 - 399\ M\^108 + 202\ M\^110 - 
                69\ M\^112 + 13\ M\^114 - M\^116)\) + 
          L\^2\ \((3\ M\^80 - 24\ M\^82 + 83\ M\^84 - 137\ M\^86 + 2\ M\^88 + 
                327\ M\^90 - 288\ M\^92 - 192\ M\^94 + 295\ M\^96 - 
                49\ M\^98 + 140\ M\^100 - 18\ M\^102 - 231\ M\^104 + 
                276\ M\^106 - 26\ M\^108 - 153\ M\^110 + 110\ M\^112 - 
                30\ M\^114 + 3\ M\^116)\)}, {"A-polynomial", Knot[9, 24], 
        M\^54 + L\^19\ M\^60 + 
          L\ \((3\ M\^44 - 13\ M\^46 + 33\ M\^48 - 33\ M\^50 - 24\ M\^52 + 
                78\ M\^54 - 15\ M\^56 - 25\ M\^58 + 20\ M\^60 - 5\ M\^62)\) + 
          L\^18\ \((\(-5\)\ M\^52 + 20\ M\^54 - 25\ M\^56 - 15\ M\^58 + 
                78\ M\^60 - 24\ M\^62 - 33\ M\^64 + 33\ M\^66 - 13\ M\^68 + 
                3\ M\^70)\) + 
          L\^2\ \((3\ M\^34 - 24\ M\^36 + 91\ M\^38 - 199\ M\^40 + 
                245\ M\^42 - 101\ M\^44 - 289\ M\^46 + 897\ M\^48 - 
                842\ M\^50 - 674\ M\^52 + 1835\ M\^54 - 543\ M\^56 - 
                907\ M\^58 + 1030\ M\^60 - 252\ M\^62 - 273\ M\^64 + 
                243\ M\^66 - 79\ M\^68 + 10\ M\^70)\) + 
          L\^3\ \((M\^24 - 11\ M\^26 + 59\ M\^28 - 203\ M\^30 + 497\ M\^32 - 
                935\ M\^34 + 1234\ M\^36 - 422\ M\^38 - 2535\ M\^40 + 
                6387\ M\^42 - 5619\ M\^44 - 5847\ M\^46 + 19995\ M\^48 - 
                11757\ M\^50 - 18196\ M\^52 + 27784\ M\^54 - 2217\ M\^56 - 
                18376\ M\^58 + 15158\ M\^60 - 667\ M\^62 - 7947\ M\^64 + 
                6383\ M\^66 - 1259\ M\^68 - 1487\ M\^70 + 1410\ M\^72 - 
                567\ M\^74 + 116\ M\^76 - 10\ M\^78)\) + 
          L\^17\ \((10\ M\^44 - 79\ M\^46 + 243\ M\^48 - 273\ M\^50 - 
                252\ M\^52 + 1030\ M\^54 - 907\ M\^56 - 543\ M\^58 + 
                1835\ M\^60 - 674\ M\^62 - 842\ M\^64 + 897\ M\^66 - 
                289\ M\^68 - 101\ M\^70 + 245\ M\^72 - 199\ M\^74 + 
                91\ M\^76 - 24\ M\^78 + 3\ M\^80)\) + 
          L\^4\ \((2\ M\^18 - 28\ M\^20 + 186\ M\^22 - 755\ M\^24 + 
                1969\ M\^26 - 3132\ M\^28 + 2018\ M\^30 + 3102\ M\^32 - 
                9493\ M\^34 + 11065\ M\^36 - 5142\ M\^38 - 12302\ M\^40 + 
                46515\ M\^42 - 55362\ M\^44 - 34299\ M\^46 + 147946\ M\^48 - 
                82077\ M\^50 - 124349\ M\^52 + 186668\ M\^54 - 30947\ M\^56 - 
                132273\ M\^58 + 146652\ M\^60 - 13022\ M\^62 - 
                101277\ M\^64 + 77765\ M\^66 + 4801\ M\^68 - 39260\ M\^70 + 
                23857\ M\^72 - 2320\ M\^74 - 5823\ M\^76 + 4758\ M\^78 - 
                2012\ M\^80 + 516\ M\^82 - 76\ M\^84 + 5\ M\^86)\) + 
          L\^16\ \((\(-10\)\ M\^36 + 116\ M\^38 - 567\ M\^40 + 1410\ M\^42 - 
                1487\ M\^44 - 1259\ M\^46 + 6383\ M\^48 - 7947\ M\^50 - 
                667\ M\^52 + 15158\ M\^54 - 18376\ M\^56 - 2217\ M\^58 + 
                27784\ M\^60 - 18196\ M\^62 - 11757\ M\^64 + 19995\ M\^66 - 
                5847\ M\^68 - 5619\ M\^70 + 6387\ M\^72 - 2535\ M\^74 - 
                422\ M\^76 + 1234\ M\^78 - 935\ M\^80 + 497\ M\^82 - 
                203\ M\^84 + 59\ M\^86 - 11\ M\^88 + M\^90)\) + 
          L\^5\ \((M\^12 - 19\ M\^14 + 178\ M\^16 - 1015\ M\^18 + 
                3694\ M\^20 - 8162\ M\^22 + 7635\ M\^24 + 11406\ M\^26 - 
                43599\ M\^28 + 32973\ M\^30 + 67477\ M\^32 - 147093\ M\^34 + 
                5587\ M\^36 + 247428\ M\^38 - 165305\ M\^40 - 270074\ M\^42 + 
                331996\ M\^44 + 281141\ M\^46 - 543051\ M\^48 - 
                317374\ M\^50 + 926222\ M\^52 + 186813\ M\^54 - 
                1321177\ M\^56 + 270071\ M\^58 + 1288151\ M\^60 - 
                733456\ M\^62 - 695881\ M\^64 + 775723\ M\^66 + 
                38320\ M\^68 - 402955\ M\^70 + 200725\ M\^72 + 46685\ M\^74 - 
                105368\ M\^76 + 52111\ M\^78 - 460\ M\^80 - 15048\ M\^82 + 
                10608\ M\^84 - 4205\ M\^86 + 1095\ M\^88 - 189\ M\^90 + 
                20\ M\^92 - M\^94)\) + 
          L\^15\ \((5\ M\^28 - 76\ M\^30 + 516\ M\^32 - 2012\ M\^34 + 
                4758\ M\^36 - 5823\ M\^38 - 2320\ M\^40 + 23857\ M\^42 - 
                39260\ M\^44 + 4801\ M\^46 + 77765\ M\^48 - 101277\ M\^50 - 
                13022\ M\^52 + 146652\ M\^54 - 132273\ M\^56 - 30947\ M\^58 + 
                186668\ M\^60 - 124349\ M\^62 - 82077\ M\^64 + 
                147946\ M\^66 - 34299\ M\^68 - 55362\ M\^70 + 46515\ M\^72 - 
                12302\ M\^74 - 5142\ M\^76 + 11065\ M\^78 - 9493\ M\^80 + 
                3102\ M\^82 + 2018\ M\^84 - 3132\ M\^86 + 1969\ M\^88 - 
                755\ M\^90 + 186\ M\^92 - 28\ M\^94 + 2\ M\^96)\) + 
          L\^6\ \((\(-3\)\ M\^8 + 56\ M\^10 - 488\ M\^12 + 2522\ M\^14 - 
                8112\ M\^16 + 14786\ M\^18 - 5515\ M\^20 - 41612\ M\^22 + 
                91245\ M\^24 - 10726\ M\^26 - 245488\ M\^28 + 329170\ M\^30 + 
                213227\ M\^32 - 906741\ M\^34 + 353261\ M\^36 + 
                1398486\ M\^38 - 1517702\ M\^40 - 1384323\ M\^42 + 
                2879991\ M\^44 + 848225\ M\^46 - 3972796\ M\^48 - 
                49986\ M\^50 + 4512594\ M\^52 - 894885\ M\^54 - 
                4159096\ M\^56 + 1877235\ M\^58 + 2712789\ M\^60 - 
                2361986\ M\^62 - 601374\ M\^64 + 1881180\ M\^66 - 
                948938\ M\^68 - 733330\ M\^70 + 1214651\ M\^72 - 
                201354\ M\^74 - 623707\ M\^76 + 412621\ M\^78 + 
                72895\ M\^80 - 192791\ M\^82 + 77022\ M\^84 + 12136\ M\^86 - 
                26750\ M\^88 + 14815\ M\^90 - 5124\ M\^92 + 1237\ M\^94 - 
                205\ M\^96 + 21\ M\^98 - M\^100)\) + 
          L\^14\ \((\(-M\^20\) + 20\ M\^22 - 189\ M\^24 + 1095\ M\^26 - 
                4205\ M\^28 + 10608\ M\^30 - 15048\ M\^32 - 460\ M\^34 + 
                52111\ M\^36 - 105368\ M\^38 + 46685\ M\^40 + 200725\ M\^42 - 
                402955\ M\^44 + 38320\ M\^46 + 775723\ M\^48 - 
                695881\ M\^50 - 733456\ M\^52 + 1288151\ M\^54 + 
                270071\ M\^56 - 1321177\ M\^58 + 186813\ M\^60 + 
                926222\ M\^62 - 317374\ M\^64 - 543051\ M\^66 + 
                281141\ M\^68 + 331996\ M\^70 - 270074\ M\^72 - 
                165305\ M\^74 + 247428\ M\^76 + 5587\ M\^78 - 147093\ M\^80 + 
                67477\ M\^82 + 32973\ M\^84 - 43599\ M\^86 + 11406\ M\^88 + 
                7635\ M\^90 - 8162\ M\^92 + 3694\ M\^94 - 1015\ M\^96 + 
                178\ M\^98 - 19\ M\^100 + M\^102)\) + 
          L\^7\ \((3\ M\^4 - 59\ M\^6 + 521\ M\^8 - 2662\ M\^10 + 
                8327\ M\^12 - 14504\ M\^14 + 4332\ M\^16 + 40110\ M\^18 - 
                78968\ M\^20 - 8053\ M\^22 + 239382\ M\^24 - 282143\ M\^26 - 
                250368\ M\^28 + 890978\ M\^30 - 321216\ M\^32 - 
                1476912\ M\^34 + 1740481\ M\^36 + 1272328\ M\^38 - 
                3471079\ M\^40 + 198703\ M\^42 + 4313576\ M\^44 - 
                2532416\ M\^46 - 3362836\ M\^48 + 4769453\ M\^50 + 
                768458\ M\^52 - 6205978\ M\^54 + 2578526\ M\^56 + 
                6623661\ M\^58 - 5661932\ M\^60 - 5627333\ M\^62 + 
                7525636\ M\^64 + 3302892\ M\^66 - 7290352\ M\^68 - 
                541914\ M\^70 + 5084759\ M\^72 - 1219685\ M\^74 - 
                2320862\ M\^76 + 1427687\ M\^78 + 433771\ M\^80 - 
                757899\ M\^82 + 207119\ M\^84 + 156810\ M\^86 - 
                157343\ M\^88 + 41733\ M\^90 + 22526\ M\^92 - 26413\ M\^94 + 
                12540\ M\^96 - 3567\ M\^98 + 632\ M\^100 - 65\ M\^102 + 
                3\ M\^104)\) + 
          L\^13\ \((\(-M\^14\) + 21\ M\^16 - 205\ M\^18 + 1237\ M\^20 - 
                5124\ M\^22 + 14815\ M\^24 - 26750\ M\^26 + 12136\ M\^28 + 
                77022\ M\^30 - 192791\ M\^32 + 72895\ M\^34 + 412621\ M\^36 - 
                623707\ M\^38 - 201354\ M\^40 + 1214651\ M\^42 - 
                733330\ M\^44 - 948938\ M\^46 + 1881180\ M\^48 - 
                601374\ M\^50 - 2361986\ M\^52 + 2712789\ M\^54 + 
                1877235\ M\^56 - 4159096\ M\^58 - 894885\ M\^60 + 
                4512594\ M\^62 - 49986\ M\^64 - 3972796\ M\^66 + 
                848225\ M\^68 + 2879991\ M\^70 - 1384323\ M\^72 - 
                1517702\ M\^74 + 1398486\ M\^76 + 353261\ M\^78 - 
                906741\ M\^80 + 213227\ M\^82 + 329170\ M\^84 - 
                245488\ M\^86 - 10726\ M\^88 + 91245\ M\^90 - 41612\ M\^92 - 
                5515\ M\^94 + 14786\ M\^96 - 8112\ M\^98 + 2522\ M\^100 - 
                488\ M\^102 + 56\ M\^104 - 3\ M\^106)\) + 
          L\^8\ \((\(-1\) + 20\ M\^2 - 181\ M\^4 + 974\ M\^6 - 3420\ M\^8 + 
                7898\ M\^10 - 10423\ M\^12 - 55\ M\^14 + 30410\ M\^16 - 
                48571\ M\^18 - 29475\ M\^20 + 207746\ M\^22 - 184259\ M\^24 - 
                382260\ M\^26 + 868332\ M\^28 + 176198\ M\^30 - 
                2166851\ M\^32 + 1301637\ M\^34 + 3518027\ M\^36 - 
                4949815\ M\^38 - 3579780\ M\^40 + 10838404\ M\^42 + 
                1102322\ M\^44 - 17664776\ M\^46 + 3673531\ M\^48 + 
                23250770\ M\^50 - 8795064\ M\^52 - 25668578\ M\^54 + 
                12216889\ M\^56 + 23975844\ M\^58 - 13018204\ M\^60 - 
                18416468\ M\^62 + 11731741\ M\^64 + 10771196\ M\^66 - 
                8843572\ M\^68 - 3701319\ M\^70 + 5118642\ M\^72 - 
                611534\ M\^74 - 1711262\ M\^76 + 1760627\ M\^78 - 
                302098\ M\^80 - 1118234\ M\^82 + 803286\ M\^84 + 
                279043\ M\^86 - 513772\ M\^88 + 87977\ M\^90 + 
                156031\ M\^92 - 93205\ M\^94 - 5739\ M\^96 + 28665\ M\^98 - 
                15458\ M\^100 + 4430\ M\^102 - 753\ M\^104 + 72\ M\^106 - 
                3\ M\^108)\) + 
          L\^12\ \((3\ M\^10 - 65\ M\^12 + 632\ M\^14 - 3567\ M\^16 + 
                12540\ M\^18 - 26413\ M\^20 + 22526\ M\^22 + 41733\ M\^24 - 
                157343\ M\^26 + 156810\ M\^28 + 207119\ M\^30 - 
                757899\ M\^32 + 433771\ M\^34 + 1427687\ M\^36 - 
                2320862\ M\^38 - 1219685\ M\^40 + 5084759\ M\^42 - 
                541914\ M\^44 - 7290352\ M\^46 + 3302892\ M\^48 + 
                7525636\ M\^50 - 5627333\ M\^52 - 5661932\ M\^54 + 
                6623661\ M\^56 + 2578526\ M\^58 - 6205978\ M\^60 + 
                768458\ M\^62 + 4769453\ M\^64 - 3362836\ M\^66 - 
                2532416\ M\^68 + 4313576\ M\^70 + 198703\ M\^72 - 
                3471079\ M\^74 + 1272328\ M\^76 + 1740481\ M\^78 - 
                1476912\ M\^80 - 321216\ M\^82 + 890978\ M\^84 - 
                250368\ M\^86 - 282143\ M\^88 + 239382\ M\^90 - 8053\ M\^92 - 
                78968\ M\^94 + 40110\ M\^96 + 4332\ M\^98 - 14504\ M\^100 + 
                8327\ M\^102 - 2662\ M\^104 + 521\ M\^106 - 59\ M\^108 + 
                3\ M\^110)\) + 
          L\^9\ \((1 - 25\ M\^2 + 266\ M\^4 - 1579\ M\^6 + 5588\ M\^8 - 
                10888\ M\^10 + 5071\ M\^12 + 24672\ M\^14 - 38011\ M\^16 - 
                59402\ M\^18 + 191324\ M\^20 + 36132\ M\^22 - 611706\ M\^24 + 
                283516\ M\^26 + 1532825\ M\^28 - 1584916\ M\^30 - 
                2912561\ M\^32 + 4809191\ M\^34 + 4227542\ M\^36 - 
                10513754\ M\^38 - 4577554\ M\^40 + 17803105\ M\^42 + 
                3458573\ M\^44 - 24074656\ M\^46 - 1346091\ M\^48 + 
                26332390\ M\^50 - 899474\ M\^52 - 23082395\ M\^54 + 
                2957002\ M\^56 + 15442884\ M\^58 - 4944367\ M\^60 - 
                5929640\ M\^62 + 6592679\ M\^64 - 2278258\ M\^66 - 
                6780111\ M\^68 + 7163140\ M\^70 + 4964390\ M\^72 - 
                8178933\ M\^74 - 2138798\ M\^76 + 6381861\ M\^78 - 
                113377\ M\^80 - 3622243\ M\^82 + 984851\ M\^84 + 
                1426666\ M\^86 - 823977\ M\^88 - 314003\ M\^90 + 
                397338\ M\^92 - 29550\ M\^94 - 108530\ M\^96 + 46216\ M\^98 + 
                12011\ M\^100 - 18709\ M\^102 + 8522\ M\^104 - 2174\ M\^106 + 
                331\ M\^108 - 28\ M\^110 + M\^112)\) + 
          L\^11\ \((\(-3\)\ M\^6 + 72\ M\^8 - 753\ M\^10 + 4430\ M\^12 - 
                15458\ M\^14 + 28665\ M\^16 - 5739\ M\^18 - 93205\ M\^20 + 
                156031\ M\^22 + 87977\ M\^24 - 513772\ M\^26 + 
                279043\ M\^28 + 803286\ M\^30 - 1118234\ M\^32 - 
                302098\ M\^34 + 1760627\ M\^36 - 1711262\ M\^38 - 
                611534\ M\^40 + 5118642\ M\^42 - 3701319\ M\^44 - 
                8843572\ M\^46 + 10771196\ M\^48 + 11731741\ M\^50 - 
                18416468\ M\^52 - 13018204\ M\^54 + 23975844\ M\^56 + 
                12216889\ M\^58 - 25668578\ M\^60 - 8795064\ M\^62 + 
                23250770\ M\^64 + 3673531\ M\^66 - 17664776\ M\^68 + 
                1102322\ M\^70 + 10838404\ M\^72 - 3579780\ M\^74 - 
                4949815\ M\^76 + 3518027\ M\^78 + 1301637\ M\^80 - 
                2166851\ M\^82 + 176198\ M\^84 + 868332\ M\^86 - 
                382260\ M\^88 - 184259\ M\^90 + 207746\ M\^92 - 
                29475\ M\^94 - 48571\ M\^96 + 30410\ M\^98 - 55\ M\^100 - 
                10423\ M\^102 + 7898\ M\^104 - 3420\ M\^106 + 974\ M\^108 - 
                181\ M\^110 + 20\ M\^112 - M\^114)\) + 
          L\^10\ \((M\^2 - 28\ M\^4 + 331\ M\^6 - 2174\ M\^8 + 8522\ M\^10 - 
                18709\ M\^12 + 12011\ M\^14 + 46216\ M\^16 - 108530\ M\^18 - 
                29550\ M\^20 + 397338\ M\^22 - 314003\ M\^24 - 
                823977\ M\^26 + 1426666\ M\^28 + 984851\ M\^30 - 
                3622243\ M\^32 - 113377\ M\^34 + 6381861\ M\^36 - 
                2138798\ M\^38 - 8178933\ M\^40 + 4964390\ M\^42 + 
                7163140\ M\^44 - 6780111\ M\^46 - 2278258\ M\^48 + 
                6592679\ M\^50 - 5929640\ M\^52 - 4944367\ M\^54 + 
                15442884\ M\^56 + 2957002\ M\^58 - 23082395\ M\^60 - 
                899474\ M\^62 + 26332390\ M\^64 - 1346091\ M\^66 - 
                24074656\ M\^68 + 3458573\ M\^70 + 17803105\ M\^72 - 
                4577554\ M\^74 - 10513754\ M\^76 + 4227542\ M\^78 + 
                4809191\ M\^80 - 2912561\ M\^82 - 1584916\ M\^84 + 
                1532825\ M\^86 + 283516\ M\^88 - 611706\ M\^90 + 
                36132\ M\^92 + 191324\ M\^94 - 59402\ M\^96 - 38011\ M\^98 + 
                24672\ M\^100 + 5071\ M\^102 - 10888\ M\^104 + 5588\ M\^106 - 
                1579\ M\^108 + 266\ M\^110 - 25\ M\^112 + 
                M\^114)\)}, {"A-polynomial", Knot[9, 25], 
        L\^27\ M\^18 + M\^124 + 
          L\^26\ \((\(-4\)\ M\^14 + 16\ M\^16 - 22\ M\^18 - 16\ M\^20 + 
                88\ M\^22 - 22\ M\^24 - 44\ M\^26 + 41\ M\^28 - 18\ M\^30 + 
                4\ M\^32)\) + 
          L\^25\ \((6\ M\^10 - 49\ M\^12 + 168\ M\^14 - 274\ M\^16 + 
                57\ M\^18 + 680\ M\^20 - 1464\ M\^22 + 639\ M\^24 + 
                1996\ M\^26 - 1931\ M\^28 - 139\ M\^30 + 1331\ M\^32 - 
                1130\ M\^34 + 158\ M\^36 + 536\ M\^38 - 504\ M\^40 + 
                217\ M\^42 - 52\ M\^44 + 6\ M\^46)\) + 
          L\^24\ \((\(-4\)\ M\^6 + 51\ M\^8 - 294\ M\^10 + 993\ M\^12 - 
                2035\ M\^14 + 1786\ M\^16 + 2753\ M\^18 - 10400\ M\^20 + 
                11092\ M\^22 + 3862\ M\^24 - 25431\ M\^26 + 21048\ M\^28 + 
                17866\ M\^30 - 29659\ M\^32 + 7206\ M\^34 + 10422\ M\^36 - 
                15450\ M\^38 + 9288\ M\^40 + 3460\ M\^42 - 8950\ M\^44 + 
                4729\ M\^46 + 771\ M\^48 - 2850\ M\^50 + 2259\ M\^52 - 
                1038\ M\^54 + 296\ M\^56 - 50\ M\^58 + 4\ M\^60)\) + 
          L\^23\ \((M\^2 - 19\ M\^4 + 172\ M\^6 - 958\ M\^8 + 3536\ M\^10 - 
                8577\ M\^12 + 12027\ M\^14 - 1694\ M\^16 - 33928\ M\^18 + 
                75146\ M\^20 - 47122\ M\^22 - 97534\ M\^24 + 218017\ M\^26 - 
                66744\ M\^28 - 293835\ M\^30 + 311187\ M\^32 + 
                184331\ M\^34 - 347970\ M\^36 - 9404\ M\^38 + 170952\ M\^40 - 
                56925\ M\^42 + 5106\ M\^44 + 6126\ M\^46 - 44137\ M\^48 + 
                41291\ M\^50 + 190\ M\^52 - 24656\ M\^54 + 18770\ M\^56 - 
                1813\ M\^58 - 8464\ M\^60 + 8841\ M\^62 - 5138\ M\^64 + 
                2072\ M\^66 - 606\ M\^68 + 124\ M\^70 - 16\ M\^72 + 
                M\^74)\) + 
          L\^22\ \((1 - 24\ M\^2 + 261\ M\^4 - 1714\ M\^6 + 7529\ M\^8 - 
                22598\ M\^10 + 43422\ M\^12 - 37437\ M\^14 - 46666\ M\^16 + 
                192226\ M\^18 - 254252\ M\^20 + 60124\ M\^22 + 
                445186\ M\^24 - 944591\ M\^26 + 422963\ M\^28 + 
                1615831\ M\^30 - 2491848\ M\^32 - 1022001\ M\^34 + 
                4331951\ M\^36 - 444076\ M\^38 - 4291466\ M\^40 + 
                1578367\ M\^42 + 2578583\ M\^44 - 1609643\ M\^46 - 
                799691\ M\^48 + 920967\ M\^50 - 64127\ M\^52 - 
                275128\ M\^54 + 171420\ M\^56 - 6262\ M\^58 - 40751\ M\^60 + 
                11886\ M\^62 - 3613\ M\^64 + 11450\ M\^66 - 2452\ M\^68 - 
                14175\ M\^70 + 19932\ M\^72 - 14532\ M\^74 + 6844\ M\^76 - 
                2158\ M\^78 + 443\ M\^80 - 54\ M\^82 + 3\ M\^84)\) + 
          L\^21\ \((\(-6\) + 139\ M\^2 - 1435\ M\^4 + 8703\ M\^6 - 
                34290\ M\^8 + 90155\ M\^10 - 146713\ M\^12 + 75466\ M\^14 + 
                283032\ M\^16 - 747982\ M\^18 + 552714\ M\^20 + 
                634291\ M\^22 - 1439434\ M\^24 + 931103\ M\^26 - 
                734536\ M\^28 - 664352\ M\^30 + 6955290\ M\^32 - 
                6400190\ M\^34 - 12669396\ M\^36 + 18598199\ M\^38 + 
                12246855\ M\^40 - 26203046\ M\^42 - 5250737\ M\^44 + 
                22730409\ M\^46 - 2125010\ M\^48 - 12158397\ M\^50 + 
                4671816\ M\^52 + 2808216\ M\^54 - 2985583\ M\^56 + 
                1115173\ M\^58 + 709880\ M\^60 - 1194051\ M\^62 + 
                203008\ M\^64 + 458344\ M\^66 - 236428\ M\^68 - 
                47195\ M\^70 + 60324\ M\^72 - 7459\ M\^74 - 9661\ M\^76 + 
                22735\ M\^78 - 38855\ M\^80 + 39575\ M\^82 - 25423\ M\^84 + 
                10799\ M\^86 - 3061\ M\^88 + 562\ M\^90 - 61\ M\^92 + 
                3\ M\^94)\) + 
          L\^20\ \((15 - 369\ M\^2 + 4022\ M\^4 - 25401\ M\^6 + 
                101498\ M\^8 - 258879\ M\^10 + 378551\ M\^12 - 99600\ M\^14 - 
                851735\ M\^16 + 1895649\ M\^18 - 1275579\ M\^20 - 
                2270178\ M\^22 + 5367348\ M\^24 - 968332\ M\^26 - 
                6562770\ M\^28 + 743576\ M\^30 + 5814762\ M\^32 + 
                15987127\ M\^34 - 16429402\ M\^36 - 48543679\ M\^38 + 
                45637735\ M\^40 + 74247169\ M\^42 - 75487088\ M\^44 - 
                69101721\ M\^46 + 80412417\ M\^48 + 37767348\ M\^50 - 
                55952381\ M\^52 - 5949309\ M\^54 + 20790770\ M\^56 - 
                8227093\ M\^58 + 3462227\ M\^60 + 6875396\ M\^62 - 
                10689100\ M\^64 - 1101892\ M\^66 + 7727205\ M\^68 - 
                1995658\ M\^70 - 2948470\ M\^72 + 1795865\ M\^74 + 
                453791\ M\^76 - 791162\ M\^78 + 201294\ M\^80 + 
                118177\ M\^82 - 70448\ M\^84 - 50825\ M\^86 + 101224\ M\^88 - 
                88185\ M\^90 + 53172\ M\^92 - 23567\ M\^94 + 7606\ M\^96 - 
                1731\ M\^98 + 263\ M\^100 - 24\ M\^102 + M\^104)\) + 
          L\^19\ \((\(-20\) + 534\ M\^2 - 6364\ M\^4 + 44236\ M\^6 - 
                195252\ M\^8 + 548702\ M\^10 - 872738\ M\^12 + 
                252626\ M\^14 + 1981412\ M\^16 - 3923089\ M\^18 + 
                1737646\ M\^20 + 4688452\ M\^22 - 9499066\ M\^24 + 
                4669593\ M\^26 + 12770607\ M\^28 - 19192636\ M\^30 - 
                15676725\ M\^32 + 27011248\ M\^34 + 52715215\ M\^36 - 
                37107627\ M\^38 - 151540493\ M\^40 + 74011071\ M\^42 + 
                271822103\ M\^44 - 135197077\ M\^46 - 318399710\ M\^48 + 
                180593769\ M\^50 + 245308159\ M\^52 - 168212279\ M\^54 - 
                105532994\ M\^56 + 98821911\ M\^58 - 8405708\ M\^60 - 
                19037474\ M\^62 + 50940156\ M\^64 - 28265409\ M\^66 - 
                40580750\ M\^68 + 36043830\ M\^70 + 17397642\ M\^72 - 
                23702929\ M\^74 - 3335811\ M\^76 + 10672720\ M\^78 - 
                1070907\ M\^80 - 3327332\ M\^82 + 1054840\ M\^84 + 
                780451\ M\^86 - 583526\ M\^88 + 39457\ M\^90 + 41578\ M\^92 + 
                105275\ M\^94 - 181909\ M\^96 + 147948\ M\^98 - 
                79993\ M\^100 + 31294\ M\^102 - 9015\ M\^104 + 1880\ M\^106 - 
                270\ M\^108 + 24\ M\^110 - M\^112)\) + 
          L\^18\ \((15 - 442\ M\^2 + 5888\ M\^4 - 46421\ M\^6 + 
                236667\ M\^8 - 790133\ M\^10 + 1600187\ M\^12 - 
                1225589\ M\^14 - 2617867\ M\^16 + 7871146\ M\^18 - 
                4752065\ M\^20 - 10314617\ M\^22 + 18508699\ M\^24 - 
                3300800\ M\^26 - 19446762\ M\^28 + 30560685\ M\^30 - 
                7041720\ M\^32 - 68857103\ M\^34 + 45549523\ M\^36 + 
                163360128\ M\^38 - 79586084\ M\^40 - 378671215\ M\^42 + 
                129042342\ M\^44 + 670060875\ M\^46 - 226435701\ M\^48 - 
                838385917\ M\^50 + 353261790\ M\^52 + 711068813\ M\^54 - 
                414678061\ M\^56 - 331934479\ M\^58 + 331696631\ M\^60 - 
                68433692\ M\^62 - 144099165\ M\^64 + 284222013\ M\^66 - 
                33412547\ M\^68 - 285697115\ M\^70 + 118469429\ M\^72 + 
                180188900\ M\^74 - 113852339\ M\^76 - 77780221\ M\^78 + 
                70743166\ M\^80 + 21249502\ M\^82 - 31635835\ M\^84 - 
                1986509\ M\^86 + 10531764\ M\^88 - 1377393\ M\^90 - 
                2305146\ M\^92 + 488367\ M\^94 + 565384\ M\^96 - 
                312949\ M\^98 + 155803\ M\^100 - 252762\ M\^102 + 
                270350\ M\^104 - 171746\ M\^106 + 72156\ M\^108 - 
                21048\ M\^110 + 4287\ M\^112 - 590\ M\^114 + 50\ M\^116 - 
                2\ M\^118)\) + 
          L\^17\ \((\(-6\) + 203\ M\^2 - 3127\ M\^4 + 28788\ M\^6 - 
                174147\ M\^8 + 711334\ M\^10 - 1901184\ M\^12 + 
                2794848\ M\^14 + 143539\ M\^16 - 9184478\ M\^18 + 
                14277817\ M\^20 + 3758586\ M\^22 - 35203519\ M\^24 + 
                28350776\ M\^26 + 28993211\ M\^28 - 67581590\ M\^30 + 
                42482223\ M\^32 + 57913769\ M\^34 - 196526283\ M\^36 + 
                33233706\ M\^38 + 463225587\ M\^40 - 183975160\ M\^42 - 
                882683219\ M\^44 + 359363107\ M\^46 + 1376355056\ M\^48 - 
                561243061\ M\^50 - 1654085637\ M\^52 + 778485916\ M\^54 + 
                1406582508\ M\^56 - 880103774\ M\^58 - 626169656\ M\^60 + 
                701701478\ M\^62 - 300158584\ M\^64 - 260251826\ M\^66 + 
                885544976\ M\^68 - 224462217\ M\^70 - 938221828\ M\^72 + 
                508712996\ M\^74 + 639119841\ M\^76 - 519689172\ M\^78 - 
                289624488\ M\^80 + 362916225\ M\^82 + 68574544\ M\^84 - 
                186554798\ M\^86 + 13556525\ M\^88 + 70639750\ M\^90 - 
                22079526\ M\^92 - 18328864\ M\^94 + 11318021\ M\^96 + 
                2873154\ M\^98 - 4184798\ M\^100 + 524617\ M\^102 + 
                706056\ M\^104 + 9702\ M\^106 - 564944\ M\^108 + 
                508987\ M\^110 - 257725\ M\^112 + 88313\ M\^114 - 
                21436\ M\^116 + 3668\ M\^118 - 424\ M\^120 + 30\ M\^122 - 
                M\^124)\) + 
          L\ \((4\ M\^110 - 18\ M\^112 + 41\ M\^114 - 44\ M\^116 - 
                22\ M\^118 + 88\ M\^120 - 16\ M\^122 - 22\ M\^124 + 
                16\ M\^126 - 4\ M\^128)\) + 
          L\^16\ \((1 - 45\ M\^2 + 874\ M\^4 - 9907\ M\^6 + 73526\ M\^8 - 
                374749\ M\^10 + 1316440\ M\^12 - 3010280\ M\^14 + 
                3414384\ M\^16 + 2600134\ M\^18 - 15653244\ M\^20 + 
                17547073\ M\^22 + 16306529\ M\^24 - 60271289\ M\^26 + 
                33813242\ M\^28 + 71219962\ M\^30 - 133112226\ M\^32 + 
                52267838\ M\^34 + 181356717\ M\^36 - 380878213\ M\^38 - 
                40926022\ M\^40 + 907064742\ M\^42 - 328318776\ M\^44 - 
                1550060749\ M\^46 + 813365558\ M\^48 + 2134579752\ M\^50 - 
                1255411116\ M\^52 - 2347798000\ M\^54 + 1530780776\ M\^56 + 
                1864858394\ M\^58 - 1476079436\ M\^60 - 653879339\ M\^62 + 
                936199735\ M\^64 - 801703351\ M\^66 + 16767180\ M\^68 + 
                1756273169\ M\^70 - 1013055713\ M\^72 - 1810027040\ M\^74 + 
                1589791805\ M\^76 + 1190755693\ M\^78 - 1552954115\ M\^80 - 
                446484099\ M\^82 + 1094444211\ M\^84 - 24281031\ M\^86 - 
                567209127\ M\^88 + 166515781\ M\^90 + 204234154\ M\^92 - 
                129046397\ M\^94 - 38460019\ M\^96 + 59102578\ M\^98 - 
                6650967\ M\^100 - 17402053\ M\^102 + 8930219\ M\^104 + 
                1568059\ M\^106 - 3031925\ M\^108 + 504420\ M\^110 + 
                1054195\ M\^112 - 1056557\ M\^114 + 558331\ M\^116 - 
                199558\ M\^118 + 50902\ M\^120 - 9232\ M\^122 + 
                1139\ M\^124 - 86\ M\^126 + 3\ M\^128)\) + 
          L\^15\ \((3\ M\^2 - 103\ M\^4 + 1644\ M\^6 - 16061\ M\^8 + 
                106178\ M\^10 - 493707\ M\^12 + 1609864\ M\^14 - 
                3452185\ M\^16 + 3615313\ M\^18 + 3323959\ M\^20 - 
                17445368\ M\^22 + 18845392\ M\^24 + 19276682\ M\^26 - 
                71329361\ M\^28 + 46600400\ M\^30 + 84908753\ M\^32 - 
                193991006\ M\^34 + 95537323\ M\^36 + 285039833\ M\^38 - 
                591049487\ M\^40 - 61485816\ M\^42 + 1345378213\ M\^44 - 
                629578362\ M\^46 - 2110666204\ M\^48 + 1611250425\ M\^50 + 
                2585883096\ M\^52 - 2455552446\ M\^54 - 2530161200\ M\^56 + 
                2757905224\ M\^58 + 1804727430\ M\^60 - 2287461571\ M\^62 - 
                467044149\ M\^64 + 1046155768\ M\^66 - 1031710033\ M\^68 + 
                676190082\ M\^70 + 1954528442\ M\^72 - 2311158904\ M\^74 - 
                1850162935\ M\^76 + 3195269504\ M\^78 + 949502298\ M\^80 - 
                3023394860\ M\^82 + 59987245\ M\^84 + 2094249644\ M\^86 - 
                613398151\ M\^88 - 1031881916\ M\^90 + 631924360\ M\^92 + 
                300848657\ M\^94 - 384286318\ M\^96 + 10867680\ M\^98 + 
                147927322\ M\^100 - 63940147\ M\^102 - 26439418\ M\^104 + 
                34647378\ M\^106 - 7630001\ M\^108 - 7400572\ M\^110 + 
                5893205\ M\^112 - 460712\ M\^114 - 1937052\ M\^116 + 
                1678610\ M\^118 - 813097\ M\^120 + 270118\ M\^122 - 
                64429\ M\^124 + 10974\ M\^126 - 1276\ M\^128 + 91\ M\^130 - 
                3\ M\^132)\) + 
          L\^2\ \((6\ M\^96 - 52\ M\^98 + 217\ M\^100 - 504\ M\^102 + 
                536\ M\^104 + 158\ M\^106 - 1130\ M\^108 + 1331\ M\^110 - 
                139\ M\^112 - 1931\ M\^114 + 1996\ M\^116 + 639\ M\^118 - 
                1464\ M\^120 + 680\ M\^122 + 57\ M\^124 - 274\ M\^126 + 
                168\ M\^128 - 49\ M\^130 + 6\ M\^132)\) + 
          L\^3\ \((4\ M\^82 - 50\ M\^84 + 296\ M\^86 - 1038\ M\^88 + 
                2259\ M\^90 - 2850\ M\^92 + 771\ M\^94 + 4729\ M\^96 - 
                8950\ M\^98 + 3460\ M\^100 + 9288\ M\^102 - 15450\ M\^104 + 
                10422\ M\^106 + 7206\ M\^108 - 29659\ M\^110 + 
                17866\ M\^112 + 21048\ M\^114 - 25431\ M\^116 + 
                3862\ M\^118 + 11092\ M\^120 - 10400\ M\^122 + 2753\ M\^124 + 
                1786\ M\^126 - 2035\ M\^128 + 993\ M\^130 - 294\ M\^132 + 
                51\ M\^134 - 4\ M\^136)\) + 
          L\^14\ \((3\ M\^4 - 97\ M\^6 + 1473\ M\^8 - 13858\ M\^10 + 
                89400\ M\^12 - 411456\ M\^14 + 1350931\ M\^16 - 
                3002317\ M\^18 + 3617481\ M\^20 + 1311971\ M\^22 - 
                13356712\ M\^24 + 18525327\ M\^26 + 8761585\ M\^28 - 
                61075922\ M\^30 + 63551742\ M\^32 + 53478886\ M\^34 - 
                213921253\ M\^36 + 164070727\ M\^38 + 291336720\ M\^40 - 
                716555067\ M\^42 + 19065184\ M\^44 + 1504399992\ M\^46 - 
                952571141\ M\^48 - 2160621485\ M\^50 + 2320526238\ M\^52 + 
                2299678529\ M\^54 - 3490884195\ M\^56 - 1813129579\ M\^58 + 
                3779817874\ M\^60 + 899568738\ M\^62 - 2884728376\ M\^64 + 
                136333387\ M\^66 + 1022247384\ M\^68 - 953184232\ M\^70 + 
                1248958356\ M\^72 + 1188733776\ M\^74 - 3230910424\ M\^76 - 
                681913008\ M\^78 + 4224090994\ M\^80 - 316274200\ M\^82 - 
                3910345046\ M\^84 + 1219325813\ M\^86 + 2641065782\ M\^88 - 
                1525137549\ M\^90 - 1195715389\ M\^92 + 1209521823\ M\^94 + 
                213877178\ M\^96 - 642314643\ M\^98 + 158074545\ M\^100 + 
                198593136\ M\^102 - 156678414\ M\^104 + 226703\ M\^106 + 
                61039816\ M\^108 - 33225348\ M\^110 - 4135398\ M\^112 + 
                13756192\ M\^114 - 6044515\ M\^116 - 1235701\ M\^118 + 
                2970047\ M\^120 - 1914145\ M\^122 + 765612\ M\^124 - 
                214325\ M\^126 + 43170\ M\^128 - 6188\ M\^130 + 603\ M\^132 - 
                36\ M\^134 + M\^136)\) + 
          L\^13\ \((M\^6 - 36\ M\^8 + 603\ M\^10 - 6188\ M\^12 + 
                43170\ M\^14 - 214325\ M\^16 + 765612\ M\^18 - 
                1914145\ M\^20 + 2970047\ M\^22 - 1235701\ M\^24 - 
                6044515\ M\^26 + 13756192\ M\^28 - 4135398\ M\^30 - 
                33225348\ M\^32 + 61039816\ M\^34 + 226703\ M\^36 - 
                156678414\ M\^38 + 198593136\ M\^40 + 158074545\ M\^42 - 
                642314643\ M\^44 + 213877178\ M\^46 + 1209521823\ M\^48 - 
                1195715389\ M\^50 - 1525137549\ M\^52 + 2641065782\ M\^54 + 
                1219325813\ M\^56 - 3910345046\ M\^58 - 316274200\ M\^60 + 
                4224090994\ M\^62 - 681913008\ M\^64 - 3230910424\ M\^66 + 
                1188733776\ M\^68 + 1248958356\ M\^70 - 953184232\ M\^72 + 
                1022247384\ M\^74 + 136333387\ M\^76 - 2884728376\ M\^78 + 
                899568738\ M\^80 + 3779817874\ M\^82 - 1813129579\ M\^84 - 
                3490884195\ M\^86 + 2299678529\ M\^88 + 2320526238\ M\^90 - 
                2160621485\ M\^92 - 952571141\ M\^94 + 1504399992\ M\^96 + 
                19065184\ M\^98 - 716555067\ M\^100 + 291336720\ M\^102 + 
                164070727\ M\^104 - 213921253\ M\^106 + 53478886\ M\^108 + 
                63551742\ M\^110 - 61075922\ M\^112 + 8761585\ M\^114 + 
                18525327\ M\^116 - 13356712\ M\^118 + 1311971\ M\^120 + 
                3617481\ M\^122 - 3002317\ M\^124 + 1350931\ M\^126 - 
                411456\ M\^128 + 89400\ M\^130 - 13858\ M\^132 + 
                1473\ M\^134 - 97\ M\^136 + 3\ M\^138)\) + 
          L\^4\ \((M\^68 - 16\ M\^70 + 124\ M\^72 - 606\ M\^74 + 
                2072\ M\^76 - 5138\ M\^78 + 8841\ M\^80 - 8464\ M\^82 - 
                1813\ M\^84 + 18770\ M\^86 - 24656\ M\^88 + 190\ M\^90 + 
                41291\ M\^92 - 44137\ M\^94 + 6126\ M\^96 + 5106\ M\^98 - 
                56925\ M\^100 + 170952\ M\^102 - 9404\ M\^104 - 
                347970\ M\^106 + 184331\ M\^108 + 311187\ M\^110 - 
                293835\ M\^112 - 66744\ M\^114 + 218017\ M\^116 - 
                97534\ M\^118 - 47122\ M\^120 + 75146\ M\^122 - 
                33928\ M\^124 - 1694\ M\^126 + 12027\ M\^128 - 8577\ M\^130 + 
                3536\ M\^132 - 958\ M\^134 + 172\ M\^136 - 19\ M\^138 + 
                M\^140)\) + 
          L\^12\ \((\(-3\)\ M\^10 + 91\ M\^12 - 1276\ M\^14 + 10974\ M\^16 - 
                64429\ M\^18 + 270118\ M\^20 - 813097\ M\^22 + 
                1678610\ M\^24 - 1937052\ M\^26 - 460712\ M\^28 + 
                5893205\ M\^30 - 7400572\ M\^32 - 7630001\ M\^34 + 
                34647378\ M\^36 - 26439418\ M\^38 - 63940147\ M\^40 + 
                147927322\ M\^42 + 10867680\ M\^44 - 384286318\ M\^46 + 
                300848657\ M\^48 + 631924360\ M\^50 - 1031881916\ M\^52 - 
                613398151\ M\^54 + 2094249644\ M\^56 + 59987245\ M\^58 - 
                3023394860\ M\^60 + 949502298\ M\^62 + 3195269504\ M\^64 - 
                1850162935\ M\^66 - 2311158904\ M\^68 + 1954528442\ M\^70 + 
                676190082\ M\^72 - 1031710033\ M\^74 + 1046155768\ M\^76 - 
                467044149\ M\^78 - 2287461571\ M\^80 + 1804727430\ M\^82 + 
                2757905224\ M\^84 - 2530161200\ M\^86 - 2455552446\ M\^88 + 
                2585883096\ M\^90 + 1611250425\ M\^92 - 2110666204\ M\^94 - 
                629578362\ M\^96 + 1345378213\ M\^98 - 61485816\ M\^100 - 
                591049487\ M\^102 + 285039833\ M\^104 + 95537323\ M\^106 - 
                193991006\ M\^108 + 84908753\ M\^110 + 46600400\ M\^112 - 
                71329361\ M\^114 + 19276682\ M\^116 + 18845392\ M\^118 - 
                17445368\ M\^120 + 3323959\ M\^122 + 3615313\ M\^124 - 
                3452185\ M\^126 + 1609864\ M\^128 - 493707\ M\^130 + 
                106178\ M\^132 - 16061\ M\^134 + 1644\ M\^136 - 103\ M\^138 + 
                3\ M\^140)\) + 
          L\^8\ \((\(-M\^30\) + 24\ M\^32 - 270\ M\^34 + 1880\ M\^36 - 
                9015\ M\^38 + 31294\ M\^40 - 79993\ M\^42 + 147948\ M\^44 - 
                181909\ M\^46 + 105275\ M\^48 + 41578\ M\^50 + 39457\ M\^52 - 
                583526\ M\^54 + 780451\ M\^56 + 1054840\ M\^58 - 
                3327332\ M\^60 - 1070907\ M\^62 + 10672720\ M\^64 - 
                3335811\ M\^66 - 23702929\ M\^68 + 17397642\ M\^70 + 
                36043830\ M\^72 - 40580750\ M\^74 - 28265409\ M\^76 + 
                50940156\ M\^78 - 19037474\ M\^80 - 8405708\ M\^82 + 
                98821911\ M\^84 - 105532994\ M\^86 - 168212279\ M\^88 + 
                245308159\ M\^90 + 180593769\ M\^92 - 318399710\ M\^94 - 
                135197077\ M\^96 + 271822103\ M\^98 + 74011071\ M\^100 - 
                151540493\ M\^102 - 37107627\ M\^104 + 52715215\ M\^106 + 
                27011248\ M\^108 - 15676725\ M\^110 - 19192636\ M\^112 + 
                12770607\ M\^114 + 4669593\ M\^116 - 9499066\ M\^118 + 
                4688452\ M\^120 + 1737646\ M\^122 - 3923089\ M\^124 + 
                1981412\ M\^126 + 252626\ M\^128 - 872738\ M\^130 + 
                548702\ M\^132 - 195252\ M\^134 + 44236\ M\^136 - 
                6364\ M\^138 + 534\ M\^140 - 20\ M\^142)\) + 
          L\^6\ \((3\ M\^48 - 61\ M\^50 + 562\ M\^52 - 3061\ M\^54 + 
                10799\ M\^56 - 25423\ M\^58 + 39575\ M\^60 - 38855\ M\^62 + 
                22735\ M\^64 - 9661\ M\^66 - 7459\ M\^68 + 60324\ M\^70 - 
                47195\ M\^72 - 236428\ M\^74 + 458344\ M\^76 + 
                203008\ M\^78 - 1194051\ M\^80 + 709880\ M\^82 + 
                1115173\ M\^84 - 2985583\ M\^86 + 2808216\ M\^88 + 
                4671816\ M\^90 - 12158397\ M\^92 - 2125010\ M\^94 + 
                22730409\ M\^96 - 5250737\ M\^98 - 26203046\ M\^100 + 
                12246855\ M\^102 + 18598199\ M\^104 - 12669396\ M\^106 - 
                6400190\ M\^108 + 6955290\ M\^110 - 664352\ M\^112 - 
                734536\ M\^114 + 931103\ M\^116 - 1439434\ M\^118 + 
                634291\ M\^120 + 552714\ M\^122 - 747982\ M\^124 + 
                283032\ M\^126 + 75466\ M\^128 - 146713\ M\^130 + 
                90155\ M\^132 - 34290\ M\^134 + 8703\ M\^136 - 1435\ M\^138 + 
                139\ M\^140 - 6\ M\^142)\) + 
          L\^10\ \((\(-M\^18\) + 30\ M\^20 - 424\ M\^22 + 3668\ M\^24 - 
                21436\ M\^26 + 88313\ M\^28 - 257725\ M\^30 + 508987\ M\^32 - 
                564944\ M\^34 + 9702\ M\^36 + 706056\ M\^38 + 524617\ M\^40 - 
                4184798\ M\^42 + 2873154\ M\^44 + 11318021\ M\^46 - 
                18328864\ M\^48 - 22079526\ M\^50 + 70639750\ M\^52 + 
                13556525\ M\^54 - 186554798\ M\^56 + 68574544\ M\^58 + 
                362916225\ M\^60 - 289624488\ M\^62 - 519689172\ M\^64 + 
                639119841\ M\^66 + 508712996\ M\^68 - 938221828\ M\^70 - 
                224462217\ M\^72 + 885544976\ M\^74 - 260251826\ M\^76 - 
                300158584\ M\^78 + 701701478\ M\^80 - 626169656\ M\^82 - 
                880103774\ M\^84 + 1406582508\ M\^86 + 778485916\ M\^88 - 
                1654085637\ M\^90 - 561243061\ M\^92 + 1376355056\ M\^94 + 
                359363107\ M\^96 - 882683219\ M\^98 - 183975160\ M\^100 + 
                463225587\ M\^102 + 33233706\ M\^104 - 196526283\ M\^106 + 
                57913769\ M\^108 + 42482223\ M\^110 - 67581590\ M\^112 + 
                28993211\ M\^114 + 28350776\ M\^116 - 35203519\ M\^118 + 
                3758586\ M\^120 + 14277817\ M\^122 - 9184478\ M\^124 + 
                143539\ M\^126 + 2794848\ M\^128 - 1901184\ M\^130 + 
                711334\ M\^132 - 174147\ M\^134 + 28788\ M\^136 - 
                3127\ M\^138 + 203\ M\^140 - 6\ M\^142)\) + 
          L\^11\ \((3\ M\^14 - 86\ M\^16 + 1139\ M\^18 - 9232\ M\^20 + 
                50902\ M\^22 - 199558\ M\^24 + 558331\ M\^26 - 
                1056557\ M\^28 + 1054195\ M\^30 + 504420\ M\^32 - 
                3031925\ M\^34 + 1568059\ M\^36 + 8930219\ M\^38 - 
                17402053\ M\^40 - 6650967\ M\^42 + 59102578\ M\^44 - 
                38460019\ M\^46 - 129046397\ M\^48 + 204234154\ M\^50 + 
                166515781\ M\^52 - 567209127\ M\^54 - 24281031\ M\^56 + 
                1094444211\ M\^58 - 446484099\ M\^60 - 1552954115\ M\^62 + 
                1190755693\ M\^64 + 1589791805\ M\^66 - 1810027040\ M\^68 - 
                1013055713\ M\^70 + 1756273169\ M\^72 + 16767180\ M\^74 - 
                801703351\ M\^76 + 936199735\ M\^78 - 653879339\ M\^80 - 
                1476079436\ M\^82 + 1864858394\ M\^84 + 1530780776\ M\^86 - 
                2347798000\ M\^88 - 1255411116\ M\^90 + 2134579752\ M\^92 + 
                813365558\ M\^94 - 1550060749\ M\^96 - 328318776\ M\^98 + 
                907064742\ M\^100 - 40926022\ M\^102 - 380878213\ M\^104 + 
                181356717\ M\^106 + 52267838\ M\^108 - 133112226\ M\^110 + 
                71219962\ M\^112 + 33813242\ M\^114 - 60271289\ M\^116 + 
                16306529\ M\^118 + 17547073\ M\^120 - 15653244\ M\^122 + 
                2600134\ M\^124 + 3414384\ M\^126 - 3010280\ M\^128 + 
                1316440\ M\^130 - 374749\ M\^132 + 73526\ M\^134 - 
                9907\ M\^136 + 874\ M\^138 - 45\ M\^140 + M\^142)\) + 
          L\^5\ \((3\ M\^58 - 54\ M\^60 + 443\ M\^62 - 2158\ M\^64 + 
                6844\ M\^66 - 14532\ M\^68 + 19932\ M\^70 - 14175\ M\^72 - 
                2452\ M\^74 + 11450\ M\^76 - 3613\ M\^78 + 11886\ M\^80 - 
                40751\ M\^82 - 6262\ M\^84 + 171420\ M\^86 - 275128\ M\^88 - 
                64127\ M\^90 + 920967\ M\^92 - 799691\ M\^94 - 
                1609643\ M\^96 + 2578583\ M\^98 + 1578367\ M\^100 - 
                4291466\ M\^102 - 444076\ M\^104 + 4331951\ M\^106 - 
                1022001\ M\^108 - 2491848\ M\^110 + 1615831\ M\^112 + 
                422963\ M\^114 - 944591\ M\^116 + 445186\ M\^118 + 
                60124\ M\^120 - 254252\ M\^122 + 192226\ M\^124 - 
                46666\ M\^126 - 37437\ M\^128 + 43422\ M\^130 - 
                22598\ M\^132 + 7529\ M\^134 - 1714\ M\^136 + 261\ M\^138 - 
                24\ M\^140 + M\^142)\) + 
          L\^9\ \((\(-2\)\ M\^24 + 50\ M\^26 - 590\ M\^28 + 4287\ M\^30 - 
                21048\ M\^32 + 72156\ M\^34 - 171746\ M\^36 + 270350\ M\^38 - 
                252762\ M\^40 + 155803\ M\^42 - 312949\ M\^44 + 
                565384\ M\^46 + 488367\ M\^48 - 2305146\ M\^50 - 
                1377393\ M\^52 + 10531764\ M\^54 - 1986509\ M\^56 - 
                31635835\ M\^58 + 21249502\ M\^60 + 70743166\ M\^62 - 
                77780221\ M\^64 - 113852339\ M\^66 + 180188900\ M\^68 + 
                118469429\ M\^70 - 285697115\ M\^72 - 33412547\ M\^74 + 
                284222013\ M\^76 - 144099165\ M\^78 - 68433692\ M\^80 + 
                331696631\ M\^82 - 331934479\ M\^84 - 414678061\ M\^86 + 
                711068813\ M\^88 + 353261790\ M\^90 - 838385917\ M\^92 - 
                226435701\ M\^94 + 670060875\ M\^96 + 129042342\ M\^98 - 
                378671215\ M\^100 - 79586084\ M\^102 + 163360128\ M\^104 + 
                45549523\ M\^106 - 68857103\ M\^108 - 7041720\ M\^110 + 
                30560685\ M\^112 - 19446762\ M\^114 - 3300800\ M\^116 + 
                18508699\ M\^118 - 10314617\ M\^120 - 4752065\ M\^122 + 
                7871146\ M\^124 - 2617867\ M\^126 - 1225589\ M\^128 + 
                1600187\ M\^130 - 790133\ M\^132 + 236667\ M\^134 - 
                46421\ M\^136 + 5888\ M\^138 - 442\ M\^140 + 15\ M\^142)\) + 
          L\^7\ \((M\^38 - 24\ M\^40 + 263\ M\^42 - 1731\ M\^44 + 
                7606\ M\^46 - 23567\ M\^48 + 53172\ M\^50 - 88185\ M\^52 + 
                101224\ M\^54 - 50825\ M\^56 - 70448\ M\^58 + 118177\ M\^60 + 
                201294\ M\^62 - 791162\ M\^64 + 453791\ M\^66 + 
                1795865\ M\^68 - 2948470\ M\^70 - 1995658\ M\^72 + 
                7727205\ M\^74 - 1101892\ M\^76 - 10689100\ M\^78 + 
                6875396\ M\^80 + 3462227\ M\^82 - 8227093\ M\^84 + 
                20790770\ M\^86 - 5949309\ M\^88 - 55952381\ M\^90 + 
                37767348\ M\^92 + 80412417\ M\^94 - 69101721\ M\^96 - 
                75487088\ M\^98 + 74247169\ M\^100 + 45637735\ M\^102 - 
                48543679\ M\^104 - 16429402\ M\^106 + 15987127\ M\^108 + 
                5814762\ M\^110 + 743576\ M\^112 - 6562770\ M\^114 - 
                968332\ M\^116 + 5367348\ M\^118 - 2270178\ M\^120 - 
                1275579\ M\^122 + 1895649\ M\^124 - 851735\ M\^126 - 
                99600\ M\^128 + 378551\ M\^130 - 258879\ M\^132 + 
                101498\ M\^134 - 25401\ M\^136 + 4022\ M\^138 - 369\ M\^140 + 
                15\ M\^142)\)}, {"A-polynomial", Knot[9, 26], 
        L\^23\ M\^24 + M\^118 + 
          L\^22\ \((4\ M\^18 - 16\ M\^20 + 26\ M\^22 - 15\ M\^24 - 
                28\ M\^26 + 72\ M\^28 - 4\ M\^30 - 35\ M\^32 + 25\ M\^34 - 
                6\ M\^36)\) + 
          L\^21\ \((6\ M\^12 - 44\ M\^14 + 132\ M\^16 - 192\ M\^18 + 
                61\ M\^20 + 202\ M\^22 - 235\ M\^24 + 236\ M\^26 - 
                556\ M\^28 - 31\ M\^30 + 1579\ M\^32 - 870\ M\^34 - 
                806\ M\^36 + 1221\ M\^38 - 301\ M\^40 - 404\ M\^42 + 
                358\ M\^44 - 118\ M\^46 + 15\ M\^48)\) + 
          L\^20\ \((4\ M\^6 - 40\ M\^8 + 170\ M\^10 - 392\ M\^12 + 
                508\ M\^14 - 444\ M\^16 + 639\ M\^18 - 419\ M\^20 - 
                2886\ M\^22 + 6719\ M\^24 - 1435\ M\^26 - 10699\ M\^28 + 
                12595\ M\^30 - 114\ M\^32 - 13547\ M\^34 + 14590\ M\^36 + 
                746\ M\^38 - 9570\ M\^40 + 9567\ M\^42 - 4173\ M\^44 - 
                4068\ M\^46 + 7090\ M\^48 - 2924\ M\^50 - 1523\ M\^52 + 
                2180\ M\^54 - 1005\ M\^56 + 222\ M\^58 - 20\ M\^60)\) + 
          L\^19\ \((1 - 12\ M\^2 + 64\ M\^4 - 204\ M\^6 + 496\ M\^8 - 
                1244\ M\^10 + 3028\ M\^12 - 4378\ M\^14 - 1096\ M\^16 + 
                14997\ M\^18 - 15969\ M\^20 - 16683\ M\^22 + 41491\ M\^24 + 
                5935\ M\^26 - 64842\ M\^28 - 3047\ M\^30 + 114004\ M\^32 - 
                12547\ M\^34 - 189986\ M\^36 + 84858\ M\^38 + 209819\ M\^40 - 
                176102\ M\^42 - 97123\ M\^44 + 195486\ M\^46 - 40327\ M\^48 - 
                101362\ M\^50 + 79812\ M\^52 - 1091\ M\^54 - 31289\ M\^56 + 
                23779\ M\^58 - 7167\ M\^60 - 4289\ M\^62 + 6697\ M\^64 - 
                3888\ M\^66 + 1227\ M\^68 - 208\ M\^70 + 15\ M\^72)\) + 
          L\^18\ \((\(-5\) + 76\ M\^2 - 513\ M\^4 + 1951\ M\^6 - 4305\ M\^8 + 
                4759\ M\^10 - 658\ M\^12 - 1561\ M\^14 - 3481\ M\^16 - 
                12457\ M\^18 + 66757\ M\^20 - 36646\ M\^22 - 200651\ M\^24 + 
                307795\ M\^26 + 222102\ M\^28 - 791023\ M\^30 + 
                94465\ M\^32 + 1186677\ M\^34 - 717829\ M\^36 - 
                1136738\ M\^38 + 1275976\ M\^40 + 659202\ M\^42 - 
                1467015\ M\^44 - 91640\ M\^46 + 1366853\ M\^48 - 
                243214\ M\^50 - 1052191\ M\^52 + 427609\ M\^54 + 
                618371\ M\^56 - 474330\ M\^58 - 188508\ M\^60 + 
                333312\ M\^62 - 63914\ M\^64 - 107555\ M\^66 + 80246\ M\^68 - 
                11416\ M\^70 - 16700\ M\^72 + 15785\ M\^74 - 8295\ M\^76 + 
                2966\ M\^78 - 699\ M\^80 + 97\ M\^82 - 6\ M\^84)\) + 
          L\^17\ \((10 - 182\ M\^2 + 1473\ M\^4 - 6724\ M\^6 + 17657\ M\^8 - 
                20452\ M\^10 - 20400\ M\^12 + 96529\ M\^14 - 57324\ M\^16 - 
                220081\ M\^18 + 357673\ M\^20 + 268549\ M\^22 - 
                1004488\ M\^24 - 37679\ M\^26 + 2139808\ M\^28 - 
                930175\ M\^30 - 3607879\ M\^32 + 3175619\ M\^34 + 
                4698927\ M\^36 - 6832254\ M\^38 - 4188938\ M\^40 + 
                10730746\ M\^42 + 1439263\ M\^44 - 12593306\ M\^46 + 
                2478935\ M\^48 + 10849115\ M\^50 - 5091906\ M\^52 - 
                6300552\ M\^54 + 5196783\ M\^56 + 2074801\ M\^58 - 
                3441149\ M\^60 - 81037\ M\^62 + 1625630\ M\^64 - 
                240705\ M\^66 - 692034\ M\^68 + 166074\ M\^70 + 
                342891\ M\^72 - 178894\ M\^74 - 103270\ M\^76 + 
                125936\ M\^78 - 23473\ M\^80 - 31462\ M\^82 + 28631\ M\^84 - 
                12936\ M\^86 + 3969\ M\^88 - 912\ M\^90 + 157\ M\^92 - 
                18\ M\^94 + M\^96)\) + 
          L\^16\ \((\(-10\) + 210\ M\^2 - 1967\ M\^4 + 10526\ M\^6 - 
                33622\ M\^8 + 55453\ M\^10 + 7631\ M\^12 - 230309\ M\^14 + 
                346901\ M\^16 + 299210\ M\^18 - 1484340\ M\^20 + 
                806935\ M\^22 + 3127268\ M\^24 - 4698508\ M\^26 - 
                3412491\ M\^28 + 12248499\ M\^30 - 1138086\ M\^32 - 
                21287549\ M\^34 + 12730577\ M\^36 + 26722216\ M\^38 - 
                28971213\ M\^40 - 23841009\ M\^42 + 42580354\ M\^44 + 
                12674847\ M\^46 - 46524147\ M\^48 + 1788171\ M\^50 + 
                39400244\ M\^52 - 13299248\ M\^54 - 25626767\ M\^56 + 
                18696877\ M\^58 + 12135720\ M\^60 - 17089575\ M\^62 - 
                2680637\ M\^64 + 11031839\ M\^66 - 1768736\ M\^68 - 
                4636367\ M\^70 + 2133496\ M\^72 + 867868\ M\^74 - 
                865745\ M\^76 + 149250\ M\^78 - 1814\ M\^80 + 24427\ M\^82 + 
                49290\ M\^84 - 95504\ M\^86 + 49393\ M\^88 + 15575\ M\^90 - 
                38757\ M\^92 + 27936\ M\^94 - 11943\ M\^96 + 3324\ M\^98 - 
                596\ M\^100 + 63\ M\^102 - 3\ M\^104)\) + 
          L\^15\ \((5 - 118\ M\^2 + 1252\ M\^4 - 7754\ M\^6 + 30155\ M\^8 - 
                71020\ M\^10 + 70947\ M\^12 + 114902\ M\^14 - 475261\ M\^16 + 
                386658\ M\^18 + 965613\ M\^20 - 2444763\ M\^22 + 
                321759\ M\^24 + 5702771\ M\^26 - 6151400\ M\^28 - 
                6853237\ M\^30 + 17511079\ M\^32 + 23475\ M\^34 - 
                29692301\ M\^36 + 17939647\ M\^38 + 33492190\ M\^40 - 
                42929609\ M\^42 - 21397113\ M\^44 + 64408258\ M\^46 - 
                5124678\ M\^48 - 73271392\ M\^50 + 35255399\ M\^52 + 
                68114970\ M\^54 - 55644803\ M\^56 - 54701815\ M\^58 + 
                60010833\ M\^60 + 40832829\ M\^62 - 50394478\ M\^64 - 
                28319034\ M\^66 + 36418411\ M\^68 + 17025437\ M\^70 - 
                24602623\ M\^72 - 7460476\ M\^74 + 15460802\ M\^76 + 
                900555\ M\^78 - 7927188\ M\^80 + 1631942\ M\^82 + 
                2711676\ M\^84 - 1306249\ M\^86 - 356514\ M\^88 + 
                294806\ M\^90 + 7663\ M\^92 + 39492\ M\^94 - 71634\ M\^96 - 
                4067\ M\^98 + 52270\ M\^100 - 41568\ M\^102 + 17502\ M\^104 - 
                4569\ M\^106 + 748\ M\^108 - 71\ M\^110 + 3\ M\^112)\) + 
          L\^14\ \((\(-1\) + 26\ M\^2 - 313\ M\^4 + 2305\ M\^6 - 
                11485\ M\^8 + 39851\ M\^10 - 92320\ M\^12 + 110744\ M\^14 + 
                82649\ M\^16 - 600044\ M\^18 + 855439\ M\^20 + 
                633705\ M\^22 - 3866299\ M\^24 + 3546340\ M\^26 + 
                6593360\ M\^28 - 16783219\ M\^30 + 87870\ M\^32 + 
                39207079\ M\^34 - 31006540\ M\^36 - 58514181\ M\^38 + 
                95588481\ M\^40 + 52226615\ M\^42 - 185603181\ M\^44 - 
                1245850\ M\^46 + 271652810\ M\^48 - 92093272\ M\^50 - 
                316895359\ M\^52 + 199393487\ M\^54 + 297928957\ M\^56 - 
                280213863\ M\^58 - 219906869\ M\^60 + 305893398\ M\^62 + 
                114550868\ M\^64 - 272303075\ M\^66 - 19103664\ M\^68 + 
                202760614\ M\^70 - 38089723\ M\^72 - 126458993\ M\^74 + 
                52807343\ M\^76 + 65142141\ M\^78 - 41399123\ M\^80 - 
                27295831\ M\^82 + 23732989\ M\^84 + 9471411\ M\^86 - 
                11173281\ M\^88 - 2661739\ M\^90 + 4795921\ M\^92 + 
                282267\ M\^94 - 1936595\ M\^96 + 505782\ M\^98 + 
                462891\ M\^100 - 361636\ M\^102 + 35913\ M\^104 + 
                94381\ M\^106 - 79238\ M\^108 + 36367\ M\^110 - 
                11172\ M\^112 + 2353\ M\^114 - 327\ M\^116 + 27\ M\^118 - 
                M\^120)\) + 
          L\ \((\(-6\)\ M\^106 + 25\ M\^108 - 35\ M\^110 - 4\ M\^112 + 
                72\ M\^114 - 28\ M\^116 - 15\ M\^118 + 26\ M\^120 - 
                16\ M\^122 + 4\ M\^124)\) + 
          L\^13\ \((4\ M\^4 - 102\ M\^6 + 1167\ M\^8 - 7867\ M\^10 + 
                34121\ M\^12 - 95169\ M\^14 + 146863\ M\^16 + 1336\ M\^18 - 
                527204\ M\^20 + 932115\ M\^22 + 216347\ M\^24 - 
                3105449\ M\^26 + 3483548\ M\^28 + 3593384\ M\^30 - 
                11590438\ M\^32 + 3307544\ M\^34 + 19521777\ M\^36 - 
                21396839\ M\^38 - 15484461\ M\^40 + 43399940\ M\^42 - 
                11594421\ M\^44 - 49575154\ M\^46 + 62880584\ M\^48 + 
                16125786\ M\^50 - 124644468\ M\^52 + 67526932\ M\^54 + 
                174465438\ M\^56 - 187353301\ M\^58 - 192054238\ M\^60 + 
                306111440\ M\^62 + 169282525\ M\^64 - 381608794\ M\^66 - 
                112641112\ M\^68 + 391271597\ M\^70 + 44970892\ M\^72 - 
                335576432\ M\^74 + 13729841\ M\^76 + 241071736\ M\^78 - 
                49531315\ M\^80 - 142845073\ M\^82 + 58506118\ M\^84 + 
                65847857\ M\^86 - 47266639\ M\^88 - 19269906\ M\^90 + 
                28293885\ M\^92 - 1121503\ M\^94 - 11771355\ M\^96 + 
                5247248\ M\^98 + 2471989\ M\^100 - 3203634\ M\^102 + 
                659274\ M\^104 + 770691\ M\^106 - 602275\ M\^108 + 
                78769\ M\^110 + 136556\ M\^112 - 113219\ M\^114 + 
                48414\ M\^116 - 13592\ M\^118 + 2622\ M\^120 - 340\ M\^122 + 
                27\ M\^124 - M\^126)\) + 
          L\^12\ \((\(-6\)\ M\^8 + 151\ M\^10 - 1701\ M\^12 + 11265\ M\^14 - 
                47790\ M\^16 + 128770\ M\^18 - 181522\ M\^20 - 69790\ M\^22 + 
                827412\ M\^24 - 1203816\ M\^26 - 976365\ M\^28 + 
                5248363\ M\^30 - 3582101\ M\^32 - 10450318\ M\^34 + 
                18563595\ M\^36 + 10137073\ M\^38 - 47831901\ M\^40 + 
                6784121\ M\^42 + 87971299\ M\^44 - 46903737\ M\^46 - 
                129920946\ M\^48 + 102971494\ M\^50 + 165281033\ M\^52 - 
                148876541\ M\^54 - 195787058\ M\^56 + 151450456\ M\^58 + 
                232156666\ M\^60 - 92921907\ M\^62 - 283219316\ M\^64 - 
                10722591\ M\^66 + 342373624\ M\^68 + 114086307\ M\^70 - 
                388084136\ M\^72 - 165750417\ M\^74 + 402843943\ M\^76 + 
                149505848\ M\^78 - 373122408\ M\^80 - 84411141\ M\^82 + 
                300370810\ M\^84 + 10257595\ M\^86 - 203661260\ M\^88 + 
                38518830\ M\^90 + 110268242\ M\^92 - 50756161\ M\^94 - 
                42507688\ M\^96 + 38084823\ M\^98 + 6800185\ M\^100 - 
                18852724\ M\^102 + 4278495\ M\^104 + 5534163\ M\^106 - 
                3873897\ M\^108 - 111556\ M\^110 + 1230590\ M\^112 - 
                552801\ M\^114 - 55941\ M\^116 + 179936\ M\^118 - 
                103874\ M\^120 + 35224\ M\^122 - 7883\ M\^124 + 
                1155\ M\^126 - 101\ M\^128 + 4\ M\^130)\) + 
          L\^2\ \((15\ M\^94 - 118\ M\^96 + 358\ M\^98 - 404\ M\^100 - 
                301\ M\^102 + 1221\ M\^104 - 806\ M\^106 - 870\ M\^108 + 
                1579\ M\^110 - 31\ M\^112 - 556\ M\^114 + 236\ M\^116 - 
                235\ M\^118 + 202\ M\^120 + 61\ M\^122 - 192\ M\^124 + 
                132\ M\^126 - 44\ M\^128 + 6\ M\^130)\) + 
          L\^11\ \((4\ M\^12 - 101\ M\^14 + 1155\ M\^16 - 7883\ M\^18 + 
                35224\ M\^20 - 103874\ M\^22 + 179936\ M\^24 - 55941\ M\^26 - 
                552801\ M\^28 + 1230590\ M\^30 - 111556\ M\^32 - 
                3873897\ M\^34 + 5534163\ M\^36 + 4278495\ M\^38 - 
                18852724\ M\^40 + 6800185\ M\^42 + 38084823\ M\^44 - 
                42507688\ M\^46 - 50756161\ M\^48 + 110268242\ M\^50 + 
                38518830\ M\^52 - 203661260\ M\^54 + 10257595\ M\^56 + 
                300370810\ M\^58 - 84411141\ M\^60 - 373122408\ M\^62 + 
                149505848\ M\^64 + 402843943\ M\^66 - 165750417\ M\^68 - 
                388084136\ M\^70 + 114086307\ M\^72 + 342373624\ M\^74 - 
                10722591\ M\^76 - 283219316\ M\^78 - 92921907\ M\^80 + 
                232156666\ M\^82 + 151450456\ M\^84 - 195787058\ M\^86 - 
                148876541\ M\^88 + 165281033\ M\^90 + 102971494\ M\^92 - 
                129920946\ M\^94 - 46903737\ M\^96 + 87971299\ M\^98 + 
                6784121\ M\^100 - 47831901\ M\^102 + 10137073\ M\^104 + 
                18563595\ M\^106 - 10450318\ M\^108 - 3582101\ M\^110 + 
                5248363\ M\^112 - 976365\ M\^114 - 1203816\ M\^116 + 
                827412\ M\^118 - 69790\ M\^120 - 181522\ M\^122 + 
                128770\ M\^124 - 47790\ M\^126 + 11265\ M\^128 - 
                1701\ M\^130 + 151\ M\^132 - 6\ M\^134)\) + 
          L\^3\ \((\(-20\)\ M\^82 + 222\ M\^84 - 1005\ M\^86 + 2180\ M\^88 - 
                1523\ M\^90 - 2924\ M\^92 + 7090\ M\^94 - 4068\ M\^96 - 
                4173\ M\^98 + 9567\ M\^100 - 9570\ M\^102 + 746\ M\^104 + 
                14590\ M\^106 - 13547\ M\^108 - 114\ M\^110 + 12595\ M\^112 - 
                10699\ M\^114 - 1435\ M\^116 + 6719\ M\^118 - 2886\ M\^120 - 
                419\ M\^122 + 639\ M\^124 - 444\ M\^126 + 508\ M\^128 - 
                392\ M\^130 + 170\ M\^132 - 40\ M\^134 + 4\ M\^136)\) + 
          L\^10\ \((\(-M\^16\) + 27\ M\^18 - 340\ M\^20 + 2622\ M\^22 - 
                13592\ M\^24 + 48414\ M\^26 - 113219\ M\^28 + 136556\ M\^30 + 
                78769\ M\^32 - 602275\ M\^34 + 770691\ M\^36 + 
                659274\ M\^38 - 3203634\ M\^40 + 2471989\ M\^42 + 
                5247248\ M\^44 - 11771355\ M\^46 - 1121503\ M\^48 + 
                28293885\ M\^50 - 19269906\ M\^52 - 47266639\ M\^54 + 
                65847857\ M\^56 + 58506118\ M\^58 - 142845073\ M\^60 - 
                49531315\ M\^62 + 241071736\ M\^64 + 13729841\ M\^66 - 
                335576432\ M\^68 + 44970892\ M\^70 + 391271597\ M\^72 - 
                112641112\ M\^74 - 381608794\ M\^76 + 169282525\ M\^78 + 
                306111440\ M\^80 - 192054238\ M\^82 - 187353301\ M\^84 + 
                174465438\ M\^86 + 67526932\ M\^88 - 124644468\ M\^90 + 
                16125786\ M\^92 + 62880584\ M\^94 - 49575154\ M\^96 - 
                11594421\ M\^98 + 43399940\ M\^100 - 15484461\ M\^102 - 
                21396839\ M\^104 + 19521777\ M\^106 + 3307544\ M\^108 - 
                11590438\ M\^110 + 3593384\ M\^112 + 3483548\ M\^114 - 
                3105449\ M\^116 + 216347\ M\^118 + 932115\ M\^120 - 
                527204\ M\^122 + 1336\ M\^124 + 146863\ M\^126 - 
                95169\ M\^128 + 34121\ M\^130 - 7867\ M\^132 + 1167\ M\^134 - 
                102\ M\^136 + 4\ M\^138)\) + 
          L\^7\ \((\(-3\)\ M\^38 + 63\ M\^40 - 596\ M\^42 + 3324\ M\^44 - 
                11943\ M\^46 + 27936\ M\^48 - 38757\ M\^50 + 15575\ M\^52 + 
                49393\ M\^54 - 95504\ M\^56 + 49290\ M\^58 + 24427\ M\^60 - 
                1814\ M\^62 + 149250\ M\^64 - 865745\ M\^66 + 867868\ M\^68 + 
                2133496\ M\^70 - 4636367\ M\^72 - 1768736\ M\^74 + 
                11031839\ M\^76 - 2680637\ M\^78 - 17089575\ M\^80 + 
                12135720\ M\^82 + 18696877\ M\^84 - 25626767\ M\^86 - 
                13299248\ M\^88 + 39400244\ M\^90 + 1788171\ M\^92 - 
                46524147\ M\^94 + 12674847\ M\^96 + 42580354\ M\^98 - 
                23841009\ M\^100 - 28971213\ M\^102 + 26722216\ M\^104 + 
                12730577\ M\^106 - 21287549\ M\^108 - 1138086\ M\^110 + 
                12248499\ M\^112 - 3412491\ M\^114 - 4698508\ M\^116 + 
                3127268\ M\^118 + 806935\ M\^120 - 1484340\ M\^122 + 
                299210\ M\^124 + 346901\ M\^126 - 230309\ M\^128 + 
                7631\ M\^130 + 55453\ M\^132 - 33622\ M\^134 + 
                10526\ M\^136 - 1967\ M\^138 + 210\ M\^140 - 10\ M\^142)\) + 
          L\^5\ \((\(-6\)\ M\^58 + 97\ M\^60 - 699\ M\^62 + 2966\ M\^64 - 
                8295\ M\^66 + 15785\ M\^68 - 16700\ M\^70 - 11416\ M\^72 + 
                80246\ M\^74 - 107555\ M\^76 - 63914\ M\^78 + 333312\ M\^80 - 
                188508\ M\^82 - 474330\ M\^84 + 618371\ M\^86 + 
                427609\ M\^88 - 1052191\ M\^90 - 243214\ M\^92 + 
                1366853\ M\^94 - 91640\ M\^96 - 1467015\ M\^98 + 
                659202\ M\^100 + 1275976\ M\^102 - 1136738\ M\^104 - 
                717829\ M\^106 + 1186677\ M\^108 + 94465\ M\^110 - 
                791023\ M\^112 + 222102\ M\^114 + 307795\ M\^116 - 
                200651\ M\^118 - 36646\ M\^120 + 66757\ M\^122 - 
                12457\ M\^124 - 3481\ M\^126 - 1561\ M\^128 - 658\ M\^130 + 
                4759\ M\^132 - 4305\ M\^134 + 1951\ M\^136 - 513\ M\^138 + 
                76\ M\^140 - 5\ M\^142)\) + 
          L\^9\ \((\(-M\^22\) + 27\ M\^24 - 327\ M\^26 + 2353\ M\^28 - 
                11172\ M\^30 + 36367\ M\^32 - 79238\ M\^34 + 94381\ M\^36 + 
                35913\ M\^38 - 361636\ M\^40 + 462891\ M\^42 + 
                505782\ M\^44 - 1936595\ M\^46 + 282267\ M\^48 + 
                4795921\ M\^50 - 2661739\ M\^52 - 11173281\ M\^54 + 
                9471411\ M\^56 + 23732989\ M\^58 - 27295831\ M\^60 - 
                41399123\ M\^62 + 65142141\ M\^64 + 52807343\ M\^66 - 
                126458993\ M\^68 - 38089723\ M\^70 + 202760614\ M\^72 - 
                19103664\ M\^74 - 272303075\ M\^76 + 114550868\ M\^78 + 
                305893398\ M\^80 - 219906869\ M\^82 - 280213863\ M\^84 + 
                297928957\ M\^86 + 199393487\ M\^88 - 316895359\ M\^90 - 
                92093272\ M\^92 + 271652810\ M\^94 - 1245850\ M\^96 - 
                185603181\ M\^98 + 52226615\ M\^100 + 95588481\ M\^102 - 
                58514181\ M\^104 - 31006540\ M\^106 + 39207079\ M\^108 + 
                87870\ M\^110 - 16783219\ M\^112 + 6593360\ M\^114 + 
                3546340\ M\^116 - 3866299\ M\^118 + 633705\ M\^120 + 
                855439\ M\^122 - 600044\ M\^124 + 82649\ M\^126 + 
                110744\ M\^128 - 92320\ M\^130 + 39851\ M\^132 - 
                11485\ M\^134 + 2305\ M\^136 - 313\ M\^138 + 26\ M\^140 - 
                M\^142)\) + 
          L\^4\ \((15\ M\^70 - 208\ M\^72 + 1227\ M\^74 - 3888\ M\^76 + 
                6697\ M\^78 - 4289\ M\^80 - 7167\ M\^82 + 23779\ M\^84 - 
                31289\ M\^86 - 1091\ M\^88 + 79812\ M\^90 - 101362\ M\^92 - 
                40327\ M\^94 + 195486\ M\^96 - 97123\ M\^98 - 
                176102\ M\^100 + 209819\ M\^102 + 84858\ M\^104 - 
                189986\ M\^106 - 12547\ M\^108 + 114004\ M\^110 - 
                3047\ M\^112 - 64842\ M\^114 + 5935\ M\^116 + 41491\ M\^118 - 
                16683\ M\^120 - 15969\ M\^122 + 14997\ M\^124 - 
                1096\ M\^126 - 4378\ M\^128 + 3028\ M\^130 - 1244\ M\^132 + 
                496\ M\^134 - 204\ M\^136 + 64\ M\^138 - 12\ M\^140 + 
                M\^142)\) + 
          L\^8\ \((3\ M\^30 - 71\ M\^32 + 748\ M\^34 - 4569\ M\^36 + 
                17502\ M\^38 - 41568\ M\^40 + 52270\ M\^42 - 4067\ M\^44 - 
                71634\ M\^46 + 39492\ M\^48 + 7663\ M\^50 + 294806\ M\^52 - 
                356514\ M\^54 - 1306249\ M\^56 + 2711676\ M\^58 + 
                1631942\ M\^60 - 7927188\ M\^62 + 900555\ M\^64 + 
                15460802\ M\^66 - 7460476\ M\^68 - 24602623\ M\^70 + 
                17025437\ M\^72 + 36418411\ M\^74 - 28319034\ M\^76 - 
                50394478\ M\^78 + 40832829\ M\^80 + 60010833\ M\^82 - 
                54701815\ M\^84 - 55644803\ M\^86 + 68114970\ M\^88 + 
                35255399\ M\^90 - 73271392\ M\^92 - 5124678\ M\^94 + 
                64408258\ M\^96 - 21397113\ M\^98 - 42929609\ M\^100 + 
                33492190\ M\^102 + 17939647\ M\^104 - 29692301\ M\^106 + 
                23475\ M\^108 + 17511079\ M\^110 - 6853237\ M\^112 - 
                6151400\ M\^114 + 5702771\ M\^116 + 321759\ M\^118 - 
                2444763\ M\^120 + 965613\ M\^122 + 386658\ M\^124 - 
                475261\ M\^126 + 114902\ M\^128 + 70947\ M\^130 - 
                71020\ M\^132 + 30155\ M\^134 - 7754\ M\^136 + 1252\ M\^138 - 
                118\ M\^140 + 5\ M\^142)\) + 
          L\^6\ \((M\^46 - 18\ M\^48 + 157\ M\^50 - 912\ M\^52 + 
                3969\ M\^54 - 12936\ M\^56 + 28631\ M\^58 - 31462\ M\^60 - 
                23473\ M\^62 + 125936\ M\^64 - 103270\ M\^66 - 
                178894\ M\^68 + 342891\ M\^70 + 166074\ M\^72 - 
                692034\ M\^74 - 240705\ M\^76 + 1625630\ M\^78 - 
                81037\ M\^80 - 3441149\ M\^82 + 2074801\ M\^84 + 
                5196783\ M\^86 - 6300552\ M\^88 - 5091906\ M\^90 + 
                10849115\ M\^92 + 2478935\ M\^94 - 12593306\ M\^96 + 
                1439263\ M\^98 + 10730746\ M\^100 - 4188938\ M\^102 - 
                6832254\ M\^104 + 4698927\ M\^106 + 3175619\ M\^108 - 
                3607879\ M\^110 - 930175\ M\^112 + 2139808\ M\^114 - 
                37679\ M\^116 - 1004488\ M\^118 + 268549\ M\^120 + 
                357673\ M\^122 - 220081\ M\^124 - 57324\ M\^126 + 
                96529\ M\^128 - 20400\ M\^130 - 20452\ M\^132 + 
                17657\ M\^134 - 6724\ M\^136 + 1473\ M\^138 - 182\ M\^140 + 
                10\ M\^142)\)}, {"A-polynomial", Knot[9, 27], 
        M\^64 + L\^24\ M\^64 + 
          L\ \((4\ M\^54 - 18\ M\^56 + 31\ M\^58 - 15\ M\^60 - 36\ M\^62 + 
                78\ M\^64 - M\^66 - 39\ M\^68 + 26\ M\^70 - 6\ M\^72)\) + 
          L\^23\ \((\(-6\)\ M\^56 + 26\ M\^58 - 39\ M\^60 - M\^62 + 
                78\ M\^64 - 36\ M\^66 - 15\ M\^68 + 31\ M\^70 - 18\ M\^72 + 
                4\ M\^74)\) + 
          L\^2\ \((6\ M\^44 - 52\ M\^46 + 193\ M\^48 - 332\ M\^50 + 
                67\ M\^52 + 607\ M\^54 - 663\ M\^56 - 67\ M\^58 + 
                187\ M\^60 - 179\ M\^62 + 1160\ M\^64 - 587\ M\^66 - 
                801\ M\^68 + 1178\ M\^70 - 269\ M\^72 - 446\ M\^74 + 
                381\ M\^76 - 122\ M\^78 + 15\ M\^80)\) + 
          L\^22\ \((15\ M\^48 - 122\ M\^50 + 381\ M\^52 - 446\ M\^54 - 
                269\ M\^56 + 1178\ M\^58 - 801\ M\^60 - 587\ M\^62 + 
                1160\ M\^64 - 179\ M\^66 + 187\ M\^68 - 67\ M\^70 - 
                663\ M\^72 + 607\ M\^74 + 67\ M\^76 - 332\ M\^78 + 
                193\ M\^80 - 52\ M\^82 + 6\ M\^84)\) + 
          L\^3\ \((4\ M\^34 - 50\ M\^36 + 278\ M\^38 - 846\ M\^40 + 
                1359\ M\^42 - 553\ M\^44 - 2019\ M\^46 + 3963\ M\^48 - 
                2302\ M\^50 - 2351\ M\^52 + 5606\ M\^54 - 4034\ M\^56 + 
                1273\ M\^58 - 64\ M\^60 - 6972\ M\^62 + 13658\ M\^64 + 
                145\ M\^66 - 12113\ M\^68 + 10233\ M\^70 - 1532\ M\^72 - 
                5896\ M\^74 + 6938\ M\^76 - 2280\ M\^78 - 1939\ M\^80 + 
                2367\ M\^82 - 1057\ M\^84 + 228\ M\^86 - 20\ M\^88)\) + 
          L\^21\ \((\(-20\)\ M\^40 + 228\ M\^42 - 1057\ M\^44 + 2367\ M\^46 - 
                1939\ M\^48 - 2280\ M\^50 + 6938\ M\^52 - 5896\ M\^54 - 
                1532\ M\^56 + 10233\ M\^58 - 12113\ M\^60 + 145\ M\^62 + 
                13658\ M\^64 - 6972\ M\^66 - 64\ M\^68 + 1273\ M\^70 - 
                4034\ M\^72 + 5606\ M\^74 - 2351\ M\^76 - 2302\ M\^78 + 
                3963\ M\^80 - 2019\ M\^82 - 553\ M\^84 + 1359\ M\^86 - 
                846\ M\^88 + 278\ M\^90 - 50\ M\^92 + 4\ M\^94)\) + 
          L\^4\ \((M\^24 - 16\ M\^26 + 120\ M\^28 - 552\ M\^30 + 
                1718\ M\^32 - 3684\ M\^34 + 4746\ M\^36 + 58\ M\^38 - 
                15114\ M\^40 + 28458\ M\^42 - 3248\ M\^44 - 69098\ M\^46 + 
                84651\ M\^48 + 66191\ M\^50 - 201690\ M\^52 + 653\ M\^54 + 
                301856\ M\^56 - 101545\ M\^58 - 349859\ M\^60 + 
                201055\ M\^62 + 315743\ M\^64 - 263945\ M\^66 - 
                162285\ M\^68 + 256893\ M\^70 - 11281\ M\^72 - 
                144727\ M\^74 + 83473\ M\^76 + 16717\ M\^78 - 43547\ M\^80 + 
                24932\ M\^82 - 3775\ M\^84 - 6986\ M\^86 + 7824\ M\^88 - 
                4192\ M\^90 + 1278\ M\^92 - 212\ M\^94 + 15\ M\^96)\) + 
          L\^5\ \((3\ M\^18 - 55\ M\^20 + 451\ M\^22 - 2140\ M\^24 + 
                6256\ M\^26 - 10484\ M\^28 + 4785\ M\^30 + 19350\ M\^32 - 
                40563\ M\^34 + 7373\ M\^36 + 74443\ M\^38 - 74523\ M\^40 - 
                71359\ M\^42 + 116563\ M\^44 + 125631\ M\^46 - 
                198170\ M\^48 - 287523\ M\^50 + 497468\ M\^52 + 
                365458\ M\^54 - 1016184\ M\^56 - 88391\ M\^58 + 
                1404123\ M\^60 - 412490\ M\^62 - 1409064\ M\^64 + 
                738182\ M\^66 + 1249518\ M\^68 - 667466\ M\^70 - 
                1037967\ M\^72 + 553325\ M\^74 + 738223\ M\^76 - 
                528412\ M\^78 - 322948\ M\^80 + 412665\ M\^82 - 
                15398\ M\^84 - 166744\ M\^86 + 91676\ M\^88 + 1014\ M\^90 - 
                27296\ M\^92 + 20004\ M\^94 - 9342\ M\^96 + 3137\ M\^98 - 
                717\ M\^100 + 98\ M\^102 - 6\ M\^104)\) + 
          L\^20\ \((15\ M\^32 - 212\ M\^34 + 1278\ M\^36 - 4192\ M\^38 + 
                7824\ M\^40 - 6986\ M\^42 - 3775\ M\^44 + 24932\ M\^46 - 
                43547\ M\^48 + 16717\ M\^50 + 83473\ M\^52 - 144727\ M\^54 - 
                11281\ M\^56 + 256893\ M\^58 - 162285\ M\^60 - 
                263945\ M\^62 + 315743\ M\^64 + 201055\ M\^66 - 
                349859\ M\^68 - 101545\ M\^70 + 301856\ M\^72 + 653\ M\^74 - 
                201690\ M\^76 + 66191\ M\^78 + 84651\ M\^80 - 69098\ M\^82 - 
                3248\ M\^84 + 28458\ M\^86 - 15114\ M\^88 + 58\ M\^90 + 
                4746\ M\^92 - 3684\ M\^94 + 1718\ M\^96 - 552\ M\^98 + 
                120\ M\^100 - 16\ M\^102 + M\^104)\) + 
          L\^19\ \((\(-6\)\ M\^24 + 98\ M\^26 - 717\ M\^28 + 3137\ M\^30 - 
                9342\ M\^32 + 20004\ M\^34 - 27296\ M\^36 + 1014\ M\^38 + 
                91676\ M\^40 - 166744\ M\^42 - 15398\ M\^44 + 412665\ M\^46 - 
                322948\ M\^48 - 528412\ M\^50 + 738223\ M\^52 + 
                553325\ M\^54 - 1037967\ M\^56 - 667466\ M\^58 + 
                1249518\ M\^60 + 738182\ M\^62 - 1409064\ M\^64 - 
                412490\ M\^66 + 1404123\ M\^68 - 88391\ M\^70 - 
                1016184\ M\^72 + 365458\ M\^74 + 497468\ M\^76 - 
                287523\ M\^78 - 198170\ M\^80 + 125631\ M\^82 + 
                116563\ M\^84 - 71359\ M\^86 - 74523\ M\^88 + 74443\ M\^90 + 
                7373\ M\^92 - 40563\ M\^94 + 19350\ M\^96 + 4785\ M\^98 - 
                10484\ M\^100 + 6256\ M\^102 - 2140\ M\^104 + 451\ M\^106 - 
                55\ M\^108 + 3\ M\^110)\) + 
          L\^6\ \((3\ M\^12 - 62\ M\^14 + 574\ M\^16 - 3095\ M\^18 + 
                10443\ M\^20 - 21203\ M\^22 + 17864\ M\^24 + 26878\ M\^26 - 
                93387\ M\^28 + 69011\ M\^30 + 130602\ M\^32 - 331088\ M\^34 + 
                166146\ M\^36 + 452298\ M\^38 - 962745\ M\^40 + 
                263173\ M\^42 + 1973861\ M\^44 - 2756254\ M\^46 - 
                1953545\ M\^48 + 7169743\ M\^50 - 833280\ M\^52 - 
                11786831\ M\^54 + 6961663\ M\^56 + 13273680\ M\^58 - 
                14011011\ M\^60 - 9514584\ M\^62 + 17626876\ M\^64 + 
                2289587\ M\^66 - 15114274\ M\^68 + 4065545\ M\^70 + 
                8761384\ M\^72 - 5892152\ M\^74 - 2865598\ M\^76 + 
                3842194\ M\^78 + 87908\ M\^80 - 1224122\ M\^82 + 
                143920\ M\^84 + 73232\ M\^86 + 224798\ M\^88 - 66996\ M\^90 - 
                180639\ M\^92 + 141590\ M\^94 + 216\ M\^96 - 54318\ M\^98 + 
                38379\ M\^100 - 15278\ M\^102 + 4280\ M\^104 - 930\ M\^106 + 
                157\ M\^108 - 18\ M\^110 + M\^112)\) + 
          L\^7\ \((M\^6 - 24\ M\^8 + 262\ M\^10 - 1712\ M\^12 + 7368\ M\^14 - 
                21361\ M\^16 + 39168\ M\^18 - 29177\ M\^20 - 58783\ M\^22 + 
                193876\ M\^24 - 130555\ M\^26 - 381960\ M\^28 + 
                825368\ M\^30 + 112467\ M\^32 - 1991745\ M\^34 + 
                951908\ M\^36 + 3878761\ M\^38 - 3524451\ M\^40 - 
                6996810\ M\^42 + 9424211\ M\^44 + 10958825\ M\^46 - 
                21294683\ M\^48 - 12178126\ M\^50 + 38428162\ M\^52 + 
                5859762\ M\^54 - 53844079\ M\^56 + 8078629\ M\^58 + 
                58097290\ M\^60 - 21987977\ M\^62 - 48440941\ M\^64 + 
                27127890\ M\^66 + 32248544\ M\^68 - 21591421\ M\^70 - 
                18029744\ M\^72 + 13067962\ M\^74 + 9685195\ M\^76 - 
                7934207\ M\^78 - 5177971\ M\^80 + 6035211\ M\^82 + 
                1681336\ M\^84 - 4307715\ M\^86 + 716699\ M\^88 + 
                1946650\ M\^90 - 1245839\ M\^92 - 166848\ M\^94 + 
                534994\ M\^96 - 257923\ M\^98 + 2283\ M\^100 + 
                80846\ M\^102 - 70213\ M\^104 + 37756\ M\^106 - 
                14172\ M\^108 + 3690\ M\^110 - 635\ M\^112 + 65\ M\^114 - 
                3\ M\^116)\) + 
          L\^18\ \((M\^16 - 18\ M\^18 + 157\ M\^20 - 930\ M\^22 + 
                4280\ M\^24 - 15278\ M\^26 + 38379\ M\^28 - 54318\ M\^30 + 
                216\ M\^32 + 141590\ M\^34 - 180639\ M\^36 - 66996\ M\^38 + 
                224798\ M\^40 + 73232\ M\^42 + 143920\ M\^44 - 
                1224122\ M\^46 + 87908\ M\^48 + 3842194\ M\^50 - 
                2865598\ M\^52 - 5892152\ M\^54 + 8761384\ M\^56 + 
                4065545\ M\^58 - 15114274\ M\^60 + 2289587\ M\^62 + 
                17626876\ M\^64 - 9514584\ M\^66 - 14011011\ M\^68 + 
                13273680\ M\^70 + 6961663\ M\^72 - 11786831\ M\^74 - 
                833280\ M\^76 + 7169743\ M\^78 - 1953545\ M\^80 - 
                2756254\ M\^82 + 1973861\ M\^84 + 263173\ M\^86 - 
                962745\ M\^88 + 452298\ M\^90 + 166146\ M\^92 - 
                331088\ M\^94 + 130602\ M\^96 + 69011\ M\^98 - 
                93387\ M\^100 + 26878\ M\^102 + 17864\ M\^104 - 
                21203\ M\^106 + 10443\ M\^108 - 3095\ M\^110 + 574\ M\^112 - 
                62\ M\^114 + 3\ M\^116)\) + 
          L\^8\ \((\(-M\^2\) + 24\ M\^4 - 265\ M\^6 + 1776\ M\^8 - 
                7971\ M\^10 + 24687\ M\^12 - 51264\ M\^14 + 60262\ M\^16 + 
                3437\ M\^18 - 157944\ M\^20 + 300692\ M\^22 - 202866\ M\^24 - 
                475839\ M\^26 + 1764380\ M\^28 - 1738951\ M\^30 - 
                2953142\ M\^32 + 8560794\ M\^34 - 932782\ M\^36 - 
                18652631\ M\^38 + 15023410\ M\^40 + 24528887\ M\^42 - 
                37716763\ M\^44 - 17880337\ M\^46 + 59252330\ M\^48 - 
                2902746\ M\^50 - 68559882\ M\^52 + 30383141\ M\^54 + 
                63204865\ M\^56 - 55783345\ M\^58 - 49226998\ M\^60 + 
                75458787\ M\^62 + 32512826\ M\^64 - 88654982\ M\^66 - 
                12458713\ M\^68 + 93125220\ M\^70 - 8846578\ M\^72 - 
                81206998\ M\^74 + 26611326\ M\^76 + 53741933\ M\^78 - 
                33317280\ M\^80 - 23136974\ M\^82 + 26715732\ M\^84 + 
                2503407\ M\^86 - 13536250\ M\^88 + 4230636\ M\^90 + 
                3312636\ M\^92 - 2825696\ M\^94 + 519945\ M\^96 + 
                327657\ M\^98 - 404767\ M\^100 + 308392\ M\^102 - 
                94232\ M\^104 - 81778\ M\^106 + 114316\ M\^108 - 
                67424\ M\^110 + 24290\ M\^112 - 5705\ M\^114 + 860\ M\^116 - 
                76\ M\^118 + 3\ M\^120)\) + 
          L\^17\ \((\(-3\)\ M\^12 + 65\ M\^14 - 635\ M\^16 + 3690\ M\^18 - 
                14172\ M\^20 + 37756\ M\^22 - 70213\ M\^24 + 80846\ M\^26 + 
                2283\ M\^28 - 257923\ M\^30 + 534994\ M\^32 - 166848\ M\^34 - 
                1245839\ M\^36 + 1946650\ M\^38 + 716699\ M\^40 - 
                4307715\ M\^42 + 1681336\ M\^44 + 6035211\ M\^46 - 
                5177971\ M\^48 - 7934207\ M\^50 + 9685195\ M\^52 + 
                13067962\ M\^54 - 18029744\ M\^56 - 21591421\ M\^58 + 
                32248544\ M\^60 + 27127890\ M\^62 - 48440941\ M\^64 - 
                21987977\ M\^66 + 58097290\ M\^68 + 8078629\ M\^70 - 
                53844079\ M\^72 + 5859762\ M\^74 + 38428162\ M\^76 - 
                12178126\ M\^78 - 21294683\ M\^80 + 10958825\ M\^82 + 
                9424211\ M\^84 - 6996810\ M\^86 - 3524451\ M\^88 + 
                3878761\ M\^90 + 951908\ M\^92 - 1991745\ M\^94 + 
                112467\ M\^96 + 825368\ M\^98 - 381960\ M\^100 - 
                130555\ M\^102 + 193876\ M\^104 - 58783\ M\^106 - 
                29177\ M\^108 + 39168\ M\^110 - 21361\ M\^112 + 
                7368\ M\^114 - 1712\ M\^116 + 262\ M\^118 - 24\ M\^120 + 
                M\^122)\) + 
          L\^9\ \((\(-1\) + 27\ M\^2 - 345\ M\^4 + 2699\ M\^6 - 14028\ M\^8 + 
                48984\ M\^10 - 107443\ M\^12 + 102196\ M\^14 + 
                149236\ M\^16 - 561823\ M\^18 + 293540\ M\^20 + 
                1233070\ M\^22 - 1768218\ M\^24 - 1599475\ M\^26 + 
                3717152\ M\^28 + 3668747\ M\^30 - 7535006\ M\^32 - 
                11679410\ M\^34 + 19694072\ M\^36 + 27383292\ M\^38 - 
                50834436\ M\^40 - 40067756\ M\^42 + 101173950\ M\^44 + 
                27696102\ M\^46 - 146013994\ M\^48 + 27012766\ M\^50 + 
                139082707\ M\^52 - 113248586\ M\^54 - 45828773\ M\^56 + 
                188647721\ M\^58 - 119994038\ M\^60 - 204589097\ M\^62 + 
                291312909\ M\^64 + 144426498\ M\^66 - 387427774\ M\^68 - 
                35589679\ M\^70 + 374680812\ M\^72 - 61332008\ M\^74 - 
                276616441\ M\^76 + 103573086\ M\^78 + 153249067\ M\^80 - 
                90646548\ M\^82 - 59151946\ M\^84 + 52237187\ M\^86 + 
                12859055\ M\^88 - 18810509\ M\^90 - 1106023\ M\^92 + 
                3109536\ M\^94 + 1566996\ M\^96 - 106207\ M\^98 - 
                1997603\ M\^100 + 850898\ M\^102 + 691166\ M\^104 - 
                720966\ M\^106 + 111054\ M\^108 + 191806\ M\^110 - 
                165587\ M\^112 + 71311\ M\^114 - 19625\ M\^116 + 
                3611\ M\^118 - 434\ M\^120 + 31\ M\^122 - M\^124)\) + 
          L\^10\ \((4 - 106\ M\^2 + 1296\ M\^4 - 9519\ M\^6 + 45528\ M\^8 - 
                141026\ M\^10 + 244119\ M\^12 - 19492\ M\^14 - 
                947647\ M\^16 + 1700081\ M\^18 + 1007272\ M\^20 - 
                7124031\ M\^22 + 4532118\ M\^24 + 16291198\ M\^26 - 
                23765338\ M\^28 - 25565283\ M\^30 + 66056354\ M\^32 + 
                27108789\ M\^34 - 133885857\ M\^36 - 18725483\ M\^38 + 
                218378762\ M\^40 + 16434461\ M\^42 - 300785632\ M\^44 - 
                60389204\ M\^46 + 367056314\ M\^48 + 190323132\ M\^50 - 
                416155798\ M\^52 - 403434133\ M\^54 + 452907593\ M\^56 + 
                630199673\ M\^58 - 469505182\ M\^60 - 767864455\ M\^62 + 
                443563297\ M\^64 + 749860796\ M\^66 - 357290905\ M\^68 - 
                588012035\ M\^70 + 231097000\ M\^72 + 371405173\ M\^74 - 
                101376263\ M\^76 - 190390849\ M\^78 + 2977407\ M\^80 + 
                84261943\ M\^82 + 46979795\ M\^84 - 42313197\ M\^86 - 
                53171919\ M\^88 + 31796803\ M\^90 + 35227534\ M\^92 - 
                26778964\ M\^94 - 13459809\ M\^96 + 17633969\ M\^98 + 
                452304\ M\^100 - 7707437\ M\^102 + 2990982\ M\^104 + 
                1390638\ M\^106 - 1600390\ M\^108 + 367399\ M\^110 + 
                262736\ M\^112 - 264058\ M\^114 + 119033\ M\^116 - 
                33971\ M\^118 + 6488\ M\^120 - 810\ M\^122 + 60\ M\^124 - 
                2\ M\^126)\) + 
          L\^16\ \((3\ M\^8 - 76\ M\^10 + 860\ M\^12 - 5705\ M\^14 + 
                24290\ M\^16 - 67424\ M\^18 + 114316\ M\^20 - 81778\ M\^22 - 
                94232\ M\^24 + 308392\ M\^26 - 404767\ M\^28 + 
                327657\ M\^30 + 519945\ M\^32 - 2825696\ M\^34 + 
                3312636\ M\^36 + 4230636\ M\^38 - 13536250\ M\^40 + 
                2503407\ M\^42 + 26715732\ M\^44 - 23136974\ M\^46 - 
                33317280\ M\^48 + 53741933\ M\^50 + 26611326\ M\^52 - 
                81206998\ M\^54 - 8846578\ M\^56 + 93125220\ M\^58 - 
                12458713\ M\^60 - 88654982\ M\^62 + 32512826\ M\^64 + 
                75458787\ M\^66 - 49226998\ M\^68 - 55783345\ M\^70 + 
                63204865\ M\^72 + 30383141\ M\^74 - 68559882\ M\^76 - 
                2902746\ M\^78 + 59252330\ M\^80 - 17880337\ M\^82 - 
                37716763\ M\^84 + 24528887\ M\^86 + 15023410\ M\^88 - 
                18652631\ M\^90 - 932782\ M\^92 + 8560794\ M\^94 - 
                2953142\ M\^96 - 1738951\ M\^98 + 1764380\ M\^100 - 
                475839\ M\^102 - 202866\ M\^104 + 300692\ M\^106 - 
                157944\ M\^108 + 3437\ M\^110 + 60262\ M\^112 - 
                51264\ M\^114 + 24687\ M\^116 - 7971\ M\^118 + 1776\ M\^120 - 
                265\ M\^122 + 24\ M\^124 - M\^126)\) + 
          L\^13\ \((\(-1\) + 35\ M\^2 - 547\ M\^4 + 5045\ M\^6 - 
                30314\ M\^8 + 121756\ M\^10 - 311581\ M\^12 + 382536\ M\^14 + 
                382240\ M\^16 - 2264506\ M\^18 + 2363041\ M\^20 + 
                4518502\ M\^22 - 13270776\ M\^24 + 844902\ M\^26 + 
                34835144\ M\^28 - 27409789\ M\^30 - 63875274\ M\^32 + 
                89534526\ M\^34 + 92684079\ M\^36 - 191313808\ M\^38 - 
                122683594\ M\^40 + 322180098\ M\^42 + 173662649\ M\^44 - 
                451488692\ M\^46 - 284215258\ M\^48 + 529368869\ M\^50 + 
                492954499\ M\^52 - 496503508\ M\^54 - 813514837\ M\^56 + 
                318190626\ M\^58 + 1211208754\ M\^60 - 22457611\ M\^62 - 
                1604569902\ M\^64 - 296112401\ M\^66 + 1891888818\ M\^68 + 
                531213838\ M\^70 - 1976144624\ M\^72 - 602561157\ M\^74 + 
                1828279014\ M\^76 + 507153194\ M\^78 - 1498325044\ M\^80 - 
                313672048\ M\^82 + 1083458291\ M\^84 + 116577128\ M\^86 - 
                683673285\ M\^88 + 16609599\ M\^90 + 367060348\ M\^92 - 
                68442866\ M\^94 - 159829715\ M\^96 + 63251578\ M\^98 + 
                50311170\ M\^100 - 36716935\ M\^102 - 7360274\ M\^104 + 
                14258081\ M\^106 - 2647654\ M\^108 - 2942062\ M\^110 + 
                1822204\ M\^112 - 75227\ M\^114 - 380437\ M\^116 + 
                230682\ M\^118 - 75040\ M\^120 + 15593\ M\^122 - 
                2088\ M\^124 + 166\ M\^126 - 6\ M\^128)\) + 
          L\^15\ \((\(-M\^4\) + 31\ M\^6 - 434\ M\^8 + 3611\ M\^10 - 
                19625\ M\^12 + 71311\ M\^14 - 165587\ M\^16 + 191806\ M\^18 + 
                111054\ M\^20 - 720966\ M\^22 + 691166\ M\^24 + 
                850898\ M\^26 - 1997603\ M\^28 - 106207\ M\^30 + 
                1566996\ M\^32 + 3109536\ M\^34 - 1106023\ M\^36 - 
                18810509\ M\^38 + 12859055\ M\^40 + 52237187\ M\^42 - 
                59151946\ M\^44 - 90646548\ M\^46 + 153249067\ M\^48 + 
                103573086\ M\^50 - 276616441\ M\^52 - 61332008\ M\^54 + 
                374680812\ M\^56 - 35589679\ M\^58 - 387427774\ M\^60 + 
                144426498\ M\^62 + 291312909\ M\^64 - 204589097\ M\^66 - 
                119994038\ M\^68 + 188647721\ M\^70 - 45828773\ M\^72 - 
                113248586\ M\^74 + 139082707\ M\^76 + 27012766\ M\^78 - 
                146013994\ M\^80 + 27696102\ M\^82 + 101173950\ M\^84 - 
                40067756\ M\^86 - 50834436\ M\^88 + 27383292\ M\^90 + 
                19694072\ M\^92 - 11679410\ M\^94 - 7535006\ M\^96 + 
                3668747\ M\^98 + 3717152\ M\^100 - 1599475\ M\^102 - 
                1768218\ M\^104 + 1233070\ M\^106 + 293540\ M\^108 - 
                561823\ M\^110 + 149236\ M\^112 + 102196\ M\^114 - 
                107443\ M\^116 + 48984\ M\^118 - 14028\ M\^120 + 
                2699\ M\^122 - 345\ M\^124 + 27\ M\^126 - M\^128)\) + 
          L\^11\ \((\(-6\) + 166\ M\^2 - 2088\ M\^4 + 15593\ M\^6 - 
                75040\ M\^8 + 230682\ M\^10 - 380437\ M\^12 - 75227\ M\^14 + 
                1822204\ M\^16 - 2942062\ M\^18 - 2647654\ M\^20 + 
                14258081\ M\^22 - 7360274\ M\^24 - 36716935\ M\^26 + 
                50311170\ M\^28 + 63251578\ M\^30 - 159829715\ M\^32 - 
                68442866\ M\^34 + 367060348\ M\^36 + 16609599\ M\^38 - 
                683673285\ M\^40 + 116577128\ M\^42 + 1083458291\ M\^44 - 
                313672048\ M\^46 - 1498325044\ M\^48 + 507153194\ M\^50 + 
                1828279014\ M\^52 - 602561157\ M\^54 - 1976144624\ M\^56 + 
                531213838\ M\^58 + 1891888818\ M\^60 - 296112401\ M\^62 - 
                1604569902\ M\^64 - 22457611\ M\^66 + 1211208754\ M\^68 + 
                318190626\ M\^70 - 813514837\ M\^72 - 496503508\ M\^74 + 
                492954499\ M\^76 + 529368869\ M\^78 - 284215258\ M\^80 - 
                451488692\ M\^82 + 173662649\ M\^84 + 322180098\ M\^86 - 
                122683594\ M\^88 - 191313808\ M\^90 + 92684079\ M\^92 + 
                89534526\ M\^94 - 63875274\ M\^96 - 27409789\ M\^98 + 
                34835144\ M\^100 + 844902\ M\^102 - 13270776\ M\^104 + 
                4518502\ M\^106 + 2363041\ M\^108 - 2264506\ M\^110 + 
                382240\ M\^112 + 382536\ M\^114 - 311581\ M\^116 + 
                121756\ M\^118 - 30314\ M\^120 + 5045\ M\^122 - 547\ M\^124 + 
                35\ M\^126 - M\^128)\) + 
          L\^12\ \((4 - 119\ M\^2 + 1600\ M\^4 - 12775\ M\^6 + 66126\ M\^8 - 
                222610\ M\^10 + 431587\ M\^12 - 141254\ M\^14 - 
                1572947\ M\^16 + 3484857\ M\^18 + 670418\ M\^20 - 
                13558981\ M\^22 + 13436243\ M\^24 + 28265688\ M\^26 - 
                60320596\ M\^28 - 34334666\ M\^30 + 165970983\ M\^32 + 
                5871231\ M\^34 - 356576348\ M\^36 + 80894971\ M\^38 + 
                661492367\ M\^40 - 230241934\ M\^42 - 1111496325\ M\^44 + 
                410888430\ M\^46 + 1724270343\ M\^48 - 558416255\ M\^50 - 
                2465408960\ M\^52 + 593745631\ M\^54 + 3218024858\ M\^56 - 
                463170826\ M\^58 - 3795704280\ M\^60 + 176948612\ M\^62 + 
                4016134010\ M\^64 + 176948612\ M\^66 - 3795704280\ M\^68 - 
                463170826\ M\^70 + 3218024858\ M\^72 + 593745631\ M\^74 - 
                2465408960\ M\^76 - 558416255\ M\^78 + 1724270343\ M\^80 + 
                410888430\ M\^82 - 1111496325\ M\^84 - 230241934\ M\^86 + 
                661492367\ M\^88 + 80894971\ M\^90 - 356576348\ M\^92 + 
                5871231\ M\^94 + 165970983\ M\^96 - 34334666\ M\^98 - 
                60320596\ M\^100 + 28265688\ M\^102 + 13436243\ M\^104 - 
                13558981\ M\^106 + 670418\ M\^108 + 3484857\ M\^110 - 
                1572947\ M\^112 - 141254\ M\^114 + 431587\ M\^116 - 
                222610\ M\^118 + 66126\ M\^120 - 12775\ M\^122 + 
                1600\ M\^124 - 119\ M\^126 + 4\ M\^128)\) + 
          L\^14\ \((\(-2\)\ M\^2 + 60\ M\^4 - 810\ M\^6 + 6488\ M\^8 - 
                33971\ M\^10 + 119033\ M\^12 - 264058\ M\^14 + 
                262736\ M\^16 + 367399\ M\^18 - 1600390\ M\^20 + 
                1390638\ M\^22 + 2990982\ M\^24 - 7707437\ M\^26 + 
                452304\ M\^28 + 17633969\ M\^30 - 13459809\ M\^32 - 
                26778964\ M\^34 + 35227534\ M\^36 + 31796803\ M\^38 - 
                53171919\ M\^40 - 42313197\ M\^42 + 46979795\ M\^44 + 
                84261943\ M\^46 + 2977407\ M\^48 - 190390849\ M\^50 - 
                101376263\ M\^52 + 371405173\ M\^54 + 231097000\ M\^56 - 
                588012035\ M\^58 - 357290905\ M\^60 + 749860796\ M\^62 + 
                443563297\ M\^64 - 767864455\ M\^66 - 469505182\ M\^68 + 
                630199673\ M\^70 + 452907593\ M\^72 - 403434133\ M\^74 - 
                416155798\ M\^76 + 190323132\ M\^78 + 367056314\ M\^80 - 
                60389204\ M\^82 - 300785632\ M\^84 + 16434461\ M\^86 + 
                218378762\ M\^88 - 18725483\ M\^90 - 133885857\ M\^92 + 
                27108789\ M\^94 + 66056354\ M\^96 - 25565283\ M\^98 - 
                23765338\ M\^100 + 16291198\ M\^102 + 4532118\ M\^104 - 
                7124031\ M\^106 + 1007272\ M\^108 + 1700081\ M\^110 - 
                947647\ M\^112 - 19492\ M\^114 + 244119\ M\^116 - 
                141026\ M\^118 + 45528\ M\^120 - 9519\ M\^122 + 
                1296\ M\^124 - 106\ M\^126 + 4\ M\^128)\)}, {"A-polynomial", 
        Knot[9, 28], 
        L\^18\ M\^28 + M\^88 + 
          L\^17\ \((3\ M\^22 - 15\ M\^24 + 27\ M\^26 - 5\ M\^28 - 56\ M\^30 + 
                81\ M\^32 + 18\ M\^34 - 55\ M\^36 + 24\ M\^38 - 4\ M\^40)\) + 
          L\^16\ \((3\ M\^16 - 30\ M\^18 + 126\ M\^20 - 249\ M\^22 + 
                116\ M\^24 + 387\ M\^26 - 699\ M\^28 + 538\ M\^30 - 
                122\ M\^32 - 1006\ M\^34 + 1802\ M\^36 + 69\ M\^38 - 
                1620\ M\^40 + 911\ M\^42 + 237\ M\^44 - 524\ M\^46 + 
                272\ M\^48 - 64\ M\^50 + 6\ M\^52)\) + 
          L\^15\ \((M\^10 - 15\ M\^12 + 101\ M\^14 - 398\ M\^16 + 
                980\ M\^18 - 1311\ M\^20 - 264\ M\^22 + 4559\ M\^24 - 
                6457\ M\^26 - 1598\ M\^28 + 11342\ M\^30 - 4635\ M\^32 - 
                6556\ M\^34 + 222\ M\^36 + 437\ M\^38 + 14239\ M\^40 - 
                4219\ M\^42 - 16329\ M\^44 + 11479\ M\^46 + 4709\ M\^48 - 
                8825\ M\^50 + 3199\ M\^52 + 1401\ M\^54 - 2064\ M\^56 + 
                1100\ M\^58 - 334\ M\^60 + 56\ M\^62 - 4\ M\^64)\) + 
          L\^14\ \((\(-M\^6\) + 18\ M\^8 - 145\ M\^10 + 675\ M\^12 - 
                1962\ M\^14 + 3485\ M\^16 - 2564\ M\^18 - 5263\ M\^20 + 
                20284\ M\^22 - 22957\ M\^24 - 23629\ M\^26 + 91325\ M\^28 - 
                33256\ M\^30 - 163844\ M\^32 + 153425\ M\^34 + 
                192743\ M\^36 - 269071\ M\^38 - 183041\ M\^40 + 
                318704\ M\^42 + 150769\ M\^44 - 289101\ M\^46 - 
                82533\ M\^48 + 215362\ M\^50 + 557\ M\^52 - 119292\ M\^54 + 
                41866\ M\^56 + 34354\ M\^58 - 30971\ M\^60 + 3857\ M\^62 + 
                6886\ M\^64 - 5241\ M\^66 + 2104\ M\^68 - 588\ M\^70 + 
                120\ M\^72 - 16\ M\^74 + M\^76)\) + 
          L\^13\ \((\(-3\)\ M\^4 + 55\ M\^6 - 447\ M\^8 + 2048\ M\^10 - 
                5431\ M\^12 + 6675\ M\^14 + 4430\ M\^16 - 27063\ M\^18 + 
                26129\ M\^20 + 28285\ M\^22 - 77687\ M\^24 + 34016\ M\^26 + 
                61406\ M\^28 - 124076\ M\^30 + 135878\ M\^32 + 80960\ M\^34 - 
                498612\ M\^36 + 216962\ M\^38 + 881293\ M\^40 - 
                696332\ M\^42 - 1106504\ M\^44 + 1089007\ M\^46 + 
                1067444\ M\^48 - 1173471\ M\^50 - 717537\ M\^52 + 
                967834\ M\^54 + 249233\ M\^56 - 604532\ M\^58 + 
                66762\ M\^60 + 243870\ M\^62 - 132909\ M\^64 - 26060\ M\^66 + 
                61488\ M\^68 - 26939\ M\^70 - 3944\ M\^72 + 11284\ M\^74 - 
                6857\ M\^76 + 2347\ M\^78 - 489\ M\^80 + 58\ M\^82 - 
                3\ M\^84)\) + 
          L\^12\ \((\(-3\)\ M\^2 + 60\ M\^4 - 528\ M\^6 + 2600\ M\^8 - 
                7319\ M\^10 + 9097\ M\^12 + 8871\ M\^14 - 44776\ M\^16 + 
                26366\ M\^18 + 105915\ M\^20 - 160064\ M\^22 - 
                144085\ M\^24 + 419883\ M\^26 + 125467\ M\^28 - 
                805397\ M\^30 - 106082\ M\^32 + 1378000\ M\^34 + 
                188281\ M\^36 - 2268220\ M\^38 - 425605\ M\^40 + 
                3539213\ M\^42 + 730732\ M\^44 - 4881717\ M\^46 - 
                988306\ M\^48 + 5664086\ M\^50 + 989350\ M\^52 - 
                5332583\ M\^54 - 503548\ M\^56 + 4090968\ M\^58 - 
                244293\ M\^60 - 2516067\ M\^62 + 729961\ M\^64 + 
                1138031\ M\^66 - 718793\ M\^68 - 273889\ M\^70 + 
                409461\ M\^72 - 57931\ M\^74 - 119892\ M\^76 + 70930\ M\^78 + 
                2066\ M\^80 - 20318\ M\^82 + 11577\ M\^84 - 3510\ M\^86 + 
                638\ M\^88 - 66\ M\^90 + 3\ M\^92)\) + 
          L\ \((\(-4\)\ M\^76 + 24\ M\^78 - 55\ M\^80 + 18\ M\^82 + 
                81\ M\^84 - 56\ M\^86 - 5\ M\^88 + 27\ M\^90 - 15\ M\^92 + 
                3\ M\^94)\) + 
          L\^11\ \((\(-1\) + 22\ M\^2 - 213\ M\^4 + 1174\ M\^6 - 3921\ M\^8 + 
                7416\ M\^10 - 4215\ M\^12 - 13589\ M\^14 + 29445\ M\^16 + 
                900\ M\^18 - 77940\ M\^20 + 93886\ M\^22 + 61349\ M\^24 - 
                267551\ M\^26 + 134210\ M\^28 + 436030\ M\^30 - 
                526873\ M\^32 - 552477\ M\^34 + 919828\ M\^36 + 
                879132\ M\^38 - 984385\ M\^40 - 2088086\ M\^42 + 
                542447\ M\^44 + 4729988\ M\^46 + 431142\ M\^48 - 
                8534659\ M\^50 - 1745322\ M\^52 + 11950195\ M\^54 + 
                2859252\ M\^56 - 13091886\ M\^58 - 2771152\ M\^60 + 
                11532800\ M\^62 + 1433716\ M\^64 - 8321163\ M\^66 + 
                141299\ M\^68 + 4844364\ M\^70 - 984276\ M\^72 - 
                2139941\ M\^74 + 956167\ M\^76 + 605008\ M\^78 - 
                532733\ M\^80 - 24236\ M\^82 + 165725\ M\^84 - 55729\ M\^86 - 
                18167\ M\^88 + 22043\ M\^90 - 9055\ M\^92 + 2141\ M\^94 - 
                310\ M\^96 + 26\ M\^98 - M\^100)\) + 
          L\^2\ \((6\ M\^64 - 64\ M\^66 + 272\ M\^68 - 524\ M\^70 + 
                237\ M\^72 + 911\ M\^74 - 1620\ M\^76 + 69\ M\^78 + 
                1802\ M\^80 - 1006\ M\^82 - 122\ M\^84 + 538\ M\^86 - 
                699\ M\^88 + 387\ M\^90 + 116\ M\^92 - 249\ M\^94 + 
                126\ M\^96 - 30\ M\^98 + 3\ M\^100)\) + 
          L\^10\ \((\(-2\)\ M\^2 + 44\ M\^4 - 426\ M\^6 + 2319\ M\^8 - 
                7321\ M\^10 + 10929\ M\^12 + 6469\ M\^14 - 51515\ M\^16 + 
                36552\ M\^18 + 157830\ M\^20 - 288176\ M\^22 - 
                236006\ M\^24 + 974873\ M\^26 - 10148\ M\^28 - 
                2264094\ M\^30 + 1130056\ M\^32 + 3890109\ M\^34 - 
                3519099\ M\^36 - 5098129\ M\^38 + 6929279\ M\^40 + 
                5120627\ M\^42 - 10416948\ M\^44 - 3888652\ M\^46 + 
                12722302\ M\^48 + 2455792\ M\^50 - 12800364\ M\^52 - 
                2219650\ M\^54 + 9984076\ M\^56 + 3507477\ M\^58 - 
                4886974\ M\^60 - 4891898\ M\^62 - 30832\ M\^64 + 
                5311613\ M\^66 + 2574551\ M\^68 - 4719159\ M\^70 - 
                2653215\ M\^72 + 3560049\ M\^74 + 1550492\ M\^76 - 
                2263028\ M\^78 - 496569\ M\^80 + 1161518\ M\^82 - 
                25627\ M\^84 - 447926\ M\^86 + 133931\ M\^88 + 98095\ M\^90 - 
                68980\ M\^92 - 1385\ M\^94 + 18480\ M\^96 - 9970\ M\^98 + 
                2820\ M\^100 - 476\ M\^102 + 46\ M\^104 - 2\ M\^106)\) + 
          L\^3\ \((\(-4\)\ M\^52 + 56\ M\^54 - 334\ M\^56 + 1100\ M\^58 - 
                2064\ M\^60 + 1401\ M\^62 + 3199\ M\^64 - 8825\ M\^66 + 
                4709\ M\^68 + 11479\ M\^70 - 16329\ M\^72 - 4219\ M\^74 + 
                14239\ M\^76 + 437\ M\^78 + 222\ M\^80 - 6556\ M\^82 - 
                4635\ M\^84 + 11342\ M\^86 - 1598\ M\^88 - 6457\ M\^90 + 
                4559\ M\^92 - 264\ M\^94 - 1311\ M\^96 + 980\ M\^98 - 
                398\ M\^100 + 101\ M\^102 - 15\ M\^104 + M\^106)\) + 
          L\^4\ \((M\^40 - 16\ M\^42 + 120\ M\^44 - 588\ M\^46 + 
                2104\ M\^48 - 5241\ M\^50 + 6886\ M\^52 + 3857\ M\^54 - 
                30971\ M\^56 + 34354\ M\^58 + 41866\ M\^60 - 119292\ M\^62 + 
                557\ M\^64 + 215362\ M\^66 - 82533\ M\^68 - 289101\ M\^70 + 
                150769\ M\^72 + 318704\ M\^74 - 183041\ M\^76 - 
                269071\ M\^78 + 192743\ M\^80 + 153425\ M\^82 - 
                163844\ M\^84 - 33256\ M\^86 + 91325\ M\^88 - 23629\ M\^90 - 
                22957\ M\^92 + 20284\ M\^94 - 5263\ M\^96 - 2564\ M\^98 + 
                3485\ M\^100 - 1962\ M\^102 + 675\ M\^104 - 145\ M\^106 + 
                18\ M\^108 - M\^110)\) + 
          L\^5\ \((\(-3\)\ M\^32 + 58\ M\^34 - 489\ M\^36 + 2347\ M\^38 - 
                6857\ M\^40 + 11284\ M\^42 - 3944\ M\^44 - 26939\ M\^46 + 
                61488\ M\^48 - 26060\ M\^50 - 132909\ M\^52 + 243870\ M\^54 + 
                66762\ M\^56 - 604532\ M\^58 + 249233\ M\^60 + 
                967834\ M\^62 - 717537\ M\^64 - 1173471\ M\^66 + 
                1067444\ M\^68 + 1089007\ M\^70 - 1106504\ M\^72 - 
                696332\ M\^74 + 881293\ M\^76 + 216962\ M\^78 - 
                498612\ M\^80 + 80960\ M\^82 + 135878\ M\^84 - 
                124076\ M\^86 + 61406\ M\^88 + 34016\ M\^90 - 77687\ M\^92 + 
                28285\ M\^94 + 26129\ M\^96 - 27063\ M\^98 + 4430\ M\^100 + 
                6675\ M\^102 - 5431\ M\^104 + 2048\ M\^106 - 447\ M\^108 + 
                55\ M\^110 - 3\ M\^112)\) + 
          L\^9\ \((\(-M\^4\) + 22\ M\^6 - 217\ M\^8 + 1236\ M\^10 - 
                4327\ M\^12 + 8822\ M\^14 - 6734\ M\^16 - 12237\ M\^18 + 
                32167\ M\^20 - 4293\ M\^22 - 73379\ M\^24 + 101271\ M\^26 + 
                902\ M\^28 - 204075\ M\^30 + 338116\ M\^32 + 108173\ M\^34 - 
                1083237\ M\^36 + 472234\ M\^38 + 2213143\ M\^40 - 
                1547543\ M\^42 - 3524837\ M\^44 + 2632043\ M\^46 + 
                4502709\ M\^48 - 3048902\ M\^50 - 4149222\ M\^52 + 
                2622072\ M\^54 + 1767787\ M\^56 - 2234766\ M\^58 + 
                1767787\ M\^60 + 2622072\ M\^62 - 4149222\ M\^64 - 
                3048902\ M\^66 + 4502709\ M\^68 + 2632043\ M\^70 - 
                3524837\ M\^72 - 1547543\ M\^74 + 2213143\ M\^76 + 
                472234\ M\^78 - 1083237\ M\^80 + 108173\ M\^82 + 
                338116\ M\^84 - 204075\ M\^86 + 902\ M\^88 + 101271\ M\^90 - 
                73379\ M\^92 - 4293\ M\^94 + 32167\ M\^96 - 12237\ M\^98 - 
                6734\ M\^100 + 8822\ M\^102 - 4327\ M\^104 + 1236\ M\^106 - 
                217\ M\^108 + 22\ M\^110 - M\^112)\) + 
          L\^6\ \((3\ M\^24 - 66\ M\^26 + 638\ M\^28 - 3510\ M\^30 + 
                11577\ M\^32 - 20318\ M\^34 + 2066\ M\^36 + 70930\ M\^38 - 
                119892\ M\^40 - 57931\ M\^42 + 409461\ M\^44 - 
                273889\ M\^46 - 718793\ M\^48 + 1138031\ M\^50 + 
                729961\ M\^52 - 2516067\ M\^54 - 244293\ M\^56 + 
                4090968\ M\^58 - 503548\ M\^60 - 5332583\ M\^62 + 
                989350\ M\^64 + 5664086\ M\^66 - 988306\ M\^68 - 
                4881717\ M\^70 + 730732\ M\^72 + 3539213\ M\^74 - 
                425605\ M\^76 - 2268220\ M\^78 + 188281\ M\^80 + 
                1378000\ M\^82 - 106082\ M\^84 - 805397\ M\^86 + 
                125467\ M\^88 + 419883\ M\^90 - 144085\ M\^92 - 
                160064\ M\^94 + 105915\ M\^96 + 26366\ M\^98 - 
                44776\ M\^100 + 8871\ M\^102 + 9097\ M\^104 - 7319\ M\^106 + 
                2600\ M\^108 - 528\ M\^110 + 60\ M\^112 - 3\ M\^114)\) + 
          L\^8\ \((\(-2\)\ M\^10 + 46\ M\^12 - 476\ M\^14 + 2820\ M\^16 - 
                9970\ M\^18 + 18480\ M\^20 - 1385\ M\^22 - 68980\ M\^24 + 
                98095\ M\^26 + 133931\ M\^28 - 447926\ M\^30 - 25627\ M\^32 + 
                1161518\ M\^34 - 496569\ M\^36 - 2263028\ M\^38 + 
                1550492\ M\^40 + 3560049\ M\^42 - 2653215\ M\^44 - 
                4719159\ M\^46 + 2574551\ M\^48 + 5311613\ M\^50 - 
                30832\ M\^52 - 4891898\ M\^54 - 4886974\ M\^56 + 
                3507477\ M\^58 + 9984076\ M\^60 - 2219650\ M\^62 - 
                12800364\ M\^64 + 2455792\ M\^66 + 12722302\ M\^68 - 
                3888652\ M\^70 - 10416948\ M\^72 + 5120627\ M\^74 + 
                6929279\ M\^76 - 5098129\ M\^78 - 3519099\ M\^80 + 
                3890109\ M\^82 + 1130056\ M\^84 - 2264094\ M\^86 - 
                10148\ M\^88 + 974873\ M\^90 - 236006\ M\^92 - 
                288176\ M\^94 + 157830\ M\^96 + 36552\ M\^98 - 
                51515\ M\^100 + 6469\ M\^102 + 10929\ M\^104 - 7321\ M\^106 + 
                2319\ M\^108 - 426\ M\^110 + 44\ M\^112 - 2\ M\^114)\) + 
          L\^7\ \((\(-M\^16\) + 26\ M\^18 - 310\ M\^20 + 2141\ M\^22 - 
                9055\ M\^24 + 22043\ M\^26 - 18167\ M\^28 - 55729\ M\^30 + 
                165725\ M\^32 - 24236\ M\^34 - 532733\ M\^36 + 
                605008\ M\^38 + 956167\ M\^40 - 2139941\ M\^42 - 
                984276\ M\^44 + 4844364\ M\^46 + 141299\ M\^48 - 
                8321163\ M\^50 + 1433716\ M\^52 + 11532800\ M\^54 - 
                2771152\ M\^56 - 13091886\ M\^58 + 2859252\ M\^60 + 
                11950195\ M\^62 - 1745322\ M\^64 - 8534659\ M\^66 + 
                431142\ M\^68 + 4729988\ M\^70 + 542447\ M\^72 - 
                2088086\ M\^74 - 984385\ M\^76 + 879132\ M\^78 + 
                919828\ M\^80 - 552477\ M\^82 - 526873\ M\^84 + 
                436030\ M\^86 + 134210\ M\^88 - 267551\ M\^90 + 
                61349\ M\^92 + 93886\ M\^94 - 77940\ M\^96 + 900\ M\^98 + 
                29445\ M\^100 - 13589\ M\^102 - 4215\ M\^104 + 7416\ M\^106 - 
                3921\ M\^108 + 1174\ M\^110 - 213\ M\^112 + 22\ M\^114 - 
                M\^116)\)}, {"A-polynomial", 
        Knot[9, 29], \(-M\^120\) - 2\ M\^122 - M\^124 + 
          L\^19\ \((M\^24 + 2\ M\^26 + M\^28)\) + 
          L\^18\ \((\(-2\)\ M\^16 + 5\ M\^18 - 5\ M\^20 - 6\ M\^22 + 
                53\ M\^24 - 32\ M\^26 - 86\ M\^28 + 84\ M\^30 + 81\ M\^32 - 
                21\ M\^34 - 11\ M\^36 + 4\ M\^38 + 4\ M\^40 + M\^42 - 
                M\^44)\) + 
          L\^17\ \((M\^8 - 7\ M\^10 + 27\ M\^12 - 52\ M\^14 + 31\ M\^16 + 
                86\ M\^18 - 253\ M\^20 + 367\ M\^22 - 10\ M\^24 - 
                1201\ M\^26 + 1318\ M\^28 + 1454\ M\^30 - 2509\ M\^32 - 
                737\ M\^34 + 2255\ M\^36 + 325\ M\^38 - 839\ M\^40 - 
                28\ M\^42 + 410\ M\^44 + 30\ M\^46 - 173\ M\^48 + 24\ M\^50 + 
                29\ M\^52 - 8\ M\^54)\) + 
          L\^16\ \((\(-2\)\ M\^4 + 21\ M\^6 - 103\ M\^8 + 265\ M\^10 - 
                261\ M\^12 - 420\ M\^14 + 1472\ M\^16 - 980\ M\^18 - 
                1707\ M\^20 + 3227\ M\^22 - 1431\ M\^24 - 1055\ M\^26 + 
                5094\ M\^28 - 8989\ M\^30 - 1829\ M\^32 + 19328\ M\^34 - 
                9039\ M\^36 - 18680\ M\^38 + 18704\ M\^40 + 9135\ M\^42 - 
                18372\ M\^44 + 464\ M\^46 + 13590\ M\^48 - 2128\ M\^50 - 
                6664\ M\^52 + 2123\ M\^54 + 1910\ M\^56 - 1025\ M\^58 - 
                144\ M\^60 + 176\ M\^62 - 28\ M\^64)\) + 
          L\^15\ \((1 - 14\ M\^2 + 91\ M\^4 - 334\ M\^6 + 656\ M\^8 - 
                232\ M\^10 - 1994\ M\^12 + 3916\ M\^14 + 1118\ M\^16 - 
                12444\ M\^18 + 8785\ M\^20 + 22100\ M\^22 - 31999\ M\^24 - 
                29348\ M\^26 + 69788\ M\^28 + 37777\ M\^30 - 129057\ M\^32 - 
                43885\ M\^34 + 214063\ M\^36 + 23872\ M\^38 - 303143\ M\^40 + 
                39112\ M\^42 + 341552\ M\^44 - 114546\ M\^46 - 
                285771\ M\^48 + 151832\ M\^50 + 173664\ M\^52 - 
                121691\ M\^54 - 68098\ M\^56 + 70088\ M\^58 + 17059\ M\^60 - 
                29416\ M\^62 - 573\ M\^64 + 8241\ M\^66 - 1478\ M\^68 - 
                1164\ M\^70 + 504\ M\^72 - 56\ M\^74)\) + 
          L\^14\ \((\(-1\) + 16\ M\^2 - 122\ M\^4 + 546\ M\^6 - 1438\ M\^8 + 
                1605\ M\^10 + 2464\ M\^12 - 10729\ M\^14 + 6899\ M\^16 + 
                28065\ M\^18 - 52835\ M\^20 - 27893\ M\^22 + 153396\ M\^24 - 
                38559\ M\^26 - 290074\ M\^28 + 209693\ M\^30 + 
                396489\ M\^32 - 477105\ M\^34 - 396184\ M\^36 + 
                795377\ M\^38 + 272298\ M\^40 - 1125409\ M\^42 - 
                74808\ M\^44 + 1402992\ M\^46 - 149848\ M\^48 - 
                1518049\ M\^50 + 363228\ M\^52 + 1383358\ M\^54 - 
                477477\ M\^56 - 998353\ M\^58 + 454308\ M\^60 + 
                554327\ M\^62 - 296047\ M\^64 - 218556\ M\^66 + 
                133228\ M\^68 + 64740\ M\^70 - 44269\ M\^72 - 15451\ M\^74 + 
                12619\ M\^76 + 1955\ M\^78 - 3120\ M\^80 + 826\ M\^82 - 
                70\ M\^84)\) + 
          L\^13\ \((\(-2\)\ M\^2 + 32\ M\^4 - 234\ M\^6 + 969\ M\^8 - 
                2226\ M\^10 + 1643\ M\^12 + 4913\ M\^14 - 11699\ M\^16 - 
                5334\ M\^18 + 44770\ M\^20 - 22642\ M\^22 - 102551\ M\^24 + 
                116138\ M\^26 + 160175\ M\^28 - 307419\ M\^30 - 
                133060\ M\^32 + 557828\ M\^34 - 83906\ M\^36 - 
                736985\ M\^38 + 492820\ M\^40 + 687523\ M\^42 - 
                947848\ M\^44 - 315148\ M\^46 + 1303124\ M\^48 - 
                320881\ M\^50 - 1615289\ M\^52 + 931579\ M\^54 + 
                1940740\ M\^56 - 1235663\ M\^58 - 2097372\ M\^60 + 
                1220508\ M\^62 + 1905108\ M\^64 - 958277\ M\^66 - 
                1333859\ M\^68 + 634945\ M\^70 + 702547\ M\^72 - 
                318011\ M\^74 - 258250\ M\^76 + 96642\ M\^78 + 75664\ M\^80 - 
                12525\ M\^82 - 26880\ M\^84 + 3487\ M\^86 + 8476\ M\^88 - 
                4356\ M\^90 + 826\ M\^92 - 56\ M\^94)\) + 
          L\^12\ \((\(-M\^4\) + 16\ M\^6 - 118\ M\^8 + 513\ M\^10 - 
                1394\ M\^12 + 2290\ M\^14 - 1812\ M\^16 - 640\ M\^18 + 
                3586\ M\^20 - 4174\ M\^22 - 6370\ M\^24 + 31254\ M\^26 - 
                17639\ M\^28 - 80134\ M\^30 + 92343\ M\^32 + 136827\ M\^34 - 
                223311\ M\^36 - 165093\ M\^38 + 383919\ M\^40 + 
                150688\ M\^42 - 575539\ M\^44 - 158767\ M\^46 + 
                803909\ M\^48 + 232116\ M\^50 - 970107\ M\^52 - 
                245029\ M\^54 + 920688\ M\^56 - 10989\ M\^58 - 
                799039\ M\^60 + 294470\ M\^62 + 763998\ M\^64 - 
                145546\ M\^66 - 697787\ M\^68 - 316116\ M\^70 + 
                596580\ M\^72 + 686773\ M\^74 - 411446\ M\^76 - 
                629544\ M\^78 + 242106\ M\^80 + 348982\ M\^82 - 
                103134\ M\^84 - 103869\ M\^86 + 1794\ M\^88 + 16781\ M\^90 + 
                27865\ M\^92 - 13184\ M\^94 - 11463\ M\^96 + 10767\ M\^98 - 
                3456\ M\^100 + 504\ M\^102 - 28\ M\^104)\) + 
          L\^11\ \((\(-2\)\ M\^10 + 35\ M\^12 - 275\ M\^14 + 1249\ M\^16 - 
                3375\ M\^18 + 4573\ M\^20 + 510\ M\^22 - 9358\ M\^24 + 
                225\ M\^26 + 32443\ M\^28 - 29025\ M\^30 - 54051\ M\^32 + 
                111300\ M\^34 + 16344\ M\^36 - 277385\ M\^38 + 
                235035\ M\^40 + 485464\ M\^42 - 911128\ M\^44 - 
                578499\ M\^46 + 2183041\ M\^48 + 365602\ M\^50 - 
                4042838\ M\^52 + 39975\ M\^54 + 6040423\ M\^56 - 
                264468\ M\^58 - 7281313\ M\^60 + 201270\ M\^62 + 
                7007445\ M\^64 - 469443\ M\^66 - 5636965\ M\^68 + 
                1250828\ M\^70 + 4083677\ M\^72 - 1828551\ M\^74 - 
                2611796\ M\^76 + 1790770\ M\^78 + 1351721\ M\^80 - 
                1206005\ M\^82 - 421391\ M\^84 + 575125\ M\^86 + 
                18272\ M\^88 - 178027\ M\^90 + 37185\ M\^92 + 22594\ M\^94 + 
                12679\ M\^96 - 7259\ M\^98 - 27114\ M\^100 + 18590\ M\^102 + 
                8387\ M\^104 - 13987\ M\^106 + 6585\ M\^108 - 1524\ M\^110 + 
                176\ M\^112 - 8\ M\^114)\) + 
          L\^10\ \((\(-M\^16\) + 20\ M\^18 - 189\ M\^20 + 1085\ M\^22 - 
                3910\ M\^24 + 8029\ M\^26 - 5578\ M\^28 - 10569\ M\^30 + 
                16996\ M\^32 + 21123\ M\^34 - 50664\ M\^36 - 20119\ M\^38 + 
                87043\ M\^40 + 13162\ M\^42 - 94948\ M\^44 - 34042\ M\^46 + 
                72304\ M\^48 + 118228\ M\^50 - 164063\ M\^52 - 
                183725\ M\^54 + 725900\ M\^56 + 101887\ M\^58 - 
                2307298\ M\^60 - 148056\ M\^62 + 4841449\ M\^64 + 
                960852\ M\^66 - 7140908\ M\^68 - 2697591\ M\^70 + 
                7495676\ M\^72 + 4044197\ M\^74 - 5935872\ M\^76 - 
                3899947\ M\^78 + 4019270\ M\^80 + 2749903\ M\^82 - 
                2579869\ M\^84 - 1452065\ M\^86 + 1624041\ M\^88 + 
                559277\ M\^90 - 872562\ M\^92 - 74512\ M\^94 + 
                328490\ M\^96 - 69943\ M\^98 - 48582\ M\^100 + 
                50264\ M\^102 - 33384\ M\^104 - 4478\ M\^106 + 
                25918\ M\^108 - 12386\ M\^110 - 7188\ M\^112 + 
                11514\ M\^114 - 6332\ M\^116 + 1897\ M\^118 - 324\ M\^120 + 
                29\ M\^122 - M\^124)\) + 
          L\^9\ \((M\^24 - 29\ M\^26 + 324\ M\^28 - 1897\ M\^30 + 
                6332\ M\^32 - 11514\ M\^34 + 7188\ M\^36 + 12386\ M\^38 - 
                25918\ M\^40 + 4478\ M\^42 + 33384\ M\^44 - 50264\ M\^46 + 
                48582\ M\^48 + 69943\ M\^50 - 328490\ M\^52 + 74512\ M\^54 + 
                872562\ M\^56 - 559277\ M\^58 - 1624041\ M\^60 + 
                1452065\ M\^62 + 2579869\ M\^64 - 2749903\ M\^66 - 
                4019270\ M\^68 + 3899947\ M\^70 + 5935872\ M\^72 - 
                4044197\ M\^74 - 7495676\ M\^76 + 2697591\ M\^78 + 
                7140908\ M\^80 - 960852\ M\^82 - 4841449\ M\^84 + 
                148056\ M\^86 + 2307298\ M\^88 - 101887\ M\^90 - 
                725900\ M\^92 + 183725\ M\^94 + 164063\ M\^96 - 
                118228\ M\^98 - 72304\ M\^100 + 34042\ M\^102 + 
                94948\ M\^104 - 13162\ M\^106 - 87043\ M\^108 + 
                20119\ M\^110 + 50664\ M\^112 - 21123\ M\^114 - 
                16996\ M\^116 + 10569\ M\^118 + 5578\ M\^120 - 8029\ M\^122 + 
                3910\ M\^124 - 1085\ M\^126 + 189\ M\^128 - 20\ M\^130 + 
                M\^132)\) + 
          L\ \((M\^104 - M\^106 - 4\ M\^108 - 4\ M\^110 + 11\ M\^112 + 
                21\ M\^114 - 81\ M\^116 - 84\ M\^118 + 86\ M\^120 + 
                32\ M\^122 - 53\ M\^124 + 6\ M\^126 + 5\ M\^128 - 5\ M\^130 + 
                2\ M\^132)\) + 
          L\^8\ \((8\ M\^34 - 176\ M\^36 + 1524\ M\^38 - 6585\ M\^40 + 
                13987\ M\^42 - 8387\ M\^44 - 18590\ M\^46 + 27114\ M\^48 + 
                7259\ M\^50 - 12679\ M\^52 - 22594\ M\^54 - 37185\ M\^56 + 
                178027\ M\^58 - 18272\ M\^60 - 575125\ M\^62 + 
                421391\ M\^64 + 1206005\ M\^66 - 1351721\ M\^68 - 
                1790770\ M\^70 + 2611796\ M\^72 + 1828551\ M\^74 - 
                4083677\ M\^76 - 1250828\ M\^78 + 5636965\ M\^80 + 
                469443\ M\^82 - 7007445\ M\^84 - 201270\ M\^86 + 
                7281313\ M\^88 + 264468\ M\^90 - 6040423\ M\^92 - 
                39975\ M\^94 + 4042838\ M\^96 - 365602\ M\^98 - 
                2183041\ M\^100 + 578499\ M\^102 + 911128\ M\^104 - 
                485464\ M\^106 - 235035\ M\^108 + 277385\ M\^110 - 
                16344\ M\^112 - 111300\ M\^114 + 54051\ M\^116 + 
                29025\ M\^118 - 32443\ M\^120 - 225\ M\^122 + 9358\ M\^124 - 
                510\ M\^126 - 4573\ M\^128 + 3375\ M\^130 - 1249\ M\^132 + 
                275\ M\^134 - 35\ M\^136 + 2\ M\^138)\) + 
          L\^2\ \((8\ M\^94 - 29\ M\^96 - 24\ M\^98 + 173\ M\^100 - 
                30\ M\^102 - 410\ M\^104 + 28\ M\^106 + 839\ M\^108 - 
                325\ M\^110 - 2255\ M\^112 + 737\ M\^114 + 2509\ M\^116 - 
                1454\ M\^118 - 1318\ M\^120 + 1201\ M\^122 + 10\ M\^124 - 
                367\ M\^126 + 253\ M\^128 - 86\ M\^130 - 31\ M\^132 + 
                52\ M\^134 - 27\ M\^136 + 7\ M\^138 - M\^140)\) + 
          L\^7\ \((28\ M\^44 - 504\ M\^46 + 3456\ M\^48 - 10767\ M\^50 + 
                11463\ M\^52 + 13184\ M\^54 - 27865\ M\^56 - 16781\ M\^58 - 
                1794\ M\^60 + 103869\ M\^62 + 103134\ M\^64 - 348982\ M\^66 - 
                242106\ M\^68 + 629544\ M\^70 + 411446\ M\^72 - 
                686773\ M\^74 - 596580\ M\^76 + 316116\ M\^78 + 
                697787\ M\^80 + 145546\ M\^82 - 763998\ M\^84 - 
                294470\ M\^86 + 799039\ M\^88 + 10989\ M\^90 - 
                920688\ M\^92 + 245029\ M\^94 + 970107\ M\^96 - 
                232116\ M\^98 - 803909\ M\^100 + 158767\ M\^102 + 
                575539\ M\^104 - 150688\ M\^106 - 383919\ M\^108 + 
                165093\ M\^110 + 223311\ M\^112 - 136827\ M\^114 - 
                92343\ M\^116 + 80134\ M\^118 + 17639\ M\^120 - 
                31254\ M\^122 + 6370\ M\^124 + 4174\ M\^126 - 3586\ M\^128 + 
                640\ M\^130 + 1812\ M\^132 - 2290\ M\^134 + 1394\ M\^136 - 
                513\ M\^138 + 118\ M\^140 - 16\ M\^142 + M\^144)\) + 
          L\^3\ \((28\ M\^84 - 176\ M\^86 + 144\ M\^88 + 1025\ M\^90 - 
                1910\ M\^92 - 2123\ M\^94 + 6664\ M\^96 + 2128\ M\^98 - 
                13590\ M\^100 - 464\ M\^102 + 18372\ M\^104 - 9135\ M\^106 - 
                18704\ M\^108 + 18680\ M\^110 + 9039\ M\^112 - 
                19328\ M\^114 + 1829\ M\^116 + 8989\ M\^118 - 5094\ M\^120 + 
                1055\ M\^122 + 1431\ M\^124 - 3227\ M\^126 + 1707\ M\^128 + 
                980\ M\^130 - 1472\ M\^132 + 420\ M\^134 + 261\ M\^136 - 
                265\ M\^138 + 103\ M\^140 - 21\ M\^142 + 2\ M\^144)\) + 
          L\^6\ \((56\ M\^54 - 826\ M\^56 + 4356\ M\^58 - 8476\ M\^60 - 
                3487\ M\^62 + 26880\ M\^64 + 12525\ M\^66 - 75664\ M\^68 - 
                96642\ M\^70 + 258250\ M\^72 + 318011\ M\^74 - 
                702547\ M\^76 - 634945\ M\^78 + 1333859\ M\^80 + 
                958277\ M\^82 - 1905108\ M\^84 - 1220508\ M\^86 + 
                2097372\ M\^88 + 1235663\ M\^90 - 1940740\ M\^92 - 
                931579\ M\^94 + 1615289\ M\^96 + 320881\ M\^98 - 
                1303124\ M\^100 + 315148\ M\^102 + 947848\ M\^104 - 
                687523\ M\^106 - 492820\ M\^108 + 736985\ M\^110 + 
                83906\ M\^112 - 557828\ M\^114 + 133060\ M\^116 + 
                307419\ M\^118 - 160175\ M\^120 - 116138\ M\^122 + 
                102551\ M\^124 + 22642\ M\^126 - 44770\ M\^128 + 
                5334\ M\^130 + 11699\ M\^132 - 4913\ M\^134 - 1643\ M\^136 + 
                2226\ M\^138 - 969\ M\^140 + 234\ M\^142 - 32\ M\^144 + 
                2\ M\^146)\) + 
          L\^4\ \((56\ M\^74 - 504\ M\^76 + 1164\ M\^78 + 1478\ M\^80 - 
                8241\ M\^82 + 573\ M\^84 + 29416\ M\^86 - 17059\ M\^88 - 
                70088\ M\^90 + 68098\ M\^92 + 121691\ M\^94 - 173664\ M\^96 - 
                151832\ M\^98 + 285771\ M\^100 + 114546\ M\^102 - 
                341552\ M\^104 - 39112\ M\^106 + 303143\ M\^108 - 
                23872\ M\^110 - 214063\ M\^112 + 43885\ M\^114 + 
                129057\ M\^116 - 37777\ M\^118 - 69788\ M\^120 + 
                29348\ M\^122 + 31999\ M\^124 - 22100\ M\^126 - 
                8785\ M\^128 + 12444\ M\^130 - 1118\ M\^132 - 3916\ M\^134 + 
                1994\ M\^136 + 232\ M\^138 - 656\ M\^140 + 334\ M\^142 - 
                91\ M\^144 + 14\ M\^146 - M\^148)\) + 
          L\^5\ \((70\ M\^64 - 826\ M\^66 + 3120\ M\^68 - 1955\ M\^70 - 
                12619\ M\^72 + 15451\ M\^74 + 44269\ M\^76 - 64740\ M\^78 - 
                133228\ M\^80 + 218556\ M\^82 + 296047\ M\^84 - 
                554327\ M\^86 - 454308\ M\^88 + 998353\ M\^90 + 
                477477\ M\^92 - 1383358\ M\^94 - 363228\ M\^96 + 
                1518049\ M\^98 + 149848\ M\^100 - 1402992\ M\^102 + 
                74808\ M\^104 + 1125409\ M\^106 - 272298\ M\^108 - 
                795377\ M\^110 + 396184\ M\^112 + 477105\ M\^114 - 
                396489\ M\^116 - 209693\ M\^118 + 290074\ M\^120 + 
                38559\ M\^122 - 153396\ M\^124 + 27893\ M\^126 + 
                52835\ M\^128 - 28065\ M\^130 - 6899\ M\^132 + 
                10729\ M\^134 - 2464\ M\^136 - 1605\ M\^138 + 1438\ M\^140 - 
                546\ M\^142 + 122\ M\^144 - 16\ M\^146 + 
                M\^148)\)}, {"A-polynomial", Knot[9, 30], 
        Null}, {"A-polynomial", Knot[9, 31], 
        L\^17\ M\^22 + M\^96 + 
          L\^16\ \((3\ M\^16 - 17\ M\^18 + 39\ M\^20 - 24\ M\^22 - 
                51\ M\^24 + 93\ M\^26 - 3\ M\^28 - 46\ M\^30 + 29\ M\^32 - 
                6\ M\^34)\) + 
          L\^15\ \((3\ M\^10 - 30\ M\^12 + 123\ M\^14 - 227\ M\^16 + 
                86\ M\^18 + 275\ M\^20 - 277\ M\^22 + 150\ M\^24 - 
                592\ M\^26 + 262\ M\^28 + 1471\ M\^30 - 1482\ M\^32 - 
                527\ M\^34 + 1641\ M\^36 - 608\ M\^38 - 455\ M\^40 + 
                442\ M\^42 - 134\ M\^44 + 15\ M\^46)\) + 
          L\^14\ \((M\^4 - 13\ M\^6 + 74\ M\^8 - 243\ M\^10 + 528\ M\^12 - 
                813\ M\^14 + 579\ M\^16 + 1490\ M\^18 - 5137\ M\^20 + 
                3730\ M\^22 + 8042\ M\^24 - 14545\ M\^26 - 3411\ M\^28 + 
                21323\ M\^30 - 3333\ M\^32 - 20344\ M\^34 + 4977\ M\^36 + 
                18742\ M\^38 - 4513\ M\^40 - 15624\ M\^42 + 7100\ M\^44 + 
                7703\ M\^46 - 6827\ M\^48 - 216\ M\^50 + 2354\ M\^52 - 
                1170\ M\^54 + 246\ M\^56 - 20\ M\^58)\) + 
          L\^13\ \((2\ M\^2 - 33\ M\^4 + 235\ M\^6 - 910\ M\^8 + 
                1872\ M\^10 - 989\ M\^12 - 4223\ M\^14 + 7855\ M\^16 + 
                4784\ M\^18 - 23494\ M\^20 - 2851\ M\^22 + 62013\ M\^24 - 
                14107\ M\^26 - 135264\ M\^28 + 78747\ M\^30 + 204880\ M\^32 - 
                188296\ M\^34 - 200565\ M\^36 + 266045\ M\^38 + 
                117992\ M\^40 - 241337\ M\^42 - 30651\ M\^44 + 
                150644\ M\^46 - 573\ M\^48 - 72788\ M\^50 - 1281\ M\^52 + 
                36447\ M\^54 - 1419\ M\^56 - 17919\ M\^58 + 6479\ M\^60 + 
                3773\ M\^62 - 3816\ M\^64 + 1337\ M\^66 - 224\ M\^68 + 
                15\ M\^70)\) + 
          L\^12\ \((1 - 19\ M\^2 + 157\ M\^4 - 729\ M\^6 + 2032\ M\^8 - 
                3237\ M\^10 + 1784\ M\^12 + 4377\ M\^14 - 14701\ M\^16 + 
                22376\ M\^18 + 4186\ M\^20 - 94824\ M\^22 + 119742\ M\^24 + 
                156406\ M\^26 - 431711\ M\^28 - 91020\ M\^30 + 
                908009\ M\^32 - 190318\ M\^34 - 1400092\ M\^36 + 
                665413\ M\^38 + 1678528\ M\^40 - 1170811\ M\^42 - 
                1549609\ M\^44 + 1453259\ M\^46 + 1044108\ M\^48 - 
                1319379\ M\^50 - 434800\ M\^52 + 865542\ M\^54 + 
                41781\ M\^56 - 387082\ M\^58 + 62202\ M\^60 + 104034\ M\^62 - 
                27829\ M\^64 - 16213\ M\^66 - 1190\ M\^68 + 7308\ M\^70 + 
                972\ M\^72 - 4375\ M\^74 + 2517\ M\^76 - 702\ M\^78 + 
                101\ M\^80 - 6\ M\^82)\) + 
          L\^11\ \((3\ M\^2 - 61\ M\^4 + 533\ M\^6 - 2546\ M\^8 + 
                6739\ M\^10 - 7067\ M\^12 - 11081\ M\^14 + 41474\ M\^16 - 
                22395\ M\^18 - 79127\ M\^20 + 117169\ M\^22 + 47970\ M\^24 - 
                167190\ M\^26 + 17364\ M\^28 - 21214\ M\^30 + 155903\ M\^32 + 
                507340\ M\^34 - 947348\ M\^36 - 1024805\ M\^38 + 
                2428791\ M\^40 + 1128037\ M\^42 - 4109135\ M\^44 - 
                572728\ M\^46 + 5181375\ M\^48 - 434493\ M\^50 - 
                5074466\ M\^52 + 1395515\ M\^54 + 3907203\ M\^56 - 
                1815733\ M\^58 - 2294399\ M\^60 + 1590106\ M\^62 + 
                930324\ M\^64 - 1000897\ M\^66 - 172367\ M\^68 + 
                438039\ M\^70 - 61692\ M\^72 - 112493\ M\^74 + 51823\ M\^76 + 
                5351\ M\^78 - 10809\ M\^80 + 4931\ M\^82 - 2601\ M\^84 + 
                1502\ M\^86 - 591\ M\^88 + 139\ M\^90 - 18\ M\^92 + 
                M\^94)\) + 
          L\^10\ \((3\ M\^4 - 65\ M\^6 + 611\ M\^8 - 3178\ M\^10 + 
                9355\ M\^12 - 11822\ M\^14 - 14756\ M\^16 + 74126\ M\^18 - 
                55866\ M\^20 - 176931\ M\^22 + 355009\ M\^24 + 
                160533\ M\^26 - 983256\ M\^28 + 245405\ M\^30 + 
                1879060\ M\^32 - 1245229\ M\^34 - 2818360\ M\^36 + 
                2739931\ M\^38 + 3617217\ M\^40 - 4338818\ M\^42 - 
                4260933\ M\^44 + 5606078\ M\^46 + 4773927\ M\^48 - 
                6289343\ M\^50 - 4982161\ M\^52 + 6315526\ M\^54 + 
                4594209\ M\^56 - 5627332\ M\^58 - 3497351\ M\^60 + 
                4365132\ M\^62 + 2107878\ M\^64 - 2858216\ M\^66 - 
                984119\ M\^68 + 1540836\ M\^70 + 361531\ M\^72 - 
                687634\ M\^74 - 111726\ M\^76 + 268894\ M\^78 + 
                27460\ M\^80 - 101922\ M\^82 + 5225\ M\^84 + 33208\ M\^86 - 
                10542\ M\^88 - 6645\ M\^90 + 6510\ M\^92 - 2474\ M\^94 + 
                520\ M\^96 - 60\ M\^98 + 3\ M\^100)\) + 
          L\ \((\(-6\)\ M\^84 + 29\ M\^86 - 46\ M\^88 - 3\ M\^90 + 
                93\ M\^92 - 51\ M\^94 - 24\ M\^96 + 39\ M\^98 - 17\ M\^100 + 
                3\ M\^102)\) + 
          L\^9\ \((M\^6 - 23\ M\^8 + 236\ M\^10 - 1395\ M\^12 + 5041\ M\^14 - 
                10280\ M\^16 + 5209\ M\^18 + 28796\ M\^20 - 66314\ M\^22 - 
                2621\ M\^24 + 207104\ M\^26 - 212969\ M\^28 - 328225\ M\^30 + 
                737429\ M\^32 + 227157\ M\^34 - 1566263\ M\^36 + 
                260705\ M\^38 + 2585912\ M\^40 - 1085658\ M\^42 - 
                3781847\ M\^44 + 2068033\ M\^46 + 5269686\ M\^48 - 
                3090501\ M\^50 - 7045607\ M\^52 + 4161233\ M\^54 + 
                8708059\ M\^56 - 5295544\ M\^58 - 9472248\ M\^60 + 
                6381043\ M\^62 + 8732600\ M\^64 - 6911720\ M\^66 - 
                6473484\ M\^68 + 6399693\ M\^70 + 3555326\ M\^72 - 
                4853408\ M\^74 - 1158767\ M\^76 + 2895477\ M\^78 - 
                87086\ M\^80 - 1285653\ M\^82 + 366744\ M\^84 + 
                379718\ M\^86 - 235646\ M\^88 - 42318\ M\^90 + 78849\ M\^92 - 
                17202\ M\^94 - 11869\ M\^96 + 9585\ M\^98 - 3231\ M\^100 + 
                615\ M\^102 - 65\ M\^104 + 3\ M\^106)\) + 
          L\^2\ \((15\ M\^72 - 134\ M\^74 + 442\ M\^76 - 455\ M\^78 - 
                608\ M\^80 + 1641\ M\^82 - 527\ M\^84 - 1482\ M\^86 + 
                1471\ M\^88 + 262\ M\^90 - 592\ M\^92 + 150\ M\^94 - 
                277\ M\^96 + 275\ M\^98 + 86\ M\^100 - 227\ M\^102 + 
                123\ M\^104 - 30\ M\^106 + 3\ M\^108)\) + 
          L\^8\ \((3\ M\^12 - 65\ M\^14 + 615\ M\^16 - 3231\ M\^18 + 
                9585\ M\^20 - 11869\ M\^22 - 17202\ M\^24 + 78849\ M\^26 - 
                42318\ M\^28 - 235646\ M\^30 + 379718\ M\^32 + 
                366744\ M\^34 - 1285653\ M\^36 - 87086\ M\^38 + 
                2895477\ M\^40 - 1158767\ M\^42 - 4853408\ M\^44 + 
                3555326\ M\^46 + 6399693\ M\^48 - 6473484\ M\^50 - 
                6911720\ M\^52 + 8732600\ M\^54 + 6381043\ M\^56 - 
                9472248\ M\^58 - 5295544\ M\^60 + 8708059\ M\^62 + 
                4161233\ M\^64 - 7045607\ M\^66 - 3090501\ M\^68 + 
                5269686\ M\^70 + 2068033\ M\^72 - 3781847\ M\^74 - 
                1085658\ M\^76 + 2585912\ M\^78 + 260705\ M\^80 - 
                1566263\ M\^82 + 227157\ M\^84 + 737429\ M\^86 - 
                328225\ M\^88 - 212969\ M\^90 + 207104\ M\^92 - 2621\ M\^94 - 
                66314\ M\^96 + 28796\ M\^98 + 5209\ M\^100 - 10280\ M\^102 + 
                5041\ M\^104 - 1395\ M\^106 + 236\ M\^108 - 23\ M\^110 + 
                M\^112)\) + 
          L\^3\ \((\(-20\)\ M\^60 + 246\ M\^62 - 1170\ M\^64 + 2354\ M\^66 - 
                216\ M\^68 - 6827\ M\^70 + 7703\ M\^72 + 7100\ M\^74 - 
                15624\ M\^76 - 4513\ M\^78 + 18742\ M\^80 + 4977\ M\^82 - 
                20344\ M\^84 - 3333\ M\^86 + 21323\ M\^88 - 3411\ M\^90 - 
                14545\ M\^92 + 8042\ M\^94 + 3730\ M\^96 - 5137\ M\^98 + 
                1490\ M\^100 + 579\ M\^102 - 813\ M\^104 + 528\ M\^106 - 
                243\ M\^108 + 74\ M\^110 - 13\ M\^112 + M\^114)\) + 
          L\^7\ \((3\ M\^18 - 60\ M\^20 + 520\ M\^22 - 2474\ M\^24 + 
                6510\ M\^26 - 6645\ M\^28 - 10542\ M\^30 + 33208\ M\^32 + 
                5225\ M\^34 - 101922\ M\^36 + 27460\ M\^38 + 268894\ M\^40 - 
                111726\ M\^42 - 687634\ M\^44 + 361531\ M\^46 + 
                1540836\ M\^48 - 984119\ M\^50 - 2858216\ M\^52 + 
                2107878\ M\^54 + 4365132\ M\^56 - 3497351\ M\^58 - 
                5627332\ M\^60 + 4594209\ M\^62 + 6315526\ M\^64 - 
                4982161\ M\^66 - 6289343\ M\^68 + 4773927\ M\^70 + 
                5606078\ M\^72 - 4260933\ M\^74 - 4338818\ M\^76 + 
                3617217\ M\^78 + 2739931\ M\^80 - 2818360\ M\^82 - 
                1245229\ M\^84 + 1879060\ M\^86 + 245405\ M\^88 - 
                983256\ M\^90 + 160533\ M\^92 + 355009\ M\^94 - 
                176931\ M\^96 - 55866\ M\^98 + 74126\ M\^100 - 
                14756\ M\^102 - 11822\ M\^104 + 9355\ M\^106 - 3178\ M\^108 + 
                611\ M\^110 - 65\ M\^112 + 3\ M\^114)\) + 
          L\^4\ \((15\ M\^48 - 224\ M\^50 + 1337\ M\^52 - 3816\ M\^54 + 
                3773\ M\^56 + 6479\ M\^58 - 17919\ M\^60 - 1419\ M\^62 + 
                36447\ M\^64 - 1281\ M\^66 - 72788\ M\^68 - 573\ M\^70 + 
                150644\ M\^72 - 30651\ M\^74 - 241337\ M\^76 + 
                117992\ M\^78 + 266045\ M\^80 - 200565\ M\^82 - 
                188296\ M\^84 + 204880\ M\^86 + 78747\ M\^88 - 
                135264\ M\^90 - 14107\ M\^92 + 62013\ M\^94 - 2851\ M\^96 - 
                23494\ M\^98 + 4784\ M\^100 + 7855\ M\^102 - 4223\ M\^104 - 
                989\ M\^106 + 1872\ M\^108 - 910\ M\^110 + 235\ M\^112 - 
                33\ M\^114 + 2\ M\^116)\) + 
          L\^6\ \((M\^24 - 18\ M\^26 + 139\ M\^28 - 591\ M\^30 + 
                1502\ M\^32 - 2601\ M\^34 + 4931\ M\^36 - 10809\ M\^38 + 
                5351\ M\^40 + 51823\ M\^42 - 112493\ M\^44 - 61692\ M\^46 + 
                438039\ M\^48 - 172367\ M\^50 - 1000897\ M\^52 + 
                930324\ M\^54 + 1590106\ M\^56 - 2294399\ M\^58 - 
                1815733\ M\^60 + 3907203\ M\^62 + 1395515\ M\^64 - 
                5074466\ M\^66 - 434493\ M\^68 + 5181375\ M\^70 - 
                572728\ M\^72 - 4109135\ M\^74 + 1128037\ M\^76 + 
                2428791\ M\^78 - 1024805\ M\^80 - 947348\ M\^82 + 
                507340\ M\^84 + 155903\ M\^86 - 21214\ M\^88 + 17364\ M\^90 - 
                167190\ M\^92 + 47970\ M\^94 + 117169\ M\^96 - 79127\ M\^98 - 
                22395\ M\^100 + 41474\ M\^102 - 11081\ M\^104 - 
                7067\ M\^106 + 6739\ M\^108 - 2546\ M\^110 + 533\ M\^112 - 
                61\ M\^114 + 3\ M\^116)\) + 
          L\^5\ \((\(-6\)\ M\^36 + 101\ M\^38 - 702\ M\^40 + 2517\ M\^42 - 
                4375\ M\^44 + 972\ M\^46 + 7308\ M\^48 - 1190\ M\^50 - 
                16213\ M\^52 - 27829\ M\^54 + 104034\ M\^56 + 62202\ M\^58 - 
                387082\ M\^60 + 41781\ M\^62 + 865542\ M\^64 - 
                434800\ M\^66 - 1319379\ M\^68 + 1044108\ M\^70 + 
                1453259\ M\^72 - 1549609\ M\^74 - 1170811\ M\^76 + 
                1678528\ M\^78 + 665413\ M\^80 - 1400092\ M\^82 - 
                190318\ M\^84 + 908009\ M\^86 - 91020\ M\^88 - 
                431711\ M\^90 + 156406\ M\^92 + 119742\ M\^94 - 
                94824\ M\^96 + 4186\ M\^98 + 22376\ M\^100 - 14701\ M\^102 + 
                4377\ M\^104 + 1784\ M\^106 - 3237\ M\^108 + 2032\ M\^110 - 
                729\ M\^112 + 157\ M\^114 - 19\ M\^116 + 
                M\^118)\)}, {"A-polynomial", Knot[9, 32], 
        Null}, {"A-polynomial", Knot[9, 33], Null}, {"A-polynomial", 
        Knot[9, 34], Null}, {"A-polynomial", Knot[9, 35], 
        1 + L\^7\ M\^34 + 
          L\ \((\(-4\) + 6\ M\^2 + 15\ M\^4 - 19\ M\^6 + 6\ M\^8 + 
                10\ M\^10 - 14\ M\^12 + 9\ M\^14 - 3\ M\^16 + M\^18)\) + 
          L\^2\ \((6 - 18\ M\^2 - 13\ M\^4 + 73\ M\^6 - 18\ M\^8 - 
                50\ M\^10 + 56\ M\^12 - 12\ M\^14 - 33\ M\^16 + 39\ M\^18 - 
                3\ M\^20 - 19\ M\^22 + 22\ M\^24 - 12\ M\^26 + 4\ M\^28 - 
                M\^30)\) + 
          L\^6\ \((M\^16 - 3\ M\^18 + 9\ M\^20 - 14\ M\^22 + 10\ M\^24 + 
                6\ M\^26 - 19\ M\^28 + 15\ M\^30 + 6\ M\^32 - 4\ M\^34)\) + 
          L\^4\ \((1 - 6\ M\^2 + 19\ M\^4 - 39\ M\^6 + 27\ M\^8 + 35\ M\^10 - 
                72\ M\^12 + 39\ M\^14 + 92\ M\^16 - 85\ M\^18 - 66\ M\^20 + 
                62\ M\^22 + 35\ M\^24 + 18\ M\^26 - 19\ M\^28 - 20\ M\^30 + 
                18\ M\^32 - 4\ M\^34)\) + 
          L\^3\ \((\(-4\) + 18\ M\^2 - 20\ M\^4 - 19\ M\^6 + 18\ M\^8 + 
                35\ M\^10 + 62\ M\^12 - 66\ M\^14 - 85\ M\^16 + 92\ M\^18 + 
                39\ M\^20 - 72\ M\^22 + 35\ M\^24 + 27\ M\^26 - 39\ M\^28 + 
                19\ M\^30 - 6\ M\^32 + M\^34)\) + 
          L\^5\ \((\(-M\^4\) + 4\ M\^6 - 12\ M\^8 + 22\ M\^10 - 19\ M\^12 - 
                3\ M\^14 + 39\ M\^16 - 33\ M\^18 - 12\ M\^20 + 56\ M\^22 - 
                50\ M\^24 - 18\ M\^26 + 73\ M\^28 - 13\ M\^30 - 18\ M\^32 + 
                6\ M\^34)\)}, {"A-polynomial", Knot[9, 36], 
        M\^8 + L\^22\ M\^184 + 
          L\ \((\(-2\)\ M\^4 + 8\ M\^6 - 19\ M\^8 + 29\ M\^10 - 9\ M\^12 - 
                30\ M\^14 + 47\ M\^16 + 5\ M\^18 - 17\ M\^20 + 6\ M\^22)\) + 
          L\^2\ \((1 - 8\ M\^2 + 32\ M\^4 - 80\ M\^6 + 125\ M\^8 - 
                110\ M\^10 + 31\ M\^12 + 88\ M\^14 - 261\ M\^16 + 
                283\ M\^18 + 192\ M\^20 - 626\ M\^22 + 383\ M\^24 + 
                473\ M\^26 - 521\ M\^28 + 64\ M\^30 + 156\ M\^32 - 
                86\ M\^34 + 15\ M\^36)\) + 
          L\^3\ \((\(-1\) + 11\ M\^2 - 59\ M\^4 + 200\ M\^6 - 458\ M\^8 + 
                672\ M\^10 - 459\ M\^12 - 316\ M\^14 + 1109\ M\^16 - 
                1050\ M\^18 - 427\ M\^20 + 2719\ M\^22 - 2800\ M\^24 - 
                1264\ M\^26 + 4970\ M\^28 - 2421\ M\^30 - 2408\ M\^32 + 
                4912\ M\^34 - 1511\ M\^36 - 2742\ M\^38 + 2973\ M\^40 - 
                521\ M\^42 - 790\ M\^44 + 596\ M\^46 - 175\ M\^48 + 
                20\ M\^50)\) + 
          L\^4\ \((\(-2\)\ M\^2 + 24\ M\^4 - 138\ M\^6 + 495\ M\^8 - 
                1173\ M\^10 + 1756\ M\^12 - 1310\ M\^14 - 184\ M\^16 + 
                907\ M\^18 - 159\ M\^20 - 188\ M\^22 + 890\ M\^24 - 
                3122\ M\^26 + 921\ M\^28 + 7049\ M\^30 - 4649\ M\^32 - 
                11708\ M\^34 + 14321\ M\^36 + 8090\ M\^38 - 19713\ M\^40 + 
                10192\ M\^42 + 11512\ M\^44 - 19820\ M\^46 + 4908\ M\^48 + 
                11859\ M\^50 - 11038\ M\^52 + 1737\ M\^54 + 3038\ M\^56 - 
                2499\ M\^58 + 924\ M\^60 - 180\ M\^62 + 15\ M\^64)\) + 
          L\^5\ \((\(-M\^4\) + 15\ M\^6 - 105\ M\^8 + 453\ M\^10 - 
                1311\ M\^12 + 2655\ M\^14 - 3963\ M\^16 + 4590\ M\^18 - 
                3039\ M\^20 - 3493\ M\^22 + 9771\ M\^24 + 2815\ M\^26 - 
                26595\ M\^28 + 7124\ M\^30 + 45523\ M\^32 - 18837\ M\^34 - 
                67726\ M\^36 + 39480\ M\^38 + 76311\ M\^40 - 65995\ M\^42 - 
                55661\ M\^44 + 87817\ M\^46 + 8425\ M\^48 - 65833\ M\^50 + 
                46787\ M\^52 + 1247\ M\^54 - 44935\ M\^56 + 44694\ M\^58 + 
                1147\ M\^60 - 33707\ M\^62 + 25412\ M\^64 - 2136\ M\^66 - 
                8750\ M\^68 + 7010\ M\^70 - 2827\ M\^72 + 681\ M\^74 - 
                95\ M\^76 + 6\ M\^78)\) + 
          L\^6\ \((M\^8 - 14\ M\^10 + 104\ M\^12 - 510\ M\^14 + 1774\ M\^16 - 
                4455\ M\^18 + 7861\ M\^20 - 8962\ M\^22 + 5766\ M\^24 - 
                3304\ M\^26 + 4004\ M\^28 + 11036\ M\^30 - 39119\ M\^32 - 
                9149\ M\^34 + 147304\ M\^36 - 89003\ M\^38 - 281248\ M\^40 + 
                312040\ M\^42 + 344522\ M\^44 - 570513\ M\^46 - 
                249800\ M\^48 + 670012\ M\^50 + 53717\ M\^52 - 
                507126\ M\^54 + 96604\ M\^56 + 224647\ M\^58 - 79821\ M\^60 - 
                16337\ M\^62 - 48526\ M\^64 + 8962\ M\^66 + 98974\ M\^68 - 
                78565\ M\^70 - 30408\ M\^72 + 77507\ M\^74 - 39312\ M\^76 - 
                7276\ M\^78 + 20749\ M\^80 - 13728\ M\^82 + 5346\ M\^84 - 
                1360\ M\^86 + 224\ M\^88 - 22\ M\^90 + M\^92)\) + 
          L\^7\ \((M\^14 - 18\ M\^16 + 156\ M\^18 - 865\ M\^20 + 
                3345\ M\^22 - 9227\ M\^24 + 17660\ M\^26 - 20253\ M\^28 + 
                4246\ M\^30 + 20842\ M\^32 - 2998\ M\^34 - 70071\ M\^36 + 
                48668\ M\^38 + 164894\ M\^40 - 183968\ M\^42 - 
                354752\ M\^44 + 595068\ M\^46 + 488332\ M\^48 - 
                1324449\ M\^50 - 328118\ M\^52 + 2070026\ M\^54 - 
                124862\ M\^56 - 2317650\ M\^58 + 493637\ M\^60 + 
                1917504\ M\^62 - 415081\ M\^64 - 1244281\ M\^66 + 
                90045\ M\^68 + 747150\ M\^70 + 84459\ M\^72 - 507712\ M\^74 + 
                43024\ M\^76 + 270827\ M\^78 - 176881\ M\^80 - 10425\ M\^82 + 
                116665\ M\^84 - 102353\ M\^86 + 23813\ M\^88 + 30844\ M\^90 - 
                35882\ M\^92 + 19456\ M\^94 - 6606\ M\^96 + 1503\ M\^98 - 
                230\ M\^100 + 22\ M\^102 - M\^104)\) + 
          L\^8\ \((\(-3\)\ M\^22 + 57\ M\^24 - 507\ M\^26 + 2780\ M\^28 - 
                10255\ M\^30 + 25877\ M\^32 - 42132\ M\^34 + 30854\ M\^36 + 
                34708\ M\^38 - 100303\ M\^40 - 2195\ M\^42 + 293967\ M\^44 - 
                277936\ M\^46 - 482134\ M\^48 + 934486\ M\^50 + 
                471247\ M\^52 - 2007692\ M\^54 + 70013\ M\^56 + 
                3131714\ M\^58 - 1193939\ M\^60 - 3699302\ M\^62 + 
                2307805\ M\^64 + 3569487\ M\^66 - 2564280\ M\^68 - 
                3217647\ M\^70 + 1901897\ M\^72 + 3053175\ M\^74 - 
                1140510\ M\^76 - 2707573\ M\^78 + 785696\ M\^80 + 
                1800972\ M\^82 - 636579\ M\^84 - 651530\ M\^86 + 
                290527\ M\^88 - 4263\ M\^90 + 89975\ M\^92 + 42404\ M\^94 - 
                183293\ M\^96 + 102344\ M\^98 + 28332\ M\^100 - 
                66946\ M\^102 + 43941\ M\^104 - 17548\ M\^106 + 
                4789\ M\^108 - 884\ M\^110 + 99\ M\^112 - 5\ M\^114)\) + 
          L\^9\ \((3\ M\^30 - 68\ M\^32 + 702\ M\^34 - 4301\ M\^36 + 
                17117\ M\^38 - 45230\ M\^40 + 74943\ M\^42 - 51780\ M\^44 - 
                80776\ M\^46 + 233365\ M\^48 - 75911\ M\^50 - 544627\ M\^52 + 
                793015\ M\^54 + 525906\ M\^56 - 2114065\ M\^58 + 
                356856\ M\^60 + 3583698\ M\^62 - 2160028\ M\^64 - 
                4460477\ M\^66 + 4064747\ M\^68 + 4478493\ M\^70 - 
                5007994\ M\^72 - 4253627\ M\^74 + 5038413\ M\^76 + 
                4447380\ M\^78 - 5008564\ M\^80 - 4639354\ M\^82 + 
                5101851\ M\^84 + 3769439\ M\^86 - 4463331\ M\^88 - 
                2103045\ M\^90 + 2754979\ M\^92 + 811484\ M\^94 - 
                919397\ M\^96 - 464348\ M\^98 + 58318\ M\^100 + 
                486973\ M\^102 - 57083\ M\^104 - 322229\ M\^106 + 
                196616\ M\^108 + 27273\ M\^110 - 99576\ M\^112 + 
                71216\ M\^114 - 30958\ M\^116 + 9085\ M\^118 - 1752\ M\^120 + 
                199\ M\^122 - 10\ M\^124)\) + 
          L\^10\ \((\(-M\^38\) + 33\ M\^40 - 446\ M\^42 + 3371\ M\^44 - 
                15959\ M\^46 + 49314\ M\^48 - 97451\ M\^50 + 100470\ M\^52 + 
                42141\ M\^54 - 313878\ M\^56 + 339320\ M\^58 + 
                390110\ M\^60 - 1298199\ M\^62 + 359376\ M\^64 + 
                2433956\ M\^66 - 2293009\ M\^68 - 3002744\ M\^70 + 
                4838864\ M\^72 + 2953148\ M\^74 - 7006167\ M\^76 - 
                3150804\ M\^78 + 8406005\ M\^80 + 4160012\ M\^82 - 
                9389611\ M\^84 - 4847902\ M\^86 + 9962966\ M\^88 + 
                3834032\ M\^90 - 9382710\ M\^92 - 1730631\ M\^94 + 
                7001984\ M\^96 + 383056\ M\^98 - 4002650\ M\^100 - 
                402710\ M\^102 + 2013722\ M\^104 + 767165\ M\^106 - 
                1290384\ M\^108 - 525127\ M\^110 + 929318\ M\^112 - 
                848\ M\^114 - 453963\ M\^116 + 236006\ M\^118 + 
                39547\ M\^120 - 123259\ M\^122 + 88538\ M\^124 - 
                38280\ M\^126 + 10865\ M\^128 - 1981\ M\^130 + 211\ M\^132 - 
                10\ M\^134)\) + 
          L\^11\ \((\(-5\)\ M\^48 + 120\ M\^50 - 1286\ M\^52 + 8009\ M\^54 - 
                31676\ M\^56 + 81308\ M\^58 - 127451\ M\^60 + 72585\ M\^62 + 
                177542\ M\^64 - 465677\ M\^66 + 193382\ M\^68 + 
                905220\ M\^70 - 1267466\ M\^72 - 880714\ M\^74 + 
                2734848\ M\^76 + 326282\ M\^78 - 4086558\ M\^80 - 
                25672\ M\^82 + 5745606\ M\^84 + 447039\ M\^86 - 
                8521369\ M\^88 - 889702\ M\^90 + 11883497\ M\^92 + 
                453026\ M\^94 - 13495368\ M\^96 + 453026\ M\^98 + 
                11883497\ M\^100 - 889702\ M\^102 - 8521369\ M\^104 + 
                447039\ M\^106 + 5745606\ M\^108 - 25672\ M\^110 - 
                4086558\ M\^112 + 326282\ M\^114 + 2734848\ M\^116 - 
                880714\ M\^118 - 1267466\ M\^120 + 905220\ M\^122 + 
                193382\ M\^124 - 465677\ M\^126 + 177542\ M\^128 + 
                72585\ M\^130 - 127451\ M\^132 + 81308\ M\^134 - 
                31676\ M\^136 + 8009\ M\^138 - 1286\ M\^140 + 120\ M\^142 - 
                5\ M\^144)\) + 
          L\^12\ \((\(-10\)\ M\^58 + 211\ M\^60 - 1981\ M\^62 + 
                10865\ M\^64 - 38280\ M\^66 + 88538\ M\^68 - 123259\ M\^70 + 
                39547\ M\^72 + 236006\ M\^74 - 453963\ M\^76 - 848\ M\^78 + 
                929318\ M\^80 - 525127\ M\^82 - 1290384\ M\^84 + 
                767165\ M\^86 + 2013722\ M\^88 - 402710\ M\^90 - 
                4002650\ M\^92 + 383056\ M\^94 + 7001984\ M\^96 - 
                1730631\ M\^98 - 9382710\ M\^100 + 3834032\ M\^102 + 
                9962966\ M\^104 - 4847902\ M\^106 - 9389611\ M\^108 + 
                4160012\ M\^110 + 8406005\ M\^112 - 3150804\ M\^114 - 
                7006167\ M\^116 + 2953148\ M\^118 + 4838864\ M\^120 - 
                3002744\ M\^122 - 2293009\ M\^124 + 2433956\ M\^126 + 
                359376\ M\^128 - 1298199\ M\^130 + 390110\ M\^132 + 
                339320\ M\^134 - 313878\ M\^136 + 42141\ M\^138 + 
                100470\ M\^140 - 97451\ M\^142 + 49314\ M\^144 - 
                15959\ M\^146 + 3371\ M\^148 - 446\ M\^150 + 33\ M\^152 - 
                M\^154)\) + 
          L\^13\ \((\(-10\)\ M\^68 + 199\ M\^70 - 1752\ M\^72 + 9085\ M\^74 - 
                30958\ M\^76 + 71216\ M\^78 - 99576\ M\^80 + 27273\ M\^82 + 
                196616\ M\^84 - 322229\ M\^86 - 57083\ M\^88 + 
                486973\ M\^90 + 58318\ M\^92 - 464348\ M\^94 - 
                919397\ M\^96 + 811484\ M\^98 + 2754979\ M\^100 - 
                2103045\ M\^102 - 4463331\ M\^104 + 3769439\ M\^106 + 
                5101851\ M\^108 - 4639354\ M\^110 - 5008564\ M\^112 + 
                4447380\ M\^114 + 5038413\ M\^116 - 4253627\ M\^118 - 
                5007994\ M\^120 + 4478493\ M\^122 + 4064747\ M\^124 - 
                4460477\ M\^126 - 2160028\ M\^128 + 3583698\ M\^130 + 
                356856\ M\^132 - 2114065\ M\^134 + 525906\ M\^136 + 
                793015\ M\^138 - 544627\ M\^140 - 75911\ M\^142 + 
                233365\ M\^144 - 80776\ M\^146 - 51780\ M\^148 + 
                74943\ M\^150 - 45230\ M\^152 + 17117\ M\^154 - 
                4301\ M\^156 + 702\ M\^158 - 68\ M\^160 + 3\ M\^162)\) + 
          L\^14\ \((\(-5\)\ M\^78 + 99\ M\^80 - 884\ M\^82 + 4789\ M\^84 - 
                17548\ M\^86 + 43941\ M\^88 - 66946\ M\^90 + 28332\ M\^92 + 
                102344\ M\^94 - 183293\ M\^96 + 42404\ M\^98 + 
                89975\ M\^100 - 4263\ M\^102 + 290527\ M\^104 - 
                651530\ M\^106 - 636579\ M\^108 + 1800972\ M\^110 + 
                785696\ M\^112 - 2707573\ M\^114 - 1140510\ M\^116 + 
                3053175\ M\^118 + 1901897\ M\^120 - 3217647\ M\^122 - 
                2564280\ M\^124 + 3569487\ M\^126 + 2307805\ M\^128 - 
                3699302\ M\^130 - 1193939\ M\^132 + 3131714\ M\^134 + 
                70013\ M\^136 - 2007692\ M\^138 + 471247\ M\^140 + 
                934486\ M\^142 - 482134\ M\^144 - 277936\ M\^146 + 
                293967\ M\^148 - 2195\ M\^150 - 100303\ M\^152 + 
                34708\ M\^154 + 30854\ M\^156 - 42132\ M\^158 + 
                25877\ M\^160 - 10255\ M\^162 + 2780\ M\^164 - 507\ M\^166 + 
                57\ M\^168 - 3\ M\^170)\) + 
          L\^15\ \((\(-M\^88\) + 22\ M\^90 - 230\ M\^92 + 1503\ M\^94 - 
                6606\ M\^96 + 19456\ M\^98 - 35882\ M\^100 + 30844\ M\^102 + 
                23813\ M\^104 - 102353\ M\^106 + 116665\ M\^108 - 
                10425\ M\^110 - 176881\ M\^112 + 270827\ M\^114 + 
                43024\ M\^116 - 507712\ M\^118 + 84459\ M\^120 + 
                747150\ M\^122 + 90045\ M\^124 - 1244281\ M\^126 - 
                415081\ M\^128 + 1917504\ M\^130 + 493637\ M\^132 - 
                2317650\ M\^134 - 124862\ M\^136 + 2070026\ M\^138 - 
                328118\ M\^140 - 1324449\ M\^142 + 488332\ M\^144 + 
                595068\ M\^146 - 354752\ M\^148 - 183968\ M\^150 + 
                164894\ M\^152 + 48668\ M\^154 - 70071\ M\^156 - 
                2998\ M\^158 + 20842\ M\^160 + 4246\ M\^162 - 20253\ M\^164 + 
                17660\ M\^166 - 9227\ M\^168 + 3345\ M\^170 - 865\ M\^172 + 
                156\ M\^174 - 18\ M\^176 + M\^178)\) + 
          L\^16\ \((M\^100 - 22\ M\^102 + 224\ M\^104 - 1360\ M\^106 + 
                5346\ M\^108 - 13728\ M\^110 + 20749\ M\^112 - 7276\ M\^114 - 
                39312\ M\^116 + 77507\ M\^118 - 30408\ M\^120 - 
                78565\ M\^122 + 98974\ M\^124 + 8962\ M\^126 - 
                48526\ M\^128 - 16337\ M\^130 - 79821\ M\^132 + 
                224647\ M\^134 + 96604\ M\^136 - 507126\ M\^138 + 
                53717\ M\^140 + 670012\ M\^142 - 249800\ M\^144 - 
                570513\ M\^146 + 344522\ M\^148 + 312040\ M\^150 - 
                281248\ M\^152 - 89003\ M\^154 + 147304\ M\^156 - 
                9149\ M\^158 - 39119\ M\^160 + 11036\ M\^162 + 4004\ M\^164 - 
                3304\ M\^166 + 5766\ M\^168 - 8962\ M\^170 + 7861\ M\^172 - 
                4455\ M\^174 + 1774\ M\^176 - 510\ M\^178 + 104\ M\^180 - 
                14\ M\^182 + M\^184)\) + 
          L\^21\ \((6\ M\^170 - 17\ M\^172 + 5\ M\^174 + 47\ M\^176 - 
                30\ M\^178 - 9\ M\^180 + 29\ M\^182 - 19\ M\^184 + 
                8\ M\^186 - 2\ M\^188)\) + 
          L\^17\ \((6\ M\^114 - 95\ M\^116 + 681\ M\^118 - 2827\ M\^120 + 
                7010\ M\^122 - 8750\ M\^124 - 2136\ M\^126 + 25412\ M\^128 - 
                33707\ M\^130 + 1147\ M\^132 + 44694\ M\^134 - 
                44935\ M\^136 + 1247\ M\^138 + 46787\ M\^140 - 
                65833\ M\^142 + 8425\ M\^144 + 87817\ M\^146 - 
                55661\ M\^148 - 65995\ M\^150 + 76311\ M\^152 + 
                39480\ M\^154 - 67726\ M\^156 - 18837\ M\^158 + 
                45523\ M\^160 + 7124\ M\^162 - 26595\ M\^164 + 2815\ M\^166 + 
                9771\ M\^168 - 3493\ M\^170 - 3039\ M\^172 + 4590\ M\^174 - 
                3963\ M\^176 + 2655\ M\^178 - 1311\ M\^180 + 453\ M\^182 - 
                105\ M\^184 + 15\ M\^186 - M\^188)\) + 
          L\^18\ \((15\ M\^128 - 180\ M\^130 + 924\ M\^132 - 2499\ M\^134 + 
                3038\ M\^136 + 1737\ M\^138 - 11038\ M\^140 + 11859\ M\^142 + 
                4908\ M\^144 - 19820\ M\^146 + 11512\ M\^148 + 
                10192\ M\^150 - 19713\ M\^152 + 8090\ M\^154 + 
                14321\ M\^156 - 11708\ M\^158 - 4649\ M\^160 + 7049\ M\^162 + 
                921\ M\^164 - 3122\ M\^166 + 890\ M\^168 - 188\ M\^170 - 
                159\ M\^172 + 907\ M\^174 - 184\ M\^176 - 1310\ M\^178 + 
                1756\ M\^180 - 1173\ M\^182 + 495\ M\^184 - 138\ M\^186 + 
                24\ M\^188 - 2\ M\^190)\) + 
          L\^19\ \((20\ M\^142 - 175\ M\^144 + 596\ M\^146 - 790\ M\^148 - 
                521\ M\^150 + 2973\ M\^152 - 2742\ M\^154 - 1511\ M\^156 + 
                4912\ M\^158 - 2408\ M\^160 - 2421\ M\^162 + 4970\ M\^164 - 
                1264\ M\^166 - 2800\ M\^168 + 2719\ M\^170 - 427\ M\^172 - 
                1050\ M\^174 + 1109\ M\^176 - 316\ M\^178 - 459\ M\^180 + 
                672\ M\^182 - 458\ M\^184 + 200\ M\^186 - 59\ M\^188 + 
                11\ M\^190 - M\^192)\) + 
          L\^20\ \((15\ M\^156 - 86\ M\^158 + 156\ M\^160 + 64\ M\^162 - 
                521\ M\^164 + 473\ M\^166 + 383\ M\^168 - 626\ M\^170 + 
                192\ M\^172 + 283\ M\^174 - 261\ M\^176 + 88\ M\^178 + 
                31\ M\^180 - 110\ M\^182 + 125\ M\^184 - 80\ M\^186 + 
                32\ M\^188 - 8\ M\^190 + M\^192)\)}, {"A-polynomial", 
        Knot[9, 37], \(-M\^42\) + L\^14\ M\^44 + 
          L\ \((\(-3\)\ M\^32 + 14\ M\^34 - 33\ M\^36 + 41\ M\^38 + 
                10\ M\^40 - 73\ M\^42 + 28\ M\^44 + 15\ M\^46 - 15\ M\^48 + 
                4\ M\^50)\) + 
          L\^13\ \((\(-4\)\ M\^36 + 15\ M\^38 - 15\ M\^40 - 28\ M\^42 + 
                73\ M\^44 - 10\ M\^46 - 41\ M\^48 + 33\ M\^50 - 14\ M\^52 + 
                3\ M\^54)\) + 
          L\^2\ \((\(-3\)\ M\^22 + 24\ M\^24 - 86\ M\^26 + 181\ M\^28 - 
                198\ M\^30 + 57\ M\^32 + 148\ M\^34 - 523\ M\^36 + 
                787\ M\^38 + 159\ M\^40 - 1321\ M\^42 + 649\ M\^44 + 
                491\ M\^46 - 683\ M\^48 + 235\ M\^50 + 104\ M\^52 - 
                125\ M\^54 + 45\ M\^56 - 6\ M\^58)\) + 
          L\^12\ \((6\ M\^28 - 45\ M\^30 + 125\ M\^32 - 104\ M\^34 - 
                235\ M\^36 + 683\ M\^38 - 491\ M\^40 - 649\ M\^42 + 
                1321\ M\^44 - 159\ M\^46 - 787\ M\^48 + 523\ M\^50 - 
                148\ M\^52 - 57\ M\^54 + 198\ M\^56 - 181\ M\^58 + 
                86\ M\^60 - 24\ M\^62 + 3\ M\^64)\) + 
          L\^3\ \((\(-M\^12\) + 10\ M\^14 - 45\ M\^16 + 123\ M\^18 - 
                244\ M\^20 + 470\ M\^22 - 930\ M\^24 + 976\ M\^26 + 
                995\ M\^28 - 4536\ M\^30 + 5189\ M\^32 + 2250\ M\^34 - 
                13688\ M\^36 + 10221\ M\^38 + 11805\ M\^40 - 19320\ M\^42 - 
                51\ M\^44 + 13131\ M\^46 - 6980\ M\^48 - 1815\ M\^50 + 
                3874\ M\^52 - 2146\ M\^54 + 320\ M\^56 + 502\ M\^58 - 
                485\ M\^60 + 208\ M\^62 - 45\ M\^64 + 4\ M\^66)\) + 
          L\^4\ \((3\ M\^8 - 40\ M\^10 + 240\ M\^12 - 799\ M\^14 + 
                1445\ M\^16 - 904\ M\^18 - 1598\ M\^20 + 3784\ M\^22 - 
                1710\ M\^24 - 3500\ M\^26 + 3410\ M\^28 - 130\ M\^30 + 
                6554\ M\^32 - 2910\ M\^34 - 25147\ M\^36 + 18139\ M\^38 + 
                32483\ M\^40 - 33654\ M\^42 - 11934\ M\^44 + 27677\ M\^46 - 
                17904\ M\^48 - 1540\ M\^50 + 21265\ M\^52 - 16016\ M\^54 - 
                4164\ M\^56 + 11029\ M\^58 - 4851\ M\^60 - 676\ M\^62 + 
                1774\ M\^64 - 1073\ M\^66 + 405\ M\^68 - 101\ M\^70 + 
                15\ M\^72 - M\^74)\) + 
          L\^11\ \((\(-4\)\ M\^20 + 45\ M\^22 - 208\ M\^24 + 485\ M\^26 - 
                502\ M\^28 - 320\ M\^30 + 2146\ M\^32 - 3874\ M\^34 + 
                1815\ M\^36 + 6980\ M\^38 - 13131\ M\^40 + 51\ M\^42 + 
                19320\ M\^44 - 11805\ M\^46 - 10221\ M\^48 + 13688\ M\^50 - 
                2250\ M\^52 - 5189\ M\^54 + 4536\ M\^56 - 995\ M\^58 - 
                976\ M\^60 + 930\ M\^62 - 470\ M\^64 + 244\ M\^66 - 
                123\ M\^68 + 45\ M\^70 - 10\ M\^72 + M\^74)\) + 
          L\^10\ \((M\^12 - 15\ M\^14 + 101\ M\^16 - 405\ M\^18 + 
                1073\ M\^20 - 1774\ M\^22 + 676\ M\^24 + 4851\ M\^26 - 
                11029\ M\^28 + 4164\ M\^30 + 16016\ M\^32 - 21265\ M\^34 + 
                1540\ M\^36 + 17904\ M\^38 - 27677\ M\^40 + 11934\ M\^42 + 
                33654\ M\^44 - 32483\ M\^46 - 18139\ M\^48 + 25147\ M\^50 + 
                2910\ M\^52 - 6554\ M\^54 + 130\ M\^56 - 3410\ M\^58 + 
                3500\ M\^60 + 1710\ M\^62 - 3784\ M\^64 + 1598\ M\^66 + 
                904\ M\^68 - 1445\ M\^70 + 799\ M\^72 - 240\ M\^74 + 
                40\ M\^76 - 3\ M\^78)\) + 
          L\^5\ \((\(-3\)\ M\^4 + 50\ M\^6 - 355\ M\^8 + 1337\ M\^10 - 
                2616\ M\^12 + 1342\ M\^14 + 5121\ M\^16 - 10298\ M\^18 - 
                668\ M\^20 + 23262\ M\^22 - 17328\ M\^24 - 26187\ M\^26 + 
                37127\ M\^28 + 7085\ M\^30 - 33273\ M\^32 + 25725\ M\^34 + 
                1608\ M\^36 - 44115\ M\^38 + 17503\ M\^40 + 39820\ M\^42 + 
                5442\ M\^44 - 29328\ M\^46 - 41376\ M\^48 + 25870\ M\^50 + 
                45938\ M\^52 - 30182\ M\^54 - 17556\ M\^56 + 24241\ M\^58 - 
                8914\ M\^60 - 5986\ M\^62 + 11195\ M\^64 - 5254\ M\^66 - 
                1605\ M\^68 + 2949\ M\^70 - 1475\ M\^72 + 381\ M\^74 - 
                52\ M\^76 + 3\ M\^78)\) + 
          L\^6\ \((1 - 20\ M\^2 + 161\ M\^4 - 676\ M\^6 + 1531\ M\^8 - 
                1435\ M\^10 - 1144\ M\^12 + 3320\ M\^14 + 1620\ M\^16 - 
                8362\ M\^18 - 2870\ M\^20 + 19952\ M\^22 - 10\ M\^24 - 
                26634\ M\^26 + 8431\ M\^28 - 534\ M\^30 - 8328\ M\^32 + 
                58609\ M\^34 - 4463\ M\^36 - 91970\ M\^38 + 13301\ M\^40 + 
                52596\ M\^42 - 1341\ M\^44 + 27434\ M\^46 - 16570\ M\^48 - 
                68211\ M\^50 + 9830\ M\^52 + 49683\ M\^54 + 12257\ M\^56 - 
                16632\ M\^58 - 26790\ M\^60 + 3153\ M\^62 + 25922\ M\^64 - 
                6111\ M\^66 - 13886\ M\^68 + 8249\ M\^70 + 1825\ M\^72 - 
                3697\ M\^74 + 1755\ M\^76 - 427\ M\^78 + 55\ M\^80 - 
                3\ M\^82)\) + 
          L\^9\ \((\(-3\)\ M\^8 + 52\ M\^10 - 381\ M\^12 + 1475\ M\^14 - 
                2949\ M\^16 + 1605\ M\^18 + 5254\ M\^20 - 11195\ M\^22 + 
                5986\ M\^24 + 8914\ M\^26 - 24241\ M\^28 + 17556\ M\^30 + 
                30182\ M\^32 - 45938\ M\^34 - 25870\ M\^36 + 41376\ M\^38 + 
                29328\ M\^40 - 5442\ M\^42 - 39820\ M\^44 - 17503\ M\^46 + 
                44115\ M\^48 - 1608\ M\^50 - 25725\ M\^52 + 33273\ M\^54 - 
                7085\ M\^56 - 37127\ M\^58 + 26187\ M\^60 + 17328\ M\^62 - 
                23262\ M\^64 + 668\ M\^66 + 10298\ M\^68 - 5121\ M\^70 - 
                1342\ M\^72 + 2616\ M\^74 - 1337\ M\^76 + 355\ M\^78 - 
                50\ M\^80 + 3\ M\^82)\) + 
          L\^8\ \((3\ M\^4 - 55\ M\^6 + 427\ M\^8 - 1755\ M\^10 + 
                3697\ M\^12 - 1825\ M\^14 - 8249\ M\^16 + 13886\ M\^18 + 
                6111\ M\^20 - 25922\ M\^22 - 3153\ M\^24 + 26790\ M\^26 + 
                16632\ M\^28 - 12257\ M\^30 - 49683\ M\^32 - 9830\ M\^34 + 
                68211\ M\^36 + 16570\ M\^38 - 27434\ M\^40 + 1341\ M\^42 - 
                52596\ M\^44 - 13301\ M\^46 + 91970\ M\^48 + 4463\ M\^50 - 
                58609\ M\^52 + 8328\ M\^54 + 534\ M\^56 - 8431\ M\^58 + 
                26634\ M\^60 + 10\ M\^62 - 19952\ M\^64 + 2870\ M\^66 + 
                8362\ M\^68 - 1620\ M\^70 - 3320\ M\^72 + 1144\ M\^74 + 
                1435\ M\^76 - 1531\ M\^78 + 676\ M\^80 - 161\ M\^82 + 
                20\ M\^84 - M\^86)\) + 
          L\^7\ \((\(-1\) + 18\ M\^2 - 140\ M\^4 + 613\ M\^6 - 1630\ M\^8 + 
                2457\ M\^10 - 688\ M\^12 - 5297\ M\^14 + 8418\ M\^16 + 
                5174\ M\^18 - 21540\ M\^20 - 1817\ M\^22 + 34387\ M\^24 - 
                173\ M\^26 - 27231\ M\^28 - 9396\ M\^30 - 17921\ M\^32 + 
                31530\ M\^34 + 68494\ M\^36 - 32198\ M\^38 - 66024\ M\^40 - 
                13071\ M\^42 + 13071\ M\^44 + 66024\ M\^46 + 32198\ M\^48 - 
                68494\ M\^50 - 31530\ M\^52 + 17921\ M\^54 + 9396\ M\^56 + 
                27231\ M\^58 + 173\ M\^60 - 34387\ M\^62 + 1817\ M\^64 + 
                21540\ M\^66 - 5174\ M\^68 - 8418\ M\^70 + 5297\ M\^72 + 
                688\ M\^74 - 2457\ M\^76 + 1630\ M\^78 - 613\ M\^80 + 
                140\ M\^82 - 18\ M\^84 + M\^86)\)}, {"A-polynomial", 
        Knot[9, 38], \(-M\^176\) + 2\ M\^180 - M\^184 + 
          L\^21\ \((1 - 2\ M\^4 + M\^8)\) + 
          L\^20\ \((\(-5\) + 18\ M\^2 - 21\ M\^4 - 72\ M\^6 + 204\ M\^8 + 
                16\ M\^10 - 346\ M\^12 + 144\ M\^14 + 176\ M\^16 - 
                133\ M\^18 - 5\ M\^20 + 36\ M\^22 - 7\ M\^24 - 8\ M\^26 + 
                M\^28 + 3\ M\^30 - M\^32)\) + 
          L\^19\ \((10 - 73\ M\^2 + 230\ M\^4 - 228\ M\^6 - 524\ M\^8 + 
                2022\ M\^10 - 1920\ M\^12 - 4067\ M\^14 + 9416\ M\^16 + 
                1833\ M\^18 - 15717\ M\^20 + 4926\ M\^22 + 12233\ M\^24 - 
                8716\ M\^26 - 3199\ M\^28 + 5775\ M\^30 - 1381\ M\^32 - 
                1735\ M\^34 + 1371\ M\^36 + 32\ M\^38 - 493\ M\^40 + 
                217\ M\^42 + 28\ M\^44 - 59\ M\^46 + 22\ M\^48 - 
                3\ M\^50)\) + 
          L\^18\ \((\(-10\) + 112\ M\^2 - 566\ M\^4 + 1551\ M\^6 - 
                2103\ M\^8 - 757\ M\^10 + 8394\ M\^12 - 12561\ M\^14 - 
                2529\ M\^16 + 43017\ M\^18 - 51976\ M\^20 - 74121\ M\^22 + 
                177461\ M\^24 + 62317\ M\^26 - 307054\ M\^28 + 15846\ M\^30 + 
                320132\ M\^32 - 106102\ M\^34 - 198683\ M\^36 + 
                123108\ M\^38 + 58747\ M\^40 - 66445\ M\^42 + 2298\ M\^44 + 
                13406\ M\^46 - 4319\ M\^48 + 2994\ M\^50 - 2426\ M\^52 - 
                1274\ M\^54 + 2498\ M\^56 - 886\ M\^58 - 369\ M\^60 + 
                444\ M\^62 - 176\ M\^64 + 35\ M\^66 - 3\ M\^68)\) + 
          L\^17\ \((5 - 78\ M\^2 + 558\ M\^4 - 2361\ M\^6 + 6321\ M\^8 - 
                9726\ M\^10 + 2471\ M\^12 + 23786\ M\^14 - 41923\ M\^16 - 
                9552\ M\^18 + 90126\ M\^20 - 45653\ M\^22 - 80313\ M\^24 + 
                191588\ M\^26 - 192466\ M\^28 - 493957\ M\^30 + 
                921843\ M\^32 + 914574\ M\^34 - 1897005\ M\^36 - 
                1210780\ M\^38 + 2526980\ M\^40 + 1122386\ M\^42 - 
                2416054\ M\^44 - 731602\ M\^46 + 1745826\ M\^48 + 
                357117\ M\^50 - 1013501\ M\^52 - 141929\ M\^54 + 
                527160\ M\^56 + 12304\ M\^58 - 244652\ M\^60 + 50195\ M\^62 + 
                80199\ M\^64 - 44828\ M\^66 - 7898\ M\^68 + 16965\ M\^70 - 
                6668\ M\^72 - 761\ M\^74 + 2328\ M\^76 - 1444\ M\^78 + 
                523\ M\^80 - 119\ M\^82 + 16\ M\^84 - M\^86)\) + 
          L\^16\ \((\(-1\) + 22\ M\^2 - 221\ M\^4 + 1345\ M\^6 - 5460\ M\^8 + 
                14996\ M\^10 - 25783\ M\^12 + 16802\ M\^14 + 36555\ M\^16 - 
                98898\ M\^18 + 26717\ M\^20 + 231732\ M\^22 - 171948\ M\^24 - 
                612219\ M\^26 + 444878\ M\^28 + 1564994\ M\^30 - 
                613026\ M\^32 - 3204878\ M\^34 - 87605\ M\^36 + 
                4698332\ M\^38 + 2428696\ M\^40 - 4605974\ M\^42 - 
                5863240\ M\^44 + 2847538\ M\^46 + 8332883\ M\^48 - 
                1111078\ M\^50 - 8273009\ M\^52 + 709076\ M\^54 + 
                6042613\ M\^56 - 1270456\ M\^58 - 3144933\ M\^60 + 
                1712629\ M\^62 + 858874\ M\^64 - 1508171\ M\^66 + 
                318244\ M\^68 + 931440\ M\^70 - 642243\ M\^72 - 
                349693\ M\^74 + 504989\ M\^76 - 6138\ M\^78 - 228935\ M\^80 + 
                90548\ M\^82 + 46221\ M\^84 - 48023\ M\^86 + 7595\ M\^88 + 
                9124\ M\^90 - 6873\ M\^92 + 2397\ M\^94 - 489\ M\^96 + 
                57\ M\^98 - 3\ M\^100)\) + 
          L\^15\ \((\(-M\^2\) + 22\ M\^4 - 239\ M\^6 + 1653\ M\^8 - 
                7710\ M\^10 + 23584\ M\^12 - 41645\ M\^14 + 22347\ M\^16 + 
                46930\ M\^18 - 37403\ M\^20 - 85988\ M\^22 - 83320\ M\^24 + 
                331041\ M\^26 + 561462\ M\^28 - 854818\ M\^30 - 
                1873391\ M\^32 + 1057289\ M\^34 + 4216755\ M\^36 + 
                1255287\ M\^38 - 6746720\ M\^40 - 7703458\ M\^42 + 
                6581005\ M\^44 + 15069595\ M\^46 - 1093839\ M\^48 - 
                15884272\ M\^50 - 7152595\ M\^52 + 6457518\ M\^54 + 
                11907280\ M\^56 + 7365800\ M\^58 - 10236707\ M\^60 - 
                16355262\ M\^62 + 4726216\ M\^64 + 17141660\ M\^66 + 
                419623\ M\^68 - 12706641\ M\^70 - 3129871\ M\^72 + 
                7374437\ M\^74 + 3797967\ M\^76 - 3766640\ M\^78 - 
                3246283\ M\^80 + 2065219\ M\^82 + 2164458\ M\^84 - 
                1394580\ M\^86 - 1033062\ M\^88 + 953828\ M\^90 + 
                220926\ M\^92 - 472101\ M\^94 + 86376\ M\^96 + 
                122488\ M\^98 - 73910\ M\^100 - 2211\ M\^102 + 
                20466\ M\^104 - 11420\ M\^106 + 3409\ M\^108 - 615\ M\^110 + 
                64\ M\^112 - 3\ M\^114)\) + 
          L\^14\ \((4\ M\^6 - 92\ M\^8 + 940\ M\^10 - 5571\ M\^12 + 
                20729\ M\^14 - 47128\ M\^16 + 49370\ M\^18 + 37109\ M\^20 - 
                133170\ M\^22 - 109665\ M\^24 + 570037\ M\^26 + 
                132028\ M\^28 - 1749661\ M\^30 - 203298\ M\^32 + 
                4332949\ M\^34 + 875732\ M\^36 - 8482907\ M\^38 - 
                2456949\ M\^40 + 10610970\ M\^42 + 3401979\ M\^44 - 
                2688072\ M\^46 + 4193679\ M\^48 - 20904346\ M\^50 - 
                33395939\ M\^52 + 50465560\ M\^54 + 87180794\ M\^56 - 
                61336283\ M\^58 - 144461415\ M\^60 + 39774195\ M\^62 + 
                172221235\ M\^64 - 1110407\ M\^66 - 156471387\ M\^68 - 
                27254372\ M\^70 + 112468494\ M\^72 + 32573053\ M\^74 - 
                65088836\ M\^76 - 21756134\ M\^78 + 29486602\ M\^80 + 
                8163765\ M\^82 - 8499228\ M\^84 + 336754\ M\^86 - 
                1617634\ M\^88 - 2500716\ M\^90 + 4832284\ M\^92 + 
                1160276\ M\^94 - 4199836\ M\^96 + 482266\ M\^98 + 
                2234101\ M\^100 - 1015629\ M\^102 - 608788\ M\^104 + 
                646432\ M\^106 - 64123\ M\^108 - 170990\ M\^110 + 
                93125\ M\^112 + 4141\ M\^114 - 28990\ M\^116 + 
                18190\ M\^118 - 6566\ M\^120 + 1561\ M\^122 - 244\ M\^124 + 
                23\ M\^126 - M\^128)\) + 
          L\^13\ \((\(-6\)\ M\^10 + 146\ M\^12 - 1523\ M\^14 + 8824\ M\^16 - 
                30063\ M\^18 + 54708\ M\^20 - 19066\ M\^22 - 108689\ M\^24 + 
                72499\ M\^26 + 451253\ M\^28 - 663084\ M\^30 - 
                887741\ M\^32 + 1935351\ M\^34 + 1957702\ M\^36 - 
                3897547\ M\^38 - 5837263\ M\^40 + 5359309\ M\^42 + 
                17485947\ M\^44 - 1176930\ M\^46 - 43153729\ M\^48 - 
                19285582\ M\^50 + 79469515\ M\^52 + 68376169\ M\^54 - 
                102506939\ M\^56 - 140532847\ M\^58 + 78022126\ M\^60 + 
                192402019\ M\^62 - 1391535\ M\^64 - 166568363\ M\^66 - 
                73987195\ M\^68 + 53638183\ M\^70 + 79840041\ M\^72 + 
                83640211\ M\^74 - 6534907\ M\^76 - 168016454\ M\^78 - 
                90252463\ M\^80 + 175127810\ M\^82 + 149515069\ M\^84 - 
                133399295\ M\^86 - 152513808\ M\^88 + 81824746\ M\^90 + 
                118976429\ M\^92 - 42769829\ M\^94 - 76802085\ M\^96 + 
                19982020\ M\^98 + 43279222\ M\^100 - 9295599\ M\^102 - 
                22063948\ M\^104 + 5194861\ M\^106 + 10054638\ M\^108 - 
                3427512\ M\^110 - 3784769\ M\^112 + 2132272\ M\^114 + 
                913755\ M\^116 - 1013948\ M\^118 + 51394\ M\^120 + 
                259412\ M\^122 - 100410\ M\^124 - 29389\ M\^126 + 
                41358\ M\^128 - 18599\ M\^130 + 4834\ M\^132 - 775\ M\^134 + 
                72\ M\^136 - 3\ M\^138)\) + 
          L\^12\ \((4\ M\^14 - 99\ M\^16 + 1057\ M\^18 - 6293\ M\^20 + 
                22091\ M\^22 - 41741\ M\^24 + 18117\ M\^26 + 72519\ M\^28 - 
                31936\ M\^30 - 372309\ M\^32 + 489306\ M\^34 + 
                685726\ M\^36 - 1270523\ M\^38 - 1527472\ M\^40 + 
                2311944\ M\^42 + 3673919\ M\^44 - 1472476\ M\^46 - 
                8726848\ M\^48 - 8446341\ M\^50 + 17183159\ M\^52 + 
                41612587\ M\^54 - 18675480\ M\^56 - 115741875\ M\^58 - 
                19312720\ M\^60 + 229515653\ M\^62 + 147670452\ M\^64 - 
                330121859\ M\^66 - 390433967\ M\^68 + 317255455\ M\^70 + 
                676511042\ M\^72 - 122595202\ M\^74 - 846815702\ M\^76 - 
                188613049\ M\^78 + 780447623\ M\^80 + 439361990\ M\^82 - 
                516386925\ M\^84 - 495997539\ M\^86 + 215361589\ M\^88 + 
                374372362\ M\^90 - 11792587\ M\^92 - 189953033\ M\^94 - 
                67089309\ M\^96 + 44118003\ M\^98 + 67339887\ M\^100 + 
                30403874\ M\^102 - 43530309\ M\^104 - 49252186\ M\^106 + 
                23965401\ M\^108 + 40761713\ M\^110 - 13873492\ M\^112 - 
                25229178\ M\^114 + 9217172\ M\^116 + 12105907\ M\^118 - 
                5973954\ M\^120 - 4249720\ M\^122 + 3221683\ M\^124 + 
                831323\ M\^126 - 1326397\ M\^128 + 164263\ M\^130 + 
                288550\ M\^132 - 122166\ M\^134 - 33215\ M\^136 + 
                48274\ M\^138 - 21389\ M\^140 + 5413\ M\^142 - 839\ M\^144 + 
                75\ M\^146 - 3\ M\^148)\) + 
          L\^11\ \((\(-M\^18\) + 24\ M\^20 - 258\ M\^22 + 1622\ M\^24 - 
                6533\ M\^26 + 17370\ M\^28 - 30251\ M\^30 + 31907\ M\^32 - 
                7183\ M\^34 - 53739\ M\^36 + 76486\ M\^38 + 274431\ M\^40 - 
                1007895\ M\^42 + 188186\ M\^44 + 3482690\ M\^46 - 
                2567060\ M\^48 - 9511350\ M\^50 + 9358286\ M\^52 + 
                22708737\ M\^54 - 22312157\ M\^56 - 49225903\ M\^58 + 
                36791156\ M\^60 + 96786432\ M\^62 - 35426139\ M\^64 - 
                162477144\ M\^66 - 13819096\ M\^68 + 212368995\ M\^70 + 
                134988748\ M\^72 - 172623725\ M\^74 - 293273409\ M\^76 - 
                29950622\ M\^78 + 368352925\ M\^80 + 370787532\ M\^82 - 
                224071374\ M\^84 - 681373570\ M\^86 - 140959166\ M\^88 + 
                761975453\ M\^90 + 541528711\ M\^92 - 569068879\ M\^94 - 
                758042306\ M\^96 + 252558451\ M\^98 + 721454098\ M\^100 - 
                165071\ M\^102 - 526974143\ M\^104 - 109212374\ M\^106 + 
                310900590\ M\^108 + 108760916\ M\^110 - 151091529\ M\^112 - 
                67597731\ M\^114 + 59906374\ M\^116 + 30802037\ M\^118 - 
                17861774\ M\^120 - 10827679\ M\^122 + 2400380\ M\^124 + 
                3422699\ M\^126 + 1256340\ M\^128 - 1352711\ M\^130 - 
                1122886\ M\^132 + 725837\ M\^134 + 486645\ M\^136 - 
                449612\ M\^138 - 12283\ M\^140 + 112092\ M\^142 - 
                11858\ M\^144 - 38709\ M\^146 + 28287\ M\^148 - 
                10339\ M\^150 + 2324\ M\^152 - 328\ M\^154 + 27\ M\^156 - 
                M\^158)\) + 
          L\^10\ \((M\^26 - 27\ M\^28 + 328\ M\^30 - 2324\ M\^32 + 
                10339\ M\^34 - 28287\ M\^36 + 38709\ M\^38 + 11858\ M\^40 - 
                112092\ M\^42 + 12283\ M\^44 + 449612\ M\^46 - 
                486645\ M\^48 - 725837\ M\^50 + 1122886\ M\^52 + 
                1352711\ M\^54 - 1256340\ M\^56 - 3422699\ M\^58 - 
                2400380\ M\^60 + 10827679\ M\^62 + 17861774\ M\^64 - 
                30802037\ M\^66 - 59906374\ M\^68 + 67597731\ M\^70 + 
                151091529\ M\^72 - 108760916\ M\^74 - 310900590\ M\^76 + 
                109212374\ M\^78 + 526974143\ M\^80 + 165071\ M\^82 - 
                721454098\ M\^84 - 252558451\ M\^86 + 758042306\ M\^88 + 
                569068879\ M\^90 - 541528711\ M\^92 - 761975453\ M\^94 + 
                140959166\ M\^96 + 681373570\ M\^98 + 224071374\ M\^100 - 
                370787532\ M\^102 - 368352925\ M\^104 + 29950622\ M\^106 + 
                293273409\ M\^108 + 172623725\ M\^110 - 134988748\ M\^112 - 
                212368995\ M\^114 + 13819096\ M\^116 + 162477144\ M\^118 + 
                35426139\ M\^120 - 96786432\ M\^122 - 36791156\ M\^124 + 
                49225903\ M\^126 + 22312157\ M\^128 - 22708737\ M\^130 - 
                9358286\ M\^132 + 9511350\ M\^134 + 2567060\ M\^136 - 
                3482690\ M\^138 - 188186\ M\^140 + 1007895\ M\^142 - 
                274431\ M\^144 - 76486\ M\^146 + 53739\ M\^148 + 
                7183\ M\^150 - 31907\ M\^152 + 30251\ M\^154 - 
                17370\ M\^156 + 6533\ M\^158 - 1622\ M\^160 + 258\ M\^162 - 
                24\ M\^164 + M\^166)\) + 
          L\^9\ \((3\ M\^36 - 75\ M\^38 + 839\ M\^40 - 5413\ M\^42 + 
                21389\ M\^44 - 48274\ M\^46 + 33215\ M\^48 + 122166\ M\^50 - 
                288550\ M\^52 - 164263\ M\^54 + 1326397\ M\^56 - 
                831323\ M\^58 - 3221683\ M\^60 + 4249720\ M\^62 + 
                5973954\ M\^64 - 12105907\ M\^66 - 9217172\ M\^68 + 
                25229178\ M\^70 + 13873492\ M\^72 - 40761713\ M\^74 - 
                23965401\ M\^76 + 49252186\ M\^78 + 43530309\ M\^80 - 
                30403874\ M\^82 - 67339887\ M\^84 - 44118003\ M\^86 + 
                67089309\ M\^88 + 189953033\ M\^90 + 11792587\ M\^92 - 
                374372362\ M\^94 - 215361589\ M\^96 + 495997539\ M\^98 + 
                516386925\ M\^100 - 439361990\ M\^102 - 780447623\ M\^104 + 
                188613049\ M\^106 + 846815702\ M\^108 + 122595202\ M\^110 - 
                676511042\ M\^112 - 317255455\ M\^114 + 390433967\ M\^116 + 
                330121859\ M\^118 - 147670452\ M\^120 - 229515653\ M\^122 + 
                19312720\ M\^124 + 115741875\ M\^126 + 18675480\ M\^128 - 
                41612587\ M\^130 - 17183159\ M\^132 + 8446341\ M\^134 + 
                8726848\ M\^136 + 1472476\ M\^138 - 3673919\ M\^140 - 
                2311944\ M\^142 + 1527472\ M\^144 + 1270523\ M\^146 - 
                685726\ M\^148 - 489306\ M\^150 + 372309\ M\^152 + 
                31936\ M\^154 - 72519\ M\^156 - 18117\ M\^158 + 
                41741\ M\^160 - 22091\ M\^162 + 6293\ M\^164 - 1057\ M\^166 + 
                99\ M\^168 - 4\ M\^170)\) + 
          L\^8\ \((3\ M\^46 - 72\ M\^48 + 775\ M\^50 - 4834\ M\^52 + 
                18599\ M\^54 - 41358\ M\^56 + 29389\ M\^58 + 100410\ M\^60 - 
                259412\ M\^62 - 51394\ M\^64 + 1013948\ M\^66 - 
                913755\ M\^68 - 2132272\ M\^70 + 3784769\ M\^72 + 
                3427512\ M\^74 - 10054638\ M\^76 - 5194861\ M\^78 + 
                22063948\ M\^80 + 9295599\ M\^82 - 43279222\ M\^84 - 
                19982020\ M\^86 + 76802085\ M\^88 + 42769829\ M\^90 - 
                118976429\ M\^92 - 81824746\ M\^94 + 152513808\ M\^96 + 
                133399295\ M\^98 - 149515069\ M\^100 - 175127810\ M\^102 + 
                90252463\ M\^104 + 168016454\ M\^106 + 6534907\ M\^108 - 
                83640211\ M\^110 - 79840041\ M\^112 - 53638183\ M\^114 + 
                73987195\ M\^116 + 166568363\ M\^118 + 1391535\ M\^120 - 
                192402019\ M\^122 - 78022126\ M\^124 + 140532847\ M\^126 + 
                102506939\ M\^128 - 68376169\ M\^130 - 79469515\ M\^132 + 
                19285582\ M\^134 + 43153729\ M\^136 + 1176930\ M\^138 - 
                17485947\ M\^140 - 5359309\ M\^142 + 5837263\ M\^144 + 
                3897547\ M\^146 - 1957702\ M\^148 - 1935351\ M\^150 + 
                887741\ M\^152 + 663084\ M\^154 - 451253\ M\^156 - 
                72499\ M\^158 + 108689\ M\^160 + 19066\ M\^162 - 
                54708\ M\^164 + 30063\ M\^166 - 8824\ M\^168 + 1523\ M\^170 - 
                146\ M\^172 + 6\ M\^174)\) + 
          L\^7\ \((M\^56 - 23\ M\^58 + 244\ M\^60 - 1561\ M\^62 + 
                6566\ M\^64 - 18190\ M\^66 + 28990\ M\^68 - 4141\ M\^70 - 
                93125\ M\^72 + 170990\ M\^74 + 64123\ M\^76 - 646432\ M\^78 + 
                608788\ M\^80 + 1015629\ M\^82 - 2234101\ M\^84 - 
                482266\ M\^86 + 4199836\ M\^88 - 1160276\ M\^90 - 
                4832284\ M\^92 + 2500716\ M\^94 + 1617634\ M\^96 - 
                336754\ M\^98 + 8499228\ M\^100 - 8163765\ M\^102 - 
                29486602\ M\^104 + 21756134\ M\^106 + 65088836\ M\^108 - 
                32573053\ M\^110 - 112468494\ M\^112 + 27254372\ M\^114 + 
                156471387\ M\^116 + 1110407\ M\^118 - 172221235\ M\^120 - 
                39774195\ M\^122 + 144461415\ M\^124 + 61336283\ M\^126 - 
                87180794\ M\^128 - 50465560\ M\^130 + 33395939\ M\^132 + 
                20904346\ M\^134 - 4193679\ M\^136 + 2688072\ M\^138 - 
                3401979\ M\^140 - 10610970\ M\^142 + 2456949\ M\^144 + 
                8482907\ M\^146 - 875732\ M\^148 - 4332949\ M\^150 + 
                203298\ M\^152 + 1749661\ M\^154 - 132028\ M\^156 - 
                570037\ M\^158 + 109665\ M\^160 + 133170\ M\^162 - 
                37109\ M\^164 - 49370\ M\^166 + 47128\ M\^168 - 
                20729\ M\^170 + 5571\ M\^172 - 940\ M\^174 + 92\ M\^176 - 
                4\ M\^178)\) + 
          L\^6\ \((3\ M\^70 - 64\ M\^72 + 615\ M\^74 - 3409\ M\^76 + 
                11420\ M\^78 - 20466\ M\^80 + 2211\ M\^82 + 73910\ M\^84 - 
                122488\ M\^86 - 86376\ M\^88 + 472101\ M\^90 - 
                220926\ M\^92 - 953828\ M\^94 + 1033062\ M\^96 + 
                1394580\ M\^98 - 2164458\ M\^100 - 2065219\ M\^102 + 
                3246283\ M\^104 + 3766640\ M\^106 - 3797967\ M\^108 - 
                7374437\ M\^110 + 3129871\ M\^112 + 12706641\ M\^114 - 
                419623\ M\^116 - 17141660\ M\^118 - 4726216\ M\^120 + 
                16355262\ M\^122 + 10236707\ M\^124 - 7365800\ M\^126 - 
                11907280\ M\^128 - 6457518\ M\^130 + 7152595\ M\^132 + 
                15884272\ M\^134 + 1093839\ M\^136 - 15069595\ M\^138 - 
                6581005\ M\^140 + 7703458\ M\^142 + 6746720\ M\^144 - 
                1255287\ M\^146 - 4216755\ M\^148 - 1057289\ M\^150 + 
                1873391\ M\^152 + 854818\ M\^154 - 561462\ M\^156 - 
                331041\ M\^158 + 83320\ M\^160 + 85988\ M\^162 + 
                37403\ M\^164 - 46930\ M\^166 - 22347\ M\^168 + 
                41645\ M\^170 - 23584\ M\^172 + 7710\ M\^174 - 1653\ M\^176 + 
                239\ M\^178 - 22\ M\^180 + M\^182)\) + 
          L\^2\ \((3\ M\^134 - 22\ M\^136 + 59\ M\^138 - 28\ M\^140 - 
                217\ M\^142 + 493\ M\^144 - 32\ M\^146 - 1371\ M\^148 + 
                1735\ M\^150 + 1381\ M\^152 - 5775\ M\^154 + 3199\ M\^156 + 
                8716\ M\^158 - 12233\ M\^160 - 4926\ M\^162 + 15717\ M\^164 - 
                1833\ M\^166 - 9416\ M\^168 + 4067\ M\^170 + 1920\ M\^172 - 
                2022\ M\^174 + 524\ M\^176 + 228\ M\^178 - 230\ M\^180 + 
                73\ M\^182 - 10\ M\^184)\) + 
          L\^4\ \((M\^98 - 16\ M\^100 + 119\ M\^102 - 523\ M\^104 + 
                1444\ M\^106 - 2328\ M\^108 + 761\ M\^110 + 6668\ M\^112 - 
                16965\ M\^114 + 7898\ M\^116 + 44828\ M\^118 - 
                80199\ M\^120 - 50195\ M\^122 + 244652\ M\^124 - 
                12304\ M\^126 - 527160\ M\^128 + 141929\ M\^130 + 
                1013501\ M\^132 - 357117\ M\^134 - 1745826\ M\^136 + 
                731602\ M\^138 + 2416054\ M\^140 - 1122386\ M\^142 - 
                2526980\ M\^144 + 1210780\ M\^146 + 1897005\ M\^148 - 
                914574\ M\^150 - 921843\ M\^152 + 493957\ M\^154 + 
                192466\ M\^156 - 191588\ M\^158 + 80313\ M\^160 + 
                45653\ M\^162 - 90126\ M\^164 + 9552\ M\^166 + 
                41923\ M\^168 - 23786\ M\^170 - 2471\ M\^172 + 9726\ M\^174 - 
                6321\ M\^176 + 2361\ M\^178 - 558\ M\^180 + 78\ M\^182 - 
                5\ M\^184)\) + 
          L\^5\ \((3\ M\^84 - 57\ M\^86 + 489\ M\^88 - 2397\ M\^90 + 
                6873\ M\^92 - 9124\ M\^94 - 7595\ M\^96 + 48023\ M\^98 - 
                46221\ M\^100 - 90548\ M\^102 + 228935\ M\^104 + 
                6138\ M\^106 - 504989\ M\^108 + 349693\ M\^110 + 
                642243\ M\^112 - 931440\ M\^114 - 318244\ M\^116 + 
                1508171\ M\^118 - 858874\ M\^120 - 1712629\ M\^122 + 
                3144933\ M\^124 + 1270456\ M\^126 - 6042613\ M\^128 - 
                709076\ M\^130 + 8273009\ M\^132 + 1111078\ M\^134 - 
                8332883\ M\^136 - 2847538\ M\^138 + 5863240\ M\^140 + 
                4605974\ M\^142 - 2428696\ M\^144 - 4698332\ M\^146 + 
                87605\ M\^148 + 3204878\ M\^150 + 613026\ M\^152 - 
                1564994\ M\^154 - 444878\ M\^156 + 612219\ M\^158 + 
                171948\ M\^160 - 231732\ M\^162 - 26717\ M\^164 + 
                98898\ M\^166 - 36555\ M\^168 - 16802\ M\^170 + 
                25783\ M\^172 - 14996\ M\^174 + 5460\ M\^176 - 1345\ M\^178 + 
                221\ M\^180 - 22\ M\^182 + M\^184)\) + 
          L\ \((M\^152 - 3\ M\^154 - M\^156 + 8\ M\^158 + 7\ M\^160 - 
                36\ M\^162 + 5\ M\^164 + 133\ M\^166 - 176\ M\^168 - 
                144\ M\^170 + 346\ M\^172 - 16\ M\^174 - 204\ M\^176 + 
                72\ M\^178 + 21\ M\^180 - 18\ M\^182 + 5\ M\^184)\) + 
          L\^3\ \((3\ M\^116 - 35\ M\^118 + 176\ M\^120 - 444\ M\^122 + 
                369\ M\^124 + 886\ M\^126 - 2498\ M\^128 + 1274\ M\^130 + 
                2426\ M\^132 - 2994\ M\^134 + 4319\ M\^136 - 13406\ M\^138 - 
                2298\ M\^140 + 66445\ M\^142 - 58747\ M\^144 - 
                123108\ M\^146 + 198683\ M\^148 + 106102\ M\^150 - 
                320132\ M\^152 - 15846\ M\^154 + 307054\ M\^156 - 
                62317\ M\^158 - 177461\ M\^160 + 74121\ M\^162 + 
                51976\ M\^164 - 43017\ M\^166 + 2529\ M\^168 + 
                12561\ M\^170 - 8394\ M\^172 + 757\ M\^174 + 2103\ M\^176 - 
                1551\ M\^178 + 566\ M\^180 - 112\ M\^182 + 
                10\ M\^184)\)}, {"A-polynomial", Knot[9, 39], 
        Null}, {"A-polynomial", Knot[9, 40], Null}, {"A-polynomial", 
        Knot[9, 41], 
        M\^38 + L\^15\ M\^52 + 
          L\ \((\(-2\)\ M\^26 + 9\ M\^28 - 22\ M\^30 + 29\ M\^32 + 9\ M\^34 - 
                61\ M\^36 + 54\ M\^38 - 11\ M\^42 + 3\ M\^48 - M\^50)\) + 
          L\^2\ \((M\^14 - 9\ M\^16 + 44\ M\^18 - 132\ M\^20 + 246\ M\^22 - 
                264\ M\^24 + 77\ M\^26 + 209\ M\^28 - 186\ M\^30 - 
                286\ M\^32 + 545\ M\^34 - 139\ M\^36 - 239\ M\^38 + 
                148\ M\^40 + 155\ M\^42 - 264\ M\^44 + M\^46 + 232\ M\^48 - 
                124\ M\^50 - 24\ M\^52 + 33\ M\^54 - 8\ M\^56 + M\^58)\) + 
          L\^14\ \((\(-M\^40\) + 3\ M\^42 - 11\ M\^48 + 54\ M\^52 - 
                61\ M\^54 + 9\ M\^56 + 29\ M\^58 - 22\ M\^60 + 9\ M\^62 - 
                2\ M\^64)\) + 
          L\^3\ \((\(-M\^6\) + 12\ M\^8 - 69\ M\^10 + 239\ M\^12 - 
                516\ M\^14 + 647\ M\^16 - 332\ M\^18 - 87\ M\^20 - 
                273\ M\^22 + 1245\ M\^24 - 950\ M\^26 - 709\ M\^28 - 
                616\ M\^30 + 4990\ M\^32 - 2686\ M\^34 - 5826\ M\^36 + 
                3868\ M\^38 + 7676\ M\^40 - 7252\ M\^42 - 4298\ M\^44 + 
                5879\ M\^46 + 2165\ M\^48 - 4390\ M\^50 + 447\ M\^52 + 
                1397\ M\^54 - 510\ M\^56 - 145\ M\^58 + 137\ M\^60 - 
                40\ M\^62 + 5\ M\^64)\) + 
          L\^4\ \((M\^2 - 13\ M\^4 + 89\ M\^6 - 390\ M\^8 + 1139\ M\^10 - 
                2060\ M\^12 + 1494\ M\^14 + 2148\ M\^16 - 4947\ M\^18 - 
                2096\ M\^20 + 16233\ M\^22 - 13994\ M\^24 - 14306\ M\^26 + 
                27056\ M\^28 + 14521\ M\^30 - 49663\ M\^32 - 1030\ M\^34 + 
                52691\ M\^36 + 6829\ M\^38 - 60070\ M\^40 - 6166\ M\^42 + 
                46774\ M\^44 + 16065\ M\^46 - 39236\ M\^48 - 8822\ M\^50 + 
                21065\ M\^52 + 5333\ M\^54 - 11570\ M\^56 + 1267\ M\^58 + 
                2353\ M\^60 - 451\ M\^62 - 581\ M\^64 + 384\ M\^66 - 
                100\ M\^68 + 10\ M\^70)\) + 
          L\^13\ \((M\^32 - 8\ M\^34 + 33\ M\^36 - 24\ M\^38 - 124\ M\^40 + 
                232\ M\^42 + M\^44 - 264\ M\^46 + 155\ M\^48 + 148\ M\^50 - 
                239\ M\^52 - 139\ M\^54 + 545\ M\^56 - 286\ M\^58 - 
                186\ M\^60 + 209\ M\^62 + 77\ M\^64 - 264\ M\^66 + 
                246\ M\^68 - 132\ M\^70 + 44\ M\^72 - 9\ M\^74 + M\^76)\) + 
          L\^5\ \((1 - 17\ M\^2 + 133\ M\^4 - 625\ M\^6 + 1854\ M\^8 - 
                3196\ M\^10 + 1502\ M\^12 + 5840\ M\^14 - 9544\ M\^16 - 
                10613\ M\^18 + 43763\ M\^20 - 25073\ M\^22 - 62387\ M\^24 + 
                87490\ M\^26 + 61364\ M\^28 - 200077\ M\^30 + 27219\ M\^32 + 
                266193\ M\^34 - 115415\ M\^36 - 303542\ M\^38 + 
                195138\ M\^40 + 241603\ M\^42 - 162238\ M\^44 - 
                173585\ M\^46 + 106498\ M\^48 + 80365\ M\^50 - 27610\ M\^52 - 
                41465\ M\^54 + 5636\ M\^56 + 10651\ M\^58 + 6845\ M\^60 - 
                7907\ M\^62 - 51\ M\^64 + 868\ M\^66 + 1252\ M\^68 - 
                1448\ M\^70 + 616\ M\^72 - 125\ M\^74 + 10\ M\^76)\) + 
          L\^6\ \((\(-2\) + 35\ M\^2 - 266\ M\^4 + 1114\ M\^6 - 2627\ M\^8 + 
                2757\ M\^10 + 1587\ M\^12 - 5893\ M\^14 - 4938\ M\^16 + 
                28577\ M\^18 - 21142\ M\^20 - 35589\ M\^22 + 54050\ M\^24 + 
                45062\ M\^26 - 151983\ M\^28 + 50814\ M\^30 + 225232\ M\^32 - 
                193958\ M\^34 - 326061\ M\^36 + 426459\ M\^38 + 
                317562\ M\^40 - 527165\ M\^42 - 296314\ M\^44 + 
                533658\ M\^46 + 164600\ M\^48 - 362660\ M\^50 - 
                79778\ M\^52 + 215317\ M\^54 - 11087\ M\^56 - 73896\ M\^58 + 
                16209\ M\^60 + 26435\ M\^62 - 22575\ M\^64 + 4705\ M\^66 + 
                4093\ M\^68 - 1815\ M\^70 - 2350\ M\^72 + 3044\ M\^74 - 
                1633\ M\^76 + 482\ M\^78 - 76\ M\^80 + 5\ M\^82)\) + 
          L\^12\ \((5\ M\^26 - 40\ M\^28 + 137\ M\^30 - 145\ M\^32 - 
                510\ M\^34 + 1397\ M\^36 + 447\ M\^38 - 4390\ M\^40 + 
                2165\ M\^42 + 5879\ M\^44 - 4298\ M\^46 - 7252\ M\^48 + 
                7676\ M\^50 + 3868\ M\^52 - 5826\ M\^54 - 2686\ M\^56 + 
                4990\ M\^58 - 616\ M\^60 - 709\ M\^62 - 950\ M\^64 + 
                1245\ M\^66 - 273\ M\^68 - 87\ M\^70 - 332\ M\^72 + 
                647\ M\^74 - 516\ M\^76 + 239\ M\^78 - 69\ M\^80 + 
                12\ M\^82 - M\^84)\) + 
          L\^7\ \((1 - 20\ M\^2 + 173\ M\^4 - 838\ M\^6 + 2411\ M\^8 - 
                3713\ M\^10 + 635\ M\^12 + 8134\ M\^14 - 8096\ M\^16 - 
                22841\ M\^18 + 60307\ M\^20 - 15106\ M\^22 - 117354\ M\^24 + 
                104878\ M\^26 + 201580\ M\^28 - 333147\ M\^30 - 
                183831\ M\^32 + 550598\ M\^34 + 152729\ M\^36 - 
                754320\ M\^38 - 13068\ M\^40 + 710497\ M\^42 - 45050\ M\^44 - 
                567723\ M\^46 + 133149\ M\^48 + 279075\ M\^50 - 
                90040\ M\^52 - 107964\ M\^54 + 73965\ M\^56 - 23756\ M\^58 + 
                2994\ M\^60 + 20096\ M\^62 - 11009\ M\^64 - 19419\ M\^66 + 
                25100\ M\^68 - 7388\ M\^70 - 4516\ M\^72 + 2738\ M\^74 + 
                2081\ M\^76 - 3206\ M\^78 + 1901\ M\^80 - 665\ M\^82 + 
                144\ M\^84 - 18\ M\^86 + M\^88)\) + 
          L\^11\ \((10\ M\^20 - 100\ M\^22 + 384\ M\^24 - 581\ M\^26 - 
                451\ M\^28 + 2353\ M\^30 + 1267\ M\^32 - 11570\ M\^34 + 
                5333\ M\^36 + 21065\ M\^38 - 8822\ M\^40 - 39236\ M\^42 + 
                16065\ M\^44 + 46774\ M\^46 - 6166\ M\^48 - 60070\ M\^50 + 
                6829\ M\^52 + 52691\ M\^54 - 1030\ M\^56 - 49663\ M\^58 + 
                14521\ M\^60 + 27056\ M\^62 - 14306\ M\^64 - 13994\ M\^66 + 
                16233\ M\^68 - 2096\ M\^70 - 4947\ M\^72 + 2148\ M\^74 + 
                1494\ M\^76 - 2060\ M\^78 + 1139\ M\^80 - 390\ M\^82 + 
                89\ M\^84 - 13\ M\^86 + M\^88)\) + 
          L\^9\ \((5\ M\^8 - 76\ M\^10 + 482\ M\^12 - 1633\ M\^14 + 
                3044\ M\^16 - 2350\ M\^18 - 1815\ M\^20 + 4093\ M\^22 + 
                4705\ M\^24 - 22575\ M\^26 + 26435\ M\^28 + 16209\ M\^30 - 
                73896\ M\^32 - 11087\ M\^34 + 215317\ M\^36 - 79778\ M\^38 - 
                362660\ M\^40 + 164600\ M\^42 + 533658\ M\^44 - 
                296314\ M\^46 - 527165\ M\^48 + 317562\ M\^50 + 
                426459\ M\^52 - 326061\ M\^54 - 193958\ M\^56 + 
                225232\ M\^58 + 50814\ M\^60 - 151983\ M\^62 + 45062\ M\^64 + 
                54050\ M\^66 - 35589\ M\^68 - 21142\ M\^70 + 28577\ M\^72 - 
                4938\ M\^74 - 5893\ M\^76 + 1587\ M\^78 + 2757\ M\^80 - 
                2627\ M\^82 + 1114\ M\^84 - 266\ M\^86 + 35\ M\^88 - 
                2\ M\^90)\) + 
          L\^8\ \((M\^2 - 18\ M\^4 + 144\ M\^6 - 665\ M\^8 + 1901\ M\^10 - 
                3206\ M\^12 + 2081\ M\^14 + 2738\ M\^16 - 4516\ M\^18 - 
                7388\ M\^20 + 25100\ M\^22 - 19419\ M\^24 - 11009\ M\^26 + 
                20096\ M\^28 + 2994\ M\^30 - 23756\ M\^32 + 73965\ M\^34 - 
                107964\ M\^36 - 90040\ M\^38 + 279075\ M\^40 + 
                133149\ M\^42 - 567723\ M\^44 - 45050\ M\^46 + 
                710497\ M\^48 - 13068\ M\^50 - 754320\ M\^52 + 
                152729\ M\^54 + 550598\ M\^56 - 183831\ M\^58 - 
                333147\ M\^60 + 201580\ M\^62 + 104878\ M\^64 - 
                117354\ M\^66 - 15106\ M\^68 + 60307\ M\^70 - 22841\ M\^72 - 
                8096\ M\^74 + 8134\ M\^76 + 635\ M\^78 - 3713\ M\^80 + 
                2411\ M\^82 - 838\ M\^84 + 173\ M\^86 - 20\ M\^88 + 
                M\^90)\) + 
          L\^10\ \((10\ M\^14 - 125\ M\^16 + 616\ M\^18 - 1448\ M\^20 + 
                1252\ M\^22 + 868\ M\^24 - 51\ M\^26 - 7907\ M\^28 + 
                6845\ M\^30 + 10651\ M\^32 + 5636\ M\^34 - 41465\ M\^36 - 
                27610\ M\^38 + 80365\ M\^40 + 106498\ M\^42 - 173585\ M\^44 - 
                162238\ M\^46 + 241603\ M\^48 + 195138\ M\^50 - 
                303542\ M\^52 - 115415\ M\^54 + 266193\ M\^56 + 
                27219\ M\^58 - 200077\ M\^60 + 61364\ M\^62 + 87490\ M\^64 - 
                62387\ M\^66 - 25073\ M\^68 + 43763\ M\^70 - 10613\ M\^72 - 
                9544\ M\^74 + 5840\ M\^76 + 1502\ M\^78 - 3196\ M\^80 + 
                1854\ M\^82 - 625\ M\^84 + 133\ M\^86 - 17\ M\^88 + 
                M\^90)\)}, {"A-polynomial", Knot[9, 42], 
        M\^8 + L\^7\ M\^26 + 
          L\ \((\(-1\) + M\^2 + 3\ M\^6 - 3\ M\^8 + 4\ M\^12 - M\^14)\) + 
          L\^2\ \((\(-M\^4\) - M\^6 + 10\ M\^8 - 3\ M\^10 - 9\ M\^12 + 
                M\^14 + 4\ M\^16 + M\^18 - M\^20)\) + 
          L\^3\ \((\(-M\^6\) - 4\ M\^8 + 14\ M\^10 + 9\ M\^12 - 18\ M\^14 - 
                18\ M\^16 + M\^18 + 13\ M\^20 + 4\ M\^22 - 6\ M\^24 + 
                M\^26)\) + 
          L\^4\ \((M\^8 - 6\ M\^10 + 4\ M\^12 + 13\ M\^14 + M\^16 - 
                18\ M\^18 - 18\ M\^20 + 9\ M\^22 + 14\ M\^24 - 4\ M\^26 - 
                M\^28)\) + 
          L\^5\ \((\(-M\^14\) + M\^16 + 4\ M\^18 + M\^20 - 9\ M\^22 - 
                3\ M\^24 + 10\ M\^26 - M\^28 - M\^30)\) + 
          L\^6\ \((\(-M\^20\) + 4\ M\^22 - 3\ M\^26 + 3\ M\^28 + M\^32 - 
                M\^34)\)}, {"A-polynomial", Knot[9, 43], 
        M\^4 + L\^10\ M\^84 + 
          L\ \((\(-1\) + 2\ M\^2 - 3\ M\^4 + 7\ M\^6 - 3\ M\^8 - 4\ M\^10 + 
                7\ M\^12 + M\^14)\) + 
          L\^2\ \((\(-2\)\ M\^6 + 9\ M\^8 - 17\ M\^10 + 14\ M\^12 + 
                2\ M\^14 + 2\ M\^16 + 7\ M\^18 - 14\ M\^20 + 8\ M\^22 + 
                4\ M\^24)\) + 
          L\^3\ \((\(-M\^12\) + 8\ M\^14 - 23\ M\^16 + 26\ M\^18 - 2\ M\^24 - 
                2\ M\^26 + 19\ M\^28 - 28\ M\^30 - 3\ M\^32 + 16\ M\^34 - 
                2\ M\^36)\) + 
          L\^4\ \((M\^20 - 6\ M\^22 + 18\ M\^24 - 28\ M\^26 + 13\ M\^28 + 
                33\ M\^30 + 14\ M\^32 - 38\ M\^34 - 59\ M\^36 + 31\ M\^38 + 
                17\ M\^40 - 22\ M\^42 + 12\ M\^44)\) + 
          L\^5\ \((\(-M\^32\) + 15\ M\^34 - 37\ M\^36 + 22\ M\^38 + 
                69\ M\^40 - 31\ M\^42 - 102\ M\^44 - 31\ M\^46 + 69\ M\^48 + 
                22\ M\^50 - 37\ M\^52 + 15\ M\^54 - M\^56)\) + 
          L\^6\ \((12\ M\^44 - 22\ M\^46 + 17\ M\^48 + 31\ M\^50 - 
                59\ M\^52 - 38\ M\^54 + 14\ M\^56 + 33\ M\^58 + 13\ M\^60 - 
                28\ M\^62 + 18\ M\^64 - 6\ M\^66 + M\^68)\) + 
          L\^7\ \((\(-2\)\ M\^52 + 16\ M\^54 - 3\ M\^56 - 28\ M\^58 + 
                19\ M\^60 - 2\ M\^62 - 2\ M\^64 + 26\ M\^70 - 23\ M\^72 + 
                8\ M\^74 - M\^76)\) + 
          L\^8\ \((4\ M\^64 + 8\ M\^66 - 14\ M\^68 + 7\ M\^70 + 2\ M\^72 + 
                2\ M\^74 + 14\ M\^76 - 17\ M\^78 + 9\ M\^80 - 2\ M\^82)\) + 
          L\^9\ \((M\^74 + 7\ M\^76 - 4\ M\^78 - 3\ M\^80 + 7\ M\^82 - 
                3\ M\^84 + 2\ M\^86 - M\^88)\)}, {"A-polynomial", 
        Knot[9, 44], 
        L\^12\ M\^20 + M\^24 + 
          L\ \((2\ M\^14 - 4\ M\^16 + 6\ M\^18 - 11\ M\^20 + 2\ M\^22 + 
                16\ M\^24 - 2\ M\^26 - M\^28)\) + 
          L\^11\ \((\(-M\^16\) - 2\ M\^18 + 16\ M\^20 + 2\ M\^22 - 
                11\ M\^24 + 6\ M\^26 - 4\ M\^28 + 2\ M\^30)\) + 
          L\^2\ \((M\^4 - 4\ M\^6 + 10\ M\^8 - 14\ M\^10 - M\^12 + 
                16\ M\^14 + 31\ M\^16 - 60\ M\^18 - 15\ M\^20 + 25\ M\^22 + 
                11\ M\^24 + 46\ M\^26 - 13\ M\^28 - 11\ M\^30 + 4\ M\^32)\) + 
          L\^3\ \((2\ M\^2 - 15\ M\^4 + 41\ M\^6 - 39\ M\^8 - M\^10 + 
                4\ M\^12 + 16\ M\^14 + 19\ M\^16 + 42\ M\^18 - 117\ M\^20 - 
                79\ M\^22 + 88\ M\^24 + 62\ M\^26 + 86\ M\^28 - 63\ M\^30 - 
                32\ M\^32 + 35\ M\^34 - 10\ M\^36 + M\^38)\) + 
          L\^10\ \((4\ M\^12 - 11\ M\^14 - 13\ M\^16 + 46\ M\^18 + 
                11\ M\^20 + 25\ M\^22 - 15\ M\^24 - 60\ M\^26 + 31\ M\^28 + 
                16\ M\^30 - M\^32 - 14\ M\^34 + 10\ M\^36 - 4\ M\^38 + 
                M\^40)\) + 
          L\^4\ \((1 - 12\ M\^2 + 49\ M\^4 - 88\ M\^6 + 64\ M\^8 + 
                18\ M\^10 - 40\ M\^12 + 7\ M\^14 + 5\ M\^16 + 27\ M\^18 + 
                76\ M\^20 - 199\ M\^22 - 195\ M\^24 + 202\ M\^26 + 
                117\ M\^28 + 77\ M\^30 - 79\ M\^32 - 67\ M\^34 + 71\ M\^36 - 
                21\ M\^38 + 2\ M\^40)\) + 
          L\^6\ \((\(-3\)\ M\^2 + 35\ M\^4 - 135\ M\^6 + 175\ M\^8 + 
                73\ M\^10 - 249\ M\^12 - 101\ M\^14 + 262\ M\^16 + 
                502\ M\^18 - 195\ M\^20 - 812\ M\^22 - 195\ M\^24 + 
                502\ M\^26 + 262\ M\^28 - 101\ M\^30 - 249\ M\^32 + 
                73\ M\^34 + 175\ M\^36 - 135\ M\^38 + 35\ M\^40 - 
                3\ M\^42)\) + 
          L\^5\ \((\(-1\) + 14\ M\^2 - 68\ M\^4 + 131\ M\^6 - 66\ M\^8 - 
                48\ M\^10 + 7\ M\^12 - 24\ M\^14 + 220\ M\^16 + 117\ M\^18 - 
                312\ M\^20 - 275\ M\^22 + 21\ M\^24 + 148\ M\^26 + 
                151\ M\^28 - 133\ M\^30 + 71\ M\^32 + 72\ M\^34 - 
                122\ M\^36 + 61\ M\^38 - 13\ M\^40 + M\^42)\) + 
          L\^9\ \((M\^6 - 10\ M\^8 + 35\ M\^10 - 32\ M\^12 - 63\ M\^14 + 
                86\ M\^16 + 62\ M\^18 + 88\ M\^20 - 79\ M\^22 - 117\ M\^24 + 
                42\ M\^26 + 19\ M\^28 + 16\ M\^30 + 4\ M\^32 - M\^34 - 
                39\ M\^36 + 41\ M\^38 - 15\ M\^40 + 2\ M\^42)\) + 
          L\^7\ \((M\^2 - 13\ M\^4 + 61\ M\^6 - 122\ M\^8 + 72\ M\^10 + 
                71\ M\^12 - 133\ M\^14 + 151\ M\^16 + 148\ M\^18 + 
                21\ M\^20 - 275\ M\^22 - 312\ M\^24 + 117\ M\^26 + 
                220\ M\^28 - 24\ M\^30 + 7\ M\^32 - 48\ M\^34 - 66\ M\^36 + 
                131\ M\^38 - 68\ M\^40 + 14\ M\^42 - M\^44)\) + 
          L\^8\ \((2\ M\^4 - 21\ M\^6 + 71\ M\^8 - 67\ M\^10 - 79\ M\^12 + 
                77\ M\^14 + 117\ M\^16 + 202\ M\^18 - 195\ M\^20 - 
                199\ M\^22 + 76\ M\^24 + 27\ M\^26 + 5\ M\^28 + 7\ M\^30 - 
                40\ M\^32 + 18\ M\^34 + 64\ M\^36 - 88\ M\^38 + 49\ M\^40 - 
                12\ M\^42 + M\^44)\)}, {"A-polynomial", Knot[9, 45], 
        L\^15\ M\^4 + M\^62 + 
          L\^14\ \((\(-4\)\ M\^4 + 2\ M\^6 + 28\ M\^8 - 6\ M\^10 - 
                18\ M\^12 + 13\ M\^14 - 6\ M\^16 + 2\ M\^18)\) + 
          L\^13\ \((\(-2\)\ M\^2 + 22\ M\^4 - 53\ M\^6 - 61\ M\^8 + 
                216\ M\^10 + 80\ M\^12 - 123\ M\^14 - 75\ M\^16 + 12\ M\^18 + 
                51\ M\^20 - 16\ M\^22 + 18\ M\^24 - 30\ M\^26 + 19\ M\^28 - 
                6\ M\^30 + M\^32)\) + 
          L\^12\ \((4\ M\^2 - 42\ M\^4 + 158\ M\^6 - 96\ M\^8 - 566\ M\^10 + 
                501\ M\^12 + 1050\ M\^14 - 240\ M\^16 - 972\ M\^18 + 
                62\ M\^20 + 216\ M\^22 + 88\ M\^24 + 168\ M\^26 - 
                238\ M\^28 + 2\ M\^30 + 28\ M\^32 + 70\ M\^34 - 77\ M\^36 + 
                34\ M\^38 - 8\ M\^40 + M\^42)\) + 
          L\^11\ \((1 - 15\ M\^2 + 91\ M\^4 - 297\ M\^6 + 392\ M\^8 + 
                642\ M\^10 - 2261\ M\^12 - 387\ M\^14 + 4669\ M\^16 + 
                534\ M\^18 - 3937\ M\^20 - 496\ M\^22 + 552\ M\^24 + 
                962\ M\^26 + 698\ M\^28 - 1035\ M\^30 - 392\ M\^32 + 
                496\ M\^34 + 248\ M\^36 - 362\ M\^38 + 173\ M\^40 - 
                110\ M\^42 + 87\ M\^44 - 41\ M\^46 + 10\ M\^48 - M\^50)\) + 
          L\^10\ \((\(-1\) + 17\ M\^2 - 125\ M\^4 + 493\ M\^6 - 998\ M\^8 + 
                211\ M\^10 + 3517\ M\^12 - 5012\ M\^14 - 4718\ M\^16 + 
                10971\ M\^18 + 4600\ M\^20 - 8609\ M\^22 - 4121\ M\^24 + 
                763\ M\^26 + 4070\ M\^28 + 2406\ M\^30 - 3343\ M\^32 - 
                2464\ M\^34 + 2219\ M\^36 + 1437\ M\^38 - 1073\ M\^40 - 
                383\ M\^42 + 70\ M\^44 + 501\ M\^46 - 402\ M\^48 + 
                140\ M\^50 - 25\ M\^52 + 2\ M\^54)\) + 
          L\^9\ \((\(-4\)\ M\^2 + 61\ M\^4 - 383\ M\^6 + 1265\ M\^8 - 
                1969\ M\^10 - 709\ M\^12 + 7696\ M\^14 - 6759\ M\^16 - 
                11407\ M\^18 + 15236\ M\^20 + 12484\ M\^22 - 11453\ M\^24 - 
                12457\ M\^26 + 485\ M\^28 + 10892\ M\^30 + 5369\ M\^32 - 
                9331\ M\^34 - 7056\ M\^36 + 7268\ M\^38 + 5265\ M\^40 - 
                4179\ M\^42 - 2658\ M\^44 + 2300\ M\^46 + 792\ M\^48 - 
                1414\ M\^50 + 660\ M\^52 - 155\ M\^54 + 19\ M\^56 - 
                M\^58)\) + 
          L\^8\ \((\(-6\)\ M\^4 + 90\ M\^6 - 560\ M\^8 + 1807\ M\^10 - 
                2658\ M\^12 - 1114\ M\^14 + 9426\ M\^16 - 6487\ M\^18 - 
                14349\ M\^20 + 14127\ M\^22 + 17978\ M\^24 - 11305\ M\^26 - 
                19637\ M\^28 + 1666\ M\^30 + 18134\ M\^32 + 5994\ M\^34 - 
                17708\ M\^36 - 11035\ M\^38 + 15101\ M\^40 + 10494\ M\^42 - 
                10461\ M\^44 - 5229\ M\^46 + 6730\ M\^48 - 51\ M\^50 - 
                2493\ M\^52 + 1476\ M\^54 - 417\ M\^56 + 62\ M\^58 - 
                4\ M\^60)\) + 
          L\^7\ \((\(-4\)\ M\^6 + 62\ M\^8 - 417\ M\^10 + 1476\ M\^12 - 
                2493\ M\^14 - 51\ M\^16 + 6730\ M\^18 - 5229\ M\^20 - 
                10461\ M\^22 + 10494\ M\^24 + 15101\ M\^26 - 11035\ M\^28 - 
                17708\ M\^30 + 5994\ M\^32 + 18134\ M\^34 + 1666\ M\^36 - 
                19637\ M\^38 - 11305\ M\^40 + 17978\ M\^42 + 14127\ M\^44 - 
                14349\ M\^46 - 6487\ M\^48 + 9426\ M\^50 - 1114\ M\^52 - 
                2658\ M\^54 + 1807\ M\^56 - 560\ M\^58 + 90\ M\^60 - 
                6\ M\^62)\) + 
          L\ \((2\ M\^48 - 6\ M\^50 + 13\ M\^52 - 18\ M\^54 - 6\ M\^56 + 
                28\ M\^58 + 2\ M\^60 - 4\ M\^62)\) + 
          L\^6\ \((\(-M\^8\) + 19\ M\^10 - 155\ M\^12 + 660\ M\^14 - 
                1414\ M\^16 + 792\ M\^18 + 2300\ M\^20 - 2658\ M\^22 - 
                4179\ M\^24 + 5265\ M\^26 + 7268\ M\^28 - 7056\ M\^30 - 
                9331\ M\^32 + 5369\ M\^34 + 10892\ M\^36 + 485\ M\^38 - 
                12457\ M\^40 - 11453\ M\^42 + 12484\ M\^44 + 15236\ M\^46 - 
                11407\ M\^48 - 6759\ M\^50 + 7696\ M\^52 - 709\ M\^54 - 
                1969\ M\^56 + 1265\ M\^58 - 383\ M\^60 + 61\ M\^62 - 
                4\ M\^64)\) + 
          L\^2\ \((M\^34 - 6\ M\^36 + 19\ M\^38 - 30\ M\^40 + 18\ M\^42 - 
                16\ M\^44 + 51\ M\^46 + 12\ M\^48 - 75\ M\^50 - 123\ M\^52 + 
                80\ M\^54 + 216\ M\^56 - 61\ M\^58 - 53\ M\^60 + 22\ M\^62 - 
                2\ M\^64)\) + 
          L\^3\ \((M\^24 - 8\ M\^26 + 34\ M\^28 - 77\ M\^30 + 70\ M\^32 + 
                28\ M\^34 + 2\ M\^36 - 238\ M\^38 + 168\ M\^40 + 88\ M\^42 + 
                216\ M\^44 + 62\ M\^46 - 972\ M\^48 - 240\ M\^50 + 
                1050\ M\^52 + 501\ M\^54 - 566\ M\^56 - 96\ M\^58 + 
                158\ M\^60 - 42\ M\^62 + 4\ M\^64)\) + 
          L\^5\ \((2\ M\^12 - 25\ M\^14 + 140\ M\^16 - 402\ M\^18 + 
                501\ M\^20 + 70\ M\^22 - 383\ M\^24 - 1073\ M\^26 + 
                1437\ M\^28 + 2219\ M\^30 - 2464\ M\^32 - 3343\ M\^34 + 
                2406\ M\^36 + 4070\ M\^38 + 763\ M\^40 - 4121\ M\^42 - 
                8609\ M\^44 + 4600\ M\^46 + 10971\ M\^48 - 4718\ M\^50 - 
                5012\ M\^52 + 3517\ M\^54 + 211\ M\^56 - 998\ M\^58 + 
                493\ M\^60 - 125\ M\^62 + 17\ M\^64 - M\^66)\) + 
          L\^4\ \((\(-M\^16\) + 10\ M\^18 - 41\ M\^20 + 87\ M\^22 - 
                110\ M\^24 + 173\ M\^26 - 362\ M\^28 + 248\ M\^30 + 
                496\ M\^32 - 392\ M\^34 - 1035\ M\^36 + 698\ M\^38 + 
                962\ M\^40 + 552\ M\^42 - 496\ M\^44 - 3937\ M\^46 + 
                534\ M\^48 + 4669\ M\^50 - 387\ M\^52 - 2261\ M\^54 + 
                642\ M\^56 + 392\ M\^58 - 297\ M\^60 + 91\ M\^62 - 
                15\ M\^64 + M\^66)\)}, {"A-polynomial", Knot[9, 46], 
        L\^5\ M\^6 + M\^12 + 
          L\ \((\(-1\) + M\^2 - 2\ M\^4 + 5\ M\^6 - 2\ M\^8 + 5\ M\^12 - 
                M\^14)\) + 
          L\^2\ \((1 - 6\ M\^2 + 4\ M\^4 + 3\ M\^6 + 3\ M\^8 + 3\ M\^10 + 
                3\ M\^12 - M\^16)\) + 
          L\^4\ \((\(-M\^4\) + 5\ M\^6 - 2\ M\^10 + 5\ M\^12 - 2\ M\^14 + 
                M\^16 - M\^18)\) + 
          L\^3\ \((\(-M\^2\) + 3\ M\^6 + 3\ M\^8 + 3\ M\^10 + 3\ M\^12 + 
                4\ M\^14 - 6\ M\^16 + M\^18)\)}, {"A-polynomial", 
        Knot[9, 47], \(-M\^12\) + L\^11\ M\^68 + 
          L\ \((\(-2\)\ M\^6 + 6\ M\^8 - 12\ M\^10 + 12\ M\^12 + 18\ M\^14 - 
                31\ M\^16 + 4\ M\^18 - M\^26 + M\^28)\) + 
          L\^2\ \((\(-1\) + 6\ M\^2 - 21\ M\^4 + 49\ M\^6 - 66\ M\^8 + 
                55\ M\^10 - 36\ M\^12 - 18\ M\^14 - 15\ M\^16 + 113\ M\^18 + 
                42\ M\^20 - 102\ M\^22 - 72\ M\^24 + 87\ M\^26 - 12\ M\^28 - 
                52\ M\^30 + 41\ M\^32 - M\^34 - 4\ M\^36)\) + 
          L\^3\ \((1 - 12\ M\^2 + 56\ M\^4 - 115\ M\^6 + 90\ M\^8 - 
                53\ M\^10 + 241\ M\^12 - 360\ M\^14 + 94\ M\^16 + 
                110\ M\^18 - 38\ M\^20 - 509\ M\^22 + 795\ M\^24 + 
                87\ M\^26 - 435\ M\^28 - 43\ M\^30 + 335\ M\^32 - 
                414\ M\^34 + 56\ M\^36 + 202\ M\^38 - 48\ M\^40 - 59\ M\^42 + 
                26\ M\^44 - 2\ M\^46)\) + 
          L\^4\ \((\(-2\)\ M\^4 + 27\ M\^6 - 142\ M\^8 + 338\ M\^10 - 
                271\ M\^12 - 196\ M\^14 + 54\ M\^16 + 986\ M\^18 - 
                972\ M\^20 - 24\ M\^22 + 3\ M\^24 + 446\ M\^26 - 916\ M\^28 + 
                677\ M\^30 - 10\ M\^32 + 846\ M\^34 - 587\ M\^36 - 
                363\ M\^38 - 532\ M\^40 + 796\ M\^42 + 84\ M\^44 - 
                161\ M\^46 - 212\ M\^48 + 207\ M\^50 - 59\ M\^52 + 
                5\ M\^54)\) + 
          L\^5\ \((M\^8 - 15\ M\^10 + 89\ M\^12 - 253\ M\^14 + 317\ M\^16 - 
                65\ M\^18 + 57\ M\^20 - 772\ M\^22 + 779\ M\^24 + 41\ M\^26 + 
                880\ M\^28 - 1446\ M\^30 - 394\ M\^32 - 298\ M\^34 + 
                2178\ M\^36 - 1163\ M\^38 + 240\ M\^40 - 146\ M\^42 + 
                260\ M\^44 - 1106\ M\^46 + 1291\ M\^48 - 270\ M\^50 - 
                310\ M\^52 - 102\ M\^54 + 455\ M\^56 - 323\ M\^58 + 
                104\ M\^60 - 16\ M\^62 + M\^64)\) + 
          L\^6\ \((\(-M\^16\) + 16\ M\^18 - 104\ M\^20 + 323\ M\^22 - 
                455\ M\^24 + 102\ M\^26 + 310\ M\^28 + 270\ M\^30 - 
                1291\ M\^32 + 1106\ M\^34 - 260\ M\^36 + 146\ M\^38 - 
                240\ M\^40 + 1163\ M\^42 - 2178\ M\^44 + 298\ M\^46 + 
                394\ M\^48 + 1446\ M\^50 - 880\ M\^52 - 41\ M\^54 - 
                779\ M\^56 + 772\ M\^58 - 57\ M\^60 + 65\ M\^62 - 
                317\ M\^64 + 253\ M\^66 - 89\ M\^68 + 15\ M\^70 - M\^72)\) + 
          L\^10\ \((\(-M\^52\) + M\^54 - 4\ M\^62 + 31\ M\^64 - 18\ M\^66 - 
                12\ M\^68 + 12\ M\^70 - 6\ M\^72 + 2\ M\^74)\) + 
          L\^7\ \((\(-5\)\ M\^26 + 59\ M\^28 - 207\ M\^30 + 212\ M\^32 + 
                161\ M\^34 - 84\ M\^36 - 796\ M\^38 + 532\ M\^40 + 
                363\ M\^42 + 587\ M\^44 - 846\ M\^46 + 10\ M\^48 - 
                677\ M\^50 + 916\ M\^52 - 446\ M\^54 - 3\ M\^56 + 24\ M\^58 + 
                972\ M\^60 - 986\ M\^62 - 54\ M\^64 + 196\ M\^66 + 
                271\ M\^68 - 338\ M\^70 + 142\ M\^72 - 27\ M\^74 + 
                2\ M\^76)\) + 
          L\^8\ \((2\ M\^34 - 26\ M\^36 + 59\ M\^38 + 48\ M\^40 - 
                202\ M\^42 - 56\ M\^44 + 414\ M\^46 - 335\ M\^48 + 
                43\ M\^50 + 435\ M\^52 - 87\ M\^54 - 795\ M\^56 + 
                509\ M\^58 + 38\ M\^60 - 110\ M\^62 - 94\ M\^64 + 
                360\ M\^66 - 241\ M\^68 + 53\ M\^70 - 90\ M\^72 + 
                115\ M\^74 - 56\ M\^76 + 12\ M\^78 - M\^80)\) + 
          L\^9\ \((4\ M\^44 + M\^46 - 41\ M\^48 + 52\ M\^50 + 12\ M\^52 - 
                87\ M\^54 + 72\ M\^56 + 102\ M\^58 - 42\ M\^60 - 113\ M\^62 + 
                15\ M\^64 + 18\ M\^66 + 36\ M\^68 - 55\ M\^70 + 66\ M\^72 - 
                49\ M\^74 + 21\ M\^76 - 6\ M\^78 + 
                M\^80)\)}, {"A-polynomial", 
        Knot[9, 48], \(-M\^12\) + L\^8\ M\^42 + 
          L\ \((3\ M\^8 - 10\ M\^10 + 24\ M\^12 - 12\ M\^14 - 27\ M\^16 + 
                18\ M\^18 + M\^20 - 3\ M\^22)\) + 
          L\^2\ \((\(-3\)\ M\^4 + 20\ M\^6 - 57\ M\^8 + 48\ M\^10 + 
                41\ M\^12 - 84\ M\^14 + 65\ M\^16 + 56\ M\^18 - 168\ M\^20 + 
                11\ M\^22 + 130\ M\^24 - 75\ M\^26 - 20\ M\^28 + 30\ M\^30 - 
                9\ M\^32 + M\^34)\) + 
          L\^3\ \((1 - 10\ M\^2 + 33\ M\^4 - 36\ M\^6 - 13\ M\^8 + 
                54\ M\^10 - 16\ M\^12 - 125\ M\^14 + 138\ M\^16 + 
                113\ M\^18 - 110\ M\^20 + 30\ M\^22 - 107\ M\^24 - 
                93\ M\^26 + 240\ M\^28 - 2\ M\^30 - 243\ M\^32 + 111\ M\^34 + 
                77\ M\^36 - 75\ M\^38 + 21\ M\^40 - 2\ M\^42)\) + 
          L\^7\ \((3\ M\^32 - M\^34 - 18\ M\^36 + 27\ M\^38 + 12\ M\^40 - 
                24\ M\^42 + 10\ M\^44 - 3\ M\^46)\) + 
          L\^4\ \((\(-M\^4\) + 12\ M\^6 - 50\ M\^8 + 72\ M\^10 + 30\ M\^12 - 
                124\ M\^14 - 22\ M\^16 + 66\ M\^18 + 52\ M\^20 + 140\ M\^22 - 
                120\ M\^24 - 242\ M\^26 + 242\ M\^28 + 120\ M\^30 - 
                140\ M\^32 - 52\ M\^34 - 66\ M\^36 + 22\ M\^38 + 124\ M\^40 - 
                30\ M\^42 - 72\ M\^44 + 50\ M\^46 - 12\ M\^48 + M\^50)\) + 
          L\^6\ \((\(-M\^20\) + 9\ M\^22 - 30\ M\^24 + 20\ M\^26 + 
                75\ M\^28 - 130\ M\^30 - 11\ M\^32 + 168\ M\^34 - 56\ M\^36 - 
                65\ M\^38 + 84\ M\^40 - 41\ M\^42 - 48\ M\^44 + 57\ M\^46 - 
                20\ M\^48 + 3\ M\^50)\) + 
          L\^5\ \((2\ M\^12 - 21\ M\^14 + 75\ M\^16 - 77\ M\^18 - 
                111\ M\^20 + 243\ M\^22 + 2\ M\^24 - 240\ M\^26 + 93\ M\^28 + 
                107\ M\^30 - 30\ M\^32 + 110\ M\^34 - 113\ M\^36 - 
                138\ M\^38 + 125\ M\^40 + 16\ M\^42 - 54\ M\^44 + 13\ M\^46 + 
                36\ M\^48 - 33\ M\^50 + 10\ M\^52 - 
                M\^54)\)}, {"A-polynomial", 
        Knot[9, 49], \(-1\) + L\^8\ M\^58 + 
          L\ \((2 - 6\ M\^2 + 13\ M\^4 + 4\ M\^6 - 29\ M\^8 + 14\ M\^10 + 
                3\ M\^12 - 3\ M\^14)\) + 
          L\^2\ \((\(-1\) + 6\ M\^2 - 23\ M\^4 + 41\ M\^6 - 44\ M\^8 - 
                3\ M\^10 + 79\ M\^12 - 14\ M\^14 - 76\ M\^16 + 5\ M\^18 + 
                64\ M\^20 - 16\ M\^22 - 41\ M\^24 + 33\ M\^26 - 9\ M\^28 + 
                M\^30)\) + 
          L\^3\ \((M\^4 - 9\ M\^6 + 40\ M\^8 - 73\ M\^10 + 49\ M\^12 + 
                3\ M\^14 - 17\ M\^16 - 55\ M\^18 + 92\ M\^20 + 46\ M\^22 - 
                114\ M\^24 - 8\ M\^26 + 118\ M\^28 - 65\ M\^30 - 21\ M\^32 - 
                4\ M\^34 + 51\ M\^36 - 37\ M\^38 + 10\ M\^40 - M\^42)\) + 
          L\^4\ \((2\ M\^10 - 18\ M\^12 + 50\ M\^14 - 54\ M\^16 + 26\ M\^18 - 
                51\ M\^20 + 86\ M\^22 - M\^24 - 61\ M\^26 - 60\ M\^28 + 
                60\ M\^30 + 61\ M\^32 + M\^34 - 86\ M\^36 + 51\ M\^38 - 
                26\ M\^40 + 54\ M\^42 - 50\ M\^44 + 18\ M\^46 - 2\ M\^48)\) + 
          L\^5\ \((M\^16 - 10\ M\^18 + 37\ M\^20 - 51\ M\^22 + 4\ M\^24 + 
                21\ M\^26 + 65\ M\^28 - 118\ M\^30 + 8\ M\^32 + 114\ M\^34 - 
                46\ M\^36 - 92\ M\^38 + 55\ M\^40 + 17\ M\^42 - 3\ M\^44 - 
                49\ M\^46 + 73\ M\^48 - 40\ M\^50 + 9\ M\^52 - M\^54)\) + 
          L\^7\ \((3\ M\^44 - 3\ M\^46 - 14\ M\^48 + 29\ M\^50 - 4\ M\^52 - 
                13\ M\^54 + 6\ M\^56 - 2\ M\^58)\) + 
          L\^6\ \((\(-M\^28\) + 9\ M\^30 - 33\ M\^32 + 41\ M\^34 + 
                16\ M\^36 - 64\ M\^38 - 5\ M\^40 + 76\ M\^42 + 14\ M\^44 - 
                79\ M\^46 + 3\ M\^48 + 44\ M\^50 - 41\ M\^52 + 23\ M\^54 - 
                6\ M\^56 + M\^58)\)}}\)], "Output"]
}, Closed]]
},
FrontEndVersion->"5.2 for Microsoft Windows",
ScreenRectangle->{{0, 1280}, {0, 713}},
WindowSize->{790, 556},
WindowMargins->{{Automatic, 102}, {Automatic, 6}}
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1754, 51, 272, 5, 52, "Text"],
Cell[2029, 58, 238, 5, 52, "Text"],

Cell[CellGroupData[{
Cell[2292, 67, 248, 5, 70, "Input"],
Cell[2543, 74, 158, 2, 44, "Print"]
}, Open  ]],
Cell[2716, 79, 49, 0, 33, "Text"],

Cell[CellGroupData[{
Cell[2790, 83, 64, 1, 30, "Input"],
Cell[2857, 86, 130, 2, 22, "Message"],
Cell[2990, 90, 145, 3, 22, "Message"],
Cell[3138, 95, 156, 2, 45, "Output"]
}, Open  ]],
Cell[3309, 100, 109, 3, 33, "Text"],

Cell[CellGroupData[{
Cell[3443, 107, 137, 2, 30, "Input"],
Cell[3583, 111, 706, 15, 135, "Output"]
}, Open  ]],
Cell[4304, 129, 324, 5, 71, "Text"],
Cell[4631, 136, 224, 4, 50, "Input"],

Cell[CellGroupData[{
Cell[4880, 144, 59, 1, 30, "Input"],
Cell[4942, 147, 43, 1, 29, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[5022, 153, 62, 1, 30, "Input"],
Cell[5087, 156, 43, 1, 29, "Output"]
}, Open  ]],
Cell[5145, 160, 239, 4, 52, "Text"],
Cell[5387, 166, 190, 3, 50, "Input"],
Cell[5580, 171, 179, 4, 52, "Text"],

Cell[CellGroupData[{
Cell[5784, 179, 119, 2, 30, "Input"],
Cell[5906, 183, 568, 14, 217, "Output"]
}, Open  ]],
Cell[6489, 200, 206, 4, 52, "Text"],
Cell[6698, 206, 134, 2, 30, "Input"],

Cell[CellGroupData[{
Cell[6857, 212, 117, 2, 30, "Input"],
Cell[6977, 216, 532, 13, 141, "Output"]
}, Open  ]],
Cell[7524, 232, 373, 8, 109, "Text"],
Cell[7900, 242, 162, 2, 50, "Input"],

Cell[CellGroupData[{
Cell[8087, 248, 119, 2, 30, "Input"],
Cell[8209, 252, 518, 12, 61, "Output"]
}, Open  ]],
Cell[8742, 267, 319, 5, 71, "Text"],
Cell[9064, 274, 356, 7, 90, "Input"],
Cell[9423, 283, 553, 9, 130, "Input"],
Cell[9979, 294, 138, 3, 33, "Text"],

Cell[CellGroupData[{
Cell[10142, 301, 166, 3, 50, "Input"],
Cell[10311, 306, 195, 4, 37, "Message"],
Cell[10509, 312, 195, 4, 37, "Message"],
Cell[10707, 318, 195, 4, 37, "Message"],
Cell[10905, 324, 335, 5, 22, "Message"],
Cell[11243, 331, 1364770, 19604, 4445, "Output"]
}, Open  ]],
Cell[1376028, 19938, 473, 8, 90, "Text"],
Cell[1376504, 19948, 95, 2, 33, "Text"],
Cell[1376602, 19952, 162, 6, 90, "Text"],
Cell[1376767, 19960, 110, 3, 33, "Text"],
Cell[1376880, 19965, 90, 1, 30, "Input"],
Cell[1376973, 19968, 317, 5, 52, "Text"],
Cell[1377293, 19975, 228, 3, 110, "Input"],
Cell[1377524, 19980, 94, 2, 33, "Text"],

Cell[CellGroupData[{
Cell[1377643, 19986, 47, 1, 30, "Input"],
Cell[1377693, 19989, 50, 1, 29, "Output"]
}, Open  ]],
Cell[1377758, 19993, 139, 3, 33, "Text"],

Cell[CellGroupData[{
Cell[1377922, 20000, 82, 1, 30, "Input"],
Cell[1378007, 20003, 36, 1, 29, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[1378080, 20009, 115, 2, 30, "Input"],
Cell[1378198, 20013, 472391, 6876, 55262, "Output"]
}, Closed]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

