(* Content-type: application/mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 6.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       145,          7]
NotebookDataLength[      9611,        273]
NotebookOptionsPosition[      8689,        238]
NotebookOutlinePosition[      9057,        254]
CellTagsIndexPosition[      9014,        251]
WindowFrame->Normal
ContainsDynamic->False*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["KnotTheory`SmallGirth` package", "Subtitle",
 InitializationCell->True,
 CellChangeTimes->{{3.441300920347542*^9, 3.441300921735297*^9}}],

Cell["\<\
A subpackage for KnotTheory`, which reorders PD presentations to minimise \
girth.
January 18, 2009, Scott Morrison\
\>", "Text",
 InitializationCell->True,
 CellChangeTimes->{{3.441302362136074*^9, 3.441302385067758*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"BeginPackage", "[", 
   RowBox[{"\"\<KnotTheory`SmallGirth`\>\"", ",", 
    RowBox[{"{", "\"\<KnotTheory`\>\"", "}"}]}], "]"}], ";"}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.4294630944324737`*^9, 3.4294630956041584`*^9}, {
  3.4413009240964403`*^9, 3.441300929612648*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"FindSmallGirthOrdering", "::", "about"}], "=", 
   "\"\<FindSmallGirthOrdering[K] tries to reorder the crossings in a PD \
presentation of K so that, when the crossings are read in order, the girth is \
minimised. It does this by repeatedly running a greedy algorithm, making \
random choices when they are available. FindSmallGirthOrdering[K, n] returns \
the best of n attempts; n defaults to 1000.\>\""}], ";"}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.441300940048771*^9, 3.44130102718477*^9}}],

Cell[BoxData[
 RowBox[{"Begin", "[", "\"\<`Private`\>\"", "]"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{
  RowBox[{"randomOrderPD", "[", "K_", "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"pd", "=", 
       RowBox[{"PD", "[", "K", "]"}]}], ",", "maxConnection", ",", 
      "availableCrossings", ",", "nextCrossing", ",", 
      RowBox[{"inside", "=", 
       RowBox[{"{", "}"}]}], ",", 
      RowBox[{"result", "=", 
       RowBox[{"PD", "[", "]"}]}], ",", 
      RowBox[{"girth", "=", "0"}], ",", 
      RowBox[{"girths", "=", 
       RowBox[{"{", "}"}]}]}], "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"While", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"Length", "[", "pd", "]"}], ">", "0"}], ",", 
       "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"maxConnection", "=", 
         RowBox[{"Max", "[", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"4", "-", 
              RowBox[{"Length", "[", 
               RowBox[{"Complement", "[", 
                RowBox[{
                 RowBox[{"List", "@@", "#1"}], ",", "inside"}], "]"}], 
               "]"}]}], "&"}], ")"}], "/@", 
           RowBox[{"List", "@@", "pd"}]}], "]"}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"availableCrossings", "=", 
         RowBox[{"Cases", "[", 
          RowBox[{
           RowBox[{"List", "@@", "pd"}], ",", 
           RowBox[{"x_", "/;", 
            RowBox[{
             RowBox[{"(", 
              RowBox[{"4", "-", 
               RowBox[{"Length", "[", 
                RowBox[{"Complement", "[", 
                 RowBox[{
                  RowBox[{"List", "@@", "x"}], ",", "inside"}], "]"}], 
                "]"}]}], ")"}], "\[Equal]", "maxConnection"}]}]}], "]"}]}], 
        ";", "\[IndentingNewLine]", 
        RowBox[{"nextCrossing", "=", 
         RowBox[{"availableCrossings", "\[LeftDoubleBracket]", 
          RowBox[{"RandomInteger", "[", 
           RowBox[{"{", 
            RowBox[{"1", ",", 
             RowBox[{"Length", "[", "availableCrossings", "]"}]}], "}"}], 
           "]"}], "\[RightDoubleBracket]"}]}], ";", "\[IndentingNewLine]", 
        RowBox[{"girth", "+=", 
         RowBox[{"4", "-", 
          RowBox[{"2", 
           RowBox[{"Length", "[", 
            RowBox[{"Intersection", "[", 
             RowBox[{
              RowBox[{"List", "@@", "nextCrossing"}], ",", "inside"}], "]"}], 
            "]"}]}]}]}], ";", "\[IndentingNewLine]", 
        RowBox[{"AppendTo", "[", 
         RowBox[{"girths", ",", "girth"}], "]"}], ";", "\[IndentingNewLine]", 
        
        RowBox[{"inside", "=", 
         RowBox[{"Union", "[", 
          RowBox[{"inside", ",", 
           RowBox[{"List", "@@", "nextCrossing"}]}], "]"}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"AppendTo", "[", 
         RowBox[{"result", ",", "nextCrossing"}], "]"}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"pd", "=", 
         RowBox[{"DeleteCases", "[", 
          RowBox[{"pd", ",", "nextCrossing"}], "]"}]}], ";"}]}], 
      "\[IndentingNewLine]", "]"}], ";", "\[IndentingNewLine]", 
     RowBox[{"{", 
      RowBox[{"girths", ",", "result"}], "}"}]}]}], "\[IndentingNewLine]", 
   "]"}]}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.426620041434458*^9, 3.426620048434523*^9}, {
  3.4266200877510576`*^9, 3.4266201973186083`*^9}, {3.4266202574650946`*^9, 
  3.4266202579558*^9}, {3.42662757522752*^9, 3.42662767546165*^9}, {
  3.4266277517713776`*^9, 3.4266277564781456`*^9}, {3.426627825056757*^9, 
  3.4266280837186947`*^9}, {3.426628242136488*^9, 3.4266282463525505`*^9}, {
  3.426628400454138*^9, 3.4266284009147997`*^9}, {3.44130084845747*^9, 
  3.441300853000552*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"FindSmallGirthOrdering", "[", "K_", "]"}], ":=", 
  RowBox[{"FindSmallGirthOrdering", "[", 
   RowBox[{"K", ",", "1000"}], "]"}]}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.441300810548665*^9, 3.441300830539405*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"FindSmallGirthOrdering", "[", 
   RowBox[{"K_", ",", "k_"}], "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"i", "=", "0"}], ",", 
      RowBox[{"bestSoFar", "=", 
       RowBox[{"randomOrderPD", "[", "K", "]"}]}], ",", "next"}], "}"}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"While", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"(", 
         RowBox[{"++", "i"}], ")"}], "\[LessEqual]", "k"}], ",", 
       "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"next", "=", 
         RowBox[{"randomOrderPD", "[", "K", "]"}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"Max", "[", 
            RowBox[{
            "next", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}], 
            "]"}], "\[LessEqual]", 
           RowBox[{"Max", "[", 
            RowBox[{
            "bestSoFar", "\[LeftDoubleBracket]", "1", 
             "\[RightDoubleBracket]"}], "]"}]}], ",", "\[IndentingNewLine]", 
          RowBox[{"If", "[", 
           RowBox[{
            RowBox[{
             RowBox[{"Count", "[", 
              RowBox[{
               RowBox[{
               "next", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}],
                ",", 
               RowBox[{"Max", "[", 
                RowBox[{
                "next", "\[LeftDoubleBracket]", "1", 
                 "\[RightDoubleBracket]"}], "]"}]}], "]"}], "<", 
             RowBox[{"Count", "[", 
              RowBox[{
               RowBox[{
               "bestSoFar", "\[LeftDoubleBracket]", "1", 
                "\[RightDoubleBracket]"}], ",", 
               RowBox[{"Max", "[", 
                RowBox[{
                "bestSoFar", "\[LeftDoubleBracket]", "1", 
                 "\[RightDoubleBracket]"}], "]"}]}], "]"}]}], ",", 
            "\[IndentingNewLine]", 
            RowBox[{
             RowBox[{"bestSoFar", "=", "next"}], ";"}]}], 
           "\[IndentingNewLine]", "]"}]}], "\[IndentingNewLine]", "]"}], 
        ";"}]}], "\[IndentingNewLine]", "]"}], ";", "\[IndentingNewLine]", 
     RowBox[{
     "bestSoFar", "\[LeftDoubleBracket]", "2", "\[RightDoubleBracket]"}]}]}], 
   "\[IndentingNewLine]", "]"}]}]], "Input",
 InitializationCell->True,
 CellChangeTimes->{{3.4266281552515535`*^9, 3.426628407053627*^9}, {
  3.4269735454603443`*^9, 3.426973551979718*^9}, {3.441300805500052*^9, 
  3.441300845638873*^9}, {3.441301541033263*^9, 3.441301581449314*^9}}],

Cell[BoxData[
 RowBox[{"End", "[", "]"}]], "Input",
 InitializationCell->True],

Cell[BoxData[
 RowBox[{"EndPackage", "[", "]"}]], "Input",
 InitializationCell->True]
}, Open  ]]
},
AutoGeneratedPackage->Automatic,
WindowSize->{640, 750},
WindowMargins->{{12, Automatic}, {Automatic, 24}},
FrontEndVersion->"6.0 for Mac OS X x86 (32-bit) (May 21, 2008)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[590, 23, 143, 2, 46, "Subtitle",
 InitializationCell->True],
Cell[736, 27, 232, 6, 41, "Text",
 InitializationCell->True],
Cell[971, 35, 327, 7, 27, "Input",
 InitializationCell->True],
Cell[1301, 44, 564, 10, 118, "Input",
 InitializationCell->True],
Cell[1868, 56, 101, 2, 27, "Input",
 InitializationCell->True],
Cell[1972, 60, 3697, 91, 298, "Input",
 InitializationCell->True],
Cell[5672, 153, 269, 6, 27, "Input",
 InitializationCell->True],
Cell[5944, 161, 2560, 66, 208, "Input",
 InitializationCell->True],
Cell[8507, 229, 78, 2, 27, "Input",
 InitializationCell->True],
Cell[8588, 233, 85, 2, 27, "Input",
 InitializationCell->True]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
